#ifndef MAT3X3_INLINE_H
#define MAT3X3_INLINE_H

#include "ts_errors.h"
#include <math.h>
#include <string.h>

#include "quat4.h"

#define MAT3X3_ZERO { .elts = {0} }
#define MAT3X3_ELT(m,i,j) ( (m)->elts[ 4*(j) + (i) ] )

static inline VecReal mat3x3_elt(const mat3x3*restrict m, int i, int j) 
{ 
  BOUNDS_CHECK(i, 0, 2);
  BOUNDS_CHECK(j, 0, 2);

  return m->elts[4*j+i]; 
}

static inline void mat3x3_zero(mat3x3*restrict m)
{
  memset(m,0,sizeof(*m));
}

static inline void mat3x3_identity(mat3x3*restrict m)
{
  memset(m,0,sizeof(*m));
  m->elts[0] = m->elts[5] = m->elts[10] = 1.0;
}

static inline void mat3x3_set(mat3x3*restrict dst, const mat3x3*restrict src)
{
  memcpy(dst,src,sizeof(*src));
  dst->elts[3] = dst->elts[7] = dst->elts[11] = 0.0;
}

static inline void mat3x3_transpose(mat3x3*restrict m)
{
  for(size_t j=0; j < 2; j++) {
    for(size_t i=j+1; i < 3; i++) {
      VecReal tmp = MAT3X3_ELT(m, i,j);
      MAT3X3_ELT(m, i,j) = MAT3X3_ELT(m, j,i);
      MAT3X3_ELT(m, j,i) = tmp;
    }
  }
}

static inline void mat3x3_set_transpose(mat3x3*restrict dst, const mat3x3*restrict src)
{
  for (size_t j=0; j < 3; j++) {
    for (size_t i=0; i < 3; i++) {
      MAT3X3_ELT(dst, i,j) = MAT3X3_ELT(src, j,i);
    }
    MAT3X3_ELT(dst, 3,j) = 0.0;
  }
}

static inline void mat3x3_add_to(mat3x3*restrict a, const mat3x3*restrict b)
{
  for (size_t i=0; i < MAT_NUM_ELTS(a); i++) {
    a->elts[i] += b->elts[i];
  }
}

static inline void mat3x3_add(mat3x3*restrict c, const mat3x3*restrict a, const mat3x3*restrict b)
{
  for (size_t i=0; i < MAT_NUM_ELTS(c); i++) {
    c->elts[i] = a->elts[i] + b->elts[i];
  }
}

static inline void mat3x3_sub(mat3x3*restrict c, const mat3x3*restrict a, const mat3x3*restrict b)
{
  for (size_t i=0; i < MAT_NUM_ELTS(c); i++) {
    c->elts[i] = a->elts[i] - b->elts[i];
  }
}

static inline void mat3x3_scale_by(mat3x3*restrict m, VecReal scl)
{
  for (size_t i=0; i < MAT_NUM_ELTS(m); i++) {
    m->elts[i] *= scl;
  }
}

/*! `dst = src*src` */
static inline void mat3x3_mul_self(mat3x3*restrict dst, const mat3x3*restrict src)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(src,i,k) * MAT3X3_ELT(src,k,j);
      }
    }
  }
}

/*! `dst = ma * mb` */
static inline void mat3x3_mul(mat3x3* restrict dst, const mat3x3*restrict ma, const mat3x3*restrict mb)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(ma,i,k) * MAT3X3_ELT(mb,k,j);
      }
    }
  }
}

/*! `dst = ma*transpose(mb)` */
static inline void mat3x3_rmul_trans(mat3x3*restrict dst, const mat3x3*restrict ma, const mat3x3*restrict mb)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(ma,i,k) * MAT3X3_ELT(mb,j,k);
      }
    }
  }
}

/*! `dst = transpose(mb)*ma` */
static inline void mat3x3_lmul_trans(mat3x3*restrict dst, const mat3x3*restrict ma, const mat3x3*restrict mb)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(mb,k,i) * MAT3X3_ELT(ma,k,j);
      }
    }
  }
}

/*! `dst = src*transpose(src)` */
static inline void mat3x3_rmul_self_trans(mat3x3*restrict dst, const mat3x3*restrict src)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(src,i,k) * MAT3X3_ELT(src,j,k);
      }
    }
  }
}

/*! `dst = transpose(src)*src` */
static inline void mat3x3_lmul_self_trans(mat3x3*restrict dst, const mat3x3*restrict src)
{
  for (int i=0; i < 3; i++) {
    for (int j=0; j < 3; j++) {
      MAT3X3_ELT(dst,i,j) = 0.0;
      for (int k=0; k < 3; k++) {
        MAT3X3_ELT(dst,i,j) += MAT3X3_ELT(src,k,i) * MAT3X3_ELT(src,k,j);
      }
    }
  }
}

/*! `mc = transpose(ma) * mb * ma */
static inline void mat3x3_orth_xform(
    mat3x3*restrict mc, 
    const mat3x3*restrict ma,
    const mat3x3*restrict mb)
{
  mat3x3 tmp;
  mat3x3_mul(&tmp,mb,ma);
  mat3x3_lmul_trans(mc, ma, &tmp);
}

/*! `mc = ma * mb * transpose(ma) */
static inline void mat3x3_orth_xform_t(
    mat3x3*restrict mc, 
    const mat3x3*restrict ma,
    const mat3x3*restrict mb)
{
  mat3x3 tmp;
  mat3x3_rmul_trans(&tmp,mb,ma);
  mat3x3_mul(mc, ma, &tmp);
}

/*! `mb = transpose(ma) * mb * ma */
static inline void mat3x3_self_orth_xform(
    mat3x3*restrict mb, 
    const mat3x3*restrict ma)
{
  mat3x3 tmp;
  mat3x3_mul(&tmp,mb,ma);
  mat3x3_lmul_trans(mb, ma, &tmp);
}

static inline void mat3x3_set_scaled(mat3x3*restrict dst, VecReal scl, const mat3x3*restrict src)
{
  mat3x3_set(dst,src);
  mat3x3_scale_by(dst,scl);
}

// M = M + transpose(M)
static inline void mat3x3_add_transpose_to_self(mat3x3*restrict M)
{
  for(int i=0; i < 3; i++) {
    MAT3X3_ELT(M,i,i) *= 2.0;
    for (int j=i+1; j < 3; j++) {
      MAT3X3_ELT(M,i,j) += MAT3X3_ELT(M,j,i);
      MAT3X3_ELT(M,j,i) = MAT3X3_ELT(M,i,j);
    }
  }
}

// M = M + a*I
static inline void mat3x3_add_scaled_identity_to_self(mat3x3*restrict M, double scl)
{
  for (int i=0; i < 3; i++) {
    MAT3X3_ELT(M,i,i) += scl;
  }
}

/*  Matrix-Vector routines */

static inline void vector3_outer(mat3x3*restrict m, const vector3* a, const vector3* b)
{
  vector3_to_array_scaled(a, vector3_x(b), m->elts+0); m->elts[3]  = 0.0;
  vector3_to_array_scaled(a, vector3_y(b), m->elts+4); m->elts[7]  = 0.0;
  vector3_to_array_scaled(a, vector3_z(b), m->elts+8); m->elts[11] = 0.0;
}

static inline void vector3_mat_rmul(vector3*restrict vr, const mat3x3*restrict mat, const vector3*restrict v)
{
  for (int i=0; i < 3; i++) {
    VECTOR3_ELT(vr,i) = 0.0;
    for (int j=0; j < 3; j++) {
      VECTOR3_ELT(vr,i) += MAT3X3_ELT(mat,i,j) * VECTOR3_ELT(v,j);
    }
  }
  VECTOR3_W(vr) = 0.0;
}

static inline void vector3_mat_lmul(vector3*restrict vr, const vector3*restrict v, const mat3x3*restrict mat)
{
  for (int j=0; j < 3; j++) {
    VECTOR3_ELT(vr,j) = 0.0;
    for (int i=0; i < 3; i++) {
      VECTOR3_ELT(vr,j) += MAT3X3_ELT(mat,i,j) * VECTOR3_ELT(v,i);
    }
  }
  VECTOR3_W(vr) = 0.0;
}

static inline void mat3x3_from_quat(const quat4* q, mat3x3*restrict mat)
{
  double xx = QUAT4_X(q) * QUAT4_X(q);
  double yy = QUAT4_Y(q) * QUAT4_Y(q);
  double zz = QUAT4_Z(q) * QUAT4_Z(q);

  double xy = QUAT4_X(q) * QUAT4_Y(q);
  double xz = QUAT4_X(q) * QUAT4_Z(q);
  double xw = QUAT4_X(q) * QUAT4_W(q);

  double yz = QUAT4_Y(q) * QUAT4_Z(q);
  double yw = QUAT4_Y(q) * QUAT4_W(q);

  double zw = QUAT4_Z(q) * QUAT4_W(q);

  MAT3X3_ELT(mat,0,0) = 1 - 2*(yy+zz);
  MAT3X3_ELT(mat,0,1) =     2*(xy-zw);
  MAT3X3_ELT(mat,0,2) =     2*(xz+yw);
  MAT3X3_ELT(mat,1,0) =     2*(xy+zw);
  MAT3X3_ELT(mat,1,1) = 1 - 2*(xx+zz);
  MAT3X3_ELT(mat,1,2) =     2*(yz-xw);
  MAT3X3_ELT(mat,2,0) =     2*(xz-yw);
  MAT3X3_ELT(mat,2,1) =     2*(yz+xw);
  MAT3X3_ELT(mat,2,2) = 1 - 2*(xx+yy);
}

static inline void mat3x3_set_col(mat3x3*restrict mat, int c, const vector3*restrict v)
{
  MAT3X3_ELT(mat,0,c) = VECTOR3_X(v);
  MAT3X3_ELT(mat,1,c) = VECTOR3_Y(v);
  MAT3X3_ELT(mat,2,c) = VECTOR3_Z(v);
}

enum MAT3X3_ROT_AXIS
{
  ROT_NONE    = 0,

  ROT_X_AXIS  = 1,
  ROT_Y_AXIS  = 2,
  ROT_Z_AXIS  = 3,

  ROT_DX_AXIS = 4,
  ROT_DY_AXIS = 5,
  ROT_DZ_AXIS = 6,

  ROT_DX2_AXIS = 7,
  ROT_DY2_AXIS = 8,
  ROT_DZ2_AXIS = 9,
};

void mat3x3_axis_rotate(mat3x3* mat, enum MAT3X3_ROT_AXIS axis, double ang);

static inline void mat3x3_rotate_three_axes(mat3x3*restrict rot, 
    enum MAT3X3_ROT_AXIS axis0,
    enum MAT3X3_ROT_AXIS axis1,
    enum MAT3X3_ROT_AXIS axis2,
    VecReal angles[3])
{
  mat3x3 xf1 = MAT3X3_ZERO;
  mat3x3 xf2 = MAT3X3_ZERO;
  mat3x3 xf3 = MAT3X3_ZERO;

  mat3x3_axis_rotate(&xf1, axis1, angles[1]);
  mat3x3_axis_rotate(&xf2, axis2, angles[2]);
  mat3x3_mul(&xf3, &xf1, &xf2);

  mat3x3_axis_rotate(&xf1, axis0, angles[0]);
  mat3x3_mul(rot, &xf1,&xf3);
}

#endif /* MAT3X3_INLINE_H */

