#ifndef QUAT4_INLINE_H
#define QUAT4_INLINE_H

#include <math.h>
#include <stdlib.h>

#include "quat4_inline_common.h"

static inline void quat4_set_scaled(quat4*restrict q1, VecReal scl, const quat4*restrict q2)
{
  quat4_set_elements(q1,
      scl*quat4_w(q2),
      scl*quat4_x(q2),
      scl*quat4_y(q2),
      scl*quat4_z(q2));
}

static inline void quat4_set_scaled_vec(quat4*restrict q, VecReal scl, const vector3*restrict v)
{
  quat4_set_elements(q,
      0.0,
      scl*vector3_x(v),
      scl*vector3_y(v),
      scl*vector3_z(v));
}

static inline void quat4_set_scl_vec(quat4*restrict q, VecReal scl, const vector3*restrict v)
{
  quat4_set_elements(q, scl, vector3_x(v),vector3_y(v),vector3_z(v));
}

static inline void quat4_add(quat4*restrict qr, const quat4*restrict qa, const quat4*restrict qb)
{
  QUAT4_W(qr) = QUAT4_W(qa) + QUAT4_W(qb);
  QUAT4_X(qr) = QUAT4_X(qa) + QUAT4_X(qb);
  QUAT4_Y(qr) = QUAT4_Y(qa) + QUAT4_Y(qb);
  QUAT4_Z(qr) = QUAT4_Z(qa) + QUAT4_Z(qb);
}

static inline void quat4_add_to(quat4*restrict qr, const quat4*restrict qa)
{
  QUAT4_W(qr) += quat4_w(qa);
  QUAT4_X(qr) += quat4_x(qa);
  QUAT4_Y(qr) += quat4_y(qa);
  QUAT4_Z(qr) += quat4_z(qa);
}

static inline void quat4_sub(quat4*restrict qr, const quat4*restrict qa, const quat4*restrict qb)
{
  QUAT4_W(qr) = QUAT4_W(qa) - QUAT4_W(qb);
  QUAT4_X(qr) = QUAT4_X(qa) - QUAT4_X(qb);
  QUAT4_Y(qr) = QUAT4_Y(qa) - QUAT4_Y(qb);
  QUAT4_Z(qr) = QUAT4_Z(qa) - QUAT4_Z(qb);
}

// a -= b
static inline void quat4_sub_from(quat4*restrict qa, const quat4*restrict qb)
{
  QUAT4_W(qa) -= QUAT4_W(qb);
  QUAT4_X(qa) -= QUAT4_X(qb);
  QUAT4_Y(qa) -= QUAT4_Y(qb);
  QUAT4_Z(qa) -= QUAT4_Z(qb);
}

static inline void quat4_set(quat4*restrict q1, const quat4*restrict q2)
{
  quat4_set_scaled(q1, 1.0, q2);
}

static inline int quat4_equal(const quat4*restrict q1, const quat4*restrict q2)
{
  return (quat4_w(q1) == quat4_w(q2)) &&
    (quat4_x(q1) == quat4_x(q2)) &&
    (quat4_y(q1) == quat4_y(q2)) &&
    (quat4_z(q1) == quat4_z(q2));
}

static inline VecReal quat4_norm( const quat4*restrict q )
{
  return sqrt(
      square( quat4_w(q) ) + 
      square( quat4_x(q) ) + 
      square( quat4_y(q) ) +
      square( quat4_z(q) ) );
}

/*! \brief Returns sqrt( (q1.w-q2.w)^2 + (q1.x-q2.x)^2 + (q1.y-q2.y)^2 +
 * (q1.z-q2.z)^2 )
 */
static inline VecReal quat4_distance_to(const quat4*restrict q1, const quat4*restrict q2)
{
  quat4 diff;
  quat4_sub(&diff, q1,q2);
  return quat4_norm(&diff);
}

/*! \brief q = scl*q */
static inline void quat4_scale_by( quat4*restrict q, VecReal scl )
{
  QUAT4_X(q) *= scl;
  QUAT4_Y(q) *= scl;
  QUAT4_Z(q) *= scl;
  QUAT4_W(q) *= scl;
}

/*! \brief q1 = q1*q2 */
static inline void quat4_rmul(quat4*restrict q1, const quat4*restrict q2)
{
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = quat4_x(q2); y2 = quat4_y(q2); z2 = quat4_z(q2);

  QUAT4_W(q1) = w1*w2 - x1*x2 - y1*y2 - z1*z2;
  QUAT4_X(q1) = w1*x2 + x1*w2 + y1*z2 - z1*y2;
  QUAT4_Y(q1) = w1*y2 + y1*w2 - x1*z2 + z1*x2;
  QUAT4_Z(q1) = w1*z2 + z1*w2 + x1*y2 - y1*x2;
}

// q1 = q1 * conj(q2)
static inline void quat4_rcmul(quat4*restrict q1, const quat4*restrict q2)
{
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = -quat4_x(q2); y2 = -quat4_y(q2); z2 = -quat4_z(q2);

  QUAT4_W(q1) = w1*w2 - x1*x2 - y1*y2 - z1*z2;
  QUAT4_X(q1) = w1*x2 + x1*w2 + y1*z2 - z1*y2;
  QUAT4_Y(q1) = w1*y2 + y1*w2 - x1*z2 + z1*x2;
  QUAT4_Z(q1) = w1*z2 + z1*w2 + x1*y2 - y1*x2;
}

// q1 = q2*q1
static inline void quat4_lmul(quat4*restrict q1, const quat4*restrict q2)
{
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = quat4_x(q2); y2 = quat4_y(q2); z2 = quat4_z(q2);

  QUAT4_W(q1) = w2*w1 - x2*x1 - y2*y1 - z2*z1;
  QUAT4_X(q1) = w2*x1 + x2*w1 + y2*z1 - z2*y1;
  QUAT4_Y(q1) = w2*y1 + y2*w1 - x2*z1 + z2*x1;
  QUAT4_Z(q1) = w2*z1 + z2*w1 + x2*y1 - y2*x1;

#ifdef QUAT_TEST_ASSERTIONS
  quat4 tmp;
  quat4_set(&tmp,q2);
  quat4_rmul(&tmp,q1);

  quat4_sub_from(&tmp,q1);
  tsAssert( fabs(quat4_norm(&tmp)) < 1e-8 );
#endif /* QUAT_TEST_ASSERTIONS */
}

// q1 = conj(q2) * q1
static inline void quat4_lcmul(quat4*restrict q1, const quat4*restrict q2)
{
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = -quat4_x(q2); y2 = -quat4_y(q2); z2 = -quat4_z(q2);

  QUAT4_W(q1) = w2*w1 - x2*x1 - y2*y1 - z2*z1;
  QUAT4_X(q1) = w2*x1 + x2*w1 + y2*z1 - z2*y1;
  QUAT4_Y(q1) = w2*y1 + y2*w1 - x2*z1 + z2*x1;
  QUAT4_Z(q1) = w2*z1 + z2*w1 + x2*y1 - y2*x1;
}

static inline void quat4_conj( quat4*restrict q )
{
  QUAT4_X(q) = -quat4_x(q);
  QUAT4_Y(q) = -quat4_y(q);
  QUAT4_Z(q) = -quat4_z(q);
}

static inline void quat4_normalize( quat4*restrict q )
{
  VecReal rn = 1.0/quat4_norm(q);

  QUAT4_W(q) *= rn;
  QUAT4_X(q) *= rn;
  QUAT4_Y(q) *= rn;
  QUAT4_Z(q) *= rn;
}

static inline void quat4_from_axis_angle_vec( quat4*restrict q, const VecReal*restrict axis_ang )
{
  VecReal ang = sqrt(
      axis_ang[0]*axis_ang[0] + 
      axis_ang[1]*axis_ang[1] + 
      axis_ang[2]*axis_ang[2]);

  VecReal cos_halfang = cos(ang/2);
  VecReal sin_halfang = sin(ang/2);
  if (ang > VEC_TOL) {
    VecReal scl = sin_halfang / ang;

    QUAT4_W(q) = cos_halfang;
    QUAT4_X(q) = scl*axis_ang[0];
    QUAT4_Y(q) = scl*axis_ang[1];
    QUAT4_Z(q) = scl*axis_ang[2];
  } else {
    quat4_identity( q );
  }
}

static inline void quat4_from_axis_angle( quat4*restrict q, const vector3*restrict axis_ang )
{
  VecReal ang = vector3_norm(axis_ang);

  if (ang > VEC_TOL) {
    VecReal cos_halfang = cos(ang/2);
    VecReal sin_halfang = sin(ang/2);

    VecReal scl = sin_halfang / ang;

    QUAT4_W(q) = cos_halfang;
    QUAT4_X(q) = scl*vector3_x(axis_ang);
    QUAT4_Y(q) = scl*vector3_y(axis_ang);
    QUAT4_Z(q) = scl*vector3_z(axis_ang);
  } else {
    quat4_identity( q );
  }
}

static inline VecReal quat4_to_axis_angle_separately(const quat4*restrict q, vector3*restrict axis)
{
  vector3_set(axis, (const vector3*)q);
  VecReal sin_halfang = vector3_normalize(axis);
  if (sin_halfang >= VEC_TOL) {
    VecReal cos_halfang = quat4_w(q);
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    return ang;
  } else {
    return 0.0;
  }
}

static inline void quat4_to_axis_angle( const quat4*restrict q, vector3*restrict axis_ang )
{
  VecReal cos_halfang = quat4_w(q);

  VecReal sin_halfang = vector3_norm((const vector3*)q);
  if (sin_halfang >= VEC_TOL) {
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    VecReal scl = ang / sin_halfang;

    vector3_set_scaled(axis_ang, scl, (const vector3*)q);
  } else {
    vector3_zero(axis_ang);
  }
}

static inline void quat4_to_axis_angle_vec( const quat4*restrict q, VecReal*restrict axis_ang )
{
  VecReal cos_halfang = quat4_w(q);
  VecReal sin_halfang = vector3_norm((const vector3*)q);

  if (sin_halfang >= VEC_TOL) {
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    VecReal scl = ang / sin_halfang;

    axis_ang[0] = quat4_x(q) * scl;
    axis_ang[1] = quat4_y(q) * scl;
    axis_ang[2] = quat4_z(q) * scl;
  } else {
    axis_ang[0] = 0;
    axis_ang[1] = 0;
    axis_ang[2] = 0;
  }
}

static inline void quat4_set_axis_angle( quat4*restrict q, const vector3*restrict axis, VecReal phi )
{
  VecReal norm = vector3_norm(axis);
  VecReal halfphi = phi*0.5;
  VecReal sin_halfphi = sin(halfphi);
  VecReal cos_halfphi = cos(halfphi);

  VecReal scl = sin_halfphi/norm;

  QUAT4_W(q) = cos_halfphi;
  QUAT4_X(q) = scl*vector3_x(axis);
  QUAT4_Y(q) = scl*vector3_y(axis);
  QUAT4_Z(q) = scl*vector3_z(axis);
}

static inline void quat4_rotate_vector(const quat4*restrict q, vector3*restrict v)
{
  quat4 qv;
  quat4_set_elements(&qv, 0.0, vector3_x(v), vector3_y(v), vector3_z(v));

  quat4_rcmul(&qv,q);
  quat4_lmul(&qv,q);

  vector3_set_with(v, quat4_x(&qv), quat4_y(&qv), quat4_z(&qv));
}

static inline void quat4_dq_from_omega(quat4*restrict dq, const quat4*restrict q, VecReal dt, const vector3*restrict omega)
{
  quat4_set_scaled_vec(dq, 0.5*dt, omega);
  quat4_rmul(dq,q);
}

static inline void quat4_rotate_delta(quat4*restrict dq, const quat4*restrict q, VecReal dt, const vector3*restrict omega)
{
  quat4_dq_from_omega(dq, q, dt, omega);

  quat4 qtmp;
  quat4_add(&qtmp, q, dq);
  quat4_normalize(&qtmp);
  quat4_sub(dq, &qtmp, q);
}

static inline void quat4_from_euler_zxz_array(quat4*restrict q, const VecReal arr[3])
{
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  QUAT4_W(q) = cos_half_phi*cos_half_theta*cos_half_psi - sin_half_phi*cos_half_theta*sin_half_psi;
  QUAT4_X(q) = cos_half_phi*sin_half_theta*cos_half_psi + sin_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_Y(q) = sin_half_phi*sin_half_theta*cos_half_psi - cos_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_Z(q) = cos_half_phi*cos_half_theta*sin_half_psi + sin_half_phi*cos_half_theta*cos_half_psi;
}

static inline void quat4_from_euler_xyz_array(quat4*restrict q, const VecReal arr[3])
{
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  QUAT4_W(q) = cos_half_phi*cos_half_theta*cos_half_psi - sin_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_X(q) = cos_half_phi*sin_half_theta*sin_half_psi + sin_half_phi*cos_half_theta*cos_half_psi;
  QUAT4_Y(q) = cos_half_phi*sin_half_theta*cos_half_psi - sin_half_phi*cos_half_theta*sin_half_psi;
  QUAT4_Z(q) = cos_half_phi*cos_half_theta*sin_half_psi + sin_half_phi*sin_half_theta*cos_half_psi;
}

#endif /* QUAT4_INLINE_H */

