#ifndef QUAT4_INLINE_COMMON_H
#define QUAT4_INLINE_COMMON_H

#include <stdio.h>

#define MODULE_NAME QUAT4_INLINE_COMMON

/* Three dimensional vector representations of rotation.
 *
 * For minimization and normal mode calculations, we need to convert to
 * a different representation for the orientation that only requires
 * three rotational parameters.
 *
 * The following parameterizations are provided:
 */
enum TS_ROT_REPR {
  TS_ROT_NONE       = 0,  // Particle (no rotational degrees of freedom)
  TS_ROT_AXIS_ANGLE = 1,  // Axis-angle (exponential map)
  TS_ROT_EULER_ZXZ  = 2,  // Euler angles: body-3-1-3 (classic Euler angles)
  TS_ROT_EULER_XYZ  = 3,  // Euler angles: body-3-2-1

  TS_ROT_BODY_AXES = 4,

  TS_ROT_BODY_RELATIVE = 5,
};

static inline VecReal quat4_w(const quat4*restrict q) { return QUAT4_W(q); }
static inline VecReal quat4_x(const quat4*restrict q) { return QUAT4_X(q); }
static inline VecReal quat4_y(const quat4*restrict q) { return QUAT4_Y(q); }
static inline VecReal quat4_z(const quat4*restrict q) { return QUAT4_Z(q); }

static inline void print_quat( FILE* f, quat4* q )
{
  fprintf(f,"%7.4f %7.4f %7.4f %7.4f\n", 
      QUAT4_W(q), QUAT4_X(q), QUAT4_Y(q), QUAT4_Z(q));
}


static inline void quat4_set_elements(quat4*restrict q, VecReal w, VecReal x, VecReal y, VecReal z)
{
  QUAT4_X(q) = x;
  QUAT4_Y(q) = y;
  QUAT4_Z(q) = z;
  QUAT4_W(q) = w;
}

static inline void quat4_to_array(const quat4*restrict q, VecReal arr[4])
{
  arr[0] = QUAT4_W(q);
  arr[1] = QUAT4_X(q);
  arr[2] = QUAT4_Y(q);
  arr[3] = QUAT4_Z(q);
}

// sets 'q' to be the identity quaternion
static inline void quat4_identity( quat4*restrict q )
{
  quat4_set_elements(q, 1.0, 0.0,0.0,0.0);
}

static inline void quat4_to_axis_angle_vec( const quat4*restrict q, VecReal*restrict axis_ang );
static inline void quat4_from_axis_angle_vec( quat4*restrict q, const VecReal*restrict axis_ang );
static inline void quat4_from_euler_zxz_array(quat4*restrict q, const VecReal arr[3]);
static inline void quat4_from_euler_xyz_array(quat4*restrict q, const VecReal arr[3]);

static inline void quat4_to_euler_zxz_array(const quat4*restrict q, VecReal arr[3])
{
  VecReal q0 = quat4_x(q);
  VecReal q1 = quat4_y(q);
  VecReal q2 = quat4_z(q);
  VecReal q3 = quat4_w(q);

  // if sin(theta) is too small, then (theta ~ 0) and we have to use a
  // different formula for the axes:
  if ((fabs(q0) < VEC_TOL) && (fabs(q1) < VEC_TOL))
  {
    // phi and psi cannot be separated if theta == 0, so we find phi+psi
    // and return it as phi
    double phi_plus_psi = atan2( 2*q3*q2, q3*q3 - q2*q2 );
    arr[0] = phi_plus_psi;
    arr[1] = 0.0;
    arr[2] = 0.0;
  }
  else
  {
    arr[0] = atan2(q0*q2 + q1*q3, q0*q3 - q1*q2);
    arr[1] = acos(q3*q3 + q2*q2 - q1*q1 - q0*q0);
    arr[2] = atan2(q0*q2 - q1*q3, q0*q3 + q1*q2);
  }
}

static inline void quat4_to_euler_xyz_array(const quat4*restrict q, VecReal arr[3])
{
  VecReal q0 = quat4_w(q);
  VecReal q1 = quat4_x(q);
  VecReal q2 = quat4_y(q);
  VecReal q3 = quat4_z(q);

  arr[0] = atan2(2.0*(q0*q1 - q2*q3), 2.0*(q0*q0+q3*q3)-1.0);
  arr[1] = asin(2.0*(q0*q2+q1*q3));
  arr[2] = atan2(2.0*(q0*q3-q1*q2), 1.0-2.0*(q2*q2+q3*q3));
}

static inline void quat4_to_euler_zxz(const quat4*restrict q, vector3*restrict v)
{
  quat4_to_euler_zxz_array(q, v->elts);
  v->elts[3] = 0.0;
}

static inline void quat4_to_euler_xyz(const quat4*restrict q, vector3*restrict v)
{
  quat4_to_euler_xyz_array(q, v->elts);
  v->elts[3] = 0.0;
}

static inline void quat4_from_euler_zxz(quat4*restrict q, const vector3*restrict euler)
{
  VecReal arr[3];
  arr[0] = vector3_x(euler);
  arr[1] = vector3_y(euler);
  arr[2] = vector3_z(euler);

  quat4_from_euler_zxz_array(q,arr);
}

static inline void quat4_from_euler_xyz(quat4*restrict q, const vector3*restrict euler)
{
  VecReal arr[3];
  arr[0] = vector3_x(euler);
  arr[1] = vector3_y(euler);
  arr[2] = vector3_z(euler);

  quat4_from_euler_xyz_array(q,arr);
}

static inline void quat4_from_euler_array(quat4* q, enum TS_ROT_REPR rot_repr, const VecReal arr[3])
{
  if (rot_repr == TS_ROT_EULER_XYZ)       { quat4_from_euler_xyz_array(q,arr); }
  else if (rot_repr == TS_ROT_EULER_ZXZ)  { quat4_from_euler_zxz_array(q,arr); }
  else if (rot_repr == TS_ROT_AXIS_ANGLE) { quat4_from_axis_angle_vec(q,arr);  }
}

static inline void quat4_to_euler_array(const quat4* q, enum TS_ROT_REPR rot_repr, VecReal arr[3])
{
  if (rot_repr == TS_ROT_EULER_XYZ)       { quat4_to_euler_xyz_array(q,arr); }
  else if (rot_repr == TS_ROT_EULER_ZXZ)  { quat4_to_euler_zxz_array(q,arr); }
  else if (rot_repr == TS_ROT_AXIS_ANGLE) { quat4_to_axis_angle_vec(q,arr);  }
}

static inline void quat4_from_euler(quat4* q, enum TS_ROT_REPR rot_repr, const vector3* v)
{
  quat4_from_euler_array(q,rot_repr,v->elts);
}

static inline void quat4_to_euler(const quat4* q, enum TS_ROT_REPR rot_repr, vector3* v)
{
  quat4_to_euler_array(q,rot_repr,v->elts);
}

#undef MODULE_NAME

#endif /* QUAT4_INLINE_COMMON_H */

