#ifndef QUAT4_INLINE_SSE_DOUBLE_H
#define QUAT4_INLINE_SSE_DOUBLE_H

#include <math.h>
#include <stdlib.h>

#define MODULE_NAME QUAT4_INLINE_SSE_DOUBLE

static inline void quat4_set_scaled(quat4*restrict q1, VecReal scl, const quat4*restrict q2)
{
  __m128d vscl = _mm_set1_pd(scl);
  __m128d qlo = q2->vec[0];
  __m128d qhi = q2->vec[1];

  q1->vec[0] = _mm_mul_pd(vscl, qlo);
  q1->vec[1] = _mm_mul_pd(vscl, qhi);
}

static inline void quat4_set_scaled_vec(quat4*restrict q, VecReal scl, const vector3*restrict v)
{
  __m128d vscl = _mm_set1_pd(scl);
  __m128d vlo = v->vec[0];
  __m128d vhi = v->vec[1];

  q->vec[0] = _mm_mul_pd(vscl, vlo);
  q->vec[1] = _mm_shuffle_pd(
      _mm_mul_pd(vscl, vhi),
      _mm_setzero_pd(),
      _MM_SHUFFLE2(0,0));
}

static inline void quat4_set_scl_vec(quat4*restrict q, VecReal scl, const vector3*restrict v)
{
  quat4_set_scaled_vec(q,scl,v);
}

static inline void quat4_add(quat4*restrict qr, const quat4*restrict qa, const quat4*restrict qb)
{
  qr->vec[0] = _mm_add_pd(qa->vec[0], qb->vec[0]);
  qr->vec[1] = _mm_add_pd(qa->vec[1], qb->vec[1]);
}

static inline void quat4_add_to(quat4*restrict qr, const quat4*restrict qa)
{
  qr->vec[0] = _mm_add_pd(qr->vec[0], qa->vec[0]);
  qr->vec[1] = _mm_add_pd(qr->vec[1], qa->vec[1]);
}

static inline void quat4_sub(quat4*restrict qr, const quat4*restrict qa, const quat4*restrict qb)
{
  qr->vec[0] = _mm_sub_pd(qa->vec[0], qb->vec[0]);
  qr->vec[1] = _mm_sub_pd(qa->vec[1], qb->vec[1]);
}

// a -= b
static inline void quat4_sub_from(quat4*restrict qa, const quat4*restrict qb)
{
  qa->vec[0] = _mm_sub_pd(qa->vec[0], qb->vec[0]);
  qa->vec[1] = _mm_sub_pd(qa->vec[1], qb->vec[1]);
}

static inline void quat4_set(quat4*restrict q1, const quat4*restrict q2)
{
  q1->vec[0] = q2->vec[0];
  q1->vec[1] = q2->vec[1];
}

static inline int quat4_equal(const quat4*restrict q1, const quat4*restrict q2)
{
  __m128d cmp = _mm_and_pd(
    _mm_cmpeq_pd(q1->vec[0],q2->vec[0]),
    _mm_cmpeq_pd(q1->vec[1],q2->vec[1]));

  return _mm_movemask_pd(cmp) == 3;
}

static inline __m128d quat4_sqnorm_sse_val(const quat4*restrict q)
{
  __m128d sqlo = _mm_mul_pd(q->vec[0],q->vec[0]);
  __m128d sqhi = _mm_mul_pd(q->vec[1],q->vec[1]);

  __m128d sqsum = _mm_add_pd(sqlo,sqhi);
  
  return _mm_hadd_pd(sqsum,sqsum);
}

static inline __m128d quat4_norm_sse( const quat4*restrict q )
{
  return _mm_sqrt_pd( quat4_sqnorm_sse_val(q) );
}

static inline VecReal quat4_norm( const quat4*restrict q )
{
  return _mm_cvtsd_f64( quat4_norm_sse(q) );
}

/*! \brief Returns sqrt( (q1.w-q2.w)^2 + (q1.x-q2.x)^2 + (q1.y-q2.y)^2 +
 * (q1.z-q2.z)^2 )
 */
static inline VecReal quat4_distance_to(const quat4*restrict q1, const quat4*restrict q2)
{
  quat4 diff;
  quat4_sub(&diff, q1,q2);
  return quat4_norm(&diff);
}

/*! \brief q = scl*q */
static inline void quat4_scale_by( quat4*restrict q, VecReal scl )
{
  __m128d vscl = _mm_set1_pd(scl);
  q->vec[0] = _mm_mul_pd(vscl, q->vec[0]);
  q->vec[1] = _mm_mul_pd(vscl, q->vec[1]);
}

/*! \brief qr = (a1,a2) * (b1,b2) */
static inline void quat4_mul_sse(quat4*restrict qr, __m128d a1, __m128d a2, __m128d b1, __m128d b2)
{
  __m128d r1,r2;
  __m128d tmp1,tmp2,tmp3,tmp4;

  // qw = (qw1,qw2)
  tmp1 = _mm_mul_pd(a1,b1);                             // X1*X2,Y1*Y2
  tmp2 = _mm_mul_pd(a2,b2);                             // Z1*Z2,W1*W2
  tmp2 = _mm_shuffle_pd(tmp2,tmp2,_MM_SHUFFLE2(0,1));   // W1*W2,Z1*Z2

  r2 = _mm_addsub_pd(tmp2, tmp1);                // W1*W2 - X1*X2, Z1*Z2 + Y1*Y2
  // r2 = _mm_shuffle_pd(r2,r2, _MM_SHUFFLE2(0,1)); // Z1*Z2 + Y1*Y2, W1*W2 - X1*X2


  // qz = (qz1,qz2)
  tmp1 = _mm_shuffle_pd(a1,a1, _MM_SHUFFLE2(0,1)); // Y1,X1
  tmp2 = _mm_shuffle_pd(b2,b2, _MM_SHUFFLE2(0,1)); // W2,Z2

  tmp1 = _mm_mul_pd(tmp1,b1); // Y1*X2,X1*Y2
  tmp2 = _mm_mul_pd(a2,tmp2); // Z1*W2,W1*Z2

  tmp3 = _mm_addsub_pd(tmp1,tmp2); // Y1*X2-Z1*W2, X1*Y2+W1*Z2
  tmp3 = _mm_shuffle_pd(tmp3,tmp3, _MM_SHUFFLE2(0,1)); // X1*Y2+W1*Z2, Y1*X2-Z2*W2

  // r2 = (qz2-qz1,qw2-qw1)
  r2 = _mm_hsub_pd(tmp3,r2); // X1*Y2+W1*Z2-Y1*X2+Z2*W2
                             // W1*W2 - X1*X2 - Z1*Z2 - Y1*Y2,

  // qx = (qx1,qx2)
  tmp1 = _mm_shuffle_pd(a1,a2, _MM_SHUFFLE2(1,0)); // X1,W1
  tmp2 = _mm_shuffle_pd(b2,b1, _MM_SHUFFLE2(0,1)); // W2,X2
  tmp3 = _mm_shuffle_pd(a1,a2, _MM_SHUFFLE2(0,1)); // Y1,Z1
  tmp4 = _mm_shuffle_pd(b2,b1, _MM_SHUFFLE2(1,0)); // Z2,Y2

  tmp1 = _mm_mul_pd(tmp1,tmp2); // X1*W2, W1*X2
  tmp2 = _mm_mul_pd(tmp3,tmp4); // Y1*Z2, Z1*Y2

  tmp2 = _mm_shuffle_pd(tmp2,tmp2, _MM_SHUFFLE2(0,1)); // Z1*Y2, Y1*Z2
  r1   = _mm_addsub_pd(tmp2,tmp1);                     // Z1*Y2 - X1*W2, Y1*Z2 + W1*X2
  r1   = _mm_shuffle_pd(r1,r1,_MM_SHUFFLE2(0,1));


  // qy = (qy1,qy2)
  tmp1 = _mm_mul_pd(a1,b2);             // X1*Z2, Y1*W2
  tmp2 = _mm_mul_pd(a2,b1);             // Z1*X2, W1*Y2

  tmp3 = _mm_addsub_pd(tmp1,tmp2);      // X1*Z2 - Z1*X2, Y1*W2 + W1*Y2
  tmp3 = _mm_shuffle_pd(tmp3,tmp3, _MM_SHUFFLE2(0,1));

  // r1 = (qx,qy)
  r1 = _mm_hsub_pd(r1,tmp3);    // Y1*Z2+W1*X2-Z1*Y2+X1*W2,
                                // Y1*W2+W1*Y2-X1*Z2+Z1*X2 

  qr->vec[0] = r1;
  qr->vec[1] = r2;
}

/*! \brief q1 = q1*q2 */
static inline void quat4_rmul(quat4*restrict q1, const quat4*restrict q2)
{
  quat4_mul_sse(q1, q1->vec[0], q1->vec[1], q2->vec[0], q2->vec[1]);
}

#define QUAT4_CONJ_SSE(qlo,qhi) do { \
  __m128d SIGNMASK = _mm_castsi128_pd(       \
      _mm_set1_epi64x(0x8000000000000000L)); \
  __m128d ZERO = _mm_setzero_pd();           \
  __m128d SIGNMASK_LO = _mm_move_sd(ZERO,SIGNMASK); \
  qlo = _mm_xor_pd(SIGNMASK, qlo);           \
  qhi = _mm_xor_pd(SIGNMASK_LO, qhi);        \
} while(0)

// q1 = q1 * conj(q2)
static inline void quat4_rcmul(quat4*restrict q1, const quat4*restrict q2)
{
  __m128d b1 = q2->vec[0]; // (X2,Y2)
  __m128d b2 = q2->vec[1]; // (Z2,W2)

  QUAT4_CONJ_SSE(b1,b2);
  quat4_mul_sse(q1, q1->vec[0], q1->vec[1], b1,b2);

#if 0
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = -quat4_x(q2); y2 = -quat4_y(q2); z2 = -quat4_z(q2);

  QUAT4_W(q1) = w1*w2 - x1*x2 - y1*y2 - z1*z2;
  QUAT4_X(q1) = w1*x2 + x1*w2 + y1*z2 - z1*y2;
  QUAT4_Y(q1) = w1*y2 + y1*w2 - x1*z2 + z1*x2;
  QUAT4_Z(q1) = w1*z2 + z1*w2 + x1*y2 - y1*x2;
#endif /* 0 */
}

// q1 = q2*q1
static inline void quat4_lmul(quat4*restrict q1, const quat4*restrict q2)
{

  quat4_mul_sse(q1, q2->vec[0], q2->vec[1], q1->vec[0], q1->vec[1]);

#if 0
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = quat4_x(q2); y2 = quat4_y(q2); z2 = quat4_z(q2);

  QUAT4_W(q1) = w2*w1 - x2*x1 - y2*y1 - z2*z1;
  QUAT4_X(q1) = w2*x1 + x2*w1 + y2*z1 - z2*y1;
  QUAT4_Y(q1) = w2*y1 + y2*w1 - x2*z1 + z2*x1;
  QUAT4_Z(q1) = w2*z1 + z2*w1 + x2*y1 - y2*x1;
#endif /* 0 */

#ifdef QUAT_TEST_ASSERTIONS
  quat4 tmp;
  quat4_set(&tmp,q2);
  quat4_rmul(&tmp,q1);

  quat4_sub_from(&tmp,q1);
  tsAssert( fabs(quat4_norm(&tmp)) < 1e-8 );
#endif /* QUAT_TEST_ASSERTIONS */
}

// q1 = conj(q2) * q1
static inline void quat4_lcmul(quat4*restrict q1, const quat4*restrict q2)
{

  __m128d b1 = q2->vec[0]; // (X2,Y2)
  __m128d b2 = q2->vec[1]; // (Z2,W2)

  QUAT4_CONJ_SSE(b1,b2);
  quat4_mul_sse(q1, b1,b2, q1->vec[0], q1->vec[1]);

#if 0
  VecReal w1,x1,y1,z1;
  VecReal w2,x2,y2,z2;
  w1 = quat4_w(q1); x1 = quat4_x(q1); y1 = quat4_y(q1); z1 = quat4_z(q1);
  w2 = quat4_w(q2); x2 = -quat4_x(q2); y2 = -quat4_y(q2); z2 = -quat4_z(q2);

  QUAT4_W(q1) = w2*w1 - x2*x1 - y2*y1 - z2*z1;
  QUAT4_X(q1) = w2*x1 + x2*w1 + y2*z1 - z2*y1;
  QUAT4_Y(q1) = w2*y1 + y2*w1 - x2*z1 + z2*x1;
  QUAT4_Z(q1) = w2*z1 + z2*w1 + x2*y1 - y2*x1;
#endif /* 0 */
}

static inline void quat4_conj( quat4*restrict q )
{
  QUAT4_CONJ_SSE(q->vec[0],q->vec[1]);

#if 0
  QUAT4_X(q) = -quat4_x(q);
  QUAT4_Y(q) = -quat4_y(q);
  QUAT4_Z(q) = -quat4_z(q);
#endif /* 0 */
}

static inline void quat4_normalize( quat4*restrict q )
{
  __m128d ONE = _mm_set1_pd(1.0);
  __m128d rn = quat4_norm_sse(q);
  rn = _mm_div_pd(ONE,rn);

  q->vec[0] = _mm_mul_pd(q->vec[0],rn);
  q->vec[1] = _mm_mul_pd(q->vec[1],rn);

#if 0
  VecReal rn = 1.0/quat4_norm(q);

  QUAT4_W(q) *= rn;
  QUAT4_X(q) *= rn;
  QUAT4_Y(q) *= rn;
  QUAT4_Z(q) *= rn;
#endif /* 0 */
}

static inline void quat4_from_axis_angle_sse( quat4*restrict q, __m128d vlo, __m128d vhi)
{
  __m128d TOL = _mm_set1_pd(VEC_TOLSQ);
  __m128d HALF = _mm_set1_pd(0.5);

  __m128d vsqn = vector3_sqnorm_sse_val(vlo,vhi);
  if (_mm_comige_sd(vsqn,TOL)) {
    __m128d ang, half_ang, cos_halfang, sin_halfang, scl, qlo, qhi;

    ang = _mm_sqrt_pd(vsqn);
    half_ang = _mm_mul_pd(HALF,ang);
    VecReal ha = _mm_cvtsd_f64(half_ang);

    VecReal cos_ha = cos(ha);
    VecReal sin_ha = sin(ha);

    cos_halfang = _mm_set1_pd(cos_ha);
    sin_halfang = _mm_set1_pd(sin_ha);

    scl = _mm_div_pd(sin_halfang,ang);
    qlo = _mm_mul_pd(scl,vlo);
    qhi = _mm_mul_pd(scl,vhi);
    qhi = _mm_shuffle_pd(qhi,cos_halfang, _MM_SHUFFLE2(1,0));

    q->vec[0] = qlo;
    q->vec[1] = qhi;
  } else {
    quat4_identity( q );
  }

#if 0
  VecReal ang = vector3_norm(axis_ang);

  if (ang > VEC_TOL) {
    VecReal cos_halfang = cos(ang/2);
    VecReal sin_halfang = sin(ang/2);

    VecReal scl = sin_halfang / ang;

    QUAT4_W(q) = cos_halfang;
    QUAT4_X(q) = scl*vector3_x(axis_ang);
    QUAT4_Y(q) = scl*vector3_y(axis_ang);
    QUAT4_Z(q) = scl*vector3_z(axis_ang);
  } else {
    quat4_identity( q );
  }
#endif /* 0 */
}

static inline void quat4_from_axis_angle_vec( quat4*restrict q, const VecReal*restrict axis_ang )
{
  __m128d vlo = _mm_loadu_pd(axis_ang+0);
  __m128d vhi = _mm_load_sd(axis_ang+2);
  quat4_from_axis_angle_sse( q, vlo, vhi);

#if 0 
  VecReal ang = sqrt(
      axis_ang[0]*axis_ang[0] + 
      axis_ang[1]*axis_ang[1] + 
      axis_ang[2]*axis_ang[2]);

  VecReal cos_halfang = cos(ang/2);
  VecReal sin_halfang = sin(ang/2);
  if (ang > VEC_TOL) {
    VecReal scl = sin_halfang / ang;

    QUAT4_W(q) = cos_halfang;
    QUAT4_X(q) = scl*axis_ang[0];
    QUAT4_Y(q) = scl*axis_ang[1];
    QUAT4_Z(q) = scl*axis_ang[2];
  } else {
    quat4_identity( q );
  }
#endif /* 0 */
}

static inline void quat4_from_axis_angle( quat4*restrict q, const vector3*restrict axis_ang )
{
  __m128d vlo,vhi;
  vlo = axis_ang->vec[0];
  vhi = axis_ang->vec[1];

  quat4_from_axis_angle_sse( q, vlo, vhi);
}

static inline VecReal quat4_to_axis_angle_separately(const quat4*restrict q, vector3*restrict axis)
{
  double ang;
  // vector3_set(axis, (const vector3*)q);
  __m128d ZERO = _mm_setzero_pd();
  __m128d ONE  = _mm_set1_pd(1.0);
  __m128d q1 = q->vec[0];
  __m128d q2 = q->vec[1];
  __m128d v1 = q1;
  __m128d v2 = _mm_move_sd(ZERO, q2);

  __m128d sqn = vector3_sqnorm_sse_val(v1,v2);
  if (_mm_cvtsd_f64(sqn) >= VEC_TOLSQ) {
    __m128d vn = _mm_sqrt_pd(sqn);
    __m128d rn = _mm_div_pd(ONE,vn);
    v1 = _mm_mul_pd(v1,rn);
    v2 = _mm_mul_pd(v2,rn);

    double sin_halfang = _mm_cvtsd_f64(vn);
    double cos_halfang = _mm_cvtsd_f64( _mm_unpackhi_pd(q2,q2) );
    ang = 2.0 * atan2(sin_halfang,cos_halfang);
  } else {
    ang = 0.0;
    v1 = _mm_setzero_pd();
    v2 = _mm_setzero_pd();
  }

  axis->vec[0] = v1;
  axis->vec[1] = v2;
  return ang;

#if 0
  vector3_set(axis, (const vector3*)q);
  VecReal sin_halfang = vector3_normalize(axis);
  if (sin_halfang >= VEC_TOL) {
    VecReal cos_halfang = quat4_w(q);
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    return ang;
  } else {
    return 0.0;
  }
#endif /* 0 */
}

static inline void quat4_to_axis_angle( const quat4*restrict q, vector3*restrict axis_ang )
{
  __m128d TOL = _mm_set1_pd(VEC_TOLSQ);//*VEC_TOL);
  __m128d ZERO = _mm_setzero_pd();
  __m128d qlo = q->vec[0];
  __m128d qhi = q->vec[1];

  __m128d cos_halfang = _mm_shuffle_pd( qhi,qhi, _MM_SHUFFLE2(1,1) );
  __m128d vsqn = vector3_sqnorm_sse_val(qlo,qhi);


  // check if we can do better...
  if (_mm_comige_sd(vsqn,TOL)) {
    __m128d sin_halfang = _mm_sqrt_pd(vsqn);
    VecReal ang = atan2(_mm_cvtsd_f64(sin_halfang),_mm_cvtsd_f64(cos_halfang));
    __m128d scl = _mm_set1_pd(ang);
    scl = _mm_add_pd(scl,scl);
    scl = _mm_div_pd(scl,sin_halfang);

    axis_ang->vec[0] = _mm_mul_pd(scl,qlo);
    __m128d tmp = _mm_mul_pd(scl,qhi);
    axis_ang->vec[1] = _mm_unpacklo_pd( tmp, ZERO );
  } else {
    axis_ang->vec[0] = ZERO;
    axis_ang->vec[1] = ZERO;
  }

#if 0
  VecReal cos_halfang = quat4_w(q);

  VecReal sin_halfang = vector3_norm((const vector3*)q);
  if (sin_halfang >= VEC_TOL) {
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    VecReal scl = ang / sin_halfang;

    vector3_set_scaled(axis_ang, scl, (const vector3*)q);
  } else {
    vector3_zero(axis_ang);
  }
#endif /* 0 */
}

static inline void quat4_to_axis_angle_vec( const quat4*restrict q, VecReal*restrict axis_ang )
{
  VecReal cos_halfang = quat4_w(q);
  VecReal sin_halfang = vector3_norm((const vector3*)q);

  if (sin_halfang >= VEC_TOL) {
    VecReal ang = 2 * atan2(sin_halfang, cos_halfang);
    VecReal scl = ang / sin_halfang;

    axis_ang[0] = quat4_x(q) * scl;
    axis_ang[1] = quat4_y(q) * scl;
    axis_ang[2] = quat4_z(q) * scl;
  } else {
    axis_ang[0] = 0;
    axis_ang[1] = 0;
    axis_ang[2] = 0;
  }
}

static inline void quat4_set_axis_angle( quat4*restrict q, const vector3*restrict axis, VecReal phi )
{
  VecReal norm = vector3_norm(axis);
  VecReal halfphi = phi*0.5;
  VecReal sin_halfphi = sin(halfphi);
  VecReal cos_halfphi = cos(halfphi);

  VecReal scl = sin_halfphi/norm;

  QUAT4_W(q) = cos_halfphi;
  QUAT4_X(q) = scl*vector3_x(axis);
  QUAT4_Y(q) = scl*vector3_y(axis);
  QUAT4_Z(q) = scl*vector3_z(axis);
}

static inline void quat4_rotate_vector(const quat4*restrict q, vector3*restrict v)
{
  __m128d vlo, vhi, qlo,qhi, qlo_c, qhi_c;
  qlo = q->vec[0];
  qhi = q->vec[1];

  vlo = v->vec[0];
  vhi = _mm_move_sd(_mm_setzero_pd(), v->vec[1]);

  quat4 qv;

  qlo_c = qlo; qhi_c = qhi; QUAT4_CONJ_SSE(qlo_c,qhi_c);

  quat4_mul_sse(&qv, vlo,vhi, qlo_c,qhi_c);
  quat4_mul_sse(&qv, qlo, qhi, qv.vec[0], qv.vec[1]);

  vhi = _mm_move_sd(_mm_setzero_pd(), qv.vec[1]);
  v->vec[0] = qv.vec[0];
  v->vec[1] = vhi;

#if 0
  quat4_set_elements(&qv, 0.0, vector3_x(v), vector3_y(v), vector3_z(v));

  quat4_rcmul(&qv,q);
  quat4_lmul(&qv,q);

  vector3_set_with(v, quat4_x(&qv), quat4_y(&qv), quat4_z(&qv));
#endif /* 0 */
}

static inline void quat4_dq_from_omega(quat4*restrict dq, const quat4*restrict q, VecReal dt, const vector3*restrict omega)
{
  quat4_set_scaled_vec(dq, 0.5*dt, omega);
  quat4_rmul(dq,q);
}

static inline void quat4_rotate_delta(quat4*restrict dq, const quat4*restrict q, VecReal dt, const vector3*restrict omega)
{
  quat4_dq_from_omega(dq, q, dt, omega);

  quat4 qtmp;
  quat4_add(&qtmp, q, dq);
  quat4_normalize(&qtmp);
  quat4_sub(dq, &qtmp, q);
}

static inline void quat4_from_euler_zxz_array(quat4*restrict q, const VecReal arr[3])
{
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  __m128d qlo,qhi, tmp1,tmp2;

  __m128d cos_htheta_scl = _mm_set1_pd(cos_half_theta);
  __m128d sin_htheta_scl = _mm_set1_pd(sin_half_theta);
  __m128d cos_hpsi_scl = _mm_set1_pd(cos_half_psi);
  __m128d sin_hpsi_scl = _mm_set1_pd(sin_half_psi);
  __m128d hphiv = _mm_set_pd(sin_half_phi,cos_half_phi); // (cos_half_phi,sin_half_phi)

  // QUAT4_Y(q) = sin_half_theta*(sin_half_phi*cos_half_psi - cos_half_phi*sin_half_psi);
  // QUAT4_X(q) = sin_half_theta*(cos_half_phi*cos_half_psi + sin_half_phi*sin_half_psi);

  tmp1 = _mm_mul_pd(hphiv, cos_hpsi_scl);
  tmp1 = _mm_shuffle_pd(tmp1,tmp1,_MM_SHUFFLE2(0,1));

  tmp2 = _mm_mul_pd(hphiv, sin_hpsi_scl);
  qlo  = _mm_addsub_pd(tmp1,tmp2);
  qlo  = _mm_shuffle_pd(qlo,qlo,_MM_SHUFFLE2(0,1));
  qlo  = _mm_mul_pd(sin_htheta_scl,qlo);

  // QUAT4_W(q) = cos_half_theta*(cos_half_phi*cos_half_psi - sin_half_phi*sin_half_psi);
  // QUAT4_Z(q) = cos_half_theta*(cos_half_phi*sin_half_psi + sin_half_phi*cos_half_psi);

  __m128d hpsiv = _mm_move_sd(sin_hpsi_scl,cos_hpsi_scl);
  tmp1 = _mm_unpacklo_pd(hphiv,hphiv); // (cos_half_phi,cos_half_phi)
  tmp2 = _mm_unpackhi_pd(hphiv,hphiv); // (sin_half_phi,sin_half_phi)

  tmp1 = _mm_mul_pd(tmp1,hpsiv);
  tmp2 = _mm_mul_pd(tmp2,_mm_shuffle_pd(hpsiv,hpsiv,_MM_SHUFFLE2(0,1)));
  qhi  = _mm_addsub_pd(tmp1,tmp2);
  qhi  = _mm_shuffle_pd(qhi,qhi,_MM_SHUFFLE2(0,1));
  qhi  = _mm_mul_pd(cos_htheta_scl,qhi);

  q->vec[0] = qlo;
  q->vec[1] = qhi;

  /*
  QUAT4_Y(q) = sin_half_theta*(sin_half_phi*cos_half_psi - cos_half_phi*sin_half_psi);
  QUAT4_X(q) = sin_half_theta*(cos_half_phi*cos_half_psi + sin_half_phi*sin_half_psi);

  QUAT4_W(q) = cos_half_theta*(cos_half_phi*cos_half_psi - sin_half_phi*sin_half_psi);
  QUAT4_Z(q) = cos_half_theta*(cos_half_phi*sin_half_psi + sin_half_phi*cos_half_psi);
  */


#if 0
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  QUAT4_W(q) = cos_half_phi*cos_half_theta*cos_half_psi - sin_half_phi*cos_half_theta*sin_half_psi;
  QUAT4_X(q) = cos_half_phi*sin_half_theta*cos_half_psi + sin_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_Y(q) = sin_half_phi*sin_half_theta*cos_half_psi - cos_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_Z(q) = cos_half_phi*cos_half_theta*sin_half_psi + sin_half_phi*cos_half_theta*cos_half_psi;
#endif /* 0 */
}

static inline void quat4_from_euler_xyz_array(quat4*restrict q, const VecReal arr[3])
{
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  __m128d hpsiv = _mm_set_pd(sin_half_psi, cos_half_psi);

  __m128d cos_hphi_scl = _mm_set1_pd(cos_half_phi);
  __m128d sin_hphi_scl = _mm_set1_pd(sin_half_phi);
  __m128d cos_htheta_scl = _mm_set1_pd(cos_half_theta);
  __m128d sin_htheta_scl = _mm_set1_pd(sin_half_theta);

  __m128d qlo,qhi,tmp1,tmp2,src1,src2;

  // QUAT4_Y(q) = sin_half_theta*(cos_half_phi*cos_half_psi) - cos_half_theta*(sin_half_phi*sin_half_psi);
  // QUAT4_X(q) = sin_half_theta*(cos_half_phi*sin_half_psi) + cos_half_theta*(sin_half_phi*cos_half_psi);
  //
  //         YX = sin_half_theta*(src1                     ) - cos_half_theta*(src2                     );

  src1 = _mm_mul_pd(cos_hphi_scl, hpsiv);
  src2 = _mm_mul_pd(sin_hphi_scl, hpsiv);
  src2 = _mm_shuffle_pd(src2,src2,_MM_SHUFFLE2(0,1));

  tmp1 = _mm_mul_pd(sin_htheta_scl,src1);
  tmp2 = _mm_mul_pd(cos_htheta_scl,src2);
  qlo  = _mm_addsub_pd(tmp1,tmp2);
  qlo  = _mm_shuffle_pd(qlo,qlo, _MM_SHUFFLE2(0,1));

  // QUAT4_W(q) = cos_half_theta*(cos_half_phi*cos_half_psi) - sin_half_theta*(sin_half_phi*sin_half_psi);
  // QUAT4_Z(q) = cos_half_theta*(cos_half_phi*sin_half_psi) + sin_half_theta*(sin_half_phi*cos_half_psi);
  //
  //         WZ = cos_half_theta*(src1                     ) - sin_half_theta*(src2                     );

  tmp1 = _mm_mul_pd(cos_htheta_scl,src1);
  tmp2 = _mm_mul_pd(sin_htheta_scl,src2);
  qhi  = _mm_addsub_pd(tmp1,tmp2);
  qhi  = _mm_shuffle_pd(qhi,qhi, _MM_SHUFFLE2(0,1));

  q->vec[0] = qlo;
  q->vec[1] = qhi;

#if 0
  VecReal cos_half_phi   = cos(0.5*arr[0]);
  VecReal sin_half_phi   = sin(0.5*arr[0]);

  VecReal cos_half_theta = cos(0.5*arr[1]);
  VecReal sin_half_theta = sin(0.5*arr[1]);

  VecReal cos_half_psi   = cos(0.5*arr[2]);
  VecReal sin_half_psi   = sin(0.5*arr[2]);

  QUAT4_W(q) = cos_half_phi*cos_half_theta*cos_half_psi - sin_half_phi*sin_half_theta*sin_half_psi;
  QUAT4_X(q) = cos_half_phi*sin_half_theta*sin_half_psi + sin_half_phi*cos_half_theta*cos_half_psi;
  QUAT4_Y(q) = cos_half_phi*sin_half_theta*cos_half_psi - sin_half_phi*cos_half_theta*sin_half_psi;
  QUAT4_Z(q) = cos_half_phi*cos_half_theta*sin_half_psi + sin_half_phi*sin_half_theta*cos_half_psi;
#endif /* 0 */
}



#undef MODULE_NAME

#endif /* QUAT4_INLINE_SSE_DOUBLE_H */

