#include "tubesheet.h"
#include "ts_bonds.h"
#include "ts_common_algo.h"

#include <string.h>

#define IMPROPER_ORDER(a,b,c) ( (a) | ((b) << 2) | ((c) << 4) )
#define IMPROPER_ORDER_OF(order, wh) (((order) >> ((wh-1)*2)) & 0x03)

static int compare_bond_patterns(const ts_bond_pattern* bt1, const ts_bond_pattern* bt2)
{
  if (bt1->type != bt2->type) return bt1->type - bt2->type;
  if (bt1->nbodies != bt2->nbodies) return bt1->nbodies - bt2->nbodies;
  for (int i=0; i < bt1->nbodies; i++) {
    int ptype_diff = bt1->ptypes[i] - bt2->ptypes[i];
    if (ptype_diff != 0) return ptype_diff;
  }

  for (int i=0; i < bt1->nbodies-1; i++) {
    int btype_diff = bondtype_to_int(bt1->btypes[i]) - bondtype_to_int(bt2->btypes[i]);
    if (btype_diff != 0) return btype_diff;
  }

  return 0;
}

static TSResult add_bond_pattern_to_list(DATAVECTOR(ts_bond_pattern)* patternlist,
    ts_bond_pattern* new_pattern, ts_bond_pattern** pattern_ptr)
{
  if (!datavector_contains(patternlist, new_pattern))
  {
    return TS_NEW_ERROR3( TS_ERR_UNSUPPORTED_OPERATION,
        "attempt to finalize a bond pattern on a list that it was not allocated on" );
  }

  // add temporary marker
  const char* saved_name = new_pattern->name;
  static char marker[5] = { 'M', 'A', 'R', 'K', 0 };
  new_pattern->name = marker+0;

  qsort(patternlist->data, patternlist->size, patternlist->elt_size,
      (int(*)(const void*,const void*))compare_bond_patterns);

  ts_bond_pattern* found = NULL;
  ts_bond_pattern* pat = DATAVECTOR_BEGIN(ts_bond_pattern,patternlist);
  for (size_t i=0; i < patternlist->size; i++)
  {
    if (pat[i].name == marker+0)
    {
      found = pat+i;
    }
  }

  if (found)
  {
    found->name = saved_name;
    if (pattern_ptr) { *pattern_ptr = found; }
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR2(TS_ERR_SIMULATION_STATE,
        ("Newly-added term '%s' has disappeared", saved_name));
  }
}

static TSResult add_bond_pattern_to_sim(ts_sim* sim,ts_bond_pattern* new_bt, ts_bond_pattern** pattern_ptr)
{
  return add_bond_pattern_to_list(&sim->bond_patterns1, new_bt, pattern_ptr);
}

TSResult ts_add_bonded_pattern(
    ts_sim* sim,
    int nbodies,
    int* ptypes,
    ts_bond_type* btypes,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp)
{
  ERROR_IF( TS_ERR_INCORRECT_NUM_BODIES, nbodies != pot->nbodies);

  enum TS_BOND_TERM pattern_type;
  switch (pot->nbodies)
  {
    case 1: pattern_type = TS_BOND_TERM_ONEBODY;   break;
    case 2: pattern_type = TS_BOND_TERM_TWOBODY;   break;
    case 3: pattern_type = TS_BOND_TERM_THREEBODY; break;
    case 4: pattern_type = TS_BOND_TERM_FOURBODY;  break;

    case 0:
    default: return TS_NEW_ERROR( TS_ERR_UNSUPPORTED_POTENTIAL );
  }

  ts_bond_pattern* pat = ts_bond_pattern_new(&sim->bond_patterns1,
      nbodies,ptypes,btypes,pattern_type,btName);
  pat->potential = ts_potential_incref(pot);

  RETURN_ON_ERROR(add_bond_pattern_to_sim(sim,pat,patp));

  return TS_OK;
}

TSResult ts_direct_add_bond_pattern(ts_sim* sim, ts_bond_pattern* pat0, ts_bond_pattern** patp)
{
  switch(pat0->type)
  {
    case TS_BOND_TERM_ZEROBODY:
    case TS_BOND_TERM_ONEBODY:
    case TS_BOND_TERM_TWOBODY:
    case TS_BOND_TERM_THREEBODY:
    case TS_BOND_TERM_FOURBODY:
      RETURN_ON_ERROR(
          ts_add_bonded_pattern(sim,
            pat0->nbodies,
            pat0->ptypes,
            pat0->btypes,
            pat0->name,
            pat0->potential,
            patp)
          );
      break;

    case TS_BOND_TERM_IMPROPER:
      RETURN_ON_ERROR( ts_add_improper_pattern(sim,
            pat0->ptypes,
            pat0->btypes,
            pat0->name,
            pat0->potential,
            patp)
          );
      break;

    case TS_BOND_TERM_NONBOND:
      RETURN_ON_ERROR(
          ts_add_nonbond_pattern(sim,
            pat0->ptypes[0],
            pat0->ptypes[1],
            pat0->name,
            pat0->potential,
            patp)
          );
      break;

    default:
      return TS_NEW_ERROR2(TS_ERR_UNSUPPORTED_BOND_PATTERN,
          ("unsupported bond pattern (%d)", (int)pat0->type));
  }

  return TS_OK;
}

TSResult ts_add_improper_pattern(
    ts_sim* sim,
    int* ptypes,
    ts_bond_type* btypes,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp)
{
  ERROR_IF( TS_ERR_INCORRECT_NUM_BODIES, pot->nbodies != 4);

  enum TS_BOND_TERM pattern_type = TS_BOND_TERM_IMPROPER;

  ts_bond_pattern* pat = ts_bond_pattern_new(&sim->bond_patterns1,
      4,ptypes,btypes,pattern_type,btName);
  pat->potential = ts_potential_incref(pot);
  
  RETURN_ON_ERROR( add_bond_pattern_to_sim(sim,pat,patp) );

  return TS_OK;
}

TSResult ts_add_nonbond_pattern(
    ts_sim* sim,
    int ptype1,
    int ptype2,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp)
{
  int ptypes[NBODY_LIMIT] = PTYPES_INIT;
  ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;

  ERROR_IF( TS_ERR_INCORRECT_NUM_BODIES, pot->nbodies != 2);

  ptypes[0] = ptype1;
  ptypes[1] = ptype2;
  btypes[0] = BT_NONE;

  ts_bond_pattern* pat = ts_bond_pattern_new(&sim->nonbond_patterns1,
      2,ptypes,btypes,TS_BOND_TERM_NONBOND,btName);
  pat->potential = ts_potential_incref(pot);

  RETURN_ON_ERROR( add_bond_pattern_to_list(&sim->nonbond_patterns1, pat, patp) );

  if (patp) { *patp = pat; }

  return TS_OK;
}

TSResult ts_add_extra_term(ts_sim * sim, const struct ts_potential_term *term)
{
  struct ts_potential_term *new_term;

  new_term = datavector_new_item(&sim->extra_terms);
  ERROR_ON_NULL_ALLOC(new_term);

  *new_term = *term;

  return TS_OK;
}

TSResult ts_add_distance_constraint(ts_sim *sim, int ptype1, int ptype2, ts_bond_type btype, const char *cName, double dist)
{
  int ptypes[NBODY_LIMIT] = PTYPES_INIT;
  ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;
  ts_ptypes_zero(ptypes);
  ptypes[0] = ptype1;
  ptypes[1] = ptype2;
  btypes[0] = btype;

  ts_bond_pattern *bond_pattern = ts_bond_pattern_new(&sim->bond_patterns1,
      2,ptypes,btypes,TS_BOND_TERM_CONSTRAINT,cName);

  DistanceConstraint *constraint = tsp_distance_constraint_new(dist);
  bond_pattern->potential = ts_potential_incref((Potential*)constraint);
  return add_bond_pattern_to_sim(sim,bond_pattern,NULL);
}

void ts_finalize_bond_patterns(DATAVECTOR(ts_bond_pattern) *patterns)
{
  ts_bond_pattern *pattern;
  size_t i,n;

  for(pattern=patterns->data,n=patterns->size, i=0; i < n; i++)
  {
    ts_potential_decref(pattern[i].potential);
  }

  datavector_free( patterns );
}

size_t ts_count_bond_patterns(ts_sim* sim)
{
  return sim->bond_patterns1.size;
}

static int has_equal_ptypes(
    int* ptypesA, ts_bond_type* btypesA, 
    int* ptypesB, ts_bond_type* btypesB, int n)
{
  return (memcmp(ptypesA,ptypesB, n*sizeof(*ptypesA)) == 0) &&
    (memcmp(btypesA,btypesB, (n-1)*sizeof(*btypesA)) == 0);
}

static inline ts_bond_pattern* bondpattern_linear_search(DATAVECTOR(ts_bond_pattern)* pattern_list,
    int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT], enum TS_BOND_TERM bondpattern)
{
  ts_bond_pattern* bt_ptr = (ts_bond_pattern*)pattern_list->data;
  for (size_t i=0; i < pattern_list->size; i++)
  {
    if ((bt_ptr[i].type == bondpattern) && (bt_ptr[i].nbodies == nbodies))
    {
      if (has_equal_ptypes(bt_ptr[i].ptypes, bt_ptr[i].btypes, ptypes, btypes, nbodies)) return bt_ptr+i;
    }
  }

  return NULL;
}

static inline ts_bond_pattern* bondpattern_binary_search(DATAVECTOR(ts_bond_pattern) *pattern_list,
    int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT], enum TS_BOND_TERM bondpattern)
{
  ts_bond_pattern key;
  key.type = bondpattern;
  key.nbodies = nbodies;
  memcpy(&key.ptypes, ptypes, nbodies*sizeof(key.ptypes[0]));
  memcpy(&key.btypes, btypes, (nbodies-1)*sizeof(key.btypes[0]));

  ts_bond_pattern* bt_ptr = bsearch(&key,
      pattern_list->data, pattern_list->size, pattern_list->elt_size,
      (int (*)(const void*,const void*))compare_bond_patterns);

  return bt_ptr;
}


#define BONDTERM_BSEARCH_LIMIT 5
static inline ts_bond_pattern* bondpattern_search(DATAVECTOR(ts_bond_pattern*) *pattern_list, int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT], enum TS_BOND_TERM bondpattern)
{
  return (pattern_list->size < BONDTERM_BSEARCH_LIMIT) 
    ? bondpattern_linear_search(pattern_list, nbodies, ptypes, btypes, bondpattern)
    : bondpattern_binary_search(pattern_list, nbodies, ptypes, btypes, bondpattern);
}
#undef BONDTERM_BSEARCH_LIMIT

static inline void swap_ints(int*restrict a, int*restrict b) { int tmp = *a; *a = *b; *b = tmp; }
static void swap_int_list(int* beg, size_t n) {
  int* end = beg+n-1;
  while(beg < end) { swap_ints(beg,end); beg++; end--; }
}
static inline void swap_bondtypes(ts_bond_type*restrict a, ts_bond_type*restrict b)
{
  ts_bond_type tmp = *a;
  *a = *b;
  *b = tmp;
}
static void swap_bondtype_list(ts_bond_type* beg, size_t n) {
  ts_bond_type* end = beg+n-1;
  while(beg < end) { swap_bondtypes(beg,end); beg++; end--; }
}

#define BT_GREATER(a,b) ( bondtype_to_int((a)) > bondtype_to_int((b)) )
#define ISGREATER(i,j) \
  ( (ptypes[i] > ptypes[j]) || ((ptypes[i] == ptypes[j]) && (BT_GREATER(btypes[i],btypes[j]))) )

#define SWAP(i,j) do { swap_ints(ptypes+i,ptypes+j); swap_bondtypes(btypes+i,btypes+j); } while (0)
#define A_GT_B 1
#define A_GT_C 2
#define B_GT_C 4
static void sort_improper_ptypes_and_btypes( int*restrict ptypes, ts_bond_type*restrict btypes, uint32_t* reordering)
{
  tsAssert( (NBODY_LIMIT == 4)  && (NBOND_LIMIT == 3) );

  int sortcase = 0;
  if (ISGREATER(0,1)) sortcase += 1;
  if (ISGREATER(0,2)) sortcase += 2;
  if (ISGREATER(1,2)) sortcase += 4;

  *reordering = 0;

  switch (sortcase) {
    case 0:
      // A <= B <= C: ABC
      *reordering = IMPROPER_ORDER(1,2,3);
      break;

    case A_GT_B:
      // A > B, A <= C, B <= C: BAC. swap 0&1
      SWAP(0,1);
      *reordering = IMPROPER_ORDER(2,1,3);
      break;

    case A_GT_B+A_GT_C:
      // A > B, A > C, B <= C: BCA
      // swap 0&1 (ABC -> BAC). swap 1&2 (BAC -> BCA).
      SWAP(0,1);
      SWAP(1,2);
      *reordering = IMPROPER_ORDER(2,3,1);
      break;

    case B_GT_C:
      // A <= B, A <= C, B > C: ACB. swap 1&2 (ABC -> ACB)
      SWAP(1,2);
      *reordering = IMPROPER_ORDER(1,3,2);
      break;

    case A_GT_B+B_GT_C:
      // A > B, B > C: CBA. swap 0&2 (ABC -> CBA)
      SWAP(0,2);
      *reordering = IMPROPER_ORDER(3,2,1);
      break;

    case A_GT_C+B_GT_C:
      // A > C, B > C: CAB. swap 0&1 (ABC -> BAC). swap 0&2 (BAC -> CAB).
      SWAP(0,1);
      SWAP(0,2);
      *reordering = IMPROPER_ORDER(3,1,2);
      break;

    case A_GT_B+A_GT_C+B_GT_C:
      // A > B, A > C, B > C: CBA. swap 0&2 (ABC -> CBA).
      SWAP(0,2);
      *reordering = IMPROPER_ORDER(3,2,1);
      break;

    case A_GT_C:
      // A <= B, A > C, B <= C: not possible
      tsAssert( 0 && "Not possible." );
      *reordering = 0;
      break;

    default:
      // should not occur
      tsAssert( 0 && "Not possible." );
      *reordering = 0;
      break;
  }
}

struct improper_sort_closure
{
  int ptypes[NBOND_LIMIT];
  int btypes[NBOND_LIMIT];
};

static int improper_compare_particle_by_types(struct improper_sort_closure* closure, int* i0, int* i1)
{
  int delta = closure->ptypes[*i0] - closure->ptypes[*i1];
  return delta != 0 ? delta : (closure->btypes[*i0] - closure->btypes[*i1]);
}

static void improper_sort_particle_indices_by_type(ts_sim* sim, size_t pinds[NBODY_LIMIT])
{
  struct improper_sort_closure closure;
  int indices[3];

  for (int i=0; i < 3; i++) {
    indices[i] = i;

    closure.ptypes[i] = TS_PTYPE_IND(sim,pinds[i+1]);
    closure.btypes[i] = bondtype_get_type( ts_particle_get_bond_type(TS_PARTICLE(sim,pinds[i+1]),TS_PARTICLE(sim,pinds[0])) );
  }

  insertion_sort_elements(indices, 3, sizeof(indices[0]), &closure, (GenericCompareFunc)improper_compare_particle_by_types);

  size_t tmp_inds[NBODY_LIMIT];
  for (int i=0; i < NBODY_LIMIT; i++) { tmp_inds[i] = pinds[i]; }
  for (int i=0; i < 3; i++) {
    pinds[i+1] = tmp_inds[indices[i]+1];
  }
}
#undef COMPARE
#undef SWAP
#undef A_GT_B
#undef A_GT_C
#undef B_GT_C

static void correct_type_order(int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT])
{
  if (ptypes[0] > ptypes[nbodies-1])
  {
    swap_int_list(ptypes,nbodies);
    swap_bondtype_list(btypes,nbodies-1);
    for (int i=0; i < nbodies-1; i++) {
      btypes[i] = bondtype_reverse_direction(btypes[i]);
    }
  }
}

static ts_bond_pattern* get_improper_pattern(ts_sim* sim, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT])
{
  uint32_t reordering = 0;
  sort_improper_ptypes_and_btypes(ptypes+1, btypes, &reordering);
  return bondpattern_search(&sim->bond_patterns1, 4, ptypes, btypes, TS_BOND_TERM_IMPROPER);
}

static ts_bond_pattern* get_proper_pattern(ts_sim* sim, int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT], enum TS_BOND_TERM bondpattern)
{
  correct_type_order(nbodies,ptypes,btypes);

  return (bondpattern != TS_BOND_TERM_NONBOND) 
    ? bondpattern_search(&sim->bond_patterns1, nbodies, ptypes, btypes, bondpattern)
    : bondpattern_search(&sim->nonbond_patterns1, nbodies, ptypes, btypes, bondpattern);
}

ts_bond_pattern* ts_get_bond_pattern(ts_sim* sim, int nbodies, 
    int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT], enum TS_BOND_TERM bondpattern)
{
  tsAssert((nbodies >= 1) && (nbodies <= NBODY_LIMIT) 
      && "Invalid nbodies passed to ts_get_bond_pattern");

  return (bondpattern != TS_BOND_TERM_IMPROPER) 
    ? get_proper_pattern(sim,nbodies,ptypes,btypes,bondpattern) 
    : get_improper_pattern(sim,ptypes,btypes);
}

ts_bond_pattern* ts_lookup_bond_pattern(ts_sim* sim, const char* name)
{
  DATAVECTOR_FOREACH(ts_bond_pattern, pat, &sim->bond_patterns1)
  {
    if (strcmp(name,pat->name) == 0) { return pat; }
  }

  return NULL;
}

static TSResult add_twobody_constraint(ts_sim* sim, size_t pi, size_t pj, DistanceConstraint* constraint)
{
  ts_constraint_entry entry;
  entry.pA = TS_PARTICLE(sim, pi);
  entry.pB = TS_PARTICLE(sim, pj);
  entry.constraint = constraint;
  vector3_zero( &entry.dr0 );

  datavector_add(&sim->constraints, &entry);

  return TS_OK;
}

static void improper_bondlist_entry_init(ts_sim* sim, ts_potential_term* entry, const ts_bond_pattern* bondpattern, size_t pinds[4])
{
  improper_sort_particle_indices_by_type(sim, pinds);
  size_t reordered_pinds[NBODY_LIMIT];
  reordered_pinds[0] = pinds[0];
  reordered_pinds[1] = pinds[IMPROPER_ORDER_OF(bondpattern->info,1)];
  reordered_pinds[2] = pinds[IMPROPER_ORDER_OF(bondpattern->info,2)];
  reordered_pinds[3] = pinds[IMPROPER_ORDER_OF(bondpattern->info,3)];

  ts_bondlist_entry_init(entry, bondpattern, 
      reordered_pinds[0],
      reordered_pinds[1],
      reordered_pinds[2],
      reordered_pinds[3]);
}

static int get_num_bodies_bondpattern_requires(enum TS_BOND_TERM bondtype)
{
  switch (bondtype) {
    case TS_BOND_TERM_ONEBODY:    return 1;
    case TS_BOND_TERM_TWOBODY:    return 2;
    case TS_BOND_TERM_THREEBODY:  return 3;

    case TS_BOND_TERM_FOURBODY:   return 4;
    case TS_BOND_TERM_IMPROPER:   return 4;

    case TS_BOND_TERM_NONBOND:    return 2;
    case TS_BOND_TERM_CONSTRAINT: return 2;

    default: return -1;
  }
}

static inline int particles_are_correctly_ordered(const ts_sim* sim, enum TS_BOND_TERM bondtype, int nbody, size_t pinds[NBODY_LIMIT])
{
  UNUSED_PARAM(sim);
  return (bondtype == TS_BOND_TERM_ONEBODY) ||
    ( (bondtype == TS_BOND_TERM_IMPROPER) 
        ? (pinds[1] < pinds[2]) && (pinds[2] < pinds[3])
        : (pinds[0] < pinds[nbody-1]) );
}

static TSResult try_to_make_bond(ts_sim* sim, enum TS_BOND_TERM bondtype, size_t pinds[NBODY_LIMIT], ts_bond_type btypes[NBOND_LIMIT])
{
  if (! sim->bonds_enabled[(int)bondtype]) { return TS_OK; }

  int nbody = get_num_bodies_bondpattern_requires(bondtype);
  ERROR_IF( TS_ERR_UNSUPPORTED_POTENTIAL, nbody < 0);
  ERROR_IF( TS_ERR_UNSUPPORTED_POTENTIAL, nbody > NBODY_LIMIT );

  if (!particles_are_correctly_ordered(sim, bondtype, nbody, pinds)) return TS_OK;

  int ptypes[NBODY_LIMIT] = {0};
  ts_ptypes_zero(ptypes);

  for (int i=0; i < nbody; i++) {
    ptypes[i] = TS_PTYPE_IND(sim,pinds[i]);
  }

  ts_bond_pattern* bondpattern = ts_get_bond_pattern(sim, nbody, ptypes, btypes, bondtype);
  if (bondpattern == NULL) return TS_OK;

  tsAssert(bondpattern->nbodies == nbody);

  if (bondtype == TS_BOND_TERM_CONSTRAINT) {
    RETURN_ON_ERROR( add_twobody_constraint(sim, pinds[0], pinds[1], 
          (DistanceConstraint*)bondpattern->potential) );
  } else {
    ts_potential_term entry;

    if (bondtype == TS_BOND_TERM_IMPROPER) {
      improper_bondlist_entry_init(sim, &entry, bondpattern, pinds);
    } else {
      ts_bondlist_entry_init(&entry, bondpattern, pinds[0], pinds[1], pinds[2], pinds[3]);
    }

    RETURN_ON_ERROR( ts_potential_term_list_add(&sim->bondlist, &entry) );
  }
  
  return TS_OK;
}

TSResult ts_nonbond_list_add(struct ts_potential_term_list* nblist, size_t pi, size_t pj, ts_bond_pattern* bondpattern)
{
  if (pi > pj) { size_t tmp = pi; pi = pj; pj = tmp; }

  ts_potential_term entry;
  ts_bondlist_entry_init(&entry,
      bondpattern, pi,pj, TS_NULL_PARTICLE, TS_NULL_PARTICLE);

  return ts_potential_term_list_add(nblist, &entry);
}

TSResult ts_build_nonbonded_potential_terms(ts_sim* sim)
{
  RETURN_ON_ERROR( ts_potential_term_list_reset(&sim->nonbond_list) );

  if (! sim->nonbonds_enabled) return TS_OK;

  int ptypes[NBODY_LIMIT] = PTYPES_INIT;
  ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;

  double rsq_nbl = ((sim->nonbond_cutoff > 0) && (sim->nonbond_padding > 0))
    ? square(sim->nonbond_cutoff + sim->nonbond_padding)
    : 0.0;

  RETURN_ON_ERROR( ts_potential_term_list_reset(&sim->nonbond_list) );
  if (sim->num_particles == 0) return TS_OK;

  for(size_t pi=0; pi < sim->num_particles-1; pi++)
  {
    int ptype_i = TS_PTYPE_IND(sim,pi);
    if (ptype_i == 0) continue;

    for (size_t pj=(pi+1); pj < sim->num_particles; pj++)
    {
      int ptype_j = TS_PTYPE_IND(sim,pj);
      if (ptype_j == 0) continue;

      if (ts_particle_is_bonded_to(TS_PARTICLE(sim,pi), TS_PARTICLE(sim,pj))) continue;

      ptypes[0] = ptype_i;
      ptypes[1] = ptype_j;

      ts_bond_pattern* nbType = ts_get_bond_pattern(sim, 2, ptypes, btypes, TS_BOND_TERM_NONBOND);
      if (nbType != NULL)
      {
        tsAssert(nbType->nbodies == 2);

        vector3 dr;
        if ( !(rsq_nbl > 0.0) || (vector3_sqdistance(PARTICLE_POS(sim,pi),PARTICLE_POS(sim,pj),&dr) <= rsq_nbl))
        {
          RETURN_ON_ERROR( ts_nonbond_list_add(&sim->nonbond_list, pi, pj, nbType) );
        }
      }
    }
  }

  RETURN_ON_ERROR( ts_potential_term_list_finished_adding(&sim->nonbond_list) );

  ts_reset_max_displacement_origins(sim);
  TS_LOG(sim, TS_LOG_INFO, TS_LOG_NONBONDS, "%lu NONBOND interactions\n", sim->nonbond_list.terms.size);

  return TS_OK;
}

static inline void set_particle_indices(size_t pinds[NBODY_LIMIT], const ts_sim*restrict sim,
    const ts_particle*restrict pA, 
    const ts_particle*restrict pB, 
    const ts_particle*restrict pC, 
    const ts_particle*restrict pD)
{
  pinds[0] = (pA != NULL) ? TS_PIND(sim, pA) : TS_NULL_PARTICLE;
  pinds[1] = (pB != NULL) ? TS_PIND(sim, pB) : TS_NULL_PARTICLE;
  pinds[2] = (pC != NULL) ? TS_PIND(sim, pC) : TS_NULL_PARTICLE;
  pinds[3] = (pD != NULL) ? TS_PIND(sim, pD) : TS_NULL_PARTICLE;
}

static inline void set_bond_types(ts_bond_type btypes[NBOND_LIMIT], 
    ts_bond_type bi, ts_bond_type bj, ts_bond_type bk) 
{ 
  btypes[0] = bi;
  btypes[1] = bj;
  btypes[2] = bk;
}

static ts_particle* get_other_particle(const ts_particle*restrict p, size_t bond_index)
{
  return p == p->bonds[bond_index].pi ? p->bonds[bond_index].pj : p->bonds[bond_index].pi;
}

TSResult ts_reset_potential(ts_sim* sim)
{
  RETURN_ON_ERROR( ts_potential_term_list_reset(&sim->bondlist) );
  RETURN_ON_ERROR( ts_potential_term_list_reset(&sim->nonbond_list) );

  return TS_OK;
}

#define GET_BONDTYPE(p,ind) ( ts_particle_bond_type_at_index((p),(ind)) )
TSResult ts_build_bonded_potential_terms(ts_sim* sim)
{
  RETURN_ON_ERROR( ts_potential_term_list_reset(&sim->bondlist) );

  size_t pinds[NBODY_LIMIT];
  ts_bond_type btypes [NBOND_LIMIT];

  ts_particle*restrict pEnd = sim->particles + sim->num_particles;
  for (ts_particle* pA = sim->particles; pA != pEnd; pA++)
  {
    set_particle_indices(pinds, sim, pA, NULL,NULL,NULL);
    set_bond_types(btypes, BT_NONE, BT_NONE, BT_NONE);

    RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_ONEBODY, pinds, btypes) );

    for (size_t bi = 0; bi < pA->num_bonds; bi++)
    {
      ts_particle* pB = get_other_particle(pA, bi);
      ts_bond_type bondtype_AB = GET_BONDTYPE(pA,bi);

      set_particle_indices(pinds, sim, pA,pB, NULL,NULL);
      set_bond_types(btypes, bondtype_AB, BT_NONE,BT_NONE);

      RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_TWOBODY, pinds, btypes) );
      RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_CONSTRAINT, pinds, btypes) );

      for (size_t bj=0; bj < pA->num_bonds; bj++)
      {
        if (bi == bj) continue;
        ts_particle* pC = get_other_particle(pA, bj);
        ts_bond_type bondtype_AC = GET_BONDTYPE(pA, bj);

        set_particle_indices(pinds, sim, pB,pA,pC, NULL);
        set_bond_types(btypes, 
            bondtype_reverse_direction(bondtype_AB),
            bondtype_AC,
            BT_NONE);

        RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_THREEBODY, pinds, btypes) );

        for(size_t bk=0; bk < pB->num_bonds; bk++)
        {
          ts_particle* pD = get_other_particle(pB, bk);
          if (pD == pA) continue;
          ts_bond_type bondtype_BD = GET_BONDTYPE(pB,bk);

          set_particle_indices(pinds, sim, pC,pA,pB,pD);
          set_bond_types(btypes,
            bondtype_reverse_direction(bondtype_AC),
            bondtype_AB,
            bondtype_BD);

          RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_FOURBODY, pinds, btypes) );
        }

        for (size_t bk=0; bk < pA->num_bonds; bk++)
        {
          if (bk == bi) continue;
          if (bk == bj) continue;

          ts_particle* pD = get_other_particle(pA, bk);
          ts_bond_type bondtype_AD = GET_BONDTYPE(pA,bk);

          set_particle_indices(pinds, sim, pA,pB,pC,pD);
          set_bond_types(btypes,
            bondtype_AB,
            bondtype_AC,
            bondtype_AD);

          RETURN_ON_ERROR( try_to_make_bond(sim, TS_BOND_TERM_IMPROPER, pinds, btypes) );
        }

      }
    }
  }

  // add any "extra terms"
  DATAVECTOR_FOREACH(struct ts_potential_term, term, &sim->extra_terms)
  {
    RETURN_ON_ERROR( ts_potential_term_list_add(&sim->bondlist, term) );
  }

  RETURN_ON_ERROR( ts_potential_term_list_finished_adding(&sim->bondlist) );

  return TS_OK;
}
#undef GET_BONDTYPE

size_t ts_count_potential_terms(ts_sim* sim)
{
  return sim->bondlist.terms.size;
}

size_t ts_count_constraints(ts_sim* sim)
{
  return sim->constraints.size;
}

ts_bond_pattern* ts_bond_pattern_new(DATAVECTOR(ts_bond_pattern)* list,
    int nbodies, int* ptypes, ts_bond_type* btypes,
    enum TS_BOND_TERM pattern_type, const char* name)
{
  ts_bond_pattern* bond_pattern = datavector_new_item(list);
  bond_pattern->type = pattern_type;
  bond_pattern->nbodies = nbodies;

  bond_pattern->info = 0;
  bond_pattern->name = ts_strdup(NO_ALLOCATOR, name);

  bond_pattern->term_group = 0;

  memcpy(bond_pattern->ptypes, ptypes, sizeof(bond_pattern->ptypes));
  memcpy(bond_pattern->btypes, btypes, sizeof(bond_pattern->btypes));

  if (pattern_type != TS_BOND_TERM_IMPROPER)
  {
    correct_type_order(nbodies,bond_pattern->ptypes,bond_pattern->btypes);
  }
  else
  {
    bond_pattern->info = 0;
    sort_improper_ptypes_and_btypes( bond_pattern->ptypes + 1, bond_pattern->btypes, &bond_pattern->info );
  }

  return bond_pattern;
}

void ts_sim_free_bond_patterns(ts_sim* sim)
{
  UNUSED_PARAM(sim);
}

int ts_rigid_body_is_bonded_to(ts_sim* sim, ts_rigid_body* b0, ts_rigid_body* b1)
{
  if (ts_rigid_body_is_bonded_to_particle(sim, b1, ts_get_particle(sim,b0->com_particle))) return 1;

  for(size_t i = 0; i < b0->type_ptr->num_sites; i++)
  {
    ts_particle* pi = ts_get_particle(sim, b0->site_indices[i]);

    if (ts_rigid_body_is_bonded_to_particle(sim, b1, pi)) return 1;
  }

  return 0;
}


static inline TSResult maybe_add_bond_info_to_list(
    ts_sim* sim,
    const ts_particle*restrict p0,
    const ts_particle*restrict p1,
    DATAVECTOR(struct ts_rigid_body_bond_info)* infolist)
{
  struct ts_rigid_body_bond_info info;
  info.btype = ts_particle_get_bond_type(p0, p1);

  if (bondtype_is_bond(info.btype))
  {
    info.pi = TS_PIND(sim,p0);
    info.pj = TS_PIND(sim,p1);
    if (!datavector_add(infolist, &info))
      return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  return TS_OK;
}

static TSResult add_any_bonds_to_particle(ts_sim* sim,
    const ts_particle*restrict p,
    const ts_rigid_body*restrict bd,
    DATAVECTOR(struct ts_rigid_body_bond_info)* infolist)
{
  const ts_particle* com_particle = ts_get_particle(sim, bd->com_particle);
  RETURN_ON_ERROR( maybe_add_bond_info_to_list(sim, p, com_particle, infolist) );

  for(size_t i = 0; i < bd->type_ptr->num_sites; i++)
  {
    ts_particle* pi = ts_get_particle(sim, bd->site_indices[i]);
    RETURN_ON_ERROR( maybe_add_bond_info_to_list(sim, p, pi, infolist) );
  }

  return TS_OK;
}

static inline ts_bond_type find_site_bonded_to_particle(const ts_sim* sim, const ts_rigid_body* rigid, const ts_particle* p, ts_particle** rpp)
{
  for(size_t i = 0; i < rigid->type_ptr->num_sites; i++)
  {
    ts_particle* p_site = ts_get_particle(sim, rigid->site_indices[i]);
    ts_bond_type bt = ts_particle_get_bond_type(p, p_site);
    if (bondtype_is_bond(bt))
    { 
      if (rpp) { *rpp = p_site; }
      return bt;
    }
  }

  if (rpp) { *rpp = NULL; }
  return bondtype_none();
}

static ts_bond_type find_rb_bonded_to_particle(const ts_sim* sim, const ts_rigid_body* rigid, const ts_particle* p, ts_particle** rpp)
{
  ts_particle* p_com = ts_get_particle(sim, rigid->com_particle);
  ts_bond_type bt = ts_particle_get_bond_type(p, p_com);
  if (bondtype_is_bond(bt))
  { 
    if (rpp) { *rpp = p_com; }
    return bt;
  }

  return find_site_bonded_to_particle(sim, rigid, p, rpp);
}

int ts_rigid_body_find_first_bond(ts_sim* sim, 
    const ts_rigid_body*restrict b0,
    const ts_rigid_body*restrict b1,
    struct ts_rigid_body_bond_info* binfo)
{
  int swapped = 0;

  if (b0 > b1)
  {
    const ts_rigid_body* btmp = b0; b0 = b1; b1 = btmp;
    swapped = 1;
  }

  ts_particle* b0_particle = NULL;
  ts_particle* b1_particle = NULL;
  ts_bond_type bt = bondtype_none();

  b0_particle = ts_get_particle(sim,b0->com_particle);
  b1_particle = NULL;
  bt = find_rb_bonded_to_particle(sim, b1, b0_particle, &b1_particle);
  if (bondtype_is_bond(bt)) { goto success; }

  b0_particle = NULL;
  b1_particle = ts_get_particle(sim,b1->com_particle);
  bt = find_site_bonded_to_particle(sim, b0, b1_particle, &b0_particle);
  if (bondtype_is_bond(bt)) { goto success; }

  for(size_t i=0; i < b0->type_ptr->num_sites; i++)
  {
    b0_particle = ts_get_particle(sim, b0->site_indices[i]);
    b1_particle = NULL;
    bt = find_site_bonded_to_particle(sim, b1, b0_particle, &b1_particle);
    if (bondtype_is_bond(bt)) { goto success; }
  }

  return 0;

success:
  if (swapped)
  {
    ts_particle* ptmp = b0_particle;
    b0_particle = b1_particle;
    b1_particle = ptmp;
  }
  binfo->pi = TS_PIND(sim,b0_particle);
  binfo->pj = TS_PIND(sim,b1_particle);
  binfo->btype = bt;

  return 1;
}

#define BOND_INFO_DEFAULT_SIZE 16
TSResult ts_rigid_body_find_bonds(ts_sim* sim,
    const ts_rigid_body*restrict b0,
    const ts_rigid_body*restrict b1,
    DATAVECTOR(struct ts_rigid_body_bond_info)* infolist)
{
  RETURN_ON_ERROR( add_any_bonds_to_particle(sim, ts_get_particle(sim, b0->com_particle), b1, infolist) );

  for(size_t i = 0; i < b0->type_ptr->num_sites; i++)
  {
    ts_particle* pi = ts_get_particle(sim, b0->site_indices[i]);

    RETURN_ON_ERROR( add_any_bonds_to_particle(sim, pi, b1, infolist) );
  }

  return TS_OK;
}
#undef BOND_INFO_DEFAULT_SIZE

int ts_rigid_body_is_bonded_to_particle(ts_sim* sim, ts_rigid_body* bd, ts_particle* p)
{
  return bondtype_is_bond( find_rb_bonded_to_particle(sim,bd,p,NULL) );
}



