#ifndef TS_BONDS_H
#define TS_BONDS_H

#include "tubesheet.h"

#define MODULE_NAME TS_BONDS

typedef struct ts_bond_pattern ts_bond_pattern;
struct ts_bond_pattern
{
  const char* name;
  Potential* potential;

  enum TS_BOND_TERM type;
  int enabled;

  uint32_t info;

  int term_group;

  int nbodies;
  int ptypes[NBODY_LIMIT];
  ts_bond_type btypes[NBOND_LIMIT];
};

typedef struct ts_constraint_entry ts_constraint_entry;
struct ts_constraint_entry
{
  ts_particle* pA;
  ts_particle* pB;

  DistanceConstraint* constraint;

  vector3 rA;
  vector3 rB;

  vector3 dr0;
};

TSResult ts_add_bond_type(ts_sim* sim, const char* btName, ts_bond_type bt);
ts_bond_type ts_get_bond_type(const ts_sim* sim, const char* btName);
size_t ts_count_bond_types(const ts_sim* sim);
TSResult ts_foreach_bond_type(ts_sim* sim,
    TSResult (*eachfunc)(void*, const char*, ts_bond_type),
    void* ud);

uint16_t ts_find_unused_bond_type(const ts_sim* sim);

struct ts_bond_type_iter { int64_t state; const char* key; ts_bond_type bt; };
void ts_bond_type_iter_init(ts_sim* sim, struct ts_bond_type_iter* iter);
int ts_bond_type_iter_next(ts_sim* sim, struct ts_bond_type_iter* iter);

struct ts_rigid_body_bond_info
{
  size_t pi;
  size_t pj;
  ts_bond_type btype;
};

TSResult ts_rigid_body_find_bonds(ts_sim* sim, 
    const ts_rigid_body*restrict b0,
    const ts_rigid_body*restrict b1,
    DATAVECTOR(struct ts_rigid_body_bond_info)* infolist);

int ts_rigid_body_find_first_bond(ts_sim* sim, 
    const ts_rigid_body*restrict b0,
    const ts_rigid_body*restrict b1,
    struct ts_rigid_body_bond_info* binfo);

int ts_rigid_body_is_bonded_to(ts_sim* sim, ts_rigid_body* b0, ts_rigid_body* b1);
int ts_rigid_body_is_bonded_to_particle(ts_sim* sim, ts_rigid_body* bd, ts_particle* pt);


TSResult ts_direct_add_bond_pattern(ts_sim* sim, ts_bond_pattern* pat0, ts_bond_pattern** patp);

TSResult ts_add_bonded_pattern(
    ts_sim* sim,
    int nbodies,
    int* ptypes,
    ts_bond_type* btypes,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp);

TSResult ts_add_improper_pattern(
    ts_sim* sim,
    int* ptypes,
    ts_bond_type* btypes,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp);

TSResult ts_add_nonbond_pattern(
    ts_sim* sim,
    int ptype1,
    int ptype2,
    const char* btName,
    Potential* pot,
    ts_bond_pattern** patp);

TSResult ts_add_extra_term(ts_sim * sim, const struct ts_potential_term *term);

TSResult ts_add_distance_constraint(ts_sim* sim, int ptype1, int ptype2, ts_bond_type btype, const char* cName, double dist);

ts_bond_pattern* ts_bond_pattern_new(DATAVECTOR(ts_bond_pattern)* list,
    int nbodies, int* ptypes, ts_bond_type* btypes,
    enum TS_BOND_TERM pattern_type, const char* name);
void ts_sim_free_bond_patterns(ts_sim* sim);
size_t ts_count_bond_patterns(ts_sim* sim);
ts_bond_pattern* ts_get_bond_pattern(ts_sim* sim, int nbodies, int ptypes[NBODY_LIMIT], ts_bond_type btypes[NBODY_LIMIT-1], enum TS_BOND_TERM bondterm);
ts_bond_pattern* ts_lookup_bond_pattern(ts_sim* sim, const char* name);

void ts_finalize_bond_patterns(DATAVECTOR(ts_bond_pattern) *patterns);

static inline void
ts_bondlist_entry_init(
    ts_potential_term* entry,
    const ts_bond_pattern* term,
    size_t pA, size_t pB, size_t pC, size_t pD)
{
  entry->name      = term->name;
  entry->potential = term->potential;
  entry->pattern   = term;

  entry->particle_indices[0] = pA;
  entry->particle_indices[1] = pB;
  entry->particle_indices[2] = pC;
  entry->particle_indices[3] = pD;
}


TSResult ts_nonbond_list_add(
    struct ts_potential_term_list* nblist,
    size_t pi, size_t pj, ts_bond_pattern* bondterm);


TSResult ts_reset_potential(ts_sim* sim);

TSResult ts_build_bonded_potential_terms(ts_sim* sim);

TSResult ts_build_nonbonded_potential_terms(ts_sim* sim);

static inline TSResult ts_build_potential(ts_sim* sim)
{
  RETURN_ON_ERROR( ts_build_bonded_potential_terms(sim) );
  RETURN_ON_ERROR( ts_build_nonbonded_potential_terms(sim) );
  return TS_OK;
}

size_t ts_count_potential_terms(ts_sim* sim);
size_t ts_count_constraints(ts_sim* sim);



#undef MODULE_NAME

#endif /* TS_BONDS_H */

