#include "ts_chem_driver.h"

#include "ts_numerical.h"

#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <stdbool.h>

#include "ts_common_algo.h"

#define TS_CHEMDRIVER_DEBUG_POLYMERIZATION 0
#define EXPLICIT_NAN_CHECK 0
#define DEBUG_UPDATES_ENABLED 0
#define RATE_CHECK_TOL 1e-8

int tsch_debug_updates = 0;

#if DEBUG_UPDATES_ENABLED
#  define UPDATE_DEBUG_LEVEL (tsch_debug_updates)
#else
#  define UPDATE_DEBUG_LEVEL (0)
#endif /* DEBUG_UPDATES_ENABLED */


#define TS_CHEMDRIVER_SITE_CODES  " CBLRSteclrs"
const char ts_chem_driver_site_codes[] = TS_CHEMDRIVER_SITE_CODES;
STATIC_ASSERT( sizeof(TS_CHEMDRIVER_SITE_CODES) == TS_CHEMDRIVER_STATE_MAX+1 );

const char* ts_chemdriver_event_names[TS_CHEMDRIVER_EVT_MAX] = {
  "none", "k_on", "k_off", "monomer", "long", "lat", "lat_seam", "hydrolysis"
};

const char* ts_chemdriver_barrier_model_names[TS_CHEMDRIVER_BARRIER_MAX] = {
  "none", "metropolis", "kawasaki", "proportional", "infinite"
};

const char* ts_chemdriver_site_type_names[TS_CHEMDRIVER_NUM_TYPES] = {
  "empty", "monomer", "long_ifc", "lat_ifc"
};

/* Several places need to return whether something is related to the
 * plus or minus ends
 */
enum TIP_DIRECTION {
  TIP_NONE  = 0x0,
  TIP_MINUS = 0x1,
  TIP_PLUS  = 0x2,
  TIP_BOTH  = TIP_MINUS | TIP_PLUS,
};

static TSResult update_seam_reaction_hook(void* data, ts_lattice* latt, const ts_lattice_move* move);
static TSResult update_tip_reaction_hook(void *data, ts_lattice *latt, const ts_lattice_move* move);

static TSResult
chemdriver_move_hooks(void* data, ts_lattice* latt, const ts_lattice_move* move)
{
  RETURN_ON_ERROR( update_seam_reaction_hook(data,latt,move) );
  RETURN_ON_ERROR( update_tip_reaction_hook(data,latt,move) );
  return TS_OK;
}

static TSResult depolymerize_move_hook(void* data, ts_lattice* drv, const ts_lattice_move* move);

static int chem_driver_gen_moves(void* closure, ts_lattice* latt, 
    const int *restrict rv, const int *restrict cv, size_t nsites,
    spin_t* states, double* rates, size_t *restrict siteinds, int move_max,
    intptr_t* cdata, TSResult* errp);

static struct ts_lattice_movetype simple_moves = {
  .name = "ChemDriver simple moves",
  .func = chem_driver_gen_moves,
  .closure = NULL
};

void chem_driver_delete(void *p);

ts_chem_driver* ts_chem_driver_new(size_t nrow, size_t ncol, TSResult* errp)
{
  ts_chem_driver* drv;
  size_t num_sites;
  TSResult err;

  drv = ts_alloc(NO_ALLOCATOR, sizeof(*drv));
  if (drv == NULL)
  {
    *errp = TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
        "cannot allocate chem_driver structure");
    return NULL;
  }

  drv->row_range = NULL;
  memset(drv,0,sizeof(*drv));

  num_sites = nrow*ncol;
  err = ts_lattice_initialize(&drv->latt, nrow,ncol,
      TS_CHEMDRIVER_STATE_MAX*num_sites);
  if (ON_ERROR_ADD_FRAME(err), !IS_OKAY(err))
  {
    goto handle_error;
  }

  err = ts_lattice_resize_changes(&drv->latt, TS_LATTICE_MAX_UPDATES + nrow);
  if (ON_ERROR_ADD_FRAME(err), !IS_OKAY(err))
  {
    goto handle_error;
  }

  drv->row_range = ts_alloc(NO_ALLOCATOR, 2*ncol*sizeof drv->row_range[0]);
  if (drv->row_range == NULL)
  {
    err = TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
        "cannot allocate chem_driver structure");
    goto handle_error;
  }

  err = ts_chem_driver_reset(drv);
  if (ON_ERROR_ADD_FRAME(err), !IS_OKAY(err))
  {
    goto handle_error;
  }

  drv->latt.destructor = chem_driver_delete;

  drv->ifclog.entries = NULL;
  drv->ifclog.n = 0;
  drv->ifclog.capa = 0;

  return drv;

handle_error:
  *errp = err;
  chem_driver_delete(drv);

  return NULL;
}

static TSResult pspin_lateral_change(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static TSResult pspin_lateral_make_bond(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static TSResult pspin_kon(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static TSResult pspin_koff(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static TSResult pspin_find_row_range(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static TSResult pspin_rollback_ifclog(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin);
static void rebuild_moves_hook(ts_lattice *latt, void *ud);

static inline TSResult
register_pspin(ts_chem_driver *drv, int spin, ts_lattice_pseudospin_func func, void* data)
{
  return ts_lattice_register_pseudospin(&drv->latt, spin_from_int(spin), func, data );
}

static void reset_parameters(ts_chem_driver* drv);

TSResult ts_chem_driver_reset(ts_chem_driver* drv)
{
  size_t i,ncol,cc_scratch_sz;
  int* alloc;

  ts_lattice_reset(&drv->latt);
  cc_scratch_sz = drv->latt.num_sites + 
    ts_lattice_connected_components_working_size(&drv->latt);

  alloc = ts_realloc(NO_ALLOCATOR, 
      drv->connected_components_scratch,
      cc_scratch_sz * sizeof(drv->connected_components_scratch[0]));
  ERROR_ON_NULL_ALLOC(alloc);
  drv->connected_components_scratch = alloc;

  drv->pitch_monomers = TS_CHEMDRIVER_DEFAULT_PITCH;

  drv->latt.boundary.row = TS_LATTICE_BOUNDARY_FIXED;
  drv->latt.boundary.col = ts_chem_driver_pitch_rows(drv->pitch_monomers);

  drv->latt.row_coupling_dist = TS_CHEMDRIVER_MAX_ROW_COUPLING;
  drv->latt.col_coupling_dist = TS_CHEMDRIVER_MAX_COL_COUPLING;

  drv->latt.rebuild_moves_hook.callback  = rebuild_moves_hook;
  drv->latt.rebuild_moves_hook.user_data = drv;

  RETURN_ON_ERROR( ts_lattice_add_movetype((ts_lattice*)drv, &simple_moves) );

  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_LATERAL_SEAM    , pspin_lateral_make_bond, NULL));

  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_LATERAL_ZERO     , pspin_lateral_change, NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_LATERAL_TUBELIKE , pspin_lateral_change, NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_LATERAL_SHEETLIKE, pspin_lateral_change, NULL));

  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_K_ON_BENT_LEFT     , pspin_kon         , NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_K_ON_BENT_RIGHT    , pspin_kon         , NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_K_ON_STRAIGHT_LEFT , pspin_kon         , NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_K_ON_STRAIGHT_RIGHT, pspin_kon         , NULL));

  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_K_OFF              , pspin_koff        , NULL));

  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_FIND_ROW_RANGE     , pspin_find_row_range, NULL));
  RETURN_ON_ERROR(register_pspin(drv, PSEUDO_ROLLBACK_IFCLOG    , pspin_rollback_ifclog, NULL));

  RETURN_ON_ERROR(
      ts_lattice_add_move_hook((ts_lattice*)drv, chemdriver_move_hooks, NULL) );
  (void)depolymerize_move_hook;

  reset_parameters(drv);

  drv->coupling_edge_dist = 0;

  drv->barrier_model = TS_CHEMDRIVER_BARRIER_NONE;
  drv->barrier_param = 0.0;

  drv->seam_allowed_thresh = ts_chem_driver_num_protofilaments(drv)-1;

  drv->dt = 0.0;
  drv->t_max = 0.0;
  drv->forbidden_energy = TS_CHEMDRIVER_FORBIDDEN_ENERGY;

  for (i=0,ncol=drv->latt.ncol; i < 2*ncol; i++)
  {
    drv->row_range[i] = 0;
  }

  return TS_OK;
}

ts_chem_driver *ts_chem_driver_retain(ts_chem_driver *driver)
{
  ts_lattice_retain(&driver->latt);
  return driver;
}

void ts_chem_driver_release(ts_chem_driver *driver)
{
  ts_lattice_release(&driver->latt);
}

void chem_driver_delete(void *p)
{
  ts_chem_driver *drv = p;

  ts_lattice_finalize(&drv->latt);
  ts_free(NO_ALLOCATOR, drv->connected_components_scratch);
  ts_free(NO_ALLOCATOR, drv->row_range);
  ts_free(NO_ALLOCATOR, drv->ifclog.entries);
  ts_free(NO_ALLOCATOR, drv);
}

static void reset_parameters(ts_chem_driver* drv)
{
  size_t i,j,k;
  double invalid = nan("2989");  /* nan with tag 0xBAD */

  /* set rates to zero (zero rates do nothing) */
  for (i=0; i < ARRAYLEN(drv->rate_prefactors); i++) {
    drv->rate_prefactors[i] = 0.0;
  }

  for (i=0; i < ARRAYLEN(drv->monomer_energies); i++) {
    drv->monomer_energies[i] = invalid;
  }

  /* set empty sites to 0.0 */
  drv->monomer_energies[0] = 0.0;

  for (i=0; i < ARRAYLEN(drv->interface_energies); i++) {
    for (j=0; j < ARRAYLEN(drv->interface_energies[i]); j++) {
      for (k=0; k < ARRAYLEN(drv->interface_energies[i][j]); k++) {

        /* set empty sites to 0.0 */
        if (i == TS_CHEMDRIVER_STATE_EMPTY ||
            j == TS_CHEMDRIVER_STATE_EMPTY || 
            k == TS_CHEMDRIVER_STATE_EMPTY)
        {
          drv->interface_energies[i][j][k] = 0.0;
        }
        else
        {
          drv->interface_energies[i][j][k] = invalid;
        }
      }
    }
  }

  drv->interface_energies[0][0][0] = 0.0;

  for (i=0; i < ARRAYLEN(lateral_states); i++) {
    int latifc = lateral_states[i];

    for (j=0; j < ARRAYLEN(monomer_states_all); j++) {
      int mer = monomer_states_all[j];

      for (k=0; k < ARRAYLEN(longitudinal_states); k++) {
        int longifc = longitudinal_states[k];

        drv->interface_energies[latifc][mer][longifc] = 0;
      }
    }
  }

  /* Zero out coupling terms */
  for (i=0; i < ARRAYLEN(drv->coupling_terms); i++) {
    drv->coupling_terms[i] = 0;
  }

  drv->num_coupling = 0;
  drv->coupling_edge_dist = 0;
  for (i=0; i < ARRAYLEN(drv->coupling_energies); i++)
  {
    drv->coupling_energies[i][0] = 0.0;
    drv->coupling_energies[i][1] = 0.0;
  }

  drv->num_dissociation = 0;
  for (i=0; i < ARRAYLEN(drv->dissociation_barrier_terms); i++)
  {
    drv->dissociation_barrier_terms[i] = 0;
  }

  for (i=0; i < ARRAYLEN(drv->dissociation_barrier_energies); i++)
  {
    drv->dissociation_barrier_energies[i] = 0.0;
  }

  for (i=0; i < (int)TSCH_NSTATES_LAT; i++)
  {
    int j,k,l;
    for (j=0; j < (int)TSCH_NSTATES_LAT; j++)
    {
      for (k=0; k < (int)TSCH_NSTATES_LONG; k++)
      {
        for (l=0; l < (int)TSCH_NSTATES_LONG; l++)
        {
          drv->coupling_lookup[i][j][k][l] = -1;
        }
      }
    }
  }
}


static void zero_populations(struct tsch_populations *pops);
static double calc_site_energy(ts_chem_driver* drv, int r0, int r1, int c0, int c1,
    struct tsch_populations *pops);

TSResult ts_chem_driver_energy(ts_chem_driver* drv, double* energyp,
    struct tsch_populations *pops)
{
  size_t r,c;
  size_t nrow, ncol;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  if (pops) { zero_populations(pops); }

  *energyp = calc_site_energy(drv, 1,nrow-2, 0,ncol-1, pops);
  return TS_OK;
}

static void calc_delta_energy(ts_chem_driver* drv, size_t r, size_t c,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict energies);

TSResult ts_chem_driver_energy_change(ts_chem_driver* latt, size_t r, size_t c, spin_t s, double* dE_p)
{
  const enum TS_CHEMDRIVER_SITE_STATE st = s.val;

  calc_delta_energy(latt, r, c, &st, &st+1, dE_p);
  return TS_OK;
}

static MAYBE_UNUSED double calc_site_dep_energy_lateral_sites(ts_chem_driver* drv, int ra, int ca);

static void calc_delta_energy_lateral_sites_0(ts_chem_driver* drv, int ra, int ca,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict energies)
{
  const enum TS_CHEMDRIVER_SITE_STATE *st;
  int old_state;
  double en_old;

  old_state = TS_LATTICE_SPIN(drv, ra, ca);
  assert(TS_LATTICE_SPIN(drv, ra+2,ca) == old_state);

  en_old = calc_site_dep_energy_lateral_sites(drv, ra,ca);

  for (st=states_begin; st != states_end; st++, energies++)
  {
    int new_state = *st;

    double en_new = 0.0;
    if (new_state == old_state)
    {
      *energies = 0.0;
      continue;
    }

    TS_LATTICE_SPIN(drv, ra+0, ca) = new_state;
    TS_LATTICE_SPIN(drv, ra+2, ca) = new_state;
    en_new = calc_site_dep_energy_lateral_sites(drv, ra,ca);

    *energies = en_new - en_old;
  }

  TS_LATTICE_SPIN(drv, ra+0, ca) = old_state;
  TS_LATTICE_SPIN(drv, ra+2, ca) = old_state;
}

static void calc_delta_energy_lateral_sites(ts_chem_driver* drv, int ra, int ca,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict energies);

TSResult ts_chem_driver_energy_change_lateral(ts_chem_driver *drv, size_t r, size_t c, spin_t s, double *energyp)
{
  const enum TS_CHEMDRIVER_SITE_STATE st = s.val;
  int is_alpha, is_seam;

  if (!tsch_is_lateral_bond(s.val))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "spin %d is not a  valid lateral bond state", s.val);
  }

  if (!ts_chem_driver_is_lateral(drv, r,c))
  {
    goto invalid_site;
  }

  is_alpha = ts_chem_driver_is_alpha_lateral(drv, r,c);
  is_seam = ts_chem_driver_is_seam(drv, r,c);

  if ((is_alpha && is_seam) || (!is_alpha && !is_seam))
  {
    goto invalid_site;
  }

  calc_delta_energy_lateral_sites(drv, r, c, &st, &st + 1, energyp);

  return TS_OK;

invalid_site:
    TS_RAISE_ERROR_MESG(TS_ERR_INVALID_ARGUMENT,
        "site must be valid for a lateral bond move");
}

TSResult ts_chem_driver_print_configuration(ts_chem_driver* drv, FILE* f)
{
  ts_lattice_print_configuration((ts_lattice*)drv,f);

  fprintf(f, "\nChemical driver configuration:\n");
  fprintf(f, "  Num pitch monomers: %d\n", drv->pitch_monomers);

  if (drv->barrier_model == TS_CHEMDRIVER_BARRIER_PROPORTIONAL) {
    fprintf(f, "\n  Barrier model: %s  a = %7.4f\n",
        ts_chemdriver_barrier_model_names[drv->barrier_model],
        drv->barrier_param);
  } else {
    fprintf(f, "\n  Barrier model: %s\n",
        ts_chemdriver_barrier_model_names[drv->barrier_model]);
  }

  fprintf(f, "\n  Rate prefactors:\n");
  for (size_t k=0; k < TS_CHEMDRIVER_EVT_MAX; k++) {
    fprintf(f, "    %10s : %7.4f\n", ts_chemdriver_event_names[k], drv->rate_prefactors[k]);
  }

  fprintf(f, "\n  Monomer state energies:\n  ");
  for(size_t k=1; k < ARRAYLEN(monomer_states_all); k++) {
    fprintf(f, "  %c   : %8.4f  ",
        ts_chem_driver_site_codes[monomer_states_all[k]],
        drv->monomer_energies[monomer_states_all[k]]);
  }
  fprintf(f, "\n");

  fprintf(f, "\n  Longitudinal interface energies:\n");
  for(size_t i=1; i < ARRAYLEN(monomer_states_all); i++) {
    int mer1 = monomer_states_all[i];
    for(size_t j=1; j < ARRAYLEN(longitudinal_states); j++) {
      int ifc = longitudinal_states[j];
      fprintf(f, "  ");
      for (size_t k=1; k < ARRAYLEN(monomer_states_all); k++) {
        int mer2 = monomer_states_all[k];
        fprintf(f, "  %c%c%c : %8.4f  ",
            ts_chem_driver_site_codes[mer1],
            ts_chem_driver_site_codes[ifc],
            ts_chem_driver_site_codes[mer2],
            drv->interface_energies[mer1][ifc][mer2]);
      }
      fprintf(f,"\n");
    }
  }

  fprintf(f, "\n  Lateral interface energies:\n");
  for(size_t i=1; i < ARRAYLEN(monomer_states_all); i++) {
    int mer1 = monomer_states_all[i];
    for(size_t j=1; j < ARRAYLEN(lateral_states); j++) {
      int ifc = lateral_states[j];
      fprintf(f, "  ");
      for (size_t k=1; k < ARRAYLEN(monomer_states_all); k++) {
        int mer2 = monomer_states_all[k];
        fprintf(f, "  %c%c%c : %8.4f  ",
            ts_chem_driver_site_codes[mer1],
            ts_chem_driver_site_codes[ifc],
            ts_chem_driver_site_codes[mer2],
            drv->interface_energies[mer1][ifc][mer2]);
      }
      fprintf(f,"\n");
    }
  }
  fprintf(f,"\n");

  return TS_OK;
}

TSResult ts_chem_driver_count_bonds(ts_chem_driver* drv,
    const int*restrict rv, const int*restrict cv,
    int*restrict nlongv, int* restrict nlatv, size_t n)
{
  int nrow = drv->latt.nrow;
  int ncol = drv->latt.ncol;

  for (size_t i=0; i < n; i++)
  {
    int r = rv[i], c = cv[i];
    nlongv[i] = nlatv[i] = 0;

    if ((r < 1) || (r >= nrow-1) ||
        (c < 0) || (c >= ncol  ) ||
        (TS_LATTICE_SPIN(drv,r,c) == 0)) { continue; }

    nlongv[i] += (TS_LATTICE_SPIN(drv,r-1,c) != 0);
    nlongv[i] += (TS_LATTICE_SPIN(drv,r+1,c) != 0);

    {
      int rm1,cm1;
      int rp1,cp1;

      rm1 = r; cm1 = c-1; ts_lattice_wrap_coord((ts_lattice*)drv, &rm1,&cm1);
      rp1 = r; cp1 = c+1; ts_lattice_wrap_coord((ts_lattice*)drv, &rp1,&cp1);

      if ((rm1 >= 0) && (rm1 < nrow) && (TS_LATTICE_SPIN(drv,rm1,cm1) != 0))
      {
        nlatv[i]++;
      }

      if ((rp1 >= 0) && (rp1 < nrow) && (TS_LATTICE_SPIN(drv,rp1,cp1) != 0))
      {
        nlatv[i]++;
      }
    }
  }

  return TS_OK;
}

/* ifclog functions */

TSResult tsch_ifclog_resize(ts_chem_driver *drv, size_t new_capa)
{
  if (new_capa != drv->ifclog.capa)
  {
    struct tsch_ifclog_entry *entries;
    size_t bytes;

    if (new_capa > SIZE_MAX/sizeof(*entries))
    {
      TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
          "too many entries requested (%zu) (size would exceed SIZE_MAX)",
          new_capa);
    }

    bytes = new_capa * sizeof(*entries);

    entries = realloc(drv->ifclog.entries, bytes);
    if (entries == NULL)
    {
      TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
          "Cannot resize ifclog (requested size is %zu entries, %zu bytes)",
          new_capa, bytes);
    }

    drv->ifclog.capa = new_capa;
    drv->ifclog.entries = entries;
  }

  if (drv->ifclog.capa > drv->ifclog.n)
  {
    size_t n,capa;

    n = drv->ifclog.n;
    capa = drv->ifclog.capa;

    memset(&drv->ifclog.entries[n], 0, (capa-n)*sizeof(drv->ifclog.entries[0]));
  }

  return TS_OK;
}


static void record_bending(struct tsch_populations *pops, int side1, int mer, int side2)
{
  if (!pops) 
  {
    return;
  }

  if (side1 == TS_CHEMDRIVER_STATE_EMPTY ||
      mer   == TS_CHEMDRIVER_STATE_EMPTY ||
      side2 == TS_CHEMDRIVER_STATE_EMPTY)
  {
    pops->bending_pop[TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY]++;
  }
  else
  {
    pops->bending_pop[side1][mer][side2]++;
  }
}

TSResult ts_chem_driver_add_coupling_energy(ts_chem_driver *drv, int terms[4], double elo, double ehi)
{
  tsch_state_t *couple;
  size_t ncouple, max;
  uint32_t packed;

  ncouple = sizeof drv->coupling_terms[0];
  max = sizeof drv->coupling_terms / ncouple;
  if ((size_t)drv->num_coupling >= max)
  {
    TS_RAISE_ERRORF(TS_ERR_MAXIMUM_ALLOCATED, "maximum number of coupling terms have been added",0);
  }

  /* check format */
  if (!tsch_is_lateral_bond(terms[0])      ||
      !tsch_is_lateral_bond(terms[1])      ||
      !tsch_is_longitudinal_bond(terms[2]) ||
      !tsch_is_longitudinal_bond(terms[3]))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "coupling parameters must be lateral,lateral,longitudinal,longitudinal",0);
  }

  packed = tsch_pack_coupling(terms[0],terms[1],terms[2],terms[3]);
  drv->coupling_terms[drv->num_coupling] = packed;
  drv->coupling_energies[drv->num_coupling][0] = elo;
  drv->coupling_energies[drv->num_coupling][1] = ehi;

  {
    int rlatm,rlatp,rlongm,rlongp;
    rlatm  = tsch_reduce_lat_states[terms[0]];
    rlatp  = tsch_reduce_lat_states[terms[1]];
    rlongm = tsch_reduce_long_states[terms[2]];
    rlongp = tsch_reduce_long_states[terms[3]];

    assert(rlatm >= 0 && rlatp >= 0 && rlongm >= 0 && rlongp >= 0);
    drv->coupling_lookup[rlatm][rlatp][rlongm][rlongp] = drv->num_coupling;
    drv->coupling_lookup[rlatm][rlatp][rlongp][rlongm] = drv->num_coupling;
    drv->coupling_lookup[rlatp][rlatm][rlongm][rlongp] = drv->num_coupling;
    drv->coupling_lookup[rlatp][rlatm][rlongp][rlongm] = drv->num_coupling;
  }

  drv->num_coupling++;
  return TS_OK;
}

static inline int compare_terms(tsch_state_t a[4], int b[4])
{
  size_t j;
  for (j=0; j < 4; j++)
  {
    if ((int)a[j] != b[j])
    {
      return 0;
    }
  }

  return 1;
}

int ts_chem_driver_find_coupling_energy(ts_chem_driver *drv, int terms[4])
{
  int rlatm,rlatp,rlongm,rlongp;

  rlatm  = tsch_reduce_lat_states[terms[0]];
  rlatp  = tsch_reduce_lat_states[terms[1]];
  rlongm = tsch_reduce_long_states[terms[2]];
  rlongp = tsch_reduce_long_states[terms[3]];

  assert(rlatm >= 0 && rlatp >= 0 && rlongm >= 0 && rlongp >= 0);

  return drv->coupling_lookup[rlatm][rlatp][rlongm][rlongp];
}

static inline uint32_t 
pack_longitudinal_interface(int alpha_mer, int long_ifc, int beta_mer)
{
  assert(alpha_mer > 0 && alpha_mer < TS_CHEMDRIVER_STATE_MAX);
  assert(long_ifc  > 0 && long_ifc  < TS_CHEMDRIVER_STATE_MAX);
  assert(beta_mer  > 0 && beta_mer  < TS_CHEMDRIVER_STATE_MAX);
  assert(TS_CHEMDRIVER_STATE_MAX <= 16);

  return alpha_mer << 8 | long_ifc << 4 | beta_mer;
}

TSResult ts_chem_driver_add_dissociation_barrier(ts_chem_driver *drv, spin_t terms[3], double en)
{
  tsch_state_t *couple;
  size_t max;
  uint32_t packed;

  max = ARRAYLEN(drv->dissociation_barrier_terms);
  assert( max == ARRAYLEN(drv->dissociation_barrier_energies) );

  if ((size_t)drv->num_dissociation >= max)
  {
    TS_RAISE_ERRORF(TS_ERR_MAXIMUM_ALLOCATED,
        "maximum number of dissociation barriers (%zu) have been added", max);
  }

  /* check format */
  if (!tsch_is_monomer_state(terms[0].val)     ||
      !tsch_is_longitudinal_bond(terms[1].val) ||
      !tsch_is_monomer_state(terms[2].val))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "dissociation barrier must be monomer-longitudinal-monomer",0);
  }

  packed = pack_longitudinal_interface(terms[0].val, terms[1].val, terms[2].val);
  drv->dissociation_barrier_terms[drv->num_dissociation] = packed;
  drv->dissociation_barrier_energies[drv->num_dissociation] = en;
  drv->num_dissociation++;

  return TS_OK;
}

static int
find_dissociation_barrier(ts_chem_driver *drv, int alpha_mer, int long_ifc, int beta_mer)
{
  int i,n;
  uint32_t packed;

  packed = pack_longitudinal_interface(alpha_mer,long_ifc,beta_mer);
  for (i=0,n=drv->num_dissociation; i < n; i++)
  {
    if (drv->dissociation_barrier_terms[i] == packed)
    {
      return i;
    }
  }

  return -1;
}

int ts_chem_driver_find_dissociation_barrier(ts_chem_driver *drv, spin_t terms[3])
{
  return find_dissociation_barrier(drv, terms[0].val, terms[1].val, terms[2].val);
}

static inline double lookup_interface_energy(ts_chem_driver *drv, int sp0, int sp1, int sp2)
{
  return drv->interface_energies[sp0][sp1][sp2];
}

static inline double lookup_bent_energy(ts_chem_driver *drv, int latsp, int mer, int longsp)
{
  return lookup_interface_energy(drv, latsp, mer, longsp);
}

static void record_monomer_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops);
static void record_long_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops);

static double calc_monomer_site_energy(ts_chem_driver *drv, int r0, int r1, int c,
    struct tsch_populations *pops)
{
  int r;
  int nrow,ncol;
  double en_sum;
  bool lat_long_bending = tsch_test_flags(drv, TSCH_ENABLE_LAT_LONG_BENDING);

  int rm1,cm1;
  int rp1,cp1;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  if (r0 < 1) { r0 = 1; }
  if (r1 > nrow-2) { r1 = nrow-2; }

  assert(ts_chem_driver_site_type(drv,r0,c) == TS_CHEMDRIVER_SITE_MONOMER);

  en_sum = 0.0;
  if (c > 0 && c < ncol-1)
  {
    for (r = r0; r <= r1; r += 2)
    {
      int latm1, latp1;
      int state,ifc,state1;
      double en_mer;

      state = TS_LATTICE_SPIN(drv,r,c);
      assert(state >= 0 && state < TS_CHEMDRIVER_STATE_MAX);

      en_mer = drv->monomer_energies[state];
      if (EXPLICIT_NAN_CHECK && !isfinite(en_mer))
      {
        fprintf(stderr, "\nmonomer state energy %c is not finite\n",
            ts_chem_driver_site_codes[state]);
      }
      assert(isfinite(en_mer));

      en_sum += drv->monomer_energies[state];

      if (r < nrow-2)
      {
        double en_ifc;

        ifc    = TS_LATTICE_SPIN(drv,r+1,c);
        state1 = TS_LATTICE_SPIN(drv,r+2,c);
        assert(ifc >= 0 && ifc < TS_CHEMDRIVER_STATE_MAX);
        assert(state1 >= 0 && state1 < TS_CHEMDRIVER_STATE_MAX);

        en_ifc = lookup_interface_energy(drv,state,ifc,state1); 
        if (EXPLICIT_NAN_CHECK && !isfinite(en_ifc))
        {
          fprintf(stderr, "\nlongitudinal interface energy %c-%c-%c is not finite\n",
              ts_chem_driver_site_codes[state],
              ts_chem_driver_site_codes[ifc],
              ts_chem_driver_site_codes[state1]);
        }
        assert(isfinite(en_ifc));

        en_sum += en_ifc;
      }

      {
        double en_lat;

        latm1 = TS_LATTICE_SPIN(drv, r,c-1);
        latp1 = TS_LATTICE_SPIN(drv, r,c+1);

        en_lat = lookup_interface_energy(drv,latm1, state, latp1);
        if (EXPLICIT_NAN_CHECK && !isfinite(en_lat))
        {
          fprintf(stderr, "\nbending energy %c-%c-%c (%3d-%3d-%3d) is not finite\n",
              ts_chem_driver_site_codes[latm1],
              ts_chem_driver_site_codes[state],
              ts_chem_driver_site_codes[latp1],
              latm1,state,latp1);
        }
        assert(isfinite(en_lat));

        en_sum += en_lat;
      }
    }
  }
  else
  {
    rm1 = r0; cm1 = c-1; ts_lattice_wrap_coord_column(&drv->latt, &rm1,&cm1);
    rp1 = r0; cp1 = c+1; ts_lattice_wrap_coord_column(&drv->latt, &rp1,&cp1);

    for (r = r0; r <= r1; r += 2, rm1 += 2, rp1 += 2)
    {
      int latm1, latp1;
      int state,ifc,state1;
      double en_mer;

      state = TS_LATTICE_SPIN(drv,r,c);
      assert(state >= 0 && state < TS_CHEMDRIVER_STATE_MAX);
      en_mer = drv->monomer_energies[state];

      if (EXPLICIT_NAN_CHECK && !isfinite(en_mer))
      {
        fprintf(stderr, "\nmonomer state energy %c is not finite\n",
            ts_chem_driver_site_codes[state]);
      }
      assert(isfinite(en_mer));

      en_sum += en_mer;

      if (r < nrow-2)
      {
        double en_ifc;

        ifc    = TS_LATTICE_SPIN(drv,r+1,c);
        state1 = TS_LATTICE_SPIN(drv,r+2,c);
        assert(ifc >= 0 && ifc < TS_CHEMDRIVER_STATE_MAX);
        assert(state1 >= 0 && state1 < TS_CHEMDRIVER_STATE_MAX);

        en_ifc = lookup_interface_energy(drv,state,ifc,state1); 
        if (EXPLICIT_NAN_CHECK && !isfinite(en_ifc))
        {
          fprintf(stderr, "\nlongitudinal interface energy %c-%c-%c is not finite\n",
              ts_chem_driver_site_codes[state],
              ts_chem_driver_site_codes[ifc],
              ts_chem_driver_site_codes[state1]);
        }
        assert(isfinite(en_ifc));

        en_sum += en_ifc;
      }

      if ((rm1 < 1) || (rp1 < 1) || (rm1+1 >= nrow) || (rp1+1 >= nrow))
      {
        continue;
      }

      {
        double en_lat;

        latm1 = TS_LATTICE_SPIN(drv, rm1,cm1);
        latp1 = TS_LATTICE_SPIN(drv, rp1,cp1);

        en_lat = lookup_interface_energy(drv,latm1, state, latp1);
        if (EXPLICIT_NAN_CHECK && !isfinite(en_lat))
        {
          fprintf(stderr, "\nbending energy %c-%c-%c (%3d-%3d-%3d) is not finite\n",
              ts_chem_driver_site_codes[latm1],
              ts_chem_driver_site_codes[state],
              ts_chem_driver_site_codes[latp1],
              latm1,state,latp1);
        }
        assert(isfinite(en_lat));

        en_sum += en_lat;
      }
    }
  }

  if (lat_long_bending)
  {
    rm1 = r0; cm1 = c-1; ts_lattice_wrap_coord_column(&drv->latt, &rm1,&cm1);
    rp1 = r0; cp1 = c+1; ts_lattice_wrap_coord_column(&drv->latt, &rp1,&cp1);

    for (r = r0; r <= r1; r += 2, rm1 += 2, rp1 += 2)
    {
      int latm1, latp1;
      int lvalp,lvalm;
      int state;

      if ((rm1 < 1) || (rp1 < 1) || (rm1+1 >= nrow) || (rp1+1 >= nrow))
      {
        continue;
      }

      lvalm = TS_LATTICE_SPIN(drv, r-1,c);
      state = TS_LATTICE_SPIN(drv, r+0,c);
      lvalp = TS_LATTICE_SPIN(drv, r+1,c);

      latm1 = TS_LATTICE_SPIN(drv, rm1,cm1);
      latp1 = TS_LATTICE_SPIN(drv, rp1,cp1);

      en_sum += lookup_bent_energy(drv, latm1, state, lvalp);
      en_sum += lookup_bent_energy(drv, latm1, state, lvalm);
      en_sum += lookup_bent_energy(drv, latp1, state, lvalp);
      en_sum += lookup_bent_energy(drv, latp1, state, lvalm);
    }
  }

  if (pops)
  {
    for (r=r0; r <= r1; r+=2)
    {
      record_monomer_site_pops(drv, r, c, pops);
    }
    for (r=r0+1; r <= r1; r+=2)
    {
      record_long_site_pops(drv, r, c, pops);
    }
  }

  return en_sum;
}

static void record_monomer_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops)
{
  int rm1,cm1;
  int rp1,cp1;
  int nrow,ncol;
  int latm1, latp1;
  int lvalp,lvalm;
  int state;

  state = TS_LATTICE_SPIN(drv,r,c);
  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  pops->monomer_pop[state]++;

  rm1 = r; cm1 = c-1;
  rp1 = r; cp1 = c+1;

  if (c == 0 || c == ncol-1)
  {
    ts_lattice_wrap_coord(&drv->latt, &rm1,&cm1);
    ts_lattice_wrap_coord(&drv->latt, &rp1,&cp1);
  }

  if ((rm1 < 1) || (rp1 < 1) || (rm1+1 >= nrow) || (rp1+1 >= nrow))
  {
    return;
  }

  lvalm = r > 0      ? TS_LATTICE_SPIN(drv, r-1,c) : 0;
  lvalp = r < nrow-1 ? TS_LATTICE_SPIN(drv, r+1,c) : 0;

  latm1 = TS_LATTICE_SPIN(drv, rm1,cm1);
  latp1 = TS_LATTICE_SPIN(drv, rp1,cp1);

  record_bending(pops, latm1, state, latp1);

  if (r+1 < nrow)
  {
    record_bending(pops, latm1, state, lvalp);
    record_bending(pops, latp1, state, lvalp);
  }

  if (r > 0)
  {
    record_bending(pops, latm1, state, lvalm);
    record_bending(pops, latp1, state, lvalm);
  }
}


static double calc_longitudinal_site_energy(ts_chem_driver *drv, int r0, int r1, int c,
    struct tsch_populations *pops)
{
  int r;
  double en_sum;

  en_sum = 0.0;
  for (r=r0; r <= r1; r+=2)
  {
    int lo,st,hi;

    lo = TS_LATTICE_SPIN(drv, r-1,c);
    st = TS_LATTICE_SPIN(drv, r+0,c);
    hi = TS_LATTICE_SPIN(drv, r+1,c);

    en_sum += lookup_interface_energy(drv,lo,st,hi); 
  }

  if (pops)
  {
    for (r=r0; r <= r1; r+=2)
    {
      record_long_site_pops(drv, r, c, pops);
    }
  }

  return en_sum;
}

static void record_long_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops)
{
  int lo,st,hi;

  lo = TS_LATTICE_SPIN(drv, r-1,c);
  st = TS_LATTICE_SPIN(drv, r+0,c);
  hi = TS_LATTICE_SPIN(drv, r+1,c);

  if ((lo != TS_CHEMDRIVER_STATE_EMPTY) &&
      (st != TS_CHEMDRIVER_STATE_EMPTY) &&
      (hi != TS_CHEMDRIVER_STATE_EMPTY))
  {
    pops->interface_pop[lo][st][hi]++;
  }
  else
  {
    pops->interface_pop[TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY]++;
  }
}

static void record_lateral_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops);
static double calc_lateral_site_energy(ts_chem_driver *drv, int r0, int r1, int c,
    struct tsch_populations *pops)
{
  int r, rm1,cm1, rp1,cp1, nrow, ncol;
  double en_sum;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  rm1 = r0; cm1 = c-1; ts_lattice_wrap_coord_column(&drv->latt, &rm1,&cm1);
  rp1 = r0; cp1 = c+1; ts_lattice_wrap_coord_column(&drv->latt, &rp1,&cp1);

  en_sum = 0.0;
  for (r=r0; r <= r1; r += 2, rm1 += 2, rp1 += 2)
  {
    int ifc, mlo,mhi;
    double en_ifc;

    if (rm1 < 1 || rp1 < 1)
    {
      continue;
    }

    if (rm1 > nrow-2 || rp1 > nrow-2)
    {
      break;
    }

    mlo = TS_LATTICE_SPIN(&drv->latt,rm1,cm1);
    ifc = TS_LATTICE_SPIN(&drv->latt,r  ,c  );
    mhi = TS_LATTICE_SPIN(&drv->latt,rp1,cp1);

    assert(mlo >= 0 && mlo < TS_CHEMDRIVER_STATE_MAX);
    assert(ifc >= 0 && ifc < TS_CHEMDRIVER_STATE_MAX);
    assert(mhi >= 0 && mhi < TS_CHEMDRIVER_STATE_MAX);

    en_ifc = drv->interface_energies[mlo][ifc][mhi];

    if (EXPLICIT_NAN_CHECK && !isfinite(en_ifc))
    {
      fprintf(stderr, "\nlateral interface energy %c-%c-%c is not finite\n",
          ts_chem_driver_site_codes[mlo],
          ts_chem_driver_site_codes[ifc],
          ts_chem_driver_site_codes[mhi]);
    }

    assert(isfinite(en_ifc));
    en_sum += en_ifc;
  }

  if (pops)
  {
    for (r=r0; r <= r1; r += 2)
    {
      record_lateral_site_pops(drv, r,c, pops);
    }
  }

  return en_sum;
}

static void record_lateral_site_pops(ts_chem_driver *drv, int r, int c, struct tsch_populations *pops)
{
  int state, rm1,cm1, rp1,cp1, rmin,rmax;
  int mlo,mhi;

  state = TS_LATTICE_SPIN(drv,r,c);

  rm1=r; cm1=c-1; ts_lattice_wrap_coord((ts_lattice*)drv, &rm1,&cm1);
  rp1=r; cp1=c+1; ts_lattice_wrap_coord((ts_lattice*)drv, &rp1,&cp1);

  if (rm1 <= rp1)
  {
    rmin = rm1; rmax = rp1;
  }
  else           
  { 
    rmin = rp1; rmax = rm1;
  }

  if ((rmin < 1) || (rmax+1 >= (int)drv->latt.nrow)) {
    return;
  }

  mlo = TS_LATTICE_SPIN(&drv->latt,rm1,cm1);
  mhi = TS_LATTICE_SPIN(&drv->latt,rp1,cp1);

  if ((mlo   != TS_CHEMDRIVER_STATE_EMPTY) &&
      (state != TS_CHEMDRIVER_STATE_EMPTY) &&
      (mhi   != TS_CHEMDRIVER_STATE_EMPTY))
  {
    pops->interface_pop[mlo][state][mhi]++;
  } else {
    pops->interface_pop[TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY]++;
  }
}

static inline int rlatind(int st)
{
  return st - TS_CHEMDRIVER_STATE_TUBELIKE;
}

static inline int rlongind(int st)
{
  assert(
      st == TS_CHEMDRIVER_STATE_CURVED   ||
      st == TS_CHEMDRIVER_STATE_BENT_IFC ||
      st == TS_CHEMDRIVER_STATE_STRAIGHT);

  return st>>1;
}

static inline int
min_edge_dist(const ts_chem_driver *drv, int r, int c, int *whp)
{
  int plus,minus, dplus, dminus, dedge;

  minus = drv->row_range[2*c+0];
  plus  = drv->row_range[2*c+1];

  dminus = r - minus;
  dplus = plus - r;

  dedge = dplus <= dminus ? dplus : dminus;
  dedge /= 2;

  if (whp != NULL)
  {
    if (dplus < dminus)
    {
      *whp = TIP_PLUS;
    }
    else if (dplus > dminus)
    {
      *whp = TIP_MINUS;
    }
    else
    {
      *whp = TIP_BOTH;
    }
  }

  return dedge;
}

static inline double
calc_coupling_site_energies_inner(ts_chem_driver *drv, int r, int c, int latm, int latp, int longm, int longp)
{
  int dedge, tind;

  latm  = rlatind(latm);
  latp  = rlatind(latp);
  longm = rlongind(longm);
  longp = rlongind(longp);

  tind = drv->coupling_lookup[latm][latp][longm][longp];
  dedge = min_edge_dist(drv, r,c, NULL);

  return (dedge <= drv->coupling_edge_dist)
    ? drv->coupling_energies[tind][0]
    : drv->coupling_energies[tind][1];
}

static double calc_coupling_site_energies(ts_chem_driver *drv, int r0, int r1, int c)
{
  int i,n, nrow,ncol, plus,minus;
  double en_sum = 0.0;
  spin_t *restrict left, *restrict center, *restrict right;
  int check;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  assert(r0 >= 0 && r1 >= 0 && (size_t)r0 < drv->latt.nrow && (size_t)r1 < drv->latt.nrow);

  if (r0 > r1)
  {
    return 0.0;
  }

  if (r0 == 1) { r0++; }
  if (r1 >= nrow-2) { r1 = nrow-3; }

  if (c > 0 && c < ncol-1)
  {
    left   = ts_lattice_ref(&drv->latt, r0  ,c-1);
    center = ts_lattice_ref(&drv->latt, r0-1,c  );
    right  = ts_lattice_ref(&drv->latt, r0  ,c+1);
  }
  else
  {
    int rm1,cm1, rp1,cp1;

    rm1 = r0; cm1 = c-1; ts_lattice_wrap_coord_column(&drv->latt, &rm1,&cm1);
    rp1 = r0; cp1 = c+1; ts_lattice_wrap_coord_column(&drv->latt, &rp1,&cp1);
    if (rm1 <= 0 || rp1 <= 0)
    {
      int rmin,off;
      rmin = rm1 < rp1 ? rm1 : rp1;
      off = 2-rmin;
      r0 += off;
      rm1 += off;
      rp1 += off;
    }

    if (r0 > r1)
    {
      return 0.0;
    }

    assert(ts_chem_driver_site_type(drv,rm1 ,cm1) == TS_CHEMDRIVER_SITE_LONG_IFC);
    assert(ts_chem_driver_site_type(drv,rp1 ,cp1) == TS_CHEMDRIVER_SITE_LONG_IFC);
    assert(ts_chem_driver_site_type(drv,r0  ,c  ) == TS_CHEMDRIVER_SITE_BLANK);
    assert(ts_chem_driver_site_type(drv,r0+1,c  ) == TS_CHEMDRIVER_SITE_LAT_IFC);
    assert(ts_chem_driver_site_type(drv,r0-1,c  ) == TS_CHEMDRIVER_SITE_LAT_IFC);

    assert(rm1 >= 0 && rm1 < nrow && cm1 >= 0 && cm1 < ncol);
    assert(rp1 >= 0 && rp1 < nrow && cp1 >= 0 && cp1 < ncol);

    left   = ts_lattice_ref(&drv->latt, rm1 ,cm1);
    center = ts_lattice_ref(&drv->latt, r0-1,c  );
    right  = ts_lattice_ref(&drv->latt, rp1 ,cp1);
  }

  for (i=0, n=r1-r0; i <= n; i+=2)
  {
    int latm,latp,longm,longp;

    if (r0 > r1)
    {
      fprintf(stderr, "r0 > r1: i=%3d, n=%3d, en_sum=%f\n",i,n,en_sum);
    }

    latm = center[i+0].val;
    latp = center[i+2].val;

    if (latm == TS_CHEMDRIVER_STATE_EMPTY ||
        latp == TS_CHEMDRIVER_STATE_EMPTY)
    {
      continue;
    }

    longm = left[i].val;
    longp = right[i].val;

    if (longm == TS_CHEMDRIVER_STATE_EMPTY ||
        longp == TS_CHEMDRIVER_STATE_EMPTY)
    {
      continue;
    }

    en_sum += calc_coupling_site_energies_inner(drv, r0+i, c, latm,latp, longm,longp);
  }

  return en_sum;
}

static void dump_lattice_config(ts_chem_driver *drv, int r, int c)
{
  int r0,r1,row, col, ncol;

  UNUSED_PARAM(c);

  ncol = drv->latt.ncol;
  r0 = r-12;
  if (r0 < 0) { r0 = 0; }

  r1 = r+12;
  if ((size_t)r1 >= drv->latt.nrow) { r1 = drv->latt.nrow-1; }

  for (row=r0; row <= r1; row++)
  {
    fprintf(stderr, "%5d | ", row);
    for (col=0; col < ncol; col++)
    {
      int ch, val;
      val = TS_LATTICE_SPIN(drv,row,col);
      ch = '*';
      if (val <= TS_CHEMDRIVER_STATE_MAX)
      {
        ch = ts_chem_driver_site_codes[val];
      }

      fprintf(stderr, "%c ", ch);
    }
    fprintf(stderr, "| %5d\n", row);
  }

  fprintf(stderr, "Cols  | ");
  for(col=0; col < ncol; col++)
  {
    fprintf(stderr, "%1d ", col%10);
  }
  fprintf(stderr, "\n\n");
}

static double calc_site_energy_inner(ts_chem_driver* drv, int r0, int r1, int c0, int c1,
    struct tsch_populations *pops)
{
  int c;
  double en;

  en = 0.0;
  for (c=c0; c < c1; c+=2)
  {
    en += calc_monomer_site_energy(drv, r0,r1,c, pops);
    assert(isfinite(en));

    en += calc_lateral_site_energy(drv, r0,r1, c+1, pops);
    assert(isfinite(en));

    en += calc_coupling_site_energies(drv, r0+1,r1, c+1);
    assert(isfinite(en));
  }
  if (c == c1)
  {
    en += calc_monomer_site_energy(drv, r0,r1,c, pops);
  }

  return en;
}

static double calc_site_energy(ts_chem_driver* drv, int r0, int r1, int c0, int c1,
    struct tsch_populations *pops)
{
  int r,c, ncol;
  double en;

  ncol = drv->latt.ncol;

  assert(r0 >= 1 && r0 <= r1 && (size_t)r1 <= drv->latt.nrow-2);
  assert(c0 >= 0 && c0 <= c1 && (size_t)c1 < drv->latt.ncol);

  en = 0.0;
  switch (ts_chem_driver_site_type(drv,r0,c0))
  {
    case TS_CHEMDRIVER_SITE_MONOMER:
      en = calc_site_energy_inner(drv, r0,r1,c0,c1, pops);
      break;

    case TS_CHEMDRIVER_SITE_LONG_IFC:
      en = calc_site_energy_inner(drv, r0-1,r1,c0,c1, pops);
      break;

    case TS_CHEMDRIVER_SITE_LAT_IFC:
      en = calc_site_energy_inner(drv, r0,r1,c0-1,c1, pops);
      break;

    case TS_CHEMDRIVER_SITE_BLANK:
      en = calc_site_energy_inner(drv, r0-1,r1,c0-1,c1, pops);
      break;

    default:
      en = 0.0;
      break;
  }

  return en;
}

static inline double calc_patch_energy(ts_chem_driver *drv, int rmin, int rmax, int c0)
{
  int rr,cc, cmin,cmax, nrow,ncol;
  double en = 0.0;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  if (rmin < 1) { rmin = 1; }
  if (rmax > nrow-2) { rmax = nrow-2; }

  assert(rmin <= rmax);

  cmin = c0-TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
  cmax = c0+TS_CHEMDRIVER_MAX_ENERGY_COUPLING;

  if (cmin >= 0 && cmax < ncol)
  {
    return calc_site_energy(drv, rmin,rmax, cmin,cmax, NULL);
  }

  if (cmin < 0)
  {
    int rr0,rr1, cc0,cc1;

    en = calc_site_energy(drv, rmin,rmax, 0,cmax, NULL);

    rr0 = rmin; cc0 = cmin; ts_lattice_wrap_coord(&drv->latt, &rr0,&cc0);
    rr1 = rmax; cc1 = -1; ts_lattice_wrap_coord(&drv->latt, &rr1,&cc1);

    if ((rr1 > 0) && (rr0 < nrow-1))
    {
      if (rr0 < 1) { rr0 = 1; }
      if (rr1 > nrow-2) { rr1 = nrow-2; }

      en += calc_site_energy(drv, rr0,rr1, cc0, cc1, NULL);
    }
  }

  if (cmax >= ncol)
  {
    int rr0,rr1, cc0,cc1;

    en = calc_site_energy(drv, rmin,rmax, cmin,ncol-1, NULL);

    rr0 = rmin; cc0 = ncol; ts_lattice_wrap_coord(&drv->latt, &rr0,&cc0);
    rr1 = rmax; cc1 = cmax; ts_lattice_wrap_coord(&drv->latt, &rr1,&cc1);

    if ((rr1 > 0) && (rr0 < nrow-1))
    {
      if (rr0 < 1) { rr0 = 1; }
      if (rr1 > nrow-2) { rr1 = nrow-2; }

      en += calc_site_energy(drv, rr0,rr1, cc0, cc1, NULL);
    }
  }

  return en;
}


static inline double calc_site_dep_energy(ts_chem_driver* drv, int r0, int c0)
{
  int rmin, rmax;

  rmin = r0-TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
  rmax = r0+TS_CHEMDRIVER_MAX_ENERGY_COUPLING;

  return calc_patch_energy(drv, rmin,rmax,c0);
}

static MAYBE_UNUSED double calc_site_dep_energy_lateral_sites(ts_chem_driver* drv, int ra, int ca)
{
  int rb, rmin,rmax;

  rb = ra + 2;

  rmin = ra-TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
  rmax = rb+TS_CHEMDRIVER_MAX_ENERGY_COUPLING;

  return calc_patch_energy(drv, rmin,rmax, ca);
}

static TSResult update_row_range(ts_chem_driver *drv, int c);

static void calc_delta_energy_lateral_sites(ts_chem_driver* drv, int ra, int ca,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict energies)
{
  int old_state, dedge, wh_edge;
  double en_old;
  size_t i, n;
  int cdist, changes_row_range, rmin,rmax, tmp;

  old_state = TS_LATTICE_SPIN(drv, ra, ca);
  assert(TS_LATTICE_SPIN(drv, ra+2,ca) == old_state);

  cdist     = drv->coupling_edge_dist;

  wh_edge = TIP_NONE;
  dedge = min_edge_dist(drv, ra,ca, &wh_edge);

  changes_row_range = dedge <= cdist;

  rmin = ra - TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
  rmax = (ra+2) + TS_CHEMDRIVER_MAX_ENERGY_COUPLING;

  if (changes_row_range)
  {
    int minus_row, plus_row;

    minus_row = drv->row_range[2*ca+0];
    plus_row  = drv->row_range[2*ca+1];
    switch (wh_edge)
    {
      case TIP_NONE:
      default:
        fprintf(stderr, "[%s:%d] problem in internal logic!\n", __FILE__, __LINE__);
        abort();

      case TIP_BOTH:
        /* equidistant... choose plus end
         */

        /* fall through */
      case TIP_PLUS:
        tmp = plus_row - 4 - TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
        if (tmp < rmin) { rmin = tmp; }
        break;

      case TIP_MINUS:
        tmp = minus_row + 4 + TS_CHEMDRIVER_MAX_ENERGY_COUPLING;
        if (tmp > rmax) { rmax = tmp; }
        break;
    }
  }

  assert(rmin <= rmax);

  if (rmin < 0) { rmin = 0; }
  if (rmax < 0) { rmax = 0; }

  if ((size_t)rmin >= drv->latt.nrow) { rmax = drv->latt.nrow-1; }
  if ((size_t)rmax >= drv->latt.nrow) { rmax = drv->latt.nrow-1; }

  en_old = calc_patch_energy(drv, rmin,rmax, ca);

  for (i = 0, n = states_end - states_begin; i < n; i++)
  {
    int new_state;
    double en_new = 0.0;

    new_state  = states_begin[i];
    if (new_state == old_state)
    {
      energies[i] = 0.0;
      continue;
    }

    TS_LATTICE_SPIN(drv, ra+0, ca) = new_state;
    TS_LATTICE_SPIN(drv, ra+2, ca) = new_state;

    if (changes_row_range)
    {
      update_row_range(drv, ca);
    }
    en_new = calc_patch_energy(drv, rmin,rmax, ca);

    energies[i] = en_new - en_old;
  }

  TS_LATTICE_SPIN(drv, ra+0, ca) = old_state;
  TS_LATTICE_SPIN(drv, ra+2, ca) = old_state;

  if (changes_row_range)
  {
    update_row_range(drv, ca);
  }
}

static void calc_delta_energy(ts_chem_driver* drv, size_t r, size_t c,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict energies)
{
  double en_old, en_new;
  spin_t old_state;
  const enum TS_CHEMDRIVER_SITE_STATE *st;

  old_state = ts_lattice_at(&drv->latt, r, c);
  en_old = 0.0;
  en_old = calc_site_dep_energy(drv, r,c);

  assert(isfinite(en_old));

  for (st=states_begin; st != states_end; st++, energies++)
  {
    double en_new = 0.0;
    if ((int)*st == old_state.val)
    {
      *energies = 0.0;
      continue;
    }

    ts_lattice_ref(&drv->latt, r, c)[0].val = *st;
    en_new = calc_site_dep_energy(drv, r,c);
    *energies = en_new - en_old;

    assert(isfinite(en_new));
    assert(isfinite(*energies));

#if DEBUG_ENERGY
    {
      double delta = en_new - en_old;

      ts_lattice_ref(&drv->latt, r, c)[0] = old_state;
      RETURN_ON_ERROR( ts_chem_driver_energy(drv, &en_old) );
      ts_lattice_ref(&drv->latt, r, c)[0].val = *st;
      RETURN_ON_ERROR( ts_chem_driver_energy(drv, &en_new) );

      double dE_check = en_new - en_old;
      if ( fabs(delta - dE_check) > 1e-8 )
      {
        fprintf(stderr, "\nchange in site-dependent energy != change in whole lattice energy\n");
        fprintf(stderr, "%7.4e != %7.4e (delta is %7.4e)\n\n",
            *dE_p, dE_check, fabs(*dE_p - dE_check));
      }
    }
#endif
  }

  ts_lattice_ref(&drv->latt, r, c)[0] = old_state;
}


static inline double prob_acceptance(const ts_chem_driver *drv, enum TS_CHEMDRIVER_BARRIER_MODEL model, double dE)
{
  switch (model)
  {
    case TS_CHEMDRIVER_BARRIER_NONE:
      return 1.0;

    case TS_CHEMDRIVER_BARRIER_INFINITE:
      return 0.0;

    case TS_CHEMDRIVER_BARRIER_PROPORTIONAL:
    case TS_CHEMDRIVER_BARRIER_METROPOLIS:
      return ts_chem_driver_prob_acceptance_metropolis(drv,dE);

    case TS_CHEMDRIVER_BARRIER_KAWASAKI:
      return ts_chem_driver_prob_acceptance_kawasaki(drv,dE);

    default:
      return 0.0;
  }
}

double ts_chem_driver_prob_acceptance(const ts_chem_driver *drv, enum TS_CHEMDRIVER_BARRIER_MODEL model, double dE)
{
  return prob_acceptance(drv,model,dE);
}

static double calc_rate(ts_chem_driver *drv, double k_pre, double dE)
{
  double factor = 1.0;

  switch (drv->barrier_model)
  {
    case TS_CHEMDRIVER_BARRIER_NONE:
      return nan("1");

    case TS_CHEMDRIVER_BARRIER_METROPOLIS:
    case TS_CHEMDRIVER_BARRIER_KAWASAKI:
      factor = prob_acceptance(drv, drv->barrier_model, dE);
      break;

    case TS_CHEMDRIVER_BARRIER_PROPORTIONAL:
      dE *= drv->latt.inv_temp;
      factor = dE <= 0
        ? exp(drv->barrier_param*dE)
        : exp(-(drv->barrier_param + 1)*dE)
        ;
      break;

    default:
      return nan("0");
  }

  return k_pre * factor;
}


static void calc_rate_single_spin(ts_chem_driver* drv, size_t r, size_t c, 
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict rates,
    double k_pre)
{
  const enum TS_CHEMDRIVER_SITE_STATE *restrict st;
  const int orig = TS_LATTICE_SPIN(&drv->latt, r,c);

  calc_delta_energy(drv, r,c, states_begin, states_end, rates);

  for (st=states_begin; st != states_end; st++, rates++)
  {
    *rates = ((int)*st != orig) ? calc_rate(drv, k_pre, *rates) : 0.0;
  }
}

static void calc_rate_lateral_spin(ts_chem_driver* drv, int ra,int ca,
    const enum TS_CHEMDRIVER_SITE_STATE *states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE *restrict states_end,
    double *restrict rates,
    double k_pre)
{
  const enum TS_CHEMDRIVER_SITE_STATE *restrict st;
  const int orig = TS_LATTICE_SPIN(&drv->latt, ra,ca);

  assert(orig == TS_LATTICE_SPIN(&drv->latt,ra+2,ca));

  calc_delta_energy_lateral_sites(drv, ra,ca, states_begin, states_end, rates);

  for (st=states_begin; st != states_end; st++, rates++)
  {
    double dE = *rates;
    *rates = ((int)*st != orig) ? calc_rate(drv, k_pre, dE) : 0.0;
  }

  assert(orig == TS_LATTICE_SPIN(&drv->latt,ra+0,ca));
  assert(orig == TS_LATTICE_SPIN(&drv->latt,ra+2,ca));
}

static inline int is_spin_gtp(spin_t state)
{
  return (state.val >= TS_CHEMDRIVER_STATE_GTP_MIN) && (state.val <= TS_CHEMDRIVER_STATE_GTP_MAX);
}

static inline int is_gtp(const ts_chem_driver* drv, size_t r, size_t c)
{
  return is_spin_gtp(ts_lattice_at((ts_lattice*)drv, r,c));
}

static TSResult site_move_single_flips(ts_chem_driver* drv,
    size_t r, size_t c, double prefactor,
    const enum TS_CHEMDRIVER_SITE_STATE* states_begin,
    const enum TS_CHEMDRIVER_SITE_STATE* states_end,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  int i,n, curr;
  size_t ind,ind0;
  ptrdiff_t nposs;
  const enum TS_CHEMDRIVER_SITE_STATE* st;

  if (states_end <= states_begin)
  {
    return TS_OK;
  }

  curr = TS_LATTICE_SPIN(drv, r,c);
  if (curr == TS_CHEMDRIVER_STATE_EMPTY)
  {
    return TS_OK;
  }

  ind = nmovesp[0];

  nposs = states_end - states_begin;
  if (ind + nposs >= TS_CHEMDRIVER_STATE_MAX) 
  {
    TS_RAISE_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
        "possible states exceed list maximum");
  }

  calc_rate_single_spin(drv, r,c, 
      states_begin, states_end, &rates[ind], prefactor);

  for (ind0=ind, i=0, n=states_end-states_begin; i < n; i++)
  {
    if (curr == (int)states_begin[i] || rates[ind0+i] == 0.0)
    {
      continue;
    }

    assert(isfinite(rates[ind0+i]));

    if (ind != ind0+i)
    {
      rates[ind] = rates[ind0+i];
    }
    states[ind] = states_begin[i];
    ind++;
  }
  nmovesp[0] = ind;

  return TS_OK;
}

static TSResult
update_row_range(ts_chem_driver *drv, int c)
{
  if (c >= 0)
  {
    ts_lattice_find_column_row_range(&drv->latt, &drv->row_range[2*c], c);
  }
  else
  {
    RETURN_ON_ERROR( ts_lattice_find_row_range(&drv->latt, drv->row_range) );
  }
  return TS_OK;
}

TSResult ts_chem_driver_update_row_range(ts_chem_driver *drv)
{
  RETURN_WITH_RESULT(update_row_range(drv,-1));
}

static void rebuild_moves_hook(ts_lattice *latt, void *ud)
{
  TSResult err;
  ts_chem_driver *drv = ud;

  UNUSED_PARAM(latt);

  if (err = update_row_range(drv, -1), !IS_OKAY(err))
  {
    abort();
  }
}

static inline TSResult
monomer_reactions_gtp(ts_chem_driver* drv,
    size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_MONOMER];
  if (!is_gtp(drv,r,c)) { return TS_OK; }

  RETURN_ON_ERROR( site_move_single_flips(drv, r,c,
        prefactor,
        monomer_states_gtp + NOT_EMPTY(monomer_states_gtp),
        monomer_states_gtp + NUM_CONST(monomer_states_gtp),
        states, rates, nmovesp) );

  return TS_OK;
}

static inline TSResult
monomer_reactions_gdp(ts_chem_driver* drv,
    size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_MONOMER];
  if (is_gtp(drv,r,c)) { return TS_OK; }

  RETURN_ON_ERROR( site_move_single_flips(drv, r,c,
        prefactor,
        monomer_states_gdp + NOT_EMPTY(monomer_states_gdp),
        monomer_states_gdp + NUM_CONST(monomer_states_gdp),
        states, rates, nmovesp) );

  return TS_OK;
}

static TSResult
monomer_reaction_hydrolysis(ts_chem_driver* drv,
    size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  int st0,if0,st1,stm1,ifm1,nrow;
  double prefactor,e_forbid;
  size_t ind;

  if (ts_chem_driver_is_alpha_monomer(drv,r,c))
  {
    return TS_OK;
  }

  prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_HYDROLYSIS];
  if (prefactor <= 0.0)
  {
    return TS_OK;
  }

  if (r+2 >= drv->latt.nrow-1)
  {
    return TS_OK;
  }

  st0  = TS_LATTICE_SPIN(drv, r+0,c);
  if0  = TS_LATTICE_SPIN(drv, r+1,c);
  st1  = TS_LATTICE_SPIN(drv, r+2,c);

  if ((st0 != TS_CHEMDRIVER_STATE_STRAIGHT) || (if0 != TS_CHEMDRIVER_STATE_STRAIGHT))
  {
    return TS_OK;
  }

  if (tsch_test_flags(drv, TSCH_HYDROLYSIS_ONLY_AT_SSS) && (st1 != TS_CHEMDRIVER_STATE_STRAIGHT))
  {
    return TS_OK;
  }

  e_forbid = drv->forbidden_energy;
  if (drv->interface_energies[TS_CHEMDRIVER_STATE_STRAIGHT_GDP][if0][st1] >= e_forbid)
  {
    return TS_OK;
  }

  if (r-2 >= 1)
  {
    stm1 = TS_LATTICE_SPIN(drv, r-2,c);
    ifm1 = TS_LATTICE_SPIN(drv, r-1,c);
    if (drv->interface_energies[stm1][ifm1][TS_CHEMDRIVER_STATE_STRAIGHT_GDP] >= e_forbid)
    {
      return TS_OK;
    }
  }

  ind = nmovesp[0];
  if (ind + 1 >= TS_CHEMDRIVER_STATE_MAX) 
  {
    TS_RAISE_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
        "state list is too small to add hydrolysis move");
  }

  states[ind] = TS_CHEMDRIVER_STATE_STRAIGHT_GDP;
  rates[ind] = prefactor;
  nmovesp[0] = ind+1;

  return TS_OK;
}

static inline TSResult
longitudinal_reactions(ts_chem_driver* drv,
    size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_LONGITUDINAL];

  RETURN_ON_ERROR( site_move_single_flips(drv, r,c,
        prefactor,
        longitudinal_states + NOT_EMPTY(longitudinal_states),
        longitudinal_states + NUM_CONST(longitudinal_states),
        states, rates, nmovesp) );

  return TS_OK;
}

static const struct {
  int pspin;
  int inter;
  int mer;
} on_reaction_table[] = {
  { PSEUDO_K_ON_BENT_LEFT     , TS_CHEMDRIVER_STATE_BENT_IFC, TS_CHEMDRIVER_STATE_BENT_LEFT  },
  { PSEUDO_K_ON_BENT_RIGHT    , TS_CHEMDRIVER_STATE_BENT_IFC, TS_CHEMDRIVER_STATE_BENT_RIGHT },

  { PSEUDO_K_ON_STRAIGHT_LEFT     , TS_CHEMDRIVER_STATE_STRAIGHT, TS_CHEMDRIVER_STATE_BENT_LEFT  },
  { PSEUDO_K_ON_STRAIGHT_RIGHT    , TS_CHEMDRIVER_STATE_STRAIGHT, TS_CHEMDRIVER_STATE_BENT_RIGHT },
};

static int can_polymerize_plus_end(ts_chem_driver *drv, int r, int c)
{
  if (r+4 >= (int)drv->latt.nrow) { return 0; }

  assert(r>0 && r+4 < (int)drv->latt.nrow);

  if (TS_LATTICE_SPIN(drv, r-1,c) == TS_CHEMDRIVER_STATE_EMPTY) { return 0; }

  if (TS_LATTICE_SPIN(drv, r+0,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r+1,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r+2,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r+3,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }

  return 1;
}

static int can_polymerize_minus_end(ts_chem_driver *drv, int r, int c)
{
  if (r < 4) { return 0; }
  if (!tsch_test_flags(drv, TSCH_MINUS_END_POLYMERIZATION)) { return 0; }

  if (TS_LATTICE_SPIN(drv, r+1,c) == TS_CHEMDRIVER_STATE_EMPTY) { return 0; }

  if (TS_LATTICE_SPIN(drv, r-0,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r-1,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r-2,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }
  if (TS_LATTICE_SPIN(drv, r-3,c) != TS_CHEMDRIVER_STATE_EMPTY) { return 0; }

  return 1;
}

static TSResult polymerize_on_reaction(ts_chem_driver* drv, int r, int c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor, e_forbid;
  int straight_allowed;
  size_t i,n,ind;
  int plus_rxn, minus_rxn, dr;

  if (!ts_chem_driver_is_interdimer_longitudinal(drv,r,c)) { return TS_OK; }

  prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_K_ON];
  if (prefactor <= 0.0) { return TS_OK; }

  e_forbid = drv->forbidden_energy;
  if ((drv->monomer_energies[TS_CHEMDRIVER_STATE_BENT_LEFT ] >= e_forbid) &&
      (drv->monomer_energies[TS_CHEMDRIVER_STATE_BENT_RIGHT] >= e_forbid))
  {
    return TS_OK;
  }

  ind = nmovesp[0];
  if (ind + ARRAYLEN(on_reaction_table) >= TS_CHEMDRIVER_STATE_MAX) 
  {
    TS_RAISE_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
        "state list is too small to add polymerization reactions");
  }

  straight_allowed = tsch_test_flags(drv, TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS) &&
    !tsch_test_flags(drv, TSCH_DISALLOW_POLYM_AT_STRAIGHT_IFCS);

  plus_rxn  = can_polymerize_plus_end(drv,r,c);
  minus_rxn = can_polymerize_minus_end(drv,r,c);

  assert(!plus_rxn || !minus_rxn);

  if      (plus_rxn ) { dr =  1; }
  else if (minus_rxn) { dr = -1; }
  else                { return TS_OK; }

  for (i=0, n=ARRAYLEN(on_reaction_table); i < n; i++)
  {
    double rate;

    if (!straight_allowed &&
        (on_reaction_table[i].inter == TS_CHEMDRIVER_STATE_STRAIGHT))
    {
      continue;
    }

    TS_LATTICE_SPIN(drv,r+1*dr,c) = on_reaction_table[i].mer;
    TS_LATTICE_SPIN(drv,r+2*dr,c) = TS_CHEMDRIVER_STATE_BENT_IFC;
    TS_LATTICE_SPIN(drv,r+3*dr,c) = on_reaction_table[i].mer;

    rate = 0.0;
    const enum TS_CHEMDRIVER_SITE_STATE st = on_reaction_table[i].inter;
    calc_rate_single_spin(drv, r,c, &st,&st+1, &rate, prefactor);

    TS_LATTICE_SPIN(drv,r+0*dr,c) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv,r+1*dr,c) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv,r+2*dr,c) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv,r+3*dr,c) = TS_CHEMDRIVER_STATE_EMPTY;

    if (rate > 0.0)
    {
      states[ind] = on_reaction_table[i].pspin;
      rates[ind] = rate;
      ind++;
    }
  }

  nmovesp[0] = ind;

  return TS_OK;
}

static TSResult polymerize_off_reaction(ts_chem_driver* drv, int r, int c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor, rate;
  int st0;
  int rmer[2],cmer[2], nlong[2], nlat[2];
  size_t ind;

  if (!ts_chem_driver_is_interdimer_longitudinal(drv,r,c)) { return TS_OK; }
  if ((r < 1) || (r+4 >= (int)drv->latt.nrow)) { return TS_OK; }

  prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_POLYMERIZE];
  if (prefactor <= 0.0) { return TS_OK; }

  st0 = TS_LATTICE_SPIN(drv, r,c);

  if (st0 == TS_CHEMDRIVER_STATE_EMPTY) { return TS_OK; }

  if ((st0 == TS_CHEMDRIVER_STATE_STRAIGHT) &&
      !tsch_test_flags(drv, TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS))
  {
    return TS_OK;
  }

  if (tsch_test_flags(drv, TSCH_DISABLE_CCS_DEPOLYMERIZATION)
      && st0 == TS_CHEMDRIVER_STATE_CURVED
      && r > 0
      && r < (int)drv->latt.nrow-1)
  {
    if (tsch_is_curved(TS_LATTICE_SPIN(&drv->latt,r-1,c)) &&
        tsch_is_straight(TS_LATTICE_SPIN(&drv->latt,r+1,c)))
    {
      return TS_OK;
    }
  }

  rmer[0] = r-1; cmer[0] = c;
  rmer[1] = r+1; cmer[1] = c;
  nlong[0] = nlong[1] = 0;
  nlat[0] = nlat[1] = 0;
  RETURN_ON_ERROR( ts_chem_driver_count_bonds(drv, rmer, cmer, nlong, nlat, 2) );
  if ((nlat[0] > 0) && (nlat[1] > 0)) { return TS_OK; }

  ind = nmovesp[0];
  if (ind + 1 >= TS_CHEMDRIVER_STATE_MAX) 
  {
    TS_RAISE_ERROR_MESG(TS_ERR_INSUFFICIENT_SIZE,
        "state list is too small to add polymerization move");
  }

  {
    const enum TS_CHEMDRIVER_SITE_STATE stz = TS_CHEMDRIVER_STATE_EMPTY;
    double delta_en = 0.0;
    int dissoc_barrier = -1;
    int alpha_mer, long_ifc, beta_mer;

    assert(r > 0 && r < (int)drv->latt.nrow-1);

    calc_delta_energy(drv, r,c, &stz, &stz+1, &delta_en);

    alpha_mer = TS_LATTICE_SPIN(&drv->latt,r-1,c);
    long_ifc  = TS_LATTICE_SPIN(&drv->latt,r  ,c);
    beta_mer  = TS_LATTICE_SPIN(&drv->latt,r+1,c);

    assert(alpha_mer >= 0 && alpha_mer < TS_CHEMDRIVER_STATE_MAX);
    assert(long_ifc  >= 0 && long_ifc  < TS_CHEMDRIVER_STATE_MAX);
    assert(beta_mer  >= 0 && beta_mer  < TS_CHEMDRIVER_STATE_MAX);

    dissoc_barrier = find_dissociation_barrier(drv, alpha_mer, long_ifc, beta_mer);
    if (dissoc_barrier >= 0)
    {
      double db;

      if (drv->barrier_model != TS_CHEMDRIVER_BARRIER_METROPOLIS)
      {
        TS_RAISE_ERROR_MESG(TS_ERR_CHEMDRIVER_BARRIER_MODEL_CONFLICT,
            "Dissociation barriers are only supported with a Metropolis barrier model");
      }

      db = drv->dissociation_barrier_energies[dissoc_barrier];

      if (db >= 0) {
        delta_en += db;
      } else {
        double delta_en1 = delta_en + db;
        if (delta_en1 >= 0) {
          delta_en = delta_en1;
        } else {
          prefactor *= exp(-db);
        }
      }
    }

    rate = calc_rate(drv, prefactor, delta_en);
    if (rate > 0.0)
    {
      rates[ind] = rate;
      states[ind] = PSEUDO_K_OFF;
      nmovesp[0] = ind+1;
    }
  }

  return TS_OK;
}

static enum TS_CHEMDRIVER_SITE_STATE lateral_reaction_states[] = {
  TS_CHEMDRIVER_STATE_EMPTY    ,
  TS_CHEMDRIVER_STATE_SHEETLIKE,
  TS_CHEMDRIVER_STATE_TUBELIKE ,
};

static int lateral_reaction_pspins[] = {
  PSEUDO_LATERAL_ZERO     ,
  PSEUDO_LATERAL_SHEETLIKE,
  PSEUDO_LATERAL_TUBELIKE ,
};

static enum TS_CHEMDRIVER_SITE_STATE lateral_nosheet_states[] = {
  TS_CHEMDRIVER_STATE_EMPTY    ,
  TS_CHEMDRIVER_STATE_TUBELIKE ,
};

static int lateral_nosheet_pspins[] = {
  PSEUDO_LATERAL_ZERO     ,
  PSEUDO_LATERAL_TUBELIKE ,
};

STATIC_ASSERT(ARRAYLEN(lateral_reaction_states) ==
    ARRAYLEN(lateral_reaction_pspins));

STATIC_ASSERT(ARRAYLEN(lateral_nosheet_states) ==
    ARRAYLEN(lateral_nosheet_pspins));

static inline bool
is_curved_dimer(ts_chem_driver *drv, int r, int c)
{
  assert( ts_chem_driver_is_alpha_monomer(drv,r,c) );
  assert( r+2 <= INT_MAX && drv->latt.nrow <= INT_MAX );
  assert( r+2 < (int)drv->latt.nrow );

  return tsch_is_curved(TS_LATTICE_SPIN(&drv->latt, r,c))
    && tsch_is_curved(TS_LATTICE_SPIN(&drv->latt, r+1,c))
    && tsch_is_curved(TS_LATTICE_SPIN(&drv->latt, r+2,c));
}

static inline bool
is_dimer_curved_and_adjacent_to_curved(ts_chem_driver *drv, int r, int c)
{
  int nrow = (int)drv->latt.nrow;

  if (r < 0 || r >= nrow)
  {
    return false;
  }

  if (!is_curved_dimer(drv,r,c))
  {
    return false;
  }

  if (r-4 >= 0)
  {
    int inter = TS_LATTICE_SPIN(&drv->latt, r-1,c);
    if (inter != TS_CHEMDRIVER_STATE_EMPTY)
    {
      if (!tsch_is_curved(inter) || !is_curved_dimer(drv,r-4,c))
      {
        return false;
      }
    }
  }

  if (r+4+2 < nrow)
  {
    int inter = TS_LATTICE_SPIN(&drv->latt, r+3,c);
    if (inter != TS_CHEMDRIVER_STATE_EMPTY)
    {
      if (!tsch_is_curved(inter) || !is_curved_dimer(drv,r+4,c))
      {
        return false;
      }
    }
  }

  return true;
}

static inline bool
lateral_to_restricted_curved(ts_chem_driver *drv, int r, int c)
{
  assert( ts_chem_driver_is_alpha_lateral(drv,r,c) );
  assert( c > 0 && c < (int)drv->latt.ncol-1 );

  return is_dimer_curved_and_adjacent_to_curved(drv,r,c-1) ||
    is_dimer_curved_and_adjacent_to_curved(drv,r,c+1);

}

static inline bool
seam_lateral_to_restricted_curved(ts_chem_driver *drv, int r, int c)
{
  int rm1,cm1, rp1,cp1;

  assert( ts_chem_driver_is_beta_lateral(drv,r,c) );
  assert(c == (int)drv->latt.ncol-1);

  rm1 = r; cm1 = c-1;
  rp1 = r; cp1 = c+1; ts_lattice_wrap_coord(&drv->latt, &rp1, &cp1);

  if (is_dimer_curved_and_adjacent_to_curved(drv,rp1,cp1))
  {
    return true;
  }

  assert(rm1 > 2);
  if (is_dimer_curved_and_adjacent_to_curved(drv,rm1-2,cm1))
  {
    return true;
  }

  if (rm1+4 < (int)drv->latt.nrow &&
      is_dimer_curved_and_adjacent_to_curved(drv,rm1+2,cm1))
  {
    return true;
  }

  return false;
}

enum {
  SHEET_ALLOWED_SKIP_CURR = 0x01,
  SHEET_ALLOWED_ADJ_ROWS  = 0x02,
};

static inline int sheet_reaction_allowed(ts_chem_driver* drv, int r, int c, unsigned int flags);

static inline TSResult lateral_rotation_reactions(ts_chem_driver *drv,
    int r, int c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t *nmovesp)
{
  int curr, i, n, nrow;
  int sheet_allowed;
  double prefactor;
  size_t ind,ind0;
  enum TS_CHEMDRIVER_SITE_STATE *st0, *st1;
  int *pspin0;

  assert(drv->latt.nrow <= INT_MAX);
  nrow = drv->latt.nrow;

  if (!ts_chem_driver_is_alpha_lateral(drv,r,c)) { return TS_OK; }

  prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_LATERAL];
  if (prefactor <= 0.0) { return TS_OK; }

  curr = ts_lattice_at(&drv->latt,r,c).val;

  if ((TS_LATTICE_SPIN(drv,r,c) != curr) || (TS_LATTICE_SPIN(drv,r+2,c) != curr))
  {
    return TS_OK;
  }

  if (curr == TS_CHEMDRIVER_STATE_EMPTY)
  {
    int rl,rr, cl,cr;
    if (!tsch_test_flags(drv, TSCH_REVERSIBLE_LATERAL_BOND_BREAKING)) { return TS_OK; }

    rl = r; cl = c-1; ts_lattice_wrap_coord(&drv->latt, &rl, &cl);
    rr = r; cr = c+1; ts_lattice_wrap_coord(&drv->latt, &rr, &cr);

    if ((rl < 1) || (rl >= nrow-1)) { return TS_OK; }
    if ((rr < 1) || (rr >= nrow-1)) { return TS_OK; }

    if ((TS_LATTICE_SPIN(drv,rl,cl) == TS_CHEMDRIVER_STATE_EMPTY) ||
        (TS_LATTICE_SPIN(drv,rr,cr) == TS_CHEMDRIVER_STATE_EMPTY))
    {
      return TS_OK;
    }
  }

  ind = nmovesp[0];

  sheet_allowed = 1;
  if (tsch_test_flags(drv, TSCH_RESTRICT_SHEET_FLUCTS))
  {
    sheet_allowed  = sheet_reaction_allowed(drv,r,c,SHEET_ALLOWED_ADJ_ROWS);
  }

  if (sheet_allowed)
  {
    st0 = &lateral_reaction_states[0];
    st1 = &lateral_reaction_states[ARRAYLEN(lateral_reaction_states)];
    pspin0 = &lateral_reaction_pspins[0];
  }
  else
  {
    st0 = &lateral_nosheet_states[0];
    st1 = &lateral_nosheet_states[ARRAYLEN(lateral_nosheet_states)];
    pspin0 = &lateral_nosheet_pspins[0];
  }

  if (tsch_test_flags(drv, TSCH_RESTRICT_CURVED_LATERAL_BONDING) &&
      lateral_to_restricted_curved(drv, r,c))
  {
    if (curr == TS_CHEMDRIVER_STATE_EMPTY)
    {
      return TS_OK;
    }

    assert(st0[0] == TS_CHEMDRIVER_STATE_EMPTY);
    if (st0[0] != TS_CHEMDRIVER_STATE_EMPTY)
    {
      TS_RAISE_ERROR_MESG(TS_ERR_SIMULATION_STATE,
          "restricted curved bonding and first item of state is not empty");
    }

    st1 = st0+1; // only evaluate * -> empty.
  }

  calc_rate_lateral_spin(drv, r,c, st0,st1, &rates[ind], prefactor);

  for (ind0=ind,i=0,n=st1-st0; i < n; i++)
  {
    if ((int)lateral_reaction_states[i] == curr || rates[ind0+i] == 0.0)
    {
      continue;
    }

    assert(ind <= ind0+i);
    if (ind != ind0+i)
    {
      rates[ind] = rates[ind0+i];
    }
    states[ind] = pspin0[i];
    ind++;
  }

  nmovesp[0] = ind;

  assert((TS_LATTICE_SPIN(drv,r,c) == curr) && (TS_LATTICE_SPIN(drv,r+2,c) == curr));

  *nmovesp = ind;

  return TS_OK;
}

static inline int seam_reaction_allowed_by_seam_row(ts_chem_driver* drv, int r, int c);
static inline int seam_reaction_allowed(ts_chem_driver* drv, int r, int c);

static TSResult lateral_seam_reactions(ts_chem_driver* drv, int r, int c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX], size_t* nmovesp)
{
  double prefactor;
  int curr;
  enum TS_CHEMDRIVER_SITE_STATE newstate;
  size_t ind;

  prefactor = drv->rate_prefactors[TS_CHEMDRIVER_EVT_LATERAL_SEAM];

  if (prefactor <= 0.0)                         { return TS_OK; }
  if (c+1 != (int)drv->latt.ncol)               { return TS_OK; }
  if (!ts_chem_driver_is_beta_lateral(drv,r,c)) { return TS_OK; }
  if (!seam_reaction_allowed(drv,r,c))          { return TS_OK; }

  curr = ts_lattice_at(&drv->latt,r,c).val;

  if (curr == TS_CHEMDRIVER_STATE_EMPTY)
  {
    if (tsch_test_flags(drv, TSCH_RESTRICT_CURVED_LATERAL_BONDING) &&
        seam_lateral_to_restricted_curved(drv, r,c))
    {
      return TS_OK;
    }

    newstate = TS_CHEMDRIVER_STATE_TUBELIKE;
  }
  else if (curr == TS_CHEMDRIVER_STATE_TUBELIKE)
  {
    newstate = TS_CHEMDRIVER_STATE_EMPTY;
  }
  else
  {
    TS_RAISE_ERRORF(TS_ERR_LATT_BAD_STATE,
        "invalid state for seam move (%d)", curr);
  }

  ind = *nmovesp;
  if (ind + 1 >= TS_CHEMDRIVER_STATE_MAX) 
  {
    TS_RAISE_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
        "possible states exceed list maximum");
  }

  calc_rate_lateral_spin(drv, r,c,
      &newstate, &newstate+1,
      &rates[ind],
      prefactor);

  if (rates[ind] > 0.0)
  {
    states[ind] = PSEUDO_LATERAL_SEAM;
    *nmovesp = ind+1;
  }

  return TS_OK;
}


TSResult
ts_chem_driver_generate_site_moves(ts_chem_driver* drv,
    size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX],
    size_t* nmovesp)
{
  for (int i=0; i < TS_CHEMDRIVER_STATE_MAX; i++) { states[i] = -1; rates[i] = 0.0; }

  *nmovesp = 0;

  switch (ts_chem_driver_site_type(drv,r,c))
  {
    case TS_CHEMDRIVER_SITE_MONOMER:
      if (is_gtp(drv,r,c))
      {
        RETURN_ON_ERROR( monomer_reactions_gtp(drv,r,c,states,rates,nmovesp) );
        RETURN_ON_ERROR( monomer_reaction_hydrolysis(drv, r,c, states,rates,nmovesp) );
      }
      else
      {
        RETURN_ON_ERROR( monomer_reactions_gdp(drv,r,c,states,rates,nmovesp) );
      }
      break;

    case TS_CHEMDRIVER_SITE_LONG_IFC:
      RETURN_ON_ERROR( longitudinal_reactions(drv,r,c,states,rates,nmovesp) );
      RETURN_ON_ERROR( polymerize_on_reaction(drv, r,c, states,rates,nmovesp) );
      RETURN_ON_ERROR( polymerize_off_reaction(drv, r,c, states,rates,nmovesp) );
      break;

    case TS_CHEMDRIVER_SITE_LAT_IFC:
      if (!ts_chem_driver_is_seam(drv,r,c))
      {
        RETURN_ON_ERROR( lateral_rotation_reactions(drv, r,c, states, rates, nmovesp) );
      }
      else
      {
        RETURN_ON_ERROR( lateral_seam_reactions(drv, r,c, states,rates,nmovesp) );
      }
      break;

    default:
      states[0] = -1;
  }

  if (UPDATE_DEBUG_LEVEL > 1)
  {
    int i;
    fprintf(stderr, "generated site moves: %3zu,%3zu  %zu moves\n", r,c,*nmovesp);
    for (i = 0; i < (int)*nmovesp; i++)
    {
      fprintf(stderr, "  [%3d] %3d %7.4e\n", i,states[i],rates[i]);
    }
  }

  return TS_OK;
}


static int chem_driver_gen_moves(void* closure, ts_lattice* latt,
    const int *restrict rv, const int *restrict cv, size_t nsites,
    spin_t* states, double* rates, size_t *restrict siteinds, int move_max,
    intptr_t* cdata, TSResult* errp)
{
  UNUSED_PARAM(closure);
  size_t site = *cdata;
  int move = 0;

  ts_chem_driver* drv = (ts_chem_driver*)latt;

  for (; (site < nsites) && (move+TS_CHEMDRIVER_STATE_MAX < move_max); site++)
  {
    int move_states[TS_CHEMDRIVER_STATE_MAX];
    double move_rates[TS_CHEMDRIVER_STATE_MAX];
    size_t nmoves = 0;
    size_t move_ind;

    if (rv[site] < 0 || (size_t)rv[site] >= drv->latt.nrow ||
        cv[site] < 0 || (size_t)cv[site] >= drv->latt.ncol)
    {
      static FILE *dbgf = NULL;
      if (!dbgf)
      {
        dbgf = fopen("dump.out","w");
      }

      fprintf(dbgf, "site %zu  r = %d, c = %d off the lattice\n",
          site, rv[site], cv[site]);
      fflush(dbgf);
    }

    assert(rv[site] >= 0 && (size_t)rv[site] < drv->latt.nrow);
    assert(cv[site] >= 0 && (size_t)cv[site] < drv->latt.ncol);
    TSResult ret = ts_chem_driver_generate_site_moves(drv,
        rv[site],cv[site],
        move_states,
        move_rates,
        &nmoves);
    if (IS_ERROR(ret))
    {
      *errp = ret;
      return -1;
    }

    memcpy(rates+move,move_rates+0,nmoves*sizeof(rates[0]));
    for (move_ind=0; move_ind < nmoves; move++,move_ind++)
    {
      states[move] = spin_from_int(move_states[move_ind]);
      siteinds[move] = site;
    }
  }

  if (site < nsites)
  {
    if (move == 0)
    {
      /* cannot return anything!  signal an error! */
      *errp = TS_NEW_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
          "move buffer size is too small to return any moves");
      return -1;
    }

    *cdata = site;
    return -move;
  }
  else
  {
    *cdata = 0;
    return move;
  }

  return move;
}

static size_t find_moveset_duplicate(int states[], size_t n)
{
  size_t mv1,mv2;
  for (mv1=0; mv1 < n; mv1++)
  {
    for (mv2=mv1+1; mv2 < n; mv2++)
    {
      if (states[mv1] == states[mv2])
      {
        return mv1;
      }
    }
  }
  return n;
}

static bool is_move_consistent(ts_lattice_move *m, int states[], double rates[], size_t n)
{
  size_t mv;
  int st;

  st = m->state.val;
  for (mv=0; mv < n; mv++)
  {
    if (states[mv] != st)
    {
      continue;
    }

    states[mv] = INT_MAX;
    return (fabs(rates[mv] - m->rate) <= RATE_CHECK_TOL);
  }

  return false;
}

TSResult
ts_chem_driver_check_move_consistency(ts_chem_driver* drv,
    size_t *rv, size_t *cv, size_t nv)
{
  int i,j;

  int states[TS_CHEMDRIVER_STATE_MAX];
  double rates[TS_CHEMDRIVER_STATE_MAX];
  size_t nmoves, num_inconsistent;
  ts_lattice_move *curr_moves;

  if (drv == NULL) TS_RAISE_ERROR_MESG(TS_ERR_INVALID_ARGUMENT, "cannot pass drv == NULL");
  if (rv == NULL)  TS_RAISE_ERROR_MESG(TS_ERR_INVALID_ARGUMENT, "cannot pass rv == NULL");
  if (cv == NULL)  TS_RAISE_ERROR_MESG(TS_ERR_INVALID_ARGUMENT, "cannot pass cv == NULL");
  if (nv <= 1)     TS_RAISE_ERROR_MESG(TS_ERR_INVALID_ARGUMENT, "requires nv >= 2");

  for (i=0; (size_t)i < nv; i++)
  {
    rv[i] = ~((size_t)0);
    cv[i] = ~((size_t)0);
  }

  for (i=0; (size_t)i < drv->latt.ncol; i++)
  {
    int range[2] = { -1, -1 };
    ts_lattice_find_column_row_range(&drv->latt, &range[0], i);
    if (drv->row_range[2*i+0] != range[0] || drv->row_range[2*i+1] != range[1])
    {
      rv[0] = cv[0] = (size_t)i;
      TS_RAISE_ERRORF(TS_ERR_CHEMDRIVER_INCONSISTENT_ROW_RANGE,
          "cached row range is inconsistent at column %d", i);
    }
  }

  num_inconsistent = 0;

  for (j=0; (size_t)j < drv->latt.ncol; j++)
  {
    for (i=0; (size_t)i < drv->latt.nrow; i++)
    {
      TSResult err;

      if (ts_chem_driver_is_blank(drv, i,j))
      {
        continue;
      }

      curr_moves = ts_lattice_get_current_site_moves(&drv->latt,i,j);

      nmoves = 0;
      err = ts_chem_driver_generate_site_moves(drv, i,j, states, rates, &nmoves);
      if (!IS_OKAY(err))
      {
        const struct ts_error_state *st;
        char buf[ERR_MESG_LEN];

        st = ts_error_get_state(err);
        rv[0] = i;
        cv[0] = j;
        memcpy(buf, st->error_message, ERR_MESG_LEN-1);
        buf[ERR_MESG_LEN-1] = '\0';

        TS_RAISE_ERRORF(TS_ERR_INVALID_RETURN,
            "error encountered getting move for site %3zu,%3zu: (%4x) %s",
            i,j, ERRCODE(err), buf);
      }

      if (!curr_moves && nmoves == 0)
      {
        continue;
      }

      {
        size_t stdup;
        stdup = find_moveset_duplicate(states,nmoves);
        if (stdup < nmoves)
        {
          rv[0] = i;
          cv[0] = j;

          TS_RAISE_ERRORF(TS_ERR_LATT_DUPLICATE_MOVE,
              "duplicate move (state %d) in generated set at site %3zu,%3zu",
              states[stdup], i,j);
        }
      }

      for (; curr_moves != NULL; curr_moves = curr_moves->next)
      {
        if (curr_moves->r != (size_t)i || curr_moves->c != (size_t)j)
        {
          rv[0] = i;
          cv[0] = j;
          rv[1] = curr_moves->r;
          cv[1] = curr_moves->c;

          TS_RAISE_ERRORF(TS_ERR_IN_ALGORITHM,
              "inconsistent move set for site %3zu,%3zu: move returned for site (%3zu,%3zu)",
              i,j, curr_moves->r, curr_moves->c);
        }

        if (!is_move_consistent(curr_moves, states, rates, nmoves))
        {
          if (num_inconsistent >= nv)
          {
            TS_RAISE_ERRORF(TS_ERR_INSUFFICIENT_SIZE,
                "more than %zu sites with inconsistent moves were identified",
                nv);
          }

          rv[num_inconsistent] = i;
          cv[num_inconsistent] = j;
          num_inconsistent++;
          goto site_inconsistent;
        }
      }

      {
        size_t mv;
        for (mv=0; mv < nmoves; mv++)
        {
          if (states[mv] != INT_MAX)
          {
            if (num_inconsistent >= nv)
            {
              TS_RAISE_ERRORF(TS_ERR_INSUFFICIENT_SIZE,
                  "more than %zu sites with inconsistent moves were identified",
                  nv);
            }

            rv[num_inconsistent] = i;
            cv[num_inconsistent] = j;
            num_inconsistent++;
            goto site_inconsistent;
          }
        }
      }
site_inconsistent:
      (void)num_inconsistent;
    }
  }

  if (num_inconsistent > 0)
  {
    TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
        "found inconsistencies at %zu sites", num_inconsistent);
  }

  return TS_OK;
}


static TSResult pspin_lateral_update_coupled_sites(
    ts_chem_driver *drv, int ra, int ca, int new_state, int old_state)
{
  bool is_plus, is_minus;
  int lb,ub,i,n;
  int prev_plus,prev_minus;

  if (new_state != 0 && old_state != 0)
  {
    return TS_OK;
  }

  if (drv->row_range[2*ca+0] < 0 ||
      drv->row_range[2*ca+1] < 0)
  {
    RETURN_ON_ERROR( update_row_range(drv, ca) );
  }

  prev_minus = drv->row_range[2*ca+0];
  prev_plus = drv->row_range[2*ca+1];

  is_plus  = (ra+2 >= prev_plus);
  is_minus = (ra+0 <= prev_minus);

  if (!is_plus && !is_minus)
  {
    return TS_OK;
  }

  RETURN_ON_ERROR( update_row_range(drv,ca) );

  n = 2*drv->coupling_edge_dist;

  if (is_plus)
  {
    int plus_min,plus_max;

    plus_min = prev_plus;
    plus_max = drv->row_range[2*ca+1];

    if (plus_min > plus_max)
    {
      int tmp = plus_min; plus_min = plus_max; plus_max = tmp;
    }

    lb = plus_min - 1 - 2*n;
    ub = (plus_max == ra+2) ? ra-1 : plus_max;
  }
  else
  {
    int minus_min, minus_max;
    minus_min = drv->row_range[2*ca+0];
    minus_max = prev_minus;

    if (minus_min > minus_max)
    {
      int tmp = minus_min; minus_min = minus_max; minus_max = tmp;
    }

    ub = minus_max + 1 + 2*n;
    lb = (minus_min == ra) ? ra+1 : minus_min;

  }

  if (lb <= 0) { lb = 1; }
  if (ub >= (int)drv->latt.nrow-1)
  {
    ub = drv->latt.nrow-2;
  }

  if (lb > ub)
  {
    int tmp = lb;
    lb = ub;
    ub = tmp;
  }

  RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, lb,ub,ca) );

  RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, ca, PSEUDO_FIND_ROW_RANGE));

  return TS_OK;
}

static TSResult pspin_lateral_change(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin)
{
  ts_chem_driver* drv;
  int old_state,new_state;

  UNUSED_PARAM(closure);

  drv = (ts_chem_driver*)latt;

  switch (pspin.val)
  {
    case PSEUDO_LATERAL_ZERO:
      new_state = TS_CHEMDRIVER_STATE_EMPTY;
      break;

    case PSEUDO_LATERAL_SHEETLIKE:
      new_state = TS_CHEMDRIVER_STATE_SHEETLIKE;
      break;

    case PSEUDO_LATERAL_TUBELIKE:
      new_state = TS_CHEMDRIVER_STATE_TUBELIKE;
      break;

    default:
      TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
          "Invalid pseudo-spin %d for LATERAL-CHANGE", pspin.val);
  }

  if (!ts_chem_driver_is_alpha_lateral(drv,ra,ca))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "LATERAL-CHANGE invoked at (%d,%d), which is not an alpha lateral site",
        ra,ca);
  }

  if (TS_LATTICE_SPIN(drv,ra,ca) != TS_LATTICE_SPIN(drv,ra+2,ca))
  {
    TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
        "LATERAL-CHANGE alpha (%d,%d) is %d != beta spin %d",
        ra,ca,
        TS_LATTICE_SPIN(drv,ra,ca),
        TS_LATTICE_SPIN(drv,ra+2,ca));
  }

  old_state = TS_LATTICE_SPIN(drv,ra+0,ca);

  RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, ra+0,ca) );
  RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, ra+2,ca) );

  TS_LATTICE_SPIN(drv,ra+0,ca) = new_state;
  TS_LATTICE_SPIN(drv,ra+2,ca) = new_state;

  if (tsch_ifclog_is_logging(drv))
  {
    int mer_left, mer_right;
    assert(ca > 0 && (size_t)ca < drv->latt.ncol-1);

    mer_left   = TS_LATTICE_SPIN(drv, ra, ca-1);
    mer_right  = TS_LATTICE_SPIN(drv, ra, ca+1);

    uint16_t packed_ifc = tsch_ifclog_pack_ifc(mer_left, old_state, mer_right);
    (void)tsch_ifclog_add(drv, packed_ifc, pspin.val, ra,ca);

    RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, ca, PSEUDO_ROLLBACK_IFCLOG));
  }

  RETURN_ON_ERROR(pspin_lateral_update_coupled_sites(drv,ra,ca,new_state,old_state));

  return TS_OK;
}

static TSResult pspin_lateral_make_bond(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin)
{
  int state, new_state;

  UNUSED_PARAM(closure);
  ts_chem_driver* drv = (ts_chem_driver*)latt;

  if (pspin.val != PSEUDO_LATERAL_SEAM)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "unrecognized pseudo-spin %d for pspin_lateral_make_bond",
        pspin.val);
  }

  if (!ts_chem_driver_is_beta_lateral(drv,ra,ca))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "pspin_lateral_make_bond requires a beta lateral site, but site %d,%d is not one",
        ra,ca);
  }

  state = TS_LATTICE_SPIN(drv,ra,ca);
  if (TS_LATTICE_SPIN(drv, ra+2,ca) != state)
  {
        TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
            "site (%zu,%zu) has state %d, not equal to state %d at (%zu,%zu)",
            ra+2,ca, TS_LATTICE_SPIN(drv,ra+2,ca), ra+0, ca, state);
  }

  switch (state)
  {
    case TS_CHEMDRIVER_STATE_EMPTY:
      new_state = TS_CHEMDRIVER_STATE_TUBELIKE;
      break;

    case TS_CHEMDRIVER_STATE_TUBELIKE:
      new_state = TS_CHEMDRIVER_STATE_EMPTY;
      break;

    default:
      TS_RAISE_ERROR2(TS_ERR_SIMULATION_STATE,
          ("Invalid state %d for lateral seam move", state ));
  }

  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+0,ca) );
  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+2,ca) );

  TS_LATTICE_SPIN(drv, ra+0, ca) = new_state;
  TS_LATTICE_SPIN(drv, ra+2, ca) = new_state;

  if (tsch_ifclog_is_logging(drv))
  {
    int rm1,cm1, rp1,cp1;
    int mer_left, mer_right;

    rm1 = ra; cm1 = ca-1; ts_lattice_wrap_coord(latt, &rm1,&cm1);
    rp1 = ra; cp1 = ca+1; ts_lattice_wrap_coord(latt, &rp1,&cp1);

    mer_left   = TS_LATTICE_SPIN(drv, rm1,cm1);
    mer_right  = TS_LATTICE_SPIN(drv, rp1,cp1);

    uint16_t packed_ifc = tsch_ifclog_pack_ifc(mer_left, state, mer_right);
    (void)tsch_ifclog_add(drv, packed_ifc, pspin.val, ra,ca);

    RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, ca, PSEUDO_ROLLBACK_IFCLOG));
  }

  RETURN_ON_ERROR(pspin_lateral_update_coupled_sites(drv, ra,ca, new_state, state));

  return TS_OK;
}

static TSResult pspin_kon(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin)
{
  ts_chem_driver* drv;
  int mer, ifc, dr;

  UNUSED_PARAM(closure);
  UNUSED_PARAM(pspin);

  drv = (ts_chem_driver*)latt;

  switch (pspin.val)
  {
    case PSEUDO_K_ON_BENT_LEFT:
      ifc = TS_CHEMDRIVER_STATE_BENT_IFC;
      mer = TS_CHEMDRIVER_STATE_BENT_LEFT;
      break;

    case PSEUDO_K_ON_BENT_RIGHT:
      ifc = TS_CHEMDRIVER_STATE_BENT_IFC;
      mer = TS_CHEMDRIVER_STATE_BENT_RIGHT;
      break;

    case PSEUDO_K_ON_STRAIGHT_LEFT:
      ifc = TS_CHEMDRIVER_STATE_STRAIGHT;
      mer = TS_CHEMDRIVER_STATE_BENT_LEFT;
      break;

    case PSEUDO_K_ON_STRAIGHT_RIGHT:
      ifc = TS_CHEMDRIVER_STATE_STRAIGHT;
      mer = TS_CHEMDRIVER_STATE_BENT_RIGHT;
      break;

    default:
      TS_RAISE_ERRORF(TS_ERR_UNSUPPORTED_OPERATION,
          "unsupported pseudospin %d passed to %s",
          (int)pspin.val, __func__);
  }

  if (TS_LATTICE_SPIN(drv,ra-1,ca) != TS_CHEMDRIVER_STATE_EMPTY)
  {
    dr = 1;
  }
  else if (TS_LATTICE_SPIN(drv,ra+1,ca) != TS_CHEMDRIVER_STATE_EMPTY)
  {
    dr = -1;
  }
  else
  {
    TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
        "Polymerization at (%d,%d), but adjacent dimers are empty", ra,ca);
  }

  if ((TS_LATTICE_SPIN(drv,ra+0*dr,ca) != TS_CHEMDRIVER_STATE_EMPTY) ||
      (TS_LATTICE_SPIN(drv,ra+1*dr,ca) != TS_CHEMDRIVER_STATE_EMPTY) ||
      (TS_LATTICE_SPIN(drv,ra+2*dr,ca) != TS_CHEMDRIVER_STATE_EMPTY) ||
      (TS_LATTICE_SPIN(drv,ra+3*dr,ca) != TS_CHEMDRIVER_STATE_EMPTY))
  {
    TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
        "Polymerization at (%d,%d) to (%d,%d), but not all sites are empty", 
        ra,ca, ra+3*dr,ca);
  }

  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+0*dr,ca) );
  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+1*dr,ca) );
  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+2*dr,ca) );
  RETURN_ON_ERROR( ts_lattice_mark_site_changed((ts_lattice*)drv, ra+3*dr,ca) );

  TS_LATTICE_SPIN(drv,ra+0*dr,ca) = ifc;
  TS_LATTICE_SPIN(drv,ra+1*dr,ca) = mer;
  TS_LATTICE_SPIN(drv,ra+2*dr,ca) = ifc;
  TS_LATTICE_SPIN(drv,ra+3*dr,ca) = mer;

  assert(dr == -1 || dr == 1);

  {
    int new_rr, rr_ind, delta;
    new_rr = ra+3*dr;
    if (dr < 0)
    {
      rr_ind = 2*ca;
      if (drv->row_range[rr_ind] > new_rr) { drv->row_range[rr_ind] = new_rr; }
    }
    else
    {
      rr_ind = 2*ca+1;
      if (drv->row_range[rr_ind] < new_rr) { drv->row_range[rr_ind] = new_rr; }
    }
  }

  RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, ca, PSEUDO_FIND_ROW_RANGE));
  return TS_OK;
}

static int find_lateral_bonds(ts_chem_driver *drv, const int rl, const int cl);

static TSResult pspin_koff(void *data, ts_lattice *latt, int r, int c, spin_t pspin)
{
  ts_chem_driver* drv;
  int oldspin;
  int lateral_bonds;
  int alpha,beta;

  UNUSED_PARAM(data);
  UNUSED_PARAM(pspin);

  drv = (ts_chem_driver*)latt;

  if (tsch_ifclog_is_logging(drv))
  {
    alpha   = TS_LATTICE_SPIN(drv,r-1,c);
    oldspin = TS_LATTICE_SPIN(drv, r,c);
    beta    = TS_LATTICE_SPIN(drv,r+1,c);
  }
  else
  {
    alpha = beta = 0;
    oldspin = TS_LATTICE_SPIN(drv, r,c);
  }

  if ((oldspin == TS_CHEMDRIVER_STATE_EMPTY) ||
      (!ts_chem_driver_is_interdimer_longitudinal(drv, r,c) &&
       !ts_chem_driver_is_lateral(drv,r,c)))
  {
    return TS_OK;
  }

  RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, r, c) );
  TS_LATTICE_SPIN(drv, r,c) = TS_CHEMDRIVER_STATE_EMPTY;

  lateral_bonds = find_lateral_bonds(drv, r,c);

  switch (lateral_bonds)
  {
    case TIP_BOTH:
      {
        RETURN_ON_ERROR( ts_lattice_rollback_changes(&drv->latt) );
        return TS_OK;
      }

    case TIP_MINUS:
      {
        int i,nr;
        nr = drv->latt.nrow;
        for(i=r+1; (i < nr) && (TS_LATTICE_SPIN(drv,i,c) != TS_CHEMDRIVER_SITE_BLANK); i++)
        {
          RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, i, c) );
          TS_LATTICE_SPIN(drv,i,c) = TS_CHEMDRIVER_SITE_BLANK;
        }

        break;
      }

    case TIP_PLUS:
      {
        int i;
        if (!tsch_test_flags(drv, TSCH_MINUS_END_POLYMERIZATION) &&
            tsch_test_flags(drv, TSCH_NO_MINUS_END_DEPOLYMERIZATION)) {
          return TS_OK;
        }

        for(i=r-1; (i >= 0) && (TS_LATTICE_SPIN(drv,i,c) != TS_CHEMDRIVER_SITE_BLANK); i--)
        {
          RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, i, c) );
          TS_LATTICE_SPIN(drv,i,c) = TS_CHEMDRIVER_SITE_BLANK;
        }

        break;
      }

    default:
      {
        TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
            "after breaking longitudinal bond at (%3d,%3d) with spin %d, "
            "protofilament has no lateral bonds!", r, c, pspin.val);
      }
  }

  if (tsch_ifclog_is_logging(drv))
  {
    uint16_t packed_ifc = tsch_ifclog_pack_ifc(alpha, oldspin, beta);
    (void)tsch_ifclog_add(drv, packed_ifc, pspin.val, r,c);
    RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, c, PSEUDO_ROLLBACK_IFCLOG));
  }

  RETURN_ON_ERROR( update_row_range(drv,c) );
  RETURN_ON_ERROR(ts_lattice_add_pspin_change(&drv->latt, -1, c, PSEUDO_FIND_ROW_RANGE));
  return TS_OK;
}

static TSResult pspin_find_row_range(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin)
{
  UNUSED_PARAM(closure);
  UNUSED_PARAM(ra);
  UNUSED_PARAM(pspin);

  RETURN_ON_ERROR(update_row_range((ts_chem_driver *)latt, ca));
  return TS_OK;
}

static TSResult pspin_rollback_ifclog(void* closure, ts_lattice* latt, int ra, int ca, spin_t pspin)
{
  int nentries = -ra;
  UNUSED_PARAM(closure);
  UNUSED_PARAM(ca);
  UNUSED_PARAM(pspin);

  ts_chem_driver *drv = (ts_chem_driver *)latt;
  if (tsch_ifclog_is_logging(drv) && nentries > 0)
  {
    if (drv->ifclog.n >= (size_t)nentries)
    {
      drv->ifclog.n -= nentries;
    }
    else
    {
      drv->ifclog.n = 0;
    }
  }

  return TS_OK;
}

static void update_wrapped_row(ts_chem_driver *drv, int wr, int wc)
{
  int nrow = drv->latt.nrow;

  if (ts_chem_driver_is_beta_lateral(drv,wr,wc))
  {
    ts_lattice_mark_site_changed(&drv->latt, wr,wc);
    return;
  }

  if (wr-2 >= 1)
  {
    ts_lattice_mark_site_changed(&drv->latt, wr-2,wc);
  }

  if (wr+2 < nrow-1)
  {
    ts_lattice_mark_site_changed(&drv->latt, wr+2,wc);
  }

  if (UPDATE_DEBUG_LEVEL > 0)
  {
    if (wr-2 >= 1) {
      fprintf(stderr, "\nmarked site %3d,%3d as changed\n", wr-2,wc);
    }
    if (wr+2 < nrow-1) {
      fprintf(stderr, "marked site %3d,%3d as changed\n", wr+2,wc);
    }
  }
}

static TSResult update_seam_reaction_hook(void* data, ts_lattice* latt, const ts_lattice_move* move)
{
  ts_chem_driver *drv;
  int r,c, site_type, state, r0,r1;
  int wr1,wc1, wr2,wc2, nrow,ncol;

  UNUSED_PARAM(data);

  drv = (ts_chem_driver *)latt;

  r = move->r;
  c = move->c;
  site_type = ts_chem_driver_site_type(drv, r,c);

  if ((site_type != TS_CHEMDRIVER_SITE_LAT_IFC) && (site_type != TS_CHEMDRIVER_SITE_MONOMER))
  {
    return TS_OK;
  }

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  wr1 = r; wc1 = drv->latt.ncol-1;
  ts_lattice_wrap_coord(latt, &wr1,&wc1);

  wr2 = r; wc2 = 0;
  ts_lattice_wrap_coord(latt, &wr2,&wc2);

  update_wrapped_row(drv, wr1,wc1);
  update_wrapped_row(drv, wr2,wc2);

  if (!tsch_test_flags(drv, TSCH_RESTRICT_SHEET_FLUCTS) || 
      (site_type != TS_CHEMDRIVER_SITE_LAT_IFC))
  {
    if (UPDATE_DEBUG_LEVEL > 0)
    {
      fprintf(stderr, "sheet fluct allowed or not lateral site, not "
          "updating rows around site %3d,%3d\n", r,c);
    }

    return TS_OK;
  }

  if (UPDATE_DEBUG_LEVEL > 0)
  {
    fprintf(stderr, "possibly updating rows around site %3d,%3d\n", r,c);
  }

  r0 = r-4;
  r1 = r+4;

  state = TS_LATTICE_SPIN(&drv->latt,r,c);
  switch (move->state.val)
  {
    case PSEUDO_LATERAL_SHEETLIKE:
      if (r >= 5 && !sheet_reaction_allowed(drv,r-4,c,SHEET_ALLOWED_ADJ_ROWS))
      {
        int j;
        if (UPDATE_DEBUG_LEVEL > 0)
        {
          fprintf(stderr, "  - *->e | updating row %3d\n", r-4);
        }

        for (j=1; j < ncol; j+=2)
        {
          RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, r-4,r-4,j) );
        }
      }

      if (r <= nrow-5 && !sheet_reaction_allowed(drv,r+4,c,SHEET_ALLOWED_ADJ_ROWS))
      {
        int j;

        if (UPDATE_DEBUG_LEVEL > 0)
        {
          fprintf(stderr, "  - *->e | updating row %3d\n", r+4);
        }

        for (j=1; j < ncol; j+=2)
        {
          RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, r+4,r+4,j) );
        }
      }

      return TS_OK;

    default:
      return TS_OK;

    case PSEUDO_LATERAL_SEAM:
      r0 = r-6;
      r1 = r+6;

      {
        int allowed0,allowed1,allowed2,allowed3;
        int flags = SHEET_ALLOWED_ADJ_ROWS;
        if (state == TS_CHEMDRIVER_STATE_EMPTY) { flags = flags | SHEET_ALLOWED_SKIP_CURR; }

        allowed0 = r < 7 || sheet_reaction_allowed(drv,r-6,c,0);
        allowed1 = sheet_reaction_allowed(drv,r-2,c, flags);
        allowed2 = r > nrow-3 || sheet_reaction_allowed(drv,r+2,c,0);
        allowed3 = r > nrow-7 || sheet_reaction_allowed(drv,r+6,c,0);

        if (allowed0 && allowed1 && allowed2 && allowed3)
        {
          if (UPDATE_DEBUG_LEVEL > 0)
          {
            fprintf(stderr, "  - seam : %3d:%1d %3d:%1d %3d:%1d %3d:%1d  not updating\n",
                r-6,allowed0, r-2,allowed1, r+2,allowed2, r+6,allowed3);
          }

          return TS_OK;
        }
      }
      break;

    case PSEUDO_LATERAL_ZERO:
      if (state == TS_CHEMDRIVER_STATE_SHEETLIKE)
      {
        return TS_OK;
      }

      if (sheet_reaction_allowed(drv,r,c,SHEET_ALLOWED_SKIP_CURR))
      {
        return TS_OK;
      }

      r0 = r;
      r1 = r;

      if (r >= 5 && !sheet_reaction_allowed(drv,r-4,c,0))
      {
        r0 = r-4;
      }

      if (r <= nrow-5 && !sheet_reaction_allowed(drv,r+4,c,0))
      {
        r1 = r+4;
      }
      break;

    case PSEUDO_LATERAL_TUBELIKE:
      if (sheet_reaction_allowed(drv,r,c,SHEET_ALLOWED_SKIP_CURR))
      {
        if (UPDATE_DEBUG_LEVEL > 0)
        {
          fprintf(stderr, "  -  *->t but row itself is still allowed, no updates.\n");
        }
        return TS_OK;
      }

      r0 = r1 = r;

      if (r >= 5 && sheet_reaction_allowed(drv,r-4,c,0))
      {
        r0 = r-4;
      }

      if (r <= nrow-5 && sheet_reaction_allowed(drv,r+4,c,0))
      {
        r1 = r+4;
      }
      break;
  }

update:
  {
    int i,j;

    if (r0 < 1) { r0 = 1; }
    if (r1 > nrow-1) { r1 = nrow-1; }

    if (UPDATE_DEBUG_LEVEL > 0)
    {
      fprintf(stderr, "  -  updating rows %3d - %3d\n", r0,r1);
    }

    for (j=1; j < ncol; j+=2)
    {

      RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, r0,r1,j) );
    }
  }

  return TS_OK;
}

static int tip_needs_update(ts_chem_driver *drv, int r, int c, int tip, int dr)
{
  int nlat,cdist, rm1,cm1, rp1,cp1;

  if (abs(tip - r) <= drv->latt.row_coupling_dist)
  {
    return 0;
  }

  cdist = drv->coupling_edge_dist;
  for (nlat = 0, r += dr; dr*(tip-r) > 0; r += 4*dr)
  {
    if (TS_LATTICE_SPIN(&drv->latt, r,c) == 0)
    {
      continue;
    }

    if (++nlat > cdist)
    {
      return 0;
    }
  }

  return 1;
}

static TSResult mark_potential_lateral_bonds_for_update(ts_chem_driver *drv, int r0, int c0, int tip_off, int dr)
{
  int cdist, rm1,cm1, rp1,cp1;
  int tip,tip_l,tip_r, nlat;
  int hl,hm,hr, hh;
  int i;

  cdist = drv->coupling_edge_dist;

  tip   = drv->row_range[2*c0  + tip_off];

  if (dr*(tip - r0)/2 > cdist)
  {
    return TS_OK;
  }

  rm1 = tip; cm1 = c0-1; ts_lattice_wrap_coord(&drv->latt, &rm1,&cm1);
  rp1 = tip; cp1 = c0+1; ts_lattice_wrap_coord(&drv->latt, &rp1,&cp1);

  tip_l = drv->row_range[2*cm1 + tip_off];
  tip_r = drv->row_range[2*cp1 + tip_off];

  hl = dr*(tip_l - rm1)/4;
  hr = dr*(tip_r - rp1)/4;
  nlat = (hl <= hr) ? hl : hr;

  if (nlat > 0)
  {
    int rmin,rmax;
    rmin = tip+4*dr;
    rmax = tip+4*nlat*dr;

    if (dr < 0)
    {
      int tmp = rmin;
      rmin = rmax;
      rmax = tmp;
    }

    RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, rmin,rmax, c0) );
  }

  return TS_OK;
}

static TSResult
update_tips_if_needed(ts_chem_driver *drv, int r, int c)
{
  int plus, minus, nposs;

  minus = drv->row_range[2*c + 0];
  plus  = drv->row_range[2*c + 1];

  if (tip_needs_update(drv, r,c, minus, -1))
  {
    RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, minus,minus,c) );
  }
  RETURN_ON_ERROR( mark_potential_lateral_bonds_for_update(drv,r,c,0,-1) );

  if (tip_needs_update(drv, r,c, plus, 1))
  {
    RETURN_ON_ERROR( ts_lattice_mark_range_for_update(&drv->latt, plus-2,plus-2,c) );
  }
  RETURN_ON_ERROR( mark_potential_lateral_bonds_for_update(drv,r,c,1,1) );

  return TS_OK;
}


static TSResult update_tip_reaction_hook(void *data, ts_lattice *latt, const ts_lattice_move* move)
{
  ts_chem_driver *drv = (ts_chem_driver *)latt;
  int r0,c0, r,c;

  UNUSED_PARAM(data);

  r0 = move->r;
  c0 = move->c;

  if (!ts_chem_driver_is_longitudinal(drv, r0,c0))
  {
    return TS_OK;
  }

  r = r0, c = c0-1;
  ts_lattice_wrap_coord((ts_lattice*)drv, &r,&c);
  if (r >= 0 && (size_t)r < drv->latt.nrow)
  {
    update_tips_if_needed(drv,r,c);
  }

  r = r0, c = c0+1;
  ts_lattice_wrap_coord((ts_lattice*)drv, &r,&c);
  if (r >= 0 && (size_t)r < drv->latt.nrow)
  {
    update_tips_if_needed(drv,r,c);
  }

  return TS_OK;
}

static TSResult zero_connected_component(ts_chem_driver *drv, int *restrict components, int component_to_zero);

static TSResult depolymerize_move_hook(void* data, ts_lattice* latt, const ts_lattice_move* move)
{
  ts_chem_driver* drv;
  size_t r,c;
  int newspin, oldspin;
  int *components, *working, ncomponents;

  UNUSED_PARAM(data);

  drv = (ts_chem_driver*)latt;
  r = move->r;
  c = move->c;
  newspin = move->state.val;

  if ((newspin != TS_CHEMDRIVER_STATE_EMPTY) ||
      (!ts_chem_driver_is_interdimer_longitudinal(drv, r,c) &&
       !ts_chem_driver_is_lateral(drv,r,c)))
  {
    return TS_OK;
  }

  oldspin = TS_LATTICE_SPIN(drv, r,c);
  TS_LATTICE_SPIN(drv, r,c) = newspin;

  components = drv->connected_components_scratch;
  working = components + drv->latt.num_sites;
  ncomponents = ts_lattice_connected_components(&drv->latt, components, working);

  TS_LATTICE_SPIN(drv, r,c) = oldspin;

  if (ncomponents <= 1)
  {
    CANCEL_OPERATION();
  }
  else if (ncomponents == 2)
  {
    int component_to_zero = (working[0] < working[1]) ? 1 : 2;
    RETURN_ON_ERROR( zero_connected_component(drv, components, component_to_zero) );
  }
  else
  {
    int cci;

    for(cci=0; cci < ncomponents; cci++)
    {
      fprintf(stderr, "component %d size %d\n", cci, working[cci]);
    }
    TS_RAISE_ERROR2(TS_ERR_SIMULATION_STATE,
        ("after depolymerization: number of connected components > 2 (%d)", ncomponents));
  }

  return TS_OK;
}


static void zero_populations(struct tsch_populations *pops)
{
  size_t i,j,k;
  for(i=0; i < TS_CHEMDRIVER_STATE_MAX; i++)
  {
    pops->monomer_pop[i] = 0;
  }

  for(i=0; i < TS_CHEMDRIVER_STATE_MAX; i++)
  {
    for(j=0; j < TS_CHEMDRIVER_STATE_MAX; j++)
    {
      for(k=0; k < TS_CHEMDRIVER_STATE_MAX; k++)
      {
        pops->interface_pop[i][j][k] = 0;
      }
    }
  }

  for(i=0; i < TS_CHEMDRIVER_STATE_MAX; i++)
  {
    for(j=0; j < TS_CHEMDRIVER_STATE_MAX; j++)
    {
      for(k=0; k < TS_CHEMDRIVER_STATE_MAX; k++)
      {
        pops->bending_pop[i][j][k] = 0;
      }
    }
  }
}

static inline int seam_reaction_allowed_by_seam_row(ts_chem_driver* drv, int r, int c)
{
  int ncol, num_defects;

  UNUSED_PARAM(c);

  ncol = drv->latt.ncol;
  num_defects = 0;

  for(int j=1; j < ncol-2; j+=2) {
    int mer1, ifc, mer2;

    ifc = ts_lattice_at(&drv->latt,r,j).val;
    if (ifc == TS_CHEMDRIVER_STATE_EMPTY)
    {
      continue;
    }

    if (ifc != TS_CHEMDRIVER_STATE_TUBELIKE)
    {
      num_defects++;
      continue;
    }

    mer1 = ts_lattice_at(&drv->latt,r,j-1).val;
    mer2 = ts_lattice_at(&drv->latt,r,j+1).val;
    if (!tsch_is_straight(mer1) || !tsch_is_straight(mer2))
    {
      num_defects++;
    }
  }

  return num_defects < (ncol/2) - drv->seam_allowed_thresh;
}

static inline int seam_reaction_allowed(ts_chem_driver* drv, int r, int c)
{
  int curr, ndimer, last_beta;
  int wr1,wc1;
  int wr2,wc2;

  curr = ts_lattice_at(&drv->latt,r,c).val;
  ndimer = ts_chem_driver_num_dimers(drv);
  last_beta = ts_chem_driver_beta_row(drv, ndimer-1);

  wr1 = r; wc1 = c+1;
  ts_lattice_wrap_coord((ts_lattice*)drv, &wr1,&wc1);

  if (wr1 < 1 || (size_t)wr1 >= drv->latt.nrow-4) { return 0; }

  switch (curr)
  {
    case TS_CHEMDRIVER_STATE_EMPTY:
      if (r == last_beta) { return 0; }

      if ((ts_lattice_at(&drv->latt, r    ,c-1).val == TS_CHEMDRIVER_STATE_EMPTY) ||
          (ts_lattice_at(&drv->latt, r+2  ,c-1).val == TS_CHEMDRIVER_STATE_EMPTY) ||
          (ts_lattice_at(&drv->latt, wr1  ,wc1).val == TS_CHEMDRIVER_STATE_EMPTY) ||
          (ts_lattice_at(&drv->latt, wr1+2,wc1).val == TS_CHEMDRIVER_STATE_EMPTY))
        return 0;

      if (wr1 <= 1 || TS_LATTICE_SPIN(drv,r-2,c) == TS_CHEMDRIVER_STATE_TUBELIKE)
        break;

      if (wr1 >= last_beta || TS_LATTICE_SPIN(drv,r+4,c) == TS_CHEMDRIVER_STATE_TUBELIKE)
        break;

      return 0;

    case TS_CHEMDRIVER_STATE_TUBELIKE:
      {
        int ok = 0, next_beta = r+4, prev_beta = r-4;

        ok = ((r == last_beta)
            || (TS_LATTICE_SPIN(drv,next_beta,c) == TS_CHEMDRIVER_STATE_EMPTY)
            || (prev_beta < 1)
            || (TS_LATTICE_SPIN(drv,prev_beta,c) == TS_CHEMDRIVER_STATE_EMPTY));

        if (!ok) { return 0; }

        if (tsch_test_flags(drv, TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE))
          return 1;
      }

      break;

    default:
      return 0;
  }

  wr2 = r+2;
  wc2 = c+1;
  ts_lattice_wrap_coord((ts_lattice*)drv, &wr2,&wc2);

  return seam_reaction_allowed_by_seam_row(drv,r,c) 
    && seam_reaction_allowed_by_seam_row(drv,wr1,wc1)
    && seam_reaction_allowed_by_seam_row(drv,r+2,c)
    && seam_reaction_allowed_by_seam_row(drv,wr2,wc2);
}

static inline int sheet_reaction_allowed(ts_chem_driver* drv, int r, int c, unsigned int flags)
{
  int curr, nrow,ncol, i,j;
  int rp,rn;
  int r0;

  int curr_flag, adj_rows;
  curr_flag = (flags & SHEET_ALLOWED_SKIP_CURR) == 0;
  adj_rows  = (flags & SHEET_ALLOWED_ADJ_ROWS) > 0;

  /* Fast out */
  if (curr_flag && TS_LATTICE_SPIN(&drv->latt,r,c) != TS_CHEMDRIVER_STATE_TUBELIKE)
  {
    return 1;
  }

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;

  assert( ts_chem_driver_is_alpha_lateral(drv,r,c) );
  if (adj_rows)
  {
    rp = r - 4;
    rn = r + 6;
  }
  else
  {
    rp = r; 
    rn = r+2;
  }

  if (rp < 1 || rn > nrow-1) { return 1; }

  r0 = (c != ncol-1) ? r+0 : r+2;

  for (j=1; j < ncol; j+= 2) {
    for (i=rp; i <= rn; i+=2) {
      if (!curr_flag && (j == c) && (r0 == i || r0+2 == i))
      {
        continue;
      }

      if (TS_LATTICE_SPIN(drv,i,j) != TS_CHEMDRIVER_STATE_TUBELIKE)
      {
        return 1;
      }
    }
  }

  return 0;
}

static TSResult zero_connected_component(ts_chem_driver *drv, int *restrict components, int component_to_zero)
{
  size_t nrow, ncol;
  int *restrict cc;

  nrow = drv->latt.nrow;
  ncol = drv->latt.ncol;
  cc = components;

  for (size_t j=0; j < ncol; j++)
  {
    for (size_t i=0; i < nrow; i++)
    {
      if (cc[i] == component_to_zero)
      {
        RETURN_ON_ERROR( ts_lattice_mark_site_changed(&drv->latt, i, j) );
        TS_LATTICE_SPIN(drv,i,j) = TS_CHEMDRIVER_STATE_EMPTY;
      }
    }
    cc += nrow;
  }

  return TS_OK;
}

static MAYBE_UNUSED inline int is_dimer_curved(ts_chem_driver* drv, int ra, int ca)
{
  ts_lattice_wrap_coord((ts_lattice*)drv, &ra, &ca);

  if ((ra < 1) && (ra+3 >= (int)drv->latt.nrow))
  {
    return 0;
  }

  int mer1 = TS_LATTICE_SPIN(drv,ra+0,ca);
  int ifc  = TS_LATTICE_SPIN(drv,ra+1,ca);
  int mer2 = TS_LATTICE_SPIN(drv,ra+2,ca);

  return (
      ((mer1 == TS_CHEMDRIVER_STATE_CURVED) ||
       (mer1 == TS_CHEMDRIVER_STATE_CURVED_GDP)) &&
      (ifc == TS_CHEMDRIVER_STATE_CURVED) &&
      ((mer2 == TS_CHEMDRIVER_STATE_CURVED) ||
       (mer2 == TS_CHEMDRIVER_STATE_CURVED_GDP))
      );
}


static int polymerize_generate_configurations(ts_chem_driver *drv, int ra, int ca,
    int bonds[TSCH_POLYM_N_BOND_POSS])
{
  int nrow;
  int ra_left , ca_left,  dimer_on_left;
  int ra_right, ca_right, dimer_on_right;
  int nconfigs;

  nrow = drv->latt.nrow;

  ra_left  = ra; ca_left  = ca-2; ts_lattice_wrap_coord(&drv->latt, &ra_left,  &ca_left);
  ra_right = ra; ca_right = ca+2; ts_lattice_wrap_coord(&drv->latt, &ra_right, &ca_right);

  dimer_on_left  = (ra_left+2  < nrow) && (TS_LATTICE_SPIN(drv, ra_left, ca_left ) != 0);
  dimer_on_right = (ra_right+2 < nrow) && (TS_LATTICE_SPIN(drv, ra_right,ca_right) != 0);

  bonds[0] = TSCH_POLYM_LEFT_MONOMER  | 0 | 0;
  bonds[1] = TSCH_POLYM_RIGHT_MONOMER | 0 | 0;
  nconfigs = 2;

  if (tsch_test_flags(drv, TSCH_POLYM_ONLY_LONGITUDINAL))
  {
    return nconfigs;
  }

  if (dimer_on_left)
  {
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_TUBE  | 0;
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_SHEET | 0;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_TUBE  | 0;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_SHEET | 0;
  }

  if (dimer_on_right)
  {
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | 0 | TSCH_POLYM_RIGHT_TUBE;
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | 0 | TSCH_POLYM_RIGHT_SHEET;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | 0 | TSCH_POLYM_RIGHT_TUBE;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | 0 | TSCH_POLYM_RIGHT_SHEET;
  }

  if (dimer_on_left && dimer_on_right)
  {
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_SHEET | TSCH_POLYM_RIGHT_TUBE;
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_TUBE  | TSCH_POLYM_RIGHT_SHEET;
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_SHEET | TSCH_POLYM_RIGHT_SHEET;
    bonds[nconfigs++] = TSCH_POLYM_LEFT_MONOMER  | TSCH_POLYM_LEFT_TUBE  | TSCH_POLYM_RIGHT_TUBE;

    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_SHEET | TSCH_POLYM_RIGHT_TUBE;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_TUBE  | TSCH_POLYM_RIGHT_SHEET;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_SHEET | TSCH_POLYM_RIGHT_SHEET;
    bonds[nconfigs++] = TSCH_POLYM_RIGHT_MONOMER | TSCH_POLYM_LEFT_TUBE  | TSCH_POLYM_RIGHT_TUBE;
  }

  return nconfigs;
}

static TSResult polymerize_unpack_configuration(int b, int *merp, int *lbp, int *rbp)
{
  if      (b & TSCH_POLYM_LEFT_MONOMER)  { *merp = TS_CHEMDRIVER_STATE_BENT_LEFT;  }
  else if (b & TSCH_POLYM_RIGHT_MONOMER) { *merp = TS_CHEMDRIVER_STATE_BENT_RIGHT; }
  else {
    TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
        "invalid monomer configuration (b=%x)", b);
  }

  if      (b & TSCH_POLYM_LEFT_TUBE)     { *lbp = TS_CHEMDRIVER_STATE_TUBELIKE; }
  else if (b & TSCH_POLYM_LEFT_SHEET)    { *lbp = TS_CHEMDRIVER_STATE_SHEETLIKE; }
  else                        { *lbp = 0; }

  if      (b & TSCH_POLYM_RIGHT_TUBE)    { *rbp = TS_CHEMDRIVER_STATE_TUBELIKE; }
  else if (b & TSCH_POLYM_RIGHT_SHEET)   { *rbp = TS_CHEMDRIVER_STATE_SHEETLIKE; }
  else                        { *rbp = 0; }

  return TS_OK;
}

TSResult tsch_polymerize_unpack_configuration(int b, int *merp, int *lbp, int *rbp)
{
  RETURN_WITH_RESULT( polymerize_unpack_configuration(b, merp, lbp, rbp) );
}


static int find_lateral_bonds(ts_chem_driver* drv, const int rl, const int cl)
{
  int nr;
  int r1,c1, r2,c2;
  int lateral_bonds;
  int i, i1, i2;

  nr = drv->latt.nrow;

  r1 = rl; c1 = cl-1; ts_lattice_wrap_coord((ts_lattice*)drv, &r1, &c1);
  r2 = rl; c2 = cl+1; ts_lattice_wrap_coord((ts_lattice*)drv, &r2, &c2);

  lateral_bonds = 0;

  /* Toward the minus end */
  for(i=rl-1, i1=r1-1, i2=r2-1;
      (i >= 0) && (TS_LATTICE_SPIN(drv,i,cl) != TS_CHEMDRIVER_SITE_BLANK);
      i-=4, i1-=4, i2-=4)
  {
    if (((i1>=0) && (TS_LATTICE_SPIN(drv,i1,c1) != TS_CHEMDRIVER_SITE_BLANK)) ||
        ((i2>=0) && (TS_LATTICE_SPIN(drv,i2,c2) != TS_CHEMDRIVER_SITE_BLANK)))
    {
      lateral_bonds = lateral_bonds | TIP_MINUS;
      break;
    }
  }

  /* Toward the plus end */
  for(i=rl+1, i1=r1+1, i2=r2+1;
      (i < nr) && (TS_LATTICE_SPIN(drv,i,cl) != TS_CHEMDRIVER_SITE_BLANK);
      i+=4, i1+=4, i2+=4)
  {
    if (((i1<nr) && (TS_LATTICE_SPIN(drv,i1,c1) != TS_CHEMDRIVER_SITE_BLANK)) ||
        ((i2<nr) && (TS_LATTICE_SPIN(drv,i2,c2) != TS_CHEMDRIVER_SITE_BLANK)))
    {
      lateral_bonds = lateral_bonds | TIP_PLUS;
      break;
    }
  }

  return lateral_bonds;
}


static TSResult sheet_length_bias_hook(void* data, ts_lattice* latt, const ts_lattice_move* move)
{
  UNUSED_PARAM(data);
  UNUSED_PARAM(latt);
  UNUSED_PARAM(move);

  return TS_OK;
}

struct tsch_sheet_length_bias *tsch_sheet_length_bias_new(ts_chem_driver *driver, size_t npts)
{
  struct tsch_sheet_length_bias *bias;
  size_t i, size;
  TSResult err;

  size = sizeof(*bias) + npts * sizeof(bias->potential[0]);
  if (bias = ts_alloc(NO_ALLOCATOR, size), bias == NULL) {
    return NULL;
  }

  memset(bias, 0, size);

  bias->driver = driver;
  bias->npts = npts;
  bias->scale = 1.0;
  bias->warned_about_extrapolation = 0;

  for (i=0; i < npts; i++)
  {
    bias->potential[i] = 0.0;
  }

  err = ts_lattice_add_move_hook(&driver->latt, sheet_length_bias_hook, bias);
  if (!IS_OKAY(err)) {
    ts_free(NO_ALLOCATOR, bias);
    return NULL;
  }

  return bias;
}

void tsch_sheet_length_bias_delete(struct tsch_sheet_length_bias *hook)
{
  ts_free(NO_ALLOCATOR, hook);
}

double tsch_sheet_length_bias_energy(struct tsch_sheet_length_bias *bias, double sheet_length)
{
  size_t hi;
  double length_hi, delta, energy, scale;

  if (sheet_length < 0.0)
  {
    /* signal error as NaN */
    return (double)NAN;
  }

  if (scale = bias->scale, scale == 0.0)
  {
    return 0.0;
  }

  length_hi = ceil(sheet_length);
  hi = (size_t)length_hi;

  if (hi < bias->npts)
  {
    energy = bias->potential[hi];
    if (length_hi != sheet_length)
    {
      delta = length_hi - sheet_length;
      energy *= (1.0-delta);
      energy += delta*bias->potential[hi-1];
    }
  }
  else
  {
    /* point is past the end of the bias potential, so we use 3-point extrapolation */
    double x0, y0,y1,y2;

    x0 = bias->npts-3;
    y0 = bias->potential[bias->npts-3];
    y1 = bias->potential[bias->npts-2];
    y2 = bias->potential[bias->npts-1];

    energy = y0 * ( (sheet_length - x0 - 1.0)*(sheet_length - x0 - 2.0) )/(-1.0 * -2.0)
      + y1 * ( (sheet_length - x0)*(sheet_length - x0 - 2.0) )/(1.0 * -1.0)
      + y2 * ( (sheet_length - x0)*(sheet_length - x0 - 1.0) )/(2.0 * 1.0);

    if (!bias->warned_about_extrapolation)
    {
      bias->warned_about_extrapolation = 1;

      fprintf(stderr, "Warning!  3-point extrapolation! (%.3f,%.3f) last point is (%.3f,%.3f)\n",
          sheet_length,energy, x0+2.0, y2);
    }
  }

  return scale * energy;
}

double tsch_sheet_length(ts_chem_driver *drv)
{
  size_t ncol, nrow, i,j, maxtip;
  size_t sheet_length;

  ncol = drv->latt.ncol;
  nrow = drv->latt.nrow;

  /* this isn't a valid lattice */
  if (nrow <= 2) { return 0.0; }

  sheet_length = 0;
  maxtip = (nrow-1)/4;
  for (j=0; j < ncol-1; j += 2)
  {
    size_t tip;
    size_t straight;

    tip = maxtip;
    for (i=3+4; i < nrow-1; i += 4)
    {
      if (TS_LATTICE_SPIN(drv,i,j) == 0)
      {
        tip = i-4;
        break;
      }
    }

    straight = 3;
    for (i=tip; i >= 3; i -= 4)
    {
      if ((TS_LATTICE_SPIN(drv,i,j) == TS_CHEMDRIVER_STATE_STRAIGHT) ||
          (TS_LATTICE_SPIN(drv,i,j) == TS_CHEMDRIVER_STATE_STRAIGHT_GDP))
      {
        straight = i;
        break;
      }
    }

    sheet_length += (tip - straight)/4;
  }

  return (double)sheet_length / (ncol/2);
}

TSResult tsch_sheet_length_bias_energy_change(struct tsch_sheet_length_bias *bias, int i, int j, int st1, double *dE)
{
  ts_chem_driver *drv;
  int st0;
  double len0, len1;
  double en0, en1;
  enum TS_CHEMDRIVER_SITE_TYPE site_type;

  drv = bias->driver;
  if ((i < 0) || ((size_t)i >= drv->latt.nrow))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_INDEX, "row index %d is invalid", i);
  }

  if ((j < 0) || ((size_t)j >= drv->latt.ncol))
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_INDEX, "column index %d is invalid", j);
  }

  if (st1 < 0)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_VALUE, "new state must be non-negative (%d is invalid)", st1);
  }

  site_type = ts_chem_driver_site_type(drv,i,j);
  if (site_type == TS_CHEMDRIVER_SITE_MONOMER)
  {
    st0 = TS_LATTICE_SPIN(drv, i,j);
    len0 = tsch_sheet_length(drv);

    TS_LATTICE_SPIN(drv,i,j) = st1;
    len1 = tsch_sheet_length(drv);

    TS_LATTICE_SPIN(drv,i,j) = st0;
  }
  else if (site_type == TS_CHEMDRIVER_SITE_LONG_IFC)
  {
    if (st1 == PSEUDO_K_ON_BENT_LEFT)
    {
      st0 = TS_CHEMDRIVER_STATE_BENT_LEFT;
    }
    else if (st1 == PSEUDO_K_ON_BENT_RIGHT)
    {
      st0 = TS_CHEMDRIVER_STATE_BENT_RIGHT;
    }
    else
    {
      *dE = 0.0;
      return TS_OK;
    }

    len0 = tsch_sheet_length(drv);

    TS_LATTICE_SPIN(drv, i+0,j) = TS_CHEMDRIVER_STATE_BENT_IFC;
    TS_LATTICE_SPIN(drv, i+1,j) = st0;
    TS_LATTICE_SPIN(drv, i+2,j) = TS_CHEMDRIVER_STATE_BENT_IFC;
    TS_LATTICE_SPIN(drv, i+3,j) = st0;

    len1 = tsch_sheet_length(drv);
    assert(len1 == len0 + (1.0/ts_chem_driver_num_protofilaments(drv)));

    TS_LATTICE_SPIN(drv, i+0,j) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv, i+1,j) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv, i+2,j) = TS_CHEMDRIVER_STATE_EMPTY;
    TS_LATTICE_SPIN(drv, i+3,j) = TS_CHEMDRIVER_STATE_EMPTY;
  }
  else
  {
    *dE = 0.0;
    return TS_OK;
  }

  if (len0 == len1)
  {
    *dE = 0.0;
    return TS_OK;
  }

  en0 = tsch_sheet_length_bias_energy(bias, len0);
  en1 = tsch_sheet_length_bias_energy(bias, len1);

  if (!isfinite(en0) || !isfinite(en1))
  {
    TS_RAISE_ERROR_MESG(TS_ERR_INVALID_RETURN, "sheet-length bias energies are not finite");
  }

  *dE = en1 - en0;
  return TS_OK;
}



