#ifndef TS_CHEM_DRIVER_H
#define TS_CHEM_DRIVER_H

#include <stdlib.h>
#include <math.h>

#include "ts_errors.h"
#include "ts_common.h"

#include "ts_lattice.h"

#define DEBUG_ENERGY 0

enum TS_CHEMDRIVER_EVENT_TYPE
{
  TS_CHEMDRIVER_EVT_NONE = 0    ,
  TS_CHEMDRIVER_EVT_K_ON  = 1   ,
  TS_CHEMDRIVER_EVT_K_OFF       ,
  TS_CHEMDRIVER_EVT_MONOMER     ,
  TS_CHEMDRIVER_EVT_LONGITUDINAL,
  TS_CHEMDRIVER_EVT_LATERAL     ,
  TS_CHEMDRIVER_EVT_LATERAL_SEAM,
  TS_CHEMDRIVER_EVT_HYDROLYSIS  ,
  TS_CHEMDRIVER_EVT_POLYMERIZE  ,

  TS_CHEMDRIVER_EVT_MAX
};

extern const char* ts_chemdriver_event_names[TS_CHEMDRIVER_EVT_MAX];


enum TS_CHEMDRIVER_BARRIER_MODEL
{
  TS_CHEMDRIVER_BARRIER_NONE         = 0,
  TS_CHEMDRIVER_BARRIER_METROPOLIS   = 1,
  TS_CHEMDRIVER_BARRIER_KAWASAKI     = 2,
  TS_CHEMDRIVER_BARRIER_PROPORTIONAL = 3,

  // This is largely for testing...
  TS_CHEMDRIVER_BARRIER_INFINITE     = 4,

  TS_CHEMDRIVER_BARRIER_MAX
};

// Some barrier aliases to better explain "none" and "infinite" barriers 
#define TS_CHEMDRIVER_BARRIER_ALWAYS_ACCEPT TS_CHEMDRIVER_BARRIER_NONE
#define TS_CHEMDRIVER_BARRIER_ALWAYS_REJECT TS_CHEMDRIVER_BARRIER_INFINITE


extern const char* ts_chemdriver_barrier_model_names[TS_CHEMDRIVER_BARRIER_MAX];

enum TS_CHEMDRIVER_SITE_TYPE
{
  TS_CHEMDRIVER_SITE_BLANK    = 0,

  TS_CHEMDRIVER_SITE_MONOMER  = 1,
  TS_CHEMDRIVER_SITE_LONG_IFC = 2,
  TS_CHEMDRIVER_SITE_LAT_IFC  = 3,

  TS_CHEMDRIVER_NUM_TYPES,
  TS_CHEMDRIVER_SITE_ANY,
};

extern const char* ts_chemdriver_site_type_names[TS_CHEMDRIVER_NUM_TYPES];


#define TS_CHEMDRIVER_STATE_GTP_MIN TS_CHEMDRIVER_STATE_CURVED
#define TS_CHEMDRIVER_STATE_GTP_MAX TS_CHEMDRIVER_STATE_STRAIGHT
#define TS_CHEMDRIVER_STATE_GDP_MIN TS_CHEMDRIVER_STATE_CURVED_GDP
#define TS_CHEMDRIVER_STATE_GDP_MAX TS_CHEMDRIVER_STATE_STRAIGHT_GDP
enum TS_CHEMDRIVER_SITE_STATE
{
  TS_CHEMDRIVER_STATE_EMPTY          =  0,
  TS_CHEMDRIVER_STATE_CURVED         =  1,
  TS_CHEMDRIVER_STATE_BENT_IFC       =  2,
  TS_CHEMDRIVER_STATE_BENT_LEFT      =  3,
  TS_CHEMDRIVER_STATE_BENT_RIGHT     =  4,
  TS_CHEMDRIVER_STATE_STRAIGHT       =  5,

  TS_CHEMDRIVER_STATE_TUBELIKE       =  6,
  TS_CHEMDRIVER_STATE_SHEETLIKE      =  7,

  TS_CHEMDRIVER_STATE_CURVED_GDP     =  8,
  TS_CHEMDRIVER_STATE_BENT_LEFT_GDP  =  9,
  TS_CHEMDRIVER_STATE_BENT_RIGHT_GDP = 10,
  TS_CHEMDRIVER_STATE_STRAIGHT_GDP   = 11,

  TS_CHEMDRIVER_STATE_MAX
};

static const signed char tsch_reduce_lat_states[TS_CHEMDRIVER_STATE_MAX] = {
  -1,
  -1,
  -1,
  -1,
  -1,
  -1,

   0,
   1,

  -1,
  -1,
  -1,
  -1,
};

static const signed char tsch_reduce_long_states[TS_CHEMDRIVER_STATE_MAX] = {
  -1,
   0,
   1,
  -1,
  -1,
   2,

  -1,
  -1,

  -1,
  -1,
  -1,
  -1,
};

static const signed char tsch_reduce_mer_states[TS_CHEMDRIVER_STATE_MAX] = {
  -1,
   0,
  -1,
   1,
   2,
   3,

  -1,
  -1,

   4,
   5,
   6,
   7,
};

extern const char ts_chem_driver_site_codes[];

static const enum TS_CHEMDRIVER_SITE_STATE monomer_states_all[] = {
  TS_CHEMDRIVER_STATE_EMPTY,
  TS_CHEMDRIVER_STATE_BENT_LEFT,
  TS_CHEMDRIVER_STATE_BENT_RIGHT,
  TS_CHEMDRIVER_STATE_CURVED,
  TS_CHEMDRIVER_STATE_STRAIGHT,

  TS_CHEMDRIVER_STATE_BENT_LEFT_GDP,
  TS_CHEMDRIVER_STATE_BENT_RIGHT_GDP,
  TS_CHEMDRIVER_STATE_CURVED_GDP,
  TS_CHEMDRIVER_STATE_STRAIGHT_GDP
};

static const enum TS_CHEMDRIVER_SITE_STATE monomer_states_gtp[] = {
  TS_CHEMDRIVER_STATE_EMPTY,
  TS_CHEMDRIVER_STATE_BENT_LEFT,
  TS_CHEMDRIVER_STATE_BENT_RIGHT,
  TS_CHEMDRIVER_STATE_CURVED,
  TS_CHEMDRIVER_STATE_STRAIGHT,
};

static const enum TS_CHEMDRIVER_SITE_STATE monomer_states_gdp[] = {
  TS_CHEMDRIVER_STATE_BENT_LEFT_GDP,
  TS_CHEMDRIVER_STATE_BENT_RIGHT_GDP,
  TS_CHEMDRIVER_STATE_CURVED_GDP,
  TS_CHEMDRIVER_STATE_STRAIGHT_GDP
};

static const enum TS_CHEMDRIVER_SITE_STATE longitudinal_states[] = {
  TS_CHEMDRIVER_STATE_EMPTY,
  TS_CHEMDRIVER_STATE_BENT_IFC,
  TS_CHEMDRIVER_STATE_CURVED,
  TS_CHEMDRIVER_STATE_STRAIGHT
};

static const enum TS_CHEMDRIVER_SITE_STATE lateral_states[] = {
  TS_CHEMDRIVER_STATE_EMPTY,
  TS_CHEMDRIVER_STATE_SHEETLIKE,
  TS_CHEMDRIVER_STATE_TUBELIKE,
};

#define NOT_EMPTY(l) (l[0] == TS_CHEMDRIVER_STATE_EMPTY ? 1 : 0)
#define NUM_CONST(l) (sizeof(l)/sizeof(l[0]))

#define TSCH_NSTATES_MONOMER  (ARRAYLEN(monomer_states_all)-1)
#define TSCH_NSTATES_LONG     (ARRAYLEN(longitudinal_states)-1)
#define TSCH_NSTATES_LAT      (ARRAYLEN(lateral_states)-1)

extern int tsch_debug_updates;

enum { TSCH_MAX_DISSOC_BARRIERS   = 16 };
enum { TSCH_MAX_COUPLING_ENERGIES = 8 };

enum { TS_CHEMDRIVER_MAX_ROW_COUPLING = 4 };
enum { TS_CHEMDRIVER_MAX_COL_COUPLING = 2 };
enum { TS_CHEMDRIVER_MAX_ENERGY_COUPLING = 1 };

enum {
  PSEUDO_LATERAL_ROTATION  = TS_LATTICE_PSPIN(1),

  PSEUDO_LATERAL_ZERO,
  PSEUDO_LATERAL_SHEETLIKE,
  PSEUDO_LATERAL_TUBELIKE,
  PSEUDO_LATERAL_SEAM,

  PSEUDO_K_ON_BENT_LEFT,
  PSEUDO_K_ON_BENT_RIGHT,

  PSEUDO_K_ON_STRAIGHT_LEFT,
  PSEUDO_K_ON_STRAIGHT_RIGHT,

  PSEUDO_K_OFF,

  PSEUDO_FIND_ROW_RANGE,

  PSEUDO_ROLLBACK_IFCLOG,
};

#define TS_CHEMDRIVER_FORBIDDEN_ENERGY 100.0

#define TS_CHEMDRIVER_DEFAULT_PITCH -3

enum TSCH_FLAGS
{
  TSCH_REVERSIBLE_LATERAL_BOND_BREAKING = 0x001,

  TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS         = 0x0002,

  TSCH_POLYM_ONLY_LONGITUDINAL                = 0x0004,

  TSCH_POLYM_METROPOLIS_WEIGHTED              = 0x0008,

  TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE = 0x0010,

  TSCH_NO_MINUS_END_DEPOLYMERIZATION          = 0x0020,

  TSCH_MINUS_END_POLYMERIZATION               = 0x0040,

  TSCH_ENABLE_LAT_LONG_BENDING                = 0x0080,

  TSCH_RESTRICT_CURVED_LATERAL_BONDING        = 0x0100,

  TSCH_DISABLE_CCS_DEPOLYMERIZATION           = 0x0200,

  TSCH_RESTRICT_SHEET_FLUCTS                  = 0x0400,

  TSCH_HYDROLYSIS_ONLY_AT_SSS                 = 0x0800,

  TSCH_DISALLOW_POLYM_AT_STRAIGHT_IFCS        = 0x1000,
};

STATIC_ASSERT(TS_CHEMDRIVER_STATE_MAX <= UCHAR_MAX);
typedef unsigned char tsch_state_t;

STATIC_ASSERT(TS_CHEMDRIVER_STATE_MAX < 16);
static inline uint32_t
tsch_pack_coupling(int lat1, int lat2, int long1, int long2)
{
  assert(lat1 >= 0 && lat1 < 16);
  assert(lat2 >= 0 && lat2 < 16);
  assert(long1 >= 0 && long1 < 16);
  assert(long2 >= 0 && long2 < 16);

  if (lat1  > lat2 ) { int sw = lat1 ; lat1  = lat2 ; lat2  = sw; }
  if (long1 > long2) { int sw = long1; long1 = long2; long2 = sw; }

  return lat1<<12 | lat2<<8 | long1<<4 | long2;
}

static inline void 
tsch_unpack_coupling(uint32_t packed, int terms[4])
{
  terms[0] = (packed >> 12) & 0xf;
  terms[1] = (packed >>  8) & 0xf;
  terms[2] = (packed >>  4) & 0xf;
  terms[3] = (packed      ) & 0xf;
}

struct tsch_populations
{
  size_t monomer_pop[TS_CHEMDRIVER_STATE_MAX];
  size_t interface_pop[TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX];
  size_t bending_pop[TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX];
};

static inline uint32_t
tsch_pack_triple(int sp1, int sp2, int sp3)
{
  assert(sp1 >= 0 && sp1 < 16);
  assert(sp2 >= 0 && sp2 < 16);
  assert(sp3 >= 0 && sp3 < 16);

  return sp1 << 8 | sp2 << 4 | sp3;
}

static inline void
tsch_unpack_triple(uint32_t triple,
    int *restrict sp1, int *restrict sp2, int *restrict sp3)
{
  *sp1 = (triple >> 8) & 0xf;
  *sp2 = (triple >> 4) & 0xf;
  *sp3 = (triple >> 0) & 0xf;
}

struct tsch_ifclog_entry;

typedef struct ts_chem_driver ts_chem_driver;
struct ts_chem_driver
{
  ts_lattice latt;

  int* connected_components_scratch;

  int pitch_monomers;
  int num_coupling;
  int num_dissociation;

  double rate_prefactors[TS_CHEMDRIVER_EVT_MAX];
  double monomer_energies[TS_CHEMDRIVER_STATE_MAX];
  double interface_energies[TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX][TS_CHEMDRIVER_STATE_MAX];

  char coupling_lookup[TSCH_NSTATES_LAT][TSCH_NSTATES_LAT][TSCH_NSTATES_LONG][TSCH_NSTATES_LONG];


  int coupling_edge_dist;
  uint32_t coupling_terms[TSCH_MAX_COUPLING_ENERGIES];
  double coupling_energies[TSCH_MAX_COUPLING_ENERGIES][2];

  uint32_t dissociation_barrier_terms[TSCH_MAX_DISSOC_BARRIERS];
  double dissociation_barrier_energies[TSCH_MAX_DISSOC_BARRIERS];

  struct {
    struct tsch_ifclog_entry *entries;
    size_t n;
    size_t capa;
  } ifclog;

  int *row_range;

  uint32_t flags;

  double (*aux_energy)(void *, ts_chem_driver *);
  void *aux_energy_data;

  int seam_allowed_thresh;

  enum TS_CHEMDRIVER_BARRIER_MODEL barrier_model;
  double barrier_param;

  double dt;

  double t_max;

  double forbidden_energy;
};

static inline void tsch_set_flags(ts_chem_driver *driver, uint32_t flags)
{
  driver->flags = driver->flags | flags;
}

static inline void tsch_unset_flags(ts_chem_driver *driver, uint32_t flags)
{
  driver->flags = driver->flags & ~flags;
}

static inline int tsch_test_flags(ts_chem_driver *driver, uint32_t flags)
{
  return (driver->flags & flags) == flags;
}

static inline int
ts_chem_driver_pitch_rows(int pitch_mers) { return 2*pitch_mers; }

ts_chem_driver *ts_chem_driver_new(size_t nrow, size_t ncol, TSResult *errp);

ts_chem_driver *ts_chem_driver_retain(ts_chem_driver *driver);
void ts_chem_driver_release(ts_chem_driver *driver);

TSResult ts_chem_driver_reset(ts_chem_driver *drv);

static inline enum TS_CHEMDRIVER_SITE_TYPE ts_chem_driver_site_type(const ts_chem_driver *drv, size_t r, size_t c)
{
  if ((r==0) || (r == drv->latt.nrow-1)) return TS_CHEMDRIVER_SITE_BLANK;

  return IS_ODD(r)
    ? ( IS_ODD(c) ? TS_CHEMDRIVER_SITE_LAT_IFC : TS_CHEMDRIVER_SITE_MONOMER  )
    : ( IS_ODD(c) ? TS_CHEMDRIVER_SITE_BLANK   : TS_CHEMDRIVER_SITE_LONG_IFC )
    ;
}

static inline int
tsch_is_valid_site(const ts_chem_driver *drv, int i, int j)
{
  return (i >= 1) && ((size_t)i < drv->latt.nrow-1) && (j >= 1) && ((size_t)j < drv->latt.ncol-1);
}

static inline int ts_chem_driver_is_alpha_monomer(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_EVEN(c) && IS_MULTIPLE((r-1), 4);
}

static inline int ts_chem_driver_is_beta_monomer(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_EVEN(c) && IS_MULTIPLE((r+1), 4);
}

static inline int ts_chem_driver_is_seam(const ts_chem_driver* drv, int r, int c)
{
  UNUSED_PARAM(r);
  return (c == (int)drv->latt.ncol-1);
}

static inline int tsch_is_straight(int spin)
{
  return (spin == TS_CHEMDRIVER_STATE_STRAIGHT) || (spin == TS_CHEMDRIVER_STATE_STRAIGHT_GDP);
}

static inline int tsch_is_curved(int spin)
{
  return (spin == TS_CHEMDRIVER_STATE_CURVED) || (spin == TS_CHEMDRIVER_STATE_CURVED_GDP);
}

static inline int tsch_is_lateral_bond(int spin)
{
  return spin == TS_CHEMDRIVER_STATE_EMPTY || 
    spin == TS_CHEMDRIVER_STATE_SHEETLIKE  || 
    spin == TS_CHEMDRIVER_STATE_TUBELIKE;
}

static inline int tsch_is_longitudinal_bond(int spin)
{
  return spin == TS_CHEMDRIVER_STATE_EMPTY || 
    spin == TS_CHEMDRIVER_STATE_CURVED     || 
    spin == TS_CHEMDRIVER_STATE_BENT_IFC   ||
    spin == TS_CHEMDRIVER_STATE_STRAIGHT;
}

static inline int tsch_is_monomer_state(int spin)
{
  return spin == TS_CHEMDRIVER_STATE_EMPTY    || 
    spin == TS_CHEMDRIVER_STATE_CURVED        || 
    spin == TS_CHEMDRIVER_STATE_BENT_LEFT     ||
    spin == TS_CHEMDRIVER_STATE_BENT_RIGHT    ||
    spin == TS_CHEMDRIVER_STATE_STRAIGHT      ||
    spin == TS_CHEMDRIVER_STATE_CURVED_GDP    || 
    spin == TS_CHEMDRIVER_STATE_BENT_LEFT_GDP ||
    spin == TS_CHEMDRIVER_STATE_BENT_RIGHT_GDP||
    spin == TS_CHEMDRIVER_STATE_STRAIGHT_GDP;
}

static inline int tsch_in_state_group(int spin, const enum TS_CHEMDRIVER_SITE_STATE states[], size_t n)
{
  size_t i;
  for (i=0; i < n; i++) { if (spin == (int)states[i]) return 1; }
  return 0;
}
#define TSCH_IN_STATE_GROUP(sp, grp) ( tsch_in_state_group((sp), (grp), sizeof(grp)/sizeof((grp)[0])) )

static inline size_t
tsch_monomer_index(const ts_chem_driver* drv, size_t pf, size_t mer)
{
  return ts_lattice_index(&drv->latt, 1 + 2*mer, pf);
}

static inline size_t 
tsch_dimer_index(const ts_chem_driver* drv, size_t pf, size_t dimer)
{
  return ts_lattice_index(&drv->latt, 1 + 4*dimer, pf);
}

static inline int
ts_chem_driver_is_lateral(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_ODD(c) && IS_ODD(r);
}

static inline int
ts_chem_driver_is_longitudinal(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_EVEN(c) && IS_EVEN(r);
}

static inline int
ts_chem_driver_is_blank(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_ODD(c) && IS_EVEN(r);
}

static inline int ts_chem_driver_is_alpha_lateral(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_ODD(c) && IS_MULTIPLE((r-1), 4);
}

static inline int ts_chem_driver_is_beta_lateral(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_ODD(c) && IS_MULTIPLE((r+1), 4);
}


static inline int ts_chem_driver_is_interdimer_longitudinal(const ts_chem_driver *drv, int r, int c)
{
  UNUSED_PARAM(drv);
  return IS_EVEN(c) && IS_MULTIPLE(r,4);
}

static inline int ts_chem_driver_num_dimers(const ts_chem_driver* drv)
{
  return (drv->latt.nrow - 1)/4;
}

static inline int ts_chem_driver_num_protofilaments(const ts_chem_driver* drv)
{
  return drv->latt.ncol / 2;
}

static inline int ts_chem_driver_alpha_row(const ts_chem_driver* drv, int alpha)
{
  UNUSED_PARAM(drv);
  return 1+4*alpha;
}

static inline int ts_chem_driver_beta_row(const ts_chem_driver* drv, int beta)
{
  UNUSED_PARAM(drv);
  return 3+4*beta;
}

static inline double ts_chem_driver_prob_acceptance_metropolis(const ts_chem_driver* drv, double dE)
{
  dE *= drv->latt.inv_temp;
  return dE <= 0 ? 1.0 : exp(-dE);
}

static inline double ts_chem_driver_prob_acceptance_kawasaki(const ts_chem_driver* drv, double dE)
{
  dE *= drv->latt.inv_temp;
  return 1.0/(1.0+exp(dE));
}

double ts_chem_driver_prob_acceptance(const ts_chem_driver* drv, enum TS_CHEMDRIVER_BARRIER_MODEL model, double dE);

TSResult ts_chem_driver_add_coupling_energy(ts_chem_driver *drv, int terms[4], double elo, double ehi);
int ts_chem_driver_find_coupling_energy(ts_chem_driver *drv, int terms[4]);

TSResult ts_chem_driver_add_dissociation_barrier(ts_chem_driver *drv, spin_t terms[3], double en);
int ts_chem_driver_find_dissociation_barrier(ts_chem_driver *drv, spin_t terms[3]);

TSResult ts_chem_driver_energy(ts_chem_driver *drv, double* energyp,
    struct tsch_populations *pops);

TSResult ts_chem_driver_energy_change(ts_chem_driver *drv, size_t r, size_t c, spin_t s, double *energyp);
TSResult ts_chem_driver_energy_change_lateral(ts_chem_driver *drv, size_t r, size_t c, spin_t s, double *energyp);

TSResult ts_chem_driver_generate_site_moves(ts_chem_driver *drv, size_t r, size_t c,
    int states[TS_CHEMDRIVER_STATE_MAX],
    double rates[TS_CHEMDRIVER_STATE_MAX],
    size_t* nmovesp);

TSResult ts_chem_driver_print_configuration(ts_chem_driver *drv, FILE* f);

TSResult ts_chem_driver_count_bonds(ts_chem_driver *drv,
    const int *restrict rv, const int *restrict cv,
    int *restrict nlongv, int *restrict nlatv, size_t n);

TSResult ts_chem_driver_update_row_range(ts_chem_driver *drv);

struct tsch_ifclog_entry {
  uint16_t ifc;
  uint16_t move;
  uint16_t rel_row;
  uint16_t col;
};

TSResult tsch_ifclog_resize(ts_chem_driver *drv, size_t new_capa);

static inline void
tsch_ifclog_clear(ts_chem_driver *drv)
{
  drv->ifclog.n = 0;
}

static inline void
tsch_ifclog_zero(ts_chem_driver *drv)
{
  size_t i,n;

  if (drv->ifclog.entries == NULL)
  {
    return;
  }

  for(i=0,n=drv->ifclog.capa; i < n; i++)
  {
    drv->ifclog.entries[i] = (struct tsch_ifclog_entry){0};
  }

  drv->ifclog.n = 0;
}

static inline int
tsch_ifclog_is_logging(const ts_chem_driver *drv)
{
  return drv->ifclog.entries != NULL && drv->ifclog.n <= drv->ifclog.capa;
}

static inline size_t
tsch_ifclog_nentries(const ts_chem_driver *drv)
{
  return (drv->ifclog.entries != NULL) ? drv->ifclog.n : 0;
}

static inline size_t
tsch_ifclog_capa(const ts_chem_driver *drv)
{
  return (drv->ifclog.entries != NULL) ? drv->ifclog.capa : 0;
}

static inline uint16_t
tsch_ifclog_pack_ifc(int sp0, int sp1, int sp2)
{
  assert(TS_CHEMDRIVER_STATE_MAX <= 0xf);
  assert(sp0 >= 0 && sp0 < TS_CHEMDRIVER_STATE_MAX);
  assert(sp1 >= 0 && sp1 < TS_CHEMDRIVER_STATE_MAX);
  assert(sp2 >= 0 && sp2 < TS_CHEMDRIVER_STATE_MAX);

  return sp0 | (sp1 << 4) | (sp2 << 8);
}

static inline int
tsch_ifclog_unpack_ifc(uint16_t ifc, int *sp0, int *sp1, int *sp2)
{
  if ((ifc & 0x8000) != 0) { return 0; }

  if (sp0) { *sp0 = ifc & 0xf; }
  if (sp1) { *sp1 = (ifc >> 4) & 0xf; }
  if (sp2) { *sp2 = (ifc >> 8) & 0xf; }

  return 1;
}

static inline struct tsch_ifclog_entry *
tsch_ifclog_at(ts_chem_driver *drv, size_t i)
{
  return i < drv->ifclog.n ? &drv->ifclog.entries[i] : NULL;
}

static inline uint16_t 
tsch_ifclog_encode_rrow(long dr, int minus)
{
  dr += (1<<14);

  /* saturate at 15 bits */
  if (dr < 0) { dr = 0; }
  if (dr > (1<<15)) { dr = 1<<15; }

  /* if minus, set high bit */
  if (minus) { dr = dr | 0x8000; }

  return dr;
}

static inline int
tsch_ifclog_decode_rrow(uint16_t rr, int *minusp)
{
  *minusp = rr & 0x8000;
  return (rr & ~0x8000) - (1<<14);
}

static inline size_t
tsch_ifclog_add(ts_chem_driver *drv, uint16_t ifc, int mv, int r, int c)
{
  struct tsch_ifclog_entry *entry;
  int dplus,dminus;
  size_t ind;

  if ((drv->ifclog.entries == NULL) || (drv->ifclog.n >= drv->ifclog.capa))
  {
    return ~(size_t)0;
  }

  assert(mv >= 0 || mv <= UINT16_MAX);
  assert(r >= 0 && (size_t)r < drv->latt.nrow && drv->latt.nrow <= UINT16_MAX);
  assert(c >= 0 && (size_t)c < drv->latt.ncol && drv->latt.ncol <= UINT16_MAX);

  ind = drv->ifclog.n++;
  entry = &drv->ifclog.entries[ind];
  entry->ifc = ifc;
  entry->move = mv;

  dminus = r - drv->row_range[2*c+0];
  dplus  = drv->row_range[2*c+1] - r;

  entry->rel_row = (abs(dminus) < abs(dplus))
    ? tsch_ifclog_encode_rrow(dminus, 1)
    : tsch_ifclog_encode_rrow(dplus, 0);
  entry->col = c;

  return ind;
}

enum { TSCH_IFCLOG_VERSION_1 = 1 };
#define TSCH_IFCLOG_VERSION_DEFAULT TSCH_IFCLOG_VERSION_1

#define TSCH_IFCLOG_FILE_MAGIC "TSCHIFCL"

TSResult tsch_ifclog_write(ts_chem_driver *drv, FILE *f);
TSResult tsch_ifclog_read(ts_chem_driver *drv, FILE *f);

TSResult ts_chem_driver_check_move_consistency(ts_chem_driver* drv,
    size_t *rv, size_t *cv, size_t nv);

enum {
  TSCH_POLYM_LEFT_TUBE    = 0x01, TSCH_POLYM_RIGHT_TUBE    = 0x10,
  TSCH_POLYM_LEFT_SHEET   = 0x02, TSCH_POLYM_RIGHT_SHEET   = 0x20,
  TSCH_POLYM_LEFT_MONOMER = 0x04, TSCH_POLYM_RIGHT_MONOMER = 0x40,

  TSCH_POLYM_N_BOND_POSS = 3*3*2,
};

struct tsch_polymerize_options
{
  double min_energy, sum_probs;
  double energies[TSCH_POLYM_N_BOND_POSS];
  double probs[TSCH_POLYM_N_BOND_POSS];

  int ra_left, ca_left, ra_right, ca_right;
  int nbonds;
  int bonds[TSCH_POLYM_N_BOND_POSS];
};

TSResult tsch_polymerize_unpack_configuration(
    int b, int *merp, int *lbp, int *rbp);


struct tsch_save_header {
  uint32_t version;
  uint32_t flags;
  uint64_t nsteps;
  double time;
  double energy;
};

struct tsch_snapshot {
  size_t num_sites;
  spin_t sites[];
};

enum { TSCH_STATE_VERSION_1 = 1 };
#define TSCH_STATE_VERSION_DEFAULT TSCH_STATE_VERSION_1

#define TSCH_STATE_FILE_MAGIC "TSCHSTFF"

enum {
  TSCH_STATE_FLAG_DELTA = 0x00000001,
};
#define TSCH_STATE_FLAG_NONE 0
#define TSCH_STATE_FLAG_DEFAULT TSCH_STATE_FLAG_NONE

TSResult tsch_write_state(ts_chem_driver *driver, FILE *outf);
TSResult tsch_read_state(ts_chem_driver *driver, FILE *inf, struct tsch_save_header *header);

TSResult tsch_read_header(FILE *inf, struct tsch_save_header *header, uint32_t *nrowp, uint32_t *ncolp);

struct tsch_snapshot *tsch_snapshot_new(ts_chem_driver *driver);
void tsch_snapshot_free(struct tsch_snapshot *snap);

TSResult tsch_snapshot_update(ts_chem_driver *driver, struct tsch_snapshot *snap);
TSResult tsch_snapshot_num_changes(ts_chem_driver *driver, struct tsch_snapshot *snap, size_t *nump);
TSResult tsch_snapshot_write_delta(ts_chem_driver *driver, const struct tsch_snapshot *snap, FILE *outf);
TSResult tsch_snapshot_read_delta(ts_chem_driver *driver, struct tsch_snapshot *snap, FILE *inf);


struct tsch_sheet_length_bias
{
  ts_chem_driver *driver;
  double scale;
  int warned_about_extrapolation;

  size_t npts;
  double potential[];
};

struct tsch_sheet_length_bias *tsch_sheet_length_bias_new(ts_chem_driver *driver, size_t npts);
void tsch_sheet_length_bias_delete(struct tsch_sheet_length_bias *bias);

double tsch_sheet_length(ts_chem_driver *drv);
double tsch_sheet_length_bias_energy(struct tsch_sheet_length_bias *bias, double sheet_length);
TSResult tsch_sheet_length_bias_energy_change(struct tsch_sheet_length_bias *bias,
    int i, int j, int st1, double *dE);


struct tsch_fast_sheet_model
{
  ts_chem_driver *driver;

  double scale;  

  ts_chem_driver *minus_copy;

  struct ts_lattice_change_list saved_changes;

  double *restrict relax_coeff;
  double *restrict no_tube_strain;
  double *restrict left_tube_strain;
  double *restrict right_tube_strain;
  double *restrict both_tube_strain;
  double *restrict dE;

  double total_strain;

  double ifc_model[6];
  double existing_model[3];

  size_t num_hook_calls;
  size_t num_hook_accepts;
  size_t num_hook_rejects;

  double sheet_fluct_penalty;
  int sheet_fluct_dist;

  int plus_minrow, plus_maxrow;
  int minus_minrow, minus_maxrow;

  int n0,n1;

  int work[];
};

#define TSCH_WKSP_FIELDS() \
  X(last_tubes , 2*npf)    \
  X(tips       , 2*npf)    \
  X(tube_bonds , 2*npf)    \
  X(sheet_bonds, 2*npf)    \
  X(segs0      , 2*npf)    \
  X(segs1      , 2*npf)    \
  X(map        , npf)      \
  X(oldifc     , npf)

struct tsch_fast_sheet_model_wksp {
  size_t npf;
#define X(field,nelts) int *restrict field;
  TSCH_WKSP_FIELDS()
#undef X
};

static inline size_t tsch_fast_sheet_model_nwork(ts_chem_driver *driver)
{
  int npf = ts_chem_driver_num_protofilaments(driver);
  size_t size = 0;

#define X(field,nelts) size += nelts;
  TSCH_WKSP_FIELDS()
#undef X

  return size*sizeof(int);
}

/* Sets up the workspace using the passed pointer to an array of
 * ints.
 */
static inline void tsch_fast_sheet_model_setup_wksp_extern(ts_chem_driver *driver,
    int *restrict work, struct tsch_fast_sheet_model_wksp *wksp)
{
  size_t npf;
  int *ptr = work;
  wksp->npf = npf = ts_chem_driver_num_protofilaments(driver);

#define X(field,nelts) wksp-> field = ptr; ptr += nelts;
  TSCH_WKSP_FIELDS()
#undef X
}

static inline void tsch_fast_sheet_model_setup_wksp(struct tsch_fast_sheet_model *model,
    struct tsch_fast_sheet_model_wksp *wksp)
{
  tsch_fast_sheet_model_setup_wksp_extern(model->driver, model->work, wksp);
}

struct tsch_fast_sheet_model *tsch_fast_sheet_model_new(ts_chem_driver *driver);
void tsch_fast_sheet_model_delete(struct tsch_fast_sheet_model *model);

void tsch_fast_sheet_model_last_tube_rows(struct tsch_fast_sheet_model *model,
    const int *restrict tips,
    struct tsch_fast_sheet_model_wksp *wksp);

void tsch_fast_sheet_model_tip_rows(struct tsch_fast_sheet_model *model, int *rows);

int tsch_fast_sheet_model_row_segments(struct tsch_fast_sheet_model *model, int row,
    const struct tsch_fast_sheet_model_wksp *wksp,
    int *restrict segs0);

TSResult tsch_fast_sheet_model_map_segments(struct tsch_fast_sheet_model *model,
    const int *restrict segs0, int n0, const int *restrict segs1, int n1, int *restrict map);

TSResult tsch_fast_sheet_model_find_new_ifcs(struct tsch_fast_sheet_model *model,
    const int *restrict segs0, int n0,
    const int *restrict segs1, int s1,
    const int *restrict map,
    int *restrict oldifc,
    int *restrict nexistp,
    const double *restrict dE,
    double *restrict strainp);

double tsch_fast_sheet_model_ifc_energy(struct tsch_fast_sheet_model *model, int ntube, int nsheet, int nexisting);

enum { TSCH_FSM_PLUS_END = 0, TSCH_FSM_MINUS_END = 1 };
TSResult tsch_fast_sheet_model_begin_rowscan(struct tsch_fast_sheet_model *model, int end);
TSResult tsch_fast_sheet_model_next_rows(struct tsch_fast_sheet_model *model,
    int row0, int n, int direc, double *strain, double *strain_err);

TSResult tsch_fast_sheet_model_energy(struct tsch_fast_sheet_model *model, double *energyp);
TSResult tsch_fast_sheet_model_energy_change(struct tsch_fast_sheet_model *model,
    int i, int j, int st1, double *dE);

TSResult tsch_fast_sheet_model_add_movehook(struct tsch_fast_sheet_model *model);

#endif /* TS_CHEM_DRIVER_H */
