#include "ts_common.h"

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>

struct ts_autopool_entry;
struct ts_autopool_header
{
  struct ts_autopool_entry* next;
};

#define POINTER_ALIGNMENT      0x10
#define ALIGN_PADDING  (POINTER_ALIGNMENT + (sizeof(struct ts_autopool_header) & (POINTER_ALIGNMENT-1)))
struct ts_autopool_entry
{
  struct ts_autopool_header header;
  char alignment_padding[ALIGN_PADDING];
  char data[1];
};

void* ts_autopool_alloc(struct ts_autopool* pool, size_t sz)
{
  struct ts_autopool_entry* entry = 
    ts_alloc(pool->allocator, 
        sizeof(struct ts_autopool_entry) + sz - 1);

  entry->header.next = pool->first;
  pool->first = entry;
  return entry->data+0;
}

void ts_autopool_release(struct ts_autopool* pool)
{
  struct ts_autopool_entry* entry = pool->first;
  while (entry)
  {
    struct ts_autopool_entry* next = entry->header.next;
    ts_free(pool->allocator, entry);
    entry = next;
  }
}

void ts_autopool_init(struct ts_autopool* pool, ts_allocator* allocator)
{
  pool->allocator = allocator;
  pool->first = NULL;

  pool->num_release_funcs = 0;
}

TSResult ts_autopool_on_release(struct ts_autopool* pool, void(*func)(void*), void* data)
{
  if (pool->num_release_funcs >= MAX_RELEASE_FUNCS)
  {
    return TS_NEW_ERROR3(TS_ERR_MAXIMUM_ALLOCATED,
        "maximum number of release functions allocated");
  }

  int ind = pool->num_release_funcs;
  pool->release_funcs[ ind ].func = func;
  pool->release_funcs[ ind ].data = data;
  pool->num_release_funcs++;

  return TS_OK;
}

void datavector_init(data_vector* vec, size_t elt_size, size_t capacity)
{
  vec->max_size = capacity;
  vec->elt_size = elt_size;

  vec->size = 0;

  vec->data = ts_alloc(NO_ALLOCATOR, elt_size*capacity);
}

void datavector_free(data_vector* vec)
{
  ts_free(NO_ALLOCATOR, vec->data );
  vec->size = 0;
  vec->max_size = 0;
  vec->elt_size = 0;
  vec->data = NULL;
}

static inline size_t maxsize(size_t a, size_t b)
{
  return a < b ? b : a;
}

static int dv_expand(data_vector* vec, size_t n)
{
  size_t newsize = vec->size + n;
  if (newsize > vec->max_size)
  {
    size_t delta = maxsize(newsize - vec->max_size, vec->size / 2);
    if (delta < 1) delta = 1;
    size_t capacity = vec->max_size + delta;
    void* new_data = ts_realloc(NO_ALLOCATOR, vec->data, vec->elt_size * capacity );
    if (new_data == NULL)
    {
      return 0;
    }

    vec->data = new_data;
    vec->max_size = capacity;
  }

  return 1;
}

void* datavector_new_item(data_vector* vec)
{
  if (!dv_expand(vec,1)) return NULL;
  void* p = (void*) (((char*)vec->data) + vec->elt_size*vec->size);
  vec->size++;
  return p;
}

int datavector_add(data_vector* vec, void* item)
{
  void* new_item = datavector_new_item(vec);
  if (new_item) {
    memcpy(new_item, item, vec->elt_size );
  }

  return (new_item != NULL);
}

int datavector_add_n(data_vector* vec, size_t n)
{
  if (!dv_expand(vec,n)) { return 0; }
  vec->size += n;
  return 1;
}

void datavector_sort(data_vector* vec, datavector_comparator func)
{
  qsort(vec->data, vec->size, vec->elt_size, func);
}

void datavector_clear(data_vector* vec)
{
  // rezero everything
  memset(vec->data, 0, vec->size * vec->elt_size);
  vec->size = 0;
}

#undef ts_free

void ts_alloc_log(const char* fmt, ...)
{
  va_list ap;
  static FILE* f = NULL;
  if (f == NULL)
  {
    f = fopen("ts_alloc.log","w");
    if (f == NULL) {
      fprintf(stderr,
          "\nERROR opening ts_alloc log: %s\n  using stderr instead\n",
          strerror(errno));
      f = stderr;
    }
  }

  va_start(ap, fmt);
  vfprintf(f, fmt, ap);
  fputs("\n",f);
  va_end(ap);

  fflush(f);
}

void ts_free(ts_allocator* A, void* p)
{
  ts_free_impl(A,p,__FILE__,__LINE__);
}


/*  Portable float/uint32_t and double/uint64_t conversions */

#define F64_SIGNBIT (1ULL << 63)
#define F64_CONST_EXP 53
#define F64_TO_SIGN(x) ((uint64_t)x << 63)
#define F64_MASK_BIT (1ULL << (F64_CONST_EXP-1))
#define F64_CONST ((double)(1ULL << F64_CONST_EXP))
#define F64_SIGNIFICAND_MASK ((1ULL << 52)-1)
#define F64_EXPN_OFFSET 1023

uint64_t portable_dbl_to_u64(double v0)
{
  int wh,sign,expn;
  uint64_t bits;
  double v = v0;

  sign = signbit(v);
  wh = fpclassify(v);
  if (wh == FP_NAN)
  {
    goto nan;
  }

  switch (wh)
  {
    case FP_NAN:
      goto nan;

    case FP_ZERO:
      bits = 0;
      goto set_sign;

    case FP_INFINITE:
      bits = 0x7ffULL << 52;
      goto set_sign;

    default:
    case FP_SUBNORMAL:
    case FP_NORMAL:
      break;
  }

  /* normals or subnormals */
  v = frexp(fabs(v),&expn);

  // -1 b/c frexp returns a fraction and the ieee 754 encoding
  // has a leading one.
  bits = (uint64_t)(v * F64_CONST);

  expn += (F64_EXPN_OFFSET-1);
  if (wh == FP_NORMAL)
  {
    // normal FP numbers have an implicit leading 1, which becomes
    // explicit in the conversion to u64.  This strips off leading
    // 1.
    bits = bits & ~F64_MASK_BIT;
    bits = bits | (((uint64_t)expn) << 52);
  }
  else
  {
    // exponent field is zero for subnormals, but we need to shift the
    // significand left
    bits = bits >> (-expn+1);
  }

set_sign:
  if (sign)
  {
    bits = bits | F64_SIGNBIT;
  }
  // print_binary_u64(">f64_1: ", bits);
  return bits;

nan:
  // can't retreive the payload, so this is good enough...
  return (0x7ffULL << 52) | (1ULL << 51) | 0xdeadbeef;
}

double portable_u64_to_dbl(uint64_t bits)
{
  int sign,expn;
  double v;
  char dig[21];

  sign = (bits & F64_SIGNBIT) == F64_SIGNBIT;
  if (sign) { bits = bits & ~F64_SIGNBIT; }

  if (bits == 0)
  {
    return sign ? -0.0 : 0.0;
  }

  expn = (bits >> 52);
  if (expn == 0x7ff)
  {
    goto nan_or_inf;
  }

  /* handle subnormals */
  if (expn == 0)
  {
    /* is there a faster way? */
    while((bits & (1ULL << 51)) == 0) {
      bits = bits << 1;
      expn -= 1;
    }
    expn++;
  }
  else
  {
    bits = bits & F64_SIGNIFICAND_MASK;
    bits = bits | F64_MASK_BIT;
  }

  expn -= (F64_EXPN_OFFSET-1);
  // XXX - can we combine these?  I'm afraid that
  // there might be an underflow case
  v = ldexp((double)bits, -F64_CONST_EXP);
  v = ldexp(v,expn);

  return sign ? -v : v;

nan_or_inf:
    bits = bits & F64_SIGNIFICAND_MASK;
    if (bits == 0)
    {
      return sign ? -HUGE_VAL : HUGE_VAL;
    }

    // NaN
    snprintf(dig,sizeof dig,"%" PRIu64, bits);
    return nan(dig);
}

static void u64_binary(uint64_t v0, char rep[static 65])
{
  int i;
  uint64_t v = v0;

  memset(rep,0,65);
  for(i=0; i < 64; i++)
  {
    rep[63-i] = (v&1) ? '1' : '0';
    v = v >> 1;
  }

  uint64_t t = 0;
  for (i=0; i < 64; i++)
  {
    t = t << 1;
    if (rep[i] == '1') { t = t | 1; }
  }

  if (t != v0)
  {
    printf("\n\nOH NOES!\n");
  }
}

int ts_common_test_f64_conversions(FILE *outf)
{
  size_t i,n,nerr;
  double vals[] = {
    0.0, -0.0,
    134143.322422,
    -134149.322422,
    23.1423, -23.1423,
    2.1423e-311, -2.1423e-311,
    HUGE_VAL, -HUGE_VAL, nan("3735928559"),
  };

  n = sizeof vals / sizeof vals[0];
  for(nerr=0,i=0; i < n; i++)
  {
    uint64_t bits, cvt_bits;
    double v,v2;
    union { uint64_t b; double d; } cvt;
    char binary[65] = {0};

    v = vals[i];
    bits = portable_dbl_to_u64(v);
    v2 = portable_u64_to_dbl(bits);

    cvt.d = v;
    cvt_bits = cvt.b;
    u64_binary(cvt_bits, binary);

    cvt.d = v2;
    // if everything is okay, skip the next part
    if (cvt_bits == bits && cvt.b == cvt_bits)
    {
      continue;
    }

    /* handle error case */
    nerr++;
    const char *mesg = "unknown error";
    if (cvt_bits != bits)
    {
      mesg = "ERROR in conversion to u64";
    }
    else if (cvt.b != cvt_bits)
    {
      mesg = "ERROR in conversion to back to f64";
    }

    if (outf)
    {
      fprintf(outf, "%s\n", mesg);
      fprintf(outf,"orig: %16e %a\n", v,v);
      fprintf(outf,"cvt : %16e %a\n", v2,v2);
      u64_binary(bits, binary);
      fprintf(outf,"                        "\
          "SEEEEEEEEEEEMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM\n");
      fprintf(outf,"orig: %16" PRIx64 "  %64s\n", bits, binary);
      u64_binary(cvt_bits, binary);
      fprintf(outf,"cmp : %16" PRIx64 "  %64s\n", cvt_bits, binary);

      u64_binary(cvt.b, binary);
      fprintf(outf,"cvt : %16" PRIx64 "  %64s\n", cvt.b, binary);
      fprintf(outf,"\n");
    }
  }

  if (outf)
  {
    fprintf(outf,"%zu tests, %zu errors\n", n, nerr);
  }

  return (nerr == 0);
}

#define F32_SIGNBIT (1ULL << 31)
#define F32_CONST_EXP 24
#define F32_MASK_BIT (1UL << (F32_CONST_EXP-1))
#define F32_CONST ((float)(1UL << F32_CONST_EXP))
#define F32_SIGNIFICAND_MASK (F32_MASK_BIT-1)
#define F32_EXPN_OFFSET 127

uint32_t portable_flt_to_u32(float v0)
{
  int wh,sign,expn;
  uint32_t bits;
  float v = v0;

  sign = signbit(v);
  wh = fpclassify(v);
  if (wh == FP_NAN)
  {
    goto nan;
  }

  switch (wh)
  {
    case FP_NAN:
      goto nan;

    case FP_ZERO:
      bits = 0;
      goto set_sign;

    case FP_INFINITE:
      bits = 0xffUL << 23;
      goto set_sign;

    default:
    case FP_SUBNORMAL:
    case FP_NORMAL:
      break;
  }

  /* normals or subnormals */
  v = frexpf(fabsf(v),&expn);

  // -1 b/c frexp returns a fraction and the ieee 754 encoding
  // has a leading one.
  bits = (uint32_t)(v * F32_CONST);

  expn += (F32_EXPN_OFFSET-1);
  // print_f32_header("      ");
  // print_binary_f32("v0  : ", v0);
  // print_binary_f32("v   : ", v);
  // print_binary_u32("exp : ", expn<<23);
  // print_binary_u32("bits: ", bits);

  if (wh == FP_NORMAL)
  {
    // normal FP numbers have an implicit leading 1, which becomes
    // explicit in the conversion to u32.  This strips off leading
    // 1.
    bits = bits & ~F32_MASK_BIT;
    bits = bits | (((uint32_t)expn) << 23);
  }
  else
  {
    // exponent field is zero for subnormals, but we need to shift the
    // significand left
    bits = bits >> (-expn+1);
  }

set_sign:
  if (sign)
  {
    bits = bits | F32_SIGNBIT;
  }
  // print_binary_u32(">f32_1: ", bits);
  return bits;

nan:
  // can't retreive the payload, so this is good enough...
  return (0xffULL << 23) | (1ULL << 22) | 0x0beef;
}

float portable_u32_to_flt(uint32_t bits)
{
  int sign,expn;
  double v;
  char dig[8];

  sign = (bits & F32_SIGNBIT) == F32_SIGNBIT;
  if (sign) { bits = bits & ~F32_SIGNBIT; }

  if (bits == 0)
  {
    return sign ? -0.0f : 0.0f;
  }

  expn = (bits >> 23);
  if (expn == 0xff)
  {
    goto nan_or_inf;
  }

  /* handle subnormals */
  if (expn == 0)
  {
    /* is there a faster way? */
    while((bits & (1UL << 22)) == 0) {
      bits = bits << 1;
      expn -= 1;
    }
    expn++;
  }
  else
  {
    bits = bits & F32_SIGNIFICAND_MASK;
    bits = bits | F32_MASK_BIT;
  }

  expn -= (F32_EXPN_OFFSET-1);
  // XXX - can we combine these?  I'm afraid that
  // there might be an underflow case
  v = ldexpf((double)bits, -F32_CONST_EXP);
  v = ldexpf(v,expn);

  return sign ? -v : v;

nan_or_inf:
    bits = bits & F32_SIGNIFICAND_MASK;
    if (bits == 0)
    {
      return sign ? -HUGE_VAL : HUGE_VAL;
    }

    // NaN
    snprintf(dig,sizeof dig,"%lu", (unsigned long)bits);
    return nanf(dig);
}

static int check_nan_bits(uint32_t v)
{
  uint32_t nan_bits = 0xffUL << 23;

  return ((v & nan_bits) == nan_bits) &&
    ((v & F32_SIGNIFICAND_MASK ) != 0);
}

static void u32_binary(uint32_t v0, char rep[static 33])
{
  int i;
  uint32_t v = v0;

  memset(rep,0,33);
  for(i=0; i < 32; i++)
  {
    rep[31-i] = (v&1) ? '1' : '0';
    v = v >> 1;
  }

  uint32_t t = 0;
  for (i=0; i < 32; i++)
  {
    t = t << 1;
    if (rep[i] == '1') { t = t | 1; }
  }

  if (t != v0)
  {
    printf("\n\nOH NOES!\n");
  }
}

int ts_common_test_f32_conversions(FILE *outf, int full)
{
  size_t i,n,nerr;
  float vals[] = {
    0.0f, -0.0f,
    134143.322422f,
    -134149.322422f,
    23.1423f, -23.1423f,
    2.1423e-40f, -2.1423e-40f,
    HUGE_VAL, -HUGE_VAL,
    nanf("48879"),
  };

  n = sizeof vals / sizeof vals[0];
  for(nerr=0,i=0; i < n; i++)
  {
    uint32_t bits, cvt_bits;
    float v,v2;
    union { uint32_t b; float f; } cvt;
    char binary[65] = {0};

    v = vals[i];
    bits = portable_flt_to_u32(v);
    v2 = portable_u32_to_flt(bits);

    cvt.f = v;
    cvt_bits = cvt.b;
    u32_binary(cvt_bits, binary);

    cvt.f = v2;
    // if everything is okay, skip the next part
    if (cvt_bits == bits && cvt.b == cvt_bits)
    {
      continue;
    }

    /* handle error case */
    nerr++;
    const char *mesg = "unknown error";
    if (cvt_bits != bits)
    {
      mesg = "ERROR in conversion to u32";
    }
    else if (cvt.b != cvt_bits)
    {
      mesg = "ERROR in conversion to back to f32";
    }

    if (outf)
    {
      fprintf(outf, "%s\n", mesg);
      fprintf(outf,"orig: %16e %a\n", v,v);
      fprintf(outf,"cvt : %16e %a\n", v2,v2);
      u32_binary(bits, binary);
      fprintf(outf,"                        "\
          "SEEEEEEEEMMMMMMMMMMMMMMMMMMMMMMM\n");
      fprintf(outf,"orig: %16lx  %32s\n", (unsigned long)bits, binary);
      u32_binary(cvt_bits, binary);
      fprintf(outf,"cmp : %16lx  %32s\n", (unsigned long)cvt_bits, binary);

      u32_binary(cvt.b, binary);
      fprintf(outf,"cvt : %16lx  %32s\n", (unsigned long)cvt.b, binary);
      fprintf(outf,"\n");
    }
  }

  if (outf)
  {
    fprintf(outf,"%zu tests, %zu errors\n", n, nerr);
  }

  if (nerr > 0)
  {
    return 0;
  }

  if (!full)
  {
    return 1;
  }

  /* full scan */

  /* for 32-bit floats, we can systematically evaluate all of them */
  if (outf)
  {
    fprintf(outf,"testing all 32-bit floats...\n");
    fflush(outf);
  }

  for (i=0; i < UINT32_MAX; i++)
  {
    uint32_t bits;
    union { uint32_t b; float f; } cvt;
    float f;

    bits = i;
    cvt.b = i;
    f = portable_u32_to_flt(bits);

    /* equality check doesn't work for NaNs, and we can't correctly
     * handle the payload (b/c C99 and C11 don't provide a portable way
     * to do that, so check for equality and then make sure that at
     * least one is not NaN
     */
    if (f != cvt.f && (!isnan(f) || !isnan(cvt.f)))
    {
      if (outf)
      {
        fprintf(outf,"disagreement with u32->f32 bits = 0x%08lx: %a and %a\n",
            (unsigned long)bits, f, cvt.f);
        return 0;
      }
    }

    bits = portable_flt_to_u32(cvt.f);
    if (bits != cvt.b)
    {
      /* if cvt.f is a nan, we can't compare the payloads directly.
       * f32_to_u32 doesn't set them reliably b/c C99 and C11 don't allow
       * us to access the payload.  Instead we check that the exponent is
       * 0xff and the significand is non-zero.
       */
      if (!isnan(cvt.f) || !check_nan_bits(bits) || !check_nan_bits(cvt.b))
      {
        if (outf)
        {
          fprintf(outf,"disagreement with f32->u32 float = %a: 0x%08lx and 0x%08lx\n",
              cvt.f, (unsigned long)bits, (unsigned long)cvt.b);
        }
        return 0;
      }
    }

    if ((i+1) % 100000000 == 0) {
      if (outf)
      {
        fprintf(outf,"%9ld / %9ld  (%6.2f%%)\n",
            (unsigned long)i+1,(unsigned long)UINT32_MAX,
            100.0*(double)(i+1) / UINT32_MAX);
        fflush(outf);
      }
    }
  }

  return 1;
}

