#ifndef TS_COMMON__H
#define TS_COMMON__H

#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include "ts_errors.h"
#include "ts_macros.h"

#define MODULE_NAME TS_COMMON

#define IS_ODD(x) ( ((x) % 2) == 1 )
#define IS_EVEN(x) ( ((x) % 2) == 0 )
#define IS_MULTIPLE(x,y) ( ((x) % (y)) == 0 )

#define ARRAYLEN(arr) ( (sizeof(arr))/(sizeof(arr[0])) )

static inline int is_between(int i, int i0, int i1) { return (i >= i0) && (i <= i1); }

static inline size_t lsb(size_t x) { return x & -x; }
static inline size_t strip_lsb(size_t x) { return x & (x-1); }
static inline size_t msb(size_t x)
{
  size_t c = x;
  while (x > 0)
  {
    c = x & -x;
    x -= c;
  }

  return c;
}

#define DATAVECTOR(type) data_vector
typedef struct data_vector data_vector;
struct data_vector
{
  size_t size;
  size_t max_size;
  size_t elt_size;
  void* data;
};

#define DATAVECTOR_INIT(type,vec,capacity) (datavector_init( &(vec), sizeof(type), (capacity) ))
#define DATAVECTOR_BEGIN(type,vec)         ((type*)((vec)->data))
#define DATAVECTOR_AT(type,vec,idx)        (((type*)((vec)->data))+idx)
#define DATAVECTOR_INDEX_OF(type,vec,elt)  (elt - DATAVECTOR_BEGIN(type,vec))
#define DATAVECTOR_END(type,vec)           DATAVECTOR_AT(type,vec,(vec)->size)
#define DATAVECTOR_FOREACH(type,var,vec) \
  for(type * var = (vec)->data; var != ((type *) ((vec)->data) + (vec)->size); ++var)

void datavector_init(data_vector* vec, size_t elt_size, size_t capacity);
void datavector_free(data_vector* vec);
int datavector_add(data_vector* vec, void* item);
int datavector_add_n(data_vector* vec, size_t n);
static inline int datavector_contains(data_vector* vec, void* p) {
  const char* vp = vec->data;
  const char* ip = (char*)p;
  size_t extent = vec->size*vec->elt_size;
  return (ip >= vp) && (ip < vp + extent);
}

void* datavector_new_item(data_vector* vec);
typedef int (*datavector_comparator)(const void* a, const void* b);
void datavector_sort(data_vector* vec, datavector_comparator func);
static inline void* datavector_at(data_vector* vec, size_t ind) { return (void*)( ((char*)vec->data) + ind*(vec->elt_size) ); }
void datavector_clear(data_vector* vec);

static inline int signof(ptrdiff_t i) { 
  if (i > 0) return  1;
  if (i < 0) return -1;
  return 0;
}

static inline double max_double(double a, double b) { return a < b ? b : a; }

static inline uint16_t swap_byteorder_16(uint16_t v) { return (v << 8) | (v >> 8); }
static inline uint32_t swap_byteorder_32(uint32_t v) { return ((uint32_t) (swap_byteorder_16(v) << 16)) | swap_byteorder_16(v>>16); }
static inline uint64_t swap_byteorder_64(uint64_t v) { return ((uint64_t) (((uint64_t)swap_byteorder_32(v)) << 32)) | swap_byteorder_32(v>>32); }

#ifdef ENDIANNESS_IS_BIG
static inline uint64_t u64_to_be(uint64_t v) { return v; }
static inline uint32_t u32_to_be(uint32_t v) { return v; }
static inline uint16_t u16_to_be(uint16_t v) { return v; }

static inline uint64_t u64_from_be(uint64_t v) { return v; }
static inline uint32_t u32_from_be(uint32_t v) { return v; }
static inline uint16_t u16_from_be(uint16_t v) { return v; }
#else /* !ENDIANNESS_IS_BIG --> endianness is little */
static inline uint64_t u64_to_be(uint64_t v) { return swap_byteorder_64(v); }
static inline uint32_t u32_to_be(uint32_t v) { return swap_byteorder_32(v); }
static inline uint16_t u16_to_be(uint16_t v) { return swap_byteorder_16(v); }

static inline uint64_t u64_from_be(uint64_t v) { return swap_byteorder_64(v); }
static inline uint32_t u32_from_be(uint32_t v) { return swap_byteorder_32(v); }
static inline uint16_t u16_from_be(uint16_t v) { return swap_byteorder_16(v); }
#endif

/* portable conversions.  these may be slow. */
int ts_common_test_f64_conversions(FILE *outf);
int ts_common_test_f32_conversions(FILE *outf, int full);

uint64_t portable_dbl_to_u64(double v);
uint32_t portable_flt_to_u32(float  v);
double   portable_u64_to_dbl(uint64_t v);
float    portable_u32_to_flt(uint32_t v);

#ifdef FLOAT_SAME_ENDIAN
ASSERT_SIZE_EQ(double,uint64_t);
#define CVT(tfrom,tto,v) \
  do { union { tfrom f; tto t; } _; _.f = (v); return _u.t } while(0)
static inline uint64_t dbl_to_u64(double v) { CVT(dbl,uint64_t,v); }
static inline uint32_t flt_to_u32(float  v) { CVT(flt,uint32_t,v); }

static inline double u64_to_dbl(uint64_t v) { CVT(uint64_t,dbl,v); }
static inline float  u32_to_flt(uint32_t v) { CVT(uint32_t,flt,v); }
#else
static inline uint64_t dbl_to_u64(double v) { return portable_dbl_to_u64(v); }
static inline uint32_t flt_to_u32(float  v) { return portable_flt_to_u32(v); }

static inline double u64_to_dbl(uint64_t v) { return portable_u64_to_dbl(v); }
static inline float  u32_to_flt(uint32_t v) { return portable_u32_to_flt(v); }
#endif

static inline uint64_t f64_to_be(double v) { return u64_to_be( dbl_to_u64(v) ); }
static inline uint32_t f32_to_be(float  v) { return u32_to_be( flt_to_u32(v) ); }

static inline double f64_from_be(uint64_t v) { return u64_to_dbl( u64_from_be(v) ); }
static inline float  f32_from_be(uint32_t v) { return u32_to_flt( u32_from_be(v) ); }

static inline double swap_dbl_bytes(double v) { return u64_to_dbl( swap_byteorder_64( dbl_to_u64(v) ) ); }
static inline float  swap_flt_bytes(float v)  { return u32_to_flt( swap_byteorder_32( flt_to_u32(v) ) ); }

#define NO_ALLOCATOR (NULL)
#define LIBC_ALLOCATOR (NULL)
typedef struct _ts_allocator ts_allocator;
// extern const ts_allocator* NO_ALLOCATOR;

/* used when a function pointer is needed for ts_free(A,p) */
static inline void ts_free_impl(ts_allocator* A, void* p, const char* fname, int line);
void ts_free(ts_allocator* A, void* p);
void ts_alloc_log(const char* fmt, ...);

#define ts_alloc(A,n)     (ts_alloc_impl(A,n,__FILE__,__LINE__))
#define ts_realloc(A,p,n) (ts_realloc_impl(A,p,n,__FILE__,__LINE__))
#define ts_calloc(A,n,sz) (ts_calloc_impl(A,n,sz,__FILE__,__LINE__))
#define ts_free(A,p)      (ts_free_impl(A,p,__FILE__,__LINE__))

#define TS_ALLOC_DEBUG 0
#if TS_ALLOC_DEBUG
#  define TS_ALLOC_LOG(fname,line,fmt,...) do {             \
    ts_alloc_log("[%s:%d] " fmt, fname, line, __VA_ARGS__); \
  } while(0)
#else
#  define TS_ALLOC_LOG(fname,line,fmt,...) do {} while(0)
#endif /* TS_ALLOC_DEBUG */

static inline void* ts_alloc_impl(ts_allocator* A, size_t n, const char* fname, int line)
{
  UNUSED_PARAM(A);
  UNUSED_PARAM(fname);
  UNUSED_PARAM(line);
  TS_ALLOC_LOG(fname,line,"ts_alloc(A=%p,%zu)",A,n);
  return malloc(n);
}

static inline void* ts_realloc_impl(ts_allocator* A, void* p, size_t n, const char* fname, int line)
{
  UNUSED_PARAM(A);
  UNUSED_PARAM(fname);
  UNUSED_PARAM(line);
  TS_ALLOC_LOG(fname,line,"ts_realloc(A=%p,%p,%zu)",A,p,n);
  return realloc(p,n);
}

static inline void* ts_calloc_impl(ts_allocator* A, size_t n, size_t sz, const char* fname, int line)
{
  UNUSED_PARAM(A);
  UNUSED_PARAM(fname);
  UNUSED_PARAM(line);
  TS_ALLOC_LOG(fname,line,"ts_calloc(A=%p,%zu,%zu)",A,n,sz);
  return calloc(n,sz);
}

static inline void ts_free_impl(ts_allocator* A, void* p, const char* fname, int line)
{
  UNUSED_PARAM(A);
  UNUSED_PARAM(fname);
  UNUSED_PARAM(line);
  TS_ALLOC_LOG(fname,line,"ts_free(A=%p,%p)",A,p);
  free(p);
}

static inline void ts_safe_free(ts_allocator* A, void* p) { UNUSED_PARAM(A); if (p) { ts_free(A,p); } }

static inline char* ts_strdup(ts_allocator* A, const char* s)
{
  UNUSED_PARAM(A);
  size_t sz = strlen(s) + 1;
  char* dup_s = ts_alloc(NO_ALLOCATOR, sz);
  return dup_s ? memcpy(dup_s, s, sz) : NULL;
}

static inline char hexnib(int nib) {
  nib &= 0xf;
  return nib < 10 ? ('0' + nib) : 'A' + nib-10;
}

static inline void i8_to_hex(int v, char s[2])
{
  s[0] = hexnib((v >> 4) & 0xf);
  s[1] = hexnib((v     ) & 0xf);
}

static inline void i16_to_hex(int v, char s[4])
{
  i8_to_hex(v>>8,s+0);
  i8_to_hex(v   ,s+2);
}

#define MAX_RELEASE_FUNCS 32
struct ts_autopool_entry;
struct ts_autopool
{
  struct ts_autopool_entry* first;
  ts_allocator* allocator;
  int num_release_funcs;
  struct {
    void (*func)(void*);
    void* data;
  } release_funcs[MAX_RELEASE_FUNCS];
};

void* ts_autopool_alloc(struct ts_autopool* pool, size_t sz);
void ts_autopool_release(struct ts_autopool* pool);
void ts_autopool_init(struct ts_autopool* pool, ts_allocator* allocator);

#define AUTOPOOL_ON_RELEASE(pool, func, data) \
  RETURN_ON_ERROR3( ts_autopool_on_release( \
        (pool), (void(*)(void*))(func), (void*)(data) ) )
TSResult ts_autopool_on_release(struct ts_autopool* pool, void(*func)(void*), void* data);

#undef MODULE_NAME

#endif /* TS_COMMON__H */

