#ifdef __GNUC__
#  if !defined(_GNU_SOURCE)
#    define _GNU_SOURCE
#  endif
#endif

#include "ts_dispatch.h"

#ifdef TS_DISPATCH_ENABLED

#include <sched.h>
#include <time.h>
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

#include "ts_common.h"

#ifdef __APPLE__
#include <sys/sysctl.h>

MAYBE_UNUSED
static inline int numcores(void) {
  // int physicalCores = -1;
  int logicalCores = -1;
  // size_t szPhysical = sizeof(physicalCores);
  size_t szLogical  = sizeof(logicalCores);

  // sysctlbyname("hw.physicalcpu", 
  //    &physicalCores, &szPhysical, NULL, 0);
  sysctlbyname("hw.logicalcpu", 
      &logicalCores, &szLogical, NULL, 0);
  return logicalCores;
}
#endif /* __APPLE__ */

#ifdef __linux
static inline int numcores(void) {
  return sysconf(_SC_NPROCESSORS_ONLN);
}
#endif /* linux */

/* thread affinity stuff */
#ifdef __APPLE__
MAYBE_UNUSED
static void attach_thread_to_core(int thread)
{
  UNUSED_PARAM(thread);
}
#endif /* __APPLE__ */

int tsd_ignore_cpuset = 0;
int tsd_cpu_offset = 0;  /* used if tsd_ignore_cpuset is true */

#ifdef __linux
MAYBE_UNUSED
static void attach_thread_to_core(int thread)
{
  int num_cores, max_cores, core_seq, core_id, i, ncpu;
  cpu_set_t cpuset, avail;

  max_cores = numcores();

  if (tsd_ignore_cpuset)
  {
    fprintf(stderr, "Dispatching ignoring the CPU set (offset=%d)\n",tsd_cpu_offset);
    core_id = (thread + tsd_cpu_offset) % max_cores;
    goto set_affinity;
  }

  CPU_ZERO(&avail);
  sched_getaffinity(0, sizeof(avail), &avail);

  num_cores = CPU_COUNT(&avail);
  core_seq = thread % num_cores;
  fprintf(stderr, "Cores available: %d\n", num_cores);

  if (num_cores == 1)
  {
    fprintf(stderr,
        "Only a single core is available (%d cores on machine)!\n"
        "Check your affinity mask.  Not attaching threads to cores\n",
        max_cores);
    return;
  }

  core_id = -1;
  ncpu = 0;

  for (i=0; i < max_cores; i++)
  {
    if (CPU_ISSET(i, &avail) && (ncpu++ == core_seq))
    {
      core_id = i;
      break;
    }
  }

  if (core_id < 0)
  {
    /* FIXME: log the failure more robustly! */
    fprintf(stderr,
        "\nERROR: Cannot set thread affinity for thread %d (could not find cpu)\n",
        thread);
    return;
  }

set_affinity:
  CPU_ZERO(&cpuset);
  CPU_SET(core_id, &cpuset);

  pthread_t current_thread = pthread_self();    
  int ok = pthread_setaffinity_np(current_thread, sizeof(cpu_set_t), &cpuset);
  if (ok == 0) {
    fprintf(stderr, "Attached thread %d to core %d\n",thread, core_id);
  } else {
    fprintf(stderr, "Error attaching thread %d to core %d: %d\n", thread, core_id, ok);
  }
}
#endif /* */

static inline int voidp_to_int(void* p0)
{
  return (int) ( (uintptr_t)p0 );
}

static inline void* int_to_voidp(int i)
{
  return (void*) ((uintptr_t)i);
}

static inline void thread_yield(void)
{
  sched_yield();
}

struct tsd_workq
{
  struct tsd_task tasks[TSD_WORKQ_MAX];

  int task_beg;
  int task_end;
  int num_tasks;
  int num_waiting;
  int num_running;

  enum WORKER_QUEUE_STATE state;

  pthread_mutex_t lock;
  pthread_cond_t  state_changed;
  pthread_cond_t  worker_waiting;
};

static struct tsd_workq workq;

void tsd_workq_initialize(void)
{
  memset(workq.tasks+0, 0, sizeof(workq.tasks));
  workq.task_beg = workq.task_end = 0;
  workq.num_tasks = 0;
  workq.num_waiting = 0;
  workq.num_running = 0;
  workq.state = WORKER_QUEUE_STOPPED;

  pthread_mutex_init(&workq.lock, NULL);
  pthread_cond_init(&workq.state_changed, NULL);
  pthread_cond_init(&workq.worker_waiting, NULL);
}

void tsd_workq_finalize(void)
{
  memset(workq.tasks+0, 0, sizeof(workq.tasks));
  workq.num_tasks = 0;

  pthread_mutex_destroy(&workq.lock);
  pthread_cond_destroy(&workq.state_changed);
}

static inline void
workq_lock(void)
{
  pthread_mutex_lock(&workq.lock);
}

static inline void
workq_unlock(void)
{
  pthread_mutex_unlock(&workq.lock);
}

static inline void
workq_unblock(void)
{
  pthread_cond_broadcast(&workq.state_changed);
}

static inline int
workq_block_until_running(void)
{
  workq.num_waiting++;
  pthread_cond_broadcast(&workq.worker_waiting);

  while ((workq.state != WORKER_QUEUE_RUNNING)
      && (workq.state != WORKER_QUEUE_QUIT))
  {
    pthread_cond_wait(
        &workq.state_changed,
        &workq.lock);
  }
  workq.num_waiting--;

  return workq.state != WORKER_QUEUE_QUIT;
}

enum WORKER_QUEUE_STATE
tsd_workq_state_info(int* nwrkp, int* nwaitp)
{
  workq_lock();

  enum WORKER_QUEUE_STATE state = workq.state;
  if (nwrkp) { *nwrkp = workq.num_tasks; }
  if (nwaitp) { *nwaitp = workq.num_waiting; }

  workq_unlock();

  return state;
}

enum WORKER_QUEUE_STATE
tsd_workq_state(void)
{
  return tsd_workq_state_info(NULL,NULL);
}

int tsd_workq_size(void)
{
  int n;

  workq_lock();
  n = workq.num_tasks;
  workq_unlock();

  return n;
}

void tsd_workq_quit(void)
{
  workq_lock();

  workq.state = WORKER_QUEUE_QUIT;
  workq_unblock();

  workq_unlock();
}

void tsd_workq_start(void)
{
  workq_lock();

  if (workq.state == WORKER_QUEUE_STOPPED)
  {
    if (workq.num_tasks > 0)
    {
      workq.state = WORKER_QUEUE_RUNNING;
      workq_unblock();
    }
    else
    {
      workq.state = WORKER_QUEUE_EMPTY;
      workq_unblock();
    }
  }

  workq_unlock();
}

static inline void workq_stop(void)
{
  if (workq.state != WORKER_QUEUE_QUIT)
    workq.state = WORKER_QUEUE_STOPPED;
}

void tsd_workq_stop(void)
{
  workq_lock();
  workq_stop();
  workq_unlock();
}

int tsd_workq_add(struct tsd_task* w)
{
  workq_lock();

  int can_add = (workq.state != WORKER_QUEUE_QUIT) &&
    (workq.num_tasks < TSD_WORKQ_MAX);
  if (can_add)
  {
    workq.tasks[workq.task_end] = *w;
    workq.num_tasks++;

    workq.task_end = (workq.task_end+1) % TSD_WORKQ_MAX;

    if (workq.state == WORKER_QUEUE_EMPTY)
    {
      workq.state = WORKER_QUEUE_RUNNING;
      workq_unblock();
    }
  }

  workq_unlock();

  return can_add;
}

static int
workq_next_inner(struct tsd_task* w)
{
  for(;;)
  {
    if (workq.state == WORKER_QUEUE_RUNNING) {
      if (workq.num_tasks > 0) {
        memcpy(w, workq.tasks + workq.task_beg, sizeof(*w));
        workq.task_beg = (workq.task_beg+1) % TSD_WORKQ_MAX;
        workq.num_tasks--;
        return 1;
      } else {
        workq.state = WORKER_QUEUE_EMPTY;
        workq_unblock();
        /*
        workq_unlock();
        thread_yield();
        workq_lock();
        */
      }
    } else if ((workq.state == WORKER_QUEUE_STOPPED) || (workq.state == WORKER_QUEUE_EMPTY)) {
      workq_block_until_running();
    } else if (workq.state == WORKER_QUEUE_QUIT) {
      return 0;
    } else {
      fprintf(stderr, "--- THREAD %d WORKQ state is unknown (%d)\n",
          ts_dispatch_thread_index(),
          (int)workq.state);
      return 0;
    }
  }
}


int tsd_workq_next(struct tsd_task* w)
{
  workq_lock();
  int ret = workq_next_inner(w);
  workq_unlock();

  return ret;
}

static inline void workq_clear_inner(void)
{
  workq.task_beg = workq.task_end = 0;
  workq.num_tasks = 0;
}

int tsd_workq_clear(void)
{
  workq_lock();

  int ret = 1;
  switch (workq.state)
  {
    default:
    case WORKER_QUEUE_QUIT:
      ret = 0;
      break;

    case WORKER_QUEUE_RUNNING:
      workq.state = WORKER_QUEUE_EMPTY;
      workq_clear_inner();
      workq_unblock();
      break;

    case WORKER_QUEUE_STOPPED:
      workq_clear_inner();
      break;

    case WORKER_QUEUE_EMPTY:
      break;
  }

  workq_unlock();

  return ret;
}

int tsd_workq_sync(void)
{
  workq_lock();

  // Wait until the state changes to empty
  while ((workq.state != WORKER_QUEUE_EMPTY)
      && (workq.state != WORKER_QUEUE_QUIT))
  {
    // fprintf(stderr, "\nworkq_sync() waiting\n");
    pthread_cond_wait(
        &workq.state_changed,
        &workq.lock);
  }

  int quitting = (workq.state == WORKER_QUEUE_QUIT);
  workq_stop();

  workq_unlock();

  return !quitting;
}






/** Dispatch queue */

struct foreach_data {
  tsd_atomic_s32 go_flag;

  TSResult (*func)(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers);
  void* arg;
  size_t i0;
  size_t i1;
  size_t worker;
  size_t nworkers;
  struct ts_error_state estate;
};

#define FOREACH_SPIN_LATENCY 10

#define MAX_THREADS  64
static struct {
  int active;

  int num_threads;
  int num_foreach_threads;
  int num_running;

  pthread_key_t index_key;
  pthread_mutex_t lock;

  pthread_t threads[MAX_THREADS];
  pthread_t foreach_threads[MAX_THREADS];

  tsd_atomic_s32 foreach_running;
  tsd_atomic_s32 foreach_completed;

  volatile int foreach_primed;
  volatile int foreach_latency;

  struct foreach_data foreach_task[MAX_THREADS];
} dispatch = { 
  .active = 1, 
  .num_threads = 0,
  .num_running = 0,
  .index_key = 0,
  .lock = PTHREAD_MUTEX_INITIALIZER,
  .threads = {0},
  .foreach_threads = {0},
  .foreach_completed = {0},
  .foreach_primed = 0,
  .foreach_latency = FOREACH_SPIN_LATENCY,
};

int ts_dispatch_is_active(void) { return dispatch.active; }
void ts_dispatch_set_active(int tf) { dispatch.active = tf; }

static inline void dispatch_lock(void)
{
  pthread_mutex_lock(&dispatch.lock);
}

static inline void dispatch_unlock(void)
{
  pthread_mutex_unlock(&dispatch.lock);
}

int ts_dispatch_num_running(void)
{
  dispatch_lock();
  int nr = dispatch.num_running;
  dispatch_unlock();

  return nr;
}

static inline void incr_num_running(void)
{
  dispatch_lock();
  dispatch.num_running++;
  dispatch_unlock();
}

MAYBE_UNUSED
static inline int
all_tasks_finished(void)
{
  int nwaiting = 0;
  int ntasks = 0;
  enum WORKER_QUEUE_STATE qstate =
    tsd_workq_state_info(&ntasks,&nwaiting);

  return ((qstate == WORKER_QUEUE_EMPTY) ||
      (qstate == WORKER_QUEUE_STOPPED)) &&
    (nwaiting == dispatch.num_threads);
}

static inline void decr_num_running(void)
{
  dispatch_lock();
  dispatch.num_running--;
  dispatch_unlock();
}

static void* thread_pool_runner(void* index_as_voidp)
{
  // int thread = voidp_to_int(index_as_voidp);
  // attach_thread_to_core(thread);
  pthread_setspecific(dispatch.index_key,index_as_voidp);

  workq_lock();
  {
    workq.num_running++;
    /*
    fprintf(stderr, "thread %d is running\n",
        ts_dispatch_thread_index());
        */
    workq_unblock();
  }
  workq_unlock();

  int running = 1;
  do {
    struct tsd_task w;
    memset(&w,0,sizeof(w));

    running = tsd_workq_next(&w);

    incr_num_running();
    if (running && w.run) { w.run( w.data ); }
    decr_num_running();
  } while(running);

  workq_lock();
  {
    workq.num_running--;
    /*
    fprintf(stderr, "thread %d has stopped\n",
        ts_dispatch_thread_index());
        */
    workq_unblock();
  }
  workq_unlock();

  return NULL;
}

static void foreach_runner(struct foreach_data* data)
{
  memset(&data->estate, 0, sizeof(data->estate));
  TSResult ret = data->func(data->arg, data->i0, data->i1, data->worker, data->nworkers);
  if (!IS_OKAY(ret)) {
    ts_error_copy_state(ret, &data->estate);
  }
}

static inline void foreach_maybe_spin(void)
{
  if (dispatch.foreach_primed || (dispatch.foreach_latency == 0)) {
    thread_yield();
  } else {
    ts_dispatch_nanosleep(dispatch.foreach_latency);
  }
}

int ts_dispatch_foreach_latency(void) { return dispatch.foreach_latency; }
void ts_dispatch_foreach_set_latency(int ns) { dispatch.foreach_latency = ns; }

static void* foreach_thread_runner(void* index_as_voidp)
{
  int thread = voidp_to_int(index_as_voidp);
  attach_thread_to_core(thread);
  pthread_setspecific(dispatch.index_key,index_as_voidp);

  while (tsd_atomic_s32_get(&dispatch.foreach_running) == 1) {
    if (tsd_atomic_s32_get(&dispatch.foreach_task[thread].go_flag)) {
      tsd_atomic_s32_set(&dispatch.foreach_task[thread].go_flag,0);
      foreach_runner(dispatch.foreach_task + thread);
      tsd_memory_barrier();
      tsd_atomic_s32_incr(&dispatch.foreach_completed);
    }
    foreach_maybe_spin();
  }

  return NULL;
}

static void create_threads(int n)
{
  if (n > MAX_THREADS) { n = MAX_THREADS; }
  // fprintf(stderr, "\n\ncreating %d threads\n",n);

  for (int i=0; i < n; i++)
  {
    if (pthread_create(
          dispatch.threads+i, NULL,
          thread_pool_runner, int_to_voidp(i+1)) == 0)
    {
      dispatch.num_threads++;
    }
    else
    {
      /* signal error in creating thread */
    }
  }
}

static void
create_foreach_threads(int n)
{
  if (n > MAX_THREADS) { n = MAX_THREADS; }

  for (int i=0; i < n; i++) {
    if (pthread_create(
          dispatch.foreach_threads+i, NULL,
          foreach_thread_runner, int_to_voidp(i+1)) == 0)
    {
      dispatch.num_foreach_threads++;
    }
    else
    {
      /* signal error in creating thread */
    }
  }
}

TSResult ts_dispatch_initialize(int num_threads)
{
  // fprintf(stderr, "--> STARTUP\n");
  if (dispatch.num_threads > 0)
  {
    return TS_NEW_ERROR2(
        TS_ERR_UNSUPPORTED_OPERATION,
        ( "ts_dispatch is already initialized (num_threads=%d)",
          dispatch.num_threads ));
  }

  memset(&dispatch,0,sizeof(dispatch));

  if (pthread_key_create(&dispatch.index_key, NULL) != 0)
  {
    /* FIXME: report appropriate error */
    return TS_NEW_ERROR3( TS_ERR_DISPATCH_INITIALIZATION,
        "could not create index key" );
  }

  /* FIXME: need to check for failure in initialization */
  tsd_workq_initialize();
  dispatch.num_threads = 0;
  dispatch.num_running = 0;

  dispatch.foreach_latency = FOREACH_SPIN_LATENCY;

  dispatch.active = 1;

  pthread_mutex_init(&dispatch.lock,NULL);

  tsd_atomic_s32_set(&dispatch.foreach_running,1);

  create_threads(num_threads);
  create_foreach_threads(num_threads-1);
  /* attach main thread to first core */
  attach_thread_to_core(0);
  if (dispatch.num_threads < 1)
  {
    tsd_atomic_s32_set(&dispatch.foreach_running,0);
    return TS_NEW_ERROR3(TS_ERR_THREAD_POOL,
        "No threads were started");
  }

  return TS_OK;
}

static void wait_until_none_running(void)
{
  workq_lock();

  if (workq.state != WORKER_QUEUE_QUIT) {
    fprintf(stderr,
        "wait_until_none_running invoked when queue is not QUITTING.\n");
  } else {
    // Wait until the state changes to empty
    while (workq.num_running > 0)
    {
      // fprintf(stderr, "\nworkq_sync() waiting\n");
      pthread_cond_wait(
          &workq.state_changed,
          &workq.lock);
    }
  }

  workq_unlock();
}

static void dispatch_signal_shutdown(void)
{
  // fprintf(stderr, "--> SHUTDOWN!\n");
  tsd_atomic_s32_set(&dispatch.foreach_running,0);
  tsd_workq_quit();
}

static void dispatch_cleanup_on_shutdown(void)
{
  /* need to implement a timeout here */
  // fprintf(stderr, "JOINing with %d QUEUE threads\n", dispatch.num_threads);
  for (int i=0; i < dispatch.num_threads; i++)
  {
    void* ret = NULL;
    pthread_join(dispatch.threads[i], &ret);
  }

  // fprintf(stderr, "JOINing with %d FOREACH threads\n", dispatch.num_foreach_threads);
  for (int i=0; i < dispatch.num_foreach_threads; i++)
  {
    void* ret = NULL;
    pthread_join(dispatch.foreach_threads[i], &ret);
  }

  pthread_key_delete(dispatch.index_key);
  dispatch.num_threads = 0;
  dispatch.num_foreach_threads = 0;

  pthread_mutex_destroy(&dispatch.lock);

  tsd_workq_finalize();
}

TSResult ts_dispatch_shutdown(void)
{
  dispatch_signal_shutdown();

  /* wait until workers have terminated... */
  wait_until_none_running();

  dispatch_cleanup_on_shutdown();

  return TS_OK;
}

/* time to wait until we start canceling threads, in seconds */
#define MAX_FORCE_DELAY 120
TSResult ts_dispatch_force_shutdown(int waitsecs)
{
  dispatch_signal_shutdown();

  int waitusecs = (waitsecs <= MAX_FORCE_DELAY)
    ? waitsecs
    : MAX_FORCE_DELAY
    ;
  waitusecs *= 1000000;

  int us = 0;
  while(ts_dispatch_num_running() > 0)
  {
    ts_dispatch_usleep(100);
    us += 100;
    if (us >= waitusecs) { break; }
  }

  if (ts_dispatch_num_running() > 0)
  {
    for (int i=0; i < dispatch.num_threads; i++)
    {
      pthread_cancel(dispatch.threads[i]);
    }
  }

  for (int i=0; i < dispatch.num_foreach_threads; i++)
  {
    pthread_cancel(dispatch.foreach_threads[i]);
  }

  dispatch_cleanup_on_shutdown();

  return TS_OK;
}

int ts_dispatch_is_initialized(void)
{
  return dispatch.num_threads > 0;
}

int ts_dispatch_thread_index(void)
{
  return (dispatch.num_threads > 0)
    ? voidp_to_int(pthread_getspecific(dispatch.index_key))
    : 0
    ;
}

int ts_dispatch_num_threads(void) { return dispatch.num_threads; }
int ts_dispatch_max_tasks(void) { return TSD_WORKQ_MAX; }

TSResult ts_dispatch_add_task(void (*run)(void*), void* arg)
{
  if (!dispatch.active) { 
    return TS_NEW_ERROR3(
        TS_ERR_DISPATCH_DISABLED,
        "Cannot add tasks: Dispatch is disabled");
  }

  struct tsd_task w = { .run = run, .data = arg };

  if (tsd_workq_add(&w))
  {
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR3(TS_ERR_THREAD_POOL,
        "Exceeded maximum number of queued tasks");
  }
}

void ts_dispatch_start(void)
{
  if (dispatch.active) { tsd_workq_start(); }
}

void ts_dispatch_stop(void)
{
  tsd_workq_stop();
}

TSResult ts_dispatch_sync(void)
{
  ts_dispatch_start();

  // tsd_workq_sync();

  workq_lock();
  for(;;) {
    if (workq.state == WORKER_QUEUE_QUIT) break;
    if ((workq.state == WORKER_QUEUE_EMPTY) &&
        (workq.num_waiting >= dispatch.num_threads)) {
      /*
      fprintf(stderr, "workq.state : %d, workq.num_waiting : %d\n",
          (int)workq.state, (int)workq.num_waiting);
          */
      break;
    }
    pthread_cond_wait(
        &workq.worker_waiting,
        &workq.lock);
  }
  workq_unlock();

  // TSD_WAIT_FOR( 0 == ts_dispatch_num_queued() );
  // TSD_WAIT_FOR( 0 == ts_dispatch_num_running() );

  ts_dispatch_stop();

  return TS_OK;
}

int ts_dispatch_num_queued(void)
{
  return tsd_workq_size();
}

void ts_dispatch_sleep(uint32_t sec)
{
  struct timespec ts = { 
    .tv_sec = sec,
    .tv_nsec = 0
  };

  nanosleep(&ts, NULL);
}

void ts_dispatch_usleep(uint32_t us)
{
  struct timespec ts = { 
    .tv_sec = us / 1000000,
    .tv_nsec = 1000 * (us % 1000000)
  };

  nanosleep(&ts, NULL);
}

void ts_dispatch_nanosleep(uint32_t ns)
{
  struct timespec ts = { .tv_sec = 0, .tv_nsec = ns, };
  nanosleep(&ts, NULL);
}

/* Allocates one more than max to provide a buffer */
struct ts_cbuf* ts_cbuf_new(size_t max)
{
  size_t size = sizeof(struct ts_cbuf) + max * sizeof(uint64_t);
  struct ts_cbuf* buf = ts_alloc(NO_ALLOCATOR, size);

  buf->head = 0;
  buf->tail = 0;
  buf->max  = max;

  return buf;
}

void ts_cbuf_delete(struct ts_cbuf* buf)
{
  ts_free(NO_ALLOCATOR, buf);
}

int ts_cbuf_push(struct ts_cbuf* buf, uint64_t element)
{
  size_t n = ts_cbuf_size(buf);
  if (n >= buf->max) { return 0; }
  buf->queue[buf->tail] = element;
  buf->tail = (buf->tail < buf->max) ? buf->tail+1 : 0;
  return 1;
}

int ts_cbuf_pop(struct ts_cbuf* buf , uint64_t* element)
{
  if (buf->tail == buf->head) return 0;
  *element = buf->queue[buf->head];
  buf->head = (buf->head < buf->max) ? buf->head+1 : 0;
  return 1;
}

struct ts_channel
{
  struct ts_cbuf* buf;

  int nwriters;

  pthread_mutex_t mutex;
  pthread_cond_t signal;
};

struct ts_channel* ts_channel_new(size_t n)
{
  struct ts_channel* ch = ts_alloc(NO_ALLOCATOR, sizeof(*ch));
  ch->buf = ts_cbuf_new(n);
  pthread_mutex_init(&ch->mutex, NULL);
  pthread_cond_init(&ch->signal, NULL);

  return ch;
}

void ts_channel_delete(struct ts_channel* ch)
{
  ts_cbuf_delete(ch->buf);

  pthread_mutex_destroy(&ch->mutex);
  pthread_cond_destroy(&ch->signal);

  ts_free(NO_ALLOCATOR, ch);
}

static inline void
channel_lock(struct ts_channel* ch)
{
  pthread_mutex_lock(&ch->mutex);
}

static inline void
channel_unlock(struct ts_channel* ch)
{
  pthread_mutex_unlock(&ch->mutex);
}

void ts_channel_writer_open(struct ts_channel* ch)
{
  channel_lock(ch);

  ch->nwriters++;

  channel_unlock(ch);
}

void ts_channel_writer_close(struct ts_channel* ch)
{
  channel_lock(ch);

  if (ch->nwriters > 0) { ch->nwriters--; }

  channel_unlock(ch);
}

static inline int channel_closed(struct ts_channel* ch)
{
  return ch->nwriters <= 0;
}

int ts_channel_closed(struct ts_channel* ch)
{
  channel_lock(ch);
  int is_closed = channel_closed(ch);
  channel_unlock(ch);

  return is_closed;
}


size_t ts_channel_size(struct ts_channel* ch)
{
  return ch->buf->max;
}

static inline size_t channel_nfill(struct ts_channel* ch)
{
  return ts_cbuf_size(ch->buf);
}

size_t ts_channel_nfill(struct ts_channel* ch)
{
  channel_lock(ch);
  size_t nfill = channel_nfill(ch);
  channel_unlock(ch);

  return nfill;
}

static inline size_t channel_available(struct ts_channel* ch)
{
  return ch->buf->max - channel_nfill(ch);
}

static inline void channel_signal_reader(struct ts_channel* ch)
{
  pthread_cond_signal(&ch->signal);
}

static inline void channel_ensure_space(struct ts_channel* ch, size_t n)
{
  while (channel_available(ch) < n)
  {
    channel_signal_reader(ch);
    channel_unlock(ch);
    sched_yield();
    channel_lock(ch);
  }
}

void ts_channel_write(struct ts_channel* ch, size_t n, uint64_t* p)
{
  channel_lock(ch);
  channel_ensure_space(ch,n);

  for(size_t i=0; i < n; i++)
  {
    ts_cbuf_push(ch->buf, p[i]);
  }

  channel_unlock(ch);
}

static inline void channel_wait_for_writer(struct ts_channel* ch)
{
  pthread_cond_wait(&ch->signal, &ch->mutex);
}

static inline int channel_ensure_content(struct ts_channel* ch, size_t n)
{
  while (channel_nfill(ch) < n)
  {
    if (channel_closed(ch)) return 0;
    channel_wait_for_writer(ch);
  }

  return 1;
}

int ts_channel_read(struct ts_channel* ch, size_t n, uint64_t* p)
{
  channel_lock(ch);
  if (!channel_ensure_content(ch,n)) return 0;

  for(size_t i=0; i < n; i++)
  {
    if (!ts_cbuf_pop(ch->buf, p+i)) return 0;
  }

  channel_unlock(ch);
  return 1;
}


struct tsd_syncpoint
{
  int count;
  pthread_mutex_t lock;
  pthread_cond_t  trigger;
};

struct tsd_syncpoint* tsd_syncpoint_new(void)
{
  struct tsd_syncpoint* sp = ts_alloc(NO_ALLOCATOR, sizeof(*sp));
  sp->count = 0;
  pthread_mutex_init(&sp->lock, NULL);
  pthread_cond_init(&sp->trigger, NULL);

  return sp;
}

void tsd_syncpoint_delete(struct tsd_syncpoint* sp)
{
  pthread_mutex_destroy(&sp->lock);
  pthread_cond_destroy(&sp->trigger);
  ts_free(NO_ALLOCATOR, sp);
}

static inline void syncpoint_lock(struct tsd_syncpoint* sp)
{
  pthread_mutex_lock(&sp->lock);
}

static inline void syncpoint_unlock(struct tsd_syncpoint* sp)
{
  pthread_mutex_unlock(&sp->lock);
}

int tsd_syncpoint_count(struct tsd_syncpoint* sp)
{
  syncpoint_lock(sp);
  int count = sp->count;
  syncpoint_unlock(sp);

  return count;
}

void tsd_syncpoint_wait(struct tsd_syncpoint* sp)
{
  syncpoint_lock(sp);

  sp->count++;
  pthread_cond_wait(
      &sp->trigger,
      &sp->lock);
  sp->count--;

  syncpoint_unlock(sp);
}

void tsd_syncpoint_trigger(struct tsd_syncpoint* sp)
{
  syncpoint_lock(sp);
  pthread_cond_broadcast(&sp->trigger);
  syncpoint_unlock(sp);
}

struct tsd_joinpoint
{
  int num_expected;
  int num_waiting;
  int num_woken;

  pthread_mutex_t lock;
  pthread_cond_t signal;
};

static inline void
joinpoint_lock(struct tsd_joinpoint* jp)
{
  pthread_mutex_lock(&jp->lock);
}

static inline void
joinpoint_unlock(struct tsd_joinpoint* jp)
{
  pthread_mutex_unlock(&jp->lock);
}

struct tsd_joinpoint* tsd_joinpoint_new(int n)
{
  struct tsd_joinpoint* jp = ts_alloc(NO_ALLOCATOR, sizeof(*jp));
  jp->num_expected = n;
  jp->num_waiting = 0;
  jp->num_woken = 0;

  pthread_mutex_init(&jp->lock, NULL);
  pthread_cond_init(&jp->signal, NULL);

  return jp;
}

void tsd_joinpoint_delete(struct tsd_joinpoint* jp)
{
  pthread_cond_destroy(&jp->signal);
  pthread_mutex_destroy(&jp->lock);
  ts_free(NO_ALLOCATOR, jp);
}

// returns number of threads currently waiting
int tsd_joinpoint_num_waiting(struct tsd_joinpoint* jp)
{
  joinpoint_lock(jp);
  int n = jp->num_waiting;
  joinpoint_unlock(jp);
  return n;
}

int tsd_joinpoint_num_woken(struct tsd_joinpoint* jp)
{
  joinpoint_lock(jp);
  int n = jp->num_woken;
  joinpoint_unlock(jp);
  return n;
}

int tsd_joinpoint_num_expected(struct tsd_joinpoint* jp)
{
  joinpoint_lock(jp);
  int n = jp->num_expected;
  joinpoint_unlock(jp);
  return n;
}

void tsd_joinpoint_join(struct tsd_joinpoint* jp)
{
  joinpoint_lock(jp);

  if (jp->num_waiting == 0) { jp->num_woken = 0; }

  jp->num_waiting++;
  if (jp->num_waiting < jp->num_expected)
  {
    while (jp->num_waiting < jp->num_expected)
      pthread_cond_wait(&jp->signal, &jp->lock);
  }
  else
  {
    jp->num_woken = 0;
    pthread_cond_broadcast(&jp->signal);
  }

  jp->num_woken++;
  if (jp->num_woken == jp->num_expected)
  {
    jp->num_waiting = 0;
  }

  joinpoint_unlock(jp);
}

/* Implements a simple barrier with a shared gate.
 *
 * (0) Barrier is (count,gate) pair of atomic integers.
 *     Initially:
 *       count <- N
 *       gate  <- 1
 *
 * (1) Threads wait until gate == 1.
 * (2) Each thread decrements count and checks the decremented value:
 *      (a) if non-zero, goto step 3.
 *      (b) if zero: increment the count, gate <- 0, goto step 5
 * (3) wait until gate == 0
 * (4) increment the count and check the incremented value
 *      (a) if value == N, set gate <- 1
 */
struct tsd_simple_barrier
{
  struct tsd_atomic_s32 gate;
  struct tsd_atomic_s32 count;
};

void tsd_simple_barrier_init(struct tsd_simple_barrier *barrier, int32_t n)
{
  tsd_atomic_s32_init(&barrier->gate,  0);
  tsd_atomic_s32_init(&barrier->count, n);
  tsd_atomic_s32_set(&barrier->gate,   1);
}

void tsd_simple_barrier_gate(struct tsd_simple_barrier *barrier, int32_t n)
{
  while (tsd_atomic_s32_get(&barrier->gate) == 0)
  {
    continue; /* spin */
  }

  if (tsd_atomic_s32_decr(&barrier->count) != 0) {
    while (tsd_atomic_s32_get(&barrier->gate) != 0)
    {
      continue; /* spin */
    }
    tsd_atomic_s32_incr(&barrier->count);
    return;
  }

  tsd_atomic_s32_incr(&barrier->count);
  tsd_atomic_s32_decr(&barrier->gate);
  while (tsd_atomic_s32_get(&barrier->count) < n)
  {
    continue; /* spin */
  }

  tsd_memory_barrier();
  tsd_atomic_s32_incr(&barrier->gate);
}

void ts_dispatch_foreach_prime(void) { tsd_memory_barrier(); dispatch.foreach_primed++; }
void ts_dispatch_foreach_unprime(void) { tsd_memory_barrier(); dispatch.foreach_primed--; }
int ts_dispatch_foreach_is_primed(void) { return dispatch.foreach_primed > 0; }

int ts_dispatch_foreach_num_threads(void) { return dispatch.active ? dispatch.num_foreach_threads+1 : 1; }

TSResult ts_dispatch_foreach(
    size_t n, 
    TSResult (*foreach)(void* data, size_t i0, size_t i1, size_t worker, size_t nworkers), 
    void* arg)
{
  // n==0 implies that there's nothing to do...
  if (n == 0) { return TS_OK; }

  uint32_t nthreads = ts_dispatch_foreach_num_threads();

#if 0
  fprintf(stderr,
      "\n---[ dispatch started ]---\n"
      "nthreads = %d, n = %zu, foreach = %p, arg =%p\n",
      nthreads, n, foreach, arg);
#endif /* 0 */

  // nthreads == 0: the dispatch system is not started
  // nthreads == 1: single-threaded operation
  //
  // in either case, directly execute the foreach function instead of
  // dispatching to the work queue
  if ( !dispatch.active || (nthreads <= 1) ) {
    RETURN_WITH_RESULT( foreach(arg,0,n,0,1) );
  }

  // if there are more threads than tasks, then limit the number of
  // threads we dispatch to to the number of tasks
  if (nthreads > n) { nthreads = n; }

  // ts_dispatch_stop();
  size_t tasks_per_thread = n / nthreads;
  if (nthreads*tasks_per_thread < n) { tasks_per_thread++; }
  if (tasks_per_thread < 1) { tasks_per_thread = 1; }

  tsd_atomic_s32_set(&dispatch.foreach_completed,0);

  size_t nrunners = 0;
  for(size_t i=0; i < nthreads; i++) {
    if (i*tasks_per_thread >= n) { break; }

    nrunners++;

    dispatch.foreach_task[i].func = foreach;
    dispatch.foreach_task[i].arg = arg;
    dispatch.foreach_task[i].worker = i;
    dispatch.foreach_task[i].i0 = i*tasks_per_thread;
    dispatch.foreach_task[i].i1 = dispatch.foreach_task[i].i0 + tasks_per_thread;
    if (dispatch.foreach_task[i].i1 > n) { dispatch.foreach_task[i].i1 = n; }
#if 0
    fprintf(stderr, "\ntask %zu\n",i);
    fprintf(stderr, "  func   = %p\n", dispatch.foreach_task[i].func);
    fprintf(stderr, "  arg    = %p\n", dispatch.foreach_task[i].arg );
    fprintf(stderr, "  worker = %zu\n", dispatch.foreach_task[i].worker);
    fprintf(stderr, "  i0     = %zu\n", dispatch.foreach_task[i].i0);
    fprintf(stderr, "  i1     = %zu\n", dispatch.foreach_task[i].i1);
#endif /* 0 */
  }

  for (size_t i=0; i < nrunners; i++) {
    dispatch.foreach_task[i].nworkers = nrunners;
  }

  tsd_memory_barrier();  // ensure writes finish
  for(size_t i=1; i < nrunners; i++) {
    tsd_atomic_s32_incr(&dispatch.foreach_task[i].go_flag);
  }

  foreach_runner(dispatch.foreach_task+0);

  while (tsd_atomic_s32_get(&dispatch.foreach_completed)+1 < (int)nrunners) {
    foreach_maybe_spin();
  }

  size_t numerrors = 0;
  // if (IS_ERROR(ret)) { numerrors++; }
  for (size_t i=0; i < nthreads; i++) {
    if (dispatch.foreach_task[i].estate.error_code != TS_OKAY) {
      numerrors++;
    }
  }

  // fprintf(stderr, "\n---[ dispatch finished ]---\n");

  if (numerrors == 0) {
    return TS_OK;
  } else {
    struct ts_error_state* earray =
      ts_alloc(NO_ALLOCATOR,numerrors*sizeof(earray[0]));
    size_t eind = 0;

    /*
    if (!IS_ERROR(ret)) {
      ts_error_copy_state(ret, earray+eind);
      eind++;
    }
    */

    for (size_t i=0; i < nthreads; i++) {
      if (dispatch.foreach_task[i].estate.error_code != TS_OKAY) {
        earray[eind] = dispatch.foreach_task[i].estate;
        eind++;
      }
    }

    TSResult err = TS_NEW_ERROR3(TS_ERR_DISPATCH_WORKER,
        "Errors encountered during dispatch foreach");
    ts_error_add_suberrors(numerrors, earray);

    return err;
  }
}


#endif /* TS_DISPATCH_ENABLED */

