#ifndef TS_DISPATCH_H
#define TS_DISPATCH_H

#ifdef TS_DISPATCH_ENABLED

#define IF_DISPATCH(x) x

#include <stdlib.h>
#include <stdint.h>

#include "ts_errors.h"

#include "ts_dispatch_atomics.h"

#define MODULE_NAME TS_DISPATCH


/* Dispatch manager
 *
 * The dispatch manager manages a worker queue and one or more threads
 * that run the workers on the queue.
 */

extern int tsd_ignore_cpuset;
extern int tsd_cpu_offset;

TSResult ts_dispatch_initialize(int num_threads);
TSResult ts_dispatch_shutdown(void);
TSResult ts_dispatch_force_shutdown(int sec);

int ts_dispatch_is_initialized(void);

int ts_dispatch_thread_index(void);

int ts_dispatch_is_active(void);
void ts_dispatch_set_active(int tf);

int ts_dispatch_num_threads(void);
int ts_dispatch_max_workers(void);
int ts_dispatch_num_running(void);

int ts_dispatch_num_queued(void);


/* Simple dispatch mechanism: add a number of workers and then start.
 * Start returns when all of the workers have finished
 */
TSResult ts_dispatch_add_task(void (*run)(void*), void* arg);

void ts_dispatch_start(void);
void ts_dispatch_stop(void);

// Sync waits until no more jobs are queued or running
TSResult ts_dispatch_sync(void);

void ts_dispatch_sleep(uint32_t sec);
void ts_dispatch_usleep(uint32_t us);
void ts_dispatch_nanosleep(uint32_t ns);

#define TSD_WAIT_FOR( cond ) \
  do{ for(;;) { if(cond) break; ts_dispatch_usleep(5); } } while(0)


/* tsd_syncpoint is a wrapper around a condition variable that has a
 * simple interface.
 *
 * Basically, a syncpoint is a point where a number of threads wait
 * until a flag is signaled.
 *
 * The syncpoint maintains a count of the number of waiting threads.  It
 * has three basic operations:
 *
 * WAIT     thread blocks until syncpoint TRIGGER is called
 * TRIGGER  unblocks all threads waiting on the syncpoint
 * COUNT    returns the number of threads blocked at the syncpoint
 */
struct tsd_syncpoint;

struct tsd_syncpoint* tsd_syncpoint_new(void);
void tsd_syncpoint_delete(struct tsd_syncpoint* sp);

int tsd_syncpoint_count(struct tsd_syncpoint* sp);
void tsd_syncpoint_wait(struct tsd_syncpoint* sp);
void tsd_syncpoint_trigger(struct tsd_syncpoint* sp);


/* tsd_joinpoint implements a portable version of a pthread barrier.
 *
 * The joinpoint waits until a specified number of threads are waiting,
 * and then signals the threads to continue
 */
struct tsd_joinpoint;

struct tsd_joinpoint* tsd_joinpoint_new(int n);
void tsd_joinpoint_delete(struct tsd_joinpoint* jp);

// returns number of threads currently waiting
int tsd_joinpoint_num_waiting(struct tsd_joinpoint* jp);
int tsd_joinpoint_num_expected(struct tsd_joinpoint* jp);
int tsd_joinpoint_num_woken(struct tsd_joinpoint* jp);
void tsd_joinpoint_join(struct tsd_joinpoint* jp);


/* Multiple dispatch of "foreach" operation.  The domain of the foreach
 * must be a range from 0 to n-1.  The worker function takes a range of
 * indices to iterate over.
 *
 * Indices are assumed to be independent and to either have independent
 * outputs or use appropriate locking to maintain data consistency.
 * Race conditions will occur if these assumptions are violated.
 */
TSResult ts_dispatch_foreach(
    size_t n, 
    TSResult (*foreach)(void* data, size_t i0, size_t i1, size_t worker, size_t nworkers), 
    void* arg);
int ts_dispatch_foreach_num_threads(void);

void ts_dispatch_foreach_prime(void);
void ts_dispatch_foreach_unprime(void);
int ts_dispatch_foreach_is_primed(void);
int ts_dispatch_foreach_latency(void);
void ts_dispatch_foreach_set_latency(int ns);

/*** worker queue public api ***/

/* tsd_workq is FIFO structure and a very simple state machine
 * with four states:
 *
 * STOPPED    the queue is stopped
 * EMPTY      the queue is empty
 * RUNNING    the queue is running
 * QUIT       all threads reading from the queue should stop execution
 *
 * The queue supports three basic operations:
 *
 *   1. ADD    : adds a task to the end of the queue
 *   2. NEXT   : removes a task from the beginning of the queue
 *   3. CLEAR  : removes all tasks from the queue
 *
 * The queue allows threads to block until it enters the RUNNING or QUIT
 * states:
 *
 *      NEXT   : waits until the queue enters the RUNNING or QUIT state.
 *               In the RUNNING state, returns the next task.
 *               In the QUIT state, returns NULL.
 *
 * The queue is initialized in the STOPPED state.  It supports three
 * operations to change the state:
 *
 *   1. START  : transitions from STOPPED to either RUNNING or EMPTY
 *               If the queue has zero entries, transitions to EMPTY.
 *               If the queue has one or more entries, transitions to
 *               RUNNING.
 *   2. STOP   : transitions from RUNNING or EMPTY to STOPPED
 *   3. QUIT   : transitions from any state to QUIT
 * 
 *
 *
 * ADD       : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : add task to queue        : STOPPED
 * RUNNING   : add task to queue        : RUNNING
 * EMPTY     : add task to queue        : RUNNING
 * QUIT      : return quit condition    : QUIT
 *
 *
 * NEXT      : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : block until state change : STOPPED
 * RUNNING   : remove task from queue   : RUNNING if new queue size > 0
 *           :                          : EMPTY   if new queue size = 0
 * EMPTY     : block until state change : EMPTY
 * QUIT      : return quit condition    : QUIT
 *
 * CLEAR     : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : clears the task queue    : STOPPED
 * RUNNING   : clears the task queue    : EMPTY
 * EMPTY     : nop                      : EMPTY
 * QUIT      : return quit condition    : QUIT
 *
 *
 * START     : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : enables removing jobs    : RUNNING if queue size > 0
 *           : from the queue           : EMPTY   if queue size = 0
 * RUNNING   : nop                      : RUNNING
 * EMPTY     : nop                      : EMPTY
 * QUIT      : return quit condition    : QUIT
 *
 * STOP      : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : nop                      : STOPPED
 * RUNNING   : nop                      : STOPPED
 * EMPTY     : nop                      : STOPPED
 * QUIT      : return quit condition    : QUIT
 *
 * QUIT      : Description              : Next state
 * ----------:--------------------------:------------------------------
 * STOPPED   : nop                      : QUIT
 * RUNNING   : nop                      : QUIT
 * EMPTY     : nop                      : QUIT
 * QUIT      : nop                      : QUIT
 *
 *
 * NEXT operations are the only operations that can block.  Transitions
 * to RUNNING will unblock all blocked NEXT operations.
 *
 */
enum WORKER_QUEUE_STATE {
  WORKER_QUEUE_STOPPED = 0,
  WORKER_QUEUE_EMPTY   = 1,
  WORKER_QUEUE_RUNNING = 2,
  WORKER_QUEUE_QUIT    = 3,
};

#define TSD_WORKQ_MAX 256

struct tsd_task
{
  void(*run)(void*);
  void* data;
};

void tsd_workq_initialize(void);
void tsd_workq_finalize(void);

enum WORKER_QUEUE_STATE tsd_workq_state(void);
enum WORKER_QUEUE_STATE
tsd_workq_state_info(int* nwrkp, int* nwaitp);

int tsd_workq_clear(void);
int tsd_workq_size(void);

int tsd_workq_add(struct tsd_task* w);
int tsd_workq_next(struct tsd_task* w);

void tsd_workq_start(void);
void tsd_workq_stop(void);
void tsd_workq_quit(void);

void tsd_workq_set_latency(int ns);

/*! \brief Waits until the task queue is empty.
 *
 * Returns 1 if the queue is empty and 0 if the queue is quitting.
 */
int tsd_workq_sync(void);

struct ts_cbuf
{
  size_t head;
  size_t tail;

  size_t max;
  uint64_t queue[1];
};

struct ts_cbuf* ts_cbuf_new(size_t max);
void ts_cbuf_delete(struct ts_cbuf* buf);

static inline size_t ts_cbuf_size(struct ts_cbuf* buf)
{
  return (buf->tail < buf->head)
    ? (buf->max + buf->tail - buf->head)
    : buf->tail - buf->head
    ;
}

int ts_cbuf_push(struct ts_cbuf* buf, uint64_t element);
int ts_cbuf_pop(struct ts_cbuf* buf , uint64_t* element);

/*! \brief Concurrent channel object designed for multiple writers and
 * one reader.
 *
 * This is a simple channel design where the channel serves as a buffer
 * between several workers (channel writers) that have read-only access
 * to global data and one worker (channel reader) that has read-write
 * access to global data.  This is only meant to be used if the data
 * written by the channel-reader is not used by the channel-writers.
 * Otherwise, race conditions may occur.
 *
 * Writers atomically write small chunks of 64-bit data (up to 4 64-bit
 * integers at a time).  Readers 
 *
 * The channel is constructed by attaching one reader and multiple
 * writers.  The reader is blocked until either the size of the data in
 * the channel exceeds some buffer or all of the writers have finished.
 */
struct ts_channel;

struct ts_channel* ts_channel_new(size_t n);
void ts_channel_delete(struct ts_channel* ch);

void ts_channel_write(struct ts_channel* ch, size_t n, uint64_t* p);
int ts_channel_read(struct ts_channel* ch, size_t n, uint64_t* p);

size_t ts_channel_size(struct ts_channel* ch);
size_t ts_channel_nfill(struct ts_channel* ch);

void ts_channel_writer_open(struct ts_channel* ch);
void ts_channel_writer_close(struct ts_channel* ch);
int ts_channel_closed(struct ts_channel* ch);

#undef MODULE_NAME

#else /* ! defined(TS_DISPATCH_ENABLED) */

#define IF_DISPATCH(x)

#endif /* TS_DISPATCH_ENABLED */

#endif /* TS_DISPATCH_H */

