#ifndef TS_DISPATCH_ATOMICS_H
#define TS_DISPATCH_ATOMICS_H

#define MODULE_NAME TS_DISPATCH_ATOMICS

/* Provides a machine-independent wrapper for atomic operations on a
 * 32-bit signed integer.
 *
 * Also provides barrier support.
 */
typedef struct tsd_atomic_s32 tsd_atomic_s32;

#ifndef USE_APPLE_ATOMICS_IF_AVAIL
#  define USE_APPLE_ATOMICS_IF_AVAIL 1
#endif

#ifndef USE_GCC_ATOMICS_IF_AVAIL
#  define USE_GCC_ATOMICS_IF_AVAIL 1
#endif

#if defined(__clang__)
#  if __has_builtin(__sync_fetch_and_add)
#    define GCC_ATOMICS_AVAIL 1
#  endif
#elif defined(__GNUC__) && defined(__GNUC_MINOR__)
#  if (__GNUC__ > 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ >= 1))
#    define GCC_ATOMICS_AVAIL 1
#  endif
#endif

#ifndef GCC_ATOMICS_AVAIL
#  define GCC_ATOMICS_AVAIL 0
#endif

#if defined(__APPLE__) && USE_APPLE_ATOMICS_IF_AVAIL

#include <libkern/OSAtomic.h>
#include <time.h>

static inline void tsd_memory_barrier(void) { OSMemoryBarrier(); }
static inline void tsd_nanosleep(int ns) {
  struct timespec ts = { .tv_sec = 0, .tv_nsec = ns };
  nanosleep(&ts, NULL);
}

struct tsd_atomic_s32 { int32_t value; };

static inline void
tsd_atomic_s32_init(struct tsd_atomic_s32* atom, int32_t value)
{
  atom->value = value;
}

static inline void
tsd_atomic_s32_finalize(struct tsd_atomic_s32* atom)
{ UNUSED_PARAM(atom); }

static inline int32_t
tsd_atomic_s32_get(struct tsd_atomic_s32* atom)
{
  return OSAtomicAdd32((int32_t)0, &atom->value);
}

static inline void
tsd_atomic_s32_set(volatile struct tsd_atomic_s32* atom, int32_t val)
{
  OSMemoryBarrier();
  atom->value = val;
}

static inline int32_t tsd_atomic_s32_incr(struct tsd_atomic_s32* atom)
{
  return OSAtomicIncrement32(&atom->value);
}

static inline int32_t tsd_atomic_s32_decr(struct tsd_atomic_s32* atom)
{
  return OSAtomicDecrement32(&atom->value);
}

static inline int tsd_atomic_s32_compare_and_swap(
    struct tsd_atomic_s32* atom, int32_t oldval, int32_t newval)
{
  return OSAtomicCompareAndSwap32(oldval, newval, &atom->value);
}

/* interface to a spinlock implementation
 */
struct tsd_spinlock {
  OSSpinLock lock;
};

static inline void
tsd_spinlock_init(struct tsd_spinlock* lock) { lock->lock = OS_SPINLOCK_INIT; }

static inline int
tsd_spinlock_try(struct tsd_spinlock* lock) { return OSSpinLockTry(&lock->lock); }

static inline void
tsd_spinlock_lock(struct tsd_spinlock* lock) { OSSpinLockLock(&lock->lock); }

static inline void
tsd_spinlock_unlock(struct tsd_spinlock* lock) { OSSpinLockUnlock(&lock->lock); }

#elif GCC_ATOMICS_AVAIL && USE_GCC_ATOMICS_IF_AVAIL

static inline void tsd_memory_barrier(void) { __sync_synchronize(); }

struct tsd_atomic_s32 { int32_t value; };

static inline void
tsd_atomic_s32_init(struct tsd_atomic_s32* atom, int32_t value)
{
  atom->value = value;
}

static inline void
tsd_atomic_s32_finalize(struct tsd_atomic_s32* atom)
{
  UNUSED_PARAM(atom);
}

static inline int32_t
tsd_atomic_s32_get(struct tsd_atomic_s32* atom)
{
  return __sync_fetch_and_add(&atom->value,0);
}

static inline void
tsd_atomic_s32_set(volatile struct tsd_atomic_s32* atom, int32_t val)
{
  __sync_synchronize();
  atom->value = val;
}

static inline int tsd_atomic_s32_incr(struct tsd_atomic_s32* atom)
{
  return __sync_fetch_and_add(&atom->value,1);
}

static inline int tsd_atomic_s32_decr(struct tsd_atomic_s32* atom)
{
  return __sync_fetch_and_sub(&atom->value,1);
}

static inline int tsd_atomic_s32_compare_and_swap(
    struct tsd_atomic_s32* atom, int32_t oldval, int32_t newval)
{
  return __sync_bool_compare_and_swap(&atom->value, oldval, newval);
}

/* interface to a spinlock implementation
 */
struct tsd_spinlock {
  int lock;
};

static inline void
tsd_spinlock_init(volatile struct tsd_spinlock* lock) { lock->lock = 0; }

static inline int
tsd_spinlock_try(volatile struct tsd_spinlock* lock) {
  return ! __sync_lock_test_and_set(&lock->lock, 1);
}

static inline void
tsd_spinlock_lock(volatile struct tsd_spinlock* lock) {
  while (__sync_lock_test_and_set(&lock->lock, 1)) {
    while (lock->lock) { /* spin */ }
  }
}

static inline void
tsd_spinlock_unlock(volatile struct tsd_spinlock* lock) {
  __sync_lock_release(&lock->lock);
  // lock->lock = 0;
}

#else /* default to pthread mutexes */
#include <sched.h>
#include <pthread.h>

/** Use pthread mutexes for synchroniztion */

/** Structure for atomic signed 32-bit integer
 *
 * Baseline implementation uses a mutex.  Can specialize if C11 atomics
 * or another atomic implementation is available.
 */

struct tsd_atomic_s32
{
  pthread_mutex_t mutex;
  int32_t value;
};

static inline void tsd_atomic_s32_init(struct tsd_atomic_s32* atom, int32_t value)
{
  pthread_mutex_init(&atom->mutex, NULL);
  atom->value = value;
}

static inline void tsd_atomic_s32_finalize(struct tsd_atomic_s32* atom)
{
  pthread_mutex_destroy(&atom->mutex);
}

static inline int32_t tsd_atomic_s32_get(struct tsd_atomic_s32* atom)
{
  pthread_mutex_lock(&atom->mutex);
  int32_t v = atom->value;
  pthread_mutex_unlock(&atom->mutex);
  return v;
}

/*
static inline void tsd_atomic_s32_set(struct tsd_atomic_s32* atom, int32_t v)
{
  pthread_mutex_lock(&atom->mutex);
  atom->value = v;
  pthread_mutex_unlock(&atom->mutex);
}
*/

static inline int tsd_atomic_s32_incr(struct tsd_atomic_s32* atom)
{
  pthread_mutex_lock(&atom->mutex);
  int32_t ret = ++atom->value;
  pthread_mutex_unlock(&atom->mutex);
  return ret;
}

static inline int tsd_atomic_s32_decr(struct tsd_atomic_s32* atom)
{
  pthread_mutex_lock(&atom->mutex);
  int32_t ret = --atom->value;
  pthread_mutex_unlock(&atom->mutex);
  return ret;
}

#endif /* __APPLE__ */


#undef MODULE_NAME

#endif /* TS_DISPATCH_ATOMICS_H */

