#include "ts_dof_map.h"
#include "ts_numerical.h"

#include "ts_dispatch.h"

/* Number of vector3s per site required for work space */
enum { VWORK_MULTIPLIER = 6 };

void ts_dof_map_init_empty(struct ts_dof_map* map)
{
  memset(map, 0, sizeof(*map));
}

TSResult ts_dof_map_init(struct ts_dof_map* map, const ts_sim* sim)
{
  RETURN_WITH_RESULT( ts_dof_map_init_with_type(map,sim,TS_DOF_ALL) );
}

TSResult ts_dof_map_init_with_type(struct ts_dof_map* map, const ts_sim* sim, enum TS_DOF_MAP_TYPE type)
{
  ts_dof_map_init_empty(map);
  map->map_type = type;
  RETURN_ON_ERROR( ts_dof_map_rebuild(map,sim) );
  return TS_OK;
}

/* Constructs maps between particles, free particles, and degrees of
 * freedom.
 */
TSResult ts_dof_map_rebuild(struct ts_dof_map* map, const ts_sim* sim)
{
  int include_trans, include_rot;
  size_t i, max_particles, max_rigids, max_rigid_types;
  size_t max_sites, vwork_size;
  size_t *index_map = NULL;
  vector3 *vwork = NULL;

  if (map == NULL) { TS_RAISE_ERROR(TS_ERR_NULL_POINTER_ARG); }

  max_particles = sim->max_particles;
  max_rigids    = sim->max_rigid_bodies;

  max_sites = 0;
  max_rigid_types = ts_count_rigid_body_types(sim);
  for (i=1; i <= max_rigid_types; i++)
  {
    struct ts_rigid_body_type* rbtype = ts_get_rigid_body_type(sim, i);
    if (rbtype->num_sites > max_sites) { max_sites = rbtype->num_sites; }
  }

  /* Allocate arrays.  Make sure all of them can be allocated before
   * assigning the fields in map.  This ensures that, in case of an out
   * of memory error, map stays in its previous state.
   */
  if (map->p_to_fp == NULL)
  {
    size_t num_inds = 2*(max_particles + max_rigids);
    index_map = ts_calloc(NO_ALLOCATOR, num_inds, sizeof(index_map[0]));
    if (index_map == NULL)
    {
      TS_RAISE_ERROR_MESG(TS_ERR_OUT_OF_MEMORY, "ts_dof_map: cannot allocate index arrays");
    }
  }

  vwork_size = VWORK_MULTIPLIER * max_sites * ts_dispatch_foreach_num_threads();
  if (map->vwork_size < vwork_size)
  {
    vwork = ts_calloc(NO_ALLOCATOR, vwork_size, sizeof(*vwork));
    if (vwork == NULL)
    {
      ts_free(NO_ALLOCATOR, index_map);
      TS_RAISE_ERROR_MESG(TS_ERR_OUT_OF_MEMORY,
          "ts_dof_map: cannot allocate vector work array");
    }
  }

  /* All arrays are now successfully allocated, now assign them */
  if (index_map)
  {
    map->p_to_fp = index_map;
    map->fp_to_p = index_map + max_particles;
    map->fr_to_r = map->fp_to_p + max_particles;
    map->rr_to_r = map->fr_to_r + max_rigids;
  }

  if (vwork)
  {
    ts_free(NO_ALLOCATOR, map->vwork);
    map->vwork = vwork;
    map->vwork_size = vwork_size;
  }

  include_trans = (map->map_type & TS_DOF_TRANSLATIONAL);
  include_rot   = (map->map_type & TS_DOF_ROTATIONAL   );

  /* Now set up the maps */
  map->num_free_particles = 0;
  if (!include_trans) { goto map_rigids; }

  for (i=0; i < sim->num_particles; i++)
  {
    ts_particle *p = ts_get_particle(sim,i);
    if (! ts_particle_is_free(p))
    {
      map->p_to_fp[i] = TS_NULL_PARTICLE;
    }
    else
    {
      map->fp_to_p[map->num_free_particles] = i;
      map->p_to_fp[i] = map->num_free_particles;
      map->num_free_particles++;
    }
  }

map_rigids:
  map->num_rot_rigids = 0;
  map->num_free_rigids = 0;
  if (!include_trans && !include_rot) { goto finish; }

  for (i=0; i < sim->num_rigid_bodies; i++)
  {
    if (ts_rigid_body_is_frozen(sim,i)) { continue; }

    if (ts_rigid_body_is_frozen_translation(sim,i) || !include_trans)
    {
      if (include_rot)
      {
        map->rr_to_r[map->num_rot_rigids] = i;
        map->num_rot_rigids++;
      }
      continue;
    }

    // not_frozen_translation && include_trans
    if (include_rot)
    {
      map->fr_to_r[map->num_free_rigids] = i;
      map->num_free_rigids++;
      continue;
    }

    // not_frozen_translation && include_trans && !include_rot
    {
      ts_rigid_body *body;
      size_t com;

      body = ts_get_rigid_body(sim,i);
      com = body->com_particle;

      map->fp_to_p[map->num_free_particles] = com;
      map->p_to_fp[com] = map->num_free_particles;
      map->num_free_particles++;
    }
  }

finish:
  map->num_objects = map->num_free_particles
    + map->num_rot_rigids
    + map->num_free_rigids;

  map->num_dof = 
    PARTICLE_NUM_DOF * map->num_free_particles + 
    RIGID_BODY_NUM_ROT_DOF * map->num_rot_rigids +
    RIGID_BODY_NUM_DOF * map->num_free_rigids;

  return TS_OK;
}

void ts_dof_map_finalize(struct ts_dof_map* map)
{
  ts_free(NO_ALLOCATOR, map->p_to_fp);
  ts_free(NO_ALLOCATOR, map->vwork);
  memset(map, 0, sizeof(*map));
}

void ts_sim_set_dof(
    ts_sim* sim,
    struct ts_dof_map* dof_map,
    size_t dof,
    double* xv,
    enum TS_ROT_REPR rot_repr)
{
  size_t num_fp_dof = 3*dof_map->num_free_particles;

  // FIXME: this doesn't handle freely-rotating rigid bodies
  if (dof < num_fp_dof)
  {
    size_t obj = dof/3;
    size_t elt = dof - 3*obj;
    size_t pind = dof_map->fp_to_p[obj];
    VECTOR3_ELT( PARTICLE_POS(sim, pind), elt ) = xv[dof];
  }
  else
  {
    dof -= num_fp_dof;
    size_t obj = dof/6;
    size_t dof_off = num_fp_dof + 6*obj;

    // tsAssert(obj >= 0);  // size_t is unsigned, so always true
    tsAssert(obj < dof_map->num_free_rigids);

    size_t rbind = dof_map->fr_to_r[obj];
    ts_rigid_body_set_dof(sim, rbind, xv+dof_off, rot_repr);
    ts_update_rigid_body(sim, rbind);
  }
}

/* -- 
 * Functions that transfer the DoFs between the simulation and 
 * an array of values
 * --
 */

struct dof_vec_data {
  ts_sim* sim;
  const ts_dof_map* dof_map;

  VecReal* xv;
  enum TS_ROT_REPR rot_repr;
};

static TSResult particle_dof_to_vec(void *p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data *data = p;
  const ts_sim *sim = data->sim;
  const ts_dof_map *dof_map = data->dof_map;
  VecReal *restrict xv = data->xv;

  for (size_t i=i0; i < i1; i++)
  {
    size_t pind = dof_map->fp_to_p[i];
    vector3_get_array(PARTICLE_POS(sim, pind),
        xv + PARTICLE_NUM_DOF*i);
  }

  return TS_OK;
}

static TSResult rigid_dof_to_vec(void *p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data *data = p;
  const ts_dof_map *dof_map = data->dof_map;
  const ts_sim *sim = data->sim;
  VecReal *restrict xv = data->xv + 
    dof_map->num_free_particles * PARTICLE_NUM_DOF;

  for (size_t i=i0; i < i1; i++)
  {
    tsAssert(i < dof_map->num_free_rigids);
    size_t rbind = dof_map->fr_to_r[i];
    ts_rigid_body_get_dof(sim, rbind, 
        xv + i*RIGID_BODY_NUM_DOF, 
        data->rot_repr);
  }

  return TS_OK;
}

static inline void
rigid_body_get_rot_dof(const ts_sim *sim, size_t rbind, VecReal*restrict dof, enum TS_ROT_REPR rot_repr);

static TSResult rot_rigid_dof_to_vec(void *p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data *data = p;
  const ts_dof_map *dof_map = data->dof_map;
  const ts_sim *sim = data->sim;
  size_t off = dof_map->num_free_particles * PARTICLE_NUM_DOF
    + dof_map->num_free_rigids * RIGID_BODY_NUM_DOF;
  VecReal *restrict xv = data->xv + off;

  for (size_t i=i0; i < i1; i++)
  {
    size_t rbind = dof_map->rr_to_r[i];
    rigid_body_get_rot_dof(sim, rbind, 
        xv + i*RIGID_BODY_NUM_ROT_DOF, 
        data->rot_repr);
  }

  return TS_OK;
}

static inline void apportion_indices(size_t imax, size_t worker, size_t nworkers, size_t *restrict i0p, size_t *restrict i1p)
{
  size_t i0,i1, ind_per_worker = imax / nworkers;
  if (nworkers*ind_per_worker < imax) { ind_per_worker++; }

  i0 = ind_per_worker*worker;
  i1 = i0 + ind_per_worker;
  if (i1 >= imax) { i1 = imax; }

  *i0p = i0;
  *i1p = i1;
}

static TSResult all_dof_to_vec(void *p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  struct dof_vec_data *data = p;
  const ts_dof_map *dof_map = data->dof_map;

  size_t num_particles  = dof_map->num_free_particles;
  size_t num_rigids     = dof_map->num_free_rigids;
  size_t num_rot_rigids = dof_map->num_rot_rigids;

  UNUSED_PARAM(i0);
  UNUSED_PARAM(i1);

  if (num_particles > 0)
  {
    size_t pind0, pind1;
    apportion_indices(num_particles, worker, nworkers, &pind0, &pind1);
    RETURN_ON_ERROR( particle_dof_to_vec(p, pind0, pind1, worker, nworkers) );
  }

  if (num_rigids > 0)
  {
    size_t rind0, rind1;
    apportion_indices(num_rigids, worker, nworkers, &rind0, &rind1);
    RETURN_ON_ERROR( rigid_dof_to_vec(p, rind0, rind1, worker, nworkers) );
  }

  if (num_rot_rigids > 0)
  {
    size_t rind0, rind1;
    apportion_indices(num_rot_rigids, worker, nworkers, &rind0, &rind1);
    RETURN_ON_ERROR( rot_rigid_dof_to_vec(p, rind0, rind1, worker, nworkers) );
  }

  return TS_OK;
}

void ts_dof_to_vec(const ts_sim *sim,
    const ts_dof_map *dof_map,
    VecReal *restrict xv,
    enum TS_ROT_REPR rot_repr)
{
  struct dof_vec_data data = {
    .sim = (ts_sim*)sim,
    .dof_map = dof_map,
    .xv  = xv,
    .rot_repr = rot_repr
  };

  ts_dispatch_foreach(ts_dispatch_foreach_num_threads(),
      all_dof_to_vec, &data);
}



static void calc_gradient_from_rigid_body_analytic(const ts_sim* sim, size_t rb_ind, VecReal*restrict g, vector3 *restrict vwork, size_t work_size);

static TSResult particle_forces_to_gradient(void* p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  const ts_sim* sim = data->sim;
  const ts_dof_map* dof_map = data->dof_map;
  VecReal*restrict gv = data->xv + i0*PARTICLE_NUM_DOF;

  for (size_t i=i0; i < i1; i++)
  {
    size_t pind = dof_map->fp_to_p[i];
    vector3_get_array(PARTICLE_FORCE(sim, pind), gv);

    gv[0] = -gv[0];
    gv[1] = -gv[1];
    gv[2] = -gv[2];

    gv += PARTICLE_NUM_DOF;
  }

  return TS_OK;
}

static TSResult rigid_forces_to_gradient(void* p,size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  const ts_dof_map* dof_map = data->dof_map;
  const ts_sim* sim = data->sim;

  // sim->vl_work is 3*max_particles
  size_t work_size = dof_map->vwork_size / nworkers;
  size_t work_off  = worker * work_size;

  tsAssert((work_off+work_size <= dof_map->vwork_size)
      && "vwork_size is not large enough!\n");

  VecReal*restrict gv = data->xv
    + dof_map->num_free_particles * PARTICLE_NUM_DOF
    + i0 * RIGID_BODY_NUM_DOF;

  for (size_t i=i0; i < i1; i++)
  {
    size_t rbind = dof_map->fr_to_r[i];
    calc_gradient_from_rigid_body_analytic(sim, rbind, gv, dof_map->vwork+work_off, work_size);
    gv += RIGID_BODY_NUM_DOF;
  }

  return TS_OK;
}

static inline void
calc_rotational_gradient_from_rigid_body_analytic(const ts_sim* sim,
    size_t rb_ind, VecReal*restrict g, vector3 *restrict vwork, size_t work_size);

static TSResult rot_rigid_forces_to_gradient(void* p,size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  const ts_dof_map* dof_map = data->dof_map;
  const ts_sim* sim = data->sim;

  // sim->vl_work is 3*max_particles
  size_t work_size = dof_map->vwork_size / nworkers;
  size_t work_off  = worker * work_size;
  VecReal*restrict gv = data->xv
    + dof_map->num_free_particles * PARTICLE_NUM_DOF
    + dof_map->num_free_rigids * RIGID_BODY_NUM_DOF
    + i0 * RIGID_BODY_NUM_ROT_DOF;

  memset(gv, 0, (i1-i0)*RIGID_BODY_NUM_ROT_DOF*sizeof(gv[0]));

  for (size_t i=i0; i < i1; i++)
  {
    size_t rbind = dof_map->rr_to_r[i];
    calc_rotational_gradient_from_rigid_body_analytic(sim, rbind, gv, dof_map->vwork+work_off, work_size);
    gv += RIGID_BODY_NUM_ROT_DOF;
  }

  return TS_OK;
}

void ts_dof_forces_to_gradient_vec(const ts_sim* sim,
    const ts_dof_map* dof_map,
    VecReal*restrict gv)
{
  struct dof_vec_data data = {
    .sim = (ts_sim*)sim, .dof_map = dof_map,
    .xv  = gv,  .rot_repr = TS_ROT_NONE /* not used */
  };

  memset(gv,0,dof_map->num_dof*sizeof(gv[0]));

  ts_dispatch_foreach(dof_map->num_free_particles,
      particle_forces_to_gradient, &data);

  ts_dispatch_foreach(dof_map->num_free_rigids,
      rigid_forces_to_gradient, &data);

  ts_dispatch_foreach(dof_map->num_rot_rigids,
      rot_rigid_forces_to_gradient, &data);
}

static inline void
calc_linear_gradient_from_rigid_body_analytic(const ts_sim* sim, size_t rb_ind, VecReal*restrict g)
{
  ts_rigid_body* body = ts_get_rigid_body(sim, rb_ind);
  size_t n = body->type_ptr->num_sites;

  vector3 com_force;
  vector3_set(&com_force, PARTICLE_FORCE(sim, body->com_particle));
  for(size_t i=0; i < n; i++)
  {
    vector3_add_to(&com_force, PARTICLE_FORCE(sim, body->site_indices[i]));
  }

  vector3_negate_to_array(&com_force, g);
}

static inline void
calc_rotational_gradient_from_rigid_body_analytic(const ts_sim* sim,
    size_t rb_ind, VecReal*restrict g, vector3 *restrict vwork, size_t work_size)
{
  ts_rigid_body* body = ts_get_rigid_body(sim, rb_ind);
  size_t n = body->type_ptr->num_sites;

  vector3 axis;
  double ang = quat4_to_axis_angle_separately(RIGID_ORI(sim,rb_ind), &axis);
  const vector3*restrict ref_vectors = body->type_ptr->ref_vectors;

  UNUSED_PARAM(work_size); // needed if assertions are disabled
  tsAssert( (work_size >= 3*n) && "work size isn't large enough" );

  vector3*restrict dq_w1 = vwork;
  vector3*restrict dq_w2 = dq_w1+n;
  vector3*restrict dq_w3 = dq_w2+n;

  dq_wj(n, dq_w1, dq_w2, dq_w3, ang, &axis, ref_vectors);

  for(size_t i=0; i < n; i++)
  {
    const vector3* fv = PARTICLE_FORCE(sim, body->site_indices[i]);

    g[0] -= vector3_dot(fv, dq_w1+i);
    g[1] -= vector3_dot(fv, dq_w2+i);
    g[2] -= vector3_dot(fv, dq_w3+i);
  }
}

static void calc_gradient_from_rigid_body_analytic(const ts_sim* sim, size_t rb_ind, VecReal*restrict g, vector3 *restrict vwork, size_t work_size)
{
  memset(g, 0, RIGID_BODY_NUM_DOF*sizeof(g[0]));

  calc_linear_gradient_from_rigid_body_analytic(sim, rb_ind, g);
  calc_rotational_gradient_from_rigid_body_analytic(sim, rb_ind, g+PARTICLE_NUM_DOF, vwork, work_size);
}

static inline void
rigid_body_get_trans_dof(const ts_sim* sim, size_t rbind, VecReal*restrict dof)
{
  const ts_rigid_body* body = ts_get_rigid_body(sim, rbind);
  const vector3* pos = PARTICLE_POS(sim, body->com_particle);

  dof[0] = vector3_x(pos);
  dof[1] = vector3_y(pos);
  dof[2] = vector3_z(pos);
}

static inline void
rigid_body_get_rot_dof(const ts_sim* sim, size_t rbind, VecReal*restrict dof, enum TS_ROT_REPR rot_repr)
{
  const quat4*restrict ori = RIGID_ORI(sim,rbind);

  switch (rot_repr)
  {
    case TS_ROT_EULER_ZXZ:
      quat4_to_euler_zxz_array(ori, dof);
      break;

    case TS_ROT_EULER_XYZ:
      quat4_to_euler_xyz_array(ori, dof);
      break;

    case TS_ROT_BODY_RELATIVE:
      {
        quat4 qtmp;
        quat4 qrot;
        quat4_set(&qrot, RIGID_PREV_ORI(sim,rbind));
        QUAT4_W(&qrot) = -QUAT4_W(&qrot);
        quat4_set(&qtmp, ori);
        quat4_lmul(&qtmp, &qrot);
        quat4_to_euler_xyz_array(&qtmp, dof);
      }
      break;

    default:
    case TS_ROT_AXIS_ANGLE:
      quat4_to_axis_angle_vec(ori, dof);
      break;
  }
}

void ts_rigid_body_get_dof(const ts_sim* sim, size_t rbind, VecReal*restrict dof, enum TS_ROT_REPR rot_repr)
{
  rigid_body_get_trans_dof(sim, rbind, dof);
  rigid_body_get_rot_dof(sim, rbind, dof+PARTICLE_NUM_DOF, rot_repr);
}

static inline void
rigid_body_set_trans_dof(ts_sim* sim, size_t rbind, const VecReal*restrict dof)
{
  ts_rigid_body* body = ts_get_rigid_body(sim, rbind);
  vector3* pos = PARTICLE_POS(sim, body->com_particle);

  VECTOR3_X(pos) = dof[0];
  VECTOR3_Y(pos) = dof[1];
  VECTOR3_Z(pos) = dof[2];
}

static inline void
rigid_body_set_rot_dof_zxz(ts_sim* sim, size_t rbind, const VecReal*restrict dof)
{
  quat4_from_euler_zxz_array(RIGID_ORI(sim,rbind), dof);
}

static inline void
rigid_body_set_rot_dof_xyz(ts_sim* sim, size_t rbind, const VecReal*restrict dof)
{
  quat4_from_euler_xyz_array(RIGID_ORI(sim,rbind), dof);
}

static inline void
rigid_body_set_rot_dof_body_rel(ts_sim* sim, size_t rbind, const VecReal*restrict dof)
{
  quat4*restrict ori = RIGID_ORI(sim,rbind);
  quat4*restrict prev_ori = RIGID_PREV_ORI(sim,rbind);
  quat4_from_euler_xyz_array(ori, dof);
  quat4_lmul(ori, prev_ori);
}

static inline void
rigid_body_set_rot_dof_axis_angle(ts_sim* sim, size_t rbind, const VecReal*restrict dof)
{
  quat4_from_axis_angle_vec(RIGID_ORI(sim,rbind), dof);
}

static inline void
rigid_body_set_rot_dof(ts_sim* sim, size_t rbind, const VecReal*restrict dof, enum TS_ROT_REPR rot_repr)
{
  switch (rot_repr)
  {
    case TS_ROT_EULER_ZXZ:
      rigid_body_set_rot_dof_zxz(sim, rbind, dof);
      break;

    case TS_ROT_EULER_XYZ:
      rigid_body_set_rot_dof_xyz(sim, rbind, dof);
      break;

    case TS_ROT_BODY_RELATIVE:
      rigid_body_set_rot_dof_body_rel(sim, rbind, dof);
      break;

    default:
    case TS_ROT_AXIS_ANGLE:
      rigid_body_set_rot_dof_axis_angle(sim, rbind, dof);
      break;
  }
}

void ts_rigid_body_set_dof(ts_sim* sim, size_t rbind, const VecReal*restrict dof, enum TS_ROT_REPR rot_repr)
{
  rigid_body_set_trans_dof(sim, rbind, dof);
  rigid_body_set_rot_dof(sim, rbind, dof+PARTICLE_NUM_DOF, rot_repr);

  ts_update_rigid_body(sim, rbind);
}


static TSResult particle_dof_from_vec(void* p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  ts_sim* sim = data->sim;
  const ts_dof_map* dof_map = data->dof_map;
  VecReal*restrict xv = data->xv;

  for (size_t i=i0; i < i1; i++)
  {
    size_t pind = dof_map->fp_to_p[i];
    vector3_set_array(PARTICLE_POS(sim, pind), 
        xv + i * PARTICLE_NUM_DOF);
  }

  return TS_OK;
}

#define SET_DOF(rotfunc) do {            \
  for(size_t i=0; i < i1; i++) {         \
    size_t rbind = dof_map->fr_to_r[i];  \
    rigid_body_set_trans_dof(sim, rbind, \
        xv + i*RIGID_BODY_NUM_DOF);      \
    rotfunc(sim, rbind,                  \
        xv + i*RIGID_BODY_NUM_DOF + PARTICLE_NUM_DOF); \
  } } while(0)
static TSResult rigid_dof_from_vec(void* p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  ts_sim* sim = data->sim;
  const ts_dof_map* dof_map = data->dof_map;
  size_t poff = dof_map->num_free_particles * PARTICLE_NUM_DOF;
  VecReal*restrict xv = data->xv + poff;

  switch (data->rot_repr)
  {
    case TS_ROT_EULER_ZXZ:
      SET_DOF(rigid_body_set_rot_dof_zxz);
      break;

    case TS_ROT_EULER_XYZ:
      SET_DOF(rigid_body_set_rot_dof_xyz);
      break;

    case TS_ROT_BODY_RELATIVE:
      SET_DOF(rigid_body_set_rot_dof_body_rel);
      break;

    default:
    case TS_ROT_AXIS_ANGLE:
      SET_DOF(rigid_body_set_rot_dof_axis_angle);
      break;
  }

  for (size_t i=i0; i < i1; i++)
  {
    size_t rbind = dof_map->fr_to_r[i];
    ts_update_rigid_body(sim, rbind);
  }

  return TS_OK;
}
#undef SET_DOF

#define SET_DOF(rotfunc) do {           \
  for(size_t i=0; i < i1; i++) {        \
    size_t rbind = dof_map->rr_to_r[i]; \
    rotfunc(sim, rbind,                 \
        xv + i*RIGID_BODY_NUM_ROT_DOF); \
  } } while(0)
static TSResult rot_rigid_dof_from_vec(void* p, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct dof_vec_data* data = p;
  ts_sim* sim = data->sim;
  const ts_dof_map* dof_map = data->dof_map;
  size_t off = dof_map->num_free_particles * PARTICLE_NUM_DOF
    + dof_map->num_free_rigids * RIGID_BODY_NUM_DOF;
  VecReal*restrict xv = data->xv + off;

  switch (data->rot_repr)
  {
    case TS_ROT_EULER_ZXZ:
      SET_DOF(rigid_body_set_rot_dof_zxz);
      break;

    case TS_ROT_EULER_XYZ:
      SET_DOF(rigid_body_set_rot_dof_xyz);
      break;

    case TS_ROT_BODY_RELATIVE:
      SET_DOF(rigid_body_set_rot_dof_body_rel);
      break;

    default:
    case TS_ROT_AXIS_ANGLE:
      SET_DOF(rigid_body_set_rot_dof_axis_angle);
      break;
  }

  for (size_t i=i0; i < i1; i++)
  {
    size_t rbind = dof_map->rr_to_r[i];
    ts_update_rigid_body(sim, rbind);
  }

  return TS_OK;
}

void ts_dof_from_vec(ts_sim* sim,
    const ts_dof_map* dof_map,
    const VecReal*restrict xv,
    double maxdisp[2],
    enum TS_ROT_REPR rot_repr)
{
  struct dof_vec_data data = {
    .sim = sim, .dof_map  = dof_map,
    .xv  = (double*restrict)xv,
    .rot_repr = rot_repr
  };

  ts_dispatch_foreach(dof_map->num_free_particles,
      particle_dof_from_vec, &data);

  ts_dispatch_foreach(dof_map->num_free_rigids,
      rigid_dof_from_vec, &data);

  ts_dispatch_foreach(dof_map->num_rot_rigids,
      rot_rigid_dof_from_vec, &data);

  maxdisp[0] = maxdisp[1] = 0.0;
  ts_update_max_displacements(sim, maxdisp);
}

void ts_calc_gradient_from_rigid_body_analytic(ts_sim* sim, size_t rb_ind, VecReal*restrict g)
{
  calc_gradient_from_rigid_body_analytic(sim, rb_ind, g, sim->vl_work, 3*sim->max_particles);
}

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
};

static TSResult calc_orientation_grad(ts_sim* sim, size_t rbind, ts_rigid_body* body, const vector3*restrict ori0, const vector3*restrict ori1, double* dE)
{
  UNUSED_PARAM(body);

  quat4 saved_ori;
  quat4*restrict ori = RIGID_ORI(sim,rbind);
  quat4_set(&saved_ori, ori);

  quat4_from_axis_angle(ori, ori0);
  ts_update_rigid_body(sim, rbind);
  double en0 = 0.0;
  RETURN_ON_ERROR( ts_calculate_energy(sim, &en0, NO_TERM_GROUPS, TERMWISE_FALSE) );

  quat4_from_axis_angle(ori, ori1);
  ts_update_rigid_body(sim, rbind);
  double en1 = 0.0;
  RETURN_ON_ERROR( ts_calculate_energy(sim, &en1, NO_TERM_GROUPS, TERMWISE_FALSE) );

  quat4_set(ori, &saved_ori);
  ts_update_rigid_body(sim, rbind);

  *dE = en1 - en0;

  return TS_OK;
}

static TSResult calc_position_grad(ts_sim* sim, size_t rbind, ts_rigid_body* body, const vector3*restrict r0, const vector3*restrict r1, double* dE)
{
  double en0 = 0.0;
  double en1 = 0.0;

  vector3* posv = PARTICLE_POS(sim, body->com_particle);

  vector3 saved_pos;
  vector3_set(&saved_pos, posv);

  vector3_set(posv, r0);
  ts_update_rigid_body(sim,rbind);
  RETURN_ON_ERROR( ts_calculate_energy(sim, &en0, NO_TERM_GROUPS, TERMWISE_FALSE) );

  vector3_set(posv, r1);
  ts_update_rigid_body(sim,rbind);
  RETURN_ON_ERROR( ts_calculate_energy(sim, &en1, NO_TERM_GROUPS, TERMWISE_FALSE) );

  vector3_set(posv, &saved_pos);
  ts_update_rigid_body(sim,rbind);

  *dE = en1 - en0;
  return TS_OK;
}

#define GDELTA 1e-6
TSResult ts_calc_gradient_from_rigid_body_numerical(ts_sim* sim, size_t rb_ind, VecReal*restrict g)
{
  ts_rigid_body* body = ts_get_rigid_body(sim, rb_ind);

  vector3 pos;
  vector3_set(&pos, PARTICLE_POS(sim,body->com_particle));

  vector3 pos0,pos1;

  vector3_set(&pos0, &pos); VECTOR3_X(&pos0) -= GDELTA;
  vector3_set(&pos1, &pos); VECTOR3_X(&pos1) += GDELTA;
  RETURN_ON_ERROR( calc_position_grad(sim, rb_ind, body, &pos0, &pos1, g+0) );
  g[0] /= 2*GDELTA;

  vector3_set(&pos0, &pos); VECTOR3_Y(&pos0) -= GDELTA;
  vector3_set(&pos1, &pos); VECTOR3_Y(&pos1) += GDELTA;
  RETURN_ON_ERROR( calc_position_grad(sim, rb_ind, body, &pos0, &pos1, g+1) );
  g[1] /= 2*GDELTA;

  vector3_set(&pos0, &pos); VECTOR3_Z(&pos0) -= GDELTA;
  vector3_set(&pos1, &pos); VECTOR3_Z(&pos1) += GDELTA;
  RETURN_ON_ERROR( calc_position_grad(sim, rb_ind, body, &pos0, &pos1, g+2) );
  g[2] /= 2*GDELTA;


  vector3 axis;
  quat4_to_axis_angle(RIGID_ORI(sim,rb_ind), &axis);

  vector3 axis0, axis1;

  vector3_set(&axis0, &axis); VECTOR3_X(&axis0) -= GDELTA;
  vector3_set(&axis1, &axis); VECTOR3_X(&axis1) += GDELTA;
  RETURN_ON_ERROR( calc_orientation_grad(sim, rb_ind, body, &axis0, &axis1, g+3) );
  g[3] /= 2*GDELTA;

  vector3_set(&axis0, &axis); VECTOR3_Y(&axis0) -= GDELTA;
  vector3_set(&axis1, &axis); VECTOR3_Y(&axis1) += GDELTA;
  RETURN_ON_ERROR( calc_orientation_grad(sim, rb_ind, body, &axis0, &axis1, g+4) );
  g[4] /= 2*GDELTA;

  vector3_set(&axis0, &axis); VECTOR3_Z(&axis0) -= GDELTA;
  vector3_set(&axis1, &axis); VECTOR3_Z(&axis1) += GDELTA;
  RETURN_ON_ERROR( calc_orientation_grad(sim, rb_ind, body, &axis0, &axis1, g+5) );
  g[5] /= 2*GDELTA;

  return TS_OK;
}
#undef GDELTA

