#ifndef TS_DOF_MAP_H
#define TS_DOF_MAP_H

#include "tubesheet.h"

#define MODULE_NAME TS_DOF_MAP

enum TS_DOF_MAP_TYPE {
  TS_DOF_NONE          = 0,
  TS_DOF_TRANSLATIONAL = 1,
  TS_DOF_ROTATIONAL    = 2,
  TS_DOF_ALL           = 3,
};

/*! \brief Map between particle indices and dof indices
 */
typedef struct ts_dof_map ts_dof_map;
struct ts_dof_map
{
  size_t num_free_particles;
  size_t num_rot_rigids;
  size_t num_free_rigids;
  size_t num_objects;
  size_t num_dof;
  size_t vwork_size;

  /* maps between particle and free particle indices */
  size_t *restrict p_to_fp;
  size_t *restrict fp_to_p;

  /* map from free rigid index to rigid body index */
  size_t *restrict fr_to_r;
  size_t *restrict rr_to_r;

  vector3 *restrict vwork;

  enum TS_DOF_MAP_TYPE map_type;
};

void ts_dof_map_init_empty(struct ts_dof_map* map);
TSResult ts_dof_map_init(struct ts_dof_map* map, const ts_sim* sim);
TSResult ts_dof_map_init_with_type(struct ts_dof_map* map, const ts_sim* sim,
    enum TS_DOF_MAP_TYPE type);
TSResult ts_dof_map_rebuild(struct ts_dof_map* map, const ts_sim* sim);

void ts_dof_map_finalize(struct ts_dof_map* map);

void ts_sim_set_dof(
    ts_sim* sim,
    struct ts_dof_map* dof_map,
    size_t dof,
    double* xv,
    enum TS_ROT_REPR rot_repr);

void ts_dof_to_vec(const ts_sim* sim,
    const ts_dof_map* dof_map,
    VecReal*restrict xv,
    enum TS_ROT_REPR rot_repr);

void ts_dof_from_vec(ts_sim* sim,
    const ts_dof_map* dof_map,
    const VecReal*restrict xv,
    double maxdisp[2],
    enum TS_ROT_REPR rot_repr);

void ts_dof_forces_to_gradient_vec(const ts_sim* sim,
    const ts_dof_map* dof_map,
    VecReal*restrict gv);

static inline void ts_particle_get_dof(const ts_sim* sim,
    size_t pind, VecReal*restrict dof)
{
  vector3_get_array( PARTICLE_POS(sim,pind), dof );
}

static inline void ts_particle_set_dof(ts_sim* sim,
    size_t pind, const VecReal*restrict dof)
{
  vector3_set_array( PARTICLE_POS(sim,pind), dof );
}

void ts_rigid_body_get_dof(const ts_sim* sim, size_t rbind, VecReal*restrict dof, enum TS_ROT_REPR rot_repr);
void ts_rigid_body_set_dof(ts_sim* sim, size_t rbind, const VecReal*restrict dof, enum TS_ROT_REPR rot_repr);

void ts_rigid_body_set_rot_dof(ts_sim* sim, size_t rbind, const VecReal*restrict dof, enum TS_ROT_REPR rot_repr);

/* Converts a particle index into an "object" index.  
 */
static inline size_t ts_dof_map_p2o(const ts_sim* sim, const struct ts_dof_map* map, size_t pind)
{
  return (map->p_to_fp[pind] != TS_NULL_PARTICLE)
    ? map->p_to_fp[pind]
    : map->num_free_particles + TS_PARTICLE(sim,pind)->rigid_body_index;
}

void ts_calc_gradient_from_rigid_body_analytic(ts_sim* sim, size_t rb_ind, VecReal*restrict g);
TSResult ts_calc_gradient_from_rigid_body_numerical(ts_sim* sim, size_t rb_ind, VecReal*restrict g);

#undef MODULE_NAME

#endif /* TS_DOF_MAP_H */

