#include "ts_errors.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <errno.h>

#include "ts_common.h"
#include "ts_malloc_protect.h"

#ifndef TS_ASSERTIONS_ENABLED

void ts_set_assert_func(tsAssertFuncType assertFunc, void *data) {
  UNUSED_PARAM(assertFunc);
  UNUSED_PARAM(data);
}
void _tsAssert_priv(int tf, const char *test, const char *fname, int fline) {
  UNUSED_PARAM(tf);
  UNUSED_PARAM(test);
  UNUSED_PARAM(fname);
  UNUSED_PARAM(fline);
}

#else

static void default_assert_handler(void *data,
    int tf, const char *test, const char *fname, int fline)
{
  UNUSED_PARAM(data);
  if (!tf)
  {
    fprintf(stderr,"assertion \"%s\" failed in %s at %d\n",
        test, fname, fline);
    exit(EXIT_FAILURE);
  }
}

static void *userAssertArg = NULL;
static tsAssertFuncType userAssertFunc = default_assert_handler;

void ts_set_assert_func(tsAssertFuncType assertFunc, void *data)
{
  UNUSED_PARAM(data);
  userAssertFunc = assertFunc;
}

void _tsAssert_priv(int tf, const char *test, const char *fname, int fline)
{
  if (!tf)
  {
    if (userAssertFunc == NULL) 
      default_assert_handler(userAssertArg,tf,test,fname,fline);
    else
      userAssertFunc(userAssertArg,tf,test,fname,fline);
  }
}

#endif /* TS_ASSERTIONS_ENABLED */


static void default_error_handler(void *data,
    const char *fname, int fline, const char *errMsg)
{
  UNUSED_PARAM(data);
  fprintf(stderr,"error \"%s\" encountered in %s at %d\n",
      errMsg, fname, fline);
  exit(EXIT_FAILURE);
}

static void *userErrorArg = NULL;
static tsErrorFuncType userErrorFunc = default_error_handler;

void ts_set_error_func(tsErrorFuncType errorFunc, void *data)
{
  UNUSED_PARAM(data);
  userErrorFunc = errorFunc;
}

void _tsError_priv(const char *fname, int fline, const char *errMsg)
{
  if (userErrorFunc == NULL)
    default_error_handler(userErrorArg,fname,fline,errMsg);
  else
    userErrorFunc(userErrorArg,fname,fline,errMsg);
}

static struct ts_error_state err_states[MAX_ERROR_STATE];

#ifdef TS_DISPATCH_ENABLED
/* FIXME: Need a better way to link error handling with the dispatch
 * mechanism
 */
extern int ts_dispatch_thread_index(void);
static inline int get_state(void)
{
  return ts_dispatch_thread_index();
}
#else /* TS_DISPATCH_ENABLED */
static inline int get_state(void)
{
  return 0;
}
#endif /* TS_DISPATCH_ENABLED */

static void init_state(int state, enum TS_ERROR_CODE code)
{
  memset(err_states+state, 0, sizeof(err_states[state]));

  err_states[state].error_code = code;
  // err_states[state].error_message[ERR_MESG_LEN-1] = '\0';
  // err_states[state].depth = 0;
}

static void add_frame_info(int state, const char *fname, int lineno, const char *func)
{
  int ind = err_states[state].depth;
  err_states[state].depth++;
  if (ind >= STACK_TRACE_LIMIT) { ind = STACK_TRACE_LIMIT-1; }

  err_states[state].trace[ind].file = fname;
  err_states[state].trace[ind].line = lineno;
  err_states[state].trace[ind].func = func;
}

TSResult ts_error_add_frame(TSResult err, const char *fname, int lineno, const char *func)
{
  int state = ERRSTATE(err);

  add_frame_info(state, fname, lineno, func);

  return err;
}

TSResult ts_error_set_userdata(TSResult err, void *data)
{
  if (IS_ERROR(err))
  {
    int state = ERRSTATE(err);
    err_states[state].user_data = data;
  }
  return err;
}

TSResult ts_error_new(enum TS_ERROR_CODE code,
    const char *fname, int lineno, const char *func,
    const char *mesg, ...)
{
  int state = get_state();
  int error = ERROR_ENCODE(code,state);

  init_state(state, code);
  if (mesg)
  {
    va_list va;
    va_start(va, mesg);
    vsnprintf(err_states[state].error_message, ERR_MESG_LEN, mesg, va);
    err_states[state].error_message[ERR_MESG_LEN-1] = '\0'; 
    va_end(va);
  }

  add_frame_info(state, fname, lineno, func);

  TSResult ret = { .error = error };

  return ret;
}

TSResult ts_error_new_errno(const char *fname, int linenum, const char *func)
{
  if (errno == 0)
  {
    return TS_OK;
  }

  return ts_error_new(TS_ERR_IO_ERROR, fname, linenum, func,
      "error (%d): %s", errno, strerror(errno));
}

const struct ts_error_state *ts_error_get_state(TSResult result)
{
  if (!IS_ERROR(result)) { return NULL; }

  int state = ERRSTATE(result);

  return err_states+state;
}

int ts_error_copy_state(TSResult result, struct ts_error_state *ecopy)
{
  const struct ts_error_state *eorig = ts_error_get_state(result);
  if (eorig != NULL) {
    memcpy(ecopy, eorig, sizeof(*ecopy));
    return 1;
  } else {
    memset(ecopy, 0, sizeof(*ecopy));
    return 0;
  }
}

void ts_error_add_suberrors(int nerr, struct ts_error_state *earray)
{
  int state = get_state();
  err_states[state].nsuberrors = nerr;
  err_states[state].suberrors = earray;
}

void ts_error_state_dump_stack(const struct ts_error_state *errstate)
{
  if (errstate == NULL) return;

  fprintf(stderr, "\n---[ ERROR encountered (0x%04x) ]---\n", errstate->error_code);
  if (errstate->error_message[0] != '\0')
    fprintf(stderr, "%s\n", errstate->error_message);

  fprintf(stderr, "\nStack trace follows\n");
  int maxtrace = errstate->depth;
  if (maxtrace >= STACK_TRACE_LIMIT) { maxtrace = STACK_TRACE_LIMIT-2; }
  for (int i=0; i < maxtrace; i++)
  {
    fprintf(stderr, "[%2d] in %s at line %d\n", 
        i, errstate->trace[i].file, errstate->trace[i].line);
  }
  if (errstate->depth >= STACK_TRACE_LIMIT)
  {
    fprintf(stderr, "...\n");
    fprintf(stderr, "[%d] in %s at line %d\n",
        errstate->depth,
        errstate->trace[STACK_TRACE_LIMIT-1].file,
        errstate->trace[STACK_TRACE_LIMIT-1].line);
  }
  fprintf(stderr, "\n");
}

void ts_error_dump_stack(TSResult result)
{
  const struct ts_error_state *errstate = ts_error_get_state(result);
  ts_error_state_dump_stack(errstate);
}

void ts_error_cleanup(TSResult err)
{
  if (IS_ERROR(err)) {
    int state = ERRSTATE(err);
    if (err_states[state].nsuberrors > 0) {
      ts_free(NO_ALLOCATOR, err_states[state].suberrors);
    }
    memset(err_states+state,0,sizeof(err_states[0]));
  }
}

