#ifndef TS_ERRORS__H
#define TS_ERRORS__H

#include "ts_macros.h"
#include <stdint.h>

#define MODULE_NAME TS_ERRORS

typedef union { uint32_t error; } TSResult;

/* Use the error codes in ts_error_codes.h to set the values in the
 * TS_ERROR_CODE enumeration.  _EC is similar to an X macro.
 */
#define _EC(esym, eval, edesc) esym = eval,
enum TS_ERROR_CODE {
#include "ts_error_codes.h"
};
#undef _EC

#define TS_NEW_ERROR(etype) ( ts_error_new(etype, __FILE__, __LINE__,__func__, NULL ) )
#define MSG_PARAMS_EXPAND(mesg,...)  mesg,__VA_ARGS__
#define TS_NEW_ERROR2(etype,msgparams) \
  ( ts_error_new(etype, __FILE__, __LINE__, __func__, MSG_PARAMS_EXPAND msgparams) )
#define TS_NEW_ERROR3(etype,mesg) ( ts_error_new(etype, __FILE__, __LINE__, __func__, mesg) )
#define TS_NEW_ERRORF(etype,...) TS_NEW_ERROR2(etype,(__VA_ARGS__,0))
#define TS_NEW_ERROR_MESG(etype,mesg) TS_NEW_ERROR3(etype,mesg)
#define TS_NEW_ERRNO_ERROR() (ts_error_new_errno(__FILE__, __LINE__, __func__))

#define TS_NEW_RESULT(etype) ( ts_new_result(etype) )

#define TS_OK ( ts_new_okay() )

#define TS_RAISE_ERROR_INNER(err) do { DO_CLEANUP_ON_ERROR(); return err; } while(0)

#define TS_RAISE_ERROR(etype) TS_RAISE_ERROR_INNER( TS_NEW_ERROR(etype) )
#define TS_RAISE_ERROR2(etype,msgparams) TS_RAISE_ERROR_INNER( TS_NEW_ERROR2(etype,msgparams) )
#define TS_RAISE_ERRORF(etype,...) TS_RAISE_ERROR2(etype,(__VA_ARGS__))
#define TS_RAISE_ERROR3(etype,mesg) TS_RAISE_ERROR_INNER( TS_NEW_ERROR3(etype,mesg) )
#define TS_RAISE_ERROR_MESG(etype,mesg) TS_RAISE_ERROR3(etype,mesg)
#define TS_RAISE_ERRNO_ERROR() TS_RAISE_ERROR_INNER( TS_NEW_ERRNO_ERROR() )

#define STOP_ITERATING()   do { return ts_new_result(TS_STOP_ITERATION); } while (0)
#define CANCEL_OPERATION() do { return ts_new_result(TS_CANCELLED);      } while (0)
#define EOF_ENCOUNTERED()  do { return ts_new_result(TS_IO_EOF);         } while (0)

static inline TSResult ts_new_result(enum TS_ERROR_CODE code) {
  TSResult ret = { .error = code }; return ret;
}

/* \brief Creates a new error and begins its stack trace */
TSResult ts_error_new(enum TS_ERROR_CODE code,
    const char *fname, int linenum, const char *func,
    const char *mesg, ...);

TSResult ts_error_new_errno(const char *fname, int linenum, const char *func);

TSResult ts_error_add_frame(TSResult err, const char *fname, int lineno, const char *func);
TSResult ts_error_set_userdata(TSResult err, void *data);

/* Returns a new no-error result.  Note that only error results have the
 * state bits set.
 */
static inline TSResult ts_new_okay(void) { TSResult ret = { .error = TS_OKAY }; return ret; }

#define ERRCODE_BITS 16
#define ERRCODE_MASK ( (1<<(ERRCODE_BITS)) - 1)
#define ERRCODE(ret) ( (ret).error & ERRCODE_MASK )
#define ERRSTATE(ret) ( (ret).error >> ERRCODE_BITS )
#define ERROR_ENCODE(code, state) ( ((code) & ERRCODE_MASK) | ((state) << ERRCODE_BITS) )

/* \brief Checks if the TSResult value is an error or not */
#define IS_ERROR(ret) ( ERRCODE(ret) >= TS_ERROR_OFFSET )

/* \brief Checks if the TSResult value is TS_OKAY (no error) */
#define IS_OKAY(ret)  ( (ret).error == TS_OKAY )

/* \brief Checks if the TSResult value is TS_STOP_ITERATION.  This
 * signals an iterator to stop.
 */
#define IS_STOP_ITERATION(ret)  ( (ret).error == TS_STOP_ITERATION )

/* \brief Checks if the TSResult value is TS_CANCELLED.  This signals
 * that an operation has been cancelled.
 */
#define IS_CANCELLED(ret)  ( (ret).error == TS_CANCELLED )

/* \brief Checks if the TSResult value is TS_IO_EOF.  This signals
 * that an end of file was encountered, and that this is a valid state
 * for the file (between records, etc.)
 */
#define IS_IO_EOF(ret) ( (ret).error == TS_IO_EOF )

/* \brief Adds a "frame" of backtrace information to the backtrace. */
#define ERROR_ADD_FRAME(err) \
  ts_error_add_frame(err, __FILE__, __LINE__, __func__)

/* \brief If the result code is an error, adds a "frame" of backtrace
 * information to the backtrace. */
#define ON_ERROR_ADD_FRAME(err) \
  ts_add_frame_on_error(err, __FILE__, __LINE__, __func__)

/* \brief Returns the result code, adding a "frame" of backtrace
 * information if the result code is an error.
 */
#define RESULT_PROPAGATE(result)   \
  do { return ON_ERROR_ADD_FRAME(result); } while(0)

#define RETURN_WITH_RESULT RESULT_PROPAGATE

#define RETURN_ON_ERROR_WITH_CLEANUP(expr, cleanup_expr) \
  do { TSResult _retOnErrVar = (expr);                   \
    if ( ! IS_OKAY(_retOnErrVar) ) {                     \
      DO_CLEANUP_ON_ERROR();                             \
      cleanup_expr;                                      \
      RESULT_PROPAGATE(_retOnErrVar);                    \
    } } while(0)

#define TS_ERR_NOP_EXPR()

#define RETURN_ON_ERROR(expr) \
  RETURN_ON_ERROR_WITH_CLEANUP(expr, TS_ERR_NOP_EXPR())

#define RETURN_ON_ERROR2(expr,cleanup_expr) \
  RETURN_ON_ERROR_WITH_CLEANUP(expr, cleanup_expr)

#define RETURN_ON_ERROR3(expr) RETURN_ON_ERROR(expr)

static inline void CLEANUP_ON_ERROR(int i) { UNUSED_PARAM(i); }
static inline int cleanup_on_error_passthru(void) { return 0; }
#define DO_CLEANUP_ON_ERROR() do { CLEANUP_ON_ERROR (cleanup_on_error_passthru()); } while(0)

#define ERROR_IF_MESG(err, cond, ...) do { \
    if ( (cond) ) { return TS_NEW_ERROR2(err, (__VA_ARGS__, 0)); } \
  } while(0)

#define ERROR_IF( err, cond ) \
  ERROR_IF_MESG(err, cond, SYMSTR(cond) " is true")

#define ERROR_ON_NULL_ALLOC( var ) \
  do { if (var == NULL) {           \
    DO_CLEANUP_ON_ERROR();          \
    return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY, SYMSTR(var) " is NULL after allocation"); \
  } } while (0)

#define UNIMPLEMENTED() do { \
    return TS_NEW_ERROR2(TS_ERR_UNIMPLEMENTED,\
        ("function %s not yet implemented", __func__)); \
  } while(0)

#define IGNORE_ERROR( expr ) expr

#define ERR_MESG_LEN 256
#define STACK_TRACE_LIMIT 64

static inline TSResult
ts_add_frame_on_error(TSResult err, const char *fname, int lineno, const char *func)
{
  return (! IS_ERROR(err)) ? err : ts_error_add_frame(err,fname,lineno,func);
}

struct ts_stack_trace
{
  const char *file;
  int line;
  const char *func;
};

struct ts_error_state
{
  enum TS_ERROR_CODE error_code;
  char error_message[ERR_MESG_LEN];

  int nsuberrors;
  struct ts_error_state *suberrors;

  void *user_data;

  int depth;
  struct ts_stack_trace trace[STACK_TRACE_LIMIT];
};

const struct ts_error_state *ts_error_get_state(TSResult result);

/*! \brief Makes a copy of the error state if result signals an error.
 * Returns 1 if an error occured and 0 if one did not.
 */
int ts_error_copy_state(TSResult result, struct ts_error_state *state);

void ts_error_dump_stack(TSResult err);
void ts_error_state_dump_stack(const struct ts_error_state* state);

void ts_error_add_suberrors(int nerr, struct ts_error_state* earray);

/*! \brief Cleans up any allocated memory or other resources stored with
 * the error state.  Currently this only applies to errors that occur in
 * the dispatch worker queue.
 */
void ts_error_cleanup(TSResult err);

#define MAX_ERROR_STATE 64
typedef void (*tsAssertFuncType)(void *data, int tf, const char *test, const char *fname, int fline);
typedef void (*tsErrorFuncType)(void *data, const char *fname, int fline, const char *errMsg);

#if defined(TS_ASSERTIONS_ENABLED)
#  define tsAssert( x ) _tsAssert_priv((int)(x), #x, __FILE__, __LINE__)
#  define SAFECHECK 1
#  define BOUNDS_CHECK(x,lo,hi) \
  do { ptrdiff_t b = (x);       \
    tsAssert( (b >= (lo)) &&    \
        (b <= (hi)) &&          \
        SYMSTR(x) " out of bounds [" SYMSTR(lo) " , " SYMSTR(hi) " ]" ); \
  } while(0)
#else
#  define tsAssert( x ) 
#  define SAFECHECK 0
#  define BOUNDS_CHECK(x,lo,hi)  do {} while(0)
#endif

#define tsError( m ) _tsError_priv(__FILE__,__LINE__,m)

void _tsAssert_priv(int tf, const char *test, const char *fname, int fline);
void _tsError_priv(const char *fname, int fline, const char *errMsg);

void ts_set_error_func(tsErrorFuncType errorFunc, void *data);
void ts_set_assert_func(tsAssertFuncType assertFunc, void *data);
#undef MODULE_NAME

/* @} */

#endif /* TS_ERRORS__H */

