#include "ts_fenwick.h"

#include "ts_common.h"
#include "ts_errors.h"

#include "khash.h"

#include <stdint.h>

/* the hash table definition assumes that a pointer will fit in a 64-bit
 * integer
 */
ASSERT_SIZE_LE(void*, uint64_t);

KHASH_MAP_INIT_INT64(index, size_t)
struct dummy_hash_eat_semi { char tmp; };

/* Opaque type for index_lookup table.
 *
 * Keep this opaque so we don't expose khash.h (keep it in the
 * implementation part).
 *
 * Use the INDEX_TABLE(tree) macro to avoid all of the redundant
 * casting.
 */
struct ts_fenwick_index_lookup
{
  /* requires that hash be the first (and only) entry */
  khash_t(index) hash;
};
#define INDEX_TABLE(tree) ( (khash_t(index)*)((tree)->index_lookup) )

// static assertion to make sure that no one has changed the definition
ASSERT_SIZE_EQ( struct ts_fenwick_index_lookup, khash_t(index) );

TSResult ts_fenwick_tree_init(ts_fenwick_tree* tree, size_t max)
{
  tree->capacity = max;
  tree->size = 0;
  tree->size_2n = 1;

  tree->total_sum = 0.0;

  // tree->elements = NULL;
  tree->counts = NULL;
  tree->data = NULL;
  tree->index_lookup = NULL;

#define CLEANUP_ON_ERROR { \
  ts_free(NO_ALLOCATOR, tree->counts);       \
  ts_free(NO_ALLOCATOR, tree->data);         \
  ts_free(NO_ALLOCATOR, tree->index_lookup); \
}

  // tree->elements = ts_calloc(NO_ALLOCATOR, max, sizeof(tree->elements[0]));
  // ERROR_ON_NULL_ALLOC( tree->elements );
  tree->counts = ts_calloc(NO_ALLOCATOR, max, sizeof tree->counts[0]);
  tree->data   = ts_calloc(NO_ALLOCATOR, max, sizeof tree->data[0]);
  ERROR_ON_NULL_ALLOC( tree->counts );
  ERROR_ON_NULL_ALLOC( tree->data   );

  tree->index_lookup = (struct ts_fenwick_index_lookup*) kh_init(index);
  ERROR_ON_NULL_ALLOC( tree->index_lookup );

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_fenwick_tree_resize(ts_fenwick_tree* tree, size_t new_max)
{
  double *new_counts;
  void **new_data;
  struct ts_fenwick_element* new_elts;

  ERROR_IF( TS_ERR_INVALID_ARGUMENT, new_max < tree->size );

#define CLEANUP_ON_ERROR {}

  new_counts = ts_realloc(NO_ALLOCATOR,
      tree->counts, new_max * sizeof new_counts[0]);
  ERROR_ON_NULL_ALLOC( new_counts );
  tree->counts = new_counts;

  new_data = ts_realloc(NO_ALLOCATOR,
      tree->data, new_max * sizeof new_data[0]);
  ERROR_ON_NULL_ALLOC( new_data );
  tree->data = new_data;

  /*
  new_elts = ts_realloc(NO_ALLOCATOR,
      tree->elements, new_max * sizeof new_elts[0]);
  ERROR_ON_NULL_ALLOC( new_elts );
  tree->elements = new_elts;
  */

  // tree->elements = new_elts;
  tree->capacity = new_max;

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_fenwick_tree_clear(ts_fenwick_tree* tree, void (*cleanup_func)(void* closure, void* data), void* closure)
{
  size_t i, size = tree->size;
  double *restrict counts = tree->counts;
  void **restrict data = tree->data;

  for(i=0; i < size; i++)
  {
    // if (cleanup_func) cleanup_func(closure, tree->elements[i].data);
    if (cleanup_func) cleanup_func(closure, data[i]);

    // tree->elements[i].count = 0.0;
    // tree->elements[i].data  = NULL;
    counts[i] = 0.0;
    data[i] = NULL;
  }

  kh_clear(index, INDEX_TABLE(tree));

  tree->size = 0;
  tree->total_sum = 0.0;

  return TS_OK;
}

TSResult ts_fenwick_tree_finalize(ts_fenwick_tree* tree)
{
  // ts_free(NO_ALLOCATOR, tree->elements);
  ts_free(NO_ALLOCATOR, tree->counts);
  ts_free(NO_ALLOCATOR, tree->data);
  if (tree->index_lookup) { kh_destroy(index, INDEX_TABLE(tree)); }

  tree->capacity = 0;
  tree->size = 0;
  tree->size_2n = 0;
  tree->total_sum = 0.0;
  // tree->elements = NULL;
  tree->counts = NULL;
  tree->data = NULL;
  tree->index_lookup = NULL;

  return TS_OK;
}

static double calc_cum_sum(const ts_fenwick_tree* tree, size_t ind)
{
  double cum_sum, *restrict counts;
  // double cum_sum = tree->elements[0].count;
  counts = tree->counts;
  cum_sum = counts[0];
  while (ind > 0)
  {
    cum_sum += counts[ind];
    ind &= ind-1;
  }

  return cum_sum;
}

static void update_total_sum(ts_fenwick_tree* tree)
{
  if (tree->size > 0)
  {
    tree->total_sum = calc_cum_sum(tree,tree->size-1);
  }
  else
  {
    tree->total_sum = 0.0;
  }
}

TSResult ts_fenwick_tree_add(ts_fenwick_tree* tree, double freq, void* data)
{
  size_t ind;
  double *restrict counts = tree->counts;
  void **restrict tree_data = tree->data;

  ERROR_IF( TS_ERR_MAXIMUM_ALLOCATED, tree->size >= tree->capacity );

  counts = tree->counts;
  tree_data = tree->data;
  ind = tree->size++;
  // struct ts_fenwick_element* elt = tree->elements + ind;
  if (tree->size > tree->size_2n) { tree->size_2n <<= 1; }

  counts[ind] = freq;
  tree_data[ind] = data;
  // elt->count = freq;
  // elt->data = data;

  int ret = 0;
  khiter_t iter = kh_put(index, INDEX_TABLE(tree), (uint64_t)data, &ret);
  if ((ret < 0) || (iter == kh_end(INDEX_TABLE(tree))))
  {
    return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
        "Could not insert data into index lookup table");
  }

  kh_value(INDEX_TABLE(tree), iter) = ind;

  if (ind > 0)
  {
    size_t i,parent;
    double accum = counts[ind];

    parent = strip_lsb(ind);
    i = ind-1;
    while (i != parent)
    {
      accum += counts[i];
      i = strip_lsb(i);
    }
    counts[ind] = accum;
  }

  update_total_sum(tree);

  return TS_OK;
}

static double get_element_count(ts_fenwick_tree* tree, size_t ind)
{
  // double count = tree->elements[ind].count;
  double elt_count, *restrict counts;

  counts = tree->counts;
  elt_count = counts[ind];

  if (ind > 0) {
    size_t i, parent;

    for(i=ind-1,parent=strip_lsb(ind); i != parent; i = strip_lsb(i))
    {
      elt_count -= counts[i];
    }
  }

  return elt_count;
}

static void add_to_count(ts_fenwick_tree* tree, size_t ind, double delta)
{
  double *restrict counts = tree->counts;

  counts[ind] += delta;
  if (ind != 0)
  {
    ind += lsb(ind);
    while (ind < tree->size)
    {
      counts[ind] += delta;
      ind += lsb(ind);
    }
  }

  update_total_sum(tree);
}

TSResult ts_fenwick_tree_update_count(ts_fenwick_tree* tree, size_t ind, double new_count)
{
  ERROR_IF( TS_ERR_INVALID_INDEX, ind >= tree->size );

  double old_count = get_element_count(tree,ind);
  double delta = new_count - old_count;

  add_to_count(tree, ind, delta);

  return TS_OK;
}

TSResult ts_fenwick_tree_get_frequency(ts_fenwick_tree* tree, size_t ind, double* freqp)
{
  ERROR_IF( TS_ERR_INVALID_INDEX, ind >= tree->size );

  *freqp = get_element_count(tree,ind);

  return TS_OK;
}

TSResult ts_fenwick_tree_get_cum_sum(ts_fenwick_tree* tree, 
    size_t ind, double* cum_sum)
{
  ERROR_IF( TS_ERR_INVALID_INDEX, ind >= tree->size );

  *cum_sum = calc_cum_sum(tree,ind);
  return TS_OK;
}

size_t ts_fenwick_tree_find_cum_sum(ts_fenwick_tree* tree, double sum)
{
  size_t ind,mask;
  double *restrict counts;

  if ((tree->size == 0) || (sum > tree->total_sum))
  {
    return tree->size;
  }

  counts = tree->counts;

  if (sum -= counts[0], sum <= 0.0) {
    return 0;
  }

  mask = tree->size_2n >> 1;
  ind = 0;

  for (ind=0,mask=tree->size_2n>>1; mask > 0; mask >>=1)
  {
    size_t test_ind = ind+mask;
    double test_count;

    if (test_ind >= tree->size)
    {
      continue;
    }

    test_count = counts[test_ind];
    if (sum == test_count)
    {
      return test_ind;
    }

    if (sum > test_count)
    {
      sum -= test_count;
      ind = test_ind;
    }
  }

  return ind+1;
}

static inline void remove_last_element(ts_fenwick_tree* tree)
{
  tree->size--;

  uint64_t key = (uint64_t) tree->data[tree->size];

  tree->counts[tree->size] = 0.0;
  tree->data[tree->size] = NULL;

  khiter_t iter = kh_get(index, INDEX_TABLE(tree), key);
  if (iter != kh_end(INDEX_TABLE(tree)))
  {
    kh_del(index, INDEX_TABLE(tree), iter);
  }

  update_total_sum(tree);
}

TSResult ts_fenwick_tree_remove_at(ts_fenwick_tree* tree, size_t ind)
{
  ERROR_IF( TS_ERR_INVALID_INDEX, ind >= tree->size );

  if (ind == tree->size-1)
  {
    remove_last_element(tree);
  }
  else
  {
    double last_freq,del_freq,delta, *restrict counts;
    void *swapped, **restrict data;
    uint64_t old_key;
    int ret;
    khiter_t iter;

    counts = tree->counts;
    data = tree->data;

    tree->size--;
    old_key = (uint64_t) data[ind];

    last_freq = get_element_count(tree,tree->size);
    del_freq  = get_element_count(tree,ind);

    /* FIXME: floating point roundoff! */
    delta = last_freq - del_freq;

    swapped = data[tree->size];
    data[ind] = swapped;

    iter = kh_get(index, INDEX_TABLE(tree), old_key);
    if (iter != kh_end(INDEX_TABLE(tree)))
    {
      kh_del(index, INDEX_TABLE(tree), iter);
    }

    ret = 0;
    iter = kh_put(index, 
        INDEX_TABLE(tree), 
        (uint64_t)swapped,
        &ret);

    if ((ret<0) || (iter == kh_end(INDEX_TABLE(tree))))
    {
      return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
          "Could not replace entry in index lookup table");
    }

    kh_value(INDEX_TABLE(tree),iter) = ind;

    data[tree->size] = NULL;
    counts[tree->size] = 0.0;

    add_to_count(tree, ind, delta);
  }

  return TS_OK;
}

TSResult ts_fenwick_tree_replace_at(ts_fenwick_tree* tree, size_t ind, double new_freq, void* data)
{
  double old_freq, delta;
  uint64_t old_key;
  khiter_t iter;
  void **restrict tree_data = tree->data;

  ERROR_IF( TS_ERR_INVALID_INDEX, ind >= tree->size );

  old_freq = get_element_count(tree,ind);
  delta = new_freq - old_freq;

  old_key = (uint64_t)tree_data[ind];
  iter = kh_get(index, INDEX_TABLE(tree), old_key);
  if (iter != kh_end(INDEX_TABLE(tree)))
  {
    kh_del(index, INDEX_TABLE(tree), iter);
  }

  add_to_count(tree, ind, delta);
  tree_data[ind] = data;

  int ret = 0;
  iter = kh_put(index, INDEX_TABLE(tree), (uint64_t)data, &ret);
  if ((ret<0) || (iter == kh_end(INDEX_TABLE(tree))))
  {
    return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
        "Could not replace entry in index table");
  }

  kh_value(INDEX_TABLE(tree), iter) = ind;

  return TS_OK;
}

size_t ts_fenwick_tree_find(ts_fenwick_tree* tree, void* data)
{
  khiter_t iter = kh_get(index, INDEX_TABLE(tree), (uint64_t)data);
  return (iter != kh_end(INDEX_TABLE(tree)))
    ? kh_value(INDEX_TABLE(tree), iter)
    : tree->size
    ;
}

