#include "ts_integrators.h"

#include "ts_common.h"
#include "ts_malloc_protect.h"

#include "ts_bonds.h"
#include "ts_shake.h"
#include "ts_dof_map.h"

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
};

enum INTEGRATOR_FLAGS {
  IFLAGS_NONE              = 0x0000,

  IFLAGS_SETUP_COEFFS      = 0x0001,

  IFLAGS_SHAKE             = 0x0002,

  IFLAGS_ALLOC_RAND_FORCES = 0x0004,

  IFLAGS_RANDOM_FORCES     = 0x0008,
  IFLAGS_RANDOM_TORQUES    = 0x0010,

  IFLAGS_PRECALC_FORCES    = 0x0020,

  IFLAGS_PRECALC_RIGIDS    = 0x0040,
};

#define NO_INTEGRATOR_FLAGS     (IFLAGS_NONE)
#define VVERLET_FLAGS           (IFLAGS_NONE)
#define LEAPFROG_FLAGS          (IFLAGS_SHAKE | IFLAGS_PRECALC_FORCES)
#define LEAPFROG_LANGEVIN_FLAGS (LEAPFROG_FLAGS | IFLAGS_RANDOM_FORCES)
#define STOCH_RUNGE_KUTTA_FLAGS \
  (IFLAGS_SETUP_COEFFS | IFLAGS_ALLOC_RAND_FORCES \
   | IFLAGS_RANDOM_FORCES | IFLAGS_RANDOM_TORQUES \
   | IFLAGS_SHAKE                                 \
   | IFLAGS_PRECALC_FORCES | IFLAGS_PRECALC_RIGIDS)

static inline TSResult setup_constraints(ts_sim* sim)
{
  return (sim->constraints.size > 0) ? shake_initialize(sim) : TS_OK;
}

static inline TSResult apply_constraints(ts_sim* sim)
{
  return (sim->constraints.size > 0) ?  shake_apply_constraints(sim) : TS_OK;
}

static TSResult get_random_forces(ts_sim* sim)
{
  if (! IS_SSE_ALIGNED(sim->rand_forces)) {
    fprintf(stderr,"\n\nsim->rand_forces = %p\n\n",
        (void*)sim->rand_forces);
  }
  tsAssert( IS_SSE_ALIGNED(sim->rand_forces) );
  if (sim->rng_state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }
  RETURN_ON_ERROR( ts_rng_gen_normals(sim->rng_state, sim->num_rand_forces, sim->rand_forces) );
  return TS_OK;
}

static TSResult get_random_torques(ts_sim* sim)
{
  assert( IS_SSE_ALIGNED(sim->rand_torques) );
  return ts_rng_gen_normals(sim->rng_state, sim->num_rand_torques, sim->rand_torques);
}

static inline TSResult integrator_start(ts_sim* sim, const uint32_t integrator_flags)
{
  sim->KE = sim->rotKE = sim->PE = 0.0;
  sim->maxdisp1 = sim->maxdisp2 = 0.0;

  if (integrator_flags & IFLAGS_SETUP_COEFFS     ) RETURN_ON_ERROR( ts_setup_collision_coeffs(sim) );

  if (integrator_flags & IFLAGS_SHAKE            ) RETURN_ON_ERROR( setup_constraints(sim) );

  if (integrator_flags & IFLAGS_ALLOC_RAND_FORCES) RETURN_ON_ERROR( ts_alloc_rand_forces(sim) );
  if (integrator_flags & IFLAGS_RANDOM_FORCES    ) RETURN_ON_ERROR( get_random_forces(sim) );
  if (integrator_flags & IFLAGS_RANDOM_TORQUES   ) RETURN_ON_ERROR( get_random_torques(sim) );

  if (integrator_flags & IFLAGS_PRECALC_FORCES   ) RETURN_ON_ERROR( ts_calculate_forces(sim, &(sim->PE), NO_TERM_GROUPS, TERMWISE_FALSE) );
  if (integrator_flags & IFLAGS_PRECALC_RIGIDS   ) RETURN_ON_ERROR( ts_rigid_bodies_calculate_all_forces_and_torques(sim) );

  return TS_OK;
}

static inline TSResult integrator_finish(ts_sim* sim, const uint32_t integrator_flags)
{
  if (integrator_flags & IFLAGS_SHAKE) RETURN_ON_ERROR( apply_constraints(sim) );

  sim->totalE = sim->PE + sim->KE + sim->rotKE;
  sim->time += sim->dt;

  if ((sim->nonbond_cutoff > 0) && (sim->nonbond_padding > 0) && (sim->maxdisp1 + sim->maxdisp2 > sim->nonbond_padding))
  {
    TS_LOG(sim, TS_LOG_INFO, TS_LOG_NONBONDS, 
        "Rebuilding nonbond list (maxdisp1 = %f, maxdisp2 = %f, padding =%f)\n",
        sim->maxdisp2, sim->maxdisp1, sim->nonbond_padding);
    RETURN_ON_ERROR( ts_build_nonbonded_potential_terms(sim) );
  }

  return TS_OK;
}

static inline void vverlet_update_positions_and_velocities(ts_sim* sim, size_t pind, double inv_mass, double dt)
{
  double force_factor = 0.5*dt*inv_mass;

  vector3_add_multiple_to(PARTICLE_VEL(sim, pind), force_factor, PARTICLE_FORCE(sim, pind));

  vector3_add_multiple_to(PARTICLE_POS(sim, pind), dt, PARTICLE_VEL(sim, pind));
}

static inline double vverlet_update_velocities(ts_sim* sim, size_t pind, double inv_mass, double dt)
{
  double force_factor = 0.5*dt*inv_mass;

  vector3_add_multiple_to(PARTICLE_VEL(sim, pind), force_factor, PARTICLE_FORCE(sim, pind));

  double vsq = vector3_sqnorm( PARTICLE_VEL(sim, pind) );
  return vsq;
}

static void update_max_displacements(ts_sim* sim, size_t pind)
{
  if ((sim->nonbond_cutoff <= 0.0) || (sim->nonbond_padding <= 0.0)) return;

  vector3 dr;
  double dist = vector3_distance(PARTICLE_MAXDISP_ORIGIN(sim,pind),PARTICLE_POS(sim,pind),&dr);
  if (dist > sim->maxdisp1)
  {
    sim->maxdisp2 = sim->maxdisp1;
    sim->maxdisp1 = dist;
  }
  else if (dist > sim->maxdisp2)
  {
    sim->maxdisp2 = dist;
  }
}

static TSResult vverlet_single_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  RETURN_ON_ERROR( integrator_start(sim, VVERLET_FLAGS) );

  for(size_t j=0; j < dof_map->num_free_particles; j++)
  {
    size_t pind = dof_map->fp_to_p[j];
    ts_particle_type* ptype = sim->particles[pind].type;
    vverlet_update_positions_and_velocities(sim, pind, ptype->inv_mass, sim->dt);
  }

  RETURN_ON_ERROR( ts_calculate_forces(sim, &(sim->PE), NO_TERM_GROUPS, TERMWISE_FALSE) );

  for(size_t j=0; j < dof_map->num_free_particles; j++)
  {
    size_t pind = dof_map->fp_to_p[j];
    ts_particle_type* ptype = sim->particles[pind].type;
    double vsq = vverlet_update_velocities(sim, pind, ptype->inv_mass, sim->dt);
    sim->KE += 0.5*ptype->mass*vsq;

    update_max_displacements(sim, pind);
  }

  return integrator_finish(sim, VVERLET_FLAGS);
}

static inline TSResult vverlet_first_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  RETURN_ON_ERROR( integrator_start(sim, VVERLET_FLAGS | IFLAGS_PRECALC_FORCES) );

  return vverlet_single_step(sim,dof_map);
}

static TSResult leapfrog_single_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  RETURN_ON_ERROR( integrator_start(sim, LEAPFROG_FLAGS) );

  for(size_t j=0; j < dof_map->num_free_particles; j++)
  {
    size_t pind = dof_map->fp_to_p[j];
    ts_particle_type* ptype = sim->particles[pind].type;

    double coeff = 0.5*sim->dt*ptype->inv_mass;
    vector3_add_multiple_to(PARTICLE_VEL( sim, pind ), coeff, PARTICLE_FORCE( sim, pind ));

    double KE = 0.5*ptype->mass*vector3_sqnorm(PARTICLE_VEL( sim, pind ));
    sim->KE += KE;

    vector3_add_multiple_to(PARTICLE_VEL( sim, pind ), coeff, PARTICLE_FORCE( sim, pind ));
    vector3_add_multiple_to(PARTICLE_POS( sim, pind ), sim->dt, PARTICLE_VEL( sim, pind ));

    update_max_displacements(sim, pind);
  }

  return integrator_finish(sim, LEAPFROG_FLAGS);
}

static TSResult leapfrog_first_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  RETURN_ON_ERROR( integrator_start(sim, LEAPFROG_FLAGS) );

  for(size_t j=0; j < dof_map->num_free_particles; j++)
  {
    size_t pind = dof_map->fp_to_p[j];
    ts_particle_type* ptype = sim->particles[pind].type;
    double KE = 0.5*ptype->mass*vector3_sqnorm(PARTICLE_VEL(sim, pind));
    sim->KE += KE;

    vector3_add_multiple_to( PARTICLE_VEL(sim, pind), 
        0.5*sim->dt*ptype->inv_mass, PARTICLE_FORCE(sim, pind));
    vector3_add_multiple_to( PARTICLE_POS(sim, pind), 
        sim->dt, PARTICLE_VEL(sim, pind));

    update_max_displacements(sim, pind);
  }

  return integrator_finish(sim, LEAPFROG_FLAGS);
}

static TSResult leapfrog_langevin_single_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  RETURN_ON_ERROR( integrator_start(sim, LEAPFROG_LANGEVIN_FLAGS) );

  for(size_t j=0; j < dof_map->num_free_particles; j++)
  {
    size_t pind = dof_map->fp_to_p[j];

    ts_particle_type* ptype = sim->particles[pind].type;

    sim->rand_forces[3*pind + 0] *= ptype->rand_coeff;
    sim->rand_forces[3*pind + 1] *= ptype->rand_coeff;
    sim->rand_forces[3*pind + 2] *= ptype->rand_coeff;

    vector3_displace(PARTICLE_FORCE(sim, pind), 
        sim->rand_forces[3*pind + 0],
        sim->rand_forces[3*pind + 1],
        sim->rand_forces[3*pind + 2]);

    vector3_scaleby(PARTICLE_VEL(sim, pind), ptype->coll_coeff1);

    double coeff = 0.5*sim->dt*ptype->inv_mass*ptype->coll_coeff2;
    vector3_add_multiple_to(PARTICLE_VEL( sim, pind ), coeff, PARTICLE_FORCE( sim, pind ));

    double KE = 0.5*ptype->mass*vector3_sqnorm(PARTICLE_VEL( sim, pind ));
    sim->KE += KE;

    vector3_add_multiple_to(PARTICLE_VEL( sim, pind ), coeff, PARTICLE_FORCE( sim, pind ));
    vector3_add_multiple_to(PARTICLE_POS( sim, pind ), sim->dt, PARTICLE_VEL( sim, pind ));

    update_max_displacements(sim, pind);
  }

  return integrator_finish(sim, LEAPFROG_LANGEVIN_FLAGS);
}

static TSResult leapfrog_langevin_first_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  UNUSED_PARAM(dof_map);

  RETURN_ON_ERROR( integrator_start(sim, LEAPFROG_LANGEVIN_FLAGS | IFLAGS_SETUP_COEFFS | IFLAGS_ALLOC_RAND_FORCES) );

  for(size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle_type* ptype = sim->particles[i].type;
    double KE = 0.5*ptype->mass*vector3_sqnorm(PARTICLE_VEL(sim, i));
    sim->KE += KE;

    double coeff1 = 1 - 0.25*ptype->collfreq*sim->dt;
    double coeff2 = 1.0 / (1 + 0.25*ptype->collfreq*sim->dt);
    coeff1 *= coeff2;

    sim->rand_forces[3*i + 0] *= ptype->rand_coeff;
    sim->rand_forces[3*i + 1] *= ptype->rand_coeff;
    sim->rand_forces[3*i + 2] *= ptype->rand_coeff;

    vector3_displace(PARTICLE_FORCE(sim, i), sim->rand_forces[3*i + 0], sim->rand_forces[3*i + 1], sim->rand_forces[3*i + 2]);

    vector3_scaleby(PARTICLE_VEL(sim, i), coeff1);

    vector3_add_multiple_to( PARTICLE_VEL(sim, i), 
        0.5*sim->dt*ptype->inv_mass*coeff2, PARTICLE_FORCE(sim, i));

    vector3_add_multiple_to( PARTICLE_POS(sim, i), 
        sim->dt, PARTICLE_VEL(sim, i));

    update_max_displacements(sim, i);
  }

  return integrator_finish(sim, LEAPFROG_LANGEVIN_FLAGS);
}

static inline double bd_ori_update(ts_sim* sim, size_t rbind, double dt)
{
  vector3 omega;
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);

  vector3_scale_and_add(&omega, 
      body->type_ptr->inv_rotfric, &body->torque, 
      1.0, &body->rand_torque);

  quat4 dq;
  quat4 ori;
  quat4_set(&ori, RIGID_ORI(sim,rbind));
  quat4_dq_from_omega(&dq, &ori, dt, &omega);
  quat4_add_to(&ori, &dq);
  quat4_normalize(&ori);
  quat4_set(RIGID_ORI(sim,rbind),&ori);

  return 0.5*body->type_ptr->I0*vector3_sqnorm(&omega);
}

static inline void SRK_ang_vel_predict(ts_sim* sim, size_t rbind, double*restrict rand_torques, double dt)
{
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);
  vector3_set_with(&body->rand_torque, 
      body->type_ptr->bd_rand_coeff*rand_torques[0],
      body->type_ptr->bd_rand_coeff*rand_torques[1],
      body->type_ptr->bd_rand_coeff*rand_torques[2]);
  quat4_set(RIGID_PREV_ORI(sim,rbind), RIGID_ORI(sim,rbind));
  vector3_set(RIGID_PREV_TORQUE(sim,rbind), RIGID_TORQUE(sim,rbind));

  bd_ori_update(sim, rbind, dt);
}

static inline double SRK_ang_vel_correct(ts_sim* sim, size_t rbind, double*restrict rand_torques, double dt)
{
  UNUSED_PARAM(rand_torques);

  vector3_set_to_midpoint(RIGID_TORQUE(sim,rbind), RIGID_PREV_TORQUE(sim,rbind));
  quat4_set(RIGID_ORI(sim,rbind), RIGID_PREV_ORI(sim,rbind));

  return bd_ori_update(sim, rbind, dt);
}

static inline void SRK_predict_angular(ts_sim*restrict sim, double*restrict rand_torques)
{
  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    SRK_ang_vel_predict(sim, i, rand_torques, sim->dt);
    rand_torques += 3;
  }

  ts_update_rigid_bodies(sim);
}

static inline double SRK_correct_angular(ts_sim*restrict sim, double*restrict rand_torques)
{
  double rotKE = 0.0;
  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    rotKE += SRK_ang_vel_correct(sim, i, rand_torques, sim->dt);
    rand_torques += 3;
  }
  ts_update_rigid_bodies(sim);

  return rotKE;
}

static TSResult stoch_runge_kutta_single_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  UNUSED_PARAM(dof_map);

  vector3* actual_forces = sim->vl_forces;
  sim->vl_forces = sim->vl_work;

  RETURN_ON_ERROR( integrator_start(sim, STOCH_RUNGE_KUTTA_FLAGS) );

  for(size_t i=0; i < sim->num_particles; i++)
  {
    if (ts_pind_is_rigid(sim,i)) continue;

    ts_particle_type* ptype = sim->particles[i].type;

    sim->rand_forces[3*i + 0] *= ptype->bd_rand_coeff;
    sim->rand_forces[3*i + 1] *= ptype->bd_rand_coeff;
    sim->rand_forces[3*i + 2] *= ptype->bd_rand_coeff;

    vector3_set(PARTICLE_VEL(sim,i), PARTICLE_POS(sim,i));
    vector3_add_multiple_to( PARTICLE_POS(sim, i), 
        ptype->brownian_coeff, PARTICLE_FORCE(sim, i));
    vector3_displace(PARTICLE_POS(sim, i), 
      sim->rand_forces[3*i + 0], sim->rand_forces[3*i + 1], sim->rand_forces[3*i + 2]);
  }

  SRK_predict_angular(sim, sim->rand_torques);

  sim->vl_forces = actual_forces;

  double PE0 = 0.0;
  RETURN_ON_ERROR( ts_calculate_forces(sim, &PE0, NO_TERM_GROUPS, TERMWISE_FALSE) );
  RETURN_ON_ERROR( ts_rigid_bodies_calculate_all_forces_and_torques(sim) );

  for(size_t i=0; i < sim->num_particles; i++)
  {
    if (ts_pind_is_rigid(sim,i)) continue;

    ts_particle_type* ptype = sim->particles[i].type;

    vector3 df;

    vector3_sub(&df, PARTICLE_FORCE(sim,i), sim->vl_work + i);
    vector3_scaleby(&df, 0.5);
    vector3_set_to_midpoint(PARTICLE_FORCE(sim,i), sim->vl_work + i);

    vector3_add_multiple_to( PARTICLE_POS(sim, i), 
        ptype->brownian_coeff, &df);

    vector3 last_pos;
    vector3_set(&last_pos, PARTICLE_VEL(sim,i));
    vector3_sub(PARTICLE_VEL(sim,i), PARTICLE_POS(sim,i), &last_pos);
    vector3_scaleby(PARTICLE_VEL(sim,i), sim->inv_dt);
    sim->KE += 0.5 * ptype->mass * vector3_sqnorm(PARTICLE_VEL(sim,i));

    vector3_add_to(&sim->net_force, PARTICLE_FORCE(sim,i));

    update_max_displacements(sim, i);
  }

  sim->rotKE = SRK_correct_angular(sim, sim->rand_torques);

  return integrator_finish( sim, STOCH_RUNGE_KUTTA_FLAGS );
}

static inline TSResult stoch_runge_kutta_first_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  return stoch_runge_kutta_single_step(sim,dof_map);
}

static inline TSResult none_single_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  UNUSED_PARAM(dof_map);

  RETURN_ON_ERROR( integrator_start(sim, NO_INTEGRATOR_FLAGS) );

  return integrator_finish(sim, NO_INTEGRATOR_FLAGS);
}

static inline TSResult none_first_step(ts_sim* sim, const ts_dof_map* dof_map)
{
  return none_single_step(sim,dof_map);
}

#define TAKE_STEP(func)                       \
  do {                                        \
    RETURN_ON_ERROR( func(sim,dof_map) );     \
    sim->integrator_steps++;                  \
    RETURN_ON_ERROR( ts_run_callbacks(sim) ); \
  } while (0)

#define RUN_NSTEPS(prefix)                      \
  do {                                          \
    size_t i=0;                                 \
    if (sim->time == 0.0)                       \
    {                                           \
      TAKE_STEP( SYMCAT(prefix,_first_step) );  \
      i++;                                      \
    }                                           \
    while( i < nsteps )                         \
    {                                           \
      TAKE_STEP( SYMCAT(prefix,_single_step) ); \
      i++;                                      \
    }                                           \
  } while(0)

static TSResult many_steps_inner(ts_sim* sim, size_t nsteps, const ts_dof_map* dof_map)
{
  ERROR_IF( TS_ERR_BAD_DT, sim->dt <= 0.0 );
  sim->inv_dt = 1.0 / sim->dt;

  if (nsteps > 0)
  {
    switch (sim->integrator)
    {
      case TS_INTEGRATOR_VELOCITY_VERLET: RUN_NSTEPS(vverlet);  break;
      case TS_INTEGRATOR_LEAPFROG_VERLET: RUN_NSTEPS(leapfrog); break;
      case TS_INTEGRATOR_NONE:            RUN_NSTEPS(none);     break;
      case TS_INTEGRATOR_LEAPFROG_LANGEVIN: RUN_NSTEPS(leapfrog_langevin); break;
      case TS_INTEGRATOR_STOCH_RUNGE_KUTTA: RUN_NSTEPS(stoch_runge_kutta); break;
      default:
        return TS_NEW_ERROR( TS_ERR_UNKNOWN_INTEGRATOR );
    }
  }

  return TS_OK;
}
#undef RUN_NSTEPS
#undef TAKE_STEP

TSResult ts_sim_many_steps(ts_sim* sim, size_t nsteps)
{
  ts_dof_map dof_map;
  RETURN_ON_ERROR( ts_dof_map_init(&dof_map, sim) );
#define CLEANUP_ON_ERROR { ts_dof_map_finalize(&dof_map); }

  RETURN_ON_ERROR3(
      many_steps_inner(sim, nsteps, &dof_map)
      );

#undef CLEANUP_ON_ERROR
  ts_dof_map_finalize(&dof_map);

  return TS_OK;
}


TSResult ts_sim_single_step(ts_sim* sim)
{
  return ts_sim_many_steps(sim,1);
}

TSResult ts_sim_step_until(ts_sim* sim, double t_end)
{
  ERROR_IF( TS_ERR_BAD_DT, sim->dt <= 0.0);

  size_t nsteps = floor( (t_end - sim->time) / sim->dt );
  return ts_sim_many_steps(sim, nsteps);
}

