#include "ts_io.h"

#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>

#include "ts_common.h"
#include "ts_integrators.h"
#include "ts_malloc_protect.h"
#include "ts_json.h"

// We depend on this throughout this file
ASSERT_SIZE_EQ( double, uint64_t );
ASSERT_SIZE_EQ( float , uint32_t );

struct tspf_pos_entry
{
  uint64_t x;
  uint64_t y;
  uint64_t z;
};

union fp_byte_access
{
  uint8_t   u8v[sizeof(double) / sizeof(uint8_t)];
  uint32_t u32v[sizeof(double) / sizeof(uint32_t)];

  uint64_t u64v;
  double fval;
};

static void vec2entry(struct tspf_pos_entry*restrict entry, const vector3*restrict v)
{
  entry->x = dbl_to_u64( VECTOR3_X(v) );
  entry->y = dbl_to_u64( VECTOR3_Y(v) );
  entry->z = dbl_to_u64( VECTOR3_Z(v) );
}

/*
static void entry2vec_noswap(vector3*restrict v, const struct tspf_pos_entry*restrict entry)
{
  VECTOR3_X(v) = u64_to_dbl( entry->x );
  VECTOR3_Y(v) = u64_to_dbl( entry->y );
  VECTOR3_Z(v) = u64_to_dbl( entry->z );
}

static void entry2vec_swap(vector3*restrict v, const struct tspf_pos_entry*restrict entry)
{
  VECTOR3_X(v) = u64_to_dbl( swap_byteorder_64(dbl_to_u64(entry->x)) );
  VECTOR3_Y(v) = u64_to_dbl( swap_byteorder_64(dbl_to_u64(entry->y)) );
  VECTOR3_Z(v) = u64_to_dbl( swap_byteorder_64(dbl_to_u64(entry->z)) );
}
*/

static void initialize_header(struct ts_file_header* hdr, ts_magic_t magic, uint8_t vmajor, uint8_t vminor)
{
  hdr->magic = magic;
  hdr->swapcheck = SWAP_CHECK;

  hdr->version[0] = vmajor;
  hdr->version[1] = vminor;

  hdr->flags = 0;

  memset(hdr->reserved, 0, sizeof(hdr->reserved));
}

static TSResult ts_file_header_check_data(struct ts_file_header* hdr, 
    ts_magic_t magic, uint8_t vmajor, uint8_t vminor, int* needswap_ptr)
{
  if (hdr->magic.i != magic.i)
  {
    return TS_NEW_ERROR(TS_ERR_BAD_FILE_FORMAT);
  }

  if ((hdr->swapcheck != SWAP_CHECK) && (hdr->swapcheck != REV_SWAP_CHECK))
  {
    return TS_NEW_ERROR(TS_ERR_BAD_FILE_FORMAT);
  }

  needswap_ptr[0] = (hdr->swapcheck == REV_SWAP_CHECK);

  if ((hdr->version[0] > vmajor) || 
      ((hdr->version[0] == vmajor) && (hdr->version[1] < vminor)))
  {
    return TS_NEW_ERROR(TS_ERR_UNKNOWN_FILE_VERSION);
  }

  return TS_OK;
}

ts_magic_t ts_str_to_magic(const char* mstr)
{
  ts_magic_t m = { .i = 0 };
  for (int i=0; i < 4; i++) { 
    if (mstr[i]) { m.c[i] = mstr[i]; }
    else { break; }
  }
  return m;
}

TSResult tspf_check_file_header(FILE* f)
{
  ts_magic_t tspf_magic = ts_str_to_magic(TSPF_MAGIC);

  RETURN_ON_ERROR( ts_check_file_header(f,
        tspf_magic,
        TSPF_MAJOR,
        TSPF_MINOR) );

  return TS_OK;
}

TSResult ts_check_file_header(FILE* f, ts_magic_t magic, uint8_t vmajor, uint8_t vminor)
{
  struct ts_file_header hdr;
  memset(&hdr, 0, sizeof(hdr));

  if (fread(&hdr, sizeof(hdr), 1, f) != 1) return TS_NEW_ERROR(TS_ERR_READ_NOT_FINISHED);

  int needswap = 0;
  {
    RETURN_ON_ERROR( ts_file_header_check_data(&hdr, magic, vmajor, vminor, &needswap) );
  }

  if (needswap) { return TS_NEW_ERROR(TS_ERR_FILE_IS_BYTESWAPPED); }

  return TS_OK;
}

TSResult ts_write_header(FILE* f, ts_magic_t magic, uint8_t vmajor, uint8_t vminor)
{
  UNUSED_PARAM(vmajor);
  UNUSED_PARAM(vminor);

  struct ts_file_header hdr;

  initialize_header(&hdr, magic, TSPF_MAJOR, TSPF_MINOR);
  if (fwrite(&hdr, sizeof(hdr), 1, f) != 1) return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED);

  return TS_OK;
}

TSResult tspf_write_header(FILE* f)
{
  RETURN_ON_ERROR( 
      ts_write_header(f, 
        ts_str_to_magic(TSPF_MAGIC), TSPF_MAJOR, TSPF_MINOR) );

  return TS_OK;
}

TSResult tspf_set_timepoint_header(struct tspf_timepoint_header* hdr, const ts_sim* sim)
{
  hdr->n = (uint32_t)sim->num_particles;

  hdr->t = sim->time;
  hdr->temp = sim->temp;
  hdr->KE = sim->KE;
  hdr->PE = sim->PE;
  hdr->totE = sim->totalE;

  return TS_OK;
}

TSResult tspf_write_positions(ts_sim* sim, FILE* out)
{
  struct tspf_timepoint_header tphdr;
  struct tspf_pos_entry entry;

  tspf_set_timepoint_header(&tphdr, sim);
  fwrite(&tphdr, sizeof(tphdr), 1, out);

  for(size_t i=0; i < sim->num_particles; i++)
  {
    vec2entry( &entry, PARTICLE_POS(sim, i) );
    fwrite(&entry, sizeof(entry), 1, out);
  }

  return TS_OK;
}

static TSResult read_header(FILE*restrict in, struct tspf_timepoint_header*restrict header)
{
  if (fread(header, sizeof(*header), 1, in) != 1) return TS_NEW_ERROR(TS_ERR_READ_NOT_FINISHED);

  return TS_OK;
}

static TSResult read_positions(FILE*restrict in, size_t n, double*restrict xv, double*restrict yv, double*restrict zv)
{
  uint32_t nrecs = n;

  for (uint32_t i = 0; i < nrecs; i++)
  {
    struct tspf_pos_entry entry;
    if (fread(&entry, sizeof(entry), 1, in) != 1) return TS_NEW_ERROR(TS_ERR_READ_NOT_FINISHED);
    xv[i] = u64_to_dbl( entry.x );
    yv[i] = u64_to_dbl( entry.y );
    zv[i] = u64_to_dbl( entry.z );
  }

  return TS_OK;
}

TSResult tspf_read_positions_into_veclist(FILE* in, struct tspf_timepoint_header* header, VectorList* veclist)
{
  tsAssert((in != NULL) && "input FILE cannot be NULL");
  tsAssert((header != NULL) && "header cannot be NULL");
  tsAssert((veclist != NULL) && "veclist cannot be NULL");

  RETURN_ON_ERROR( read_header(in, header) );

  if (veclist_init(veclist, header->n) < 0) {
    TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  return read_positions(in, veclist->n, veclist->vX, veclist->vY, veclist->vZ);
}

TSResult tspf_read_positions(FILE* in, struct tspf_timepoint_header* header, size_t n, double*restrict xv, double*restrict yv, double*restrict zv)
{
  tsAssert((in != NULL) && "input FILE cannot be NULL");
  tsAssert((header != NULL) && "header cannot be NULL");
  tsAssert((xv != NULL) && "xv cannot be NULL");
  tsAssert((yv != NULL) && "yv cannot be NULL");
  tsAssert((zv != NULL) && "zv cannot be NULL");

  RETURN_ON_ERROR( read_header(in, header) );

  if (n != header->n) return TS_NEW_ERROR(TS_ERR_INSUFFICIENT_ARRAY_SIZE);

  return read_positions(in, n, xv,yv,zv);
}

#if 0
/*
*  Function:  wrpdb
  *  ------------------------------------------------
  *  Writes out the current particle data in PDB format
  *  for visualization.
 */

  bool_t wrpdb(char *ofil, int nx, int ny, int npart, pd_t pd[],
species_t splist[], char *msg) {

  int i, ix, iy, iz, nxp2, npln ;
  FILE *fp ;

  nxp2 =  nx + 2*LATTADD ;
  npln = (ny + 2*LATTADD) * nxp2 ;

  if (((fp = fopen(ofil,"w")) == NULL)) {
    strcat(strcpy(msg, "File error for file: "), ofil) ;
    return FALSE ;
  }

  fprintf (fp, "%d\n", npart) ;
  for (i = 1 ; i <= npart ; i++) {
    getxyz(pd[i].ind, nxp2, npln, &ix, &iy, &iz) ;

    fprintf (fp, "ATOM %6d  %-4s%-5s%4d", i, "X", splist[pd[i].spec].name, i) ;
    fprintf (fp, "    %8.3f%8.3f%8.3f %5.2f%6.2f         TCR \n",
      3.0*ix, 3.0*iy, 3.0*iz, (float) pd[i].state, (float) pd[i].nbond) ;
  }

  fclose(fp) ;
  return TRUE ;
}
#endif /* 0 */

TSResult ts_write_pdb(ts_sim* sim, FILE* out)
{
  fprintf(out, "%lu\n", sim->num_particles);

  for (size_t i=0; i < sim->num_particles; i++) {
    ts_particle* particle = ts_get_particle(sim, i);
    ts_particle_type* ptype = particle->type;
    const vector3* pos = PARTICLE_POS(sim, i);

    const char* fullname = ptype->name;
    char shortname[6] = {0};
    strncpy(shortname, fullname, 5);
    shortname[5] = '\0';

    fprintf(out, "ATOM %6lu  %-4s%-5s%4lu", i+1, "X", shortname, i+1);
    fprintf(out, "    %8.3f%8.3f%8.3f %5.2f%6.2f         TCR \n",
        vector3_x(pos), 
        vector3_y(pos), 
        vector3_z(pos),
        (double)1,  /* particle state */
        (double)particle->num_bonds);
  }

  return TS_OK;
}

/* removes any non-alphanumeric characters */
static void cleanup_particle_name(const char* orig, char* new, size_t n)
{
  char* end = new+n-1;
  char* s = new;

  while (*orig && s < end)
  {
    if (isalnum(*orig))
    {
      *s = *orig;
      s++;
    }
    orig++;
  }

  *s = '\0';
}

TSResult ts_write_xyz(ts_sim* sim, FILE* out, const char* comment)
{
  fprintf(out, "%lu\n", (unsigned long)sim->num_particles);
  fprintf(out, "time %f step %lu : %s\n",
      (double)sim->time, (unsigned long)sim->integrator_steps, comment);

  for (size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle* particle = ts_get_particle(sim, i);
    ts_particle_type* ptype = particle->type;
    const vector3* pos = PARTICLE_POS(sim, i);

    char clean_name[9];
    const char* fullname = ptype->name;
    cleanup_particle_name(fullname, clean_name, sizeof(clean_name));

    /* if the name is bad (no alphanumeric characters), default to 'X'.
     * NB: We should probably log this to warn the user.
     */
    if (clean_name[0] == '\0') { clean_name[0] = 'X'; clean_name[1] = '\0'; }

    fprintf(out, "%-8s %10.5f %10.5f %10.5f\n",
        clean_name, vector3_x(pos), vector3_y(pos), vector3_z(pos));
  }

  return TS_OK;
}

#if 0
bool_t wrpsf(char *ofil, int nx, int ny, int npart, pd_t pd[],
species_t splist[], char *msg) {

  int i, j, nb = 0 ;
  FILE *fp ;

  if (((fp = fopen(ofil,"w")) == NULL)) {
    strcat(strcpy(msg, "File error for file: "), ofil) ;
    return FALSE ;
  }

  fprintf (fp, "PSF\n\n1 !NTITLE\n*\n\n") ;
  fprintf (fp, "%d !NATOM\n", npart) ;
  for (i = 1 ; i <= npart ; i++) {

    fprintf (fp, "%8d TCR %4d  %-5s%-2s  %3d",
      i, i, splist[pd[i].spec].name, "X", pd[i].spec) ;
    fprintf (fp, "   %8.6f       %7.5f           0\n",
      (float) pd[i].state, (float) pd[i].nbond) ;
    nb += pd[i].nbond ;
  }
  nb /= 2 ;
  fprintf (fp, "\n%d !NBOND\n", nb) ;
  nb = 0 ;
  for (i = 1 ; i <= npart ; i++) {
    for (j = 0 ; j < pd[i].nbond ; j++) {
      if (i < *(pd[i].bondp + j)) {
        fprintf (fp, "%8d%8d", i, *(pd[i].bondp + j)) ;
        ++nb ;
        if (nb % 4 == 0)
          fprintf(fp, "\n") ;
      }
    }
  }
  fprintf(fp, "\n") ;

  fclose(fp) ;
  return TRUE ;
}
#endif /* 0 */

TSResult ts_write_psf(ts_sim* sim, FILE* out)
{
  fprintf (out, "PSF\n\n1 !NTITLE\n*\n\n") ;
  fprintf (out, "%lu !NATOM\n", sim->num_particles) ;

  size_t num_bonds = 0;
  for (size_t i = 0; i < sim->num_particles; i++) {
    ts_particle* particle = ts_get_particle(sim, i);
    ts_particle_type* ptype = particle->type;

    if (ptype == NULL) {
      TS_LOG(sim, TS_LOG_ERROR, TS_LOG_IO, "particle %lu has NO type\n");
      return TS_NEW_ERROR( TS_ERR_SIMULATION_STATE );
    }

    const char* fullname = ptype->name;
    char shortname[6] = {0};
    strncpy(shortname, fullname, 5);
    shortname[5] = '\0';

    fprintf(out, "%8lu TCR %4lu  %-5s%-2s  %3lu",
        i+1,i+1, shortname, "X", ts_particle_type_index(sim,particle->type));
    fprintf(out, "   %8.6f       %7.5f           0\n",
        (double)1, /* state*/
        (double)particle->num_bonds);
    num_bonds += particle->num_bonds;
  }

  tsAssert( num_bonds % 2 == 0 );
  num_bonds /= 2;

  fprintf(out, "\n%lu !NBOND\n", num_bonds) ;
  num_bonds = 0 ;
  for (size_t pi=0; pi < sim->num_particles; pi++) {
    ts_particle* particle = ts_get_particle(sim, pi);
    for (size_t b=0; b < particle->num_bonds; b++) {
      if (particle == particle->bonds[b].pi) {
        size_t pj = TS_PIND(sim, particle->bonds[b].pj);
        // if (pi < pj) {
        fprintf(out, "%8lu%8lu", pi+1, pj+1);
        if (++num_bonds % 4 == 0) fprintf(out, "\n");
      }
    }
  }

  fprintf(out, "\n") ;

  return TS_OK;
}

static TSResult f77_unformatted_write(void* p, uint32_t size, FILE* f)
{
  fwrite(&size, sizeof(size), 1, f);
  fwrite(p,size,1,f);
  fwrite(&size, sizeof(size), 1, f);

  return TS_OK;
}

static TSResult write_title(FILE* f, const char* title0, const char* title1)
{
  union { char tdata[164]; uint32_t ntitle; } data;
  memset(data.tdata+4, ' ', 160);

  data.ntitle = 2;

  int l0 = strlen(title0);
  memcpy(data.tdata+4 , title0, l0 <= 80 ? l0 : 80);

  int l1 = strlen(title1);
  memcpy(data.tdata+84, title1, l1 <= 80 ? l1 : 80);

  f77_unformatted_write(data.tdata, 164, f);

  return TS_OK;
}

TSResult ts_write_dcd_header(ts_sim* sim, FILE* out, 
    uint32_t step0, uint32_t dstep, const char* title0, const char* title1)
{
  union { char cmagic[4]; uint32_t imagic; } magic;
  union { float fdt; uint32_t idt; } dt;

  magic.cmagic[0] = 'C';
  magic.cmagic[1] = 'O';
  magic.cmagic[2] = 'R';
  magic.cmagic[3] = 'D';

  uint32_t hdr[21] = { 0 };
  hdr[0] = magic.imagic;

  hdr[1] = 1; /* num files */
  hdr[2] = step0;
  hdr[3] = dstep;
  hdr[9] = 0; /* natom - nfreat */

  dt.fdt = sim->dt;
  hdr[10] = dt.idt;

  f77_unformatted_write(hdr, sizeof(hdr), out);
  write_title(out, title0, title1);

  uint32_t natoms = sim->num_particles;
  f77_unformatted_write(&natoms, sizeof(natoms), out);

  return TS_OK;
}

TSResult ts_write_dcd_entry(ts_sim* sim, FILE* out)
{
  const uint32_t rec_size = sizeof(float) * sim->num_particles;
  float* tmp = ts_alloc(NO_ALLOCATOR, rec_size);
#define CLEANUP_ON_ERROR { if (tmp) { ts_free(NO_ALLOCATOR, tmp); } }
  ERROR_ON_NULL_ALLOC(tmp);

  for(size_t i=0; i < sim->num_particles; i++) {
    tmp[i] = vector3_x(PARTICLE_POS(sim,i));
  }
  fwrite(&rec_size, sizeof(rec_size), 1, out);
  fwrite(tmp, sizeof(float), sim->num_particles, out);
  fwrite(&rec_size, sizeof(rec_size), 1, out);

  for(size_t i=0; i < sim->num_particles; i++) {
    tmp[i] = vector3_y(PARTICLE_POS(sim,i));
  }
  fwrite(&rec_size, sizeof(rec_size), 1, out);
  fwrite(tmp, sizeof(float), sim->num_particles, out);
  fwrite(&rec_size, sizeof(rec_size), 1, out);

  for(size_t i=0; i < sim->num_particles; i++) {
    tmp[i] = vector3_z(PARTICLE_POS(sim,i));
  }
  fwrite(&rec_size, sizeof(rec_size), 1, out);
  fwrite(tmp, sizeof(float), sim->num_particles, out);
  fwrite(&rec_size, sizeof(rec_size), 1, out);

#undef CLEANUP_ON_ERROR

  ts_free(NO_ALLOCATOR, tmp);

  return TS_OK;
}

struct dcd_callback_data {
  ts_integrator_callback parent;
  FILE* out;
};

static TSResult dcd_writer_callback(ts_sim* sim, struct dcd_callback_data* data)
{
  if (data->out != NULL)
  {
    TSResult err = ts_write_dcd_entry(sim, data->out);
    if (!IS_OKAY(err))
    {
      fprintf(stderr, 
          "Error in DCD writer (callback): %x "
          "[Subsequent writes suppressed]\n", err.error);
      data->out = NULL;

      RETURN_ON_ERROR(err);
    }
  }

  return TS_OK;
}

TSResult ts_add_dcd_callback(ts_sim* sim, FILE* out,
    uint32_t freq, const char* title0, const char* title1)
{
  RETURN_ON_ERROR( ts_write_dcd_header(sim, out, 1, freq, title0, title1) );

#define CLEANUP_ON_ERROR { ts_safe_free(NO_ALLOCATOR, cbdata); }
  struct dcd_callback_data* cbdata = ts_alloc(NO_ALLOCATOR, sizeof(*cbdata));
  ERROR_ON_NULL_ALLOC(cbdata);

  cbdata->parent.countdown = 0;
  cbdata->parent.freq      = freq;
  cbdata->parent.callback  = (TSIntegratorCallbackFunc)dcd_writer_callback;
  cbdata->out = out;

  RETURN_ON_ERROR3(ts_add_callback(sim, (ts_integrator_callback*)cbdata));
#undef CLEANUP_ON_ERROR

  return TS_OK;
}


TSResult ts_write_all_bonds(ts_sim* sim, FILE* outf)
{
  fprintf(outf,"Particles\n");
  for(size_t i=0; i < sim->num_particles; i++)
  {
    fprintf(outf, "%lu %s\n", i+1, TS_PARTICLE(sim,i)->type->name);
  }

  fprintf(outf,"\nBonds\n");
  const char* dtbl = "UFB*";

  for (size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle* p = TS_PARTICLE(sim,i);
    for (size_t b=0; b < p->num_bonds; b++) {
      if (p->bonds[b].pi == p) {
        fprintf(outf, "%lu %lu 0x%04x [ 0x%04x %c ]\n",
            TS_PIND(sim,p->bonds[b].pi),
            TS_PIND(sim,p->bonds[b].pj),
            bondtype_to_int(p->bonds[b].type),
            bondtype_get_type(p->bonds[b].type),
            dtbl[(int)bondtype_get_direction(p->bonds[b].type)]);
      }
    }
  }

  return TS_OK;
}

static TSResult print_potential_list_summary(
    ts_sim* sim, FILE* f, const char* title, struct ts_potential_term_list* plist)
{
  if (plist->terms.size == 0) return TS_OK;

  double en = 0.0;
  RETURN_ON_ERROR( ts_potential_energies_resum(sim, plist, &en, 0) );

  fprintf(f, "---[ %s ]---\n", title);
  size_t ind0 = 0;
  double total_energy = 0.0;
  size_t total_count = 0;
  fprintf(f, "|%-15s|%-37s|%5s|%7s|%7s|\n",
      "Potential Type",
      "Term name",
      "Count",
      "TotalE",
      "E/term");
  for (size_t i=0; i < plist->potential_indices.size; i++)
  {
    ts_potential_term* term = DATAVECTOR_AT(ts_potential_term, &plist->terms, ind0);
    size_t next_ind = *(DATAVECTOR_AT(size_t, &plist->potential_indices, i));

    ptrdiff_t count = next_ind - ind0;
    fprintf(f, "|%-15s|%-37s|%5ld|%7.2f|%7.4e|\n",
        term->potential->name,
        term->name,
        count,
        term->potential->potential_energy,
        term->potential->potential_energy/count);
    ind0 = next_ind;
    total_energy += term->potential->potential_energy;
    total_count += count;
  }
  fprintf(f, "|%-15s|%-37s|%5ld|%7.2f|%7.4e|\n",
      "TOTAL",
      "Total potential energy",
      total_count,
      total_energy,
      total_energy/total_count);
  fprintf(f,"\n");

  return TS_OK;
}

TSResult ts_print_potential_summary(ts_sim* sim, FILE* f)
{
  RETURN_ON_ERROR( print_potential_list_summary(sim, f, "BONDED", &sim->bondlist) );
  RETURN_ON_ERROR( print_potential_list_summary(sim, f, "BONDED", &sim->nonbond_list) );

  return TS_OK;
}

TSResult ts_sim_print_summary(ts_sim* sim, FILE* f)
{
  fprintf(f,"--------------------------------------------------------------------------------\n");
  fprintf(f,"Step: %lu  Time: %f  Temp: %f\n", sim->integrator_steps, sim->time, sim->temp);

  if (ts_integrator_is_brownian(sim)) {
    fprintf(f,"PE: %f\n", sim->PE);
  } else {
    fprintf(f, "KE: %f  KE(rot): %f  PE: %f  TotE: %f\n", sim->KE, sim->rotKE, sim->PE, sim->totalE);
  }

  RETURN_ON_ERROR( ts_print_potential_summary(sim,f) );

  return TS_OK;
}

static TSResult ts_fprintf(FILE* f, const char* fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  int ret = vfprintf(f, fmt, args);
  va_end(args);

  if (ret >= 0)
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("Error encountered in write: %s", strerror(errno)));
}

MAYBE_UNUSED static TSResult
write_json_string(FILE* f, const char* key, const char* value)
{
  if (key)
    RETURN_WITH_RESULT( ts_fprintf(f, "\"%s\" : \"%s\"", key, value) );
  else
    RETURN_WITH_RESULT( ts_fprintf(f, "\"%s\"", value) );

  return TS_OK;
}

MAYBE_UNUSED static TSResult
write_json_literal(FILE* f, const char* key, enum TS_JSON_LITERAL lit)
{
  const char* litstr = NULL;
  if (lit == TS_JSON_NULL)
    litstr = "null";
  else if (lit == TS_JSON_FALSE)
    litstr = "false";
  else if (lit == TS_JSON_TRUE)
    litstr = "true";
  else
    return TS_NEW_ERROR2(TS_ERR_INVALID_VALUE,
        ("invalid value for JSON literal (%d)", (int)lit));

  if (key)
    RETURN_WITH_RESULT( ts_fprintf(f, "\"%s\" : %s", key, litstr) );
  else
    RETURN_WITH_RESULT( ts_fprintf(f, "%s", litstr) );

  return TS_OK;
}

MAYBE_UNUSED static TSResult
write_json_number(FILE* f, const char* key, double n)
{
  if (key)
    RETURN_WITH_RESULT( ts_fprintf(f, "\"%s\" : %f", key, n) );
  else
    RETURN_WITH_RESULT( ts_fprintf(f, "%f", n) );

  return TS_OK;
}

#define JS_MAP_OPEN(k) \
    RETURN_ON_ERROR( ts_fprintf(f, "\"%s\" : {\n", k) )

#define JS_MAP_CLOSE(k) \
    RETURN_ON_ERROR( ts_fprintf(f, "}") )

#define JS_ARRAY_OPEN(k) \
    RETURN_ON_ERROR( ts_fprintf(f, "\"%s\" : [\n", k) )

#define JS_ARRAY_CLOSE(k) \
    RETURN_ON_ERROR( ts_fprintf(f, "]") )

#define JS_COMMA   RETURN_ON_ERROR( ts_fprintf(f, ",\n") )

#define JS_WRITE(type,k,v) \
  RETURN_ON_ERROR( SYMCAT(write_json_,type)(f, k, v) )

TSResult ts_write_json_state(ts_sim* sim, FILE* f)
{
  RETURN_ON_ERROR( ts_fprintf(f, "{\n") );

  JS_MAP_OPEN( "global" );

  JS_WRITE(number, "n"   , sim->num_particles ); JS_COMMA;
  JS_WRITE(number, "t"   , sim->time          ); JS_COMMA;
  JS_WRITE(number, "temp", sim->temp          ); JS_COMMA;
  JS_WRITE(number, "KE"  , sim->KE            ); JS_COMMA;
  JS_WRITE(number, "PE"  , sim->PE            ); JS_COMMA;
  JS_WRITE(number, "totE", sim->totalE        );

  JS_MAP_CLOSE("state");
  JS_COMMA;

  JS_ARRAY_OPEN( "bonds" );
  JS_ARRAY_CLOSE( "bonds" );

  RETURN_ON_ERROR( ts_fprintf(f, "}\n") ); // global

  return TS_OK;
}

