#ifndef TS_IO__H
#define TS_IO__H

#include <stdio.h>

#include "tubesheet.h"
#include "ts_veclist.h"

#define MODULE_NAME TS_IO

// TSpf -- T(ube) S(heet) p(osition) f(ile)
#define TSPF_MAGIC "TSpf"
#define TSPF_MAJOR 0x01
#define TSPF_MINOR 0x01

// TSdf -- T(ube) S(heet) d(ata) f(ile)
#define TSDF_MAGIC "TSdf"
#define TSDF_MAJOR 0x01
#define TSDF_MINOR 0x01

// TSsf -- _T_ube _S_heet s_t_ate _f_ile
#define TSTF_MAGIC "TStf"
#define TSTF_MAJOR 0x01
#define TSTF_MINOR 0x01


// swap check -- 16-bit entry that is written as 0x4142.  If the entry
// is read as 0x4241, then the position file was written on a computer
// with a different endian order than the computer that is reading the
// file.  The byte order of all entries after the version numbers must
// be swapped.
#define SWAP_CHECK 0x4142
#define REV_SWAP_CHECK 0x4241

typedef union {
  char c[4];
  uint32_t i;
} ts_magic_t;

struct ts_file_header
{
  ts_magic_t magic;      // mark our file turf
  uint16_t swapcheck;    // make sure endianness of reader agrees with writer
  uint8_t version[2];    // bytes for major and minor version

  uint32_t flags;        // options for file (should currently be zero)
  uint32_t reserved[4];  // space reserved for expansion
};

typedef struct _tspf TSPF;

struct tspf_timepoint_header
{
  uint32_t n;            // number of entries

  double t;
  double temp;

  double KE;
  double PE;
  double totE;
};

ts_magic_t ts_str_to_magic(const char *mstr);

TSResult ts_check_file_header(FILE *f, ts_magic_t magic, uint8_t vmajor, uint8_t vminor);
TSResult ts_write_header(FILE *f, ts_magic_t magic, uint8_t vmajor, uint8_t vminor);

TSResult tspf_check_file_header(FILE *f);
TSResult tspf_write_header(FILE *f);

TSResult tspf_set_timepoint_header(struct tspf_timepoint_header *hdr, const ts_sim *sim);

TSResult tspf_read_positions_into_veclist(FILE *in, struct tspf_timepoint_header *header, VectorList *veclist);
TSResult tspf_read_positions(FILE *in, struct tspf_timepoint_header *header, size_t n,
    double *restrict xv, double *restrict yv, double *restrict zv);
TSResult tspf_write_positions(ts_sim *sim, FILE *out);


TSResult ts_write_psf(ts_sim *sim, FILE *out);
TSResult ts_write_pdb(ts_sim *sim, FILE *out);
TSResult ts_write_xyz(ts_sim *sim, FILE *out, const char *comment);

TSResult ts_write_dcd_header(ts_sim *sim, FILE *out, 
    uint32_t step0, uint32_t dstep, const char *title0, const char *title1);
TSResult ts_write_dcd_entry(ts_sim *sim, FILE *out);


TSResult ts_add_dcd_callback(ts_sim *sim, FILE *out,
    uint32_t freq, const char *title0, const char *title1);


TSResult ts_write_all_bonds(ts_sim *sim, FILE *outf);

TSResult ts_print_potential_summary(ts_sim *sim, FILE *f);
TSResult ts_sim_print_summary(ts_sim *sim, FILE *f);

TSResult ts_write(ts_sim *sim, FILE *f);
ts_sim* ts_read(FILE *f);

TSResult ts_write_json_state(ts_sim *sim, FILE *f);

enum TSTF_HEADER_FLAGS {
  TSTF_HDRFLAG_NONE = 0,
};


struct tsif_bond_info
{
  size_t obj_i;
  size_t obj_j;
  size_t pi;
  size_t pj;
  ts_bond_type bt;
};

struct tsif_state
{
  size_t num_bonds;
  size_t num_particles;
  size_t num_rigids;

  struct tsif_bond_info* bonds;

  float *particle_pos;
  
  float *rigid_pos;
  float *rigid_ori;  // Orientations are stored as axis-angle
};

TSResult tsif_write_state(ts_sim *sim, FILE *f);
TSResult tsif_read_state(FILE *f, struct tsif_state *state, ts_allocator *A);


struct tstf_state_header
{
  struct tspf_timepoint_header sim_state;

  uint32_t num_free_particles;
  uint32_t num_rigid_bodies;

  uint32_t flags;
  char reserved[16];  /* reserved for future expansion */
};

TSResult tstf_check_header(FILE *f);
TSResult tstf_write_header(FILE *f);

/*! \brief Writes the state of the simulation to a file stream
 *
 * Note that a TSTF header must be written before a state for the file
 * to be a valid TSTF file.  The header is not required to reconstruct
 * the state information, but validates the file format.
 *
 * The structure of the state information is as follows:
 *
 *    Header section
 *    Bonding section
 *    Position information section
 *    Additional information section
 * 
 */
TSResult tstf_write_state(ts_sim *sim, FILE *f);

struct tstf_file;

struct tstf_file *tstf_open(FILE *f, ts_allocator *alloc, TSResult *errp);
void tstf_close(struct tstf_file *f);

TSResult tstf_read_state_header(struct tstf_file *f,
    struct tstf_state_header *hdr);

struct tstf_bond_info
{
  size_t obj_i;
  size_t obj_j;

  size_t pi;
  size_t pj;

  ts_bond_type bt;
};

TSResult tstf_start_reading_bonds(struct tstf_file *f);
TSResult tstf_read_bonds(struct tstf_file *f,
    size_t n, struct tstf_bond_info *binfo, size_t *nread);
TSResult tstf_end_reading_bonds(struct tstf_file *f);

typedef struct TSTF TSTF;
struct TSTF
{
  size_t max_bonds;
  size_t max_particles;
  size_t max_rigids;

  size_t num_bonds;
  size_t num_particles;
  size_t num_rigids;

  size_t *bonds;

  vector3 *ppos;

  vector3 *rpos;
  vector3 *rori;  // Orientations are stored as axis-angle
};

// TSResult tstf_init(TSTF *tstf, size_t maxb, size_t maxp, size_t maxr);
// TSResult tstf_finalize(TSTF *tstf);

// TSResult tstf_read_state(TSTF *tstf, FILE *f);


// TSmf -- _T_ube _S_heet _m_echanical _f_ile
#define TSMF_MAGIC "TSmf"
#define TSMF_SECTION_MAGIC "MF*S"
#define TSMF_MAJOR 0x01
#define TSMF_MINOR 0x01

enum TSMF_FLAGS {
  TSMF_FLAG_NONE = 0x0000,
};

/* TSMF header
 *
 * size type    desc
 * 4    chars   magic
 * 1    u8      major
 * 1    u8      minor
 * 2    u16     flags
 * 8    u64     ioff - offset of the first index
 * 16   chars   padding (for future versions)
 *
 * total: 32 bytes
 *
 * note that if ioff is zero, the header has no pointer to the first
 * index
 */

#define TSMF_INDEX_SUFFIX "IDX*"
struct tsmf_file_header
{
  char magic[4];
  uint8_t version[2];
  uint16_t flags;
  uint64_t ioff;
  uint8_t padding[16];
};

TSResult tsmf_read_header(FILE *f, struct tsmf_file_header *hdr);

struct tsmf_index;

enum TSMF_FILE_MODE {
  // First bit specifies the file mode:
  TSMF_READWRITE = 0x0000,       // Read-write: modifies an existing file
  TSMF_READONLY  = 0x0001,       // Read-only: only read the file.
};

struct tsmf_info
{
  ts_sim *sim;
  FILE *f;

  uint16_t mode;

  uint16_t flags;      // flags set in the file
  uint64_t ioff;

  uint32_t seqno;

  struct tsmf_index *indices;

  // for delta encoding of connectivity
  struct tsmf_conn_list *clist;
};

/* Opens a tsmf file, associating it with a mechanical simulation.
 *
 * The mode argument affects how the file is treated:
 *
 *   TSMF_READWRITE
 *      If the file is empty, creates a new tsmf file.
 *      Otherwise the file must be valid (with a valid header).  New
 *      data is appended, although existing incomplete indices may be
 *      updated.
 *
 *   TSMF_READONLY
 *      Only opens the file for reading.  No modifications are allowed.
 *
 * The mode argument may also contain flags.
 */
struct tsmf_info *tsmf_new(ts_sim *sim, FILE *f, uint16_t mode, TSResult *errp);

/* Closes the tsmf file, cleans up any resources
 *
 * All data should be written to disk before the various write calls
 * return, so this isn't necessary for data integrity.
 */
TSResult tsmf_close(struct tsmf_info *tsmf);

// TSResult tsmf_write_header(struct tsmf_info *hdr);

/* TSMF section header -- all sections start with this
 *
 * size type    desc
 * 4    chars   section header magic
 * 2    u16     section header type
 * 2    u16     section header flags
 * 4    u32     length
 * 4    chars   unused
 *
 * 16 bytes
 */

enum
{
  TSMF_SXN_HEADER_SIZE = 16,
};

enum TSMF_SECTION_TYPE
{
  TSMF_SEC_INDEX = 1,
  TSMF_SEC_METADATA,
  TSMF_SEC_PARTICLE_TYPES,
  TSMF_SEC_RB_TYPES,
  TSMF_SEC_PARTICLE_LIST,
  TSMF_SEC_RB_LIST,
  TSMF_SEC_CONNECTIVITY,
  TSMF_SEC_COORDINATES,
};

struct tsmf_section
{
  uint16_t type;
  uint16_t flags;
  uint32_t length;
  uint64_t soff;
};
#define TSMF_SECTION_ZERO ((struct tsmf_section){0})

/* TSMF index section
 *
 * data after the section header:
 * size         desc
 * 4            num entries
 * 12           unused
 *
 * following this are num_entries of the following entries:
 *
 * 2            section type
 * 2            unused
 * 4            sequence
 * 8            section offset (from beginning of file)
 *
 * this is followed by:
 * size type    desc
 * 4    bytes   index magic
 * 4    bytes   unused
 * 8    bytes   index offset
 *
 * notes:
 *
 * The sequence number is used by the simulation to sequence related
 * sections together.
 *
 * 1. entries must be in ascending order, ordered by section offset
 * 2. section sequences must be in ascending order in the file.
 * 3. entries with the same sequence number must be in the same index
 *    section, and must be stored together in the file.
 * 4. The index may have empty entries.  The data type for these must
 *    be zero
 */

// enum { TSMF_INDEX_NENTRIES = 256 };

struct tsmf_index_entry
{
  uint16_t type;
  uint16_t unused;
  uint32_t seqno;
  uint64_t soff;
};

struct tsmf_index_section
{
  uint32_t num_entries;
  uint32_t max_entries;
  uint64_t ioff;
  uint64_t fsize;
  unsigned char unused[8];

  struct tsmf_index_entry entries[];
};

struct tsmf_index
{
  uint32_t len;
  uint32_t cap;
  int dirty;

  struct tsmf_index_entry entries[];
};

/* TSMF metadata section
 *
 * stores length-encoded metadata entries
 *
 * metadata entries are key-value pairs.
 *
 * data after section header:
 * size         desc
 * 4            num short entries
 * 4            num kv entries
 * 4            unused
 *
 * following this are `num short entries` of short metadata entries:
 *
 * size         desc
 * 1            key - key code
 * 1            vtype - value type
 * 4            vcount - value length
 * vlen         value data
 *
 * following this are `num kv entries` of kv metadata:
 * 
 * size         desc
 * 4            klen  - key length
 * 4            vlen  - value length
 * 4            unused
 * klen         key data
 * vlen         value data
 *
 * following this, the metadata section is padded
 * with 0-15 bytes of 'M' so that its length will
 * be a multiple of 16.
 */

enum TSMF_METADATA_KEY
{
  TSMF_METADATA_COMMENT = 0,            // Unkeyed "comment"
  TSMF_METADATA_TIME    = 1,            // Time entry
  TSMF_METADATA_NSTEPS  = 2,            // Number of steps
  TSMF_METADATA_POT_ENERGY = 3,         // Potential energy
  TSMF_METADATA_KIN_ENERGY = 4,         // Kinetic energy
  TSMF_METADATA_TOT_ENERGY = 5,         // Total energy
  TSMF_METADATA_NET_FORCE  = 6,         // Net force
  TSMF_METADATA_NET_TORQUE = 7,         // Net torque
};

enum TSMF_METADATA_VALUE_TYPE
{
  TSMF_METADATA_VALUE_EMPTY   =  0,
  TSMF_METADATA_VALUE_STRING  =  1,
  TSMF_METADATA_VALUE_BYTES   =  2,

  // For fixed-width types, the count is implicit in the length
  TSMF_METADATA_VALUE_S16     =  3,
  TSMF_METADATA_VALUE_S32     =  4,
  TSMF_METADATA_VALUE_S64     =  5,
  TSMF_METADATA_VALUE_U16     =  6,
  TSMF_METADATA_VALUE_U32     =  7,
  TSMF_METADATA_VALUE_U64     =  8,
  TSMF_METADATA_VALUE_F32     =  9,
  TSMF_METADATA_VALUE_F64     = 10,

  // CURRENTLY UNSUPPORTED:
  //
  // Varint and Uvarint encodings are count-encoded as well as
  // length-encoded.  The first value is a uvarint that gives the count.
  // TSMF_METADATA_VALUE_VARINT  = 12,
  // TSMF_METADATA_VALUE_UVARINT = 13,
};

struct tsmf_metadata_header
{
  uint32_t doff;        // offset to data
  uint16_t nshort;      // number of short entries
  uint16_t nkv;         // number of kv entries
  uint32_t unused;
};

struct tsmf_metadata_short_entry
{
  uint8_t key_code;
  uint8_t vtype;
  uint32_t vcount;
  unsigned char vdata[];
};

struct tsmf_metadata_kv_entry
{
  uint32_t klen;
  uint32_t vlen;
  uint8_t unused[4];
  unsigned char kvdata[];
};

struct tsmf_metadata_info
{
  struct tsmf_info *tsmf;
  TSResult err;

  size_t slen;
  size_t scap;
  struct tsmf_metadata_short_entry **sdata;

  size_t kvlen;
  size_t kvcap;
  struct tsmf_metadata_kv_entry **kvdata;
};

TSResult tsmf_metadata_initialize(
    struct tsmf_info *tsmf,
    struct tsmf_metadata_info *md);

void tsmf_metadata_finalize(struct tsmf_metadata_info *md);

static inline int tsmf_metadata_error(struct tsmf_metadata_info *md)
{
  return IS_ERROR(md->err);
}

TSResult tsmf_metadata_add_short(
    struct tsmf_metadata_info *md,
    enum TSMF_METADATA_KEY key,
    enum TSMF_METADATA_VALUE_TYPE vtype,
    void *base,
    size_t vcount);

TSResult tsmf_metadata_add_kv(
    struct tsmf_metadata_info *md,
    const unsigned char *key, size_t klen,
    const unsigned char *val, size_t vlen);

TSResult tsmf_metadata_read(struct tsmf_metadata_info *md);
TSResult tsmf_metadata_write(struct tsmf_metadata_info *md);

/* TSMF particle types section
 *
 * data after section header:
 * size         desc
 * 4            num_types
 * 2            flags
 * 10           unused
 * 16 bytes
 *
 * then num_types entries:
 * size         desc
 * 2            lnamelen - length of long name
 * 1            namelen  - length of name
 * 5            unused
 * 8            mass   (double)
 * 8            charge (double)
 * 8            radius (double)
 * 8            collfreq (double)
 * namelen      name data
 * lnamelen     long name data
 *
 * 40+namelen+lnamelen bytes
 *
 * then padding with 0-15 bytes of '$' so the length of the section is a
 * multiple of 16.
 */

enum 
{
  TSMF_PTYPES_HEADER_SIZE      = 16,
  TSMF_PTYPES_ENTRY_FIXED_SIZE = 40,
};

void tsmf_ptypes_free(struct ts_particle_type *ptypes, uint32_t n);

TSResult tsmf_ptypes_write(struct tsmf_info *tsmf);
TSResult tsmf_ptypes_read(struct tsmf_info *tsmf,
    struct ts_particle_type **ptypesp, uint32_t *np);

/* TSMF rigid types section
 *
 * data after section header:
 * size         desc
 * 4            num_types
 * 4            num_sites - total number of sites
 * 2            flags
 * 6            unused
 * 16 bytes total
 *
 * then num_types entries:
 * size         desc
 * 2            namelen  - length of name
 * 2            rigid body type
 * 4            num_sites
 * 8            I0 (double)
 * 8            rotfric (double)
 * 24 bytes total
 *
 * 8            site1_dx
 * 8            site1_dy
 * 8            site1_dz
 * ...
 * 8            siteN_dx        for num_sites sites
 * 8            siteN_dy
 * 8            siteN_dz
 * namelen      name data
 * 0-7          padding to align to 8 bytes
 *
 * then padding with 0-15 bytes of '$' so the length of the section is a
 * multiple of 16.
 */

enum 
{
  TSMF_RBTYPES_HEADER_SIZE      = 16,
  TSMF_RBTYPES_ENTRY_FIXED_SIZE = 24,
  TSMF_RBTYPES_SITE_SIZE        = 24,
};

struct tsmf_rbtypes_buffer
{
  uint32_t tlen;
  uint32_t tcap;
  struct ts_rigid_body_type *rbtypes;

  uint32_t vlen;
  uint32_t vcap;
  vector3 *vlist;
};

TSResult tsmf_btypes_buffer_initialize(struct tsmf_rbtypes_buffer *buf,
    uint32_t tcap, uint32_t vcap);
void tsmf_btypes_buffer_finalize(struct tsmf_rbtypes_buffer *buf);
TSResult tsmf_btypes_buffer_resize(struct tsmf_rbtypes_buffer *buf,
    uint32_t tcap, uint32_t vcap);

TSResult tsmf_rbtypes_write(struct tsmf_info *tsmf);
TSResult tsmf_rbtypes_read(struct tsmf_info *tsmf,
    struct tsmf_rbtypes_buffer *buf);

/* TSMF particle list section
 *
 * after header:
 * size         desc
 * 4            num_particles
 * 2            flags
 * 10           unused
 * 16 bytes total
 *
 * then num_particles entries of:
 * 4    particle_type
 * 4    particle_flags
 * 
 * then padding with 0-15 bytes of '$' so the length of the section is a
 * multiple of 16.
 */

enum
{
  TSMF_PLIST_HEADER_SIZE = 16,
  TSMF_PLIST_ENTRY_SIZE  =  8,
};

struct tsmf_plist_buf
{
  size_t n;
  uint32_t flags;

  struct {
    uint32_t type;
    uint32_t flags;
  } entries[];
};

TSResult tsmf_plist_write(struct tsmf_info *tsmf);
TSResult tsmf_plist_read(struct tsmf_info *tsmf,
    struct tsmf_plist_buf **plistp);

/* TSMF rigid body list section
 *
 * after header:
 * size         desc
 * 4            num_rigid_bodies
 * 4            num_particle_entries
 * 2            flags
 * 6            unused
 *
 * then num_rigid_bodies entries:
 * 4            rigid body type
 * 4            flags
 * 4            particle index 1
 * 4            particle index 2
 * ...
 * 4            particle index N  <-- N is specified by the rigid body type
 *
 * then padding with 0-15 bytes of '$' so the length of the section is a
 * multiple of 16.
 */

enum 
{
  TSMF_RBLIST_HEADER_SIZE      = 16,
  TSMF_RBLIST_ENTRY_FIXED_SIZE = 8,
  TSMF_RBLIST_ENTRY_SITE_SIZE  = 4,
};

struct tsmf_rblist_buf
{
  size_t nrigid;
  uint32_t flags;

  size_t npart;
  uint32_t *plist;

  struct {
    uint32_t type;
    uint32_t flags;
    uint32_t pind0;
  } entries[];
};

struct tsmf_rblist_buf *tsmf_rblist_buf_new(uint32_t nrigid, uint32_t npart);
void tsmf_rblist_buf_delete(struct tsmf_rblist_buf *buf);

TSResult tsmf_rblist_write(struct tsmf_info *tsmf);
TSResult tsmf_rblist_read(struct tsmf_info *tsmf,
    struct tsmf_rblist_buf **rblistp);

/* TSMF connectivity lists
 *
 * after header:
 * size         desc
 * 4            num_bonds
 * 4            flags
 * 8            unused
 *
 * then num_particle_connections entries:
 * size         desc
 * uint32_t     particle index i
 * uint32_t     particle index j
 * uint32_t     bond type
 *
 * then padding with 0-15 bytes of '$' so the length of the section is a
 * multiple of 16.
 */

enum
{
  TSMF_CONN_FLAGS_DELTA = 0x0001,
};

enum 
{
  TSMF_CONN_HEADER_SIZE = 16,
  TSMF_CONN_ENTRY_SIZE  = 12,
};

struct tsmf_conn_list
{
  uint32_t nbonds;
  uint32_t flags;
  struct {
    uint32_t pi;
    uint32_t pj;
    uint32_t bt;
  } bonds[];
};

struct tsmf_conn_list *tsmf_conn_list_new(uint32_t nbonds, uint32_t flags);
void tsmf_conn_list_delete(struct tsmf_conn_list *clist);

TSResult tsmf_conn_write(struct tsmf_info *tsmf);
TSResult tsmf_conn_write_delta(struct tsmf_info *tsmf);
TSResult tsmf_conn_read(struct tsmf_info *tsmf,
    struct tsmf_conn_list **clistp);

/* TSMF coordinate lists
 *
 * after header:
 * size         desc
 * 4            num_particle_dofs
 * 4            num_rigid_body_dofs
 * 4            num_rot_rigid_body_dofs
 * 2            flags
 * 10            unused
 *
 * then num_particle_dofs entries of:
 * size         desc
 * 8            x       - x coordinate
 * 8            y       - y coordinate
 * 8            z       - z coordinate
 *
 * then num_rigid_body_dofs entries of:
 * size         desc
 * 8            x       - com x coordinate
 * 8            y       - com y coordinate
 * 8            z       - com z coordinate
 * 8            rx      - axis-angle x component
 * 8            ry      - axis-angle y component
 * 8            rz      - axis-angle z component
 *
 * then num_rot_rigid_body_dofs entries of:
 * size         desc
 * 8            rx      - axis-angle x component
 * 8            ry      - axis-angle y component
 * 8            rz      - axis-angle z component
 */

enum 
{
  TSMF_COORDS_HEADER_SIZE   = 24,
  TSMF_COORDS_RIGID_SIZE    = 6*sizeof (double),
  TSMF_COORDS_PARTICLE_SIZE = 3*sizeof (double),
  TSMF_COORDS_ROTATE_SIZE   = 3*sizeof (double),
};

enum 
{
  TSMF_COORDS_FLAG_ALL = 0x0001,  /* all degrees (including frozen) */
};

struct tsmf_coords_header
{
  uint32_t num_particle_dof;
  uint32_t num_rigid_dof;
  uint32_t num_rot_rigid_dof;
  uint16_t flags;
};

struct tsmf_coords_buffer
{
  uint16_t flags;

  uint32_t nparticle;
  uint32_t nrigid;
  uint32_t nrotate;
  uint32_t cap;
  double *coords;
};

struct tsmf_coords_buffer *tsmf_coords_buffer_new(uint32_t cap);
int tsmf_coords_buffer_resize(struct tsmf_coords_buffer *cbuf, uint32_t cap);
void tsmf_coords_buffer_release(struct tsmf_coords_buffer *cbuf);
void tsmf_coords_buffer_delete(struct tsmf_coords_buffer *cbuf);
uint32_t tsmf_coords_buffer_ndof(struct tsmf_coords_buffer *cbuf);

TSResult tsmf_coords_write_sim(struct tsmf_info *tsmf);
TSResult tsmf_coords_write_sim_all(struct tsmf_info *tsmf);
TSResult tsmf_coords_write_coords(struct tsmf_info *tsmf,
    struct tsmf_coords_buffer *cbuf);
TSResult tsmf_coords_read(struct tsmf_info *tsmf,
    struct tsmf_coords_buffer *cbuf);
TSResult tsmf_coords_read_into_sim(struct tsmf_info *tsmf);


#undef MODULE_NAME

#endif /* TS_IO__H */

