#include "ts_io.h"
#include "tubesheet.h"
#include "ts_bonds.h"

static TSResult tsdf_check_file_header(FILE* f)
{
  ts_magic_t tsdf_magic = ts_str_to_magic(TSDF_MAGIC);

  RETURN_ON_ERROR( ts_check_file_header(f,
        tsdf_magic,
        TSDF_MAJOR,
        TSDF_MINOR) );

  return TS_OK;
}

static TSResult tsdf_write_header(FILE* f)
{
  RETURN_ON_ERROR( ts_write_header(f, 
        ts_str_to_magic(TSDF_MAGIC),
        TSDF_MAJOR,
        TSDF_MINOR) );

  return TS_OK;
}

#define OP_READ(op)  SYMCAT(READ_,op)
#define OP_WRITE(op) SYMCAT(WRITE_,op)

#define READ_FIELD(f,elt) \
  do { if (fread(&(elt), sizeof(elt), 1, f) != 1) { \
    return TS_NEW_ERROR(TS_ERR_READ_NOT_FINISHED); \
  } } while(0)

#define WRITE_FIELD(f,elt) \
  do { if (fwrite(&(elt), sizeof(elt), 1, f) != 1) { \
    return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED); \
  } } while(0)

#define READ_PARTICLES(f,sim) \
  do { RETURN_ON_ERROR( read_particles(sim,f) ); } while(0)

#define WRITE_PARTICLES(f,sim) \
  do { RETURN_ON_ERROR( write_particles(sim,f) ); } while(0)


#define READ_PARTICLE_TYPES(f,sim) \
  do { RETURN_ON_ERROR( read_particle_types(sim,f) ); } while(0)

#define WRITE_PARTICLE_TYPES(f,sim) \
  do { RETURN_ON_ERROR( write_particle_types(sim,f) ); } while(0)


#define READ_BOND_TYPES(f,sim) \
  do { RETURN_ON_ERROR( read_bond_types(sim,f) ); } while(0)

#define WRITE_BOND_TYPES(f,sim) \
  do { RETURN_ON_ERROR( write_bond_types(sim,f) ); } while(0)


#define READ_BONDS(f,sim) \
  do { RETURN_ON_ERROR( read_bonds(sim,f) ); } while(0)

#define WRITE_BONDS(f,sim) \
  do { RETURN_ON_ERROR( write_bonds(sim,f) ); } while(0)


#define READ_ARRAY(f,elt,n) \
  do { if (fread(elt, sizeof(elt[0]), n, f) != n) { \
    return TS_NEW_ERROR(TS_ERR_READ_NOT_FINISHED); \
  } } while(0)

#define WRITE_ARRAY(f,elt,n) \
  do { if (fwrite(elt, sizeof(elt[0]), n, f) != n) { \
    return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED); \
  } } while(0)


#define READ_VECTOR(f, v) \
  do { if (fread((v), sizeof(*(v)), 1, f) != 1) { \
    return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED); \
  } } while(0)

#define WRITE_VECTOR(f, v) \
  do { if (fwrite((v), sizeof(*(v)), 1, f) != 1) { \
    return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED); \
  } } while(0)

static TSResult read_vector(FILE* f, vector3* v)
{
  double x = 0.0;
  double y = 0.0;
  double z = 0.0;

  READ_FIELD(f, x);
  READ_FIELD(f, y);
  READ_FIELD(f, z);

  VECTOR3_X(v) = x;
  VECTOR3_Y(v) = y;
  VECTOR3_Z(v) = z;

  return TS_OK;
}

static TSResult write_vector(FILE* f, const vector3* v)
{
  double x = VECTOR3_X(v);
  double y = VECTOR3_Y(v);
  double z = VECTOR3_Z(v);

  WRITE_FIELD(f,x);
  WRITE_FIELD(f,y);
  WRITE_FIELD(f,z);

  return TS_OK;
}


static TSResult read_string(FILE* f, char** sp)
{
  size_t nlen = 0;
  READ_FIELD(f,nlen);

  char* s = ts_alloc(NO_ALLOCATOR, (nlen+1)*sizeof(char));
  READ_ARRAY(f,s,nlen);
  s[nlen] = '\0';

  *sp = s;

  return TS_OK;
}

static TSResult write_string(FILE* f, const char* s)
{
  size_t nlen = strlen(s);
  WRITE_FIELD(f,nlen);
  WRITE_ARRAY(f,s,nlen);

  return TS_OK;
}


static TSResult read_particle_types(ts_sim* sim, FILE* f)
{
  size_t ntypes = 0;
  READ_FIELD(f, ntypes);

  for (size_t i=0; i < ntypes; i++)
  {
    char* name = NULL;
    char* long_name = NULL;
    RETURN_ON_ERROR( read_string(f, &name) );
    RETURN_ON_ERROR( read_string(f, &long_name) );

    ts_particle_type pt;
    READ_FIELD(f, pt);

    pt.name = name;
    pt.long_name = long_name;

    size_t ptind = 0;
    RETURN_ON_ERROR(
        ts_add_particle_type(sim, &pt, &ptind) );
  }

  return TS_OK;
}

static TSResult write_particle_types(ts_sim* sim, FILE* f)
{
  size_t ntypes = sim->particle_types.size;
  WRITE_FIELD(f, ntypes);

  for (size_t i=0; i < ntypes; i++)
  {
    ts_particle_type* ptype = ts_get_particle_type(sim, i+1);

    RETURN_ON_ERROR( write_string(f, ptype->name) );
    RETURN_ON_ERROR( write_string(f, ptype->long_name) );

    if (fwrite(ptype, sizeof(*ptype), 1, f) != 1) 
    {
      return TS_NEW_ERROR(TS_ERR_WRITE_NOT_FINISHED);
    }
  }

  return TS_OK;
}

static TSResult read_bond_types(ts_sim* sim, FILE* f)
{
  size_t num = 0;
  READ_FIELD(f, num);
  for (size_t i=0; i < num; i++)
  {
    char* btName = NULL;
    uint16_t btype = 0;

    RETURN_ON_ERROR( read_string(f, &btName) );
    READ_FIELD(f, btype);

    RETURN_ON_ERROR(
        ts_add_bond_type(sim, btName, 
          bondtype_from_int(btype)) );
  }

  return TS_OK;
}

struct write_bond_type_closure
{
  ts_sim* sim;
  FILE* f;
  TSResult result;
};

union bondtype_typecast_t { 
  void* ptr;
  ts_bond_type bt;
};

static TSResult write_bond_entry_data(FILE* f, 
    const char* key, ts_bond_type bt)
{
  RETURN_ON_ERROR( write_string(f, key) );
  WRITE_FIELD(f, bt);

  return TS_OK;
}

static TSResult
do_write_bond_type(void* ud, const char* btName, ts_bond_type bt)
{
  FILE* f = (FILE*) ud;
  RETURN_ON_ERROR(
      write_bond_entry_data(f, btName, bt)
      );

  return TS_OK;
}

static TSResult write_bond_types(ts_sim* sim, FILE* f)
{
  size_t nbonds = ts_count_bond_types(sim);
  WRITE_FIELD(f, nbonds);

  RETURN_ON_ERROR(
      ts_foreach_bond_type(sim, do_write_bond_type, (void*)f)
      );

  return TS_OK;
}

static TSResult read_particles(ts_sim* sim, FILE* f)
{
  size_t max_p = 0;
  READ_FIELD(f, max_p);
  ts_alloc_particles(sim,max_p);
  READ_FIELD(f, sim->num_particles);

  for(size_t i=0; i < sim->num_particles; i++)
  {
    size_t ptype = 0;
    READ_FIELD(f,ptype);
    ts_particle* p = TS_PARTICLE(sim,i);
    p->type = ts_get_particle_type(sim, ptype);

    READ_FIELD(f, p->flags);
    READ_FIELD(f, p->rigid_body_index);

    read_vector(f, PARTICLE_POS(sim,i));
    read_vector(f, PARTICLE_VEL(sim,i));
  }

  return TS_OK;
}

static TSResult write_particles(ts_sim* sim, FILE* f)
{
  WRITE_FIELD(f, sim->max_particles);
  WRITE_FIELD(f, sim->num_particles);

  for (size_t i=0; i < sim->num_particles; i++)
  {
    size_t ptype = TS_PTYPE_IND(sim,i);

    WRITE_FIELD(f, ptype);

    ts_particle* p = TS_PARTICLE(sim,i);

    WRITE_FIELD(f, p->flags);
    WRITE_FIELD(f, p->rigid_body_index);

    write_vector(f, PARTICLE_POS(sim,i));
    write_vector(f, PARTICLE_VEL(sim,i));
  }

  return TS_OK;
}

static TSResult read_bonds(ts_sim* sim, FILE* f)
{
  size_t count = 0;
  READ_FIELD(f, count);

  for (size_t i=0; i < count; i++)
  {
    size_t pi = 0;
    size_t pj = 0;
    ts_bond_type bt = bondtype_none();

    READ_FIELD(f, pi);
    READ_FIELD(f, pj);
    READ_FIELD(f, bt);

    ts_particle* pt_i = TS_PARTICLE(sim,pi);
    ts_particle* pt_j = TS_PARTICLE(sim,pj);

    RETURN_ON_ERROR( ts_particle_add_bond(pt_i, pt_j, bt) );
  }

  return TS_OK;
}

static size_t count_bonds(ts_sim* sim)
{
  size_t count = 0;

  for (size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle* p = TS_PARTICLE(sim,i);
    count += p->num_bonds;
  }

  return count / 2;
}

static TSResult write_bonds(ts_sim* sim, FILE* f)
{
  size_t count = count_bonds(sim);
  WRITE_FIELD(f, count);

  for (size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle* p = TS_PARTICLE(sim,i);
    for (size_t b=0; b < p->num_bonds; b++)
    {
      if (p->bonds[b].pi == p)
      {
        size_t pi = TS_PIND(sim, p->bonds[b].pi);
        size_t pj = TS_PIND(sim, p->bonds[b].pj);
        ts_bond_type bt = p->bonds[b].type;

        WRITE_FIELD(f, pi);
        WRITE_FIELD(f, pj);
        WRITE_FIELD(f, bt);
      }
    }
  }

  return TS_OK;
}

#define DO_FIELDS do { \
    OP(FIELD)(f, sim->temp             ); \
    OP(FIELD)(f, sim->integrator       ); \
    OP(FIELD)(f, sim->dt               ); \
    OP(FIELD)(f, sim->time             ); \
    OP(FIELD)(f, sim->integrator_steps ); \
    OP(FIELD)(f, sim->nonbond_cutoff   ); \
    OP(FIELD)(f, sim->nonbond_padding  ); \
    OP(FIELD)(f, sim->nonbonds_enabled ); \
    OP(FIELD)(f, sim->bonds_enabled    ); \
    OP(FIELD)(f, sim->logging_flags    ); \
    OP(PARTICLE_TYPES)(f, sim ); \
    OP(BOND_TYPES)    (f, sim ); \
    OP(PARTICLES)     (f, sim ); \
    OP(BONDS)         (f, sim ); \
  } while(0)

#define OP OP_WRITE
TSResult ts_write(ts_sim* sim, FILE* f)
{
  tsdf_write_header(f);

  DO_FIELDS;

  return TS_OK;
}
#undef OP

#define OP OP_READ
TSResult ts_read_into(ts_sim* sim, FILE* f)
{
  tsdf_check_file_header(f);

  DO_FIELDS;

  return TS_OK;
}
#undef OP

ts_sim* ts_read(FILE* f)
{
  ts_sim* sim = ts_sim_new();

  if ( IS_OKAY(ts_read_into(sim,f)) )
  {
    return sim;
  }
  else
  {
    ts_sim_delete(sim);
    return NULL;
  }
}



