#include "ts_json.h"

#include <ctype.h>
#include <string.h>
#include "ts_common_algo.h"

#define PARSE_ERROR(p,mesg, ...) do {        \
    return TS_NEW_ERROR2(TS_ERR_PARSE_ERROR, \
        ("error parsing JSON at line %d, column %d: " mesg, \
         p->line, p->col, __VA_ARGS__));     \
  } while(0)

static int buf_getc(ts_json_parser* parser)
{
  return (parser->bpos < parser->blen)
    ? (unsigned char)parser->buf[parser->bpos++]
    : EOF
    ;
}

static void buf_ungetc(ts_json_parser* parser, int c)
{
  if ((c != EOF) && (parser->bpos > 0)) { parser->bpos--; }
}

static inline int lit_eq(const char* s, size_t len, const char* lit, size_t llen)
{
  return (len == llen) && ( memcmp(s,lit,len) == 0 );
}

#define LIT_EQ(s, len, lstr) \
  ( lit_eq((s), (len), (lstr), sizeof(lstr)-1) )

TSResult ts_json_init_buf(ts_json_parser* parser,
    char* p, size_t len,
    void (*freefunc)(void*,void*),void* cl)
{
  parser->buf  = p;
  parser->blen = len;
  parser->bpos = 0;

  parser->parser_getc   = buf_getc;
  parser->parser_ungetc = buf_ungetc;

  parser->enable_literal_strings = 0;

  for (int i=0; i < TS_JSON_STACKSIZE; i++)
    parser->state[i] = TS_JSON_PARSER_BARE;

  parser->stack_top = 0;

  parser->prev_line = parser->line = 0;
  parser->prev_col  = parser->col  = 0;

  parser->free_func = freefunc;
  parser->free_func_data = cl;

  return TS_OK;
}

void ts_json_finalize(ts_json_parser* parser)
{
  if (parser->free_func)
  { 
    parser->free_func(parser->free_func_data, parser->buf);
  }
  else
  {
    free(parser->buf);
  }
}

static inline TSResult
json_push_state(ts_json_parser* parser, enum TS_JSON_PARSER_STATE state)
{
  if (parser->stack_top >= TS_JSON_STACKSIZE-1) {
    return TS_NEW_ERROR2(TS_ERR_INSUFFICIENT_SIZE,
        ("parser is out of stack space (max %d levels of nesting)", TS_JSON_STACKSIZE));
  }

  parser->stack_top++;
  parser->state[parser->stack_top] = state;
  return TS_OK;
}
#define JSON_PUSH_STATE(p,st) do { \
    RETURN_ON_ERROR( json_push_state(p,st) ); \
  } while(0)

#define JSON_SET_STATE(p,st) do { \
    (p)->state[(p)->stack_top] = (st); \
  } while(0)

static inline enum TS_JSON_PARSER_STATE
json_pop_state(ts_json_parser* parser)
{
  if (parser->stack_top == 0) { return TS_JSON_PARSER_BARE; }

  parser->stack_top--;
  return parser->state[ parser->stack_top ];
}
#define JSON_POP_STATE(p) ( json_pop_state(p) )

static inline enum TS_JSON_PARSER_STATE
json_peek_state(ts_json_parser* parser)
{
  return parser->state[ parser->stack_top ];
}
#define JSON_PEEK_STATE(p) ( json_peek_state(p) )

static inline int
json_getc(ts_json_parser* parser)
{
  int c = parser->parser_getc(parser);

  parser->prev_line = parser->line;
  parser->prev_col  = parser->col;

  parser->col++;
  if (c == '\n') { parser->line++; parser->col = 0; }

  return c;
}

static inline void
json_ungetc(ts_json_parser* parser, int c)
{
  parser->parser_ungetc(parser,c);
  parser->line = parser->prev_line;
  parser->col  = parser->prev_col;
}

static int skipws(ts_json_parser* parser)
{
  int c;
  do { c = json_getc(parser); } while((c != EOF) && isspace(c));
  return c;
}

static void set_literal_event(ts_json_event* event, enum TS_JSON_LITERAL lit)
{
  event->type = TS_JSON_LITERAL;
  event->value.lit = lit;
}

static void set_string_event(ts_json_event* event, char* s, size_t len)
{
  event->type = TS_JSON_STRING;
  event->value.str.p = s;
  event->value.str.len = len;
}

static int
uescape(const char* s)
{
  int uval = 0;
  int i;
  for (i=0; i<4; i++) {
    uval <<= 4;

    int u = toupper( s[i] );
    if (u >= '0' && u <= '9') 
    {
      uval |= (u - '0');
    }
    else if (u >= 'A' && u <= 'F')
    {
      uval |= (u - 'A' + 10);
    }
  }

  return uval;
}

static int to_utf8(char* s, int cp)
{
  if (cp < 0x80) {
    s[0] = cp;
    return 1;
  }
  else if (cp < 0x800) { 
    s[0] = 0x80|0x40|(cp>>6);
    s[1] = 0x80|(cp & 0x3f);
    return 2;
  }
  else if (cp < 0xffff) {
    s[0] = 0x80|0x40|0x20|(cp>>12);
    s[1] = 0x80|((cp>>6) & 0x3f);
    s[2] = 0x80|(cp & 0x3f);
    return 3;
  }
  else if (cp < 0x1ffff) {
    s[0] = 0x80|0x40|0x20|0x10|(cp>>18);
    s[1] = 0x80|((cp>>12) & 0x3f);
    s[2] = 0x80|((cp>> 6) & 0x3f);
    s[3] = 0x80|(cp & 0x3f);
    return 4;
  }
  else {
    return -1;
  }
}

static size_t
process_string_escapes(char* s, char* send)
{
  char* sinp = s;

  while ((sinp < send) && (sinp[0] != '\\')) { sinp++; }

  char* sout = sinp;
  while (sinp < send)
  {
    if (*sinp != '\\')
    {
      sout[0] = sinp[0];
      sinp++;
      sout++;
    }
    else
    {
      switch (sinp[1]) {
        case '\\': break;

        case '/': sout[0] = '/';  sout++; sinp += 2; break;
        case '"': sout[0] = '"';  sout++; sinp += 2; break;
        case 'n': sout[0] = '\n'; sout++; sinp += 2; break;
        case 'r': sout[0] = '\r'; sout++; sinp += 2; break;
        case 't': sout[0] = '\t'; sout++; sinp += 2; break;
        case 'b': sout[0] = '\b'; sout++; sinp += 2; break;
        case 'f': sout[0] = '\f'; sout++; sinp += 2; break;

        case 'u':
          sout += to_utf8(sout, uescape(s+2));
          sinp += 6;
          break;
      }
    }
  }

  return sout-s;
}

static TSResult
parse_string_inner(ts_json_parser* parser, ts_json_event* event)
{
  parser->spos = parser->bpos;

  int sline = parser->line;
  int scol  = parser->col;

  int has_escapes = 0;

  for(;;) {
    int c = json_getc(parser);
    if (c == EOF)
    {
      PARSE_ERROR(parser,
          "string is possibly unterminated (begins at line %d, col %d)",
          sline,scol);
    }
    else if (c == '"')
    {
      goto ret_string;
    }
    else if (c == '\\')
    {
      has_escapes = 1;
      c = json_getc(parser);
      switch (c) {
        case '\\':
        case '/':
        case '"':
        case 'n':
        case 'r':
        case 't':
        case 'b':
        case 'f':
          break;

        case 'u':
          {
            int i;
            for (i=0; i<4; i++) {
              int u = toupper( json_getc(parser) );
              if ((u >= '0' && u <= '9') || (u >= 'A' && u <= 'F')) 
              {}
              else
              {
                PARSE_ERROR(parser, "invalid character in \\uXXXX sequence: %c", u);
              }
            }
          }
          break;

        default:
          PARSE_ERROR(parser, "invalid escape character: %c", c);
      }
    }
  }

ret_string:
  {
    char* s0 = parser->buf + parser->spos;
    size_t new_len = has_escapes
      ? process_string_escapes(s0, parser->buf+parser->bpos-1)
      : parser->bpos - parser->spos - 1
      ;

    set_string_event(event, s0, new_len);
  }

  return TS_OK;
}

static TSResult
parse_string(ts_json_parser* parser, ts_json_event* event)
{
  JSON_PUSH_STATE(parser, TS_JSON_PARSER_STRING);
  TSResult err = parse_string_inner(parser,event);
  JSON_POP_STATE(parser);

  return err;
}

static inline int
is_starting_number(int c)
{
  return isdigit(c) || (c == '-');
}

static inline int
is_starting_literal_char(int c)
{
  return isalpha(c) || (c == '_');
}

static inline int
is_literal_char(int c)
{
  return isalnum(c) || (c == '_');
}

static TSResult
parse_literal_inner(ts_json_parser* parser, ts_json_event* event)
{
  int lline = parser->line;
  int lcol  = parser->col;

  size_t lpos = parser->bpos;
  for(;;)
  {
    int c = json_getc(parser);
    if (c == EOF) { break; }
    else if (!is_literal_char(c)) {
      json_ungetc(parser,c);
      break;
    }
  }

  char* s = parser->buf+lpos;
  size_t n = parser->bpos - lpos;
  if (LIT_EQ(s,n, "null")) {
    set_literal_event(event, TS_JSON_NULL);
  } else if (LIT_EQ(s,n,"true")) {
    set_literal_event(event, TS_JSON_TRUE);
  } else if (LIT_EQ(s,n,"false")) {
    set_literal_event(event, TS_JSON_FALSE);
  } else if (parser->enable_literal_strings) {
    set_string_event(event, s, n);
  } else {
    PARSE_ERROR(parser,
        "unknown literal encountered at line %d, col %d", lline, lcol);
  }

  return TS_OK;
}

static TSResult
parse_literal(ts_json_parser* parser, ts_json_event* event)
{
  JSON_PUSH_STATE(parser, TS_JSON_PARSER_LITERAL);
  TSResult err = parse_literal_inner(parser,event);
  JSON_POP_STATE(parser);

  return err;
}

#define INVALID_NUMBER() PARSE_ERROR(parser, \
    "invalid character encountered while parsing number", 0)

static TSResult
parse_number_inner(ts_json_parser* parser, ts_json_event* event)
{
  /* need to determine maximum string size for double */
  char buf[96];
  int bloc = 0;

  int c = json_getc(parser);

  if (c == '-') {
    buf[bloc++] = c;
    c = json_getc(parser);
  }
  
  if (!isdigit(c)) {
    INVALID_NUMBER();
  }

  while (isdigit(c)) {
    buf[bloc++] = c;
    c = json_getc(parser);
  }

  if (c == '.') {
    buf[bloc++] = c;

    c = json_getc(parser);
    if (!isdigit(c)) {
      INVALID_NUMBER();
    }

    while(isdigit(c)) {
      buf[bloc++] = c;
      c = json_getc(parser);
    }
  }

  if ((c == 'e') || (c == 'E'))
  {
    buf[bloc++] = c;
    c = json_getc(parser);

    if (c == '-') { buf[bloc++] = c; c = json_getc(parser); }
    if (c == '+') { c = json_getc(parser); }

    if (!isdigit(c)) {
      INVALID_NUMBER();
    }

    while(isdigit(c)) {
      buf[bloc++] = c;
      c = json_getc(parser);
    }
  }

  json_ungetc(parser,c);

  buf[bloc] = '\0';
  char* endptr = NULL;

  errno = 0;
  double num = strtod(buf, &endptr);
  if (errno == ERANGE) {
    PARSE_ERROR(parser, "error in parsing number: number out of range for double", 0);
  }
  if (errno != 0 || *endptr != '\0') {
    PARSE_ERROR(parser, "error in final parsing!", 0);
  }

  event->type = TS_JSON_NUMBER;
  event->value.num = num;

  return TS_OK;
}

static TSResult
parse_number(ts_json_parser* parser, ts_json_event* event)
{
  JSON_PUSH_STATE(parser, TS_JSON_PARSER_NUMBER);
  TSResult err = parse_number_inner(parser,event);
  JSON_POP_STATE(parser);

  return err;
}

static TSResult
parse_array(ts_json_parser* parser, ts_json_event* event);

static TSResult
parse_bare(ts_json_parser* parser, ts_json_event* event)
{
  int c = skipws(parser);

  if (c == EOF) { event->type = TS_JSON_EOF; return TS_OK; }

  if (c == '"') {
    RETURN_WITH_RESULT( parse_string(parser,event) );
  }

  if (c == '[') {
    JSON_PUSH_STATE(parser, TS_JSON_PARSER_ARRAY_OBJ);
    event->type = TS_JSON_ARRAY_BEGIN;
    return TS_OK;
  }

  if (c == '{') {
    JSON_PUSH_STATE(parser, TS_JSON_PARSER_OBJECT_KEY);
    event->type = TS_JSON_OBJECT_BEGIN;
    return TS_OK;
  }

  if (is_starting_number(c)) {
    json_ungetc(parser,c);
    RETURN_WITH_RESULT( parse_number(parser,event) );
  }

  if (is_starting_literal_char(c)) {
    json_ungetc(parser,c);
    RETURN_WITH_RESULT( parse_literal(parser,event) );
  }

  PARSE_ERROR(parser, "unexpected character %d", c);

  return TS_OK;
}

static TSResult
parse_array(ts_json_parser* parser, ts_json_event* event)
{
  int c;

restart:
  c = skipws(parser);

  if (c == ']') {
    JSON_POP_STATE(parser);
    event->type = TS_JSON_ARRAY_END;
    return TS_OK;
  }

  enum TS_JSON_PARSER_STATE state = JSON_PEEK_STATE(parser);
  switch (state)
  {
    case TS_JSON_PARSER_ARRAY_OBJ:
      json_ungetc(parser, c);
      JSON_SET_STATE(parser, TS_JSON_PARSER_ARRAY_DELIM);
      RETURN_WITH_RESULT( parse_bare(parser,event) );

    case TS_JSON_PARSER_ARRAY_DELIM:
      if (c != ',') { 
        PARSE_ERROR(parser, "expected comma delimeter", 0);
      }
      JSON_SET_STATE(parser, TS_JSON_PARSER_ARRAY_OBJ);
      goto restart;

    default:
      PARSE_ERROR(parser,
          "invalid state %d for parsing array", (int)state);
  }

  return TS_OK;
}

static TSResult
parse_object(ts_json_parser* parser, ts_json_event* event)
{
  int c;
  enum TS_JSON_PARSER_STATE state;

restart:
  c = skipws(parser);
  state = JSON_PEEK_STATE(parser);

  if (c == '}')
  {
    if ((state == TS_JSON_PARSER_OBJECT_KEY) ||
        (state == TS_JSON_PARSER_OBJECT_PAIR_DELIM))
    {
      JSON_POP_STATE(parser);
      event->type = TS_JSON_OBJECT_END;
      return TS_OK;
    }
    else
    {
      PARSE_ERROR(parser, "invalid closing brace '}'", 0);
    }
  }

  switch (state)
  {
    case TS_JSON_PARSER_OBJECT_KEY:
      if (c == '"') { 
        RETURN_ON_ERROR( parse_string(parser, event) );
      } else if (parser->enable_literal_strings && is_starting_literal_char(c)) {
        json_ungetc(parser,c);
        RETURN_ON_ERROR( parse_literal(parser, event) );
      } else {
        PARSE_ERROR(parser, "expected object key (string)", 0);
      }

      event->type = TS_JSON_OBJECT_KEY;
      JSON_SET_STATE(parser, TS_JSON_PARSER_OBJECT_KV_DELIM);
      return TS_OK;

    case TS_JSON_PARSER_OBJECT_KV_DELIM:
      if (c != ':') { 
        PARSE_ERROR(parser, "expected ':' after object key", 0);
      }
      JSON_SET_STATE(parser, TS_JSON_PARSER_OBJECT_VALUE);
      goto restart;

    case TS_JSON_PARSER_OBJECT_VALUE:
      json_ungetc(parser, c);
      JSON_SET_STATE(parser, TS_JSON_PARSER_OBJECT_PAIR_DELIM);
      RETURN_WITH_RESULT( parse_bare(parser,event) );

    case TS_JSON_PARSER_OBJECT_PAIR_DELIM:
      if (c != ',') {
        PARSE_ERROR(parser, "expected ',' between key-value pairs", 0);
      }
      JSON_SET_STATE(parser,TS_JSON_PARSER_OBJECT_KEY);
      goto restart;

    default:
      PARSE_ERROR(parser,
          "invalid state %d while parsing object", (int)state);
  }

  return TS_OK;
}

TSResult ts_json_next(ts_json_parser* parser, ts_json_event* event)
{
  enum TS_JSON_PARSER_STATE state = JSON_PEEK_STATE(parser);
  switch (state)
  {
    case TS_JSON_PARSER_BARE:
      RETURN_ON_ERROR( parse_bare(parser,event) );
      break;

    case TS_JSON_PARSER_ARRAY_OBJ:
    case TS_JSON_PARSER_ARRAY_DELIM:
      RETURN_ON_ERROR( parse_array(parser,event) );
      break;

    case TS_JSON_PARSER_OBJECT_KEY:
    case TS_JSON_PARSER_OBJECT_KV_DELIM:
    case TS_JSON_PARSER_OBJECT_VALUE:
    case TS_JSON_PARSER_OBJECT_PAIR_DELIM:
      RETURN_ON_ERROR( parse_object(parser,event) );
      break;

    default:
      PARSE_ERROR(parser, "unsupported state %d", (int)state);
  }

  return TS_OK;
}
