#include "ts_lattice.h"

#include <math.h>
#include <errno.h>

#include "ts_common.h"
#include "ts_common_algo.h"
#include "khash.h"

/* Some defines used to trigger debugging modes */
#define LATTICE_DEBUG_STEPS 0        /* disable logging of each step */
#undef LATTICE_DEBUG_STEPS_MINSTEP   /* define if logging steps */
#define LATTICE_DEBUG_UPDATES 0

int ts_latt_debug_updates = 0;

#if LATTICE_DEBUG_UPDATES
#  define LATT_UPDATE_DEBUG_LEVEL  (ts_latt_debug_updates)
#else
#  define LATT_UPDATE_DEBUG_LEVEL  (0)
#endif /* LATTICE_DEBUG_UPDATES */

enum { MOVETYPE_SIZE_INITIAL = 16 };

enum{ HOOK_SIZE_INITIAL     =  4 };

enum { CHECK_MOVELISTS_FOR_CONSISTENCY = 0 };
  
enum { CHECK_EACH_MOVE_FOR_CONSISTENCY = 0 };

enum { UPDATE_BATCH_SIZE = 64 };

enum { MOVEPOOL_SIZE_INITIAL = 8192 };

struct ts_lattice_movepool
{
  struct ts_lattice_movepool *next;

  size_t capa;
  size_t top;
  ts_lattice_move* free_list;

  ts_lattice_move base[];
};

struct ts_lattice_movepool* movepool_new(size_t capa, struct ts_lattice_movepool *next)
{
  struct ts_lattice_movepool *pool = 
    ts_alloc(NO_ALLOCATOR, sizeof(*pool)+capa*sizeof(pool->base[0]));

  if (pool != NULL)
  {
    pool->top = 0;
    pool->capa = capa;
    pool->free_list = NULL;
    pool->next = next;
  }

  return pool;
}

static void movepool_delete(struct ts_lattice_movepool *pool)
{
  ts_free(NO_ALLOCATOR, pool);
}

static void movepool_delete_chain(struct ts_lattice_movepool *head)
{
  while (head)
  {
    struct ts_lattice_movepool *next = head->next;
    movepool_delete(head);
    head = next;
  }
}

static int movepool_request(struct ts_lattice_movepool *pool, ts_lattice_move** movep)
{
  if (pool->top < pool->capa)
  {
    *movep = pool->base+pool->top;
    pool->top++;
    return 1;
  }

  if (pool->free_list != NULL)
  {
    *movep = pool->free_list;
    pool->free_list = pool->free_list->next;
    (*movep)->next = NULL;
    return 1;
  }

  return 0;
}

static void movepool_release(struct ts_lattice_movepool *pool, ts_lattice_move* move)
{
  move->next = pool->free_list;
  pool->free_list = move;
}

struct pseudospin_data
{
  ts_lattice_pseudospin_func func;
  void* closure;
  spin_t pseudo;
};

struct ts_lattice_hook
{
  ts_lattice_hook_func func;
  void* closure;
};

KHASH_MAP_INIT_INT64(site_move, ts_lattice_move*)
KHASH_MAP_INIT_INT(pspin, struct pseudospin_data*)

/* silences CLang and gcc warnings about kh_del_* functions */
static MAYBE_UNUSED void elim_warnings(void)
{
  (void)kh_del_site_move(NULL,0);
  (void)kh_del_pspin(NULL,0);
}

struct ts_lattice_pspins
{
  khash_t(pspin) hash;
};

struct ts_lattice_moves
{
  khash_t(site_move) hash;
};

#define PSPIN_TABLE(latt) \
  ( (khash_t(pspin)*) (latt)->pseudospins )

#define MOVES_TABLE(latt) \
  ( (khash_t(site_move)*) (latt)->site_moves )

TSResult ts_lattice_request_move(ts_lattice* latt, ts_lattice_move** movesp)
{
  while (!movepool_request(latt->movepool, movesp))
  {
    size_t new_size = latt->movepool->capa + latt->movepool->capa/2;
    struct ts_lattice_movepool *pool =
      movepool_new(new_size,latt->movepool);
    ERROR_ON_NULL_ALLOC(pool);

    latt->movepool = pool;
  }

  return TS_OK;
}

void ts_lattice_release_move(struct ts_lattice* latt, ts_lattice_move* move)
{
  movepool_release(latt->movepool, move);
}

TSResult ts_lattice_initialize(ts_lattice* latt, size_t nrow, size_t ncol, size_t rate_size)
{
  memset(latt,0,sizeof(*latt));

#define CLEANUP_ON_ERROR { ts_lattice_finalize(latt); }

  latt->refcount = 1;
  latt->destructor = NULL;

  latt->sites = NULL;
  memset(&latt->rates, 0, sizeof(latt->rates));
  latt->site_moves  = NULL;
  latt->pseudospins = NULL;
  memset(&latt->movetypes, 0, sizeof(latt->movetypes));

  latt->sites = NULL;
  latt->num_sites = 0;
  latt->nrow = latt->ncol = 0;

  latt->row_coupling_dist = TS_LATTICE_DEFAULT_COUPLING_DIST;
  latt->col_coupling_dist = TS_LATTICE_DEFAULT_COUPLING_DIST;

  latt->rng = NULL;
  latt->rebuild_moves_hook.callback  = NULL;
  latt->rebuild_moves_hook.user_data = NULL;

  latt->changes.rv = NULL;
  latt->changes.cv = NULL;
  latt->changes.vv = NULL;
  latt->changes.n = latt->changes.max = 0;

  latt->num_sites = nrow*ncol;
  latt->sites = ts_calloc(NO_ALLOCATOR, latt->num_sites, sizeof latt->sites[0]);
  ERROR_ON_NULL_ALLOC( latt->sites );

  latt->nrow = nrow;
  latt->ncol = ncol;

  RETURN_ON_ERROR3( ts_fenwick_tree_init(&latt->rates, rate_size) );

  latt->site_moves = (struct ts_lattice_moves*)kh_init(site_move);
  ERROR_ON_NULL_ALLOC(latt->site_moves);

  latt->pseudospins = (struct ts_lattice_pspins*)kh_init(pspin);
  ERROR_ON_NULL_ALLOC(latt->pseudospins);

  DATAVECTOR_INIT(struct ts_lattice_movetype, latt->movetypes, MOVETYPE_SIZE_INITIAL);

  DATAVECTOR_INIT(struct ts_lattice_hook, latt->move_hooks,      HOOK_SIZE_INITIAL);
  DATAVECTOR_INIT(struct ts_lattice_hook, latt->post_move_hooks, HOOK_SIZE_INITIAL);

  latt->movepool = movepool_new( MOVEPOOL_SIZE_INITIAL, NULL );
  ERROR_ON_NULL_ALLOC( latt->movepool );

  RETURN_ON_ERROR( ts_lattice_resize_changes(latt, TS_LATTICE_MAX_UPDATES) );

  ts_lattice_reset(latt);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

ts_lattice *ts_lattice_retain(ts_lattice *latt)
{
  assert(latt->refcount > 0);
  assert(latt->refcount < INT_MAX); /* avoid overflow */
  latt->refcount++;
  return latt;
}

int ts_lattice_refcount(ts_lattice *latt)
{
  return latt->refcount;
}

void ts_lattice_release(ts_lattice *latt)
{
  assert(latt->refcount > 0);
  if (--latt->refcount == 0)
  {
    if (latt->destructor != NULL)
    {
      latt->destructor(latt);
    }
  }
}

static void lattice_free(void *p);

ts_lattice *ts_lattice_new(size_t nrow, size_t ncol, size_t rate_size, TSResult *errp)
{
  ts_lattice *latt;

  assert(errp != NULL);

  latt = ts_alloc(NO_ALLOCATOR, sizeof *latt);
  if (latt == NULL)
  {
    *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    return NULL;
  }

  *errp = ts_lattice_initialize(latt, nrow,ncol,rate_size), ON_ERROR_ADD_FRAME(*errp);
  if (IS_ERROR(*errp))
  {
    ts_free(NO_ALLOCATOR, latt);
    return NULL;
  }

  latt->destructor = lattice_free;
  return latt;
}

static void lattice_free(void *p)
{
  ts_lattice_finalize(p);
  ts_free(NO_ALLOCATOR, p);
}

static void free_each_move(void *closure, void *data)
{
  ts_lattice *latt = closure;
  ts_lattice_move *m = data;
  ts_lattice_release_move(latt,m);
}

void ts_lattice_finalize(ts_lattice *latt)
{
  ts_free(NO_ALLOCATOR,latt->sites);
  ts_fenwick_tree_clear(&latt->rates, free_each_move, latt);

  ts_fenwick_tree_finalize(&latt->rates);

  if (latt->site_moves)  { kh_destroy(site_move, MOVES_TABLE(latt)); }
  if (latt->pseudospins) {
    khiter_t k;

    for (k=kh_begin(PSPIN_TABLE(latt)); k != kh_end(PSPIN_TABLE(latt)); k++)
    {
      if (kh_exist(PSPIN_TABLE(latt), k))
        ts_free(NO_ALLOCATOR, kh_value(PSPIN_TABLE(latt), k));
    }

    kh_destroy(pspin, PSPIN_TABLE(latt));
  }

  if (latt->movepool)
  {
    movepool_delete_chain(latt->movepool);
    latt->movepool = NULL;
  }

  datavector_free(&latt->movetypes);
  datavector_free(&latt->move_hooks);
  datavector_free(&latt->post_move_hooks);

  ts_free(NO_ALLOCATOR, latt->changes.rv);
}

static void reset_rate_tree(ts_lattice* latt)
{
  kh_clear(site_move, MOVES_TABLE(latt) );
  ts_fenwick_tree_clear(&latt->rates, free_each_move, latt);
}

void ts_lattice_reset_rate_tree(ts_lattice* latt) { reset_rate_tree(latt); }

static inline void clear_updates(ts_lattice* latt)
{
  if (LATT_UPDATE_DEBUG_LEVEL > 0)
  {
    fprintf(stderr, "clearing %zu changes and %d updates\n", latt->changes.n, latt->update_list.n);
  }

  latt->changes.n = 0;
  latt->update_list.n = 0;
}

void ts_lattice_zero(ts_lattice *latt)
{
  size_t i,n;
  spin_t *restrict sites;

  for (i=0,n=latt->num_sites,sites = &latt->sites[0]; i < n; i++)
  {
    sites[i].val = 0;
  }
}

void ts_lattice_clear_history(ts_lattice *latt)
{
  size_t i,n;
  struct ts_lattice_move zero = { 0 };

  for (i=0,n=ARRAYLEN(latt->move_history.moves); i < n; i++)
  {
    latt->move_history.moves[i] = zero;
  }

  latt->move_history.last = 0;
}

void ts_lattice_reset(ts_lattice* latt)
{
  ts_lattice_zero(latt);

  latt->time = 0.0;
  ts_lattice_set_temp(latt, 1.0);
  latt->nsteps = 0;

  reset_rate_tree(latt);

  struct pseudospin_data* pspin = NULL;
  kh_foreach_value( PSPIN_TABLE(latt), pspin,
      ts_safe_free(NO_ALLOCATOR, pspin)
      );

  kh_clear(pspin, PSPIN_TABLE(latt));

  clear_updates(latt);
  ts_lattice_clear_history(latt);

  datavector_clear(&latt->movetypes);
  datavector_clear(&latt->move_hooks);
  datavector_clear(&latt->post_move_hooks);
}

#if defined(HAS_ALIGNOF)
enum { SPIN_ALIGN = alignof(spin_t) };
#else /* defined(HAS_ALIGNOF) */
struct spin_align { char x; spin_t s; };
enum { SPIN_ALIGN = offsetof(struct spin_align, s) };
#endif /* defined(HAS_ALIGNOF) */

TSResult ts_lattice_change_list_resize(struct ts_lattice_change_list *changes, size_t max)
{
  size_t bsz, align, i, n;
  char *block;
  int *rv, *cv;
  spin_t *vv;

  if (max <= changes->max)
  {
    return TS_OK;
  }

  bsz = 2 * max * sizeof *rv;

  /* correct for alignment */
  align = bsz % SPIN_ALIGN;
  if (align > 0) { align = SPIN_ALIGN - align; }
  bsz += align;
  bsz += max * sizeof *vv;

  block = ts_alloc(NO_ALLOCATOR, bsz);
  ERROR_ON_NULL_ALLOC(block);

  memset(block, 0, bsz);

  rv = (int *)block;
  cv = (int *)(block + max * sizeof *rv);
  vv = (spin_t *)(block + 2 * max * sizeof *rv + align);

  for (i=0, n=changes->n; i < n; i++)
  {
    rv[i] = changes->rv[i];
    cv[i] = changes->cv[i];
    vv[i] = changes->vv[i];
  }

  ts_free(NO_ALLOCATOR, changes->rv);

  changes->rv = rv;
  changes->cv = cv;
  changes->vv = vv;

  changes->max = max;

  return TS_OK;
}

TSResult ts_lattice_resize_changes(ts_lattice *latt, size_t max)
{
  RETURN_ON_ERROR(
      ts_lattice_change_list_resize(&latt->changes, max) );
  return TS_OK;
}

TSResult ts_lattice_save_changes(ts_lattice *latt,
    struct ts_lattice_change_list *changes,
    struct ts_lattice_update_list *updates)
{
  size_t i,n;

  n = latt->changes.n;
  if (changes->max < n)
  {
    RETURN_ON_ERROR(
        ts_lattice_change_list_resize(changes, latt->changes.n)
        );
  }

  for (i=0; i < n; i++)
  {
    changes->rv[i] = latt->changes.rv[i];
    changes->cv[i] = latt->changes.cv[i];
    changes->vv[i] = latt->changes.vv[i];
  }

  changes->n = latt->changes.n;

  assert(sizeof *updates == sizeof latt->update_list);
  memcpy(updates, &latt->update_list, sizeof latt->update_list);

  if (LATT_UPDATE_DEBUG_LEVEL > 0)
  {
    fprintf(stderr, "latt: saved %zu changes and %d updates\n", latt->changes.n, latt->update_list.n);
  }

  return TS_OK;
}

TSResult ts_lattice_restore_changes(ts_lattice *latt,
    struct ts_lattice_change_list *changes,
    struct ts_lattice_update_list *updates)
{
  size_t i,n;

  n = changes->n;
  RETURN_ON_ERROR(ts_lattice_resize_changes(latt, n));

  for (i=0; i < n; i++)
  {
    latt->changes.rv[i] = changes->rv[i];
    latt->changes.cv[i] = changes->cv[i];
    latt->changes.vv[i] = changes->vv[i];
  }

  latt->changes.n = changes->n;

  assert(sizeof *updates == sizeof latt->update_list);
  memcpy(&latt->update_list, updates, sizeof *updates);

  if (LATT_UPDATE_DEBUG_LEVEL > 0)
  {
    fprintf(stderr, "latt: restored %zu changes and %d updates\n", latt->changes.n, latt->update_list.n);
  }

  return TS_OK;
}

void ts_lattice_change_list_finalize(struct ts_lattice_change_list *changes)
{
  ts_free(NO_ALLOCATOR, changes->rv);
  changes->rv = NULL;
  changes->cv = NULL;
  changes->vv = NULL;
  changes->n = changes->max = 0;
}

TSResult ts_lattice_add_movetype(ts_lattice* latt, struct ts_lattice_movetype* movetype)
{
  int succ = datavector_add(&latt->movetypes, movetype);
  if (!succ)
  {
    return TS_NEW_ERROR3(
        TS_ERR_OUT_OF_MEMORY,
        "Cannot resize movetypes datavector");
  }

  return TS_OK;
}

static inline TSResult add_move_hook(
    DATAVECTOR(struct ts_lattice_hook)* hook_list,
    ts_lattice_hook_func func, void* closure)
{
  struct ts_lattice_hook hook = { .func = func, .closure = closure };

  if (!datavector_add(hook_list, &hook))
  {
    return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY, "cannot add move hook");
  }

  return TS_OK;
}

static inline int remove_move_hook(
    DATAVECTOR(struct ts_lattice_hook)* hook_list,
    ts_lattice_hook_func func, void* closure)
{
  struct ts_lattice_hook *hook_data = hook_list->data;
  size_t i, n = hook_list->size;

  for(i=0; i < n; i++)
  {
    if ((hook_data[i].func == func) && (hook_data[i].closure == closure))
    {
      break;
    }
  }

  if (i == n) 
  {
    /* not found */
    return 0;
  }

  /* delete element and move all elements down by one */
  for(; i < n-1; i++)
  {
    hook_data[i] = hook_data[i+1];
  }

  hook_data[n-1].func = NULL;
  hook_data[n-1].closure = NULL;
  hook_list->size--;

  return 1;
}

TSResult ts_lattice_add_move_hook(ts_lattice* latt, ts_lattice_hook_func func, void* closure)
{
  RETURN_WITH_RESULT(
      add_move_hook(&latt->move_hooks, func, closure)
      );
}

int ts_lattice_remove_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure)
{
  return remove_move_hook(&latt->move_hooks, func, closure);
}

TSResult ts_lattice_add_post_move_hook(ts_lattice* latt, ts_lattice_hook_func func, void* closure)
{
  RETURN_WITH_RESULT(
      add_move_hook(&latt->post_move_hooks, func, closure)
      );
}

int ts_lattice_remove_post_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure)
{
  return remove_move_hook(&latt->post_move_hooks, func, closure);
}

static inline TSResult run_hooks(ts_lattice* latt,
    DATAVECTOR(struct ts_lattice_hook)* hooks,
    ts_lattice_move* move)
{
  ts_lattice_move orig = *move;
  orig.next = NULL;

  int cancelled = 0;

  DATAVECTOR_FOREACH(struct ts_lattice_hook, hook, hooks)
  {
    ts_lattice_move copy = orig;
    TSResult ret = hook->func( hook->closure, latt, &copy);
    if (IS_ERROR(ret))
    {
      RETURN_WITH_RESULT(ret);
    }

    if (IS_CANCELLED(ret) && !cancelled)
    {
      cancelled = 1;
    }
  }

  if (cancelled)
  {
    CANCEL_OPERATION();
  }

  return TS_OK;
}

TSResult ts_lattice_register_pseudospin(ts_lattice* latt, spin_t pseudo, ts_lattice_pseudospin_func func, void* closure)
{
  struct pseudospin_data* data = ts_alloc(NO_ALLOCATOR, sizeof(*data));
  memset(data,0,sizeof(*data));
  data->func = func;
  data->closure = closure;
  data->pseudo = pseudo;

  int ret = 0;
  khiter_t iter = kh_put(pspin, PSPIN_TABLE(latt), (int)pseudo.val, &ret);

  if (iter != kh_end( PSPIN_TABLE(latt) ))
  {
    kh_value(PSPIN_TABLE(latt), iter) = data;
  }

  if ((ret < 0) || (iter == kh_end( PSPIN_TABLE(latt) )))
  {
    return TS_NEW_ERROR2(TS_ERR_OUT_OF_MEMORY,
        ("cannot register pseudospin %d", (int)pseudo.val));
  }
  else
  {
    return TS_OK;
  }
}

int ts_lattice_num_pseudospins(const ts_lattice* latt)
{
  return kh_size( PSPIN_TABLE(latt) );
}

static TSResult invoke_pseudospin_without_update(ts_lattice* latt, spin_t pseudo, int r, int c)
{
  khiter_t iter = kh_get(pspin, PSPIN_TABLE(latt), (int)pseudo.val);

  if (iter == kh_end( PSPIN_TABLE(latt) ))
  {
    return TS_NEW_ERROR2(TS_ERR_LATT_PSEUDOSPIN_NOT_REGISTERED,
        ("pseudospin %d not registered", pseudo.val));
  }

  struct pseudospin_data* pspin = kh_value( PSPIN_TABLE(latt), iter );

  RETURN_WITH_RESULT( pspin->func(pspin->closure, latt,r,c, pseudo) );
}

static TSResult invoke_pseudospin(ts_lattice* latt, spin_t pseudo, int r, int c)
{
  RETURN_ON_ERROR( invoke_pseudospin_without_update(latt,pseudo,r,c) );
  RETURN_ON_ERROR( ts_lattice_update_marked_sites(latt) );
  return TS_OK;
}

TSResult ts_lattice_invoke_pseudospin(ts_lattice* latt, spin_t pseudo, int r, int c)
{
  clear_updates(latt);
  RETURN_WITH_RESULT( invoke_pseudospin(latt, pseudo, r,c) );
}

TSResult ts_lattice_invoke_pseudospin_without_update(ts_lattice* latt, spin_t pseudo, int r, int c)
{
  clear_updates(latt);
  RETURN_WITH_RESULT( invoke_pseudospin_without_update(latt, pseudo, r,c) );
}

static TSResult update_moves(ts_lattice* latt, const int* rv, const int* cv, size_t n);
TSResult ts_lattice_rebuild_moves(ts_lattice* latt)
{
  int r,c, nr, nc, rv[UPDATE_BATCH_SIZE], cv[UPDATE_BATCH_SIZE];

  ts_lattice_reset_rate_tree(latt);
  nr = latt->nrow;
  nc = latt->ncol;

  if (latt->rebuild_moves_hook.callback)
  {
    latt->rebuild_moves_hook.callback(latt, latt->rebuild_moves_hook.user_data);
  }

  for (c = 0; c < nc; c++)
  {
    for (r = 0; r < nr; r += UPDATE_BATCH_SIZE)
    {
      int ind, maxind = nr - r;
      if (maxind > UPDATE_BATCH_SIZE) { maxind = UPDATE_BATCH_SIZE; }

      for (ind=0; ind < maxind; ind++)
      {
        rv[ind] = r+ind;
        cv[ind] = c;
      }

      RETURN_ON_ERROR( update_moves(latt,rv,cv,maxind) );
    }
  }

  return TS_OK;
}

TSResult ts_lattice_get_k_total(ts_lattice* latt, double* ktotp)
{
  if (latt->rates.size == 0)
  {
    RETURN_ON_ERROR( ts_lattice_rebuild_moves(latt) );
  }

  *ktotp = latt->rates.total_sum;
  return TS_OK;
}

TSResult ts_lattice_move_iterator_init(ts_lattice* latt, ts_lattice_move_iterator* iter)
{
  UNUSED_PARAM(latt);

  iter->curr.r = iter->curr.c = 0;
  iter->curr.state = INVALID_SPIN;
  iter->curr.rate = -1.0;

  iter->ind = 0;

  return TS_OK;
}

TSResult ts_lattice_moves_iterate(ts_lattice* latt, ts_lattice_move_iterator* iter)
{
  if (iter->ind >= latt->rates.size) { goto stop; }

  memcpy(&iter->curr, latt->rates.data[iter->ind++], sizeof iter->curr);

  return TS_OK;

stop:
  return TS_NEW_RESULT(TS_STOP_ITERATION);
}

TSResult ts_lattice_get_all_moves(ts_lattice* latt,
    ts_lattice_move** moves_ptr, size_t* nmovesp, double* ktotp)
{
  void **restrict data = latt->rates.data;
  size_t i,n;
#define CLEANUP_ON_ERROR {}

  if (n = latt->rates.size, n == 0)
  {
    RETURN_ON_ERROR( ts_lattice_rebuild_moves(latt) );
  }

  *nmovesp = n;
  *ktotp = latt->rates.total_sum;

  ts_lattice_move* moves = ts_calloc(NO_ALLOCATOR, n, sizeof moves[0]);
  ERROR_ON_NULL_ALLOC(moves);

  for (size_t i=0; i < latt->rates.size; i++)
  {
    memcpy(&moves[i], data[i], sizeof moves[i]);
  }

#undef CLEANUP_ON_ERROR

  *moves_ptr = moves;
  return TS_OK;
}

ts_lattice_move* ts_lattice_get_current_site_moves(ts_lattice* latt, size_t r, size_t c)
{
  size_t index = ts_lattice_index(latt, r,c);

  khiter_t iter = kh_get(site_move, MOVES_TABLE(latt), index);

  return (iter != kh_end(MOVES_TABLE(latt)))
    ? kh_value(MOVES_TABLE(latt),iter)
    : NULL
    ;
}

TSResult ts_lattice_set_current_site_moves(ts_lattice* latt,
    size_t r, size_t c, ts_lattice_move* m)
{
  size_t index = ts_lattice_index(latt, r,c);
  int ret = 0;
  khiter_t iter = kh_put(site_move, MOVES_TABLE(latt), index, &ret);

  if (UNLIKELY((ret < 0) || (iter == kh_end( MOVES_TABLE(latt) ))))
  {
    return TS_NEW_ERROR2( TS_ERR_OUT_OF_MEMORY,
        ("cannot set moves for site %d", (int)index) );
  }
  else
  {
    kh_value(MOVES_TABLE(latt), iter) = m;
    return TS_OK;
  }
}

static TSResult ensure_enough_moves(ts_lattice *latt, ts_lattice_move **mlistp, size_t nmoves)
{
  size_t i;
  ts_lattice_move *mv;

  tsAssert((mlistp != NULL) && "pointer to move list cannot be NULL");

  mv = *mlistp;
  while ((mv != NULL) && (nmoves > 0))
  {
    mv = mv ->next;
    nmoves--;
  }

  for (i=0; i < nmoves; i++)
  {
    RETURN_ON_ERROR( ts_lattice_request_move(latt, &mv) );
    mv->next = *mlistp;
    *mlistp = mv;
  }

  return TS_OK;
}

TSResult ts_lattice_update_current_site_moves(ts_lattice* latt, size_t r, size_t c, spin_t* states, double* rates, size_t nmoves)
{
  ts_lattice_move* mlist = ts_lattice_get_current_site_moves(latt,r,c);
  ts_lattice_move* new_list = NULL;

  RETURN_ON_ERROR( ensure_enough_moves(latt, &mlist, nmoves) );

  for (size_t i=0; i < nmoves; i++)
  {
    ts_lattice_move* m = mlist;
    mlist = mlist->next;

    m->state = states[i];
    m->rate = rates[i];

    m->next = new_list;
    new_list = m;

    size_t tree_ind = ts_fenwick_tree_find(&latt->rates, m);
    if (tree_ind < latt->rates.size)
    {
      RETURN_ON_ERROR( ts_fenwick_tree_update_count(&latt->rates, tree_ind, m->rate) );
    }
    else
    {
      RETURN_ON_ERROR( ts_fenwick_tree_add(&latt->rates, m->rate, m) );
    }
  }

  while (mlist != NULL)
  {
    ts_lattice_move *mv = mlist;
    mlist = mlist->next;

    size_t tree_ind = ts_fenwick_tree_find(&latt->rates, mv);
    ERROR_IF(TS_ERR_SIMULATION_STATE, tree_ind >= latt->rates.size);

    RETURN_ON_ERROR( ts_fenwick_tree_remove_at(&latt->rates, tree_ind) );

    ts_lattice_release_move(latt, mv);
  }

  RETURN_ON_ERROR(
      ts_lattice_set_current_site_moves(latt,r,c, new_list) );

  return TS_OK;
}

void ts_lattice_print_movelist(ts_lattice_move* m)
{
  while(m)
  {
    fprintf(stderr, "move %p (%3zu,%3zu)  state = %3d  rate = %7.4f\n",
        m, m->r, m->c, (int)m->state.val, m->rate);
    m = m->next;
  }
  fprintf(stderr, "\n");
}

void ts_lattice_print_moves(ts_lattice* latt, int r, int c)
{
  ts_lattice_move *movelist = ts_lattice_get_current_site_moves(latt, r, c);
  ts_lattice_print_movelist(movelist);
}

static TSResult check_that_movelists_are_consistent(ts_lattice* latt);
static TSResult update_moves(ts_lattice* latt, const int* rv, const int* cv, size_t nsites)
{
  spin_t states[UPDATE_BATCH_SIZE];
  double rates[UPDATE_BATCH_SIZE];
  size_t siteinds[UPDATE_BATCH_SIZE];
  ts_lattice_move *mvlists_old[UPDATE_BATCH_SIZE];
  ts_lattice_move *mvlists_new[UPDATE_BATCH_SIZE];
  size_t site;

  for (site = 0; site < nsites; site += UPDATE_BATCH_SIZE)
  {
    size_t si, batch_max;
    batch_max = site + UPDATE_BATCH_SIZE;

    if (batch_max > nsites) { batch_max = nsites; }
    for (si=site; si < batch_max; si++)
    {
      mvlists_old[si] = ts_lattice_get_current_site_moves(latt, rv[si], cv[si]);
      mvlists_new[si] = NULL;
    }

    for(size_t i=0; i < latt->movetypes.size; i++)
    {
      struct ts_lattice_movetype* mt =
        DATAVECTOR_AT(struct ts_lattice_movetype, &latt->movetypes,i);
      intptr_t cdata = 0;
      int mv, incomplete;

      incomplete = 1;
      while(incomplete)
      {
        int nmoves = 0;
        TSResult ret;

        incomplete = 0;

        ret = TS_OK;
        nmoves = mt->func(mt->closure, latt, rv,cv, nsites,
            states, rates, siteinds, UPDATE_BATCH_SIZE, &cdata, &ret);

        if (nmoves < 0)
        {
          RETURN_ON_ERROR(ret);

          /* no error if we made it to here */
          incomplete = 1;
          nmoves = -nmoves;
        }

        for (mv=0; mv < nmoves; mv++)
        {
          if (rates[mv] <= 0.0) { continue; }

          si = siteinds[mv];
          ts_lattice_move* m = mvlists_old[si];
          if (m != NULL)
          {
            mvlists_old[si] = m->next;
          }
          else
          {
            RETURN_ON_ERROR( ts_lattice_request_move(latt, &m) );
          }

          m->r = rv[si];
          m->c = cv[si];
          m->state = states[mv];
          m->rate  = rates[mv];

          assert(isfinite(m->rate));

          m->next = mvlists_new[si];
          mvlists_new[si] = m;
        }
      }
    }

    for (si=site; si < batch_max; si++)
    {
      ts_lattice_move* m = mvlists_old[si];
      while (m)
      {
        ts_lattice_move* next = m->next;
        size_t tree_ind = ts_fenwick_tree_find(&latt->rates, m);
        ERROR_IF(TS_ERR_SIMULATION_STATE, tree_ind >= latt->rates.size);

        RETURN_ON_ERROR( ts_fenwick_tree_remove_at(&latt->rates, tree_ind) );

        memset(m,0,sizeof(*m));
        ts_lattice_release_move(latt,m);

        m = next;
      }

      RETURN_ON_ERROR(
          ts_lattice_set_current_site_moves(latt,
            rv[ si ], cv[ si], mvlists_new[si]) );

      for(m = mvlists_new[si]; m != NULL; m = m->next)
      {
        size_t tree_ind = ts_fenwick_tree_find(&latt->rates, m);
        if (tree_ind < latt->rates.size)
        {
          RETURN_ON_ERROR( ts_fenwick_tree_update_count(&latt->rates, tree_ind, m->rate) );
        }
        else
        {
          RETURN_ON_ERROR( ts_fenwick_tree_add(&latt->rates, m->rate, m) );
        }
      }
    }
  }

  RETURN_ON_ERROR(check_that_movelists_are_consistent(latt));

  return TS_OK;
}

static inline void get_row_range(const ts_lattice* latt, const int r, int* rmin, int* rmax)
{
  *rmin = r-(latt->row_coupling_dist);
  *rmax = r+(latt->row_coupling_dist);

  if (*rmin < 0) { *rmin = 0; }
  if (*rmax >= (int)latt->nrow) { *rmax = ((int)latt->nrow)-1; }
}

static inline TSResult batch_update_inds(ts_lattice* latt, size_t *indv, size_t n)
{
  size_t tmp[UPDATE_BATCH_SIZE] = { ~((size_t)0) };
  size_t *restrict sorted_indv;
  int rv[UPDATE_BATCH_SIZE];
  int cv[UPDATE_BATCH_SIZE];
  int nr;
  size_t i, count;

  if (n==0) { return TS_OK; }

  /* sort indv and remove duplicates! */
  if (n>1)
  {
    sorted_indv = merge_sort_sizes(indv,tmp,n);
  }
  else
  {
    sorted_indv = indv;
  }

  for (i=1; i < n; i++)
  {
    if ((sorted_indv[i] < sorted_indv[i-1]) || (sorted_indv[i] == ~((size_t)0)))
    {
      TS_RAISE_ERROR_MESG(TS_ERR_SIMULATION_STATE, "failure in merge sort!\n");
    }
  }

  nr = latt->nrow;
  cv[0] = sorted_indv[0] / nr;
  rv[0] = sorted_indv[0] % nr;
  for(count=1,i = 1; i < n; i++)
  {
    if (sorted_indv[i] != sorted_indv[i-1])
    {
      cv[count] = sorted_indv[i] / nr;
      rv[count] = sorted_indv[i] % nr;
      count++;
    }
  }

  RETURN_ON_ERROR( update_moves(latt, rv,cv, count) );

  return TS_OK;
}

static inline TSResult update_sites_and_related(ts_lattice* latt, const int *rminv0, const int *rmaxv0, const int *restrict cv0, size_t nranges)
{
  size_t n,i;
  size_t indv[UPDATE_BATCH_SIZE];
  int nr = latt->nrow;

  n = 0;
  for(i=0; i < nranges; i++)
  {
    int rmin,rmax,tmp;
    int cmin,cmax;
    int rr,cc,rowcount;

    if (rminv0[i] < 0 || cv0[i] < 0)
    {
      continue;
    }

    rmin = rmax = tmp = 0;
    get_row_range(latt, rminv0[i], &rmin, &tmp);
    get_row_range(latt, rmaxv0[i], &tmp, &rmax);

    if (rmin > rmax)
    {
      int rl0, rh0, rl1,rh1;
      rl0 = rh0 = rl1 = rh1 = 0;
      get_row_range(latt, rminv0[i], &rl0, &rh0);
      get_row_range(latt, rmaxv0[i], &rl1, &rh1);

      fprintf(stderr, "[%3zu] rmin = %3d, rmax = %3d, rminv0[i] = %3d, rmaxv0[i] = %3d\n",
          i,rmin,rmax,rminv0[i],rmaxv0[i]);
      fprintf(stderr, "       rminv0[%3zu] range: %3d,%3d\n", i, rl0,rh0);
      fprintf(stderr, "       rmaxv0[%3zu] range: %3d,%3d\n", i, rl1,rh1);
    }
    assert(rmin <= rmax);

    cmin = cv0[i]-(latt->col_coupling_dist);
    cmax = cv0[i]+(latt->col_coupling_dist);

    rowcount = rmax-rmin+1;
    for(cc = cmin; cc <= cmax; cc++)
    {
      int rw0,cw,dr,off;
      rw0 = rmin;
      cw = cc;
      ts_lattice_wrap_coord(latt,&rw0,&cw);
      
      off = nr*cw;

      for (dr = 0; dr < rowcount; dr++)
      {
        int rw = rw0+dr;

        if (rw < 0 || rw >= nr)
        {
          continue;
        }

        indv[n] = rw + off;
        assert(indv[n] < latt->num_sites);

        if (++n == UPDATE_BATCH_SIZE)
        {
          batch_update_inds(latt, indv, n);
          n = 0;
        }
      }
    }
  }

  if (n > 0)
  {
    batch_update_inds(latt,indv,n);
  }

  return TS_OK;
}

static inline TSResult change_spin_and_update(ts_lattice *latt, int r, int c, spin_t new_state)
{
  RETURN_ON_ERROR( ts_lattice_mark_site_changed(latt, r, c) );
  ts_lattice_ref(latt, r,c)[0] = new_state;

  RETURN_ON_ERROR( ts_lattice_update_marked_sites(latt) );

  return TS_OK;
}

TSResult ts_lattice_rollback_changes(ts_lattice *latt)
{
  size_t i,n=latt->changes.n;

  if (n > latt->changes.max)
  {
    TS_RAISE_ERRORF(TS_ERR_UNSUPPORTED_OPERATION,
        "%zu changes: too many to rollback (max is %zu)",
        n, TS_LATTICE_MAX_UPDATES);
  }

  for (i=0; i < n; i++)
  {
    int r,c;
    spin_t v;
    r = latt->changes.rv[i];
    c = latt->changes.cv[i];
    v = latt->changes.vv[i];

    if (!ts_lattice_is_pseudo_spin_value(latt,v.val))
    {
      TS_LATTICE_AT(latt, r,c) = v;
    }
    else
    {
      RETURN_ON_ERROR(ts_lattice_invoke_pseudospin(latt, v, r,c));
    }
  }

  latt->changes.n = 0;

  return TS_OK;
}

TSResult ts_lattice_update_marked_sites(ts_lattice *latt)
{
  if (latt->changes.n <= 0 && latt->update_list.n <= 0)
  {
    return TS_OK;
  }

  if (latt->changes.n <= latt->changes.max && latt->update_list.n <= TS_LATTICE_MAX_UPDATES)
  {
    if (latt->changes.n > 0)
    {
      RETURN_ON_ERROR(
          update_sites_and_related(latt,
            latt->changes.rv, latt->changes.rv, latt->changes.cv, latt->changes.n) );
    }

    if (latt->update_list.n > 0)
    {
      RETURN_ON_ERROR(
          update_sites_and_related(latt,
            latt->update_list.rminv,
            latt->update_list.rmaxv,
            latt->update_list.cv,
            latt->update_list.n) );
    }
  }
  else
  {
    fprintf(stderr, "\ntoo many changes!  rebuilding everything!\n");
    RETURN_ON_ERROR( ts_lattice_rebuild_moves(latt) );
  }
  clear_updates(latt);

  return TS_OK;
}

static inline int site_moves_have_move(ts_lattice* latt, ts_lattice_move* move)
{
  ts_lattice_move *m, *sitemoves;
  sitemoves = ts_lattice_get_current_site_moves(latt, move->r, move->c);
  for(m=sitemoves; m != NULL; m = m->next)
  {
    if (m == move) return 1;
  }

  return 0;
}

static TSResult check_that_movelists_are_consistent(ts_lattice* latt)
{
  size_t i,n;
  void **restrict data;

  if (!CHECK_MOVELISTS_FOR_CONSISTENCY) return TS_OK;

  data = latt->rates.data;
  n = latt->rates.size;
  for(i=0; i < n; i++)
  {
    ts_lattice_move* move = data[i];
    if (!site_moves_have_move(latt,move))
    {
      ts_lattice_move *movelist;
      movelist = ts_lattice_get_current_site_moves(latt, move->r, move->c);

      fprintf(stderr,
          "\n[%7d  %7.4f] move: %p (%3zu,%3zu) state = %3d  rate = %7.4f  next = %p\n",
          latt->nsteps, latt->time, move, move->r, move->c,
          (int)move->state.val, move->rate, move->next);

      ts_lattice_print_movelist(movelist);

      TS_RAISE_ERROR_MESG(TS_ERR_SIMULATION_STATE,
          "move is not part of the site move list!\n");
    }
  }

  return TS_OK;
}

static struct ts_lattice_move *
record_move(ts_lattice *latt, struct ts_lattice_move *mv)
{
  struct ts_lattice_move *rec;

  assert((int)latt->move_history.last >= 0 && latt->move_history.last < ARRAYLEN(latt->move_history.moves));

  rec = &latt->move_history.moves[latt->move_history.last++];
  if (latt->move_history.last == ARRAYLEN(latt->move_history.moves))
  {
    latt->move_history.last = 0;
  }

  *rec = *mv;
  rec->next = NULL;

  return rec;
}

static inline TSResult run_hooks_and_execute_move(ts_lattice *latt, ts_lattice_move *move)
{
  TSResult ret;
  struct ts_lattice_move *rec;

#if LATTICE_DEBUG_STEPS
  if (move->rate == 0.0)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_SIMULATION_STATE, "move selected with rate == 0.0");
  }

  if (latt->nsteps >= LATTICE_DEBUG_STEP_MINSTEP)
  {
    ts_lattice_move *movelist = ts_lattice_get_current_site_moves(latt, move->r, move->c);

    fprintf(stderr, "\n[%7d  %7.4f] move: %p (%3zu,%3zu) state = %3d  rate = %7.4f  next = %p\n",
        latt->nsteps, latt->time, move, move->r, move->c, (int)move->state.val, move->rate, move->next);

    ts_lattice_print_movelist(movelist);
  }
#endif /* LATTICE_DEBUG_STEPS */

  clear_updates(latt);

  rec = record_move(latt, move);
  rec->next = rec;

  ret = run_hooks(latt, &latt->move_hooks, move);
  if (IS_ERROR(ret)) {
    RETURN_WITH_RESULT(ret);
  }

  if (IS_CANCELLED(ret))
  {
    rec->next = NULL;
    return TS_OK;
  }

  ts_lattice_move saved_move = *move;
  saved_move.state = TS_LATTICE_AT(latt, move->r,move->c);

  if (!ts_lattice_is_pseudo_spin(latt, move->state))
  {
    RETURN_ON_ERROR( change_spin_and_update(latt, move->r, move->c, move->state) );
  }
  else
  {
    RETURN_ON_ERROR( invoke_pseudospin(latt, move->state, move->r, move->c) );
  }

  ret = run_hooks(latt, &latt->post_move_hooks, &saved_move);
  if (IS_ERROR(ret)) {
    RETURN_ON_ERROR(ret);
  }
  RETURN_ON_ERROR( ts_lattice_update_marked_sites(latt) );

  return TS_OK;
}

TSResult ts_lattice_change_spin(ts_lattice *latt, int r, int c, spin_t new_state)
{
  ts_lattice_move mv = { .r = r, .c = c, .state = new_state, .rate = 0.0, .next = NULL };
  RETURN_ON_ERROR( run_hooks_and_execute_move(latt, &mv) );
  return TS_OK;
}

static TSResult single_step(ts_lattice *latt, double *restrict ktotp)
{
  RETURN_ON_ERROR( ts_lattice_get_k_total(latt, ktotp) );
  ERROR_IF( TS_ERR_LATT_RATE_IS_ZERO, *ktotp == 0.0);

  double a[2] = {0.0, 0.0};
  RETURN_ON_ERROR( ts_rng_gen_uniforms(latt->rng, 2, a) );
  double dt = -log( a[0] ) / ktotp[0];
  latt->time += dt;
  latt->nsteps++;

  a[1] *= ktotp[0];

  size_t move_ind = ts_fenwick_tree_find_cum_sum(&latt->rates, a[1]);
  if (move_ind >= latt->rates.size)
  {
    return TS_NEW_ERROR3( TS_ERR_SIMULATION_STATE,
        "ERROR in rate tree??? no move matches chosen rate");
  }

  ts_lattice_move* move = latt->rates.data[move_ind];

  assert(move->r < latt->nrow || move->c < latt->ncol);

  if (CHECK_EACH_MOVE_FOR_CONSISTENCY && !site_moves_have_move(latt, move))
  {
    ts_lattice_move *movelist = ts_lattice_get_current_site_moves(latt, move->r, move->c);

    fprintf(stderr, "\n[%7d  %7.4f] move: %p (%3zu,%3zu) state = %3d  rate = %7.4f  next = %p\n",
        latt->nsteps, latt->time, move, move->r, move->c, (int)move->state.val, move->rate, move->next);

    ts_lattice_print_movelist(movelist);
    TS_RAISE_ERROR_MESG(TS_ERR_SIMULATION_STATE, "move is not part of the site move list!\n");
  }

  RETURN_ON_ERROR( check_that_movelists_are_consistent(latt) );
  RETURN_ON_ERROR( run_hooks_and_execute_move(latt, move) );
  RETURN_ON_ERROR( check_that_movelists_are_consistent(latt) );
  return TS_OK;
}

TSResult ts_lattice_single_step(ts_lattice* latt)
{
  double ktot = 0.0;

  RETURN_ON_ERROR( single_step(latt,&ktot) );
  return TS_OK;
}

TSResult ts_lattice_simulate_until(ts_lattice* latt, double t_end)
{
  while (latt->time < t_end)
  {
    double ktot = 0.0;
    TSResult ret = single_step(latt,&ktot);
    if (!IS_OKAY(ret))
    {
      if (ERRCODE(ret) == TS_ERR_LATT_RATE_IS_ZERO)
      {
        latt->time = t_end;
      }
      else
      {
        return ret;
      }
    }
  }

  return TS_OK;
}

TSResult ts_lattice_add_pspin_change(ts_lattice *latt, int r, int c, int pspin)
{
  size_t n = latt->changes.n++;

  if (n >= latt->changes.max)
  {
    size_t i;
    fprintf(stderr,
        "\n\n"
        "cannot add pseudospin update %3d,%3d,%4d : too many updates\n"
        "Current updates:\n",
        r,c,pspin);
    fprintf(stderr, "      %3s %3s %3s\n", "row", "col", "val");
    fprintf(stderr, "      %3s %3s %3s\n", "---", "---", "---");
    for (i=0; i < TS_LATTICE_MAX_UPDATES; i++)
    {
      fprintf(stderr, "[%3zu] %3d %3d %3d\n",
          i,
          latt->changes.rv[i],
          latt->changes.cv[i],
          latt->changes.vv[i].val);
    }
    fprintf(stderr, "\n");

    TS_RAISE_ERROR_MESG(TS_ERR_INSUFFICIENT_SIZE,
        "cannot add a pseudospin update: too many updates");
  }

  latt->changes.rv[n] = r;
  latt->changes.cv[n] = c;
  latt->changes.vv[n] = spin_from_int(pspin);

  return TS_OK;
}

TSResult ts_lattice_mark_site_changed(ts_lattice *latt, int r, int c)
{
  size_t n;

  n = latt->changes.n++;
  if (n < latt->changes.max)
  {
    latt->changes.rv[n] = r;
    latt->changes.cv[n] = c;
    latt->changes.vv[n] = TS_LATTICE_AT(latt,r,c);
  }

  return TS_OK;
}

TSResult ts_lattice_mark_range_for_update(ts_lattice *latt, int rmin, int rmax, int c)
{
  size_t n;

  STATIC_ASSERT(
      ARRAYLEN(latt->update_list.rminv) == ARRAYLEN(latt->update_list.rmaxv) &&
      ARRAYLEN(latt->update_list.rminv) == ARRAYLEN(latt->update_list.cv));

  assert(latt->update_list.n >= 0);
  assert(rmin <= rmax);

  n = latt->update_list.n++;

  if (n < ARRAYLEN(latt->update_list.rminv))
  {
    latt->update_list.rminv[n] = rmin;
    latt->update_list.rmaxv[n] = rmax;
    latt->update_list.cv[n] = c;
  }

  return TS_OK;
}

static inline const char* boundary_desc(int boundary)
{
  return (boundary < 0)
    ? "fixed"
    : (boundary == 0
        ? "periodic"
        : "helical");
}

TSResult ts_lattice_print_configuration(ts_lattice* latt, FILE* f)
{
  fprintf(f, "Lattice: %4lu x %4lu (%7lu sites)\n", latt->nrow, latt->ncol, latt->num_sites);
  fprintf(f, "  RNG: ");
  RETURN_ON_ERROR( ts_rng_print_configuration(latt->rng, f) );
  fprintf(f, "  Coupling distance: %d x %d sites\n", latt->row_coupling_dist, latt->col_coupling_dist);
  fprintf(f, "  Row    boundaries: %s\n", boundary_desc(latt->boundary.row));
  fprintf(f, "  Column boundaries: %s\n", boundary_desc(latt->boundary.col));
  fprintf(f, "  Num. pseudospins : %d\n", ts_lattice_num_pseudospins(latt));
  double ktot = 0.0;
  RETURN_ON_ERROR( ts_lattice_get_k_total(latt, &ktot) );

  fprintf(f,
      "\n"
      "  Simulation:\n"
      "    Temperature %7.4f T0\n"
      "    Time        %7.4f t0\n"
      "    k_tot       %7.4f 1/t0\n",
      latt->temp, latt->time, ktot);

  return TS_OK;
}


/** Connected component labeling
 *
 * The algorithm w use is described in:
 *
 * He, L., Chao, Y., Suzuki, K., "A Run-Based Two-Scan Labeling Algorithm."
 * Lecture Notes in Computer Science 4633: 131–142 (2007).
 *
 * This is a two-pass algorithm that maintains label equivalence using
 * disjoint sets, managed by linked lists. 
 */

STATIC_ASSERT( sizeof(spin_t) == sizeof(int) );
STATIC_ASSERT( offsetof(spin_t,val) == 0 );

static inline int new_label(int lbl, int*restrict repr_labels,
    int*restrict prov_list, int*restrict prov_tail);

static inline MAYBE_UNUSED
void dump_label_state(int nlabels, 
    const int*restrict repr_labels, 
    const int*restrict prov_next, 
    const int*restrict prov_tail);

static inline int assign_labels(int* nlabelsp, int nconn, int*restrict clbls, int*restrict repr_labels, int*restrict prov_list, int*restrict prov_tail);

static inline int merge_labels(int nlabels, int l1, int l2,
    int*restrict repr_labels, int*restrict prov_next, int*restrict prov_tail);

int ts_lattice_connected_components(const ts_lattice* latt, int*restrict components, int*restrict working)
{
  /* WARNING! This does not handle helical or periodic boundaries at
   * all!
   */
  size_t nrow = latt->nrow;
  size_t ncol = latt->ncol;
  size_t num_working = (nrow*ncol/4);
  memset(components, 0, latt->num_sites * sizeof(components[0]));
  memset(working, 0, 3*num_working * sizeof(working[0]));

  // at the end, we store the count of each representative label in
  // prov_list. to make this easy for callers to use, we put prov_list
  // first so callers can access the count of `lbl` via `working[lbl-1]`
  int*restrict prov_list = working;
  int*restrict repr_labels = prov_list+num_working;
  int*restrict prov_tail = repr_labels+num_working;

  const int*restrict elements = (const int*)latt->sites;
  int nlabels = 0;

#define NEWLABEL() \
  (new_label(++nlabels, repr_labels, prov_list, prov_tail))

#define MERGELABELS(l1,l2) \
    (merge_labels(nlabels, l1, l2, repr_labels, prov_list, prov_tail))

#define ASSIGNLABEL() \
  (assign_labels(&nlabels, nconn, clbls, repr_labels, prov_list, prov_tail))

#define GATHERLABEL(poss) do { \
  int _lbl = (poss);           \
  if (_lbl != 0) { clbls[nconn] = repr_labels[_lbl-1]; nconn++; } \
} while(0)

  /* special pass over first column */
  if (elements[0] != 0)
  {
    components[0] = NEWLABEL();
  }

  for (size_t i=1; i < nrow; i++)
  {
    if (elements[i] != 0)
    {
      if (components[i-1] != 0)
      {
        components[i] = components[i-1];
      }
      else
      {
        components[i] = NEWLABEL();
      }
    }
  }

  {
    const int* col = elements+nrow;
    int*restrict cc0 = components;
    int*restrict cc1 = components+nrow;
    for (size_t j=1; j < ncol; j++)
    {
      int nconn = 0;
      int clbls[4] = { 0, 0, 0, 0 };

      if (col[0] != 0)
      {
        GATHERLABEL(cc0[0]);
        GATHERLABEL(cc0[1]);

        cc1[0] = ASSIGNLABEL();
      }

      for (size_t i=1; i < nrow-1; i++)
      {
        if (col[i] != 0)
        {
          nconn = 0;
          GATHERLABEL(cc0[i-1]);
          GATHERLABEL(cc0[i  ]);
          GATHERLABEL(cc0[i+1]);
          GATHERLABEL(cc1[i-1]);

          cc1[i] = ASSIGNLABEL();
        }
      }

      if (col[nrow-1] != 0)
      {
        nconn = 0;
        GATHERLABEL(cc0[(nrow-1)-1]);
        GATHERLABEL(cc0[(nrow-1)  ]);
        GATHERLABEL(cc1[(nrow-1)-1]);

        cc1[nrow-1] = ASSIGNLABEL();
      }

      col += nrow;

      cc0 = cc1;
      cc1 += nrow;
    }
  }

#undef GATHERLABEL
#undef ASSIGNLABEL
#undef MERGELABELS
#undef NEWLABEL

  int nrepr = 0;
  // Reassign representative labels so they're consecutively numbered
  // and start at 1. Count the number of unique representative labels
  // while we do this.
  //
  // We use prov_tail[] to store the temporary mapping. The final
  // mapping is stored in repr_labels[]
  memset(prov_tail, 0, num_working*sizeof(*prov_tail));
  for(int i=0; i < nlabels; i++)
  {
    int repr = repr_labels[i];
    int final = prov_tail[repr-1];
    if (final == 0)
    {
      final = ++nrepr;
      prov_tail[repr-1] = final;
    }
    repr_labels[i] = final;
  }

  // zero out prov_list and use this to keep a count of each label
  memset(prov_list, 0, nrepr*sizeof(*prov_list));

  {
    int*restrict cc = components;
    // final pass to assign labels
    for(size_t j=0; j < ncol; j++)
    {
      for(size_t i=0; i < nrow; i++)
      {
        if (cc[i] > 0) {
          int rl = repr_labels[ cc[i] - 1 ];
          cc[i] = rl;
          prov_list[rl-1]++;
        }
      }

      cc += nrow;
    }
  }

  return nrepr;
}

static inline int new_label(int lbl, int*restrict repr_labels,
    int*restrict prov_list, int*restrict prov_tail)
{
  repr_labels[lbl-1] = lbl;
  prov_list[lbl-1] = -1;
  prov_tail[lbl-1] = lbl;
  return lbl;
}

static inline MAYBE_UNUSED
void dump_label_state(int nlabels, 
    const int*restrict repr_labels, 
    const int*restrict prov_next, 
    const int*restrict prov_tail)
{
  fprintf(stderr, "Label RLabel Next Tail\n");
  for (int i=0; i < nlabels; i++)
  {
    fprintf(stderr, "%5d %6d %4d %4d\n",
        i+1, repr_labels[i], prov_next[i], prov_tail[i]);
  }
}

static inline int merge_labels(int nlabels, int l1, int l2,
    int*restrict repr_labels, int*restrict prov_next, int*restrict prov_tail)
{
  UNUSED_PARAM(nlabels);

  int r1 = repr_labels[l1-1];
  int r2 = repr_labels[l2-1];

  int t1 = prov_tail[r1-1];
  int t2 = prov_tail[r2-1];

  int n1 = prov_next[t1-1];
  int n2 = prov_next[t2-1];

  if (r1 == r2) {
    return r1;
  }

  /* we merge the smaller list into the larger list
   *
   * to keep the code below simple, we always merge list1 into list2,
   * but we initially swap the lists if the original list2 is smaller
   * than list1.
   */
  if ((-n2 < -n1) || ((-n1 == -n2) && (r1 < r2)))
  {
    int tmp;
    tmp = r1; r1 = r2; r2 = tmp; // swap r1,r2
    tmp = t1; t1 = t2; t2 = tmp; // swap t1,t2
    tmp = l1; l1 = l2; l2 = tmp; // swap l1,l2
    // no need to swap n1 and n2
  }

  // merge l1 into l2:
  //
  // store item after the head of the l2. we merge the lists by
  // inserting the list at l1 after the head of the list l1 and then
  // appending next2 (and all subsequent items) to the end of l1.
  int next2 = prov_next[r2-1];
  prov_next[r2-1] = r1;

  // iterate over provisional labels of l1, set their representative
  // label to the representative label of l2
  int lbl = r1;
  do {
    repr_labels[lbl-1] = r2;
    prov_tail[lbl-1] = t2;
    lbl = prov_next[lbl-1];
  } while (lbl >= 0);

  // append next2 to the end of the list of l1 and update the count of
  // the merged list size
  if (next2 < 0) {
    // l2 is a single-item list
    prov_tail[r1-1] = t1;
    prov_tail[r2-1] = t1;
    prov_next[t1-1] = n1+n2;
  } else {
    prov_next[t1-1] = next2;
    prov_tail[r1-1] = t2;
    prov_next[t2-1] = n1 + n2;
  }

  return r2;
}

static inline int assign_labels(int* nlabelsp, int nconn, int*restrict clbls, int*restrict repr_labels, int*restrict prov_list, int*restrict prov_tail)
{
  if (nconn == 0) {
    (*nlabelsp)++;
    return new_label(*nlabelsp, repr_labels, prov_list, prov_tail);
  } else {
    int r0 = repr_labels[clbls[0]-1];
    for (int k=1; k < nconn; k++) {
      int r1 = repr_labels[clbls[k]-1];
      if (r1 != r0) {
        r0 = merge_labels(*nlabelsp,clbls[0],clbls[k],repr_labels, prov_list,prov_tail);
      }
    }
    return r0;
  }
}

TSResult ts_lattice_write_plot(ts_lattice* latt, FILE* f)
{
  if ((latt->nrow == 0) || (latt->ncol == 0)) {
    return TS_OK;
  }

  for(size_t i=0; i < latt->nrow; i++) {
    int val = ts_lattice_at(latt, i,0).val;
    if (fprintf(f, "%d", val) < 0) {
      goto error;
    }

    for(size_t j=1; j < latt->ncol; j++) {
      val = ts_lattice_at(latt, i,j).val;
      if (fprintf(f, " %d", val) < 0) {
        goto error;
      }
    }
    if (fprintf(f,"\n") < 0) {
      goto error;
    }
  }

  if (fprintf(f,"\n\n") < 0) {
    goto error;
  }

  return TS_OK;

error:
  TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
      "Error writing lattice plot: %s", strerror(errno));
}

struct plot_reader
{
  /* for error reporting */
  size_t bytes;
  size_t line;
  size_t lpos;

  int row;
  int col;
};

static TSResult process_row(ts_lattice* latt, size_t row, size_t line_off, char* buf, size_t n, size_t* nreadp)
{
  size_t nread = 0;
  size_t col = 0;

  size_t ncol = latt->ncol;

  char ws[2];
  do
  {
    int val,nc = 0;

    if (2 != sscanf(buf+nread, "%d%1[ \n]%n", &val, ws, &nc))
    {
      TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
          "Error reading lattice plot at line %zu, position %zu: %s", 
          line_off + row + 1, nread+1, buf);
    }

    TS_LATTICE_SPIN(latt,row,col) = val;
    col++;
    nread += nc;
  } while ((ws[0] != '\n') && (col < ncol) && (nread < n));

  if ((ws[0] != '\n') || (col < ncol))
  {
    TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
        "Line %zu disagrees with number of columns (%d sites expected, %d sites read):\n%s", 
        line_off+row+1, ncol, col, buf);
  }

  *nreadp = nread;
  return TS_OK;
}

TSResult ts_lattice_read_plot(ts_lattice* latt, FILE* f)
{
  static char buf[1024];
  long pos;

  int line = 0;
  while (line < (int)latt->nrow)
  {
    size_t nread = 0;
    size_t n;

    pos = ftell(f);
    memset(buf,0,sizeof(buf));
    if (fgets(buf,sizeof(buf),f) == NULL)
    {
      if (feof(f))
      {
        TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
            "End of file encountered before lattice was read", 0);
      }
      else
      {
        TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
            "Error reading lattice plot: %s", strerror(errno));
      }
    }

    n = strlen(buf);
    if (buf[n-1] != '\n')
    {
      TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
          "Input buffer is too small for the current assumptions of the code. Fix the code.", 0);
    }

    RETURN_ON_ERROR( process_row(latt, line, 0, buf, n, &nread) );
    pos += nread;
    line++;
  }

  // read next two lines, which should be empty
  {
    int i, ch;
    for (i=0; i < 2; i++)
    {
      pos = ftell(f);
      memset(buf,0,sizeof(buf));
      if (fgets(buf,sizeof(buf),f) == NULL)
      {
        TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
            "End of file encountered before lattice delimiters were read", 0);
      }
      line++;

      if (buf[0] != '\n' && buf[1] != '\0')
      {
        TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
            "Invalid delimiters.  Expected a blank line, found: %s", buf);
      }
    }

    // look for excess delimitters
    do {
      ch = getc(f);
    } while (ch == '\n');
    if (ch != EOF) { ungetc(ch,f); }
  }

  return TS_OK;
}

TSResult ts_lattice_find_last_rows(ts_lattice* latt, int*rows)
{
  for (int j=0; j < (int)latt->ncol; j++)
  {
    rows[j] = -1;
    for (int i=latt->nrow-1; i >= 0; i--)
    {
      if (TS_LATTICE_SPIN(latt,i,j) != 0)
      {
        rows[j] = i;
        break;
      }
    }
  }

  return TS_OK;
}

void ts_lattice_find_column_row_range(ts_lattice *latt, int rows[static 2], int col)
{
  int nrow,ncol,minr,maxr;
  spin_t *restrict cvals;

  cvals = ts_lattice_ref(latt,0,col);
  nrow = latt->nrow;

  for (minr=0; (minr < nrow) && (cvals[minr].val == 0); minr++)
  {
    continue;
  }

  if (minr == nrow)
  {
    rows[0] = rows[1] = -1;
    return;
  }

  for (maxr=nrow-1; (maxr > minr) && (cvals[maxr].val == 0); maxr--)
  {
    continue;
  }

  rows[0] = minr;
  rows[1] = maxr;
}

void ts_lattice_find_row_range_fast(ts_lattice *latt, int *rows)
{
  int nrow,ncol;

  nrow = latt->nrow;
  ncol = latt->ncol;

  for (int j=0; j < ncol; j++)
  {
    ts_lattice_find_column_row_range(latt, &rows[2*j], j);
  }
}

TSResult ts_lattice_find_row_range(ts_lattice *latt, int *rows)
{
  int nrow,ncol;
  if (latt->ncol > INT_MAX)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_VALUE,
        "%zu columns exceeds what can be held in an `int` value", latt->ncol);
  }

  if (latt->nrow > INT_MAX)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_VALUE,
        "%zu rows exceeds what can be held in an `int` value", latt->nrow);
  }

  ts_lattice_find_row_range_fast(latt, rows);
  return TS_OK;
}

