#ifndef TS_LATTICE__H
#define TS_LATTICE__H

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>

#include "ts_common.h"
#include "ts_errors.h"
#include "ts_rng.h"
#include "ts_fenwick.h"

#define MODULE_NAME TS_LATTICE

typedef union { signed int val; } spin_t;
static inline spin_t spin_from_int(int v) { spin_t s = { .val = v }; return s; }
static const spin_t INVALID_SPIN = { .val = INT_MIN };
static const spin_t ZERO_SPIN    = { .val = 0 };

#define TS_LATTICE_DEFAULT_COUPLING_DIST 1
#define TS_LATTICE_PSEUDOSPIN_OFFSET     1023

#define TS_LATTICE_PSPIN(ind)   ( TS_LATTICE_PSEUDOSPIN_OFFSET + (ind) )

extern int ts_latt_debug_updates;

typedef struct ts_lattice ts_lattice;

struct ts_lattice_coord { int r; int c; };

typedef struct ts_lattice_move ts_lattice_move;
struct ts_lattice_move
{
  size_t r;
  size_t c;

  spin_t state;
  double rate;

  ts_lattice_move *next;
};

TSResult ts_lattice_request_move(ts_lattice *latt, ts_lattice_move **movesp);
void ts_lattice_release_move(ts_lattice *latt, ts_lattice_move *move);

typedef int (*ts_lattice_gen_move_func)(void *data,
    ts_lattice *latt, const int *restrict rv, const int *restrict cv, size_t nsites,
    spin_t *states, double *rates, size_t *restrict siteinds, int move_max,
    intptr_t *cdata, TSResult *errp);

struct ts_lattice_movetype
{
  const char *name;
  ts_lattice_gen_move_func func;
  void *closure;
};

typedef struct ts_lattice_move_iterator ts_lattice_move_iterator;
struct ts_lattice_move_iterator
{
  ts_lattice_move curr;
  size_t ind;
};

#define TS_LATTICE_BOUNDARY_FIXED    -1
#define TS_LATTICE_BOUNDARY_PERIODIC  0

typedef TSResult (*ts_lattice_pseudospin_func)(void *closure, ts_lattice *latt, int r, int c, spin_t pspin);

struct ts_lattice_hook;

struct ts_lattice_pspins;
struct ts_lattice_moves;
struct ts_lattice_movepool;

enum { TS_LATTICE_MAX_UPDATES = 64 };
enum { TS_LATTICE_MOVE_HISTORY_SIZE = 16 };

struct ts_lattice_update_list {
  int n;
  int rminv[TS_LATTICE_MAX_UPDATES];
  int rmaxv[TS_LATTICE_MAX_UPDATES];
  int cv[TS_LATTICE_MAX_UPDATES];
};

struct ts_lattice_change_list {
  size_t n;
  size_t max;
  int *rv;
  int *cv;
  spin_t *vv;
};

struct ts_lattice
{
  size_t nrow;
  size_t ncol;
  size_t num_sites;
  spin_t *sites;

  int refcount;

  int row_coupling_dist;
  int col_coupling_dist;

  /* boundary conditions:
   *
   * <0   fixed boundary
   *  0   periodic boundary
   * >0   helical boundary (value is the pitch in the other dimension)
   */
  struct {
    int row;
    int col;
  } boundary;

  double time;
  double temp;
  double inv_temp;

  int nsteps;
  struct ts_rng *rng;

  ts_fenwick_tree rates;

  /* called when lattice is released */
  void (*destructor)(void *p);

  struct {
    void (*callback)(struct ts_lattice *, void *);
    void *user_data;
  } rebuild_moves_hook;

  struct ts_lattice_moves *site_moves;

  DATAVECTOR(struct ts_lattice_movetype) movetypes;
  struct ts_lattice_pspins *pseudospins;

  struct ts_lattice_change_list changes;

  struct ts_lattice_update_list update_list;

  struct ts_lattice_movepool *movepool;

  DATAVECTOR(struct ts_lattice_hook) move_hooks;
  DATAVECTOR(struct ts_lattice_hook) post_move_hooks;

  struct {
    size_t last;
    struct ts_lattice_move moves[TS_LATTICE_MOVE_HISTORY_SIZE];
  } move_history;
};

typedef TSResult (*ts_lattice_hook_func)(
    void *closure,
    ts_lattice *latt, 
    const ts_lattice_move *move);

TSResult ts_lattice_add_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure);
int ts_lattice_remove_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure);

TSResult ts_lattice_add_post_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure);
int ts_lattice_remove_post_move_hook(ts_lattice *latt, ts_lattice_hook_func func, void *closure);

static inline void ts_lattice_set_temp(ts_lattice *latt, double temp)
{
  latt->temp = temp;
  latt->inv_temp = 1.0/temp;
}

void
ts_lattice_wrap_coord_inner(const ts_lattice *latt, int *restrict rp, int *restrict cp);

static inline void
ts_lattice_wrap_coord_column(const ts_lattice *latt, int *restrict rp, int *restrict cp)
{
  int c,r, ncol,nrow, bcol,brow;

  c = *cp;
  r = *rp;

  ncol = latt->ncol;
  nrow = latt->nrow;

  bcol = latt->boundary.col;
  brow = latt->boundary.row;

  /* fast out */
  if ((c >= 0) && (c < ncol))
  {
    return;
  }

  if (c < 0)
  {
    *cp = (bcol != TS_LATTICE_BOUNDARY_FIXED) ? c+ncol : 0;
    *rp = (bcol != TS_LATTICE_BOUNDARY_FIXED) ? r-bcol : r;
  } 
  else
  {
    *cp = (bcol != TS_LATTICE_BOUNDARY_FIXED) ? c-ncol : ncol-1;
    *rp = (bcol != TS_LATTICE_BOUNDARY_FIXED) ? r+bcol : r;
  }
}

#define TS_LATTICE_WRAP_WITH_BRANCHING
static inline void
ts_lattice_wrap_coord(const ts_lattice *latt, int *restrict rp, int *restrict cp)
{
  int c,r, ncol,nrow, bcol,brow;
  int bmask, mask1, mask2;
  int fixed;

  ncol = latt->ncol;
  nrow = latt->nrow;

  bcol = latt->boundary.col;
  brow = latt->boundary.row;

#ifdef TS_LATTICE_WRAP_WITH_BRANCHING
  UNUSED_PARAM(bmask);
  UNUSED_PARAM(mask1);
  UNUSED_PARAM(mask2);
  UNUSED_PARAM(fixed);
  UNUSED_PARAM(c);

  ts_lattice_wrap_coord_column(latt, rp,cp);

  r = *rp;

  /* fast out */
  if ((r >= 0) && (r < nrow)) { return; }

  if (r < 0)
  {
    *rp = (brow != TS_LATTICE_BOUNDARY_FIXED) ? nrow - r : 0;
  }
  else
  {
    *rp = (brow != TS_LATTICE_BOUNDARY_FIXED) ? r - nrow : nrow-1;
  }
#else
  bmask = -(bcol == TS_LATTICE_BOUNDARY_FIXED);
  mask1 = -(c<0);
  mask2 = -(c>=ncol);

  fixed = (mask1 & -c) | (mask2 & (ncol-1-c)) | (~mask1 & ~mask2 & 0);

  c = c + ((bmask & fixed) | (~bmask & ((mask1 &  ncol) | (mask2 & -ncol))));
  r = r + (~bmask & ((mask1 & -bcol) | (mask2 &  bcol)));

  bmask = -(brow == TS_LATTICE_BOUNDARY_FIXED);
  mask1 = -(r<0);
  mask2 = -(r>= nrow);

  fixed = (mask1 & 0) | (mask2 & (nrow-1)) | (~mask1 & ~mask2 & r);

  r = (~mask1 & ~mask2 & r) |
    (mask1 & ( (fixed & 0       ) | (~fixed & (nrow-1)) )) |
    (mask2 & ( (fixed & (nrow-1)) | (~fixed & 0       ) ))
    ;

  *rp = r;
  *cp = c;
#endif /* TS_LATTICE_WRAP_WITH_BRANCHING */
}

TSResult ts_lattice_initialize(ts_lattice *latt, size_t nrow, size_t ncol, size_t rate_size);
void ts_lattice_finalize(ts_lattice *latt);

ts_lattice *ts_lattice_new(size_t nrow, size_t ncol, size_t rate_size, TSResult *errp);

ts_lattice *ts_lattice_retain(ts_lattice *latt);

void ts_lattice_release(ts_lattice *latt);

int ts_lattice_refcount(ts_lattice *latt);

void ts_lattice_zero(ts_lattice *latt);

/* Resets the lattice to its initial state. */
void ts_lattice_reset(ts_lattice *latt);

TSResult ts_lattice_resize_changes(ts_lattice *latt, size_t max);

TSResult ts_lattice_save_changes(ts_lattice *latt,
    struct ts_lattice_change_list *changes,
    struct ts_lattice_update_list *updates);

TSResult ts_lattice_restore_changes(ts_lattice *latt,
    struct ts_lattice_change_list *changes,
    struct ts_lattice_update_list *updates);

TSResult ts_lattice_change_list_resize(struct ts_lattice_change_list *changes, size_t max);

void ts_lattice_change_list_finalize(struct ts_lattice_change_list *changes);

static inline size_t ts_lattice_index(const ts_lattice *latt, size_t r, size_t c) { return r + latt->nrow*c; }
static inline spin_t* ts_lattice_ref(const ts_lattice *latt, size_t r, size_t c) { return latt->sites + ts_lattice_index(latt,r,c); }
static inline spin_t ts_lattice_at(const ts_lattice *latt, size_t r, size_t c) { return ts_lattice_ref(latt,r,c)[0]; }
#define TS_LATTICE_AT(l,r,c) ( ts_lattice_ref((ts_lattice*)(l),(r),(c))[0] )
#define TS_LATTICE_SPIN(l,r,c) ( (TS_LATTICE_AT(l,r,c)).val )

TSResult ts_lattice_add_movetype(ts_lattice *latt, struct ts_lattice_movetype *movetype);
TSResult ts_lattice_register_pseudospin(ts_lattice *latt, spin_t pseudo, ts_lattice_pseudospin_func func, void *closure);
int ts_lattice_num_pseudospins(const ts_lattice *latt);

static inline int ts_lattice_is_pseudo_spin_value(ts_lattice *latt, int val)
{
  UNUSED_PARAM(latt);
  return val >= TS_LATTICE_PSEUDOSPIN_OFFSET;
}

static inline int ts_lattice_is_pseudo_spin(ts_lattice *latt, spin_t spin) {
  return ts_lattice_is_pseudo_spin_value(latt,spin.val);
}

TSResult ts_lattice_invoke_pseudospin(ts_lattice *latt, spin_t pseudo, int r, int c);
TSResult ts_lattice_invoke_pseudospin_without_update(ts_lattice *latt, spin_t pseudo, int r, int c);

TSResult ts_lattice_rollback_changes(ts_lattice *latt);

TSResult ts_lattice_get_all_moves(ts_lattice *latt, 
    ts_lattice_move **moves_ptr, size_t *nmovesp, double *ktotp);
TSResult ts_lattice_move_iterator_init(ts_lattice *latt, ts_lattice_move_iterator *iter);
TSResult ts_lattice_moves_iterate(ts_lattice *latt, ts_lattice_move_iterator *iter);

TSResult ts_lattice_rebuild_moves(ts_lattice *latt);

TSResult ts_lattice_get_k_total(ts_lattice *latt, double *ktotp);

ts_lattice_move *ts_lattice_get_current_site_moves(ts_lattice *latt,
    size_t r, size_t c);
TSResult ts_lattice_set_current_site_moves(ts_lattice *drv,
    size_t r, size_t c, ts_lattice_move *m);

TSResult ts_lattice_update_current_site_moves(ts_lattice *latt,
    size_t r, size_t c, spin_t *states, double *rates, size_t nmoves);

TSResult ts_lattice_change_spin(ts_lattice *latt,
    int r, int c, spin_t new_state);

void ts_lattice_reset_rate_tree(ts_lattice *latt);

TSResult ts_lattice_mark_site_changed(ts_lattice *latt, int r, int c);
TSResult ts_lattice_add_pspin_change(ts_lattice *latt, int r, int c, int pspin);

TSResult ts_lattice_mark_range_for_update(ts_lattice *latt, int rmin, int rmax, int c);

TSResult ts_lattice_update_marked_sites(ts_lattice *latt);

TSResult ts_lattice_single_step(ts_lattice *latt);
TSResult ts_lattice_simulate_until(ts_lattice *latt, double t_end);

TSResult ts_lattice_print_configuration(ts_lattice *latt, FILE *f);

// components must have NROW x NCOL elements
// working must have 3 x (NROW/2) x (NCOL/2) elements.
int ts_lattice_connected_components(const ts_lattice *latt, int *restrict components, int *restrict working);
static inline size_t ts_lattice_connected_components_working_size(const ts_lattice *latt)
{
  return 3 * (latt->nrow * latt->ncol) / 4;
}

TSResult ts_lattice_write_plot(ts_lattice *latt, FILE *f);

// WARNING: ts_lattice_read_plot can only be called from one thread at a
// time
TSResult ts_lattice_read_plot(ts_lattice *latt, FILE *f);

// WARNING: ts_lattice_read_plot_from_stream can only be called from one thread at a
// time
TSResult ts_lattice_read_plot_from_stream(ts_lattice *latt, char *data, size_t len);

TSResult ts_lattice_find_last_rows(ts_lattice *latt, int *rows);
TSResult ts_lattice_find_row_range(ts_lattice *latt, int *rows);

void ts_lattice_find_column_row_range(ts_lattice *latt, int rows[static 2], int col);
void ts_lattice_find_row_range_fast(ts_lattice *latt, int *rows);

#undef MODULE_NAME

#endif /* TS_LATTICE__H */

