#include "ts_minimize.h"

#include <limits.h>

#include "ts_common.h"
#include "ts_errors.h"
#include "ts_units.h"
#include "ts_bonds.h"
#include "ts_dof_map.h"
#include "ts_dispatch.h"

#include "lbfgs.h"

#include "ts_malloc_protect.h"

#include "ts_machdeps.h"

#define STEEPEST_DEBUG 0

#define PARTICLE_NUM_DOF 3
#define RIGID_BODY_NUM_DOF 6

#define STEEPEST_ROT_FACTOR 0.05

#define _L(...) TS_LOG( sim, TS_LOG_INFO, TS_LOG_MINIMIZER, __VA_ARGS__ )

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
};

void ts_minimize_initialize_params(struct ts_minimize_params* params)
{
  /* reuse some of the defaults from liblbfgs */
  lbfgs_parameter_t lbfgs_params;
  lbfgs_parameter_init(&lbfgs_params);

  params->method = TS_MINIMIZE_DEFAULT;
  params->what   = TS_MINIMIZE_ALL;

  params->use_nonbonds = 0;
  params->rescale_dofs = 0;

  params->report_freq = 100;
  params->max_iter = lbfgs_params.max_iterations;

  /* epsilon convergence criteria: ||g|| < epsilon * max(1,||x||)
   *
   * default is 1e-5
   */
  params->eps = lbfgs_params.epsilon;

  params->minstep = lbfgs_params.min_step;
  params->maxstep = lbfgs_params.max_step;

  params->prec = lbfgs_params.xtol;

  params->step0 = 1e-4; /* initial step size for steepest-descent minimizer */
};

struct ts_minimization_state {
  ts_dof_map dof_map;
  ts_sim* sim;

  VecReal *origPos;
  VecReal *newPos;
  VecReal *prevPos;
  VecReal *currPos;
  VecReal *bestPos;

  VecReal* currGrad;

  VecReal *scales;
  VecReal trans[3];
  VecReal scale[3];

  VecReal linstep;
  VecReal rotstep;

  int report_freq;
  int rescale;
  int num_fevals;
};

static inline void ts_minimize_copy_positions_to_xvec(
    struct ts_minimization_state *state,
    lbfgsfloatval_t*restrict x)
{
  ts_dof_to_vec(state->sim,&state->dof_map, x, TS_ROT_AXIS_ANGLE);
}

static inline void ts_minimize_copy_positions_from_xvec(
    struct ts_minimization_state *state,
    const lbfgsfloatval_t*restrict x, double maxdisp[2])
{
  ts_dof_from_vec(state->sim,&state->dof_map, x, maxdisp, TS_ROT_AXIS_ANGLE);
}

static inline void ts_minimize_copy_forces_to_gvec(
    struct ts_minimization_state *state,
    lbfgsfloatval_t*restrict g)
{
  ts_sim *sim;
  double en = 0.0;

  sim = state->sim;
  ts_calculate_forces(sim, &en, NO_TERM_GROUPS, TERMWISE_FALSE);
  ts_update_rigid_bodies(sim);
  ts_dof_forces_to_gradient_vec(sim, &state->dof_map, g);

  if (state->rescale)
  {
  }
}

static inline void minimize_eval_set_pos(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step,
    double maxdisp[2])
{
  UNUSED_PARAM(g);
  UNUSED_PARAM(step);
  UNUSED_PARAM(n);
  tsAssert( n == (int)state->dof_map.num_dof );

  ts_minimize_copy_positions_from_xvec(
      state,
      x,
      maxdisp);
}

static inline void minimize_eval_maybe_rebuild_nonbonds(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step,
    double maxdisp[2])
{
  UNUSED_PARAM(x);
  UNUSED_PARAM(g);
  UNUSED_PARAM(n);
  UNUSED_PARAM(step);

  int rebuild_nonbonds = 0;
  if ((state->sim->nonbond_cutoff == 0.0) || (state->sim->nonbond_padding = 0.0))
  {
    rebuild_nonbonds = 1;
  }
  else if ((maxdisp[0] + maxdisp[1]) >= (state->sim->nonbond_cutoff + state->sim->nonbond_padding))
  {
    rebuild_nonbonds = 1;
  }

  if (rebuild_nonbonds)
  {
    ts_build_nonbonded_potential_terms(state->sim);
  }
}

static inline double minimize_eval_calc_grad_and_energy(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step)
{
  UNUSED_PARAM(x);
  UNUSED_PARAM(n);
  UNUSED_PARAM(step);

  double en = 0.0;
  ts_calculate_forces(state->sim, &en, NO_TERM_GROUPS, TERMWISE_FALSE);
  ts_minimize_copy_forces_to_gvec(state, g);

  return en;
}

static lbfgsfloatval_t minimize_evaluate_no_nonbonds(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step)
{
  tsAssert( (state->dof_map.num_dof <= INT_MAX) && (n == (int)state->dof_map.num_dof) );

  double maxdisp[2] = { 0.0, 0.0 };
  minimize_eval_set_pos(state, x, g, n, step, maxdisp);
  double en = minimize_eval_calc_grad_and_energy(state, x, g, n, step);

  return (lbfgsfloatval_t) en;
}

static lbfgsfloatval_t minimize_evaluate_with_nonbonds(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step)
{
  tsAssert( (state->dof_map.num_dof <= INT_MAX) && (n == (int)state->dof_map.num_dof) );

  double maxdisp[2] = { 0.0, 0.0 };
  minimize_eval_set_pos(state, x, g, n, step, maxdisp);
  minimize_eval_maybe_rebuild_nonbonds(state, x, g, n, step, maxdisp);
  double en = minimize_eval_calc_grad_and_energy(state, x, g, n, step);

  return (lbfgsfloatval_t) en;
}

static int lbfgs_progress(
    struct ts_minimization_state* state,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step,
    int n,
    int k,
    int ls
    )
{
  UNUSED_PARAM(x);
  UNUSED_PARAM(g);
  UNUSED_PARAM(xnorm);
  UNUSED_PARAM(gnorm);
  UNUSED_PARAM(step);
  UNUSED_PARAM(n);

  state->num_fevals += ls;
  if ((state->report_freq == 0) || (k % state->report_freq == 0))
  {
    TS_LOG(state->sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
        "L-BFGS iter %d  energy %f  fevals %d  |grad|=%e  |x|=%e\n",
        k, fx, state->num_fevals, gnorm,xnorm);
  }

  return 0;
}

static TSResult minimizer_init_dof(
    const struct ts_minimize_params *params,
    ts_sim *sim,
    ts_dof_map *dof_map)
{
  switch (params->what)
  {
    case TS_MINIMIZE_POS:
      RETURN_ON_ERROR( ts_dof_map_init_with_type(dof_map, sim, TS_DOF_ROTATIONAL) );
      break;

    case TS_MINIMIZE_ORI:
      RETURN_ON_ERROR( ts_dof_map_init_with_type(dof_map, sim, TS_DOF_ROTATIONAL) );
      break;

    case TS_MINIMIZE_ALL:
      RETURN_ON_ERROR( ts_dof_map_init(dof_map, sim) );
      break;

    default:
      TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
          "unknown minimizer WHAT value: 0x%02x\n",
          params->what);
  }

  return TS_OK;
}

#define HALF 0.5

static TSResult setup_state_rescale(struct ts_minimization_state *state)
{
  ts_sim *sim;
  VecReal mincoord[3], maxcoord[3];
  VecReal *scales;
  size_t i;

  sim = state->sim;

  (void)scales;
  /*
  scales = ts_alloc(NO_ALLOCATOR, state->dof_map.num_dof * sizeof(*scales));
  if (scales == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_OUT_OF_MEMORY, "could not allocate space for rescaling");
  }
  */

  mincoord[0] = mincoord[1] = mincoord[2] =  HUGE_VAL;
  maxcoord[0] = maxcoord[1] = maxcoord[2] = -HUGE_VAL;

  ts_update_rigid_bodies(sim);
  for (i=0; i < sim->num_particles; i++)
  {
    ts_particle *p = ts_get_particle(sim,i);
    if (ts_particle_is_free(p) || ts_particle_is_rigid_com(p))
    {
      VecReal x,y,z;

      x = PARTICLE_PX(sim,i);
      y = PARTICLE_PY(sim,i);
      z = PARTICLE_PZ(sim,i);

      if (x < mincoord[0]) { mincoord[0] = x; }
      if (x > maxcoord[0]) { maxcoord[0] = x; }

      if (y < mincoord[1]) { mincoord[1] = y; }
      if (y > maxcoord[1]) { maxcoord[1] = y; }

      if (z < mincoord[2]) { mincoord[2] = z; }
      if (z > maxcoord[2]) { maxcoord[2] = z; }
    }
  }

  for (i=0; i < 3; i++)
  {
    state->trans[i] = HALF*(mincoord[i] + maxcoord[i]);
    state->scale[i] = maxcoord[i] - mincoord[i];
  }

  /* FIXME: this should log somewhere! */
  fprintf(stderr, "System size: %f x %f x %f   Center: %f, %f, %f\n",
      state->scale[0], state->scale[1], state->scale[2],
      state->trans[0], state->trans[1], state->trans[2]);

  return TS_OK;
}

static TSResult initialize_state(struct ts_minimization_state *state,
    const struct ts_minimize_params *params, ts_sim *sim, int report_freq)
{
  size_t ndof;
  int rescale;

  memset(state, 0, sizeof(*state));

  state->sim = sim;
  state->num_fevals = 0;
  ndof = state->dof_map.num_dof;

  RETURN_ON_ERROR( minimizer_init_dof(params, sim, &state->dof_map) );

  state->bestPos = lbfgs_malloc(ndof * sizeof(state->bestPos[0]));
  if (state->bestPos == NULL)
  {
    ts_dof_map_finalize(&(state->dof_map));
    return TS_NEW_ERROR3(TS_ERR_OUT_OF_MEMORY,
        "minimizer could not allocate XV vector");
  }

  rescale = params->rescale_dofs;
  state->rescale = rescale;
  state->scales = NULL;
  if (rescale)
  {
    TSResult err = setup_state_rescale(state);
    if (IS_ERROR(err))
    {
      /* FIXME: this should be logged! */
      fprintf(stderr, "Error when initializing rescaling.  Rescaling disabled!\n");
      state->rescale = 0;
    }
  }

  state->report_freq = report_freq;

  return TS_OK;
}

static void finalize_state(struct ts_minimization_state* state)
{
  ts_dof_map_finalize(&(state->dof_map));

  if (state->rescale && state->scales)
  {
    ts_free(NO_ALLOCATOR, state->scales);
  }

  lbfgs_free(state->bestPos);
}

static TSResult report_minimizer_result(const ts_sim* sim, int ret);
static void report_minimizer_params(const ts_sim* sim, const lbfgs_parameter_t* params);

// TSResult ts_minimize_energy(ts_sim* sim, int max_iter, int nonbonds, int report_freq)
TSResult ts_minimize_lbfgs(ts_sim* sim, const struct ts_minimize_params* params)
{
  struct ts_minimization_state state;
  lbfgs_parameter_t lbfgs_params;
  size_t ndof;
  double prev_padding, maxdisp[2];
  int prev_nonbonds_enabled, ret;

  if (sim->num_particles == 0) return TS_OK;

  // Force full generation of nonbond lists
  prev_padding = sim->nonbond_padding;
  sim->nonbond_padding = 0.0;

  prev_nonbonds_enabled = sim->nonbonds_enabled;
  sim->nonbonds_enabled = params->use_nonbonds;

  RETURN_ON_ERROR( ts_build_potential(sim) );
  RETURN_ON_ERROR( initialize_state(&state, params, sim, params->report_freq) );

  ndof = state.dof_map.num_dof;

  lbfgs_parameter_init(&lbfgs_params);
  lbfgs_params.max_iterations = params->max_iter;
  lbfgs_params.epsilon = params->eps;

  report_minimizer_params(sim, &lbfgs_params);

  ts_minimize_copy_positions_to_xvec(&state, state.bestPos);

  enable_user_interrupt();
  lbfgsfloatval_t fx = 0.0;

  lbfgs_evaluate_t evalfunc = sim->nonbonds_enabled
    ? (lbfgs_evaluate_t) &minimize_evaluate_with_nonbonds
    : (lbfgs_evaluate_t) &minimize_evaluate_no_nonbonds;

  ret = lbfgs(ndof,
      state.bestPos, &fx,
      (lbfgs_evaluate_t)evalfunc,
      (lbfgs_progress_t)&lbfgs_progress,
      &state,
      &lbfgs_params);
  restore_user_interrupt();

  maxdisp[0] = maxdisp[1] = 0.0;
  ts_minimize_copy_positions_from_xvec(&state, state.bestPos, maxdisp);

  finalize_state(&state);

  RETURN_ON_ERROR( report_minimizer_result(sim, ret) );

  // Restore any previous padding
  sim->nonbond_padding = prev_padding;
  sim->nonbonds_enabled = prev_nonbonds_enabled;
  RETURN_ON_ERROR( ts_build_potential(sim) );

  return TS_OK;
}

static TSResult report_minimizer_result(const ts_sim* sim, int ret)
{
#define _E(retval, err) \
    case retval: return TS_NEW_ERROR(err)

#define _EL(retval, err, mesg) \
    case retval:               \
      TS_LOG(sim,TS_LOG_ERROR,TS_LOG_MINIMIZER, mesg "\n"); \
      return TS_NEW_ERROR3(err,mesg)
#define _ELmin(retval, mesg) _EL(retval, TS_ERR_MINIMIZATION_ERROR, mesg)
#define _ELinc(retval, mesg) _EL(retval, TS_ERR_MINIMIZATION_INCOMPLETE, mesg)

#define _OK(retval,mesg) \
    case retval:         \
      TS_LOG(sim, TS_LOG_INFO, TS_LOG_MINIMIZER, mesg); \
      return TS_OK

  TS_LOG(sim, TS_LOG_INFO,TS_LOG_MINIMIZER, "L-BFGS returned %d\n", ret);
  switch (ret) {
    _OK(LBFGS_CONVERGENCE,      "L-BFGS converged\n");
    _OK(LBFGS_STOP,             "L-BFGS stopped\n");
    _OK(LBFGS_ALREADY_MINIMIZED, "L-BFGS already minimized\n");

    _E(LBFGSERR_LOGICERROR, TS_ERR_LOGIC_ERROR);

    _E(LBFGSERR_OUTOFMEMORY, TS_ERR_OUT_OF_MEMORY);
    _E(LBFGSERR_CANCELED   , TS_CANCELLED);

    _ELmin(LBFGSERR_INVALID_N,                  "error is INVALID_N"                );
    _ELmin(LBFGSERR_INVALID_N_SSE,              "error is INVALID_N_SSE"            );
    _ELmin(LBFGSERR_INVALID_X_SSE,              "error is INVALID_X_SSE"            );
    _ELmin(LBFGSERR_INVALID_EPSILON,            "error is INVALID_EPSILON"          );
    _ELmin(LBFGSERR_INVALID_TESTPERIOD,         "error is INVALID_TESTPERIOD"       );
    _ELmin(LBFGSERR_INVALID_DELTA,              "error is INVALID_DELTA"            );
    _ELmin(LBFGSERR_INVALID_LINESEARCH,         "error is INVALID_LINESEARCH"       );
    _ELmin(LBFGSERR_INVALID_MINSTEP,            "error is INVALID_MINSTEP"          );
    _ELmin(LBFGSERR_INVALID_MAXSTEP,            "error is INVALID_MAXSTEP"          );
    _ELmin(LBFGSERR_INVALID_FTOL,               "error is INVALID_FTOL"             );
    _ELmin(LBFGSERR_INVALID_WOLFE,              "error is INVALID_WOLFE"            );
    _ELmin(LBFGSERR_INVALID_GTOL,               "error is INVALID_GTOL"             );
    _ELmin(LBFGSERR_INVALID_XTOL,               "error is INVALID_XTOL"             );
    _ELmin(LBFGSERR_INVALID_MAXLINESEARCH,      "error is INVALID_MAXLINESEARCH"    );
    _ELmin(LBFGSERR_INVALID_ORTHANTWISE,        "error is INVALID_ORTHANTWISE"      );
    _ELmin(LBFGSERR_INVALID_ORTHANTWISE_START,  "error is INVALID_ORTHANTWISE_START");
    _ELmin(LBFGSERR_INVALID_ORTHANTWISE_END,    "error is INVALID_ORTHANTWISE_END"  );
    _ELmin(LBFGSERR_OUTOFINTERVAL,              "error is OUTOFINTERVAL"            );
    _ELmin(LBFGSERR_INCORRECT_TMINMAX,          "error is INCORRECT_TMINMAX"        );

    _ELmin(LBFGSERR_MINIMUMSTEP,                "error is MINIMUMSTEP"              );
    _ELmin(LBFGSERR_MAXIMUMSTEP,                "error is MAXIMUMSTEP"              );
    _ELmin(LBFGSERR_MAXIMUMLINESEARCH,          "error is MAXIMUMLINESEARCH"        );
    _ELmin(LBFGSERR_WIDTHTOOSMALL,              "error is WIDTHTOOSMALL"            );
    _ELmin(LBFGSERR_INVALIDPARAMETERS,          "error is INVALIDPARAMETERS"        );
    _ELmin(LBFGSERR_INCREASEGRADIENT,           "error is INCREASEGRADIENT"         );

    _ELinc(LBFGSERR_ROUNDING_ERROR,             "incomplete because of ROUNDING_ERROR"   );
    _ELinc(LBFGSERR_MAXIMUMITERATION,           "incomplete because of MAXIMUMITERATION" );

    case LBFGSERR_UNKNOWNERROR:   
    default:
      TS_LOG(sim, TS_LOG_ERROR, TS_LOG_MINIMIZER, "error is UNKNOWN");
      return TS_NEW_ERROR( TS_ERR_UNKNOWN_ERROR );
  }

  tsAssert( ret >= 0 );

#undef _OK
#undef _E
#undef _EL
#undef _ELmin

}

static void report_minimizer_params(const ts_sim* sim, const lbfgs_parameter_t* params)
{
  _L("---------------[ Minimizer parameters ]---------------\n");
  _L("m        Corrections to inverse Hessian matrix  %d\n", params->m);
  _L("\n");
  _L("epsilon  Epsilon for convergence test           %e\n", params->epsilon);
  _L("delta    Delta for convergence test             %e\n", params->delta);
  _L("past     Distance for delta-convergence test    %d\n", params->past);
  _L("\n");
  _L("max_iterations  Maximum number of iterations    %d\n", params->max_iterations);
  _L("\n");
  _L("linesearch   Linesearch algorithm               %d\n", params->linesearch);
  _L("min_step     Linesearch - minimum step          %e\n", params->min_step);
  _L("max_step     Linesearch - maximum step          %e\n", params->max_step);
  _L("ftol         Linesearch - ftol accuracy param   %f\n", params->ftol);
  _L("gtol         Linesearch - gtol accuracy param   %f\n", params->gtol);
  _L("wolfe        Linesearch - Coeff for Wolfe cond  %f\n", params->wolfe);
  _L("\n");
  _L("xtol         Floating point machine precision   %e\n", params->xtol);
  _L("\n");
  _L("orthantwise_c      L1 norm - coeff              %f\n", params->orthantwise_c);
  _L("orthantwise_start  L1 norm - start index        %d\n", params->orthantwise_start);
  _L("orthantwise_end    L1 norm - end index          %d\n", params->orthantwise_end);
  _L("\n");
}

#define ALLOC_ARRAY(var,n) do { \
    var = ts_autopool_alloc(pool, n * sizeof((var)[0])); \
    ERROR_ON_NULL_ALLOC(var); \
  } while(0)
static TSResult sd_step_data_init(
    struct ts_autopool *pool,
    struct ts_minimization_state *data,
    ts_sim* sim,
    const struct ts_minimize_params *params)
{
  size_t ndof;

  memset(data,0,sizeof(*data));
  data->sim = sim;

  RETURN_ON_ERROR( minimizer_init_dof(params, sim, &data->dof_map) );
  AUTOPOOL_ON_RELEASE(pool, ts_dof_map_finalize, &data->dof_map);

#define CLEANUP_ON_ERROR {}

  ndof = data->dof_map.num_dof;
  ALLOC_ARRAY(data->origPos, ndof);
  ALLOC_ARRAY(data->newPos, ndof);
  ALLOC_ARRAY(data->bestPos, ndof);
  ALLOC_ARRAY(data->currGrad, ndof);

  data->linstep = params->step0;                     // linear motion step
  data->rotstep = params->step0*STEEPEST_ROT_FACTOR; // rotation motion step

#undef CLEANUP_ON_ERROR

  return TS_OK;
}
#undef ALLOC_ARRAY

static MAYBE_UNUSED TSResult steepest_particle_take_step(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct ts_minimization_state *stepdata = arg;
  VecReal*restrict xv0 = stepdata->origPos;
  VecReal*restrict xv1 = stepdata->newPos;
  VecReal*restrict gv  = stepdata->currGrad;
  register double step = stepdata->linstep;

  for (size_t i=i0; i < i1; i++) {
    size_t ind = i*PARTICLE_NUM_DOF;
    xv1[ind+0] = xv0[ind+0] - step*gv[ind+0];
    xv1[ind+1] = xv0[ind+1] - step*gv[ind+1];
    xv1[ind+2] = xv0[ind+2] - step*gv[ind+2];
  }

  return TS_OK;
}

static MAYBE_UNUSED TSResult steepest_rigid_take_step(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  struct ts_minimization_state *stepdata = arg;
  size_t poff = stepdata->dof_map.num_free_particles * PARTICLE_NUM_DOF;
  VecReal*restrict xv0 = stepdata->origPos + poff;
  VecReal*restrict xv1 = stepdata->newPos + poff;
  VecReal*restrict gv  = stepdata->currGrad + poff;

  register double linstep = stepdata->linstep;
  register double rotstep = stepdata->rotstep;

  for (size_t i=i0; i < i1; i++) {
    size_t ind = i*RIGID_BODY_NUM_DOF;
    xv1[ind+0] = xv0[ind+0] - linstep*gv[ind+0];
    xv1[ind+1] = xv0[ind+1] - linstep*gv[ind+1];
    xv1[ind+2] = xv0[ind+2] - linstep*gv[ind+2];

    xv1[ind+3] = xv0[ind+3] - rotstep*gv[ind+3];
    xv1[ind+4] = xv0[ind+4] - rotstep*gv[ind+4];
    xv1[ind+5] = xv0[ind+5] - rotstep*gv[ind+5];
  }

  return TS_OK;
}

static MAYBE_UNUSED TSResult steepest_rot_rigid_take_step(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);
  struct ts_minimization_state *stepdata = arg;
  size_t num_fp = stepdata->dof_map.num_free_particles;
  size_t num_fr = stepdata->dof_map.num_free_rigids;
  size_t poff = num_fp * PARTICLE_NUM_DOF + num_fr * RIGID_BODY_NUM_DOF;

  VecReal*restrict xv0 = stepdata->origPos + poff;
  VecReal*restrict xv1 = stepdata->newPos + poff;
  VecReal*restrict gv  = stepdata->currGrad + poff;

  register double rotstep = stepdata->rotstep;

  for (size_t i=i0; i < i1; i++) {
    size_t ind = i*RIGID_BODY_NUM_ROT_DOF;
    xv1[ind+0] = xv0[ind+0] - rotstep*gv[ind+0];
    xv1[ind+1] = xv0[ind+1] - rotstep*gv[ind+1];
    xv1[ind+2] = xv0[ind+2] - rotstep*gv[ind+2];
  }

  return TS_OK;
}

static TSResult steepest_update_dofs(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  struct ts_minimization_state *data;
  VecReal *restrict xv0, *restrict xv1, *restrict gv;
  register double linstep,rotstep;
  size_t i, n, ind0, ind1;
  size_t num_fp, num_fr, num_rr;

  UNUSED_PARAM(i0);
  UNUSED_PARAM(i1);

  data = arg;
  xv0 = data->origPos;
  xv1 = data->newPos;
  gv  = data->currGrad;

  linstep = data->linstep;
  rotstep = data->rotstep;

  num_fp = data->dof_map.num_free_particles;
  num_fr = data->dof_map.num_free_rigids;
  num_rr = data->dof_map.num_rot_rigids;

  if (num_fp > worker)
  {
    n = num_fp / nworkers;
    if ((num_fp % nworkers) > 0)
    {
      n++;
    }

    ind0 = n*worker;
    ind1 = ind0+n;
    if (ind1 > num_fp)
    {
      ind1 = num_fp;
    }

    for (i=ind0; i < ind1; i++) {
      size_t ind = i*PARTICLE_NUM_DOF;
      xv1[ind+0] = xv0[ind+0] - linstep*gv[ind+0];
      xv1[ind+1] = xv0[ind+1] - linstep*gv[ind+1];
      xv1[ind+2] = xv0[ind+2] - linstep*gv[ind+2];
    }
  }

  xv0 += PARTICLE_NUM_DOF*num_fp;
  xv1 += PARTICLE_NUM_DOF*num_fp;
  gv  += PARTICLE_NUM_DOF*num_fp;

  if (num_fr > worker)
  {
    n = num_fr / nworkers;
    if ((num_fr % nworkers) > 0)
    {
      n++;
    }

    ind0 = n*worker;
    ind1 = ind0+n;
    if (ind1 > num_fr)
    {
      ind1 = num_fr;
    }

    for (i=ind0; i < ind1; i++) {
      size_t ind = i*RIGID_BODY_NUM_DOF;
      xv1[ind+0] = xv0[ind+0] - linstep*gv[ind+0];
      xv1[ind+1] = xv0[ind+1] - linstep*gv[ind+1];
      xv1[ind+2] = xv0[ind+2] - linstep*gv[ind+2];

      xv1[ind+3] = xv0[ind+3] - rotstep*gv[ind+3];
      xv1[ind+4] = xv0[ind+4] - rotstep*gv[ind+4];
      xv1[ind+5] = xv0[ind+5] - rotstep*gv[ind+5];
    }
  }

  xv0 += RIGID_BODY_NUM_DOF*num_fr;
  xv1 += RIGID_BODY_NUM_DOF*num_fr;
  gv  += RIGID_BODY_NUM_DOF*num_fr;

  if (num_rr > worker)
  {
    n = num_rr / nworkers;
    if ((num_rr % nworkers) > 0)
    {
      n++;
    }

    ind0 = n*worker;
    ind1 = ind0+n;
    if (ind1 > num_rr)
    {
      ind1 = num_rr;
    }

    for (i=ind0; i < ind1; i++) {
      size_t ind = i*RIGID_BODY_NUM_ROT_DOF;
      xv1[ind+0] = xv0[ind+0] - rotstep*gv[ind+0];
      xv1[ind+1] = xv0[ind+1] - rotstep*gv[ind+1];
      xv1[ind+2] = xv0[ind+2] - rotstep*gv[ind+2];
    }
  }

  return TS_OK;
}

static TSResult steepest_take_step(struct ts_minimization_state *data, double maxdisp[2])
{
  int nthreads = ts_dispatch_foreach_num_threads();

  RETURN_ON_ERROR( ts_dispatch_foreach(nthreads, steepest_update_dofs, data) );

  ts_dof_from_vec(data->sim,
      &data->dof_map, 
      data->newPos,
      maxdisp,
      TS_ROT_AXIS_ANGLE);

  return TS_OK;
}

#define NSUCC_STEP_INCR 0  /* always increase step if energy decreases */
#define STEP_INCR 1.2
#define STEP_DECR 0.5
#define GRAD_NORM_TOL 1e-6

/*
static VecReal normv(size_t n, const VecReal*restrict v)
{
  VecReal acc = 0.0;
  for(size_t i=0; i < n; i++) { acc += v[i]*v[i]; }
  return sqrt(acc);
}
*/

static inline void swap_xv(VecReal** xvp, VecReal** yvp)
{
  VecReal* tmp = *xvp;
  *xvp = *yvp;
  *yvp = tmp;
}

static inline void copy_dof(const ts_dof_map* dof_map, const VecReal*restrict v0, VecReal*restrict v1)
{
  memcpy(v1, v0, dof_map->num_dof * sizeof(*v0));
}

enum TS_MINIMIZE_CONVERGED {
  TS_MINIMIZE_NOT_CONVERGED      = 0,

  TS_MINIMIZE_MAXED_ITERATIONS   = 1, 
  TS_MINIMIZE_GRADIENT_CONDITION = 2, // |grad| < eps * |pos|
  TS_MINIMIZE_ENERGY_CONDITION   = 3, // |en(new) - en(old)| < delta
  TS_MINIMIZE_STEP_TOO_SMALL     = 4, // avg(ds) < dstol
};

static const char* converged_as_text(enum TS_MINIMIZE_CONVERGED converged)
{
  switch (converged) {
    case TS_MINIMIZE_MAXED_ITERATIONS:   return "max iterations";
    case TS_MINIMIZE_GRADIENT_CONDITION: return "gradient converged";
    case TS_MINIMIZE_ENERGY_CONDITION:   return "energy converged";
    case TS_MINIMIZE_STEP_TOO_SMALL:     return "step too small";
    default: return "unknown";
  }
}

struct minimizer_progress_data {
  ts_sim* sim;
  const char* name;  // minimizer name

  size_t iter;
  int total_succ;
  int accum_succ;

  int report_freq;

  double energy;
  double best_energy;

  double frac_succ;
  double acc_linstep;
  double acc_rotstep;

  double grad_norm;
  double pos_norm;
};

MAYBE_UNUSED
static inline void
print_minimizer_progress(struct minimizer_progress_data* data, int report_freq)
{
  UNUSED_PARAM(report_freq);

  int recent_succ = data->total_succ - data->accum_succ;
  double avg_linstep = data->acc_linstep / data->report_freq;
  double avg_rotstep = data->acc_rotstep / data->report_freq;
  double prct_succ = 100.0*((VecReal)recent_succ)/(data->iter);
  TS_LOG(data->sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "%s> iter %d  energy %7.4f best %7.4f succ=%5.2f%%\n"
      "%s> avg(lin)=%7.4e  avg(rot)=%7.4e\n"
      "%s> |grad| = %7.4e  |x| = %7.4f  |grad|/|x| = %7.4e\n"
      "\n",
      data->name, (int)data->iter, data->energy, data->best_energy, prct_succ,
      data->name, avg_linstep, avg_rotstep,
      data->name, data->grad_norm, data->pos_norm, data->grad_norm/data->pos_norm);

  data->accum_succ = data->total_succ;
  data->acc_linstep = data->acc_rotstep = 0.0;
}

static TSResult minimize_steepest(struct ts_minimization_state *data, const struct ts_minimize_params *params)
{
  ts_sim *sim;
  ts_dof_map *dof_map;
  double en_old, en_best, grad_norm, pos_norm;
  int niter, nsucc, accum_succ, total_succ;
  double avg_linstep, avg_rotstep, max_linstep, max_rotstep;
  double maxdisp[2];
  enum TS_MINIMIZE_CONVERGED converged;
  size_t ndof;

  sim = data->sim;
  dof_map = &data->dof_map;
  ndof = dof_map->num_dof;

  // calculate energy and forces of current position
  en_old = 0.0;
  RETURN_ON_ERROR( ts_calculate_forces(sim, &en_old, NO_TERM_GROUPS, TERMWISE_FALSE) );
  TS_LOG(sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "STEEPEST> Starting energy = %7.4f\n", en_old);
  en_best = en_old;

#if STEEPEST_DEBUG
  static FILE* fdof = NULL;
  static FILE* fener = NULL;

  if (fdof == NULL) {
    fdof = fopen("steepest_dof.log","w");
  }
  if (fener == NULL) {
    fener = fopen("steepest_ener.log","w");
  }
#endif /* STEEPEST_DEBUG */

  //   1. save position, forces
  ts_minimize_copy_positions_to_xvec(data, data->origPos);
  ts_minimize_copy_forces_to_gvec(data, data->currGrad);

  copy_dof(dof_map, data->origPos, data->bestPos);

  grad_norm = normalizev(ndof, data->currGrad, GRAD_NORM_TOL);
  pos_norm  = normv(ndof, data->origPos);
  if (pos_norm < 1.0) { pos_norm = 1.0; }

  nsucc = accum_succ = total_succ = 0;
  avg_linstep = avg_rotstep = 0.0;

  max_linstep = params->maxstep;
  max_rotstep = params->maxstep * STEEPEST_ROT_FACTOR;

  maxdisp[0] = maxdisp[1] = 0.0;

  converged = TS_MINIMIZE_NOT_CONVERGED;
  niter = 0;
  // LOOP:
  while(converged == TS_MINIMIZE_NOT_CONVERGED)
  {
    double en_new, edelta;

    niter++;

#if STEEPEST_DEBUG
    for(size_t dof_ind = 0; dof_ind < ndof; dof_ind++) {
      fprintf(fdof,"%16.10f  %16.10f  %16.10f\n",
          data->bestPos[dof_ind],
          data->origPos[dof_ind],
          data->currGrad[dof_ind]);
    }
    fprintf(fdof,"\n");
    fprintf(fener, "%16.10f %16.10f %16.10f\n", data->linstep, data->rotstep, en_old);
#endif /* STEEPEST_DEBUG */

    // Take step against gradient to new position
    avg_linstep += data->linstep;
    avg_rotstep += data->rotstep;
    RETURN_ON_ERROR( steepest_take_step(data, maxdisp) );

    // Calculate energy, forces
    en_new = 0.0;
    RETURN_ON_ERROR( ts_calculate_forces(sim, &en_new, NO_TERM_GROUPS, TERMWISE_FALSE) );

    if (en_new < en_best) {
      copy_dof(dof_map, data->newPos, data->bestPos);
      en_best = en_new;
    }

    edelta = en_new - en_old; 

    //   4a. if energy decreases:
    if (edelta <= 0) {
      // increase step size
      nsucc++;
      total_succ++;

      data->linstep *= STEP_INCR;
      data->rotstep *= STEP_INCR;

      if (data->linstep > max_linstep) data->linstep = max_linstep;
      if (data->rotstep > max_rotstep) data->rotstep = max_rotstep;

      // save position, forces
      swap_xv(&data->origPos, &data->newPos);

      // update gradient, update norms, normalize gradient
      ts_minimize_copy_forces_to_gvec(data, data->currGrad);
      grad_norm = normalizev(ndof, data->currGrad, GRAD_NORM_TOL);
      pos_norm  = normv(ndof, data->origPos);

      en_old = en_new;
    } else {
      // energy increases
      nsucc = 0; // step not successful

      // decrease step size
      data->linstep *= STEP_DECR;
      data->rotstep *= STEP_DECR;
    }

    // periodically report progress
    if ((params->report_freq == 0) || (niter % params->report_freq == 0))
    {
      int recent_succ = total_succ - accum_succ;
      TS_LOG(sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
          "STEEPEST> iter %d  energy %7.4f best %7.4f succ=%5.2f%%\n"
          "STEEPEST> avg(lin)=%7.4e avg(rot)=%7.4e\n"
          "STEEPEST> |grad|/|x| = %7.4e  (eps = %7.4e)\n"
          "\n",
          niter,
          en_old, en_best,
          100.0*((VecReal)recent_succ)/niter,
          avg_linstep/params->report_freq,
          avg_rotstep/params->report_freq,
          grad_norm/pos_norm,params->eps);
      avg_linstep = avg_rotstep = 0.0;
      accum_succ = total_succ;
    }

    //   6. check remaining termination conditions:
    //      - iter >= max_iter
    //      - step < `step_min` (for linear and rotational steps)
    //

    // Check if energy change is less than machine precision...
    if (niter >= params->max_iter)                      { converged = TS_MINIMIZE_MAXED_ITERATIONS;   }
    else if ((edelta <= 0) && (-edelta < params->prec)) { converged = TS_MINIMIZE_ENERGY_CONDITION;   }
    else if (grad_norm < params->eps * pos_norm)        { converged = TS_MINIMIZE_GRADIENT_CONDITION; }
    else if (data->linstep < params->minstep)           { converged = TS_MINIMIZE_STEP_TOO_SMALL;     }
    else if (data->rotstep < params->minstep)           { converged = TS_MINIMIZE_STEP_TOO_SMALL;     }
  }

  ts_minimize_copy_positions_from_xvec(data, data->bestPos, maxdisp);

  const char* why = converged_as_text(converged);
  TS_LOG(sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "STEEPEST> Stopping (%s).  iter %d  energy %7.4f  succ=%5.2f%%\n"
      "STEEPEST> avg(lin)=%7.4e avg(rot)=%7.4e\n"
      "STEEPEST> |grad|/|x| = %7.4e  (eps = %7.4e)\n"
      "\n",
      why, niter, en_best,
      100.0*((VecReal)total_succ)/niter,
      avg_linstep/params->report_freq,
      avg_rotstep/params->report_freq,
      grad_norm/pos_norm,params->eps);

  return TS_OK;
}

#define PALLOC(var,n) do {    \
    var = ts_autopool_alloc(&pool, (n)*sizeof(var[0])); \
    ERROR_ON_NULL_ALLOC(var); \
  } while(0)

static TSResult abnr_steepest_step(
    struct ts_minimization_state *data,
    VecReal*restrict posv,
    VecReal*restrict gradv,
    VecReal*restrict newv,
    VecReal linstep,
    VecReal rotstep,
    double maxdisp[2])
{
  VecReal *origPos, *newPos, *currGrad, saved_linstep, saved_rotstep;
  TSResult err;

  origPos  = data->origPos;
  newPos   = data->newPos;
  currGrad = data->currGrad;

  saved_linstep = data->linstep;
  saved_rotstep = data->rotstep;

  data->origPos  = posv;
  data->currGrad = gradv;
  data->newPos   = newv;

  data->linstep = linstep;
  data->rotstep = rotstep;

  err = steepest_take_step(data, maxdisp);

  data->origPos  = origPos;
  data->newPos   = newPos;
  data->currGrad = currGrad;

  data->linstep = saved_linstep;
  data->rotstep = saved_rotstep;

  RETURN_ON_ERROR( err );

  return TS_OK;
}

MAYBE_UNUSED
static TSResult minimize_abnr(ts_sim *sim, struct ts_minimize_params *params)
{
  struct ts_minimization_state data;
  double max_linstep, max_rotstep;
  int nsucc, accum_succ, total_succ;
  double avg_linstep, avg_rotstep;
  enum TS_MINIMIZE_CONVERGED converged;
  struct ts_autopool pool;

  max_linstep = params->maxstep;
  max_rotstep = params->maxstep * STEEPEST_ROT_FACTOR;

  nsucc = accum_succ = total_succ = 0;
  avg_linstep = avg_rotstep = 0.0;

  data.sim = sim;

  converged = TS_MINIMIZE_NOT_CONVERGED;

#define CLEANUP_ON_ERROR { ts_autopool_release(&pool); }
  ts_autopool_init(&pool, NO_ALLOCATOR);

  RETURN_ON_ERROR3( ts_dof_map_init(&data.dof_map, sim) );
  AUTOPOOL_ON_RELEASE(&pool, ts_dof_map_finalize, &data.dof_map);

  int subdim = params->dim_subsp;
  size_t ndof = data.dof_map.num_dof;

  // Need to allocate space for:

#if VECTOR_SINGLE_PRECISION
#error  Code needs to be revised to work with single precision
#endif /* VECTOR_SINGLE_PRECISION */

  VecReal*restrict subhess;  PALLOC(subhess , subdim*subdim);   // reduced Hessian
  VecReal*restrict eigvals;  PALLOC(eigvals , subdim);          // eigenvalues of reduced Hessian
  VecReal*restrict gproj;    PALLOC(gproj   , subdim);          // gradient projected onto subspace
  VecReal*restrict cvals;    PALLOC(cvals   , subdim);          // c_lk coefficients

  VecReal*restrict posvecs;  PALLOC(posvecs , (subdim+1)*ndof); // subdim prev and current positions
  VecReal*restrict gradvecs; PALLOC(gradvecs, (subdim+1)*ndof); // subdim-prev and current gradients

  VecReal*restrict dposNR;   PALLOC(dposNR  , ndof);            // reduced dim Newton-Raphson step
  VecReal*restrict dgradNR;  PALLOC(dgradNR , ndof);            // estimated gradient after NR step

  VecReal*restrict dpos;     PALLOC(dpos,    subdim*ndof);      // diffs in positions
  VecReal*restrict dgrads;   PALLOC(dgrads,  subdim*ndof);      // diffs in gradients
  VecReal*restrict bestvec;  PALLOC(bestvec, ndof);             // best position

  int lda = subdim;
  double tmp = 0.0;
  int lwork = -1;
  int info = 0;
  dsyev( "V", "U", &subdim, NULL, NULL, NULL, &tmp, &lwork, &info);
  lwork = tmp;

  VecReal*restrict work;
  PALLOC(work, lwork);

#define POSVEC(ind) ( posvecs + (ndof*(ind)) )
#define GRADVEC(ind) ( gradvecs + (ndof*(ind)) )
#define DPOSVEC(ind) ( dpos + (ndof*(ind)) )
#define DGRADVEC(ind) ( dgrads + (ndof*(ind)) )
#define SUBHESS(i,j) (subhess[(i) + subdim*(j)])

  int curr = 0;

  double linstep = data.linstep;
  double rotstep = data.rotstep;

  // 
  // 0. Take M_s successful steps of steepest descent (SD)
  //
  //    r(k+1) = r(k) + dr_sd(k)
  //
  //    dr_sd(k) = F(k)/|F(k)| ds(k)
  //

  double en0 = 0.0;
  double en1 = 0.0;
  double enBest = 0.0;
  RETURN_ON_ERROR3( ts_calculate_forces(data.sim, &en0, NO_TERM_GROUPS, TERMWISE_FALSE) );
  TS_LOG(data.sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "ABNR> Starting energy = %7.4f\n", en0);

  ts_minimize_copy_positions_to_xvec(&data, POSVEC(0));
  ts_minimize_copy_forces_to_gvec(&data, GRADVEC(0));

  enBest = en0;
  copy_dof(&data.dof_map, POSVEC(0), bestvec);

  VecReal gnorm = normv(ndof, GRADVEC(0));
  VecReal pnorm = normv(ndof, POSVEC(0));

  double maxdisp[2] = { 0.0, 0.0 };

  for(int i=0; i < subdim; i++) {
    // 1. take a steepest descent step
    RETURN_ON_ERROR( abnr_steepest_step(&data,
          POSVEC(i),
          GRADVEC(i),
          POSVEC(i+1),
          data.linstep,
          data.rotstep,
          maxdisp) );
    RETURN_ON_ERROR3( ts_calculate_forces(sim, &en1, NO_TERM_GROUPS, TERMWISE_FALSE) );
    ts_minimize_copy_forces_to_gvec(&data, GRADVEC(i+1));

    if (en1 < en0) {
      if (en1 < enBest) {
        copy_dof(&data.dof_map, POSVEC(i+1), bestvec);
        enBest = en1;
      }

      linstep *= STEP_INCR;
      rotstep *= STEP_INCR;

      if (linstep > max_linstep) { linstep = max_linstep; }
      if (rotstep > max_rotstep) { rotstep = max_rotstep; }
    } else if (en1 > en0) {
      linstep *= STEP_DECR;
      rotstep *= STEP_DECR;
    }

    en0 = en1;
  }

  TS_LOG(data.sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "ABNR> Starting energy = %7.4f\n", en0);

  curr = subdim;

  int iter = 0;
  while(converged == TS_MINIMIZE_NOT_CONVERGED) {
    /*
     * construct M_s-dimensional approximation to the Hessian matrix:
     * 
     *    Let |l k> = r(k-l) - r(k)
     *
     *    A_sm = <l' k| A_full |l k>  ~=  <l' k|[ G(k-l) - G(k) ]
     *        where  A_full |l k> ~= [G(k-l) - G(k)]
     *    then, we want to solve Newton-Raphson in the M_s approximation:
     *
     *                  M_s
     *                  ---
     *                  \
     *    <l' k|F(k) =  /   <l' k| [ F(k) - F(k-l) ] c_lk
     *                  ---
     *                  l=1
     *
     *    Instead of solving for c_lk directly, first solve the
     *    eigensystem for A_sm.  If any of the eigenvalues are negative,
     *    then only use steepest descent for this step.
     *
     *    (a) Construct A_sm
     */

    for(int i=0; i < subdim; i++) {
      int ii = (curr+i+1) % subdim;
      subv(ndof, DPOSVEC(i) , POSVEC(ii) , POSVEC(curr));
      subv(ndof, DGRADVEC(i), GRADVEC(ii), GRADVEC(curr));
    }

    // project current gradient onto subspace
    for(int i=0; i < subdim; i++) {
      gproj[i] = dotv(ndof, DPOSVEC(i), GRADVEC(curr));
    }

    for (int i=0; i < subdim; i++) {
      SUBHESS(i,i) = dotv(ndof, DPOSVEC(i), DGRADVEC(i));

      for (int j=i+1; j < subdim; j++) {
        SUBHESS(i,j) = dotv(ndof, DPOSVEC(j), DGRADVEC(i));
      }
    }

    //
    //    (b) Solve eigensystem for A_sm:  A_sm |v_l> = a_l |v_l>
    //
    dsyev( "V", "U", &subdim, subhess, &lda, eigvals, work, &lwork, &info);
    if (info != 0) {
      TS_RAISE_ERROR2(TS_ERR_IN_ALGORITHM,
          ("Error solving eigensystem in Hessian subspace (info=%d)", info));
    }

    /*
     *    (c) construct dNR(k):
     *
     *    let |F'(k)> be F(k) projected onto M_s subspace
     *
     *    if any a_l not positive (or close to negative), then omit it
     *    from the following
     *
     *              M_s      
     *              ---    <v_l | F'(k)>
     *              \     --------------- |v_l>
     *    c_lk   =  /           a_l
     *              ---      
     *              l=1
     *              a_l>0
     * 
     *   In full space:
     *
     *              M_s      
     *              ---      
     *              \
     *    dNR(k) =  /   c_lk  |l k>
     *              ---      
     *              l=1      
     *
     *   NB: if all eigenvalues (a_l) are non-positive, then skip dNR(k)
     *       step and only use steepest descent step
     */

    // memset(dposNR,0,ndof*sizeof(dposNR[0]));
    memcpy(dposNR ,POSVEC(curr) ,ndof*sizeof(dposNR[0]));
    memcpy(dgradNR,GRADVEC(curr),ndof*sizeof(dgradNR[0]));

    if (cvals[subdim-1] >= params->mineig) {
      memset(cvals,0,subdim*sizeof(cvals[0]));
      for (int i=subdim-1; (i >= 0) && (eigvals[i] >= params->mineig); i--) {
        // build C vector by determining coefficients of eigenvector
        // components and summing in |l k> basis

        // project gproj onto eigenvec(i) and divide by eigenvalue

        double eigproj = 0.0;
        for (int j=0; j < subdim; j++) {
          eigproj += gproj[j] * SUBHESS(i,j);
        }

        double coeff = eigproj / eigvals[i];

        for (int j=0; j < subdim; j++) {
          cvals[j] += coeff * SUBHESS(i,j);
        }
      }

      // Build dposNR vector from coefficients
      for(size_t i=0; i < ndof; i++) {
        for (int j=0; j < subdim; j++) {
          dposNR[i] += cvals[j] * DPOSVEC(j)[i];
        }
      }
    }

    /*
     * 2. Construct residual and use steepest descent...
     *
     *                               M_s      
     *                               ---   
     *                               \
     *   F(r(k) + dNR(k)) ~= F(k) +  /    [ F(k-l) - F(k) ] c_lk
     *                               ---      
     *                               l=1
     *                               a_l>0
     *
     *   Use steepest descent on F(k+1/2) = F( r(k) + dNR(k) ):
     */

    // Estimate gradient vector at new position
    // memset(dgradNR,0,ndof*sizeof(dgradNR[0]));
    for(size_t i=0; i < ndof; i++) {
      for (int j=0; j < subdim; j++) {
        dgradNR[i] += cvals[j] * DGRADVEC(j)[i];
      }
    }

    double inv_gnorm = 1.0/normv(ndof,dgradNR);
    int next = (curr+1) % subdim;

    //
    //              F(k+1/2)
    //   dSD(k) = ------------  ds
    //             |F(k+1/2)|
    //
    // 3. Adjust step size:
    //
    //    if only SD was used, step is adjusted based on whether the
    //    energy decreases or increases
    //

    avg_linstep += linstep;
    avg_rotstep += rotstep;

    double maxdisp[2] = { 0.0, 0.0 };
    RETURN_ON_ERROR(
        abnr_steepest_step(&data,
          dposNR,
          dgradNR,
          POSVEC(next),
          linstep*inv_gnorm,
          rotstep*inv_gnorm,
          maxdisp)
        );

    // Calculate energy, forces
    en1 = 0.0;
    RETURN_ON_ERROR( ts_calculate_forces(sim, &en1, NO_TERM_GROUPS, TERMWISE_FALSE) );

    if (en1 < enBest) {
      copy_dof(&data.dof_map, POSVEC(next), bestvec);
      enBest = en1;
    }

    if (en1 < en0) { // energy decreases
      nsucc++;
      total_succ++;

      linstep *= STEP_INCR;
      rotstep *= STEP_INCR;

      // increase step size
      if (linstep > max_linstep) linstep = max_linstep;
      if (rotstep > max_rotstep) rotstep = max_rotstep;
    } else { // energy increases
      nsucc = 0; // step not successful

      // decrease step size
      linstep *= STEP_DECR;
      rotstep *= STEP_DECR;
    }

    double edelta = fabs(en1-en0);
    en0 = en1;

    gnorm = normv(ndof,GRADVEC(next));
    pnorm = normv(ndof,POSVEC(next));
    if (pnorm < 1.0) { pnorm = 1.0; }

    curr = next;

    // periodically report progress
    if ((params->report_freq == 0) || (iter % params->report_freq == 0))
    {
      int recent_succ = total_succ - accum_succ;
      TS_LOG(data.sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
          "ABNR> iter %d  energy %7.4f best %7.4f succ=%5.2f%%\n"
          "ABNR> avg(lin)=%7.4e avg(rot)=%7.4e\n"
          "ABNR> |grad|/max(1,|x|) = %7.4e\n"
          "\n",
          iter,
          en0, enBest,
          100.0*((VecReal)recent_succ)/iter,
          avg_linstep/params->report_freq,
          avg_rotstep/params->report_freq,
          gnorm/pnorm);
      avg_linstep = avg_rotstep = 0.0;
      accum_succ = total_succ;
    }

    // Check convergence criteria
    if (++iter > params->max_iter) { converged = TS_MINIMIZE_MAXED_ITERATIONS;   }
    if (edelta < params->prec)     { converged = TS_MINIMIZE_ENERGY_CONDITION;   }
    if (gnorm < pnorm*params->eps) { converged = TS_MINIMIZE_GRADIENT_CONDITION; }
    if (linstep < params->minstep) { converged = TS_MINIMIZE_STEP_TOO_SMALL;     }
    if (rotstep < params->minstep) { converged = TS_MINIMIZE_STEP_TOO_SMALL;     }
  }

  ts_minimize_copy_positions_from_xvec(&data, bestvec, maxdisp);

  const char* why = converged_as_text(converged);
  TS_LOG(data.sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "ABNR> Stopping (%s).  energy %7.4f  succ=%5.2f%%\n"
      "ABNR> avg(lin)=%7.4e avg(rot)=%7.4e\n"
      "ABNR> |grad|/|x| = %7.4e\n"
      "\n",
      why, enBest,
      100.0*((VecReal)total_succ)/iter,
      avg_linstep/params->report_freq,
      avg_rotstep/params->report_freq,
      gnorm/pnorm);

#undef POSVEC
#undef GRADVEC
#undef DGRADVEC

#undef CLEANUP_ON_ERROR
  ts_autopool_release(&pool);

  return TS_OK;
}

TSResult ts_minimize_steepest(ts_sim* sim, const struct ts_minimize_params* params)
{
  struct ts_autopool pool;
  ts_autopool_init(&pool, NO_ALLOCATOR);
#define CLEANUP_ON_ERROR { ts_autopool_release(&pool); }

  // Force full generation of nonbond lists
  double prev_padding = sim->nonbond_padding;
  int prev_nonbonds_enabled = sim->nonbonds_enabled;

  sim->nonbond_padding = 0.0;
  sim->nonbonds_enabled = 0;

  // Regenerate potential, excluding nonbonded terms
  ts_update_rigid_bodies(sim);
  RETURN_ON_ERROR3( ts_build_potential(sim) );

  struct ts_minimization_state data;
  memset(&data,0,sizeof(data));
  RETURN_ON_ERROR3( sd_step_data_init(&pool, &data, sim, params) );

  enable_user_interrupt();

  RETURN_ON_ERROR3( minimize_steepest(&data,params) );

  restore_user_interrupt();

  // RETURN_ON_ERROR( report_minimizer_result(sim, ret) );

  // Restore any previous padding
  sim->nonbond_padding = prev_padding;
  sim->nonbonds_enabled = prev_nonbonds_enabled;
  RETURN_ON_ERROR3( ts_build_potential(sim) );

  ts_autopool_release(&pool);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

struct minimization_method {
  const char* name;
  TSResult (*func)(ts_sim*, const struct ts_minimize_params*);
};

static const struct minimization_method methods[] = {
  { "default (lbfgs)", ts_minimize_lbfgs    }, // TS_MINIMIZE_UNSPECIFIED
  { "steepest"       , ts_minimize_steepest }, // TS_MINIMIZE_STEEPEST
  { "lbfgs"          , ts_minimize_lbfgs    }, // TS_MINIMIZE_STEEPEST
  { 0 , 0 }
};

TSResult ts_minimize_energy(ts_sim* sim, const struct ts_minimize_params* params)
{
  if (((int)params->method < TS_MINIMIZE_DEFAULT) ||
      ((int)params->method >= TS_MINIMIZE_MAXMETHOD))
  {
    TS_RAISE_ERROR2(TS_ERR_INVALID_ARGUMENT,
        ("unknown minimization method %d", (int)(params->method)));
  }

  // fprintf(stderr, "\nuse_nonbonds? %d\n", params->use_nonbonds);
  // fprintf(stderr, "\nnum_iterations = %d\n", params->max_iter);

  const struct minimization_method* method = methods + params->method;

  // FIXME: LOGGING/OUTPUT!
  TS_LOG(sim,TS_LOG_INFO,TS_LOG_MINIMIZER,
      "--[ minimizing with '%s' ]--\n", method->name);

#define CLEANUP_ON_ERROR { ts_dispatch_foreach_unprime(); }
  ts_dispatch_foreach_prime();
  RETURN_ON_ERROR3(method->func(sim,params));
  ts_dispatch_foreach_unprime();
#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_minimization_protocol_run(ts_sim* sim, const ts_minimization_protocol* protocol)
{
  TSResult ret = TS_OK;

  if (protocol == NULL) { TS_RAISE_ERROR3(TS_ERR_NULL_POINTER_ARG, "protocol is NULL"); }

  do {
    int maxproto = TS_MINIMIZATION_PROTOCOL_DEFAULT_SIZE;
    if (protocol->nminimizations < maxproto) { maxproto = protocol->nminimizations; }

    for(int i = 0; i < maxproto; i++) {
      ret = ts_minimize_energy(sim, protocol->params+i);
      if (IS_ERROR(ret) && (ERRCODE(ret) != TS_ERR_MINIMIZATION_INCOMPLETE)) {
        RETURN_ON_ERROR(ret);
      }
    }

    protocol = protocol->next;
  } while(protocol != NULL);

  return ret;
}

