#include "ts_normalmode.h"

#include "ts_units.h"
#include "ts_numerical.h"
#include "ts_sparse.h"
#include "ts_bonds.h"

#define DELTA 1e-6

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
};

void ts_sim_nma_init(ts_sim_nma* nma, ts_sim* sim)
{
  nma->sim = sim;

  ts_autopool_init(&nma->pool, NO_ALLOCATOR);

  nma->method = TS_NMA_METHOD_DEFAULT;

  for (int i=0; i < TS_BOND_TERM_MAX; i++)
  {
    nma->enabled[i] = 1;
  }

  nma->flags = TS_NMA_FL_NONE;

  nma->nfreqs = 0;
  nma->freqs = NULL;
  nma->modes = NULL;

  nma->freq_skip = 0;
  nma->freq_off  = 0;

  nma->enthalpic_energy = 0.0;
  nma->entropic_energy  = 0.0;
}

void ts_sim_nma_finalize(ts_sim_nma* nma)
{
  ts_autopool_release(&nma->pool);
}

static void MAYBE_UNUSED print_matrix(size_t m, size_t n, const VecReal* mat)
{
  fprintf(stderr, "\n");
  for (size_t i=0; i < m; i++) {
    fprintf(stderr, "%5.2f", mat[i]);
    for (size_t j=1; j < n; j++) {
      fprintf(stderr, " %5.2f", mat[m*j+i]);
    }
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n");
}

static void MAYBE_UNUSED print_mat3x3(const mat3x3* mat)
{
  fprintf(stderr, "\n");
  for (int i=0; i < 3; i++)
  {
    fprintf(stderr, "%5.2f %5.2f %5.2f\n",
        MAT3X3_ELT(mat, i, 0),
        MAT3X3_ELT(mat, i, 1),
        MAT3X3_ELT(mat, i, 2));
  }
  fprintf(stderr, "\n");
}

static void MAYBE_UNUSED print_hessian(size_t n, const VecReal* hess)
{
  fprintf(stderr, "\n");
  for (size_t i=0; i < n; i++) {
    fprintf(stderr, "%9.3f", hess[ hessian_index(n,i,0) ]);
    for (size_t j=1; j < n; j++) {
      fprintf(stderr, " %9.3f", hess[ hessian_index(n,i,j) ]);
    }
    fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n");
}

/*! \internal Argument passed to `energy_at_coord` function that serves
 * as closure for the function.
 */
struct energy_at_coord_closure {
  size_t dof;
  struct ts_dof_map* dof_map;
  ts_sim* sim;
  VecReal* xv;
  enum TS_ROT_REPR rot_repr;
};

/*! \internal Evaluates the energy of the system when a single degree of
 * freedom is set to `x`.  This allows the energy of the system to be
 * treated as a function of a degree of freedom, holding all other
 * degrees of freedom constant.
 */
static double energy_at_coord(struct energy_at_coord_closure* closure, double x)
{
  double en = 0.0;

  VecReal x0 = closure->xv[ closure->dof ];

  closure->xv[ closure->dof ] = x;

  ts_sim_set_dof(closure->sim, closure->dof_map, closure->dof, closure->xv, closure->rot_repr);

  ts_calculate_energy(closure->sim, &en, NO_TERM_GROUPS, TERMWISE_FALSE);
  closure->xv[ closure->dof ] = x0;

  return en;
}

/*! \brief Numerically calculates \f$ \partial^2 V / \partial
 * \eta_i^2\f$, the partial second derivative of the potential with
 * respect to a single degree of freedom.  
 *
 * Derivatives are calculated with Ridders method for greater accuracy.
 * Use `#calculate_fij` to calculate the second derivative with respect
 * to two degrees of freedom.
 */
#define H0 1e-2
static inline TSResult calculate_fii(ts_sim* sim, struct ts_dof_map* dof_map, VecReal* xv,
    size_t i, VecReal* hess_entry, VecReal* hess_err, enum TS_ROT_REPR rot_repr)
{
  struct energy_at_coord_closure closure = {
    .sim = sim,
    .dof = i,
    .dof_map = dof_map,
    .xv  = xv,
    .rot_repr = rot_repr
  };

  VecReal xv_i = xv[i];

#define CLEANUP_ON_ERROR { xv[i] = xv_i; }

  *hess_entry = 0.0;
  *hess_err = 0.0;
  RETURN_ON_ERROR3( deriv2_ridders(
        (single_var_func)energy_at_coord, &closure, xv_i, H0, hess_entry, hess_err) );

#undef CLEANUP_ON_ERROR

  xv[i] = xv_i;

  ts_sim_set_dof(sim, dof_map, i, xv, rot_repr);

  return TS_OK;
}

/*! \internal Argument passed to `energy_at_two_coord` function that
 * serves as closure for the function.
 */
struct energy_at_two_coord_closure {
  size_t dof1;
  size_t dof2;

  struct ts_dof_map* dof_map;

  ts_sim* sim;
  VecReal* xv;
  enum TS_ROT_REPR rot_repr1;
  enum TS_ROT_REPR rot_repr2;
};

/*! \internal Evaluates the energy of the system when a two degrees of
 * freedom are set to `x` and `y`, respectively.  This allows the energy
 * of the system to be treated as a function of two degrees of freedom,
 * holding all other degrees of freedom constant.
 */
static double energy_at_two_coord(struct energy_at_two_coord_closure* closure, double x, double y)
{
  double en = 0.0;

  VecReal x0 = closure->xv[ closure->dof1 ];
  VecReal y0 = closure->xv[ closure->dof2 ];

  closure->xv[ closure->dof1 ] = x;
  closure->xv[ closure->dof2 ] = y;

  ts_sim_set_dof(closure->sim, closure->dof_map, closure->dof1, closure->xv, closure->rot_repr1);
  ts_sim_set_dof(closure->sim, closure->dof_map, closure->dof2, closure->xv, closure->rot_repr2);

  ts_calculate_energy(closure->sim, &en, NO_TERM_GROUPS, TERMWISE_FALSE);

  closure->xv[ closure->dof1 ] = x0;
  closure->xv[ closure->dof2 ] = y0;

  return en;
}

/*! \brief Numerically calculates \f$ \partial^2 V / \partial \eta_i
 * \partial\eta_j \f$, the partial second derivative of the potential with
 * respect to two degrees of freedom.  
 *
 * Derivatives are calculated with Ridders method for greater accuracy.
 * This method requires the degrees of freedom to be different.  Use
 * `#calculate_fii` to calculate the second derivative with respect to a
 * single degree of freedom.
 */
static inline TSResult calculate_fij(ts_sim* sim, struct ts_dof_map* dof_map, VecReal* xv,
    size_t i, size_t j, VecReal* hess_entry, VecReal* hess_err,
    enum TS_ROT_REPR rot_repr1,
    enum TS_ROT_REPR rot_repr2)
{
  struct energy_at_two_coord_closure closure = {
    .sim = sim,
    .dof1 = i,
    .dof2 = j,
    .dof_map = dof_map,
    .xv  = xv,
    .rot_repr1 = rot_repr1,
    .rot_repr2 = rot_repr2,
  };

  VecReal xv_i = xv[i];
  VecReal xv_j = xv[j];

#define CLEANUP_ON_ERROR { xv[i] = xv_i;  xv[j] = xv_j; }

  *hess_entry = 0.0;
  *hess_err   = 0.0;
  RETURN_ON_ERROR3( deriv_xy_ridders(
        (double_var_func)energy_at_two_coord, &closure, xv_i, xv_j, H0, hess_entry, hess_err) );

#undef CLEANUP_ON_ERROR

  xv[i] = xv_i;
  xv[j] = xv_j;

  ts_sim_set_dof(sim, dof_map, i, xv, rot_repr1);
  ts_sim_set_dof(sim, dof_map, j, xv, rot_repr2);

  return TS_OK;
}
#undef H0

/*! \internal Driver loop that calculates the Hessian for all degrees of
 * freedom
 */
static MAYBE_UNUSED TSResult calculate_hessian_all_dof(
    ts_sim* sim,
    struct ts_dof_map* dof_map,
    VecReal* xv, VecReal* hess,
    enum TS_ROT_REPR rot_repr)
{
  for(size_t i=0; i < dof_map->num_dof; i++) {
    double err = 0.0;
    RETURN_ON_ERROR(
        calculate_fii(sim, dof_map, 
          xv, i, hess + hessian_index(dof_map->num_dof,i,i), &err, rot_repr) );

    for(size_t j=i+1; j < dof_map->num_dof; j++) {
      RETURN_ON_ERROR( calculate_fij(sim, dof_map,
            xv, i, j, hess + hessian_index(dof_map->num_dof,i,j), &err, rot_repr, rot_repr) );
    }
  }

  return TS_OK;
}

static TSResult numerical_hessian_by_linked_objects(
    ts_sim* sim, 
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    struct ts_dof_map* dof_map,
    const char*restrict links,
    enum TS_ROT_REPR rot_repr);

/*! \brief Calculates the Hessian matrix for the system in the current
 * configuration by numerically calculating all of its elements.
 *
 * Deriviatives are calculated using Ridders method, to obtain greater
 * numerical accuracy.  
 *
 * NB: This calculation is very slow compared to analytical methods, and
 * is primarily for testing.
 */
TSResult ts_sim_numerical_hessian_check_all_dof(ts_sim* sim, 
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    size_t* num_dof_ptr, 
    enum TS_ROT_REPR rot_repr)
{
  struct ts_dof_map dof_map;
  ts_dof_map_init(&dof_map, sim);

  *hess_ptr = NULL;
  *hess_err_ptr = NULL;
  RETURN_ON_ERROR2(
      numerical_hessian_by_linked_objects(
        sim, hess_ptr, hess_err_ptr, &dof_map, NULL, rot_repr),
      ts_dof_map_finalize(&dof_map));

  *num_dof_ptr = dof_map.num_dof;
  ts_dof_map_finalize(&dof_map);

  return TS_OK;
}

/*! \brief Determines which components actually interact via the system
 * potential.
 *
 * If two degrees of freedom have no potential term in common, then that
 * element of the Hessian will be zero.  This function is used to find
 * the elements of the Hessian that may not be zero because two degrees
 * of freedom share the same potential.  This reduces the cost of the
 * numerical Hessian considerably compared to
 * #ts_sim_numerical_hessian_check_all_dof.
 */
static TSResult determine_pairwise_links(
    const ts_sim* sim, 
    struct ts_potential_term_list* term_list, 
    char* links,
    struct ts_dof_map* dof_map)
{
  if (term_list->terms.size == 0) return TS_OK;

  const ts_potential_term* terms = term_list->terms.data;
  const ts_potential_term* end = terms + term_list->terms.size;

  // fprintf(stderr, "\n...pairwise links...\n");
  do {
    for (int i=0; i < terms->potential->nbodies; i++) {
      size_t p_ind_i = terms->particle_indices[i];
      size_t ind_i = ts_dof_map_p2o(sim, dof_map, p_ind_i);

      for (int j=i+1; j < terms->potential->nbodies; j++) {
        size_t p_ind_j = terms->particle_indices[j];
        size_t ind_j = ts_dof_map_p2o(sim, dof_map, p_ind_j);

        /*
        fprintf(stderr, "[%lu] (%lu,%lu) (%lu,%lu) LINKED\n",
            terms - ((const ts_potential_term*)term_list->terms.data),
            p_ind_i, p_ind_j,
            ind_i, ind_j);
            */

        links[ hessian_index(dof_map->num_objects, ind_i, ind_j) ] = 1;
      }
    }
  } while(++terms < end);

  return TS_OK;
}

static TSResult get_linked_elements(ts_sim* sim, struct ts_dof_map* dof_map, char** links_ptr)
{
  size_t link_sz = dof_map->num_objects*(dof_map->num_objects+1)*sizeof(char)/2;
  char* links = ts_alloc(NO_ALLOCATOR, link_sz);

#define CLEANUP_ON_ERROR { ts_safe_free(NO_ALLOCATOR, links); }

  ERROR_ON_NULL_ALLOC( links );
  memset(links, 0, link_sz);

  RETURN_ON_ERROR3(
      determine_pairwise_links(sim, &sim->bondlist, links, dof_map) );
  *links_ptr = links;

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

/*! \brief Determines which objects (free particles or rigid bodies) are
 * "linked" by a term in the system potential.
 */
TSResult ts_sim_numerical_hessian_linked_elements(ts_sim* sim, char** links_ptr, size_t* num_objs_ptr)
{
  struct ts_dof_map dof_map;
  RETURN_ON_ERROR(
      ts_dof_map_init(&dof_map, sim) );

#define CLEANUP_ON_ERROR { ts_dof_map_finalize(&dof_map); }

  RETURN_ON_ERROR3(
      get_linked_elements(sim, &dof_map, links_ptr) );

  *num_objs_ptr = dof_map.num_objects;

#undef CLEANUP_ON_ERROR

  ts_dof_map_finalize(&dof_map);

  return TS_OK;
}

/*! \brief Returns the DoFs associated with the object index `oind`. */
static inline void get_dof_info(size_t oind, size_t num_free_particles, size_t* beg_dof, size_t* end_dof)
{
  if (oind < num_free_particles) {
    *beg_dof = PARTICLE_NUM_DOF * oind;
    *end_dof = *beg_dof + PARTICLE_NUM_DOF;
  } else {
    *beg_dof = PARTICLE_NUM_DOF * num_free_particles + 
      RIGID_BODY_NUM_DOF * (oind-num_free_particles);
    *end_dof = *beg_dof + RIGID_BODY_NUM_DOF;
  }
}

static void body_to_euler(mat3x3*restrict xform, VecReal*restrict euler, enum TS_ROT_REPR repr)
{
  if (repr == TS_ROT_EULER_ZXZ)
  {
    VecReal sin_b = sin(euler[1]);
    VecReal tan_b = tan(euler[1]);
    VecReal cos_c = cos(euler[2]);
    VecReal sin_c = sin(euler[2]);

    mat3x3_zero(xform);

    MAT3X3_ELT(xform, 0,0) =  sin_c / sin_b;
    MAT3X3_ELT(xform, 1,0) =  cos_c;
    MAT3X3_ELT(xform, 2,0) = -sin_c / tan_b;

    MAT3X3_ELT(xform, 0,1) =  cos_c / sin_b;
    MAT3X3_ELT(xform, 1,1) = -sin_c;
    MAT3X3_ELT(xform, 2,1) = -cos_c / tan_b;

    MAT3X3_ELT(xform, 0,2) = 0.0;
    MAT3X3_ELT(xform, 1,2) = 0.0;
    MAT3X3_ELT(xform, 2,2) = 1.0;
  }
  else if (repr == TS_ROT_EULER_XYZ)
  {
    VecReal cos_b = cos(euler[1]);
    VecReal tan_b = tan(euler[1]);
    VecReal cos_c = cos(euler[2]);
    VecReal sin_c = sin(euler[2]);

    mat3x3_zero(xform);

    MAT3X3_ELT(xform, 0,0) = cos_c / cos_b;
    MAT3X3_ELT(xform, 1,0) = sin_c;
    MAT3X3_ELT(xform, 2,0) = -cos_c * tan_b;

    MAT3X3_ELT(xform, 0,1) = -sin_c / cos_b;
    MAT3X3_ELT(xform, 1,1) = cos_c;
    MAT3X3_ELT(xform, 2,1) = sin_c * tan_b;

    MAT3X3_ELT(xform, 0,2) = 0.0;
    MAT3X3_ELT(xform, 1,2) = 0.0;
    MAT3X3_ELT(xform, 2,2) = 1.0;
  }
  else if (repr == TS_ROT_NONE)
  {
    mat3x3_identity(xform);
  }
  else
  {
    fprintf(stderr, "ERROR! Bad rot_repr (%d) for body_to_euler\n", (int)repr);
    mat3x3_identity(xform);
  }
}


static void convert_euler_derivs_to_body_axes(
    ts_sim* sim,
    VecReal*restrict hess,
    VecReal*restrict xv,
    struct ts_dof_map* dof_map,
    size_t dof_i0,
    size_t dof_j0,
    enum TS_ROT_REPR repr_i,
    enum TS_ROT_REPR repr_j)
{
  UNUSED_PARAM(sim);

  mat3x3 body_to_euler_i;
  mat3x3 body_to_euler_j;

  /* Generate the transformation matrices that convert the body axes
   * into the Euler axes.
   */
  body_to_euler( &body_to_euler_i, xv+dof_i0+RB_ROTATIONAL_OFFSET, repr_i );
  body_to_euler( &body_to_euler_j, xv+dof_j0+RB_ROTATIONAL_OFFSET, repr_j );

  mat3x3 old_dof;

  /* The next two macros handle copying between the Hessian matrix and
   * old_dof.  They rely on the HESS(i,j) macro being defined.  This is
   * done to simplify the code and eliminate some of the error-prone
   * repetition required to determine the correct offsets into the
   * Hessian matrix
   */
#define COPY_FROM_HESS(mat) do { \
    for (int i=0; i < 3; i++) {   \
      for (int j=0; j < 3; j++) { \
        MAT3X3_ELT((mat),i,j) = HESS(i,j); \
      } } } while(0) 
#define COPY_TO_HESS(mat) do { \
    for (int i=0; i < 3; i++) {   \
      for (int j=0; j < 3; j++) { \
        HESS(i,j) = MAT3X3_ELT((mat),i,j); \
      } } } while(0) 

  /* handle rot(i) X trans(j) */
#define HESS(i,j) \
  ( hess[ hessian_index(dof_map->num_dof, dof_i0+RB_ROTATIONAL_OFFSET+(i),dof_j0+(j)) ] )

  /* volatile */ int debug = 0; /* (dof_i0 == 0) && (dof_j0 == 0); */
  if (repr_i != TS_ROT_NONE)
  {
    COPY_FROM_HESS(&old_dof);

    mat3x3 tmp = MAT3X3_ZERO;
    mat3x3_lmul_trans(&tmp, &old_dof, &body_to_euler_i);

    if (debug) {
      printf("rot(i) x trans(j)\n");
      printf("orig hess:\n");
      print_mat3x3(&old_dof);
      printf("XF:\n");
      print_mat3x3(&body_to_euler_i);
      printf("new hess:\n");
      print_mat3x3(&tmp);
      printf("\n");
      printf("\n");
    }

    COPY_TO_HESS(&tmp);
  }

  /* handle trans(i) X rot(j) */
#undef HESS
#define HESS(i,j) \
  ( hess[ hessian_index(dof_map->num_dof, dof_i0+(i), dof_j0+RB_ROTATIONAL_OFFSET+(j)) ] )

  if ((dof_i0 != dof_j0) &&(repr_j != TS_ROT_NONE))
  {
    COPY_FROM_HESS(&old_dof);

    mat3x3 tmp = MAT3X3_ZERO;
    mat3x3_mul(&tmp, &old_dof, &body_to_euler_j);

    if (debug) {
      printf("trans(i) x rot(j)\n");
      printf("orig hess:\n");
      print_mat3x3(&old_dof);
      printf("XF:\n");
      print_mat3x3(&body_to_euler_i);
      printf("new hess:\n");
      print_mat3x3(&tmp);
      printf("\n");
      printf("\n");
    }

    COPY_TO_HESS(&tmp);
  }

  /* handle rot(i) X rot(j) */
#undef HESS
#define HESS(i,j) \
  ( hess[ hessian_index(dof_map->num_dof, \
                        dof_i0+RB_ROTATIONAL_OFFSET+(i), \
                        dof_j0+RB_ROTATIONAL_OFFSET+(j)) ] )

  if ((repr_i != TS_ROT_NONE) && (repr_j != TS_ROT_NONE))
  {
    COPY_FROM_HESS(&old_dof);

    mat3x3 tmp = MAT3X3_ZERO;
    mat3x3 tmp1 = MAT3X3_ZERO;
    mat3x3_mul(&tmp1, &old_dof, &body_to_euler_j);
    mat3x3_lmul_trans(&tmp, &tmp1, &body_to_euler_i);

    if (debug) {
      printf("rot(i) x rot(j)\n");
      printf("orig hess:\n");
      print_mat3x3(&old_dof);
      printf("XF:\n");
      print_mat3x3(&body_to_euler_i);
      printf("new hess:\n");
      print_mat3x3(&tmp);
      printf("\n");
      printf("\n");
    }

    COPY_TO_HESS(&tmp);
  }
#undef HESS
}
#undef COPY_FROM_HESS
#undef COPY_TO_HESS

int zxz_default_coords = 1;

static enum TS_ROT_REPR rigid_body_euler_coords(const ts_sim*sim, size_t rbind, VecReal euler[3])
{
  const quat4*restrict ori = RIGID_ORI(sim,rbind);
  if (zxz_default_coords)
  {
    quat4_to_euler_zxz_array(ori, euler);

    if ((euler[1] > ZXZ_THETA_CUTOFF)      && 
        (euler[0] < M_PI-ZXZ_THETA_CUTOFF) &&
        (euler[2] < M_PI-ZXZ_THETA_CUTOFF))
    {
      return TS_ROT_EULER_ZXZ;
    }
    else
    {
      quat4_to_euler_xyz_array(ori, euler);
      return TS_ROT_EULER_XYZ;
    }
  }
  else
  {
    quat4_to_euler_xyz_array(ori, euler);

    if (fabs(euler[1]) <= XYZ_BETA_CUTOFF)
    {
      return TS_ROT_EULER_XYZ;
    }
    else
    {
      quat4_to_euler_zxz_array(ori, euler);
      return TS_ROT_EULER_ZXZ;
    }
  }
}

void ts_sim_copy_vec_to_dof_in_body_axes(
    ts_sim* sim, 
    struct ts_dof_map* dof_map,
    const VecReal*restrict xv, 
    const enum TS_ROT_REPR* rot_conventions)
{
  size_t dof_off = 0;

  for (size_t i=0; i < dof_map->num_free_particles; i++)
  {
    vector3* pos = PARTICLE_POS(sim, dof_map->fp_to_p[i]);
    VECTOR3_X(pos) = xv[dof_off + 0];
    VECTOR3_Y(pos) = xv[dof_off + 1];
    VECTOR3_Z(pos) = xv[dof_off + 2];

    dof_off += PARTICLE_NUM_DOF;
  }

  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    ts_rigid_body_set_dof(
        sim, i, xv+dof_off, rot_conventions[i]);

    dof_off += RIGID_BODY_NUM_DOF;
  }
}

void ts_sim_copy_dof_to_vec_in_body_axes(
    ts_sim* sim, 
    struct ts_dof_map* dof_map,
    VecReal*restrict xv, 
    enum TS_ROT_REPR* rot_conventions)
{
  size_t dof_off = 0;
  for (size_t i=0; i < dof_map->num_free_particles; i++)
  {
    const vector3*restrict pos = PARTICLE_POS(sim,dof_map->fp_to_p[i]);
    xv[dof_off + 0] = VECTOR3_X( pos );
    xv[dof_off + 1] = VECTOR3_Y( pos );
    xv[dof_off + 2] = VECTOR3_Z( pos );

    dof_off += PARTICLE_NUM_DOF;
  }

  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    const vector3*restrict pos = ts_rigid_body_pos(sim, i);
    xv[dof_off + 0] = VECTOR3_X( pos );
    xv[dof_off + 1] = VECTOR3_Y( pos );
    xv[dof_off + 2] = VECTOR3_Z( pos );

    rot_conventions[i] = rigid_body_euler_coords(sim, i,
        xv + dof_off + PARTICLE_NUM_DOF);

    dof_off += RIGID_BODY_NUM_DOF;
  }
}

static MAYBE_UNUSED void print_dof(const ts_sim* sim, const enum TS_ROT_REPR* rot_conventions)
{
  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    VecReal sixdof[6];
    ts_rigid_body_get_dof(
        sim, i, sixdof, 
        (rot_conventions != NULL) 
          ? rot_conventions[i] 
          : TS_ROT_AXIS_ANGLE
      );
    TS_LOG(sim,TS_LOG_INFO,TS_LOG_NORMALMODE,
        "%7.4f %7.4f %7.4f %7.4f %7.4f %7.4f\n", 
        sixdof[0], sixdof[1], sixdof[2],
        sixdof[3], sixdof[4], sixdof[5]);
  }
}

static TSResult numerical_hessian_by_linked_objects(
    ts_sim* sim, 
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    struct ts_dof_map* dof_map,
    const char*restrict links,
    enum TS_ROT_REPR rot_repr)
{
  size_t num_dof = dof_map->num_dof;
  size_t hess_sz = hessian_size(num_dof);

  VecReal* xv = NULL;
  VecReal* hess = NULL;
  VecReal* hess_err = NULL;
  enum TS_ROT_REPR *rot_conventions = NULL;

#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR, xv);    \
    ts_safe_free(NO_ALLOCATOR, hess);  \
    ts_safe_free(NO_ALLOCATOR, hess_err);  \
    ts_safe_free(NO_ALLOCATOR, rot_conventions); }

  hess = ts_calloc(NO_ALLOCATOR, hess_sz, sizeof(hess[0]));
  ERROR_ON_NULL_ALLOC(hess);

  hess_err = ts_calloc(NO_ALLOCATOR, hess_sz, sizeof(hess_err[0]));
  ERROR_ON_NULL_ALLOC(hess_err);

  xv = ts_calloc(NO_ALLOCATOR, num_dof, sizeof(xv[0]));
  ERROR_ON_NULL_ALLOC(xv);

  rot_conventions = 
    ts_alloc(NO_ALLOCATOR, sim->num_rigid_bodies * sizeof(rot_conventions[0]));
  ERROR_ON_NULL_ALLOC(rot_conventions);

  memset(hess, 0, hess_sz*sizeof(hess[0]));
  memset(hess_err, 0, hess_sz*sizeof(hess_err[0]));

  /* Save orientations and determine the appropriate Euler convention
   */
  if (rot_repr != TS_ROT_BODY_AXES)
  {
    if (rot_repr == TS_ROT_BODY_RELATIVE)
    {
      for(size_t i=0; i < sim->num_rigid_bodies; i++)
      {
        quat4_set(RIGID_PREV_ORI(sim,i), RIGID_ORI(sim,i));
      }
    }

    ts_dof_to_vec(sim, dof_map, xv, rot_repr);
    for (size_t i = 0; i < sim->num_rigid_bodies; i++)
    {
      rot_conventions[i] = rot_repr;
    }
  }
  else
  {
    ts_sim_copy_dof_to_vec_in_body_axes(sim, dof_map, xv, rot_conventions);
  }

  for (size_t i=0; i < dof_map->num_objects; i++) {
    int oi_is_rigid = (i >= dof_map->num_free_particles);
    size_t rb_i = oi_is_rigid ? i-dof_map->num_free_particles : TS_NULL_RIGID_BODY;
    enum TS_ROT_REPR oi_rot_repr = oi_is_rigid ? rot_conventions[rb_i] : TS_ROT_NONE;

    size_t beg_dof_i = 0;
    size_t end_dof_i = 0;
    get_dof_info(i, dof_map->num_free_particles, &beg_dof_i, &end_dof_i);

    for (size_t dof_i1 = beg_dof_i; dof_i1 < end_dof_i; dof_i1++) {
      size_t off_ii = hessian_index(num_dof, dof_i1, dof_i1);
      RETURN_ON_ERROR3( calculate_fii(sim, dof_map, xv, dof_i1,
            hess + off_ii, hess_err + off_ii, oi_rot_repr) );

      for (size_t dof_i2 = dof_i1+1; dof_i2 < end_dof_i; dof_i2++) {
        size_t off_ij = hessian_index(num_dof, dof_i1, dof_i2);
        RETURN_ON_ERROR3( calculate_fij(sim, dof_map, xv, dof_i1, dof_i2,
              hess + off_ij, hess_err + off_ij, oi_rot_repr, oi_rot_repr) );
      }
    }

    if (oi_is_rigid && (rot_repr == TS_ROT_BODY_AXES))
    {
      convert_euler_derivs_to_body_axes( sim, 
          hess, 
          xv, 
          dof_map, 
          beg_dof_i, 
          beg_dof_i, 
          oi_rot_repr, 
          oi_rot_repr );
    }

    for (size_t j=i+1; j < dof_map->num_objects; j++) {
      if (!links || links[ hessian_index(dof_map->num_objects,i,j) ]) {
        int oj_is_rigid = (j >= dof_map->num_free_particles);
        size_t rb_j = oj_is_rigid ? j-dof_map->num_free_particles : TS_NULL_RIGID_BODY;
        enum TS_ROT_REPR oj_rot_repr = oj_is_rigid ? rot_conventions[rb_j] : TS_ROT_NONE;

        size_t beg_dof_j = 0;
        size_t end_dof_j = 0;
        get_dof_info(j, dof_map->num_free_particles, &beg_dof_j, &end_dof_j);

        for (size_t dof_i = beg_dof_i; dof_i < end_dof_i; dof_i++) {
          for (size_t dof_j = beg_dof_j; dof_j < end_dof_j; dof_j++) {
            size_t off_ij = hessian_index(num_dof, dof_i, dof_j);
            RETURN_ON_ERROR3( calculate_fij(sim, dof_map, xv, dof_i, dof_j,
                  hess + off_ij,
                  hess_err + off_ij,
                  oi_rot_repr, 
                  oj_rot_repr) );
          }
        }

        if ((oi_is_rigid || oj_is_rigid) && (rot_repr == TS_ROT_BODY_AXES))
        {
          convert_euler_derivs_to_body_axes( sim, 
              hess, 
              xv, 
              dof_map, 
              beg_dof_i, 
              beg_dof_j, 
              oi_rot_repr, 
              oj_rot_repr );
        }
      }
    }
  }

#undef CLEANUP_ON_ERROR

  *hess_ptr = hess;
  *hess_err_ptr = hess_err;

  double maxdisp[2] = { 0.0, 0.0 };
  if (rot_repr != TS_ROT_BODY_AXES)
  {
    ts_dof_from_vec(sim, dof_map, xv, maxdisp, rot_repr);
  }
  else
  {
    ts_sim_copy_vec_to_dof_in_body_axes(
        sim, dof_map, xv, rot_conventions);
  }

  ts_free(NO_ALLOCATOR, xv);
  ts_free(NO_ALLOCATOR, rot_conventions);

  return TS_OK;
}

/*! \brief Calculates the Hessian matrix for the system in the current
 * configuration by numerically calculating all elements that interact
 * via a term in the system potential.
 *
 * Deriviatives are calculated using Ridders method, to obtain greater
 * numerical accuracy.  
 *
 */
TSResult ts_sim_numerical_hessian_by_potential_terms(
    ts_sim* sim, 
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    size_t* num_dof_ptr, 
    enum TS_ROT_REPR rot_repr)
{
  struct ts_dof_map dof_map;
  RETURN_ON_ERROR( ts_dof_map_init(&dof_map, sim) );

  char* links = NULL;
#define CLEANUP_ON_ERROR { \
  ts_safe_free(NO_ALLOCATOR, links); \
  ts_dof_map_finalize(&dof_map); }

  RETURN_ON_ERROR3( get_linked_elements(sim, &dof_map, &links) );

  *hess_ptr = NULL;
  RETURN_ON_ERROR3(
      numerical_hessian_by_linked_objects(
        sim, hess_ptr, hess_err_ptr, &dof_map, links, rot_repr));

  *num_dof_ptr = dof_map.num_dof;

  ts_free(NO_ALLOCATOR, links);
  ts_dof_map_finalize(&dof_map);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

static inline size_t site_for_particle(const ts_rigid_body*restrict body, size_t pind)
{
  for(size_t i=0; i < body->type_ptr->num_sites; i++) {
    if (body->site_indices[i] == pind) return i;
  }

  return TS_NULL_PARTICLE;
}

static void MAYBE_UNUSED print_jacobian(const vector3 dq_w[3])
{
  fprintf(stderr, "\nJacobian:\n");
  for (int i=0; i < PARTICLE_NUM_DOF; i++) {
    fprintf(stderr, 
        "w = %1d [ %7.4f %7.4f %7.4f ]\n",
        i, 
        vector3_x(dq_w+i),
        vector3_y(dq_w+i),
        vector3_z(dq_w+i));
  }
  fprintf(stderr, "\n");
}

/* Cartesian X Cartesian */
static inline MAYBE_UNUSED void transform_cartesian_terms(ts_sim* sim,
    VecReal*restrict hess,
    size_t nbodies, int body_i, int body_j,
    size_t num_dof, size_t dof_i0, size_t dof_j0, 
    const VecReal*restrict weights,
    const VecReal*restrict cartesian_hess, int diffobj)
{
  TS_LOG(sim,TS_LOG_INFO,TS_LOG_NORMALMODE, "\n");
  for (int i = 0; i < PARTICLE_NUM_DOF; i++) {
    VecReal wi = weights[dof_i0+i];
    int j0 = diffobj? 0 : i;
    for (int j = j0; j < PARTICLE_NUM_DOF; j++) {
      VecReal wj = weights[dof_j0+j];
      int ci = PARTICLE_NUM_DOF*body_i + i;
      int cj = PARTICLE_NUM_DOF*body_j + j;
      size_t c_ind = hessian_index( PARTICLE_NUM_DOF*nbodies, ci, cj);
      size_t h_ind = hessian_index(num_dof, dof_i0+i, dof_j0+j);

      if ((dof_i0 == dof_j0) && (body_i != body_j)) {
        hess[ h_ind ] += 2.0*wi*wj*cartesian_hess[c_ind];
      } else {
        hess[ h_ind ] += wi*wj*cartesian_hess[c_ind];
      }
    }
  }
}

#define ROTATION_NUM_DOF  (RIGID_BODY_NUM_DOF-PARTICLE_NUM_DOF)
static inline MAYBE_UNUSED void transform_cartesian_rotation_terms(ts_sim* sim,
    VecReal*restrict hess,
    size_t nbodies, int body_i, int body_j,
    size_t num_dof, size_t dof_i0, size_t dof_j0, 
    const VecReal*restrict weights,
    const VecReal*restrict cartesian_hess,
    vector3 dq_w[3])
{
  TS_LOG(sim,TS_LOG_INFO,TS_LOG_NORMALMODE, "\n");
  for (int i = 0; i < ROTATION_NUM_DOF; i++) {
    size_t iind = dof_i0+i;
    VecReal wi = weights[iind];
    for (int ci = 0; ci < PARTICLE_NUM_DOF; ci++) {
      for (int j = 0; j < PARTICLE_NUM_DOF; j++) {
        size_t jind = dof_j0+j;
        VecReal wj = weights[jind];
        size_t hess_ind = hessian_index(num_dof, iind, jind);

        size_t c_ind_ij = hessian_index( PARTICLE_NUM_DOF*nbodies, PARTICLE_NUM_DOF*body_i+ci, PARTICLE_NUM_DOF*body_j+j);
        size_t c_ind_ji = hessian_index( PARTICLE_NUM_DOF*nbodies, PARTICLE_NUM_DOF*body_j+j, PARTICLE_NUM_DOF*body_i+ci);

        VecReal delta = VECTOR3_ELT(&dq_w[i], ci) * wi * wj * 
          (cartesian_hess[ c_ind_ij ] + (c_ind_ij != c_ind_ji ? cartesian_hess[ c_ind_ji ] : 0.0));

        hess[ hess_ind ] += delta;
      }
    }
  }
}

static inline MAYBE_UNUSED void transform_rotation_rotation_terms(ts_sim* sim,
    VecReal*restrict hess,
    size_t nbodies, int body_i, int body_j,
    size_t num_dof, size_t dof_i0, size_t dof_j0, 
    const VecReal*restrict weights,
    const VecReal*restrict cartesian_hess,
    vector3 dqi_w[3], vector3 dqj_w[3], int diff_obj)
{
  UNUSED_PARAM(sim);

  for (int i = 0; i < ROTATION_NUM_DOF; i++) {
    VecReal wi = weights[dof_i0+i];
    int j0 = diff_obj ? 0 : i; // avoid double-counting if the same object
    for (int j = j0; j < ROTATION_NUM_DOF; j++) {
      VecReal wj = weights[dof_j0+j];

      for (int ci = 0; ci < PARTICLE_NUM_DOF; ci++) {
        for (int cj = 0; cj < PARTICLE_NUM_DOF; cj++) {
          size_t c_hess_ind = hessian_index( PARTICLE_NUM_DOF*nbodies, PARTICLE_NUM_DOF*body_i+ci, PARTICLE_NUM_DOF*body_j+cj );
          VecReal c_hess_elt = cartesian_hess[ c_hess_ind ];

          size_t hess_ind = hessian_index(num_dof, dof_i0+i, dof_j0+j);
          hess[ hess_ind ] += VECTOR3_ELT(&dqi_w[i], ci) * VECTOR3_ELT(&dqj_w[j], cj) * wi * wj * c_hess_elt;
        }
      }

    }
  }
}

static inline MAYBE_UNUSED TSResult transform_gradient_terms(const ts_sim* sim,
    ts_sparse_builder_t* builder,
    size_t num_dof, size_t dof_i0,
    const VecReal*restrict weights,
    const vector3*restrict forces,
    vector3 d2q[PARTICLE_HESSIAN_SZ])
{
  UNUSED_PARAM(sim);
  UNUSED_PARAM(num_dof);

  for (int i1 = 0; i1 < ROTATION_NUM_DOF; i1++) {
    VecReal wi1 = weights[dof_i0+i1];
    for (int i2 = i1; i2 < ROTATION_NUM_DOF; i2++) {
      VecReal wi2 = weights[dof_i0+i2];
      size_t hess_i = dof_i0+i1;
      size_t hess_j = dof_i0+i2;
      size_t wind = hessian_index(PARTICLE_NUM_DOF, i1, i2);
      for (int fi=0; fi < PARTICLE_NUM_DOF; fi++) {
        RETURN_ON_ERROR(
            ts_sparse_builder_sub_from(builder, hess_i, hess_j,
              VECTOR3_ELT( d2q+wind, fi ) * wi1 * wi2 * VECTOR3_ELT(forces, fi)) );
        if (hess_i != hess_j)
        {
          RETURN_ON_ERROR(
              ts_sparse_builder_sub_from(builder, hess_j, hess_i,
                VECTOR3_ELT( d2q+wind, fi ) * wi1 * wi2 * VECTOR3_ELT(forces, fi)) );
        }
      }
    }
  }

  return TS_OK;
}

static inline void get_obj_info(ts_sim* sim, 
    struct ts_dof_map* dof_map,
    size_t pind, 
    size_t* oind_p,
    size_t*restrict dof0_p, 
    size_t*restrict rbind_p,
    size_t*restrict site_p)
{
  *oind_p = ts_dof_map_p2o(sim,dof_map,pind);
  size_t dof1 = 0;
  get_dof_info(*oind_p, dof_map->num_free_particles, dof0_p, &dof1);
  if (dof1 > *dof0_p + PARTICLE_NUM_DOF) {
    size_t rbind = TS_PARTICLE(sim,pind)->rigid_body_index;
    ts_rigid_body* body = ts_get_rigid_body(sim, rbind);
    *rbind_p = rbind;
    *site_p = site_for_particle(body, pind);
  }
}

static const enum MAT3X3_ROT_AXIS xyz_jacob_axes[3][3] = {
  { ROT_DX_AXIS, ROT_Y_AXIS, ROT_Z_AXIS },
  { ROT_X_AXIS, ROT_DY_AXIS, ROT_Z_AXIS },
  { ROT_X_AXIS, ROT_Y_AXIS, ROT_DZ_AXIS },
};

static const enum MAT3X3_ROT_AXIS zxz_jacob_axes[3][3] = {
  { ROT_DZ_AXIS, ROT_X_AXIS, ROT_Z_AXIS },
  { ROT_Z_AXIS, ROT_DX_AXIS, ROT_Z_AXIS },
  { ROT_Z_AXIS, ROT_X_AXIS, ROT_DZ_AXIS },
};

static inline void make_sitevec_jacobian(
    mat3x3*restrict jacob,
    const vector3*restrict sitevec,
    VecReal euler[3],
    const enum MAT3X3_ROT_AXIS axes[3][3])
{
  mat3x3 xf;
  vector3 vtmp;

  for (int i=0; i < 3; i++)
  {
    mat3x3_rotate_three_axes(&xf, 
        axes[i][0],
        axes[i][1],
        axes[i][2],
        euler);

    vector3_mat_rmul(&vtmp, &xf, sitevec);
    mat3x3_set_col(jacob, i, &vtmp);
  }
}

void ts_rigid_body_xyz_body_axes_jacobian(const ts_sim* sim,
    size_t rbind, size_t site,
    mat3x3*restrict jacobian)
{
  mat3x3 xf = MAT3X3_ZERO;
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);
  vector3*restrict sitevec = body->type_ptr->ref_vectors + site;

  mat3x3_zero(jacobian);

  VecReal euler[3] = {0,0,0};
  enum TS_ROT_REPR rot_repr = rigid_body_euler_coords(sim, rbind, euler);

  mat3x3 xf_body_to_euler;
  body_to_euler( &xf_body_to_euler, euler, rot_repr );

  mat3x3_zero(&xf);

  make_sitevec_jacobian(&xf, 
      sitevec, 
      euler, 
      (rot_repr == TS_ROT_EULER_XYZ) 
        ? xyz_jacob_axes 
        : zxz_jacob_axes );

  mat3x3_mul(jacobian, &xf, &xf_body_to_euler);
}

static const enum MAT3X3_ROT_AXIS xyz_hess_axes[3][3][3] = {
  { { ROT_DX2_AXIS, ROT_Y_AXIS  , ROT_Z_AXIS   }, 
    { ROT_DX_AXIS , ROT_DY_AXIS , ROT_Z_AXIS   },
    { ROT_DX_AXIS , ROT_Y_AXIS  , ROT_DZ_AXIS  } },

  { { ROT_DX_AXIS , ROT_DY_AXIS , ROT_Z_AXIS   },
    { ROT_X_AXIS  , ROT_DY2_AXIS, ROT_Z_AXIS   },
    { ROT_X_AXIS  , ROT_DY_AXIS , ROT_DZ_AXIS  } },

  { { ROT_DX_AXIS , ROT_Y_AXIS  , ROT_DZ_AXIS  },
    { ROT_X_AXIS  , ROT_DY_AXIS , ROT_DZ_AXIS  },
    { ROT_X_AXIS  , ROT_Y_AXIS  , ROT_DZ2_AXIS } },
};

static const enum MAT3X3_ROT_AXIS zxz_hess_axes[3][3][3] = {
  { { ROT_DZ2_AXIS, ROT_X_AXIS  , ROT_Z_AXIS   }, 
    { ROT_DZ_AXIS , ROT_DX_AXIS , ROT_Z_AXIS   },
    { ROT_DZ_AXIS , ROT_X_AXIS  , ROT_DZ_AXIS  } },

  { { ROT_DZ_AXIS , ROT_DX_AXIS , ROT_Z_AXIS   },
    { ROT_Z_AXIS  , ROT_DX2_AXIS, ROT_Z_AXIS   },
    { ROT_Z_AXIS  , ROT_DX_AXIS , ROT_DZ_AXIS  } },

  { { ROT_DZ_AXIS , ROT_X_AXIS  , ROT_DZ_AXIS  },
    { ROT_Z_AXIS  , ROT_DX_AXIS , ROT_DZ_AXIS  },
    { ROT_Z_AXIS  , ROT_X_AXIS  , ROT_DZ2_AXIS } },
};

// Second derivatives of the site coordinates wrt to the angular terms
// form a 3rd order tensor D(l,m,n) for a rigid body site (b,k)
//
//   l -- element of site vector r_{b,k}
//   m -- element of angular coordinate Omega_{b}
//   n -- element of angular coordinate Omega_{b}
//
// This can be stored either as a vector of three symmetric matrices
// (VoM), where each matrix represents the derivative of r_{b,k,l} wrt
// two angular coordinates, or it can be stored as a matrix of
// 3-vectors (MoV).
//
// The VoM form is easier to calculate using the mat3x3 structures.  The
// MoV form is easier to use in subsequent calculations.  This converts
// between the two, applying an optional orthogonal transformation
// (e.g.: to change variables) in the process
//
static inline void xform_angular_hessian_to_vector_repr(
    vector3*restrict hessv,  // vector3 hessv[PARTICLE_HESSIAN_SZ]
    const mat3x3*restrict orth_xform,
    mat3x3*restrict hess_x,
    mat3x3*restrict hess_y,
    mat3x3*restrict hess_z)
{
  if (orth_xform)
  {
    mat3x3_self_orth_xform(hess_x, orth_xform);
    mat3x3_self_orth_xform(hess_y, orth_xform);
    mat3x3_self_orth_xform(hess_z, orth_xform);
  }

  for (int i=0; i < PARTICLE_NUM_DOF; i++)
  {
    for (int j=i; j < PARTICLE_NUM_DOF; j++)
    {
      size_t off = hessian_index(PARTICLE_NUM_DOF, i,j);
      vector3*restrict vec = hessv + off;
      VECTOR3_X(vec) = MAT3X3_ELT(hess_x, i,j);
      VECTOR3_Y(vec) = MAT3X3_ELT(hess_y, i,j);
      VECTOR3_Z(vec) = MAT3X3_ELT(hess_z, i,j);
    }
  }
}

static inline void gen_hess_matrices(
    mat3x3*restrict hess_x, 
    mat3x3*restrict hess_y,
    mat3x3*restrict hess_z,
    const vector3*restrict sitevec,
    VecReal euler[3],
    const enum MAT3X3_ROT_AXIS axes[3][3][3])
{
  mat3x3 xf1;
  vector3 vtmp;
  for(int i=0; i < 3; i++) {                      
    for (int j=0; j < 3; j++) {                   
      mat3x3_rotate_three_axes(&xf1,              
          axes[i][j][0],                          
          axes[i][j][1],                          
          axes[i][j][2],                          
          euler);                                 
      vector3_mat_rmul(&vtmp, &xf1, sitevec);     
      MAT3X3_ELT(hess_x, i,j) = VECTOR3_X(&vtmp);
      MAT3X3_ELT(hess_y, i,j) = VECTOR3_Y(&vtmp);
      MAT3X3_ELT(hess_z, i,j) = VECTOR3_Z(&vtmp);
    }
  } 
}

void ts_rigid_body_xyz_body_axes_hessian(const ts_sim* sim,
    size_t rbind, size_t site,
    vector3*restrict hessv /*[PARTICLE_HESSIAN_SZ]*/)
{
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);
  vector3*restrict sitevec = body->type_ptr->ref_vectors + site;

  mat3x3 hess_x = MAT3X3_ZERO;
  mat3x3 hess_y = MAT3X3_ZERO;
  mat3x3 hess_z = MAT3X3_ZERO;

  for (int i=0; i < PARTICLE_HESSIAN_SZ; i++)
  {
    vector3_zero(hessv+i);
  }

  VecReal euler[3] = {0,0,0};
  enum TS_ROT_REPR rot_repr = rigid_body_euler_coords(sim, rbind, euler);

  mat3x3 euler_xform = MAT3X3_ZERO;
  body_to_euler(&euler_xform, euler, rot_repr);

  if (rot_repr == TS_ROT_EULER_XYZ)
  {
    gen_hess_matrices(&hess_x, &hess_y, &hess_z, sitevec, euler, xyz_hess_axes);
  }
  else if (rot_repr == TS_ROT_EULER_ZXZ)
  {
    gen_hess_matrices(&hess_x, &hess_y, &hess_z, sitevec, euler, zxz_hess_axes);
  }

  xform_angular_hessian_to_vector_repr(
      hessv,
      &euler_xform,
      &hess_x,
      &hess_y,
      &hess_z);
}


#define HCART_SIZE   (NBODY_LIMIT*PARTICLE_NUM_DOF)
#define HPROJ_SIZE   (NBODY_LIMIT*RIGID_BODY_NUM_DOF)
#define XFORM_SIZE   HCART_SIZE*HPROJ_SIZE
#define RESULT_SIZE  HPROJ_SIZE*HPROJ_SIZE


#define _ELT(m,i,j) ( (m)*(j) + (i) )
#define XFORM(pind, ci,gj) \
  ( xform[ _ELT(PARTICLE_NUM_DOF*nbodies,     \
                PARTICLE_NUM_DOF*(pind) + ci, \
                RIGID_BODY_NUM_DOF*(pind)+(gj)) ] )
#define HPROJ(pind, gi) \
  ( hproj[ ( RIGID_BODY_NUM_DOF*(pind) + (gi) ) ] )
static void MAYBE_UNUSED build_xform_matrix(ts_sim*restrict sim,
    struct ts_dof_map* dof_map, 
    const ts_potential_term*restrict term, 
    const VecReal*restrict dofvec,
    const enum TS_ROT_REPR* rot_reprs,
    VecReal xform[XFORM_SIZE],
    size_t hproj[HPROJ_SIZE])
{
  UNUSED_PARAM(dofvec);
  UNUSED_PARAM(rot_reprs);

  memset(xform, 0, XFORM_SIZE*sizeof(xform[0]));
  memset(hproj, 0, HPROJ_SIZE*sizeof(hproj[0]));

  int nbodies = term->potential->nbodies;
  for (int i=0; i < nbodies; i++) {
    size_t pi = term->particle_indices[i];
    size_t oi = 0;
    size_t beg_dof_i = 0;
    size_t rb_i = TS_NULL_RIGID_BODY;
    size_t site_i = TS_NULL_PARTICLE;

    get_obj_info(sim, dof_map, pi, &oi, &beg_dof_i, &rb_i, &site_i);

    XFORM(i, 0,0) = 1.0;
    XFORM(i, 1,1) = 1.0;
    XFORM(i, 2,2) = 1.0;

    HPROJ(i, 0) = beg_dof_i+0;
    HPROJ(i, 1) = beg_dof_i+1;
    HPROJ(i, 2) = beg_dof_i+2;

    if (site_i != TS_NULL_PARTICLE) {
      mat3x3 jacobian;

      ts_rigid_body_xyz_body_axes_jacobian(sim, rb_i, site_i, &jacobian);

      for (int ci=0; ci < PARTICLE_NUM_DOF; ci++) {
        for (int gj=0; gj < ROTATION_NUM_DOF; gj++) {
          XFORM(i, ci,PARTICLE_NUM_DOF+gj) = MAT3X3_ELT(&jacobian, ci, gj);
        }
      }

      HPROJ(i,3) = beg_dof_i+3;
      HPROJ(i,4) = beg_dof_i+4;
      HPROJ(i,5) = beg_dof_i+5;
    } else {
      HPROJ(i,3) = TS_NULL_PARTICLE;
      HPROJ(i,4) = TS_NULL_PARTICLE;
      HPROJ(i,5) = TS_NULL_PARTICLE;
    }
  }
}

static inline TSResult add_gradient_terms(const ts_sim* sim, size_t num_dof,
    size_t rbind,
    size_t site_i, size_t beg_dof_i,
    const VecReal*restrict weights,
    const vector3*restrict f,
    ts_sparse_builder_t* builder)
{
  vector3 hessv[PARTICLE_HESSIAN_SZ];
  ts_rigid_body_xyz_body_axes_hessian(sim, rbind, site_i, hessv);

  RETURN_ON_ERROR(
      transform_gradient_terms(sim, builder,
        num_dof, beg_dof_i+PARTICLE_NUM_DOF,
        weights, f, hessv) );
  return TS_OK;
}

static TSResult convert_cartesian_hessian_to_hessian_mat(ts_sim* sim,
    struct ts_dof_map* dof_map,
    const ts_potential_term* term, 
    vector3*restrict f, // f[NBODY_LIMIT]
    const VecReal*restrict cartesian_hessian, 
    const VecReal*restrict weights,
    const VecReal*restrict dofvec,
    const enum TS_ROT_REPR* rot_reprs,
    ts_sparse_builder_t* builder)
{
  VecReal xform[XFORM_SIZE];
  VecReal result[RESULT_SIZE];
  size_t hproj[HPROJ_SIZE];

  build_xform_matrix(sim, dof_map, term, dofvec, rot_reprs, xform, hproj);

  // Project Cartesian Hessian onto 6N x 6N RB Hessian
  VecReal tmp[PARTICLE_NUM_DOF*NBODY_LIMIT];
  const int xform_m = PARTICLE_NUM_DOF*term->potential->nbodies;
  const int xform_n = RIGID_BODY_NUM_DOF*term->potential->nbodies;

  const double alpha = 1.0;
  const double beta = 0.0;
  int inc = 1;
  for (int i=0; i < xform_n; i++) {
    dspmv("L",                          // UPLO
        &xform_m,                       // N
        &alpha,                         // ALPHA
        cartesian_hessian,              // AP
        xform+i*xform_m,                // X
        &inc,                           // INCX
        &beta,                          // BETA
        tmp,                            // Y
        &inc                            // INCY
        );

    dgemv("T",                          // TRANS
        &xform_m,                       // M
        &xform_n,                       // N
        &alpha,                         // ALPHA
        xform,                          // A
        &xform_m,                       // LDA
        tmp,                            // X
        &inc,                           // INCX
        &beta,                          // BETA
        result+i*xform_n,               // Y
        &inc                            // INCY
        );
  }

  // Use projection map to project RB Hessian onto the full simulation
  // Hessian
  for (int i=0; i < xform_n; i++)
  {
    if (hproj[i] == TS_NULL_PARTICLE) continue;
    VecReal wi = weights[ hproj[i] ];

    for (int j=0; j < xform_n; j++) {
      if (hproj[j] == TS_NULL_PARTICLE) continue;
      VecReal wj = weights[ hproj[j] ];
      if (hproj[i] >= hproj[j])
      {
        double element = wi*wj*result[xform_n*i+j];
        if (element != 0.0)
        {
          RETURN_ON_ERROR(
              ts_sparse_builder_add_to(builder, hproj[i],hproj[j], 
                wi*wj*result[xform_n*i + j])
              );
        }

        if (hproj[i] != hproj[j])
        {
          RETURN_ON_ERROR(
              ts_sparse_builder_add_to(builder, hproj[j],hproj[i], 
                wi*wj*result[xform_n*i + j])
              );
        }
      }
    }
  }

  for (int i=0; i < term->potential->nbodies; i++)
  {
    size_t pi = term->particle_indices[i];
    size_t oi = 0;
    size_t beg_dof_i = 0;
    size_t rb_i = TS_NULL_RIGID_BODY;
    size_t site_i = TS_NULL_PARTICLE;

    get_obj_info(sim, dof_map, pi, &oi, &beg_dof_i, &rb_i, &site_i);
    if (site_i != TS_NULL_PARTICLE)
    {
      RETURN_ON_ERROR(
          add_gradient_terms(sim, dof_map->num_dof, 
            rb_i, site_i, beg_dof_i, weights, f+i, builder) );
    }
  }

  return TS_OK;
}

static inline void init_weights_uniform(size_t num_dof, VecReal* weights)
{
  for (size_t i=0; i < num_dof; i++) weights[i] = 1.0;
}

static inline void init_weights_by_mass(const ts_sim* sim,
    const struct ts_dof_map* dof_map,
    VecReal* weights)
{
  size_t off = 0;
  for (size_t i=0; i < dof_map->num_free_particles; i++) {
    const ts_particle_type*restrict ptype = 
      TS_PARTICLE(sim,dof_map->fp_to_p[i])->type;
    for(int j=0; j < 3; j++) weights[off+j] = ptype->inv_sqrt_mass;
    off += PARTICLE_NUM_DOF;
  }

  for (size_t i=0; i < sim->num_rigid_bodies; i++) {
    const struct ts_rigid_body_type*restrict rbtype = 
      ts_get_rigid_body(sim,i)->type_ptr;
    for (int j=0; j < 6; j++) weights[off+j] = rbtype->inv_sqrt_I0;
    off += RIGID_BODY_NUM_DOF;
  }
}

static inline void zero_vectors(vector3* vv, int n)
{
  for(int i=0; i < n; i++) { vector3_zero(vv+i); }
}

static TSResult calc_weighted_hessian_inner(
    struct ts_autopool* pool, ts_sim* sim, 
    ts_sparse_builder_t* builder, 
    struct ts_dof_map* dof_map,
    int weight_by_mass,
    int enabled[TS_BOND_TERM_MAX])
{
  VecReal cartesian_hessian[ HESSIAN_SIZE(3*NBODY_LIMIT) ] = { 0.0 };

#define CLEANUP_ON_ERROR {}

  VecReal* weights = ts_autopool_alloc(pool, dof_map->num_dof * sizeof(*weights));
  ERROR_ON_NULL_ALLOC(weights);

  enum TS_ROT_REPR* rot_conventions = 
    ts_autopool_alloc(pool, sim->num_rigid_bodies * sizeof(*rot_conventions));
  ERROR_ON_NULL_ALLOC(rot_conventions);

  VecReal* xv = ts_autopool_alloc(pool, dof_map->num_dof * sizeof(*xv));
  ERROR_ON_NULL_ALLOC(xv);

  ts_sim_copy_dof_to_vec_in_body_axes(sim, dof_map, xv, rot_conventions);

  memset(weights,0, dof_map->num_dof*sizeof(*weights));
  if (weight_by_mass) {
    init_weights_by_mass(sim, dof_map, weights);
  } else {
    init_weights_uniform(dof_map->num_dof, weights);
  }

  if (sim->bondlist.terms.size == 0) return TS_OK;

  const ts_potential_term* terms = sim->bondlist.terms.data;
  const ts_potential_term* end = terms + sim->bondlist.terms.size;

  vector3 r[NBODY_LIMIT];
  vector3 f[NBODY_LIMIT];

  do {
    memset( cartesian_hessian, 0, sizeof(cartesian_hessian) );
    if (enabled && terms->pattern && !enabled[terms->pattern->type])
      continue;

    zero_vectors(f, NBODY_LIMIT);
    for (int i=0; i < terms->potential->nbodies; i++)
    {
      vector3_set(&r[i], PARTICLE_POS(sim, terms->particle_indices[i]));
    }

    ts_potential_hessian(terms->potential, r, cartesian_hessian);
    terms->potential->forces_func(terms->potential, r, f);

    convert_cartesian_hessian_to_hessian_mat(sim,
        dof_map, terms, 
        f, 
        cartesian_hessian, 
        weights,
        xv,
        rot_conventions,
        builder);
  } while(++terms < end);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

static TSResult calc_weighted_hessian(ts_sim* sim, 
    ts_sparse_builder_t* hess,
    int weight_by_mass,
    int enabled[TS_BOND_TERM_MAX])
{
  struct ts_autopool pool;
  struct ts_dof_map dof_map;

  ts_autopool_init(&pool, NO_ALLOCATOR);

#define CLEANUP_ON_ERROR \
  { ts_autopool_release(&pool); \
    ts_safe_free(NO_ALLOCATOR, hess); }

  RETURN_ON_ERROR3( ts_dof_map_init(&dof_map, sim) );
  AUTOPOOL_ON_RELEASE(&pool, ts_dof_map_finalize, &dof_map);

  RETURN_ON_ERROR3( ts_sparse_builder_initialize(hess, 
        dof_map.num_dof, dof_map.num_dof) );

  RETURN_ON_ERROR3( calc_weighted_hessian_inner(&pool,sim,
        hess, &dof_map, weight_by_mass, enabled) );

#undef CLEANUP_ON_ERROR

  ts_autopool_release(&pool);
  return TS_OK;
}

TSResult ts_sim_hessian(ts_sim* sim, ts_sparse_builder_t* hess,
  int enabled[TS_BOND_TERM_MAX])
{
  RETURN_ON_ERROR( calc_weighted_hessian(sim, hess, 0, enabled) );
  return TS_OK;
}

TSResult ts_sim_mass_weighted_hessian(ts_sim* sim, ts_sparse_builder_t* hess,
  int enabled[TS_BOND_TERM_MAX])
{
  RETURN_ON_ERROR( calc_weighted_hessian(sim, hess, 1, enabled) );

  return TS_OK;
}


TSResult ts_sim_nma_build_hessian(ts_sim_nma* nma)
{

#define CLEANUP_ON_ERROR {}

  RETURN_ON_ERROR( ts_sim_mass_weighted_hessian(nma->sim, &nma->hess_builder, nma->enabled) );
  AUTOPOOL_ON_RELEASE( &nma->pool, ts_sparse_builder_finalize, &nma->hess_builder);

  RETURN_ON_ERROR( ts_sparse_builder_build(&nma->hess_builder,&nma->hess_mat) );
  AUTOPOOL_ON_RELEASE( &nma->pool, ts_sparse_finalize, &nma->hess_mat );

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

static TSResult nma_build_sparse_matrix(ts_sim_nma* nma)
{

#define CLEANUP_ON_ERROR {}

  RETURN_ON_ERROR( ts_sparse_builder_build(&nma->hess_builder,&nma->hess_mat) );
  AUTOPOOL_ON_RELEASE( &nma->pool, ts_sparse_finalize, &nma->hess_mat );

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

static TSResult nma_calc_eigensystem(ts_sim_nma* nma)
{

  VecReal** freqsp = (nma->flags & TS_NMA_FL_CALC_FREQS) ? &nma->freqs : NULL;
  VecReal** modesp = (nma->flags & TS_NMA_FL_CALC_MODES) ? &nma->modes : NULL;

  switch (nma->method)
  {
    case TS_NMA_DENSE:
      {
        RETURN_ON_ERROR( nma_build_sparse_matrix(nma) );
        RETURN_ON_ERROR( ts_sparse_eigensystem_dense(
              &nma->pool, &nma->hess_mat, 
              freqsp, modesp) );
        nma->freq_off = nma->freq_skip;
      }
      break;
    case TS_NMA_SPARSE:
      {
        RETURN_ON_ERROR( nma_build_sparse_matrix(nma) );
        RETURN_ON_ERROR( ts_sparse_eigensystem_sparse(
              &nma->pool, &nma->hess_mat, 
              freqsp, modesp,
              nma->freq_skip) );
        nma->freq_off = 0;
      }
      break;
    case TS_NMA_BANDED:
      {
        RETURN_ON_ERROR(
            ts_sparse_builder_eigensystem_banded(
              &nma->pool, &nma->hess_builder, 
              freqsp, modesp) );
        nma->freq_off = nma->freq_skip;
      }
      break;
  }

  nma->nfreqs = nma->hess_mat.m - nma->freq_skip;

  return TS_OK;
}

TSResult ts_sim_nma_calc_freqs(ts_sim_nma* nma)
{
  RETURN_ON_ERROR( ts_calculate_forces(nma->sim, &nma->enthalpic_energy, NO_TERM_GROUPS, TERMWISE_FALSE) );
  RETURN_ON_ERROR( ts_sim_nma_build_hessian(nma) );
  RETURN_ON_ERROR( nma_calc_eigensystem(nma) );

  return TS_OK;
}

static inline TSResult setup_nma(ts_sim_nma* nma,
    ts_sim* sim, uint64_t flags,
    int enabled[TS_BOND_TERM_MAX], size_t skip)
{
  ts_sim_nma_init(nma, sim);

  if (enabled)
  {
    memcpy(nma->enabled, enabled, TS_BOND_TERM_MAX*sizeof(enabled[0]));
  }

  nma->flags = flags;
  nma->freq_skip = skip;

  return TS_OK;
}

TSResult ts_sim_normal_mode_square_freqs(ts_sim* sim, 
    VecReal** freqp, VecReal** modep, size_t* np, 
    double* potE_ptr, int enabled[TS_BOND_TERM_MAX], 
    size_t skip, size_t* offp)
{
  ts_sim_nma nma;

  uint64_t flags = TS_NMA_FL_NONE;
  if (freqp) { flags |= TS_NMA_FL_CALC_FREQS; }
  if (modep) { flags |= TS_NMA_FL_CALC_MODES; }

  RETURN_ON_ERROR( setup_nma(&nma, sim, flags, enabled, skip) );

  RETURN_ON_ERROR2(
      ts_sim_nma_calc_freqs(&nma),
      ts_sim_nma_finalize(&nma));

  if (freqp) { *freqp = nma.freqs; }
  if (modep) { *modep = nma.modes; }

  if (potE_ptr) { *potE_ptr = nma.enthalpic_energy; }
  if (np)       { *np       = nma.nfreqs;           }
  if (offp)     { *offp     = nma.freq_off;         }

  ts_sim_nma_finalize(&nma);

  return TS_OK;
}

TSResult ts_sim_nma_calc_free_energy(ts_sim_nma* nma)
{
  double entropic_energy = 0.0;
  for(size_t ind = 0; ind < nma->nfreqs; ind++)
  {
    double freq = sqrt(nma->freqs[ind+nma->freq_off]);
    entropic_energy += log( HARMONIC_FREE_ENERGY_COEFF * freq / nma->sim->temp );
  }

  entropic_energy *= nma->sim->temp;
  nma->entropic_energy = entropic_energy;
  nma->free_energy = nma->enthalpic_energy + nma->entropic_energy;

  return TS_OK;
}

TSResult ts_sim_normal_mode_free_energy(ts_sim* sim, VecReal* energyp, int linear)
{
  ts_sim_nma nma;

  if (sim->num_particles == 2) { linear = 1; }

  size_t skip = linear ? 5 : 6;

  RETURN_ON_ERROR( setup_nma(&nma, sim, TS_NMA_FL_CALC_FREQS, NULL, skip) );
  RETURN_ON_ERROR2(
      ts_sim_nma_free_energy(&nma),
      ts_sim_nma_finalize(&nma));

  *energyp = nma.free_energy;

  ts_safe_free(NO_ALLOCATOR, nma.freqs);
  ts_sim_nma_finalize(&nma);

  return TS_OK;
}


