#ifndef TS_NORMALMODE_H
#define TS_NORMALMODE_H

#include "tubesheet.h"
#include "ts_dof_map.h"
#include "ts_sparse.h"

/*! For values of theta less than this, the Euler ZXZ convention starts
 * to becomes singular (Gimble lock), and the XYZ convention is a better
 * choice */
#define ZXZ_THETA_CUTOFF 0.261799387799149 /* 15 degrees */

#define XYZ_BETA_CUTOFF 1.30899693899575  /* 90-15=75 degrees */

extern int zxz_default_coords;

void ts_sim_copy_dof_to_vec_in_body_axes(
    ts_sim* sim, 
    struct ts_dof_map* dof_map,
    VecReal*restrict xv, 
    enum TS_ROT_REPR* rot_conventions);

TSResult ts_sim_numerical_hessian_linked_elements(ts_sim* sim,
    char** links_ptr, size_t* num_objs_ptr);

TSResult ts_sim_numerical_hessian_check_all_dof(ts_sim* sim,
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    size_t* num_dof_ptr,
    enum TS_ROT_REPR rot_repr);

TSResult ts_sim_numerical_hessian_by_potential_terms(ts_sim* sim,
    VecReal** hess_ptr, 
    VecReal** hess_err_ptr, 
    size_t* num_dof_ptr,
    enum TS_ROT_REPR rot_repr);

TSResult ts_sim_hessian(ts_sim* sim,
    ts_sparse_builder_t* hess,
    int enabled[TS_BOND_TERM_MAX]);

TSResult ts_sim_mass_weighted_hessian(ts_sim* sim,
    ts_sparse_builder_t* hess,
    int enabled[TS_BOND_TERM_MAX]);

TSResult ts_sim_normal_mode_square_freqs(ts_sim* sim,
    VecReal** freqp, VecReal** modep, size_t* np, 
    double* potE_ptr,
    int enabled[TS_BOND_TERM_MAX],
    size_t skip, size_t* off);

TSResult ts_sim_normal_mode_free_energy(ts_sim* sim,
    VecReal* energyp, int linear);

enum TS_NMA_METHOD
{
  TS_NMA_DENSE  = 0,
  TS_NMA_SPARSE = 1,
  TS_NMA_BANDED = 2,
};

#define TS_NMA_METHOD_DEFAULT TS_NMA_BANDED

enum TS_NMA_FLAGS
{
  TS_NMA_FL_NONE       = 0,
  TS_NMA_FL_CALC_FREQS = 1,
  TS_NMA_FL_CALC_MODES = 2,
};

typedef struct ts_sim_nma ts_sim_nma;
struct ts_sim_nma
{
  ts_sim* sim;

  struct ts_autopool pool;
  struct ts_dof_map dof_map;

  enum TS_NMA_METHOD method;

  int enabled[TS_BOND_TERM_MAX];

  size_t freq_skip;
  size_t freq_off;

  uint64_t flags;

  size_t nfreqs;
  VecReal* freqs;
  VecReal* modes;

  VecReal enthalpic_energy;
  VecReal entropic_energy;
  VecReal free_energy;

  ts_sparse_builder_t hess_builder;
  ts_sparse_t hess_mat;
};

void ts_sim_nma_init(ts_sim_nma* nma, ts_sim* sim);
void ts_sim_nma_finalize(ts_sim_nma* nma);

TSResult ts_sim_nma_build_hessian(ts_sim_nma* nma);
TSResult ts_sim_nma_calc_freqs(ts_sim_nma* nma);
TSResult ts_sim_nma_calc_free_energy(ts_sim_nma* nma);

static inline TSResult ts_sim_nma_free_energy(ts_sim_nma* nma)
{
  RETURN_ON_ERROR( ts_sim_nma_calc_freqs(nma) );
  RETURN_ON_ERROR( ts_sim_nma_calc_free_energy(nma) );

  return TS_OK;
}


#endif /* TS_NORMALMODE_H */
