#include "ts_numerical.h"
#include <math.h>

#if __STDC_VERSION__ >= 199901L && __STDC_IEC_559__  /* C99, use fenv */
#  define USE_FENV_ACCESS 1
#endif

#if defined(__clang__) 
#  undef USE_FENV_ACCESS
#endif

#if defined(__GNUC__) && __GNUC__ <= 4 && __GNUC_MINOR__ <= 9
#  undef USE_FENV_ACCESS
#endif

#if __STDC_VERSION__ >= 199901L   /* C99, use fenv */
#  include <fenv.h>
#  if defined(USE_FENV_ACCESS)
#    pragma STDC FENV_ACCESS ON
#  endif

enum TS_ROUNDING_MODES ts_get_rounding(void)
{
  const int originalRounding = fegetround();
  switch (originalRounding) {
    case FE_TONEAREST:  return TS_ROUND_NEAREST;
    case FE_TOWARDZERO: return TS_ROUND_ZERO;
    case FE_UPWARD:     return TS_ROUND_UPWARD;   /* to +infinity */
    case FE_DOWNWARD:   return TS_ROUND_DOWNWARD; /* to -infinity */
    default: return TS_ROUND_INVALID;
  }
}

TSResult ts_set_rounding(enum TS_ROUNDING_MODES mode)
{
  int rmode = 0;
  switch (mode) {
    default:
    case TS_ROUND_INVALID: 
      TS_RAISE_ERROR(TS_ERR_INVALID_VALUE); break; 

    case TS_ROUND_NEAREST:  rmode = FE_TONEAREST;  break;
    case TS_ROUND_ZERO:     rmode = FE_TOWARDZERO; break;
    case TS_ROUND_UPWARD:   rmode = FE_UPWARD;     break;
    case TS_ROUND_DOWNWARD: rmode = FE_DOWNWARD;   break;
  }

  if (0 == fesetround(rmode)) {
    return TS_OK;
  } else {
    TS_RAISE_ERROR2(TS_ERR_UNSUPPORTED_OPERATION,
        ("Could not set rounding mode to %d (0x%x)", mode, rmode));
  }
}
void ts_set_rounding_default(void) { ts_set_rounding(TS_ROUND_NEAREST); }
#else
/*Not C99*/
enum TS_ROUNDING_MODES ts_get_rounding(void) { return TS_ROUND_INVALID; }
TSResult ts_set_rounding(enum TS_ROUNDING_MODES mode)
{ 
  TS_RAISE_ERROR2(TS_ERR_UNIMPLEMENTED,
      "IEEE rounding modes are not implemented for this compiler/processor"
      );
}
#endif

#define DERIV_NTAB 10
#define CONTRACT_STEP 1.4
#define CONTRACT_STEP_SQ (CONTRACT_STEP*CONTRACT_STEP)
#define BIG  INFINITY
#define SAFE 2.0

TSResult deriv_ridders(single_var_func func, void* closure, double x, double h0, 
    double* ansp, double* errp)
{
  double a[DERIV_NTAB][DERIV_NTAB];
  double h = h0;

  ERROR_IF( TS_ERR_INVALID_ARGUMENT, h0 <= 0.0);

  a[0][0] = (func(closure, x+h) - func(closure, x-h))/(2.0*h);
  *errp = BIG;
  *ansp = a[0][0];

  for (int i=1; i < DERIV_NTAB; i++) {
    h /= CONTRACT_STEP;
    a[0][i] = (func(closure, x+h) - func(closure, x-h))/(2.0*h);
    double fac = CONTRACT_STEP_SQ;
    for (int j=1; j <= i; j++) {
      a[j][i] = (a[j-1][i] * fac - a[j-1][i-1])/(fac-1.0);
      fac *= CONTRACT_STEP_SQ;
      double errt = max_double( 
          fabs(a[j][i] - a[j-1][i]),
          fabs(a[j][i] - a[j-1][i-1]));
      if (errt <= *errp) {
        *errp = errt;
        *ansp = a[j][i];
      }
    }
    if (fabs(a[i][i] - a[i-1][i-1]) >= SAFE * (*errp)) break;
  }

  return TS_OK;
}


TSResult deriv2_ridders(single_var_func func, void* closure, double x, double h0, 
    double* ansp, double* errp)
{
  double a[DERIV_NTAB][DERIV_NTAB] = {{0}};
  double h = h0;
  size_t i;

  ERROR_IF( TS_ERR_INVALID_ARGUMENT, h0 <= 0.0);

  for (i=0; i < DERIV_NTAB; i++)
  {
    size_t j;
    for (j=0; j < DERIV_NTAB; j++)
    {
      a[i][j] = 0.0;
    }
  }

  a[0][0] = (
      func(closure, x+h) - 2.0*func(closure,x) + func(closure, x-h)
      )/(h*h);
  *errp = BIG;
  *ansp = a[0][0];

  for (int i=1; i < DERIV_NTAB; i++) {
    h /= CONTRACT_STEP;
    a[0][i] = (
        func(closure, x+h) - 2.0*func(closure,x) + func(closure, x-h)
        )/(h*h);
    double fac = CONTRACT_STEP_SQ;
    for (int j=1; j <= i; j++) {
      double a1 = a[j-1][i], a0 = a[j-1][i-1];
      double anew = (a1 * fac - a0)/(fac-1.0);
      a[j][i] = anew;
      fac *= CONTRACT_STEP_SQ;
      double errt = max_double( fabs(anew-a1), fabs(anew-a0) );
      if (errt <= *errp) {
        *errp = errt;
        *ansp = a[j][i];
      }
    }
    if (fabs(a[i][i] - a[i-1][i-1]) >= SAFE * (*errp)) break;
  }

  return TS_OK;
}

TSResult deriv_xy_ridders(double_var_func func, void* closure, double x,
    double y, double h0, double* ansp, double* errp)
{
  double a[DERIV_NTAB][DERIV_NTAB] = {{0}};
  double h = h0;

  ERROR_IF( TS_ERR_INVALID_ARGUMENT, h0 <= 0.0);

  a[0][0] = (
      (func(closure, x+h, y+h) - func(closure, x-h, y+h)) - 
      (func(closure, x+h, y-h) - func(closure, x-h, y-h))
      )/(4.0*h*h);
  *errp = BIG;
  *ansp = a[0][0];

  for (int i=1; i < DERIV_NTAB; i++) {
    h /= CONTRACT_STEP;
    a[0][i] = (
        (func(closure, x+h, y+h) - func(closure, x-h, y+h)) - 
        (func(closure, x+h, y-h) - func(closure, x-h, y-h))
        )/(4.0*h*h);
    double fac = CONTRACT_STEP_SQ;
    for (int j=1; j <= i; j++) {
      a[j][i] = (a[j-1][i] * fac - a[j-1][i-1])/(fac-1.0);
      fac *= CONTRACT_STEP_SQ;
      double errt = max_double( 
          fabs(a[j][i] - a[j-1][i]),
          fabs(a[j][i] - a[j-1][i-1]));
      if (errt <= *errp) {
        *errp = errt;
        *ansp = a[j][i];
      }
    }
    if (fabs(a[i][i] - a[i-1][i-1]) >= SAFE * (*errp)) break;
  }

  return TS_OK;
}

#undef DERIV_NTAB
#undef CONTRACT_STEP
#undef CONTRACT_STEP_SQ
#undef BIG 
#undef SAFE


static inline void deriv_cross_x(vector3*restrict result, const vector3*restrict v)
{
  vector3_displace(result, 0.0, -1.0*vector3_z(v), vector3_y(v));
}

static inline void deriv_cross_y(vector3*restrict result, const vector3*restrict v)
{
  vector3_displace(result, vector3_z(v), 0, -1.0*vector3_x(v));
}

static inline void deriv_cross_z(vector3*restrict result, const vector3*restrict v)
{
  vector3_displace(result, -1.0*vector3_y(v), vector3_x(v), 0.0);
}

static inline void deriv_double_cross_x(vector3*restrict result, const vector3*restrict aa, const vector3*restrict v)
{
  vector3_displace(result,
      vector3_y(aa)*vector3_y(v) +   vector3_z(aa)*vector3_z(v),
      vector3_y(aa)*vector3_x(v) - 2*vector3_x(aa)*vector3_y(v),
      vector3_z(aa)*vector3_x(v) - 2*vector3_x(aa)*vector3_z(v));
}

static inline void deriv_double_cross_y(vector3*restrict result, const vector3*restrict aa, const vector3*restrict v)
{
  vector3_displace(result,
      -2 * vector3_y(aa)*vector3_x(v) +     vector3_x(aa)*vector3_y(v),
           vector3_x(aa)*vector3_x(v) +     vector3_z(aa)*vector3_z(v),
           vector3_z(aa)*vector3_y(v) - 2 * vector3_y(aa)*vector3_z(v));
}

static inline void deriv_double_cross_z(vector3*restrict result, const vector3*restrict aa, const vector3*restrict v)
{
  vector3_displace(result,
      -2 * vector3_z(aa)*vector3_x(v) + vector3_x(aa)*vector3_z(v),
      -2 * vector3_z(aa)*vector3_y(v) + vector3_y(aa)*vector3_z(v),
           vector3_x(aa)*vector3_x(v) + vector3_y(aa)*vector3_y(v));
}

static inline void d2_double_cross_i2(vector3*restrict result, int i, const vector3*restrict v)
{
  vector3_set_scaled(result,-2.0,v);
  VECTOR3_ELT(result, i) = 0.0;
}

static inline void d2_double_cross_ij(vector3*restrict result, int i, int j, const vector3*restrict v)
{
  vector3_zero(result);
  VECTOR3_ELT(result, i) = VECTOR3_ELT(v, j);
  VECTOR3_ELT(result, j) = VECTOR3_ELT(v, i);
}


void dq_wj(size_t n, vector3*restrict dq_w1, vector3*restrict dq_w2, vector3*restrict dq_w3, 
    VecReal ang, const vector3*restrict axis, const vector3*restrict dv)
{
  vector3 tmp,tmp2;

  const double NEG_THIRD = -1.0/3.0;
  const double NEG_TWELTH = -1.0/12.0;

  if (ang > SMALL)
  {
    VecReal inv_ang = 1.0/ang;

    VecReal cos_ang = cos(ang);
    VecReal sin_ang = sin(ang);

    VecReal sin_ratio = sin_ang*inv_ang;
    VecReal cos_ratio = (1.0-cos_ang)*inv_ang;

    double aa_x = vector3_x(axis);
    double aa_y = vector3_y(axis);
    double aa_z = vector3_z(axis);

    VecReal coeff1 = cos_ang - sin_ratio;
    VecReal coeff2 = sin_ang - 2.0*cos_ratio;

    for (size_t i=0; i < n; i++)
    {
      vector3_cross(&tmp, axis, dv+i);
      vector3_cross(&tmp2, axis, &tmp);

      vector3_scale_by(&tmp, coeff1);
      vector3_set_scaled(dq_w1+i, aa_x, &tmp);
      vector3_set_scaled(dq_w2+i, aa_y, &tmp);
      vector3_set_scaled(dq_w3+i, aa_z, &tmp);

      vector3_scale_by(&tmp2, coeff2);
      vector3_add_multiple_to(dq_w1+i, aa_x, &tmp2);
      vector3_add_multiple_to(dq_w2+i, aa_y, &tmp2);
      vector3_add_multiple_to(dq_w3+i, aa_z, &tmp2);

      vector3_zero(&tmp); deriv_cross_x(&tmp, dv+i); vector3_add_multiple_to(dq_w1+i, sin_ratio, &tmp);
      vector3_zero(&tmp); deriv_cross_y(&tmp, dv+i); vector3_add_multiple_to(dq_w2+i, sin_ratio, &tmp);
      vector3_zero(&tmp); deriv_cross_z(&tmp, dv+i); vector3_add_multiple_to(dq_w3+i, sin_ratio, &tmp);

      vector3_zero(&tmp); deriv_double_cross_x(&tmp, axis, dv+i); vector3_add_multiple_to(dq_w1+i, cos_ratio, &tmp);
      vector3_zero(&tmp); deriv_double_cross_y(&tmp, axis, dv+i); vector3_add_multiple_to(dq_w2+i, cos_ratio, &tmp);
      vector3_zero(&tmp); deriv_double_cross_z(&tmp, axis, dv+i); vector3_add_multiple_to(dq_w3+i, cos_ratio, &tmp);
    }
  }
  else
  {
    vector3 axis_ang;
    vector3_set_scaled(&axis_ang, ang, axis);

    double aa_x = vector3_x(&axis_ang);
    double aa_y = vector3_y(&axis_ang);
    double aa_z = vector3_z(&axis_ang);

    double coeff1 = 1.0-ang*ang/6.0;
    double coeff2 = 0.5*(1.0-ang*ang/12.0);

    for (size_t i=0; i < n; i++)
    {
      vector3_cross(&tmp, &axis_ang, dv+i);
      vector3_cross(&tmp2, &axis_ang, &tmp);

      // vector3_cross(&tmp, axis, dv);
      // vector3_cross(&tmp2, axis, &tmp);

      // vector3_scaleby(&tmp, ang);
      // vector3_scaleby(&tmp2, ang*ang*ang);

      vector3_scale_by(&tmp, NEG_THIRD);
      vector3_set_scaled(dq_w1+i, aa_x, &tmp);
      vector3_set_scaled(dq_w2+i, aa_y, &tmp);
      vector3_set_scaled(dq_w3+i, aa_z, &tmp);

      vector3_add_multiple_to(dq_w1+i, NEG_TWELTH*aa_x, &tmp2);
      vector3_add_multiple_to(dq_w2+i, NEG_TWELTH*aa_y, &tmp2);
      vector3_add_multiple_to(dq_w3+i, NEG_TWELTH*aa_z, &tmp2);

      vector3_zero(&tmp); deriv_cross_x(&tmp, dv+i); vector3_add_multiple_to(dq_w1+i, coeff1, &tmp);
      vector3_zero(&tmp); deriv_cross_y(&tmp, dv+i); vector3_add_multiple_to(dq_w2+i, coeff1, &tmp);
      vector3_zero(&tmp); deriv_cross_z(&tmp, dv+i); vector3_add_multiple_to(dq_w3+i, coeff1, &tmp);

      vector3_zero(&tmp); deriv_double_cross_x(&tmp, &axis_ang, dv+i); vector3_add_multiple_to(dq_w1+i, coeff2, &tmp);
      vector3_zero(&tmp); deriv_double_cross_y(&tmp, &axis_ang, dv+i); vector3_add_multiple_to(dq_w2+i, coeff2, &tmp);
      vector3_zero(&tmp); deriv_double_cross_z(&tmp, &axis_ang, dv+i); vector3_add_multiple_to(dq_w3+i, coeff2, &tmp);
    }
  }
}


void d2q_dwi_dwj(vector3 d2q[D2Q_SIZE],
    const vector3*restrict axis_angle, 
    const vector3*restrict dv)
{
  vector3 axis;
  vector3 unit_axis;
  vector3_set(&axis, axis_angle);
  VecReal ang = vector3_norm(&axis);

  VecReal f, df, df_ang_ratio, d2f;
  VecReal g, dg, dg_ang_ratio, d2g;

  f = df = d2f = 0.0;
  g = dg = d2g = 0.0;

  if (fabs(ang) > SMALL)
  {
    VecReal inv_ang = 1.0/ang;

    VecReal sin_a = sin(ang);
    VecReal cos_a = cos(ang);

    f = inv_ang * sin_a;
    df = inv_ang*(cos_a - f);
    d2f = -f - 2.0*inv_ang*df;

    g = (1.0-cos_a)/(ang*ang);
    dg = inv_ang*(f-2.0*g);
    d2g = inv_ang*(df - 3.0*dg);

    df_ang_ratio = inv_ang*df;
    dg_ang_ratio = inv_ang*dg;

    vector3_set_scaled(&unit_axis, inv_ang, &axis);
  }
  else
  {
    VecReal a2 = ang*ang;
    f = 1.0 - a2/6.0;
    df = -ang/3.0;
    d2f = -1.0/3.0;

    g = 0.5 - a2/24.0;
    dg = -ang/12.0;
    d2g = -1.0/12.0;

    df_ang_ratio = d2f;
    dg_ang_ratio = d2g;

    vector3_set_with(&unit_axis, 1.0, 1.0, 1.0);
  }

  for (int a=0; a < DW_NUM_ELTS; a++) {
    for (int b=a; b < DW_NUM_ELTS; b++) {
      vector3_zero(d2q+symm_index(DW_NUM_ELTS,a,b));
    }
  }

  vector3 wXdv;
  vector3_zero(&wXdv);
  vector3_cross(&wXdv, &axis, dv);

  vector3 wXwXdv;
  vector3_zero(&wXwXdv);
  vector3_cross(&wXwXdv, &axis, &wXdv);

  vector3 d_wXdv[3];
  vector3_zero(d_wXdv+0);
  vector3_zero(d_wXdv+1);
  vector3_zero(d_wXdv+2);

  deriv_cross_x(d_wXdv+0, dv);
  deriv_cross_y(d_wXdv+1, dv);
  deriv_cross_z(d_wXdv+2, dv);
  

  vector3 d_wXwXdv[3];
  vector3_zero(d_wXwXdv+0);
  vector3_zero(d_wXwXdv+1);
  vector3_zero(d_wXwXdv+2);

  deriv_double_cross_x(d_wXwXdv+0, &axis, dv);
  deriv_double_cross_y(d_wXwXdv+1, &axis, dv);
  deriv_double_cross_z(d_wXwXdv+2, &axis, dv);


  /* Do the diagonal entries first */
  for (int a=0; a < DW_NUM_ELTS; a++) {
    size_t ind = symm_index(DW_NUM_ELTS,a,a);
    VecReal ua = VECTOR3_ELT(&unit_axis,a);
    VecReal ua2 = square( ua );

    vector3_add_multiple_to(
        &d2q[ind], 
        d2f*ua2 + df_ang_ratio*(1.0 - ua2),
        &wXdv);

    vector3_add_multiple_to(
        &d2q[ind],
        2.0*df*ua,
        &d_wXdv[a]);

    vector3_add_multiple_to(
        &d2q[ind],
        d2g*ua2 + dg_ang_ratio*(1.0 - ua2),
        &wXwXdv);

    vector3_add_multiple_to(
        &d2q[ind],
        2.0*dg*ua,
        &d_wXwXdv[a]);

    vector3 d2_wXwXdv;
    d2_double_cross_i2(&d2_wXwXdv, a, dv);

    vector3_add_multiple_to(&d2q[ind], g, &d2_wXwXdv);
  }

  /* Off-diagonal entries */
  for (int a=0; a < DW_NUM_ELTS; a++) {
    VecReal ua = VECTOR3_ELT(&unit_axis,a);
    for (int b=a+1; b < DW_NUM_ELTS; b++) {
      size_t ind = symm_index(DW_NUM_ELTS,a,b);
      VecReal ub = VECTOR3_ELT(&unit_axis,b);
      VecReal ab = ua*ub;

      vector3_add_multiple_to(
          &d2q[ind],
          ab*(d2f - df_ang_ratio),
          &wXdv);

      vector3_add_multiple_to(&d2q[ind], df*ub, &d_wXdv[a]);
      vector3_add_multiple_to(&d2q[ind], df*ua, &d_wXdv[b]);

      vector3_add_multiple_to(
          &d2q[ind],
          ab*(d2g - dg_ang_ratio),
          &wXwXdv);

      vector3_add_multiple_to(&d2q[ind], dg*ub, &d_wXwXdv[a]);
      vector3_add_multiple_to(&d2q[ind], dg*ua, &d_wXwXdv[b]);

      vector3 d2_wXwXdv;
      d2_double_cross_ij(&d2_wXwXdv, a,b, dv);

      vector3_add_multiple_to(&d2q[ind], g, &d2_wXwXdv);
    }
  }
}

void upper_sym_matmult(const int N,
    double*restrict y,
    const double*restrict mat,
    const double*restrict x, enum TS_MATRIXORDER order)
{
  int incr = 1;
  double alpha = 1.0;
  double beta  = 0.0;
  const char *uplo = (order == TS_ROWMAJOR) ? "U" : "L";

  dsymv( uplo, &N, &alpha, mat, &N, x, &incr, &beta, y, &incr );
}


// Rescales values so xnew[i] = (x[i] - mean(x))/std(x)
static void rescale_values(size_t n, double*restrict x, double mu[2])
{
  register double avg, var, davg;

  avg = 0.0;
  for (size_t i=0; i < n; i++)
  {
    avg += x[i];
  }
  avg /= n;

  davg = var = 0.0;
  for (size_t i=0; i < n; i++)
  {
    x[i] -= avg;
    var += x[i]*x[i];
    davg += x[i];
  }

  mu[0] = avg;
  var = (var - davg*davg/n)/(n-1);
  register double sd = sqrt(var);
  mu[1] = sd;

  register double inv_sd = 1.0/sd;
  for (size_t i=0; i < n; i++)
  {
    x[i] *= inv_sd;
  }
}

static double pow_int(double x, int n)
{
  double y = x;
  for(int k=1; k < n; k++) { y *= x; }
  return y;
}

static MAYBE_UNUSED void
printmat(const char* mname, int m, int n, double* M)
{
  fprintf(stderr, "%s = (%d x %d)\n", mname, (int)m,(int)n);
  for (int i=0; i < m; i++)
  {
    if (i==0)
      fprintf(stderr, "  [ ");
    else
      fprintf(stderr, "    ");

    if ((fabs(M[i+m*0]) > 1e-3) || (fabs(M[i+m*0]) < 1e-8))
      fprintf(stderr, "%7.4f", M[i+m*0]);
    else
      fprintf(stderr, "%7.4e", M[i+m*0]);

    for (int j=1; j < n; j++)
    {
      if ((fabs(M[i+m*j]) > 1e-3) || (fabs(M[i+m*j]) < 1e-8))
        fprintf(stderr, " %7.4f", M[i+m*j]);
      else
        fprintf(stderr, " %7.4e", M[i+m*j]);
    }
    if (i < m-1)
      fprintf(stderr, "\n\n");
    else
      fprintf(stderr, " ]\n\n");
  }
}

void ts_poly_init(ts_poly* P, int ndeg, double coeff[])
{
  P->ncoeff = ndeg+1;
  memcpy(P->coeff, coeff, (P->ncoeff)*sizeof(coeff[0]));

  P->rescaled = 0;
  P->scales[0] = P->scales[1] = 0.0;

  P->covar = NULL;
}

double ts_poly_value(ts_poly* P, double x)
{
  return polyval(P->ncoeff-1, P->coeff, x);
}

void ts_poly_finalize(ts_poly* P)
{
  if (P->covar != NULL)
    ts_free(NO_ALLOCATOR, P->covar);
}

TSResult ts_poly_deriv(ts_poly*restrict dP, const ts_poly*restrict P)
{
  memset(dP, 0, sizeof(*dP));
  dP->ncoeff = P->ncoeff-1;
  polyder(P->ncoeff-1, P->coeff, dP->coeff);
  return TS_OK;
}

TSResult ts_poly_fit(ts_poly* P, int ndeg, int npts, double*restrict x, double*restrict y, int rescale)
{
  UNUSED_PARAM(rescale);

  size_t ncoeff = ndeg+1;
  double* dcoeff = ts_calloc(NO_ALLOCATOR, ncoeff*ncoeff, sizeof(*dcoeff));
  P->covar  = dcoeff;
  P->ncoeff = ndeg+1;
  double mu[2] = {0.0,0.0};

  RETURN_ON_ERROR(
      polyfit(npts, x,y, ndeg,
        P->coeff, P->covar, mu, 0));

  return TS_OK;
}

int polyder(int ndeg, const double*restrict coeff, double*restrict dcoeff)
{
  if (ndeg == 0)
  {
    dcoeff[0] = 0.0;
    return 0;
  }

  for (int i=0; i < ndeg; i++)
  {
    int pn = ndeg-i;
    dcoeff[i] = pn*coeff[i];
  }

  return ndeg-1;
}

double polyval(int ndeg, const double* coeff, double x)
{
  register double val = 0.0;
  for(register int i=0; i < ndeg; i++)
  {
    val += coeff[i];
    val *= x;
  }

  val += coeff[ndeg];

  return val;
}

TSResult polyfit(int n, double*restrict x, double*restrict y, int ndeg, double*restrict coeff, double*restrict dcoeff, double mu[2], int rescale)
{
  int ncol = ndeg + 1;

  if (ndeg < 1)
  {
    return TS_NEW_ERROR3( TS_ERR_INVALID_ARGUMENT,
        "fit_polynomial: ndeg < 1" );
  }

  if (n < ncol)
  {
    return TS_NEW_ERROR3( TS_ERR_INVALID_ARGUMENT,
        "fit_polynomial: n < ndeg+1" );
  }

  mu[0] = mu[1] = 0.0;
  if (rescale)
  {
    rescale_values(n, x, mu);
  }

  double*restrict A = ts_alloc(NO_ALLOCATOR, n*(ndeg+1)*sizeof(*A));
  for (int deg=0; deg < ndeg; deg++)
  {
    for (int i=0; i < n; i++)
    {
      A[i+n*deg] = pow_int(x[i], ndeg-deg);
    }
  }

  for (int i=0; i < n; i++)
  {
    A[i+n*ndeg] = 1.0;
  }

  /*
  printmat("x", 1, n, x);
  printmat("y", 1, n, y);
  printmat("A", n, ncol, A);
  */

  // compute SVD of A

  // A = U * W * V**T
  //
  // where:
  //   A is M x N,
  //   U is M x M, orthogonal
  //   W is N x N, diagonal
  //   V is N x N, orthogonal

  double*restrict VT = ts_alloc(NO_ALLOCATOR, ncol*ncol*sizeof(*VT));
  double*restrict W  = ts_alloc(NO_ALLOCATOR, ncol*sizeof(*W));

  int lda = n;
  int ldu = 1;
  int ldvt = ncol;
  int info = 0;

  double wtmp = 0;
  int lwork = -1;

  dgesvd(
      "O",    // JOBU:  "O" -- overwrite A with first M cols of U
      "A",    // JOBVT: "A" -- all rows of VT returned in VT
      &n,     // M
      &ncol,  // N
      A,      // A: MxN matrix
      &lda,   // LDA
      W,      // singular value array
      NULL,   // U matrix: not referenced when JOBU == "O"
      &ldu,   // leading dimension of U
      VT,     // VT: NxN orthogonal matrix V**T
      &ldvt,  // LDVT: leading dimension of VT
      &wtmp,  // output contains optimal size of work array
      &lwork, // -1 : query optimal size of work array
      &info); // info: result of SVD operation
  if (info != 0)
    return TS_NEW_ERROR2( TS_ERR_IN_ALGORITHM,
        ("SVD returned error %d", info) );

  lwork = wtmp;
  double* work = ts_alloc(NO_ALLOCATOR, lwork*sizeof(*work));

  dgesvd(
      "O",    // JOBU:  "O" -- overwrite A with first M cols of U
      "A",    // JOBVT: "A" -- all rows of VT returned in VT
      &n,     // M
      &ncol,  // N
      A,      // A: MxN matrix
      &lda,   // LDA
      W,      // singular value array
      NULL,   // U matrix: not referenced when JOBU == "O"
      &ldu,   // leading dimension of U
      VT,     // VT: NxN orthogonal matrix V**T
      &ldvt,  // LDVT: leading dimension of VT
      work,   // output contains optimal size of work array
      &lwork, // -1 : query optimal size of work array
      &info); // info: result of SVD operation
  if (info != 0)
    return TS_NEW_ERROR2( TS_ERR_IN_ALGORITHM,
        ("SVD returned error %d", info) );

  /*
  printmat("U", n, ncol, A);
  printmat("W", 1, ncol, W);
  printmat("VT", ncol, ncol, VT);
  */

  memset(coeff , 0, ncol*sizeof(coeff[0]));
  memset(dcoeff, 0, ncol*ncol*sizeof(coeff[0]));

  for (int i=0; i < ncol; i++)
  {
    if (fabs(W[i]) >= POLYFIT_TOL)
      W[i] = 1.0/W[i];
    else
      W[i] = 0.0;
  }

  for (int i=0; i < ncol; i++)
  {
    int inc = 1;
    double Ub = ddot(&n, A+n*i, &inc, y, &inc)*W[i];
    for (int j=0; j < ncol; j++)
    {
      coeff[j] += Ub*VT[i+ncol*j];
    }
  }

  for (int i=0; i < ncol; i++)
    W[i] = W[i] * W[i];

  for(int j=0; j < ncol; j++)
  {
    for (int k=j; k < ncol; k++)
    {
      for (int i=0; i < ncol; i++)
      {
        dcoeff[k+ncol*j] += 
          W[i] * VT[i+ncol*j] * VT[i+ncol*k];
      }
    }
  }

  /*
  for (int j=0; j < ncol; j++)
  {
    for (int i=0; i < ncol; i++)
    {
      double ratio = VT[i+ncol*j] / W[i];
      dcoeff[j] += ratio*ratio;
    }
  }
  */

  // estimate parameter errors using chi-sq approach
  double chisq = 0.0;
  for (int i=0; i < n; i++)
  {
    double resid = y[i] - coeff[ndeg];
    double xval = x[i];
    resid -= coeff[ndeg-1]*xval;
    for (int j=1; j < ndeg; j++)
    {
      xval *= x[i];
      resid -= coeff[ndeg-j-1]*xval;
    }

    chisq += resid*resid;
  }

  /*
  printmat("coeff", 1, ncol, coeff);
  printmat("dcoeff(orig)", 1, ncol, dcoeff);
  fprintf(stderr, "ndof = %d / chisq == %12.7e\n", n-ncol, chisq);
  */

  // chisq = sqrt(chisq / (n-ncol));
  chisq /= n-ncol;
  // fprintf(stderr, "sqrt(chisq/ndof) == %12.7e\n", chisq);
  for (int j=0; j < ncol; j++)
  {
    size_t col_off = j*ncol;
    dcoeff[j+col_off] *= chisq;
    for (int i=j+1; i < ncol; i++)
    {
      double elt = dcoeff[i+col_off] * chisq;
      dcoeff[i+col_off] = dcoeff[j+i*ncol] = elt;
    }
  }

  ts_free(NO_ALLOCATOR, work);
  ts_free(NO_ALLOCATOR, W);
  ts_free(NO_ALLOCATOR, VT);
  ts_free(NO_ALLOCATOR, A);

  return TS_OK;
}

