#ifndef TS_NUMERICAL_H
#define TS_NUMERICAL_H

#include <math.h>

#include "ts_common.h"
#include "ts_errors.h"
#include "ts_machdeps.h"

#include "fortran_mangle.h"

/* Set IEEE 754 FP rounding mode -- useful for numerical debugging 
 * 
 * We define our own enumeration to allow cross-platform portability.
 * In particular, MSVC does not provide C99 but provides its own
 * non-standard way to set the FP rounding mode.
 */
enum TS_ROUNDING_MODES {
  TS_ROUND_INVALID  = 0, // signals library returned bad value
  TS_ROUND_NEAREST  = 1, // IEEE round to nearest
  TS_ROUND_ZERO     = 2, // IEEE round to zero
  TS_ROUND_UPWARD   = 3, // IEEE round to +infinity
  TS_ROUND_DOWNWARD = 4, // IEEE round to -infinity
};

enum TS_ROUNDING_MODES ts_get_rounding(void);
TSResult ts_set_rounding(enum TS_ROUNDING_MODES rounding);
void ts_set_rounding_default(void);


/** Inline functions to handle packed symmetric matrix storage */

/* see xlinux.nist.gov/dads/HTML/upperTriangularMatrix.html */
#define SYMM_SIZE(n) ( (n)*(n+1)/2 )
static inline size_t symm_size(size_t n) { return SYMM_SIZE(n); }
static inline size_t symm_index(size_t n, size_t i, size_t j)
{
  return i > j 
    ? (n*j + i - j*(j+1)/2) 
    : (n*i + j - i*(i+1)/2);
}

typedef double (*single_var_func)(void* closure, double x);
typedef double (*double_var_func)(void* closure, double x, double y);

TSResult deriv_ridders(single_var_func func, void* closure, double x, double h0, 
    double* ansp, double* errp);

TSResult deriv2_ridders(single_var_func func, void* closure, double x, double h0, 
    double* ansp, double* errp);

TSResult deriv_xy_ridders(double_var_func func, void* closure, double x,
    double y, double h0, double* ansp, double* errp);

enum {
  TS_POLY_MAX_DEG = 10,
  TS_POLY_MAX_COEFF = TS_POLY_MAX_DEG+1
};

typedef struct ts_poly ts_poly;
struct ts_poly
{
  int ncoeff;
  int rescaled;
  double scales[2];

  double coeff[TS_POLY_MAX_COEFF];
  double* covar;
};

// n -- number of data points
// x -- x values of data points
// y -- y values of data points
//
// ndeg   -- degree of polynomial (ndeg >= 1)
// coeff  -- array of ndeg+1 values for coefficients
// dcoeff -- array of ndeg+1 values for std errors of coefficients
//
// mu     -- shift and scale for x values:
//
//   X[i] = (x[i] - mu[0])/mu[1]
//
// NB: The x vectors are modified during the fit
void ts_poly_init(ts_poly* P, int ndeg, double coeff[]);
void ts_poly_finalize(ts_poly* P);

TSResult ts_poly_fit(ts_poly* P, int ndeg, int npts, double*restrict x, double*restrict y, int rescale);
TSResult ts_poly_deriv(ts_poly*restrict dP, const ts_poly*restrict P);
double ts_poly_value(ts_poly* P, double x);

#define POLYFIT_TOL 1e-6
TSResult polyfit(int n, double*restrict x, double*restrict y, int ndeg, double*restrict coeff, double*restrict dcoeff, double mu[2], int rescale);
int polyder(int ndeg, const double*restrict coeff, double*restrict dcoeff);
double polyval(int ndeg, const double* coeff, double x);

// For rigid body sites defined by: q_k = r + R(w) . v_k,
// where:
//   q_k is the location of the site
//   r is the CoM of the rigid body
//   R(w) is the rotation operator for axis-angle vector w
//   v_k is the vector from the CoM to the site in a reference orientation

// first derivatives of q_k wrt w
void dq_wj(size_t n, vector3*restrict dq_w1, vector3*restrict dq_w2, vector3*restrict dq_w3, 
    VecReal angle, const vector3*restrict axis, const vector3*restrict dv);

#define DW_NUM_ELTS 3
#define D2Q_SIZE ( SYMM_SIZE(DW_NUM_ELTS) )
// second derivatives of q_k wrt w
void d2q_dwi_dwj(vector3 d2q[D2Q_SIZE],
    const vector3*restrict axis_angle, 
    const vector3*restrict dv);

// Returns M if nelts == M*(M+1)/2 where M is an integer, and zero
// otherwise
static inline int get_symmetric_size(int nelts)
{
  int two_n = 2*nelts;
  int m_lo = floor( sqrt(two_n) );
  return m_lo*(m_lo+1) == two_n ? m_lo : 0;
}

enum TS_MATRIXORDER { TS_ROWMAJOR = 1, TS_COLMAJOR = 2 };

// calculates y = M.x where M is a full matrix, stored in column-major
// format, and assumed to be symmetric.  Uses only the upper half for
// the multiplication.
void upper_sym_matmult(const int N,
    double*restrict y,
    const double*restrict M,
    const double*restrict x, enum TS_MATRIXORDER order);

#define _F77P(plist) 

/* QUADPACK routines*/
#define dqags   FC_GLOBAL( dqags , DQAGS  )

extern void dqags( _F77P((f,a,b,epsabs,epsrel,
        result,abserr,neval,ier,limit,lenw,last,
        iwork,work)) );

/* LAPACK routines */
#define dgeev	FC_GLOBAL( dgeev , DGEEV  )
#define dsyev	FC_GLOBAL( dsyev , DSYEV  )
#define dsytrd	FC_GLOBAL( dsytrd, DSYTRD )
#define dstegr	FC_GLOBAL( dstegr, DSTEGR )
#define dsptrd	FC_GLOBAL( dsptrd, DSPTRD )
#define dgesvd  FC_GLOBAL( dgesvd, DGESVD )
#define dpotrf  FC_GLOBAL( dpotrf, DPOTRF )
#define dgesv   FC_GLOBAL( dgesv , DGESV  )

// dgeev -- solve general eigensystem
extern void dgeev( _F77P((jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr,
      ldvr, work, lwork, info)) );

// SSYEV computes all eigenvalues and, optionally, eigenvectors of a
// real symmetric matrix A.
extern void dsyev(
    char* jobz,   // JOBZ    (input) CHARACTER*1
                  //         = 'N':  Compute eigenvalues only;
                  //         = 'V':  Compute eigenvalues and eigenvectors.
                  //
    char* uplo,   // UPLO    (input) CHARACTER*1
                  //         = 'U':  Upper triangle of A is stored;
                  //         = 'L':  Lower triangle of A is stored.
                  //
    int*  n,      // N       (input) INTEGER
                  //         The order of the matrix A.  N >= 0.
                  //
    double* A,    // A       (input/output) REAL array, dimension (LDA, N)
                  //         On entry, the symmetric matrix A.  If UPLO = 'U', the
                  //         leading N-by-N upper triangular part of A contains the
                  //         upper triangular part of the matrix A.  If UPLO = 'L',
                  //         the leading N-by-N lower triangular part of A contains
                  //         the lower triangular part of the matrix A.
                  //         On exit, if JOBZ = 'V', then if INFO = 0, A contains the
                  //         orthonormal eigenvectors of the matrix A.
                  //         If JOBZ = 'N', then on exit the lower triangle (if UPLO='L')
                  //         or the upper triangle (if UPLO='U') of A, including the
                  //         diagonal, is destroyed.
                  //
    int* lda,     // LDA     (input) INTEGER
                  //         The leading dimension of the array A.  LDA >= max(1,N).
                  //
    double* W,    // W       (output) REAL array, dimension (N)
                  //         If INFO = 0, the eigenvalues in ascending order.
                  //
    double* work, // WORK    (workspace/output) REAL array, dimension (MAX(1,LWORK))
                  //         On exit, if INFO = 0, WORK(1) returns the optimal LWORK.
                  //
    int* lwork,   // LWORK   (input) INTEGER
                  //         The length of the array WORK.  LWORK >= max(1,3*N-1).
                  //         For optimal efficiency, LWORK >= (NB+2)*N,
                  //         where NB is the blocksize for SSYTRD returned by ILAENV.
                  //
                  //         If LWORK = -1, then a workspace query is assumed; the routine
                  //         only calculates the optimal size of the WORK array, returns
                  //         this value as the first entry of the WORK array, and no error
                  //         message related to LWORK is issued by XERBLA.
                  //
    int* info);   // INFO    (output) INTEGER
                  //         = 0:  successful exit
                  //         < 0:  if INFO = -i, the i-th argument had an illegal value
                  //         > 0:  if INFO = i, the algorithm failed to converge; i
                  //               off-diagonal elements of an intermediate tridiagonal
                  //               form did not converge to zero.

// dsytrd -- reduce dense symmetric to tridiagonal form
extern void dsytrd( _F77P((uplo, n, a, lda, d, e, tau, work, lwork, info)) );
// dsptrd -- reduce packed symmetric matrix to tridiagonal form
extern void dsptrd( _F77P((uplo, n, ap, d, e, tau, info)) );
// dstegr -- solve eigenvalues/vectors of tridiagonal matrix
extern void dstegr( _F77P((jobz, range, n, d, e, vl, vu, il, iu,
     abstol, m, w, z, ldz, isuppz, work, lwork, iwork,
     liwork, info)) );

// dgesv -- solve systems of equations for general matrix
extern void dgesv( _F77P((n, nrhs, a, lda, ipiv, b, ldb, info)) );

// dgesvd -- compute SVD of matrix
extern void dgesvd( _F77P((jobu, jobvt, m, n, A, 
        lda, s, U, ldu, VT, ldvt, work, lwork, info)) );

extern void dpotrf( _F77P((uplo, n, A, lda, info)) );


/* BLAS routines */
#define ddot    FC_GLOBAL(ddot ,DDOT )
#define dspmv	FC_GLOBAL(dspmv,DSPMV)
#define dsymv	FC_GLOBAL(dsymv,DSYMV)
#define dgemv	FC_GLOBAL(dgemv,DGEMV)

// ddot -- BLAS routine for dot product
extern double ddot( _F77P((n,x,incx,y,incy)) );

// y <- alpha * A * x + beta*y   A is a packed symmetric matrix
extern void dspmv( _F77P(( uplo,n,alpha,ap,x,incx,beta,y,incy )) );
extern void dsymv( _F77P(( uplo,n,alpha,a,lda,x,incx,beta,y,incy )) );
extern void dgemv( _F77P((trans,m,n,alpha,a,lda,x,incx,beta,y,incy)) );

extern void dstemr_();
// extern void dstemr();

extern void dsbtrd_();
// extern void dsbtrd();

/* ARPACK routines */
#define dseupd	FC_GLOBAL(dseupd,DSEUPD)
#define dsaupd	FC_GLOBAL(dsaupd,DSAUPD)

extern void dsaupd();
/*
extern void dsaupd( int* ido, const char* bmat, int* n, const char* which,
    int* nev, double* tol, double* resid, int* ncv, double* v,
    int* ldv, int* iparam, int* ipntr, double* workd, double* workl,
    int* lworkl, int* info, long bmat_len, long which_len );
    */

/*
extern void dsaupd( _F77P(( ido, bmat, n, which, nev, tol, resid, 
        ncv, v, ldv, iparam, ipntr, workd, workl, lworkl, info )) );
        */
extern void dseupd( _F77P(( rvec, howmny, select, d, z, ldz, 
        sigma, bmat, n, which, nev, tol, resid, ncv, v, ldv, 
        iparam, ipntr, workd, workl, lworkl, info )) );

#undef _F77P


/* double-double accumulators
 *
 * Use Knuth's trick to maintain an extra (approx) 53 bits of precision
 * when summing quantities.
 *
 */
typedef union {
  double elts[2];
  struct { double hi,lo; } parts;
#if defined(USE_AVX) || defined(USE_SSE)
  __m128d svec;
#endif
} dbldbl;

static inline dbldbl zero_dbldbl(void) { dbldbl zero = { .elts = { 0.0, 0.0 } }; return zero; }
static inline dbldbl set_dbldbl(double v) { dbldbl zero = { .elts = { v, 0.0 } }; return zero; }
static inline double dbldbl_to_double(dbldbl a) { return a.parts.hi; }
static inline double dbldbl_roundoff(dbldbl a) { return a.parts.lo; }

static inline dbldbl accum_dbldbl(dbldbl a, double b)
{
  double sa = a.parts.hi + b;
  double d1 = sa - a.parts.hi;

  double d2 = b - d1;
  double d3 = a.parts.hi - (sa - d1);

  double sb = a.parts.lo + d2 + d3;

  dbldbl result = { .parts = { .hi = sa+sb, .lo = sb - ((sa+sb) - sa) } };
  return result;
}

static inline int dbldbl_not_zero(dbldbl a) { return a.parts.hi != 0.0 || a.parts.lo != 0.0; }

#if TS_USE_AVX || TS_USE_SSE
static inline dbldbl sum_dbldbl(dbldbl a, dbldbl b)
{
  double sa = a.parts.hi + b.parts.hi;
  double d1 = sa - a.parts.hi;
  double d2 = b.parts.hi - d1;
  double d3 = sa - d1;
  d3 = a.parts.hi - d3;

  double sb = d2 + d3 + a.parts.lo + b.parts.lo;

  dbldbl result = { .elts = { sa+sb, sb - ((sa+sb)-sa) } };
  return result;
}
#else /* !TS_USE_AVX && !TS_USE_SSE */
static inline dbldbl sum_dbldbl(dbldbl a, dbldbl b)
{
  double sa = a.parts.hi + b.parts.hi;
  double d1 = sa - a.parts.hi;
  double d2 = b.parts.hi - d1;
  double d3 = sa - d1;
  d3 = a.parts.hi - d3;

  double sb = d2 + d3 + a.parts.lo + b.parts.lo;

  dbldbl result = { .elts = { sa+sb, sb - ((sa+sb)-sa) } };
  return result;
}
#endif

static inline double kahan_sum(const double *restrict vals, size_t n, double *restrict errp)
{
  size_t i;
  double sum,err;

  if (n==0) { return 0.0; }

  for (i=1, sum=vals[0], err = 0.0; i < n; i++)
  {
    double tmp, acc;

    tmp = vals[i] - err;
    acc = sum + tmp;
    err = (acc - sum) - tmp;
    sum = acc;
  }

  if (errp) { *errp = err; }
  return sum;
}

static inline double dotv(size_t n, const double*restrict u, const double*restrict v)
{
#if TS_USE_AVX || TS_USE_SSE
  // assert that u and v are aligned on 16-byte boundaries
  tsAssert( IS_SSE_ALIGNED( u ) );
  tsAssert( IS_SSE_ALIGNED( v ) );

  __m128d acc = _mm_setzero_pd();
  if (n>1) {
    for(size_t i=0; i+1 < n; i+=2) {
      __m128d a = _mm_load_pd(u+i);
      __m128d b = _mm_load_pd(v+i);

      __m128d c = _mm_mul_pd(a,b);
      acc = _mm_add_pd(acc, c);
    }
  }

  if (n%2) {
    __m128d a = _mm_load_sd(u+n-1);
    __m128d b = _mm_load_sd(v+n-1);
    __m128d c = _mm_mul_pd(a,b);
    acc = _mm_add_pd(acc,c);
  }

  __m128d dv = _mm_hadd_pd(acc, acc);
  return _mm_cvtsd_f64(dv);

#else /* ! TS_USE_AVX && ! TS_USE_SSE2 */

  double acc = 0.0;
  for (size_t i=0; i < n; i++) {
    acc += u[i]*v[i];
  }
  return acc;

#endif
}

#if defined(TS_USE_AVX)
static inline __m128d normv_avx(size_t n, const double*restrict u)
{
  tsAssert( IS_AVX_ALIGNED( u ) );

  __m256d acc = _mm256_setzero_pd();
  for(size_t i=0; i < n; i+=4) {
    __m256d a = _mm256_load_pd(u+i);
    __m256d c = _mm256_mul_pd(a,a);
    acc = _mm256_add_pd(acc, c);
  }

  __m128d ans = _mm_add_pd(
      _mm256_extractf128_pd(acc,0),
      _mm256_extractf128_pd(acc,1) );

  int rem = n&3;
  if (rem&2) {
    __m128d xtra = _mm_load_pd(u+n-rem);
    __m128d xsq = _mm_mul_pd(xtra,xtra);
    ans = _mm_add_pd(ans,xsq);
  }

  if (rem&1) {
    __m128d xtra = _mm_load_sd(u+n-1);
    __m128d xsq = _mm_mul_pd(xtra,xtra);
    ans = _mm_add_pd(ans,xsq);
  }

  __m128d vsq = _mm_hadd_pd(ans, ans);
  return _mm_sqrt_pd(vsq);
}
#endif

#if defined(TS_USE_AVX) || defined(TS_USE_SSE)
static inline __m128d normv_sse(size_t n, const double*restrict u)
{
  tsAssert( IS_SSE_ALIGNED( u ) );

  __m128d acc = _mm_setzero_pd();
  for(size_t i=0; i < n-1; i+=2) {
    __m128d a = _mm_load_pd(u+i);
    __m128d c = _mm_mul_pd(a,a);
    acc = _mm_add_pd(acc, c);
  }

  if (n%2) {
    __m128d xtra = _mm_load_sd(u+n-1);
    __m128d xsq = _mm_mul_pd(xtra,xtra);
    acc = _mm_add_pd(acc,xsq);
  }

  __m128d vsq = _mm_hadd_pd(acc, acc);
  return _mm_sqrt_pd(vsq);
}
#endif

static inline double normv(size_t n, const double*restrict u)
{
#if defined(TS_USE_AVX) || defined(TS_USE_SSE)

  __m128d vn = normv_sse(n,u);
  return _mm_cvtsd_f64(vn);

#else /* ! TS_USE_AVX && ! TS_USE_SSE2 */

  double acc = 0.0;
  for (size_t i=0; i < n; i++) {
    acc += u[i]*u[i];
  }
  return sqrt(acc);

#endif
}

static inline double normalizev(size_t n, VecReal*restrict v, double tol)
{
#if defined(TS_USE_AVX)

  __m128d vn = normv_avx(n,v);

  if (_mm_cvtsd_f64(vn) > tol) {
    __m128d one = _mm_set1_pd(1.0);
    __m128d hinv = _mm_div_pd(one,vn);

    __m256d inv = ts_mm256_set_m128d(hinv,hinv);

    for (size_t i=0; i < n; i+=4) {
      __m256d a = _mm256_load_pd(v+i);
      __m256d b = _mm256_mul_pd(a,inv);
      _mm256_store_pd(v+i,b);
    }

    if (n%2) {
      v[n-1] *= _mm_cvtsd_f64(hinv);
    }
  }

  return _mm_cvtsd_f64(vn);

#elif defined(TS_USE_SSE)

  __m128d vn = normv_sse(n,v);

  if (_mm_cvtsd_f64(vn) > tol) {
    __m128d one = _mm_set1_pd(1.0);
    __m128d inv = _mm_div_pd(one,vn);

    for (size_t i=0; i+1 < n; i+=2) {
      __m128d a = _mm_load_pd(v+i);
      __m128d b = _mm_mul_pd(a,inv);
      _mm_store_pd(v+i,b);
    }

    if (n%2) {
      v[n-1] *= _mm_cvtsd_f64(inv);
    }
  }

  return _mm_cvtsd_f64(vn);

#else
  VecReal norm = normv(n,v);
  if (norm > tol) {
    register VecReal inv_norm = 1.0/norm;
    for(size_t i=0; i < n; i++) { v[i] *= inv_norm; }
  }
  return norm;
#endif
}

static inline void subv(size_t n, double*restrict w, const double*restrict u, const double*restrict v)
{
#if defined(TS_USE_AVX) || defined(TS_USE_SSE)

  // assert that u and v are aligned on 16-byte boundaries
  tsAssert( IS_SSE_ALIGNED( u ) );
  tsAssert( IS_SSE_ALIGNED( v ) );
  tsAssert( IS_SSE_ALIGNED( w ) );

  for(size_t i=0; i < n-1; i+=2) {
    __m128d a = _mm_load_pd(u+i);
    __m128d b = _mm_load_pd(v+i);
    __m128d c = _mm_sub_pd(a,b);

    _mm_store_pd(w+i,c);
  }

  if (n&1) {
    w[n-1] = u[n-1] - v[n-1];
  }

#else /* ! TS_USE_AVX && ! TS_USE_SSE2 */

  for (size_t i=0; i < n; i++) {
    w[i] = u[i] - v[i];
  }

#endif
}

#endif /* TS_NUMERICAL_H */

