#include "ts_parameters.h"

#include "ts_common.h"
#include "ts_tables_priv.h"

/* Simulation parameter hashtable functions 
 *
 * struct ts_param_table is really an alias for khash_t(sim_param) to
 * keep the visibility of khash.h private
 */

DEFINE_TABLE(sim_param, param, ts_param_table, double);
#define HASH_PARAM_TABLE(tbl) ((khash_t(sim_param)*)(tbl))
#define PARAM_FOREACH(tbl, kvar, vvar, code) \
  kh_foreach(HASH_PARAM_TABLE(tbl), kvar, vvar, code)

ts_param_table* ts_param_table_new() { return param_table_new(); }
void ts_param_table_delete(ts_param_table* tbl) { param_table_delete(tbl); }

int ts_param_table_exists(ts_param_table* tbl, const char* key)
{
  return param_table_exists(tbl, key);
}

TSResult ts_param_table_get(ts_param_table* tbl, const char* key, double* val_ptr)
{
  if (param_table_get(tbl, key, val_ptr))
  {
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR( TS_ERR_NO_SUCH_PARAMETER );
  }
}

TSResult ts_param_table_set(ts_param_table* tbl, const char* key, double val)
{
  char* kdup = ts_strdup(NO_ALLOCATOR, key);
  if (kdup == NULL)
  {
    return TS_NEW_ERROR3(
        TS_ERR_OUT_OF_MEMORY,
        "out of memory duplicating parameter string" );
  }

  if (!param_table_set(tbl, kdup, val))
  {
    ts_free(NO_ALLOCATOR, kdup);
    return TS_NEW_ERROR( TS_ERR_ADDING_PARAMETER );
  }

  return TS_OK;
}

TSResult ts_param_table_foreach(ts_param_table* tbl, ts_parameter_func func, void* data)
{
  const char* key;
  double val;
  PARAM_FOREACH(tbl, key, val,
      RETURN_ON_ERROR(func(data,key,val)));

  return TS_OK;
}

void ts_param_iter_init(ts_param_table* tbl, struct ts_param_iter* iter)
{
  UNUSED_PARAM(tbl);
  iter->state = 0;
  iter->key = NULL;
  iter->val = 0.0;
}

int ts_param_iter_next(ts_param_table* tbl, struct ts_param_iter* iter)
{
  khint_t _i = (khint_t)iter->state;
  khash_t(sim_param)* ht = HASH_PARAM_TABLE(tbl);
  khint_t _end = kh_end(ht);
  while (!kh_exist(ht, _i) && (_i < _end)) { _i++; }
  if (_i == _end) { return 0; }

  iter->key = kh_key(ht, _i);
  iter->val = kh_val(ht, _i);
  iter->state = (int64_t)(_i+1);
  return 1;
}

