#include "ts_particle.h"

#include "ts_common.h"
#include "ts_malloc_protect.h"

void ts_particle_init(ts_particle* particle)
{
  particle->type = NULL;
  particle->flags = PARTICLE_FLAG_NONE;

  particle->max_bonds = DEFAULT_MAX_BONDS;
  particle->bonds = ts_calloc(NO_ALLOCATOR, particle->max_bonds, sizeof(particle->bonds[0]));

  particle->num_bonds = 0;

  particle->rigid_body_index = TS_NULL_RIGID_BODY;
}

void ts_particle_finalize(ts_particle* particle)
{
  ts_free(NO_ALLOCATOR, particle->bonds);
}

void ts_rigid_body_init(ts_rigid_body* rigid_body)
{
  rigid_body->type_ptr = NULL;
}

void ts_rigid_body_finalize(ts_rigid_body* rigid_body)
{
  UNUSED_PARAM(rigid_body);
}

static struct ts_bond_info* find_bond_type(const ts_particle*restrict p0, const ts_particle*restrict p1)
{
  if (p0 == p1) return NULL;

  for(size_t i=0; i < p0->num_bonds; i++) {
    if ((p1 == p0->bonds[i].pi) || (p1 == p0->bonds[i].pj))
      return p0->bonds+i;
  }

  return NULL;
}

static TSResult expand_max_bonds(ts_particle* p, size_t new_max)
{
  if (p->max_bonds < new_max)
  {
    struct ts_bond_info* new_bonds = ts_realloc(NO_ALLOCATOR, p->bonds, new_max*sizeof(p->bonds[0]));
    ERROR_IF( TS_ERR_OUT_OF_MEMORY, new_bonds == NULL);

    p->bonds = new_bonds;
    p->max_bonds = new_max;
  }

  return TS_OK;
}

static inline void set_bond_info(struct ts_bond_info* info, ts_particle*restrict pi, ts_particle*restrict pj, ts_bond_type bondtype)
{
  info->pi = pi;
  info->pj = pj;
  info->type = bondtype;
}

TSResult ts_particle_add_bond(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype)
{
  if (find_bond_type(p0,p1) != NULL)
  {
    return TS_NEW_ERROR(TS_ERR_PARTICLES_ALREADY_BONDED);
  }

  RETURN_ON_ERROR( expand_max_bonds(p0,p0->num_bonds+1) );
  RETURN_ON_ERROR( expand_max_bonds(p1,p1->num_bonds+1) );

  tsAssert( p0->max_bonds >= p0->num_bonds+1 );
  tsAssert( p1->max_bonds >= p1->num_bonds+1 );

  if (p0 > p1) {
    ts_particle* tmp = p0;
    p0 = p1;
    p1 = tmp;
    bondtype = bondtype_reverse_direction(bondtype);
  }

  set_bond_info(p0->bonds + p0->num_bonds, p0,p1,bondtype); p0->num_bonds++;
  set_bond_info(p1->bonds + p1->num_bonds, p0,p1,bondtype); p1->num_bonds++;

  return TS_OK;
}

static inline TSResult remove_bond_info(ts_particle* pi, ts_particle* pj)
{
  struct ts_bond_info* info = find_bond_type(pi,pj);
  if (info == NULL) {
    return TS_NEW_ERROR(TS_ERR_PARTICLES_NOT_BONDED);
  }

  ptrdiff_t ind = info - pi->bonds;
  for (size_t i=ind+1; i < pi->num_bonds; i++) {
    pi->bonds[i-1] = pi->bonds[i];
  }
  pi->num_bonds--;

  return TS_OK;
}

TSResult ts_particle_remove_bond(ts_particle* p0, ts_particle* p1)
{
  ERROR_IF(TS_ERR_INVALID_ARGUMENT, p0 == p1);

  if (p0 > p1) {
    ts_particle* tmp = p0;
    p0 = p1;
    p1 = tmp;
  }

  RETURN_ON_ERROR( remove_bond_info(p0,p1) );
  RETURN_ON_ERROR( remove_bond_info(p1,p0) );

  return TS_OK;
}

static TSResult modify_bond_type_on_particle(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype)
{
  struct ts_bond_info* info = find_bond_type(p0,p1);

  ERROR_IF( TS_ERR_PARTICLES_NOT_BONDED, info == NULL );

  info->type = (p0 < p1) ? bondtype : bondtype_reverse_direction(bondtype);

  return TS_OK;
}

TSResult ts_particle_modify_bond(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype)
{
  RETURN_ON_ERROR( modify_bond_type_on_particle(p0,p1,bondtype) );
  RETURN_ON_ERROR( modify_bond_type_on_particle(p1,p0,bondtype_reverse_direction(bondtype)) );

  return TS_OK;
}

ts_bond_type ts_particle_get_bond_type(const ts_particle*restrict p0, const ts_particle*restrict p1)
{
  struct ts_bond_info* info = find_bond_type(p0,p1);

  if (info == NULL) return bondtype_none();
  if (info->pi == p0) return info->type;

  ts_bond_type rev = bondtype_reverse_direction(info->type);
  return rev;
}

int ts_particle_is_bonded_to(const ts_particle*restrict p0, const ts_particle*restrict p1)
{
  return bondtype_is_bond( ts_particle_get_bond_type(p0,p1) );
}

TSResult ts_particle_set_bond_type(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype)
{
  struct ts_bond_info* info0 = find_bond_type(p0,p1);
  struct ts_bond_info* info1 = find_bond_type(p1,p0);
  ERROR_IF(TS_ERR_PARTICLES_NOT_BONDED, info0 == NULL);
  ERROR_IF(TS_ERR_PARTICLES_NOT_BONDED, info1 == NULL);

  info0->type = bondtype;
  info1->type = bondtype;

  return TS_OK;
}
