#ifndef TS_PARTICLE_H
#define TS_PARTICLE_H

#include <stdint.h>
#include <stdlib.h>
#include "ts_common.h"
#include "ts_errors.h"
#include "vector3.h"

#define MODULE_NAME TS_PARTICLE

#define PARTICLE_NUM_DOF 3

#define RIGID_BODY_NUM_DOF 6

#define RIGID_BODY_NUM_ROT_DOF 3

#define RB_TRANSLATION_OFFSET 0

#define RB_ROTATIONAL_OFFSET 3

#define DEFAULT_MAX_BONDS 4

struct ts_sim;

typedef struct ts_particle_type ts_particle_type;
struct ts_particle_type
{
  struct ts_sim* sim;

  const char* name;
  const char* long_name;

  double mass;
  double charge;
  double radius;

  double collfreq;

  double inv_collfreq;
  double inv_mass;
  double inv_sqrt_mass;

  double brownian_coeff;
  double coll_coeff1;
  double coll_coeff2;

  double rand_coeff;
  double bd_rand_coeff;
};

enum PARTICLE_FLAGS {
  PARTICLE_FLAG_NONE      = 0,
  PARTICLE_FLAG_FROZEN    = BIT_VALUE(1),
  PARTICLE_FLAG_RIGID     = BIT_VALUE(2),
  PARTICLE_FLAG_RIGID_COM = BIT_VALUE(3),
};

struct ts_particle;

typedef union {
  uint16_t value;
} ts_bond_type;

#define BONDTYPE_RESERVED_OFFSET 1000
#define BONDTYPE_MINIMIZER_CONSTRAINT  (BONDTYPE_RESERVED_OFFSET+1)

#define BONDTYPE_FLAG_MASK  0xC000
#define BONDTYPE_VALUE_MASK 0x3FFF

#define BONDTYPE_UNDIRECTED_BIT 0x0000
#define BONDTYPE_BACK_BIT 0x4000
#define BONDTYPE_FORW_BIT 0x8000
#define BONDTYPE_BIDIREC_BIT (BONDTYPE_BACK_BIT | BONDTYPE_FORW_BIT)

#define BONDTYPE_MAX  0x3FFF

enum BONDTYPE_DIRECTION {
  BONDTYPE_UNDIRECTED    = 0,
  BONDTYPE_BACKWARD      = 1,
  BONDTYPE_FORWARD       = 2,
  BONDTYPE_BIDIRECTIONAL = 3,
};
#define BONDTYPE_DIREC_FLAGS(direc) ( (uint16_t)(direc) * BONDTYPE_BACK_BIT )

static inline uint16_t bondtype_to_int(ts_bond_type bt) { return bt.value; }
static inline ts_bond_type bondtype_from_int(uint16_t v) { ts_bond_type bt = {v}; return bt; }
static inline ts_bond_type bondtype_none(void) { ts_bond_type bt = {0}; return bt; }

static inline uint16_t bondtype_new_as_int(enum BONDTYPE_DIRECTION direc, uint16_t type) 
{
  uint16_t flags = BONDTYPE_DIREC_FLAGS(direc);

  tsAssert( ((flags & BONDTYPE_FLAG_MASK) == flags) || "bad flags value passed to bondtype_new" );
  tsAssert( ((type & BONDTYPE_VALUE_MASK) == type) || "bad type value passed to bondtype_new" );

  return flags | type;
}

static inline ts_bond_type bondtype_new(enum BONDTYPE_DIRECTION direc, uint16_t type) 
{
  ts_bond_type bt = { 0 };
  bt.value = bondtype_new_as_int(direc,type);
  return bt;
}

static inline uint16_t bondtype_get_type(ts_bond_type bt)  { return bt.value & BONDTYPE_VALUE_MASK; }
static inline uint16_t bondtype_get_flags(ts_bond_type bt) { return bt.value & BONDTYPE_FLAG_MASK; }

static inline int bondtype_is_bond(ts_bond_type bt) { return bondtype_get_type(bt) > 0; }

#define BT_IS_NONE(bt) ( (bt & BONDTYPE_VALUE_MASK) == 0 )
static inline int bondtype_is_none(ts_bond_type bt) { return bondtype_get_type(bt) == 0; }

static inline int bondtype_is_forw(ts_bond_type bt) { return (bt.value & BONDTYPE_FORW_BIT) == BONDTYPE_FORW_BIT; }
static inline int bondtype_is_back(ts_bond_type bt) { return (bt.value & BONDTYPE_BACK_BIT) == BONDTYPE_BACK_BIT; }

static inline enum BONDTYPE_DIRECTION bondtype_get_direction(ts_bond_type bt)
{
  return bondtype_is_none(bt) ? BONDTYPE_UNDIRECTED : ((enum BONDTYPE_DIRECTION) (bondtype_get_flags(bt) / BONDTYPE_BACK_BIT));
}

static inline int bondtype_is_omni(ts_bond_type bt) { 
  return (bt.value & (BONDTYPE_FORW_BIT | BONDTYPE_BACK_BIT)) == (BONDTYPE_FORW_BIT | BONDTYPE_BACK_BIT);
}

static inline int bondtype_is_undirected(ts_bond_type bt) { 
  return (bt.value & (BONDTYPE_FORW_BIT | BONDTYPE_BACK_BIT)) == 0;
}

#define BT_REVERSE(x) \
  ( ( (((x) & BONDTYPE_FLAG_MASK) == 0) || (((x) & BONDTYPE_FLAG_MASK) == BONDTYPE_FLAG_MASK) ) ? (x) : ((x) ^ BONDTYPE_FLAG_MASK) )
static inline ts_bond_type bondtype_reverse_direction(ts_bond_type bt) 
{
  ts_bond_type rev = { BT_REVERSE(bt.value) };
  return rev;
}

static inline uint16_t bondtype_reverse_direction_as_int(uint16_t bt)
{
  return BT_REVERSE(bt);
}

struct ts_bond_info
{
  ts_bond_type type;

  /* must maintain invariant: pi < pj */
  struct ts_particle* pi;
  struct ts_particle* pj;
};

#define TS_NULL_PARTICLE   ( ~((size_t)0) )

typedef struct ts_particle ts_particle;
struct ts_particle
{
  ts_particle_type* type;
  size_t rigid_body_index;

  size_t max_bonds;
  size_t num_bonds;
  struct ts_bond_info* bonds;

  uint32_t flags;
};

void ts_particle_init(ts_particle* particle);
void ts_particle_finalize(ts_particle* particle);

TSResult ts_particle_add_bond(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype);
TSResult ts_particle_remove_bond(ts_particle* p0, ts_particle* p1);
TSResult ts_particle_modify_bond(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype);
int ts_particle_is_bonded_to(const ts_particle*restrict p0, const ts_particle*restrict p1);
ts_bond_type ts_particle_get_bond_type(const ts_particle*restrict p0, const ts_particle*restrict p1);

static inline void ts_particle_set_frozen(ts_particle* p, int frozen)
{
  if (frozen)
    p->flags |= PARTICLE_FLAG_FROZEN;
  else
    p->flags = (p->flags & ~PARTICLE_FLAG_FROZEN);
}

static inline int ts_particle_is_frozen(const ts_particle* p)
{
  return (p->flags & PARTICLE_FLAG_FROZEN);
}

static inline int ts_particle_is_rigid(const ts_particle* p)
{
  return (p->flags & PARTICLE_FLAG_RIGID);
}

static inline int ts_particle_is_rigid_com(const ts_particle* p)
{
  return (p->flags & PARTICLE_FLAG_RIGID_COM);
}

static inline int ts_particle_in_rigid_body(const ts_particle* p)
{
  return (p->flags & (PARTICLE_FLAG_RIGID | PARTICLE_FLAG_RIGID_COM));
}

static inline int ts_particle_is_free(const ts_particle* p)
{
  return !ts_particle_is_frozen(p) && !ts_particle_in_rigid_body(p);
}

static inline ts_bond_type ts_particle_bond_type_at_index(const ts_particle*restrict p, size_t ind)
{
  return (ind < p->num_bonds) 
    ? (p == p->bonds[ind].pi ? p->bonds[ind].type : bondtype_reverse_direction(p->bonds[ind].type))
    : bondtype_none();
}

TSResult ts_particle_set_bond_type(ts_particle* p0, ts_particle* p1, ts_bond_type bondtype);

enum TS_RIGID_BODY_TYPE { 
  TS_RIGID_NOT_DEFINED = 0, 
  TS_RIGID_SITE_INTERACTION,

  TS_RIGID_MAX_TYPE
};

struct ts_rigid_body_type
{
  const char* name;

  enum TS_RIGID_BODY_TYPE type;

  double I0;
  double rotfric;

  double inv_rotfric;
  double inv_I0;
  double inv_sqrt_I0;
  double bd_rand_coeff;

  size_t num_sites;
  vector3* ref_vectors;
};

#define TS_NULL_RIGID_BODY ( ~((size_t)0) )

enum RIGID_FLAGS {
  RIGID_FLAG_NONE       = 0,
  RIGID_FLAG_FROZEN_POS = BIT_VALUE(1),
  RIGID_FLAG_FROZEN_ORI = BIT_VALUE(2),
  RIGID_FLAG_FROZEN_ALL = BIT_VALUE(1) | BIT_VALUE(2),
};

typedef struct ts_rigid_body ts_rigid_body;
struct ts_rigid_body
{
  struct ts_rigid_body_type* type_ptr;
  uint32_t flags;

  vector3 torque;
  vector3 prev_torque;
  vector3 rand_torque;

  size_t com_particle;
  size_t* site_indices;
};

void ts_rigid_body_init(ts_rigid_body* rigid_body);
void ts_rigid_body_finalize(ts_rigid_body* rigid_body);

#undef MODULE_NAME

#endif /* TS_PARTICLE_H */

