#include "ts_potentials.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ts_common.h"
#include "ts_units.h"
#include "ts_errors.h"
#include "ts_malloc_protect.h"
#include "ts_dispatch.h"
#include "tubesheet.h"
#include "ts_numerical.h"

#include "ts_bonds.h"

#if TS_USE_SSE
#include <immintrin.h>
#endif /* TS_USE_SSE */

#undef USE_TS_FASTMATH

#if defined(USE_TS_FASTMATH)
#  include "ts_fastmath.h"
#endif /* defined(USE_TS_FASTMATH) */

#define FORCE_EPS 1e-12

#define HESSIAN_EPS 1e-8

static Potential* new_harmonic(void) { return (Potential*)tsp_harmonic_new(0.0, 0.0); }
static Potential* new_wca(void)      { return (Potential*)tsp_wca_new(0.0, 0.0); }
static Potential* new_angle_harmonic(void) { return (Potential*)tsp_angle_harmonic_new(0.0, 0.0); }
static Potential* new_dihedral(void) { return (Potential*)tsp_dihedral_new(DIHEDRAL_TORSIONAL,0.0, 0.0); }
static Potential* new_dipole(void)   { return (Potential*)tsp_dipole_new(0.0, 0.0); }

/* constants that make the code more readable */
enum {
  ONE_BODY   = 1,
  TWO_BODY   = 2,
  THREE_BODY = 3,
  FOUR_BODY  = 4,
};

struct named_potential_constructor
{
  const char* name; 
  Potential* (*func)(void);
};

static const struct named_potential_constructor
named_constructors[] = {
  { "harmonic"      , new_harmonic       },
  { "wca"           , new_wca            },
  { "angle_harmonic", new_angle_harmonic },
  { "dihedral"      , new_dihedral       },
  { "dipole"        , new_dipole         },
  { 0, 0 }
};

Potential* ts_potential_new(const char* name)
{
  /* For a small table, just do a linear search */
  for (int i=0; named_constructors[i].name != NULL; i++)
  {
    if (strcmp(name, named_constructors[i].name) == 0) {
      return named_constructors[i].func();
    }
  }

  return NULL;
}

void ts_potential_init(Potential* pot, const char* name, int nbodies)
{
  pot->name = name;
  pot->nbodies = nbodies;
  pot->num_refs = 1;

#if POTENTIAL_REF_LOGGING
  fprintf(stderr, "made  %p %d %s:%d @ %s\n",
      (void *)pot, pot->num_refs, __FILE__, __LINE__, __func__);
#endif
  
  ts_potential_incref(pot);

  pot->potential_energy = 0.0;

  pot->energy_func = NULL;
  pot->forces_func = NULL;
  pot->hessian_func = NULL;
  pot->multiple_forces_func = NULL;

  pot->get_params_func = NULL;
  pot->set_param_func = NULL;

  pot->delete_func = ts_potential_delete;
}

void ts_potential_delete(Potential *pot)
{
  ts_potential_finalize(pot);
  ts_free(NO_ALLOCATOR, pot);
}

void ts_potential_finalize(Potential* pot) { UNUSED_PARAM(pot); }

TSResult ts_potential_get_params(Potential* pot, struct ts_potential_params** params_ptr)
{
  ERROR_IF( TS_ERR_NULL_POINTER_ARG, params_ptr == NULL);

  if (pot->get_params_func != NULL)
  {
    return pot->get_params_func(pot, params_ptr);
  }
  else
  {
    *params_ptr = NULL;
    return TS_OK;
  }
}

TSResult ts_potential_term_list_reset(struct ts_potential_term_list* tl)
{
  tl->terms.size = 0;
  tl->potential_indices.size = 0;
  tl->particle_indices.size = 0;
  tl->term_data.size = 0;
  tl->posv.size = 0;
  tl->forcev.size = 0;
  return TS_OK;
}

TSResult ts_potential_set_param(Potential* pot, struct ts_potential_params* param)
{
  ERROR_IF( TS_ERR_NULL_POINTER_ARG,      param == NULL);
  ERROR_IF( TS_ERR_UNSUPPORTED_OPERATION, pot->set_param_func == NULL);

  return pot->set_param_func(pot, param);
}

TSResult ts_potential_term_list_init(struct ts_potential_term_list* tl)
{
  DATAVECTOR_INIT(ts_potential_term, tl->terms, DEFAULT_TERM_LIST_SIZE);
  DATAVECTOR_INIT(struct ts_potential_term_data, tl->term_data, DEFAULT_TERM_LIST_SIZE);
  DATAVECTOR_INIT(size_t, tl->potential_indices, DEFAULT_TERM_LIST_SIZE);
  DATAVECTOR_INIT(size_t, tl->particle_indices, DEFAULT_TERM_LIST_SIZE);

  DATAVECTOR_INIT(vector3, tl->posv, NBODY_LIMIT*DEFAULT_TERM_LIST_SIZE);
  DATAVECTOR_INIT(vector3, tl->forcev, NBODY_LIMIT*DEFAULT_TERM_LIST_SIZE);

  return TS_OK;
}

TSResult ts_potential_term_list_finalize(struct ts_potential_term_list* tl)
{
  datavector_free(&tl->terms);
  datavector_free(&tl->term_data);
  datavector_free(&tl->potential_indices);
  datavector_free(&tl->particle_indices);
  datavector_free(&tl->posv);
  datavector_free(&tl->forcev);

  return TS_OK;
}

static inline ptrdiff_t compare_pind(const ts_potential_term*restrict e1, const ts_potential_term*restrict e2, size_t ind)
{
  return e1->particle_indices[ind] - e2->particle_indices[ind];
}

/* Need to change the compare_pind code if NBODY_LIMIT changes */
STATIC_ASSERT( NBODY_LIMIT == 4);
static int bondlist_entry_compare(const void* a, const void* b)
{
  const ts_potential_term *e1, *e2;
  ptrdiff_t delta_pat, delta;

  e1 = a;
  e2 = b;

  delta_pat = ((char*)e1->pattern) - ((char*)e2->pattern);
  if (delta_pat != 0) return signof(delta_pat);

  delta = compare_pind(e1,e2,0); if (delta != 0) return signof(delta);
  delta = compare_pind(e1,e2,1); if (delta != 0) return signof(delta);
  delta = compare_pind(e1,e2,2); if (delta != 0) return signof(delta);
  return signof(compare_pind(e1,e2,3));
}

TSResult ts_potential_term_list_finished_adding(struct ts_potential_term_list* tl)
{
  tl->potential_indices.size = 0;
  tl->term_data.size = 0;

  size_t nterms = tl->terms.size;
  if (nterms == 0) return TS_OK;

  qsort(tl->terms.data, nterms, sizeof(ts_potential_term), 
      (int(*)(const void*,const void*))&bondlist_entry_compare);

  if (!datavector_add_n(&tl->term_data, nterms)) {
    return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  ts_potential_term*restrict term = (ts_potential_term*)tl->terms.data;
  struct ts_potential_term_data*restrict data = (struct ts_potential_term_data*)tl->term_data.data;

  for (size_t i=1; i < nterms; i++)
  {
    if (term[i].potential != term[i-1].potential)
    {
      if (!datavector_add(&tl->potential_indices, &i))
        return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    }
  }

  if (!datavector_add(&tl->potential_indices, &nterms)) {
    return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  size_t vec_off = 0;
  for (size_t i=0; i < nterms; i++) {
    data[i].vec_off = vec_off;
    data[i].potential = term[i].potential;
    data[i].nbodies = term[i].potential->nbodies;
    vec_off += data[i].nbodies;
  }

  if (!datavector_add_n(&tl->posv, vec_off)) {
    return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  if (!datavector_add_n(&tl->forcev, vec_off)) {
    return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  if (!datavector_add_n(&tl->particle_indices, vec_off)) {
    return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  size_t*restrict pindv = tl->particle_indices.data;
  for (size_t i=0; i < nterms; i++) {
    int nbodies = data[i].nbodies;
    for (int j=0; j < nbodies; j++) {
      pindv[j] = term[i].particle_indices[j];
    }
    pindv += nbodies;
  }

  return TS_OK;
}

static TSResult calc_potential_forces(struct ts_sim* sim, struct ts_potential_term_data* term_data,
    const vector3*restrict rv, vector3*restrict fv, double* energy_p)
{
  UNUSED_PARAM(sim);
  Potential*restrict potential = term_data->potential;

  ERROR_IF(TS_ERR_UNSUPPORTED_POTENTIAL, 
      (potential->nbodies < 1) || (potential->nbodies > NBODY_LIMIT));

  double en = potential->forces_func(potential, rv, fv);
  term_data->term_energy = set_dbldbl(en);

  if (energy_p) *energy_p = en;

  return TS_OK;
}

extern TSResult calc_forces_individually(ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n, double* energy_ptr);

TSResult calc_forces_individually(ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n, double* energy_ptr)
{
  double en_total = 0.0;
  const vector3*restrict posv = DATAVECTOR_BEGIN(vector3,&termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3,&termlist->forcev);
  struct ts_potential_term_data* term_data = termlist->term_data.data;

  term_data += off;
  posv += term_data->vec_off;
  forcev += term_data->vec_off;

  for (size_t i=0; i < n; i++)
  {
    double en = 0.0;
    RETURN_ON_ERROR( calc_potential_forces(sim, term_data+i, posv, forcev, &en) );
    int nbodies = term_data[i].nbodies;

    posv += nbodies;
    forcev += nbodies;
    en_total += en;
  }

  if (energy_ptr) *energy_ptr += en_total;

  return TS_OK;
}

struct foreach_gather_data {
  struct ts_sim* sim;
  struct ts_potential_term_list* termlist;
};

static inline TSResult gather_particle_positions(struct ts_sim* sim,
    const size_t*restrict pindv,
    size_t i0, size_t i1,
    vector3*restrict rv)
{
  for(size_t i=i0; i < i1; i++) {
    size_t pind = pindv[i];
    vector3_set( rv+i, PARTICLE_POS(sim,pind) );
  }

  return TS_OK;
}

static TSResult foreach_gather_positions(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  struct foreach_gather_data* data = arg;

  ts_sim* sim = data->sim;
  struct ts_potential_term_list* termlist = data->termlist;
  struct ts_potential_term_data* term_data = termlist->term_data.data;
  vector3*restrict posv = DATAVECTOR_BEGIN(vector3,&termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3,&termlist->forcev);
  size_t*restrict pindv = termlist->particle_indices.data;

  size_t pind0 = term_data[i0].vec_off;
  for (size_t i=i0; i < i1; i++) {
    term_data[i].term_energy = zero_dbldbl();
  }
  size_t pind1 = term_data[i1-1].vec_off + term_data[i1-1].nbodies;

  RETURN_ON_ERROR(gather_particle_positions(sim, pindv, pind0,pind1, posv));

  size_t nthreads = ts_dispatch_foreach_num_threads();
  size_t nparticles = sim->num_particles / nthreads;
  if (nparticles*nthreads < sim->num_particles) { nparticles++; }
  size_t pi0 = worker*nparticles;
  size_t pi1 = i0 + nparticles;

  if (pi0 < sim->num_particles) {
    if (pi1 > sim->num_particles) { pi1 = sim->num_particles; }
    for (size_t pi=pi0; pi < pi1; pi++) {
      vector3_zero( sim->vl_eforces + pi );
    }
  }

  size_t nforcev = termlist->forcev.size;
  if (nworkers > 1) {
    size_t each_forcev =  nforcev / nworkers;
    if (each_forcev*nworkers < nforcev) { each_forcev++; }
    size_t fv0 = each_forcev*worker;
    size_t fv1 = fv0 + each_forcev;
    if (fv1 > nforcev) { fv1 = nforcev; }
    for(size_t fvi=fv0; fvi < fv1; fvi++) {
      vector3_zero(forcev+fvi);
    }
  } else {
    for(size_t i=0; i < nforcev; i++) {
      vector3_zero(forcev+i);
    }
  }

  return TS_OK;
}

static TSResult gather_positions(struct ts_sim* sim, struct ts_potential_term_list* term_list)
{
  struct foreach_gather_data data = { 
    .sim = sim,
    .termlist = term_list
  };

  size_t n = term_list->terms.size;
  RETURN_WITH_RESULT( ts_dispatch_foreach(n, foreach_gather_positions, &data) );
}

struct scatter_data {
  ts_sim* sim;
  struct ts_potential_term_data* term_data;
  size_t* pindv;

  vector3* forcev;

  dbldbl* total_energies;
  int nthreads;
};

#if TS_USE_SSE && 0
static inline void particle_add_forces(ts_sim* sim, size_t pind, vector3*restrict eforces, const vector3*restrict f)
{
#if 0
  vector3*restrict pfv = PARTICLE_FORCE(sim,pind);
  vector3*restrict efv = eforces+pind;

  __m256d pf, ef, ff;
  __m256d d1,d2,d3, sa,sb;

  pf = _mm256_loadu_pd(pfv->elts+0);
  ff = _mm256_loadu_pd(f->elts+0);
  ef = _mm256_loadu_pd(efv->elts+0);

  sa = _mm256_add_pd(pf,ff);

  d1 = _mm256_sub_pd(sa,pf);
  d2 = _mm256_sub_pd(ff,d1);
  d3 = _mm256_sub_pd(sa,d1);
  d3 = _mm256_sub_pd(pf,d3);

  sb = _mm256_add_pd(d2,d3);
  sb = _mm256_add_pd(sb,ef);

  pf = _mm256_add_pd(sa,sb);
  ef = _mm256_sub_pd(pf,sa);
  ef = _mm256_sub_pd(sb,ef);

  _mm256_storeu_pd(pfv->elts+0,pf);
  _mm256_storeu_pd(efv->elts+0,ef);
#endif /* 0 */

  vector3*restrict pf = PARTICLE_FORCE(sim,pind);
  vector3*restrict ef = eforces+pind;
  // PREFETCH(pf);
  // PREFETCH(ef);

  __m128d pflo, pfhi, fflo, ffhi, eflo, efhi;

  fflo = f->vec[0];
  ffhi = f->vec[1];
  // fflo = _mm_load_pd(f->elts+0);
  // ffhi = _mm_load_pd(f->elts+2);

  pflo = pf->vec[0];
  pfhi = pf->vec[1];
  // pflo = _mm_load_pd(pf->elts+0);
  // pfhi = _mm_load_pd(pf->elts+2);

  __m128d d1, d2, d3, sa, sb;
  sa = _mm_add_pd(pflo,fflo);

  d1 = _mm_sub_pd(sa,pflo);
  d2 = _mm_sub_pd(fflo,d1);
  d3 = _mm_sub_pd(sa,d1);
  d3 = _mm_sub_pd(pflo,d3);

  eflo = ef->vec[0];
  efhi = ef->vec[1];
  // eflo = _mm_load_pd(ef->elts+0);
  // efhi = _mm_load_pd(ef->elts+2);

  sb = _mm_add_pd(d2,d3);
  sb = _mm_add_pd(sb,eflo);
  pflo = _mm_add_pd(sa,sb);
  eflo = _mm_sub_pd(pflo,sa);
  eflo = _mm_sub_pd(sb,eflo);

  pf->vec[0] = pflo;
  ef->vec[0] = eflo;
  // _mm_store_pd(pf->elts+0,pflo);
  // _mm_store_pd(ef->elts+0,eflo);

  sa = _mm_add_pd(pfhi,ffhi);
  
  d1 = _mm_sub_pd(sa,pfhi);
  d2 = _mm_sub_pd(ffhi,d1);
  d3 = _mm_sub_pd(sa,d1);
  d3 = _mm_sub_pd(pfhi,d3);

  sb = _mm_add_pd(d2,d3);
  sb = _mm_add_pd(sb,efhi);
  pfhi = _mm_add_pd(sa,sb);
  efhi = _mm_sub_pd(pfhi,sa);
  efhi = _mm_sub_pd(sb,efhi);

  pf->vec[1] = pfhi;
  ef->vec[1] = efhi;
  // _mm_store_pd(pf->elts+2,pfhi);
  // _mm_store_pd(ef->elts+2,efhi);

#if 0
  // use Knuth's trick to achieve double-double precision
  vector3 pf;
  vector3 delta1, delta2, delta3;

  vector3 sum, err, ftotal;

  PREFETCH(eforces+pind);
  vector3_set(&pf, PARTICLE_FORCE(sim,pind));
  vector3_add(&sum, &pf, f);

  vector3_sub(&delta1, &sum, &pf);

  vector3_sub(&delta2, f, &delta1);
  vector3_sub(&delta3, &delta1, &sum);
  vector3_add_to(&delta3, &pf);

  vector3_add(&err, &delta2, &delta3);
  vector3_add_to(&err, eforces+pind);

  vector3_add(&ftotal, &sum, &err);
  vector3_sub(&delta1, &ftotal,&sum);

  vector3_set(PARTICLE_FORCE(sim,pind), &ftotal);
  vector3_sub(eforces+pind, &err, &delta1);
#endif /* 0 */
}
#else
static inline void particle_add_forces(ts_sim* sim, size_t pind, vector3*restrict eforces, const vector3*restrict f)
{
  // use Knuth's trick to achieve double-double precision
  vector3 pf;
  vector3 delta1, delta2, delta3;

  vector3 sum, err, ftotal;

  PREFETCH(eforces+pind);
  vector3_set(&pf, PARTICLE_FORCE(sim,pind));
  vector3_add(&sum, &pf, f);

  vector3_sub(&delta1, &sum, &pf);

  vector3_sub(&delta2, f, &delta1);
  vector3_sub(&delta3, &delta1, &sum);
  vector3_add_to(&delta3, &pf);

  vector3_add(&err, &delta2, &delta3);
  vector3_add_to(&err, eforces+pind);

  vector3_add(&ftotal, &sum, &err);
  vector3_sub(&delta1, &ftotal,&sum);

  vector3_set(PARTICLE_FORCE(sim,pind), &ftotal);
  vector3_sub(eforces+pind, &err, &delta1);
}
#endif /* TS_USE_SSE */

static inline TSResult term_scatter_forces(ts_sim* sim,
    struct ts_potential_term_data* term_data,
    const size_t*restrict pindv,
    const vector3*restrict forcev, size_t n)
{
  vector3*restrict eforces = sim->vl_eforces;

  size_t i0 = term_data[0].vec_off;
  size_t i1 = term_data[n-1].vec_off + term_data[n-1].nbodies;
  for (size_t i = i0; i < i1; i++) {
    size_t pind = pindv[i];
    ts_sim_lock_particle(sim,pind);
    particle_add_forces(sim, pind, eforces, forcev+i);
    ts_sim_unlock_particle(sim,pind);
  }

  return TS_OK;
}

static TSResult foreach_scatter_forces(void* arg, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(nworkers);

  struct scatter_data* data = arg;
  struct ts_potential_term_data* term_data = data->term_data;
  const size_t*restrict pindv = data->pindv;
  const vector3*restrict forcev = data->forcev;

  RETURN_ON_ERROR( term_scatter_forces(data->sim, term_data+i0, pindv, forcev, i1-i0) );

  dbldbl accum = zero_dbldbl();

  for(size_t i=i0; i < i1; i++) {
    if (dbldbl_not_zero(term_data[i].term_energy)) {
      accum = sum_dbldbl(accum, term_data[i].term_energy);
    }
  }

  data->total_energies[worker] = accum;

  return TS_OK;
}

MAYBE_UNUSED
static TSResult scatter_forces_multi_threaded(struct ts_sim* sim, const struct ts_potential_term_list* term_list, double* energyp)
{
  int nthreads = ts_dispatch_foreach_num_threads();
  struct scatter_data data = {
    .sim    = sim,
    .term_data = term_list->term_data.data,
    .pindv  = term_list->particle_indices.data,
    .forcev = term_list->forcev.data,
    .total_energies = NULL,
    .nthreads = nthreads
  };

  dbldbl total_energies[nthreads];
  memset(total_energies, 0, nthreads*sizeof(total_energies[0]));

  data.total_energies = total_energies;

  RETURN_ON_ERROR3(
      ts_dispatch_foreach(
        term_list->terms.size,
        foreach_scatter_forces, 
        &data) );

  if (energyp) {
    for (int i=1; i < nthreads; i++) {
      total_energies[0] = sum_dbldbl(total_energies[0],total_energies[i]);
    }

    *energyp += dbldbl_to_double(total_energies[0]);
  }

  return TS_OK;
}

static TSResult scatter_forces_single_threaded(struct ts_sim* sim, const struct ts_potential_term_list* term_list, double* energyp)
{
  struct ts_potential_term_data* term_data = term_list->term_data.data;
  size_t n = term_list->term_data.size;

  if (n > 0) {
    size_t*restrict pindv = term_list->particle_indices.data;
    vector3*restrict forcev = term_list->forcev.data;
    vector3*restrict eforces = sim->vl_eforces;

    size_t i0 = term_data[0].vec_off;
    dbldbl en_accum = zero_dbldbl();
    for (size_t i=0; i < n; i++) {
      en_accum = sum_dbldbl(en_accum, term_data[i].term_energy);
    }
    size_t i1 = term_data[n-1].vec_off + term_data[n-1].nbodies;

    if (energyp) {
      *energyp += dbldbl_to_double(en_accum);
    }

    for (size_t i=i0; i < i1; i++) {
      particle_add_forces(sim, pindv[i], eforces, forcev+i);
    }
  }

  return TS_OK;
}

static TSResult scatter_forces(struct ts_sim* sim, const struct ts_potential_term_list* term_list, double* energyp)
{
  int nthreads = ts_dispatch_foreach_num_threads();
  if (nthreads <= 1) {
    RETURN_ON_ERROR( scatter_forces_single_threaded(sim, term_list, energyp) );
  } else {
    RETURN_ON_ERROR( scatter_forces_multi_threaded(sim, term_list, energyp) );
  }
  return TS_OK;
}

struct dispatch_forces_data {
  ts_sim* sim;

  size_t* potential_index_base;
  size_t* potential_index_end;

  struct ts_potential_term_list* termlist;

  size_t nworkers;

  int termwise_energies;
};

TSResult dispatch_forces_from_list(void* arg, size_t i0, size_t i1, size_t worker_ind, size_t nworkers)
{
  UNUSED_PARAM(i0);
  UNUSED_PARAM(i1);

  struct dispatch_forces_data* data0 = arg;
  struct dispatch_forces_data* data = data0+worker_ind;

  size_t*restrict pind = data->potential_index_base;
  size_t*restrict pind_end = data->potential_index_end;

  size_t pind_last = 0;
  ts_sim* sim = data->sim;
  struct ts_potential_term_list* termlist = data->termlist;
  ts_potential_term* terms = DATAVECTOR_BEGIN(ts_potential_term,&termlist->terms);

  while (pind < pind_end) {
    size_t total_terms = *pind - pind_last;
    size_t worker_terms = total_terms / nworkers;
    if (worker_terms * nworkers < total_terms) { worker_terms++; }

    size_t worker_start = worker_terms*worker_ind;
    size_t worker_n = worker_terms;
    if (worker_start + worker_n > total_terms) { 
      worker_n = total_terms > worker_start ?  total_terms - worker_start : 0;
    }

    if (worker_n > 0) {
      Potential* pot = terms[pind_last].potential;
      double en = 0.0;
      if (!data->termwise_energies && (pot->multiple_forces_func != NULL)) {
        dbldbl en1 = pot->multiple_forces_func(pot, sim, termlist, pind_last+worker_start, worker_n);
        en = dbldbl_to_double(en1);
      }
      else
      {
        RETURN_ON_ERROR( 
            calc_forces_individually(sim, termlist, pind_last+worker_start, worker_n, &en)
            );
      }
    }

    pind_last = *pind;
    pind++;
  }

  return TS_OK;
}

TSResult ts_potential_forces_from_list(struct ts_sim* sim,
    struct ts_potential_term_list* term_list,
    double*restrict energy_ptr,
    int termwise_energies)
{
#define CLEANUP_ON_ERROR { ts_dispatch_foreach_unprime(); }

  ts_dispatch_foreach_prime();
  RETURN_ON_ERROR3( gather_positions(sim,term_list) );

  size_t nthreads = ts_dispatch_foreach_num_threads();
  if (nthreads < 1) { nthreads = 1; }

  struct dispatch_forces_data data[nthreads];
  memset(data,0,nthreads*sizeof(data[0]));

  for (size_t i=0; i < nthreads; i++) {
    data[i].sim = sim;
    data[i].potential_index_base = term_list->potential_indices.data;
    data[i].potential_index_end  = data[i].potential_index_base +
      term_list->potential_indices.size;

    data[i].termlist = term_list;

    data[i].nworkers = nthreads;

    data[i].termwise_energies = termwise_energies;
  }

  RETURN_ON_ERROR3(
      ts_dispatch_foreach(nthreads, 
        dispatch_forces_from_list, data+0) );

  RETURN_ON_ERROR3( scatter_forces(sim,term_list, energy_ptr) );

  ts_dispatch_foreach_unprime();
#undef CLEANUP_ON_ERROR
  return TS_OK;
}

TSResult ts_potential_energies_resum(struct ts_sim* sim, struct ts_potential_term_list* term_list, double energies[], int max_term_groups)
{
  UNUSED_PARAM(sim);

  const size_t*restrict potInd_cur = (const size_t*) term_list->potential_indices.data;
  const size_t*restrict potInd_end = potInd_cur + term_list->potential_indices.size;

  size_t nenergies = max_term_groups+1;
  dbldbl accum[nenergies];
  memset(accum, 0, nenergies*sizeof(accum[0]));

  size_t ind0 = 0;
  ts_potential_term* terms = term_list->terms.data;
  struct ts_potential_term_data* term_data = term_list->term_data.data;
  while (potInd_cur < potInd_end)
  {
    dbldbl en = zero_dbldbl();
    for (size_t i=ind0; i < *potInd_cur; i++) {
      dbldbl term_energy = term_data[i].term_energy;
      if (dbldbl_not_zero(term_energy)) {
        accum[0] = sum_dbldbl(accum[0], term_energy);

        if ((max_term_groups > 0) && (terms[i].pattern != NULL)) {
          int term_group = terms[i].pattern->term_group;
          if ((term_group >= 0) && (term_group < max_term_groups)) {
            accum[term_group+1] = sum_dbldbl(accum[term_group+1],term_energy);
          }
        }

        en = sum_dbldbl(en, term_energy);
      }
    }

    terms[ind0].potential->potential_energy = dbldbl_to_double(en);

    ind0 = *potInd_cur;
    potInd_cur++;
  }

  for(size_t i=0; i < nenergies; i++) {
    energies[i] += dbldbl_to_double(accum[i]);
  }

  return TS_OK;
}

TSResult ts_potential_forces(struct ts_sim* sim, struct ts_potential_term_data* term_data,
    const vector3*restrict rv, vector3*restrict fv, double* energy_p)
{
  double en = 0.0;
  RETURN_ON_ERROR( calc_potential_forces(sim, term_data, rv, fv, &en) );
  energy_p[0] += en;

  return TS_OK;
}

static TSResult dist_constraint_get_params(DistanceConstraint* constraint, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 2, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "dist"; params[0].value = constraint->distance;
  params[1].name = NULL; params[1].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

DistanceConstraint* tsp_distance_constraint_new(double dist)
{
  DistanceConstraint* constraint = ts_alloc(NO_ALLOCATOR, sizeof(DistanceConstraint) );
  ts_potential_init((Potential*)constraint, "DIST-CONSTRAINT", TWO_BODY);

  constraint->parent.get_params_func = (PotentialGetParametersFunc) dist_constraint_get_params;

  constraint->distance = dist;
  constraint->sqdist   = dist*dist;

  return constraint;
}

void tsp_distance_constraint_delete(DistanceConstraint *constraint)
{
  ts_potential_delete((Potential *)constraint);
}

static double const_field_energy(ConstantField* potential, vector3 r[NBODY_LIMIT]);
static double const_field_forces(ConstantField* potential, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static TSResult const_field_get_params(ConstantField* potential, struct ts_potential_params** params);
ConstantField* tsp_const_field_new(const vector3*restrict const_force, double ref_point)
{
  ConstantField* potential = ts_alloc(NO_ALLOCATOR, sizeof(ConstantField) );
  ts_potential_init((Potential*)potential, "CONST-FIELD", ONE_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc) const_field_energy;
  potential->parent.forces_func = (PotentialForcesFunc) const_field_forces;
  potential->parent.get_params_func = (PotentialGetParametersFunc) const_field_get_params;

  vector3_set(&potential->const_force, const_force);
  potential->ref_point = ref_point;

  return potential;
}

static TSResult const_field_get_params(ConstantField* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 5, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "fx"; params[0].value = vector3_x(&potential->const_force);
  params[1].name = "fy"; params[1].value = vector3_y(&potential->const_force);
  params[2].name = "fz"; params[2].value = vector3_z(&potential->const_force);
  params[3].name = "ref_point"; params[3].value = potential->ref_point;
  params[4].name = NULL;  params[4].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static double const_field_energy(ConstantField* potential, vector3 r[NBODY_LIMIT])
{
  vector3 fnorm;
  vector3_set(&fnorm, &potential->const_force);
  vector3_normalize(&fnorm);

  double A = vector3_dot(&fnorm,r+0);
  return A - potential->ref_point;
}

static double const_field_forces(ConstantField* potential, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  vector3_set(f+0, &potential->const_force);
  return const_field_energy(potential, r);
}

void tsp_const_field_delete(ConstantField* field)
{
  ts_potential_delete((Potential *)field);
}

static double spin_around_particle_energy(SpinAroundParticle* potential, vector3 r[NBODY_LIMIT])
{
  UNUSED_PARAM(potential);
  UNUSED_PARAM(r);
  return 0.0;
}
static double spin_around_particle_forces(SpinAroundParticle* spinner, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static TSResult spin_around_particle_get_params(SpinAroundParticle* potential, struct ts_potential_params** params_ptr);
SpinAroundParticle* tsp_spin_around_particle_new(struct ts_sim* sim, size_t pind, const vector3*restrict axis)
{
  SpinAroundParticle* spinner = ts_alloc(NO_ALLOCATOR, sizeof(*spinner));
  ts_potential_init((Potential*)spinner, "SPIN-AROUND-PARTICLE", ONE_BODY);

  spinner->parent.energy_func = (PotentialEnergyFunc) spin_around_particle_energy;
  spinner->parent.forces_func = (PotentialForcesFunc) spin_around_particle_forces;
  spinner->parent.get_params_func = (PotentialGetParametersFunc) spin_around_particle_get_params;

  spinner->sim = sim;
  spinner->pind = pind;
  vector3_set(&spinner->axis, axis);

  return spinner;
}

void tsp_spin_around_particle_delete(SpinAroundParticle* potential)
{
  ts_potential_delete((Potential *)potential);
}

static double spin_around_particle_forces(SpinAroundParticle* spinner, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  vector3 dr;
  vector3_sub(&dr, r+0, PARTICLE_POS(spinner->sim,spinner->pind));

  vector3_normalize(&dr);
  vector3_cross(f+0, &dr, &spinner->axis);

  return 0.0;
}

static TSResult spin_around_particle_get_params(SpinAroundParticle* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 4, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "axis_x"; params[0].value = vector3_x(&potential->axis);
  params[1].name = "axis_y"; params[1].value = vector3_y(&potential->axis);
  params[2].name = "axis_z"; params[2].value = vector3_z(&potential->axis);
  params[3].name = NULL;  params[3].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static double plane_barrier_energy(PlaneBarrier *potential, vector3 r[NBODY_LIMIT]);
static double plane_barrier_forces(PlaneBarrier *spinner, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static TSResult plane_barrier_get_params(PlaneBarrier *potential, struct ts_potential_params **params_ptr);
static TSResult plane_barrier_set_param(PlaneBarrier *potential, const struct ts_potential_params *param);

PlaneBarrier *tsp_plane_barrier_new(const vector3 *pos, const vector3 *norm, double dist, double eps)
{
  PlaneBarrier* barrier = ts_alloc(NO_ALLOCATOR, sizeof(*barrier));
  ts_potential_init((Potential*)barrier, "PLANE-BARRIER", ONE_BODY);

  barrier->parent.energy_func = (PotentialEnergyFunc) plane_barrier_energy;
  barrier->parent.forces_func = (PotentialForcesFunc) plane_barrier_forces;
  barrier->parent.get_params_func = (PotentialGetParametersFunc) plane_barrier_get_params;
  barrier->parent.set_param_func = (PotentialSetParameterFunc) plane_barrier_set_param;

  vector3_set(&barrier->pos , pos );
  vector3_set(&barrier->norm, norm);

  tsp_plane_barrier_set_params(barrier, dist, eps);

  return barrier;
}

void tsp_plane_barrier_delete(PlaneBarrier *potential)
{
  ts_potential_delete((Potential *)potential);
}

#define INV_TWO_TO_SIXTH 0.8908987181403393
void tsp_plane_barrier_set_params(PlaneBarrier *potential, double dist, double eps)
{
  potential->eps = eps;
  potential->dist = dist;
  potential->sigma = INV_TWO_TO_SIXTH*dist;
  potential->f_coeff = 24.0*eps/potential->sigma;
}

/* slightly different version of the WCA purely repulsive potential that
 * takes advantage of simplifications given the plane barrier potential
 * form
 */
static inline double plane_barrier_force(PlaneBarrier *potential, const vector3 *restrict r, vector3 *f)
{
  vector3 dr;
  vector3_sub(&dr, r, &potential->pos);
  double dpl = vector3_dot(&dr,&potential->norm);
  if (dpl >= potential->dist) { vector3_zero(f); return 0.0; }

  double rat = potential->sigma/dpl;
  double rat2 = rat*rat;
  double rat6 = rat2*rat2*rat2;
  double fmag = potential->f_coeff * (rat6*rat)*(2.0*rat6 - 1.0);
  vector3_set_scaled(f, fmag, &potential->norm);

  return potential->eps*(1.0 + 4.0*rat6*(rat6 - 1.0));
}

static double plane_barrier_energy(PlaneBarrier *potential, vector3 r[NBODY_LIMIT])
{
  vector3 f = VEC3_ZERO;
  return plane_barrier_force(potential, r+0, &f);
}

static double plane_barrier_forces(PlaneBarrier *potential, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  vector3 df;
  double en = plane_barrier_force(potential, r+0, &df);
  vector3_add_to(f+0,&df);
  return en;
}

static TSResult plane_barrier_set_param(PlaneBarrier *potential, const struct ts_potential_params *param)
{
  if ( strcmp(param->name, "pos_x") == 0) {
    VECTOR3_X(&potential->pos) = param->value;
  } else if (strcmp(param->name, "pos_y") == 0) {
    VECTOR3_Y(&potential->pos) = param->value;
  } else if (strcmp(param->name, "pos_z") == 0) {
    VECTOR3_Z(&potential->pos) = param->value;
  } else if (strcmp(param->name, "norm_x") == 0) {
    VECTOR3_X(&potential->norm) = param->value;
  } else if (strcmp(param->name, "norm_y") == 0) {
    VECTOR3_Y(&potential->norm) = param->value;
  } else if (strcmp(param->name, "norm_z") == 0) {
    VECTOR3_Z(&potential->norm) = param->value;
  } else if (strcmp(param->name, "eps") == 0) {
    potential->eps = param->value;
  } else if (strcmp(param->name, "dist") == 0) {
    potential->dist = param->value;
  } else {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  tsp_plane_barrier_set_params(potential, potential->dist, potential->eps);

  return TS_OK;
}

static TSResult plane_barrier_get_params(PlaneBarrier *potential, struct ts_potential_params **params_ptr)
{
  struct ts_potential_params *params = ts_calloc(NO_ALLOCATOR, 9, sizeof(*params));
  if (params == NULL) { TS_RAISE_ERROR( TS_ERR_OUT_OF_MEMORY ); }

  params[0].name = "pos_x";  params[0].value = vector3_x(&potential->pos);
  params[1].name = "pos_y";  params[1].value = vector3_y(&potential->pos);
  params[2].name = "pos_z";  params[2].value = vector3_z(&potential->pos);

  params[3].name = "norm_x"; params[3].value = vector3_x(&potential->norm);
  params[4].name = "norm_y"; params[4].value = vector3_y(&potential->norm);
  params[5].name = "norm_z"; params[5].value = vector3_z(&potential->norm);

  params[6].name = "eps";  params[6].value = potential->eps;
  params[7].name = "dist"; params[7].value = potential->dist;

  params[8].name = NULL;  params[8].value = 0.0;

  *params_ptr = params;

  return TS_OK;
}


static double harmonic_energy(HarmonicPotential* potential, const vector3 r[NBODY_LIMIT]);
static double harmonic_forces(HarmonicPotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static void harmonic_hessian(HarmonicPotential* potential, const vector3 r[NBODY_LIMIT],
    VecReal* hessian_upper_tri);

static dbldbl harmonic_multiple_forces(HarmonicPotential* potential, struct ts_sim* sim, 
    struct ts_potential_term_list* termlist, size_t off, size_t n);
static TSResult harmonic_get_params(HarmonicPotential* potential, struct ts_potential_params** params_ptr);
static TSResult harmonic_set_param(HarmonicPotential* potential, struct ts_potential_params* param);

HarmonicPotential* tsp_harmonic_new(double k, double req)
{
  HarmonicPotential* potential = ts_alloc(NO_ALLOCATOR, sizeof(HarmonicPotential) );
  ts_potential_init((Potential*)potential, "HARMONIC", TWO_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc) harmonic_energy;
  potential->parent.forces_func = (PotentialForcesFunc) harmonic_forces;

  potential->parent.hessian_func = (PotentialHessianFunc) harmonic_hessian;

  potential->parent.multiple_forces_func = (PotentialMultipleForcesFunc)harmonic_multiple_forces;

  potential->parent.get_params_func = (PotentialGetParametersFunc) harmonic_get_params;
  potential->parent.set_param_func = (PotentialSetParameterFunc) harmonic_set_param;

  potential->k = k;
  potential->req = req;

  return potential;
}

void tsp_harmonic_delete(HarmonicPotential* potential)
{
  ts_potential_delete((Potential *)potential);
}

static TSResult harmonic_get_params(HarmonicPotential* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 3, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "k";   params[0].value = potential->k;
  params[1].name = "r_eq"; params[1].value = potential->req;
  params[2].name = NULL;  params[2].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static TSResult harmonic_set_param(HarmonicPotential* potential, struct ts_potential_params* param)
{
  if ( strcmp(param->name, "k") == 0) {
    potential->k = param->value;
  } else if ( strcmp(param->name, "r_eq") == 0 ) {
    potential->req= param->value;
  } else {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  return TS_OK;
}

static double harmonic_energy(HarmonicPotential* potential, const vector3 r[NBODY_LIMIT])
{
  vector3 f[2] = { VEC3_ZERO, VEC3_ZERO };
  return harmonic_forces(potential,r,f);
}

static inline dbldbl calc_harmonic_forces(const HarmonicPotential*restrict pot, size_t n, const vector3*restrict rv, vector3*restrict fv)
{
  double k = pot->k;
  double req = pot->req;

  dbldbl en_tot = zero_dbldbl();

  for (size_t i=0; i < n; i++) {
    vector3 dr;
    double r = vector3_distance(rv+1, rv+0, &dr);

    double dx = r - req;
    double fmag = k*dx;
    double en = 0.5*fmag*dx;

    en_tot = accum_dbldbl(en_tot, en);

    if (r > VEC_TOL)
    {
      fmag /= r;
      vector3_scaleby(&dr, fmag);
      vector3_set_negative(fv+0,&dr);
      vector3_set(fv+1, &dr);
    }
    else
    {
      vector3_zero(fv+0);
      vector3_zero(fv+1);
    }

    rv += 2;
    fv += 2;
  }

  return en_tot;
}

static dbldbl harmonic_multiple_forces(HarmonicPotential* potential, struct ts_sim* sim, 
    struct ts_potential_term_list* termlist, size_t off, size_t n)
{
  UNUSED_PARAM(sim);
  if (n==0) return zero_dbldbl();

  struct ts_potential_term_data* term_data = 
    DATAVECTOR_AT(struct ts_potential_term_data,&termlist->term_data,off);
  vector3*restrict posv = DATAVECTOR_BEGIN(vector3, &termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3, &termlist->forcev);

  posv   += term_data->vec_off;
  forcev += term_data->vec_off;

  dbldbl en = calc_harmonic_forces(potential, n, posv, forcev);

  term_data->term_energy = en;

  return en;
}

static double harmonic_forces(HarmonicPotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  dbldbl en = calc_harmonic_forces(potential, 1, r, f);
  return dbldbl_to_double(en);
}

#define DOF (2*PARTICLE_NUM_DOF)
static void harmonic_hessian(HarmonicPotential* pot, const vector3 r[NBODY_LIMIT],
    VecReal* hess)
{
  vector3 dr;
  double Rij = vector3_distance(r+0,r+1, &dr);
  double rRij = 1.0/Rij;

  vector3_scale_by(&dr, rRij);

  double dRij = Rij - pot->req;
  double du = rRij * dRij;

  for (int i=0; i < PARTICLE_NUM_DOF; i++) {
    double a2 = square(VECTOR3_ELT(&dr,i));
    double h_ii = pot->k*((1.0-du)*a2 + du);
    hess[ hessian_index(DOF, i                 ,i                 ) ] =  h_ii;
    hess[ hessian_index(DOF, i+PARTICLE_NUM_DOF,i+PARTICLE_NUM_DOF) ] =  h_ii;
    hess[ hessian_index(DOF, i                 ,i+PARTICLE_NUM_DOF) ] = -h_ii;

    for (int j=i+1; j < PARTICLE_NUM_DOF; j++) {
      VecReal ab = VECTOR3_ELT(&dr,i)*VECTOR3_ELT(&dr,j);
      double h_ij = (pot->k)*(1.0-du)*ab;
      hess[ hessian_index(DOF, i                 ,j                 ) ] =  h_ij;
      hess[ hessian_index(DOF, i+PARTICLE_NUM_DOF,j+PARTICLE_NUM_DOF) ] =  h_ij;
      hess[ hessian_index(DOF, i                 ,j+PARTICLE_NUM_DOF) ] = -h_ij;
      hess[ hessian_index(DOF, j                 ,i+PARTICLE_NUM_DOF) ] = -h_ij;
    }
  }

}
#undef DOF

static inline double calc_wca_forces(WCAPotential* potential, const vector3*restrict r0, const vector3*restrict r1, vector3*restrict f1)
{
  double rsq = vector3_sqdistance(r0,r1,f1);

  if (rsq >= potential->r_min_sq) { vector3_zero(f1); return 0.0; }

  double ratio2 = potential->sigma_sq / rsq;
  double ratio6 = ratio2*ratio2*ratio2;
  double ratio8 = ratio2*ratio6;

  double fvec_coeff = potential->f_coeff * (2.0*ratio6 - 1.0)*ratio8;

  vector3_scaleby(f1, fvec_coeff);

  return potential->eps*(1 + 4.0*ratio6*(ratio6 - 1));
}

static double wca_energy(WCAPotential* potential, const vector3 r[NBODY_LIMIT])
{
  vector3 df;
  return calc_wca_forces(potential, r+0, r+1, &df); 
}

static double wca_forces(WCAPotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  vector3 df;
  double en = calc_wca_forces(potential, r+0, r+1, &df); 

  vector3_sub_from(f+0, &df);
  vector3_add_to  (f+1, &df);

  return en;
}

static inline void check_high_energies(ts_sim* sim, ts_potential_term* term,
    const vector3*restrict posv, double en)
{
  if (en <= NONBOND_ENERGY_THRESH) return;

  vector3 dr;
  double dist = vector3_distance(posv+0, posv+1, &dr);

#define _L(...) TS_LOG(sim, TS_LOG_WARN, TS_LOG_NONBONDS, __VA_ARGS__)
  _L( "NONBOND energies are high (%f)\n", en);
  _L( "particle indices: %lu %lu\n", term->particle_indices[0], term->particle_indices[1]);
  _L( "  P0 @ %f %f %f\n", vector3_x(posv+0), vector3_y(posv+0), vector3_z(posv+0));
  _L( "  P1 @ %f %f %f\n", vector3_x(posv+1), vector3_y(posv+1), vector3_z(posv+1));
  _L( "  separation is %f\n\n", dist);
#undef _L
}

static dbldbl wca_multiple_forces(WCAPotential* potential, struct ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n)
{
  ts_potential_term* terms = DATAVECTOR_AT(ts_potential_term,&termlist->terms,off);

  struct ts_potential_term_data* term_data = 
    DATAVECTOR_AT(struct ts_potential_term_data,&termlist->term_data,off);

  vector3*restrict posv = DATAVECTOR_BEGIN(vector3, &termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3, &termlist->forcev);
  struct ts_potential_term_data*restrict end = term_data+n;
  dbldbl total_en = zero_dbldbl();

  if (term_data == end) { return total_en; }

  posv   += term_data->vec_off;
  forcev += term_data->vec_off;

  do {
    tsAssert(term_data->potential == (Potential*)potential);

    vector3 df;
    double en = calc_wca_forces(potential, posv+0, posv+1, &df);
    check_high_energies(sim, terms, posv, en);

    vector3_sub_from(forcev+0, &df);
    vector3_add_to(forcev+1, &df);

    term_data->term_energy = set_dbldbl(en);
    total_en = accum_dbldbl(total_en,en);

    posv   += 2;
    forcev += 2;
    terms++;
    term_data++;
  } while(term_data < end);

  return total_en;
}

static TSResult wca_get_params(WCAPotential* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 3, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "sigma"; params[0].value = potential->sigma;
  params[1].name = "eps";   params[1].value = potential->eps;
  params[2].name = NULL;    params[2].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static TSResult wca_set_param(WCAPotential* potential, struct ts_potential_params* param)
{
  if ( strcmp(param->name, "sigma") == 0) {
    potential->sigma = param->value;
  } else if ( strcmp(param->name, "eps") == 0 ) {
    potential->eps = param->value;
  } else {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  tsp_wca_set_params(potential, potential->sigma, potential->eps);

  return TS_OK;
}

WCAPotential* tsp_wca_new(double sigma, double eps)
{
  WCAPotential* potential = ts_alloc( NO_ALLOCATOR, sizeof(WCAPotential) );
  ts_potential_init((Potential*)potential, "WCA_REPULSIVE", TWO_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc) wca_energy;
  potential->parent.forces_func = (PotentialForcesFunc) wca_forces;
  potential->parent.multiple_forces_func = (PotentialMultipleForcesFunc) wca_multiple_forces;
  potential->parent.get_params_func = (PotentialGetParametersFunc) wca_get_params;
  potential->parent.set_param_func = (PotentialSetParameterFunc) wca_set_param;

  tsp_wca_set_params(potential, sigma, eps);

  return potential;
}

#define TWO_TO_SIXTH 1.12246204830937
void tsp_wca_set_params(WCAPotential* potential, double sigma, double eps)
{
  potential->sigma = sigma;
  potential->eps   = eps;

  double sigma_sq = sigma*sigma;
  potential->sigma_sq = sigma_sq;

  double r_min = sigma*TWO_TO_SIXTH;
  potential->r_min = r_min;
  potential->r_min_sq = r_min*r_min;
  potential->f_coeff = 24.0*eps/sigma_sq;
}

void tsp_wca_delete(WCAPotential* potential)
{
  ts_potential_delete((Potential *)potential);
}

static inline MAYBE_UNUSED double closest_integer(double x)
{
  return ceil(x-0.5);
}

static inline double regularize_angle(double ang)
{
  int n = (int) ( ang/(2.0*M_PI) );
  return ang - 2*M_PI*n;
}

// Switch function 1:
//               { +1 if m == a
// sw1(m, a,b) = { -1 if m == b
//               {  0 otherwise
static inline MAYBE_UNUSED int sw1(int m, int a, int b)
{
  return (m == a) ? 1 : ( (m == b) ? -1 : 0 );
}

// Calculates d(|v_ji|)/dr
static inline void calc_dvji_dr(vector3*restrict dv, int m, const vector3*restrict v_ji)
{
  if (m < 2)
  {
    vector3_set(dv, v_ji);
    vector3_normalize(dv);
    if (m==0) vector3_negate(dv);
  }
  else // m == 2
  { 
    vector3_zero(dv); // d(v_ji)/dr_k == 0
  }
}

// Calculates d(|v_ji|)/dr
static inline void calc_dvkj_dr(vector3*restrict dv, int m, const vector3*restrict v_kj)
{
  if (m > 0)
  {
    vector3_set(dv, v_kj);
    vector3_normalize(dv);
    if (m==1) vector3_negate(dv);
  }
  else // m == 0
  { 
    vector3_zero(dv); // d(v_kj)/dr_i == 0
  }
}

static inline MAYBE_UNUSED int calc_d2v_sign_ind(int ind) { return -2*ind*(ind-2)-1; }

static inline void calc_d2vji(mat3x3* d2v, int m, int n, const vector3*restrict v_ji, double r_sqd_ji, double r_d_ji)
{
  if ((m == 2) || (n == 2)) {
    mat3x3_zero(d2v);
  } else {
    VecReal sign_m = (m == 1) ? 1.0 : -1.0;
    VecReal sign_n = (n == 1) ? 1.0 : -1.0;
    vector3_outer(d2v, v_ji, v_ji);
    mat3x3_scale_by(d2v, r_sqd_ji);
    mat3x3_add_scaled_identity_to_self(d2v, -1.0);
    mat3x3_scale_by(d2v, -(sign_m)*(sign_n)*(r_d_ji));
  }
}

static inline void calc_d2vkj(mat3x3* d2v, int m, int n, const vector3*restrict v_kj, double r_sqd_kj, double r_d_kj)
{
  if ((m == 0) || (n == 0)) {
    mat3x3_zero(d2v);
  } else {
    VecReal sign_m = (m == 2) ? 1.0 : -1.0;
    VecReal sign_n = (n == 2) ? 1.0 : -1.0;
    vector3_outer(d2v, v_kj, v_kj);
    mat3x3_scale_by(d2v, r_sqd_kj);
    mat3x3_add_scaled_identity_to_self(d2v, -1.0);
    mat3x3_scale_by(d2v, -(sign_m*sign_n)*(r_d_kj));
  }
}

static inline void calc_d2cos(mat3x3*restrict d2cos, int m, int n,
    const vector3*restrict v_ji, const vector3*restrict v_kj,
    const vector3*restrict dcos_drm, const vector3*restrict dcos_drn,
    double cos_a, double r_norms,
    double r_d_ji, double r_d_kj,
    double r_sqd_ji, double r_sqd_kj)
{
  vector3 dv_ji_m, dv_kj_m, dv_ji_n, dv_kj_n;

  calc_dvji_dr(&dv_ji_m, m, v_ji);
  calc_dvji_dr(&dv_ji_n, n, v_ji);

  calc_dvkj_dr(&dv_kj_m, m, v_kj);
  calc_dvkj_dr(&dv_kj_n, n, v_kj);

  vector3_outer(d2cos, dcos_drm, &dv_ji_n);
  mat3x3_scale_by(d2cos, -r_d_ji);

  mat3x3 tmp;
  vector3_outer(&tmp, dcos_drm, &dv_kj_n);
  mat3x3_scale_by(&tmp, -r_d_kj);
  mat3x3_add_to(d2cos, &tmp);

  vector3_outer(&tmp, &dv_ji_m, dcos_drn);
  mat3x3_scale_by(&tmp, -r_d_ji);
  mat3x3_add_to(d2cos, &tmp);

  vector3_outer(&tmp, &dv_kj_m, dcos_drn);
  mat3x3_scale_by(&tmp, -r_d_kj);
  mat3x3_add_to(d2cos, &tmp);

  calc_d2vji(&tmp, m,n, v_ji, r_sqd_ji, r_d_ji);
  mat3x3_scale_by(&tmp, -cos_a*r_d_ji);
  mat3x3_add_to(d2cos, &tmp);

  calc_d2vkj(&tmp, m,n, v_kj, r_sqd_kj, r_d_kj);
  mat3x3_scale_by(&tmp, -cos_a*r_d_kj);
  mat3x3_add_to(d2cos, &tmp);

  vector3_outer(&tmp, &dv_ji_m, &dv_kj_n);
  mat3x3_scale_by(&tmp, -cos_a*r_norms);
  mat3x3_add_to(d2cos, &tmp);

  vector3_outer(&tmp, &dv_kj_m, &dv_ji_n);
  mat3x3_scale_by(&tmp, -cos_a*r_norms);
  mat3x3_add_to(d2cos, &tmp);

  int ji_kj = sw1(m,1,0) * sw1(n,2,1);
  int kj_ji = sw1(m,2,1) * sw1(n,1,0);

  if (ji_kj != 0) { 
    mat3x3_add_scaled_identity_to_self(d2cos, ji_kj*r_norms);
  }
  if (kj_ji != 0) {
    mat3x3_add_scaled_identity_to_self(d2cos, kj_ji*r_norms);
  }

}

static inline void mat3x3_add_hess(const int m, const int n, VecReal*restrict hess, const mat3x3*restrict mat)
{
  for (int i=0; i < 3; i++) {
    int j0 = (m==n) ? i : 0;
    for (int j=j0; j < 3; j++) {
      hess[ hessian_index(3*PARTICLE_NUM_DOF,
          m*PARTICLE_NUM_DOF + i,
          n*PARTICLE_NUM_DOF + j) ] += MAT3X3_ELT(mat, i,j);
    }
  }
}

static dbldbl angle_harmonic_calculate(
    AngleHarmonicPotential* potential,
    size_t n,
    const vector3*restrict r,
    vector3*restrict f,
    VecReal*restrict hess)
{
  double theta_eq = potential->theta_eq;
  double k = potential->k;

  dbldbl en_tot = zero_dbldbl();
  for (size_t i=0; i < n; i++) {
    vector3 v_ji;
    vector3 v_kj;

    vector3_sub(&v_ji, r+1, r+0);
    vector3_sub(&v_kj, r+2, r+1);
    r += 3;

    double sqd_ji = vector3_sqnorm(&v_ji);
    double sqd_kj = vector3_sqnorm(&v_kj);
    double dot_kji = vector3_dot(&v_ji,&v_kj);

    double r_sqd_ji = 1.0 / sqd_ji;
    double r_sqd_kj = 1.0 / sqd_kj;

    double r_norms = sqrt(r_sqd_ji * r_sqd_kj);
    double r_d_ji = sqrt(r_sqd_ji);
    double r_d_kj = sqrt(r_sqd_kj);

    double cos_a = dot_kji * r_norms;
    if (fabs(cos_a) > 1.0) { 
      cos_a = copysign(1.0,cos_a); 
    }

#ifdef USE_TS_FASTMATH
    double a = my_acos( cos_a );
#else
    double a = acos( cos_a );
#endif /* USE_TS_FASTMATH */
    double da = a - theta_eq;

    en_tot = accum_dbldbl(en_tot, 0.5*k*da*da);

    if (f == NULL && hess == NULL) {
      continue;
    }

    double r_sinsq_a, r_sin_a, df;
    if (fabs(cos_a) <= 0.99999) {
      r_sinsq_a = 1.0/(1.0 - cos_a*cos_a);
      r_sin_a = sqrt(r_sinsq_a);

      df = da*r_sin_a;
    } else {
      r_sinsq_a = 1.0/(1.0 - cos_a*cos_a+FORCE_EPS);
      r_sin_a = sqrt(r_sinsq_a);

      if (theta_eq < 0.00001)
      {
        df = 1.0 + da*da/6.0;
      }
      else if (M_PI - theta_eq < 0.00001)
      {
        df = -1.0 - da*da/6.0;
      }
      else
      {
        df = da*r_sin_a;
      }
    }

    df *= k;

    tsAssert(k >= 0.0);
    tsAssert(df * (a - theta_eq) >= 0.0);
    tsAssert(r_norms >= 0.0);
    tsAssert(r_sqd_ji >= 0.0);
    tsAssert(r_sqd_kj >= 0.0);

    vector3 dcos_dr0;
    vector3 dcos_dr1;
    vector3 dcos_dr2;
    vector3_scale_and_add(&dcos_dr0, -r_norms, &v_kj,  cos_a*r_sqd_ji, &v_ji);
    vector3_scale_and_add(&dcos_dr2,  r_norms, &v_ji, -cos_a*r_sqd_kj, &v_kj);

    vector3_scale_and_add(&dcos_dr1, -1.0, &dcos_dr0, -1.0, &dcos_dr2);

    if (f != NULL) {
      vector3_add_multiple_to(f+0, df, &dcos_dr0);
      vector3_add_multiple_to(f+1, df, &dcos_dr1);
      vector3_add_multiple_to(f+2, df, &dcos_dr2);

      f += 3;
    }

    if (hess == NULL) {
      continue;
    }

    double ddf = r_sinsq_a * (k - df*cos_a);

    mat3x3 subhess;
    mat3x3 d2cos;

#define DCOS_DR(ind) (&(dcos_dr ## ind))
#define LOOP(m,n) do {                   \
    vector3_outer(&subhess, DCOS_DR(m), DCOS_DR(n)); \
    mat3x3_scale_by(&subhess, ddf);        \
    calc_d2cos(&d2cos, m,n,                \
        &v_ji, &v_kj,                      \
        DCOS_DR(m), DCOS_DR(n),            \
        cos_a, r_norms,                    \
        r_d_ji, r_d_kj,                    \
        r_sqd_ji, r_sqd_kj);               \
    mat3x3_scale_by(&d2cos, -df);          \
    mat3x3_add_to(&subhess, &d2cos);       \
    mat3x3_add_hess(m, n, hess, &subhess); \
  } while(0)

    LOOP(0,0);
    LOOP(0,1);
    LOOP(0,2);

    LOOP(1,1);
    LOOP(1,2);

    LOOP(2,2);

#undef LOOP
#undef DCOS_DR

    hess += hessian_size(9);
  }

  return en_tot;
}


static double angle_harmonic_energy(AngleHarmonicPotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  UNUSED_PARAM(f);
  dbldbl en = angle_harmonic_calculate(potential, 1, r, NULL, NULL);
  return dbldbl_to_double(en);
}

static double angle_harmonic_forces(AngleHarmonicPotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  dbldbl en = angle_harmonic_calculate(potential, 1, r, f, NULL);
  return dbldbl_to_double(en);
}

static void angle_harmonic_hessian(AngleHarmonicPotential* potential, vector3 r[NBODY_LIMIT],
    VecReal*restrict hess)
{
  (void)angle_harmonic_calculate(potential, 1, r, NULL, hess);
}

static dbldbl angle_harmonic_multiple_forces(AngleHarmonicPotential* potential, struct ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n)
{
  UNUSED_PARAM(sim);
  if (n == 0) { return zero_dbldbl(); }

  struct ts_potential_term_data* term_data = 
    DATAVECTOR_AT(struct ts_potential_term_data,&termlist->term_data,off);
  vector3*restrict posv = DATAVECTOR_BEGIN(vector3, &termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3, &termlist->forcev);

  posv   += term_data->vec_off;
  forcev += term_data->vec_off;

  dbldbl en = angle_harmonic_calculate(potential, n,
      posv+0, forcev+0, NULL);

  term_data->term_energy = en;

  return en;
}

static TSResult angle_harmonic_get_params(AngleHarmonicPotential* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 3, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "k";        params[0].value = potential->k;
  params[1].name = "theta_eq"; params[1].value = potential->theta_eq;
  params[2].name = NULL;       params[2].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static TSResult angle_harmonic_set_param(AngleHarmonicPotential* potential, const struct ts_potential_params* param)
{
  if ( strcmp(param->name, "k") == 0) {
    potential->k = param->value;
  } else if ( strcmp(param->name, "theta_eq") == 0 ) {
    potential->theta_eq = param->value;
  } else {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  return TS_OK;
}

AngleHarmonicPotential* tsp_angle_harmonic_new(double k, double theta_eq)
{
  AngleHarmonicPotential* potential = ts_alloc( NO_ALLOCATOR, sizeof(AngleHarmonicPotential) );
  ts_potential_init((Potential*)potential, "ANGLE-HARMONIC", THREE_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc) angle_harmonic_energy;
  potential->parent.forces_func = (PotentialForcesFunc) angle_harmonic_forces;
  potential->parent.multiple_forces_func = (PotentialMultipleForcesFunc) angle_harmonic_multiple_forces;

  potential->parent.hessian_func = (PotentialHessianFunc) angle_harmonic_hessian;

  potential->parent.get_params_func = (PotentialGetParametersFunc)angle_harmonic_get_params;
  potential->parent.set_param_func = (PotentialSetParameterFunc) angle_harmonic_set_param;

  potential->parent.delete_func = (PotentialDeleteFunc) tsp_angle_harmonic_delete;

  potential->k = k;
  potential->theta_eq = theta_eq;

  return potential;
}

void tsp_angle_harmonic_delete(AngleHarmonicPotential* potential)
{
  ts_potential_delete((Potential *)potential);
}

static TSResult dihedral_get_params(DihedralPotential* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 4, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "k";      params[0].value = potential->k;
  params[1].name = "phi_eq"; params[1].value = potential->phi_eq;
  params[2].name = "func";   params[2].value = (double)potential->func;
  params[3].name = NULL;     params[3].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static TSResult dihedral_set_param(DihedralPotential* potential, const struct ts_potential_params* param)
{
  if ( strcmp(param->name, "k") == 0) {
    potential->k = param->value;
  } else if ( strcmp(param->name, "phi_eq") == 0 ) {
    potential->phi_eq= param->value;
  } else if ( strcmp(param->name, "func") == 0 ) {
    potential->func = (enum DIHEDRAL_FUNC)param->value;
  } else {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  tsp_dihedral_set_params(potential, potential->k, potential->phi_eq, potential->func);

  return TS_OK;
}

static dbldbl dihedral_multiple_forces(DihedralPotential* potential, struct ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n);
static double dihedral_forces(DihedralPotential* potential, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static double dihedral_energy(DihedralPotential* potential, vector3 r[NBODY_LIMIT]);

DihedralPotential* tsp_dihedral_new(enum DIHEDRAL_FUNC func, double k, double phi_eq)
{
  DihedralPotential* potential = ts_alloc(NO_ALLOCATOR, sizeof(DihedralPotential) );
  ts_potential_init((Potential*)potential, "DIHEDRAL", FOUR_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc)dihedral_energy;
  potential->parent.forces_func = (PotentialForcesFunc)dihedral_forces;
  potential->parent.multiple_forces_func = (PotentialMultipleForcesFunc) dihedral_multiple_forces;
  potential->parent.get_params_func = (PotentialGetParametersFunc) dihedral_get_params;
  potential->parent.set_param_func = (PotentialSetParameterFunc) dihedral_set_param;

  potential->parent.delete_func = (PotentialDeleteFunc) tsp_dihedral_delete;

  potential->k = k;
  potential->phi_eq = phi_eq;
  potential->func = func;

  return potential;
}

void tsp_dihedral_set_params(DihedralPotential* potential, double k, double phi_eq, enum DIHEDRAL_FUNC func)
{
  potential->k = k;
  potential->phi_eq = phi_eq;
  potential->func = func;
}

void tsp_dihedral_delete(DihedralPotential* potential)
{
  ts_potential_delete((Potential *)potential);
}

static inline double maxval(double a, double b)
{
  return a > b ? a : b;
}


/* This code was adapted from the DL_POLY_CLASSIC dihedral routine
 * (original Fortran shown in the comments), which is released under a
 * BSD license
 */
static inline dbldbl calc_dihedral_forces(DihedralPotential* potential, size_t n,
    const vector3*restrict rv, vector3*restrict fv)
{
  enum DIHEDRAL_FUNC func = potential->func;
  double phi_eq = potential->phi_eq;
  double k = potential->k;
  dbldbl en_tot = zero_dbldbl();

  for(size_t i=0; i < n; i++) {
    // define components of bond vectors

    // xdab(ii)=xxx(ia)-xxx(ib)
    // ydab(ii)=yyy(ia)-yyy(ib)
    // zdab(ii)=zzz(ia)-zzz(ib)

    vector3 d_ab;
    vector3_sub(&d_ab, rv+0,rv+1);

    // xdbc(ii)=xxx(ib)-xxx(ic)
    // ydbc(ii)=yyy(ib)-yyy(ic)
    // zdbc(ii)=zzz(ib)-zzz(ic)

    vector3 d_bc;
    vector3_sub(&d_bc, rv+1,rv+2);

    // xdcd(ii)=xxx(ic)-xxx(id)
    // ydcd(ii)=yyy(ic)-yyy(id)
    // zdcd(ii)=zzz(ic)-zzz(id)

    vector3 d_cd;
    vector3_sub(&d_cd, rv+2,rv+3);

    rv += 4; // dihedral potentials are 4-body potentials

    // define components of bond vectors
    double rrbc=1.0/vector3_norm(&d_bc);

    // construct first dihedral vector

    // pbx=yab*zbc-zab*ybc
    // pby=zab*xbc-xab*zbc
    // pbz=xab*ybc-yab*xbc
    // pb2=pbx*pbx+pby*pby+pbz*pbz
    // rpb1=1.d0/sqrt(pb2)
    // rpb2=rpb1*rpb1

    vector3 pb;
    vector3_cross(&pb, &d_ab, &d_bc);
    double pb2=vector3_sqnorm(&pb); // pbx*pbx+pby*pby+pbz*pbz;
    double rpb1=1.0/sqrt(pb2);
    double rpb2=rpb1*rpb1;

    // construct second dihedral vector

    // pcx=ybc*zcd-zbc*ycd
    // pcy=zbc*xcd-xbc*zcd
    // pcz=xbc*ycd-ybc*xcd
    // pc2=pcx*pcx+pcy*pcy+pcz*pcz
    // rpc1=1.d0/sqrt(pc2)
    // rpc2=rpc1*rpc1

    vector3 pc;
    vector3_cross(&pc, &d_bc, &d_cd);
    double pc2=vector3_sqnorm(&pc); // pcx*pcx+pcy*pcy+pcz*pcz;
    double rpc1=1.0/sqrt(pc2);
    double rpc2=rpc1*rpc1;

    // determine dihedral angle 

    //    pbpc=pbx*pcx+pby*pcy+pbz*pcz
    //    cosp=pbpc*rpb1*rpc1
    //    sinp=(xbc*(pcy*pbz-pcz*pby)+ybc*(pbx*pcz-pbz*pcx)+
    // x    zbc*(pcx*pby-pcy*pbx))*(rpb1*rpc1*rrbc)

    //    phi=atan2(sinp,cosp)

    double pbpc=vector3_dot(&pb,&pc); // pbx*pcx+pby*pcy+pbz*pcz;
    double cosp=pbpc*rpb1*rpc1;
    double sinp=vector3_scalar_triple(&d_bc,&pc,&pb)*rpb1*rpc1*rrbc;
    // (xbc*(pcy*pbz-pcz*pby)+ybc*(pbx*pcz-pbz*pcx)+
    // zbc*(pcx*pby-pcy*pbx))*(rpb1*rpc1*rrbc);

    double phi=atan2(sinp,cosp);

    // avoid singularity in sinp

    sinp=copysign(maxval(1e-8,fabs(sinp)),sinp);
    double rsinp=1.0/sinp;

    // selection of potential energy function type
    //
    // calculate potential energy and scalar force term
    // 
    // SFS: Currently, only a modified form of the torsion dihedral
    // potential is used.
    //
    // key=1 for torsion dihedral potential
    //

    //      dterm=prmdih(kk,1)*(1.d0+cos(prmdih(kk,3)*phi-
    // x      prmdih(kk,2)))
    //      gamma=-rpb1*rpc1*rsinp*prmdih(kk,1)*prmdih(kk,3)*
    // x      sin(prmdih(kk,3)*phi-prmdih(kk,2))

    double dterm = 0.0;
    double gamma = 0.0;

    if (func == DIHEDRAL_TORSIONAL) {
      dterm=(k) * (1.0 - cos(phi - phi_eq));
      gamma=+rpb1*rpc1*rsinp*(k)*sin(phi - phi_eq);
    } else if (func == DIHEDRAL_HARMONIC) {
      dterm = phi - phi_eq;
      gamma=+rpb1*rpc1*rsinp*(k)*dterm;
      dterm = 0.5 * k * dterm * dterm;
    }

    en_tot = accum_dbldbl(en_tot, dterm);

    if (fv == NULL) {
      continue;
    }

    //  <cut out a lot of the DL_POLY code that is not relevant to our
    //  case...

    // calculate atomic forces for dihedral term

    // fax=gamma*((-pcy*zbc+pcz*ybc)-pbpc*rpb2*(-pby*zbc+pbz*ybc))
    // fay=gamma*(( pcx*zbc-pcz*xbc)-pbpc*rpb2*( pbx*zbc-pbz*xbc))
    // faz=gamma*((-pcx*ybc+pcy*xbc)-pbpc*rpb2*(-pbx*ybc+pby*xbc))

    vector3 vtmp;

    vector3 dfa;
    vector3_cross(&dfa, &d_bc, &pc);
    vector3_cross(&vtmp, &d_bc, &pb);
    vector3_add_multiple_to(&dfa, -pbpc*rpb2, &vtmp);
    vector3_scaleby(&dfa,gamma);

    // fcx=gamma*((-pcy*zab+pcz*yab)-pbpc*rpb2*(-pby*zab+pbz*yab))
    // fcy=gamma*(( pcx*zab-pcz*xab)-pbpc*rpb2*( pbx*zab-pbz*xab))
    // fcz=gamma*((-pcx*yab+pcy*xab)-pbpc*rpb2*(-pbx*yab+pby*xab))

    vector3 dfc;
    vector3_cross(&dfc, &d_ab, &pc);
    vector3_cross(&vtmp, &d_ab, &pb);
    vector3_add_multiple_to(&dfc, -pbpc*rpb2, &vtmp);
    vector3_scaleby(&dfc,gamma);

    // fb1x=gamma*((-pby*zcd+pbz*ycd)-pbpc*rpc2*(-pcy*zcd+pcz*ycd))
    // fb1y=gamma*(( pbx*zcd-pbz*xcd)-pbpc*rpc2*( pcx*zcd-pcz*xcd))
    // fb1z=gamma*((-pbx*ycd+pby*xcd)-pbpc*rpc2*(-pcx*ycd+pcy*xcd))

    vector3 dfb1;
    vector3_cross(&dfb1, &d_cd, &pb);
    vector3_cross(&vtmp, &d_cd, &pc);
    vector3_add_multiple_to(&dfb1, -pbpc*rpc2, &vtmp);
    vector3_scaleby(&dfb1,gamma);

    // fd1x=gamma*((-pby*zbc+pbz*ybc)-pbpc*rpc2*(-pcy*zbc+pcz*ybc))
    // fd1y=gamma*(( pbx*zbc-pbz*xbc)-pbpc*rpc2*( pcx*zbc-pcz*xbc))
    // fd1z=gamma*((-pbx*ybc+pby*xbc)-pbpc*rpc2*(-pcx*ybc+pcy*xbc))

    vector3 dfd1;
    vector3_cross(&dfd1, &d_bc, &pb);
    vector3_cross(&vtmp, &d_bc, &pc);
    vector3_add_multiple_to(&dfd1, -pbpc*rpc2, &vtmp);
    vector3_scaleby(&dfd1, gamma);

    // fxx(ia)=fxx(ia)+fax
    // fyy(ia)=fyy(ia)+fay
    // fzz(ia)=fzz(ia)+faz
    vector3_set(fv+0, &dfa);

    // fxx(ib)=fxx(ib)-fax-fcx+fb1x
    // fyy(ib)=fyy(ib)-fay-fcy+fb1y
    // fzz(ib)=fzz(ib)-faz-fcz+fb1z
    vector3_set(fv+1, &dfb1);
    vector3_sub_from(fv+1,&dfa);
    vector3_sub_from(fv+1,&dfc);

    // fxx(ic)=fxx(ic)+fcx-fb1x-fd1x
    // fyy(ic)=fyy(ic)+fcy-fb1y-fd1y
    // fzz(ic)=fzz(ic)+fcz-fb1z-fd1z
    vector3_set(fv+2, &dfc);
    vector3_sub_from(fv+2,&dfb1);
    vector3_sub_from(fv+2,&dfd1);

    // fxx(id)=fxx(id)+fd1x
    // fyy(id)=fyy(id)+fd1y
    // fzz(id)=fzz(id)+fd1z
    vector3_set(fv+3, &dfd1);

    fv += 4;
  }

  return en_tot;
}


static double dihedral_forces(DihedralPotential* potential, vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  dbldbl en = calc_dihedral_forces(potential, 1, r+0, f+0);
  return dbldbl_to_double(en);
}

static double dihedral_energy(DihedralPotential* potential, vector3 r[NBODY_LIMIT])
{
  dbldbl en = calc_dihedral_forces(potential, 1, r+0, NULL); // f+0);
  return dbldbl_to_double(en);
}

static dbldbl dihedral_multiple_forces(DihedralPotential* potential, struct ts_sim* sim,
    struct ts_potential_term_list* termlist, size_t off, size_t n)
{
  UNUSED_PARAM(sim);
  if (n == 0) { return zero_dbldbl(); }

  struct ts_potential_term_data* term_data = 
    DATAVECTOR_AT(struct ts_potential_term_data,&termlist->term_data,off);

  vector3*restrict posv   = DATAVECTOR_BEGIN(vector3, &termlist->posv);
  vector3*restrict forcev = DATAVECTOR_BEGIN(vector3, &termlist->forcev);

  posv   += term_data->vec_off;
  forcev += term_data->vec_off;

  dbldbl en = calc_dihedral_forces(potential, n, posv+0, forcev+0);

  term_data->term_energy = en;

  return en;
}

static double dipole_energy(DipolePotential* potential, const vector3 r[NBODY_LIMIT]);
static double dipole_forces(DipolePotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT]);
static TSResult dipole_get_params(DipolePotential* potential, struct ts_potential_params** params_ptr);
static TSResult dipole_set_param(DipolePotential* potential, struct ts_potential_params* param);
DipolePotential* tsp_dipole_new(double k, double theta_eq)
{
  DipolePotential* potential = ts_alloc(NO_ALLOCATOR, sizeof(*potential) );
  ts_potential_init((Potential*)potential, "DIPOLE", FOUR_BODY);

  potential->parent.energy_func = (PotentialEnergyFunc)dipole_energy;
  potential->parent.forces_func = (PotentialForcesFunc)dipole_forces;
  potential->parent.get_params_func = (PotentialGetParametersFunc) dipole_get_params;
  potential->parent.set_param_func = (PotentialSetParameterFunc) dipole_set_param;

  potential->k = k;
  potential->theta_eq = theta_eq;

  return potential;
}

static double dipole_energy(DipolePotential* potential, const vector3 r[NBODY_LIMIT])
{
  vector3 f[NBODY_LIMIT];
  return dipole_forces(potential, r, f);
}

static double dipole_forces(DipolePotential* potential, const vector3 r[NBODY_LIMIT], vector3 f[NBODY_LIMIT])
{
  vector3 ua, ub;
  vector3_sub(&ua, r+0, r+1);
  vector3_sub(&ub, r+2, r+3);

  double mag_a = vector3_normalize(&ua);
  double mag_b = vector3_normalize(&ub);
  if (mag_a < VEC_TOL) { vector3_zero(&ua); mag_a = 1.0; }
  if (mag_b < VEC_TOL) { vector3_zero(&ub); mag_b = 1.0; }

  double rmaga = 1.0 / mag_a;
  double rmagb = 1.0 / mag_b;

  double cosa = vector3_dot(&ua, &ub);
  double sina = sqrt( 1.0 - cosa*cosa );

  if (sina < 1e-8) sina = 1e-8;  // avoid singularity in sin_theta
  double rsina = 1.0 / sina;

  double theta = acos(cosa);

  double dtheta = regularize_angle(theta - potential->theta_eq);
  double fmag = potential->k * (dtheta) * rsina;
  dtheta = 0.5*potential->k * dtheta * dtheta;

  vector3_scale_and_add(f+0,  1.0, &ub, -cosa, &ua);
  vector3_scaleby(f+0, fmag*rmaga);
  vector3_set_negative(f+1, f+0);

  vector3_scale_and_add(f+2,  1.0, &ua, -cosa, &ub);
  vector3_scaleby(f+2, fmag*rmagb);
  vector3_set_negative(f+3, f+2);

  return dtheta;
}

static TSResult dipole_get_params(DipolePotential* potential, struct ts_potential_params** params_ptr)
{
  struct ts_potential_params* params = ts_calloc(NO_ALLOCATOR, 3, sizeof(*params));
  ERROR_IF( TS_ERR_OUT_OF_MEMORY, params == NULL);

  params[0].name = "k";      params[0].value = potential->k;
  params[1].name = "theta_eq"; params[1].value = potential->theta_eq;
  params[3].name = NULL;     params[3].value = 0.0;

  *params_ptr = params;
  return TS_OK;
}

static TSResult dipole_set_param(DipolePotential* potential, struct ts_potential_params* param)
{
  if ( strcmp(param->name, "k") == 0)
  {
    potential->k = param->value;
  } 
  else if ( strcmp(param->name, "theta_eq") == 0 )
  {
    potential->theta_eq= param->value;
  } 
  else
  {
    return TS_NEW_ERROR( TS_ERR_UNKNOWN_POTENTIAL_PARAM );
  }

  tsp_dipole_set_params(potential, potential->k, potential->theta_eq);

  return TS_OK;
}

void tsp_dipole_set_params(DipolePotential* potential, double k, double theta_eq)
{
  potential->k = k;
  potential->theta_eq = theta_eq;
}

void tsp_dipole_delete(DipolePotential* potential)
{
  ts_potential_delete((Potential *)potential);
}

static inline MAYBE_UNUSED
void zero_vector_set(int n, vector3** vv)
{
  for (int i=0; i < n; i++) { vector3_zero(vv[i]); }
}

static inline MAYBE_UNUSED
void copy_vector_set(int n, vector3** dv, vector3** sv)
{
  for (int i=0; i < n; i++) { vector3_set(dv[i],sv[i]); }
}

static inline MAYBE_UNUSED
void diff_vector_set(int n, double scl, vector3** v1, vector3** v0)
{
  for (int i=0; i < n; i++) {
    vector3_sub_from(v1[i], v0[i]);
    vector3_scale_by(v1[i], scl);
  }
}

void ts_potential_hessian(Potential* pot, vector3 r[], VecReal* hess)
{
  if (pot->hessian_func != NULL) {
    pot->hessian_func(pot,r,hess);
  } else {
    ts_potential_numerical_hessian(pot, r, hess);
  }
}

struct potential_single_coord {
  Potential* potential;

  int pind;
  int cind;

  int find;
  int fpind;

  vector3 r[NBODY_LIMIT];
  vector3 f[NBODY_LIMIT];
};

static MAYBE_UNUSED double potential_at_one_coord(struct potential_single_coord* closure, double x)
{
  VecReal x0 = VECTOR3_ELT(closure->r+ closure->pind, closure->cind);

  VECTOR3_ELT(closure->r + closure->pind, closure->cind) = x;
  VecReal en = closure->potential->energy_func(closure->potential, closure->r);
  VECTOR3_ELT(closure->r + closure->pind, closure->cind) = x0;

  return en;
}

static MAYBE_UNUSED double forces_at_one_coord(struct potential_single_coord* closure, double x)
{
  for(int i=0; i < closure->potential->nbodies; i++) {
    vector3_zero(closure->f+i);
  }

  VecReal x0 = VECTOR3_ELT(closure->r+ closure->pind, closure->cind);

  VECTOR3_ELT(closure->r + closure->pind, closure->cind) = x;
  closure->potential->forces_func(closure->potential, closure->r, closure->f);
  VECTOR3_ELT(closure->r + closure->pind, closure->cind) = x0;

  return VECTOR3_ELT(closure->f + closure->fpind, closure->find);
}

#define H0 0.001

struct num_hess_closure
{
  Potential* pot;
  vector3 r[NBODY_LIMIT];

  int body1;
  int dof1;

  int body2;
  int dof2;
};

static double numhess_potential_at_dof(struct num_hess_closure* cl, double x)
{
  double*restrict dofp = &( VECTOR3_ELT(cl->r + cl->body1, cl->dof1) );

  double x0 = *dofp;

  *dofp = x;
  double en = cl->pot->energy_func(cl->pot, cl->r);
  *dofp = x0;

  return en;
}

static double numhess_potential_at_two_dof(struct num_hess_closure* cl, double x, double y)
{
  double*restrict dof1p = &( VECTOR3_ELT(cl->r + cl->body1, cl->dof1) );
  double*restrict dof2p = &( VECTOR3_ELT(cl->r + cl->body2, cl->dof2) );

  double x0 = *dof1p;
  double y0 = *dof2p;

  *dof1p = x;
  *dof2p = y;
  double en = cl->pot->energy_func(cl->pot, cl->r);
  *dof1p = x0;
  *dof2p = y0;

  return en;
}

void ts_potential_numerical_hessian(Potential* pot, vector3 r0[NBODY_LIMIT], VecReal* hess)
{
  struct num_hess_closure cl;
  cl.pot = pot;

  size_t num_dof = 3*pot->nbodies;

  memset(cl.r+0, 0, sizeof(cl.r));
  memset(hess, 0, sizeof(hess[0]) * hessian_size(num_dof));

  for (int i=0; i < pot->nbodies; i++) { vector3_set(cl.r+i, r0+i); }

  double err = 0.0;
  for (int i=0; i < pot->nbodies; i++) {
    cl.body1 = i;
    cl.body2 = i;

    for (int k=0; k < PARTICLE_NUM_DOF; k++) {
      size_t ind1 = PARTICLE_NUM_DOF*i + k;
      cl.dof1 = k;
      deriv2_ridders(
          (single_var_func)numhess_potential_at_dof,
          &cl, VECTOR3_ELT(r0+i,k), H0,
          hess + hessian_index(num_dof, ind1, ind1),
          &err);

      for (int l=(k+1); l < PARTICLE_NUM_DOF; l++) {
        size_t ind2 = PARTICLE_NUM_DOF*i + l;
        cl.dof2 = l;
        deriv_xy_ridders(
            (double_var_func)numhess_potential_at_two_dof,
            &cl, VECTOR3_ELT(r0+i,k), VECTOR3_ELT(r0+i,l), H0,
            hess + hessian_index(num_dof, ind1, ind2),
            &err);
      }
    }

    for (int j=(i+1); j < pot->nbodies; j++) {
      cl.body2 = j;

      for (int k=0; k < PARTICLE_NUM_DOF; k++) {
        size_t ind1 = PARTICLE_NUM_DOF*i + k;
        cl.dof1 = k;

        for (int l=0; l < PARTICLE_NUM_DOF; l++) {
          size_t ind2 = PARTICLE_NUM_DOF*j + l;
          cl.dof2 = l;

          deriv_xy_ridders(
              (double_var_func)numhess_potential_at_two_dof,
              &cl, VECTOR3_ELT(r0+i,k), VECTOR3_ELT(r0+j,l), H0,
              hess + hessian_index(num_dof, ind1, ind2),
              &err);
        }
      }
    }
  }
}


