#ifndef TUBESHEET_POTENTIALS__H
#define TUBESHEET_POTENTIALS__H

#include "ts_common.h"
#include "ts_errors.h"
#include "vector3.h"
#include "ts_particle.h"
#include "ts_numerical.h"

#define MODULE_NAME TUBESHEET_POTENTIALS

#define NBODY_LIMIT 4
#define PTYPES_INIT { 0,0,0,0 }

#define BT_NONE  bondtype_none()
#define NBOND_LIMIT 3
#define BTYPES_INIT { BT_NONE, BT_NONE, BT_NONE }

#define NONBOND_ENERGY_THRESH 100.0

#define DEFAULT_TERM_LIST_SIZE 1000

struct ts_sim;
typedef struct Potential Potential;
typedef struct ts_potential_term ts_potential_term;
struct ts_potential_term_list;

struct ts_potential_params
{
  const char* name;
  double value;
};

typedef double (*PotentialEnergyFunc)(Potential *potential, vector3 r[NBODY_LIMIT]);
typedef double (*PotentialForcesFunc)(Potential *potential, const vector3 *restrict r, vector3 *restrict f);
typedef double (*PotentialHessianFunc)(Potential *potential, vector3 r[NBODY_LIMIT], VecReal *hessian_upper_tri);
typedef dbldbl (*PotentialMultipleForcesFunc)(Potential *potential, struct ts_sim *sim, const struct ts_potential_term_list *termlist, size_t off, size_t n);
typedef TSResult (*PotentialGetParametersFunc)(Potential *potential, struct ts_potential_params **params_ptr);
typedef TSResult (*PotentialSetParameterFunc)(Potential *potential, const struct ts_potential_params *params_ptr);
typedef void (*PotentialDeleteFunc)(Potential *potential);

#define HESSIAN_SIZE(num_dof) ( SYMM_SIZE(num_dof) )
#define PARTICLE_HESSIAN_SZ  HESSIAN_SIZE(PARTICLE_NUM_DOF)
static inline size_t hessian_size(size_t num_dof) { return symm_size(num_dof); }
static inline size_t hessian_index(size_t num_dof, size_t i, size_t j)
{
  return symm_index(num_dof,i,j);
}

struct ts_bond_pattern;
struct ts_potential_term 
{
  const char* name;
  Potential* potential;
  const struct ts_bond_pattern* pattern;
  size_t particle_indices[NBODY_LIMIT];
};

struct ts_potential_term_data
{
  Potential* potential;
  dbldbl term_energy;
  size_t vec_off;
  int nbodies;
};

struct ts_potential_term_list
{
  DATAVECTOR(ts_potential_term) terms;
  DATAVECTOR(struct ts_potential_term_data) term_data;
  DATAVECTOR(size_t) potential_indices;
  DATAVECTOR(size_t) particle_indices;
  DATAVECTOR(vector3) posv;
  DATAVECTOR(vector3) forcev;
};

struct Potential
{
  const char* name;
  int nbodies;
  int num_refs;

  double potential_energy;

  PotentialEnergyFunc energy_func;
  PotentialForcesFunc forces_func;
  PotentialHessianFunc hessian_func;
  PotentialMultipleForcesFunc multiple_forces_func;

  PotentialGetParametersFunc get_params_func;
  PotentialSetParameterFunc set_param_func;

  PotentialDeleteFunc delete_func;
};

typedef struct DistanceConstraint DistanceConstraint;
struct DistanceConstraint
{
  Potential parent;
  double distance;
  double sqdist;
};

typedef struct ConstantField ConstantField;
struct ConstantField
{
  Potential parent;
  vector3 const_force;
  double ref_point;   // zero-point of potential
};

typedef struct SpinAroundParticle SpinAroundParticle;
struct SpinAroundParticle
{
  Potential parent;
  struct ts_sim* sim;
  size_t pind;

  vector3 axis;
};

typedef struct PlaneBarrier PlaneBarrier;
struct PlaneBarrier
{
  Potential parent;
  vector3 pos;
  vector3 norm;

  double dist;
  double eps;

  double sigma;
  double f_coeff;
};

typedef struct HarmonicPotential HarmonicPotential;
struct HarmonicPotential
{
  Potential parent;

  double k;
  double req;
};

typedef struct WCAPotential WCAPotential;
struct WCAPotential
{
  Potential parent;

  double sigma;
  double eps;

  double sigma_sq;
  double r_min;
  double r_min_sq;
  double f_coeff;
};

typedef struct AngleHarmonicPotential AngleHarmonicPotential;
struct AngleHarmonicPotential
{
  Potential parent;

  double k;
  double theta_eq;
};

enum DIHEDRAL_FUNC {
  DIHEDRAL_TORSIONAL = 1,
  DIHEDRAL_HARMONIC  = 2,
};

typedef struct DihedralPotential DihedralPotential;
struct DihedralPotential
{
  Potential parent;

  enum DIHEDRAL_FUNC func;

  double k;
  double phi_eq;
};

typedef struct DipolePotential DipolePotential;
struct DipolePotential
{
  Potential parent;

  double k;
  double theta_eq;
};


TSResult ts_potential_get_params(Potential* pot, struct ts_potential_params** params_ptr);
TSResult ts_potential_set_param(Potential* pot, struct ts_potential_params* param);

static inline TSResult
ts_potential_set_param_value(Potential* pot, const char* name, double value)
{
  struct ts_potential_params p = { .name = name, .value = value };
  return ts_potential_set_param(pot,&p);
}


void ts_potential_init(Potential* pot, const char* name, int nbodies);
void ts_potential_finalize(Potential* pot);

static inline Potential *ts_potential_incref_inner(Potential *pot)
{
  pot->num_refs++;
  return pot;
}

static inline Potential *ts_potential_incref_logged(Potential *pot,
    const char *fname, int line, const char *funcname)
{
  fprintf(stderr, "incref %p %d %s:%d @ %s\n",
      (void *)pot, pot->num_refs, fname, line, funcname);
  return ts_potential_incref_inner(pot);
}

static inline void ts_potential_decref_inner(Potential *pot)
{
  if (--pot->num_refs == 0)
  {
    pot->delete_func(pot);
  }
}

static inline void ts_potential_decref_logged(Potential *pot,
    const char *fname, int line, const char *funcname)
{
  fprintf(stderr, "decref %p %d %s:%d @ %s\n",
      (void *)pot, pot->num_refs, fname, line, funcname);
  ts_potential_decref_inner(pot);
}

#define POTENTIAL_REF_LOGGING 0
#if POTENTIAL_REF_LOGGING
#  define ts_potential_decref(pot) \
  (ts_potential_decref_logged((pot), __FILE__, __LINE__, __func__))
#  define ts_potential_incref(pot) \
  (ts_potential_incref_logged((pot), __FILE__, __LINE__, __func__))
#else
#  define ts_potential_incref(pot) (ts_potential_incref_inner((pot)))
#  define ts_potential_decref(pot) (ts_potential_decref_inner((pot)))
#endif

Potential* ts_potential_new(const char* name);
void ts_potential_delete(Potential *pot);

TSResult ts_potential_term_list_init(struct ts_potential_term_list* tl);
TSResult ts_potential_term_list_finalize(struct ts_potential_term_list* tl);

TSResult ts_potential_term_list_reset(struct ts_potential_term_list* tl);

static inline TSResult ts_potential_term_list_add(struct ts_potential_term_list* tl, ts_potential_term* term)
{
  tsAssert(tl->terms.elt_size == sizeof(*term));
  return datavector_add(&tl->terms, term) ? TS_OK : TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
}
TSResult ts_potential_term_list_finished_adding(struct ts_potential_term_list* tl);

TSResult ts_potential_forces(struct ts_sim* sim, struct ts_potential_term_data* termdata,
    const vector3*restrict rv, vector3*restrict fv, double* energy_p);

TSResult ts_potential_forces_from_list(struct ts_sim* sim,
    struct ts_potential_term_list* term_list, double*restrict energy_ptr, int termwise_energies);

TSResult ts_potential_energies_resum(struct ts_sim* sim, struct ts_potential_term_list* term_list, double energies[], int max_term_groups);

ConstantField* tsp_const_field_new(const vector3*restrict const_force, double ref_point);
void tsp_const_field_delete(ConstantField* potential);

SpinAroundParticle* tsp_spin_around_particle_new(struct ts_sim* sim, size_t pind, const vector3*restrict axis);
void tsp_spin_around_particle_delete(SpinAroundParticle* potential);

PlaneBarrier* tsp_plane_barrier_new(const vector3* pos, const vector3* norm, double dist, double eps);
void tsp_plane_barrier_set_params(PlaneBarrier* potential, double dist, double eps);
void tsp_plane_barrier_delete(PlaneBarrier* potential);

HarmonicPotential* tsp_harmonic_new(double k, double req);
void tsp_harmonic_delete(HarmonicPotential* potential);

DistanceConstraint* tsp_distance_constraint_new(double dist);
void tsp_distance_constraint_delete(DistanceConstraint* constraint);

WCAPotential* tsp_wca_new(double sigma, double eps);
void tsp_wca_set_params(WCAPotential* potential, double sigma, double eps);
void tsp_wca_delete(WCAPotential* potential);

AngleHarmonicPotential* tsp_angle_harmonic_new(double k, double theta_eq);
void tsp_angle_harmonic_delete(AngleHarmonicPotential* potential);

DihedralPotential* tsp_dihedral_new(enum DIHEDRAL_FUNC func, double k, double phi_eq);
void tsp_dihedral_set_params(DihedralPotential* potential, double k, double phi_eq, enum DIHEDRAL_FUNC func);
void tsp_dihedral_delete(DihedralPotential* potential);

DipolePotential* tsp_dipole_new(double k, double theta_eq);
void tsp_dipole_set_params(DipolePotential* potential, double k, double theta_eq);
void tsp_dipole_delete(DipolePotential* potential);


void ts_potential_hessian(Potential* pot, vector3 r[NBODY_LIMIT], VecReal* hess);
void ts_potential_numerical_hessian(Potential* pot, vector3 r[NBODY_LIMIT], VecReal* hess);

#undef MODULE_NAME

#endif /* TUBESHEET_POTENTIALS__H */

