#include "ts_rng.h"

#include <math.h>

#include "ts_common.h"
#include "ts_malloc_protect.h"

/* Merseinne-Twister 19937 RNG */
typedef struct
{
  dsfmt_t dsfmt_state;
  uint32_t seed;
} RNG_mt19937_state;

/* Test RNG */
typedef struct
{
  size_t seq_len;
  size_t seq_idx;
  double *seq_vals;
} RNG_test_state;

static TSResult RNG_mt19937_init(RNG_mt19937_state *state, uint32_t seed);
static TSResult RNG_mt19937(size_t n, double *vals, RNG_mt19937_state *state);
static TSResult RNG_mt19937_free(RNG_mt19937_state *state);
static TSResult RNG_mt19937_print(RNG_mt19937_state *state, FILE *f);

static void RNG_test_init(RNG_test_state *state, size_t n, double *vals);
static TSResult RNG_test(size_t n, double *vals, RNG_test_state *state);
static TSResult RNG_test_free(RNG_test_state *state);
static TSResult RNG_test_print(RNG_test_state *state, FILE *f);

TSResult ts_rng_print_configuration(const struct ts_rng *state, FILE *f)
{
  if ((state != NULL) && (state->rng_print)) {
    return state->rng_print(state->state, f);
  }
  return TS_OK;
}

TSResult ts_rng_free(struct ts_rng* state)
{
  if (state != NULL)
  {
    if ((state->gen_uniforms != NULL) 
        && (state->rng_free != NULL) 
        && (state->state != NULL)) 
    {
      state->rng_free(state->state);
    }

    state->gen_uniforms = NULL;
    state->rng_free = NULL;
    state->state = NULL;
  }

  return TS_OK;
}

TSResult ts_rng_gen_uniforms(struct ts_rng *state, size_t n, double *v)
{
  if (state == NULL) return TS_NEW_ERROR(TS_ERR_NO_RNG);
  if (state->gen_uniforms == NULL) return TS_NEW_ERROR(TS_ERR_NO_RNG);

  return state->gen_uniforms(n,v,state->state);
}

static void rng_state_finish_initialization(struct ts_rng *rng)
{
  rng->gauss_has_extra = 0;
  rng->gauss_extra = 0.0;
}

TSResult RNG_mt19937_new(uint32_t seed, struct ts_rng **rngp)
{
#define CLEANUP_ON_ERROR { \
  ts_safe_free(NO_ALLOCATOR, state);     \
  ts_safe_free(NO_ALLOCATOR, rng_state); \
}

  RNG_mt19937_state *state = ts_alloc(NO_ALLOCATOR, sizeof(*state));
  struct ts_rng *rng_state = ts_alloc(NO_ALLOCATOR, sizeof(*rng_state));

  ERROR_ON_NULL_ALLOC(state);
  ERROR_ON_NULL_ALLOC(rng_state);

  RETURN_ON_ERROR3( RNG_mt19937_init(state, seed) );

  rng_state->type = TS_RNG_MT19937;
  rng_state->gen_uniforms = (RNG_call_func)RNG_mt19937;
  rng_state->gen_normals  = NULL; // use default routines to generate normals
  rng_state->rng_free = (RNG_free_func)RNG_mt19937_free;
  rng_state->rng_print = (RNG_print_func)RNG_mt19937_print;
  rng_state->state = state;

  rng_state_finish_initialization(rng_state);

  *rngp = rng_state;

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult RNG_mt19937_get_info(struct ts_rng *state, uint32_t *seed)
{
  ERROR_IF(TS_ERR_WRONG_RNG, state->gen_uniforms != (RNG_call_func)RNG_mt19937);
  RNG_mt19937_state *mtState = state->state;

  ERROR_IF(TS_ERR_RNG_INCONSISTENT_STATE, mtState == NULL);

  if (seed) *seed = mtState->seed;
  return TS_OK;
}

TSResult RNG_mt19937_init(RNG_mt19937_state *state, uint32_t seed)
{
  dsfmt_init_gen_rand(&(state->dsfmt_state), seed);
  state->seed = seed;

  return TS_OK;
}

/* Merseinne-Twister 19937 RNG */
TSResult RNG_mt19937(size_t n, double* vals, RNG_mt19937_state* state)
{
  /* potential speed increase here... look into 
   * dsfmt_fill_array_open_close
   */
  for (size_t i=0; i < n; i++)
  {
    vals[i] = dsfmt_genrand_open_close( &(state->dsfmt_state) );
  }

  return TS_OK;
}

TSResult RNG_mt19937_free(RNG_mt19937_state *state)
{
  ts_free(NO_ALLOCATOR, state);
  return TS_OK;
}

static TSResult RNG_mt19937_print(RNG_mt19937_state *state, FILE *f)
{
  fprintf(f, "type: MT 19937  seed: %lu\n", (unsigned long) state->seed);
  return TS_OK;
}

TSResult RNG_test_new(size_t n, double *vals, struct ts_rng **rngp)
{
#define CLEANUP_ON_ERROR { \
  ts_safe_free(NO_ALLOCATOR, state); \
  ts_safe_free(NO_ALLOCATOR, rng_state); \
}

  RNG_test_state* state = ts_alloc(NO_ALLOCATOR, sizeof(*state));
  struct ts_rng* rng_state = ts_alloc(NO_ALLOCATOR, sizeof(*rng_state));
  ERROR_ON_NULL_ALLOC(state);
  ERROR_ON_NULL_ALLOC(rng_state);

  rng_state->type = TS_RNG_TEST;
  rng_state->gen_uniforms = (RNG_call_func)RNG_test;
  rng_state->gen_normals  = (RNG_call_func)RNG_test;   // same seq to generate normals (for testing)
  rng_state->rng_free = (RNG_free_func)RNG_test_free;
  rng_state->rng_print = (RNG_print_func)RNG_test_print;
  rng_state->state = state;

  rng_state_finish_initialization(rng_state);

  RNG_test_init(state, n, vals);
  *rngp = rng_state;

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult RNG_test_get_info(struct ts_rng *state, size_t *seq_len, size_t *seq_idx, double **seq_vals)
{
  ERROR_IF(TS_ERR_WRONG_RNG, state->gen_uniforms != (RNG_call_func)RNG_test);
  RNG_test_state* testState = state->state;

  ERROR_IF(TS_ERR_RNG_INCONSISTENT_STATE, testState == NULL);

  if (seq_len)  *seq_len  = testState->seq_len;
  if (seq_idx)  *seq_idx  = testState->seq_idx;
  if (seq_vals) *seq_vals = testState->seq_vals;

  return TS_OK;
}

void RNG_test_init(RNG_test_state *state, size_t n, double *vals)
{
  state->seq_len  = n;
  state->seq_vals = vals;
  state->seq_idx = 0;
}

/* Test RNG */
TSResult RNG_test(size_t n, double *vals, RNG_test_state *state)
{
  for (size_t i=0; i < n; i++)
  {
    vals[i] = state->seq_vals[ state->seq_idx ];
    state->seq_idx = (state->seq_idx + 1) % state->seq_len;
  }

  return TS_OK;
}

static TSResult RNG_test_print(RNG_test_state* state, FILE* f)
{
  fprintf(f, "type: Test RNG  ind: %d seq(%d):", (int)state->seq_idx, (int)state->seq_len);
  for(size_t i=0; i < state->seq_len; i++) {
    fprintf(f, " %7.4f", state->seq_vals[i]);
  }
  fprintf(f,"\n");

  return TS_OK;
}

TSResult RNG_test_free(RNG_test_state *state)
{
  ts_free(NO_ALLOCATOR, state->seq_vals);
  ts_free(NO_ALLOCATOR, state);

  return TS_OK;
}


#if 0
/*** CODE BELOW IS BY SFS (Thu Sep 15 14:22:19 CDT 2005) 
 * Taken from snippets by ARD (genrand.c) originally from Numerical
 * Recipes
 */
double genrand_res53(void) { return 0.0; }

/* generates a random number from the normal distribution N(0,1) */
double genrand_gaussian(void)
{
    static int iset = 0 ;
    static double gset ;
    double fac, rsq, v1, v2 ;
    if (iset == 0) {
        do {
            v1 = 2.0*genrand_res53()-1.0 ;
            v2 = 2.0*genrand_res53()-1.0 ;
            rsq = v1*v1 + v2*v2 ;
        } while (rsq >= 1.0 || rsq == 0.0) ;
        fac = sqrt(-2.0*log(rsq)/rsq) ;
        gset = v1*fac ;
        iset = 1 ;
        return v2*fac ;
    } else {
        iset=0 ;
        return gset ;
    }
}

/* generates a random unit vector, originally from Computer Simulations
 * of Liquids */
void genrand_unit( double* xp, double* yp, double* zp )
{
  double a,b,sq,t;

  for(;;) {
    a = genrand_res53();
    b = genrand_res53();

    a = 1 - 2*a;
    b = 1 - 2*b;
    sq = a*a + b*b;

    if (sq <= 1.0) {
      t = sqrt(1-sq);
      *xp = 2*a*t;
      *yp = 2*b*t;
      *zp = 1-2*sq;
      return;
    }
  }
}

/* generates a random unit vector in the positive z hemisphere, modified
 * from the unit vector code in Computer Simulations of Liquids */
void genrand_hemisphere( double* xp, double* yp, double* zp )
{
  double a,b,sq,t;

  for(;;) {
    a = genrand_res53();
    b = genrand_res53();

    a = 1 - 2*a;
    b = 1 - 2*b;
    sq = a*a + b*b;

    if (sq <= 0.5) {
      t = sqrt(1-sq);
      *xp = 2*a*t;
      *yp = 2*b*t;
      *zp = 1-2*sq;
      return;
    }
  }
}

void gaussian_bivariate(double *r1, double *r2, double s1, double s2, double c12)
{
  double v1 = genrand_gaussian();
  double v2 = genrand_gaussian();
  double c12x = sqrt(1.0-c12*c12);
  *r1 = s1*v1;
  *r2 = s2*(c12*v1+c12x*v2);
}
#endif /* 0 */

TSResult ts_rng_gen_normals(struct ts_rng* state, size_t n, double* v)
{
  if (state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }

  if (state->gen_normals == NULL) {
    return ts_rng_gen_normals_block(state, n, v);
  } else {
    return state->gen_normals(n,v,state->state);
  }
}

static TSResult generate_one_normal(struct ts_rng* state, double* out)
{
  double v[2];

  if (state->gauss_has_extra == 0) {
    double rsq;
    do {
      RETURN_ON_ERROR( ts_rng_gen_uniforms(state, 2, v) );

      v[0] = 2.0*v[0] - 1.0;
      v[1] = 2.0*v[1] - 1.0;

      rsq = v[0]*v[0] + v[1]*v[1];
    } while (rsq >= 1.0 || rsq == 0.0);
    double fac = sqrt(-2.0*log(rsq)/rsq);

    state->gauss_extra = v[0]*fac;
    state->gauss_has_extra = 1;

    *out = v[1]*fac;
    return TS_OK;
  } else {
    state->gauss_has_extra = 0;
    *out = state->gauss_extra;
    return TS_OK;
  }
}

TSResult ts_rng_gen_normals_sequential(struct ts_rng* state, size_t n, double* v)
{
  if (state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }

  for(size_t i=0; i < n; i++)
  {
    RETURN_ON_ERROR( generate_one_normal(state, v+i) );
  }

  return TS_OK;
}

#define ASSERT_NO_UNUSED(state) \
  tsAssert( (state->gauss_has_extra == 0) && \
      "cannot call ts_rng_gen_normals_block_* functions when there is an unused extra gaussian variable" )
static TSResult ts_rng_gen_normals_block_even_size(struct ts_rng* state, size_t n, double* v)
{
  if (state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }

  tsAssert( ((n % 2) == 0) && "ts_rng_gen_normals_block_even_size called but n is odd" );
  ASSERT_NO_UNUSED(state);

  RETURN_ON_ERROR( ts_rng_gen_uniforms(state, n, v) );

  double* uni = v;
  double* nrm = v;
  const double* end = v+n;
  while(nrm < end) {
    double rsq;
    double a,b;

    do {
      if (uni >= end) {
        RETURN_ON_ERROR( ts_rng_gen_uniforms(state, end-nrm, nrm) );
        uni = nrm;
      }

      a = *uni++;
      b = *uni++;

      a = 2.0*a - 1.0;
      b = 2.0*b - 1.0;

      rsq = a*a + b*b;
    } while (rsq >= 1.0 || rsq == 0.0);
    double fac = sqrt(-2.0*log(rsq)/rsq);

    *nrm++ = b*fac;
    *nrm++ = a*fac;
  }

  return TS_OK;
}

TSResult ts_rng_gen_normals_block_odd_size(struct ts_rng* state, size_t n, double* v)
{
  if (state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }
  tsAssert( ((n % 2) == 1) && "ts_rng_gen_normals_block_odd_size called but n is even" );
  ASSERT_NO_UNUSED(state);

  if (n == 1) return ts_rng_gen_normals_sequential(state, n, v);

  RETURN_ON_ERROR( ts_rng_gen_normals_block_even_size(state, n-1, v) );

  return ts_rng_gen_normals_sequential(state, 1, v+n-1);
}
#undef ASSERT_NO_UNUSED

TSResult ts_rng_gen_normals_block(struct ts_rng* state, size_t n, double* v)
{
  if (state == NULL) { return TS_NEW_ERROR(TS_ERR_NO_RNG); }
  if (state->gauss_has_extra) {
    RETURN_ON_ERROR( ts_rng_gen_normals_sequential(state, 1, v) );
    n--;
    v++;
  }

  if (n == 0) return TS_OK;

  if ((n % 2) == 0) {
    return ts_rng_gen_normals_block_even_size(state,n,v);
  } else {
    return ts_rng_gen_normals_block_odd_size(state,n,v);
  }
}


