#include "ts_shake.h"

#include "ts_bonds.h"
#include "ts_malloc_protect.h"

TSResult shake_initialize(ts_sim* sim)
{
  for(size_t i=0; i < sim->constraints.size; i++)
  {
    ts_constraint_entry* entry = DATAVECTOR_AT(ts_constraint_entry, &(sim->constraints), i);
    DistanceConstraint* constraint = (DistanceConstraint*) entry->constraint;

    double sqdist = vector3_sqdistance(PARTICLE_POS(sim, TS_PIND(sim,entry->pA)), PARTICLE_POS(sim, TS_PIND(sim,entry->pB)), &(entry->dr0));
    double delta = sqdist - constraint->sqdist;
    if (fabs(delta) > sim->shake_tol)
    {
      return TS_NEW_ERROR( TS_ERR_SHAKE_BOND_TOO_LONG );
    }

    vector3_set(&(entry->rA), PARTICLE_POS(sim, TS_PIND(sim,entry->pA))); 
    vector3_set(&(entry->rB), PARTICLE_POS(sim, TS_PIND(sim,entry->pB))); 
  }

  return TS_OK;
}

TSResult shake_iteration(ts_sim* sim, double* max_ptr, double* mean_ptr)
{
  double max_delta = 0.0;
  double mean_delta = 0.0;

  for(size_t i=0; i < sim->constraints.size; i++)
  {
    ts_constraint_entry* entry = DATAVECTOR_AT(ts_constraint_entry, &(sim->constraints), i);
    DistanceConstraint* constraint = (DistanceConstraint*) entry->constraint;

    vector3 dr;
    double sqdist = vector3_sqdistance(PARTICLE_POS(sim, TS_PIND(sim, entry->pA)), PARTICLE_POS(sim, TS_PIND(sim, entry->pB)), &dr);

    double delta = sqdist - constraint->sqdist;
    double abs_delta = fabs(delta);
    if ( abs_delta > sim->shake_tol)
    {
      ts_particle_type* ptype0 = entry->pA->type;
      ts_particle_type* ptype1 = entry->pB->type;

      double dr0_dr = vector3_dot(&dr, &(entry->dr0));
      double sum_inv_mass = ptype0->inv_mass + ptype1->inv_mass;

      double lambda_ij = 0.5*delta/(dr0_dr * sum_inv_mass);

      vector3_add_multiple_to( PARTICLE_POS(sim, TS_PIND(sim, entry->pA)),  lambda_ij*ptype0->inv_mass, &(entry->dr0) );
      vector3_add_multiple_to( PARTICLE_POS(sim, TS_PIND(sim, entry->pB)), -lambda_ij*ptype0->inv_mass, &(entry->dr0) );
    }

    if (abs_delta > max_delta) { max_delta = abs_delta; }
    mean_delta += abs_delta;
  }

  if (sim->constraints.size > 0)
  {
    mean_delta /= sim->constraints.size;
  }

  if (max_ptr  != NULL) { *max_ptr  = max_delta;  }
  if (mean_ptr != NULL) { *mean_ptr = mean_delta; }

  return TS_OK;
}

TSResult shake_adjust_velocities(ts_sim* sim)
{
  double inv_dt = 1.0/sim->dt;

  for(size_t i=0; i < sim->constraints.size; i++)
  {
    ts_constraint_entry* entry = DATAVECTOR_AT(ts_constraint_entry, &(sim->constraints), i);

    vector3_sub(PARTICLE_VEL(sim, TS_PIND(sim, entry->pA)), PARTICLE_POS(sim, TS_PIND(sim,entry->pA)), &(entry->rA));
    vector3_multiply_by(PARTICLE_VEL(sim, TS_PIND(sim,entry->pA)), inv_dt);
  }

  return TS_OK;
}

TSResult shake_apply_constraints(ts_sim* sim)
{
  double max_delta = 0.0;
  double mean_delta = 0.0;

  for(int i=0; i < MAX_SHAKE_ITERATIONS; i++)
  {
    RETURN_ON_ERROR( shake_iteration(sim, &max_delta, &mean_delta) );

    if (max_delta < sim->shake_tol)
    {
      return shake_adjust_velocities(sim);
    }
  }

  return TS_NEW_ERROR( TS_ERR_SHAKE_MAX_ITERATIONS );
}

