#include "ts_sparse.h"

#include "ts_numerical.h"

#include "tree.h"
#include "rcm.h"

#define ARPACK_TOL 1e-6

typedef struct bentry bentry;
struct bentry
{
  SPLAY_ENTRY(bentry) tree;

  size_t i;
  size_t j;
  double v;
};

SPLAY_HEAD(bpriv, bentry);

struct ts_sparse_builder_priv
{
  struct bpriv priv;
  size_t nnz;
};

static int entry_cmp(const bentry*restrict a, const bentry*restrict b)
{
  ptrdiff_t dj = a->j - b->j;
  if (dj==0) { dj  = a->i - b->i; }
  if (dj < 0) return -1;
  if (dj > 0) return  1;
  return 0;
}

SPLAY_PROTOTYPE( bpriv, bentry, tree, entry_cmp )
SPLAY_GENERATE(  bpriv, bentry, tree, entry_cmp )

static inline int tree_empty(struct ts_sparse_builder_priv* priv)
{
  return SPLAY_EMPTY(&priv->priv);
}

static inline bentry* tree_first_bentry(struct ts_sparse_builder_priv* priv)
{
  return SPLAY_MIN(bpriv, &priv->priv);
}

static inline bentry* tree_next_bentry(struct ts_sparse_builder_priv* priv, bentry* ent)
{
  return SPLAY_NEXT(bpriv, &priv->priv, ent);
}

static inline void tree_remove_bentry(struct ts_sparse_builder_priv* priv, bentry* ent)
{
  SPLAY_REMOVE(bpriv, &priv->priv, ent);
  priv->nnz--;
}

static inline void tree_insert_bentry(struct ts_sparse_builder_priv* priv, bentry* ent)
{
  SPLAY_INSERT(bpriv, &priv->priv, ent);
  priv->nnz++;
}

static inline void builder_priv_init(struct ts_sparse_builder_priv* prv)
{
  SPLAY_INIT(&prv->priv);
  prv->nnz = 0;
}

static struct ts_sparse_builder_priv* alloc_builder_priv(size_t m, size_t n)
{
  UNUSED_PARAM(m);
  UNUSED_PARAM(n);
  struct ts_sparse_builder_priv* prv = ts_alloc(NO_ALLOCATOR, sizeof(prv[0]));
  builder_priv_init(prv);
  return prv;
}

static void dealloc_builder_priv(struct ts_sparse_builder_priv* priv)
{
  if (!SPLAY_EMPTY(&priv->priv))
  {
    bentry* entry = SPLAY_MIN(bpriv, &priv->priv);
    while (entry)
    {
      bentry* next = SPLAY_NEXT(bpriv, &priv->priv, entry);
      tree_remove_bentry(priv, entry);
      ts_free(NO_ALLOCATOR, entry);
      entry = next;
    }
  }

  ts_free(NO_ALLOCATOR, priv);
}

TSResult ts_sparse_builder_initialize(ts_sparse_builder_t* builder, size_t m, size_t n)
{
  builder->m = m;
  builder->n = n;

  builder->priv = alloc_builder_priv(m,n);
  if (builder->priv == NULL) return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);

  return TS_OK;
}

void ts_sparse_builder_finalize(ts_sparse_builder_t* builder)
{
  dealloc_builder_priv( builder->priv );
}

double ts_sparse_builder_get(ts_sparse_builder_t* builder, size_t i,size_t j)
{
  bentry k = { .i = i, .j = j };
  bentry* v = SPLAY_FIND( bpriv, &builder->priv->priv, &k );
  return (v != NULL) ? (v->v) : 0.0;
}

static bentry* insert_bentry(struct ts_sparse_builder_priv* priv, size_t i, size_t j, double val)
{
  bentry* v = ts_alloc(NO_ALLOCATOR, sizeof(*v));
  if (v == NULL) return NULL;

  v->i = i;
  v->j = j;
  v->v = val;
  tree_insert_bentry(priv, v);

  return v;
}

TSResult ts_sparse_builder_set(ts_sparse_builder_t* builder, size_t i,size_t j, double val)
{
  bentry k = { .i = i, .j = j };
  bentry* v = SPLAY_FIND( bpriv, &builder->priv->priv, &k );

  if (v == NULL)
  {
    if (val != 0.0)
    {
      v = insert_bentry(builder->priv, i,j,val);
      if (v == NULL) return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    }
  }
  else
  {
    v->v = val;
  }

  return TS_OK;
}

TSResult ts_sparse_builder_add_to(ts_sparse_builder_t* builder, size_t i,size_t j, double val)
{
  if (val == 0.0) return TS_OK;

  bentry k = { .i = i, .j = j };
  bentry* v = SPLAY_FIND( bpriv, &builder->priv->priv, &k );

  if (v == NULL)
  {
    v = insert_bentry(builder->priv, i,j,val);
    if (v == NULL) return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  }
  else
  {
    v->v += val;
  }

  return TS_OK;
}

size_t ts_sparse_builder_nnz(ts_sparse_builder_t* builder)
{
  return builder->priv->nnz;
}

static inline void builder_load_iter(ts_sparse_iter* iter, bentry* ent)
{
  iter->data2 = ent;
  if (ent)
  {
    iter->i = ent->i;
    iter->j = ent->j;
    iter->v = ent->v;
  }
  else
  {
    iter->i = 0;
    iter->j = 0;
    iter->v = 0.0;
  }
}

void ts_sparse_builder_iter_init(ts_sparse_builder_t* builder, ts_sparse_iter* iter)
{
  iter->data1 = builder->priv;
  iter->data2 = NULL;
  iter->i = iter->j = 0;
  iter->v = 0.0;

  if (!SPLAY_EMPTY(&builder->priv->priv))
  {
    bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );
    builder_load_iter( iter, ent );
  }
}

int ts_sparse_builder_iter_has_next(ts_sparse_iter* iter)
{
  return iter->data2 != NULL;
}

void ts_sparse_builder_iter_advance(ts_sparse_iter* iter)
{
  struct ts_sparse_builder_priv* priv = iter->data1;
  bentry* ent = tree_next_bentry(priv, (bentry*) iter->data2);
  builder_load_iter( iter, ent );
}

static TSResult builder_build_arrays(ts_sparse_builder_t* builder,
    size_t nnz, size_t*restrict jc, size_t*restrict ir, double*restrict pr)
{
  memset(jc, 0, (builder->n+1)*sizeof(jc[0]));
  memset(ir, 0, (nnz)*sizeof(ir[0]));
  if (pr) { memset(pr, 0, (nnz)*sizeof(pr[0])); }

  if (!SPLAY_EMPTY(&builder->priv->priv))
  {
    bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );

    size_t n = builder->n;
    size_t ind = 0;
    size_t col = ent->j;

    while (ent)
    {
      if (ent->j != col)
      {
        size_t j0 = col+1;
        while ((j0 <= ent->j) && (j0 <= n))
        {
          jc[j0++] = ind;
        }
        col = ent->j;
      }

      ir[ind] = ent->i;
      if (pr) { pr[ind] = ent->v; }
      ind++;

      ent = SPLAY_NEXT(bpriv, &builder->priv->priv, ent);
    }

    ++col;
    while(col <= n)
    {
      jc[col++] = nnz;
    }
  }

  return TS_OK;
}

TSResult ts_sparse_builder_build(ts_sparse_builder_t* builder, ts_sparse_t* matrix)
{
  RETURN_ON_ERROR(
      ts_sparse_initialize(matrix,
        builder->m,
        builder->n,
        ts_sparse_builder_nnz(builder)) );

  RETURN_ON_ERROR(
      builder_build_arrays(builder, matrix->nnz,
        matrix->jc, matrix->ir, matrix->pr) );

  return TS_OK;
}

ASSERT_SIZE_EQ( long, size_t );
TSResult ts_sparse_builder_rcm(ts_sparse_builder_t* builder, size_t* perm)
{
  if (builder->m != builder->n)
  {
    return TS_NEW_ERROR3( TS_ERR_UNSUPPORTED_OPERATION,
        "banded representation only available for square matrices" );
  }

  struct ts_autopool pool;
  ts_autopool_init(&pool, NO_ALLOCATOR);

#define CLEANUP_ON_ERROR \
  { ts_autopool_release(&pool); }

  size_t nnz = ts_sparse_builder_nnz(builder);

  size_t* jc = ts_autopool_alloc(&pool, (builder->n+1)*sizeof(jc[0]));
  size_t* ir = ts_autopool_alloc(&pool, nnz*sizeof(jc[0]));

  RETURN_ON_ERROR( builder_build_arrays(builder, nnz, jc, ir, NULL) );

  long* deg = ts_autopool_alloc(&pool, builder->n*sizeof(deg[0]));
  signed char* mask = ts_autopool_alloc(&pool, builder->n*sizeof(mask[0]));

  memset(  deg, 0, builder->n * sizeof( deg[0]) );
  memset( mask, 0, builder->n * sizeof(mask[0]) );

  int flags = 0;
  genrcml(
      (long)builder->n,
      flags,
      (long*)jc,
      (long*)ir,
      (long*)perm,
      mask,
      deg);

  ts_autopool_release(&pool);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_sparse_builder_permute_self(ts_sparse_builder_t* origbld, size_t* perm)
{
#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR,revp); }

  size_t* revp = revp = ts_alloc(NO_ALLOCATOR, origbld->n * sizeof(revp[0]));
  for (size_t i=0; i < origbld->n; i++)
  {
    revp[ perm[i] ] = i;
  }

  struct ts_sparse_builder_priv priv;
  builder_priv_init(&priv);

  while (! tree_empty(origbld->priv))
  {
    bentry* ent = tree_first_bentry(origbld->priv);
    tree_remove_bentry(origbld->priv, ent);

    ent->i = revp[ent->i];
    ent->j = revp[ent->j];
    tree_insert_bentry(&priv, ent);
  }

  origbld->priv[0] = priv;

#undef CLEANUP_ON_ERROR

  ts_free(NO_ALLOCATOR, revp);

  return TS_OK;
}

TSResult ts_sparse_builder_permute(ts_sparse_builder_t* origbld,
    ts_sparse_builder_t* permbld, size_t* perm)
{
  RETURN_ON_ERROR(
      ts_sparse_builder_initialize(permbld, origbld->m, origbld->n) );

  size_t* revp = ts_alloc(NO_ALLOCATOR, origbld->n * sizeof(revp[0]));
  ERROR_ON_NULL_ALLOC(revp);

#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR,revp); }

  for (size_t i=0; i < origbld->n; i++)
  {
    revp[ perm[i] ] = i;
  }

  if (!SPLAY_EMPTY(&origbld->priv->priv))
  {
    bentry* ent = SPLAY_MIN( bpriv, &origbld->priv->priv );

    while (ent)
    {
      if (ent->v != 0.0)
      {
        RETURN_ON_ERROR3(
            ts_sparse_builder_set(permbld, revp[ent->i], revp[ent->j], ent->v) );
      }

      ent = SPLAY_NEXT(bpriv, &origbld->priv->priv, ent);
    }
  }

#undef CLEANUP_ON_ERROR

  ts_free(NO_ALLOCATOR, revp);

  return TS_OK;
}

static int get_outer_bands(ts_sparse_builder_t* builder, int bands[2])
{
  if (builder->m != builder->n) return -1;
  if (builder->m < 1) return 0;

  bands[0] = 0;
  bands[1] = 0;

  if (SPLAY_EMPTY(&builder->priv->priv)) return 1;

  bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );
  while(ent)
  {
    if (ent->v != 0.0)
    {
      int delta = ent->j - ent->i;
      if (delta > bands[0]) bands[0] = delta;
      if (delta < bands[1]) bands[1] = delta;
    }

    ent = SPLAY_NEXT(bpriv, &builder->priv->priv, ent);
  }

  bands[1] = -bands[1];

  return 1;
}

int ts_sparse_builder_get_bands(ts_sparse_builder_t* builder, int* bands)
{
  if (builder->m != builder->n) return -1;
  if (builder->m < 1) return 0;

  const ptrdiff_t m = builder->m;

  if (SPLAY_EMPTY(&builder->priv->priv)) return 0;

  memset(bands,0,(2*m-1)*sizeof(bands[0]));
  bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );
  ptrdiff_t min_band = 1-m;
  while(ent)
  {
    if (ent->v != 0.0)
    {
      ptrdiff_t delta = ent->j - ent->i;
      ptrdiff_t idx = delta-min_band;
      bands[idx] = 1;
    }

    ent = SPLAY_NEXT(bpriv, &builder->priv->priv, ent);
  }

  int nbands = 0;
  for (ptrdiff_t i=0; i < 2*m-1; i++)
  {
    if (bands[i])
    {
      bands[nbands] = i + min_band;
      ++nbands;
    }
  }

  return nbands;
}

TSResult ts_sparse_builder_as_banded(ts_sparse_builder_t* builder, int bands[2], double** elementsp, int which)
{
  int ret = get_outer_bands(builder,bands);
  switch (ret)
  {
    case -1:
      return TS_NEW_ERROR3(
          TS_ERR_UNSUPPORTED_OPERATION,
          "banded representation only available for square matrices");
      break;
    case  0:
      return TS_NEW_ERROR3(
          TS_ERR_INSUFFICIENT_SIZE, "matrix has invalid size");
      break;
    case  1:
      break;
    default:
      return TS_NEW_ERROR2(
          TS_ERR_INVALID_RETURN,
          ("unknown return %d from get_outer_bands", ret));
  }

  if      (which == -1) { bands[0] = 0; }
  else if (which ==  1) { bands[1] = 0; }

  int nb = bands[0] + bands[1] + 1;
  int ncol = builder->n;
  size_t nelts = nb * ncol;

#define CLEANUP_ON_ERROR { ts_safe_free(NO_ALLOCATOR, elts); }

  double*restrict elts = ts_calloc(NO_ALLOCATOR, nelts, sizeof(elts[0]));
  ERROR_ON_NULL_ALLOC(elts);

  memset(elts, 0, nelts*sizeof(elts[0]));
  *elementsp = elts;

  if (SPLAY_EMPTY(&builder->priv->priv))
  {
    return TS_OK;
  }

  bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );
  while(ent)
  {
    if (ent->v != 0.0)
    {
      int delta = ent->j - ent->i;
      if (delta*which >= 0)
      {
        int b = bands[0] - delta;
        elts[ nb*(ent->j) + b ] = ent->v;
      }
    }

    ent = SPLAY_NEXT(bpriv, &builder->priv->priv, ent);
  }

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_sparse_initialize(ts_sparse_t* matrix, size_t m, size_t n, size_t nnz)
{
  matrix->m = m;
  matrix->n = n;
  matrix->nnz = nnz;

  matrix->ir = NULL;
  matrix->jc = NULL;
  matrix->pr = NULL;

#define CLEANUP_ON_ERROR \
  { ts_safe_free( NO_ALLOCATOR, matrix->ir ); \
    ts_safe_free( NO_ALLOCATOR, matrix->jc ); \
    ts_safe_free( NO_ALLOCATOR, matrix->pr ); }

  matrix->ir = ts_alloc(NO_ALLOCATOR, nnz*sizeof(matrix->ir[0]));
  ERROR_ON_NULL_ALLOC(matrix->ir);

  matrix->jc = ts_alloc(NO_ALLOCATOR, (n+1)*sizeof(matrix->jc[0]));
  ERROR_ON_NULL_ALLOC(matrix->jc);

  matrix->pr = ts_alloc(NO_ALLOCATOR, nnz*sizeof(matrix->pr[0]));
  ERROR_ON_NULL_ALLOC(matrix->pr);

  memset(matrix->ir, 0, nnz*sizeof(matrix->ir[0]));
  memset(matrix->pr, 0, nnz*sizeof(matrix->pr[0]));
  memset(matrix->jc, 0, (n+1)*sizeof(matrix->jc[0]));

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

void ts_sparse_finalize(ts_sparse_t* matrix)
{
  ts_safe_free(NO_ALLOCATOR, matrix->ir);
  ts_safe_free(NO_ALLOCATOR, matrix->jc);
  ts_safe_free(NO_ALLOCATOR, matrix->pr);
}

static size_t* spbsearch(size_t elt, size_t* beg, size_t* end)
{
  if (beg == end) return NULL;
  if ((elt < *beg) || (elt > *(end-1))) return NULL;

  do
  {
    size_t* p = beg+(end-beg)/2;

    if (elt == *p)
    {
      return p;
    }
    else if (elt < *p)
    {
      end = p-1;
    }
    else
    {
      beg = p+1;
    }
  }
  while (beg <= end);

  return NULL;
}

double ts_sparse_get(ts_sparse_t* matrix, size_t i, size_t j)
{
  size_t ind0 = matrix->jc[j];
  size_t ind1 = matrix->jc[j+1];

  size_t* ip = spbsearch(i, matrix->ir+ind0, matrix->ir+ind1);
  return (ip != NULL) ? matrix->pr[ (ip - matrix->ir) ] : 0.0;
}

void ts_sparse_full(ts_sparse_t* matrix, double* elts, enum TS_MATRIXORDER order)
{
  memset(elts,0,matrix->m*matrix->n*sizeof(elts[0]));

  if (order == TS_ROWMAJOR)
  {
    for(size_t j=0; j < matrix->n; j++)
    {
      for (size_t rr = matrix->jc[j] ; rr < matrix->jc[j+1]; rr++)
      {
        elts[ matrix->ir[rr] + (matrix->m)*j ] = matrix->pr[rr];
      }
    }
  }
  else
  {
    for(size_t j=0; j < matrix->n; j++)
    {
      for (size_t rr = matrix->jc[j] ; rr < matrix->jc[j+1]; rr++)
      {
        elts[ j + (matrix->n)*(matrix->ir[rr]) ] = matrix->pr[rr];
      }
    }
  }
}

void ts_sparse_symm(ts_sparse_t* matrix, double* elts)
{
  memset(elts, 0, SYMM_SIZE(matrix->n)*sizeof(elts[0]));

  for(size_t j=0; j < matrix->n; j++)
  {
    for (size_t rr = matrix->jc[j] ; rr < matrix->jc[j+1]; rr++)
    {
      size_t i = matrix->ir[rr];
      if (i >= j)
      {
        elts[ symm_index(matrix->n,i,j)] = matrix->pr[rr];
      }
    }
  }
}

void ts_sparse_vec_rmul(ts_sparse_t* matrix, const double*restrict v0, double*restrict v1)
{
  memset(v1, 0, (matrix->m) * sizeof(v1[0]));

  for (size_t j=0; j < matrix->n; j++)
  {
    for (size_t rr=matrix->jc[j]; rr < matrix->jc[j+1]; rr++)
    {
      size_t i = matrix->ir[rr];
      v1[i] += v0[j] * matrix->pr[rr];
    }
  }
}

void ts_sparse_vec_lmul(ts_sparse_t* matrix, const double*restrict v0, double*restrict v1)
{
  memset(v1, 0, (matrix->n) * sizeof(v1[0]));

  for (size_t j=0; j < matrix->n; j++)
  {
    for (size_t rr=matrix->jc[j]; rr < matrix->jc[j+1]; rr++)
    {
      size_t i = matrix->ir[rr];
      v1[j] += matrix->pr[rr] * v0[i];
    }
  }
}

static TSResult sparse_eigenvalues_inner(
    struct ts_autopool* pool,
    struct ts_sparse_eigen* eigen)
{
  ts_sparse_t*restrict matrix = eigen->matrix;
  int nev = eigen->nev;

  if (matrix->m != matrix->n)
  {
    return TS_NEW_ERROR(TS_ERR_INVALID_MATRIX_SIZE);
  }

  memset(eigen->vals, 0, matrix->n * sizeof(eigen->vals[0]));
  if (eigen->vecs)
  {
    memset(eigen->vecs, 0, matrix->n * eigen->nev * sizeof(eigen->vecs[0]));
  }

  int n = matrix->m;
  int ncv = matrix->m;

  if (nev > n) { nev = n;}

  int pad0 = (nev == n);
  if (pad0) n++;

  if (ncv <= nev) { ncv = nev+1; }
  if (ncv >  n  ) { ncv = n;     }

  int lworkl = ncv*(ncv+8);

  double* resid = ts_autopool_alloc(pool,      n * sizeof(resid[0]));
  double* v     = ts_autopool_alloc(pool,  n*ncv * sizeof(v[0]    ));
  double* workd = ts_autopool_alloc(pool,    3*n * sizeof(workd[0]));
  double* workl = ts_autopool_alloc(pool, lworkl * sizeof(workl[0]));
  int* select   = ts_autopool_alloc(pool,    ncv * sizeof(select[0]));

  int rvec = 0;
  int ldz = 0;
  double* z = NULL;
  if (eigen->vecs) {
    rvec = 1;
    ldz = n;
    if (pad0)
    {
      z = ts_autopool_alloc(pool, n*nev * sizeof(double));
      memset(z, 0, n*nev * sizeof(z[0]));
    }
    else
    {
      z = eigen->vecs;
    }
  }

  memset(resid, 0,      n * sizeof(*resid));
  memset(v    , 0,  n*ncv * sizeof(*v));
  memset(workd, 0,    3*n * sizeof(*workd));
  memset(workl, 0, lworkl * sizeof(*workl));

  int ido = 0;
  const char* bmat = "I";
  const char* which = "LM";
  double tol = ARPACK_TOL; // 0.0;
  int info = 0;

  int iparam[11];
  int ipntr[11];

  iparam[0] = 1;                // ISHIFT
  iparam[2] = eigen->max_iter;  // max iterations
  iparam[3] = 1;                // NB (blocksize)
  iparam[6] = 1;                // solve A.x == lambda*x for symmetric A

  // parameters that are no longer used
  iparam[1] = 0;
  iparam[5] = 0;

  // outputs for iparam
  iparam[4] = 0;                // number of "converged" Ritz values
  iparam[7] = 0;                // NP: number of shifts the user is to provide
  iparam[8] = 0;                // NUMOP
  iparam[9] = 0;                // NUMOPB
  iparam[10] = 0;               // NUMREO

  memset(ipntr, 0, sizeof(ipntr));

  //long bmat_len = 1;
  //long which_len = 2;

  for(;;) {
    dsaupd(
        &ido,    // reverse communication flag for performing
                 // matrix operations (OP = matrix)
                 //  0 : input to first call
                 // -1 : compute Y = OP*X
                 //      X given by WORKD+IPNTR[0]-1
                 //      Y given by WORKD+IPNTR[1]-1
                 //
                 // +1 : same as -1 but vector B*X is already
                 //      available at WORKD+ipntr[2]-1
                 //
                 // +2 : compute Y = B*X
                 //      X given by WORKD+IPNTR[0]-1
                 //      Y given by WORKD+IPNTR[1]-1
                 //
                 // +3 : compute the IPARAM[7] shifts, and
                 //      place the results into workl+ipntr[10]
                 //
                 // 99 : done

        bmat,    // how matrix B is defined:
                 // 'I' -- std eigen problem: A.x == lambda*x
                 // 'G' -- gen eigen problem: A.x == lambda*B.x

        &n,      // dimension of the problem

        which,   // which values to compute:
                 // "SA" / "LA" -- smallest/largest algebraic
                 // "SM" / "LM" -- smallest/largest magnitude
                 // "BE"        -- compute half from each end

        &nev,    // number of eigenvalues to compute

        &tol,    // tolerance.  if (*tol == 0), then the machine
                 // precision is used

        resid,   // array of length N
                 //
                 // if INFO == 0, a random initial residual vector
                 // if INFO == 1, the initial residual vector
                 //                from a previous run

        &ncv,    // number of columns of matrix V
        v,       // N x NCV array containing the Lanczos basis vectors

        &n,      // leading dimension of V as declared in calling program

        iparam,  // iparam[0]    ISHIFT: method for selecting implicit shifts
                 //              ISHIFT = 0: the shifts are provided by the user via
                 //                          reverse communication.  The NCV eigenvalues of
                 //                          the current tridiagonal matrix T are returned in
                 //                          the part of WORKL array corresponding to RITZ.
                 //                          See remark 6 below.
                 //              ISHIFT = 1: exact shifts with respect to the reduced
                 //                          tridiagonal matrix T.  This is equivalent to
                 //                          restarting the iteration with a starting vector
                 //                          that is a linear combination of Ritz vectors
                 //                          associated with the "wanted" Ritz values.
                 // iparam[1]    LEVEC:  (no longer referenced)
                 // iparam[2]    MXITER: INPUT - max iterations
                 //                      OUTPUT - actual number of iterations
                 // iparam[3]    NB:     blocksize to use in recurrence (only works for NB=1)
                 // iparam[4]    NCONV:  OUTPUT - number of "converged" Ritz values
                 // iparam[5]    IUPD:   (no longer used)
                 // iparam[6]    MODE:   determines what type of eigenvalue problem is
                 //                      being solved
                 // iparam[7]    NP:     number of shifts the user is to provide
                 //
                 // iparam[8]    NUMOP   total # of OP*x operations
                 // iparam[9]    NUMOPB  total # of B*x operations (BMAT='G')
                 // iparam[10]   NUMREO  total # of steps of reorthogonalization

        ipntr,   // IPNTR[0]: pointer to the current operand vector X in WORKD.
                 // IPNTR[1]: pointer to the current result vector Y in WORKD.
                 // IPNTR[2]: pointer to the vector B * X in WORKD when used in
                 //           the shift-and-invert mode.
                 // IPNTR[3]: pointer to the next available location in WORKL
                 //           that is untouched by the program.
                 // IPNTR[4]: pointer to the NCV by 2 tridiagonal matrix T in WORKL.
                 // IPNTR[5]: pointer to the NCV RITZ values array in WORKL.
                 // IPNTR[6]: pointer to the Ritz estimates in array WORKL associated
                 //           with the Ritz values located in RITZ in WORKL.
                 // IPNTR[10]: pointer to the NP shifts in WORKL. See Remark 6 below.
                 //
                 // Note: IPNTR[7:9] is only referenced by dseupd . See Remark 2.
                 // IPNTR[7]: pointer to the NCV RITZ values of the original system.
                 // IPNTR[8]: pointer to the NCV corresponding error bounds.
                 // IPNTR[9]: pointer to the NCV by NCV matrix of eigenvectors
                 //           of the tridiagonal matrix T. Only referenced by
                 //           dseupd  if RVEC = .TRUE. See Remarks.

        workd,   // Work array of length 3*N.  Used in reverse communication interface.
                 // OUTPUT - workd[0] thru workd[N-1] have B*resid[0] .. resid[N-1]
                 // (dseupd uses this output if the Ritz vectors
                 // are desired)
        workl,   // Work array of length lworkl.  (NB: Used for distributed ARPACK)
        &lworkl, // Number of elements in workl.  Must be at least NCV*NCV + 8*NCV

        &info,   // INPUT: 0 - random initial residual vector is used
                 //        1 - RESID contains the residual vector
                 //            (possibly from a previous run)
                 //
                 // OUTPUT: Result of the calculation
                 //     =  0: Normal exit.
                 //     =  1: Maximum number of iterations taken.
                 //           All possible eigenvalues of OP has been found. IPARAM(5)
                 //           returns the number of wanted converged Ritz values.
                 //     =  2: No longer an informational error. Deprecated starting
                 //           with release 2 of ARPACK.
                 //     =  3: No shifts could be applied during a cycle of the
                 //           Implicitly restarted Arnoldi iteration. One possibility
                 //           is to increase the size of NCV relative to NEV.
                 //           See remark 4 below.
                 //     = -1: N must be positive.
                 //     = -2: NEV must be positive.
                 //     = -3: NCV must be greater than NEV and less than or equal to N.
                 //     = -4: The maximum number of Arnoldi update iterations allowed
                 //           must be greater than zero.
                 //     = -5: WHICH must be one of 'LM', 'SM', 'LA', 'SA' or 'BE'.
                 //     = -6: BMAT must be one of 'I' or 'G'.
                 //     = -7: Length of private work array WORKL is not sufficient.
                 //     = -8: Error return from trid. eigenvalue calculation;
                 //           Informatinal error from LAPACK routine dsteqr .
                 //     = -9: Starting vector is zero.
                 //     = -10: IPARAM(7) must be 1,2,3,4,5.
                 //     = -11: IPARAM(7) = 1 and BMAT = 'G' are incompatable.
                 //     = -12: IPARAM(1) must be equal to 0 or 1.
                 //     = -13: NEV and WHICH = 'BE' are incompatable.
                 //     = -9999: Could not build an Arnoldi factorization.
                 //              IPARAM(5) returns the size of the current Arnoldi
                 //              factorization. The user is advised to check that
                 //              enough workspace and array storage has been allocated.
        (long)strlen(bmat),
        (long)strlen(which));

    if ((ido == 1)||(ido == -1))
    {

      ts_sparse_vec_lmul(matrix, workd + ipntr[0]-1, workd + ipntr[1]-1);

      if (pad0) {
        workd[ ipntr[1]+n-1 ] = 0.0;
      }
    }
    else if (ido == 99)
    {
      break;
    }
    else
    {
      return TS_NEW_ERROR2(TS_ERR_IN_ALGORITHM,
          ("Unexpected ARPACK value: ido=%d", ido));
    }
  }

  eigen->num_iter      = iparam[2];
  eigen->num_converged = iparam[4];
  eigen->num_ops       = iparam[8];
  eigen->num_reorth    = iparam[10];

  if (info != 0) {
    return TS_NEW_ERROR2(TS_ERR_IN_ALGORITHM,
        ("ARPACK encountered error %d in routine dsaupd (check documentation)", info));
  }

  const char* howmny = "A";
  double sigma = 0; // should not be referenced
  for(int i=0; i < ncv; i++) { select[i] = 1; }

  dseupd(
      &rvec,
      howmny,
      select,
      eigen->vals,
      z,
      &ldz,
      &sigma,

      bmat,
      &n,
      which,
      &nev,
      &tol,
      resid,
      &ncv,
      v,
      &n,
      iparam,
      ipntr,
      workd,
      workl,
      &lworkl,
      &info,
      (long)strlen(howmny),
      (long)strlen(bmat),
      (long)strlen(which)
      );

  if (rvec)
  {
    eigen->max_resid = 0;
    for (int i=0; i < nev; i++)
    {
      // used workd as scratch space to save the vector
      ts_sparse_vec_lmul(matrix, z+n*i, workd);
      for (size_t j=0; j < matrix->n; j++)
      {
        double delta = z[n*i + j] * eigen->vals[i] - workd[j];
        eigen->max_resid += delta*delta;
      }
    }
    // fprintf(stderr, "\nmax residual is %7.4e\n", eigen->max_resid);
  }
  else
  {
    // if we don't calculate the eigenvectors, then we cannot determine
    // the maximum residual.  Instead, set max_resid to an invalid value
    // (-1.0).
    eigen->max_resid = -1.0;
  }

  if (rvec && pad0)
  {
    for (size_t j=0; j < matrix->n; j++)
    {
      double*restrict row0 = z + n*j;
      double*restrict row1 = eigen->vecs + matrix->n*j;
      for (size_t i=0; i < matrix->n; i++)
      {
        row1[i] = row0[i];
      }
    }
  }

  if (info != 0) {
    return TS_NEW_ERROR2(TS_ERR_IN_ALGORITHM,
        ("ARPACK encountered error %d in routine dseupd (check documentation)", info));
  }

  return TS_OK;
}

TSResult ts_sparse_eigenvalues_calc(struct ts_sparse_eigen* eigen)
{
  struct ts_autopool pool;
  ts_autopool_init(&pool, NO_ALLOCATOR);

  RETURN_ON_ERROR2(
      sparse_eigenvalues_inner(&pool, eigen),
      ts_autopool_release(&pool)
      );

  ts_autopool_release(&pool);

  return TS_OK;
}

void ts_sparse_builder_print(ts_sparse_builder_t* builder, FILE* f)
{
  fprintf(f, "sparse builder (%lu, %lu) nnz=%lu\n",
      builder->m,builder->n,ts_sparse_builder_nnz(builder));
  if (!SPLAY_EMPTY(&builder->priv->priv))
  {
    bentry* ent = SPLAY_MIN( bpriv, &builder->priv->priv );

    while (ent)
    {
      fprintf(f, "    (%4lu,%4lu)  %7.4f\n",
          ent->i, ent->j, ent->v);

      ent = SPLAY_NEXT(bpriv, &builder->priv->priv, ent);
    }
  }
}
