#ifndef TS_SPARSE_H
#define TS_SPARSE_H

#include <stddef.h>
#include <stdio.h>

#include "ts_errors.h"
#include "ts_common.h"
#include "ts_numerical.h"

#define MODULE_NAME TS_SPARSE

/* Sparse matrix builder 
 *
 * Used to efficiently build a matrix
 */
typedef struct ts_sparse_builder ts_sparse_builder_t;
struct ts_sparse_builder_priv;
struct ts_sparse_builder
{
  size_t m;
  size_t n;

  struct ts_sparse_builder_priv* priv;
};

TSResult ts_sparse_builder_initialize(ts_sparse_builder_t* builder, size_t m, size_t n);
void ts_sparse_builder_finalize(ts_sparse_builder_t* builder);
static inline size_t ts_sparse_builder_size(ts_sparse_builder_t* builder) { return builder->m*builder->n; }

double ts_sparse_builder_get(ts_sparse_builder_t* builder, size_t i,size_t j);
TSResult ts_sparse_builder_set(ts_sparse_builder_t* builder, size_t i,size_t j, double val);
TSResult ts_sparse_builder_add_to(ts_sparse_builder_t* builder, size_t i,size_t j, double val);
static inline TSResult ts_sparse_builder_sub_from(ts_sparse_builder_t* builder, size_t i, size_t j, double val)
{
  return ts_sparse_builder_add_to(builder, i,j, -val);
}

size_t ts_sparse_builder_nnz(ts_sparse_builder_t* builder);

void ts_sparse_builder_print(ts_sparse_builder_t* builder, FILE* f);

typedef struct ts_sparse_iter ts_sparse_iter;
struct ts_sparse_iter
{
  void* data1;
  void* data2;

  size_t i;
  size_t j;
  double v;
};

void ts_sparse_builder_iter_init(ts_sparse_builder_t* builder, ts_sparse_iter* iter);
int ts_sparse_builder_iter_has_next(ts_sparse_iter* iter);
void ts_sparse_builder_iter_advance(ts_sparse_iter* iter);

/* Sparse matrix structure */
typedef struct ts_sparse ts_sparse_t;
struct ts_sparse
{
  size_t m;
  size_t n;
  size_t nnz;

  size_t* ir;
  size_t* jc;

  double* pr;
};

TSResult ts_sparse_initialize(ts_sparse_t* matrix, size_t m, size_t n, size_t nnz);
void ts_sparse_finalize(ts_sparse_t* matrix);

TSResult ts_sparse_builder_build(ts_sparse_builder_t* builder, ts_sparse_t* matrix);

TSResult ts_sparse_builder_permute(ts_sparse_builder_t* origbld, 
    ts_sparse_builder_t* permbld, size_t* perm);

TSResult ts_sparse_builder_permute_self(ts_sparse_builder_t* origbld, size_t* perm);
TSResult ts_sparse_builder_rcm(ts_sparse_builder_t* origbld, size_t* perm);

/*! expects builder to be a square matrix (NxN), and 'bands' to have
 * 2N-1 elements
 */
int ts_sparse_builder_get_bands(ts_sparse_builder_t* builder, int* bands);

TSResult ts_sparse_builder_as_banded(ts_sparse_builder_t* builder, int bands[2], double** elementsp, int which);

double ts_sparse_get(ts_sparse_t* matrix, size_t i, size_t j);

void ts_sparse_full(ts_sparse_t* matrix, double* elts, enum TS_MATRIXORDER order);
void ts_sparse_symm(ts_sparse_t* matrix, double* elts);

// v1 = matrix * v0
void ts_sparse_vec_rmul(ts_sparse_t* matrix, 
    const double*restrict v0, double*restrict v1);

// v1 = v0 * matrix
void ts_sparse_vec_lmul(ts_sparse_t* matrix, 
    const double*restrict v0, double*restrict v1);

#define TS_SPARSE_EIGEN_MAXITER 300

/*! \brief Used set parameters of the sparse eigenvalue method, and to
 * return eigenvalues/eigenvectors and information about the algorithm.
 *
 * The field 'vals' must be not be NULL (and must point to valid
 * memory).  If the field 'vecs' is NULL, no eigenvectors are
 * calculated.
 *
 */
struct ts_sparse_eigen
{
  ts_sparse_t* matrix; /*! sparse matrix to determine eigenvalue */
  int nev;             /*! number of eigenvalues */
  double* vals;        /*! array of 'nev' right eigenvalues */
  double* vecs;        /*! matrix of 'nev' x 'matrix->n' eigenvectors */
  int max_iter;

  int num_iter;
  int num_converged;
  int num_ops;
  int num_reorth;

  double max_resid;
};

TSResult ts_sparse_eigenvalues_calc(struct ts_sparse_eigen* eigen);

TSResult ts_sparse_builder_eigensystem_banded(struct ts_autopool* pool,
    ts_sparse_builder_t* builder, double** eigvalsp, double** eigvecsp);

TSResult ts_sparse_eigensystem_sparse(struct ts_autopool* pool,
    ts_sparse_t* hess_mat, double** eigvalsp, double** eigvecsp, size_t skip);

TSResult ts_sparse_eigensystem_dense(struct ts_autopool* pool,
    ts_sparse_t* hess_mat, double** eigvalsp, double** eigvecsp);

#undef MODULE_NAME

#endif /* TS_SPARSE_H */

