#include "ts_sparse.h"
#include "ts_numerical.h"

TSResult ts_sparse_builder_eigensystem_banded(struct ts_autopool* pool,
    ts_sparse_builder_t* builder, double** eigvalsp, double** eigvecsp)
{
  int bands[2] = { 0, 0 };
  double* elts = NULL;
  int ndof = builder->m;

#define CLEANUP_ON_ERROR { ts_safe_free(NO_ALLOCATOR, elts); }

  size_t* perm = ts_autopool_alloc(pool, (ndof)*sizeof(pool[0]));
  ERROR_ON_NULL_ALLOC( perm );

  double* D = ts_autopool_alloc(pool, (ndof)*sizeof(D[0]));
  ERROR_ON_NULL_ALLOC( D );

  double* E = ts_autopool_alloc(pool, (ndof)*sizeof(E[0]));
  ERROR_ON_NULL_ALLOC( E );

  double* work  = ts_autopool_alloc(pool, ndof*sizeof(work[0]));
  ERROR_ON_NULL_ALLOC( work );

  /* FIXME: This should be fixed to work with eigenvectors */
  if (eigvecsp == NULL)
  {
    RETURN_ON_ERROR3( ts_sparse_builder_rcm(builder, perm) );
    RETURN_ON_ERROR3( ts_sparse_builder_permute_self(builder, perm) );
  }

  RETURN_ON_ERROR3( ts_sparse_builder_as_banded(builder, bands, &elts, 1) );

  int nb = 1 + bands[0];
  int ldab = nb;

  int ldq = 1;

  int info = 0;

  dsbtrd_(
      "N",        //  VECT    (input) CHARACTER*1
                  //          = 'N':  do not form Q;
                  //          = 'V':  form Q;
                  //          = 'U':  update a matrix X, by forming X*Q.

      "U",        //  UPLO    (input) CHARACTER*1
                  //          = 'U':  Upper triangle of A is stored;
                  //          = 'L':  Lower triangle of A is stored.

      &ndof,      //  N       (input) INTEGER
                  //          The order of the matrix A.  N >= 0.

      &bands[0],  //  KD      (input) INTEGER
                  //          The number of superdiagonals of the matrix A if UPLO = 'U',
                  //          or the number of subdiagonals if UPLO = 'L'.  KD >= 0.

      elts,       //  AB      (input/output) DOUBLE PRECISION array, dimension (LDAB,N)
                  //          On entry, the upper or lower triangle of the symmetric band
                  //          matrix A, stored in the first KD+1 rows of the array.  The
                  //          j-th column of A is stored in the j-th column of the array AB
                  //          as follows:
                  //          if UPLO = 'U', AB(kd+1+i-j,j) = A(i,j) for max(1,j-kd)<=i<=j;
                  //          if UPLO = 'L', AB(1+i-j,j)    = A(i,j) for j<=i<=min(n,j+kd).
                  //          On exit, the diagonal elements of AB are overwritten by the
                  //          diagonal elements of the tridiagonal matrix T; if KD > 0, the
                  //          elements on the first superdiagonal (if UPLO = 'U') or the
                  //          first subdiagonal (if UPLO = 'L') are overwritten by the
                  //          off-diagonal elements of T; the rest of AB is overwritten by
                  //          values generated during the reduction.

      &ldab,      //  LDAB    (input) INTEGER
                  //          The leading dimension of the array AB.  LDAB >= KD+1.

      D,          //  D       (output) DOUBLE PRECISION array, dimension (N)
                  //          The diagonal elements of the tridiagonal matrix T.

      E,          //  E       (output) DOUBLE PRECISION array, dimension (N-1)
                  //          The off-diagonal elements of the tridiagonal matrix T:
                  //          E(i) = T(i,i+1) if UPLO = 'U'; E(i) = T(i+1,i) if UPLO = 'L'.

      NULL,       //  Q       (input/output) DOUBLE PRECISION array, dimension (LDQ,N)
                  //          On entry, if VECT = 'U', then Q must contain an N-by-N
                  //          matrix X; if VECT = 'N' or 'V', then Q need not be set.
                  //
                  //          On exit:
                  //          if VECT = 'V', Q contains the N-by-N orthogonal matrix Q;
                  //          if VECT = 'U', Q contains the product X*Q;
                  //          if VECT = 'N', the array Q is not referenced.

      &ldq,       //   LDQ     (input) INTEGER
                  //           The leading dimension of the array Q.
                  //           LDQ >= 1, and LDQ >= N if VECT = 'V' or 'U'.

      work,       //  WORK    (workspace) DOUBLE PRECISION array, dimension (N)

      &info       //  INFO    (output) INTEGER
                  //          = 0:  successful exit
                  //          < 0:  if INFO = -i, the i-th argument had an illegal value
        );

  if (info != 0) {
    return TS_NEW_ERROR2(TS_ERR_INVALID_ARGUMENT, ("dsptrd failed (see arg %d)", -info));
  }

  // STEP 2: solve eigenvectors of tridiagonal matrix...
  double vl = 0.0;
  double vu = 0.0;
  int il = 0;
  int iu = 0;

  int neigen = 0;
  int ldz = 1; // ndof;

#undef CLEANUP_ON_ERROR

  double* eigvals = NULL;
  double* eigvecs = NULL;
#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR, elts); \
    ts_safe_free(NO_ALLOCATOR, eigvals); \
    ts_safe_free(NO_ALLOCATOR, eigvecs); }

  eigvals = ts_alloc(NO_ALLOCATOR, (ndof)*sizeof(eigvals[0]));
  ERROR_ON_NULL_ALLOC( eigvals );

  // eigvecs = ts_alloc(NO_ALLOCATOR, (ndof*ndof)*sizeof(eigvecs[0]));
  // ERROR_ON_NULL_ALLOC( eigvecs );

  int* isuppz = ts_autopool_alloc(pool, 2*ndof*sizeof(isuppz[0]));
  ERROR_ON_NULL_ALLOC( isuppz );

  int tryrac = 1;
  int lwork = 18*ndof;
  double* work2 = ts_autopool_alloc(pool, lwork*sizeof(work2[0]));

  int liwork = 10*ndof;
  int* iwork = ts_autopool_alloc(pool, liwork*sizeof(iwork[0]));

  int nzc = ndof;

  dstemr_( 
      "N",              //  JOBZ    (input) CHARACTER*1
                        //          = 'N':  Compute eigenvalues only;
                        //          = 'V':  Compute eigenvalues and eigenvectors.

      "A",              //  RANGE   (input) CHARACTER*1
                        //          = 'A': all eigenvalues will be found.
                        //          = 'V': all eigenvalues in the half-open interval (VL,VU]
                        //                 will be found.
                        //          = 'I': the IL-th through IU-th eigenvalues will be found.

    &ndof,              //  N       (input) INTEGER
                        //          The order of the matrix.  N >= 0.

    D,                  //  D       (input/output) DOUBLE PRECISION array, dimension (N)
                        //          On entry, the N diagonal elements of the tridiagonal matrix
                        //          T. On exit, D is overwritten.

    E,                  //  E       (input/output) DOUBLE PRECISION array, dimension (N)
                        //          On entry, the (N-1) subdiagonal elements of the tridiagonal
                        //          matrix T in elements 1 to N-1 of E. E(N) need not be set on
                        //          input, but is used internally as workspace.
                        //          On exit, E is overwritten.

    &vl,                //  VL      (input) DOUBLE PRECISION
    &vu,                //  VU      (input) DOUBLE PRECISION
                        //          If RANGE='V', the lower and upper bounds of the interval to
                        //          be searched for eigenvalues. VL < VU.
                        //          Not referenced if RANGE = 'A' or 'I'.

    &il,                //  IL      (input) INTEGER

    &iu,                //  IU      (input) INTEGER
                        //          If RANGE='I', the indices (in ascending order) of the
                        //          smallest and largest eigenvalues to be returned.
                        //          1 <= IL <= IU <= N, if N > 0.
                        //          Not referenced if RANGE = 'A' or 'V'.

    &neigen,            //  M       (output) INTEGER
                        //          The total number of eigenvalues found.  0 <= M <= N.
                        //          If RANGE = 'A', M = N, and if RANGE = 'I', M = IU-IL+1.

    eigvals,            //  W       (output) DOUBLE PRECISION array, dimension (N)
                        //          The first M elements contain the selected eigenvalues in
                        //          ascending order.

    eigvecs,            //  Z       (output) DOUBLE PRECISION array, dimension (LDZ, max(1,M) )
                        //          If JOBZ = 'V', and if INFO = 0, then the first M columns of Z
                        //          contain the orthonormal eigenvectors of the matrix T
                        //          corresponding to the selected eigenvalues, with the i-th
                        //          column of Z holding the eigenvector associated with W(i).
                        //          If JOBZ = 'N', then Z is not referenced.
                        //          Note: the user must ensure that at least max(1,M) columns are
                        //          supplied in the array Z; if RANGE = 'V', the exact value of M
                        //          is not known in advance and can be computed with a workspace
                        //          query by setting NZC = -1, see below.

    &ldz,               //  LDZ     (input) INTEGER
                        //          The leading dimension of the array Z.  LDZ >= 1, and if
                        //          JOBZ = 'V', then LDZ >= max(1,N).

    &nzc,               //  NZC     (input) INTEGER
                        //          The number of eigenvectors to be held in the array Z.
                        //          If RANGE = 'A', then NZC >= max(1,N).
                        //          If RANGE = 'V', then NZC >= the number of eigenvalues in (VL,VU].
                        //          If RANGE = 'I', then NZC >= IU-IL+1.
                        //          If NZC = -1, then a workspace query is assumed; the
                        //          routine calculates the number of columns of the array Z that
                        //          are needed to hold the eigenvectors.
                        //          This value is returned as the first entry of the Z array, and
                        //          no error message related to NZC is issued by XERBLA.

    isuppz,             //  ISUPPZ  (output) INTEGER ARRAY, dimension ( 2*max(1,M) )
                        //          The support of the eigenvectors in Z, i.e., the indices
                        //          indicating the nonzero elements in Z. The i-th computed eigenvector
                        //          is nonzero only in elements ISUPPZ( 2*i-1 ) through
                        //          ISUPPZ( 2*i ). This is relevant in the case when the matrix
                        //          is split. ISUPPZ is only accessed when JOBZ is 'V' and N > 0.

    &tryrac,            //  TRYRAC  (input/output) LOGICAL
                        //          If TRYRAC.EQ..TRUE., indicates that the code should check whether
                        //          the tridiagonal matrix defines its eigenvalues to high relative
                        //          accuracy.  If so, the code uses relative-accuracy preserving
                        //          algorithms that might be (a bit) slower depending on the matrix.
                        //          If the matrix does not define its eigenvalues to high relative
                        //          accuracy, the code can uses possibly faster algorithms.
                        //          If TRYRAC.EQ..FALSE., the code is not required to guarantee
                        //          relatively accurate eigenvalues and can use the fastest possible
                        //          techniques.
                        //          On exit, a .TRUE. TRYRAC will be set to .FALSE. if the matrix
                        //          does not define its eigenvalues to high relative accuracy.

    work2,              //  WORK    (workspace/output) DOUBLE PRECISION array, dimension (LWORK)
                        //          On exit, if INFO = 0, WORK(1) returns the optimal
                        //          (and minimal) LWORK.

    &lwork,             //  LWORK   (input) INTEGER
                        //          The dimension of the array WORK. LWORK >= max(1,18*N)
                        //          if JOBZ = 'V', and LWORK >= max(1,12*N) if JOBZ = 'N'.
                        //          If LWORK = -1, then a workspace query is assumed; the routine
                        //          only calculates the optimal size of the WORK array, returns
                        //          this value as the first entry of the WORK array, and no error
                        //          message related to LWORK is issued by XERBLA.

    iwork,              //  IWORK   (workspace/output) INTEGER array, dimension (LIWORK)
                        //          On exit, if INFO = 0, IWORK(1) returns the optimal LIWORK.

    &liwork,            //  LIWORK  (input) INTEGER
                        //          The dimension of the array IWORK.  LIWORK >= max(1,10*N)
                        //          if the eigenvectors are desired, and LIWORK >= max(1,8*N)
                        //          if only the eigenvalues are to be computed.
                        //          If LIWORK = -1, then a workspace query is assumed; the
                        //          routine only calculates the optimal size of the IWORK array,
                        //          returns this value as the first entry of the IWORK array, and
                        //          no error message related to LIWORK is issued by XERBLA.

    &info               //  INFO    (output) INTEGER
                        //          On exit, INFO
                        //          = 0:  successful exit
                        //          < 0:  if INFO = -i, the i-th argument had an illegal value
                        //          > 0:  if INFO = 1X, internal error in DLARRE,
                        //                if INFO = 2X, internal error in DLARRV.
                        //                Here, the digit X = ABS( IINFO ) < 10, where IINFO is
                        //                the nonzero error code returned by DLARRE or
                        //                DLARRV, respectively.
      );

  if (info != 0) {
    DO_CLEANUP_ON_ERROR();
    return TS_NEW_ERROR2(TS_ERR_INVALID_ARGUMENT, ("dstemr failed (info=%d)", info));
  }

  *eigvalsp = eigvals;
  if (eigvecsp) { *eigvecsp = eigvecs; }

  ts_safe_free(NO_ALLOCATOR, elts);

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_sparse_eigensystem_sparse(struct ts_autopool* pool,
    ts_sparse_t* hess_mat, double** eigvalsp, double** eigvecsp, size_t skip)
{
  UNUSED_PARAM(pool);

  int ndof = hess_mat->m;

  double* eigvals = NULL;
  double* eigvecs = NULL;

#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR, eigvals); \
    ts_safe_free(NO_ALLOCATOR, eigvecs); }

  eigvals = ts_alloc(NO_ALLOCATOR, (ndof)*sizeof(eigvals[0]));
  ERROR_ON_NULL_ALLOC(eigvals);

  if (eigvecsp)
  {
    eigvecs = ts_alloc(NO_ALLOCATOR, (ndof*ndof)*sizeof(eigvecs[0]));
    ERROR_ON_NULL_ALLOC(eigvecs);
  }

  struct ts_sparse_eigen eigen;
  memset(&eigen,0,sizeof(eigen));
  eigen.matrix = hess_mat;
  eigen.nev = hess_mat->m - skip;
  eigen.max_iter = TS_SPARSE_EIGEN_MAXITER;
  eigen.vals = eigvals;
  eigen.vecs = eigvecs;

  RETURN_ON_ERROR3( ts_sparse_eigenvalues_calc(&eigen) );

  *eigvalsp = eigvals;
  if (eigvecsp) { *eigvecsp = eigvecs; }

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult ts_sparse_eigensystem_dense(struct ts_autopool* pool,
    ts_sparse_t* hess_mat, double** eigvalsp, double** eigvecsp)
{

#define CLEANUP_ON_ERROR {}

  int ndof = hess_mat->m;

  size_t hess_sz = symm_size(ndof);
  double* hess = ts_autopool_alloc(pool, hess_sz*sizeof(*hess));
  ERROR_ON_NULL_ALLOC( hess );

  int lwork = 12*ndof;
  double* work  = ts_autopool_alloc(pool, lwork*sizeof(work[0]));
  ERROR_ON_NULL_ALLOC( work );

  int liwork = 10*ndof;
  int* iwork = ts_autopool_alloc(pool, liwork*sizeof(iwork[0]));
  ERROR_ON_NULL_ALLOC( iwork );

  int* isuppz = ts_autopool_alloc(pool, 2*ndof*sizeof(*isuppz));
  ERROR_ON_NULL_ALLOC( isuppz );

  double* D = ts_autopool_alloc(pool, (ndof)*sizeof(D[0]));
  ERROR_ON_NULL_ALLOC( D );

  double* E = ts_autopool_alloc(pool, (ndof)*sizeof(E[0]));
  ERROR_ON_NULL_ALLOC( E );

  double* tau = ts_autopool_alloc(pool, (ndof-1)*sizeof(tau[0]));
  ERROR_ON_NULL_ALLOC( tau );

  ts_sparse_symm(hess_mat, hess);

  // Reduce packed symmetric matrix A to tridiagonal matrix...
  int info = 0;
  dsptrd("L", &ndof, hess, D, E, tau, &info);
  if (info != 0) {
    return TS_NEW_ERROR2(TS_ERR_INVALID_ARGUMENT, ("dsptrd failed (see arg %d)", -info));
  }

  // STEP 2: solve eigenvectors of tridiagonal matrix...
  double vl = 0.0;
  double vu = 0.0;
  int il = 0;
  int iu = 0;

  double d_unused = 0.0;

  int neigen = 0;
  int ldz = ndof;

#undef CLEANUP_ON_ERROR

  double* eigvals = NULL;
  double* eigvecs = NULL;
#define CLEANUP_ON_ERROR \
  { ts_safe_free(NO_ALLOCATOR, eigvals); \
    ts_safe_free(NO_ALLOCATOR, eigvecs); }

  eigvals = ts_alloc(NO_ALLOCATOR, (ndof)*sizeof(eigvals[0]));
  ERROR_ON_NULL_ALLOC( eigvals );

  eigvecs = ts_alloc(NO_ALLOCATOR, (ndof*ndof)*sizeof(eigvecs[0]));
  ERROR_ON_NULL_ALLOC( eigvecs );

  dstegr("N", "A", 
      &ndof, D,E, 
      &vl, &vu, 
      &il, &iu, 
      &d_unused,        // abstol (no longer used)
      &neigen,          // M -- number of eigenvalues
      eigvals,          // W array (results)
      eigvecs,          // Z array (unused for eigenvalues)
      &ldz,
      isuppz,           // isuppz (unused for eigenvalues)
      work,
      &lwork,
      iwork,
      &liwork,
      &info);

  if (info != 0) {
    DO_CLEANUP_ON_ERROR();
    return TS_NEW_ERROR2(TS_ERR_INVALID_ARGUMENT, ("dstegr failed (info=%d)", info));
  }

  *eigvalsp = eigvals;
  *eigvecsp = eigvecs;

#undef CLEANUP_ON_ERROR

  return TS_OK;
}




