#include "ts_toml.h"

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <math.h>

#include "ts_tables_priv.h"

DEFINE_TABLE(toml, toml, ts_toml_table, ts_toml_value);

void ts_toml_table_finalize(ts_toml_table* table)
{
  toml_table_delete(table);
}

const ts_toml_value* ts_toml_table_lookup(ts_toml_table* table, const char* key)
{
  ts_toml_value* val = NULL;
  toml_table_get(table, key, val);
  return val;
}

void ts_toml_table_iter_init(ts_toml_table* tbl, struct ts_toml_table_iter* iter);
int ts_toml_table_iter_next(ts_toml_table* tbl, struct ts_toml_table_iter* iter);

void ts_toml_scanner_initialize(ts_toml_scanner* scanner, unsigned char* buf, size_t len)
{
  memset(scanner, 0, sizeof(*scanner));
  scanner->buf = buf;
  scanner->buflen = len;
}

static inline MAYBE_UNUSED int is_toml_whitespace(int ch)
{
  return (ch == ' ') || (ch == '\t');
}

static inline MAYBE_UNUSED int is_toml_key(int ch)
{
  return !is_toml_whitespace(ch) && (ch != '\n') && (ch != '=');
}

static inline void
scanner_set_state(ts_toml_scanner* scanner, enum TS_TOML_SCANNER_STATE st)
{
  scanner->state[scanner->top] = st;
}

static inline TSResult
scanner_push_state(ts_toml_scanner* scanner, enum TS_TOML_SCANNER_STATE st)
{
  scanner->top++;
  if (scanner->top < TS_TOML_SCANNER_STACKSIZE )
  {
    scanner->state[scanner->top] = st;
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR2(TS_ERR_INSUFFICIENT_SIZE,
        ("stack OVERFLOW in TOML scanner (max stack size is %d)",
         TS_TOML_SCANNER_STACKSIZE));
  }
}

static inline TSResult
scanner_pop_state(ts_toml_scanner* scanner)
{
  scanner->top--;
  if (scanner->top >= 0)
  {
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR3(TS_ERR_IN_ALGORITHM,
        ("stack UNDERFLOW in TOML scanner"));
  }
}

#define SCANNER_ERROR(scanner,off,mesg, ...) do {  \
  scanner_set_state(scanner,TS_TOML_SCANNING_END); \
    return TS_NEW_ERROR2(TS_ERR_PARSE_ERROR,       \
        ("error scanning TOML at line %zd, column %zd: " mesg, \
         scanner->line+1, (1 + off - scanner->line0), __VA_ARGS__)); \
  } while(0)

/* Scans a string for a newline '\n' character, returns either the
 * newline or the end of the string.
 */
static inline unsigned char*
find_newline(unsigned char* beg, unsigned char* end)
{
  while ((beg < end) && (*beg != '\n')) { beg++; }
  return beg;
}

/* searches for a newline or a comment character '#'.  If neither is
 * encountered before the end of the string, return the end of the
 * string.
 *
 * This is a tentative newline because it doesn't detect strings or
 * string literals.
 */
static inline unsigned char*
find_tentative_eol(unsigned char* beg, unsigned char* end)
{
  while ((beg < end) && (*beg != '\n') && (*beg != '#')) { beg++; }
  return beg;
}

static inline unsigned char*
skip_whitespace(unsigned char* beg, unsigned char* end)
{
  while ((beg < end) && ((*beg == ' ') || (*beg == '\t'))) { beg++; }
  return beg;
}

static inline void scanner_update(ts_toml_scanner* scanner, unsigned char* beg)
{
  scanner->off = beg - scanner->buf;
}

static inline TSResult
scanner_eat_rest_of_line(ts_toml_scanner* scanner, unsigned char* beg)
{
  unsigned char* buf = scanner->buf;
  unsigned char* end = buf + scanner->buflen;

  if (beg == NULL) { beg = buf + scanner->off; }

  beg = skip_whitespace(beg,end);
  scanner_update(scanner,beg);

  if (beg >= end) { return TS_OK; }
  if (*beg == '#') {
    beg = find_newline(beg,end);
  } else if (*beg != '\n') {
     SCANNER_ERROR(scanner,beg, "unexpected data after key-value pair", 0);
  }

  if (beg < end) {
    beg++; 
    scanner->line++;
    scanner->line0 = beg;
  }
  scanner_update(scanner,beg);
  return TS_OK;
}

static inline unsigned char*
scanner_next_line(ts_toml_scanner* scanner, unsigned char* beg, unsigned char* end)
{
  beg = find_newline(beg,end);
  if (beg < end) {
    beg++;
    scanner_update(scanner,beg);
    scanner->line++;
    scanner->line0 = beg;
  }
  return beg;
}

static inline unsigned char*
scanner_next_token(ts_toml_scanner* scanner, unsigned char* beg, unsigned char* end)
{
  unsigned char* endl;

  while (beg < end) {
    endl = find_tentative_eol(beg,end);
    beg = skip_whitespace(beg,endl);

    if (beg < endl) { break; }

    beg = scanner_next_line(scanner,beg,end);
  }

  return (beg < end) ? beg : NULL;
}

static inline TSResult
scanner_scan_table(ts_toml_scanner* scanner, unsigned char* beg, unsigned char* end)
{
  int array_of_tables = 0;

  if (*beg != '[') { SCANNER_ERROR(scanner,beg,"expected a table",0); }
  beg++;

  if (beg >= end) { SCANNER_ERROR(scanner,beg,"table name did not end before end of line", 0); }

  if (*beg == '[')
  {
    array_of_tables = 1;
    beg++;
  }

  unsigned char* tbeg = beg;

  unsigned char* endl = find_tentative_eol(beg,end);
  while ((beg < endl) && (*beg != ']')) { beg++; }

  if ((beg >= endl) || (*beg != ']'))
  {
    SCANNER_ERROR(scanner,beg,"table name did not end before end of line", 0);
  }

  scanner->toklen = beg - tbeg;
  scanner->tok_semantics.s = tbeg;
  scanner->toktype = array_of_tables ? TS_TOML_EVENT_ARRAY_OF_TABLES_BEGIN : TS_TOML_EVENT_TABLE_BEGIN;

  beg++;

  if (array_of_tables)
  {
   if (beg >= endl)
   {
     SCANNER_ERROR(scanner,beg,"table name did not end before end of line", 0);
   }
   else if (*beg != ']')
   {
     SCANNER_ERROR(scanner,beg-1,"expected ]] to close key for array of tables, but only ] found", 0);
   }

   beg++;
  }

  scanner_update(scanner,beg);
  RETURN_ON_ERROR( scanner_eat_rest_of_line(scanner,beg) );

  return TS_OK;
}

static inline TSResult
scanner_try_key(ts_toml_scanner* scanner)
{
  unsigned char* buf = scanner->buf;
  unsigned char* beg = buf + scanner->off;
  unsigned char* end = buf + scanner->buflen;

  beg = scanner_next_token(scanner,beg,end);
  if (beg == NULL) {
    scanner_update(scanner,end);
    STOP_ITERATING();
  }

  if (*beg == '[')
  {
    RETURN_WITH_RESULT( scanner_scan_table(scanner,beg,end) );
  }

  unsigned char* endl = find_tentative_eol(beg,end);
  unsigned char* kbeg = beg;
  while ((beg < endl) && (*beg != ' ') && (*beg != '\t') && (*beg != '=')) { beg++; }
  unsigned char* kend = beg;
  beg = skip_whitespace(beg,endl);

  if (beg >= endl) {
    SCANNER_ERROR(scanner,kbeg,"key-value pair did not finish before end of line", 0);
  }

  if (*beg != '=') {
    SCANNER_ERROR(scanner,beg,"key-value pair is missing '=' sign", 0);
  }

  beg = skip_whitespace(beg+1,endl);
  if (beg >= endl) {
    SCANNER_ERROR(scanner,kbeg,"key-value pair did not finish before end of line", 0);
  }

  scanner_update(scanner,beg);
  scanner->toklen = kend-kbeg;
  scanner->tok_semantics.s = kbeg;
  scanner->toktype = TS_TOML_EVENT_KEY;
  RETURN_ON_ERROR( scanner_push_state(scanner, TS_TOML_SCANNING_VALUE) );

  return TS_OK;
}

static inline TSResult
scanner_skip_to_value(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  unsigned char* endl = find_tentative_eol(beg,end);
  beg = skip_whitespace(beg,endl);
  if (beg >= endl) {
    SCANNER_ERROR(scanner, beg+scanner->off, "value missing from line", 0);
  }

  scanner_update(scanner,beg);

  return TS_OK;
}

static inline int
ishex(int c)
{
  if (isdigit(c)) { return 1; }
  c = toupper(c);
  return c >= 'A' && c <= 'F';
}

static inline char hex2dec(int c)
{
  if (isdigit(c)) { return c - '0'; }

  c = toupper(c);
  return 10 + c - 'A';
}

// dest is expected to be an array of size 4 or greater.  The encoding
// writes 1-4 bytes into dest.  The number of elements written is
// returned.  dest is not zero terminated by this function.
static inline int codepoint_to_utf8(unsigned long cp, unsigned char* dest)
{
  if (cp <= 0x7F)
  {
    dest[0] = cp;
    return 1;
  }
  else if (cp <= 0x7FF)
  {
    dest[1] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[0] = 0xc0 | cp;
    return 2;
  }
  else if (cp <= 0xFFFF)
  {
    dest[2] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[1] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[0] = 0xe0 | cp;
    return 3;
  }
  else
  {
    dest[3] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[2] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[1] = 0x80 | (cp & 0x3f); cp >>= 6;
    dest[0] = 0xf0 | cp;
    return 4;
  }
}

static TSResult
scanner_try_string(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  if ((*beg != '\'') && (*beg != '"')) {
    CANCEL_OPERATION();
  }

  int literal = (*beg == '\'');

  unsigned char* endl = end;

  unsigned char* spos = beg;
  unsigned char schar = literal ? '\'' : '"';
  int multiline = (beg+2 < endl) && (beg[1] == schar) && (beg[2] == schar);
  if (!multiline) {
    beg++;
    endl = find_newline(beg, end);
  } else {
    beg += 3;
    if ((beg < endl) && (beg[0] == '\n'))
    {
      /* strip newline after """ */
      beg++;
    }
  }

  unsigned char* sbeg = beg;
  unsigned char* send = beg;

  while (beg < endl) {
    if (*beg == schar) {
      if (!multiline)
      {
        goto string_end;
      }
      /* test multiline end */
      else if ((beg+2 < endl) && (beg[1] == schar) && (beg[2] == schar))
      {
        beg += 2;
        goto string_end;
      }
    }
    else if ((*beg == '\\') && !literal)
    {
      if (++beg >= endl) { goto string_end; }

      if (multiline && (*beg == '\n'))
      {
        while ((beg < endl) && (is_toml_whitespace(*beg) || *beg == '\n')) { beg++; }
        continue;
      }
      else
      {
        switch (beg[0]) {
          case 'b': *send = '\b'; break;
          case 't': *send = '\t'; break;
          case 'r': *send = '\r'; break;
          case 'n': *send = '\n'; break;
          case 'f': *send = '\f'; break;
          case '"': *send = '"';  break;
          case '\\': *send = '\\'; break;
          case '/': *send = '/'; break;
          case 'u':
            {
              if ((beg+4 < endl) && ishex(beg[1]) && ishex(beg[2]) && ishex(beg[3]) && ishex(beg[4]))
              {
                unsigned long codepoint = (hex2dec(beg[1]) << 12)
                  | (hex2dec(beg[2]) << 8)
                  | (hex2dec(beg[3]) << 4)
                  | (hex2dec(beg[4]));

                int n = codepoint_to_utf8(codepoint, send);

                beg += 5;
                send += n;
                continue;
              }
              else
              {
                SCANNER_ERROR(scanner,beg, "invalid Unicode escape", 0);
              }
            }

          case 'U':
            {
              unsigned long codepoint = 0;
              if (beg+8 >= endl) { SCANNER_ERROR(scanner,beg, "invalid Unicode escape", 0); }
              for (int i=1; i <= 8; i++)
              {
                if (!ishex(beg[i])) { SCANNER_ERROR(scanner,beg, "invalid Unicode escape", 0); }
                codepoint <<= 4;
                codepoint |= hex2dec(beg[i]);
              }
              if (codepoint > 0x1FFFFF)
              {
                SCANNER_ERROR(scanner,beg, "Unicode escape exceeds valid Unicode range of [0,1FFFFF]", 0);
              }

              int n = codepoint_to_utf8(codepoint, send);
              beg += 9;
              send += n;
              continue;
            }

          default: SCANNER_ERROR(scanner,beg, "invalid escape sequence", 0);
        }
      }
    }
    else
    {
      if (multiline && (*beg == '\n')) {
        scanner->line++;
        scanner->line0 = beg+1;
      }
      *send = *beg;
    }

    beg++;
    send++;
  }

string_end:
  if (beg >= endl) {
    SCANNER_ERROR(scanner,spos, "string does not end before end of line", 0);
  }

  beg++;

  scanner->toklen = send - sbeg;
  scanner->tokoff = sbeg - buf0;
  scanner->tok_semantics.s = sbeg;
  scanner->toktype = TS_TOML_EVENT_STRING;

  scanner_update(scanner,beg);

  return TS_OK;
}

static inline TSResult
scanner_try_bool(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  unsigned char* endl = find_tentative_eol(beg,end);
  unsigned char* bbeg = beg;

  int tf = 0;
  if ((beg+3 < endl) && (memcmp(beg,"true",4)==0)) {
    tf = 1;
    beg += 4;
  } else if ((beg+4 < endl) && (memcmp(beg,"false",5)==0)) {
    tf = 0;
    beg += 5;
  } else {
    CANCEL_OPERATION();
  }

  scanner->toklen = beg - bbeg;
  scanner->tokoff = bbeg - buf0;
  scanner->tok_semantics.b = tf;
  scanner->toktype = TS_TOML_EVENT_BOOLEAN;

  scanner_update(scanner,beg);

  return TS_OK;
}

static TSResult
scanner_parse_double(ts_toml_scanner* scanner, unsigned char* beg, unsigned char* end)
{
  char buf[64];
  ssize_t dlen = end - beg;
  if (dlen+1 > (ssize_t)sizeof(buf)) {
    SCANNER_ERROR(scanner,beg,
        "double representation is too long (max of %zu characters)", sizeof(buf)-1);
  }
  memcpy(buf+0,beg,dlen);
  buf[dlen] = '\0';

  char* eptr = NULL;
  errno = 0; /* Clear errono before call to strtod */
  double val = strtod(buf+0, &eptr);
  if (eptr && *eptr != '\0')
  {
    SCANNER_ERROR(scanner, beg+(eptr-buf), "invalid double", 0);
  }

  if (errno == ERANGE) {
    if (val == 0.0) {
      SCANNER_ERROR(scanner,beg, "double underflow", 0);
    } else if (val == HUGE_VAL) {
      SCANNER_ERROR(scanner,beg, "double overflow", 0);
    } else {
      SCANNER_ERROR(scanner,beg, "double out of range", 0);
    }
  }

  scanner->toklen = end-beg;
  scanner->tokoff = beg - scanner->buf;
  scanner->tok_semantics.f = val;
  scanner->toktype = TS_TOML_EVENT_FLOAT;

  return TS_OK;
}

static TSResult
scanner_parse_integer(ts_toml_scanner* scanner, unsigned char* beg, unsigned char* end)
{
  uint64_t val = 0;

  unsigned char* nbeg = beg;
  int minus = (beg[0] == '-');
  if (minus || (beg[0] == '+')) { beg++; }
  while ((beg < end) && isdigit(beg[0])) {
    uint64_t prev = val;
    unsigned char dig = beg[0] - '0';
    val *= 10;
    val += (dig);
    if (val < prev) { /* overflow happened! */
      SCANNER_ERROR(scanner,beg,
          "integer value out of valid range (integers are signed 64-bit numbers)", 0);
    }
    beg++;
  }

  /* need to check for overflow or underflow */
  if ((val != 0) && ((val - minus) > INT64_MAX))
  {
    SCANNER_ERROR(scanner,beg,
        "integer value out of valid range (integers are signed 64-bit numbers)", 0);
  }

  int64_t sval;
  if (minus && (val != 0))
  {
    sval = -((int64_t)(val-1)) - 1;
  }
  else
  {
    sval = val;
  }

  scanner->toklen = beg - nbeg;
  scanner->tokoff = nbeg - scanner->buf;
  scanner->tok_semantics.i = sval;
  scanner->toktype = TS_TOML_EVENT_INTEGER;

  return TS_OK;
}

static TSResult
scanner_try_number(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  unsigned char* endl = find_tentative_eol(beg,end);
  unsigned char* nbeg = beg;

  if ((beg[0] == '-') || (beg[0] == '+')) { beg++; }
  if (!isdigit(*beg))
  {
    CANCEL_OPERATION();
  }

  if ((beg[0] == '0') && (beg+1 < endl) && isdigit(beg[1]))
  {
    SCANNER_ERROR(scanner,beg, "number has illegal zero padding", 0);
  }

  while ((beg < endl) && isdigit(*beg)) { beg++; }

  int is_double = 0;
  if ((beg < endl) && (beg[0] == '.'))
  {
    beg++;
    if ((beg >= endl) || (!isdigit(*beg)))
    {
      SCANNER_ERROR(scanner,beg, "double has no digits following decimal point", 0);
    }

    while((beg < endl) && isdigit(*beg)) { beg++; }
    is_double = 1;
  }
  
  if ((beg < endl) && ((beg[0] == 'e') || (beg[0] == 'E')))
  {
    beg++;
    if ((beg < endl) && (beg[0] == '-')) { beg++; }
    if ((beg >= endl) || (!isdigit(*beg)))
    {
      SCANNER_ERROR(scanner,beg, "double has no digits following exponent", 0);
    }
    while ((beg < endl) && isdigit(*beg)) { beg++; }
    is_double = 1;
  }

  if (is_double)
  {
    RETURN_ON_ERROR( scanner_parse_double(scanner, nbeg, beg) );
  }
  else
  {
    RETURN_ON_ERROR( scanner_parse_integer(scanner, nbeg, beg) );
  }

  scanner_update(scanner,beg);

  return TS_OK;
}

static inline int parse_digits(unsigned char* beg, unsigned char* end, int ndig)
{
  if (ndig == 0) { return 0; }

  int val = 0;
  if (beg+(ndig-1) >= end) { return -1; }
  for(int i=0; i < ndig; i++)
  {
    if (!isdigit(beg[i])) { return -1; }
    val *= 10;
    val += (beg[i] - '0');
  }

  return val;
}

static TSResult
scanner_try_datetime(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  unsigned char* endl = find_tentative_eol(beg,end);
  unsigned char* dbeg = beg;

  struct ts_toml_datetime dt;
  memset(&dt,0,sizeof(dt));

  if ((dt.year = parse_digits(beg,endl,4)) < 0)
  {
    CANCEL_OPERATION();
  }
  beg += 4;

  if ((beg >= endl) || (*beg != '-')) { CANCEL_OPERATION(); }
  beg++;

  if ((dt.month = parse_digits(beg,endl,2)) < 0)
  {
    CANCEL_OPERATION();
  }
  beg += 2;

  if ((beg >= endl) || (*beg != '-')) { CANCEL_OPERATION(); }
  beg++;

  /* if XXXX-XX- encountered, treat this as a date, so bad parsing is
   * treated as an error below
   */
  if ((dt.day = parse_digits(beg,endl,2)) < 0)
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: day field should be two digits", 0);
  }
  beg += 2;

  if ((beg >= endl) || (*beg != 'T'))
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: expected RFC3339 'T' delimitter for time", 0);
  }
  beg++;

  if ((dt.hour = parse_digits(beg,endl,2)) < 0)
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: hour field should be two digits", 0);
  }
  beg += 2;

  if ((beg >= endl) || (*beg != ':'))
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: expected RFC3339 ':' delimitter for time fields", 0);
  }
  beg++;

  if ((dt.min = parse_digits(beg,endl,2)) < 0)
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: minute field should be two digits", 0);
  }
  beg += 2;

  if ((beg >= endl) || (*beg != ':'))
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: expected RFC3339 ':' delimitter for time fields", 0);
  }
  beg++;

  if ((dt.sec = parse_digits(beg,endl,2)) < 0)
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: second field should be two digits", 0);
  }
  beg += 2;

  dt.nanosec = 0;
  if ((beg < endl) && (*beg == '.'))
  {
    beg++;
    if ((beg >= endl) || !isdigit(*beg))
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: fractional second field has no digits", 0);
    }
    int ndig = 0;
    while((beg+ndig < endl) && isdigit(beg[ndig])) {
      if (ndig < 9) {
        dt.nanosec *= 10;
        dt.nanosec += (beg[ndig] - '0');
      }
      ndig++;
    }
    beg += ndig;
    if (dt.nanosec > 0) {
      while (ndig < 9) { dt.nanosec *= 10; ndig++; }
    }
  }

  if (beg >= endl)
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: missing time offset", 0);
  }

  if (*beg == 'Z')
  {
    dt.utc_off = 0;
    beg++;
  }
  else if ((*beg == '+') || (*beg == '-'))
  {
    int off_hours, off_mins;
    int neg = (*beg == '-');
    beg++;
    if ((off_hours = parse_digits(beg,endl,2)) < 0)
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: time offset field expected to have 2-digit hours", 0);
    }
    if (off_hours > 23)
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: time offset hours outside of valid range [0,23]: %d",
          off_hours);
    }
    beg += 2;

    if ((beg >= endl) || (beg[0] != ':'))
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: time offset field is missing ':' delimitter", 0);
    }
    beg++;

    if ((off_mins = parse_digits(beg,endl,2)) < 0)
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: time offset field expected to have 2-digit minutes", 0);
    }
    if (off_mins > 59)
    {
      SCANNER_ERROR(scanner,beg,
          "invalid datetime: time offset hours outside of valid range [0,59]: %d",
          off_mins);
    }
    beg += 2;

    dt.utc_off = 60*off_hours + off_mins;
    if (neg) { dt.utc_off = -dt.utc_off; }
  }
  else
  {
    SCANNER_ERROR(scanner,beg,
        "invalid datetime: invalid time offset", 0);
  }

  /* Now that we've parsed the date, validate each field */
  if ((dt.month < 1) || (dt.month > 12))
  {
    SCANNER_ERROR(scanner,dbeg,
        "invalid datetime: month field out of valid range [1,12]: %d", dt.day);
  }
  /* FIXME: better validation for days */
  if (dt.day > 31)
  {
    SCANNER_ERROR(scanner,dbeg,
        "invalid datetime: day field out of valid range [0,31: %d", dt.day);
  }
  if (dt.hour > 23)
  {
    SCANNER_ERROR(scanner,dbeg,
        "invalid datetime: hour field out of valid range [0,23]: %d", dt.hour);
  }
  if (dt.min > 59)
  {
    SCANNER_ERROR(scanner,dbeg,
        "invalid datetime: minutes out of valid range [0,59]: %d", dt.min);
  }
  /* FIXME: better validation for leap seconds */
  if (dt.sec > 59)
  {
    SCANNER_ERROR(scanner,dbeg,
        "invalid datetime: seconds out of valid range [0,59]: %d", dt.min);
  }

  scanner->toklen = beg - dbeg;
  scanner->tokoff = dbeg - scanner->buf;
  scanner->tok_semantics.dt = dt;
  scanner->toktype = TS_TOML_EVENT_DATETIME;

  scanner_update(scanner,beg);

  return TS_OK;
}

static inline TSResult
scanner_try_scalar_value(ts_toml_scanner* scanner)
{
  TSResult ret;
  if (!IS_CANCELLED(ret = scanner_try_string(scanner)))   { RETURN_WITH_RESULT(ret); }
  if (!IS_CANCELLED(ret = scanner_try_bool(scanner)))     { RETURN_WITH_RESULT(ret); }
  if (!IS_CANCELLED(ret = scanner_try_datetime(scanner))) { RETURN_WITH_RESULT(ret); }
  if (!IS_CANCELLED(ret = scanner_try_number(scanner)))   { RETURN_WITH_RESULT(ret); }

  CANCEL_OPERATION();
}

static inline TSResult
scanner_try_array(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;

  if (*beg != '[') { CANCEL_OPERATION(); }

  scanner->toklen = 1;
  scanner->tok_semantics.s = beg;
  scanner->toktype = TS_TOML_EVENT_ARRAY_BEGIN;

  beg++;
  scanner_update(scanner,beg);
  scanner_set_state(scanner, TS_TOML_SCANNING_ARRAY);

  return TS_OK;
}

static inline TSResult
scanner_expect_value(ts_toml_scanner* scanner)
{
  RETURN_ON_ERROR( scanner_skip_to_value(scanner) );

  TSResult ret;
  if (!IS_CANCELLED(ret = scanner_try_array(scanner))) { RETURN_WITH_RESULT(ret); }

  ret = scanner_try_scalar_value(scanner);
  RETURN_ON_ERROR(ret);
  if (IS_OKAY(ret))
  {
    RETURN_ON_ERROR( scanner_pop_state(scanner) );
    if (scanner->state[scanner->top] == TS_TOML_SCANNING_KEY)
    {
      RETURN_ON_ERROR( scanner_eat_rest_of_line(scanner,NULL) );
    }
    return TS_OK;
  }

  SCANNER_ERROR(scanner,(scanner->buf+scanner->off),
      "unknown value", 0);
}

static inline TSResult
scanner_try_close_array(ts_toml_scanner* scanner, unsigned char* beg)
{
  if (*beg != ']') { CANCEL_OPERATION(); }

  scanner->toklen = 1;
  scanner->tok_semantics.s = beg;
  scanner->toktype = TS_TOML_EVENT_ARRAY_END;

  beg++;
  scanner_update(scanner,beg);

  RETURN_ON_ERROR( scanner_pop_state(scanner) );
  return TS_OK;
}

static inline TSResult
scanner_scan_array(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  beg = scanner_next_token(scanner,beg,end);
  if (beg == NULL) {
    scanner_update(scanner,end);
    STOP_ITERATING();
  }

  TSResult ret;
  if (!IS_CANCELLED(ret = scanner_try_close_array(scanner,beg))) { RETURN_WITH_RESULT(ret); }

  // Set the current state at this stack level to SCANNING_ARRAY_COMMA
  // to ensure that we scan for a comma on the next element of *this*
  // array
  scanner_set_state(scanner, TS_TOML_SCANNING_ARRAY_COMMA);

  // Push the state here.  scanner_expect_value() either pops the state
  // stack or, if it find an array, sets the top to
  // TS_TOML_SCANNING_ARRAY.
  RETURN_ON_ERROR( scanner_push_state(scanner, TS_TOML_SCANNING_VALUE) );

  // Update the scanner position so scanner_expect_value() starts from
  // the right place
  scanner_update(scanner,beg);
  RETURN_ON_ERROR( scanner_expect_value(scanner) );

  return TS_OK;
}

static inline TSResult
scanner_scan_array_comma(ts_toml_scanner* scanner)
{
  unsigned char* buf0 = scanner->buf;
  unsigned char* beg = buf0 + scanner->off;
  unsigned char* end = buf0 + scanner->buflen;

  beg = scanner_next_token(scanner,beg,end);
  if (beg == NULL) {
    scanner_update(scanner,end);
    STOP_ITERATING();
  }

  TSResult ret;
  if (!IS_CANCELLED(ret = scanner_try_close_array(scanner,beg))) { RETURN_WITH_RESULT(ret); }

  if (*beg != ',')
  {
    SCANNER_ERROR(scanner,beg,
        "expected a comma delimitter between array elements", 0);
  }

  scanner_update(scanner,beg+1);
  RETURN_WITH_RESULT(scanner_scan_array(scanner));
}

TSResult ts_toml_scanner_next(ts_toml_scanner* scanner)
{
  switch (scanner->state[scanner->top]) {
    case TS_TOML_SCANNING_START:
      {
        scanner->toktype = TS_TOML_EVENT_TABLE_BEGIN;
        scanner->tokoff = scanner->off;
        scanner->toklen = 0;
        scanner->line0 = scanner->tok_semantics.s = scanner->buf + scanner->off;
        scanner_set_state(scanner,TS_TOML_SCANNING_KEY);
        return TS_OK;
      }

    case TS_TOML_SCANNING_KEY:
      {
        RETURN_WITH_RESULT(scanner_try_key(scanner));
      }

    case TS_TOML_SCANNING_VALUE:
      {
        RETURN_WITH_RESULT( scanner_expect_value(scanner) );
      }

    case TS_TOML_SCANNING_ARRAY:
      {
        RETURN_WITH_RESULT( scanner_scan_array(scanner) );
      }

    case TS_TOML_SCANNING_ARRAY_COMMA:
      {
        RETURN_WITH_RESULT( scanner_scan_array_comma(scanner) );
      }

    default:
      {
        STOP_ITERATING();
      }
  }
}

