#ifndef TS_TOML_H
#define TS_TOML_H

#include <stdio.h>
#include <time.h>

#include "ts_errors.h"

#define MODULE_NAME TS_TOML

enum TS_TOML_TYPE {
  TS_TOML_NONE = 0,
  TS_TOML_STRING ,
  TS_TOML_BOOLEAN,
  TS_TOML_INTEGER,
  TS_TOML_DOUBLE,
  TS_TOML_DATE,
  TS_TOML_ARRAY,
  TS_TOML_TABLE,
};

struct ts_toml_value;
typedef struct ts_toml_value ts_toml_value;

struct ts_toml_table;
typedef struct ts_toml_table ts_toml_table;

void ts_toml_table_finalize(ts_toml_table* table);
const ts_toml_value* ts_toml_table_lookup(ts_toml_table* table, const char* key);

struct ts_toml_table_iter { int64_t state; const char* key; ts_toml_value* val; };
void ts_toml_table_iter_init(ts_toml_table* tbl, struct ts_toml_table_iter* iter);
int ts_toml_table_iter_next(ts_toml_table* tbl, struct ts_toml_table_iter* iter);


typedef struct ts_toml_array ts_toml_array;
struct ts_toml_array
{
  size_t len;
  ts_toml_value* values;
};

struct ts_toml_value
{
  enum TS_TOML_TYPE type;
  union { 
    const char* vstring;
    int vbool;
    int64_t vint;
    double vfloat;
    // needed: DATE
    ts_toml_array* array;
    ts_toml_table* table;
  };
};

ts_toml_table* ts_toml_parse_string(const unsigned char* s, size_t len, TSResult* errp);
ts_toml_table* ts_toml_parse_file(FILE* f, TSResult* errp);

enum TS_TOML_EVENT_TYPE {
  TS_TOML_EVENT_NONE = 0,

  TS_TOML_EVENT_KEY,
  TS_TOML_EVENT_BOOLEAN,
  TS_TOML_EVENT_STRING,
  TS_TOML_EVENT_INTEGER,
  TS_TOML_EVENT_FLOAT,
  TS_TOML_EVENT_DATETIME,
  TS_TOML_EVENT_ARRAY_BEGIN,
  TS_TOML_EVENT_ARRAY_END,
  TS_TOML_EVENT_ARRAY_OF_TABLES_BEGIN,
  TS_TOML_EVENT_TABLE_BEGIN,
};

enum TS_TOML_SCANNER_STATE {
  TS_TOML_SCANNING_START  = 0,
  TS_TOML_SCANNING_KEY  ,
  TS_TOML_SCANNING_KVEQUAL,
  TS_TOML_SCANNING_VALUE,
  TS_TOML_SCANNING_ARRAY,
  TS_TOML_SCANNING_ARRAY_COMMA,
  TS_TOML_SCANNING_TABLE,

  TS_TOML_SCANNING_END
};

/* Holds all fields of an RFC3339 datetime
 */
struct ts_toml_datetime
{
  int year;  /* negative years are considered BC/BCE */
  int month;
  int day;
  int hour;
  int min;
  int sec;
  uint64_t nanosec;
  int utc_off;  /* offset from UTC in minutes */
};

/* fills in relevant fields to construct a struct tm.
 * Note that this ignores nanoseconds
 */
void ts_toml_datetime_to_tm(const struct ts_toml_datetime* dt, struct tm* timeptr);

/* returns a time_t from the ts_toml_datetime, if possible */
time_t ts_toml_datetime_to_time_t;

enum { TS_TOML_SCANNER_STACKSIZE = 64 };
typedef struct ts_toml_scanner ts_toml_scanner;
struct ts_toml_scanner
{
  unsigned char* buf;
  size_t buflen;
  size_t off;

  size_t line;
  const unsigned char* line0;

  size_t tokoff;
  size_t toklen;

  enum TS_TOML_SCANNER_STATE state[TS_TOML_SCANNER_STACKSIZE];
  int top;

  enum TS_TOML_EVENT_TYPE toktype;
  union {
    int64_t i;
    double  f;
    int     b;
    const unsigned char* s;
    struct ts_toml_datetime dt;
  } tok_semantics;;
};

void ts_toml_scanner_initialize(ts_toml_scanner* scanner, unsigned char* buf, size_t len);

// Returns TS_OK if a token is available, TS_STOP_ITERATION if the
// scanner has reached the end of the stream, or an error code
// if the scanner encounters an error.
TSResult ts_toml_scanner_next(ts_toml_scanner* scanner);

#undef MODULE_NAME

#endif /* TS_TOML_H */

