#include "ts_io.h"
#include "ts_errors.h"
#include "ts_dof_map.h"
#include "ts_machdeps.h"

#include <errno.h>

/* default 16K */
enum { PBUF_SIZE = 16384 };
// use small odd sizes for testing
// enum { PBUF_SIZE = 41 };

static TSResult write_header(struct tsmf_info *tsmf);
static TSResult write_index(struct tsmf_info *tsmf, struct tsmf_index *idx);

static TSResult read_buffer(FILE *f, unsigned char *buf, size_t n)
{
  size_t nb;

retry:
  nb = fread(buf, 1, n, f);
  if (nb == n)
  {
    return TS_OK;
  }

  if (feof(f))
  {
    EOF_ENCOUNTERED();
  }

  if (errno == EINTR)
  {
    buf += nb;
    n -= nb;

    goto retry;
  }

  TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
      "error (%d): %s", errno, strerror(errno));
}

static TSResult req_read_buffer(FILE *f, unsigned char *buf, size_t n)
{
  TSResult err;

  err = read_buffer(f, buf, n), ON_ERROR_ADD_FRAME(err);
  if (IS_IO_EOF(err))
  {
    TS_RAISE_ERROR_MESG(TS_ERR_BAD_FILE_FORMAT,
        "cannot read TSMF file: file is too short");
  }
  return err;
}

static TSResult write_buffer(FILE *f, unsigned char *buf, size_t n)
{
  size_t nb;

retry:
  nb = fwrite(buf, 1, n, f);
  if (nb == n)
  {
    return TS_OK;
  }

  if (errno == EINTR)
  {
    buf += nb;
    n -= nb;
    goto retry;
  }

  TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
      "error (%d): %s", errno, strerror(errno));
}

/**** Packing buffer utility code ****/
struct pbuf
{
  FILE *f;
  uint32_t len;
  uint32_t off;
  uint32_t cap;
  size_t nread;
  TSResult error;
  unsigned char data[];
};

static inline int pbuf_err(struct pbuf *buf)
{
  return !IS_OKAY(buf->error);
}

static struct pbuf *pbuf_new(struct tsmf_info *tsmf, uint32_t sz)
{
  size_t nb;
  struct pbuf *buf;

  /* XXX - consider caching these within tsmf */
  nb = sizeof *buf + sz;
  buf = ts_alloc(NO_ALLOCATOR,nb);
  if (buf == NULL)
  {
    return NULL;
  }

  buf->f = tsmf->f;
  buf->len = 0;
  buf->off = 0;
  buf->cap = sz;
  buf->error = TS_OK;
  buf->nread = 0;

  return buf;
}

static void pbuf_clear(struct pbuf *buf)
{
  if (pbuf_err(buf))
  {
    return;
  }

  buf->off = 0;
  buf->len = 0;
}

static TSResult pbuf_flush(struct pbuf *buf)
{
  if (!IS_OKAY(buf->error))
  {
    return buf->error;
  }

  if (buf->len == 0)
  {
    return TS_OK;
  }

  buf->error = write_buffer(buf->f, buf->data, buf->len),
    ON_ERROR_ADD_FRAME(buf->error);

  pbuf_clear(buf);

  return buf->error;
}

static TSResult pbuf_readn(struct pbuf *buf, uint32_t n)
{
  uint32_t maxr,nr;

  if (!IS_OKAY(buf->error))
  {
    return buf->error;
  }

  // invariants
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  // fast exit if enough has been read
  if (n <= buf->len - buf->off)
  {
    return TS_OK;
  }

  if (n > buf->cap)
  {
    buf->error = TS_NEW_ERRORF(TS_ERR_INSUFFICIENT_ARRAY_SIZE,
        "cannot read %zu bytes into pbuf with capacity %zu",
        (size_t)n, (size_t)buf->cap);
    return buf->error;
  }

  // n <= capacity, but we have to move the current contents to the
  // beginning.
  if (n > buf->cap - buf->off)
  {
    uint32_t new_off;
    new_off = buf->len - buf->off;
    if (new_off > 0)
    {
      memmove(&buf->data[0], &buf->data[buf->off], new_off);
    }
    buf->len = new_off;
    buf->off = 0;
    assert(buf->len <= buf->cap);
  }

  n -= (buf->len - buf->off);
  maxr = buf->cap - buf->len;
  if (n < maxr)
  {
    maxr = n;
  }

retry:
  nr = fread(&buf->data[buf->len], 1, maxr, buf->f);
  buf->len += nr;
  maxr -= nr;
  buf->nread += nr;

  /* case 1: read full amount */
  if (maxr == 0)
  {
    return TS_OK;
  }

  /* case 2: EOF */
  if (feof(buf->f))
  {
    buf->error = TS_NEW_ERROR_MESG(TS_ERR_BAD_FILE_FORMAT,
        "file is too short");
    return buf->error;
  }

  /* case 3: EINTR */
  if (nr == 0 && errno == EINTR)
  {
    goto retry;
  }

  /* case 4: partial read but no error or eof, try to finish
   * the read
   */
  if (maxr > 0 && !ferror(buf->f))
  {
    goto retry;
  }

  /* case 5: error */
  buf->error = TS_NEW_ERRNO_ERROR();
  return buf->error;
}

static TSResult pbuf_readfull(struct pbuf *buf)
{
  return pbuf_readn(buf,buf->cap);
}

static void pbuf_skip(struct pbuf *buf, uint32_t n)
{
  uint32_t nfill;

  if (pbuf_err(buf))
  {
    return;
  }

  while (buf->cap < n)
  {
    pbuf_readfull(buf);
    if (pbuf_err(buf))
    {
      return;
    }

    pbuf_clear(buf);
    n -= buf->cap;
  }

  pbuf_readn(buf,n);
  if (pbuf_err(buf))
  {
    return;
  }

  // skip
  buf->off += n;
}

static void pbuf_release(struct pbuf *buf)
{
  ts_free(NO_ALLOCATOR, buf);
}


/**** Common packing and unpacking routines ****/

static void pack_bytes(struct pbuf *buf, const unsigned char *restrict p, uint32_t n)
{
  uint32_t nr;

  if (pbuf_err(buf))
  {
    return;
  }

  nr = buf->cap - buf->len;
  while (nr < n)
  {
    memcpy(&buf->data[buf->len], p, nr);
    buf->len += nr;
    assert(buf->len <= buf->cap);

    pbuf_flush(buf);
    if (pbuf_err(buf))
    {
      return;
    }

    n -= nr;
    p += nr;
    nr = buf->cap;
  }

  assert(buf->len + n <= buf->cap);
  memcpy(&buf->data[buf->len], p, n);
  buf->len += n;
}

static void pack_chars(struct pbuf *buf, const char *restrict p, uint32_t n)
{
  pack_bytes(buf, (const unsigned char *)p, n);
}

static void unpack_bytes(struct pbuf *buf, unsigned char *restrict p, uint32_t n)
{
  uint32_t off = 0, n0 = n;

  while (n>0)
  {
    uint32_t nfill;
    nfill = (n < buf->cap) ? n : buf->cap;

    pbuf_readn(buf, nfill);
    if (pbuf_err(buf))
    {
      return;
    }

    memcpy(p, &buf->data[buf->off], nfill);
    p += nfill;
    n -= nfill;
    buf->off += nfill;
  }
}

static uint8_t unpack_u8(struct pbuf *buf)
{
  if (pbuf_err(buf))
  {
    return 0;
  }

  pbuf_readn(buf,1);
  return (uint8_t)buf->data[buf->off++];
}

static void pack_fill(struct pbuf *buf, unsigned char c, uint32_t n)
{
  uint32_t nw,nr;

  if (!IS_OKAY(buf->error))
  {
    return;
  }

  nr = buf->cap - buf->len;
  while (nr < n)
  {
    memset(&buf->data[buf->len], c, nr);
    buf->len += nr;

    pbuf_flush(buf);
    if (pbuf_err(buf))
    {
      return;
    }

    n -= nr;
    nr = buf->cap;
  }

  assert(buf->len + n <= buf->cap);
  memset(&buf->data[buf->len], c, n);
  buf->len += n;
}

static void unpack_chars(struct pbuf *buf, char *p, uint32_t n)
{
  // n includes NUL as the last character
  unpack_bytes(buf, (unsigned char *)p, n-1);

  if (IS_ERROR(buf->error))
  {
    return;
  }

  // add NUL
  p[n-1] = '\0';
}

static void pack_unused(struct pbuf *buf, uint32_t n)
{
  pack_fill(buf, 0, n);
}

static void unpack_unused(struct pbuf *buf, uint32_t n)
{
  pbuf_skip(buf,n);
}

static void pack_u8v(struct pbuf *buf, const uint8_t *restrict vv, size_t n)
{
  // XXX - some way to check unsigned char and uint8_t compatibility
  // (for CHAR_BIT > 8 or other funky systems)
  pack_bytes(buf, (const unsigned char *)vv, n);
}

static void unpack_u8v(struct pbuf *buf, uint8_t *restrict vv, uint32_t n)
{
  // XXX - some way to check unsigned char and uint8_t compatibility
  // (for CHAR_BIT > 8 or other funky systems)
  unpack_bytes(buf,(unsigned char*)vv, n);
}

static void pack_u16v(struct pbuf *buf, const uint16_t *restrict vv, size_t n)
{
  size_t i;

  if (!IS_OKAY(buf->error))
  {
    return;
  }

  for (i=0; i < n; i++)
  {
    uint16_t v = vv[i];
    if (buf->len + 2 > buf->cap)
    {
      pbuf_flush(buf);
      if (pbuf_err(buf))
      {
        return;
      }
    }

    buf->data[buf->len++] = v&0xff;
    buf->data[buf->len++] = (v>>8)&0xff;
  }

  assert(buf->len <= buf->cap);
}

static void pack_u32v(struct pbuf *buf, const uint32_t *restrict vv, size_t n)
{
  size_t i;

  if (!IS_OKAY(buf->error))
  {
    return;
  }

  for (i=0; i < n; i++)
  {
    uint32_t v = vv[i];
    if (buf->len + 4 > buf->cap)
    {
      pbuf_flush(buf);
      if (pbuf_err(buf))
      {
        return;
      }
    }

    buf->data[buf->len++] = v&0xff;
    buf->data[buf->len++] = (v>>8)&0xff;
    buf->data[buf->len++] = (v>>16)&0xff;
    buf->data[buf->len++] = (v>>24)&0xff;
  }

  assert(buf->len <= buf->cap);
}

static void pack_u64v(struct pbuf *buf, const uint64_t *restrict vv, size_t n)
{
  size_t i;

  if (!IS_OKAY(buf->error))
  {
    return;
  }

  for (i=0; i < n; i++)
  {
    uint64_t v = vv[i];
    if (buf->len + 8 > buf->cap)
    {
      pbuf_flush(buf);
      if (pbuf_err(buf))
      {
        return;
      }
    }

    buf->data[buf->len++] = v&0xff;
    buf->data[buf->len++] = (v>>8)&0xff;
    buf->data[buf->len++] = (v>>16)&0xff;
    buf->data[buf->len++] = (v>>24)&0xff;

    buf->data[buf->len++] = (v>>32)&0xff;
    buf->data[buf->len++] = (v>>40)&0xff;
    buf->data[buf->len++] = (v>>48)&0xff;
    buf->data[buf->len++] = (v>>56)&0xff;
  }

  assert(buf->len <= buf->cap);
}

static void pack_f32v(struct pbuf *buf, const float *restrict vv, size_t n)
{
  size_t i;

  if (!IS_OKAY(buf->error))
  {
    return;
  }

  for (i=0; i < n; i++)
  {
    uint32_t v = vv[i];
    if (buf->len + 4 > buf->cap)
    {
      pbuf_flush(buf);
      if (pbuf_err(buf))
      {
        return;
      }
    }

    buf->data[buf->len++] = v&0xff;
    buf->data[buf->len++] = (v>>8)&0xff;
    buf->data[buf->len++] = (v>>16)&0xff;
    buf->data[buf->len++] = (v>>24)&0xff;
  }

  assert(buf->len <= buf->cap);
}

static void pack_f64v(struct pbuf *buf, const double *restrict vv, size_t n)
{
  size_t i;

  for (i=0; i < n; i+=8)
  {
    size_t j,nj;
    uint64_t bits[8];

    if (pbuf_err(buf))
    {
      return;
    }

    for(j=0; j < 8; j++)
    {
      if (i+j >= n) break;
      double d = vv[i+j];
      bits[j] = dbl_to_u64(d);
    }

    pack_u64v(buf, bits, j);
  }

  assert(buf->len <= buf->cap);
}

static void pack_u8(struct pbuf *buf, const uint8_t v)
{
  pack_u8v(buf, &v, 1);
}

static void pack_u16(struct pbuf *buf, const uint16_t v)
{
  pack_u16v(buf, &v, 1);
}

static void pack_u32(struct pbuf *buf, const uint32_t v)
{
  pack_u32v(buf, &v, 1);
}

static void pack_u64(struct pbuf *buf, const uint64_t v)
{
  pack_u64v(buf, &v, 1);
}

static void pack_f32(struct pbuf *buf, const float v)
{
  pack_f32v(buf, &v, 1);
}

static void pack_f64(struct pbuf *buf, const double v)
{
  pack_f64v(buf, &v, 1);
}

static void pack_value(struct pbuf *buf,
    enum TSMF_METADATA_VALUE_TYPE vtype,
    void *vbase,
    uint32_t vcount)
{
  switch (vtype)
  {
    case TSMF_METADATA_VALUE_EMPTY:
      /* nop */
      return;

    case TSMF_METADATA_VALUE_STRING:
    case TSMF_METADATA_VALUE_BYTES:
      pack_bytes(buf, vbase, vcount);
      return;

      // For fixed-width types, the count is implicit in the length
    case TSMF_METADATA_VALUE_S16:
    case TSMF_METADATA_VALUE_U16:
      // XXX - portable for signed?
      pack_u16v(buf, (uint16_t *)vbase, vcount);
      return;

    case TSMF_METADATA_VALUE_S32:
    case TSMF_METADATA_VALUE_U32:
      // XXX - portable for signed?
      pack_u32v(buf, (uint32_t *)vbase, vcount);
      return;

    case TSMF_METADATA_VALUE_S64:
    case TSMF_METADATA_VALUE_U64:
      pack_u64v(buf, (uint64_t *)vbase, vcount);
      return;

    case TSMF_METADATA_VALUE_F32:
      pack_f32v(buf, (float *)vbase, vcount);
      return;

    case TSMF_METADATA_VALUE_F64:
      pack_f64v(buf, (double *)vbase, vcount);
      return;

    default:
      buf->error = TS_NEW_ERRORF(TS_ERR_INVALID_ARGUMENT,
          "invalid value type 0x%02x", vtype);
      return;
  }
}

static uint16_t raw_unpack_u16(const unsigned char *buf)
{
  return (uint16_t)buf[0] | ((uint16_t)buf[1] << 8);
}

static void raw_unpack_u16v(uint16_t *restrict p,
    const unsigned char *restrict buf,
    size_t n)
{
  size_t i;
  for (i=0; i < n; i++, buf += sizeof *p)
  {
    p[i] = raw_unpack_u16(buf);
  }
}

static uint32_t raw_unpack_u32(const unsigned char *buf)
{
  uint32_t v32;
  v32 = buf[0];
  v32 = v32 | ((uint32_t)buf[1] << 8);
  v32 = v32 | ((uint32_t)buf[2] << 16);
  v32 = v32 | ((uint32_t)buf[3] << 24);
  return v32;
}

static void raw_unpack_u32v(uint32_t *restrict p,
    const unsigned char *restrict buf,
    size_t n)
{
  size_t i;
  for (i=0; i < n; i++, buf += sizeof *p)
  {
    p[i] = raw_unpack_u32(buf);
  }
}

static float raw_unpack_f32(const unsigned char *buf)
{
  return u32_to_flt(raw_unpack_u32(buf));
}

static void raw_unpack_f32v(float *restrict p,
    const unsigned char *restrict buf,
    size_t n)
{
  size_t i;
  for (i=0; i < n; i++, buf += sizeof *p)
  {
    p[i] = raw_unpack_f32(buf);
  }
}

static uint64_t raw_unpack_u64(const unsigned char *buf)
{
  uint64_t v64;
  v64 = buf[0];
  v64 = v64 | ((uint64_t)buf[1] << 8);
  v64 = v64 | ((uint64_t)buf[2] << 16);
  v64 = v64 | ((uint64_t)buf[3] << 24);
  v64 = v64 | ((uint64_t)buf[4] << 32);
  v64 = v64 | ((uint64_t)buf[5] << 40);
  v64 = v64 | ((uint64_t)buf[6] << 48);
  v64 = v64 | ((uint64_t)buf[7] << 56);

  return v64;
}

static void raw_unpack_u64v(uint64_t *restrict p,
    const unsigned char *restrict buf,
    size_t n)
{
  size_t i;
  for (i=0; i < n; i++, buf += sizeof *p)
  {
    p[i] = raw_unpack_u64(buf);
  }
}

static double raw_unpack_f64(const unsigned char *buf)
{
  return u64_to_dbl(raw_unpack_u64(buf));
}

static void raw_unpack_f64v(double *restrict p,
    const unsigned char *restrict buf,
    size_t n)
{
  size_t i;
  for (i=0; i < n; i++, buf += sizeof *p)
  {
    p[i] = raw_unpack_f64(buf);
  }
}


static void unpack_u16v(struct pbuf *buf, uint16_t *restrict vv, uint32_t n)
{
  uint32_t nb;

  if (pbuf_err(buf))
  {
    return;
  }

  // XXX - check overflow
  nb = n*sizeof *vv;

  while (nb > 0)
  {
    uint32_t i, nfill, nitems;
    nfill = (nb < buf->cap) ? nb : buf->cap;
    assert(nfill <= nb);

    pbuf_readn(buf, nfill);
    if (pbuf_err(buf))
    {
      return;
    }

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    nitems = nfill/sizeof *vv;
    nfill = nitems * sizeof *vv;
    raw_unpack_u16v(vv, &buf->data[buf->off], nitems);
    buf->off += nfill;

    assert(nfill < nb);
    nb -= nfill;

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    vv += nitems;
  }
}

static uint16_t unpack_u16(struct pbuf *buf)
{
  uint16_t v;

  pbuf_readn(buf,sizeof v);
  if (pbuf_err(buf))
  {
    return 0;
  }

  v = raw_unpack_u16(&buf->data[buf->off]);
  buf->off += sizeof v;
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  return v;
}

static void unpack_u32v(struct pbuf *buf, uint32_t *restrict vv, uint32_t n)
{
  uint32_t nb;

  if (pbuf_err(buf))
  {
    return;
  }

  // XXX - check overflow
  nb = n*sizeof *vv;

  while (nb > 0)
  {
    uint32_t i, nfill, nitems;
    nfill = (nb < buf->cap) ? nb : buf->cap;
    assert(nfill <= nb);

    pbuf_readn(buf, nfill);
    if (pbuf_err(buf))
    {
      return;
    }

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    nitems = nfill/sizeof *vv;
    nfill = nitems * sizeof *vv;
    raw_unpack_u32v(vv, &buf->data[buf->off], nitems);
    buf->off += nfill;

    assert(nfill < nb);
    nb -= nfill;

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    vv += nitems;
  }
}

static uint32_t unpack_u32(struct pbuf *buf)
{
  uint32_t v;

  pbuf_readn(buf,sizeof v);
  if (pbuf_err(buf))
  {
    return 0;
  }

  v = raw_unpack_u32(&buf->data[buf->off]);
  buf->off += sizeof v;
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  return v;
}

static void unpack_u64v(struct pbuf *buf, uint64_t *restrict vv, uint32_t n)
{
  uint32_t nb;

  if (pbuf_err(buf))
  {
    return;
  }

  // XXX - check overflow
  nb = n*sizeof *vv;

  while (nb > 0)
  {
    uint32_t i, nfill, nitems;
    nfill = (nb < buf->cap) ? nb : buf->cap;
    assert(nfill <= nb);

    pbuf_readn(buf, nfill);
    if (pbuf_err(buf))
    {
      return;
    }

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    nitems = nfill/sizeof *vv;
    nfill = nitems * sizeof *vv;
    raw_unpack_u64v(vv, &buf->data[buf->off], nitems);
    buf->off += nfill;

    assert(nfill <= nb);
    nb -= nfill;

    assert(buf->off <= buf->len);
    assert(buf->len <= buf->cap);

    vv += nitems;
  }
}

static uint64_t unpack_u64(struct pbuf *buf)
{
  uint64_t v;

  pbuf_readn(buf,sizeof v);
  if (pbuf_err(buf))
  {
    return 0;
  }

  v = raw_unpack_u64(&buf->data[buf->off]);
  buf->off += sizeof v;
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  return v;
}


static void unpack_f32v(struct pbuf *buf, float *restrict vv, uint32_t n)
{
  uint32_t nr, raw[64];

  if (pbuf_err(buf))
  {
    return;
  }

  while (n > 0)
  {
    uint32_t i;

    nr = (n < ARRAYLEN(raw)) ? n : ARRAYLEN(raw);
    unpack_u32v(buf, raw, nr);
    if (pbuf_err(buf))
    {
      return;
    }

    for (i=0; i < nr; i++, vv++)
    {
      *vv = u32_to_flt(raw[i]);
    }

    n -= nr;
  }
}

static float unpack_f32(struct pbuf *buf)
{
  float v;

  pbuf_readn(buf,sizeof v);
  if (pbuf_err(buf))
  {
    return 0;
  }

  v = raw_unpack_f32(&buf->data[buf->off]);
  buf->off += sizeof v;
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  return v;
}

static void unpack_f64v(struct pbuf *buf, double *restrict vv, uint32_t n)
{
  uint32_t nr;
  uint64_t raw[64];

  if (pbuf_err(buf))
  {
    return;
  }

  while (n > 0)
  {
    uint32_t i;

    nr = (n < ARRAYLEN(raw)) ? n : ARRAYLEN(raw);
    unpack_u64v(buf, raw, nr);
    if (pbuf_err(buf))
    {
      return;
    }

    for (i=0; i < nr; i++, vv++)
    {
      *vv = u64_to_dbl(raw[i]);
    }

    n -= nr;
  }
}

static double unpack_f64(struct pbuf *buf)
{
  double v;

  pbuf_readn(buf,sizeof v);
  if (pbuf_err(buf))
  {
    return 0;
  }

  v = raw_unpack_f64(&buf->data[buf->off]);
  buf->off += sizeof v;
  assert(buf->off <= buf->len);
  assert(buf->len <= buf->cap);

  return v;
}


static void pack_section_header(struct pbuf *buf, uint16_t type, uint16_t flags, uint32_t size)
{
  if (!IS_OKAY(buf->error)) return;

  pack_chars(buf, TSMF_SECTION_MAGIC, strlen(TSMF_SECTION_MAGIC));
  pack_u16(buf, type);
  pack_u16(buf, flags);
  pack_u32(buf, size);
  pack_unused(buf, 4);
}

static TSResult unpack_section_header(struct pbuf *buf, long soff, struct tsmf_section *sxn)
{
  assert(strlen(TSMF_SECTION_MAGIC) == 4);

  pbuf_readn(buf, TSMF_SXN_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    return buf->error;
  }

  *sxn = TSMF_SECTION_ZERO;
  if (memcmp(&buf->data[buf->off], TSMF_SECTION_MAGIC, 4) != 0)
  {
    buf->error = TS_NEW_ERRORF(TS_ERR_BAD_FILE_FORMAT,
        "invalid file section magic: found %c%c%c%c, expected %.4s",
        buf->data[buf->off+0], buf->data[buf->off+1], buf->data[buf->off+2], buf->data[buf->off+3],
        TSMF_SECTION_MAGIC);
    return buf->error;
  }

  pbuf_skip(buf, 4);

  sxn->type = unpack_u16(buf);
  sxn->flags = unpack_u16(buf);
  sxn->length = unpack_u32(buf);
  unpack_unused(buf,4);
  sxn->soff   = soff;

  return TS_OK;
}


/**** Various utility TSMF routines ****/

static inline int allows_writes(struct tsmf_info *tsmf)
{
  return (tsmf->mode & 0x1) == TSMF_READWRITE;
}

static long seek_and_tell(FILE *f, long off, int wh)
{
  if (fseek(f, off, wh) != 0)
  {
    return -1;
  }

  return ftell(f);
}

/*** Section utilities ***/

struct section_info
{
  struct tsmf_info *tsmf;
  struct pbuf *buf;
  struct tsmf_section section;
};

static TSResult section_begin(
    struct section_info *sxn,
    struct tsmf_info *tsmf,
    enum TSMF_SECTION_TYPE stype,
    uint16_t sflags,
    uint32_t data_size)
{
  struct pbuf *buf;
  uint32_t sec_size;
  long soff;
  FILE *f;
  TSResult err;

  buf = NULL;
  f = tsmf->f;
  err = TS_OK;

  assert(f != NULL);

  if (!allows_writes(tsmf))
  {
    err = TS_NEW_ERROR_MESG(
        TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write index: TSMF file is read-only");
    goto error;
  }

  if (soff = seek_and_tell(f, 0, SEEK_END), soff < 0)
  {
    err = TS_NEW_ERRNO_ERROR();
    goto error;
  }

  if (buf = pbuf_new(tsmf, PBUF_SIZE), buf == NULL)
  {
    err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu bytes for writing the index",
        PBUF_SIZE);
    goto error;
  }

  sec_size = TSMF_SXN_HEADER_SIZE + data_size;

  pack_section_header(buf,
      stype, sflags, sec_size);

  if (IS_ERROR(buf->error))
  {
    err = buf->error;
    goto error;
  }

  sxn->tsmf = tsmf;
  sxn->buf = buf;
  sxn->section = (struct tsmf_section){
    .type = stype,
    .flags = sflags,
    .length = sec_size,
    .soff   = soff
  };

  return TS_OK;

error:
  if (buf != NULL)
  {
    pbuf_release(buf);
  }

  return err;
}

static TSResult add_index(struct tsmf_info *tsmf,
    uint16_t section, uint32_t seqno, uint64_t soff);

static TSResult section_end(struct section_info *sxn)
{
  TSResult err;
  struct pbuf *buf;
  struct tsmf_info *tsmf;
  long off, len, cur, end;

  buf  = sxn->buf;
  tsmf = sxn->tsmf;
  assert(buf  != NULL);
  assert(tsmf != NULL);

  err = pbuf_flush(buf), ON_ERROR_ADD_FRAME(err);
  if (IS_ERROR(err))
  {
    goto cleanup;
  }

  off = sxn->section.soff;
  len = sxn->section.length;
  end = off + len;

  cur = ftell(tsmf->f);
  if (cur < 0)
  {
    err = TS_NEW_ERRNO_ERROR();
    goto cleanup;
  }

  if (cur != end)
  {
    err = TS_NEW_ERRORF(TS_ERR_IO_ENCODE_ERROR,
        "section encode disagrees with length: off0=%lu, len=%lu, end=%lu, curr=%lu",
        off, len, end, cur);
    goto cleanup;
  }

  add_index(sxn->tsmf, sxn->section.type, tsmf->seqno, sxn->section.soff);

cleanup:
  pbuf_release(buf);
  return err;
}

static const char *section_name(int stype)
{
  switch (stype)
  {
  case TSMF_SEC_INDEX:
    return "index";

  case TSMF_SEC_METADATA:
    return "metadata";

  case TSMF_SEC_PARTICLE_TYPES:
    return "particle types";

  case TSMF_SEC_RB_TYPES:
    return "rigid body types";

  case TSMF_SEC_PARTICLE_LIST:
    return "particle list";

  case TSMF_SEC_RB_LIST:
    return "rigid body list";

  case TSMF_SEC_CONNECTIVITY:
    return "connectivity";

  case TSMF_SEC_COORDINATES:
    return "coordinates";

  default:
    return "unknown";
  }
}

static TSResult section_begin_read(
    struct section_info *sxn,
    struct tsmf_info *tsmf,
    enum TSMF_SECTION_TYPE stype)
{
  struct pbuf *buf;
  uint32_t sec_size;
  long soff;
  FILE *f;
  TSResult err;

  buf = NULL;
  f = tsmf->f;
  err = TS_OK;

  assert(sxn != NULL);
  assert(f != NULL);

  if (soff = ftell(f), soff < 0)
  {
    err = TS_NEW_ERRNO_ERROR();
    goto error;
  }

  if (buf = pbuf_new(tsmf, PBUF_SIZE), buf == NULL)
  {
    err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu bytes for writing the index",
        PBUF_SIZE);
    goto error;
  }

  unpack_section_header(buf, soff, &sxn->section);
  if (pbuf_err(buf))
  {
    err = buf->error;
    goto error;
  }

  if (sxn->section.type != stype)
  {
    err = TS_NEW_ERRORF(TS_ERR_INVALID_FILE_SECTION,
        "expected %s section (0x%02x), but found %s 0x%02x",
        section_name(stype), stype,
        section_name(sxn->section.type),
        sxn->section.type);
    goto error;
  }

  sxn->tsmf = tsmf;
  sxn->buf = buf;

  return TS_OK;

error:
  if (buf != NULL)
  {
    pbuf_release(buf);
  }

  return err;
}

static TSResult section_end_read(struct section_info *sxn)
{
  TSResult err;
  struct pbuf *buf;
  struct tsmf_info *tsmf;

  assert(sxn != NULL);
  tsmf = sxn->tsmf;

  buf  = sxn->buf;
  if (buf == NULL)
  {
    return TS_OK;
  }

  if (IS_ERROR(buf->error))
  {
    goto cleanup;
  }

  // on success, try to skip to the end
  if (tsmf != NULL && tsmf->f != NULL)
  {
    long off, len, cur, end;
    off = sxn->section.soff;
    len = sxn->section.length;
    end = off + len;

    cur = ftell(tsmf->f);
    if (cur < 0)
    {
      buf->error = TS_NEW_ERRNO_ERROR();
      goto cleanup;
    }

    if (cur > end)
    {
      buf->error = TS_NEW_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "section decode read past end of section: off0=%lu, len=%lu, end=%lu, curr=%lu",
          off, len, end, cur);
      goto cleanup;
    }

    if (cur < end)
    {
      // seek to end of the section, silently ignoring any erorrs
      fseek(sxn->tsmf->f, end, SEEK_SET);
    }
  }

cleanup:
  err = buf->error;
  pbuf_release(buf);
  sxn->buf = NULL;

  return err;
}


/**** TSMF writing routines ****/

enum
{
  IDXHDR_SIZE = 16,
  IDXENT_SIZE = 16,
  IDXSFX_SIZE = 16,
};

static TSResult new_file(struct tsmf_info *tsmf);
static TSResult read_header(FILE *f, struct tsmf_file_header *hdr);
static struct tsmf_index *new_indices(void);
static void free_indices(struct tsmf_index *idx);
static TSResult read_index(struct tsmf_info *tsmf);

static TSResult try_reading_index(struct tsmf_info *tsmf)
{
  long off;
  unsigned char buf[IDXSFX_SIZE];
  uint64_t idx_off;
  TSResult err;
  FILE *f;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);
  f = tsmf->f;

  if (off = seek_and_tell(f, -IDXSFX_SIZE, SEEK_END), off < 0)
  {
    TS_RAISE_ERRNO_ERROR();
  }

  err = read_buffer(f, buf, sizeof buf), ON_ERROR_ADD_FRAME(err);
  if (!IS_OKAY(err))
  {
    if (IS_IO_EOF(err))
    {
      err = TS_NEW_ERROR_MESG(TS_ERR_BAD_FILE_FORMAT,
          "cannot read TSMF index header: file is too short");
    }

    return err;
  }

  if (memcmp(&buf[0], TSMF_INDEX_SUFFIX, 4) != 0)
  {
    return TS_OK;
  }

  idx_off = raw_unpack_u64(&buf[8]);
  if (fseek(f, idx_off, SEEK_SET) != 0)
  {
    TS_RAISE_ERRNO_ERROR();
  }

  RETURN_ON_ERROR( read_index(tsmf) );

  return TS_OK;
}

struct tsmf_info *tsmf_new(ts_sim *sim, FILE *f, uint16_t mode, TSResult *errp)
{
  struct tsmf_info *tsmf;
  int basic_mode, empty_file;
  long off;

  assert(errp != NULL);
  assert(sim != NULL);

  tsmf = ts_alloc(NO_ALLOCATOR, sizeof *tsmf);
  if (tsmf == NULL)
  {
    *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    return NULL;
  }

  tsmf->sim = sim;
  tsmf->f = f;
  tsmf->mode = mode;
  tsmf->flags = 0;
  tsmf->ioff = 0;
  tsmf->seqno = 0;

  tsmf->clist   = NULL;
  tsmf->indices = NULL;

  if (off = seek_and_tell(f, 0, SEEK_END), off < 0)
  {
    goto errno_error;
  }

  empty_file = (off == 0);

  if (!empty_file)
  {
    *errp = try_reading_index(tsmf);
    if (!IS_OKAY(*errp))
    {
      /* log error reading indices? */
    }
  }

  if (tsmf->indices == NULL)
  {
    tsmf->indices = new_indices();
    if (tsmf->indices == NULL)
    {
      *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
      goto cleanup_on_error;
    }
  }

  if (fseek(tsmf->f,0,SEEK_SET) != 0)
  {
    goto errno_error;
  }

  if (((mode & 0x1) == TSMF_READWRITE) && empty_file)
  {
    *errp = new_file(tsmf), ON_ERROR_ADD_FRAME(*errp);
  }
  else
  {
    struct tsmf_file_header hdr;
    *errp = read_header(tsmf->f,&hdr), ON_ERROR_ADD_FRAME(*errp);
    tsmf->flags = hdr.flags;
  }

  if (IS_ERROR(*errp))
  {
    ERROR_ADD_FRAME(*errp);
    goto cleanup_on_error;
  }

  return tsmf;

errno_error:
  *errp = TS_NEW_ERRNO_ERROR();

cleanup_on_error:
  ts_free(NO_ALLOCATOR, tsmf);
  return NULL;
}

/* Closes the tsmf file, reclaiming memory */
TSResult tsmf_close(struct tsmf_info *tsmf)
{
  TSResult err = TS_OK;

  if (tsmf->indices != NULL && tsmf->indices->dirty && allows_writes(tsmf))
  {
    if (fseek(tsmf->f, 0, SEEK_END) != 0)
    {
      err = TS_NEW_ERRORF(TS_ERR_IO_ERROR,
          "error (%d): %s", errno, strerror(errno));
      goto finish_closing;
    }
    err = write_index(tsmf, tsmf->indices);
  }

finish_closing:
  fflush(tsmf->f);
  free_indices(tsmf->indices);
  tsmf_conn_list_delete(tsmf->clist);

  ts_free(NO_ALLOCATOR, tsmf);
  return err;
}

static TSResult read_header(FILE *f, struct tsmf_file_header *hdr)
{
  unsigned char buf[32] = {0};
  size_t n;
  TSResult err;

  err = read_buffer(f, &buf[0], sizeof buf), ON_ERROR_ADD_FRAME(err);
  if (IS_IO_EOF(err))
  {
    TS_RAISE_ERROR_MESG(TS_ERR_BAD_FILE_FORMAT,
        "cannot read TSMF header: file is too short");
  }
  if (IS_ERROR(err))
  {
    return err;
  }

  assert(strlen(TSMF_MAGIC) == 4);
  memcpy(&hdr->magic[0], &buf[0], 4);
  hdr->version[0] = buf[4];
  hdr->version[1] = buf[5];
  hdr->flags = raw_unpack_u16(&buf[6]);
  hdr->ioff = raw_unpack_u64(&buf[8]);
  memcpy(hdr->padding, &buf[16], 16);

  if (memcmp(hdr->magic, TSMF_MAGIC, 4) != 0)
  {
    TS_RAISE_ERRORF(TS_ERR_BAD_FILE_FORMAT,
        "invalid TSMF magic: %c%c%c%c, expected %s",
        hdr->magic[0], hdr->magic[1], hdr->magic[2], hdr->magic[3],
        TSMF_MAGIC);
  }

  if (hdr->version[0] != TSMF_MAJOR || hdr->version[1] != TSMF_MINOR)
  {
    TS_RAISE_ERRORF(TS_ERR_BAD_FILE_FORMAT,
        "invalid TSMF file version: %d.%d, but expected %d.%d",
        (int)buf[4], (int)buf[5], TSMF_MAJOR, TSMF_MINOR);
  }

  return TS_OK;
}

/* Writes the index to disk */
static TSResult write_index(struct tsmf_info *tsmf, struct tsmf_index *idx)
{
  struct pbuf *buf = NULL;
  size_t index_size, i,n, nb;
  long ioff;
  TSResult err;

  if (!allows_writes(tsmf))
  {
    TS_RAISE_ERRORF(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write index: TSMF file is read-only", 0);
  }

  index_size = TSMF_SXN_HEADER_SIZE + IDXHDR_SIZE +
    idx->len * IDXENT_SIZE + IDXSFX_SIZE;

  if (ioff = seek_and_tell(tsmf->f, 0, SEEK_END), ioff < 0)
  {
    TS_RAISE_ERRNO_ERROR();
  }

  if (buf = pbuf_new(tsmf, PBUF_SIZE), buf == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu bytes for writing the index",
        PBUF_SIZE);
  }

  pack_section_header(buf, TSMF_SEC_INDEX, 0, index_size);

  /* pack index header */
  pack_u32(buf, idx->len);
  pack_unused(buf, 12);

  n = idx->len;
  for (i=0; i < n; i++)
  {
    /* pack each index entry */
    pack_u16(buf, idx->entries[i].type);
    pack_unused(buf, 2);
    pack_u32(buf, idx->entries[i].seqno);
    pack_u64(buf, idx->entries[i].soff);
  }

  /* write the section suffix */
  pack_chars(buf, TSMF_INDEX_SUFFIX, 4);
  pack_unused(buf, 4);
  pack_u64(buf, ioff);

  err = pbuf_flush(buf), ON_ERROR_ADD_FRAME(err);
  pbuf_release(buf);

  idx->dirty = 0;

  return err;
}

static TSResult read_index(struct tsmf_info *tsmf)
{
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  struct pbuf *buf;
  uint32_t i,nentries;
  size_t off;
  long soff;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_INDEX));

  buf = sxn.buf;
  pbuf_readn(buf, IDXHDR_SIZE);
  if (pbuf_err(buf))
  {
    err = buf->error;
    goto cleanup;
  }

  nentries = unpack_u32(buf);
  unpack_unused(buf, 12);

  if (tsmf->indices == NULL)
  {
    tsmf->indices = new_indices();
  }

  tsmf->indices->len = 0;
  for (i=0; i < nentries; i++)
  {
    uint16_t stype;
    uint32_t seqno;
    uint64_t soff;

    pbuf_readn(buf, IDXENT_SIZE);
    if (pbuf_err(buf))
    {
      err = buf->error;
      goto cleanup;
    }

    stype = unpack_u16(buf);
    unpack_unused(buf, 2);
    seqno = unpack_u32(buf);
    soff = unpack_u64(buf);

    err = add_index(tsmf, stype, seqno, soff), ON_ERROR_ADD_FRAME(err);
    if (IS_ERROR(err))
    {
      goto cleanup;
    }
  }

cleanup:
  section_end_read(&sxn);

  return err;
}

static TSResult new_file(struct tsmf_info *tsmf)
{
  RETURN_ON_ERROR( write_header(tsmf) );
  return TS_OK;
}

static TSResult write_header(struct tsmf_info *tsmf)
{
  unsigned char buf[32] = {0};
  size_t n;

  assert(strlen(TSMF_MAGIC) == 4);
  memcpy(buf, TSMF_MAGIC, 4);
  buf[4] = TSMF_MAJOR;
  buf[5] = TSMF_MINOR;
  buf[6] = tsmf->flags & 0xff;
  buf[7] = (tsmf->flags >> 8) & 0xff;

  n = fwrite(buf, 1, sizeof buf, tsmf->f);
  // XXX - handle error case

  return TS_OK;
}

TSResult tsmf_read_header(FILE *f, struct tsmf_file_header *hdr)
{
  return read_header(f, hdr);
}

/* Metadata handling */

TSResult tsmf_metadata_initialize(
    struct tsmf_info *tsmf,
    struct tsmf_metadata_info *md)
{
  *md = (struct tsmf_metadata_info){0};
  md->tsmf = tsmf;
  md->err = TS_OK;

  return TS_OK;
}

void tsmf_metadata_finalize(struct tsmf_metadata_info *md)
{
  if (md->sdata)
  {
    size_t i;
    for (i=0; i < md->slen; i++)
    {
      ts_free(NO_ALLOCATOR, md->sdata[i]);
      md->sdata[i] = NULL;
    }

    ts_free(NO_ALLOCATOR, md->sdata);
    md->slen = md->scap = 0;
    md->sdata = NULL;
  }

  ts_free(NO_ALLOCATOR, md->kvdata);
  md->kvlen = md->kvcap = 0;
  md->kvdata = NULL;
}

enum
{
  MD_SHORT_CAP = 1,
  MD_KV_CAP = 1,
  INDEX_CAP = 1,
};

static struct tsmf_index *new_indices(void)
{
  struct tsmf_index *index;
  size_t cap = INDEX_CAP;

  index = ts_alloc(NO_ALLOCATOR,
      sizeof *index + cap * sizeof index->entries[0]);

  if (index != NULL)
  {
    index->len = 0;
    index->dirty = 0;
    index->cap = cap;
  }

  return index;
}

static void free_indices(struct tsmf_index *idx)
{
  ts_free(NO_ALLOCATOR, idx);
}

static TSResult add_index(struct tsmf_info *tsmf,
    uint16_t section, uint32_t seqno, uint64_t soff)
{
  struct tsmf_index *index;
  index = tsmf->indices;
  if (index->len >= index->cap)
  {
    size_t cap = 2*index->cap;

    index = ts_realloc(NO_ALLOCATOR, index,
        sizeof *index + cap * sizeof index->entries[0]);

    if (index == NULL)
    {
      TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
    }

    index->cap = cap;
    tsmf->indices = index;
  }

  assert(index != NULL);
  assert(tsmf->indices == index);
  assert(index->len < index->cap);

  index->entries[index->len++] = ((struct tsmf_index_entry){
      .type = section,
      .unused = 0,
      .seqno = seqno,
      .soff = soff,
  });

  index->dirty = 1;

  return TS_OK;
}

static int md_resize_short(struct tsmf_metadata_info *md, size_t n)
{
  struct tsmf_metadata_short_entry **new_sdata;

  new_sdata = ts_realloc(NO_ALLOCATOR, md->sdata, n* sizeof *new_sdata);
  if (new_sdata == NULL)
  {
    return 0;
  }

  md->scap = n;
  md->sdata = new_sdata;

  return 1;
}

static int md_resize_kv(struct tsmf_metadata_info *md, size_t n)
{
  struct tsmf_metadata_kv_entry **new_kv;

  new_kv = ts_realloc(NO_ALLOCATOR, md->kvdata, n* sizeof *new_kv);
  if (new_kv == NULL)
  {
    return 0;
  }

  md->kvcap = n;
  md->kvdata = new_kv;

  return 1;
}

static TSResult metadata_size(
    enum TSMF_METADATA_VALUE_TYPE vtype,
    uint32_t *szp)
{
  switch (vtype)
  {
    case TSMF_METADATA_VALUE_EMPTY:
      *szp = 0;
      return TS_OK;

    case TSMF_METADATA_VALUE_STRING:
    case TSMF_METADATA_VALUE_BYTES:
      *szp = 1;
      return TS_OK;

      // For fixed-width types, the count is implicit in the length
    case TSMF_METADATA_VALUE_S16:
    case TSMF_METADATA_VALUE_U16:
      *szp = 2;
      return TS_OK;

    case TSMF_METADATA_VALUE_S32:
    case TSMF_METADATA_VALUE_U32:
    case TSMF_METADATA_VALUE_F32:
      *szp = 4;
      return TS_OK;

    case TSMF_METADATA_VALUE_S64:
    case TSMF_METADATA_VALUE_U64:
    case TSMF_METADATA_VALUE_F64:
      *szp = 8;
      return TS_OK;

    default:
      TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
          "invalid value type 0x%02x", vtype);
  }
}

TSResult tsmf_metadata_add_short(
    struct tsmf_metadata_info *md,
    enum TSMF_METADATA_KEY key,
    enum TSMF_METADATA_VALUE_TYPE vtype,
    void *base,
    size_t vcount)
{
  struct tsmf_metadata_short_entry *elt;
  uint32_t elt_size;
  size_t vlen;

  elt_size = 0;
  RETURN_ON_ERROR(metadata_size(vtype, &elt_size));

  /* nop for now */
  if (md->slen >= md->scap)
  {
    size_t new_cap;
    new_cap = (md->scap > 0) ? md->scap * 2 : MD_SHORT_CAP;

    if (!md_resize_short(md,new_cap))
    {
      TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
    }
  }

  vlen = elt_size > 0 ? elt_size * vcount : 0;
  elt = ts_alloc(NO_ALLOCATOR, sizeof *elt + vlen);
  if (elt == NULL)
  {
    TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  elt->key_code = key;
  elt->vtype = vtype;

  elt->vcount = vcount;
  if (elt_size > 0)
  {
    memcpy(elt->vdata, base, vlen);
  }

  md->sdata[md->slen++] = elt;

  return TS_OK;
}

static TSResult metadata_add_kv(struct tsmf_metadata_info *md,
    struct tsmf_metadata_kv_entry *elt)
{
  if (md->kvlen >= md->kvcap)
  {
    size_t new_cap;
    new_cap = (md->kvcap > 0) ? md->kvcap * 2 : MD_KV_CAP;

    if (!md_resize_kv(md,new_cap))
    {
      TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
    }
  }

  md->kvdata[md->kvlen++] = elt;
  assert(md->kvlen <= md->kvcap);

  return TS_OK;
}

TSResult tsmf_metadata_add_kv(
    struct tsmf_metadata_info *md,
    const unsigned char *key, size_t klen,
    const unsigned char *val, size_t vlen)
{
  struct tsmf_metadata_kv_entry *elt;
  size_t kvlen;

  // allocate space first
  RETURN_ON_ERROR( metadata_add_kv(md,NULL) );

  kvlen = klen+vlen;
  elt = ts_alloc(NO_ALLOCATOR, sizeof *elt + kvlen);
  if (elt == NULL)
  {
    TS_RAISE_ERROR(TS_ERR_OUT_OF_MEMORY);
  }

  elt->klen = klen;
  elt->vlen = vlen;
  memset(&elt->unused[0],0,sizeof elt->unused);

  memcpy(&elt->kvdata[0], key, klen);
  memcpy(&elt->kvdata[klen], val, vlen);

  // now fill the space we allocates
  md->kvdata[md->kvlen-1] = elt;

  return TS_OK;
}

enum
{
  MD_HDR_SIZE = 12,
  MD_SHDR_SIZE = 6,
  MD_KVHDR_SIZE = 12,
};

TSResult metadata_write_inner(struct tsmf_metadata_info *md)
{
  struct pbuf *buf;
  struct tsmf_info *tsmf;
  struct section_info sxn = {0};
  FILE *f;
  size_t npad,data_size, i, nb;
  TSResult err;
  long soff;

  tsmf = md->tsmf;
  f = tsmf->f;

  data_size = MD_HDR_SIZE;
  for (i=0; i < md->slen; i++)
  {
    uint32_t esz = 0;
    RETURN_ON_ERROR( metadata_size(md->sdata[i]->vtype, &esz) );
    data_size += MD_SHDR_SIZE;
    data_size += esz * md->sdata[i]->vcount;
  }

  for (i=0; i < md->kvlen; i++)
  {
    data_size += MD_KVHDR_SIZE;
    data_size += md->kvdata[i]->klen;
    data_size += md->kvdata[i]->vlen;
  }

  npad = (16 - (data_size % 16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_METADATA, 0, data_size));

  buf = sxn.buf;
  assert( buf != NULL );

  /* pack metadata header */
  pack_u32(buf, md->slen);
  pack_u32(buf, md->kvlen);
  pack_unused(buf, 4);

  /* pack short entries */
  for (i=0; i < md->slen; i++)
  {
    pack_u8(buf, md->sdata[i]->key_code);
    pack_u8(buf, md->sdata[i]->vtype);
    pack_u32(buf, md->sdata[i]->vcount);
    pack_value(buf, md->sdata[i]->vtype, md->sdata[i]->vdata, md->sdata[i]->vcount);
  }

  /* pack kv entries */
  for (i=0; i < md->kvlen; i++)
  {
    uint32_t nb,klen,vlen;
    klen = md->kvdata[i]->klen;
    vlen = md->kvdata[i]->vlen;
    nb = klen + vlen;

    pack_u32(buf, klen);
    pack_u32(buf, vlen);
    pack_unused(buf, 4);
    pack_bytes(buf, md->kvdata[i]->kvdata, nb);
  }

  /* pad until a multiple of 16 bytes */
  pack_fill(buf, 'm', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

TSResult tsmf_metadata_write(struct tsmf_metadata_info *md)
{
  if (IS_OKAY(md->err))
  {
    md->err = metadata_write_inner(md);
  }

  return md->err;
}

TSResult metadata_read_inner(struct tsmf_metadata_info *md)
{
  struct tsmf_info *tsmf;
  struct section_info sxn = {0};
  struct tsmf_metadata_header md_hdr = {0};
  TSResult err = TS_OK;
  struct pbuf *buf;
  size_t i, nb;
  uint32_t nshort, nkv;

  assert(md != NULL);

  tsmf = md->tsmf;
  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_METADATA));

  buf = sxn.buf;
  pbuf_readn(buf, MD_HDR_SIZE);

  nshort = unpack_u32(buf);
  nkv    = unpack_u32(buf);
  unpack_unused(buf,4);

  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  tsmf_metadata_finalize(md);
  if (!md_resize_short(md,nshort) || !md_resize_kv(md,nkv))
  {
    err = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    goto error;
  }

  // size         desc
  // 1            key - key code
  // 1            vtype - value type
  // 4            vlen - value length
  // vlen         value data
  for (i=0; i < nshort; i++)
  {
    uint32_t vlen, vcount, eltsize, maxread, off;
    uint8_t key, vtype;
    struct tsmf_metadata_short_entry *elt;
    size_t n;

    pbuf_readn(buf, MD_SHDR_SIZE);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    key = unpack_u8(buf);
    vtype = unpack_u8(buf);
    vcount = unpack_u32(buf);
    
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    eltsize = 0;
    // XXX - this leaks:
    err = metadata_size(vtype, &eltsize), ON_ERROR_ADD_FRAME(err);
    if (IS_ERROR(err))
    {
      goto error;
    }

    // XXX - check for overflow!
    vlen = vcount * eltsize;

    n = sizeof *elt + vlen;
    elt = ts_alloc(NO_ALLOCATOR, n);
    if (elt == NULL)
    {
      err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
          "could not allocate %zu bytes for metadata entry", n); 
      goto error;
    }

    elt->key_code = key;
    elt->vtype = vtype;
    elt->vcount = vcount;

    if (vlen > 0)
    {
      switch (vtype)
      {
        // For fixed-width types, the count is implicit in the length
        case TSMF_METADATA_VALUE_S16:
        case TSMF_METADATA_VALUE_U16:
          // XXX - signed case?
          unpack_u16v(buf,(uint16_t *)elt->vdata, vcount);
          break;

        case TSMF_METADATA_VALUE_S32:
        case TSMF_METADATA_VALUE_U32:
          // XXX - signed case?
          unpack_u32v(buf,(uint32_t *)elt->vdata, vcount);
          break;

        case TSMF_METADATA_VALUE_S64:
        case TSMF_METADATA_VALUE_U64:
          // XXX - signed case?
          unpack_u64v(buf,(uint64_t *)elt->vdata, vcount);
          break;

        case TSMF_METADATA_VALUE_F32:
          unpack_f32v(buf,(float *)elt->vdata, vcount);
          break;

        case TSMF_METADATA_VALUE_F64:
          unpack_f64v(buf,(double *)elt->vdata, vcount);
          break;

        case TSMF_METADATA_VALUE_STRING:
        case TSMF_METADATA_VALUE_BYTES:
          unpack_bytes(buf, elt->vdata, vlen);
          break;

        case TSMF_METADATA_VALUE_EMPTY:
          /* should skip this! */
          TS_RAISE_ERRORF(TS_ERR_LOGIC_ERROR,
              "value type 0x%02x already handled", vtype);
          break;

        default:
          TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
              "invalid value type 0x%02x", vtype);
      }
    }

    assert(i < md->scap);
    md->sdata[i] = elt;
  }
  md->slen = nshort;

  // 4            klen  - key length
  // 4            vlen  - value length
  // 4            unused
  // klen         key data
  // vlen         value data
  for (i=0; i < nkv; i++)
  {
    uint32_t klen,vlen;
    size_t kvlen;
    struct tsmf_metadata_kv_entry *elt;

    pbuf_readn(buf, MD_KVHDR_SIZE);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    klen = unpack_u32(buf);
    vlen = unpack_u32(buf);
    unpack_unused(buf,4);
    kvlen = klen+vlen;

    elt = ts_alloc(NO_ALLOCATOR, sizeof *elt + kvlen);
    if (elt == NULL)
    {
      err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
          "could not allocate %zu bytes for a kv element",
          kvlen);
      goto error;
    }
    elt->klen = klen;
    elt->vlen = vlen;
    memset(&elt->unused[0], 0, sizeof elt->unused);

    unpack_bytes(buf, elt->kvdata, kvlen);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    md->kvdata[i] = elt;
  }
  md->kvlen = nkv;

  section_end_read(&sxn);
  return TS_OK;

buf_error:
  err = buf->error;
  /* continue to error: */

error:
  section_end_read(&sxn);
  tsmf_metadata_finalize(md);
  return err;
}

TSResult tsmf_metadata_read(struct tsmf_metadata_info *md)
{
  if (IS_OKAY(md->err))
  {
    md->err = metadata_read_inner(md);
  }

  return md->err;
}


/** Particle types **/

TSResult tsmf_ptypes_write(struct tsmf_info *tsmf)
{
  ts_sim *sim;
  struct section_info sxn;
  uint32_t i, data_size, npad, ntypes;
  struct ts_particle_type *ptypes;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write particle types: no simulation");
  }

  data_size = TSMF_PTYPES_HEADER_SIZE;
  ntypes = sim->particle_types.size;
  ptypes = sim->particle_types.data;
  for (i=0; i < ntypes; i++)
  {
    // XXX - check that the name and long_name lengths are correct
    data_size += TSMF_PTYPES_ENTRY_FIXED_SIZE;
    data_size += strlen(ptypes[i].name);
    data_size += strlen(ptypes[i].long_name);
  }

  npad = (16 - (data_size %16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_PARTICLE_TYPES, 0, data_size));

  pack_u32(sxn.buf, ntypes);
  pack_u16(sxn.buf, 0); /* flags */
  pack_unused(sxn.buf, 10);

  for (i=0; i < ntypes; i++)
  {
    uint16_t lname_len;
    uint8_t name_len;

    lname_len = strlen(ptypes[i].long_name);
    name_len = strlen(ptypes[i].name);
    pack_u16(sxn.buf,lname_len);
    pack_u8(sxn.buf,name_len);
    pack_unused(sxn.buf,5);

    pack_f64(sxn.buf,ptypes[i].mass);
    pack_f64(sxn.buf,ptypes[i].charge);
    pack_f64(sxn.buf,ptypes[i].radius);
    pack_f64(sxn.buf,ptypes[i].collfreq);

    pack_chars(sxn.buf, ptypes[i].name, name_len);
    pack_chars(sxn.buf, ptypes[i].long_name, lname_len);
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

static void free_ptypes(struct ts_particle_type *ptypes, uint32_t n)
{
  uint32_t i;
  for (i=0; i < n; i++)
  {
    ts_free(NO_ALLOCATOR, (char *)ptypes[i].name);
    ts_free(NO_ALLOCATOR, (char *)ptypes[i].long_name);
  }

  ts_free(NO_ALLOCATOR, ptypes);
}

void tsmf_ptypes_free(struct ts_particle_type *ptypes, uint32_t n)
{
  free_ptypes(ptypes,n);
}

TSResult tsmf_ptypes_read(struct tsmf_info *tsmf,
    struct ts_particle_type **ptypesp, uint32_t *np)
{
  struct ts_particle_type *ptypes = NULL;
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  struct pbuf *buf = NULL;
  uint32_t i, ntypes = 0;
  uint16_t flags;
  long soff;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_PARTICLE_TYPES));

  buf = sxn.buf;
  pbuf_readn(buf, TSMF_PTYPES_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  /* XXX - ignore flags for now */
  ntypes = unpack_u32(buf);
  flags  = unpack_u16(buf);
  unpack_unused(buf,10);

  if (pbuf_err(buf))
  {
    err = buf->error;
    goto error;
  }

  ptypes = ts_calloc(NO_ALLOCATOR, ntypes, sizeof *ptypes);
  if (ptypes == NULL)
  {
    err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "error: could not allocate %zu bytes for particle types",
        ntypes*sizeof *ptypes);
    goto error;
  }

  for (i=0; i < ntypes; i++)
  {
    uint16_t lname_len;
    uint8_t name_len;
    char *name, *lname;

    pbuf_readn(buf, TSMF_PTYPES_ENTRY_FIXED_SIZE);
    if (pbuf_err(buf))
    {
      err = buf->error;
      goto error;
    }

    lname_len = unpack_u16(buf);
    name_len  = unpack_u8(buf);
    unpack_unused(buf, 5);

    ptypes[i].mass     = unpack_f64(buf);
    ptypes[i].charge   = unpack_f64(buf);
    ptypes[i].radius   = unpack_f64(buf);
    ptypes[i].collfreq = unpack_f64(buf);

    name  = ts_alloc(NO_ALLOCATOR, name_len+1);
    lname = ts_alloc(NO_ALLOCATOR, lname_len+1);

    if (name == NULL || lname == NULL)
    {
      ts_free(NO_ALLOCATOR, name);
      ts_free(NO_ALLOCATOR, lname);
      err = TS_NEW_ERROR_MESG(TS_ERR_OUT_OF_MEMORY,
          "error: could not allocate memory for particle type names");
      goto error;
    }

    ptypes[i].name = name;
    ptypes[i].long_name = lname;

    unpack_chars(buf, name, name_len+1);
    unpack_chars(buf, lname, lname_len+1);
    if (pbuf_err(buf))
    {
      err = buf->error;
      goto error;
    }
  }

  *np = ntypes;
  *ptypesp = ptypes;

  section_end_read(&sxn);
  return TS_OK;

buf_error:
  err = buf->error;
  /* continue to error: */

error:
  free_ptypes(ptypes, ntypes);
  section_end_read(&sxn);
  return err;
}



/** Coordinate lists **/

/* All of the code below assumes that VecReal is a double.  If this
 * changes, we need to modify the code to handle either and introduce a
 * flag to indicate that the data is stored as floats.
 */
ASSERT_SIZE_EQ(VecReal, double);

static uint32_t coords_ndof(uint32_t nparticle, uint32_t nrigid, uint32_t nrotate)
{
  return PARTICLE_NUM_DOF*nparticle +
    RIGID_BODY_NUM_DOF*nrigid +
    RIGID_BODY_NUM_ROT_DOF*nrotate;
}

static inline uint32_t cbuf_ndof(struct tsmf_coords_buffer *cbuf)
{
  return coords_ndof(
      cbuf->nparticle,
      cbuf->nrigid,
      cbuf->nrotate);
}

uint32_t tsmf_coords_buffer_ndof(struct tsmf_coords_buffer *cbuf)
{
  return cbuf_ndof(cbuf);
}

static TSResult coords_begin(struct section_info *sxn, struct tsmf_info *tsmf,
    uint32_t nparticle, uint32_t nrigid, uint32_t nrotate, uint16_t flags)
{
  uint32_t data_size;
  uint32_t n;

  n = coords_ndof(nparticle,nrigid,nrotate);
  data_size = TSMF_COORDS_HEADER_SIZE + n * sizeof (double);

  RETURN_ON_ERROR(
      section_begin(sxn, tsmf, TSMF_SEC_COORDINATES, 0, data_size));

  pack_u32(sxn->buf, nparticle);
  pack_u32(sxn->buf, nrigid);
  pack_u32(sxn->buf, nrotate);
  pack_u16(sxn->buf, flags);
  pack_unused(sxn->buf, 10);

  return TS_OK;
}

TSResult tsmf_coords_write_sim_all(struct tsmf_info *tsmf)
{
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  ts_sim *sim;
  uint32_t i,n,nrigid,nparticle;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write simulation coordinates: no simulation");
  }

  nrigid = sim->num_rigid_bodies;
  nparticle = 0;

  for (n=sim->num_particles, i=0; i < n; i++)
  {
    ts_particle *p;
    p = ts_get_particle(sim, i);

    if (p->rigid_body_index == TS_NULL_RIGID_BODY)
    {
      nparticle++;
    }
  }

  RETURN_ON_ERROR(
      coords_begin(&sxn, tsmf,
        nparticle, nrigid, 0 /* nrotate */,
        TSMF_COORDS_FLAG_ALL /* flags */));

  for (i=0; i < nrigid; i++)
  {
    const vector3 *pos;
    const quat4 *ori;
    double aa[3];

    pos = RIGID_POS(sim,i);
    ori = RIGID_ORI(sim,i);
    quat4_to_axis_angle_vec(ori, &aa[0]);

    pack_f64(sxn.buf, VECTOR3_X(pos));
    pack_f64(sxn.buf, VECTOR3_Y(pos));
    pack_f64(sxn.buf, VECTOR3_Z(pos));

    pack_f64v(sxn.buf, aa, 3);
  }

  for (n=sim->num_particles, i=0; i < n; i++)
  {
    ts_particle *p;
    const vector3 *pos;

    p = ts_get_particle(sim, i);
    if (p->rigid_body_index != TS_NULL_RIGID_BODY)
    {
      continue;
    }

    pos = PARTICLE_POS(sim,i);

    pack_f64(sxn.buf, VECTOR3_X(pos));
    pack_f64(sxn.buf, VECTOR3_Y(pos));
    pack_f64(sxn.buf, VECTOR3_Z(pos));
  }

  RETURN_ON_ERROR(
      section_end(&sxn));

  return err;
}

TSResult tsmf_coords_write_sim(struct tsmf_info *tsmf)
{
  struct section_info sxn = {0};
  ts_sim *sim;
  struct ts_dof_map dof_map = {0};
  struct tsmf_coords_buffer cbuf = {0};
  TSResult err = TS_OK;

  if (sim = tsmf->sim, sim == NULL)
  {
    err = TS_NEW_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write simulation coordinates: no simulation");
    goto cleanup;
  }

  err = ts_dof_map_init(&dof_map, sim), ON_ERROR_ADD_FRAME(err);
  if (IS_ERROR(err))
  {
    goto cleanup;
  }

  if (!tsmf_coords_buffer_resize(&cbuf, dof_map.num_dof))
  {
    err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "cannot allocate coord buffer for %zu degrees of freedom",
        dof_map.num_dof);
    goto cleanup;
  }

  ts_dof_to_vec(sim, &dof_map, cbuf.coords, TS_ROT_AXIS_ANGLE);
  cbuf.nparticle = dof_map.num_free_particles;
  cbuf.nrigid = dof_map.num_free_rigids;
  cbuf.nrotate = dof_map.num_rot_rigids;

  err = tsmf_coords_write_coords(tsmf, &cbuf), ON_ERROR_ADD_FRAME(err);
  if (IS_ERROR(err))
  {
    goto cleanup;
  }

cleanup:
  ts_dof_map_finalize(&dof_map);
  tsmf_coords_buffer_release(&cbuf);

  return err;
}

TSResult tsmf_coords_write_coords(struct tsmf_info *tsmf,
    struct tsmf_coords_buffer *cbuf)
{
  struct section_info sxn;
  uint32_t n;

  n = cbuf_ndof(cbuf);

  RETURN_ON_ERROR(
      coords_begin(&sxn, tsmf,
        cbuf->nparticle, cbuf->nrigid, cbuf->nrotate,
        0 /* flags */));

  pack_f64v(sxn.buf, cbuf->coords, n);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

TSResult tsmf_coords_read(struct tsmf_info *tsmf,
    struct tsmf_coords_buffer *cbuf)
{
  struct pbuf *buf = NULL;
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  uint32_t i,n;
  uint32_t nrigid,nparticle,nrotate;
  uint16_t flags;
  long soff;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_COORDINATES));

  buf = sxn.buf;
  pbuf_readn(buf, TSMF_COORDS_HEADER_SIZE);

  /* XXX - ignore flags for now */
  nparticle = unpack_u32(buf);
  nrigid    = unpack_u32(buf);
  nrotate   = unpack_u32(buf);
  flags     = unpack_u16(buf);
  unpack_unused(buf, 10);

  n = coords_ndof(nparticle,nrigid,nrotate);
  if (cbuf->cap < n && !tsmf_coords_buffer_resize(cbuf, n))
  {
    err = TS_NEW_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate a coordinate buffer for %zu coordinates",
        (size_t)n);
    goto error;
  }

  unpack_f64v(buf, cbuf->coords, n);
  cbuf->flags = flags;
  cbuf->nparticle = nparticle;
  cbuf->nrigid = nrigid;
  cbuf->nrotate = nrotate;

  section_end_read(&sxn);
  return TS_OK;

buf_error:
  err = buf->error;
  /* continue to error: */

error:
  section_end_read(&sxn);
  return err;
}

TSResult tsmf_coords_read_into_sim(struct tsmf_info *tsmf)
{
  struct pbuf *buf = NULL;
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  ts_sim *sim;
  uint32_t i,n;
  uint32_t nrigid,nparticle,nrotate;
  uint16_t flags;
  long soff;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot read coords into simulation: no simulation");
  }

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_COORDINATES));

  buf = sxn.buf;
  pbuf_readn(buf, TSMF_COORDS_HEADER_SIZE);

  /* XXX - ignore flags for now */
  nparticle = unpack_u32(buf);
  nrigid    = unpack_u32(buf);
  nrotate   = unpack_u32(buf);
  flags     = unpack_u16(buf);
  unpack_unused(buf, 10);

  if ((flags & TSMF_COORDS_FLAG_ALL) == 0)
  {
    err = TS_NEW_ERROR_MESG(TS_ERR_UNIMPLEMENTED,
        "coords error: reading DoF coordinates (save as ALL)");
    goto error;
  }

  if (nrigid != sim->num_rigid_bodies)
  {
    err = TS_NEW_ERRORF(TS_ERR_SIMULATION_STATE,
        "invalid coordinates: simulation has %zu rigid bodies, coordinates have %zu",
        sim->num_rigid_bodies, (size_t)nrigid);
    goto error;
  }

  if (nparticle != ts_sim_num_free_particles(sim))
  {
    err = TS_NEW_ERRORF(TS_ERR_SIMULATION_STATE,
        "invalid coordinates: simulation has %zu free particles, coordinates have %zu",
        ts_sim_num_free_particles(sim), (size_t)nparticle);
    goto error;
  }

  if (nrotate != 0)
  {
    err = TS_NEW_ERROR_MESG(TS_ERR_SIMULATION_STATE,
        "invalid coordinates: require nrotate == 0 for ALL coordinates");
    goto error;
  }

  for (i=0; i < nrigid; i++)
  {
    vector3 *pos;
    quat4 *ori;
    double coords[6] = {0};

    pbuf_readn(buf, TSMF_COORDS_RIGID_SIZE);
    if (pbuf_err(buf))
    {
      goto error;
    }

    unpack_f64v(sxn.buf, &coords[0], ARRAYLEN(coords));
    pos = RIGID_POS(sim,i);
    ori = RIGID_ORI(sim,i);

    vector3_set_with(pos, coords[0], coords[1], coords[2]);
    quat4_from_axis_angle_vec(ori, &coords[3]);
  }

  for (n=sim->num_particles, i=0; i < n; i++)
  {
    ts_particle *p;
    vector3 *pos;
    double pv[3] = {0};

    p = ts_get_particle(sim, i);
    if (p->rigid_body_index != TS_NULL_RIGID_BODY)
    {
      continue;
    }

    pbuf_readn(buf, TSMF_COORDS_PARTICLE_SIZE);
    unpack_f64v(sxn.buf, &pv[0], ARRAYLEN(pv));

    pos = PARTICLE_POS(sim,i);
    vector3_set_with(pos, pv[0], pv[1], pv[2]);
  }

  RETURN_ON_ERROR(section_end_read(&sxn));
  return TS_OK;

error:
  section_end_read(&sxn);
  return err;
}

struct tsmf_coords_buffer *tsmf_coords_buffer_new(uint32_t cap)
{
  struct tsmf_coords_buffer *cbuf = NULL;
  double *v = NULL;

  if (v = ts_calloc(NO_ALLOCATOR, cap, sizeof *v), v == NULL)
  {
    goto error;
  }

  if (cbuf = ts_alloc(NO_ALLOCATOR, sizeof *cbuf), cbuf == NULL)
  {
    goto error;
  }

  cbuf->nrigid = 0;
  cbuf->nparticle = 0;
  cbuf->cap = cap;
  cbuf->coords = v;

  return cbuf;

error:
  ts_free(NO_ALLOCATOR, cbuf);
  ts_free(NO_ALLOCATOR, v);

  return NULL;
}

int tsmf_coords_buffer_resize(struct tsmf_coords_buffer *cbuf, uint32_t cap)
{
  double *v;
  uint32_t n;

  if (cbuf == NULL)
  {
    return 0;
  }

  n = cbuf_ndof(cbuf);
  if (cap < n)
  {
    return 0;
  }

  v = ts_realloc(NO_ALLOCATOR, cbuf->coords, cap * sizeof *v);
  if (v == NULL)
  {
    return 0;
  }

  cbuf->coords = v;
  cbuf->cap = cap;
  return 1;
}

void tsmf_coords_buffer_release(struct tsmf_coords_buffer *cbuf)
{
  ts_free(NO_ALLOCATOR, cbuf->coords);
  *cbuf = (struct tsmf_coords_buffer){0};
}

void tsmf_coords_buffer_delete(struct tsmf_coords_buffer *cbuf)
{
  if (cbuf)
  {
    tsmf_coords_buffer_release(cbuf);
    ts_free(NO_ALLOCATOR, cbuf);
  }
}


TSResult tsmf_btypes_buffer_initialize(struct tsmf_rbtypes_buffer *buf,
    uint32_t tcap, uint32_t vcap)
{
  struct ts_rigid_body_type *rbtypes;
  vector3 *vlist;

  rbtypes = ts_calloc(NO_ALLOCATOR, tcap, sizeof *rbtypes);
  if (rbtypes == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu elements for rigid body types",
        (size_t)tcap);
  }

  vlist   = ts_calloc(NO_ALLOCATOR, vcap, sizeof *vlist);
  if (vlist == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu elements for rigid body site positions",
        (size_t)vcap);
  }

  buf->rbtypes = rbtypes;
  buf->tlen = 0;
  buf->tcap = tcap;

  buf->vlist = vlist;
  buf->vlen = 0;
  buf->vcap = vcap;

  return TS_OK;
}

void tsmf_btypes_buffer_finalize(struct tsmf_rbtypes_buffer *buf)
{
  assert(buf != NULL);
  ts_free(NO_ALLOCATOR, buf->rbtypes);
  buf->rbtypes = NULL;
  buf->tlen = buf->tcap = 0;

  ts_free(NO_ALLOCATOR, buf->vlist);
  buf->vlist = NULL;
  buf->vlen = buf->vcap = 0;
}

TSResult tsmf_btypes_buffer_resize(struct tsmf_rbtypes_buffer *buf,
    uint32_t tcap, uint32_t vcap)
{
  struct ts_rigid_body_type *rbtypes;
  vector3 *vlist;

  rbtypes = ts_realloc(NO_ALLOCATOR, buf->rbtypes, tcap * sizeof *rbtypes);
  if (rbtypes == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu elements for rigid body types",
        (size_t)tcap);
  }
  buf->rbtypes = rbtypes;
  buf->tcap = tcap;
  if (buf->tlen > tcap)
  {
    buf->tlen = tcap;
  }

  vlist   = ts_realloc(NO_ALLOCATOR, buf->vlist, vcap * sizeof *vlist);
  if (vlist == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate %zu elements for rigid body site positions",
        (size_t)vcap);
  }
  buf->vlist = vlist;
  buf->vcap = vcap;
  if (buf->vlen > vcap)
  {
    buf->vlen = vcap;
  }

  return TS_OK;
}

TSResult tsmf_rbtypes_write(struct tsmf_info *tsmf)
{
  ts_sim *sim;
  struct section_info sxn;
  uint32_t i, data_size, npad, ntypes, nsites;
  struct ts_rigid_body_type *rbtypes;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write rigid body types: no simulation");
  }

  data_size = TSMF_RBTYPES_HEADER_SIZE;
  ntypes = sim->rigid_body_types.size;
  nsites = 0;
  rbtypes = sim->rigid_body_types.data;
  for (i=0; i < ntypes; i++)
  {
    uint32_t ns,namelen,namepad;

    // XXX - check name length
    namelen = strlen(rbtypes[i].name);


    ns = rbtypes[i].num_sites;
    data_size += TSMF_RBTYPES_ENTRY_FIXED_SIZE;
    data_size += ns * TSMF_RBTYPES_SITE_SIZE;
    data_size += namelen;

    namepad = (8 - namelen % 8) % 8;
    data_size += namepad;

    nsites += ns;
  }

  npad = (16 - (data_size %16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_RB_TYPES, 0, data_size));

  pack_u32(sxn.buf, ntypes);
  pack_u32(sxn.buf, nsites);
  pack_u16(sxn.buf, 0); /* flags */
  pack_unused(sxn.buf, 6);

  for (i=0; i < ntypes; i++)
  {
    uint32_t j,ns;
    uint16_t namelen,namepad;

    ns = rbtypes[i].num_sites;
    namelen = strlen(rbtypes[i].name);
    pack_u16(sxn.buf,namelen);
    pack_u16(sxn.buf,rbtypes[i].type);
    pack_u32(sxn.buf,ns);

    pack_f64(sxn.buf,rbtypes[i].I0);
    pack_f64(sxn.buf,rbtypes[i].rotfric);

    for (j=0; j < ns; j++)
    {
      pack_f64(sxn.buf,VECTOR3_X(&rbtypes[i].ref_vectors[j]));
      pack_f64(sxn.buf,VECTOR3_Y(&rbtypes[i].ref_vectors[j]));
      pack_f64(sxn.buf,VECTOR3_Z(&rbtypes[i].ref_vectors[j]));
    }

    pack_chars(sxn.buf,rbtypes[i].name,namelen);
    namepad = (8 - namelen % 8) % 8;
    pack_fill(sxn.buf,'%',namepad);
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

TSResult tsmf_rbtypes_read(struct tsmf_info *tsmf,
    struct tsmf_rbtypes_buffer *rbbuf)
{
  struct ts_rigid_body_type *rbtypes = NULL;
  struct section_info sxn = {0};
  TSResult err = TS_OK;
  struct pbuf *buf = NULL;
  uint32_t i, ntypes = 0, site_off = 0, nsites = 0;;
  uint16_t flags;
  long soff;

  assert(tsmf != NULL);
  assert(tsmf->f != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_RB_TYPES));

  buf = sxn.buf;
  pbuf_readn(buf, TSMF_RBTYPES_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  /* XXX - ignore flags for now */
  ntypes = unpack_u32(buf);
  nsites = unpack_u32(buf);
  flags  = unpack_u16(buf);
  unpack_unused(buf,6);

  if (pbuf_err(buf))
  {
    err = buf->error;
    goto error;
  }

  err = tsmf_btypes_buffer_initialize(rbbuf,ntypes,nsites);
  if (IS_ERROR(err))
  {
    goto error;
  }

  for (i=0; i < ntypes; i++)
  {
    uint32_t j,num_sites;
    uint16_t namelen,namepad;
    char *name;

    pbuf_readn(buf, TSMF_RBTYPES_ENTRY_FIXED_SIZE);
    if (pbuf_err(buf))
    {
      err = buf->error;
      goto error;
    }

    namelen                     = unpack_u16(buf);
    rbbuf->rbtypes[i].type      = unpack_u16(buf);
    num_sites                   = unpack_u32(buf);
    rbbuf->rbtypes[i].num_sites = num_sites;
    rbbuf->rbtypes[i].I0        = unpack_f64(buf);
    rbbuf->rbtypes[i].rotfric   = unpack_f64(buf);

    namepad = (8 - namelen % 8) % 8;

    rbbuf->rbtypes[i].ref_vectors = &rbbuf->vlist[rbbuf->vlen];

    for (j=0; j < num_sites; j++)
    {
      VECTOR3_X(&rbbuf->vlist[rbbuf->vlen]) = unpack_f64(buf);
      VECTOR3_Y(&rbbuf->vlist[rbbuf->vlen]) = unpack_f64(buf);
      VECTOR3_Z(&rbbuf->vlist[rbbuf->vlen]) = unpack_f64(buf);

      if (pbuf_err(buf))
      {
        goto buf_error;
      }

      rbbuf->vlen++;
      assert(rbbuf->vlen <= rbbuf->vcap);
    }

    name  = ts_alloc(NO_ALLOCATOR, namelen+1);
    if (name == NULL)
    {
      err = TS_NEW_ERROR_MESG(TS_ERR_OUT_OF_MEMORY,
          "error: could not allocate memory for rigid body type names");
      goto error;
    }

    rbbuf->rbtypes[i].name = name;

    unpack_chars(buf, name, namelen+1);
    unpack_unused(buf, namepad);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    rbbuf->tlen++;
  }

  if (err = section_end_read(&sxn), IS_ERROR(err))
  {
    goto error;
  }

  return TS_OK;

buf_error:
  err = buf->error;
  /* continue to error: */

error:
  tsmf_btypes_buffer_finalize(rbbuf);
  section_end_read(&sxn);
  return err;
}


TSResult tsmf_plist_write(struct tsmf_info *tsmf)
{
  ts_sim *sim;
  struct section_info sxn = {0};
  uint32_t i, npart, npad, data_size;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write particle list: no simulation");
  }

  npart = sim->num_particles;
  data_size = TSMF_PLIST_HEADER_SIZE + npart * TSMF_PLIST_ENTRY_SIZE;
  npad = (16 - (data_size % 16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_PARTICLE_LIST, 0, data_size));

  pack_u32(sxn.buf, npart);
  pack_u16(sxn.buf, 0); /* flags */
  pack_unused(sxn.buf, 10);

  for (i=0; i < npart; i++)
  {
    size_t ptype;

    // XXX - check range (size_t vs uint32_t)
    ptype = ts_particle_type_index(sim, sim->particles[i].type);
    pack_u32(sxn.buf, ptype);
    pack_u32(sxn.buf, sim->particles[i].flags);
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

TSResult tsmf_plist_read(struct tsmf_info *tsmf,
    struct tsmf_plist_buf **plistp)
{
  struct section_info sxn = {0};
  struct tsmf_plist_buf *plist = NULL;
  TSResult err = TS_OK;
  struct pbuf *buf;
  uint32_t flags, i, npart, npad, data_size;

  assert(tsmf != NULL);
  assert(plistp != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_PARTICLE_LIST));

  buf = sxn.buf;
  assert(buf != NULL);

  pbuf_readn(buf, TSMF_PLIST_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  npart = unpack_u32(buf);
  flags = unpack_u16(buf); /* flags */
  unpack_unused(buf, 10);

  plist = ts_alloc(NO_ALLOCATOR,
      sizeof *plist + npart * sizeof plist->entries[0]);
  plist->n = npart;
  plist->flags = flags;

  for (i=0; i < npart; i++)
  {
    pbuf_readn(buf, TSMF_PLIST_ENTRY_SIZE);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    plist->entries[i].type  = unpack_u32(buf);
    plist->entries[i].flags = unpack_u32(buf);
  }

  npad = (16 - (TSMF_PLIST_HEADER_SIZE + npart*TSMF_PLIST_ENTRY_SIZE)%16)%16;
  if (npad > 0)
  {
    unpack_unused(buf,npad);
  }

  if (err = section_end_read(&sxn), IS_ERROR(err))
  {
    goto error;
  }

  *plistp = plist;
  return TS_OK;

buf_error:
  err = buf->error;

error:
  ts_free(NO_ALLOCATOR, plist);
  section_end_read(&sxn);
  return err;
}

TSResult tsmf_rblist_write(struct tsmf_info *tsmf)
{
  ts_sim *sim;
  struct section_info sxn = {0};
  uint32_t i, nrigid, npart, npad, data_size;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write rigid body list: no simulation");
  }

  nrigid  = sim->num_rigid_bodies;
  data_size = TSMF_RBLIST_HEADER_SIZE + nrigid * TSMF_RBLIST_ENTRY_FIXED_SIZE;

  npart = 0;
  for (i=0; i < nrigid; i++)
  {
    uint32_t np;
    np = 1 + sim->rigid_bodies[i].type_ptr->num_sites;

    npart += np;
    data_size += np * TSMF_RBLIST_ENTRY_SITE_SIZE;
  }

  npad = (16 - (data_size % 16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_RB_LIST, 0, data_size));

  pack_u32(sxn.buf, nrigid);
  pack_u32(sxn.buf, npart);
  pack_u16(sxn.buf, 0); /* flags */
  pack_unused(sxn.buf, 6);

  for (i=0; i < nrigid; i++)
  {
    struct ts_rigid_body_type *rbtype;
    uint32_t j,nsites;
    uint32_t rbt;

    rbtype = sim->rigid_bodies[i].type_ptr;
    // FIXME - check for uint32_t overflow!
    rbt = ts_rigid_body_type_index(sim, rbtype);

    pack_u32(sxn.buf, rbt);
    pack_u32(sxn.buf, sim->rigid_bodies[i].flags);
    // XXX - check for uint32_t overflow
    pack_u32(sxn.buf, sim->rigid_bodies[i].com_particle);

    nsites = rbtype->num_sites;
    for (j=0; j < nsites; j++)
    {
      // XXX - check for uint32_t overflow
      pack_u32(sxn.buf, sim->rigid_bodies[i].site_indices[j]);
    }
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

struct tsmf_rblist_buf *tsmf_rblist_buf_new(uint32_t nrigid, uint32_t npart)
{
  struct tsmf_rblist_buf *buf;
  size_t sz,off,npad;

  sz = sizeof *buf + nrigid * sizeof buf->entries[0];
  
  npad = (U32_ALIGN - (sz % U32_ALIGN)) % U32_ALIGN;
  sz += npad;
  off = sz;
  sz += npart * sizeof (uint32_t);

  buf = ts_alloc(NO_ALLOCATOR, sz);
  if (buf != NULL)
  {
    buf->nrigid = nrigid;
    buf->flags = 0;
    buf->npart = npart;
    buf->plist = (uint32_t *)((char *)buf + off);
  }
  return buf;
}

void tsmf_rblist_buf_delete(struct tsmf_rblist_buf *buf)
{
  ts_free(NO_ALLOCATOR, buf);
}

TSResult tsmf_rblist_read(struct tsmf_info *tsmf,
    struct tsmf_rblist_buf **rblistp)
{
  struct section_info sxn = {0};
  struct tsmf_rblist_buf *rblist = NULL;
  TSResult err = TS_OK;
  ts_sim *sim;
  struct pbuf *buf;
  uint32_t flags, i, nrigid, poff, npart, npad;

  assert(tsmf != NULL);
  assert(rblistp != NULL);

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot read rigid body list: no simulation");
  }

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_RB_LIST));

  buf = sxn.buf;
  assert(buf != NULL);

  pbuf_readn(buf, TSMF_RBLIST_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  nrigid = unpack_u32(buf);
  npart  = unpack_u32(buf);
  flags  = unpack_u16(buf); /* flags */
  unpack_unused(buf, 6);

  rblist = tsmf_rblist_buf_new(nrigid,npart);
  rblist->flags = flags;

  for (poff=0, i=0; i < nrigid; i++)
  {
    uint32_t rbt,j,nsites;
    struct ts_rigid_body_type *rbtype;

    pbuf_readn(buf, TSMF_RBLIST_ENTRY_FIXED_SIZE);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    rbt = unpack_u32(sxn.buf);
    rblist->entries[i].type = rbt;
    rblist->entries[i].flags = unpack_u32(sxn.buf);

    if (rbtype = ts_get_rigid_body_type(sim, rbt), rbtype == NULL)
    {
      err = TS_NEW_ERRORF(TS_ERR_INVALID_INDEX,
          "rigid body %lu: invalid rigid body type: %lu",
          (unsigned long)i, (unsigned long)rbt);
      goto error;
    }

    nsites = 1+rbtype->num_sites;
    rblist->entries[i].pind0 = poff;
    for (j=0; j < nsites; j++, poff++)
    {
      rblist->plist[poff] = unpack_u32(sxn.buf);
    }

    if (pbuf_err(buf))
    {
      goto buf_error;
    }
  }

  npad = TSMF_RBLIST_HEADER_SIZE + nrigid*TSMF_RBLIST_ENTRY_FIXED_SIZE + npart*TSMF_RBLIST_ENTRY_SITE_SIZE;
  npad = (16 - (npad%16))%16;
  if (npad > 0)
  {
    unpack_unused(buf,npad);
  }

  if (err = section_end_read(&sxn), IS_ERROR(err))
  {
    goto error;
  }

  *rblistp = rblist;
  return TS_OK;

buf_error:
  err = buf->error;

error:
  tsmf_rblist_buf_delete(rblist);
  section_end_read(&sxn);
  return err;
}

struct tsmf_conn_list *tsmf_conn_list_new(uint32_t nbonds, uint32_t flags)
{
  struct tsmf_conn_list *clist = NULL;

  clist = ts_alloc(NO_ALLOCATOR, sizeof *clist +
      nbonds * sizeof clist->bonds[0]);

  if (clist == NULL)
  {
    return clist;
  }

  clist->flags = flags;
  clist->nbonds = nbonds;

  return clist;
}

void tsmf_conn_list_delete(struct tsmf_conn_list *clist)
{
  ts_free(NO_ALLOCATOR, clist);
}

TSResult tsmf_conn_write(struct tsmf_info *tsmf)
{
  struct section_info sxn = {0};
  struct tsmf_conn_list *clist = NULL;
  ts_sim *sim;
  uint32_t i, bi, nbonds, npad, data_size;
  size_t pind;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write rigid body list: no simulation");
  }

  nbonds = 0;
  for (pind=0; pind < sim->num_particles; pind++)
  {
    ts_particle *p = ts_get_particle(sim,pind);
    nbonds += p->num_bonds;
  }

  // the above sum double counts the bonds
  nbonds /= 2;

  if (clist = tsmf->clist, clist != NULL)
  {
    if (clist->nbonds != nbonds)
    {
      tsmf_conn_list_delete(clist);
      tsmf->clist = clist = NULL;
    }
  }

  if (clist == NULL)
  {
    clist = tsmf_conn_list_new(nbonds, 0);
    tsmf->clist = clist;
  }

  if (clist == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_OUT_OF_MEMORY,
        "could not allocate connectivity list");
  }

  clist->nbonds = nbonds;
  clist->flags  = 0;

  data_size = TSMF_CONN_HEADER_SIZE + nbonds * TSMF_CONN_ENTRY_SIZE;
  npad = (16 - (data_size % 16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_CONNECTIVITY, 0, data_size));

  pack_u32(sxn.buf, nbonds);
  pack_u32(sxn.buf, 0); /* flags */
  pack_unused(sxn.buf, 8);

  for (bi=0,pind=0; pind < sim->num_particles; pind++)
  {
    unsigned int bj;

    ts_particle *p = ts_get_particle(sim,pind);
    struct ts_bond_info *bonds = p->bonds;

    for (bj=0; bj < p->num_bonds; bj++)
    {
      size_t pi,pj;
      pi = ts_particle_index(sim, p->bonds[bj].pi);
      pj = ts_particle_index(sim, p->bonds[bj].pj);

      if (pi > pj)
      {
        size_t ptmp = pi;
        pi = pj;
        pj = ptmp;
      }

      if (pi != pind)
      {
        continue;
      }

      // FIXME - check for uint32_t overflow
      pack_u32(sxn.buf, pi);
      pack_u32(sxn.buf, pj);
      pack_u32(sxn.buf, bonds[bj].type.value);

      clist->bonds[bi].pi = pi;
      clist->bonds[bi].pj = pj;
      clist->bonds[bi].bt = bonds[bj].type.value;
      bi++;
    }
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

struct bond_iter
{
  struct ts_bond_info *bond;
  ts_sim *sim;
  size_t pind;
  int bind;
};

static inline const struct ts_bond_info *bond_iter_bond(struct bond_iter *iter)
{
  return iter->bond;
}

static int bond_iter_nextbond(struct bond_iter *iter, size_t pind0, int bind0)
{
  ts_sim *sim;
  ts_particle *p;
  struct ts_bond_info *bond;
  size_t pind,np;
  int bind,nb;

  sim = iter->sim;
  assert(sim != NULL);

  bond = NULL;
  np = sim->num_particles;
  for (pind=pind0, bind=bind0; pind < np; bind=0, pind++)
  {
    p = ts_get_particle(sim, pind);
    nb = p->num_bonds;

    for (; bind < nb; bind++)
    {
      size_t pi,pj,tmp;
      pi = TS_PIND(iter->sim,p->bonds[bind].pi);
      pj = TS_PIND(iter->sim,p->bonds[bind].pj);

      if (pi > pj)
      {
        tmp = pi; pi = pj; pj = tmp;
      }

      if (pi == pind)
      {
        bond = &p->bonds[bind];
        goto finish;
      }
    }
  }

finish:
  iter->pind = pind;
  iter->bind = bind;
  iter->bond = bond;
  return bond != NULL;
}

static void bond_iter_init(struct bond_iter *iter, ts_sim *sim)
{
  iter->sim = sim;
  bond_iter_nextbond(iter, 0, 0);
}

static int bond_iter_next(struct bond_iter *iter)
{
  return bond_iter_nextbond(iter, iter->pind, iter->bind+1);
}

struct bond_delta_iter
{
  struct bond_iter blist;
  struct tsmf_conn_list *clist;
  uint32_t ci;

  uint32_t pi;
  uint32_t pj;
  uint32_t bt;
};

static void bond_delta_iter_init(struct bond_delta_iter *iter,
    ts_sim *sim,
    struct tsmf_conn_list *clist)
{
  bond_iter_init(&iter->blist, sim);
  iter->clist = clist;
  iter->ci = 0;
  iter->pi = 0;
  iter->pj = 0;
  iter->bt = 0;
}

static int bond_delta_iter_next(struct bond_delta_iter *iter)
{
  ts_sim *sim;
  const struct ts_bond_info *bond;
  struct tsmf_conn_list *clist;
  uint32_t pi0,pj0,bt0, pi1,pj1,bt1;

  clist = iter->clist;
  sim = iter->blist.sim;

restart:
  // exit conditions
  if (bond = bond_iter_bond(&iter->blist), bond == NULL)
  {
    goto finish_clist;
  }

  if (iter->ci >= iter->clist->nbonds)
  {
    goto finish_blist;
  }

  pi0 = clist->bonds[iter->ci].pi;
  pj0 = clist->bonds[iter->ci].pj;
  bt0 = clist->bonds[iter->ci].bt;

  pi1 = TS_PIND(sim,bond->pi);
  pj1 = TS_PIND(sim,bond->pj);
  bt1 = bond->type.value;
  if (pi1 > pj1)
  {
    uint32_t tmp = pi1;
    pi1 = pj1;
    pj1 = tmp;
  }

  // identical entries... either skip or update the bond.
  // in either case, we advance both the blist and clist iterators
  if (pi0 == pi1 && pj0 == pj1)
  {
    iter->ci++;
    bond_iter_next(&iter->blist);

    if (bt0 == bt1)
    {
      // skip this entry...
      goto restart;
    }

    iter->pi = pi1;
    iter->pj = pj1;
    iter->bt = bt1;

    return 1;
  }

  // pi0 < pi1 || (pi0 == pi1 && pj0 < pj1): bond exists in clist but
  // not in blist, which means that it has been deleted.
  //
  // Report that the bond type has changed to zero, advance clist.
  if (pi0 < pi1 || (pi0 == pi1 && pj0 < pj1))
  {
    iter->pi = pi0;
    iter->pj = pj0;
    iter->bt = 0;

    iter->ci++;
    return 1;
  }

  // pi0 >= pi1 || (pi0 == pi1 && pj0 > pj1): bond exists in blist but
  // not in clist, which means that the bond is new.
  //
  // report the new bond and advance blist
  if (pi0 >= pi1 || (pi0 == pi1 && pj0 > pj1))
  {
    iter->pi = pi1;
    iter->pj = pj1;
    iter->bt = bt1;

    bond_iter_next(&iter->blist);
    return 1;
  }

  // reaching this assertion indicates faulty logic
  assert(0 && "should never reach here");

finish_blist:
  iter->pi = TS_PIND(sim, bond->pi);
  iter->pj = TS_PIND(sim, bond->pj);
  if (iter->pi > iter->pj)
  {
    uint32_t tmp = iter->pi;
    iter->pi = iter->pj;
    iter->pj = tmp;
  }

  iter->bt = bond->type.value;
  bond_iter_next(&iter->blist);
  return 1;

finish_clist:
  if (iter->ci >= iter->clist->nbonds)
  {
    return 0;
  }

  // if we're finishing the clist, the bonds have been deleted, so bt = 0
  iter->pi = clist->bonds[iter->ci].pi;
  iter->pj = clist->bonds[iter->ci].pj;
  iter->bt = 0;
  iter->ci++;
  return 1;
}

TSResult tsmf_conn_write_delta(struct tsmf_info *tsmf)
{
  struct section_info sxn = {0};
  struct tsmf_conn_list *clist = NULL;
  struct bond_delta_iter iter = {{0},0};
  struct ts_bond_info *bond;
  ts_sim *sim;
  uint32_t i, bi, ci,cn, nbonds, npad, data_size;
  size_t pind;

  if (sim = tsmf->sim, sim == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write rigid body list: no simulation");
  }

  if (clist = tsmf->clist, clist == NULL)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "missing reference connectivity list");
  }

  nbonds = 0;
  bond_delta_iter_init(&iter, sim, clist);
  while (bond_delta_iter_next(&iter))
  {
    nbonds++;
  }

  data_size = TSMF_CONN_HEADER_SIZE + nbonds * TSMF_CONN_ENTRY_SIZE;
  npad = (16 - (data_size % 16)) % 16;
  data_size += npad;

  RETURN_ON_ERROR(
      section_begin(&sxn, tsmf, TSMF_SEC_CONNECTIVITY, 0, data_size));

  pack_u32(sxn.buf, nbonds);
  pack_u32(sxn.buf, TSMF_CONN_FLAGS_DELTA); /* flags */
  pack_unused(sxn.buf, 8);

  bond_delta_iter_init(&iter, sim, clist);
  while (bond_delta_iter_next(&iter))
  {
    pack_u32(sxn.buf, iter.pi);
    pack_u32(sxn.buf, iter.pj);
    pack_u32(sxn.buf, iter.bt);
  }

  pack_fill(sxn.buf, '$', npad);

  RETURN_ON_ERROR(
      section_end(&sxn));

  return TS_OK;
}

TSResult tsmf_conn_read(struct tsmf_info *tsmf,
    struct tsmf_conn_list **clistp)
{
  struct section_info sxn = {0};
  struct tsmf_conn_list *clist = NULL;
  TSResult err = TS_OK;
  struct pbuf *buf;
  uint32_t flags, i, nbonds, npad;

  assert(tsmf != NULL);
  assert(clistp != NULL);

  RETURN_ON_ERROR(
      section_begin_read(&sxn, tsmf, TSMF_SEC_CONNECTIVITY));

  buf = sxn.buf;
  assert(buf != NULL);

  pbuf_readn(buf, TSMF_CONN_HEADER_SIZE);
  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  nbonds = unpack_u32(buf);
  flags  = unpack_u32(buf);
  unpack_unused(buf, 8);

  if (pbuf_err(buf))
  {
    goto buf_error;
  }

  clist = tsmf_conn_list_new(nbonds,flags);

  if (clist == NULL)
  {
    err = TS_NEW_ERROR_MESG(TS_ERR_OUT_OF_MEMORY,
        "could not allocate connectivity list");
    goto error;
  }

  clist->nbonds = nbonds;
  clist->flags  = flags;

  for (i=0; i < nbonds; i++)
  {
    pbuf_readn(buf, TSMF_CONN_ENTRY_SIZE);
    if (pbuf_err(buf))
    {
      goto buf_error;
    }

    clist->bonds[i].pi = unpack_u32(sxn.buf);
    clist->bonds[i].pj = unpack_u32(sxn.buf);
    clist->bonds[i].bt = unpack_u32(sxn.buf);
  }

  npad = TSMF_CONN_HEADER_SIZE + nbonds*TSMF_CONN_ENTRY_SIZE;
  npad = (16 - (npad%16))%16;
  if (npad > 0)
  {
    unpack_unused(buf,npad);
  }

  if (err = section_end_read(&sxn), IS_ERROR(err))
  {
    goto error;
  }

  *clistp = clist;
  return TS_OK;

buf_error:
  err = buf->error;

error:
  ts_free(NO_ALLOCATOR, clist);
  section_end_read(&sxn);
  return err;
}

