#include "ts_io.h"
#include "ts_dof_map.h"
#include "ts_bonds.h"

enum FILE_MARKERS {
  STATE_BEGIN_MARKER = '['     ,
  STATE_END_MARKER   = ']'     ,

  STATE_HEADER       = '1'     ,

  BONDLIST_BEGIN_MARKER ,
  BONDLIST_END_MARKER   ,

  PARTICLE_BONDS_BEGIN_MARKER ,
  PARTICLE_BONDS_END_MARKER   ,

  RIGID_BONDS_BEGIN_MARKER    ,
  RIGID_BONDS_END_MARKER      ,
};

enum READING_STATE {
  READNING_NONE = 0,

  READING_BONDS ,
  READING_PARTICLE_BONDS,
  READING_RIGID_BODY_BONDS,
};

struct tstf_file
{
  FILE* f;
  ts_allocator* alloc;

  enum READING_STATE reading;
  size_t reading_counter;
};

struct tstf_file* tstf_open(FILE* f, ts_allocator* alloc, TSResult* errp)
{
  struct tstf_file* tstf_f = ts_alloc(alloc, sizeof(*tstf_f));
  if (tstf_f == NULL)
  {
    *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    return NULL;
  }

  tstf_f->f = f;
  tstf_f->alloc = alloc;

  return tstf_f;
}

void tstf_close(struct tstf_file* f)
{
  ts_free(f->alloc, f);
}

TSResult tstf_write_header(FILE* f)
{
  RETURN_ON_ERROR( 
      ts_write_header(f, 
        ts_str_to_magic(TSTF_MAGIC), TSTF_MAJOR, TSTF_MINOR) );

  return TS_OK;
}

TSResult tstf_check_header(FILE* f)
{
  RETURN_ON_ERROR( ts_check_file_header(f,
        ts_str_to_magic(TSTF_MAGIC),
        TSTF_MAJOR,
        TSTF_MINOR) );

  return TS_OK;
}

static inline void setup_file_mark_array(char mark[8], enum FILE_MARKERS marker)
{
  mark[0] = 'M';
  mark[1] = 'A';
  mark[2] = 'R';
  mark[3] = 'K';
  mark[4] = '_';
  mark[5] = '0';
  mark[6] = '0';
  mark[7] = marker;
}

static inline TSResult write_marker(FILE* f, enum FILE_MARKERS marker)
{
  char mstr[8];
  setup_file_mark_array(mstr, marker);

  if (fwrite(mstr, sizeof(mstr), 1, f) == 1)
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing file marker %d", (int)marker));
}

static inline TSResult check_marker(FILE* f, enum FILE_MARKERS marker)
{
  char mark[8];
  if (fread(mark, sizeof(mark), 1, f) != 1)
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error reading file marker %d", (int)marker));

  char mexpected[8];
  setup_file_mark_array(mexpected, marker);

  for (int i=0; i < 8; i++)
  {
    if (mark[i] != mexpected[i]) 
      return TS_NEW_ERROR2(TS_ERR_INVALID_VALUE,
          ("error while reading file marker",0));
  }

  return TS_OK;
}

struct write_bond_data {
  ts_sim* sim;
  struct ts_dof_map* dof_map;
  FILE* f;
};

TSResult write_bond_pair(void* arg, struct ts_bond_info* bi)
{
  struct write_bond_data* data = arg;

  uint32_t info[3];
  info[0] = ts_dof_map_p2o(data->sim, data->dof_map, TS_PIND(data->sim, bi->pi));
  info[1] = ts_dof_map_p2o(data->sim, data->dof_map, TS_PIND(data->sim, bi->pj));
  info[2] = (uint32_t)bondtype_to_int(bi->type);

  if (fwrite(info, sizeof(info[0]), ARRAYLEN(info), data->f) == ARRAYLEN(info))
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing particle bond info for particles %lu and %lu",
         (long)info[0], (long)info[1]));
}

static size_t p2site(const ts_rigid_body* rigid, size_t pind)
{
  if (pind == rigid->com_particle) return 0;
  for (size_t i=0; i < rigid->type_ptr->num_sites; i++)
  {
    if (pind == rigid->site_indices[i]) return i+1;
  }

  return TS_NULL_PARTICLE;
}

TSResult count_rigid_pair(void* arg, size_t ri, size_t rj,
    struct ts_rigid_body_bond_info* binfo)
{
  UNUSED_PARAM(ri);
  UNUSED_PARAM(rj);
  UNUSED_PARAM(binfo);
  uint32_t* np = arg;
  (*np)++;
  return TS_OK;
}

TSResult write_rigid_pair(void* arg, size_t ri, size_t rj,
    struct ts_rigid_body_bond_info* binfo)
{
  struct write_bond_data* data = arg;

  uint32_t info[5];
  info[0] = data->dof_map->num_free_particles + ri;
  info[1] = data->dof_map->num_free_particles + rj;
  info[2] = p2site(ts_get_rigid_body(data->sim,ri),binfo->pi);
  info[3] = p2site(ts_get_rigid_body(data->sim,rj),binfo->pj);
  info[4] = (uint32_t)bondtype_to_int(binfo->btype);

  if (fwrite(info, sizeof(info[0]), ARRAYLEN(info), data->f) == ARRAYLEN(info))
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing rigid bond info for rigid bodies %lu and %lu",
         (long)info[0], (long)info[1]));
}

static TSResult
foreach_particle_bond(ts_sim* sim, struct ts_dof_map* dof_map,
    TSResult(*func)(void* arg, struct ts_bond_info* bond), void* arg)
{
  for (size_t fi=0; fi < dof_map->num_free_particles; fi++)
  {
    size_t i = dof_map->fp_to_p[fi];
    ts_particle* p = ts_get_particle(sim,i);

    for (size_t b=0; b < p->num_bonds; b++)
    {
      if (p == p->bonds[b].pi)
      {
        RETURN_ON_ERROR3( func(arg, p->bonds+b) );
      }
    }
  }

  return TS_OK;
}

TSResult count_bond_pair(void* arg, struct ts_bond_info* bi)
{
  UNUSED_PARAM(bi);
  uint32_t* np = arg;
  (*np)++;
  return TS_OK;
}

static TSResult
write_free_particle_bonds(ts_sim* sim, FILE* f, struct ts_dof_map* dof_map)
{
  uint32_t nbonds = 0;
  RETURN_ON_ERROR( foreach_particle_bond(sim,dof_map,count_bond_pair,&nbonds) );

  RETURN_ON_ERROR( write_marker(f, PARTICLE_BONDS_BEGIN_MARKER) );

  if (fwrite(&nbonds, sizeof(nbonds), 1, f) == 1)
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing bond count for particles", 0));

  struct write_bond_data data = { .sim = sim, .dof_map = dof_map, .f = f };
  RETURN_ON_ERROR( foreach_particle_bond(sim,dof_map,write_bond_pair,&data) );

  RETURN_ON_ERROR( write_marker(f, PARTICLE_BONDS_END_MARKER) );

  return TS_OK;
}

static TSResult
foreach_rigid_body_bond(ts_sim* sim, struct ts_dof_map* dof_map,
    TSResult (*func)(void* arg, size_t ri, size_t rj,
      struct ts_rigid_body_bond_info* binfo), void* arg)
{
  UNUSED_PARAM(dof_map);
  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    ts_rigid_body* ri = ts_get_rigid_body(sim,i);
    for (size_t j=i+1; j < sim->num_rigid_bodies; j++)
    {
      ts_rigid_body* rj = ts_get_rigid_body(sim,j);

      struct ts_rigid_body_bond_info binfo;
      if (ts_rigid_body_find_first_bond(sim, ri,rj, &binfo))
      {
        RETURN_ON_ERROR3( func(arg,
              TS_RIGID_IND(sim,ri),
              TS_RIGID_IND(sim,rj),
              &binfo) );
      }
    }
  }

  return TS_OK;
}

TSResult write_rigid_body_bonds(ts_sim* sim, FILE* f, struct ts_dof_map* dof_map)
{
  uint32_t nbonds = 0;
  RETURN_ON_ERROR( foreach_rigid_body_bond(sim,dof_map,count_rigid_pair,&nbonds) );

  RETURN_ON_ERROR( write_marker(f, RIGID_BONDS_BEGIN_MARKER) );

  if (fwrite(&nbonds, sizeof(nbonds), 1, f) == 1)
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing bond count for rigid bodies", 0));

  struct write_bond_data data = { .sim = sim, .dof_map = dof_map, .f = f };
  RETURN_ON_ERROR( foreach_rigid_body_bond(sim,dof_map,write_rigid_pair,&data) );

  RETURN_ON_ERROR( write_marker(f, RIGID_BONDS_END_MARKER) );

  return TS_OK;
}

static TSResult write_state_header(ts_sim* sim, FILE* f, struct ts_dof_map* dof_map)
{
  struct tstf_state_header hdr;
  RETURN_ON_ERROR3( tspf_set_timepoint_header( &(hdr.sim_state), sim ) );

  hdr.num_free_particles = dof_map->num_free_particles;
  hdr.num_rigid_bodies   = sim->num_rigid_bodies;

  hdr.flags = TSTF_HDRFLAG_NONE;
  memset(hdr.reserved, 0, sizeof(hdr.reserved));

  RETURN_ON_ERROR( write_marker(f, STATE_HEADER) );

  if (fwrite(&hdr, sizeof(hdr), 1, f) == 1)
    return TS_OK;
  else
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error writing state header", 0));

  return TS_OK;
}

TSResult tstf_read_state_header(struct tstf_file* f, struct tstf_state_header* hdr)
{
  memset(hdr,0,sizeof(*hdr));

  RETURN_ON_ERROR( check_marker(f->f, STATE_HEADER) );

  if (fread(hdr,sizeof(*hdr),1,f->f) != 1)
    return TS_NEW_ERROR2(TS_ERR_WRITE_NOT_FINISHED,
        ("error reading state header", 0));

  return TS_OK;
}

static TSResult write_bonds(ts_sim* sim, FILE* f, struct ts_dof_map* dof_map)
{
  RETURN_ON_ERROR3( write_marker(f, BONDLIST_BEGIN_MARKER) );

  RETURN_ON_ERROR3( write_free_particle_bonds(sim,f,dof_map) );
  RETURN_ON_ERROR3( write_rigid_body_bonds(sim,f,dof_map) );

  RETURN_ON_ERROR3( write_marker(f, BONDLIST_END_MARKER) );

  return TS_OK;
}

TSResult tstf_start_reading_bonds(struct tstf_file* f)
{
  RETURN_ON_ERROR( check_marker(f->f, BONDLIST_BEGIN_MARKER) );
  return TS_OK;
}

TSResult tstf_end_reading_bonds(struct tstf_file* f)
{
  RETURN_ON_ERROR( check_marker(f->f, BONDLIST_END_MARKER) );
  return TS_OK;
}

TSResult tstf_read_bonds(struct tstf_file* f,
    size_t n, struct tstf_bond_info* binfo, size_t* nread)
{
  UNUSED_PARAM(f);
  UNUSED_PARAM(n);
  UNUSED_PARAM(binfo);
  UNUSED_PARAM(nread);
  /*
  switch (f->reading)
  {
    case READING_NONE:
      break;
    case READING_BONDS:
      break;

    case READING_PARTICLE_BONDS:
      break;
    case READING_RIGID_BODY_BONDS:
      break;
  }
  */

  STOP_ITERATING();
}

static TSResult write_positions(ts_sim* sim, FILE* f, struct ts_dof_map* dof_map)
{
  UNUSED_PARAM(sim);
  UNUSED_PARAM(f);
  UNUSED_PARAM(dof_map);

  return TS_OK;
}


TSResult tstf_write_state(ts_sim* sim, FILE* f)
{
  struct ts_dof_map dof_map;
  RETURN_ON_ERROR3( ts_dof_map_init(&dof_map, sim) );

#define CLEANUP_ON_ERROR { ts_dof_map_finalize(&dof_map); }

  RETURN_ON_ERROR3( write_state_header(sim,f,&dof_map) );
  RETURN_ON_ERROR3( write_bonds(sim,f,&dof_map) );
  RETURN_ON_ERROR3( write_positions(sim,f,&dof_map) );

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

TSResult tstf_init(TSTF* tstf, size_t maxb, size_t maxp, size_t maxr)
{
  UNUSED_PARAM(tstf);
  UNUSED_PARAM(maxb);
  UNUSED_PARAM(maxp);
  UNUSED_PARAM(maxr);
  return TS_OK;
}

TSResult tstf_finalize(TSTF* tstf)
{
  UNUSED_PARAM(tstf);
  return TS_OK;
}

TSResult tstf_read_state(TSTF* tstf, FILE* f)
{
  UNUSED_PARAM(tstf);
  UNUSED_PARAM(f);

  return TS_OK;
}

