#include "ts_veclist.h"

#include "ts_common.h"
#include "ts_errors.h"

#include <math.h>
#include <string.h>
#include <vector3.h>

#define NORM_TOL 1e-8

#ifdef TS_USE_SSE
#  include <xmmintrin.h>
#endif

#include "ts_malloc_protect.h"

// IF_SSE and IF_NOT_SSE are short convenience macros meant to remove
// some of the unsightly ifdef...endif pairs when they surround short
// statements that select code based on SSE availability or not.
//
// Don't use more than a single statement for their arguments
//
// SELECT_SSE(sse,not_sse) takes two statments and issues code for one
// or the other, depending on whether TS_USE_SSE has been set.  Again,
// this is for single statements.  Use ifdef...else...endif for anything
// complex
//
#ifdef TS_USE_SSE
#  define IF_SSE( stmt ) stmt
#  define IF_NOT_SSE( stmt ) do {} while(0)
#  define SELECT_SSE( stmt_sse, stmt_not_sse ) stmt_sse
#  define USING_SSE 1
#else
#  define IF_SSE( stmt ) do {} while(0)
#  define IF_NOT_SSE( stmt ) stmt
#  define SELECT_SSE( stmt_sse, stmt_not_sse ) stmt_not_sse
#  define USING_SSE 0
#endif

#ifdef TS_USE_SSE
static void veclist_add_sse2(VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;

  double*restrict aptr = a->vX;
  double*restrict bptr = b->vX;

  for(size_t i = 0; i+1 < n; i += 2)
  {
    __m128d av = _mm_load_pd(aptr+i); 
    __m128d bv = _mm_load_pd(bptr+i); 

    _mm_store_pd(aptr+i, _mm_add_pd(av,bv));
  }

  if (n % 2 == 1) {
    aptr[n-1] += bptr[n-1];
  }
}

static void veclist_add3_sse2(VectorList*restrict dst, const VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;
  double*restrict dptr = dst->vX;
  const double*restrict aptr = a->vX;
  const double*restrict bptr = b->vX;

  for(size_t i = 0; i+1 < n; i += 2)
  {
    __m128d av = _mm_load_pd(aptr+i);
    __m128d bv = _mm_load_pd(bptr+i);

    _mm_store_pd(dptr+i, _mm_add_pd(av,bv));
  }

  if (n % 2 == 1) {
    dptr[n-1] = aptr[n-1] + bptr[n-1];
  }
}
#endif

MAYBE_UNUSED
static void veclist_add_nonvec(VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;
  double*restrict av = a->vX;
  const double*restrict bv = b->vX;
  for(size_t i=0; i < n; i++)
  {
    av[i] += bv[i];
  }
}

MAYBE_UNUSED
static void veclist_add3_nonvec(VectorList*restrict dst, const VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;
  double*restrict dv = dst->vX;
  const double*restrict av = a->vX;
  const double*restrict bv = b->vX;

  for(size_t i=0; i < n; i++) {
    dv[i] = av[i] + bv[i];
  }
}

MAYBE_UNUSED
static void veclist_sub3_nonvec(VectorList*restrict dst, const VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;
  double*restrict dv = dst->vX;
  const double*restrict av = a->vX;
  const double*restrict bv = b->vX;

  for(size_t i=0; i < n; i++) {
    dv[i] = av[i] - bv[i];
  }
}

void veclist_add(VectorList*restrict a, const VectorList*restrict b)
{
  SELECT_SSE( veclist_add_sse2(a,b),  veclist_add_nonvec(a,b) );
}

void veclist_add3(VectorList*restrict dst, const VectorList*restrict a, const VectorList*restrict b)
{
  SELECT_SSE( veclist_add3_sse2(dst,a,b),  veclist_add3_nonvec(dst,a,b) );
}

void veclist_sub3(VectorList*restrict dst, const VectorList*restrict a, const VectorList*restrict b)
{
  // SELECT_SSE( veclist_add3_sse2(dst,a,b),  veclist_add3_nonvec(dst,a,b) );
  veclist_sub3_nonvec(dst,a,b);
}

void veclist_sub(VectorList*restrict a, const VectorList*restrict b)
{
  size_t n = 3*a->n;
  double*restrict aptr = a->vX;
  const double*restrict bptr = b->vX;
  for(size_t i=0; i < n; i++)
  {
    aptr[i] -= bptr[i];
  }
}

void veclist_zero(VectorList* vecList)
{
  if (vecList->n > 0)
  {
    memset(vecList->vX, 0, 3*(vecList->n)*sizeof(*(vecList->vX)));
  }
}

void veclist_dealloc(VectorList* vl)
{
  if (vl->n == 0) return;
  if (vl->vX != NULL) ts_free(NO_ALLOCATOR, vl->vX);

  vl->n = 0;
  vl->vX = NULL;
  vl->vY = NULL;
  vl->vZ = NULL;
}

VectorList* veclist_alloc(size_t n)
{
  VectorList* vl;
  
  if (vl = ts_alloc(NO_ALLOCATOR, sizeof(*vl) ), vl == NULL) {
    goto error;
  }

  memset(vl,0,sizeof(*vl));

  if (veclist_init(vl, n) < 0) {
    goto error;
  }

  return vl;

error:
  if (vl != NULL) {
    veclist_dealloc(vl);
  }

  return NULL;
}

int veclist_init(VectorList* vecList, size_t n)
{
  double* all_elts;

  if (n <= vecList->n) {
    veclist_zero(vecList);
    return 0;
  }

  all_elts = ts_realloc(NO_ALLOCATOR, vecList->vX, 3*n*sizeof(vecList->vX[0]));

  if (all_elts == NULL) {
    return -1;
  }

  vecList->n = n;
  vecList->vX = all_elts;
  vecList->vY = vecList->vX + n;
  vecList->vZ = vecList->vY + n;

  return 0;
}

void veclist_scaleby(VectorList*restrict vecList, double scl)
{
  for(size_t i=0; i < vecList->n; i++)
  {
    vecList->vX[i] *= scl;
    vecList->vY[i] *= scl;
    vecList->vZ[i] *= scl;
  }
}

void veclist_normalize(VectorList* vecList)
{
  for(size_t i=0; i < vecList->n; i++)
  {
    double mag = sqrt(
        square(vecList->vX[i]) +
        square(vecList->vY[i]) +
        square(vecList->vZ[i]));

    if (mag > NORM_TOL)
    {
      double inv_mag = 1.0/mag;
      vecList->vX[i] *= inv_mag;
      vecList->vY[i] *= inv_mag;
      vecList->vZ[i] *= inv_mag;
    }
  }
}

void veclist_distance_to(VectorList*restrict v0, const VectorList*restrict v1)
{
  tsAssert( (v0->n == v1->n) && "vector lists are not the same size" );

  for(size_t i = 0; i < v0->n; i++)
  {
    v0->vX[i] -= v1->vX[i];
    v0->vY[i] -= v1->vY[i];
    v0->vZ[i] -= v1->vZ[i];

    v0->vX[i] = (v0->vX[i] * v0->vX[i])
      + (v0->vY[i] * v0->vY[i])
      + (v0->vZ[i] * v0->vZ[i]);
    v0->vX[i] = sqrt( v0->vX[i] );
    // v0->vY[i] = v0->vZ[i] = 0.0;
  }
}

void veclist_sqdist_to(VectorList*restrict v0, const VectorList*restrict v1)
{
  tsAssert( (v0->n == v1->n) && "vector lists are not the same size" );

  for(size_t i = 0; i < v0->n; i++)
  {
    v0->vX[i] -= v1->vX[i];
    v0->vY[i] -= v1->vY[i];
    v0->vZ[i] -= v1->vZ[i];

    v0->vX[i] = (v0->vX[i] * v0->vX[i])
      + (v0->vY[i] * v0->vY[i])
      + (v0->vZ[i] * v0->vZ[i]);
  }
}

void veclist_dot_with(VectorList*restrict v0, const VectorList*restrict v1)
{
  tsAssert( (v0->n == v1->n) && "vector lists are not the same size" );

  for(size_t i = 0; i < v0->n; i++)
  {
    v0->vX[i] *= v1->vX[i];
    v0->vY[i] *= v1->vY[i];
    v0->vZ[i] *= v1->vZ[i];

    v0->vX[i] += v0->vY[i];
    v0->vX[i] += v0->vZ[i];

    v0->vY[i] = v0->vZ[i] = 0.0;
  }
}

int veclist_set(VectorList*restrict a, const VectorList*restrict b)
{
  if (a->n < b->n) {
    int err;

    if (err = veclist_init(a, b->n), err < 0) {
      return err;
    }
  }

  {
    double *restrict ax = a->vX;
    double *restrict ay = a->vY;
    double *restrict az = a->vZ;

    const double *restrict bx = b->vX;
    const double *restrict by = b->vY;
    const double *restrict bz = b->vZ;

    memcpy(ax,bx,3*b->n);
    memcpy(ay,by,3*b->n);
    memcpy(az,bz,3*b->n);
  }

  return 0;
}

