#include "tubesheet.h"
#include "ts_common_algo.h"
#include "ts_bonds.h"

#include "ts_yaml.h"

#include <ctype.h>
#include <yaml.h>

#define TSY_DEBUG 0

#define DEFAULT_POOL_CAPACITY 128

#define DEPAREN(...) __VA_ARGS__

#define EVENT_LOOP_ERROR(evar,err0,emsg,...) do {      \
    return TS_NEW_YAML_ERROR(err0,evar,emsg,__VA_ARGS__); \
  } while(0)

#define MAP_EVENT_LOOP(pvar,evar,err0,err1,emsg1,on_each)    \
  TSY_EXPECT_EVENT(pvar, YAML_MAPPING_START_EVENT, err0);   \
  for(;;) { yaml_event_t* evar = NULL; TSY_GET_EVENT(pvar,evar); \
    if (event->type == YAML_MAPPING_END_EVENT) { break; }    \
    if (event->type != YAML_SCALAR_EVENT) {                  \
      EVENT_LOOP_ERROR(evar,err1,DEPAREN emsg1,0);           \
    } on_each ; }

struct autorelease_pool* tsy_autorelease_pool_new(struct autorelease_pool* prev, 
    size_t elt_size, size_t n,
    autorelease_pool_finalize_func finalizer, 
    void* finalizer_closure)
{
  struct autorelease_pool* pool = ts_alloc(NO_ALLOCATOR, sizeof(*pool));
  pool->prev = prev;
  pool->finalizer = finalizer;
  pool->finalizer_closure = finalizer_closure;
  datavector_init(&pool->elements, elt_size, n);

  return pool;
}

void* tsy_autorelease_pool_new_item(struct autorelease_pool* pool)
{
  if (pool == NULL) return NULL;
  return datavector_new_item(&pool->elements);
}

void tsy_autorelease_pool_release(struct autorelease_pool* pool)
{
  if (pool == NULL) return;

  if (pool->finalizer != NULL) { 
    for (size_t i=0; i < pool->elements.size; i++) {
      pool->finalizer(pool->finalizer_closure, datavector_at(&pool->elements, i));
    }
  }

  pool->elements.size = 0;
}

static void autorelease_pool_delete(struct autorelease_pool* pool)
{
  tsy_autorelease_pool_release(pool);
  datavector_free(&pool->elements);
  ts_free(NO_ALLOCATOR, pool);
}

static inline yaml_event_t* new_event(struct ts_yaml_parser* yp)
{
  return (yaml_event_t*)tsy_autorelease_pool_new_item(yp->event_pool);
}

static void print_all_events(struct ts_yaml_parser* yp)
{
  for (size_t i=0; i < yp->event_pool->elements.size; i++) {
    tsy_print_event( yp, DATAVECTOR_AT(yaml_event_t, &yp->event_pool->elements, i) );
  }
}

static void release_event_pool(struct ts_yaml_parser* yp)
{
  struct autorelease_pool* pool = tsy_get_event_pool(yp);
  if (pool == NULL) return;

  if (TSY_DEBUG) {
    print_all_events(yp);
  }

  yp->event_pool = pool->prev;
  autorelease_pool_delete(pool);
}

void release_event(void* closure, void* data)
{
  UNUSED_PARAM(closure);

  yaml_event_delete( (yaml_event_t*) data );
}

static void new_event_pool(struct ts_yaml_parser* yp)
{
  struct autorelease_pool* pool = tsy_autorelease_pool_new(yp->event_pool,
      sizeof(yaml_event_t), DEFAULT_POOL_CAPACITY,
      release_event,
      NULL);

  if (pool != NULL) {
    yp->event_pool = pool;
  }
}

static void new_string_pool(struct ts_yaml_parser* yp)
{
  struct autorelease_pool* pool = tsy_autorelease_pool_new(yp->string_pool,
      sizeof(char*), DEFAULT_POOL_CAPACITY,
      (autorelease_pool_finalize_func)ts_free,
      NO_ALLOCATOR);

  if (pool != NULL) {
    yp->string_pool = pool;
  }
}

void ts_yaml_parser_init(struct ts_yaml_parser* yp, yaml_parser_t* parser)
{
  yp->parser = parser;
  yp->lookahead_event = NULL;

  new_event_pool(yp);
  new_string_pool(yp);

  yp->print_event_indent = 0;
}

void ts_yaml_parse_finalize(struct ts_yaml_parser* yp)
{
  yp->lookahead_event = NULL;
  release_event_pool(yp);
  autorelease_pool_delete(yp->string_pool);
}

const char* ts_yaml_event_type_name(yaml_event_type_t type)
{
  switch (type) 
  {
    case YAML_NO_EVENT: return "<NO EVENT>"; break;

    case YAML_STREAM_START_EVENT:   return "stream-start event";   break;
    case YAML_STREAM_END_EVENT:     return "stream-end event";     break;

    case YAML_DOCUMENT_START_EVENT: return "document-start event"; break;
    case YAML_DOCUMENT_END_EVENT:   return "document-end event";   break;

    case YAML_ALIAS_EVENT:          return "alias event";          break;
                                                                  
    case YAML_SCALAR_EVENT:         return "scalar event";         break;
                                                                  
    case YAML_SEQUENCE_START_EVENT: return "seq-start event";      break;
    case YAML_SEQUENCE_END_EVENT:   return "seq-end event";        break;
                                                                  
    case YAML_MAPPING_START_EVENT:  return "mapping-start event";  break;
    case YAML_MAPPING_END_EVENT:    return "mapping-end event";    break;

    default: return "unknown event"; break;
  }
}

#define DO_INDENT(ilim) \
  do { for (int i=0; i < (ilim); i++) fprintf(stderr, "  "); } while(0)
#define PRINT_EVENT_TYPE(ilim,s) \
  do { DO_INDENT(ilim); fprintf(stderr,"%s\n",(s)); } while(0)
void tsy_print_event(struct ts_yaml_parser* yp, yaml_event_t* evt)
{
  switch (evt->type) 
  {
    case YAML_NO_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent, "<NO EVENT>");
      break;

    case YAML_STREAM_START_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent, "stream-start event");
      yp->print_event_indent++;
      break;

    case YAML_STREAM_END_EVENT:
      yp->print_event_indent--;
      PRINT_EVENT_TYPE(yp->print_event_indent, "stream-end event");
      break;

    case YAML_DOCUMENT_START_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent, "document-start event");
      yp->print_event_indent++;
      break;

    case YAML_DOCUMENT_END_EVENT:
      yp->print_event_indent--;
      PRINT_EVENT_TYPE(yp->print_event_indent,"document-end event");
      break;

    case YAML_ALIAS_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent, "alias event"); break;

    case YAML_SCALAR_EVENT:
      DO_INDENT(yp->print_event_indent);
      fprintf(stderr, "scalar event: %s\n", evt->data.scalar.value); 
      break;

    case YAML_SEQUENCE_START_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent, "seq-start event");
      yp->print_event_indent++; 
      break;
    case YAML_SEQUENCE_END_EVENT:
      yp->print_event_indent--; 
      PRINT_EVENT_TYPE(yp->print_event_indent, "seq-end event");
      break;

    case YAML_MAPPING_START_EVENT:
      PRINT_EVENT_TYPE(yp->print_event_indent,"mapping-start event"); 
      yp->print_event_indent++; 
      break;
    case YAML_MAPPING_END_EVENT:
      yp->print_event_indent--; 
      PRINT_EVENT_TYPE(yp->print_event_indent,"mapping-end event"); 
      break;

    default:
      PRINT_EVENT_TYPE(yp->print_event_indent, "unknown event"); break;
  }
}
#undef PRINT_EVENT_TYPE
#undef DO_INDENT

TSResult tsy_get_event(struct ts_yaml_parser* yp, yaml_event_t** event_ptr)
{
  tsAssert(event_ptr != NULL);
  if (yp->lookahead_event != NULL) {
    *event_ptr = yp->lookahead_event;
    yp->lookahead_event = NULL;
    return TS_OK;
  } else {
    *event_ptr = new_event(yp);
    if (yaml_parser_parse(yp->parser,*event_ptr)) {
      if (0) tsy_print_event(yp,*event_ptr);
      return TS_OK;
    } else {
      return TS_NEW_ERROR2( TS_ERR_PARSE_ERROR,
          ("YAML parser error at line %d, column %d",
           yp->parser->mark.line,
           yp->parser->mark.column));
    }
  }
}

TSResult tsy_unget_event(struct ts_yaml_parser* yp, yaml_event_t* event)
{
  ERROR_IF( TS_ERR_YAML_LOOKAHEAD_EXCEEDED, yp->lookahead_event != NULL);

  yp->lookahead_event = event;
  return TS_OK;
}

enum map_value_type {
  MAP_NULL = 0,
  MAP_STRING = 1,
  MAP_FLOAT = 2,
  MAP_INT = 3,
};

struct map_entry {
  const char* key;
  enum map_value_type type;
  ptrdiff_t offset;
};

static const struct map_entry particle_type_entries[] = {
  { "name"    , MAP_STRING, offsetof(ts_particle_type, long_name) },
  { "mass"    , MAP_FLOAT , offsetof(ts_particle_type, mass    ) },
  { "radius"  , MAP_FLOAT , offsetof(ts_particle_type, radius   ) },
  { "collfreq", MAP_FLOAT , offsetof(ts_particle_type, collfreq ) },
  { "charge"  , MAP_FLOAT , offsetof(ts_particle_type, charge   ) },
  { NULL, MAP_NULL, 0 }
};

union map_types {
  void* generic;
  char** sptr;
  int* iptr;
  double* dptr;
};

static inline union map_types entry_pointer(void* data, const struct map_entry* entry)
{
  union map_types t;
  t.generic = (void*) (((char*)data) + entry->offset);
  return t;
}

static inline int is_id_char(int c) { return isalnum(c) || (c == '_'); }
static TSResult scalar_subst_param(ts_sim* sim, yaml_event_t* event, int* succ, double* vp)
{
  const unsigned char* s = event->data.scalar.value;
  const unsigned char* end = s+event->data.scalar.length;

  *succ = 0;

  while (isspace(*s) && (++s < end));

  if ((s >= end) || (*s != '$')) { return TS_OK; }
  const unsigned char* p = ++s;

  while (is_id_char(*s) && (++s < end));
  if (p == s) { return TS_OK; }

  size_t len = s-p;
  char* stmp = ts_alloc(NO_ALLOCATOR, sizeof(char) * (len+1));
  if (stmp == NULL) { return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY); }

  memcpy(stmp,p,len);
  stmp[s-p] = '\0';

  RETURN_ON_ERROR2(
      ts_get_parameter(sim, stmp, vp),
      ts_safe_free(NO_ALLOCATOR, stmp) );

  *succ = 1;

  return TS_OK;
}

static TSResult scalar_to_double(ts_sim* sim, yaml_event_t* event, double* dptr)
{
  int subst = 0;
  RETURN_ON_ERROR( scalar_subst_param(sim,event,&subst,dptr) );
  if (!subst)
  {
    RETURN_ON_ERROR(
      string_to_double((char*)event->data.scalar.value, dptr)
      );
  }

  return TS_OK;
}

static TSResult scalar_to_size(ts_sim* sim, yaml_event_t* event, size_t* szp)
{
  UNUSED_PARAM(sim);

  return string_to_size((char*)event->data.scalar.value, szp);
}

static TSResult scalar_to_bool(yaml_event_t* event, int* bptr)
{
  if (TSY_SCALAR_EQ(event, "true"))  { *bptr = 1; return TS_OK; }
  if (TSY_SCALAR_EQ(event, "false")) { *bptr = 0; return TS_OK; }

  return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_VALUE, event,
      "expected boolean value but found \"%s\"", TSY_SCALAR_STRING(event));
}

static TSResult parse_entry_value(ts_sim* sim, struct ts_yaml_parser* yp, yaml_event_t* event, void* data, const struct map_entry* entry)
{
  UNUSED_PARAM(yp);

  union map_types entry_ptr = entry_pointer(data, entry);

  switch (entry->type) {
    case MAP_NULL:
      return TS_OK;

    case MAP_STRING:
      entry_ptr.sptr[0] = ts_strdup(NO_ALLOCATOR, (char*)event->data.scalar.value);
      return TS_OK;
    case MAP_FLOAT:
      return scalar_to_double(sim, event, entry_ptr.dptr);
    case MAP_INT:
      return TS_OK;
  }

  return TS_OK;
}

static TSResult parse_map_entry(ts_sim* sim, struct ts_yaml_parser* yp, void* data, const struct map_entry* entries)
{
  yaml_event_t* key_event = NULL;
  yaml_event_t* val_event = NULL;

  TSY_GET_EVENT(yp,key_event);
  if (key_event->type == YAML_MAPPING_END_EVENT) return TS_NEW_RESULT( TS_STOP_ITERATION );
  if (key_event->type != YAML_SCALAR_EVENT)      return TS_NEW_ERROR( TS_ERR_YAML_INVALID_PTYPE );

  TSY_GET_EVENT(yp,val_event);
  if (val_event->type != YAML_SCALAR_EVENT)      return TS_NEW_ERROR( TS_ERR_YAML_INVALID_PTYPE );

  for (size_t i=0; entries[i].key != NULL; i++) {
    if (TSY_SCALAR_EQ(key_event, entries[i].key)) {
      return parse_entry_value(sim, yp,val_event, data, entries+i);
    }
  }

  return TS_NEW_ERROR( TS_ERR_YAML_INVALID_KEY );
}

#define PARSE_VEC_COMPONENT(component)                                       \
  do { yaml_event_t* evt = NULL; TSY_GET_EVENT(yp,evt);                          \
  if (evt->type != YAML_SCALAR_EVENT) {                                      \
    return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_VECTOR3, evt,                  \
        "expected scalar for vector " SYMSTR(component) " component", 0); }  \
  if (!IS_OKAY(scalar_to_double(sim, evt, &( SYMCAT(VECTOR3_,component)(vec) )))) \
    return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_VECTOR3, evt,                  \
        "expected double for vector " SYMSTR(component)                      \
          " component (found \"%s\")", TSY_SCALAR_STRING(evt));                  \
  } while(0)

static TSResult
parse_vector(ts_sim* sim, struct ts_yaml_parser* yp, vector3* vec)
{
  TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_START_EVENT, TS_ERR_YAML_INVALID_VECTOR3);

  PARSE_VEC_COMPONENT(X);
  PARSE_VEC_COMPONENT(Y);
  PARSE_VEC_COMPONENT(Z);

  TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_END_EVENT, TS_ERR_YAML_INVALID_VECTOR3);

  return TS_OK;
}
#undef PARSE_VEC_COMPONENT

static TSResult parse_particle_type(ts_sim* sim, struct ts_yaml_parser* yp, const char* name)
{
  yaml_event_t* event = NULL;
  TSY_GET_EVENT(yp,event);

  if (event->type != YAML_MAPPING_START_EVENT) return TS_NEW_ERROR( TS_ERR_YAML_INVALID_PTYPE );

  ts_particle_type ptype;
  memset(&ptype,0,sizeof(ptype));
  ptype.sim = sim;
  ptype.name = name;
  ptype.long_name = name;

  for(;;) {
    TSResult ret = parse_map_entry(sim, yp, &ptype, particle_type_entries);

    if (IS_STOP_ITERATION(ret))
    {
      size_t ptInd = 0;
      return ts_add_particle_type(sim, &ptype, &ptInd);
    }

    if (IS_ERROR(ret)) { return ret; }
  }

  return TS_OK;
}

static TSResult parse_particle_type_list(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_PTYPE_LIST);

  for(;;) {
    yaml_event_t* event = NULL;

    TSY_GET_EVENT(yp,event);
    if (event->type == YAML_MAPPING_END_EVENT) return TS_OK;

    if (event->type != YAML_SCALAR_EVENT) return TS_NEW_ERROR(TS_ERR_YAML_INVALID_PTYPE);
    const char* name = ts_strdup(NO_ALLOCATOR, (char*)event->data.scalar.value);

    TSY_GET_EVENT(yp, event);
    if (event->type != YAML_MAPPING_START_EVENT)
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE, event,
          "expected map (MAPPING-START) for particle type but found %s",
          ts_yaml_event_name(event));
    TSY_UNGET_EVENT(yp,event);

    RETURN_ON_ERROR( parse_particle_type(sim, yp, name) );
  }

  return TS_OK;
}

static TSResult
parse_rbtype_type(const yaml_event_t* evt, struct ts_rigid_body_type* rbtype)
{
  if (evt->type != YAML_SCALAR_EVENT)
    return TS_NEW_ERROR3( TS_ERR_YAML_INVALID_RBTYPE,
        "rigid body \"type\": expected a scalar" );

  if (TSY_SCALAR_EQ(evt, "Site interaction"))
  {
    rbtype->type = TS_RIGID_SITE_INTERACTION;
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR3( TS_ERR_YAML_INVALID_RBTYPE,
        "rigid body \"type\": unknown type" );
  }
}

#define ERRCOND(...) \
  ERROR_IF_MESG(TS_ERR_YAML_INVALID_RBTYPE, __VA_ARGS__)
static TSResult check_rbtype(struct ts_rigid_body_type* rbtype)
{
  ERRCOND(   rbtype->name == NULL, "rigid body requires a name" );
  ERRCOND(     rbtype->I0 <=  0.0, "rigid body I0 must be positive");
  ERRCOND(rbtype->rotfric <=  0.0, "rigid body rotfric must be positive");

  ERRCOND(
      (rbtype->type <= TS_RIGID_NOT_DEFINED) ||
      (rbtype->type >= TS_RIGID_MAX_TYPE),
      "invalid rigid body type (%d)", (int)rbtype->type );

  ERRCOND(rbtype->num_sites == 0, "rigid body requires site vectors");
  ERRCOND(rbtype->ref_vectors == NULL, "rigid body site vectors are NULL");

  return TS_OK;
}
#undef ERRCOND

#define MAX_SITES 64
static TSResult parse_rbtype_sites(ts_sim* sim, struct ts_yaml_parser* yp, struct ts_rigid_body_type* rbtype)
{
  vector3 vecs[MAX_SITES];
  size_t num_sites = 0;

  for(;;)
  {
    yaml_event_t* evt = NULL;

    TSY_GET_EVENT(yp, evt);
    if (evt->type == YAML_SEQUENCE_END_EVENT) { goto set_vectors; }

    if (evt->type != YAML_SEQUENCE_START_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_RBTYPE, evt,
          "expected vector (sequence-start) in rigid body site vector list (found event %s)",
          ts_yaml_event_name(evt));
    }

    TSY_UNGET_EVENT(yp,evt);
    RETURN_ON_ERROR( parse_vector(sim, yp, vecs+num_sites) );
    num_sites++;
  }

set_vectors:
  rbtype->num_sites = num_sites;
  rbtype->ref_vectors = ts_calloc(NO_ALLOCATOR, num_sites, sizeof(rbtype->ref_vectors[0]));
  memcpy(rbtype->ref_vectors, vecs, num_sites*sizeof(rbtype->ref_vectors[0]));

  return TS_OK;
}

static TSResult parse_rigid_body_type(ts_sim* sim, struct ts_yaml_parser* yp, const char* name)
{
  TSY_EXPECT_EVENT(yp,YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_PTYPE);

  struct ts_rigid_body_type rbtype;
  memset(&rbtype,0,sizeof(rbtype));
  rbtype.name = name;

  for(;;) {
    yaml_event_t* key_event = NULL;
    yaml_event_t* val_event = NULL;

    TSY_GET_EVENT(yp,key_event);
    if (key_event->type == YAML_MAPPING_END_EVENT) { goto check_and_add; }
    if (key_event->type != YAML_SCALAR_EVENT)
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_RBTYPE, key_event,
          "expected scalar while parsing rigid body type, found %s",
          ts_yaml_event_name(key_event));

    TSY_GET_EVENT(yp,val_event);
    if (TSY_SCALAR_EQ(key_event, "sites"))
    {
      if (val_event->type != YAML_SEQUENCE_START_EVENT)
        return TS_NEW_ERROR3( TS_ERR_YAML_INVALID_RBTYPE,
            "rigid body \"sites\": expected list of site locations" );

      RETURN_ON_ERROR( parse_rbtype_sites(sim, yp, &rbtype) );
    }
    else if (TSY_SCALAR_EQ(key_event, "type"))
    {
      RETURN_ON_ERROR( parse_rbtype_type(val_event, &rbtype) );
    }
    else if (TSY_SCALAR_EQ(key_event, "I0"))
    {
      if (val_event->type != YAML_SCALAR_EVENT)
        return TS_NEW_ERROR3( TS_ERR_YAML_INVALID_RBTYPE,
            "rigid body \"I0\": expected a scalar" );

      RETURN_ON_ERROR( scalar_to_double(sim, val_event, &rbtype.I0) );
    }
    else if (TSY_SCALAR_EQ(key_event, "rotfric"))
    {
      if (val_event->type != YAML_SCALAR_EVENT)
        return TS_NEW_ERROR3( TS_ERR_YAML_INVALID_RBTYPE,
            "rigid body \"rotfric\": expected a scalar" );

      RETURN_ON_ERROR( scalar_to_double(sim, val_event, &rbtype.rotfric) );
    }
    else
    {
      return TS_NEW_YAML_ERROR( TS_ERR_YAML_INVALID_KEY,
          key_event, "unknown rigid body type key \"%s\"", TSY_SCALAR_STRING(key_event) );
    }
  }

check_and_add:
  RETURN_ON_ERROR( check_rbtype(&rbtype) );
  ts_rigid_body_type_update_parameters(sim, &rbtype);

  size_t rbTypeInd = 0;
  RETURN_ON_ERROR( ts_add_rigid_body_type(sim, &rbtype, &rbTypeInd) );

  return TS_OK;
}

static TSResult parse_rigid_body_type_list(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_RBTYPE_LIST);

  for(;;) {
    yaml_event_t* event = NULL;

    TSY_GET_EVENT(yp,event);
    if (event->type == YAML_MAPPING_END_EVENT) return TS_OK;

    if (event->type != YAML_SCALAR_EVENT)
    {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_RBTYPE, event,
          "bad YAML while parsing rigid body list (expected SCALAR-EVENT, encountered %s)",
          ts_yaml_event_name(event));
    }
    const char* name = ts_strdup(NO_ALLOCATOR, (char*)event->data.scalar.value);
    RETURN_ON_ERROR( parse_rigid_body_type(sim, yp, name) );
  }

  return TS_OK;
}

static inline TSResult add_another_bondtype(ts_sim* sim, const char* name, const char* direcName)
{
  enum BONDTYPE_DIRECTION direc;

  if (direcName == NULL) {
    direc = BONDTYPE_UNDIRECTED;
  } else {
    if (STRINGS_EQ(direcName, "none")) {
      direc = BONDTYPE_UNDIRECTED;
    }
    else if (STRINGS_EQ(direcName, "forw")) {
      direc = BONDTYPE_FORWARD;
    }
    else if (STRINGS_EQ(direcName, "back")) {
      direc = BONDTYPE_BACKWARD;
    }
    else if (STRINGS_EQ(direcName, "bidirec")) {
      direc = BONDTYPE_BIDIRECTIONAL;
    }
    else {
      return TS_NEW_ERROR(TS_ERR_YAML_INVALID_BONDTYPE);
    }
  }

  int num_bond_types = ts_count_bond_types(sim);
  return ts_add_bond_type(sim, name, bondtype_new(direc, num_bond_types+1));
}

static TSResult parse_bond_type_list(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_START_EVENT, TS_ERR_YAML_INVALID_BONDTYPE_LIST);

  for(;;) {
    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_SEQUENCE_END_EVENT) 
    {
      return TS_OK;
    }
    else if (event->type == YAML_SCALAR_EVENT)
    {
      RETURN_ON_ERROR( add_another_bondtype(sim, (char*)event->data.scalar.value, NULL) );
    }
    else if (event->type == YAML_SEQUENCE_START_EVENT) 
    {
      yaml_event_t* name_event = NULL;
      yaml_event_t* direc_event = NULL;

      TSY_GET_EVENT(yp,name_event);
      if (name_event->type != YAML_SCALAR_EVENT)
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_BONDTYPE,name_event,
            "expected bondtype name", 0);

      TSY_GET_EVENT(yp,direc_event);
      if (direc_event->type != YAML_SCALAR_EVENT)
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_BONDTYPE, direc_event,
            "expected bondtype direction", 0);

      TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_END_EVENT, TS_ERR_YAML_INVALID_BONDTYPE);

      RETURN_ON_ERROR( add_another_bondtype(sim, 
          (char*)name_event->data.scalar.value, 
          (char*)direc_event->data.scalar.value) );
    }
    else 
    {
      return TS_NEW_ERROR(TS_ERR_YAML_INVALID_BONDTYPE);
    }
  }

  return TS_OK;
}

static TSResult parse_dynamics_integrator_type(ts_sim* sim, yaml_event_t* val)
{
  if (TSY_SCALAR_EQ(val, "SRK"))
  {
    sim->integrator = TS_INTEGRATOR_STOCH_RUNGE_KUTTA;
    return TS_OK;
  }

  return TS_NEW_ERROR(TS_ERR_YAML_INVALID_DYNAMICS_VALUE);
}

static TSResult parse_dynamics_kvpair(ts_sim* sim, struct ts_yaml_parser* yp)
{
  yaml_event_t* key = NULL;
  yaml_event_t* val = NULL;

  TSY_GET_SCALAR(yp,key, TS_ERR_YAML_INVALID_DYNAMICS_KEY);
  TSY_GET_SCALAR(yp,val, TS_ERR_YAML_INVALID_DYNAMICS_VALUE);

  if (TSY_SCALAR_EQ(key, "integrator"))
  {
    RETURN_ON_ERROR( parse_dynamics_integrator_type(sim,val) );
  }
  else if (TSY_SCALAR_EQ(key, "temp"))
  {
    RETURN_ON_ERROR( scalar_to_double(sim, val, &sim->temp) );
  }
  else if (TSY_SCALAR_EQ(key, "timestep"))
  {
    RETURN_ON_ERROR( scalar_to_double(sim, val, &sim->dt) );
  }
  else if (TSY_SCALAR_EQ(key, "nonbond_cutoff"))
  {
    RETURN_ON_ERROR( scalar_to_double(sim, val, &sim->nonbond_cutoff) );
  }
  else if (TSY_SCALAR_EQ(key, "nonbond_padding"))
  {
    RETURN_ON_ERROR( scalar_to_double(sim, val, &sim->nonbond_padding) );
  }
  else if (TSY_SCALAR_EQ(key, "max_particles"))
  {
    size_t max_particles = 0;
    RETURN_ON_ERROR( scalar_to_size(sim, val, &max_particles) );
    RETURN_ON_ERROR( ts_alloc_particles(sim, max_particles) );
  }
  else if (TSY_SCALAR_EQ(key, "max_rigid_bodies"))
  {
    size_t max_rigid_bodies = 0;
    RETURN_ON_ERROR( scalar_to_size(sim, val, &max_rigid_bodies) );
    RETURN_ON_ERROR( ts_alloc_rigid_bodies(sim, max_rigid_bodies) );
  }
  else
  {
    return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY,key,
        "invalid key \"%s\"", TSY_SCALAR_STRING(key));
  }

  return TS_OK;
}

static TSResult parse_dynamics(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_DYNAMICS_SECTION);

  for(;;) {
    yaml_event_t* event = NULL;

    TSY_GET_EVENT(yp,event);
    if (event->type == YAML_MAPPING_END_EVENT)
    {
      return TS_OK;
    }
    else
    {
      TSY_UNGET_EVENT(yp,event);
      RETURN_ON_ERROR( parse_dynamics_kvpair(sim, yp) );
    }
  }
}

static TSResult parse_parameter_kvpair(ts_sim* sim, struct ts_yaml_parser* yp)
{
  yaml_event_t* key = NULL;
  yaml_event_t* val = NULL;

  TSY_GET_SCALAR(yp,key, TS_ERR_YAML_INVALID_DYNAMICS_KEY);
  TSY_GET_SCALAR(yp,val, TS_ERR_YAML_INVALID_DYNAMICS_VALUE);

  double d_val = 0.0;
  RETURN_ON_ERROR( scalar_to_double(sim, val,&d_val) );
  RETURN_ON_ERROR( ts_set_parameter(sim, TSY_SCALAR_STRING(key), d_val) );

  return TS_OK;
}

static TSResult parse_parameters(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_PARAMETERS_SECTION);

  for(;;) {
    yaml_event_t* event = NULL;

    TSY_GET_EVENT(yp,event);
    if (event->type == YAML_MAPPING_END_EVENT) 
    {
      return TS_OK;
    }
    else
    {
      TSY_UNGET_EVENT(yp,event);
      RETURN_ON_ERROR( parse_parameter_kvpair(sim, yp) );
    }
  }
}

static TSResult get_nbody_term(yaml_event_t* evt, enum TS_BOND_TERM* btp)
{
  if (TSY_SCALAR_EQ(evt,"zero_body" )) { *btp = TS_BOND_TERM_ZEROBODY;   return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"one_body"  )) { *btp = TS_BOND_TERM_ONEBODY;    return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"two_body"  )) { *btp = TS_BOND_TERM_TWOBODY;    return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"three_body")) { *btp = TS_BOND_TERM_THREEBODY;  return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"four_body" )) { *btp = TS_BOND_TERM_FOURBODY;   return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"improper"  )) { *btp = TS_BOND_TERM_IMPROPER;   return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"nonbond"   )) { *btp = TS_BOND_TERM_NONBOND;    return TS_OK; };
  if (TSY_SCALAR_EQ(evt,"constraint")) { *btp = TS_BOND_TERM_CONSTRAINT; return TS_OK; };

  return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, evt,
      "expected a bonded term type", 0);
}

static TSResult parse_bonds_enabled(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_SECTION);

  for (;;)
  {
    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);
    if (event->type == YAML_MAPPING_END_EVENT) { return TS_OK; }
    if (event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
          "expected scalar key", 0);
    }

    enum TS_BOND_TERM bt = TS_BOND_TERM_MAX;
    RETURN_ON_ERROR( get_nbody_term(event, &bt) );

    int enabled = 0;
    TSY_GET_EVENT(yp,event);
    RETURN_ON_ERROR( scalar_to_bool(event, &enabled) );

    sim->bonds_enabled[bt] = enabled;
    if (bt == TS_BOND_TERM_NONBOND) { sim->nonbonds_enabled = enabled; }
  }
}

static TSResult parse_potential_definition(ts_sim* sim, struct ts_yaml_parser* yp, Potential** potp);

static TSResult parse_list_of_particle_types(ts_sim* sim, struct ts_yaml_parser* yp, ts_bond_pattern* pat)
{
  TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_START_EVENT, TS_ERR_YAML_INVALID_BONDED_PATTERN);

  int ind = 0;
  for(;;)
  {
    yaml_event_t* event;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_SEQUENCE_END_EVENT)
    {
      break;
    }
    else if (event->type == YAML_SCALAR_EVENT)
    {
      if (ind >= NBODY_LIMIT) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE_LIST, event,
            "too many particle indices", 0);
      }
      size_t pind = ts_find_particle_type(sim, TSY_SCALAR_STRING(event));
      if (pind == 0)
      {
        TSResult ret = string_to_size(TSY_SCALAR_STRING(event), &pind);
        if (IS_ERROR(ret)) { pind = 0; }
      }

      if (pind > 0) {
        pat->ptypes[ind] = pind;
        ind++;
      }
      else {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE, event,
            "invalid particle type: \"%s\"", TSY_SCALAR_STRING(event));
      }
    }
    else
    {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE, event,
          "invalid bonded term type \"%s\"", TSY_SCALAR_STRING(event));
    }
  }

  pat->nbodies = ind;

  return TS_OK;
}

static TSResult parse_list_of_bond_types(ts_sim* sim, struct ts_yaml_parser* yp, ts_bond_pattern* pat)
{
  TSY_EXPECT_EVENT(yp, YAML_SEQUENCE_START_EVENT, TS_ERR_YAML_INVALID_BONDED_PATTERN);

  int ind = 0;
  for(;;)
  {
    yaml_event_t* event;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_SEQUENCE_END_EVENT)
    {
      break;
    }
    else if (event->type == YAML_SCALAR_EVENT)
    {
      if (ind >= NBOND_LIMIT) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE_LIST, event,
            "too many bond types", 0);
      }

      ts_bond_type bt = ts_get_bond_type(sim, TSY_SCALAR_STRING(event));
      if (bondtype_is_none(bt)) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_BONDTYPE, event,
            "invalid bond type: \"%s\"", TSY_SCALAR_STRING(event));
      }

      pat->btypes[ind] = bt;
      ind++;
    }
    else
    {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_PTYPE, event,
          "invalid bonded term type \"%s\"", TSY_SCALAR_STRING(event));
    }
  }

  return TS_OK;
}

static TSResult parse_bond_pattern_def(ts_sim* sim,
    struct ts_yaml_parser* yp,
    ts_bond_pattern* pat,
    const yaml_event_t* key_event)
{
  yaml_event_t* val_event = NULL;

  if (TSY_SCALAR_EQ(key_event, "potential"))
  {
    TSY_GET_EVENT(yp, val_event);

    Potential* pot = NULL;
    if (val_event->type == YAML_MAPPING_START_EVENT)
    {
      TSY_UNGET_EVENT(yp,val_event);
      RETURN_ON_ERROR( parse_potential_definition(sim, yp, &pot) );
    }
    else if (val_event->type == YAML_SCALAR_EVENT)
    {
      const char* pot_name = TSY_SCALAR_STRING(val_event);
      pot = ts_get_potential(sim, pot_name);
      if (pot == NULL) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_VALUE, val_event,
            "unknown potential name: %s", pot_name);
      }
    }
    else
    {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, val_event,
          "unknown potential value: %s", ts_yaml_event_name(val_event));
    }

    if (pot != NULL) { pat->potential = pot; }
  }
  else if (TSY_SCALAR_EQ(key_event, "type"))
  {
    TSY_GET_EVENT(yp, val_event);
    if (TSY_SCALAR_EQ(val_event, "bonded"))
    {
      pat->type = TS_BOND_TERM_ZEROBODY;
    }
    else if (TSY_SCALAR_EQ(val_event, "improper"))
    {
      pat->type = TS_BOND_TERM_IMPROPER;
    }
    else if (TSY_SCALAR_EQ(val_event, "nonbonded") || TSY_SCALAR_EQ(val_event, "non-bonded"))
    {
      pat->type = TS_BOND_TERM_NONBOND;
    }
    else
    {
      return TS_NEW_YAML_ERROR(TS_ERR_UNSUPPORTED_BOND_PATTERN, val_event,
          "invalid bonded term type \"%s\"", TSY_SCALAR_STRING(val_event));
    }
  }
  else if (TSY_SCALAR_EQ(key_event, "particle_types"))
  {
    RETURN_ON_ERROR( parse_list_of_particle_types(sim, yp, pat) );
    if (pat->type == TS_BOND_TERM_ZEROBODY)
    {
      switch (pat->nbodies) {
        case 0: pat->type = TS_BOND_TERM_ZEROBODY;  break;
        case 1: pat->type = TS_BOND_TERM_ONEBODY;   break;
        case 2: pat->type = TS_BOND_TERM_TWOBODY;   break;
        case 3: pat->type = TS_BOND_TERM_THREEBODY; break;
        case 4: pat->type = TS_BOND_TERM_FOURBODY;  break;
        default:
          return TS_NEW_YAML_ERROR(TS_ERR_UNSUPPORTED_BOND_PATTERN, val_event,
              "invalid number of particle types for bonded term (%d)", pat->nbodies);
      }
    }
  }
  else if (TSY_SCALAR_EQ(key_event, "bond_types"))
  {
    RETURN_ON_ERROR( parse_list_of_bond_types(sim, yp, pat) );
  }
  else
  {
    return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, key_event, 
        "unknown bond pattern key: %s", TSY_SCALAR_STRING(key_event));
  }

  return TS_OK;
}


static TSResult parse_bond_pattern(ts_sim* sim, struct ts_yaml_parser* yp, const char* name)
{
  ts_bond_pattern pattern;
  memset(&pattern, 0, sizeof(pattern));
  pattern.name = ts_strdup(NO_ALLOCATOR, name);
  pattern.potential = NULL;
  pattern.type = TS_BOND_TERM_ZEROBODY;

  MAP_EVENT_LOOP(yp, event,
      TS_ERR_YAML_INVALID_SECTION,
      TS_ERR_YAML_INVALID_KEY,
      (
        "expected string key (bond pattern definition), found %s",
        ts_yaml_event_name(event)
      ),
      RETURN_ON_ERROR( parse_bond_pattern_def(sim, yp, &pattern, event) )
    );

  RETURN_ON_ERROR( ts_direct_add_bond_pattern(sim, &pattern, NULL) );

  return TS_OK;
}

static TSResult parse_bond_patterns(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_SECTION);

  for (;;)
  {
    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_MAPPING_END_EVENT) { break; }

    if (event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
          "expected string key (bond pattern name), found %s",
          ts_yaml_event_name(event));
    }

    RETURN_ON_ERROR( parse_bond_pattern(sim, yp, TSY_SCALAR_STRING(event)) );
  }

  return TS_OK;
}

static TSResult new_potential(yaml_event_t* event, Potential** potp)
{
  *potp = ts_potential_new(TSY_SCALAR_STRING(event));
  return (*potp != NULL)
    ? TS_OK
    : TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_VALUE, event,
        "potential type \"%s\" is unknown", TSY_SCALAR_STRING(event))
    ;

  return TS_OK;
}

static TSResult set_potential_params(ts_sim* sim, Potential* pot, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_SECTION);
  for(;;)
  {
    yaml_event_t* key_event = NULL;
    yaml_event_t* val_event = NULL;
    TSY_GET_EVENT(yp,key_event);

    if (key_event->type == YAML_MAPPING_END_EVENT) { return TS_OK; }

    if (key_event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, key_event,
          "expected scalar key while parsing potential parameters", 0);
    }

    TSY_GET_EVENT(yp,val_event);
    if (val_event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, val_event,
          "expected scalar value while parsing potential parameters", 0);
    }

    double val = 0.0;
    TSResult ret = scalar_to_double(sim, val_event, &val);
    if (IS_ERROR(ret)) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, val_event,
          "expected scalar double but found \"%s\" while parsing potential parameters", 
          TSY_SCALAR_STRING(val_event));
    }

    RETURN_ON_ERROR( ts_potential_set_param_value(pot, TSY_SCALAR_STRING(key_event), val) );
  }
}

static TSResult parse_potential_definition(ts_sim* sim, struct ts_yaml_parser* yp, Potential** potp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_SECTION);

  if (potp == NULL) {
    return TS_NEW_ERROR3(TS_ERR_NULL_POINTER_ARG, "potp variable is NULL");
  }

  for(;;)
  {
    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_MAPPING_END_EVENT) {
      return TS_OK;
    }

    if (event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
          "expected scalar key while parsing potential", 0);
    }

    if (TSY_SCALAR_EQ(event, "type")) {
      if (*potp != NULL) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
            "potential \"type\" encountered twice", 0);

      }

      TSY_GET_EVENT(yp,event);
      if (event->type != YAML_SCALAR_EVENT) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
            "potential \"type\" expected to have a scalar value", 0);
      }

      RETURN_ON_ERROR( new_potential(event, potp) );
    }
    else if (TSY_SCALAR_EQ(event, "params")) {
      if (*potp == NULL) {
        return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
            "potential \"params\" encountered before \"type\"", 0);
      }

      RETURN_ON_ERROR( set_potential_params(sim, *potp, yp) );
    }
  }
}

static TSResult parse_potential(ts_sim* sim, struct ts_yaml_parser* yp, const char* name)
{
  Potential* pot = NULL;

  RETURN_ON_ERROR( parse_potential_definition(sim, yp, &pot) );

  if (pot != NULL)
  {
    RETURN_ON_ERROR( ts_add_potential(sim, name, pot) );
  }

  return TS_OK;
}

static TSResult parse_potential_list(ts_sim* sim, struct ts_yaml_parser* yp)
{
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_SECTION);

  for (;;)
  {
    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_MAPPING_END_EVENT) { return TS_OK; }

    if (event->type != YAML_SCALAR_EVENT) {
      return TS_NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY, event,
          "expected scalar key", 0);
    }

    RETURN_ON_ERROR( parse_potential(sim, yp, TSY_SCALAR_STRING(event)) );
  }
  return TS_OK;
}

#define DISPATCH_SECTION(str, func)   \
  do { if (TSY_SCALAR_EQ(event, str)) {   \
    RESULT_PROPAGATE( func(sim,yp) ); \
  } } while(0)
static TSResult dispatch_top_map(ts_sim* sim, struct ts_yaml_parser* yp)
{
  yaml_event_t* event = NULL;

  TSY_GET_SCALAR(yp,event,TS_ERR_YAML_INVALID_KEY);

  DISPATCH_SECTION( "particle_types"  , parse_particle_type_list   );
  DISPATCH_SECTION( "bond_types"      , parse_bond_type_list       );
  DISPATCH_SECTION( "simulation"      , parse_dynamics             );
  DISPATCH_SECTION( "parameters"      , parse_parameters           );
  DISPATCH_SECTION( "rigid_body_types", parse_rigid_body_type_list );
  DISPATCH_SECTION( "bonds_enabled"   , parse_bonds_enabled        );
  DISPATCH_SECTION( "bond_patterns"   , parse_bond_patterns        );
  DISPATCH_SECTION( "potentials"      , parse_potential_list       );

  return TS_NEW_ERROR2(
      TS_ERR_YAML_INVALID_KEY,
      ("\"%s\"", TSY_SCALAR_STRING(event)) );
}
#undef DISPATCH_SECTION

static TSResult parse_config_file(ts_sim* sim, struct ts_yaml_parser* yp)
{
  yp->print_event_indent = 0;

  TSY_EXPECT_EVENT(yp, YAML_STREAM_START_EVENT, TS_ERR_YAML_INVALID_CONFIG_DOC);
  TSY_EXPECT_EVENT(yp, YAML_DOCUMENT_START_EVENT, TS_ERR_YAML_INVALID_CONFIG_DOC);
  TSY_EXPECT_EVENT(yp, YAML_MAPPING_START_EVENT, TS_ERR_YAML_INVALID_CONFIG_DOC);

  /* Read the event sequence. */
  for(;;) {

    yaml_event_t* event = NULL;
    TSY_GET_EVENT(yp,event);

    if (event->type == YAML_MAPPING_END_EVENT) { break; }
    TSY_UNGET_EVENT(yp,event);

    RETURN_ON_ERROR(dispatch_top_map(sim, yp));
  }

  TSY_EXPECT_EVENT(yp, YAML_DOCUMENT_END_EVENT, TS_ERR_YAML_INVALID_CONFIG_DOC);
  TSY_EXPECT_EVENT(yp, YAML_STREAM_END_EVENT, TS_ERR_YAML_INVALID_CONFIG_DOC);

  return TS_OK;
}


#define STRNULL(s,dft) ( (s != NULL) ? ((char*)s) : dft )
TSResult ts_yaml_load_string(ts_sim* sim, const char* s, size_t slen)
{
  yaml_parser_t parser;
  struct ts_yaml_parser yp;
  memset(&yp,0,sizeof(yp));

  /* Create the Parser object. */
  yaml_parser_initialize(&parser);
  yaml_parser_set_input_string(&parser, (const unsigned char*)s, slen);

  yp.parser = &parser;

  ts_yaml_parser_init(&yp,&parser);
  TSResult err = parse_config_file(sim, &yp);
  ts_yaml_parse_finalize(&yp);

  /* Destroy the Parser object. */
  yaml_parser_delete(&parser);
  return err;
}
#undef STRNULL


