#include "tsch_config.h"

struct tsch_config_pool* tsch_config_pool_alloc(struct tsch_config_pool* prev, size_t n)
{
  struct tsch_config_pool* pool = ts_alloc(NO_ALLOCATOR, sizeof(*pool) + n*sizeof(pool->entries[0]));
  if (pool == NULL) {
    /* cannot allocate the pool */
    return NULL;
  }

  pool->prev_pool = prev;
  pool->n = 0;
  pool->capa = n;
  memset(pool->entries+0,0,n*sizeof(pool->entries[0]));

  return pool;
}

void tsch_config_pool_dealloc(struct tsch_config_pool* top)
{
  struct tsch_config_pool* p = top;
  while(p) {
    struct tsch_config_pool* pp = p->prev_pool;
    ts_free(NO_ALLOCATOR, p);
    p = pp;
  }
}

struct tsch_config* tsch_config_pool_entry(struct tsch_config_pool** poolp)
{
  if ((*poolp)->n >= (*poolp)->capa) {
    /* try to allocate a new pool */
    struct tsch_config_pool* new_pool = tsch_config_pool_alloc(*poolp, TSCH_CFG_POOL_SIZE);
    if (new_pool == NULL) { return NULL; }

    *poolp = new_pool;
  }

  struct tsch_config* entry = (*poolp)->entries + (*poolp)->n;
  memset(entry,0,sizeof(*entry));
  (*poolp)->n++;
  return entry;
}

/* DJB2 string hash, from Dan J. Bernstein */
static unsigned long djb2_hash(const char *str)
{
  unsigned long hash = 5381;
  int c;

  /* NB: variant uses XOR instead of ADD, see:
   * http://www.cse.yorku.ca/~oz/hash.html
   */
  while (c = (unsigned char)*str++, c) {
    hash = ((hash << 5) + hash) ^ c; /* hash * 33 ^ c */
  }
  return hash;
}

/* wrapper around djb2_hash to avoid potential collisions (in case some
 * other code includes djb2_hash as well.
 */
unsigned long tsch_string_hash(const char *s)
{
  return djb2_hash(s);
}

static inline size_t calc_rehash_limit(size_t capa, int max_load)
{
  if (capa > SIZE_MAX/1024) {
    return (capa/1024)*((size_t)max_load);
  }

  return capa * (size_t)max_load / 1024;
}

TSResult tsch_string_table_initialize(struct tsch_string_table *tbl, size_t capa, bool uses_values)
{
  tbl->keys   = ts_calloc(NO_ALLOCATOR, capa, sizeof tbl->keys[0]);
  tbl->hashes = ts_calloc(NO_ALLOCATOR, capa, sizeof tbl->hashes[0]);

  if (uses_values) {
    tbl->values = ts_calloc(NO_ALLOCATOR, capa, sizeof tbl->values[0]);
  } else {
    tbl->values = NULL;
  }

  if (tbl->keys == NULL   ||
      tbl->hashes == NULL ||
      (uses_values && tbl->values == NULL)) {
    ts_free(NO_ALLOCATOR, tbl->keys);
    ts_free(NO_ALLOCATOR, tbl->hashes);
    ts_free(NO_ALLOCATOR, tbl->values);
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "could not allocate strings table with %zu entries", capa);
  }

  tbl->capa = capa;
  tbl->size = 0;
  tbl->max_load = TSCH_STRINGS_DEFAULT_MAXLOAD;

  tbl->rehash_limit = calc_rehash_limit(capa, tbl->max_load);
  return TS_OK;
}

void tsch_string_table_finalize(struct tsch_string_table *tbl)
{
  ts_free(NO_ALLOCATOR, tbl->keys);
  ts_free(NO_ALLOCATOR, tbl->hashes);
  ts_free(NO_ALLOCATOR, tbl->values);
  tbl->keys   = NULL;
  tbl->hashes = NULL;
  tbl->size = 0;
  tbl->capa = 0;
  tbl->rehash_limit = 0;
}

static inline char *strings_dup(const char *s)
{
  size_t len = strlen(s);
  char *ns = ts_alloc(NO_ALLOCATOR, len+1);
  if (!ns) { return NULL; }

  return strcpy(ns,s);
}

static size_t add_string(struct tsch_string_table *tbl, const char *s, unsigned long hash)
{
  size_t bucket;

  if (tbl->size == tbl->capa) {
    return 0;
  }

  bucket = hash % tbl->capa;
  for (;;) {
    /* Currently use simple linear probing */
    if (!tbl->keys[bucket]) {
      const char *ns = strings_dup(s);
      tbl->keys[bucket] = ns;
      tbl->hashes[bucket] = hash;

      tbl->size++;

      return bucket+1;
    }

    if (tbl->hashes[bucket] == hash &&
        strcmp(tbl->keys[bucket],s) == 0) {
      return bucket+1;
    }

    if (++bucket >= tbl->capa) {
      bucket = 0;
    }
    // bucket++;
  }

}

static int strings_resize_and_rehash(struct tsch_string_table *tbl)
{
  size_t i, sz, new_capa, old_capa;
  const char **new_str, **old_str;
  unsigned long *new_hash, *old_hash;
  void **old_values, **new_values;

  new_capa = tbl->capa * 2;
  new_str  = ts_calloc(NO_ALLOCATOR, new_capa, sizeof *new_str);
  new_hash = ts_calloc(NO_ALLOCATOR, new_capa, sizeof *new_hash);

  if (tbl->values) {
    new_values = ts_calloc(NO_ALLOCATOR, new_capa, sizeof *new_values);
  } else {
    new_values = NULL;
  }

  if (new_str == NULL  ||
      new_hash == NULL ||
      (tbl->values && new_values == NULL)) {
    ts_free(NO_ALLOCATOR, new_str);
    ts_free(NO_ALLOCATOR, new_hash);
    ts_free(NO_ALLOCATOR, new_values);
    return 0;
  }

  old_capa = tbl->capa;
  sz = tbl->size;
  old_hash = tbl->hashes;
  old_str  = tbl->keys;
  old_values = tbl->values;

  tbl->size = 0;
  tbl->capa = new_capa;
  tbl->keys = new_str;
  tbl->hashes = new_hash;
  tbl->values = new_values;

  for (i=0; i < old_capa; i++) {
    size_t b;

    if (!old_str[i])
      continue;

    b = add_string(tbl, old_str[i], old_hash[i]);
    assert(b != 0);
    if (new_values) {
      new_values[b-1] = old_values[i];
    }
  }

  assert(tbl->size == sz);

  ts_free(NO_ALLOCATOR, old_str);
  ts_free(NO_ALLOCATOR, old_hash);
  ts_free(NO_ALLOCATOR, old_values);

  tbl->rehash_limit = calc_rehash_limit(tbl->capa, tbl->max_load);

  return 1;
}

unsigned int tsch_string_table_load_factor(struct tsch_string_table *tbl)
{
  if (tbl->size > SIZE_MAX / 1000) {
    return tbl->size / (tbl->capa/1000);
  }

  return tbl->size * 1000 / tbl->capa;
}

size_t tsch_string_table_lookup(struct tsch_string_table *tbl, struct tsch_string str)
{
  size_t b,b0;

  b0 = b = str.h % tbl->capa;
  for(;;) {
    if (tbl->hashes[b] == str.h &&
        tbl->keys[b] &&
        strcmp(tbl->keys[b],str.s) == 0)
      return b+1;

    if (++b == tbl->capa) { b = 0; }
    if (b == b0) {
      return 0;
    }
  }
}

size_t tsch_string_table_find(struct tsch_string_table *tbl, const char *s)
{
  struct tsch_string str;
  str.s = s;
  str.h = tsch_string_hash(s);
  return tsch_string_table_lookup(tbl,str);
}

struct tsch_string tsch_string_table_add(struct tsch_string_table *tbl, const char *k, void *v)
{
  size_t b;
  struct tsch_string result;
  unsigned long h;

  if (tbl->size >= tbl->rehash_limit) {
    /* no error handling here because the table tries to proceed if it
     * cannot rehash.  if it cannot add the string, it returns a
     * tsch_string will NULL and no hash
     */
    // fprintf(stderr, "rehashing (size=%zu, capa=%zu, rehash_limit=%zu!\n",
    //     tbl->size, tbl->capa, tbl->rehash_limit);
    strings_resize_and_rehash(tbl);
    // fprintf(stderr, "done.     (size=%zu, capa=%zu, rehash_limit=%zu!\n",
    //     tbl->size, tbl->capa, tbl->rehash_limit);
  }

  h = tsch_string_hash(k);
  b = add_string(tbl,k,h);
  if (b == 0) {
    result.s = NULL;
    result.h = 0;
  } else {
    result.s = tbl->keys[b-1];
    result.h = tbl->hashes[b-1];
    if (tbl->values) { tbl->values[b-1] = v; }
  }

  return result;
}

const char *tsch_string_table_intern(struct tsch_string_table *tbl, const char *s)
{
  if (!s) return NULL;
  struct tsch_string str = tsch_string_table_add(tbl,s,NULL);
  return str.s;
}

