#ifndef TSCH_CONFIG_H
#define TSCH_CONFIG_H

#include <stdbool.h>
#include <stdlib.h>
#include <stdarg.h>

#include "ts_errors.h"
#include "ts_chem_driver.h"

enum TSCH_CFG_TYPE { 
  TSCH_CFG_NONE = 0,
  TSCH_CFG_STRING,
  TSCH_CFG_NUM,
  TSCH_CFG_BOOL,
  TSCH_CFG_MAPPING,
  TSCH_CFG_SEQUENCE,
};

struct tsch_config
{
  struct tsch_config* next;
  const char* key;

  enum TSCH_CFG_TYPE type;
  union {
    const char *s;
    double num;
    bool b;
    struct tsch_config *val;
  };
};

/* Use a pool of tsch_config entries so we can just free the pool after
 * the driver is built.  There's no real reason to keep the remaining
 * entries around.
 */
enum { TSCH_CFG_POOL_SIZE = 256 };
struct tsch_config_pool
{
  struct tsch_config_pool *prev_pool;
  size_t n;
  size_t capa;
  struct tsch_config entries[];
};

struct tsch_config_pool *tsch_config_pool_alloc(struct tsch_config_pool *prev, size_t n);
struct tsch_config *tsch_config_pool_entry(struct tsch_config_pool **poolp);
void tsch_config_pool_dealloc(struct tsch_config_pool *top);

TSResult tsch_config_build_driver(struct tsch_config *root, ts_chem_driver **driverp, int ignore_missing);

/* interned string */
struct tsch_string
{
  const char *s;
  unsigned long h;
};

struct tsch_string_table
{
  size_t capa;
  size_t size;

  /* when size > rehash_limit (set by max_load),
   * then resize and rehash
   */
  size_t rehash_limit;

  const char **keys;
  unsigned long *hashes;
  void **values;

  int max_load; /* fraction of max load factor * 1024 */
};

enum {
  TSCH_STRINGS_DEFAULT_CAPA    = 1024,
  TSCH_STRINGS_DEFAULT_MAXLOAD = 683, /* 683/1024 == 0.6669 ~ 0.667 or 2/3 */
};

TSResult tsch_string_table_initialize(struct tsch_string_table *tbl, size_t capa, bool uses_values);
void tsch_string_table_finalize(struct tsch_string_table *tbl);

/* returns load as a fraction multiplied by 1000 */
unsigned int tsch_string_table_load_factor(struct tsch_string_table *tbl);

size_t tsch_string_table_lookup(struct tsch_string_table *tbl, struct tsch_string str);
size_t tsch_string_table_find(struct tsch_string_table *tbl, const char *s);
unsigned long tsch_string_hash(const char *s);

struct tsch_string tsch_string_table_add(struct tsch_string_table *tbl, const char *k, void* v);
const char *tsch_string_table_intern(struct tsch_string_table *tbl, const char *s);

static inline struct tsch_string
tsch_string_table_intern_str(struct tsch_string_table *tbl, const char *s)
{
  if (!s) return (struct tsch_string){ NULL, 0 };
  return tsch_string_table_add(tbl,s,NULL);
}

static inline struct tsch_config **
tsch_config_tail(struct tsch_config *item)
{
  struct tsch_config **tail;

  for (tail=&item; *tail; tail = &(*tail)->next)
    continue;

  return tail;
}

#endif /* TSCH_CONFIG_H */

