#include "tsch_config.h"

#include "ts_common.h"
#include "ts_common_algo.h"

#include <ctype.h>

enum { OPTIONAL = 1, REQUIRED = 0 };

#define OPTIONAL_DEFAULT(val) (OPTIONAL | (val << 1))

static inline void print_indent(int indent)
{
  char istr[64];
  if (indent > 64) { indent = 64; }
  memset(istr,' ', indent);
  fwrite(istr+0, sizeof(char), indent, stdout);
}

static MAYBE_UNUSED void print_config(struct tsch_config* entry, int indent)
{
  while(entry != NULL) {
    if (indent < 0) {
      indent = -indent;
    } else {
      print_indent(indent);
    }
    if (entry->key != NULL) {
      printf("%s : ", entry->key);
      print_config(entry->val,-indent);
    } else {
      switch (entry->type) {
        case TSCH_CFG_NONE:
          printf("<none>\n");
          break;
        case TSCH_CFG_NUM:
          printf("%f\n",entry->num);
          break;
        case TSCH_CFG_BOOL:
          printf("%s\n",entry->b ? "true" : "false");
          break;
        case TSCH_CFG_STRING:
          printf("\"%s\"\n", entry->s);
          break;
        case TSCH_CFG_SEQUENCE:
          printf("SEQ [\n");
          print_config(entry->val, indent+2);
          print_indent(indent);
          printf("]\n");
          break;
        case TSCH_CFG_MAPPING:
          printf("MAP {\n");
          print_config(entry->val, indent+2);
          print_indent(indent);
          printf("}\n");
          break;
      }
    }

    entry = entry->next;
  }
}

enum { CONFIG_OK = 0, CONFIG_MISSING = -1, CONFIG_INVALID = -2 };
static struct tsch_config* find_map_entry(struct tsch_config* entry, const char* key)
{
  struct tsch_config* item;

  if (entry->type != TSCH_CFG_MAPPING) {
    return NULL;
  }

  for (item=entry->val; item; item = item->next) {
    if ((item->key != NULL) && (strcmp(item->key,key) == 0)) {
      // return item->val;
      return item;
    }
  }

  return NULL;
}

static inline int find_str_with_key(struct tsch_config* entry, const char* key, const char** val)
{
  struct tsch_config* value = find_map_entry(entry,key);
  if (value == NULL) { return CONFIG_MISSING; }
  if ((value->type != TSCH_CFG_STRING) || (value->s == NULL)) {
    return CONFIG_INVALID;
  }
  *val = value->s;
  return CONFIG_OK;
}

static inline int find_int_with_key(struct tsch_config* entry, const char* key, long* val)
{
  struct tsch_config *item = find_map_entry(entry, key);
  if (!item) { return CONFIG_MISSING; }

  if (item->type == TSCH_CFG_NUM) {
    *val = item->num;
    return CONFIG_OK;
  }

  if (item->type == TSCH_CFG_STRING) {
    char* end;
    long v;

    if (!item->s) {
      return CONFIG_INVALID;
    }

    end = NULL;
    errno = 0;
    v = strtol(item->s, &end, 0);
    if (errno == 0 && end && *end == '\0') {
      *val = v;
      return CONFIG_OK;
    }
  }

  return CONFIG_INVALID;
}

static const char *true_values[] = {
  "true", "TRUE", "yes", "on", "T", "t", "1", NULL
};

static const char *false_values[] = {
  "false", "FALSE", "no", "off", "F", "f", "0", NULL
};

static inline int find_bool_with_key(struct tsch_config* entry, const char* key, int* val)
{
  struct tsch_config *item= find_map_entry(entry, key);
  if (!item) { return CONFIG_MISSING; }

  if (item->type == TSCH_CFG_BOOL) {
    *val = item->b;
    return CONFIG_OK;
  }

  if (item->type == TSCH_CFG_NUM) {
    *val = !!(long)item->num;
    return CONFIG_OK;
  }

  if (item->type == TSCH_CFG_STRING) {
    int i;
    const char *s;
    char* end;

    s = item->s;

    for (i=0; true_values[i] != NULL; i++) {
      if (STRINGS_EQ(s,true_values[i])) {
        *val = 1;
        return CONFIG_OK;
      }
    }

    for (i=0; false_values[i] != NULL; i++) {
      if (STRINGS_EQ(s,false_values[i])) {
        *val = 0;
        return CONFIG_OK;
      }
    }
  }

  return CONFIG_INVALID;
}

static inline int find_dbl_with_key(struct tsch_config* entry, const char* key, double* val)
{
  struct tsch_config *item= find_map_entry(entry, key);
  if (!item) { return CONFIG_MISSING; }

  if (item->type == TSCH_CFG_NUM) {
    *val = item->num;
    return CONFIG_OK;
  }

  if (item->type == TSCH_CFG_STRING) {
    char* end;
    double v;

    if (!item->s) {
      return CONFIG_INVALID;
    }

    end = NULL;
    v = strtod(item->s, &end);
    if (end && *end == '\0') {
      *val = v;
      return CONFIG_OK;
    }
  }

  return CONFIG_INVALID;
}

static TSResult
config_bool_value(struct tsch_config *entry,
    const char *key, const char *desc, int *valp, int opt)
{
  int res = find_bool_with_key(entry,key,valp);
  switch(res) {
    case CONFIG_MISSING:
      if (opt) { *valp = opt >> 1; }
      else {
        TS_RAISE_ERROR2(TS_ERR_CONFIG_MISSING_VALUE,
            ("missing required '%s'",desc));
      }
      return TS_OK;

    default:
    case CONFIG_OK:
      return TS_OK;

    case CONFIG_INVALID:
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("'%s' value is not a valid boolean",desc));
  }
}

#define CONFIG_BOOL_VALUE(_root,_optname,_desc,_var,_opt) \
  RETURN_ON_ERROR( config_bool_value(_root,_optname,_desc,&(_var),_opt) )

static TSResult
config_int_value(struct tsch_config *entry,
    const char *key, const char *desc, long *valp, int opt)
{
  int res = find_int_with_key(entry,key,valp);
  switch(res) {
    case CONFIG_MISSING:
      if (opt) { *valp = 0; }
      else {
        TS_RAISE_ERROR2(TS_ERR_CONFIG_MISSING_VALUE,
            ("missing required '%s'",desc));
      }
      return TS_OK;

    default:
    case CONFIG_OK:
      return TS_OK;

    case CONFIG_INVALID:
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("'%s' value is not a valid number",desc));
  }
}

#define CONFIG_INT_VALUE(_root,_optname,_desc,_var,_opt) \
  RETURN_ON_ERROR( config_int_value(_root,_optname,_desc,&(_var),_opt) )

static TSResult
config_dbl_value(struct tsch_config *entry,
    const char *key, const char *desc, double *valp, int opt)
{
  int res = find_dbl_with_key(entry,key,valp);
  switch(res) {
    case CONFIG_MISSING:
      if (opt) { *valp = 0.0; }
      else {
        TS_RAISE_ERROR2(TS_ERR_CONFIG_MISSING_VALUE,
            ("missing required '%s'",desc));
      }
      return TS_OK;

    default:
    case CONFIG_OK:
      return TS_OK;

    case CONFIG_INVALID:
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("'%s' value is not a valid number",desc));
  }
}

#define CONFIG_DBL_VALUE(_root,_optname,_desc,_var,_opt) \
  RETURN_ON_ERROR( config_dbl_value(_root,_optname,_desc,&(_var),_opt) )

static TSResult
config_str_value(struct tsch_config *entry,
    const char *key, const char *desc, const char **valp, int opt)
{
  int res = find_str_with_key(entry,key,valp);
  switch(res) {
    case CONFIG_MISSING:
      if (opt) { *valp = NULL; }
      else {
        TS_RAISE_ERROR2(TS_ERR_CONFIG_MISSING_VALUE,
            ("missing required '%s'",desc));
      }
      return TS_OK;

    default:
    case CONFIG_OK:
      return TS_OK;

    case CONFIG_INVALID:
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("'%s' value is not a valid string",desc));
  }
}

#define REQUIRE_STR_VALUE(_root,_optname,_desc,_var) \
  RETURN_ON_ERROR( config_str_value(_root,_optname,_desc,&(_var), ignore_missing) )

static inline int check_valid_state(int state,
    const enum TS_CHEMDRIVER_SITE_STATE* state_beg, 
    const enum TS_CHEMDRIVER_SITE_STATE* state_end)
{
  const enum TS_CHEMDRIVER_SITE_STATE* statep = state_beg;
  while(statep < state_end) {
    if ((int)*statep == state) { return 1; }
    statep++;
  }
  return 0;
}

static inline int decode_monomer_state(char s)
{
  char* which_mer = strchr(ts_chem_driver_site_codes, s);
  if (which_mer == NULL) { return -1; }
  
  int mer1 = which_mer - ts_chem_driver_site_codes;
  if (check_valid_state(mer1,
        monomer_states_all + NOT_EMPTY(monomer_states_all),
        monomer_states_all + NUM_CONST(monomer_states_all))) {
    return mer1;
  } else {
    return -1;
  }
}

static inline int decode_interface_state(char s)
{
  char* which_ifc = strchr(ts_chem_driver_site_codes, s);
  if (which_ifc == NULL) { return -1; }
  
  int ifc = which_ifc - ts_chem_driver_site_codes;
  if (check_valid_state(ifc,
        longitudinal_states + NOT_EMPTY(longitudinal_states),
        longitudinal_states + NUM_CONST(longitudinal_states)) ||
      check_valid_state(ifc,
        lateral_states + NOT_EMPTY(lateral_states),
        lateral_states + NUM_CONST(lateral_states))) {
    return ifc;
  } else {
    return -1;
  }
}

static inline void driver_set_flag(ts_chem_driver *drv, uint32_t flag, int tf)
{
  if (tf) { tsch_set_flags(drv, flag); }
  else    { tsch_unset_flags(drv, flag); }
}

static int read_energy(ts_chem_driver *drv, const char *s, double *enp)
{
  char* end;

  if (STRINGS_EQ(s, "FORBIDDEN")) {
    *enp = drv->forbidden_energy;
    return 1;
  }

  end = NULL;
  *enp = strtod(s, &end);
  return (end && *end == '\0');
}

static bool parse_energy(ts_chem_driver *driver, struct tsch_config *entry, double *valp)
{
  if (entry->type == TSCH_CFG_NUM)
  {
    *valp = entry->num;
    return true;
  }

  if (entry->type == TSCH_CFG_STRING)
  {
    return read_energy(driver,entry->s, valp);
  }

  return false;
}

struct driver_flag {
  const char *name;
  const char *desc;
  enum TSCH_FLAGS flag;
  int req;
};

static const struct driver_flag driver_flags[] = {
  {
    "reversible_lateral_bond_breaking",
    "reversible_lateral_bond_breaking", 
    TSCH_REVERSIBLE_LATERAL_BOND_BREAKING, OPTIONAL,
  },
  {
    "restrict_curved_lateral_bonding", 
    "restrict_curved_lateral_bonding", 
    TSCH_RESTRICT_CURVED_LATERAL_BONDING, OPTIONAL,
  },
  {
    "disable_ccs_depolymerization", 
    "disable_ccs_depolymerization", 
    TSCH_DISABLE_CCS_DEPOLYMERIZATION, OPTIONAL,
  },
  {
    "allow_depolymerization_at_straight_interfaces",
    "allow_depolymerization_at_straight_interfaces", 
    TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS, OPTIONAL,
  },
  {
    "disallow_polymerization_at_straight_interfaces",
    "disallow_polymerization_at_straight_interfaces", 
    TSCH_DISALLOW_POLYM_AT_STRAIGHT_IFCS, OPTIONAL,
  },
  {
    "polym_only_longitudinal",
    "polym_only_longitudinal", 
    TSCH_POLYM_ONLY_LONGITUDINAL, OPTIONAL,
  },
  {
    "no_minus_end_depolymerization",
    "no_minus_end_depolymerization",
    TSCH_NO_MINUS_END_DEPOLYMERIZATION, OPTIONAL,
  },
  {
    "minus_end_polymerization",
    "minus_end_polymerization",
    TSCH_MINUS_END_POLYMERIZATION, OPTIONAL,
  },
  {
    "polym_metropolis_weighted",
    "polym_metropolis_weighted", 
    TSCH_POLYM_METROPOLIS_WEIGHTED, OPTIONAL,
  },
  {
    "hydrolysis_only_at_sss",
    "hydrolysis_only_at_sss",
    TSCH_HYDROLYSIS_ONLY_AT_SSS, OPTIONAL,
  },
  { 0 }
};


static const struct {
  const char *name;
  const char *desc;
  enum TS_CHEMDRIVER_EVENT_TYPE which;
} prefactors[] = {
  { "monomer"     , "rates.monomer"     , TS_CHEMDRIVER_EVT_MONOMER     },
  { "longitudinal", "rates.longitudinal", TS_CHEMDRIVER_EVT_LONGITUDINAL},
  { "lateral"     , "rates.lateral"     , TS_CHEMDRIVER_EVT_LATERAL     },
  { "seam"        , "rates.seam"        , TS_CHEMDRIVER_EVT_LATERAL_SEAM},
  { "polymerize"  , "rates.polymerize"  , TS_CHEMDRIVER_EVT_POLYMERIZE  },
  { "hydrolysis"  , "rates.hydrolysis"  , TS_CHEMDRIVER_EVT_HYDROLYSIS  },
  { "k_on"        , "rates.k_on"        , TS_CHEMDRIVER_EVT_K_ON        },
  { 0 },
};


static TSResult parse_root_settings(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing);
static TSResult parse_barrier_model(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing);
static TSResult parse_prefactors(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing);
static TSResult parse_monomer_energies(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing);
static TSResult parse_ifc_energies(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing);

static TSResult parse_bending_energies(ts_chem_driver *driver, struct tsch_config *root);
static TSResult parse_coupling_energies(ts_chem_driver *driver, struct tsch_config *root);
static TSResult parse_dissoc_barriers(ts_chem_driver *driver, struct tsch_config *root);

TSResult tsch_config_build_driver(struct tsch_config *root, ts_chem_driver **driverp, int ignore_missing)
{
  TSResult err;
  ts_chem_driver* driver;

  if (root->type != TSCH_CFG_MAPPING) {
    TS_RAISE_ERROR3(TS_ERR_PARSE_ERROR,
        "Configuration file was not a key-value dictionary");
  }

  long max_dimers = 0;
  long num_pfs = 0;

  CONFIG_INT_VALUE(root, "num_protofilaments", "num_protofilaments", num_pfs, REQUIRED);
  CONFIG_INT_VALUE(root, "max_dimers", "max_dimers", max_dimers, REQUIRED);

  if (max_dimers <= 0) {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE, "max_dimers missing or <= 0");
  }

  if (num_pfs <= 0) {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE, "num_protofilaments missing or <= 0");
  }

  int nrow = max_dimers*4 + 1;  // first and last row are kept empty
  int ncol = num_pfs*2;

  // Allocate driver
  {
    err = TS_OK;
    driver = ts_chem_driver_new(nrow, ncol, &err);
    if (IS_ERROR(err)) {
      return err;
    }
  }

#define CLEANUP_ON_ERROR { ts_chem_driver_release(driver); }

  /* Driver settings in the root map */
  RETURN_ON_ERROR(parse_root_settings(driver, root, ignore_missing));

  RETURN_ON_ERROR(parse_barrier_model(driver, root, ignore_missing));
  RETURN_ON_ERROR(parse_prefactors(driver, root, ignore_missing));
  RETURN_ON_ERROR(parse_monomer_energies(driver, root, ignore_missing));
  RETURN_ON_ERROR(parse_ifc_energies(driver, root, ignore_missing));

  RETURN_ON_ERROR(parse_bending_energies(driver, root));
  RETURN_ON_ERROR(parse_coupling_energies(driver, root));
  RETURN_ON_ERROR(parse_dissoc_barriers(driver, root));

#undef CLEANUP_ON_ERROR

  *driverp = driver;
  return TS_OK;
}

static TSResult parse_root_settings(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing)
{
  double t_max;
  double dt;
  double temp;
  double forbidden_energy;
  long rng_seed = 0;
  long seam_allowed_thresh = 0;
  int seam_detailed_balance = 1;
  long coupling_edge_distance = 0;
  int i;

  CONFIG_DBL_VALUE(root, "t_max", "t_max", t_max, ignore_missing);
  CONFIG_DBL_VALUE(root, "dt", "dt", dt, ignore_missing);
  CONFIG_DBL_VALUE(root, "temp", "temp", temp, ignore_missing);
  CONFIG_DBL_VALUE(root, "FORBIDDEN", "FORBIDDEN", forbidden_energy, ignore_missing);

  for (i=0; driver_flags[i].name; i++)
  {
    int b=0;
    RETURN_ON_ERROR(config_bool_value(
          root,driver_flags[i].name,driver_flags[i].desc,&b,driver_flags[i].req));
    driver_set_flag(driver, driver_flags[i].flag, b);
  }

  CONFIG_BOOL_VALUE(root,
      "seam_detailed_balance",
      "seam_detailed_balance", 
      seam_detailed_balance, OPTIONAL_DEFAULT(1));

  if (find_map_entry(root,"no_minus_end_depolymerization") != NULL)
  {
    /* FIXME: need a better way to warn about this! */
    fprintf(stderr, "option no_minus_end_depolymerization is DEPRECATED\n");
  }

  driver->dt = dt;
  driver->t_max = t_max;
  driver->latt.temp = temp;
  driver->forbidden_energy = forbidden_energy;

  driver_set_flag(driver,
      TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE,
      !seam_detailed_balance);

  CONFIG_INT_VALUE(root, "rng_seed", "rng_seed", rng_seed, ignore_missing);
  if (rng_seed > 0) {
    struct ts_rng* rng = NULL;
    RETURN_ON_ERROR( RNG_mt19937_new((uint32_t)rng_seed, &rng) );

    driver->latt.rng = rng;
  } else if (!ignore_missing) {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE, "'rng_seed' must be > 0");
  }

  /* This is an optional setting, with a different default */
  if (find_map_entry(root, "seam_allowed_thresh") != NULL) {
    CONFIG_INT_VALUE(root, "seam_allowed_thresh", "seam_allowed_thresh", seam_allowed_thresh, OPTIONAL);
    driver->seam_allowed_thresh = seam_allowed_thresh;
  }

  if (find_map_entry(root, "coupling_edge_distance") != NULL) {
    CONFIG_INT_VALUE(root, "coupling_edge_distance",
        "coupling_edge_distance", coupling_edge_distance, OPTIONAL);
    driver->coupling_edge_dist = coupling_edge_distance;
  }

  return TS_OK;
}

static TSResult parse_barrier_model(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing)
{
  struct tsch_config *barrier_cfg, *bparam;
  enum TS_CHEMDRIVER_BARRIER_MODEL bmodel;
  double val;

  if (barrier_cfg = find_map_entry(root, "barrier"), !barrier_cfg)
  {
    if (ignore_missing)
    {
      return TS_OK;
    }

    TS_RAISE_ERROR3(TS_ERR_CONFIG_MISSING_SECTION, "Missing 'barrier' section");
  }

  const char* btype = NULL;
  REQUIRE_STR_VALUE(barrier_cfg, "type", "barrier.type", btype);
  if (btype == NULL)
  {
    return TS_OK;
  }

  if      (STRINGS_EQ(btype, "none"          )) { bmodel = TS_CHEMDRIVER_BARRIER_NONE;          }
  else if (STRINGS_EQ(btype, "metropolis"    )) { bmodel = TS_CHEMDRIVER_BARRIER_METROPOLIS;    }
  else if (STRINGS_EQ(btype, "kawasaki"      )) { bmodel = TS_CHEMDRIVER_BARRIER_KAWASAKI;      }
  else if (STRINGS_EQ(btype, "proportional"  )) { bmodel = TS_CHEMDRIVER_BARRIER_PROPORTIONAL;  }
  else if (STRINGS_EQ(btype, "infinite"      )) { bmodel = TS_CHEMDRIVER_BARRIER_INFINITE;      }
  else if (STRINGS_EQ(btype, "always_accept" )) { bmodel = TS_CHEMDRIVER_BARRIER_ALWAYS_ACCEPT; }
  else if (STRINGS_EQ(btype, "always_reject" )) { bmodel = TS_CHEMDRIVER_BARRIER_ALWAYS_REJECT; }
  else {
    TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE, ("unknown barrier model '%s'", btype));
  }
  driver->barrier_model = bmodel;

  bparam = find_map_entry(barrier_cfg, "params");
  if (bparam == NULL)
  {
    return TS_OK;
  }

  if (bparam->type == TSCH_CFG_SEQUENCE) {
    bparam = bparam->val;
  }

  if (bparam->type == TSCH_CFG_NUM) {
    val = bparam->num;
  } else if (bparam->type == TSCH_CFG_STRING) {
    char* end = NULL;

    val = strtod(bparam->s, &end);
    if (!end || (*end != '\0')) {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid entry in 'barrier.params': %s", bparam->s));
    }
  } else {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "barrier.params must be either a single number of a sequence of numbers");
  }
  driver->barrier_param = val;

  return TS_OK;
}

/* parses the rates section to configure prefactors */
static TSResult parse_prefactors(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing)
{
  struct tsch_config *rates_cfg;
  int i;

  if (rates_cfg = find_map_entry(root, "rates"), !rates_cfg)
  {
    if (ignore_missing)
    {
      return TS_OK;
    }

    TS_RAISE_ERROR3(TS_ERR_CONFIG_MISSING_SECTION, "Missing 'rates' section");
  }

  for (i=0; prefactors[i].name; i++)
  {
    double rate = 0.0;
    CONFIG_DBL_VALUE(rates_cfg,
        prefactors[i].name, prefactors[i].desc, rate, ignore_missing);
    driver->rate_prefactors[prefactors[i].which] = rate;
  }

  return TS_OK;
}

static TSResult
parse_energy_entry(ts_chem_driver *driver, const char *section,
    struct tsch_config *entry, double *enp)
{
  if (!parse_energy(driver, entry, enp))
  {
    if (entry->type == TSCH_CFG_STRING)
    {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
          "%s[%s] is not a number", section,entry->key);
    }
    else
    {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
          "%s[%s] has an invalid (non-scalar) value (type %d)",
          section, entry->key, (int)entry->type);
    }
  }

  return TS_OK;
}

static TSResult
parse_monomer_energies(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing)
{
  struct tsch_config *mer_energies, *entry;

  if (mer_energies = find_map_entry(root, "monomer_energies"), !mer_energies)
  {
    if (ignore_missing)
    {
      return TS_OK;
    }

    TS_RAISE_ERROR3(TS_ERR_CONFIG_MISSING_SECTION, "Missing 'monomer_energies' section");
  }

  if (mer_energies->type != TSCH_CFG_MAPPING) {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "'monomer_energies' is not a key-value dictionary");
  }

  for (entry = mer_energies->val; entry; entry = entry->next)
  {
    const char *key;
    int monomer_state;
    double *enp;

    key = entry->key;
    if (key == NULL)
    {
      continue;
    }

    if (strlen(key) != 1)
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid monomer energy key '%s'", key));
    }

    if (monomer_state = decode_monomer_state(key[0]), monomer_state < 0)
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid monomer energy key '%s'", key));
    }

    RETURN_ON_ERROR(parse_energy_entry(driver, "monomer_energies", entry, 
          &driver->monomer_energies[monomer_state]));
  }

  return TS_OK;
}

static inline size_t
ifc_ind(int mer1, int ifc, int mer2)
{
  return mer1+(ifc + mer2*TS_CHEMDRIVER_STATE_MAX)*TS_CHEMDRIVER_STATE_MAX;
}

static TSResult
parse_ifc_energies_inner(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing,
    char *ifc_set)
{
  struct tsch_config *ifc_energies, *entry;
  if (ifc_energies = find_map_entry(root, "interface_energies"), !ifc_energies)
  {
    if (ignore_missing)
    {
      return TS_OK;
    }

    TS_RAISE_ERROR3(TS_ERR_CONFIG_MISSING_SECTION, "Missing 'interface_energies' section");
  }

  if (ifc_energies->type != TSCH_CFG_MAPPING)
  {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "'interface_energies' is not a key-value dictionary");
  }

  for (entry = ifc_energies->val; entry; entry = entry->next)
  {
    double *enp;
    int mer1, ifc, mer2;
    size_t ind;
    const char* key;

    if (key = entry->key, key == NULL)
    {
      continue;
    }

    if (strlen(key) != 3)
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid interface key '%s'", key));
    }

    mer1 = decode_monomer_state(key[0]);
    ifc  = decode_interface_state(key[1]);
    mer2 = decode_monomer_state(key[2]);

    if ((mer1 < 0) || (mer1 >= TS_CHEMDRIVER_STATE_MAX) ||
        (ifc < 0)  || (ifc  >= TS_CHEMDRIVER_STATE_MAX) ||
        (mer2 < 0) || (mer2 >= TS_CHEMDRIVER_STATE_MAX))
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid interface key '%s'", key));
    }

    ind = ifc_ind(mer1,ifc,mer2);
    if (ifc_set[ind]) {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_DUPLICATE_VALUE,
          "duplicate entry for interface key '%s'", key);
    }
    ifc_set[ind] = 1;

    RETURN_ON_ERROR(parse_energy_entry(driver, "interface_energies", entry, 
          &driver->interface_energies[mer1][ifc][mer2]));
  }

  return TS_OK;
}

static TSResult
parse_ifc_energies(ts_chem_driver *driver, struct tsch_config *root, int ignore_missing)
{
  TSResult err;
  size_t nmers, nifcs, nb;
  char *ifc_set;

  nmers = ARRAYLEN(driver->interface_energies);
  nifcs = ARRAYLEN(driver->interface_energies[0]);
  nb = nmers*nmers*nifcs;

  assert(nb == sizeof driver->interface_energies / sizeof driver->interface_energies[0][0][0]);
  ifc_set = ts_alloc(NO_ALLOCATOR, nmers*nifcs*nmers);
  memset(ifc_set, 0, nb);

  err = ON_ERROR_ADD_FRAME(
      parse_ifc_energies_inner(driver, root, ignore_missing, ifc_set)
    );

  ts_free(NO_ALLOCATOR, ifc_set);

  return err;
}

static TSResult
parse_bending_energies(ts_chem_driver *driver, struct tsch_config *root)
{
  struct tsch_config *bending_energies, *entry;

  if (bending_energies = find_map_entry(root, "bending_energies"), !bending_energies)
  {
    return TS_OK;
  }

  if (bending_energies->type != TSCH_CFG_MAPPING)
  {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "'bending_energies' is not a key-value dictionary");
  }

  for (entry = bending_energies->val; entry; entry = entry->next)
  {
    double *enp;
    int ifc1,mer,ifc2;
    const char* key;

    key = entry->key;
    if (!key)
    {
      continue;
    }

    if (strlen(key) != 3)
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid bending key '%s'", key));
    }

    ifc1 = decode_interface_state(key[0]);
    mer  = decode_monomer_state(key[1]);
    ifc2 = decode_interface_state(key[2]);

    if ((ifc1 < 0) || (mer < 0) || (ifc2 < 0))
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid bending key '%s'", key));
    }

    RETURN_ON_ERROR(parse_energy_entry(driver, "bending_energies", entry, 
          &driver->interface_energies[ifc1][mer][ifc2]));
  }

  return TS_OK;
}

static TSResult parse_coupling_energy_pair(ts_chem_driver *driver,
    struct tsch_config *entry, double *elop, double *ehip)
{
  struct tsch_config *item;

  assert(driver);
  assert(entry);
  assert(elop && ehip);

  if (entry->type == TSCH_CFG_SEQUENCE)
  {
    item = entry->val;
    if (!item)
    {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
          "empty sequence for key '%s'", entry->key);
    }
  }
  else
  {
    item = entry;
  }

  if (!parse_energy(driver, item, elop))
  {
    TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
        "coupling_energies[%s] first energy is not a number", entry->key);
  }

  if (item == entry || !item->next)
  {
    *ehip = *elop;
    return TS_OK;
  }

  item = item->next;
  if (item->next)
  {
    TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
        "coupling_energies[%s] has too many items (expect 1-2)", entry->key);
  }

  if (!parse_energy(driver, item, ehip))
  {
    TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
        "coupling_energies[%s] second item is not a number", entry->key);
  }

  return TS_OK;
}

static TSResult parse_coupling_energies(ts_chem_driver *driver, struct tsch_config *root)
{
  struct tsch_config *coupling_energies, *entry;
  int terms[4];

  coupling_energies = find_map_entry(root, "coupling_energies");
  if (!coupling_energies)
  {
    return TS_OK;
  }

  if (coupling_energies->type != TSCH_CFG_MAPPING)
  {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "'coupling_energies' is not a key-value dictionary");
  }

  for(entry = coupling_energies->val; entry != NULL; entry = entry->next)
  {
    double elo = 0.0, ehi = 0.0;
    const char *key = entry->key;

    if (key == NULL)
    {
      continue;
    }

    if (strlen(key) != 4)
    {
      TS_RAISE_ERROR2(TS_ERR_CONFIG_INVALID_VALUE,
          ("invalid coupling key '%s'", key));
    }

    terms[0] = decode_interface_state(key[0]);
    terms[1] = decode_interface_state(key[1]);
    terms[2] = decode_interface_state(key[2]);
    terms[3] = decode_interface_state(key[3]);

    if (!TSCH_IN_STATE_GROUP(terms[0],lateral_states) || 
        !TSCH_IN_STATE_GROUP(terms[1],lateral_states) || 
        !TSCH_IN_STATE_GROUP(terms[2],longitudinal_states) || 
        !TSCH_IN_STATE_GROUP(terms[3],longitudinal_states))
    {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
          "invalid coupling key '%s' (expected lat-lat-long-long)", key);
    }

    RETURN_ON_ERROR(parse_coupling_energy_pair(driver, entry, &elo,&ehi));
    RETURN_ON_ERROR(ts_chem_driver_add_coupling_energy(driver, terms, elo, ehi));
  }

  return TS_OK;
}

static TSResult parse_dissoc_barriers(ts_chem_driver *driver, struct tsch_config *root)
{
  struct tsch_config *dissoc_barriers, *entry;

  dissoc_barriers = find_map_entry(root, "dissociation_barriers");
  if (dissoc_barriers == NULL)
  {
    return TS_OK;
  }

  if (dissoc_barriers->type != TSCH_CFG_MAPPING)
  {
    TS_RAISE_ERROR3(TS_ERR_CONFIG_INVALID_VALUE,
        "'dissociation_barriers' is not a key-value dictionary");
  }

  for(entry = dissoc_barriers->val; entry != NULL; entry = entry->next)
  {
    double en = 0.0;
    const char* key;
    spin_t terms[3];

    if (key = entry->key, !key)
    {
      continue;
    }

    if (strlen(key) != 3) {
      TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
          "invalid longitudinal interface '%s' for dissociation barrier", key);
    }

    terms[0] = spin_from_int(decode_monomer_state(key[0]));
    terms[1] = spin_from_int(decode_interface_state(key[1]));
    terms[2] = spin_from_int(decode_monomer_state(key[2]));

  if (!TSCH_IN_STATE_GROUP(terms[0].val,monomer_states_all) || 
      !TSCH_IN_STATE_GROUP(terms[1].val,longitudinal_states) || 
      !TSCH_IN_STATE_GROUP(terms[2].val,monomer_states_all))
  {
    fprintf(stderr, "\n\nbad longitudinal interface '%s' (%1d %1d %1d): %1d %1d %1d\n\n",
        key, terms[0].val, terms[1].val, terms[2].val,
        TSCH_IN_STATE_GROUP(terms[0].val,monomer_states_all),
        TSCH_IN_STATE_GROUP(terms[1].val,longitudinal_states),
        TSCH_IN_STATE_GROUP(terms[2].val,monomer_states_all));
    TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
        "invalid longitudinal interface '%s' for dissociation barrier", key);
  }

    if (!parse_energy(driver, entry, &en))
    {
      if (entry->type == TSCH_CFG_STRING)
      {
        TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
            "dissociation_barriers[%s] is not a number", key);
      }
      else
      {
        TS_RAISE_ERRORF(TS_ERR_CONFIG_INVALID_VALUE,
            "dissociation_barriers[%s] has an invalid (non-scalar) value (type %d)",
            key, (int)entry->type);
      }
    }

    RETURN_ON_ERROR(ts_chem_driver_add_dissociation_barrier(driver, terms, en));
  }

  return TS_OK;
}





