#include "tsch_config.h"
#include "tsch_config_parse.h"
#include "tsch_config_priv.h"

#include <ctype.h>

static const char *skipws(const char *s)
{
  for (; *s && *s != '\n' && isspace(*s); s++)
    continue;

  return s;
}

static void record_location(struct tsch_cfg_lex *lx, size_t *linep, size_t *colp)
{
  *linep = tsch_cfg_lex_line(lx);
  *colp  = tsch_cfg_lex_col(lx);
}

static inline int  lex_set_vbuf_char(struct tsch_cfg_lex *lx, char c)
{
  lx->vbuf[0] = c;
  lx->vbuf[1] = '\0';
  return (unsigned char)c;
}

static char *lex_set_vbuf(struct tsch_cfg_lex *lx, const char *s, ptrdiff_t len)
{
  if (len < 0) {
    snprintf(lx->ebuf,sizeof lx->ebuf, "identifier length is negative");
    return NULL;
  }

  if ((size_t)len >= sizeof(lx->vbuf)) {
    snprintf(lx->ebuf,sizeof lx->ebuf,
        "identifier length is %zd but max length is %zu",
        len, sizeof(lx->vbuf)-1);
    return NULL;
  }

  memcpy(lx->vbuf, s, len);
  lx->vbuf[len] = '\0';
  return lx->vbuf;
}

static inline bool is_id_begin(int ch)
{
  return ch == '@' || ch == '$' || ch == '_' || isalpha(ch);
}

static inline bool is_id_rest(int ch)
{
  return ch == '_' || isalnum(ch);
}

static inline void update_pos(struct tsch_cfg_lex *lx, const char *pos)
{
  ptrdiff_t delta = pos - lx->buf;
  assert(delta >= 0);
  lx->state[1].pos = (size_t)delta;
  record_location(lx, &lx->pos1.line, &lx->pos1.col);
}

static int lex_set_error(struct tsch_cfg_lex *lx, const char *pos, const char *fmt, ...)
{
  va_list args;

  update_pos(lx,pos);
  va_start(args, fmt);
  vsnprintf(lx->ebuf,sizeof lx->ebuf,fmt,args);
  va_end(args);

  return TOK_ERR;
}

static const struct {
  const char *kw;
  int token;
} keywords[] = {
  { "true"  , KW_TRUE  },
  { "false" , KW_FALSE },
  { "shared", KW_SHARED },
  { "SHARED", KW_SHARED },
  { 0 },
};

static int lex_id(struct tsch_cfg_lex *lx, const char *beg)
{
  const char *pos, *stok;
  ptrdiff_t len;
  int tok;

  /* first character should be fine */
  for(pos=beg+1; is_id_rest(*pos); pos++)
    continue;

  len = pos-beg;
  update_pos(lx, pos);

  if (stok = lex_set_vbuf(lx, beg, len), stok == NULL) {
    return TOK_ERR;
  }

  lx->tok_value.str = stok;

  for (tok=0; keywords[tok].kw; tok++) {
    if (strcmp(stok, keywords[tok].kw) == 0) {
      return keywords[tok].token;
    }
  }

  return TOK_ID;
}

static int lex_string(struct tsch_cfg_lex *lx, const char *beg)
{
  const char *pos, *beg0;
  int ch0;
  size_t len;
  bool escapes, newlines;

  assert(*beg == '\'' || *beg == '\"');
  beg0 = beg;
  ch0 = *beg0;

  escapes = (ch0 == '\"');
  if (beg[1] == ch0 && beg[2] == ch0) {
    newlines = true;
    beg += 3;
  } else {
    newlines = false;
    beg++;
  }

  len = 0;
  pos = beg;
  for (;;) {
    int ch = *pos;
    if (escapes && ch == '\\') {
      int esc = 0;
      switch (pos[1]) {
        case 't': esc = '\t'; break;
        case 'n': esc = '\n'; break;
        case 'r': esc = '\r'; break;
        case '\"': esc = '\"'; break;
        case '\'': esc = '\''; break;
        /* FIXME: more escape sequences! */
        default:
          return lex_set_error(lx, pos, "invalid escape sequence in string");
      }
      lx->vbuf[len++] = esc;
      pos += 2;
      continue;
    } 

    if (ch == ch0) {
      if (!newlines) {
        pos++;
        break;
      }

      if (pos[1] == ch0 && pos[2] == ch0) {
        pos += 3;
        break;
      }
    }

    if (!newlines && ch == '\n') {
      return lex_set_error(lx, pos,
        "invalid newline in string (use triple-quoted strings)");
    } 

    if (ch == '\0') {
      return lex_set_error(lx, pos, "string does not end");
    }

    if (len >= sizeof(lx->vbuf)) {
        return lex_set_error(lx, pos,
          "string exceeds maximum size (%zu characters)",
          sizeof(lx->vbuf));
    }

    lx->vbuf[len++] = *pos++;
  }

  update_pos(lx, pos);

  lx->vbuf[len] = '\0';
  lx->tok_value.str = lx->vbuf;
  return TOK_STR;
}

static int lex_number(struct tsch_cfg_lex *lx, const char *beg)
{
  double v;
  char *end = NULL;

  assert(beg != NULL);
  v = strtod(beg, &end);

  if (is_id_begin(*end) || is_id_rest(*end) || *end == '\"' || *end == '\'') {
    return lex_set_error(lx, end, "invalid number");
  }

  update_pos(lx, end);

  lx->tok_value.num = v;
  return TOK_NUM;
}

TSResult tsch_cfg_lex_initialize(struct tsch_cfg_lex *lx, const char *s)
{
  *lx = (struct tsch_cfg_lex) { 0 };
  lx->buf = s;
  return TS_OK;
}

void tsch_cfg_lex_finalize(struct tsch_cfg_lex *lx)
{
  *lx = (struct tsch_cfg_lex) { 0 };
}

int tsch_config_lex(struct tsch_cfg_lex *lx)
{
  const char *beg;
  lx->state[0] = lx->state[1];
  beg = skipws(lx->buf + lx->state[1].pos);

  record_location(lx, &lx->pos0.line, &lx->pos0.col);

restart:
  switch (*beg) {
  case '\0':
    return EOF;

  case '\n':
    lx->state[1].line++;
    update_pos(lx, beg+1);
    lx->state[1].line0 = lx->state[1].pos;
    return *beg;

  case '-':
  case '[': case ']':
  case '(': case ')':
  case '{': case '}':
  case ':': case ',': case ';':
  case '*': case '/':
  case '+': case '%':
    lex_set_vbuf_char(lx, *beg);
    update_pos(lx, beg+1);
    return *beg;

  case '&':
  case '|':
    if (beg[1] != beg[0]) {
      goto invalid_char;
    }
    lex_set_vbuf(lx, beg, 2);
    update_pos(lx, beg+2);
    return beg[0] == '&' ? TOK_AND : TOK_OR;

  case '!':
    if (beg[1] == '=') {
      lex_set_vbuf(lx, beg, 2);
      update_pos(lx, beg+2);
      return TOK_NE;
    }
    lex_set_vbuf_char(lx, *beg);
    update_pos(lx, beg+1);
    return *beg;

  case '=':
    if (beg[1] != '=') {
      goto invalid_char;
    }
    lex_set_vbuf(lx, beg, 2);
    update_pos(lx, beg+2);
    return TOK_EQ;

  case '<':
  case '>':
    if (beg[1] == '?') {
      lex_set_vbuf(lx, beg, 2);
      update_pos(lx, beg+2);
      return *beg == '<' ? TOK_MIN : TOK_MAX;
    }
    if (beg[1] == '=') {
      lex_set_vbuf(lx, beg, 2);
      update_pos(lx, beg+2);
      return *beg == '<' ? TOK_LE : TOK_GE;
    }
    update_pos(lx, beg+1);
    return lex_set_vbuf_char(lx, *beg);

  case '\'':
  case '\"':
    return lex_string(lx, beg);

  case '\\':
    {
      const char *next, *eol;
      next = skipws(beg+1);

      switch (*next)
      {
        case '\0':
          beg = next;
          break;

        case '\n':
          beg = next+1;
          break;

        case '#':
          eol = strchr(next, '\n');
          if (eol) {
            beg = eol+1;
          } else {
            beg += strlen(beg);
          }
          break;

        default:
          return '\\';
      }
      beg = skipws(beg);
      goto restart;
    }

  case '#':
    {
      const char *eol = strchr(beg,'\n');
      if (eol) {
        beg = eol;
      } else {
        beg += strlen(beg);
      }
      goto restart;
    }

  default:
    if (isdigit(*beg)) {
      return lex_number(lx, beg);
    }
    if (is_id_begin(*beg)) {
      return lex_id(lx, beg);
    }

    goto invalid_char;
  }

invalid_char:
  lex_set_vbuf_char(lx, *beg);
  return lex_set_error(lx, beg, "invalid character");
}

