#include "tsch_config.h"
#include "tsch_config_parse.h"

#include <setjmp.h>
#include <ctype.h>

#include "tsch_config_priv.h"
#include "ts_errors.h"

/* Various debugging flags.  These cause the expression parser to dump
 * lots of stuff to stderr
 */
#define DEBUG_FUNC_CALL 0
#define DEBUG_READ_EXPR 0

#define PARSE_ERROR(p, expected, tok) do { \
  TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR, \
      "Expected %s but found unexpected token %d '%s' at %zu, %zu", \
      (expected), (tok), tsch_cfg_lex_tok(&p->lex), \
      tsch_cfg_lex_line(&p->lex), tsch_cfg_lex_col(&p->lex)); \
  } while(0)

void
tsch_parse_set_error_va(struct tsch_config_parser *p, const char *fmt, va_list args)
{
  vsnprintf(p->errbuf,sizeof p->errbuf, fmt, args);
}

void
tsch_parse_set_error(struct tsch_config_parser *p, const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  tsch_parse_set_error_va(p,fmt,args);
  va_end(args);
}

enum {
  V_NUM  = 1,
  V_BOOL = 2,
  V_STR  = 3,
};

static const char *value_type_name(int t);
static void print_parser_state(const struct tsch_config_parser *p);

const struct {
  const char *name;
  double value;
} global_constants[] = {
  { "PI", M_PI },
  { "E",  M_E  },

  { NULL, 0.0 },  /* end sentinel */
};

TSResult func_dbl_func(
    struct tsch_config_parser *p,
    const char *name,
    double (*f)(double),
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp)
{
  double v;

  UNUSED_PARAM(p);

#if DEBUG_FUNC_CALL
  fprintf(stderr, "function '%s'  argc=%zu, stack offset = %td, output offset = %td\n",
      name, argc, argv - p->vstack.values, outp - p->vstack.values);
  fprintf(stderr, "state:\n");
  print_parser_state(p);
#endif /* DEBUG_FUNC_CALL */

  if (argc != 1)
  {
    TS_RAISE_ERRORF(TS_ERR_WRONG_NUMBER_OF_ARGUMENTS,
        "expected one argument to %s but %zu given", name, argc);
  }

  if (argv[0].type != V_NUM)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "argument to %s should be a number (%s given)",
        name, value_type_name(argv[0].type));
  }

  /* in case outp aliases the arguments, capture the value separately */
  v = f(argv[0].num);

#if DEBUG_FUNC_CALL
  fprintf(stderr, "args: %f\n", argv[0].num);
  fprintf(stderr, "result: %f\n", v);
#endif /* DEBUG_FUNC_CALL */

  outp->type = V_NUM;
  outp->num = v;

  return TS_OK;
}

TSResult func_pow(struct tsch_config_parser *p,
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp)
{
  double v;

  UNUSED_PARAM(p);

#if DEBUG_FUNC_CALL
  fprintf(stderr, "function 'pow'  argc=%zu, stack offset = %td, output offset = %td\n",
      argc, argv - p->vstack.values, outp - p->vstack.values);
  fprintf(stderr, "state:\n");
  print_parser_state(p);
#endif /* DEBUG_FUNC_CALL */

  if (argc != 2)
  {
    TS_RAISE_ERRORF(TS_ERR_WRONG_NUMBER_OF_ARGUMENTS,
        "expected two arguments to pow but %zu given", argc);
  }

  if (argv[0].type != V_NUM || argv[1].type != V_NUM)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "arguments to pow should be numbers (%s and %s given)",
        value_type_name(argv[0].type),
        value_type_name(argv[1].type));
  }

  /* in case outp aliases the arguments, capture the value separately */
  v = pow(argv[0].num, argv[1].num);

#if DEBUG_FUNC_CALL
  fprintf(stderr, "args: %f %f\n", argv[0].num, argv[1].num);
  fprintf(stderr, "result: %f\n", v);
#endif /* DEBUG_FUNC_CALL */

  outp->type = V_NUM;
  outp->num = v;

  return TS_OK;
}

static TSResult func_reduce(struct tsch_config_parser *p,
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp,
    double (*reduce)(double,double),
    const char *fname)
{
  double val;
  size_t i;

  UNUSED_PARAM(p);

  if (argc == 0)
  {
    TS_RAISE_ERRORF(TS_ERR_WRONG_NUMBER_OF_ARGUMENTS,
        "expected at least one argument to %s(), none given", fname);
  }

  if (argv[0].type != V_NUM)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
        "arguments to %s() should be numbers (arg 1 is %s)",
        fname,value_type_name(argv[0].type));
  }

  val = argv[0].num;
  for (i=1; i < argc; i++)
  {
    if (argv[i].type != V_NUM)
    {
      TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT,
          "arguments to %s() should be numbers (arg %zu is %s)",
          fname, i+1,value_type_name(argv[i].type));
    }

    val = reduce(val,argv[i].num);
  }

  outp->type = V_NUM;
  outp->num = val;

  return TS_OK;
}


static double reduce_min(double x, double y)
{
  return x <= y ? x : y;
}

static double reduce_max(double x, double y)
{
  return x >= y ? x : y;
}

static TSResult func_min(struct tsch_config_parser *p,
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp)
{
  RETURN_WITH_RESULT(
      func_reduce(p, argc, argv, outp, reduce_min, "min")
      );
}

static TSResult func_max(struct tsch_config_parser *p,
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp)
{
  RETURN_WITH_RESULT(
      func_reduce(p, argc, argv, outp, reduce_max, "max")
      );
}

#define FUNC_DBL_WRAP(name_,f_)                                  \
  TSResult name_ (struct tsch_config_parser *p,                  \
      size_t argc, const struct tsch_config_expr_value *argv,    \
      struct tsch_config_expr_value *outp)                       \
  { return func_dbl_func(p, SYMSTR(f_), f_, argc, argv, outp); } \
  DEFINITION_EAT_SEMI 

FUNC_DBL_WRAP( func_exp, exp );
FUNC_DBL_WRAP( func_sin, sin );
FUNC_DBL_WRAP( func_cos, cos );
FUNC_DBL_WRAP( func_log, log );
FUNC_DBL_WRAP( func_log10, log10 );

typedef TSResult (*expr_func_t)(
    struct tsch_config_parser *p,
    size_t argc,
    const struct tsch_config_expr_value *argv,
    struct tsch_config_expr_value *outp);

const struct {
  const char *name;
  expr_func_t func;
} builtin_funcs[] = {
  { "exp", func_exp },
  { "pow", func_pow },
  { "sin", func_sin },
  { "cos", func_cos },
  { "log", func_log },
  { "min", func_min },
  { "max", func_max },

  { NULL, NULL } /* end sentinel */
};

/* adds constants to globals table */
static TSResult
add_global_consts(struct tsch_config_parser *p)
{
  size_t i;
  for (i=0; global_constants[i].name != NULL; i++)
  {
    struct tsch_string result;
    struct tsch_config *v;

    if (v = tsch_config_num(p, global_constants[i].value), v == NULL)
    {
      TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
          "Could not allocate a value for constant '%s' (%f)",
          global_constants[i].name, global_constants[i].value);
    }

    result = tsch_string_table_add(&p->globals, global_constants[i].name, v);
    if (result.s == NULL)
    {
      TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
          "Could not add constant '%s' to the globals table",
          global_constants[i].name);
    }
  }

  return TS_OK;
}

static TSResult
parsing_initialize(struct tsch_config_parser *p, const char *buf)
{
  RETURN_ON_ERROR(tsch_cfg_lex_initialize(&p->lex,buf));
  return TS_OK;
}

static void
parsing_finalize(struct tsch_config_parser *p)
{
  UNUSED_PARAM(p);
}

static inline int
next_token(struct tsch_config_parser *p)
{
  return tsch_config_lex(&p->lex);
}

static TSResult
new_section(struct tsch_config_parser *p, const struct tsch_cfg_id_path *path)
{
  const char *name = path->id.sym.s;
  struct tsch_config *map;

  if (map = tsch_config_map(p, name, NULL), map == NULL) {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "Cannot allocate memory for a configuration section", 0);
  }

  if (p->section_next == NULL) {
    p->section_next = p->kvtail;
  }

  tsch_config_parse_set_flag(p, TSCH_CFG_PARSE_EXPORT_SECTION, path->id.exported);
  p->section = map;
  if (path->parent == NULL) {
    *(p->section_next) = map;
    p->section_next = &map->next;
    tsch_string_table_add(&p->globals, name, map);
  } else {
    map->next = path->parent->val;
    path->parent->val = map;
  }
  p->kvtail = &map->val;
  return TS_OK;
}

static struct tsch_config *find_in_map(struct tsch_config *map, struct tsch_string key)
{
  struct tsch_config *entry;

  if (!map || map->type != TSCH_CFG_MAPPING) {
    return NULL;
  }

  for (entry = map->val; entry; entry = entry->next) {
    if (strcmp(entry->key, key.s) == 0) {
      return entry;
    }
  }

  return NULL;
}

static struct tsch_config *
find_submap(
  struct tsch_config_parser *p,
  struct tsch_config *map,
  struct tsch_cfg_id id)
{
  struct tsch_config *submap;
  size_t bucket;

  assert(id.sym.s != NULL);

  if (map) {
    return find_in_map(map, id.sym);
  }

  bucket = tsch_string_table_lookup(&p->globals, id.sym);
  if (bucket > 0) {
    return p->globals.values[bucket-1];
  }

  return NULL;
}

static TSResult
read_section(struct tsch_config_parser *p)
{
  int tok;
  bool shared = false;
  struct tsch_cfg_id sec;
  struct tsch_cfg_id_path path = { 0 };

  /* FIXME: this needs a serious refactoring */

  tok = next_token(p);
  switch (tok) {
    case KW_SHARED:
      shared = true;
      break;

    // case KW_CHECKS:
    case TOK_ID:
      path.parent = NULL;
      path.id = tsch_config_make_id(p,tsch_cfg_lex_tok(&p->lex));
      break;

    case '/':
      {
        tok = next_token(p);
        if (tok != ']')
        {
          PARSE_ERROR(p, "']'", tok);
        }
        if (p->section_next)
        {
          p->kvtail = p->section_next;
          p->section_next = NULL;
          p->section = p->head;
        }
        return TS_OK;
      }

    default:
      PARSE_ERROR(p, "section name or keyword", tok);
  }

  for(;;) {
    switch (tok = next_token(p), tok) {
      case ']':
        goto end_section;

      case KW_SHARED:
        shared = true;
        break;

      /* case KW_CHECKS: */

      case '/':
        {
          struct tsch_cfg_id_path newpath;
          struct tsch_config *parent;
          struct tsch_cfg_id id;

          tok = next_token(p);
          if (tok != TOK_ID) {
            PARSE_ERROR(p, "section name", tok);
          }

          parent = find_submap(p,path.parent,path.id);
          if (parent == NULL) {
            if (path.parent) {
              TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                  "cannot find map %s under map %s",
                  path.parent->key, path.id.sym.s);
            } else {
              TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                  "cannot find map %s", path.id.sym.s);
            }
          }
          path.parent = parent;
          path.id = tsch_config_make_id(p, tsch_cfg_lex_tok(&p->lex));
        }
        break;

      default:
        PARSE_ERROR(p, "section path or keyword", tok);
    }
  }

end_section:
  {
    struct tsch_config *map;

    map = find_submap(p,path.parent,path.id);
    if (map != NULL)
    {
      p->kvtail = tsch_config_tail(map->val);
      p->section = map;
    }
    else
    {
      path.id.exported = shared;
      RETURN_ON_ERROR(new_section(p, &path));
    }
  }

  return TS_OK;
}

static void
add_kv_pair(struct tsch_config_parser *p, struct tsch_config *kv)
{
  assert(p->kvtail  != NULL);

  *(p->kvtail) = kv;
  p->kvtail = &kv->next;

  if (p->section->key == NULL || tsch_config_parse_test_flag(p,TSCH_CFG_PARSE_EXPORT_SECTION))
  {
    tsch_string_table_add(&p->globals, kv->key, kv);
  }
}

enum {
  OP_NOP = 0,

  VALUE_NUM  = 1,
  VALUE_BOOL,
  VALUE_STR,

  OP_ADD,
  OP_SUB,
  OP_MUL,
  OP_DIV,
  OP_UMINUS,
  OP_UNEGATE,

  OP_MIN,
  OP_MAX,

  OP_LT,
  OP_LE,
  OP_GT,
  OP_GE,
  OP_EQ,
  OP_NE,

  OP_CAT,

  OP_IF,
  OP_IFELSE,

};

/* BNF-ish notation for expressions:
 *
 * expr :: comma_separated_exprs
 *
 * comma_separated_exprs :: comma_separated_exprs ',' expr_scalar
 *                        | expr_scalar
 *
 * expr_scalar :: expr_logic_or
 *
 * expr_logic_or :: expr_compare TOK_AND expr_compare
 *                | expr_compare
 *
 * expr_logic_and :: expr_compare TOK_AND expr_compare
 *                 | expr_compare
 *
 * expr_compare :: expr_minmax LT expr_minmax
 *               | expr_minmax LE expr_minmax
 *               | expr_minmax GT expr_minmax
 *               | expr_minmax GE expr_minmax
 *               | expr_minmax EQ expr_minmax
 *               | expr_minmax NE expr_minmax
 *               | expr_minmax
 *
 * expr_minmax :: expr_minmax TOK_MIN expr_term
 *              | expr_minmax TOK_MAX expr_term
 *              | expr_term
 *
 * expr_term   :: expr_term '+' expr_factor
 *              | expr_term '-' expr_factor
 *              | expr_factor
 *
 * expr_factor :: expr_factor '*' expr_unary
 *              | expr_factor '/' expr_unary
 *              | expr_unary
 *
 * expr_unary  :: '-' expr_simple
 *              | '~' expr_simple
 *              | expr_simple
 *
 * expr_simple :: TOK_NUM
 *              | '(' expr_scalar ')'
 *              | expr_id_or_function
 *
 * expr_id_or_function :: id '(' comma_separated_exprs ')'
 *                      | id
 *                      | '[' '[' id_path ']' ']'
 *
 */

enum EXPR_STATES {
  ST_EXPR = 1,
  ST_EXPR_SCALAR,
  ST_EXPR_LOGICAL_OR_1,
  ST_EXPR_LOGICAL_OR_2,
  ST_EXPR_LOGICAL_AND_1,
  ST_EXPR_LOGICAL_AND_2,
  ST_EXPR_COMPARE1,
  ST_EXPR_COMPARE2,
  ST_EXPR_FACTOR1,
  ST_EXPR_FACTOR2,
  ST_EXPR_MINMAX1,
  ST_EXPR_MINMAX2,
  ST_EXPR_SIMPLE,
  ST_EXPR_TERM1,
  ST_EXPR_TERM2,
  ST_EXPR_UNARY,
  ST_OP_AND,
  ST_OP_OR,
  ST_OP_LT,
  ST_OP_LE,
  ST_OP_GT,
  ST_OP_GE,
  ST_OP_EQ,
  ST_OP_NE,
  ST_OP_MIN,
  ST_OP_MAX,
  ST_OP_ADD,
  ST_OP_SUB,
  ST_OP_MUL,
  ST_OP_DIV,
  ST_OP_UMINUS,
  ST_OP_UNEGATE,
  ST_PARSE_ID,
  ST_PARENS,
  ST_COMMA_SEPARATED_EXPRS,
  ST_COMMA_SEPARATED_EXPRS_NEXT,
  ST_FUNC_CALL,
  ST_FINISH,
};

static inline const char *
expr_state_name(int state)
{
  switch (tsch_config_decode_state(state))
  {
    case ST_EXPR:               return "EXPR";
    case ST_EXPR_SCALAR:        return "EXPR_SCALAR";
    case ST_EXPR_LOGICAL_AND_1: return "EXPR_LOGICAL_AND";
    case ST_EXPR_LOGICAL_AND_2: return "EXPR_LOGICAL_AND";
    case ST_EXPR_LOGICAL_OR_1:  return "EXPR_LOGICAL_OR";
    case ST_EXPR_LOGICAL_OR_2:  return "EXPR_LOGICAL_OR";
    case ST_EXPR_COMPARE1:      return "EXPR_COMPARE1";
    case ST_EXPR_COMPARE2:      return "EXPR_COMPARE2";
    case ST_EXPR_FACTOR1:       return "EXPR_FACTOR1";
    case ST_EXPR_FACTOR2:       return "EXPR_FACTOR2";
    case ST_EXPR_MINMAX1:       return "EXPR_MINMAX1";
    case ST_EXPR_MINMAX2:       return "EXPR_MINMAX2";
    case ST_EXPR_SIMPLE:        return "EXPR_SIMPLE";
    case ST_EXPR_TERM1:         return "EXPR_TERM1";
    case ST_EXPR_TERM2:         return "EXPR_TERM2";
    case ST_EXPR_UNARY:         return "EXPR_UNARY";
    case ST_OP_AND:             return "OP_AND";
    case ST_OP_OR:              return "OP_OR";
    case ST_OP_LT:              return "OP_LT";
    case ST_OP_LE:              return "OP_LE";
    case ST_OP_GT:              return "OP_GT";
    case ST_OP_GE:              return "OP_GE";
    case ST_OP_EQ:              return "OP_EQ";
    case ST_OP_NE:              return "OP_NE";
    case ST_OP_MIN:             return "OP_MIN";
    case ST_OP_MAX:             return "OP_MAX";
    case ST_OP_ADD:             return "OP_ADD";
    case ST_OP_SUB:             return "OP_SUB";
    case ST_OP_MUL:             return "OP_MUL";
    case ST_OP_DIV:             return "OP_DIV";
    case ST_OP_UMINUS:          return "OP_UMINUS";
    case ST_OP_UNEGATE:         return "OP_UNEGATE";
    case ST_PARSE_ID:           return "PARSE_ID";
    case ST_PARENS:             return "PARENS";
    case ST_COMMA_SEPARATED_EXPRS:     return "COMMA_SEPARATED_EXPRS";
    case ST_COMMA_SEPARATED_EXPRS_NEXT: return "COMMA_SEPARATED_EXPRS_NEXT";
    case ST_FUNC_CALL:          return "FUNC_CALL";
    case ST_FINISH:             return "FINISH";
    default:                    return "<unknown>";
  }
}

#define SSTACK_SIZE(p) ((sizeof (p)->sstack.states) / (sizeof (p)->sstack.states[0]))
#define VSTACK_SIZE(p) ((sizeof (p)->vstack.values) / (sizeof (p)->vstack.values[0]))

static inline void print_state_stack(struct tsch_config_parser *p)
{
  size_t i;
  fprintf(stderr, "state stack:\n");
  for (i=p->sstack.top; i > 0; i--) {
    int st = p->sstack.states[i-1];
    fprintf(stderr, "[%3zu] %3d  %s\n",
        i,st, expr_state_name(st));
  }
  fprintf(stderr, "\n");
}

static void print_parser_state(const struct tsch_config_parser *p)
{
  size_t i;
  fprintf(stderr, "location: %zu %zu %zu | %zu %zu %zu\n'%s'\n'%s'\n",
      p->lex.state[0].pos, p->lex.state[0].line, p->lex.state[0].line0,
      p->lex.state[1].pos, p->lex.state[1].line, p->lex.state[1].line0,
      p->lex.buf + p->lex.state[0].pos,
      p->lex.buf + p->lex.state[1].pos);
  fprintf(stderr, "state stack:\n");
  for (i=p->sstack.top; i > 0; i--) {
    int st = p->sstack.states[i-1];
    fprintf(stderr, "[%3zu] %3d  %s\n",
        i,st, expr_state_name(st));
  }
  fprintf(stderr, "\nvalue stack:\n");
  for (i=p->vstack.top; i > 0; i--) {
    fprintf(stderr, "[%3zu] ", i);
    switch (p->vstack.values[i-1].type) {
      case V_BOOL:
        fprintf(stderr, "BOOL %s\n", p->vstack.values[i-1].b ? "true" : "false");
        break;

      case V_STR:
        fprintf(stderr, "STR  %s\n", p->vstack.values[i-1].s);
        break;

      case V_NUM:
        fprintf(stderr, "NUM  %f\n", p->vstack.values[i-1].num);
        break;

      default:
        fprintf(stderr, "unknown %d\n", p->vstack.values[i-1].type);
        break;
    }
  }
  fprintf(stderr, "\n");
}

static inline TSResult expr_push_state(struct tsch_config_parser *p, unsigned int st)
{
  if (p->sstack.top >= SSTACK_SIZE(p)) {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "State stack size exceeded while parsing expression", 0);
  }

  p->sstack.states[p->sstack.top++] = st;

  /*
  fprintf(stderr, "PUSH_STATE(%s):\n", expr_state_name(st));
  print_state_stack(p);
  */

  return TS_OK;
}

static inline TSResult expr_pop_state(struct tsch_config_parser *p, unsigned int *stp)
{
  if (p->sstack.top == 0) {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "Cannot pop next state off the state stack: stack is empty", 0);
  }

  *stp = p->sstack.states[--p->sstack.top];

  /*
  fprintf(stderr, "POP_STATE() = %s\n", expr_state_name(*stp));
  print_state_stack(p);
  */

  return TS_OK;
}

static inline TSResult expr_push_value(struct tsch_config_parser *p,
    const struct tsch_config_expr_value val)
{
  if (p->vstack.top >= VSTACK_SIZE(p)) {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "Value stack size exceeded while parsing expression", 0);
  }
  p->vstack.values[p->vstack.top++] = val;
  return TS_OK;
}

static inline TSResult expr_push_bool(struct tsch_config_parser *p, bool b)
{
  struct tsch_config_expr_value v;
  v.type = V_BOOL;
  v.b = b;

  RETURN_WITH_RESULT(expr_push_value(p, v));
}

static inline TSResult expr_push_num(struct tsch_config_parser *p, double n)
{
  struct tsch_config_expr_value v;
  v.type = V_NUM;
  v.num = n;

  RETURN_WITH_RESULT(expr_push_value(p, v));
}

static inline TSResult expr_push_str(struct tsch_config_parser *p, const char *s)
{
  struct tsch_config_expr_value v;
  v.type = V_STR;
  v.s = s;

  RETURN_WITH_RESULT(expr_push_value(p, v));
}

static const char *
value_type_name(int t) {
  switch (t) {
    case V_BOOL: return "bool";
    case V_NUM: return "number";
    case V_STR: return "string";
    default: return "unknown";
  }
}

static inline TSResult
check_value_stack(struct tsch_config_parser *p, unsigned int type)
{
  size_t ind;

  if (p->vstack.top == 0) {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "Cannot pop a %s off the value stack: stack is empty",
        value_type_name(type));
  }

  ind = p->vstack.top-1;
  if (p->vstack.values[ind].type != type) {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "Expected %s at top of stack, but found %s",
        value_type_name(type),
        value_type_name(p->vstack.values[ind].type));
  }

  return  TS_OK;
}

static inline TSResult expr_pop_bool(struct tsch_config_parser *p, bool *bp)
{
  RETURN_ON_ERROR(check_value_stack(p,V_BOOL));
  *bp = p->vstack.values[--p->vstack.top].b;
  return TS_OK;
}

static inline TSResult expr_pop_num(struct tsch_config_parser *p, double *nump)
{
  RETURN_ON_ERROR(check_value_stack(p,V_NUM));
  *nump = p->vstack.values[--p->vstack.top].num;
  return TS_OK;
}

static inline TSResult expr_pop_str(struct tsch_config_parser *p, const char **sp)
{
  RETURN_ON_ERROR(check_value_stack(p,V_STR));
  *sp = p->vstack.values[--p->vstack.top].s;
  return TS_OK;
}

static bool
config_to_expr_value(
    const struct tsch_config *c,
    struct tsch_config_expr_value *val)
{
  switch (c->type)
  {
    case TSCH_CFG_STRING:
      val->type = V_STR;
      val->s = c->s;
      return true;

    case TSCH_CFG_NUM:
      val->type = V_NUM;
      val->num = c->num;
      return true;

    case TSCH_CFG_BOOL:
      val->type = V_BOOL;
      val->b = c->b;
      return true;

    case TSCH_CFG_NONE:
    case TSCH_CFG_MAPPING:
    case TSCH_CFG_SEQUENCE:
    default:
      return false;
  }
}

enum { FOUND = 0, E_MISSING = -1, E_TYPE = -2 };
static int expr_lookup_in_map(
    struct tsch_string key,
    struct tsch_config *map, 
    struct tsch_config_expr_value *valp)
{
  struct tsch_config *entry;

  assert(map && map->type == TSCH_CFG_MAPPING);
  assert(valp != NULL);

  entry = find_in_map(map, key);
  if (!entry) {
    return E_MISSING;
  }

  if (!config_to_expr_value(entry, valp)) {
    return E_TYPE;
  }

  return FOUND;
}

static expr_func_t expr_lookup_func(struct tsch_config_parser *p, const char *name)
{
  size_t i;

  UNUSED_PARAM(p);  /* future: parsers may have their own function tables */

  for (i=0; builtin_funcs[i].name != NULL; i++)
  {
    if (strcmp(name, builtin_funcs[i].name) == 0)
    {
      return builtin_funcs[i].func;
    }
  }

  return NULL;
}

struct tsch_config *
expr_value_to_config(struct tsch_config_parser *p, struct tsch_config_expr_value *v)
{
  switch (v->type)
  {
    case V_BOOL:
      return tsch_config_bool(p,v->b);

    case V_NUM:
      return tsch_config_num(p,v->num);

    case V_STR:
      return tsch_config_str(p,v->s);

    default:
      return NULL;
  }
}

static TSResult
expr_lookup(
    struct tsch_config_parser *p,
    struct tsch_string key, 
    struct tsch_config_expr_value *valp)
{
  int ret;
  size_t bucket;
  struct tsch_config *entry;
  TSResult err;

  switch (ret = expr_lookup_in_map(key,p->section,valp), ret) {
    case FOUND:
      return TS_OK;
    case E_MISSING:
      break;
    case E_TYPE:
      goto type_error;
    default:
      /* should not reach here */
      TS_RAISE_ERRORF(TS_ERR_INVALID_RETURN,
          "Unrecognized return %d during lookup while parsing config file",
          ret);
  }

  bucket = tsch_string_table_lookup(&p->globals, key);
  if (bucket == 0) {
    goto missing;
  }

  if (entry = p->globals.values[bucket-1], !entry) {
    goto missing;
  }

  if (!config_to_expr_value(entry, valp)) {
    goto type_error;
  }

  return TS_OK;

missing:
  TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
      "Could not find an identifier '%s'", key.s);

type_error:
  TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
    "Expected '%s' to be a scalar value", key.s);
}

static inline double mindbl(double x, double y) { return x <= y ? x : y; }
static inline double maxdbl(double x, double y) { return x >= y ? x : y; }

#define PUSH_STATE(st) RETURN_ON_ERROR(expr_push_state(p, (st)))
#define PUSH_VALUE_BOOL(b) RETURN_ON_ERROR(expr_push_bool(p,(b)))
#define PUSH_VALUE_NUM(b)  RETURN_ON_ERROR(expr_push_num(p,(b)))
#define PUSH_VALUE_STR(s)  RETURN_ON_ERROR(expr_push_str(p,(s)))
#define POP_VALUE_BOOL(v)  RETURN_ON_ERROR(expr_pop_bool(p, &(v)))
#define POP_VALUE_NUM(v)   RETURN_ON_ERROR(expr_pop_num(p, &(v)))
#define POP_VALUE_STR(v)   RETURN_ON_ERROR(expr_pop_str(p, &(v)))
#define GOTO_STATE(st) do { state=(st); goto next_state; } while(0)
#define NEXT_STATE()   do {                  \
  RETURN_ON_ERROR(expr_pop_state(p,&state)); \
  goto next_state;                           \
} while(0)

#define NEXT_TOKEN()  (tsch_config_lex(&p->lex))
#define UNGET_TOKEN(t)  (tsch_config_unlex(&p->lex))

static TSResult
read_expr(struct tsch_config_parser *p, struct tsch_config **cfgp)
{
  int tok;
  unsigned int state;

  PUSH_STATE(ST_FINISH);
  state = ST_EXPR; /* starting state */

#if DEBUG_READ_EXPR
  fprintf(stderr, "\n\nread_expr()\n");
#endif /* DEBUG_READ_EXPR */

  for(;;) {
next_state:

#if DEBUG_READ_EXPR
    {
      fprintf(stderr, "state == %s (%u)  arg: %u\n",
          expr_state_name(state),tsch_config_decode_state(state), tsch_config_decode_state_arg(state));
      print_parser_state(p);
    }
#else
    (void)print_parser_state;  /* silence unused warnings */
#endif /* DEBUG_READ_EXPR */

    switch(tsch_config_decode_state(state)) {
      case ST_EXPR:
        GOTO_STATE(ST_COMMA_SEPARATED_EXPRS);
        // GOTO_STATE(ST_EXPR_SCALAR);

      case ST_COMMA_SEPARATED_EXPRS:
        PUSH_STATE(ST_COMMA_SEPARATED_EXPRS_NEXT);
        GOTO_STATE(ST_EXPR_SCALAR);

      case ST_COMMA_SEPARATED_EXPRS_NEXT:
        tok = NEXT_TOKEN();
        if (DEBUG_READ_EXPR) {
          fprintf(stderr, "tok : %d (%c)\n",
              tok, tok >= CHAR_MIN && tok <= CHAR_MAX ? tok : 0);
        }
        if (tok == ',') {
          PUSH_STATE(ST_COMMA_SEPARATED_EXPRS_NEXT);
          GOTO_STATE(ST_EXPR_SCALAR);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_EXPR_SCALAR:
        GOTO_STATE(ST_EXPR_LOGICAL_OR_1);

      case ST_EXPR_LOGICAL_OR_1:
        PUSH_STATE(ST_EXPR_LOGICAL_OR_2);
        GOTO_STATE(ST_EXPR_LOGICAL_AND_1);

      case ST_EXPR_LOGICAL_OR_2:
        tok = NEXT_TOKEN();
        if (tok == TOK_OR) {
          PUSH_STATE(ST_OP_OR);
          GOTO_STATE(ST_EXPR_LOGICAL_AND_1);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_EXPR_LOGICAL_AND_1:
        PUSH_STATE(ST_EXPR_LOGICAL_AND_2);
        GOTO_STATE(ST_EXPR_COMPARE1);

      case ST_EXPR_LOGICAL_AND_2:
        tok = NEXT_TOKEN();
        if (tok == TOK_AND) {
          PUSH_STATE(ST_OP_AND);
          GOTO_STATE(ST_EXPR_COMPARE1);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_OP_OR:
        {
          bool b1,b2;
          b1=b2=false;
          POP_VALUE_BOOL(b2);
          POP_VALUE_BOOL(b1);
          PUSH_VALUE_BOOL(b1 || b2);
          GOTO_STATE(ST_EXPR_LOGICAL_OR_2);
        }

      case ST_OP_AND:
        {
          bool b1,b2;
          b1=b2=false;
          POP_VALUE_BOOL(b2);
          POP_VALUE_BOOL(b1);
          PUSH_VALUE_BOOL(b1 && b2);
          GOTO_STATE(ST_EXPR_LOGICAL_AND_2);
        }

      case ST_EXPR_COMPARE1:
        PUSH_STATE(ST_EXPR_COMPARE2);
        GOTO_STATE(ST_EXPR_MINMAX1);

      case ST_EXPR_COMPARE2:
        tok = NEXT_TOKEN();
        switch (tok) {
          case '<':    PUSH_STATE(ST_OP_LT); GOTO_STATE(ST_EXPR_MINMAX1);
          case '>':    PUSH_STATE(ST_OP_GT); GOTO_STATE(ST_EXPR_MINMAX1);
          case TOK_LE: PUSH_STATE(ST_OP_LE); GOTO_STATE(ST_EXPR_MINMAX1);
          case TOK_GE: PUSH_STATE(ST_OP_GE); GOTO_STATE(ST_EXPR_MINMAX1);
          case TOK_EQ: PUSH_STATE(ST_OP_EQ); GOTO_STATE(ST_EXPR_MINMAX1);
          case TOK_NE: PUSH_STATE(ST_OP_NE); GOTO_STATE(ST_EXPR_MINMAX1);
          default:
                       UNGET_TOKEN(tok);
                       NEXT_STATE();
        }

#define COMPARE_OP(cmp) do { double v1,v2; \
  v1=v2=-0.0;                              \
  POP_VALUE_NUM(v2); POP_VALUE_NUM(v1);    \
  PUSH_VALUE_BOOL(cmp(v1,v2));             \
  NEXT_STATE();                            \
} while (0)
#define LT(x,y) ((x) <  (y))
#define LE(x,y) ((x) <= (y))
#define GT(x,y) ((x) >  (y))
#define GE(x,y) ((x) >= (y))
#define EQ(x,y) ((x) == (y))
#define NE(x,y) ((x) != (y))

      case ST_OP_LT: COMPARE_OP(LT);
      case ST_OP_LE: COMPARE_OP(LE);
      case ST_OP_GT: COMPARE_OP(GT);
      case ST_OP_GE: COMPARE_OP(GE);
      case ST_OP_EQ: COMPARE_OP(EQ);
      case ST_OP_NE: COMPARE_OP(NE);

#undef COMPARE_OP
#undef LT
#undef LE
#undef GT
#undef GE
#undef EQ
#undef NE

#define NUM_OP(op,nst) do { double v1,v2; \
  v1=v2=0.0;                              \
  POP_VALUE_NUM(v2); POP_VALUE_NUM(v1);   \
  PUSH_VALUE_NUM(op(v1,v2));              \
  GOTO_STATE(nst);                        \
} while (0)
#define ADD(x,y) ((x) + (y))
#define SUB(x,y) ((x) - (y))
#define MUL(x,y) ((x) * (y))
#define DIV(x,y) ((x) / (y))

      case ST_OP_ADD: NUM_OP(ADD, ST_EXPR_TERM2);
      case ST_OP_SUB: NUM_OP(SUB, ST_EXPR_TERM2);
      case ST_OP_MUL: NUM_OP(MUL, ST_EXPR_FACTOR2);
      case ST_OP_DIV: NUM_OP(DIV, ST_EXPR_FACTOR2);
      case ST_OP_MIN: NUM_OP(mindbl, ST_EXPR_MINMAX2);
      case ST_OP_MAX: NUM_OP(maxdbl, ST_EXPR_MINMAX2);

#undef NUM_OP
#undef ADD
#undef SUB
#undef MUL
#undef DIV


      case ST_EXPR_MINMAX1:
        {
          PUSH_STATE(ST_EXPR_MINMAX2);
          GOTO_STATE(ST_EXPR_TERM1);
        }

      case ST_EXPR_MINMAX2:
        tok = NEXT_TOKEN();
        if (tok == TOK_MIN) {
          PUSH_STATE(ST_OP_MIN);
          GOTO_STATE(ST_EXPR_TERM1);
        }
        if (tok == TOK_MAX) {
          PUSH_STATE(ST_OP_MAX);
          GOTO_STATE(ST_EXPR_TERM1);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_EXPR_TERM1:
        PUSH_STATE(ST_EXPR_TERM2);
        GOTO_STATE(ST_EXPR_FACTOR1);

      case ST_EXPR_TERM2:
        tok = NEXT_TOKEN();
        if (DEBUG_READ_EXPR) {
          fprintf(stderr, "tok = %d (%c)\n", tok, isprint(tok) ? tok : ' ');
        }
        if (tok == '+') {
          PUSH_STATE(ST_OP_ADD);
          GOTO_STATE(ST_EXPR_FACTOR1);
        }
        if (tok == '-') {
          PUSH_STATE(ST_OP_SUB);
          GOTO_STATE(ST_EXPR_FACTOR1);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_EXPR_FACTOR1:
        PUSH_STATE(ST_EXPR_FACTOR2);
        GOTO_STATE(ST_EXPR_UNARY);

      case ST_EXPR_FACTOR2:
        tok = NEXT_TOKEN();
        if (tok == '*') {
          PUSH_STATE(ST_OP_MUL);
          GOTO_STATE(ST_EXPR_UNARY);
        }
        if (tok == '/') {
          PUSH_STATE(ST_OP_DIV);
          GOTO_STATE(ST_EXPR_UNARY);
        }
        UNGET_TOKEN(tok);
        NEXT_STATE();

      case ST_EXPR_UNARY:
        tok = NEXT_TOKEN();
        if (tok == '-') {
          PUSH_STATE(ST_OP_UMINUS);
          GOTO_STATE(ST_EXPR_SIMPLE);
        }
        if (tok == '!') {
          PUSH_STATE(ST_OP_UNEGATE);
          GOTO_STATE(ST_EXPR_SIMPLE);
        }
        UNGET_TOKEN(tok);
        GOTO_STATE(ST_EXPR_SIMPLE);

      case ST_OP_UMINUS:
        {
          double v = 0.0;
          POP_VALUE_NUM(v);
          PUSH_VALUE_NUM( -v );
          NEXT_STATE();
        }

      case ST_OP_UNEGATE:
        {
          bool b = false;
          POP_VALUE_BOOL(b);
          PUSH_VALUE_BOOL( !b );
          NEXT_STATE();
        }

      case ST_EXPR_SIMPLE:
        tok = NEXT_TOKEN();
        switch (tok) {
          case TOK_NUM:
            RETURN_ON_ERROR(expr_push_num(p,p->lex.tok_value.num));
            break;

          case TOK_STR:
            RETURN_ON_ERROR(expr_push_str(p,p->lex.tok_value.str));
            break;

          case KW_TRUE:
            RETURN_ON_ERROR(expr_push_bool(p,true));
            break;

          case KW_FALSE:
            RETURN_ON_ERROR(expr_push_bool(p,false));
            break;

          case '(':
            PUSH_STATE(ST_PARENS);
            GOTO_STATE(ST_EXPR_SCALAR);

          case TOK_ID:
            GOTO_STATE(ST_PARSE_ID);

          default:
            TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                "Invalid token %d ('%s')",
                tok, tsch_cfg_lex_tok(&p->lex));
        }

        NEXT_STATE();

      case ST_PARSE_ID:
        {
          int tok2;
          const char *sem;
          struct tsch_string id;
          struct tsch_config_expr_value val;

          /* token string buffer is overwritten after we fetch the next
           * token, so we have to save the id before we get the next
           * token...
           */
          sem = tsch_cfg_lex_tok(&p->lex);
          assert(sem != NULL);

          id = tsch_string_table_intern_str(&p->strings,sem);
          assert(id.s != NULL);

          tok2 = NEXT_TOKEN();
          if (tok2 == '(') {
            if (p->vstack.top > 0xff)
            {
              TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                  "Cannot parse function call: value stack size exceeds %d", 0xff);
            }

            PUSH_VALUE_STR( id.s );
            PUSH_STATE(tsch_config_encode_state(ST_FUNC_CALL,
                  (unsigned int)(p->vstack.top & 0xff)));
            tok2 = NEXT_TOKEN();
            if (tok2 == ')')
            {
              UNGET_TOKEN(tok2);
              NEXT_STATE();
            }
            UNGET_TOKEN(tok2);
            GOTO_STATE(ST_COMMA_SEPARATED_EXPRS);
          }
          UNGET_TOKEN(tok2);

          val = (struct tsch_config_expr_value) { 0 };
          RETURN_ON_ERROR(expr_lookup(p,id,&val));
          RETURN_ON_ERROR(expr_push_value(p, val));
          NEXT_STATE();
        }

      case ST_FUNC_CALL:
        {
          size_t arg_top, stk_top, nargs;
          const char *funcname;
          expr_func_t func;

          tok = NEXT_TOKEN();
          if (tok != ')')
          {
            TS_RAISE_ERROR_MESG(TS_ERR_PARSE_ERROR,
                "invalid stack state at function call!");
          }

          arg_top = tsch_config_decode_state_arg(state);
          stk_top = p->vstack.top;
          if (arg_top > stk_top)
          {
            TS_RAISE_ERROR_MESG(TS_ERR_PARSE_ERROR,
                "invalid stack state at function call!");
          }
          nargs = stk_top - arg_top;

          if (p->vstack.values[arg_top-1].type != V_STR)
          {
            TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                "stack does not contain a function name at element %zu!", arg_top);
          }

          funcname = p->vstack.values[arg_top-1].s;
          if (DEBUG_READ_EXPR)
          {
            fprintf(stderr, "FUNC_CALL(%s): popped %zu args\n", funcname, nargs);
          }

          if (func = expr_lookup_func(p, funcname), func == NULL)
          {
            TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                "unknown function '%s'", funcname);
          }

          RETURN_ON_ERROR(
              func(p, nargs, &p->vstack.values[arg_top],
                &p->vstack.values[arg_top-1]) );

          p->vstack.top = arg_top;
          NEXT_STATE();
        }

      case ST_PARENS:
        tok = NEXT_TOKEN();
        if (tok != ')') {
            TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
                "expected a ')', but found '%s'", tsch_cfg_lex_tok(&p->lex));
        }
        NEXT_STATE();

      case ST_FINISH:
        goto finish;
    }
  }

finish:
  if (p->sstack.top != 0)
  {
    TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
        "State stack is not empty at end of expression!", 0);
  }
  if (p->vstack.top < 1)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_PARSE_ERROR,
        "Value stack should contain at least one item, but no items found");
  }

  if (p->vstack.top == 1)
  {
    *cfgp = expr_value_to_config(p, &p->vstack.values[0]);
    if (*cfgp == NULL)
    {
      TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
          "Unknown value type %u for the expression",
          p->vstack.values[0].type);
    }
  }
  else
  {
    struct tsch_config *head, **curr;
    size_t i;

    head = tsch_config_list_empty(p);
    curr = &head->val;

    for (i=0; i < p->vstack.top; i++)
    {
      *curr = expr_value_to_config(p, &p->vstack.values[i]);
      if (*curr == NULL)
      {
        TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
            "Unknown value type %u for item %zu in the expression",
            p->vstack.values[i].type, i);
      }
      curr = &(*curr)->next;
    }
    *cfgp = head;
  }

  p->vstack.top = 0;
  return TS_OK;
}

static TSResult 
read_multiline_list(struct tsch_config_parser *p, struct tsch_config **nodep)
{
  struct tsch_config *list, **tail;
  list = tsch_config_list_empty(p);
  tail = &list->val;

  // fprintf(stderr, "\nmulti-line list\n");

  for (;;)
  {
    int tok;
    struct tsch_config *item;

    do {
      tok = next_token(p);
    } while (tok == '\n');

    if (tok == '}')
    {
      goto end_list;
    }

    if (!tsch_config_unlex(&p->lex))
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IN_ALGORITHM, "cannot unget token!");
    }

    RETURN_ON_ERROR(read_expr(p, &item));
    if (item->type == TSCH_CFG_SEQUENCE)
    {
      *tail = item->val;
      if (*tail) {
        tail = tsch_config_tail(item->val);
      }
    }
    else
    {
      *tail = item;
      tail = &item->next;
    }
  }

end_list:
  *nodep = list;
  return TS_OK;
}

static TSResult
include_file(struct tsch_config_parser *p, const char *fname)
{
  UNUSED_PARAM(p);
  UNUSED_PARAM(fname);

  return TS_OK;
}

static TSResult
read_directive(struct tsch_config_parser *p)
{
  int tok;
  struct tsch_cfg_id direc;
  TSResult ret;

  tok = next_token(p);
  if (tok != TOK_ID)
  {
    PARSE_ERROR(p, "directive", tok);
  }

  direc = tsch_config_make_id(p,tsch_cfg_lex_tok(&p->lex));
  /* Currently we have no directives */
  TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR,
      "unknown directive '%s'", direc.sym.s);

  return TS_OK;
}

static TSResult
read_kvpair(struct tsch_config_parser *p, struct tsch_cfg_id key)
{
  int tok;
  struct tsch_config *config;

  tok = next_token(p);
  if (tok != ':') {
    PARSE_ERROR(p, "colon (:)", tok);
  }

  config = NULL;
  tok = next_token(p);
  switch (tok)
  {
    case '{':
      /* multiline list */
      RETURN_ON_ERROR(read_multiline_list(p, &config));
      break;

    default:
      tsch_config_unlex(&p->lex);
      RETURN_ON_ERROR(read_expr(p, &config));
      break;
  }

  tok = next_token(p);
  if (tok != '\n' && tok != ';') {
    PARSE_ERROR(p, "newline or semicolon (;)", tok);
  }

  config->key = key.sym.s;
  add_kv_pair(p, config);

  return TS_OK;
}

static TSResult read_directive(struct tsch_config_parser *p);

static TSResult
parsing_parse(struct tsch_config_parser *p)
{
  for(;;) {
    int tok = tsch_config_lex(&p->lex);
    switch (tok) {
      case '[':
        RETURN_ON_ERROR(read_section(p));
        break;

      case TOK_ID:
        {
          struct tsch_cfg_id id;
          id = tsch_config_make_id(p, p->lex.tok_value.str);
          RETURN_ON_ERROR(read_kvpair(p,id));
        }
        break;

      case '%':
        RETURN_ON_ERROR(read_directive(p));
        break;

      case '\n':
        break;

      case EOF:
        goto finished;

      default:
       TS_RAISE_ERRORF(TS_ERR_PARSE_ERROR, 
           "Unexpected symbol %s encountered on line %zu at col %zu",
           tsch_cfg_lex_tok(&p->lex),
           tsch_cfg_lex_line(&p->lex),
           tsch_cfg_lex_col(&p->lex));
    }
  }

finished:
  return TS_OK;
}

TSResult
tsch_parse_config(
    struct tsch_config_parser *p,
    struct tsch_config **rootp, 
    const char *buf)
{
  TSResult result;

  result = parsing_initialize(p,buf), ON_ERROR_ADD_FRAME(result);
  if (IS_ERROR(result)) {
    goto finish;
  }

  result = parsing_parse(p), ON_ERROR_ADD_FRAME(result);
  if (!IS_ERROR(result)) {
    tsch_cfg_lex_finalize(&p->lex);
    *rootp = p->head;
  }

finish:
  parsing_finalize(p);
  return result;
}

TSResult tsch_config_add_parameter(struct tsch_config_parser *p,
    const char *param, double val)
{
  struct tsch_string pkey;
  struct tsch_config *pval;

  assert(param != NULL && "param != NULL");

  /* Enforce parameter prefix... Not sure if this is necessary/useful.
   */
  if (param[0] != TSCH_CFG_PARAM_PREFIX)
  {
    TS_RAISE_ERRORF(TS_ERR_INVALID_PARAM_NAME,
        "invalid parameter name '%s': parameter names must start with '%c'",
        param, TSCH_CFG_PARAM_PREFIX);
  }

  if (pval = tsch_config_num(p, val), !pval)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "out of memory allocating configuration parameter '%s'",
        param);
  }

  pkey = tsch_string_table_add(&p->globals, param, pval);
  if (pkey.s == NULL)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "out of memory adding configuration parameter '%s'",
        param);
  }

  return TS_OK;
}

static TSResult config_initialize(struct tsch_config_parser *p)
{
  struct tsch_config *root = NULL;

  p->pool = tsch_config_pool_alloc(NULL,TSCH_CFG_POOL_SIZE);
  if (!p->pool) {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "cannot allocate configuration item pool",0);
  }

  RETURN_ON_ERROR(
      tsch_string_table_initialize(&p->strings,
        TSCH_STRINGS_DEFAULT_CAPA, false)
      );

  RETURN_ON_ERROR(tsch_string_table_initialize(
        &p->globals, TSCH_STRINGS_DEFAULT_CAPA, true));
  RETURN_ON_ERROR(add_global_consts(p));

  if (root = tsch_config_map(p, NULL, NULL), !root)
  {
    TS_RAISE_ERRORF(TS_ERR_OUT_OF_MEMORY,
        "Could not allocate global configuration map", 0);
  }

  p->section = p->head = root;
  p->section_next = NULL;
  p->kvtail = &p->head->val;

  return TS_OK;
}

TSResult tsch_config_initialize(struct tsch_config_parser *p)
{
  TSResult err;

  *p = (struct tsch_config_parser){ {0} };

  err = config_initialize(p), ON_ERROR_ADD_FRAME(err);
  if (IS_ERROR(err))
  {
    tsch_config_finalize(p);
  }

  return err;
}

void tsch_config_finalize(struct tsch_config_parser *p)
{
  if (p->lex.buf)
  {
    tsch_cfg_lex_finalize(&p->lex);
  }

  p->section = p->head = NULL;
  p->section_next = NULL;
  p->kvtail = NULL;

  if (p->globals.keys)
  {
    tsch_string_table_finalize(&p->globals);
  }

  if (p->strings.keys)
  {
    tsch_string_table_finalize(&p->strings);
  }

  if (p->pool)
  {
    tsch_config_pool_dealloc(p->pool);
  }
}

