#ifndef TSCH_CONFIG_PARSE_H
#define TSCH_CONFIG_PARSE_H

#include "tsch_config.h"

#define MODULE_NAME TSCH_CONFIG_PARSE

struct tsch_cfg_lex {
  const char *buf;
  struct {
    size_t pos;
    size_t line;
    size_t line0;
  } state[2];

  struct {
    size_t line, col;
  } pos0,pos1;

  int tok;
  union {
    const char *str;
    double num;
  } tok_value;

  char ebuf[128];
  char vbuf[8192];
};

static inline size_t 
tsch_cfg_lex_line(const struct tsch_cfg_lex *lex)
{
  return lex->state[1].line;
}

static inline size_t 
tsch_cfg_lex_col(const struct tsch_cfg_lex *lex)
{
  size_t pos, line0;
  pos = lex->state[1].pos;
  line0 = lex->state[1].line0;
  return pos >= line0 ? pos - line0 : 0;
}

static inline const char *
tsch_cfg_lex_tok(const struct tsch_cfg_lex *lex)
{
  return lex->vbuf;
}

static inline const char *
tsch_cfg_lex_current_line(const struct tsch_cfg_lex *lex)
{
  return lex->buf + lex->state[1].pos;
}

struct tsch_config_expr_value
{
  unsigned int type;
  union {
    double num;
    bool b;
    const char *s;
  };
};

static inline size_t 
tsch_cfg_lex_last_lineno(const struct tsch_cfg_lex *lex)
{
  return lex->state[0].line;
}

static inline size_t 
tsch_cfg_lex_last_col(const struct tsch_cfg_lex *lex)
{
  size_t pos, line0;
  pos = lex->state[0].pos;
  line0 = lex->state[0].line0;
  return pos >= line0 ? pos - line0 : 0;
}

static inline const char *
tsch_cfg_lex_last_line(const struct tsch_cfg_lex *lex)
{
  return lex->buf + lex->state[0].line0;
}

/* prefix in front of config identifiers that indicates that the
 * identifier is an external parameter
 */
enum { TSCH_CFG_PARAM_PREFIX = '$' };

enum {
  TSCH_CFG_PARSE_EXPORT_SECTION = 0x0001,
};

enum { TSCH_CFG_SSTACKSIZE = 128, TSCH_CFG_VSTACKSIZE = 32 };
struct tsch_config_parser {
  struct tsch_cfg_lex lex;
  struct tsch_config_pool *pool;
  struct tsch_string_table strings;
  struct tsch_string_table globals;

  struct tsch_config *head;
  struct tsch_config **kvtail;
  struct tsch_config *section;
  struct tsch_config **section_next;
  unsigned int flags;

  struct {
    size_t top;
    unsigned int states[TSCH_CFG_SSTACKSIZE];
  } sstack; /* state stack */

  struct {
    size_t top;
    struct tsch_config_expr_value values[TSCH_CFG_VSTACKSIZE];
  } vstack; /* value stack */

  char errbuf[1024];
};

#define TSCH_CONFIG_PARSER_INIT \
  ((struct tsch_config_parser) { .lex = { 0 }, 0 })

static inline unsigned int tsch_config_decode_state(unsigned int state)
{
  return state & 0xff;
}

static inline unsigned int tsch_config_decode_state_arg(unsigned int state)
{
  return (state >> 8) & 0xff;
}

static inline unsigned int
tsch_config_encode_state(unsigned int state, unsigned int arg)
{
  return (state) | (arg << 8);
}

static inline void
tsch_config_parse_set_flag(struct tsch_config_parser *p, unsigned int flag, bool set)
{
  p->flags = set ? (p->flags | flag) : (p->flags & ~flag);
}

static inline bool
tsch_config_parse_test_flag(struct tsch_config_parser *p, unsigned int flag)
{
  return p->flags & flag;
}

TSResult tsch_config_initialize(struct tsch_config_parser *p);
void tsch_config_finalize(struct tsch_config_parser *p);

/* TODO: Extend parameters to other types.  Currently parameters are
 * limited to numbers */
TSResult tsch_config_add_parameter(struct tsch_config_parser *p,
    const char *param, double val);

TSResult tsch_parse_config(struct tsch_config_parser *p,
    struct tsch_config **configp, const char *buf);

void tsch_parse_set_error_va(struct tsch_config_parser *p, const char *fmt, va_list args);
void tsch_parse_set_error(struct tsch_config_parser *p, const char *fmt, ...);

TSResult tsch_cfg_lex_initialize(struct tsch_cfg_lex *lx, const char *s);
void tsch_cfg_lex_finalize(struct tsch_cfg_lex *lx);
int tsch_config_lex(struct tsch_cfg_lex *lx);

/* tries to unget a token */
static inline bool
tsch_config_unlex(struct tsch_cfg_lex *lx)
{
  if (lx->state[0].pos == lx->state[1].pos) {
    return false;
  }
  lx->state[1] = lx->state[0];
  return true;
}

static inline struct tsch_config *
tsch_config_bool(struct tsch_config_parser *p, bool tf)
{
  struct tsch_config *item;
  if (item = tsch_config_pool_entry(&p->pool), !item) {
    /* this should probably signal an error! */
    return NULL;
  }
  item->next = NULL;
  item->key  = NULL;
  item->type = TSCH_CFG_BOOL;
  item->b    = tf;

  return item;
}

static inline struct tsch_config *
tsch_config_str(struct tsch_config_parser *p, const char *str)
{
  struct tsch_config *item;
  if (item = tsch_config_pool_entry(&p->pool), !item) {
    /* this should probably signal an error! */
    return NULL;
  }
  item->next = NULL;
  item->key  = NULL;
  item->type = TSCH_CFG_STRING;
  item->s    = tsch_string_table_intern(&p->strings,str);

  return item;
}

static inline struct tsch_config *
tsch_config_num(struct tsch_config_parser *p, double num)
{
  struct tsch_config *item;
  if (item = tsch_config_pool_entry(&p->pool), !item) {
    /* this should probably signal an error! */
    return NULL;
  }
  item->next = NULL;
  item->key  = NULL;
  item->type = TSCH_CFG_NUM;
  item->num  = num;

  return item;
}

static inline struct tsch_config *
tsch_config_map(
    struct tsch_config_parser *p,
    const char *name,
    struct tsch_config *value)
{
  struct tsch_config *pair;

  if (pair = tsch_config_pool_entry(&p->pool), !pair) {
    /* this should probably signal an error! */
    return NULL;
  }

  pair->next = NULL;
  pair->key  = tsch_string_table_intern(&p->strings,name);
  pair->type = TSCH_CFG_MAPPING;
  pair->val  = value;

  return pair;
}

static inline struct tsch_config *
tsch_config_list_empty(struct tsch_config_parser *p)
{
  struct tsch_config *item;
  if (item = tsch_config_pool_entry(&p->pool), !item) {
    /* this should probably signal an error! */
    return NULL;
  }
  item->next = NULL;
  item->key  = NULL;
  item->type = TSCH_CFG_SEQUENCE;
  item->val  = NULL;

  return item;
}


#undef MODULE_NAME

#endif /* TSCH_CONFIG_PARSE_H */

