#include "tsch_config_yaml.h"
#include "tsch_config.h"

#include <yaml.h>

#define NEW_YAML_ERROR(etype,mesg,...) ( \
    TS_NEW_ERROR2(etype,                 \
      ("line %d, column %d: " mesg,      \
       parser->mark.line, parser->mark.column, \
       __VA_ARGS__) ) )

static struct tsch_config* parse_entry(yaml_parser_t* parser, struct tsch_config_pool** poolp, TSResult* errp);
static struct tsch_config* parse_event(yaml_parser_t* parser, yaml_event_t* event, struct tsch_config_pool** poolp, TSResult* errp);

static struct tsch_config* parse_mapping(yaml_parser_t* parser, struct tsch_config_pool** poolp, TSResult* errp)
{
  struct tsch_config* mapping = tsch_config_pool_entry(poolp);
  if (mapping == NULL) {
    *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    return NULL;
  }

  mapping->type = TSCH_CFG_MAPPING;
  mapping->key = NULL;
  mapping->next = NULL;
  mapping->val = NULL;

  struct tsch_config* items = NULL;
  *errp = TS_OK;
  /* mappings are a sequence of key-value pairs */
  for(;;) {
    yaml_event_t event;
    struct tsch_config* kvpair;

    if (!yaml_parser_parse(parser,&event)) {
      *errp = NEW_YAML_ERROR(TS_ERR_PARSE_ERROR,"error parsing YAML",0);
      return NULL;
    }

    if (event.type == YAML_MAPPING_END_EVENT) {
      mapping->val = items;
      return mapping;
    } else if (event.type == YAML_SCALAR_EVENT) {
      /* get key */
      const char *key;

      if (kvpair = parse_entry(parser,poolp,errp), !kvpair) {
        return NULL;
      }

      key = ts_strdup(NO_ALLOCATOR, (const char*)event.data.scalar.value);
      kvpair->key = key;
      kvpair->next = items;
      items = kvpair;
    } else {
      *errp = NEW_YAML_ERROR(TS_ERR_YAML_INVALID_KEY,
          "keys expected to be scalars", 0);
      return NULL;
    }
  }
}

static struct tsch_config* parse_sequence(yaml_parser_t* parser, struct tsch_config_pool** poolp, TSResult* errp)
{
  struct tsch_config* seq = tsch_config_pool_entry(poolp);
  if (seq == NULL) {
    *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
    return NULL;
  }

  seq->type = TSCH_CFG_SEQUENCE;
  seq->key  = NULL;
  seq->next = NULL;
  seq->val  = NULL;

  struct tsch_config head;
  memset(&head,0,sizeof(head));
  struct tsch_config* tail = &head;

  *errp = TS_OK;
  /* mappings are a sequence of key-value pairs */
  for(;;) {
    yaml_event_t event;
    struct tsch_config* element = NULL;
    if (!yaml_parser_parse(parser,&event)) {
      *errp = NEW_YAML_ERROR(TS_ERR_PARSE_ERROR,"error parsing YAML",0);
      return NULL;
    }

    if (event.type == YAML_SEQUENCE_END_EVENT) {
      seq->val = head.next;
      return seq;
    } else {
      element = parse_event(parser, &event, poolp, errp);
      if (element != NULL) {
        tail->next = element;
        tail = element;
      } else {
        return NULL;
      }
    }
  }
}

static struct tsch_config* parse_event(yaml_parser_t* parser, yaml_event_t* event, struct tsch_config_pool** poolp, TSResult* errp)
{
restart:
  switch (event->type) {
    case YAML_NO_EVENT:
    case YAML_STREAM_START_EVENT:
    case YAML_DOCUMENT_START_EVENT:
    case YAML_ALIAS_EVENT:
      if (!yaml_parser_parse(parser,event)) {
        *errp = TS_NEW_ERROR2( TS_ERR_PARSE_ERROR,
            ("YAML parser error at line %d, column %d",
             parser->mark.line,
             parser->mark.column));
        return NULL;
      }

      goto restart;

    case YAML_STREAM_END_EVENT:
    case YAML_DOCUMENT_END_EVENT:
      *errp = TS_OK;
      return NULL;

    case YAML_SEQUENCE_END_EVENT:
    case YAML_MAPPING_END_EVENT:
      *errp = TS_NEW_ERROR2( TS_ERR_PARSE_ERROR,
          ("Unexpected YAML event %s at line %d, column %d",
           (event->type == YAML_SEQUENCE_END_EVENT) ? "SEQ-END" : "MAP-END",
           parser->mark.line,
           parser->mark.column));
      return NULL;

    case YAML_SCALAR_EVENT:
      {
        struct tsch_config* scalar = tsch_config_pool_entry(poolp);
        if (scalar == NULL) {
          *errp = TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
          return NULL;
        }

        scalar->type = TSCH_CFG_STRING;
        scalar->s = ts_strdup(NO_ALLOCATOR, (const char*)event->data.scalar.value);
        return scalar;
      }

    case YAML_SEQUENCE_START_EVENT:
      return parse_sequence(parser, poolp, errp);

    case YAML_MAPPING_START_EVENT:
      return parse_mapping(parser, poolp, errp);
  }

  return NULL; /* suppresses GCC 4.9 warning */
}

static struct tsch_config* parse_entry(yaml_parser_t* parser, struct tsch_config_pool** poolp, TSResult* errp)
{
  yaml_event_t event;

  if (yaml_parser_parse(parser,&event)) {
    return parse_event(parser, &event, poolp, errp);
  } else {
    *errp = TS_NEW_ERROR2( TS_ERR_PARSE_ERROR,
        ("YAML parser error at line %d, column %d",
         parser->mark.line,
         parser->mark.column));
    return NULL;
  }
}

struct tsch_config *
tsch_parse_yaml(struct tsch_config_pool *pool, const char *s, size_t slen, TSResult *errp)
{
  yaml_parser_t parser;
  struct tsch_config *root;

  yaml_parser_initialize(&parser);
  yaml_parser_set_input_string(&parser, (const unsigned char*)s, slen);

  root = parse_entry(&parser, &pool, errp);

  yaml_parser_delete(&parser);

  return root;
}

ts_chem_driver* ts_chem_driver_from_yaml(const char* s, size_t slen, TSResult* errp, int ignore_missing)
{
  struct tsch_config_pool* pool;
  struct tsch_config *root;
  ts_chem_driver* drv;

  pool = tsch_config_pool_alloc(NULL, TSCH_CFG_POOL_SIZE);

  root = tsch_parse_yaml(pool, s,slen, errp);

  drv = NULL;
  if (root)
  {
    assert( IS_OKAY(*errp) );
    *errp = tsch_config_build_driver(root, &drv, ignore_missing);
  }

  tsch_config_pool_dealloc(pool);

  return drv;
}

