#include "ts_chem_driver.h"
#include "ts_errors.h"
#include "ts_machdeps.h"
#include "ts_macros.h"

#define DEBUG_LOGGING 0
#if DEBUG_LOGGING
#  define DEBUG_LOG(...) fprintf(stderr, __VA_ARGS__)
#else
#  define DEBUG_LOG(...) 
#endif

#define DEFAULT_CHANGE_LIST_SIZE 32

/* fast sheet model */
struct tsch_fast_sheet_model *tsch_fast_sheet_model_new(ts_chem_driver *driver)
{
  struct tsch_fast_sheet_model *model;
  size_t nwork, model_size;
  size_t dbl_off;
  double *dptr;
  int ncol;
  TSResult ret;

  ncol = ts_chem_driver_num_protofilaments(driver);
  nwork  = tsch_fast_sheet_model_nwork(driver);
  model_size = sizeof(*model) + nwork * sizeof(model->work[0]);

  dbl_off = model_size % DBL_ALIGN;
  if (dbl_off > 0) { dbl_off = DBL_ALIGN - dbl_off; }
  dbl_off += model_size;

  model_size = dbl_off + (2 + 7*ncol)*sizeof(double);

  if (model = ts_alloc(NO_ALLOCATOR, model_size), model == NULL) {
err1:
    ts_free(NO_ALLOCATOR,model);
    return NULL;
  }

  memset(model, 0, model_size);

  memset(model, 0, sizeof(*model));
  model->driver = ts_chem_driver_retain(driver);
  model->scale = 1.0;
  model->minus_copy = NULL;
  model->total_strain = 0.0;

  model->saved_changes = (struct ts_lattice_change_list){0};
  ret = ts_lattice_change_list_resize(&model->saved_changes,
      DEFAULT_CHANGE_LIST_SIZE), ON_ERROR_ADD_FRAME(ret);
  if (IS_ERROR(ret))
  {
err2:
    ts_lattice_change_list_finalize(&model->saved_changes);
    goto err1;
  }
  
  assert(model->saved_changes.n == 0);
  assert(model->saved_changes.max == DEFAULT_CHANGE_LIST_SIZE);

  dptr = (double*) ((char *)model + dbl_off);

  model->relax_coeff       = dptr; dptr += ncol;
  model->no_tube_strain    = dptr; dptr += ncol;
  model->left_tube_strain  = dptr; dptr += ncol;
  model->right_tube_strain = dptr; dptr += ncol;
  model->both_tube_strain  = dptr; dptr += ncol;
  model->dE                = dptr;

  model->sheet_fluct_dist = -1;
  model->sheet_fluct_penalty = 0.0;

  return model;
}

static TSResult fast_sheet_model_move_hook(void *closure, ts_lattice *latt, const ts_lattice_move *move);

void tsch_fast_sheet_model_delete(struct tsch_fast_sheet_model *model)
{
  if (model->driver)
  {
    ts_lattice_remove_move_hook(&model->driver->latt, fast_sheet_model_move_hook, model);
    ts_chem_driver_release(model->driver);
    model->driver = NULL;
  }

  ts_free(NO_ALLOCATOR, model);
}

static void find_sheet_and_tube_bonds(struct tsch_fast_sheet_model *model,
    const int *restrict tips,
    int *restrict tube_bonds,
    int *restrict sheet_bonds)
{
  ts_chem_driver *drv;
  int npfs, j;

  drv = model->driver;
  npfs    = ts_chem_driver_num_protofilaments(drv);

  for(j=0; j < npfs-1; j++)
  {
    int i, dimer;
    spin_t *restrict col;
    col = ts_lattice_ref(&drv->latt, 0, 2*j+1);

    for (dimer=tips[j],i=4*tips[j]-1; i >= 3; i -= 4, dimer--)
    {
      if ((col[i-0].val != TS_CHEMDRIVER_STATE_EMPTY) ||
          (col[i-2].val != TS_CHEMDRIVER_STATE_EMPTY))
      {
        break;
      }
    }
    sheet_bonds[j] = dimer;

    for (; i >= 3; i -= 4, dimer--)
    {
      if ((col[i-0].val == TS_CHEMDRIVER_STATE_TUBELIKE) &&
          (col[i-2].val == TS_CHEMDRIVER_STATE_TUBELIKE))
      {
        break;
      }
    }
    tube_bonds[j] = dimer;

    for (dimer=tips[npfs+j],i=4*tips[npfs+j]-1; dimer <= tips[j]; i += 4, dimer++)
    {
      if ((col[i-0].val != TS_CHEMDRIVER_STATE_EMPTY) ||
          (col[i-2].val != TS_CHEMDRIVER_STATE_EMPTY))
      {
        break;
      }
    }
    sheet_bonds[npfs+j] = dimer;

    for (; dimer <= tips[j]; i += 4, dimer++)
    {
      if ((col[i-0].val == TS_CHEMDRIVER_STATE_TUBELIKE) &&
          (col[i-2].val == TS_CHEMDRIVER_STATE_TUBELIKE))
      {
        break;
      }
    }
    tube_bonds[npfs+j] = dimer;
  }

  {
    int i,dimer;
    spin_t *restrict col;
    col = ts_lattice_ref(&drv->latt, 0, 2*npfs-1);

    for (dimer=tips[0],i=4*tips[0]-3; i >= 3; i -= 4, dimer--)
    {
      if ((col[i-0].val != TS_CHEMDRIVER_STATE_EMPTY) ||
          (col[i-2].val != TS_CHEMDRIVER_STATE_EMPTY))
      {
        break;
      }
    }
    sheet_bonds[npfs-1] = dimer;

    for (; i >= 3; i -= 4, dimer--)
    {
      if ((col[i-0].val == TS_CHEMDRIVER_STATE_TUBELIKE) ||
          (col[i-2].val == TS_CHEMDRIVER_STATE_TUBELIKE))
      {
        break;
      }
    }
    tube_bonds[npfs-1] = dimer;

    for (dimer=tips[npfs+0],i=4*tips[npfs+0]-3; dimer <= tips[0]; i += 4, dimer++)
    {
      if ((col[i-0].val != TS_CHEMDRIVER_STATE_EMPTY) ||
          (col[i-2].val != TS_CHEMDRIVER_STATE_EMPTY))
      {
        break;
      }
    }
    sheet_bonds[2*npfs-1] = dimer;

    for (; dimer < tips[0]; i += 4, dimer++)
    {
      if ((col[i-0].val == TS_CHEMDRIVER_STATE_TUBELIKE) ||
          (col[i-2].val == TS_CHEMDRIVER_STATE_TUBELIKE))
      {
        break;
      }
    }
    tube_bonds[2*npfs-1] = dimer;
  }

#if DEBUG_LOGGING
  if (npfs > 0)
  {
    DEBUG_LOG("\npf tip+   sheet+ tube+  tip-   sheet- tube-\n");
    for(j=0; j < npfs; j++)
    {
      DEBUG_LOG("%2d %-6d %-6d %-6d %-6d %-6d %-6d\n", j,
          tips[j],sheet_bonds[j],tube_bonds[j],
          tips[npfs+j],sheet_bonds[npfs+j],tube_bonds[npfs+j]);
    }
    DEBUG_LOG("\n");
  }
#endif

}

void tsch_fast_sheet_model_last_tube_rows(struct tsch_fast_sheet_model *model,
    const int *restrict tips,
    struct tsch_fast_sheet_model_wksp *wksp)
{
  ts_chem_driver *drv;
  int ndimers, npfs, j, min_plus, max_minus;
  int *restrict tube_bonds  = wksp->tube_bonds;
  int *restrict sheet_bonds = wksp->sheet_bonds;
  int *restrict rows        = wksp->last_tubes;

  drv = model->driver;
  ndimers = ts_chem_driver_num_dimers(drv);
  npfs    = ts_chem_driver_num_protofilaments(drv);

  min_plus  = ndimers;
  max_minus = 1;

  find_sheet_and_tube_bonds(model, tips, tube_bonds, sheet_bonds);

  for (j=1; j < npfs; j++)
  {
    int i, dimer, plus_dimer;
    const spin_t *restrict col;

    col  = ts_lattice_ref(&drv->latt, 0, 2*j);

    for (dimer=tips[j],i=4*tips[j]-1; i >= 3; i -= 4, dimer--)
    {
      int sheet_left  = (dimer <= sheet_bonds[j-1]) && (dimer > tube_bonds[j-1]);
      int sheet_right = (dimer <= sheet_bonds[j-0]) && (dimer > tube_bonds[j-0]);

      if (!sheet_left && !sheet_right &&
          tsch_is_straight(col[i-0].val) &&
          tsch_is_straight(col[i-1].val) &&
          tsch_is_straight(col[i-2].val) &&
          ((i<=3) || tsch_is_straight(col[i-3].val)))
      {
        break;
      }
    }

    rows[j] = plus_dimer = dimer;
    if (dimer < min_plus) { min_plus = dimer; }

    for (dimer=tips[npfs+j], i=4*tips[npfs+j]-3; dimer <= plus_dimer; i+= 4, dimer++)
    {
      int sheet_left  = (dimer >= sheet_bonds[npfs+j-1]) && (dimer < tube_bonds[npfs+j-1]);
      int sheet_right = (dimer >= sheet_bonds[npfs+j-0]) && (dimer < tube_bonds[npfs+j-0]);

      if (DEBUG_LOGGING && j >= 6 && j <= 8)
      {
        DEBUG_LOG("[%3d | %3d] "
            "(%3d in %3d ... %3d) = %1d  (%3d in %3d ... %3d) = %1d  %1d %1d %1d %1d\n",
            dimer, j, dimer,sheet_bonds[npfs+j-1], tube_bonds[npfs+j-1], sheet_left,
            dimer,sheet_bonds[npfs+j-0], tube_bonds[npfs+j-0], sheet_right,
            tsch_is_straight(col[i-0].val),
            tsch_is_straight(col[i-1].val),
            tsch_is_straight(col[i-2].val),
            (i<=3) || tsch_is_straight(col[i-3].val));
      }

      if (!sheet_left && !sheet_right &&
          tsch_is_straight(col[i+0].val) &&
          tsch_is_straight(col[i+1].val) &&
          tsch_is_straight(col[i+2].val) &&
          ((i<=3) || tsch_is_straight(col[i+3].val)))
      {
        break;
      }
    }

    rows[npfs+j] = dimer;
    if (dimer > max_minus) { max_minus = dimer; }
  }

  {
    int i, dimer, plus_dimer;
    const spin_t *restrict col;

    col  = ts_lattice_ref(&drv->latt, 0,  0);

    for (dimer=tips[0], i=4*tips[0]-1; i >= 3; i -= 4, dimer--)
    {
      int ldimer = dimer+1;
      int sheet_seam  = (ldimer <= sheet_bonds[npfs-1]) && (ldimer > tube_bonds[npfs-1]);
      int sheet_right = ( dimer <= sheet_bonds[0]     ) && ( dimer > tube_bonds[0]     );

      if (!sheet_seam && !sheet_right &&
          tsch_is_straight(col[i-0].val) &&
          tsch_is_straight(col[i-1].val) &&
          tsch_is_straight(col[i-2].val) &&
          ((i<=3) || tsch_is_straight(col[i-3].val)))
      {
        break;
      }
    }

    rows[0] = plus_dimer = dimer;
    if (dimer < min_plus) { min_plus = dimer; }

    for (dimer=tips[npfs+0], i=4*tips[npfs+0]-3; dimer <= plus_dimer; i+= 4, dimer++)
    {
      int ldimer = dimer+1;
      int sheet_seam  = (ldimer >= sheet_bonds[2*npfs-1]) && (ldimer < tube_bonds[2*npfs-1]);
      int sheet_right = (dimer >= sheet_bonds[npfs+0]) && (dimer < tube_bonds[npfs+0]);

      DEBUG_LOG("[%3d | %3d] (%3d >= %3d) && (%3d < %3d) ? %1d  "
          "(%3d >= %3d) && (%3d < %3d) ? %1d\n",
          dimer, 0, ldimer,sheet_bonds[2*npfs-1], ldimer,tube_bonds[2*npfs-1], sheet_seam,
                     dimer,sheet_bonds[2*npfs-0],  dimer,tube_bonds[2*npfs-0], sheet_right);


      if (!sheet_seam && !sheet_right &&
          tsch_is_straight(col[i+0].val) &&
          tsch_is_straight(col[i+1].val) &&
          tsch_is_straight(col[i+2].val) &&
          ((i<=3) || tsch_is_straight(col[i+3].val)))
      {
        break;
      }
    }

    rows[npfs+0] = dimer;
    if (dimer > max_minus) { max_minus = dimer; }
  }

  model->plus_minrow  = min_plus;
  model->minus_maxrow = max_minus;
}

static int is_bent_or_straight(int spin)
{
  return (spin == TS_CHEMDRIVER_STATE_BENT_IFC) ||
    (spin == TS_CHEMDRIVER_STATE_BENT_LEFT) ||
    (spin == TS_CHEMDRIVER_STATE_BENT_RIGHT) ||
    (spin == TS_CHEMDRIVER_STATE_STRAIGHT) ||
    (spin == TS_CHEMDRIVER_STATE_STRAIGHT_GDP);
}

void tsch_fast_sheet_model_tip_rows(struct tsch_fast_sheet_model *model, int *rows)
{
  ts_chem_driver *drv;
  int ndimers, npfs, nrow, j, maxrow, minrow;

  drv = model->driver;
  nrow = drv->latt.nrow;
  ndimers = ts_chem_driver_num_dimers(drv);
  npfs    = ts_chem_driver_num_protofilaments(drv);

  maxrow = 0;
  minrow = ndimers;

  for (j=0; j < npfs; j++)
  {
    int i, dimer_row, i0, plus_row;
    spin_t *restrict col;

    col = ts_lattice_ref(&drv->latt, 0,2*j);

    i0 = (rows[j] > 0) && (rows[j] <= ndimers) ? 4*rows[j]-1 : nrow-2;

plus_retry:
    for (i=i0; (i < nrow-1) && is_bent_or_straight(col[i].val); i++)
    {
      continue;
    }

    for (; (i > 0) && !is_bent_or_straight(col[i].val); i--)
    {
      continue;
    }

    if (i0 != nrow-2)
    {
      if ((i <= 0) || !is_bent_or_straight(col[i].val) || is_bent_or_straight(col[i+1].val))
      {
        i0 = nrow-2;
        goto plus_retry;
      }
    }

    dimer_row = (i < nrow-1) ? (i+1)/4 : ndimers;
    rows[j] = dimer_row;
    if (maxrow < dimer_row) { maxrow = dimer_row; }

    plus_row = i0 = i;
    if ((rows[npfs+j] > 0) && (rows[npfs+j] <= ndimers))
    {
      i = i0 = 4*rows[npfs+j]-1;
    }

minus_retry:
    for (; (i > 0) && is_bent_or_straight(col[i].val); i--)
    {
      continue;
    }

    for (; (i < nrow-1) && !is_bent_or_straight(col[i].val); i++)
    {
      continue;
    }

    if (i0 != plus_row)
    {
      if ((i <= 0) || !is_bent_or_straight(col[i].val) || is_bent_or_straight(col[i-1].val))
      {
        i = i0 = plus_row;
        goto minus_retry;
      }
    }

    dimer_row = (i < nrow-1) ? (i+3)/4 : ndimers;
    rows[npfs+j] = dimer_row;
    if (minrow > dimer_row) { minrow = dimer_row; }
  }

  model->plus_maxrow = maxrow;
  model->minus_minrow = minrow;
}

#define STATE_SHEET 0
#define STATE_TUBE  1
int tsch_fast_sheet_model_row_segments(struct tsch_fast_sheet_model *model, int row,
    const struct tsch_fast_sheet_model_wksp *wksp,
    int *restrict segs0)
{
  ts_chem_driver *drv = model->driver;
  int j, npf, state, lbl, *restrict segs1;

  const int *restrict last_tubes  = wksp->last_tubes;
  const int *restrict tips        = wksp->tips;
  const int *restrict sheet_bonds = wksp->sheet_bonds;

  state = STATE_TUBE;
  lbl = 0;

  npf = ts_chem_driver_num_protofilaments(drv);
  segs1 = segs0+npf;

  for(j=0; j < npf; j++)
  {
    int in_sheet, off, dir;

    off = j;
    dir = 1;
    if (row < last_tubes[j])
    {
      off += npf;
      dir = -1;
    }

    in_sheet = (dir*(row - last_tubes[off]) > 0) && (dir*(tips[off] - row) >= 0);

    /* last_tubes[off] < row   ==> last_tubes[off] - row < 0  ==> row - last_tubes[off] > 0 */
    /*       tips[off] >= row  ==>      tips[off] - row >= 0  */

    /* last_tubes[off] > row   ==> last_tubes[off] - row > 0  ==> -(row - last_tubes[off]) >  0 */
    /*       tips[off] <= row  ==>      tips[off] - row <= 0  ==> -(tips[off] - row)       >= 0 */

    DEBUG_LOG(
        "[ %2d, %2d ] last_tubes[%2d] = %2d, last_tubes[%2d] = %2d, "
        "tips[%2d] = %2d, tips[%2d] = %2d, in_sheet = %1d\n",
        row, j,
        j, last_tubes[j], j+npf, last_tubes[j+npf],
        j, tips[j], j+npf, tips[j+npf],
        in_sheet);

    switch (state) {
      case STATE_TUBE:
        if (in_sheet)
        {
          state = STATE_SHEET;
          segs0[lbl] = j;
        }
        break;

      case STATE_SHEET:
        {
          int change = !in_sheet;
          if (!change && j>0) {
            change = dir*(sheet_bonds[off-1] - row) < 0;
            // change = (off == j) ? sheet_bonds[j-1] < row : sheet_bonds[off-1] > row;

            /* sheet_bonds[j-1]   < row ==> sheet_bonds[j-1]   - row < 0
             * sheet_bonds[off-1] > row ==> sheet_bonds[off-1] - row > 0 
             *                          ==> -(sheet_bonds[off-1] - row) < 0
             */
          }
          if (change)
          {
            state = in_sheet ? STATE_SHEET : STATE_TUBE;
            segs1[lbl] = j-1;
            segs0[++lbl] = j;
          }
        }
        break;
    }
  }

  if (state == STATE_SHEET)
  {
    segs1[lbl] = npf-1;
    lbl++;
  }
  return lbl;
}
#undef STATE_SHEET
#undef STATE_TUBE

TSResult tsch_fast_sheet_model_map_segments(struct tsch_fast_sheet_model *model,
    const int *restrict segs0, int n0, const int *restrict segs1, int n1, int *restrict map)
{
  ts_chem_driver *drv = model->driver;
  int j, j1, npf = ts_chem_driver_num_protofilaments(drv);

  for (j=0; j < n0; j++)
  {
    map[j] = -1;
  }

  for (j1=0; j1 < n1; j1++)
  {
    int left,right;
    left  = segs1[j1];
    right = segs1[j1+npf];

    for (j=0; j < n0; j++)
    {
      int include_left, include_right;

      include_left  = (segs0[j] >= left) && (segs0[j] <= right);
      include_right = (segs0[j+npf] >= left) && (segs0[j+npf] <= right);

      if (include_left || include_right)
      {
        if (map[j] >= 0)
        {
          TS_RAISE_ERRORF(TS_ERR_SIMULATION_STATE,
              "segment map from row n to row n+1 is not one-to-one",0);
        }

        map[j] = j1;
      }
    }
  }

  return TS_OK;
}

TSResult tsch_fast_sheet_model_begin_rowscan(struct tsch_fast_sheet_model *model, int end)
{
  size_t npf   = ts_chem_driver_num_protofilaments(model->driver);

  struct tsch_fast_sheet_model_wksp wksp;
  int n0,n1;

  memset(model->dE, 0, (2*npf+2)*sizeof(model->dE[0]));

  tsch_fast_sheet_model_setup_wksp(model, &wksp);
  model->total_strain = 0.0;

  tsch_fast_sheet_model_tip_rows(model, wksp.tips);
  tsch_fast_sheet_model_last_tube_rows(model, wksp.tips, &wksp);

  switch (end)
  {
    case TSCH_FSM_PLUS_END:
      n0 = tsch_fast_sheet_model_row_segments(model, model->plus_minrow+0, &wksp, wksp.segs0);
      n1 = tsch_fast_sheet_model_row_segments(model, model->plus_minrow+1, &wksp, wksp.segs1);
      break;

    case TSCH_FSM_MINUS_END:
      n0 = tsch_fast_sheet_model_row_segments(model, model->minus_maxrow-0, &wksp, wksp.segs0);
      n1 = tsch_fast_sheet_model_row_segments(model, model->minus_maxrow-1, &wksp, wksp.segs1);

      break;

    default:
      TS_RAISE_ERRORF(TS_ERR_INVALID_ARGUMENT, "argument 'end' is invalid", 0);
      break;
  }

  model->n0 = n0;
  model->n1 = n1;

  return TS_OK;
}

static inline double fast_sheet_ifc_model(struct tsch_fast_sheet_model *model, int ntube, int nsheet)
{
  double nt=ntube, ns=nsheet;

  if (nsheet <= 0) { return 0.0; }

  return model->ifc_model[0]  +
    model->ifc_model[1]*ns    +
    model->ifc_model[2]*ns*ns +
    model->ifc_model[3]*nt    +
    model->ifc_model[4]*nt*nt +
    model->ifc_model[5]*ns*nt;
}

static inline double fast_sheet_existing_model(struct tsch_fast_sheet_model *model, int nexisting)
{
  double ne;

  if (nexisting <= 0) { return 0.0; }

  ne = nexisting;
  return model->existing_model[0] +
    model->existing_model[1]*ne   +
    model->existing_model[2]*ne*ne;
}

double tsch_fast_sheet_model_ifc_energy(struct tsch_fast_sheet_model *model, int ntube, int nsheet, int nexisting)
{
  return fast_sheet_ifc_model(model, ntube, nsheet) +
    fast_sheet_existing_model(model,nexisting);
}

static inline void reverse_ints(int *restrict vals, int n)
{
  int i,j,sw;
  for (i=0, j=n-1; i < j; i++,j--)
  {
    sw = vals[i], vals[i] = vals[j], vals[j] = sw;
  }
}

static inline void reverse_seglist(int *restrict segs, int n, int npf)
{
  int i,sw;

  reverse_ints(segs,n);
  reverse_ints(segs+npf,n);

  for (i=0; i < n; i++)
  {
    sw = npf-1-segs[i];
    segs[i] = npf-1-segs[i+npf];
    segs[i+npf] = sw;
  }
}

static MAYBE_UNUSED void print_intlist(FILE *f, const char *pfx, int *vals, int n)
{
  int i;

  if (!f) { f = stdout; }
  if (pfx) { fprintf(f, "%8.8s:", pfx); }
  for (i=0; i < n; i++) { fprintf(f, " %d", vals[i]); }
  fprintf(f, "\n");
}

static MAYBE_UNUSED void print_seglist(FILE *f,
    const char *pfx, int *restrict segs, int n, int npf)
{
  int i;

  if (!f) { f = stdout; }
  if (pfx) { fprintf(f, "%8.8s:", pfx); }
  for (i=0; i < n; i++) { fprintf(f, " (%d,%d)", segs[i], segs[i+npf]); }
  fprintf(f, "\n");
}

static double fast_sheet_model_next_row(struct tsch_fast_sheet_model *model,
    int row,
    struct tsch_fast_sheet_model_wksp *wksp,
    int *restrict segs0, int n0,
    int *restrict segs1, int n1,
    double *restrict dE0,
    double *restrict dE1,
    int direc)
{
  ts_chem_driver *driver;
  int s0,i, npf;
  double sum;

  int *restrict last_tubes = wksp->last_tubes;
  int *restrict tube_bonds = wksp->tube_bonds;
  const double *restrict left_tube_strain  = model->left_tube_strain;
  const double *restrict right_tube_strain = model->right_tube_strain;
  int *restrict oldifc = wksp->oldifc;

  driver = model->driver;
  npf = ts_chem_driver_num_protofilaments(driver);

  if (direc < 0)
  {
    /* swap stuff ... */

    last_tubes += npf;
    tube_bonds += npf;
  }

  DEBUG_LOG( "\n---[row %d]---\n",row);

  if (DEBUG_LOGGING)
  {
    DEBUG_LOG("last_tubes:");
    for (i=0; i < npf; i++)
    {
      DEBUG_LOG(" %3d", last_tubes[i]);
    }

    DEBUG_LOG("\n");
  }

  for(sum=0.0, s0=0, i=0; i < n1; i++)
  {
    int j, nexist, beg,end,len, left_tube,right_tube, width, lsheet,rsheet, ntube,nsheet;
    double relax_energy, internal_strain, new_ifc_energy, frac_retained;

    beg = segs1[i];

    end = segs1[i+npf];
    len = end-beg+1;
    new_ifc_energy = relax_energy = 0.0;

    for (j=0; j < len; j++) { oldifc[j] = 0; }

    lsheet = direc*(tube_bonds[beg] - row) < 0;
    rsheet = (len % 2 == 0) ? lsheet : !lsheet;

    DEBUG_LOG(
        "seg %2d [%2d,%2d] len=%2d, %d*(%d - %d) < 0 ? %d\n"
        "  lsheet=%d, rsheet=%d\n"
        "  tube_bonds[%d]=%d row=%d\n",
        i, beg,end, len,
        direc,tube_bonds[beg],row, direc*(tube_bonds[beg]-row) < 0,
        lsheet,rsheet, 
        beg,tube_bonds[beg],row);

    for (nexist=0,j=s0; j < n0; j++)
    {
      int beg0 = segs0[j], end0 = segs0[j+npf], len0 = end0-beg0+1;
      int ovl0,ovl1, olen, k;

      if (end0 < beg) { continue; }

      if (beg0 > end) { break; }

      ovl0 = (beg0 > beg) ? beg0 : beg;
      ovl1 = (end0 < end) ? end0 : end;
      olen = ovl1-ovl0+1;

      nexist += olen;
      relax_energy += (olen == len0) ? dE0[j] : (dE0[j]*olen)/len0;

      DEBUG_LOG("  segs0[%d] overlap: %2d/%2d (dE[%2d] = %7.4f, frac=%7.4f)\n",
          j, olen, len0, j, dE0[j], (dE0[j]*olen)/len0);

      for (k=ovl0; k <= ovl1; k++) { oldifc[k-beg] = j+1; }

      if (end0 <= end) { s0 = j+1; }
    }

    DEBUG_LOG("oldifc:");
    for (j=0; j < len; j++)
    {
      DEBUG_LOG(" %2d", oldifc[j]);
    }
    DEBUG_LOG("\n");

    DEBUG_LOG("cumulative energy that can relax: %7.4f\n", relax_energy);


    left_tube  = (beg > 0)     && (direc*(last_tubes[beg-1] - row) >= 0);
    right_tube = (end < npf-1) && (direc*(last_tubes[end+1] - row) >= 0);
    new_ifc_energy = 0.0;

    if ((nexist < len) && (row > 1))
    {
      int j0;

      ntube = !oldifc[0] && left_tube && lsheet;
      nsheet = 0;

      for (j0=0,j=1; j < len; j++)
      {
        if (!oldifc[j] == !oldifc[j-1]) { continue; }

        if (oldifc[j]) /* end of a new piece the new segment */
        {
          int slen,nb,half,extra;
          double en;

          slen = j-j0+1;

          nb = slen-1;

          half = nb/2;
          extra = nb%2;

          DEBUG_LOG("  segment ending.  before calc: half = %d, extra = %d, ntube = %d, nsheet = %d\n", 
              half, extra, ntube, nsheet);

          nsheet += half;
          ntube  += half;

          if (extra)
          {
            int sheet_border;

            sheet_border = (j % 2 == 1) ? lsheet : !lsheet;

            DEBUG_LOG("  segment dist %d  sheet at border? %d\n", j, sheet_border);
            nsheet += sheet_border;
            ntube  += !sheet_border;
          }

          en = fast_sheet_ifc_model(model, ntube,nsheet);
          DEBUG_LOG("[%2d,%2d] slen=%d, nb=%d, half=%d, extra=%d, nsheet=%2d,ntube=%2d, lsheet=%1d, en=%7.4f\n", 
              j0,j-1, slen,nb,half,extra, nsheet,ntube, lsheet, en);
          new_ifc_energy += en;
        }
        else /* segment beginning */
        {
          int sheet_border;

          j0 = j;
          nsheet = ntube = 0;

          sheet_border = (j%2==1) ? lsheet : !lsheet;
          ntube  += !sheet_border;
          nsheet += sheet_border;
        }
      }

      /* take care of any existing */
      if (!oldifc[len-1])
      {
        int slen = j-j0, nb = slen-1, half = nb/2, extra = nb%2;
        double en;

        DEBUG_LOG("  segment ending.  before calc: half = %d, extra = %d, ntube = %d, nsheet = %d\n", 
            half, extra, ntube, nsheet);

        nsheet += half;
        ntube  += half;
        if (extra)
        {
          nsheet += rsheet;
          ntube += !rsheet;
        }

        ntube += right_tube && rsheet;

        en = fast_sheet_ifc_model(model, ntube,nsheet);
        DEBUG_LOG("[%2d,%2d] slen=%d, nb=%d, half=%d, extra=%d, nsheet=%2d,ntube=%2d, lsheet=%1d, en=%7.4f\n", 
            j0,j-1, slen,nb,half,extra, nsheet,ntube, lsheet, en);
        new_ifc_energy += en;
      }
    }

    DEBUG_LOG("ifc energy: %7.4f\n", new_ifc_energy);

    if ((nexist > 0) && (nexist < len))
    {
      double exist_en = fast_sheet_existing_model(model, nexist);
      DEBUG_LOG("exist_en(nexist=%d): %7.4f\n", nexist, exist_en);
      new_ifc_energy += exist_en;
    }

    DEBUG_LOG("tot new energy: %7.4f\n", new_ifc_energy);

    /* Adjust effective width */
    width = len;
    if (left_tube && lsheet)
    {
      width++;
    }

    if (right_tube)
    {
      width += 2;
      if (rsheet) { width++; }
    }

    if (width > npf) { width = npf; }
    frac_retained = (width >= 1) ? model->relax_coeff[width-1] : 0.0;

    if (left_tube && right_tube)
    {
      internal_strain = model->both_tube_strain[width-1];
    }
    else if (left_tube)
    {
      internal_strain = left_tube_strain[width-1];
    }
    else if (right_tube)
    {
      internal_strain = right_tube_strain[width-1];
    }
    else
    {
      internal_strain = model->no_tube_strain[width-1];
    }

    DEBUG_LOG(
        "eff.wid: %2d %2d %1d%1d%1d%1d old: %7.4f f.retain: %7.4f int: %7.4f tot: %7.4f\n",
        width, len, left_tube, lsheet, right_tube, rsheet,
        relax_energy, frac_retained, internal_strain,
        internal_strain + frac_retained*relax_energy);

    dE1[i] = frac_retained*relax_energy + internal_strain + new_ifc_energy;

    sum += dE1[i];
  }

  DEBUG_LOG("sum = %7.4f\n", sum);

  return sum;
}

static inline MAYBE_UNUSED
void swap_int(int *restrict x, int *restrict y, size_t _)
{ int tmp; (void)_; tmp = *x; *x = *y; *y = tmp; }

static inline MAYBE_UNUSED
void swap_int_ptr(int **restrict x, int ** restrict y, size_t _)
{ int *tmp; (void)_; tmp = *x; *x = *y; *y = tmp; }

static inline MAYBE_UNUSED
void swap_double(double *restrict x, double *restrict y, size_t _)
{ double tmp; (void)_; tmp = *x; *x = *y; *y = tmp; }

static inline MAYBE_UNUSED
void swap_double_ptr(double **restrict x, double **restrict y, size_t _)
{ double *tmp; (void)_; tmp = *x; *x = *y; *y = tmp; }

static inline MAYBE_UNUSED
void swap_generic(void *x, void *y, size_t sz)
{ unsigned char tmp[sz]; memcpy(tmp,x,sz); memcpy(x,y,sz); memcpy(y,tmp,sz); }

TSResult tsch_fast_sheet_model_next_rows(struct tsch_fast_sheet_model *model, int row0, int n, int direc, double *strain, double *strain_err)
{
  ts_chem_driver *driver = model->driver;
  int npf = ts_chem_driver_num_protofilaments(driver);
  int r,row,n0 = model->n0, n1 = model->n1;

  struct tsch_fast_sheet_model_wksp wksp;
  tsch_fast_sheet_model_setup_wksp(model, &wksp);

  int *segs0      = wksp.segs0;
  int *segs1      = wksp.segs1;

  if (direc < 0)
  {
    /* swap stuff ... */
    reverse_ints(wksp.tube_bonds+npf,npf-1);
    reverse_seglist(segs0, n0, npf);
    reverse_seglist(segs1, n1, npf);
  }

  double *dE0 = model->dE+0, *dE1 = dE0+npf, *sump = dE1+npf;
  (void)sump;
  double sum = 0.0;

  for (r=1,row=row0+direc*r; r <= n; r++)
  {

    if (direc < 0)
    {
      /* swap stuff ... */
      reverse_ints(wksp.last_tubes+npf,npf);
    }

    sum += fast_sheet_model_next_row(model,row, &wksp, segs0,n0, segs1,n1, dE0, dE1, direc); 

    if (direc < 0)
    {
      /* swap things back ... */
      reverse_ints(wksp.last_tubes+npf,npf);
    }

    /* swap segs0, segs1 and find next row */
    n0 = n1;
    swap_int_ptr(&segs0,&segs1,0);
    swap_double_ptr(&dE0,&dE1,0);

    row += direc;
    n1=tsch_fast_sheet_model_row_segments(model,row, &wksp, segs1);

    if (direc < 0)
    {
      reverse_seglist(segs1, n1, npf);
    }
  }

  *strain = sum;
  *strain_err = 0.0;

  model->n0 = n0;
  model->n1 = n1;

  if (direc < 0)
  {
    /* swap things back ... */
    reverse_ints(wksp.tube_bonds+npf,npf-1);

    reverse_seglist(segs0, n0, npf);
    reverse_seglist(segs1, n1, npf);
  }

  /* check if we've switched, and switch back */
  if (segs0 != wksp.segs0)
  {
    int i;
    for (i=0; i < 2*npf; i++)
    {
      swap_int(&segs0[i],&segs1[i],0);
    }
  }

  if (dE0 != model->dE)
  {
    int i;
    for (i=0; i < npf; i++)
    {
      swap_double(&dE0[i], &dE1[i],0);
    }
  }

  return TS_OK;
}

#define PLUS_DIREC   1
#define MINUS_DIREC -1
static TSResult
calc_minus_energy_by_copying(struct tsch_fast_sheet_model *model, double *energyp)
{
  int j,off,n,ncol, minus_min,minus_max;
  ts_chem_driver *orig, *copy;
  TSResult err;

  orig = model->driver;
  copy = model->minus_copy;

  assert(orig != NULL);
  assert(copy != NULL);

  minus_min = model->minus_minrow;
  minus_max = model->minus_maxrow;
  ncol = orig->latt.ncol;

  if (minus_max <= minus_min)
  {
    return TS_OK;
  }

  off = 4*minus_max-1;
  n = off - 4*(minus_min-1);

  /* first clear everything in the copy */
  ts_lattice_zero(&copy->latt);

  for (j=0; j < ncol-1; j++)
  {
    int i,j1;

    j1 = ncol-2-j;

    for (i=0; i < n; i++)
    {
      int i0 = off-i;
      int spin = TS_LATTICE_SPIN(orig,i0,j);

      if (spin == TS_CHEMDRIVER_STATE_BENT_LEFT)
      {
        spin = TS_CHEMDRIVER_STATE_BENT_RIGHT;
      }
      else if (spin == TS_CHEMDRIVER_STATE_BENT_RIGHT)
      {
        spin = TS_CHEMDRIVER_STATE_BENT_LEFT;
      }

      TS_LATTICE_SPIN(copy, i+1, j1) = spin;
    }
  }

  {
    int i;
    for (i=0; i < n; i++)
    {
      int i0 = off-i;
      int spin = TS_LATTICE_SPIN(orig,i0,ncol-1);

      TS_LATTICE_SPIN(copy, i+1, ncol-1) = spin;
    }
  }

  model->driver = copy;
  model->minus_copy = NULL;

  err = tsch_fast_sheet_model_energy(model, energyp);

  model->driver = orig;
  model->minus_copy = copy;

  RETURN_WITH_RESULT(err);
}

TSResult tsch_fast_sheet_model_energy(struct tsch_fast_sheet_model *model, double *energyp)
{
  double strain, err;
  int minus_min,minus_max, plus_min,plus_max;

  RETURN_ON_ERROR( tsch_fast_sheet_model_begin_rowscan(model, TSCH_FSM_MINUS_END) );

  minus_min = model->minus_minrow; minus_max = model->minus_maxrow;
  plus_min  = model->plus_minrow;  plus_max  = model->plus_maxrow;

  *energyp = 0.0;

  if (plus_min <= minus_min)
  {
    goto plus_end;
  }

  strain = err = 0.0;
  if (model->minus_copy)
  {
    RETURN_ON_ERROR( calc_minus_energy_by_copying(model, &strain) );
  }
  else
  {
    RETURN_ON_ERROR( 
        tsch_fast_sheet_model_next_rows(model,
          minus_max, minus_max - minus_min + 1,
          MINUS_DIREC, &strain, &err)
        );
  }

  *energyp += strain;

plus_end:
  RETURN_ON_ERROR( tsch_fast_sheet_model_begin_rowscan(model, TSCH_FSM_PLUS_END) );

  assert( minus_min == model->minus_minrow );
  assert( minus_max == model->minus_maxrow );
  assert( plus_min == model->plus_minrow );
  assert( plus_max == model->plus_maxrow );

  strain = err = 0.0;
  RETURN_ON_ERROR( 
      tsch_fast_sheet_model_next_rows(model,
        plus_min, plus_max - plus_min + 1,
        PLUS_DIREC, &strain, &err)
      );
  *energyp += strain;
  *energyp *= model->scale;

  return TS_OK;
}
#undef PLUS_DIREC
#undef MINUS_DIREC

TSResult tsch_fast_sheet_model_energy_change(struct tsch_fast_sheet_model *model,
    int i, int j, int st1, double *dE)
{
  double energy0, energy1;
  TSResult err;

  energy0 = energy1 = 0.0;
  RETURN_ON_ERROR( tsch_fast_sheet_model_energy(model, &energy0) );

  if (!ts_lattice_is_pseudo_spin(&model->driver->latt, spin_from_int(st1)))
  {
    int st0;
    st0 = TS_LATTICE_SPIN(model->driver, i,j);

    TS_LATTICE_SPIN(model->driver,i,j) = st1;
    err = tsch_fast_sheet_model_energy(model, &energy1);
    TS_LATTICE_SPIN(model->driver,i,j) = st0;

    RETURN_ON_ERROR(err);
  }
  else
  {
    err = TS_OK;
    switch(st1)
    {
      case PSEUDO_LATERAL_ZERO:
      case PSEUDO_LATERAL_SHEETLIKE:
      case PSEUDO_LATERAL_TUBELIKE:
      case PSEUDO_LATERAL_SEAM:
        if (tsch_debug_updates)
        {
          fprintf(stderr, "FSM: invoking pspin %5d at site %3d,%3d\n", st1,i,j);
        }

        RETURN_ON_ERROR(
            ts_lattice_invoke_pseudospin_without_update(
              &model->driver->latt, spin_from_int(st1), i,j));
        err = tsch_fast_sheet_model_energy(model, &energy1);
        RETURN_ON_ERROR( ts_lattice_rollback_changes(&model->driver->latt) );
        break;

      case PSEUDO_K_ON_BENT_LEFT:
      case PSEUDO_K_ON_BENT_RIGHT:
      case PSEUDO_K_ON_STRAIGHT_LEFT:
      case PSEUDO_K_ON_STRAIGHT_RIGHT:
        RETURN_ON_ERROR(
            ts_lattice_invoke_pseudospin_without_update(
              &model->driver->latt, spin_from_int(st1), i,j));
        err = tsch_fast_sheet_model_energy(model, &energy1);
        RETURN_ON_ERROR( ts_lattice_rollback_changes(&model->driver->latt) );
        break;

      case PSEUDO_K_OFF:
        *dE = 0.0;
        return TS_OK;

      default:
        TS_RAISE_ERRORF(TS_ERR_UNSUPPORTED_OPERATION, "unsupported pseudo spin: %d", st1);
    }

    RETURN_ON_ERROR(err);
  }

  *dE = energy1-energy0;
  return TS_OK;
}

static TSResult fast_sheet_model_move_hook(void *closure, ts_lattice *latt, const ts_lattice_move *move)
{
  struct tsch_fast_sheet_model *model;
  ts_chem_driver *driver;
  double dE, pacc, prand;
  int fluct_dist;
  TSResult err;
  struct ts_lattice_update_list updates = { 0 };

  driver = (ts_chem_driver *)latt;
  model = closure;
  model->num_hook_calls++;

  RETURN_ON_ERROR(ts_lattice_save_changes(latt,
        &model->saved_changes,&updates));

  dE = 0.0;
  err = ON_ERROR_ADD_FRAME( tsch_fast_sheet_model_energy_change(model,
        (int)move->r, (int)move->c, move->state.val, &dE)
      );

  RETURN_ON_ERROR(ts_lattice_restore_changes(latt,
        &model->saved_changes,&updates));

  if (IS_ERROR(err))
  {
    if (ERRCODE(err) != TS_ERR_SIMULATION_STATE)
    {
      return err;
    }

    if (1)
    {
      fprintf(stderr, "ERROR in fast sheet move hook r=%zu, c=%zu, st1=%d\n",
          move->r,move->c,(int)(move->state.val));
      ts_error_dump_stack(err);
    }
    model->num_hook_rejects++;
    CANCEL_OPERATION();
  }

  fluct_dist = model->sheet_fluct_dist;
  if (fluct_dist >= 0 && 
      move->state.val == PSEUDO_LATERAL_SHEETLIKE &&
      move->r > 0)
  {
    struct tsch_fast_sheet_model_wksp wksp;
    int dist, min_dist, pf, npf, dimer;

    tsch_fast_sheet_model_setup_wksp(model, &wksp);

    npf = ts_chem_driver_num_protofilaments(driver);
    pf = move->c/2;
    dimer = 1+(move->r-1)/4;

    if (move->r < 1 || move->r >= driver->latt.nrow-1 ||
        move->c < 1 || move->c >= driver->latt.ncol-1)
    {
      TS_RAISE_ERRORF(TS_ERR_INVALID_INDEX,
          "site (%zu,%zu) is not a valid lattice site",
          move->r,move->c);
    }

    assert(pf>=0 && pf+1 < npf);

    dist = wksp.last_tubes[pf] - dimer;
    min_dist = dist;

    dist = wksp.last_tubes[pf+1] - dimer;
    min_dist = dist < min_dist ? dist : min_dist;

    dist = dimer - wksp.last_tubes[npf+pf];
    min_dist = dist < min_dist ? dist : min_dist;

    dist = dimer - wksp.last_tubes[npf+pf+1];
    min_dist = dist < min_dist ? dist : min_dist;

    if (min_dist > fluct_dist)
    {
      dE += model->sheet_fluct_penalty;
    }

  }

  if (dE <= 0)
  {
    goto accept;
  }

  pacc = ts_chem_driver_prob_acceptance_metropolis(model->driver, dE);
  RETURN_ON_ERROR( ts_rng_gen_uniforms(driver->latt.rng, 1, &prand) );

  if (prand > pacc)
  {
    if (tsch_debug_updates)
    {
      fprintf(stderr,
          "fast sheet model hook canceling move %d at (%3zu,%3zu): prand=%f, pacc=%f\n",
          move->state.val,move->r,move->c,prand,pacc);
    }

    model->num_hook_rejects++;
    CANCEL_OPERATION();
  } 

accept:
  model->num_hook_accepts++;
  return TS_OK;
}

TSResult tsch_fast_sheet_model_add_movehook(struct tsch_fast_sheet_model *model)
{
  RETURN_WITH_RESULT(
      ts_lattice_add_move_hook(&model->driver->latt,
        fast_sheet_model_move_hook, model)
      );
}



