#include "ts_chem_driver.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>

/* if an implemention provides an arithmetic right shift for signed types, #define ARITH_RSHIFT here
 */
#if defined(__clang__) || defined(__GNUC__)
#  if defined(__i386__) || defined(__x86_64__)
#    define ARITH_RSHIFT 1
#    define DBL_IS_INT_ALIGNED 1
#  endif
#endif

#if ARITH_RSHIFT
static inline int32_t ashift32(int32_t x, unsigned int n)
{
  return x>>n;
}

static inline int64_t ashift64(int64_t x, unsigned int n)
{
  return x>>n;
}
#else
static inline int32_t ashift32(int32_t x, unsigned int n)
{
  uint32_t u = x;
  return (u >> n) | (~( ~((uint32_t)0) >> n ) );
}

static inline int64_t ashift64(int64_t x, unsigned int n)
{
  uint64_t u = x;
  return (u >> n) | (~( ~((uint64_t)0) >> n ) );
}
#endif /* ARITH_RSHIFT */

static inline uint32_t zigzag_enc32(int32_t x)
{
  return (x<<1) ^ ashift32(x,sizeof(x)*CHAR_BIT - 1);
}

static inline int32_t zigzag_dec32(uint32_t x)
{
  int32_t y = x>>1;
  return (x&1) ? (-y-1) : y;
}

static inline uint32_t zigzag_enc64(int64_t x)
{
  return (x<<1) ^ ashift64(x,sizeof(x)*CHAR_BIT - 1);
}

static inline MAYBE_UNUSED int64_t zigzag_dec64(uint64_t x)
{
  int64_t y = x>>1;
  return (x&1) ? (-y-1) : y;
}

static TSResult read_u64(FILE *f, uint64_t *valp)
{
  uint64_t val = 0;
  int b,sh,lower7;

  sh = 0;
  do {
    uint64_t bits;

    if (b = fgetc(f), b == EOF)
    {
      goto io_error;
    }

    lower7 = b & ~0x80;
    bits = (uint64_t)lower7 << sh;

    /* check for overflow! */
    if ((int)(bits >> sh) != lower7)
    {
      goto overflow_error;
    }

    val = val | bits;
    sh += 7;
  } while (b != lower7);

  *valp = val;

  return TS_OK;

io_error:
  if (feof(f))
  {
    TS_RAISE_ERROR3(TS_ERR_IO_DECODE_ERROR,
        "Unexpected end of stream while decoding varint");
  }
  else
  {
    TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
        "Unexpected error while decoding varint: %s", strerror(errno));
  }

overflow_error:
  TS_RAISE_ERROR3(TS_ERR_IO_DECODE_ERROR,
      "Overflow while reading 64-bit varint");
}

static TSResult read_u32(FILE *f, uint32_t *valp)
{
  uint64_t val = 0;

  RETURN_ON_ERROR( read_u64(f, &val) );

  if (val > UINT32_MAX)
  {
    TS_RAISE_ERROR3(TS_ERR_IO_DECODE_ERROR,
        "Expected a 32-bit varint, but value is outside of range");
  }

  *valp = (uint32_t)val;
  return TS_OK;
}

static TSResult read_s32(FILE *f, int32_t *valp)
{
  uint32_t val = 0;
  RETURN_ON_ERROR( read_u32(f, &val) );

  *valp = zigzag_dec32(val);
  return TS_OK;
}

static TSResult write_u64(FILE *f, uint64_t val)
{
  unsigned char bytes[sizeof(val) + ((CHAR_BIT + sizeof(val))/CHAR_BIT)];
  size_t n;

  if (val == 0) {
    bytes[0] = 0;
    n = 1;
    goto write;
  }

  for (n=0; val > 0; n++)
  {
    bytes[n] = 0x80 | (val & 0x7f);  /* mask off lowest 7 bits */
    val = val >> 7;
  }

  bytes[n-1] = bytes[n-1] & ~(0x80);

write:
  if (fwrite(bytes,sizeof(bytes[0]),n,f) != n)
  {
    TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED,
        "Error encountered in write: %s", strerror(errno));
  }
  return TS_OK;
}

static inline TSResult write_u32(FILE *f, uint32_t val)
{
  RETURN_WITH_RESULT( write_u64(f, (uint64_t)val) );
}

static inline TSResult write_s32(FILE *f, int32_t val)
{
  RETURN_WITH_RESULT( write_u32(f, zigzag_enc32(val)) );
}

static inline MAYBE_UNUSED TSResult write_s64(FILE *f, int64_t val)
{
  RETURN_WITH_RESULT( write_u64(f, zigzag_enc64(val)) );
}

static TSResult write_fixed_u32(FILE *f, uint32_t u)
{
  unsigned char bytes[ sizeof(u) ];

  bytes[0] = u & 0xff;
  bytes[1] = (u>> 8) & 0xff;
  bytes[2] = (u>>16) & 0xff;
  bytes[3] = (u>>24) & 0xff;

  if (fwrite(bytes,sizeof(bytes[0]), sizeof(u), f) != sizeof(u))
  {
    TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED,
        "Error encountered in write: %s", strerror(errno));
  }

  return TS_OK;
}

static TSResult read_fixed_u32(FILE *f, uint32_t *valp)
{
  unsigned char bytes[ sizeof(uint32_t) ];

  if (fread(bytes, 1, sizeof(bytes), f) != sizeof(bytes))
  {
    TS_RAISE_ERRORF(TS_ERR_READ_NOT_FINISHED,
        "Error encountered in read: %s", strerror(errno));
  }

  *valp = 0;
  *valp = *valp | bytes[3]; *valp = *valp << 8;
  *valp = *valp | bytes[2]; *valp = *valp << 8;
  *valp = *valp | bytes[1]; *valp = *valp << 8;
  *valp = *valp | bytes[0];

  return TS_OK;
}

static MAYBE_UNUSED TSResult write_fixed_i32(FILE *f, int32_t i)
{
  RETURN_WITH_RESULT( write_fixed_u32(f,(uint32_t)i) );
}

static TSResult write_fixed_u64(FILE *f, int64_t u)
{
  unsigned char bytes[ sizeof(u) ];

  bytes[0] = u & 0xff;
  bytes[1] = (u>> 8) & 0xff;
  bytes[2] = (u>>16) & 0xff;
  bytes[3] = (u>>24) & 0xff;
  bytes[4] = (u>>32) & 0xff;
  bytes[5] = (u>>40) & 0xff;
  bytes[6] = (u>>48) & 0xff;
  bytes[7] = (u>>56) & 0xff;

  if (fwrite(bytes,sizeof(bytes[0]), sizeof(u), f) != sizeof(u))
  {
    TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED,
        "Error encountered in write: %s", strerror(errno));
  }

  return TS_OK;
}

static TSResult read_fixed_u64(FILE *f, uint64_t *valp)
{
  unsigned char bytes[ sizeof(uint64_t) ];

  if (fread(bytes, 1, sizeof(bytes), f) != sizeof(bytes))
  {
    TS_RAISE_ERRORF(TS_ERR_READ_NOT_FINISHED,
        "Error encountered in read: %s", strerror(errno));
  }

  *valp = 0;
  *valp = *valp | bytes[7]; *valp = *valp << 8;
  *valp = *valp | bytes[6]; *valp = *valp << 8;
  *valp = *valp | bytes[5]; *valp = *valp << 8;
  *valp = *valp | bytes[4]; *valp = *valp << 8;
  *valp = *valp | bytes[3]; *valp = *valp << 8;
  *valp = *valp | bytes[2]; *valp = *valp << 8;
  *valp = *valp | bytes[1]; *valp = *valp << 8;
  *valp = *valp | bytes[0];

  return TS_OK;
}

static TSResult write_size(FILE *f, size_t val)
{
#if SIZE_MAX > UINT64_MAX
  /* This is only true if size_t is larger than 64-bits, in which case
   * someone needs to add a flag to specify the larger size and dump a
   * 128-bit (or whatevever) size entry to the file.
   */
#  error Code expects sizes to fit within 64-bits unsigned integer
#endif

  RETURN_WITH_RESULT( write_u64(f,(uint64_t)val) );
}

static TSResult read_size(FILE *f, size_t *valp)
{
  uint64_t val;

#if SIZE_MAX > UINT64_MAX
  /* This is only true if size_t is larger than 64-bits, in which case
   * someone needs to add a flag to specify the larger size and read a
   * 128-bit (or whatevever) size entry to the file.
   */
#  error Code expects sizes to fit within 64-bits unsigned integer
#endif

  RETURN_ON_ERROR( read_u64(f,&val) );

  if ((SIZE_MAX < UINT64_MAX) && (val > SIZE_MAX))
  {
    TS_RAISE_ERROR3(TS_ERR_IO_DECODE_ERROR,
        "Expected a size varint, but value is outside of range");
  }

  *valp = (size_t)val;

  return TS_OK;
}

static MAYBE_UNUSED TSResult write_fixed_s32(FILE *f, int32_t i)
{
  RETURN_WITH_RESULT(write_fixed_u32(f, (uint32_t)i));
}

static MAYBE_UNUSED TSResult write_fixed_s64(FILE *f, int64_t i)
{
  RETURN_WITH_RESULT(write_fixed_u64(f, (uint64_t)i));
}

/* FIXME: this doesn't check for systems that store doubles in a
 * different endian format from integers.  These are rare.  A better
 * solution is probably to marshall the FP bits in that case (which is
 * also not implemented).
 */
#if DBL_IS_INT_ALIGNED
STATIC_ASSERT( sizeof(uint64_t) == sizeof(double) );
static TSResult write_dbl(FILE *f, double val) 
{
  uint64_t bits;
  memcpy(&bits, &val, sizeof(val));

  RETURN_WITH_RESULT( write_fixed_u64(f,bits) );
}

static TSResult read_dbl(FILE *f, double *valp) 
{
  uint64_t bits = 0;

  RETURN_ON_ERROR( read_fixed_u64(f,&bits) );
  memcpy(valp, &bits, sizeof(*valp));

  return TS_OK;
}
#else
#  error Need to implement way to marshall FP bits
#endif

static TSResult write_state_magic(FILE *outf)
{
  size_t n = sizeof(TSCH_STATE_FILE_MAGIC)-1;
  if (fwrite(TSCH_STATE_FILE_MAGIC, sizeof(char), n, outf) != n)
  {
    TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED,
        "Error encountered in write: %s", strerror(errno));
  }
  return TS_OK;
}

static TSResult encode_column(ts_chem_driver *driver, FILE *outf, size_t col, size_t skip)
{
  size_t nrow, i, i0 = 1;
  int v0 = TS_LATTICE_SPIN(driver,i0,col);

#if (INT_MIN < INT32_MIN) || (INT_MAX > INT32_MAX)
  /* Ensure that lattice values fit in 32-bit signed integers.
   * If not, we need to either increase the version number and 
   * have the read/write code expect larger signed integers, or
   * add a flag which tells the read/write code to expect larger
   * signed integers.
   */
#  error Code expects lattice values to fit within 32-bit signed integers
#endif

  nrow = driver->latt.nrow;

  for (i=i0+skip; i < nrow-1; i+=skip)
  {
    int v = TS_LATTICE_SPIN(driver,i,col);

    if (v == v0)
    {
      continue;
    }

    RETURN_ON_ERROR(write_s32(outf, v0));
    RETURN_ON_ERROR(write_size(outf, (i-i0)/skip));

    i0 = i;
    v0 = v;
  }

  if (i-i0 >= skip)
  {
    RETURN_ON_ERROR(write_s32(outf, v0));
    RETURN_ON_ERROR(write_size(outf, (i-i0)/skip));
  }

  /* write two zeros to indicate the end of a column */
  RETURN_ON_ERROR(write_u32(outf, 0));
  RETURN_ON_ERROR(write_size(outf, 0));

  return TS_OK;
}

static TSResult decode_column(ts_chem_driver *driver, FILE *inf, size_t col, size_t skip)
{
  size_t nrow, i;

#if (INT_MIN < INT32_MIN) || (INT_MAX > INT32_MAX)
  /* Ensure that lattice values fit in 32-bit signed integers.
   * If not, we need to either increase the version number and 
   * have the read/write code expect larger signed integers, or
   * add a flag which tells the read/write code to expect larger
   * signed integers.
   */
#  error Code expects lattice values to fit within 32-bit signed integers
#endif

  nrow = driver->latt.nrow;

  /* if skip == 1, we should fill the entire column, otherwise we zero it out 
   * just to give it a consistent state
   */
  if (skip > 1)
  {
    for (i=0; i < nrow; i++)
    {
      TS_LATTICE_SPIN(driver,i,col) = 0;
    }
  }

  i = 1;
  for(;;)
  {
    int32_t spin;
    size_t count, j;
    spin = 0;
    count = 0;
    RETURN_ON_ERROR(read_s32(inf, &spin));
    RETURN_ON_ERROR(read_size(inf, &count));

    if (spin == 0 && count == 0)
    {
      if (i >= nrow-1)
      {
        return TS_OK;
      }

      TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "RLE for column %zu with skip %zu reaches row %zu instead of column length (%zu)",
          col, skip, i, nrow-1);
    }

    if (i+skip*count > nrow)
    {
      TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "RLE for column %zu with skip %zu exceeds allowed column length %zu",
          col, skip, nrow-1);
    }

    for(j=0; j < count; j++)
    {
      assert( i < nrow );
      TS_LATTICE_SPIN(driver, i, col) = spin;
      i += skip;
    }
  }

  return TS_OK;
}

static TSResult write_state_header(ts_chem_driver *driver, FILE *outf, uint32_t version, uint32_t flags)
{
  double energy;

  energy = 0.0;
  RETURN_ON_ERROR(ts_chem_driver_energy(driver, &energy, 0));

  RETURN_ON_ERROR(write_state_magic(outf));
  RETURN_ON_ERROR(write_fixed_u32(outf, version));
  RETURN_ON_ERROR(write_fixed_u32(outf, flags));
  RETURN_ON_ERROR(write_u64(outf, driver->latt.nsteps));
  RETURN_ON_ERROR(write_dbl(outf, driver->latt.time));
  RETURN_ON_ERROR(write_dbl(outf, energy));

  if ((driver->latt.nrow > UINT64_MAX) || (driver->latt.ncol > UINT64_MAX))
  {
    TS_RAISE_ERROR3(TS_ERR_UNSUPPORTED_OPERATION,
        "Lattice is too large for the state I/O code");
  }

  RETURN_ON_ERROR(write_u32(outf, driver->latt.nrow));
  RETURN_ON_ERROR(write_u32(outf, driver->latt.ncol));

  return TS_OK;
}

TSResult tsch_write_state(ts_chem_driver *driver, FILE *outf)
{
  size_t j,npf;
  uint32_t version,flags;

  npf = ts_chem_driver_num_protofilaments(driver);

  version = TSCH_STATE_VERSION_DEFAULT;
  flags = TSCH_STATE_FLAG_DEFAULT;

  RETURN_ON_ERROR( write_state_header(driver, outf, version, flags) );

  for(j=0; j < npf; j++)
  {
    RETURN_ON_ERROR(encode_column(driver, outf, 2*j+0, 1));
    RETURN_ON_ERROR(encode_column(driver, outf, 2*j+1, 2));
  }

  /* ideally we'd write a checksum or something here... */

  return TS_OK;
}

TSResult check_state_magic(FILE *inf)
{
  char buf[sizeof(TSCH_STATE_FILE_MAGIC)-1];
  size_t nread, n = sizeof(buf);

  if (nread = fread(buf, sizeof(char), n, inf), nread != n)
  {
    if ((nread == 0) && feof(inf))
    {
      EOF_ENCOUNTERED();
    }
    else
    {
      TS_RAISE_ERRORF(TS_ERR_READ_NOT_FINISHED,
          "Error encountered in read: %s", strerror(errno));
    }
  }

  if (memcmp(buf,TSCH_STATE_FILE_MAGIC,sizeof(buf)) != 0)
  {
    TS_RAISE_ERROR3(TS_ERR_IO_DECODE_ERROR,
        "Invalid header: did not recognize magic string");
  }

  return TS_OK;
}

TSResult tsch_read_header(FILE *inf, struct tsch_save_header *header, uint32_t *nrowp, uint32_t *ncolp)
{
  uint32_t nrow,ncol;
  TSResult err;

  err = ON_ERROR_ADD_FRAME(check_state_magic(inf));
  if (!IS_OKAY(err))
  {
    return err;
  }

  RETURN_ON_ERROR(read_fixed_u32(inf, &header->version));
  RETURN_ON_ERROR(read_fixed_u32(inf, &header->flags));

  RETURN_ON_ERROR(read_u64(inf, &header->nsteps));
  RETURN_ON_ERROR(read_dbl(inf, &header->time));
  RETURN_ON_ERROR(read_dbl(inf, &header->energy));

  RETURN_ON_ERROR(read_u32(inf, &nrow));
  RETURN_ON_ERROR(read_u32(inf, &ncol));

  if (nrowp)
  {
    *nrowp = nrow;
  }

  if (ncolp)
  {
    *ncolp = ncol;
  }

  return TS_OK;
}

TSResult tsch_read_state(ts_chem_driver *driver, FILE *inf, struct tsch_save_header *header)
{
  size_t j,npf;
  uint32_t nrow,ncol;

  npf = ts_chem_driver_num_protofilaments(driver);

  /**/
  nrow = ncol = 0;
  RETURN_ON_ERROR( tsch_read_header(inf, header, &nrow, &ncol) );

  if ((nrow != driver->latt.nrow) || (ncol != driver->latt.ncol))
  {
    TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
        "Expected %zu rows, %zu columns, but save has %lu rows, %lu columns",
        driver->latt.nrow, driver->latt.ncol, (unsigned long)nrow,(unsigned long)ncol);
  }

  npf = ncol/2;
  for(j=0; j < npf; j++)
  {
    RETURN_ON_ERROR(decode_column(driver, inf, 2*j+0, 1));
    RETURN_ON_ERROR(decode_column(driver, inf, 2*j+1, 2));
  }

  /* ideally we'd read and verify a checksum or something here... */

  return TS_OK;
}

struct tsch_snapshot *tsch_snapshot_new(ts_chem_driver *driver)
{
  struct tsch_snapshot *snap;
  size_t nsites = driver->latt.num_sites;

  snap = ts_alloc(NO_ALLOCATOR, sizeof(*snap) + nsites * sizeof(snap->sites[0]));
  if (snap == NULL)
  {
    return NULL;
  }

  snap->num_sites = nsites;
  memcpy(snap->sites, driver->latt.sites, nsites * sizeof(snap->sites[0]));

  return snap;
}

void tsch_snapshot_free(struct tsch_snapshot *snap)
{
  ts_free(NO_ALLOCATOR, snap);
}

TSResult tsch_snapshot_update(ts_chem_driver *driver, struct tsch_snapshot *snap)
{
  size_t nsites = driver->latt.num_sites;
  if (snap->num_sites != nsites)
  {
    TS_RAISE_ERROR3(TS_ERR_INVALID_ARRAY_SIZE,
        "snapshot size does not match ChemDriver size");
  }

  memcpy(snap->sites, driver->latt.sites, nsites * sizeof(snap->sites[0]));
  return TS_OK;
}

TSResult tsch_snapshot_num_changes(ts_chem_driver *driver, struct tsch_snapshot *snap, size_t *nump)
{
  const spin_t *restrict dspins = driver->latt.sites;
  const spin_t *restrict sspins = snap->sites;
  size_t i,count,n= driver->latt.num_sites;

  if (snap->num_sites != n)
  {
    TS_RAISE_ERROR3(TS_ERR_INVALID_ARRAY_SIZE,
        "snapshot size does not match ChemDriver size");
  }

  for (count=0,i=0; i < n; i++)
  {
    if (dspins[i].val != sspins[i].val) { count++; }
  }

  *nump = count;

  return TS_OK;
}

enum TSCH_DELTA_STATE { TSCH_DELTA_SAME, TSCH_DELTA_DIFF };
TSResult tsch_snapshot_write_delta(ts_chem_driver *driver, const struct tsch_snapshot *snap, FILE *outf)
{
  size_t i0,i;
  uint32_t version,flags;
  size_t nsites = driver->latt.num_sites;
  const spin_t *restrict dspins = driver->latt.sites;
  const spin_t *restrict sspins = snap->sites;
  enum TSCH_DELTA_STATE state;
  spin_t v0;

  if (snap->num_sites != nsites)
  {
    TS_RAISE_ERROR3(TS_ERR_INVALID_ARRAY_SIZE,
        "snapshot size does not match ChemDriver size");
  }

  version = TSCH_STATE_VERSION_DEFAULT;
  flags = TSCH_STATE_FLAG_DEFAULT | TSCH_STATE_FLAG_DELTA;

  RETURN_ON_ERROR( write_state_header(driver, outf, version, flags) );
  RETURN_ON_ERROR( write_size(outf, nsites) );

  state = TSCH_DELTA_SAME;
  for (i0=0,i=1; i < nsites; i++)
  {
    if (state == TSCH_DELTA_SAME)
    {
      if (dspins[i].val == sspins[i].val)
      {
        continue;
      }

      write_s32(outf, -1);
      write_size(outf, i-i0);
      state = TSCH_DELTA_DIFF;
      i0 = i;
      v0 = dspins[i];
    }
    else
    {
      assert( state == TSCH_DELTA_DIFF );
      if (dspins[i].val == v0.val)
      {
        continue;
      }

      write_s32(outf, v0.val);
      write_size(outf,i-i0);
      i0 = i;
      v0 = dspins[i];
      if (dspins[i].val == sspins[i].val)
      {
        state = TSCH_DELTA_SAME;
      }
    }
  }

  if (i - i0 > 1)
  {
    if (state == TSCH_DELTA_SAME)
    {
      write_s32(outf, -1);
      write_size(outf, i-i0);
    }
    else
    {
      write_s32(outf, v0.val);
      write_size(outf, i-i0);
    }
  }

  /* ideally we'd write a checksum or something here... */

  return TS_OK;
}

TSResult read_delta(ts_chem_driver *driver, const struct tsch_snapshot *snap, FILE *inf)
{
  (void)driver;
  (void)snap;
  (void)inf;
  return TS_OK;
}

#if 0
TSResult tsch_snapshot_take(ts_chem_driver *driver, struct tsch_snapshot **snapp);
TSResult tsch_snapshot_num_changes(ts_chem_driver *driver, struct tsch_snapshot *snap, size_t *nchangesp);
TSResult tsch_snapshot_write_delta(ts_chem_driver *driver, struct tsch_snapshot *snap, FILE *outf);
TSResult tsch_snapshot_read_delta(ts_chem_driver *driver, struct tsch_snapshot *snap, FILE *inf);
#endif /* 0 */

static inline unsigned char *
u16_to_buf(unsigned char *buf, uint16_t x)
{
  buf[0] = (x >> 8) & 0xff;
  buf[1] = x & 0xff;
  return buf+2;
}

static inline unsigned char *
buf_to_u16(unsigned char *buf, uint16_t *restrict xp)
{
  *xp = (buf[0] << 8) | buf[1];
  return buf+2;
}

static inline unsigned char *
u64_to_buf(unsigned char *buf, uint64_t x)
{
  buf[0] = (x >> 56) & 0xff;
  buf[1] = (x >> 48) & 0xff;
  buf[2] = (x >> 40) & 0xff;
  buf[3] = (x >> 32) & 0xff;
  buf[4] = (x >> 24) & 0xff;
  buf[5] = (x >> 16) & 0xff;
  buf[6] = (x >>  8) & 0xff;
  buf[7] = (x      ) & 0xff;
  return buf+8;
}

static inline unsigned char *
buf_to_u64(unsigned char *buf, uint64_t *restrict xp)
{
  *xp =       (uint64_t)buf[0] << 56;
  *xp = *xp | (uint64_t)buf[1] << 48;
  *xp = *xp | (uint64_t)buf[2] << 40;
  *xp = *xp | (uint64_t)buf[3] << 32;
  *xp = *xp | (uint64_t)buf[4] << 24;
  *xp = *xp | (uint64_t)buf[5] << 16;
  *xp = *xp | (uint64_t)buf[6] <<  8;
  *xp = *xp | (uint64_t)buf[7]      ;
  return buf+8;
}

static inline unsigned char *
pad_buf(unsigned char *buf, size_t n)
{
  size_t i;
  for (i=0; i < n; i++) { buf[i] = 0; }
  return buf+n;
}

static TSResult
wrbuf(FILE *f, unsigned char *buf, size_t n)
{
  size_t nw;

  errno = 0;
  nw = fwrite(buf, sizeof *buf, n, f);
  if (nw < n)
  {
    if (errno != 0)
    {
      TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED, "write did not finish (error 0x%x): %s",
          errno, strerror(errno));
    }
    else
    {
      TS_RAISE_ERROR_MESG(TS_ERR_WRITE_NOT_FINISHED, "write did not finish (unknown error)");
    }
  }
  return TS_OK;
}

static TSResult
rdbuf(FILE *f, unsigned char *buf, size_t n)
{
  size_t nw;

  errno = 0;
  while (n > 0)
  {
    nw = fread(buf, sizeof *buf, n, f);
    if (nw == n)
    {
      break;
    }

    if (feof(f))
    {
      return ts_new_result(TS_IO_EOF);
    }

    if (ferror(f))
    {
      TS_RAISE_ERRORF(TS_ERR_WRITE_NOT_FINISHED, "read did not finish (error 0x%x): %s",
          errno, strerror(errno));
    }

    buf += nw;
    n -= nw;
  }

  return TS_OK;
}

static inline unsigned char *
ifclog_to_buf(unsigned char *buf, struct tsch_ifclog_entry *restrict entry)
{
  buf = u16_to_buf(buf, entry->ifc);
  buf = u16_to_buf(buf, entry->move);
  buf = u16_to_buf(buf, entry->rel_row);
  buf = u16_to_buf(buf, entry->col);
  return buf;
}

static inline unsigned char *
buf_to_ifclog(unsigned char *buf, struct tsch_ifclog_entry *restrict entry)
{
  buf = buf_to_u16(buf, &entry->ifc);
  buf = buf_to_u16(buf, &entry->move);
  buf = buf_to_u16(buf, &entry->rel_row);
  buf = buf_to_u16(buf, &entry->col);
  return buf;
}

TSResult
tsch_ifclog_write(ts_chem_driver *drv, FILE *f)
{
  size_t i, n, off, entry_sz;
  unsigned char buf[1024], *b, *spill;

  if (!tsch_ifclog_is_logging(drv))
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "cannot write ifclog: ifclog does not exist");
  }

  b = &buf[0];

  /* calculate size of entry on the wire and the
   * buffer spill line (when we need to write the buffer)
   */
  {
    ptrdiff_t batch;
    struct tsch_ifclog_entry entry = { 0 };
    entry_sz = ifclog_to_buf(b, &entry) - b;
    batch = sizeof buf / entry_sz;
    spill = &buf[batch * entry_sz];
  }

  /* check constraints */
  n = drv->ifclog.n;
  if (n > UINT64_MAX)
  {
    TS_RAISE_ERRORF(TS_ERR_UNIMPLEMENTED,
        "size of ifclog (%zu) exceeds 64-bits",
        drv->ifclog.n);
  }

  if (entry_sz > UINT16_MAX)
  {
    TS_RAISE_ERROR_MESG(TS_ERR_UNIMPLEMENTED,
        "entry size exceeds 16 bits");
  }

  /* write header:
   *    magic                   8 bytes
   *    version                 2 bytes
   *    record size             2 bytes
   *    <padding>               4 bytes
   *    number of records       8 bytes
   */

  // write magic
  const char *magic = TSCH_IFCLOG_FILE_MAGIC;
  size_t magic_len = strlen(magic);
  memcpy(b, magic, magic_len);
  b += magic_len;

  b = u16_to_buf(b, 0x0001);                    // version
  b = u16_to_buf(b, (uint16_t)entry_sz);        // record size
  b = pad_buf(b, 4);                            // padding
  b = u64_to_buf(b, drv->ifclog.n);             // number of records

  RETURN_ON_ERROR(wrbuf(f, buf, b - &buf[0]));

  b = &buf[0];
  for (i=0; i < n; i++)
  {
    struct tsch_ifclog_entry *entry;
    entry = tsch_ifclog_at(drv, i);

    b = ifclog_to_buf(b, entry);
    assert(b <= &buf[sizeof buf]);

    if (b >= spill)
    {
      RETURN_ON_ERROR(wrbuf(f, buf, b - &buf[0]));
      b = &buf[0];
    }
  }

  /* flush any remaining bits */
  if (b > &buf[0])
  {
    RETURN_ON_ERROR(wrbuf(f, buf, b - &buf[0]));
  }

  if (fflush(f) != 0)
  {
    TS_RAISE_ERRORF(TS_ERR_IO_ERROR,
        "Unexpected error 0x%02x flushing output: %s",
        (unsigned int)errno, strerror(errno));
  }

  return TS_OK;
}

TSResult
tsch_ifclog_read(ts_chem_driver *drv, FILE *f)
{
  size_t i, n, off, entry_sz;
  unsigned char buf[64], *b, *spill;
  TSResult ret;

  {
    struct tsch_ifclog_entry etmp = { 0 };
    entry_sz = buf_to_ifclog(&buf[0], &etmp) - &buf[0];
  }

  /* check the magic string */
  {
    size_t msz = strlen(TSCH_IFCLOG_FILE_MAGIC);
    ret = rdbuf(f, buf, msz);
    RETURN_ON_ERROR(ret);

    if (IS_IO_EOF(ret))
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: file is too short");
    }

    if (memcmp(buf, TSCH_IFCLOG_FILE_MAGIC, msz) != 0)
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: invalid file magic string");
    }
  }

  /* read the version, entry size, 4 bytes of padding, and the number of
   * records
   */
  {
    uint16_t vers, esz;
    uint64_t nrec;

    b = &buf[0];
    ret = rdbuf(f, b, 16);
    RETURN_ON_ERROR(ret);

    if (IS_IO_EOF(ret))
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: could not read header");
    }

    vers = esz = 0;
    b = buf_to_u16(b, &vers);
    b = buf_to_u16(b, &esz);

    if (vers != 1)
    {
      TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: unsupported ifclog version %d", vers);
    }

    if (esz != entry_sz)
    {
      TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: entry size mismatch (expected %zu, found %d)\n",
          entry_sz, (int)esz);
    }

    if (entry_sz > sizeof buf)
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: internal buffer is too small");
    }

    b += 4; /* four bytes of padding */

    nrec = 0;
    b = buf_to_u64(b, &nrec);

    if (nrec > SIZE_MAX)
    {
      TS_RAISE_ERROR_MESG(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: too many entries to decode on this system");
    }

    n = nrec;
  }

  if (!tsch_ifclog_is_logging(drv) || (tsch_ifclog_capa(drv) < n))
  {
    RETURN_ON_ERROR(tsch_ifclog_resize(drv, n));
  }
  tsch_ifclog_clear(drv);

  for (i=0; i < n; i++)
  {
    struct tsch_ifclog_entry *entry;

    b = &buf[0];
    ret = rdbuf(f, b, entry_sz);
    RETURN_ON_ERROR(ret);

    if (IS_IO_EOF(ret))
    {
      TS_RAISE_ERRORF(TS_ERR_IO_DECODE_ERROR,
          "error decoding stream: stream cut short after %zu entries (expected %zu)",
          i+1,n);
    }

    drv->ifclog.n++;
    entry = tsch_ifclog_at(drv,i);
    b = buf_to_ifclog(b, entry);
  }

  return TS_OK;
}




