#include "tubesheet.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>

#include "ts_common.h"
#include "ts_common_algo.h"
#include "ts_numerical.h"
#include "ts_dispatch.h"

#define kcalloc(N,Z)  ts_calloc(NO_ALLOCATOR,N,Z)
#define kmalloc(Z)    ts_alloc(NO_ALLOCATOR,Z)
#define krealloc(P,Z) ts_realloc(NO_ALLOCATOR,P,Z)
#define kfree(P)      ts_free(NO_ALLOCATOR,P)

#include "ts_units.h"
#include "ts_errors.h"
#include "ts_dof_map.h"
#include "ts_bonds.h"
#include "ts_integrators.h"

#include "ts_tables_priv.h"

#include "ts_malloc_protect.h"

#define FREE_AND_NULLIFY(A,x) \
  do { ts_safe_free(A,x); x = NULL; } while(0)

DEFINE_TABLE(bond_type, bt, ts_bond_type_table, ts_bond_type);
#define HASH_BT_TABLE(tbl) ((khash_t(bond_type)*)(tbl))
#define BT_FOREACH(tbl, kvar, vvar, code) \
  kh_foreach(HASH_BT_TABLE(tbl), kvar, vvar, code)



DEFINE_TABLE(potentials, potentials, ts_potential_table, Potential*);
#define POTENTIALS_FOREACH(tbl, kvar, vvar, code) \
  kh_foreach((khash_t(potentials)*)tbl, kvar, vvar, code)


void ts_log_message(const ts_sim* sim,
    enum TS_LOGGING_FLAG component,
    enum TS_LOGGING_LEVEL level,
    const char* mesg, ...)
{
  va_list ap;
  va_start(ap, mesg);
  const char* component_str = NULL;
  const char* loglevel_str = NULL;

  switch(component)
  {
    case TS_LOG_INTEGRATOR: component_str = "DYNA"; break;
    case TS_LOG_MINIMIZER:  component_str = "MINI"; break;
    case TS_LOG_NONBONDS:   component_str = "NONB"; break;
    case TS_LOG_IO:         component_str = "IO  "; break;
    default: component_str = "????";
  }

  switch (level) {
    case TS_LOG_VERBOSE: loglevel_str = "V"; break;
    case TS_LOG_INFO:    loglevel_str = "I"; break;
    case TS_LOG_WARN:    loglevel_str = "W"; break;
    case TS_LOG_ERROR:   loglevel_str = "E"; break;
    case TS_LOG_ABORT:   loglevel_str = "A"; break;
    default: component_str = "?";
  }

  fprintf(sim->logging_output, "[%4s:%1s] ",component_str,loglevel_str);
  vfprintf(sim->logging_output, mesg, ap);
  va_end(ap);
}

STATIC_ASSERT( sizeof(uint32_t) == sizeof(unsigned int) );

ts_sim* ts_sim_new(void)
{
  ts_sim* sim = ts_alloc(NO_ALLOCATOR, sizeof(*sim) );

  sim->logging_flags = TS_DEFAULT_LOGGING_FLAGS;
  sim->logging_output = stderr;

  sim->rng_state = NULL;

  sim->rand_forces = NULL;
  sim->num_rand_forces = 0;

  sim->rand_torques = NULL;
  sim->num_rand_torques = 0;

  sim->max_particles = 0;
  sim->num_particles = 0;
  sim->particles = NULL;

  sim->max_rigid_bodies = 0;
  sim->num_rigid_bodies = 0;
  sim->rigid_bodies = NULL;

  sim->vl_pos = NULL;
  sim->vl_vel = NULL;
  sim->vl_forces = NULL;
  sim->vl_eforces = NULL;
  sim->vl_work = NULL;
  sim->vl_maxdisp_origin = NULL;

  sim->ql_ori = NULL;
  sim->ql_prev_ori = NULL;
  sim->ql_work = NULL;

  sim->work_array = NULL;

#ifdef TS_DISPATCH_ENABLED
  sim->particle_locks = NULL;
#endif /* TS_DISPATCH_ENABLED */

  sim->dof_map = ts_alloc(NO_ALLOCATOR, sizeof(*sim->dof_map));
  ts_dof_map_init_empty(sim->dof_map);

  sim->integrator = TS_INTEGRATOR_NONE;

  sim->parameters = ts_param_table_new();
  sim->bond_types = bt_table_new();
  sim->potentials = potentials_table_new();

  datavector_init( &(sim->particle_types)  , sizeof(ts_particle_type)         , DEFAULT_PTYPE_TABLE_SIZE );
  datavector_init( &(sim->rigid_body_types), sizeof(struct ts_rigid_body_type), DEFAULT_PTYPE_TABLE_SIZE );

  datavector_init( &(sim->bond_patterns1)   , sizeof(ts_bond_pattern)      , DEFAULT_BONDLIST_SIZE    );
  datavector_init( &(sim->nonbond_patterns1), sizeof(ts_bond_pattern)      , DEFAULT_BONDLIST_SIZE    );

  datavector_init( &sim->extra_terms, sizeof (struct ts_potential_term), DEFAULT_BONDLIST_SIZE );

  datavector_init( &(sim->constraints)     , sizeof(ts_constraint_entry)      , DEFAULT_BONDLIST_SIZE    );
  datavector_init( &(sim->callbacks)       , sizeof(ts_integrator_callback*)  , DEFAULT_PTYPE_TABLE_SIZE );

  datavector_init( &sim->particle_sets, sizeof (struct ts_index_set), TS_INDEX_SET_DEFAULT_CAP );

  ts_potential_term_list_init(&sim->bondlist);
  ts_potential_term_list_init(&sim->nonbond_list);

  sim->dt   = 0.0;
  sim->temp = 0.0;

  sim->time = 0.0;
  sim->integrator_steps = 0;

  sim->shake_tol = DEFAULT_SHAKE_TOL;

  sim->nonbond_cutoff = 0.0;
  sim->nonbond_padding = 0.0;

  sim->nonbonds_enabled = 1;
  for (int i=0; i < TS_BOND_TERM_MAX; i++)
  {
    sim->bonds_enabled[i] = 1;
  }

  sim->PE     = 0.0;
  sim->KE     = 0.0;
  sim->totalE = 0.0;

  return sim;
}

static void remove_all_bond_types(ts_sim* sim);

void ts_sim_delete(ts_sim* sim)
{
  const char *key;
  Potential *pot;

  remove_all_bond_types(sim);

  bt_table_delete(sim->bond_types);

  POTENTIALS_FOREACH(sim->potentials,
      key,pot,
      (ts_free(NO_ALLOCATOR,(char*)key), ts_potential_decref(pot)));
  potentials_table_delete(sim->potentials);

  ts_param_table_delete(sim->parameters);

  ts_rng_free(sim->rng_state);
  ts_free_rand_forces(sim);
  ts_dealloc_particles(sim);

  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_pos);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_vel);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_forces);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_eforces);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_work);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_maxdisp_origin);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->work_array);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->ql_ori);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->ql_prev_ori);
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->ql_work);

#ifdef TS_DISPATCH_ENABLED
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->particle_locks);
#endif /* TS_DISPATCH_ENABLED */

  if (sim->dof_map)
  {
    ts_dof_map_finalize(sim->dof_map);
    ts_free(NO_ALLOCATOR, sim->dof_map);
    sim->dof_map = NULL;
  }

  ts_sim_free_bond_patterns(sim);

  ts_potential_term_list_finalize(&sim->bondlist);
  ts_potential_term_list_finalize(&sim->nonbond_list);

  datavector_free( &(sim->particle_types) );
  datavector_free( &(sim->callbacks)      );

  {
    DATAVECTOR_FOREACH(ts_constraint_entry, cons, &sim->constraints)
    {
      ts_potential_decref((Potential *)cons->constraint);
    }

    datavector_free( &(sim->constraints)    );
  }

  {
    DATAVECTOR_FOREACH(struct ts_index_set, set, &sim->particle_sets)
    {
      ts_index_set_finalize(set);
    }
    datavector_free( &sim->particle_sets );
  }

  datavector_free( &sim->extra_terms );

  ts_finalize_bond_patterns(&sim->bond_patterns1);
  ts_finalize_bond_patterns(&sim->nonbond_patterns1);
}

TSResult ts_set_parameter(ts_sim* sim, const char* key, double val)
{
  RETURN_WITH_RESULT( ts_param_table_set(sim->parameters, key, val) );
}

TSResult ts_get_parameter(ts_sim* sim, const char* key, double* val_ptr)
{
  RETURN_WITH_RESULT( ts_param_table_get(sim->parameters, key, val_ptr) );
}

int ts_has_parameter(ts_sim* sim, const char* key)
{
  return ts_param_table_exists(sim->parameters, key);
}

TSResult ts_param_table_get(ts_param_table* tbl, const char* key, double *valp);
int ts_param_table_exists(ts_param_table* tbl, const char* key);

TSResult ts_foreach_parameter(ts_sim* sim, ts_parameter_func func, void* data)
{
  RETURN_WITH_RESULT( ts_param_table_foreach(sim->parameters,func,data) );
}

TSResult ts_add_potential(ts_sim* sim, const char* name, Potential* pot)
{
  char *kdup;
  Potential *pswap = pot;

  if (potentials_table_replace(sim->potentials, name, &pswap))
  {
    ts_potential_incref(pot);
    ts_potential_decref(pswap);
    return TS_OK;
  }

  if (kdup = ts_strdup(NO_ALLOCATOR, name), !kdup)
  {
    return TS_NEW_ERROR3(
        TS_ERR_OUT_OF_MEMORY,
        "out of memory duplicating parameter string");
  }

  if (!potentials_table_set(sim->potentials, kdup, pot))
  {
    ts_free(NO_ALLOCATOR, kdup);
    return TS_NEW_ERROR( TS_ERR_ADDING_PARAMETER );
  }
  ts_potential_incref(pot);

  return TS_OK;
}

Potential *ts_get_potential(ts_sim *sim, const char *name)
{
  Potential *pot = NULL;
  potentials_table_get(sim->potentials, name, &pot);
  return pot;
}

TSResult ts_foreach_potential(ts_sim* sim,
    TSResult (*func)(void*, const char*, Potential*))
{
  UNUSED_PARAM(sim);
  UNUSED_PARAM(func);
  return TS_OK;
}

void ts_sim_potential_iter_init(ts_sim *sim, struct ts_sim_potential_iter *iter)
{
  UNUSED_PARAM(sim);
  iter->state = 0;
  iter->key = NULL;
  iter->pot = NULL;
}

int ts_sim_potential_iter_next(ts_sim *sim, struct ts_sim_potential_iter *iter)
{
  khint_t i,end;
  khash_t(potentials)* ht;

  ht = (khash_t(potentials)*)sim->potentials;
  for (i = (khint_t)iter->state, end = kh_end(ht); i < end; i++)
  {
    if (!kh_exist(ht,i))
    {
      continue;
    }

    iter->key = kh_key(ht, i);
    iter->pot = kh_val(ht, i);
    iter->state = (int64_t)(i+1);
    return 1;
  }

  return 0;
}

TSResult ts_alloc_rand_forces(ts_sim* sim)
{

#define CLEANUP_ON_ERROR {}

  if (sim->num_rand_forces < 3*sim->num_particles)
  {
    sim->rand_forces = ts_realloc(NO_ALLOCATOR, sim->rand_forces, 3*sim->num_particles * sizeof(double));
    ERROR_ON_NULL_ALLOC( sim->rand_forces );
    tsAssert( IS_SSE_ALIGNED( sim->rand_forces ) );
    sim->num_rand_forces = 3*sim->num_particles;
  }

  if (sim->num_rand_torques < 3*sim->num_rigid_bodies)
  {
    sim->rand_torques = ts_realloc(NO_ALLOCATOR, sim->rand_torques, 3*sim->num_rigid_bodies*sizeof(double));
    ERROR_ON_NULL_ALLOC( sim->rand_torques );
    tsAssert( IS_SSE_ALIGNED( sim->rand_torques ) );
    sim->num_rand_torques = 3*sim->num_rigid_bodies;
  }

#undef CLEANUP_ON_ERROR

  return TS_OK;
}

void ts_free_rand_forces(ts_sim* sim)
{
  if (sim->rand_forces != NULL)
  {
    ts_free(NO_ALLOCATOR, sim->rand_forces);
  }
}

static TSResult ptype_setup_collision_coeffs(ts_sim* sim, ts_particle_type* ptData)
{
  if (ptData->collfreq > 0.0)
  {
    ERROR_IF( TS_ERR_BAD_DT,   sim->dt   <= 0.0 );
    ERROR_IF( TS_ERR_BAD_TEMP, sim->temp <= 0.0 );

    double mod = 0.5*ptData->collfreq*sim->dt;
    double coll_numer = 1.0 - mod;
    double coll_denom = 1.0 + mod;
    ptData->coll_coeff2 = 1.0/coll_denom;
    ptData->coll_coeff1 = coll_numer * ptData->coll_coeff2;
    ptData->rand_coeff = sqrt(2.0*ptData->mass*ptData->collfreq*sim->temp);

    ptData->brownian_coeff = sim->dt * ptData->inv_mass * ptData->inv_collfreq;

    ptData->bd_rand_coeff = sqrt(2.0 * sim->temp * ptData->brownian_coeff);
  }
  else
  {
    ptData->coll_coeff1 = 1.0;
    ptData->coll_coeff2 = 1.0;
    ptData->rand_coeff  = 0.0;
    ptData->brownian_coeff = 0.0;
  }

  return TS_OK;
}

TSResult ts_setup_collision_coeffs(ts_sim *sim)
{
  for (size_t i=1; i <= sim->particle_types.size; i++)
  {
    ts_particle_type* ptype = ts_get_particle_type(sim, i);
    RETURN_ON_ERROR( ptype_setup_collision_coeffs(sim, ptype) );
  }

  for (size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    ts_rigid_body* body = ts_get_rigid_body(sim, i);
    ts_rigid_body_update_parameters(sim, body);
  }

  return TS_OK;
}

TSResult ts_add_particle_type(ts_sim *sim, ts_particle_type *ptData, size_t *nptr)
{
  ptData->inv_mass = 1.0 / ptData->mass;
  ptData->inv_sqrt_mass = sqrt(ptData->inv_mass);
  ptData->inv_collfreq = ptData->collfreq > 0.0 ?  1.0 / ptData->collfreq : 0.0;

  ptype_setup_collision_coeffs(sim,ptData);

  if (!datavector_add(&(sim->particle_types), ptData)) return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  if (nptr != NULL) { *nptr = sim->particle_types.size; }

  return TS_OK;
}

MAYBE_UNUSED
static inline void* bondtype_as_voidp(ts_bond_type bt)
{
  union { void* ptr; ts_bond_type bt; } value = { .ptr = NULL };
  value.bt = bt;
  return value.ptr;
}

MAYBE_UNUSED
static inline ts_bond_type voidp_as_bondtype(void* ptr)
{
  union { void* ptr; ts_bond_type bt; } value = { .ptr = ptr };
  return value.bt;
}

STATIC_ASSERT( sizeof(void*) >= sizeof(ts_bond_type) );
TSResult ts_add_bond_type(ts_sim* sim, const char* btName, ts_bond_type bt)
{
  char* key = ts_strdup(NO_ALLOCATOR, btName);
  if (key == NULL) {
    return TS_NEW_ERROR3( TS_ERR_OUT_OF_MEMORY,
        "out of memory allocating key for bond type");
  }

  if (bt_table_set(sim->bond_types, key, bt))
  {
    return TS_OK;
  }
  else
  {
    return TS_NEW_ERROR( TS_ERR_ADDING_BONDTYPE );
  }
}

static void remove_all_bond_types(ts_sim* sim)
{
  bt_table_clear(sim->bond_types);
}

ts_bond_type ts_get_bond_type(const ts_sim* sim, const char* btName)
{
  ts_bond_type bt = bondtype_none();
  bt_table_get(sim->bond_types, btName, &bt);
  return bt;
}

TSResult ts_foreach_bond_type(ts_sim* sim,
    TSResult (*eachfunc)(void*, const char*, ts_bond_type),
    void* ud)
{
  const char* key;
  ts_bond_type val;
  BT_FOREACH(sim->bond_types,
      key, val,
      RETURN_ON_ERROR(eachfunc(ud, key, val))
      );

  return TS_OK;
}

void ts_bond_type_iter_init(ts_sim* sim, struct ts_bond_type_iter* iter)
{
  UNUSED_PARAM(sim);
  iter->state = 0;
  iter->key = NULL;
  iter->bt  = bondtype_none();
}

int ts_bond_type_iter_next(ts_sim* sim, struct ts_bond_type_iter* iter)
{
  khint_t i = (khint_t)iter->state;
  khash_t(bond_type)* ht = HASH_BT_TABLE(sim->bond_types);
  khint_t end = kh_end(ht);
  fprintf(stderr, "i=%zd  end=%zd  key=\"%s\" @ %p bt=%d\n",
      (ptrdiff_t)iter->state, (ptrdiff_t) end,
      iter->key, iter->key, (int)iter->bt.value);
  for(; i < end; i++)
  {
    if (kh_exist(ht,i))
    {
      break;
    }
  }
  if (i == end) { return 0; }

  iter->key = kh_key(ht, i);
  iter->bt  = kh_val(ht, i);
  iter->state = (int64_t)(i+1);
  return 1;
}

size_t ts_count_bond_types(const ts_sim* sim)
{
  return bt_table_size(sim->bond_types);
}

uint16_t ts_find_unused_bond_type(const ts_sim* sim)
{
  uint16_t last_used = 0;

  /* first scan bond type table to find maximum bond type */
  khash_t(bond_type)* ht = HASH_BT_TABLE(sim->bond_types);
  khint_t _end = kh_end(ht);

  for(khint_t _i = 0; _i < _end; _i++) {
    if (kh_exist(ht,_i)) {
      uint16_t bt = bondtype_get_type(kh_val(ht,_i));
      if (bt > last_used) { last_used = bt; }
    }
  }

  /* next scan through the bond patterns to see if an unregistered type
   * has been used */
  ts_bond_pattern* pat = DATAVECTOR_BEGIN(ts_bond_pattern, &sim->bond_patterns1);
  ts_bond_pattern* pat_end = DATAVECTOR_END(ts_bond_pattern, &sim->bond_patterns1);
  for( ; pat < pat_end ; pat++) {
    for (int i=0; i < pat->nbodies-1; i++) {
      uint16_t bt = bondtype_get_type(pat->btypes[i]);
      if (bt > last_used) { last_used = bt; }
    }
  }

  if (last_used < BONDTYPE_VALUE_MASK)
    return last_used+1;
  else
    return 0; /* all used up */
}

TSResult ts_add_rigid_body_type(ts_sim* sim, struct ts_rigid_body_type* rbType, size_t* nptr)
{
  if (!datavector_add(&(sim->rigid_body_types), rbType)) return TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
  if (nptr != NULL) { *nptr = sim->rigid_body_types.size; }

  return TS_OK;
}

struct ts_rigid_body_type* ts_get_rigid_body_type(const ts_sim* sim, size_t rbTypeInd)
{
  if ((rbTypeInd < 1) || (rbTypeInd > sim->rigid_body_types.size))
  {
    return NULL;
  }

  return DATAVECTOR_AT(struct ts_rigid_body_type,&(sim->rigid_body_types),rbTypeInd-1);
}

static inline ts_particle_type* 
get_ptype(const ts_sim* sim, size_t type)
{
  return DATAVECTOR_AT(ts_particle_type,&(sim->particle_types),type-1);
}

ts_particle_type* ts_get_particle_type(const ts_sim* sim, size_t type)
{
  return ((type >= 1) && (type <= sim->particle_types.size))
    ? get_ptype(sim,type)
    : NULL
    ;
}

size_t ts_find_particle_type(const ts_sim* sim, const char* name)
{
  size_t ntypes = ts_count_particle_types(sim);
  for (size_t i=1; i <= ntypes; i++)
  {
    ts_particle_type* ptype = get_ptype(sim, i);
    if (STRINGS_EQ(ptype->name, name)) { return i; }
  }
  return 0;
}

size_t ts_find_rigid_body_type(const ts_sim* sim, const char* name)
{
  size_t ntypes = ts_count_rigid_body_types(sim);
  for (size_t i=1; i <= ntypes; i++) {
    struct ts_rigid_body_type* rbtype = ts_get_rigid_body_type(sim, i);
    if (STRINGS_EQ(rbtype->name, name)) {
      return i;
    }
  }

  return 0;
}

TSResult ts_alloc_particles(ts_sim* sim, size_t n)
{
  if (sim->particles != NULL)
  {
    return TS_NEW_ERROR3(
        TS_ERR_UNSUPPORTED_OPERATION,
        "particles already allocated"
        );
  }

#define CLEANUP_ON_ERROR { \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->particles); \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_pos);    \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_vel);    \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_forces); \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_eforces); \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_work);   \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->vl_maxdisp_origin); \
  FREE_AND_NULLIFY(NO_ALLOCATOR, sim->work_array); \
  IF_DISPATCH( FREE_AND_NULLIFY(NO_ALLOCATOR, sim->particle_locks) ); }

  sim->particles = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->particles[0]) );
  ERROR_ON_NULL_ALLOC(sim->particles);

  sim->vl_pos = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->vl_pos[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_pos);

  sim->vl_vel = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->vl_vel[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_vel);

  sim->vl_forces = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->vl_forces[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_forces);

  sim->vl_eforces = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->vl_eforces[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_eforces);

  sim->vl_work = ts_calloc(NO_ALLOCATOR, 3*n, sizeof(sim->vl_work[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_work);

  sim->vl_maxdisp_origin = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->vl_maxdisp_origin[0]));
  ERROR_ON_NULL_ALLOC(sim->vl_maxdisp_origin);

  sim->work_array = ts_calloc(NO_ALLOCATOR, 12*n, sizeof(sim->work_array[0]));
  ERROR_ON_NULL_ALLOC(sim->work_array);

#ifdef TS_DISPATCH_ENABLED
  sim->particle_locks = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->particle_locks[0]));
  ERROR_ON_NULL_ALLOC(sim->particle_locks);

  for(size_t i=0; i < n; i++) { tsd_spinlock_init(sim->particle_locks+i); }
#endif /* TS_DISPATCH_ENABLED */

  sim->work_size = 12*n;
  sim->max_particles = n;

#undef CLEANUP_ON_ERROR

  RETURN_ON_ERROR( ts_dof_map_init(sim->dof_map, sim) );

  return TS_OK;
}

ts_particle* ts_define_particle(ts_sim* sim, size_t ptype)
{
  if (sim->num_particles >= sim->max_particles) return NULL;

  ts_particle* p = sim->particles + sim->num_particles;
  ts_particle_init(p);

  p->type = ts_get_particle_type(sim,ptype);

  sim->num_particles++;

  return p;
}

void ts_dealloc_particles(ts_sim* sim)
{
  if (sim->particles == NULL) return;

  for (size_t i=0; i < sim->num_particles; i++)
  {
    ts_particle_finalize(sim->particles + i);
  }

  ts_free(NO_ALLOCATOR, sim->particles);

  sim->particles = NULL;
  sim->num_particles = 0;
}

TSResult ts_alloc_rigid_bodies(ts_sim* sim, size_t n)
{
  if (sim->rigid_bodies != NULL)
  {
    return TS_NEW_ERROR3(
        TS_ERR_UNSUPPORTED_OPERATION,
        "rigid bodies already allocated"
        );
  }

#define CLEANUP_ON_ERROR {}
  sim->rigid_bodies = ts_calloc(NO_ALLOCATOR, n, sizeof(ts_rigid_body));
  ERROR_ON_NULL_ALLOC(sim->rigid_bodies);

  sim->ql_ori = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->ql_ori[0]));
  ERROR_ON_NULL_ALLOC(sim->ql_ori);

  sim->ql_prev_ori = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->ql_ori[0]));
  ERROR_ON_NULL_ALLOC(sim->ql_ori);

  sim->ql_work = ts_calloc(NO_ALLOCATOR, n, sizeof(sim->ql_work[0]));
  ERROR_ON_NULL_ALLOC(sim->ql_ori);

  sim->max_rigid_bodies = n;
  sim->num_rigid_bodies = 0;
  for(size_t i = 0; i < n; i++)
  {
    ts_rigid_body_init(sim->rigid_bodies + i);
  }
#undef CLEANUP_ON_ERROR

  return TS_OK;
}

void ts_dealloc_rigid_bodies(ts_sim* sim)
{
  if (sim->rigid_bodies == NULL) return;

  for(size_t i = 0; i < sim->max_rigid_bodies; i++)
  {
    ts_rigid_body_finalize(sim->rigid_bodies+i);
  }

  ts_free(NO_ALLOCATOR, sim->rigid_bodies);

  sim->num_rigid_bodies = 0;
  sim->max_rigid_bodies = 0;
  sim->rigid_bodies = NULL;
}

ts_rigid_body* ts_define_rigid_body(ts_sim* sim, struct ts_rigid_body_type* type_ptr,
    size_t com_particle_ind, size_t site_inds[])
{
  size_t* inds = NULL;

  if (sim->num_rigid_bodies >= sim->max_rigid_bodies) return NULL;

  size_t rbind = sim->num_rigid_bodies;

  ts_rigid_body* body = sim->rigid_bodies + rbind;
  body->type_ptr = type_ptr;

  body->com_particle = com_particle_ind;
  quat4_identity(RIGID_ORI(sim,rbind));
  vector3_zero(RIGID_TORQUE(sim,rbind));

  inds = ts_calloc(NO_ALLOCATOR, type_ptr->num_sites, sizeof(body->site_indices[0]));

  if (inds == NULL)
  {
    goto error;
  }

  memcpy(inds, site_inds, type_ptr->num_sites * sizeof(inds[0]));
  body->site_indices = inds;

  ts_particle* com_particle = ts_get_particle(sim, body->com_particle);

  com_particle->flags |= PARTICLE_FLAG_RIGID_COM;
  com_particle->rigid_body_index = sim->num_rigid_bodies;

  for (size_t i=0; i < type_ptr->num_sites; i++)
  {
    ts_particle* site_particle = ts_get_particle(sim, site_inds[i]);
    site_particle->flags |= PARTICLE_FLAG_RIGID;
    site_particle->rigid_body_index = sim->num_rigid_bodies;
  }

  ts_rigid_body_update_parameters(sim,body);

  sim->num_rigid_bodies++;
  return body;

error:
  if (inds != NULL) ts_free(NO_ALLOCATOR, inds);

  return NULL;
}

static inline double safe_inverse(double x) { return x > 0.0 ? 1.0/x : 0.0; }

void ts_rigid_body_type_update_parameters(ts_sim* sim, struct ts_rigid_body_type* body_type)
{
  body_type->inv_rotfric = safe_inverse(body_type->rotfric);
  body_type->inv_I0 = safe_inverse(body_type->I0);
  body_type->inv_sqrt_I0 = sqrt(body_type->inv_I0);
  body_type->bd_rand_coeff = sqrt( 2.0 * sim->temp * sim->dt * body_type->inv_rotfric );
}

void ts_rigid_body_update_parameters(ts_sim* sim, ts_rigid_body* body)
{
  ts_rigid_body_type_update_parameters(sim, body->type_ptr);
}

static TSResult do_update_rigid_bodies(void* data, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  UNUSED_PARAM(worker);
  UNUSED_PARAM(nworkers);

  ts_sim* sim = data;
  for (size_t i=i0; i < i1; i++) {
    ts_update_rigid_body(sim,i);
  }
  return TS_OK;
}

void ts_update_rigid_bodies(ts_sim* sim)
{
  ts_dispatch_foreach(
      sim->num_rigid_bodies,
      do_update_rigid_bodies,
      sim);
}

void ts_update_rigid_body(ts_sim*restrict sim, size_t rbind)
{
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);

  vector3*restrict com_pos = PARTICLE_POS(sim, body->com_particle);
  const quat4*restrict ori = RIGID_ORI(sim,rbind);
  for (size_t i=0; i < body->type_ptr->num_sites; i++)
  {
    vector3 off;
    vector3_set(&off, body->type_ptr->ref_vectors+i);
    quat4_rotate_vector(ori, &off);

    vector3_add(PARTICLE_POS(sim, body->site_indices[i]), com_pos, &off);
  }
}

void ts_rigid_body_calculate_force_and_torque(ts_sim*restrict sim, ts_rigid_body*restrict body)
{
  const vector3*restrict com_pos = PARTICLE_POS(sim, body->com_particle);
  vector3*restrict com_force = PARTICLE_FORCE(sim, body->com_particle);
  vector3_zero(&(body->torque));

  for (size_t i=0; i < body->type_ptr->num_sites; i++)
  {
    vector3 dr;
    vector3_sub(&dr, PARTICLE_POS(sim, body->site_indices[i]), com_pos);

    const vector3*restrict pf = PARTICLE_FORCE(sim, body->site_indices[i]);

    vector3 site_torque;
    vector3_cross(&site_torque, &dr, pf);
    vector3_add_to(&body->torque, &site_torque);

    vector3_add_to( com_force, pf );
  }
}

TSResult ts_set_rng(ts_sim* sim, struct ts_rng* rng_state)
{
  if (sim->rng_state != NULL)
  {
    RETURN_ON_ERROR( ts_rng_free(sim->rng_state) );
  }

  sim->rng_state = rng_state;
  return TS_OK;
}

TSResult ts_call_rng(ts_sim* sim, int n, double* v)
{
  RETURN_ON_ERROR( ts_rng_gen_uniforms(sim->rng_state, n, v) );
  return TS_OK;
}

TSResult ts_reset_max_displacement_origins(ts_sim* sim)
{
  memcpy(sim->vl_maxdisp_origin, sim->vl_pos, sim->num_particles*sizeof(sim->vl_pos[0]));

  return TS_OK;
}

void ts_clear_forces(ts_sim* sim)
{
  size_t i=0;
  for (i=0; i < sim->num_particles; i++)
  {
    vector3_zero(PARTICLE_FORCE(sim, i));
  }
}

static inline TSResult calc_bonded_forces(ts_sim*restrict sim, double*restrict energy_ptr, int termwise)
{
  return ts_potential_forces_from_list(sim, &sim->bondlist, energy_ptr, termwise);
}

#define NONBOND_ENERGY_THRESH 100.0

static inline TSResult calc_nonbond_forces(ts_sim*restrict sim, double*restrict energy_ptr, int termwise)
{
  RETURN_ON_ERROR( ts_potential_forces_from_list(sim, &sim->nonbond_list, energy_ptr, termwise) );
  return TS_OK;
}

TSResult ts_calculate_forces(ts_sim* sim, double* energy_ptr, int max_term_groups, int termwise)
{
  if (sim->num_particles == 0) return TS_OK;

  ts_clear_forces(sim);
  double energy = 0.0;
  RETURN_ON_ERROR( calc_bonded_forces(sim, &energy, termwise) );
  RETURN_ON_ERROR( calc_nonbond_forces(sim, &energy, termwise) );

  if (max_term_groups == 0) {
    if (energy_ptr) { *energy_ptr = energy; }
  } else {
    int i;

    for (i=0; i < max_term_groups+1; i++)
    {
      energy_ptr[i] = 0.0;
    }

    RETURN_ON_ERROR( 
        ts_potential_energies_resum(sim, &sim->bondlist, energy_ptr, max_term_groups)
        );
    RETURN_ON_ERROR( 
        ts_potential_energies_resum(sim, &sim->nonbond_list, energy_ptr, max_term_groups)
        );
  }

  return TS_OK;
}

TSResult ts_rigid_bodies_calculate_all_forces_and_torques(ts_sim* sim)
{
  ts_rigid_body* body = sim->rigid_bodies;
  ts_rigid_body* end = body + sim->num_rigid_bodies;
  while (body != end) {
    ts_rigid_body_calculate_force_and_torque(sim,body);
    body++;
  }

  return TS_OK;
}

TSResult ts_calculate_energy(ts_sim* sim, double* energy_ptr, int max_term_groups, int termwise)
{
  RETURN_ON_ERROR( ts_calculate_forces(sim, energy_ptr, max_term_groups, termwise) );
  return TS_OK;
}

TSResult ts_run_callbacks(ts_sim* sim)
{
  for (size_t i=0; i < sim->callbacks.size; i++)
  {
    ts_integrator_callback* cb_at_i = *(DATAVECTOR_AT(ts_integrator_callback*, &(sim->callbacks), i));

    if ((cb_at_i->callback == NULL) || (cb_at_i->freq < 0)) { continue; }

    cb_at_i->countdown++;
    if ((cb_at_i->freq == 1) || (cb_at_i->countdown == cb_at_i->freq))
    {
      cb_at_i->countdown = 0;
      RETURN_ON_ERROR( cb_at_i->callback(sim, cb_at_i) );
    }
  }

  return TS_OK;
}

TSResult ts_add_callback(ts_sim* sim, ts_integrator_callback* callback)
{
  return datavector_add( &(sim->callbacks), &callback ) ? TS_OK : TS_NEW_ERROR(TS_ERR_OUT_OF_MEMORY);
}

size_t ts_sim_num_dof(const ts_sim* sim)
{
  return PARTICLE_NUM_DOF*ts_sim_num_free_particles(sim) + RIGID_BODY_NUM_DOF*sim->num_rigid_bodies;
}

size_t ts_sim_num_free_particles(const ts_sim* sim)
{
  size_t num_free_particles = 0;
  for (size_t i=0; i < sim->num_particles; i++) {
    if (!ts_pind_in_rigid_body(sim, i)) {
      num_free_particles++;
    }
  }

  return num_free_particles;
}

static inline void update_maxdisp_values(double v, double maxdisp[2])
{
  if (v > maxdisp[0]) {
    maxdisp[1] = maxdisp[0];
    maxdisp[0] = v;
  } else if (v > maxdisp[1]) {
    maxdisp[1] = v;
  }
}

void ts_update_max_displacements(const ts_sim* sim, double maxdisp[2])
{
  maxdisp[0] = maxdisp[1] = 0.0;

  for (size_t i=0; i < sim->num_particles; i++) {
    if (!ts_pind_in_rigid_body(sim, i)) {
      vector3 dr;
      double dist = vector3_sqdistance(PARTICLE_POS(sim,i), PARTICLE_MAXDISP_ORIGIN(sim,i), &dr);

      update_maxdisp_values(dist,maxdisp);
    }

  }

  for (size_t i=0; i < sim->num_rigid_bodies; i++) {
    ts_rigid_body* body = ts_get_rigid_body(sim, i);

    vector3 dr;
    double dist = vector3_sqdistance(PARTICLE_POS(sim,body->com_particle), PARTICLE_MAXDISP_ORIGIN(sim,body->com_particle), &dr);

    update_maxdisp_values(dist,maxdisp);
  }

  maxdisp[0] = sqrt(maxdisp[0]);
  maxdisp[1] = sqrt(maxdisp[1]);
}

TSResult ts_rigid_body_xyz_axis_angle_jacobian(const ts_sim* sim,
    size_t rbind, size_t site,
    vector3*restrict dq_w1,
    vector3*restrict dq_w2,
    vector3*restrict dq_w3)
{
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);
  if (site >= body->type_ptr->num_sites) return TS_NEW_ERROR(TS_ERR_INVALID_INDEX);

  vector3 axis;
  double ang = quat4_to_axis_angle_separately(RIGID_ORI(sim,rbind),&axis);
  dq_wj(1,dq_w1,dq_w2,dq_w3, ang, &axis, body->type_ptr->ref_vectors+site);
  return TS_OK;
}

TSResult ts_rigid_body_xyz_axis_angle_hessian(const ts_sim* sim,
    size_t rbind, size_t site,
    vector3 d2q[PARTICLE_HESSIAN_SZ])
{
  ts_rigid_body* body = ts_get_rigid_body(sim,rbind);
  if (site >= body->type_ptr->num_sites) return TS_NEW_ERROR(TS_ERR_INVALID_INDEX);

  vector3 axis_angle;
  quat4_to_axis_angle(RIGID_ORI(sim,rbind), &axis_angle);
  d2q_dwi_dwj(d2q, &axis_angle, body->type_ptr->ref_vectors+site);
  return TS_OK;
}

size_t ts_lookup_particle_set(ts_sim *sim, struct ts_index_set *set)
{
  if (!datavector_contains(&sim->particle_sets, set))
  {
    return 0;
  }

  return DATAVECTOR_INDEX_OF(struct ts_index_set, &sim->particle_sets, set) + 1;
}

struct ts_index_set *ts_add_particle_set(ts_sim *sim, size_t cap)
{
  static struct ts_index_set zero;
  struct ts_index_set *set;
  TSResult err;

  if (set = datavector_new_item(&sim->particle_sets), set == NULL)
  {
    return NULL;
  }

  *set = zero;

  if (err = ts_index_set_initialize(set, cap), IS_ERROR(err))
  {
    sim->particle_sets.size--;
    return NULL;
  }

  return set;
}

struct ts_index_set *ts_get_particle_set(ts_sim *sim, size_t ind)
{
  if (ind < 1 || ind > sim->particle_sets.size)
  {
    return NULL;
  }

  return DATAVECTOR_AT(struct ts_index_set, &sim->particle_sets, ind-1);
}

enum { TS_INDEX_SET_2X_THRESH = 2048 };

TSResult ts_index_set_initialize(struct ts_index_set *set, size_t cap)
{
  size_t *inds;

  if (cap == 0)
  {
    cap = TS_INDEX_SET_DEFAULT_CAP;
  }

  inds = ts_alloc(NO_ALLOCATOR, cap * sizeof *inds);
  ERROR_ON_NULL_ALLOC(inds);

  set->indices = inds;
  set->len = 0;
  set->cap = cap;

  return TS_OK;
}

static TSResult index_set_ensure(struct ts_index_set *set, size_t len)
{
  size_t cap;
  size_t *inds;

  cap = set->cap;
  if (cap == 0)
  {
    cap = TS_INDEX_SET_DEFAULT_CAP;
  }

  if (len > cap)
  {
    cap += (cap < TS_INDEX_SET_2X_THRESH) ? cap : cap/4;

    if (len > cap)
    {
      cap = len;
    }
  }

  if (cap == set->cap)
  {
    return TS_OK;
  }

  assert(cap > set->cap);
  inds = ts_realloc(NO_ALLOCATOR, set->indices, cap * sizeof *inds);
  ERROR_ON_NULL_ALLOC(inds);

  set->indices = inds;
  set->cap = cap;

  return TS_OK;
}

/* Ensures that an index set can hold at least len items */
TSResult ts_index_set_reserve(struct ts_index_set *set, size_t len)
{
  if (len > set->cap)
  {
    RETURN_ON_ERROR( index_set_ensure(set,len) );
  }

  return TS_OK;
}

/* Finalizes an index set */
void ts_index_set_finalize(struct ts_index_set *set)
{
  ts_free(NO_ALLOCATOR, set->indices);
  set->indices = NULL;
  set->len = 0;
  set->cap = 0;
}

/* Adds an index to the index set, resizing if necessary */
TSResult ts_index_set_add(struct ts_index_set *set, size_t ind)
{
  size_t i,n;

  /* replace with a binary search */
  n = set->len;
  for (i=0; i < n; i++)
  {
    if (set->indices[i] == ind)
    {
      return TS_OK;
    }

    if (set->indices[i] > ind)
    {
      break;
    }
  }

  if (set->len >= set->cap)
  {
    RETURN_ON_ERROR( index_set_ensure(set, set->len+1) );
  }

  assert(set->len < set->cap);

  // move items
  if (i < n)
  {
    memmove(&set->indices[i+1], &set->indices[i], (n-i) * sizeof set->indices[0]);
  }

  set->indices[i] = ind;
  set->len++;

  return TS_OK;
}

/* Clears the index set */
void ts_index_set_clear(struct ts_index_set *set)
{
  set->len = 0;
}



