#ifndef TUBESHEET__H
#define TUBESHEET__H

#include <stddef.h>
#include <stdlib.h>

#include "ts_common.h"

#include "vector3.h"
#include "ts_particle.h"
#include "ts_potentials.h"
#include "ts_errors.h"
#include "ts_rng.h"
#include "ts_dispatch.h"
#include "ts_parameters.h"

#define MODULE_NAME TUBESHEET

#define DEFAULT_BONDLIST_SIZE 1024
#define DEFAULT_PTYPE_TABLE_SIZE 64
#define DEFAULT_BONDTYPE_TABLE_SIZE 16

#define DEFAULT_SHAKE_TOL 1e-12
#define MAX_SHAKE_ITERATIONS 100

extern const char* tubesheet_repo_sha1;
extern const char* tubesheet_repo_version;

enum TS_BOND_TERM {
  TS_BOND_TERM_ZEROBODY   = 0,

  TS_BOND_TERM_ONEBODY    = 1,
  TS_BOND_TERM_TWOBODY    = 2,
  TS_BOND_TERM_THREEBODY  = 3,

  TS_BOND_TERM_FOURBODY   = 4,
  TS_BOND_TERM_IMPROPER   = 5,

  TS_BOND_TERM_NONBOND    = 6,
  TS_BOND_TERM_CONSTRAINT = 7,

  TS_BOND_TERM_MAX
};

struct ts_bond_pattern;
struct ts_constraint_entry;

enum TS_LOGGING_LEVEL {
  TS_LOG_VERBOSE = 0,
  TS_LOG_INFO    = 1,
  TS_LOG_WARN    = 2,
  TS_LOG_ERROR   = 3,
  TS_LOG_ABORT   = 4,
};

enum TS_LOGGING_FLAG {
  TS_LOG_NONE       = 0,
  TS_LOG_GENERAL    = BIT_VALUE(1),
  TS_LOG_INTEGRATOR = BIT_VALUE(2),
  TS_LOG_MINIMIZER  = BIT_VALUE(3),
  TS_LOG_NONBONDS   = BIT_VALUE(4),
  TS_LOG_IO         = BIT_VALUE(5),
  TS_LOG_NORMALMODE = BIT_VALUE(6),
  TS_LOG_ENERGY     = BIT_VALUE(7),

  TS_LOG_MAX_COMPONENT
};

#define TS_DEFAULT_LOGGING_FLAGS  ( ((TS_LOG_MAX_COMPONENT-1) << 1) - 1 )

enum TS_INTEGRATOR_TYPE {
  TS_INTEGRATOR_NONE = 0,

  TS_INTEGRATOR_VELOCITY_VERLET,
  TS_INTEGRATOR_LEAPFROG_VERLET,

  TS_INTEGRATOR_LEAPFROG_LANGEVIN,

  TS_INTEGRATOR_BROWNIAN_DYNAMICS,
  TS_INTEGRATOR_STOCH_RUNGE_KUTTA,
};

#define TS_PARTICLE(s,ind) ( (s)->particles + (ind) )
#define TS_PTYPE_IND(s,ind) ( ts_particle_type_index_from_pind((s),(ind)) )

#define TS_PIND(s,p) ( ts_particle_index((s),(p)) )
#define PARTICLE_POS(s,pind) ( ts_particle_pos((s),(pind)) )
#define PARTICLE_VEL(s,pind) ( ts_particle_vel((s),(pind)) )
#define PARTICLE_FORCE(s,pind) ( ts_particle_force((s),(pind)) )
#define PARTICLE_PX(s,pind) (VECTOR3_X(PARTICLE_POS(s,pind)))
#define PARTICLE_PY(s,pind) (VECTOR3_Y(PARTICLE_POS(s,pind)))
#define PARTICLE_PZ(s,pind) (VECTOR3_Z(PARTICLE_POS(s,pind)))

#define PARTICLE_MAXDISP_ORIGIN(s,pind) ( ts_particle_maxdisp_origin((s),(pind)) )

#define TS_RIGID_IND(s,rb) ( ts_rigid_body_index((s),(rb)) )
#define RIGID_ORI(s,rind) ( ts_rigid_body_orientation((s), (rind)) )
#define RIGID_PREV_ORI(s,rind) ( ts_rigid_body_prev_orientation((s), (rind)) )
#define RIGID_POS(s,rind) ( ts_rigid_body_pos((s), (rind)) )
#define RIGID_VEL(s,rind) ( ts_rigid_body_vel((s), (rind)) )
#define RIGID_FORCE(s,rind) ( ts_rigid_body_force((s), (rind)) )
#define RIGID_TORQUE(s,rind) ( ts_rigid_body_torque((s), (rind)) )
#define RIGID_PREV_TORQUE(s,rind) ( ts_rigid_body_prev_torque((s), (rind)) )


struct ts_index_set;

typedef struct ts_sim ts_sim;
struct ts_sim {
  uint32_t logging_flags;
  FILE* logging_output;

  struct ts_param_table *parameters;
  struct ts_potential_table *potentials;

  struct ts_rng *rng_state;

  double *rand_forces;
  size_t num_rand_forces;

  double *rand_torques;
  size_t num_rand_torques;

  size_t max_particles;
  size_t num_particles;
  ts_particle *particles;

#ifdef TS_DISPATCH_ENABLED
  struct tsd_spinlock* particle_locks;
#endif /* TS_DISPATCH_ENABLED */

  vector3 *vl_pos;
  vector3 *vl_vel;
  vector3 *vl_forces;
  vector3 *vl_eforces;

  vector3 *vl_work;

  vector3 *vl_maxdisp_origin;

  quat4 *ql_ori;
  quat4 *ql_prev_ori;
  quat4 *ql_work;

  double *work_array;
  size_t work_size;

  double maxdisp1;
  double maxdisp2;

  size_t max_rigid_bodies;
  size_t num_rigid_bodies;
  ts_rigid_body *rigid_bodies;

  struct ts_bond_type_table *bond_types;
  DATAVECTOR(ts_particle_type) particle_types;
  DATAVECTOR(struct ts_rigid_body_type) rigid_body_types;

  DATAVECTOR(ts_bond_pattern) bond_patterns1;
  DATAVECTOR(ts_bond_pattern) nonbond_patterns1;

  DATAVECTOR(struct ts_potential_term) extra_terms;

  int bonds_enabled[TS_BOND_TERM_MAX];
  int nonbonds_enabled;

  DATAVECTOR(ts_constraint_entry) constraints;

  DATAVECTOR(ts_integrator_callback *) callbacks;

  struct ts_dof_map *dof_map;

  DATAVECTOR(struct ts_index_set) particle_sets;

  vector3 net_force;
  vector3 net_torque;

  struct ts_potential_term_list bondlist;
  struct ts_potential_term_list nonbond_list;

  double nonbond_cutoff;
  double nonbond_padding;

  enum TS_INTEGRATOR_TYPE integrator;
  double dt;
  double inv_dt;
  double temp;
  double shake_tol;

  double time;
  size_t integrator_steps;

  double KE;
  double rotKE;
  double PE;
  double totalE;
};


/* Simulation constructor and destructor */
ts_sim *ts_sim_new(void);
void ts_sim_delete(ts_sim *sim);


/* Simulation logging */
static inline int ts_is_component_logging(const ts_sim *sim, enum TS_LOGGING_LEVEL level, uint32_t component) 
{ 
  return (level >= TS_LOG_ABORT) || (sim->logging_flags & component); 
}

static inline void ts_set_component_logging(ts_sim *sim, uint32_t component, int tf) { 
  sim->logging_flags = (tf) ? (sim->logging_flags | component) : (sim->logging_flags & (~component));
}

void ts_log_message(const ts_sim *sim, enum TS_LOGGING_FLAG component, enum TS_LOGGING_LEVEL level, const char *mesg, ...);

#define TS_LOG(sim_param,log_level,component_param,...)  \
  do {                                                   \
    const ts_sim *_sim = (sim_param);                    \
    uint32_t _component = (component_param);             \
    enum TS_LOGGING_LEVEL _level = (log_level);          \
    if (ts_is_component_logging(_sim, _level, _component)) { \
      ts_log_message(_sim, _component, _level, __VA_ARGS__); \
    }                                                    \
  } while(0)


/* Simulation parameters */
TSResult ts_set_parameter(ts_sim *sim, const char *key, double val);
TSResult ts_get_parameter(ts_sim *sim, const char *key, double *val_ptr);
int ts_has_parameter(ts_sim *sim, const char *key);

TSResult ts_foreach_parameter(ts_sim *sim, ts_parameter_func func, void *data);

TSResult ts_add_potential(ts_sim *sim, const char *name, Potential *pot);
Potential *ts_get_potential(ts_sim *sim, const char *name);

TSResult ts_foreach_potential(ts_sim *sim,
    TSResult (*func)(void *, const char *, Potential *));

struct ts_sim_potential_iter { int64_t state; const char* key; Potential *pot; };
void ts_sim_potential_iter_init(ts_sim *sim, struct ts_sim_potential_iter *iter);
int ts_sim_potential_iter_next(ts_sim *sim, struct ts_sim_potential_iter *iter);

static inline vector3 *ts_particle_pos(const ts_sim *sim, size_t pind) { return sim->vl_pos + pind; }
static inline vector3 *ts_particle_vel(const ts_sim *sim, size_t pind) { return sim->vl_vel + pind; }
static inline vector3 *ts_particle_force(const ts_sim *sim, size_t pind) { return sim->vl_forces + pind; }
static inline vector3 *ts_particle_maxdisp_origin(const ts_sim *sim, size_t pind) { return sim->vl_maxdisp_origin + pind; }
static inline size_t ts_particle_index(const ts_sim *sim, const ts_particle *p)
{
  if (SAFECHECK &&
      (p < sim->particles || p >= sim->particles+sim->num_particles))
  {
      return TS_NULL_RIGID_BODY;
  }
  return (size_t)(p - sim->particles);
}

static inline int ts_pind_is_free(const ts_sim *sim, size_t pind)
{
  return ts_particle_is_free(sim->particles+pind);
}

static inline int ts_pind_is_rigid(const ts_sim *sim, size_t pind)
{
  return ts_particle_is_rigid(sim->particles+pind);
}

static inline int ts_pind_in_rigid_body(const ts_sim *sim, size_t pind)
{
  return ts_particle_in_rigid_body(sim->particles+pind);
}

static inline void ts_sim_lock_particle(ts_sim* sim, size_t pind)
{ 
  tsAssert(pind < sim->num_particles);
#ifdef TS_DISPATCH_ENABLED
  tsd_spinlock_lock(sim->particle_locks+pind);
#endif /* TS_DISPATCH_ENABLED */
}

static inline void ts_sim_unlock_particle(ts_sim* sim, size_t pind)
{
  tsAssert(pind < sim->num_particles);
#ifdef TS_DISPATCH_ENABLED
  tsd_spinlock_unlock(sim->particle_locks+pind);
#endif /* TS_DISPATCH_ENABLED */
}

TSResult ts_add_particle_type(ts_sim* sim, ts_particle_type* ptData, size_t* nptr);
ts_particle_type* ts_get_particle_type(const ts_sim* sim, size_t pType);
size_t ts_find_particle_type(const ts_sim* sim, const char* name);

static inline size_t ts_count_particle_types(const ts_sim* sim) { return sim->particle_types.size; }
static inline size_t ts_particle_type_index(const ts_sim *sim, ts_particle_type *ptype)
{
  return (ptype != NULL) 
    ? ((size_t) DATAVECTOR_INDEX_OF(ts_particle_type, &sim->particle_types, ptype) + 1)
    : 0;
}

static inline size_t ts_particle_type_index_from_pind(const ts_sim* sim, size_t pind)
{
  return ts_particle_type_index(sim, TS_PARTICLE(sim,pind)->type);
}


TSResult ts_add_rigid_body_type(ts_sim *sim, struct ts_rigid_body_type *rbData, size_t *nptr);
size_t ts_find_rigid_body_type(const ts_sim *sim, const char *name);
static inline size_t ts_count_rigid_body_types(const ts_sim *sim) { return sim->rigid_body_types.size; }
struct ts_rigid_body_type* ts_get_rigid_body_type(const ts_sim *sim, size_t rbTypeInd);
void ts_rigid_body_type_update_parameters(ts_sim *sim, struct ts_rigid_body_type *body);

static inline size_t ts_rigid_body_type_index(const ts_sim *sim,
    struct ts_rigid_body_type *rbtype)
{
  return (rbtype != NULL) 
    ? ((size_t) DATAVECTOR_INDEX_OF(struct ts_rigid_body_type,
          &sim->rigid_body_types, rbtype) + 1)
    : 0;
}

TSResult ts_alloc_particles(ts_sim *sim, size_t n);
void ts_dealloc_particles(ts_sim *sim);
ts_particle* ts_define_particle(ts_sim *sim, size_t ptype);

static inline ts_particle *ts_get_particle(const ts_sim *sim, size_t pidx) { return sim->particles + pidx; }
static inline int ts_has_particle(const ts_sim *restrict sim, const ts_particle *restrict p)
{
  return (p >= sim->particles) && (p < sim->particles+sim->num_particles);
}

TSResult ts_alloc_rigid_bodies(ts_sim* sim, size_t n);
void ts_dealloc_rigid_bodies(ts_sim* sim);
ts_rigid_body* ts_define_rigid_body(ts_sim* sim, struct ts_rigid_body_type* type_ptr,
    size_t com_particle_ind, size_t site_inds[]);

static inline size_t ts_num_particles(const ts_sim* sim) { return sim->num_particles; }
static inline size_t ts_max_particles(const ts_sim* sim) { return sim->max_particles; }

static inline size_t ts_rigid_body_index(ts_sim* sim, ts_rigid_body* body)
{
  if(SAFECHECK) {
    if ((body < sim->rigid_bodies) || (body >= sim->rigid_bodies+sim->num_rigid_bodies))
      return TS_NULL_RIGID_BODY;
  }

  return (size_t)(body - sim->rigid_bodies);

}

static inline ts_rigid_body *ts_get_rigid_body(const ts_sim *sim, size_t body_idx) { return sim->rigid_bodies + body_idx; }


void ts_rigid_body_update_parameters(ts_sim *sim, ts_rigid_body *body);

static inline vector3* ts_rigid_body_torque(ts_sim* sim, size_t body_idx) 
{ 
  ts_rigid_body* body = ts_get_rigid_body(sim,body_idx); 
  return &body->torque; 
}

static inline vector3 *ts_rigid_body_prev_torque(ts_sim *sim, size_t body_idx) 
{ 
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx); 
  return &body->prev_torque; 
}

static inline quat4 *ts_rigid_body_orientation(const ts_sim *sim, size_t rbind) {
  return sim->ql_ori + rbind;
}

static inline quat4* ts_rigid_body_prev_orientation(const ts_sim* sim, size_t rbind) {
  return sim->ql_prev_ori + rbind;
}

static inline size_t ts_rigid_body_com_particle(ts_sim *sim, size_t body_idx)
{
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return body->com_particle;
}

static inline vector3 *ts_rigid_body_pos(ts_sim *sim, size_t body_idx)
{
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return PARTICLE_POS(sim,body->com_particle);
}

static inline vector3 *ts_rigid_body_force(ts_sim *sim, size_t body_idx) 
{ 
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return PARTICLE_FORCE(sim,body->com_particle);
}

static inline int ts_rigid_body_is_frozen(const ts_sim *sim, size_t body_idx)
{
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return (body->flags & RIGID_FLAG_FROZEN_ALL) == RIGID_FLAG_FROZEN_ALL;
}

static inline int ts_rigid_body_is_frozen_translation(const ts_sim *sim, size_t body_idx)
{
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return (body->flags & RIGID_FLAG_FROZEN_POS) == RIGID_FLAG_FROZEN_POS;
}

static inline int ts_rigid_body_is_frozen_rotation(const ts_sim *sim, size_t body_idx)
{
  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  return (body->flags & RIGID_FLAG_FROZEN_ORI) == RIGID_FLAG_FROZEN_ORI;
}

static inline void ts_rigid_body_set_frozen(ts_sim *sim, size_t body_idx, int trans, int rot)
{
  /* freezing rotation but not translation is not supported
   *
   * !trans && rot  --> rotation frozen but not translation
   *
   * assert !(!trans && rot) = trans || !rot
   */
  tsAssert( trans || !rot );

  ts_rigid_body *body = ts_get_rigid_body(sim,body_idx);
  ts_particle *com = ts_get_particle(sim, body->com_particle);
  ts_particle_set_frozen(com,trans);

  if (trans) {
    body->flags |= RIGID_FLAG_FROZEN_POS;
  } else {
    body->flags &= ~RIGID_FLAG_FROZEN_POS;
  }
  if (rot) {
    body->flags |= RIGID_FLAG_FROZEN_ORI;
  } else {
    body->flags &= ~RIGID_FLAG_FROZEN_ORI;
  }
}


void ts_update_rigid_body(ts_sim*restrict sim, size_t rbind);
void ts_rigid_body_calculate_force_and_torque(ts_sim*restrict sim, ts_rigid_body*restrict body);

void ts_update_rigid_bodies(ts_sim* sim);


TSResult ts_set_rng(ts_sim* sim, struct ts_rng* rng_state);
TSResult ts_call_rng(ts_sim* sim, int n, double* v);

TSResult ts_alloc_rand_forces(ts_sim* sim);
void ts_free_rand_forces(ts_sim* sim);


void ts_clear_forces(ts_sim* sim);

TSResult ts_calculate_energy(ts_sim* sim, double* energy_ptr, int max_term_groups, int termwise);

TSResult ts_calculate_forces(ts_sim* sim, double* energy_ptr, int max_term_groups, int termwise);
TSResult ts_rigid_bodies_calculate_all_forces_and_torques(ts_sim* sim);


TSResult ts_reset_max_displacement_origins(ts_sim* sim);
void ts_update_max_displacements(const ts_sim* sim, double maxdisp[2]);


size_t ts_sim_num_dof(const ts_sim* sim);
size_t ts_sim_num_free_particles(const ts_sim* sim);

TSResult ts_rigid_body_xyz_axis_angle_jacobian(const ts_sim* sim,
    size_t rbind, size_t site,
    vector3*restrict dq_w1,
    vector3*restrict dq_w2,
    vector3*restrict dq_w3);

TSResult ts_rigid_body_xyz_axis_angle_hessian(const ts_sim* sim,
    size_t rbind, size_t site,
    vector3 d2q[PARTICLE_HESSIAN_SZ]);


TSResult ts_yaml_load_string(ts_sim* sim, const char* s, size_t slen);


TSResult ts_setup_collision_coeffs(ts_sim* sim);

static inline void ts_ptypes_zero(int ptypes[NBODY_LIMIT])
{
  memset(ptypes, 0, NBODY_LIMIT*sizeof(ptypes[0]));
}

struct ts_index_set *ts_add_particle_set(ts_sim *sim, size_t cap);
struct ts_index_set *ts_get_particle_set(ts_sim *sim, size_t ind);
size_t ts_lookup_particle_set(ts_sim *sim, struct ts_index_set *set);


struct ts_index_set {
  size_t len;
  size_t cap;
  size_t *indices;
};

enum { TS_INDEX_SET_DEFAULT_CAP = 16 };

/* Initializes an index set */
TSResult ts_index_set_initialize(struct ts_index_set *set, size_t cap);

/* Ensures that an index set can hold at least len items */
TSResult ts_index_set_reserve(struct ts_index_set *set, size_t len);

/* Finalizes an index set */
void ts_index_set_finalize(struct ts_index_set *set);

/* Adds an index to the index set, resizing if necessary */
TSResult ts_index_set_add(struct ts_index_set *set, size_t ind);

/* Clears the index set */
void ts_index_set_clear(struct ts_index_set *set);

#undef MODULE_NAME

#endif /* TUBESHEET__H */

