#include "vector3.h"

#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "ts_errors.h"
#include "ts_common.h"

typedef struct v3_pool v3_pool;
typedef struct v3_chunk_header v3_chunk_header;
typedef struct v3_free_list v3_free_list;

struct v3_chunk_header
{
  size_t size;  // Total size of chunk, including header
  v3_chunk_header* next;

  void* element0;  // first element
};

struct v3_free_list
{
  v3_free_list* next;
  size_t num_vectors;
};

struct v3_pool
{
  size_t total_bytes;
  size_t total_vectors;

  vector3_allocator_func allocator;
  vector3_deallocator_func deallocator;

  v3_chunk_header chunk_list;
  v3_free_list free_list;
};

static v3_pool pool = { 0,0, NULL,NULL, {0,NULL,NULL}, {NULL,0} };

// returns the number of vectors supported by the chunk
static size_t v3_init_chunk(v3_pool* pool, v3_chunk_header* chunk, size_t nbytes)
{
  chunk->size = nbytes;

  char* first_byte = (char*)chunk;
  char* after_header = first_byte + sizeof(v3_chunk_header);
  size_t align_off = SSE_ALIGNMENT - ( ((size_t)after_header) & (SSE_ALIGNMENT - 1) );
  char* element0 = after_header + align_off;
  chunk->element0 = (void*)element0;
  size_t nvecs = ((first_byte+nbytes) - element0) / sizeof(vector3);

  chunk->next = pool->chunk_list.next;
  pool->chunk_list.next = chunk;

  return nvecs;
}

void vector3_init_pool(void* base, size_t nbytes, vector3_allocator_func vec_alloc, vector3_deallocator_func vec_dealloc)
{
  if (pool.total_bytes != 0) return;

  pool.total_bytes = nbytes;
  pool.allocator = vec_alloc;
  pool.deallocator = vec_dealloc;
  pool.chunk_list.next = &(pool.chunk_list);

  pool.total_vectors += v3_init_chunk(&pool, (v3_chunk_header*)base, nbytes);
}

void vector3_free_pool(void)
{
}

size_t vector3_pool_num_vectors(void)
{
  return pool.total_vectors;
}

size_t vector3_pool_num_bytes(void)
{
  return pool.total_bytes;
}

#if 0
// Adds memory to the vector pool
void vector3_add_pool(void* base, size_t nbytes);

// Frees the vector pool.

// Allocates n contiguous vectors from the pool
vector3* vector3_alloc(size_t n);

// Releases n contiguous vectors back to the pool
void vector3_release(vector3* v0, size_t n);

// Allocates a temporary vector, which disappears when the current
// function exits
vector3* vector3_alloc_temp(void);
#endif /* 0 */

#ifndef TS_ASSERTIONS_ENABLED
void vector3_vector_tests(void) {}
#else /* ! TS_ASSERTIONS_ENABLED */
#define TEST_POOL_SIZE 2048
static char v3_test_pool[TEST_POOL_SIZE];

static void v3_test_can_init_pool(void)
{
  vector3_init_pool(v3_test_pool, sizeof(v3_test_pool), 0, 0);
  vector3_free_pool();
}

#define ASSERT_VEC_EQ( x, y, z, vec ) \
  do { const vector3* _temp_vec_ = vec; \
    tsAssert( fabs(vector3_x(_temp_vec_) - (x)) < VEC_TOL ); \
    tsAssert( fabs(vector3_y(_temp_vec_) - (y)) < VEC_TOL ); \
    tsAssert( fabs(vector3_z(_temp_vec_) - (z)) < VEC_TOL ); \
  } while(0)


static void v3_test_init_pool_is_correct_size(void)
{
  vector3_init_pool(v3_test_pool, sizeof(v3_test_pool), 0, 0);
  size_t overhead = sizeof(v3_chunk_header);
  char* after_overhead = v3_test_pool + overhead;
  char* after_aligned  = after_overhead + (0x10 - (((size_t)after_overhead) & (0x10 - 1)));
  size_t restsize = sizeof(v3_test_pool) - (after_aligned-v3_test_pool);
  size_t numvectors = restsize / sizeof(vector3);
  fprintf(stderr, "\n\n");
  fprintf(stderr, "test_pool      @ %p\n", v3_test_pool);
  fprintf(stderr, "after_overhead @ %p\n", after_overhead);
  fprintf(stderr, "after_aligned  @ %p\n", after_aligned);
  fprintf(stderr, "size of v3_test_pool is %lu\n",sizeof(v3_test_pool));
  fprintf(stderr, "size of overhead     is %lu\n",overhead);
  fprintf(stderr, "size of restsize     is %lu\n",restsize);
  fprintf(stderr, "numvectors           is %lu\n",numvectors);

  tsAssert( sizeof(v3_test_pool) == vector3_pool_num_bytes() );
  tsAssert( numvectors == vector3_pool_num_vectors() );
  vector3_free_pool();
}

static inline void dummy_use_variable(void* v) { UNUSED_PARAM(v); }

static void v3_test_set_with(void)
{
  vector3 a;
  vector3_set_with(&a, 1.0, 3.5, 7.25);
  tsAssert( 1.0  == vector3_x(&a) );
  tsAssert( 3.5  == vector3_y(&a) );
  tsAssert( 7.25 == vector3_z(&a) );
}

static void v3_test_zero(void)
{
  vector3 a;
  vector3_set_with(&a, 1.0, 3.5, 7.25);
  vector3_zero(&a);
  tsAssert( 0.0 == vector3_x(&a) );
  tsAssert( 0.0 == vector3_y(&a) );
  tsAssert( 0.0 == vector3_z(&a) );
}

static void v3_test_set(void)
{
  vector3 a;
  vector3 b;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 2.0, 4.0, 6.0);

  vector3_set(&b, &a);

  tsAssert( 1.0 == vector3_x(&b) );
  tsAssert( 2.0 == vector3_y(&b) );
  tsAssert( 3.0 == vector3_z(&b) );
}

static void v3_test_set_scaled(void)
{
  vector3 a;
  vector3 b;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 2.0, 4.0, 6.0);

  vector3_set_scaled(&b, 4.0, &a);

  tsAssert(  4.0 == vector3_x(&b) );
  tsAssert(  8.0 == vector3_y(&b) );
  tsAssert( 12.0 == vector3_z(&b) );
}

static void v3_test_set_scaled_array(void)
{
  vector3 a;
  VecReal arr1[4] = { 1.0, 2.0, 3.0, 0.0 };
  VecReal arr2[3] = { 3.0, 2.0, 1.0 };

  vector3_set_scaled_array(&a, 1.5, arr1);
  tsAssert( 1.5 == vector3_x(&a) );
  tsAssert( 3.0 == vector3_y(&a) );
  tsAssert( 4.5 == vector3_z(&a) );

  vector3_set_scaled_array(&a, 2.0, arr2);
  tsAssert( 6.0 == vector3_x(&a) );
  tsAssert( 4.0 == vector3_y(&a) );
  tsAssert( 2.0 == vector3_z(&a) );
}

static void v3_test_set_array(void)
{
  vector3 a;
  VecReal arr1[4] = { 1.0, 2.0, 3.0, 0.0 };
  VecReal arr2[3] = { 3.0, 2.0, 1.0 };

  vector3_set_array(&a, arr1);
  tsAssert( 1.0 == vector3_x(&a) );
  tsAssert( 2.0 == vector3_y(&a) );
  tsAssert( 3.0 == vector3_z(&a) );

  vector3_set_array(&a, arr2);
  tsAssert( 3.0 == vector3_x(&a) );
  tsAssert( 2.0 == vector3_y(&a) );
  tsAssert( 1.0 == vector3_z(&a) );
}

static void v3_test_get_elements(void)
{
  VecReal x,y,z;
  vector3 a,b;
  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set(&b, &a);
  x = y = z = 0.0;

  vector3_get_elements(&b, &x,&y,&z);

  tsAssert( 1.0 == x );
  tsAssert( 2.0 == y );
  tsAssert( 3.0 == z );
}

static void v3_test_add_to(void)
{
  vector3 a;
  vector3 b;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 1.5, 1.6, 1.7);

  vector3_add_to(&a, &b);
  ASSERT_VEC_EQ( 2.5, 3.6, 4.7, &a );

  ASSERT_VEC_EQ( 1.5, 1.6, 1.7, &b );
}

static void test_v3_sub_from(void)
{
  vector3 a;
  vector3 b;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 1.5, 1.6, 1.7);

  vector3_sub_from(&a, &b);
  ASSERT_VEC_EQ( -0.5, 0.4, 1.3, &a );

  ASSERT_VEC_EQ( 1.5, 1.6, 1.7, &b );
}


static void v3_test_add(void)
{
  vector3 a;
  vector3 b;
  vector3 c;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 1.5, 1.6, 1.7);

  vector3_add(&c, &a, &b);
  tsAssert( fabs(2.5 - vector3_x(&c)) < VEC_TOL );
  tsAssert( fabs(3.6 - vector3_y(&c)) < VEC_TOL );
  tsAssert( fabs(4.7 - vector3_z(&c)) < VEC_TOL );
}

static void v3_test_sub(void)
{
  vector3 a;
  vector3 b;
  vector3 c;

  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_set_with(&b, 1.5, 1.6, 1.7);

  vector3_sub(&c, &a, &b);
  tsAssert( fabs(vector3_x(&c) - (-0.5)) < VEC_TOL );
  tsAssert( fabs(vector3_y(&c) - ( 0.4)) < VEC_TOL );
  tsAssert( fabs(vector3_z(&c) - ( 1.3)) < VEC_TOL );
}

static void v3_test_displace(void)
{
  vector3 a;
  vector3_set_with(&a, 1.0, 2.0, 3.0);
  vector3_displace(&a, 1.5, 1.6, 1.7);

  ASSERT_VEC_EQ( 2.5, 3.6, 4.7, &a );
}

static void v3_test_scaleby(void)
{
  vector3 a;

  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_scaleby(&a, 2.0);
  ASSERT_VEC_EQ( 6.0, 4.0, 2.0, &a );

  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_multiply_by(&a, 2.0);
  ASSERT_VEC_EQ( 6.0, 4.0, 2.0, &a );
}

static void v3_test_set_to_midpoint(void)
{
  vector3 a;
  vector3 b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);
  vector3_set_to_midpoint(&a, &b);

  tsAssert( 2.0 == vector3_x(&a) );
  tsAssert( 2.0 == vector3_y(&a) );
  tsAssert( 2.0 == vector3_z(&a) );

  tsAssert( 1.0 == vector3_x(&b) );
  tsAssert( 2.0 == vector3_y(&b) );
  tsAssert( 3.0 == vector3_z(&b) );
}

static void v3_test_add_multiple_to(void)
{
  vector3 a;
  vector3 b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  vector3_add_multiple_to(&b, 2.0, &a);

  tsAssert( 3.0 == vector3_x(&a) );
  tsAssert( 2.0 == vector3_y(&a) );
  tsAssert( 1.0 == vector3_z(&a) );

  tsAssert( 7.0 == vector3_x(&b) );
  tsAssert( 6.0 == vector3_y(&b) );
  tsAssert( 5.0 == vector3_z(&b) );
}

static void v3_test_scale_and_add(void)
{
  vector3 a,b,c;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);
  vector3_zero(&c);

  vector3_scale_and_add(&c, 0.5, &a, 2.0, &b);

  tsAssert( 3.0 == vector3_x(&a) );
  tsAssert( 2.0 == vector3_y(&a) );
  tsAssert( 1.0 == vector3_z(&a) );

  tsAssert( 1.0 == vector3_x(&b) );
  tsAssert( 2.0 == vector3_y(&b) );
  tsAssert( 3.0 == vector3_z(&b) );

  tsAssert( 3.5 == vector3_x(&c) );
  tsAssert( 5.0 == vector3_y(&c) );
  tsAssert( 6.5 == vector3_z(&c) );
}

static void v3_test_sqnorm(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  tsAssert( 14.0 == vector3_sqnorm(&a) );
  tsAssert( 14.0 == vector3_sqnorm(&b) );
}

static void v3_test_norm(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  tsAssert( fabs(sqrt(14.0) - vector3_norm(&a)) < VEC_TOL );
  tsAssert( fabs(sqrt(14.0) - vector3_norm(&b)) < VEC_TOL );
}

static void v3_test_normalize(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  VecReal na = vector3_normalize(&a);
  VecReal nb = vector3_normalize(&b);

  tsAssert( fabs(sqrt(14.0) - na) < VEC_TOL );
  tsAssert( fabs(sqrt(14.0) - nb) < VEC_TOL );

  tsAssert( fabs(3.0/sqrt(14.0) - vector3_x(&a)) < VEC_TOL );
  tsAssert( fabs(2.0/sqrt(14.0) - vector3_y(&a)) < VEC_TOL );
  tsAssert( fabs(1.0/sqrt(14.0) - vector3_z(&a)) < VEC_TOL );

  tsAssert( fabs(1.0/sqrt(14.0) - vector3_x(&b)) < VEC_TOL );
  tsAssert( fabs(2.0/sqrt(14.0) - vector3_y(&b)) < VEC_TOL );
  tsAssert( fabs(3.0/sqrt(14.0) - vector3_z(&b)) < VEC_TOL );
}

static void v3_test_dot(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  tsAssert( 10.0 == vector3_dot(&a,&b) );
}

static void v3_test_sqdistance(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  vector3 dr;
  VecReal sqdist = vector3_sqdistance(&a,&b, &dr);
  tsAssert( 8.0 == sqdist );

  tsAssert( fabs(-2.0 - vector3_x(&dr)) < VEC_TOL );
  tsAssert( fabs( 0.0 - vector3_y(&dr)) < VEC_TOL );
  tsAssert( fabs( 2.0 - vector3_z(&dr)) < VEC_TOL );
}

static void v3_test_distance(void)
{
  vector3 a,b;
  vector3_set_with(&a, 3.0, 2.0, 1.0);
  vector3_set_with(&b, 1.0, 2.0, 3.0);

  vector3 dr;
  VecReal dist = vector3_distance(&a,&b, &dr);
  tsAssert( fabs(sqrt(8.0) - dist) < VEC_TOL );

  tsAssert( fabs(-2.0 - vector3_x(&dr)) < VEC_TOL );
  tsAssert( fabs( 0.0 - vector3_y(&dr)) < VEC_TOL );
  tsAssert( fabs( 2.0 - vector3_z(&dr)) < VEC_TOL );
}

void v3_test_cross_1(void)
{
  vector3 a,b,c;
  vector3_set_with(&b, 1.0, 0.0, 0.0);
  vector3_set_with(&c, 0.0, 1.0, 0.0);

  vector3_cross(&a, &b,&c);
  tsAssert( (0.0 == vector3_x(&a)) && "A = uX x uY.  A_x != 0.0" );
  tsAssert( (0.0 == vector3_y(&a)) && "A = uX x uY.  A_y != 0.0" );
  tsAssert( (1.0 == vector3_z(&a)) && "A = uX x uY.  A_z != 1.0" );

  vector3_cross(&a, &c,&b);
  tsAssert( ( 0.0 == vector3_x(&a)) && "A = uX x uY.  A_x !=  0.0" );
  tsAssert( ( 0.0 == vector3_y(&a)) && "A = uX x uY.  A_y !=  0.0" );
  tsAssert( (-1.0 == vector3_z(&a)) && "A = uX x uY.  A_z != -1.0" );

  VecReal rrt2 = 1.0/sqrt(2.0);
  vector3_set_with(&b,  rrt2,rrt2,0.0);
  vector3_set_with(&c, -rrt2,rrt2,0.0);
  vector3_cross(&a, &b,&c);
  tsAssert( ( fabs(vector3_x(&a) - 0.0) < VEC_TOL ) && "A = uX x uY.  A_x != 0.0" );
  tsAssert( ( fabs(vector3_y(&a) - 0.0) < VEC_TOL ) && "A = uX x uY.  A_y != 0.0" );
  tsAssert( ( fabs(vector3_z(&a) - 1.0) < VEC_TOL ) && "A = uX x uY.  A_z != 1.0" );

  vector3_set(&c,&b);
  vector3_cross(&a, &b,&c);
  tsAssert( ( fabs(vector3_x(&a) - 0.0) < VEC_TOL ) && "A = uX x uY.  A_x != 0.0" );
  tsAssert( ( fabs(vector3_y(&a) - 0.0) < VEC_TOL ) && "A = uX x uY.  A_y != 0.0" );
  tsAssert( ( fabs(vector3_z(&a) - 0.0) < VEC_TOL ) && "A = uX x uY.  A_z != 0.0" );

  VecReal rrt3 = 1.0/sqrt(3.0);
  vector3_set_with(&c, rrt3, rrt3, rrt3);
  vector3_cross(&a, &b,&c);
  VecReal rrt6 = 1.0/sqrt(6.0);
  dummy_use_variable(&rrt6);
  tsAssert( ( fabs(vector3_x(&a) - rrt6) < VEC_TOL ) && "A = uX x uY.  A_x !=  1/sqrt(6)" );
  tsAssert( ( fabs(vector3_y(&a) + rrt6) < VEC_TOL ) && "A = uX x uY.  A_y != -1/sqrt(6)" );
  tsAssert( ( fabs(vector3_z(&a) -  0.0) < VEC_TOL ) && "A = uX x uY.  A_z != 0.0" );

  vector3_set_with(&b,  1.0, 2.0, 3.0);
  vector3_set_with(&c, -1.0, 2.0, 1.0);
  vector3_cross(&a, &b,&c);
  tsAssert( ( fabs(vector3_x(&a) + 4.0) < VEC_TOL ) && "A = uX x uY.  A_x != -4.0" );
  tsAssert( ( fabs(vector3_y(&a) + 4.0) < VEC_TOL ) && "A = uX x uY.  A_y != -4.0" );
  tsAssert( ( fabs(vector3_z(&a) - 4.0) < VEC_TOL ) && "A = uX x uY.  A_z !=  4.0" );
}

void v3_test_triple_scalar_1(void)
{
  vector3 a,b,c;
  vector3_set_with(&a, 0.0, 0.0, 1.0);
  vector3_set_with(&b, 1.0, 0.0, 0.0);
  vector3_set_with(&c, 0.0, 1.0, 0.0);

  tsAssert( (fabs(vector3_scalar_triple(&a,&b,&c) - 1.0) < VEC_TOL) && "A.BxC != 1.0" );
  // vector3_z(&a) - 4.0) < VEC_TOL ) && "A = uX x uY.  A_z !=  4.0" );
}

void vector3_vector_tests(void)
{
  v3_test_can_init_pool();
  v3_test_init_pool_is_correct_size();

  v3_test_set_with();
  v3_test_zero();

  v3_test_set();
  v3_test_set_scaled();

  v3_test_set_scaled_array();
  v3_test_set_array();

  v3_test_get_elements();
  
  v3_test_add();
  v3_test_sub();

  v3_test_add_to();
  test_v3_sub_from();

  v3_test_displace();

  v3_test_scaleby();

  v3_test_set_to_midpoint();
  v3_test_add_multiple_to();
  v3_test_scale_and_add();

  v3_test_sqnorm();
  v3_test_norm();
  v3_test_normalize();
  v3_test_dot();

  v3_test_sqdistance();
  v3_test_distance();

  v3_test_cross_1();
  v3_test_triple_scalar_1();
}
#endif /* TS_ASSERTIONS_ENABLED */

