#ifndef VECTOR3_H
#define VECTOR3_H

#include <stdlib.h>
#include <assert.h>

#undef DEBUG_ALIGNMENT
#undef DEVEL_AVX

/*! \defgroup VectorMath Vector math library
 * @{
 */

/*! \def VECTOR_SINGLE_PRECISION
 *  \brief Macro which, if defined, signals that the vector math library
 *         uses single-precision elements instead of double-precision
 *         elements.  Default is to use double precision.
 */
#ifndef VECTOR_SINGLE_PRECISION
#  define VECTOR_SINGLE_PRECISION 0
#endif

#if defined(TS_USE_AVX)
#  include <immintrin.h>
#endif /* TS_USE_AVX */

#if defined(TS_USE_SSE)
#  include <emmintrin.h>
#  include <pmmintrin.h>
#endif /* TS_USE_SSE */

#define MODULE_NAME VECTOR3

#define SSE_ALIGNMENT 0x10
#define AVX_ALIGNMENT 0x20
#define HAS_ALIGNMENT( ptr, align ) ( (((uintptr_t)(ptr)) & (align-1) ) == 0 )
#define IS_SSE_ALIGNED( ptr ) (HAS_ALIGNMENT( ptr, SSE_ALIGNMENT ))
#define IS_AVX_ALIGNED( ptr ) (HAS_ALIGNMENT( ptr, AVX_ALIGNMENT ))

#if defined(DEBUG_ALIGNMENT)
#  define ASSERT_SSE_ALIGNED(ptr) do{assert(IS_SSE_ALIGNED(ptr));}while(0)
#  define ASSERT_AVX_ALIGNED(ptr) do{assert(IS_AVX_ALIGNED(ptr));}while(0)
#else
#  define ASSERT_SSE_ALIGNED(ptr) do{}while(0)
#  define ASSERT_AVX_ALIGNED(ptr) do{}while(0)
#endif /* defined(DEBUG_ALIGNMENT) */

#if defined(TS_USE_AVX)
#  if !defined(DEVEL_AVX)
#    error AVX is very development and should not be used
#  endif /* !defined(DEVEL_AVX) */
#    define DEF_SSE(x) x
#    define DEF_AVX(x) x
#elif defined(TS_USE_SSE)
#  define DEF_SSE(x) x
#  define DEF_AVX(x)
#else
#  define DEF_SSE(x)
#  define DEF_AVX(x)
#endif

/*
typedef union vector4 vector3;
typedef union vector4 vector4;
typedef union vector4 quat4;
*/

/*! \typedef VecReal
 *  \brief Floating point type, used to allow code to be either single-
 *  or double-precision.
 *
 *  The #VECTOR_SINGLE_PRECISION macro determines whether the vector3,
 *  quat4, mat3x3 and mat4x4 types are single- or double-precision.
 *  This type definition allows functions to be written for either
 *  precision.
 *  To allow numerical code to be either single- or double-precision, 
 */
#if VECTOR_SINGLE_PRECISION

typedef float VecReal;
#  define VEC_TOL 1e-4
#  define VEC_TOLSQ 1e-8

#else

typedef double VecReal;
#  define VEC_TOL 1e-8
#  define VEC_TOLSQ 1e-16

#endif

static inline VecReal square( VecReal x ) { return x*x; }


#if VECTOR_SINGLE_PRECISION
#  define VEC4_SSE_FIELDS DEF_SSE( __m128 vec )
#else /* !VECTOR_DOUBLE_PRECISION */
#  define VEC4_SSE_FIELDS \
  DEF_SSE( __m128d vec[2]; ) \
  DEF_AVX( __m256d vvec; )
#endif /* VECTOR_DOUBLE_PRECISION */


/*! \brief Macro defines common layout for 3- and 4-component math types.
 *
 * The vector math library uses a unified 4-component layout for all 3-
 * and 4-component values.  We use a macro to define to layout so it
 * remains the same for all 3- and 4-component types. The C type system
 * sees the types as different, which helps detect errors, but their
 * memory layout is the same, so they can be interconverted with a
 * simple cast.
 *
 * This also allows certain operations (eg: quaternion rotation of a
 * vector) to be performed more efficiently.
 *
 * This typically takes no additional memory.  Most modern systems have
 * 16-byte alignment for optimal memory access.  This means that
 * structures that contain doubles are typically padded up to 2N
 * doubles, and structures that contain floats are padded up to 4N
 * floats.  This essentially means that 3-component objects like 3-D
 * vectors take up the same amount of space as 4-component objects like
 * 4-D vectors and quaternions.  
 */
#define VEC4_FIELDS \
  struct {          \
    VecReal x; VecReal y; \
    VecReal z; VecReal w; \
  } xyz;            \
  VecReal elts[4];  \
  VEC4_SSE_FIELDS

/*! \brief vector4 defines a 4-element vector using VEC4_FIELDS layout */
typedef union { VEC4_FIELDS } vector4;

/*! \brief vector3 defines a 3-element vector using VEC4_FIELDS layout */
typedef union { VEC4_FIELDS } vector3;

/*! \brief quat4 defines a 4-element quaternion using VEC4_FIELDS layout */
typedef union { VEC4_FIELDS } quat4;


/*! \brief Initializes a vector3 to (0,0,0) */
#define VEC3_ZERO { .elts = {0} }

/*! \brief Initializes a vector3 to (x,y,z) */
#define VEC3_INIT(x,y,z) { .elts = { (x), (y), (z), 0.0 } }

#define QUAT4_IDENTITY  { .elts = { 0.0, 0.0, 0.0, 1.0 } }

typedef struct mat4x4 mat4x4;
/*! \brief A 4x4 matrix of VecReal */
struct mat4x4 {
  VecReal elts[16];
};

/*! \brief Represents a 3x3 matrix of VecReal. */
typedef struct mat3x3 mat3x3;
/*! \brief A 4x4 matrix of VecReal */
struct mat3x3 {
  VecReal elts[12];
};

#define MAT_NUM_ELTS(m) ( sizeof((m)->elts) / sizeof((m)->elts[0]) )

#define VECTOR3_X(v) ((v)->xyz.x) //!< Returns the X-component of the vector
#define VECTOR3_Y(v) ((v)->xyz.y) //!< Returns the Y-component of the vector
#define VECTOR3_Z(v) ((v)->xyz.z) //!< Returns the Z-component of the vector
#define VECTOR3_W(v) ((v)->xyz.w)

#define VECTOR4_X(v) ((v)->xyz.x) //!< Returns the X-component of the vector
#define VECTOR4_Y(v) ((v)->xyz.y) //!< Returns the Y-component of the vector
#define VECTOR4_Z(v) ((v)->xyz.z) //!< Returns the Z-component of the vector
#define VECTOR4_W(v) ((v)->xyz.w) //!< Returns the W-component of the vector

/*! \brief Returns the i-th component of the vector (i=0,1,2) */
#define VECTOR3_ELT(v,i) ( (v)->elts[ (i) ] )

#define QUAT4_X(v) ((v)->xyz.x) //!< Returns the X-component of the quaternion
#define QUAT4_Y(v) ((v)->xyz.y) //!< Returns the Y-component of the quaternion
#define QUAT4_Z(v) ((v)->xyz.z) //!< Returns the Z-component of the quaternion
#define QUAT4_W(v) ((v)->xyz.w) //!< Returns the W-component of the quaternion

/*! \brief Returns the i-th component of the quaternion (i=0,1,2,3) */
#define QUAT4_ELT(q,i) ( (q)->elts[ (i) ] )

#define VECTOR3_NUM_TEMPS 16

// Function called if the vector pool is exhausted
typedef void (*vector3_allocator_func)(size_t nvectors);

// Function called when the vector pool is freed
typedef void (*vector3_deallocator_func)(void* base, size_t nbytes);

// Initializes the vector pool
void vector3_init_pool(void* base, size_t nbytes, vector3_allocator_func vec_alloc, vector3_deallocator_func vec_dealloc);

// Adds memory to the vector pool
void vector3_add_pool(void* base, size_t nbytes);

// Returns the number of vectors in the pool
size_t vector3_pool_num_vectors(void);

// Returns the number of bytes used by the pool.  This includes bytes
// for the vectors and overhead.
size_t vector3_pool_num_bytes(void);

// Frees the vector pool.
void vector3_free_pool(void);

// Allocates n contiguous vectors from the pool
vector3* vector3_alloc(size_t n);

// Releases n contiguous vectors back to the pool
void vector3_release(vector3* v0, size_t n);

// Allocates a temporary vector, which disappears when the current
// function exits
vector3* vector3_alloc_temp(void);

void vector3_vector_tests(void);

#undef MODULE_NAME

#if defined(TS_USE_AVX)
#  include "vector3_inline_avx.h"
#  include "vector4_inline.h"
#  include "quat4_inline.h"
#  include "mat4x4_inline.h"
#  include "mat3x3_inline.h"
#elif defined(TS_USE_SSE)
#  include "vector3_inline_sse.h"
#  include "vector4_inline.h"
#  include "quat4_inline_sse.h"
#  include "mat4x4_inline.h"
#  include "mat3x3_inline.h"
#else
#  include "vector3_inline.h"
#  include "vector4_inline.h"
#  include "quat4_inline.h"
#  include "mat4x4_inline.h"
#  include "mat3x3_inline.h"
#endif /* TS_USE_AVX and TS_USE_SSE */

/** @} */

#endif /* VECTOR3_H */

