#ifndef VECTOR3_INLINE_H
#define VECTOR3_INLINE_H

#include "ts_errors.h"

#include <math.h>
#include <stddef.h>
#include <string.h>

#include "vector3_inline_common.h"

/*! \addtogroup VectorMath
 * @{
 */

/*! \brief Sets the components of vector v */
static inline void vector3_set_with(vector3*restrict v, VecReal x, VecReal y, VecReal z)
{
  v->xyz.x = x; 
  v->xyz.y = y; 
  v->xyz.z = z; 
  v->xyz.w = 0.0; 
}

/*! \brief Sets the componets of vector v to the elements of array arr
 * multiplied by scl
 *
 * Equivalent to 
 *
 *     VECTOR3_X(v) = scl * arr[0];
 *     VECTOR3_Y(v) = scl * arr[1];
 *     VECTOR3_Z(v) = scl * arr[2];
 *
 */
static inline void vector3_set_scaled_array(vector3*restrict v, VecReal scl, const VecReal*restrict arr)
{
  v->xyz.x = scl * arr[0];
  v->xyz.y = scl * arr[1];
  v->xyz.z = scl * arr[2];
  v->xyz.w = 0.0;
}

/*! \brief Sets the elements of array arr to the components of vector v
 * multiplied by scl
 *
 * Equivalent to
 *
 *     arr[0] = scl * VECTOR3_X(v);
 *     arr[1] = scl * VECTOR3_Y(v);
 *     arr[2] = scl * VECTOR3_Z(v);
 */
static inline void vector3_to_array_scaled(const vector3*restrict v, VecReal scl, VecReal arr[3])
{
  arr[0] = scl * vector3_x(v);
  arr[1] = scl * vector3_y(v);
  arr[2] = scl * vector3_z(v);
}

/*! \brief `v0 = scl*v1`
 *
 * Equivalent to
 *
 *     VECTOR3_X(v0) = scl * VECTOR3_X(v1);
 *     VECTOR3_Y(v0) = scl * VECTOR3_Y(v1);
 *     VECTOR3_Z(v0) = scl * VECTOR3_Z(v1);
 */
static inline void vector3_set_scaled(vector3*restrict v0, VecReal scl, const vector3*restrict v1) 
{ 
  v0->xyz.x = scl * v1->xyz.x; 
  v0->xyz.y = scl * v1->xyz.y; 
  v0->xyz.z = scl * v1->xyz.z; 
  v0->xyz.w = 0.0; // scl * v1->xyz.w; 
}

/*! \brief `v0 = v1`
 *
 * Equivalent to
 *
 *     VECTOR3_X(v0) = VECTOR3_X(v1);
 *     VECTOR3_Y(v0) = VECTOR3_Y(v1);
 *     VECTOR3_Z(v0) = VECTOR3_Z(v1);
 */
static inline void vector3_set(vector3*restrict v0, const vector3*restrict v1) 
{
  vector3_set_scaled(v0, 1.0, v1);
}

/*! \brief `v0 = -v1`
 *
 * Equivalent to
 *
 *     VECTOR3_X(v0) = -VECTOR3_X(v1);
 *     VECTOR3_Y(v0) = -VECTOR3_Y(v1);
 *     VECTOR3_Z(v0) = -VECTOR3_Z(v1);
 */
static inline void vector3_set_negative(vector3*restrict v0, const vector3*restrict v1) 
{
  vector3_set_scaled(v0, -1.0, v1);
}

/*! \brief In place negation: `v = -v` */
static inline void vector3_negate(vector3*restrict v)
{
  VECTOR3_X(v) = -vector3_x(v);
  VECTOR3_Y(v) = -vector3_y(v);
  VECTOR3_Z(v) = -vector3_z(v);
}

/*! \brief `v *= scl` */
static inline void vector3_scaleby(vector3*restrict v, VecReal scl)
{
  v->xyz.x *= scl;
  v->xyz.y *= scl;
  v->xyz.z *= scl;
}

/*! \brief `v += (dx,dy,dz)` */
static inline void vector3_displace(vector3*restrict v, VecReal dx, VecReal dy, VecReal dz)
{
  v->xyz.x += dx;
  v->xyz.y += dy;
  v->xyz.z += dz;
}

/*! \brief `v0 = (v0+v1)/2` */
static inline void vector3_set_to_midpoint(vector3*restrict v0, const vector3*restrict v1)
{
  v0->xyz.x = 0.5*(v0->xyz.x + v1->xyz.x);
  v0->xyz.y = 0.5*(v0->xyz.y + v1->xyz.y);
  v0->xyz.z = 0.5*(v0->xyz.z + v1->xyz.z);
  v0->xyz.w = 0.0;
}

/*! \brief `v0 += scl * v1` */
static inline void vector3_add_multiple_to(vector3*restrict v0, VecReal scl, const vector3*restrict v1)
{
  v0->xyz.x += scl * v1->xyz.x;
  v0->xyz.y += scl * v1->xyz.y;
  v0->xyz.z += scl * v1->xyz.z;
}

/*! \brief `v = sa*va + sb*vb` */
static inline void vector3_scale_and_add(vector3*restrict v, VecReal sa, const vector3* va, VecReal sb, const vector3* vb)
{
  v->xyz.x = sa*va->xyz.x + sb*vb->xyz.x;
  v->xyz.y = sa*va->xyz.y + sb*vb->xyz.y;
  v->xyz.z = sa*va->xyz.z + sb*vb->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief `v = v0 + v1` */
static inline void vector3_add(vector3*restrict v, const vector3*restrict v0, const vector3*restrict v1)
{
  v->xyz.x = v0->xyz.x + v1->xyz.x;
  v->xyz.y = v0->xyz.y + v1->xyz.y;
  v->xyz.z = v0->xyz.z + v1->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief `v += v0` */
static inline void vector3_add_to(vector3*restrict v, const vector3*restrict v0)
{
  v->xyz.x += v0->xyz.x;
  v->xyz.y += v0->xyz.y;
  v->xyz.z += v0->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief `v = v0 - v` */
static inline void vector3_subtract_from_and_set(vector3*restrict v, const vector3* v0)
{
  v->xyz.x = v0->xyz.x - v->xyz.x;
  v->xyz.y = v0->xyz.y - v->xyz.y;
  v->xyz.z = v0->xyz.z - v->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief `v -= v0 */
static inline void vector3_sub_from(vector3*restrict v, const vector3*restrict v0)
{
  v->xyz.x -= v0->xyz.x;
  v->xyz.y -= v0->xyz.y;
  v->xyz.z -= v0->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief `v = v0 - v1` */
static inline void vector3_sub(vector3*restrict v, const vector3*restrict v0, const vector3*restrict v1)
{
  v->xyz.x = v0->xyz.x - v1->xyz.x;
  v->xyz.y = v0->xyz.y - v1->xyz.y;
  v->xyz.z = v0->xyz.z - v1->xyz.z;
  v->xyz.w = 0.0;
}

/*! \brief Returns the squared norm of vector v */
static inline VecReal vector3_sqnorm(const vector3*restrict dr)
{
  return (dr->xyz.x * dr->xyz.x) + (dr->xyz.y * dr->xyz.y) + (dr->xyz.z * dr->xyz.z);
}

/*! \brief Returns the norm of vector dr */
static inline VecReal vector3_norm(const vector3*restrict dr)
{
  return sqrt( vector3_sqnorm(dr) );
}

/*! \brief Normalizes vector dr and returns the norm of the original
 * vector
 */
static inline VecReal vector3_normalize(vector3*restrict dr)
{
  VecReal mag = vector3_norm(dr);
  if (mag > VEC_TOL)
  {
    dr->xyz.x /= mag;
    dr->xyz.y /= mag;
    dr->xyz.z /= mag;
  }

  return mag;
}

/*! \brief `v *= scl` */
static inline void vector3_multiply_by(vector3*restrict v, VecReal scl)
{
  v->xyz.x *= scl;
  v->xyz.y *= scl;
  v->xyz.z *= scl;
}

/*! \brief Returns the dot product (inner product) of `va` and `vb`.
 * These cannot be the same vector.
 */
static inline VecReal vector3_dot(const vector3*restrict va, const vector3*restrict vb)
{
  tsAssert(va != vb && "va and vb cannot be the same vector");
  return 
    va->xyz.x * vb->xyz.x + 
    va->xyz.y * vb->xyz.y + 
    va->xyz.z * vb->xyz.z;
}

static inline VecReal vector3_sqdistance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  vector3_sub(dr, r1,r0);
  return vector3_sqnorm(dr);
}

static inline VecReal vector3_distance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  vector3_sub(dr, r1,r0);

  return vector3_norm(dr);
}

static inline void vector3_cross(vector3*restrict vc, const vector3*restrict va, const vector3*restrict vb)
{
  VECTOR3_X(vc) = vector3_y(va) * vector3_z(vb) - vector3_z(va) * vector3_y(vb);
  VECTOR3_Y(vc) = vector3_z(va) * vector3_x(vb) - vector3_x(va) * vector3_z(vb);
  VECTOR3_Z(vc) = vector3_x(va) * vector3_y(vb) - vector3_y(va) * vector3_x(vb);
}

static inline VecReal vector3_scalar_triple(const vector3*restrict va, const vector3*restrict vb, const vector3*restrict vc)
{
  vector3 tmp;
  vector3_cross(&tmp, vb,vc);
  return vector3_dot(va,&tmp);
}

/* @} */

#endif /* VECTOR3_INLINE_H */

