#ifndef VECTOR3_INLINE_COMMON_H
#define VECTOR3_INLINE_COMMON_H

#include <math.h>
#include <stddef.h>
#include <string.h>

#define MODULE_NAME VECTOR3_INLINE_COMMON

static inline VecReal vector3_x(const vector3*restrict v) { return v->xyz.x; }
static inline VecReal vector3_y(const vector3*restrict v) { return v->xyz.y; }
static inline VecReal vector3_z(const vector3*restrict v) { return v->xyz.z; }
static inline VecReal vector3_w(const vector3*restrict v) { return v->xyz.w; }

/*! \addtogroup VectorMath
 * @{
 */

/*! \brief Zeros vector v */
static inline void vector3_zero(vector3*restrict v) 
{ 
  memset(v->elts,0,sizeof(v->elts));
}

/*! \brief Extracts the of vector v into the memory locations xp, yp,
 * and zp
 */
static inline void vector3_get_elements(const vector3*restrict v, VecReal*restrict xp, VecReal*restrict yp, VecReal*restrict zp)
{
  *xp = vector3_x(v);
  *yp = vector3_y(v);
  *zp = vector3_z(v);
}

/*! \brief Extracts the of vector v into the memory locations xp, yp,
 * and zp
 */
static inline void vector3_get_array(const vector3*restrict v, VecReal elts[3])
{
  memcpy(elts,v->elts,3*sizeof(VecReal));
}

/*! \brief Sets the components of vector v to the elements of array arr
 *
 * Equivalent to 
 *
 *     VECTOR3_X(v) = arr[0];
 *     VECTOR3_Y(v) = arr[1];
 *     VECTOR3_Z(v) = arr[2];
 */
static inline void vector3_set_array(vector3*restrict v, const double*restrict arr)
{
  memcpy(v->elts,arr,3*sizeof(arr[0]));
  v->elts[3] = 0.0;
}

/*! \brief Sets the elements of array arr to the components of vector v
 *
 * Equivalent to
 *
 *     arr[0] = VECTOR3_X(v);
 *     arr[1] = VECTOR3_Y(v);
 *     arr[2] = VECTOR3_Z(v);
 */
static inline void vector3_to_array(const vector3*restrict v, VecReal arr[3])
{
  memcpy(arr,v->elts,3*sizeof(arr[0]));
}

static inline void vector3_to_array_scaled(const vector3*restrict v, VecReal scl, VecReal arr[3]);

/*! \brief Sets the elements of array arr to the components of vector -v
 *
 * Equivalent to
 * 
 *     arr[0] = -VECTOR3_X(v);
 *     arr[1] = -VECTOR3_Y(v);
 *     arr[2] = -VECTOR3_Z(v);
 */
static inline void vector3_negate_to_array(const vector3*restrict v, VecReal arr[3])
{
  vector3_to_array_scaled(v,-1.0,arr);
}

static inline void vector3_scaleby(vector3*restrict v, VecReal scl);

/*! \brief `v *= scl` */
static inline void vector3_scale_by(vector3*restrict v, VecReal scl) { vector3_scaleby(v,scl); }

#undef MODULE_NAME

#endif /* VECTOR3_INLINE_COMMON_H */

