#ifndef VECTOR3_INLINE_SSE_SINGLE_H
#define VECTOR3_INLINE_SSE_SINGLE_H

#include <string.h>
#include <math.h>

static inline void vector3_set_scaled_array(vector3*restrict v, double scl, const double*restrict arr)
{
  __m128d s = _mm_set1_pd(scl);
  __m128d vlo = _mm_loadu_pd(arr);
  __m128d vhi = _mm_load_sd(arr+2);
  
  vlo = _mm_mul_pd(s,vlo);
  vhi = _mm_mul_pd(s,vhi);

  v->vec[0] = vlo;
  v->vec[1] = vhi;
}

static inline void vector3_to_array_scaled(const vector3*restrict v, VecReal scl, VecReal arr[3])
{
  __m128d vscl = _mm_set1_pd(scl);
  _mm_storeu_pd(arr, _mm_mul_pd(v->vec[0],vscl));
  _mm_store_sd(arr+2, _mm_mul_pd(v->vec[1],vscl));
}

static inline void vector3_set_with(vector3*restrict v, double x, double y, double z)
{
  v->vec[0] = _mm_setr_pd(x,y); 
  v->vec[1] = _mm_set_sd(z);
}

static inline void vector3_set(vector3*restrict v0, const vector3*restrict v1) 
{
  v0->vec[0] = v1->vec[0];
  v0->vec[1] = v1->vec[1];
}

static inline void vector3_set_scaled(vector3*restrict dest, double scl, const vector3*restrict src) 
{ 
  __m128d vscl = _mm_set1_pd(scl);
  __m128d vlo = src->vec[0];
  __m128d vhi = src->vec[1];

  vlo = _mm_mul_pd(vscl,vlo);
  vhi = _mm_mul_pd(vscl,vhi);

  dest->vec[0] = vlo;
  dest->vec[1] = vhi;
}

static inline void vector3_set_negative(vector3*restrict v0, const vector3*restrict v1) 
{
  v0->vec[0] = -v1->vec[0];
  v0->vec[1] = -v1->vec[1];
}

static inline void vector3_negate(vector3*restrict v)
{
  v->vec[0] = -v->vec[0];
  v->vec[1] = -v->vec[1];
}

static inline void vector3_scaleby(vector3*restrict v, double scl)
{
  __m128d vscl = _mm_set1_pd(scl);
  __m128d vlo = v->vec[0];
  __m128d vhi = v->vec[1];
  vlo = _mm_mul_pd(vscl,vlo);
  vhi = _mm_mul_pd(vscl,vhi);
  v->vec[0] = vlo;
  v->vec[1] = vhi;
}

static inline void vector3_displace(vector3*restrict v, double dx, double dy, double dz)
{
  v->vec[0] += _mm_setr_pd(dx,dy);
  v->vec[1] += _mm_set_sd(dz);
}

static inline void vector3_set_to_midpoint(vector3*restrict dst, const vector3*restrict src)
{
  __m128d HALF = _mm_set1_pd(0.5);

  __m128d dlo = dst->vec[0];
  __m128d dhi = dst->vec[1];

  __m128d slo = src->vec[0];
  __m128d shi = src->vec[1];

  dlo = _mm_mul_pd(HALF,dlo);
  dhi = _mm_mul_pd(HALF,dhi);

  slo = _mm_mul_pd(HALF,slo);
  shi = _mm_mul_pd(HALF,shi);

  dlo = _mm_add_pd(dlo,slo);
  dhi = _mm_add_pd(dhi,shi);

  dst->vec[0] = dlo;
  dst->vec[1] = dhi;
}

static inline void vector3_add_to(vector3*restrict dest, const vector3*restrict v0)
{
  dest->vec[0] += v0->vec[0];
  dest->vec[1] += v0->vec[1];
}

static inline void vector3_add_multiple_to(vector3*restrict v0, double mul, const vector3*restrict v1)
{
  __m128d s = _mm_set1_pd(mul);

  __m128d v00 = v0->vec[0];
  __m128d v01 = v0->vec[1];

  __m128d v10 = v1->vec[0];
  __m128d v11 = v1->vec[1];

  v10 = _mm_mul_pd(s,v10);
  v11 = _mm_mul_pd(s,v11);

  v00 = _mm_add_pd(v00,v10);
  v01 = _mm_add_pd(v01,v11);

  v0->vec[0] = v00;
  v0->vec[1] = v01;
}

static inline void vector3_scale_and_add(vector3*restrict dst, double sa, const vector3*restrict va, double sb, const vector3*restrict vb)
{
  __m128d vsa = _mm_set1_pd(sa);
  __m128d vsb = _mm_set1_pd(sb);

  __m128d a0 = va->vec[0];
  __m128d a1 = va->vec[1];

  __m128d b0 = vb->vec[0];
  __m128d b1 = vb->vec[1];

  __m128d d0 = _mm_add_pd( _mm_mul_pd(a0,vsa), _mm_mul_pd(b0,vsb) );
  __m128d d1 = _mm_add_pd( _mm_mul_pd(a1,vsa), _mm_mul_pd(b1,vsb) );

  dst->vec[0] = d0;
  dst->vec[1] = d1;
}

static inline void vector3_add(vector3*restrict dest, const vector3*restrict v0, const vector3*restrict v1)
{
  dest->vec[0] = v0->vec[0] + v1->vec[0];
  dest->vec[1] = v0->vec[1] + v1->vec[1];
}

static inline void vector3_sub_from(vector3*restrict dest, const vector3*restrict vec)
{
  __m128d d0 = dest->vec[0];
  __m128d d1 = dest->vec[1];

  __m128d v0 = vec->vec[0];
  __m128d v1 = vec->vec[1];

  d0 = _mm_sub_pd(d0,v0);
  d1 = _mm_sub_pd(d1,v1);

  dest->vec[0] = d0;
  dest->vec[1] = d1;
}

static inline void vector3_sub(vector3*restrict dest, const vector3*restrict v0, const vector3*restrict v1)
{
  __m128d v00 = v0->vec[0];
  __m128d v01 = v0->vec[1];

  __m128d v10 = v1->vec[0];
  __m128d v11 = v1->vec[1];

  __m128d dv0 = _mm_sub_pd(v00,v10);
  __m128d dv1 = _mm_sub_pd(v01,v11);

  dest->vec[0] = dv0;
  dest->vec[1] = dv1;
}

static inline __m128d vector3_sqnorm_sse_val(__m128d v0, __m128d v1)
{
  __m128d zero = _mm_setzero_pd();

  __m128d sq0 = _mm_mul_pd(v0,v0);
  __m128d sq1 = _mm_mul_pd(v1,v1);

  sq1 = _mm_shuffle_pd(zero,sq1, _MM_SHUFFLE2(0,1));

  __m128d ssq = _mm_add_pd(sq0,sq1);

  return _mm_hadd_pd(ssq,ssq);
}

static inline __m128d vector3_sqnorm_sse(const vector3*restrict dr)
{
  __m128d v0 = dr->vec[0];
  __m128d v1 = dr->vec[1];

  return vector3_sqnorm_sse_val(v0,v1);
}

static inline double vector3_sqnorm(const vector3*restrict dr)
{
  return _mm_cvtsd_f64( vector3_sqnorm_sse(dr) );
}

static inline double vector3_norm(const vector3*restrict dr)
{
  return _mm_cvtsd_f64( _mm_sqrt_pd(vector3_sqnorm_sse(dr)) );
}

static inline double vector3_normalize(vector3*restrict dr)
{
  __m128d a = dr->vec[0];
  __m128d b = dr->vec[1];

  __m128d vn = _mm_sqrt_pd(vector3_sqnorm_sse_val(a,b));

  if (_mm_cvtsd_f64(vn) > VEC_TOL) {
    __m128d one = _mm_set1_pd(1.0);
    __m128d inv = _mm_div_pd(one,vn);

    dr->vec[0] = _mm_mul_pd(inv, a);
    dr->vec[1] = _mm_mul_pd(inv, b);
  }

  return _mm_cvtsd_f64( vn );
}

static inline void vector3_multiply_by(vector3*restrict v, double scl)
{
  vector3_scaleby(v,scl);
}

static inline double vector3_dot(const vector3*restrict a, const vector3*restrict b)
{
  __m128d a0 = a->vec[0];
  __m128d a1 = a->vec[1];

  __m128d b0 = b->vec[0];
  __m128d b1 = b->vec[1];

  __m128d v0 = _mm_mul_pd(a0,b0);
  __m128d v1 = _mm_mul_pd(a1,b1);

  __m128d v2 = _mm_add_pd(v0,v1);

  v2 = _mm_hadd_pd(v2,v2);
  // v2 += _mm_unpackhi_pd(v0,v1);

  return _mm_cvtsd_f64( v2 );
}

static inline __m128d vector3_sqdistance_sse(
    const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  __m128d r00 = r0->vec[0];  __m128d r01 = r0->vec[1];
  __m128d r10 = r1->vec[0];  __m128d r11 = r1->vec[1];

  __m128d d0 = _mm_sub_pd(r10,r00);
  __m128d d1 = _mm_sub_pd(r11,r01);

  __m128d sqd = _mm_add_pd(
      _mm_mul_pd(d0,d0),
      _mm_mul_pd(d1,d1));

  dr->vec[0] = d0;
  dr->vec[1] = d1;
  return _mm_hadd_pd(sqd,sqd);
}

static inline double vector3_sqdistance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  return _mm_cvtsd_f64( vector3_sqdistance_sse(r0,r1,dr) );
}

static inline double vector3_distance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  __m128d sqd = vector3_sqdistance_sse(r0,r1,dr);
  return _mm_cvtsd_f64( _mm_sqrt_pd(sqd) );
}

static inline void vector3_cross(vector3*restrict vc, const vector3*restrict va, const vector3*restrict vb)
{
  __m128d a0,a1, b0,b1;
  __m128d cx, cy, cz;
  a0 = va->vec[0];
  a1 = va->vec[1];

  b0 = vb->vec[0];
  b1 = vb->vec[1];

  cz = _mm_mul_pd(
      a0,                                        // (X1,Y1)
      _mm_shuffle_pd(b0,b0, _MM_SHUFFLE2(0,1))); // (Y2,X2)

  cx = _mm_mul_pd(
      _mm_shuffle_pd(a0,a1, _MM_SHUFFLE2(0,1)),  // (Y1,Z1)
      _mm_shuffle_pd(b1,b0, _MM_SHUFFLE2(1,0))); // (Z2,Y2)

  cy = _mm_mul_pd(
      _mm_shuffle_pd(a1,a0, _MM_SHUFFLE2(0,0)),  // (Z1,X1)
      _mm_shuffle_pd(b0,b1, _MM_SHUFFLE2(0,0))); // (X2,Z2)

  vc->vec[0] = _mm_hsub_pd(cx,cy);
  cz = _mm_hsub_pd(cz,cz);
  vc->vec[1] = _mm_move_sd(_mm_setzero_pd(), cz);

#if 0
  /*
  __m128d a0 = va->vec[0];  __m128d a1 = va->vec[1];
  __m128d b0 = vb->vec[0];  __m128d b1 = vb->vec[1];

  __m128d tmp = _mm_shuffle_pd(a0,a1, _MM_SHUFFLE2(0,
  */

  VECTOR3_X(vc) = vector3_y(va) * vector3_z(vb) - vector3_z(va) * vector3_y(vb);
  VECTOR3_Y(vc) = vector3_z(va) * vector3_x(vb) - vector3_x(va) * vector3_z(vb);
  VECTOR3_Z(vc) = vector3_x(va) * vector3_y(vb) - vector3_y(va) * vector3_x(vb);
#endif /* 0 */
}

static inline double vector3_scalar_triple(const vector3*restrict va, const vector3*restrict vb, const vector3*restrict vc)
{
  vector3 tmp;
  vector3_cross(&tmp, vb,vc);
  return vector3_dot(va,&tmp);
}

#endif /* VECTOR3_INLINE_SSE_SINGLE_H */

