#ifndef VECTOR3_INLINE_SSE_SINGLE_H
#define VECTOR3_INLINE_SSE_SINGLE_H

static inline void vector3_set_scaled_array(vector3*restrict v, float scl, const float*restrict arr)
{
  double tmp[2] = {arr[2], 0.0};

  __m128d vscl = _mm_set1_pd(scl);

  v->vec[0] = vscl * _mm_load_pd(arr);
  v->vec[1] = vscl * _mm_load_pd(tmp);
}

static inline void vector3_set_array(vector3*restrict v, const double*restrict arr)
{
  double tmp[2] = {arr[2], 0.0};

  v->vec[0] = _mm_load_pd(arr);
  v->vec[1] = _mm_load_pd(tmp);
}

static inline void vector3_zero(vector3*restrict v) 
{ 
  v->vec[0] = _mm_setzero_pd();
  v->vec[1] = _mm_setzero_pd();
}

static inline void vector3_set_with(vector3*restrict v, double x, double y, double z)
{
  v->vec[0] = _mm_setr_pd(x,y); 
  v->vec[1] = _mm_set_sd(z);
}

static inline void vector3_set(vector3*restrict v0, const vector3*restrict v1) 
{
  v0->vec[0] = v1->vec[0];
  v0->vec[1] = v1->vec[1];
}

static inline void vector3_set_scaled(vector3*restrict v0, double scl, vector3*restrict v1) 
{ 
  __m128d vscl = _mm_set1_pd(scl);
  v0->vec[0] = vscl * v1->vec[0];
  v0->vec[1] = vscl * v1->vec[1];
}

static inline void vector3_scaleby(vector3*restrict v, double scl)
{
  __m128d vscl = _mm_set1_pd(scl);
  v->vec[0] *= vscl;
  v->vec[1] *= vscl;
}

static inline void vector3_displace(vector3*restrict v, double dx, double dy, double dz)
{
  v->vec[0] += _mm_setr_pd(dx,dy);
  v->vec[1] += _mm_set_sd(dz);
}

static inline void vector3_set_to_midpoint(vector3*restrict v0, const vector3*restrict v1)
{
  __m128d vscl = _mm_set1_pd(0.5);

  v0->vec[0] += v1->vec[0];
  v0->vec[0] *= vscl;
  v0->vec[1] += v1->vec[1];
  v0->vec[1] *= vscl;
}

static inline void vector3_add_to(vector3*restrict dest, const vector3*restrict v0)
{
  dest->vec[0] += v0->vec[0];
  dest->vec[1] += v0->vec[1];
}

static inline void vector3_add_multiple_to(vector3*restrict v0, double mul, const vector3*restrict v1)
{
  __m128d vmul = _mm_set1_pd(mul);
  v0->vec[0] += vmul * v1->vec[0];
  v0->vec[1] += vmul * v1->vec[1];
}

static inline void vector3_scale_and_add(vector3*restrict dst, double sa, vector3* va, double sb, vector3* vb)
{
  __m128d vsa = _mm_set1_pd(sa);
  __m128d vsb = _mm_set1_pd(sb);
  dst->vec[0] = vsa * va->vec[0] + vsb * vb->vec[0];
  dst->vec[1] = vsa * va->vec[1] + vsb * vb->vec[1];
}

static inline void vector3_add(vector3*restrict dest, vector3*restrict v0, vector3*restrict v1)
{
  dest->vec[0] = v0->vec[0] + v1->vec[0];
  dest->vec[1] = v0->vec[1] + v1->vec[1];
}

static inline void vector3_sub_from(vector3*restrict dest, const vector3*restrict v0)
{
  dest->vec[0] -= v0->vec[0];
  dest->vec[1] -= v0->vec[1];
}

static inline void vector3_sub(vector3*restrict dest, const vector3*restrict v0, const vector3*restrict v1)
{
  dest->vec[0] = v0->vec[0] - v1->vec[0];
  dest->vec[1] = v0->vec[1] - v1->vec[1];
}

static inline double vector3_sqnorm(const vector3*restrict dr)
{
  __m128d v0 = dr->vec[0] * dr->vec[0];
  __m128d v1 = dr->vec[1] * dr->vec[1];

  __m128d v2 = v0 + v1;

  v2 += _mm_unpackhi_pd(v0,v1);
  return _mm_cvtsd_f64(v2);
}

static inline double vector3_norm(const vector3*restrict dr)
{
  __m128d v0 = dr->vec[0] * dr->vec[0];
  __m128d v1 = dr->vec[1] * dr->vec[1];

  __m128d v2 = v0 + v1;

  v2 += _mm_unpackhi_pd(v0,v1);
  return _mm_cvtsd_f64( _mm_sqrt_pd(v2) );
}

static inline double vector3_normalize(vector3*restrict dr)
{
  __m128d v0 = dr->vec[0] * dr->vec[0];
  __m128d v1 = dr->vec[1] * dr->vec[1];

  __m128d v2 = v0 + v1;

  v2 += _mm_unpackhi_pd(v0,v1);
  v1 = _mm_unpacklo_pd(v2,v2);

  v2 = _mm_sqrt_pd(v1);
  if (_mm_comigt_sd(v2, _mm_set1_pd(VEC_TOL))) {
    dr->vec[0] /= v2;
    dr->vec[1] /= v2;
  }

  return _mm_cvtsd_f64( v2 );
}

static inline void vector3_multiply_by(vector3*restrict v, double scl)
{
  vector3_scaleby(v,scl);
}

static inline double vector3_dot(const vector3*restrict a, const vector3*restrict b)
{
  __m128d v0 = a->vec[0] * b->vec[0];
  __m128d v1 = a->vec[1] * b->vec[1];

  __m128d v2 = v0 + v1;

  v2 += _mm_unpackhi_pd(v0,v1);

  return _mm_cvtsd_f64( v2 );
}

static inline double vector3_sqdistance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  vector3_sub(dr, r1,r0);
  return vector3_sqnorm(dr);
}

static inline double vector3_distance(const vector3*restrict r0, const vector3*restrict r1, vector3*restrict dr)
{
  vector3_sub(dr, r1,r0);

  return vector3_norm(dr);
}

static inline void vector3_cross(vector3*restrict vc, const vector3*restrict va, const vector3*restrict vb)
{
  VECTOR3_X(vc) = vector3_y(va) * vector3_z(vb) - vector3_z(va) * vector3_y(vb);
  VECTOR3_Y(vc) = vector3_z(va) * vector3_x(vb) - vector3_x(va) * vector3_z(vb);
  VECTOR3_Z(vc) = vector3_x(va) * vector3_y(vb) - vector3_y(va) * vector3_x(vb);
}

static inline double vector3_scalar_triple(const vector3*restrict va, const vector3*restrict vb, const vector3*restrict vc)
{
  vector3 tmp;
  vector3_cross(&tmp, vb,vc);
  return vector3_dot(va,&tmp);
}


#endif /* VECTOR3_INLINE_SSE_SINGLE_H */

