#include <string.h>
#include <linenoise.h>
#include <mruby.h>
#include <mruby/compile.h>
#include <mruby/string.h>

static mrb_value repl_readline(mrb_state *mrb, mrb_value mod)
{
  const char *prompt;
  char *in;
  mrb_value input;

  prompt = NULL;
  mrb_get_args(mrb, "z", &prompt);

  if (in = linenoise(prompt), in == NULL) {
    return mrb_nil_value();
  }

  linenoiseHistoryAdd(in);
  input = mrb_str_new_cstr(mrb, in);
  free(in);

  return input;
}

static void (*prev_code_fetch_hook)(struct mrb_state* mrb, struct
    mrb_irep *irep, mrb_code *pc, mrb_value *regs) = NULL;
static char prompt[64];

/* adapted from mrdb debugger */

static void
check_syntax(mrb_state *mrb, const char *expr, size_t len)
{
  mrbc_context *c;

  c = mrbc_context_new(mrb);
  c->no_exec = TRUE;
  c->capture_errors = TRUE;
  c->filename = "<input>";
  c->lineno = 1;

  /* Load program */
  mrb_load_nstring_cxt(mrb, expr, len, c);

  mrbc_context_free(mrb, c);
}

static mrb_value
do_eval(mrb_state *mrb, mrb_value recv, const char *expr, size_t len, mrb_bool *exc)
{
  void (*tmp)(struct mrb_state *, struct mrb_irep *, mrb_code *, mrb_value *);
  mrb_value ruby_code, s,v;

  /* disable code_fetch_hook */
  tmp = mrb->code_fetch_hook;
  mrb->code_fetch_hook = NULL;

  check_syntax(mrb, expr, len);
  if (mrb->exc) {
    v = mrb_obj_value(mrb->exc);
    mrb->exc = NULL;
    goto report;
  }

  /*
   * begin
   *   expr
   * rescue => e
   *   e
   * end
   */
  ruby_code = mrb_str_new_lit(mrb, "begin\n");
  ruby_code = mrb_str_cat(mrb, ruby_code, expr, len);
  ruby_code = mrb_str_cat_lit(mrb, ruby_code, "\nrescue => e\ne\nend");

  v =  mrb_funcall(mrb, recv, "instance_eval", 1, ruby_code);

  if (exc) {
    *exc = mrb_obj_is_kind_of(mrb, v, mrb->eException_class);
  }

report:
  s = mrb_funcall(mrb, v, "inspect", 0);

  /* enable code_fetch_hook */
  mrb->code_fetch_hook = tmp;

  return s;
}

static void break_hook(struct mrb_state* mrb,
    struct mrb_irep *irep, mrb_code *pc, mrb_value *regs)
{
  char *in;
  mrb_value recv,result;

  mrb->code_fetch_hook = prev_code_fetch_hook;
  prev_code_fetch_hook = NULL;

  recv = regs[0];

  for(;;) {
    int ai;
    mrb_bool exc;

    if (in = linenoise(prompt), in == NULL) {
      return;
    }

    // fprintf(stderr, "input: z%sz\n",in);
    if (strcmp(in, "\\\\q") == 0) {
      free(in);
      return;
    }

    if (strcmp(in,"!!") == 0) {
      /* call Kernel.exit */
      struct RClass *kernel_mod;

      free(in); /* be responsible... */

      kernel_mod = mrb_module_get(mrb, "Kernel");
      if (kernel_mod == NULL) {
        mrb_raise(mrb, E_RUNTIME_ERROR, "Cannot find Kernel module!");
      }

      mrb_funcall(mrb, mrb_obj_value(kernel_mod), "exit", 0);
      return; /* should not reach this point */
    }

    linenoiseHistoryAdd(in);

    ai = mrb_gc_arena_save(mrb);
    exc = 0;
    result = do_eval(mrb, recv, in, strlen(in), &exc);
    if (exc) {
      printf("Error!:\n  %s\n", mrb_string_value_cstr(mrb, &result));
    } else {
      printf("%s\n", mrb_string_value_cstr(mrb, &result));
    }

    free(in);
    mrb_gc_arena_restore(mrb,ai);
  }
}

static mrb_value repl_break(mrb_state *mrb, mrb_value mod)
{
  const char *p;
  size_t len;

  p = NULL;
  mrb_get_args(mrb, "|z", &p);

  if (p == NULL) { p = "> "; }
  len = strlen(p);

  if (len >= sizeof prompt ) { len = sizeof prompt - 1; }

  memcpy(prompt,p,len);
  prompt[len] = '\0';

  prev_code_fetch_hook = mrb->code_fetch_hook;
  mrb->code_fetch_hook = &break_hook;
  return mrb_nil_value();
}

void
mrb_mruby_repl_gem_init(mrb_state *mrb)
{
  struct RClass *mod;
  int ai;

  ai = mrb_gc_arena_save(mrb);
  mod = mrb_define_module(mrb, "REPL");

  mrb_define_module_function(mrb, mod, "readline", repl_readline, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mod, "break", repl_break, MRB_ARGS_NONE());
  /* mrb_define_module_function(mrb, mod, "history", repl_get_history, MRB_ARGS_NONE()); */
  /* mrb_define_module_function(mrb, mod, "clear_history!", repl_clear_history, MRB_ARGS_NONE()); */
  /* mrb_define_module_function(mrb, mod, "save_history", repl_save_history, MRB_ARGS_REQ(1)); */
  /* mrb_define_module_function(mrb, mod, "load_history", repl_load_history, MRB_ARGS_REQ(1)); */
  /* mrb_define_module_function(mrb, mod, "max_history", repl_get_max_history, MRB_ARGS_NONE()); */
  /* mrb_define_module_function(mrb, mod, "max_history=", repl_set_max_history, MRB_ARGS_REQ(1)); */
}

void
mrb_mruby_repl_gem_final(mrb_state *mrb)
{
}
