#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "mruby.h"
#include "mruby/array.h"
#include "mruby/class.h"
#include "mruby/data.h"
#include "mruby/string.h"
#include "mruby/value.h"
#include "mruby/variable.h"

#define E_IO_ERROR      (mrb_class_get(mrb, "IOError"))
#define E_EOF_ERROR     (mrb_class_get(mrb, "IOError"))

#if defined(_WIN32)
#  define USE_WIN32
#elif defined(__unix__) || defined(__unix) || (defined(__APPLE__) && defined(__MACH__))
#  include <unistd.h>
#  include <sys/stat.h>
#  if defined(_POSIX_VERSION)
#    include <libgen.h>
#    include <sys/param.h>
#    define USE_POSIX 1
#  endif  /* defined(_POSIX_VERSION) */
#endif /* defined(_WIN32) */

#ifdef __GNUC__
#  define noreturn  __attribute__((noreturn))
#else
#  define noreturn 
#endif

struct mruby_io {
  FILE *f;
  int sync;
  void *user;
};

static void
io_free(mrb_state* mrb, void* p)
{
  if (p) {
    struct mruby_io* ios = p;
    if (ios->f) {
      fclose(ios->f);
    }
    mrb_free(mrb, p);
  }
}

static const mrb_data_type io_type = {
  .struct_name = "IO",
  .dfree = io_free,
};

static void noreturn
io_raise(mrb_state *mrb, const char *mesg)
{
  mrb_raise(mrb, E_IO_ERROR, mesg);
}

static void noreturn
io_raise_errno(mrb_state *mrb)
{
  const char *mesg = strerror(errno);
  io_raise(mrb, mesg);
}

static void noreturn
eof_raise(mrb_state *mrb)
{
  mrb_raise(mrb, E_EOF_ERROR, "end of file encountered");
}

static struct RClass *
io_class(mrb_state *mrb)
{
  return mrb_class_get(mrb, "IO");
}

static struct RClass *
file_class(mrb_state *mrb)
{
  return mrb_class_get(mrb, "File");
}

static void
io_file_check(mrb_state *mrb, mrb_value obj)
{
  if (!mrb_obj_is_kind_of(mrb, obj, io_class(mrb))) {
    mrb_raise(mrb, E_TYPE_ERROR, "expected an IO object");
  }
}

static struct mruby_io *
io_get_ios(mrb_state *mrb, mrb_value io)
{
  io_file_check(mrb,io);
  return DATA_PTR(io);
}

static FILE *
io_file_get(mrb_state *mrb, mrb_value io)
{
  struct mruby_io *ios;

  ios = io_get_ios(mrb, io);
  return ios->f;
}

int
mrb_io_closed_p(mrb_state *mrb, mrb_value io)
{
  return io_file_get(mrb, io) == NULL;
}

FILE *
mrb_io_file_noerr(mrb_state *mrb, mrb_value io)
{
  return io_file_get(mrb, io);
}

static struct mruby_io *
ios_new(mrb_state *mrb, FILE *f, int sync)
{
  struct mruby_io *ios;

  ios = mrb_malloc(mrb, sizeof(*ios));
  memset(ios,0,sizeof(*ios));

  ios->f = f;
  ios->sync = sync;

  return ios;
}

static int
io_sync_default(FILE *f)
{
#if USE_POSIX
  int fd;

  fd = fileno(f);
  if (fd < 0) {
    return 0;
  }

  return isatty(fd);
#else
  return 0;
#endif
}

static void
io_close(mrb_state *mrb, mrb_value io)
{
  struct mruby_io *ios;

  io_file_check(mrb,io);
  ios = DATA_PTR(io);

  if (ios->f == NULL) {
    return;
  }

  if (fclose(ios->f) == EOF) {
    io_raise_errno(mrb);
  }

  ios->f = NULL;
  ios->sync = 0;
}

static void
io_file_set(mrb_state *mrb, mrb_value io, FILE *f, int sync)
{
  struct mruby_io *ios;

  io_file_check(mrb,io);
  ios = DATA_PTR(io);

  if (ios == NULL) {
    ios = ios_new(mrb, f, sync);
    DATA_PTR(io) = ios;
    return;
  }

  if ((ios->f != NULL) && (ios->f != f)) {
    if (errno = 0, fclose(ios->f) != 0) {
      io_raise_errno(mrb);
    }
  }

  ios->f = f;
  ios->sync = sync;
}

static mrb_value
char2str(mrb_state *mrb, char c)
{
  return mrb_str_new(mrb, &c, 1);
}

FILE *
mrb_io_file(mrb_state *mrb, mrb_value io)
{
  FILE *f;

  f = io_file_get(mrb, io);
  if (f == NULL) {
    io_raise(mrb, "already closed");
  }

  return f;
}

static mrb_value
io_initialize(mrb_state *mrb, mrb_value self, FILE *f)
{
  io_file_set(mrb, self, f, io_sync_default(f));
  return self;
}

static mrb_value
io_open_with_class(mrb_state *mrb, struct RClass *cls, const mrb_data_type *dtype, FILE *f)
{
  struct RData *obj;
  struct mruby_io *ios;

  ios = ios_new(mrb, f, io_sync_default(f));
  obj = mrb_data_object_alloc(mrb, cls, ios, dtype);
  return mrb_obj_value(obj);
}

mrb_value
mrb_io_open(mrb_state *mrb, FILE *f)
{
  return io_open_with_class(mrb, io_class(mrb), &io_type, f);
}

mrb_value
mrb_file_open(mrb_state *mrb, const char *path, const char *mode)
{
  FILE *f;

  if (errno = 0, f = fopen(path,mode), f == NULL) {
    io_raise_errno(mrb);
  }

  return io_open_with_class(mrb, file_class(mrb), &io_type, f);
}

/** IO Methods **/

static mrb_value
mIO_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_raise(mrb, E_NOTIMP_ERROR,
      "cannot directly instantiate an object of class IO");
  return mrb_nil_value();
}

/* IO.open(*args, &block) singleton
 *
 * implemnented in ruby
 *
 */

/* close
 *
 * a) If the receiver is closed, raise a direct instance of the class IOError.
 *
 * b) If the buffering flag of the receiver is true, and the receiver is
 *    buffering any output, immediately write all the buffered data to
 *    the stream which the receiver represents.
 *
 * c) Set the openness flag of the receiver to false.
 *
 * d) Return an implementation-defined value.
 *
 */

static mrb_value
mIO_close(mrb_state *mrb, mrb_value self)
{
  io_close(mrb,self);
  return mrb_nil_value();
}

/* closed?
 *
 * a) If the receiver is closed, return true.
 * b) Otherwise, return false.
 * 
 */
static mrb_value
mIO_closed_p(mrb_state *mrb, mrb_value self)
{
  return mrb_bool_value( mrb_io_closed_p(mrb, self) );
}

static mrb_value
io_readline(mrb_state *mrb, FILE *f)
{
  char buf[255], *s;
  size_t len;
  mrb_value str;

  str = mrb_str_buf_new(mrb, sizeof(buf));
  do {
    s = fgets(buf,sizeof(buf),f);
    if (s == NULL) {
      goto error;
    }

    str = mrb_str_cat_cstr(mrb, str, s);

    len = RSTRING_LEN(str);
  } while (RSTRING_PTR(str)[len-1] != '\n');

  return str;

error:
  if (feof(f)) {
    if (RSTRING_LEN(str) > 0) {
      return str;
    }

    return mrb_nil_value();
  }

  io_raise_errno(mrb);
  return mrb_nil_value();
}

/* each(&block)
 *
 * a) If block is not given, the behavior is unspecified.
 * b) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * c) If the receiver has reached its end, return the receiver.
 * d) Otherwise, read characters from the receiver until 0x0a is read or
 *    the receiver reaches its end.
 * e) Create a direct instance of the class String whose content is the
 *    sequence of characters read in Step d). Call block with this
 *    instance as an argument.
 * f) Continue processing from Step c).
 *
 */
static mrb_value
mIO_each(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  int ai;
  mrb_value str, block;

  f = mrb_io_file(mrb, self);

  block = mrb_nil_value();
  mrb_get_args(mrb,"&", &block);
  if (mrb_nil_p(block)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires a block");
  }

  for(;;) {
    ai = mrb_gc_arena_save(mrb);
    str = io_readline(mrb, f);

    if (mrb_nil_p(str)) {
      break;
    }

    mrb_yield(mrb, block, str);
    mrb_gc_arena_restore(mrb,ai);
  }

  return self;
}

/* each_byte(&block)
 *
 * a) If block is not given, the behavior is unspecified.
 * b) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * c) If the receiver has reached its end, return the receiver.
 * d) Otherwise, read a single byte from the receiver. Call block with
 *    an argument, an instance of the class Integer whose value is the
 *    byte.
 * e) Continue processing from Step c).
 *
 */
static mrb_value
mIO_each_byte(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  int ch;
  mrb_value block;

  f = mrb_io_file(mrb, self);

  block = mrb_nil_value();
  mrb_get_args(mrb,"&", &block);
  if (mrb_nil_p(block)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires a block");
  }

  while (ch = fgetc(f), ch != EOF) {
    mrb_yield(mrb, block, mrb_fixnum_value(ch));
  }

  if (ferror(f)) {
    io_raise_errno(mrb);
  }

  return self;
}

/* each_line(&block)
 *
 * Same as the method each (see 15.2.20.5.3).
 *
 */

/* eof?
 *
 * a) If the receiver is not readable, raise a direct instance of the class
 *    IOError.
 * b) If the receiver has reached its end, return true. Otherwise,
 *    return false.
 * 
 */
static mrb_value
mIO_eof_p(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  f = mrb_io_file(mrb, self);
  return mrb_bool_value( feof(f) );
}

/* flush
 *
 * a) If the receiver is not writable, raise a direct instance of the
 *    class IOError.
 * b) If the buffering flag of the receiver is true, and the receiver is
 *    buffering any output, immediately write all the buffered data to
 *    the stream which the receiver represents. 
 * c) Return the receiver.
 *
 */
static mrb_value
mIO_flush(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  f = mrb_io_file(mrb, self);

  if (errno = 0, fflush(f) == EOF) {
    io_raise_errno(mrb);
  }

  return mrb_bool_value( feof(f) );
}

/* getc
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError. If the receiver has reached its end, return nil.
 * b) Otherwise, read a character from the receiver. 
 * c) Return an instance of the class Object which represents the
 *    character (see 15.2.10.1).
 *
 */
static mrb_value
mIO_getc(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  int ch;

  f = mrb_io_file(mrb, self);

  errno = 0;
  ch = fgetc(f);
  if (ch != EOF) {
    return char2str(mrb,ch);
  }

  if (ferror(f)) {
    io_raise_errno(mrb);
  }

  return mrb_nil_value();
}

/* gets
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * b) If the receiver has reached its end, return nil.
 * c) Otherwise, read characters from the receiver until 0x0a is read or
 *    the receiver reaches its end.
 * d) Create a direct instance of the class String whose content is the
 *    sequence of characters read in Step c) and return this instance.
 *
 */
static mrb_value
mIO_gets(mrb_state *mrb, mrb_value self)
{
  FILE *f;

  f = mrb_io_file(mrb, self);
  if (!feof(f)) { return io_readline(mrb, f); }

  return mrb_nil_value();
}

/* initialize_copy(orig)  <-- unspecified */

/* print(*args)
 *
 * a) For each element of args in the indexing order:
 *    1) Let V be the element. If the element is nil, a conforming
 *       processor may create a direct instance of the class String
 *       whose content is “nil” and let V be the instance.
 *    2) Invoke the method write on the receiver with V as the argument.
 * b) Return an implementation-defined value.
 *
 * Implemented in Ruby
 */

/* putc(obj)
 *
 * a) If obj is not an instance of the class Integer or an instance of
 *    the class String, the behavior is unspecified. If obj is an
 *    instance of the class Integer whose value is smaller than 0 or
 *    larger than 255, the behavior is unspecified.
 * b) If obj is an instance of the class Integer, create a direct
 *    instance S of the class String whose content is a single
 *    character, whose character code is the value of obj.
 * c) If obj is an instance of the class String, create a direct
 *    instance S of the class String whose content is the first
 *    character of obj .
 * d) Invoke the method write on the receiver with S as the argument.
 * e) Return obj.
 *
 * Implemented in Ruby
 */


/* puts(*args)
 *
 * a) If the length of args is 0, create a direct instance of the class
 *    String whose content is a single character 0x0a and invoke the
 *    method write on the receiver with this instance as an argument.
 * b) Otherwise, for each element E of args in the indexing order:
 *    1) If E is an instance of the class Array, for each element X of E
 *       in the indexing order:
 *       i)  If X is the same object as E, i.e. if E contains itself,
 *           invoke the method write on the receiver with an instance of
 *           the class String, whose content is implementation-defined.
 *       ii) Otherwise, invoke the method write on the receiver with X
 *           as the argument.
 *    2) Otherwise:
 *       i)   If E is nil, a conforming processor may create a direct
 *            instance of the class String whose content is “nil” and
 *            let E be the instance.
 *       ii)  If E is not an instance of the class String, invoke the
 *            method to s on the E. If the resulting value is an
 *            instance of the class String, let E be the resulting
 *            value. Otherwise, the behavior is unspecified.
 *       iii) Invoke the method write on the receiver with E as the
 *            argument.
 *       iv)  If the last character of E is not 0x0a, create a direct
 *            instance of the class String whose content is a single
 *            character 0x0a and invoke the method write on the receiver
 *            with this instance as an argument.
 * c) Return an implementation-defined value.
 *
 * Implemented in Ruby
 */

/* read(length=nil)
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * b) If the receiver has reached its end:
 *    1) If length is nil, create an empty instance of the class String
 *       and return that instance.
 *    2) If length is not nil, return nil.
 * c) Otherwise:
 *    1) If length is nil, read characters from the receiver until the
 *       receiver reaches its end.
 *    2) If length is an instance of the class Integer, let N be the
 *       value of length. Other- wise, the behavior is unspecified.
 *    3) If N is smaller than 0, raise a direct instance of the class
 *       ArgumentError.
 *    4) Read bytes from the receiver until N bytes are read or the
 *       receiver reaches its end.
 * d) Create a direct instance of the class String whose content is the
 *    sequence of characters read in Step c) and return this instance.
 *
 */
static mrb_value
mIO_read(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  int has_len;
  char buf[1024];
  mrb_int len;
  mrb_value s;

  f = mrb_io_file(mrb, self);

  len = 0;
  has_len = (mrb_get_args(mrb, "|i", &len) > 0);

  if (feof(f)) {
    return has_len ? mrb_nil_value() : mrb_str_new_lit(mrb,"");
  }

  if (has_len < 0) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "length must be >= 0");
  }

  if (has_len) {
    size_t max, count, n = 0;

    max = len;
    s = mrb_str_buf_new(mrb, max);

    for (n=0; !feof(f) && (n+sizeof(buf) < max); n += count) {
      errno = 0;
      count = fread(buf, 1, sizeof(buf), f);
      if ((count < sizeof(buf)) && ferror(f)) {
        io_raise_errno(mrb);
      }

      s = mrb_str_cat(mrb, s, buf, count);
    }

    if (!feof(f) && (n < max)) {
      size_t spill = max - n;
      count = fread(buf, 1, spill, f);
      if ((count < spill) && ferror(f)) {
        io_raise_errno(mrb);
      }

      s = mrb_str_cat(mrb, s, buf, count);
    }
  } else {
    s = mrb_str_buf_new(mrb, sizeof(buf));
    while (!feof(f)) {
      size_t count = fread(buf, 1, sizeof(buf), f);
      if ((count < sizeof(buf)) && ferror(f)) {
        io_raise_errno(mrb);
      }
      s = mrb_str_cat(mrb, s, buf, count);
    }
  }

  return s;
}

/* readchar
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * b) If the receiver has reached its end, raise a direct instance of
 *    the class EOFError.
 * c) Otherwise, read a character from the receiver. Return an instance
 *    of the class Object which represents the character.
 *
 */
static mrb_value
mIO_readchar(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  int ch;

  f = mrb_io_file(mrb, self);

  ch = getc(f);
  if (ch == EOF) {
    if (ferror(f)) {
      io_raise_errno(mrb);
    }
    else {
      eof_raise(mrb);
    }
  }

  return char2str(mrb,ch);
}

/* readline
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * b) If the receiver has reached its end, raise a direct instance of
 *    the class EOFError.
 * c) Otherwise, read characters from the receiver until 0x0a is read or
 *    the receiver reaches its end.
 * d) Create a direct instance of the class String whose content is the
 *    sequence of characters read in Step c) and return this instance.
 *
 */
static mrb_value
mIO_readline(mrb_state *mrb, mrb_value self)
{
  FILE *f;

  f = mrb_io_file(mrb, self);

  if (feof(f)) {
    eof_raise(mrb);
  }

  return io_readline(mrb, f);
}

/* readlines
 *
 * a) If the receiver is not readable, raise a direct instance of the
 *    class IOError.
 * b) Create an empty direct instance A of the class Array.
 * c) If the receiver has reached to its end, return A.
 * d) Otherwise, read characters from the receiver until 0x0a is read or
 *    the receiver reaches its end.
 * e) Create a direct instance of the class String whose content is the
 *    sequence of characters read in Step d) and append this instance to
 *    A.
 * f) Continue processing from Step c).
 *
 */
static mrb_value
mIO_readlines(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  mrb_value arr;

  f = mrb_io_file(mrb, self);
  arr = mrb_ary_new(mrb);

  while (!feof(f)) {
    int ai;
    ai = mrb_gc_arena_save(mrb);

    mrb_value s = io_readline(mrb, f);
    mrb_ary_push(mrb, arr, s);

    mrb_gc_arena_restore(mrb,ai);
  }

  return arr;
}

/* sync */
static mrb_value
mIO_get_sync(mrb_state *mrb, mrb_value self)
{
  struct mruby_io *ios;
  ios = io_get_ios(mrb,self);
  return mrb_bool_value(ios->sync);
}

/* sync=(bool) */
static mrb_value
mIO_set_sync(mrb_state *mrb, mrb_value self)
{
  struct mruby_io *ios;
  mrb_bool b;

  b = 0;
  mrb_get_args(mrb, "b", &b);

  ios = io_get_ios(mrb,self);
  ios->sync = (int)b;

  return mrb_bool_value(0);
}

/* write(str) */
static mrb_value
mIO_write(mrb_state *mrb, mrb_value self)
{
  struct mruby_io *ios;
  mrb_value obj;
  const char *s;
  size_t count,len;

  ios = io_get_ios(mrb,self);

  if (ios->f == NULL) {
    io_raise(mrb, "already closed");
  }

  obj = mrb_nil_value();
  mrb_get_args(mrb, "o", &obj);

  if (mrb_type(obj) != MRB_TT_STRING) {
    obj = mrb_any_to_s(mrb, obj);
    if (mrb_type(obj) != MRB_TT_STRING) {
      mrb_raise(mrb, E_TYPE_ERROR, "Object return by to_s must be a string");
    }
  }

  s   = RSTRING_PTR(obj);
  len = RSTRING_LEN(obj);

  if (len == 0) {
    return mrb_fixnum_value(0);
  }

  errno = 0;
  count = fwrite(s, 1, len, ios->f);
  if (count < len) {
    if (ferror(ios->f)) {
      io_raise_errno(mrb);
    }
    eof_raise(mrb);
  }

  if (ios->sync) {
    fflush(ios->f);
  }

  return mrb_fixnum_value(count);
}

static mrb_value
mIO_seek(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  mrb_int off, wh; 

  f = mrb_io_file(mrb, self);

  wh = SEEK_SET;
  mrb_get_args(mrb, "i|i", &off,&wh);

  if (fseek(f, off, wh) < 0) {
    io_raise_errno(mrb);
  }

  return mrb_fixnum_value(0);
}

static mrb_value
mIO_rewind(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  f = mrb_io_file(mrb, self);

  if (fseek(f, 0, SEEK_SET) < 0) {
    io_raise_errno(mrb);
  }

  return mrb_fixnum_value(0);
}

static mrb_value
mIO_ungetc(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  char *s;
  mrb_int slen;
  f = mrb_io_file(mrb, self);

  s = NULL;
  slen = 0;
  mrb_get_args(mrb, "s", &s, &slen);
  if (slen > 0) {
    ungetc(s[0],f);
  }

  return mrb_nil_value();
}

static mrb_value
mIO_tell(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  long off;
  f = mrb_io_file(mrb, self);

  off = ftell(f);
  if (off < 0) {
    io_raise_errno(mrb);
  }

  // FIXME: double check range!
  return mrb_fixnum_value(off);
}


#if USE_POSIX
static mrb_value
mIO_fileno(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  mrb_int fd;
  f = mrb_io_file(mrb, self);

  fd = fileno(f);
  if (fd == -1) {
    io_raise_errno(mrb);
  }

  return mrb_fixnum_value(fd);
}

static mrb_value
mIO_fsync(mrb_state *mrb, mrb_value self)
{
  FILE *f;
  mrb_int fd;
  f = mrb_io_file(mrb, self);

  fd = fileno(f);
  if (fd == -1) {
    io_raise_errno(mrb);
  }
  
  if (fsync(fd) < 0) {
    io_raise_errno(mrb);
  }

  return mrb_nil_value();
}
#endif /* USE_POSIX */


static void
mrb_init_io(mrb_state *mrb)
{
  struct RClass *io;

  io = mrb_define_class(mrb, "IO", mrb->object_class);
  MRB_SET_INSTANCE_TT(io, MRB_TT_DATA);

  mrb_include_module(mrb, io, mrb_module_get(mrb, "Enumerable")); /* 15.2.20.3 */

  mrb_define_method(mrb, io, "initialize", mIO_initialize, MRB_ARGS_ANY());    /* 15.2.20.5.21 (x)*/
  mrb_define_method(mrb, io, "close", mIO_close, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "closed?", mIO_closed_p, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "each", mIO_each, MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_method(mrb, io, "each_byte", mIO_each_byte, MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_method(mrb, io, "each_line", mIO_each, MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_method(mrb, io, "eof?", mIO_eof_p, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "flush", mIO_flush, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "getc", mIO_getc, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "gets", mIO_gets, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "read", mIO_read, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, io, "readchar", mIO_readchar, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "readline", mIO_readline, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "readlines", mIO_readlines, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "sync", mIO_get_sync, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "sync=", mIO_set_sync, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, io, "write", mIO_write, MRB_ARGS_REQ(1));

  mrb_define_const(mrb, io, "SEEK_SET", mrb_fixnum_value(SEEK_SET));
  mrb_define_const(mrb, io, "SEEK_CUR", mrb_fixnum_value(SEEK_CUR));
  mrb_define_const(mrb, io, "SEEK_END", mrb_fixnum_value(SEEK_END));

  mrb_define_method(mrb, io, "tell", mIO_tell, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "seek", mIO_seek, MRB_ARGS_REQ(1) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, io, "rewind", mIO_rewind, MRB_ARGS_NONE());

  mrb_define_alias(mrb, io, "pos",  "tell");
  mrb_define_alias(mrb, io, "pos=", "seek");

  mrb_define_method(mrb, io, "ungetc", mIO_ungetc, MRB_ARGS_REQ(1));

  {
    mrb_value io_stdin,io_stdout,io_stderr;

    io_stdin  = mrb_io_open(mrb, stdin);
    io_stdout = mrb_io_open(mrb, stdout);
    io_stderr = mrb_io_open(mrb, stderr);

    mrb_define_global_const(mrb, "STDIN" , io_stdin);
    mrb_define_global_const(mrb, "STDOUT", io_stdout);
    mrb_define_global_const(mrb, "STDERR", io_stderr);

    mrb_gv_set(mrb, mrb_intern_lit(mrb, "$stdin" ), io_stdin );
    mrb_gv_set(mrb, mrb_intern_lit(mrb, "$stdout"), io_stdout);
    mrb_gv_set(mrb, mrb_intern_lit(mrb, "$stderr"), io_stderr);
  }

#if 0
  mrb_define_method(mrb, io, "reopen", mIO_reopen, MRB_ARGS_REQ(1) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, io, "sync", mIO_get_sync, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "sync=", mIO_set_sync, MRB_ARGS_NONE());
#endif /* 0 */

#if USE_POSIX
  mrb_define_method(mrb, io, "fileno", mIO_fileno, MRB_ARGS_NONE());
  mrb_define_method(mrb, io, "fsync", mIO_fsync, MRB_ARGS_NONE());
#  if 0
  mrb_define_method(mrb, io, "stat", mIO_stat, MRB_ARGS_NONE());
#  endif /* 0 */
#endif /* USE_POSIX */
}

/** File methods **/

/* File.exist?(path) */

static mrb_sym
file_iv_path(mrb_state *mrb)
{
  return mrb_intern_lit(mrb, "_path");
}

static void
file_set_path(mrb_state *mrb, mrb_value fobj, const char *path)
{
  mrb_value mpath;

  if (path != NULL) {
    mpath = mrb_str_new_cstr(mrb, path);
  } else {
    mpath = mrb_nil_value();
  }

  mrb_iv_set(mrb, fobj, file_iv_path(mrb), mpath);
}

/* File.initialize(path,mode="r")
 *
 * a) If path is not an instance of the class String, the behavior is
 *    unspecified.
 * b) If mode is not an instance of the class String whose content is a
 *    single character “r” or “w”, the behavior is unspecified.
 * c) Open the file specified by path in an implementation-defined way,
 *    and associate it with the receiver.
 * d) Set the path of the receiver to the content of path.
 * e) Set the openness flag and the buffering flag of the receiver to
 *    true.
 * f) Set the readability flag and the writability flag of the receiver
 *    as follows:
 *    1) If mode is an instance of the class String whose content is a
 *       single character “r”, set the readability flag of the receiver
 *       to true and set the writability flag of the receiver to false.
 *    2) If mode is an instance of the class String whose content is a
 *       single character “w”, set the readability flag of the receiver
 *       to false and set the writability flag of the receiver to true.
 * g) Return an implementation-defined value.
 * 
 */
static mrb_value
cFile_initialize(mrb_state *mrb, mrb_value self)
{
  char *path, *mode;
  FILE *f;

  path = NULL;
  mode = "r";
  mrb_get_args(mrb, "z|z", &path, &mode);

  if (errno = 0, f = fopen(path,mode), f == NULL) {
    io_raise_errno(mrb);
  }

  io_initialize(mrb, self, f);

  file_set_path(mrb, self, path);

  return self;
}

/* path */
static mrb_value
cFile_path(mrb_state *mrb, mrb_value self)
{
  return mrb_iv_get(mrb, self, file_iv_path(mrb));
}

static mrb_value
cFile_exists_p(mrb_state *mrb, mrb_value mod)
{
  const char* path;
  path = NULL;
  mrb_get_args(mrb, "z", &path);

#if USE_POSIX
  {
    struct stat buf;
    return mrb_bool_value(stat(path,&buf) == 0);
  }
#else
  /* in standard C, all we can do is try to open the file */
  {
    FILE *f;
    f = fopen(path, "r");
    if (f) { fclose(f); }
    return mrb_bool_value(f != NULL);
  }
#endif
}

static mrb_value
cFile_delete(mrb_state *mrb, mrb_value mod)
{
  mrb_value *argv = NULL;
  mrb_int i, argc = 0;

  mrb_get_args(mrb, "*", &argv, &argc);

  for(i=0; i < argc; i++) {
    if (mrb_type(argv[i]) != MRB_TT_STRING) {
      mrb_raisef(mrb, E_TYPE_ERROR, "argument %S is not a string", mrb_fixnum_value(argc+1));
    }
  }

  for(i=0; i < argc; i++) {
    const char *path;
    mrb_value str = argv[i];
    path = mrb_string_value_cstr(mrb, &str);
    if (remove(path) < 0) {
      io_raise_errno(mrb);
    }
  }

  return mrb_fixnum_value(argc);
}

static mrb_value
cFile_rename(mrb_state *mrb, mrb_value mod)
{
  const char *oldpath, *newpath;
  mrb_get_args(mrb, "zz", &oldpath, &newpath);

  if (rename(oldpath, newpath) < 0) {
    io_raise_errno(mrb);
  }

  return mrb_fixnum_value(0);
}

#if USE_POSIX
static mrb_value
cFile_absolute_path(mrb_state *mrb, mrb_value mod)
{
  /* okay, POSIX screwed up with realpath().  PATH_MAX may not be a
   * defined constant, so we can't allocate a sensible buffer based on
   * it.
   *
   * POSIX.1-2008 fixes this by letting realpath() malloc() the string
   * if the destination path is NULL.  So if we're using a system that
   * supports POSX.1-2008, we use realpath.
   *
   * On Mac OS X and iOS, PATH_MAX is a defined constant, so we
   * use the PATH_MAX buffer approach
   *
   * Otherwise we throw an exception.
   */
#if _POSIX_VERSION >= 200809L
  {
    const char *s;
    char *rp;
    mrb_value ret;

    s = NULL;
    mrb_get_args(mrb, "z", &s);

    if (rp = realpath(s,NULL), rp == NULL) {
      io_raise_errno(mrb);
    }

    ret = mrb_str_new_cstr(mrb,rp);
    free(rp);

    return ret;
  }
#elif (defined(__APPLE__) && defined(__MACH__))
  {
    const char *s;
    char rp[PATH_MAX+1];

    s = NULL;
    mrb_get_args(mrb, "z", &s);
    if (realpath(s,rp) == NULL) {
      io_raise_errno(mrb);
    }

    return mrb_str_new_cstr(mrb,rp);
  }
#else
  {
    char mesg[256];
    snprintf(mesg,sizeof(mesg)-1,
        "absolute_path only implemented on POSIX.1-2008 systems (POSIX version %ld)",
        _POSIX_VERSION);
    mrb_raise(mrb, E_NOTIMP_ERROR, mesg);

    return mrb_nil_value();
  }
#endif
}

static mrb_value
cFile_basename(mrb_state *mrb, mrb_value mod)
{
  const char *s;
  mrb_int slen;
  char path[MAXPATHLEN];

  s = NULL;
  slen = 0;
  mrb_get_args(mrb, "s", &s, &slen);

  if (slen >= sizeof(path)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "path exceeds maximum path size");
  }

  if (s) {
    memcpy(path,s,slen);
  } else {
    slen = 0;
  }

  path[slen] = '\0';
  return mrb_str_new_cstr(mrb, basename(path));
}

static mrb_value
cFile_dirname(mrb_state *mrb, mrb_value mod)
{
  const char *s;
  mrb_int slen;
  char path[MAXPATHLEN];

  s = NULL;
  slen = 0;
  mrb_get_args(mrb, "s", &s, &slen);

  if (slen >= sizeof(path)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "path exceeds maximum path size");
  }

  if (s) {
    memcpy(path,s,slen);
  } else {
    slen = 0;
  }

  path[slen] = '\0';
  return mrb_str_new_cstr(mrb, dirname(path));
}
#endif /* USE_POSIX */

static void
mrb_init_file(mrb_state *mrb)
{
  struct RClass *file;
  file = mrb_define_class(mrb, "File", io_class(mrb));

  mrb_define_method(mrb, file, "initialize", cFile_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, file, "path", cFile_path, MRB_ARGS_NONE());

  mrb_define_singleton_method(mrb, (struct RObject *)file, "exists?", cFile_exists_p, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject *)file, "delete", cFile_delete, MRB_ARGS_REST());
  mrb_define_singleton_method(mrb, (struct RObject *)file, "unlink", cFile_delete, MRB_ARGS_REST());
  mrb_define_singleton_method(mrb, (struct RObject *)file, "rename", cFile_rename, MRB_ARGS_REQ(2));

#if USE_POSIX
  mrb_define_singleton_method(mrb, (struct RObject *)file, "absolute_path", cFile_absolute_path, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject *)file, "basename", cFile_basename, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject *)file, "dirname", cFile_dirname, MRB_ARGS_REQ(1));
#endif /* USE_POSIX */

#if 0 && defined(POSIX)
  mrb_define_singleton_method(mrb, io, "umask", cFile_umask, MRB_ARGS_OPT(1));
  mrb_define_singleton_method(mrb, file, "atime", cFile_atime, MRB_ARGS_NONE());
  mrb_define_singleton_method(mrb, file, "ctime", cFile_ctime, MRB_ARGS_NONE());
  mrb_define_singleton_method(mrb, file, "mtime", cFile_mtime, MRB_ARGS_NONE());

  mrb_define_method(mrb, file, "atime", cFile_atime, MRB_ARGS_NONE());
  mrb_define_method(mrb, file, "ctime", cFile_ctime, MRB_ARGS_NONE());
  mrb_define_method(mrb, file, "mtime", cFile_mtime, MRB_ARGS_NONE());
  mrb_define_method(mrb, file, "lstat", cFile_lstat, MRB_ARGS_NONE());
  mrb_define_method(mrb, file, "flock", cFile_flock, MRB_ARGS_REQ(1));
#endif /* 0 && defined(POSIX) */
}

static void tempfile_free(mrb_state* mrb, void* p);

static const mrb_data_type tempfile_type = {
  .struct_name = "Tempfile",
  .dfree = tempfile_free,
};


static mrb_value
tempfile_open(mrb_state *mrb, FILE *f, const char *path)
{
  mrb_value tmpfile;

  if (f == NULL) {
    io_raise_errno(mrb);
  }

  tmpfile = io_open_with_class(mrb, mrb_class_get(mrb, "Tempfile"), &tempfile_type, f);
  if (path) {
    struct mruby_io *ios;
    char *pathcopy;
    size_t len;

    ios = io_get_ios(mrb,tmpfile);
    len = strlen(path);
    pathcopy = mrb_malloc(mrb, len+1);
    memcpy(pathcopy, path, len+1);

    ios->user = pathcopy;

    file_set_path(mrb, tmpfile, path);
  }

  return tmpfile;
}

static mrb_value
create_temp(mrb_state *mrb, mrb_value path, mrb_value dir, mrb_int argc, mrb_value *argv)
{
#if USE_POSIX
  char *buf;
  FILE *f;
  int fd, ai;
  mrb_value s, tmpfile;

  ai = mrb_gc_arena_save(mrb);

  s = path;
  if (!mrb_nil_p(dir)) {
    s = mrb_format(mrb, "%S/%SXXXXXX", dir, path);
  } else {
    s = mrb_str_cat_cstr(mrb, path, "XXXXXX");
  }

  buf = mrb_str_to_cstr(mrb, s);
  fd = mkstemp(buf);
  if (fd < 0) {
    io_raise_errno(mrb);
  }

  f = fdopen(fd, "r+");

  if (f == NULL) {
    int prev_err = errno;
    close(fd);
    errno = prev_err;
    io_raise_errno(mrb);
  }

  tmpfile = tempfile_open(mrb, f, buf);
  mrb_gc_arena_restore(mrb, ai);
  return tmpfile;
#else
  mrb_raise(mrb, E_NOTIMP_ERROR,
      "temporary filename prefixes are not supported on this architecture");

  return mrb_nil_value();
#endif
}

static void
tempfile_free(mrb_state* mrb, void* p)
{
  struct mruby_io* ios = p;
  if (ios == NULL) { return; }

  if (ios->user != NULL) {
    char *path = ios->user;
    remove(path);
    mrb_free(mrb, path);
    ios->user = NULL;
  }

  io_free(mrb, p);
}

static mrb_value
cTempfile_new(mrb_state *mrb, mrb_value c)
{
  mrb_value path, dir;
  mrb_value *argv;
  mrb_int argc;
  mrb_int nargs;

  path = dir = mrb_nil_value();
  argv = NULL;
  argc = 0;
  nargs = mrb_get_args(mrb, "|SS*", &path, &dir, &argv, &argc);

  if (nargs == 0) {
    return tempfile_open(mrb, tmpfile(), NULL);
  } else {
    return create_temp(mrb, path, dir, argc, argv);
  }
}

static mrb_value
cTempfile_unlink(mrb_state *mrb, mrb_value self)
{
  mrb_value path;
  char *cpath;
  int ai;

  path = cFile_path(mrb, self);
  if (mrb_nil_p(path)) { return mrb_nil_value(); }

  ai = mrb_gc_arena_save(mrb);
  cpath = mrb_str_to_cstr(mrb, path);

  if (remove(cpath) != 0) {
    io_raise_errno(mrb);
  }

  file_set_path(mrb, self, NULL);

  mrb_gc_arena_restore(mrb, ai);
  return mrb_bool_value(1);
}

static void
mrb_init_tempfile(mrb_state *mrb)
{
  struct RClass *tempfile;
  tempfile = mrb_define_class(mrb, "Tempfile", mrb_class_get(mrb, "File"));

  mrb_define_singleton_method(mrb, (struct RObject*)tempfile, "new", cTempfile_new, MRB_ARGS_REST());

  mrb_define_method(mrb, tempfile, "unlink", cTempfile_unlink, MRB_ARGS_NONE());
}

void
mrb_mruby_stdio_gem_init(mrb_state *mrb)
{
  int ai;

  ai = mrb_gc_arena_save(mrb);
  mrb_init_io(mrb);
  mrb_gc_arena_restore(mrb,ai);

  ai = mrb_gc_arena_save(mrb);
  mrb_init_file(mrb);
  mrb_gc_arena_restore(mrb,ai);

  ai = mrb_gc_arena_save(mrb);
  mrb_init_tempfile(mrb);
  mrb_gc_arena_restore(mrb,ai);
}

void
mrb_mruby_stdio_gem_final(mrb_state *mrb)
{
}
