#include "tsmrb_callbacks.h"

#include "tsmrb_sim.h"
#include "tsmrb_classes.h"
#include "tsmrb_helpers.h"

#include "mruby/data.h"
#include "mruby/numeric.h"
#include "mruby/hash.h"
#include "mruby/value.h"
#include "mruby/variable.h"
#include "mruby/proc.h"

#include "ts_integrators.h"

typedef struct {
  ts_integrator_callback parent;
  mrb_state *mrb;
  mrb_value rb_callback;
  int num_calls;
} cCallback_data;

static const mrb_data_type callback_type = {
  .struct_name = "Callback",
  .dfree = NULL,
};

static TSResult cCallback_callback(ts_sim* sim, cCallback_data* closure);

static void cCallback_init_data(cCallback_data* data, mrb_state *mrb, int freq, mrb_value proc)
{
  memset(data, 0, sizeof(*data));
  data->parent.callback  = (TSIntegratorCallbackFunc)cCallback_callback;
  data->parent.freq      = freq;
  data->parent.countdown = 0;

  data->mrb = mrb;
  data->rb_callback = proc;
}

static mrb_value cCallback_allocate(mrb_state *mrb, int freq, mrb_value rproc)
{
  cCallback_data* cbdata;
  struct RData* data;

  cbdata = mrb_malloc(mrb, sizeof(*data));
  cCallback_init_data(cbdata, mrb, freq, rproc);

  data = mrb_data_object_alloc(mrb, 
      tsmrb_callback_class(mrb), cbdata, &callback_type);
  return mrb_obj_value(data);
}

static TSResult cCallback_callback(ts_sim* sim, cCallback_data* closure)
{
  closure->num_calls++;

  if (!mrb_nil_p(closure->rb_callback))
  {
    mrb_state *mrb;
    mrb_value ret;
    struct RProc *cb;
    int state;

    mrb = closure->mrb;
    cb = mrb_proc_ptr(closure->rb_callback);
    state = 0;
    ret = mrb_call_protected_proc(mrb, cb, 0, NULL, &state);

    if (state)
    {
      RETURN_WITH_RESULT( tsmrb_handle_mruby_error(mrb) );
    }

    if (mrb_symbol_p(ret) && (mrb_symbol(ret) == TSMRB_SYM(STOP_SIMULATION)))
    {
      return TS_NEW_RESULT(TS_CANCELLED);
    }
  }

  return TS_OK;
}


static mrb_value cCallback_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_value kwArgs, rfreq, rproc;
  mrb_int freq;
  mrb_get_args(mrb, "H", &kwArgs);

  rfreq = mrb_hash_get(mrb, kwArgs, mrb_symbol_value(TSMRB_SYM(freq)));
  rproc = mrb_hash_get(mrb, kwArgs, mrb_symbol_value(TSMRB_SYM(proc)));

  if (mrb_nil_p(rfreq))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "missing :freq keyword");
  }
  
  if (!mrb_fixnum_p(rfreq))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, ":freq keyword must be an integer");
  }

  freq = mrb_fixnum(rfreq);

  if (mrb_type(rproc) == MRB_TT_PROC)
  {
    mrb_iv_set(mrb, self, mrb_intern_lit(mrb, "@proc"), rproc);
  }
  else if (!mrb_nil_p(rproc))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, ":proc keyword must be a proc");
  }

  cCallback_data* data = mrb_malloc(mrb, sizeof(*data));

  DATA_TYPE(self) = &callback_type;
  DATA_PTR(self) = data;

  cCallback_init_data(data, mrb, freq, rproc);

  return self;
}

static mrb_value cCallback_with_freq(mrb_state *mrb, mrb_value klass)
{
  mrb_int freq;
  mrb_value rproc;

  rproc = mrb_nil_value();
  mrb_get_args(mrb, "i&", &freq, &rproc);

  return cCallback_allocate(mrb, freq, rproc);
}

static mrb_value cCallback_freq(mrb_state *mrb, mrb_value self)
{
  cCallback_data *cbdata = DATA_PTR(self);
  return mrb_fixnum_value(cbdata->parent.freq);
}

static mrb_value cCallback_num_calls(mrb_state *mrb, mrb_value self)
{
  cCallback_data *cbdata = DATA_PTR(self);

  return mrb_fixnum_value(cbdata->num_calls);
}

static mrb_value cCallback_call(mrb_state *mrb, mrb_value self)
{
  cCallback_data *cbdata;
  mrb_value simobj;
  ts_sim *sim;

  cbdata = DATA_PTR(self);
  mrb_get_args(mrb, "o", &simobj);
  sim = tsmrb_to_sim(mrb, simobj);

  tsmrb_handle_error(mrb, 
      cbdata->parent.callback(sim, &(cbdata->parent)));

  return self;
}

static mrb_value cCallback_proc(mrb_state *mrb, mrb_value self)
{
  cCallback_data *cbdata = DATA_PTR(self);
  return cbdata->rb_callback;
}

static mrb_value cSim_add_callback(mrb_state *mrb, mrb_value self)
{
  mrb_value cbobj;
  cCallback_data *cbdata;

  ts_sim* sim = tsmrb_to_sim(mrb, self);

  mrb_get_args(mrb, "o", &cbobj);
  mrb_data_check_type(mrb, cbobj, &callback_type);
  cbdata = mrb_data_get_ptr(mrb, cbobj, &callback_type);

  mrb_sym at_callbacks = mrb_intern_lit(mrb, "@callbacks");
  mrb_value cblist = mrb_iv_get(mrb, self, at_callbacks);
  if (mrb_nil_p(cblist))
  {
    cblist = mrb_ary_new(mrb);
    mrb_iv_set(mrb, self, at_callbacks, cblist);
  }

  mrb_ary_push(mrb, cblist, cbobj);

  tsmrb_handle_error(mrb,
      ts_add_callback(sim, (ts_integrator_callback*)cbdata));

  return self;
}

static mrb_value cSim_callback_p(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim;
  void *data;
  cCallback_data *cbdata;
  size_t i;

  mrb_get_args(mrb, "d", &data, &callback_type);
  sim = tsmrb_to_sim(mrb, self);
  cbdata = data;

  for (i=0; i < sim->callbacks.size; i++)
  {
    ts_integrator_callback** cb_at_i = DATAVECTOR_AT(ts_integrator_callback*, &(sim->callbacks), i);
    if (&(cbdata->parent) == *cb_at_i)
    {
      return mrb_bool_value(1);
    }
  }

  return mrb_bool_value(0);
}

struct RClass* tsmrb_callback_class(mrb_state* mrb)
{
  return tsmrb_classes(mrb)->cCallback;
}

void tsmrb_define_callbacks(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_callback_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Callback class has already been initialized!");

  struct RClass* cCallback = mrb_define_class_under(mrb, mod,
      "Callback", mrb->object_class);
  tsmrb_classes(mrb)->cCallback = cCallback;
  MRB_SET_INSTANCE_TT(cCallback, MRB_TT_DATA);

  mrb_define_method(mrb, cCallback, "initialize", cCallback_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cCallback, "freq", cCallback_freq, MRB_ARGS_NONE());
  mrb_define_method(mrb, cCallback, "proc", cCallback_proc, MRB_ARGS_NONE());
  mrb_define_method(mrb, cCallback, "num_calls", cCallback_num_calls, MRB_ARGS_NONE());

  mrb_define_method(mrb, cCallback, "call", cCallback_call, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cCallback, "with_freq", cCallback_with_freq,
      MRB_ARGS_REQ(1) | MRB_ARGS_BLOCK());

  mrb_define_const(mrb, cCallback, "STOP_SIMULATION", 
      mrb_symbol_value(TSMRB_SYM(STOP_SIMULATION)));

  struct RClass *cSim = tsmrb_sim_class(mrb);
  mrb_define_method(mrb, cSim, "add_callback"    , cSim_add_callback    ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "callback?"       , cSim_callback_p      ,  MRB_ARGS_REQ(1));
}
