#include "tsmrb_chem_driver.h"
#include "tsmrb_lattice.h"

#include <string.h>
#include <errno.h>

#include "tsch_config.h"
#include "tsch_config_yaml.h"
#include "tsch_config_parse.h"

#include "mruby/data.h"
#include "mruby/string.h"
#include "mruby/variable.h"
#include "mruby/error.h"
#include "mruby/numeric.h"

#include "tsmrb_helpers.h"
#include "tsmrb_lattice.h"
#include "tsmrb_io.h"

static void chem_driver_free(mrb_state *mrb, void* p)
{
  ts_chem_driver_release(p);
}

static const mrb_data_type chem_driver_type = {
  .struct_name = "ChemDriver",
  .dfree = chem_driver_free,
};

ts_chem_driver* tsmrb_to_chem_driver(mrb_state *mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_chem_driver_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a ChemDriver object");

  return DATA_PTR(obj);
}

static inline enum TS_CHEMDRIVER_SITE_TYPE sym_to_site_type(mrb_state *mrb, mrb_sym sym)
{
  if (sym == TSMRB_SYM(BLANK  )) { return TS_CHEMDRIVER_SITE_BLANK;    }
  if (sym == TSMRB_SYM(MONOMER)) { return TS_CHEMDRIVER_SITE_MONOMER;  }
  if (sym == TSMRB_SYM(LONGIFC)) { return TS_CHEMDRIVER_SITE_LONG_IFC; }
  if (sym == TSMRB_SYM(LATIFC )) { return TS_CHEMDRIVER_SITE_LAT_IFC;  }

  tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
      "symbol :%s is not a valid site type", mrb_sym2name(mrb,sym));
  return TS_CHEMDRIVER_SITE_BLANK;
}

#define RAISE_BAD_STATE(sitedesc) do {          \
  if (argnum > 0) {                             \
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,         \
        "invalid state in arg %d (%s site cannot have state %d)", \
        argnum, SYMSTR(sitedesc), state);       \
  } else {                                      \
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,         \
        "invalid state in (%s site cannot have state %d)", \
        SYMSTR(sitedesc), state);               \
  }                                             \
} while (0)

#define DECLARE_BAD_STATE(sitedesc, st) \
  do { if (st == state) {       \
    RAISE_BAD_STATE(sitedesc);  \
    return INVALID_SPIN;        \
  } } while (0)

#define DECLARE_GOOD_STATE(sitedesc, st) \
  do { if (st == state) { return sp; } } while(0)

static inline spin_t check_site_and_return(mrb_state *mrb,
    enum TS_CHEMDRIVER_SITE_TYPE stype, int argnum, int state)
{
  spin_t sp = spin_from_int(state);

  if (ts_lattice_is_pseudo_spin(NULL, sp)) { return sp; }

  switch (stype)
  {
    case TS_CHEMDRIVER_SITE_ANY:
      break;

    case TS_CHEMDRIVER_SITE_BLANK:
      DECLARE_GOOD_STATE(EMPTY, TS_CHEMDRIVER_STATE_EMPTY);
      RAISE_BAD_STATE(EMPTY);
      break;

    case TS_CHEMDRIVER_SITE_MONOMER:
      DECLARE_BAD_STATE(MONOMER, TS_CHEMDRIVER_STATE_TUBELIKE );
      DECLARE_BAD_STATE(MONOMER, TS_CHEMDRIVER_STATE_SHEETLIKE);
      DECLARE_BAD_STATE(MONOMER, TS_CHEMDRIVER_STATE_BENT_IFC );
      break;

    case TS_CHEMDRIVER_SITE_LONG_IFC:
      DECLARE_BAD_STATE(LONGIFC, TS_CHEMDRIVER_STATE_TUBELIKE   );
      DECLARE_BAD_STATE(LONGIFC, TS_CHEMDRIVER_STATE_SHEETLIKE  );
      DECLARE_BAD_STATE(MONOMER, TS_CHEMDRIVER_STATE_BENT_LEFT  );
      DECLARE_BAD_STATE(MONOMER, TS_CHEMDRIVER_STATE_BENT_RIGHT );
      break;

    case TS_CHEMDRIVER_SITE_LAT_IFC:
      DECLARE_GOOD_STATE(LATIFC, TS_CHEMDRIVER_STATE_EMPTY);
      DECLARE_GOOD_STATE(LATIFC, TS_CHEMDRIVER_STATE_TUBELIKE);
      DECLARE_GOOD_STATE(LATIFC, TS_CHEMDRIVER_STATE_SHEETLIKE);
      RAISE_BAD_STATE(LATIFC);
      break;

    default:
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
          "invalid site type %d", (int)stype);
  }

  return sp;
}
#undef RAISE_BAD_STATE
#undef DECLARE_BAD_STATE
#undef DECLARE_GOOD_STATE

static spin_t spin_value_from_char(mrb_state *mrb, enum TS_CHEMDRIVER_SITE_TYPE stype, int argnum, char c)
{
  if (c == '\0')
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid string length");
  else if (c == '_')
    return check_site_and_return(mrb, stype, argnum, TS_CHEMDRIVER_STATE_EMPTY   );
  else
  {
    int sp = strchr(ts_chem_driver_site_codes, c) - ts_chem_driver_site_codes;
    if ((sp < 0) || (sp >= TS_CHEMDRIVER_STATE_MAX))
      mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid string type");

    return check_site_and_return(mrb, stype, argnum, (enum TS_CHEMDRIVER_SITE_STATE)sp );
  }

  return INVALID_SPIN;
}

static inline spin_t spin_value_from_string(mrb_state *mrb, enum TS_CHEMDRIVER_SITE_TYPE stype, int argnum, const char* s)
{
  if (s==NULL) { mrb_raise(mrb, E_ARGUMENT_ERROR, "NULL string is invalid"); }

  return spin_value_from_char(mrb, stype,argnum,s[0]);
}

static void spin_triplet_from_string(mrb_state *mrb, int argnum, const char s[3], spin_t triplet[3])
{
  triplet[0] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_MONOMER, argnum, s[0]);
  triplet[1] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_ANY    , argnum, s[1]);
  triplet[2] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_MONOMER, argnum, s[2]);
}

static void spin_bending_triplet_from_string(mrb_state *mrb, int argnum, const char s[3], spin_t triplet[3])
{
  triplet[0] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_ANY    , argnum, s[0]);
  triplet[1] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_MONOMER, argnum, s[1]);
  triplet[2] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_ANY    , argnum, s[2]);
}

static void spin_coupling_term_from_string(mrb_state *mrb, int argnum, const char s[4], int term[4])
{
  term[0] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_LAT_IFC , argnum, s[0]).val;
  term[1] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_LAT_IFC , argnum, s[1]).val;
  term[2] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_LONG_IFC, argnum, s[2]).val;
  term[3] = spin_value_from_char(mrb, TS_CHEMDRIVER_SITE_LONG_IFC, argnum, s[3]).val;
}

static spin_t spin_value(mrb_state *mrb, enum TS_CHEMDRIVER_SITE_TYPE stype, int argnum, mrb_value r_val)
{
  switch (mrb_type(r_val))
  {
    case MRB_TT_SYMBOL:
      return spin_value_from_string(mrb, stype,
          argnum, mrb_sym2name(mrb,mrb_symbol(r_val)));

    case MRB_TT_STRING:
      return spin_value_from_string(mrb, stype,
          argnum, mrb_string_value_ptr(mrb,r_val));

    case MRB_TT_FIXNUM:
      return check_site_and_return(mrb, stype,
          argnum, mrb_fixnum(r_val));

    default:
      mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown value type");
  }

  return spin_from_int(0);
}

static mrb_value chem_driver_char_value(mrb_state *mrb, mrb_value mod)
{
  mrb_int state = 0;
  mrb_get_args(mrb, "i", &state);

  if ((state < 0) || (state >= TS_CHEMDRIVER_STATE_MAX))
    tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid state %d", state);

  char c = ts_chem_driver_site_codes[state];
  return mrb_str_new(mrb, &c, 1);
}

static mrb_value chem_driver_spin_value(mrb_state *mrb, mrb_value mod)
{
  mrb_sym stype = 0;
  mrb_value rval = mrb_nil_value();
  mrb_get_args(mrb, "no", &stype, &rval);

  spin_t s = spin_value(mrb, sym_to_site_type(mrb,stype), 2, rval);
  return mrb_fixnum_value(s.val);
}

static mrb_value chem_driver_get_debug_updates(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(tsch_debug_updates);
}

static mrb_value chem_driver_set_debug_updates(mrb_state *mrb, mrb_value self)
{
  mrb_int tf = 0;
  mrb_get_args(mrb, "i", &tf);
  tsch_debug_updates = tf;

  return mrb_nil_value();
}

static mrb_value cChemDriver_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_int nrow = 0;
  mrb_int ncol = 0;
  mrb_get_args(mrb, "ii", &nrow, &ncol);

  TSResult err = TS_OK;
  ts_chem_driver* driver = ts_chem_driver_new( nrow, ncol, &err );

  DATA_PTR(self) = driver;
  DATA_TYPE(self) = &chem_driver_type;

  tsmrb_lattice_setup_callback_refs(mrb, self);

  tsmrb_handle_error(mrb, err);

  return self;
}

static mrb_value cChemDriver_num_dimers(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_fixnum_value( ts_chem_driver_num_dimers(drv) );
}

static mrb_value cChemDriver_alpha_row(mrb_state *mrb, mrb_value self)
{
  mrb_int dimer;
  mrb_get_args(mrb, "i", &dimer);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_fixnum_value(ts_chem_driver_alpha_row(drv,dimer));
}

static mrb_value cChemDriver_beta_row(mrb_state *mrb, mrb_value self)
{
  mrb_int dimer;
  mrb_get_args(mrb, "i", &dimer);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_fixnum_value(ts_chem_driver_beta_row(drv,dimer));
}

static mrb_value cChemDriver_get_pitch_monomers(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_fixnum_value(drv->pitch_monomers);
}

static mrb_value cChemDriver_get_seam_allowed_thresh(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_fixnum_value(drv->seam_allowed_thresh);
}

static mrb_value cChemDriver_set_seam_allowed_thresh(mrb_state *mrb, mrb_value self)
{
  mrb_int thresh;
  mrb_get_args(mrb, "i", &thresh);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  drv->seam_allowed_thresh = thresh;
  return self;
}

static mrb_value cChemDriver_get_dt(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_float_value(mrb,drv->dt);
}

static mrb_value cChemDriver_set_dt(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_float val;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "f", &val);
  drv->dt = val;

  return mrb_nil_value();
}

static mrb_value cChemDriver_get_t_max(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_float_value(mrb,drv->t_max);
}

static mrb_value cChemDriver_set_t_max(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_float val;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "f", &val);

  drv->t_max = val;
  return mrb_nil_value();
}

static mrb_value cChemDriver_get_forbidden_energy(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_float_value(mrb,drv->forbidden_energy);
}

static mrb_value cChemDriver_set_forbidden_energy(mrb_state *mrb, mrb_value self)
{
  mrb_float val;
  ts_chem_driver* drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "f", &val);
  drv->forbidden_energy = val;
  return mrb_nil_value();
}

static mrb_value set_driver_flags(mrb_state *mrb, mrb_value self, uint32_t flags)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  mrb_bool tf;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "b", &tf);

  if (tf)
  {
    tsch_set_flags(drv, flags);
  }
  else
  {
    tsch_unset_flags(drv, flags);
  }

  return mrb_nil_value();
}

static mrb_value cChemDriver_get_reversible_lateral_bond_breaking(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_REVERSIBLE_LATERAL_BOND_BREAKING));
}

static mrb_value cChemDriver_set_reversible_lateral_bond_breaking(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_REVERSIBLE_LATERAL_BOND_BREAKING);
}

static mrb_value
cChemDriver_get_restrict_curved_lateral_bonding(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_RESTRICT_CURVED_LATERAL_BONDING));
}

static mrb_value
cChemDriver_set_restrict_curved_lateral_bonding(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_RESTRICT_CURVED_LATERAL_BONDING);
}

static mrb_value
cChemDriver_get_disable_ccs_depolymerization(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_DISABLE_CCS_DEPOLYMERIZATION));
}

static mrb_value
cChemDriver_set_disable_ccs_depolymerization(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_DISABLE_CCS_DEPOLYMERIZATION);
}

static mrb_value cChemDriver_get_minus_end_polymerization(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_MINUS_END_POLYMERIZATION));
}

static mrb_value cChemDriver_set_minus_end_polymerization(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_MINUS_END_POLYMERIZATION);
}

static mrb_value cChemDriver_get_enable_lat_long_bending(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_ENABLE_LAT_LONG_BENDING));
}

static mrb_value cChemDriver_set_enable_lat_long_bending(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_ENABLE_LAT_LONG_BENDING);
}

static mrb_value cChemDriver_get_allow_depolymerization_at_straight_interfaces(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS));
}

static mrb_value cChemDriver_set_allow_depolymerization_at_straight_interfaces(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_ALLOW_DEPOLYM_AT_STRAIGHT_IFCS);
}

static mrb_value cChemDriver_get_disallow_polymerization_at_straight_interfaces(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_DISALLOW_POLYM_AT_STRAIGHT_IFCS));
}

static mrb_value cChemDriver_set_disallow_polymerization_at_straight_interfaces(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb,self, TSCH_DISALLOW_POLYM_AT_STRAIGHT_IFCS);
}

static mrb_value cChemDriver_get_polym_only_longitudinal(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_POLYM_ONLY_LONGITUDINAL));
}

static mrb_value cChemDriver_set_polym_only_longitudinal(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb, self, TSCH_POLYM_ONLY_LONGITUDINAL);
}

static mrb_value cChemDriver_get_seam_detailed_balance(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  return mrb_bool_value(!tsch_test_flags(drv, TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE));
}

static mrb_value cChemDriver_set_seam_detailed_balance(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_bool tf;

  drv = tsmrb_to_chem_driver(mrb, self);
  tf = 0;
  mrb_get_args(mrb, "b", &tf);

  if (tf)
  {
    tsch_unset_flags(drv, TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE);
  }
  else
  {
    tsch_set_flags(drv, TSCH_SEAM_BREAKING_IGNORES_DETAILED_BALANCE);
  }
  return mrb_nil_value();
}

static mrb_value cChemDriver_get_polym_metropolis_weighted(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb, self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_POLYM_METROPOLIS_WEIGHTED));
}

static mrb_value cChemDriver_set_polym_metropolis_weighted(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb, self, TSCH_POLYM_METROPOLIS_WEIGHTED);
}

static mrb_value
cChemDriver_get_restrict_sheet_fluctuations(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_RESTRICT_SHEET_FLUCTS));
}

static mrb_value
cChemDriver_set_restrict_sheet_fluctuations(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb, self, TSCH_RESTRICT_SHEET_FLUCTS);
}

static mrb_value
cChemDriver_get_hydrolysis_only_at_sss(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value(tsch_test_flags(drv, TSCH_HYDROLYSIS_ONLY_AT_SSS));
}

static mrb_value
cChemDriver_set_hydrolysis_only_at_sss(mrb_state *mrb, mrb_value self)
{
  return set_driver_flags(mrb, self, TSCH_HYDROLYSIS_ONLY_AT_SSS);
}

static mrb_bool ignore_missing_config = 0;

static mrb_value cChemDriver_get_ignore_missing_config(mrb_state *mrb, mrb_value mod)
{
  return mrb_bool_value(ignore_missing_config);
}

static mrb_value cChemDriver_set_ignore_missing_config(mrb_state *mrb, mrb_value mod)
{
  mrb_get_args(mrb, "b", &ignore_missing_config);
  return mrb_nil_value();
}

static mrb_value chem_driver_from_yaml(mrb_state *mrb, mrb_value mod)
{
  int ai = mrb_gc_arena_save(mrb);
  char* s = NULL;
  mrb_int slen = 0;
  mrb_get_args(mrb, "s", &s, &slen);
  mrb_value r_drv;

  TSResult res = TS_OK;
  ts_chem_driver* drv = ts_chem_driver_from_yaml(s,slen,&res,ignore_missing_config);
  tsmrb_handle_error(mrb,res);

  struct RData* obj = Data_Wrap_Struct(mrb,
      tsmrb_chem_driver_class(mrb), &chem_driver_type, drv);

  r_drv = mrb_obj_value(obj);
  tsmrb_lattice_setup_callback_refs(mrb, r_drv);

  mrb_gc_arena_restore(mrb,ai);
  return r_drv;
}

static mrb_value cChemDriver_set(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_value v = mrb_nil_value();
  mrb_get_args(mrb, "iio", &i, &j, &v);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb, (ts_lattice*)drv, i,j);

  spin_t spin = spin_value(mrb, ts_chem_driver_site_type(drv, i,j), 1, v);
  ts_lattice_ref((ts_lattice*)drv, i,j)[0] = spin;
  return self;
}

static mrb_value cChemDriver_change_spin(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_value v = mrb_nil_value();
  mrb_get_args(mrb, "iio", &i, &j, &v);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb, (ts_lattice*)drv, i,j);

  spin_t new_state =
    spin_value(mrb, ts_chem_driver_site_type(drv, i,j), 3, v);
  tsmrb_handle_error(mrb, 
      ts_lattice_change_spin((ts_lattice*)drv, i,j, new_state) );

  return self;
}

#define is_valid_prefactor(latt, prefactor_index) \
  (is_between(prefactor_index, 0, TS_CHEMDRIVER_EVT_MAX))

#define is_valid_monomer_energy(latt, ener_index) \
  (is_between(ener_index, 0, TS_CHEMDRIVER_STATE_MAX))

static mrb_value cChemDriver_site_type(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb,(ts_lattice*)drv, i,j);

  enum TS_CHEMDRIVER_SITE_TYPE type = ts_chem_driver_site_type(drv, i,j);

  switch (type) {
    case TS_CHEMDRIVER_SITE_BLANK:         return TSMRB_SYM_OBJ( BLANK );
    case TS_CHEMDRIVER_SITE_MONOMER:       return TSMRB_SYM_OBJ( MONOMER );
    case TS_CHEMDRIVER_SITE_LONG_IFC:      return TSMRB_SYM_OBJ( LONGIFC );
    case TS_CHEMDRIVER_SITE_LAT_IFC:       return TSMRB_SYM_OBJ( LATIFC );

    default:
      return mrb_nil_value();
  }
}

static mrb_value cChemDriver_is_alpha_lateral(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_int i = 0, j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb,&drv->latt, i,j);

  return mrb_bool_value( ts_chem_driver_is_alpha_lateral(drv, i,j) );
}

static mrb_value cChemDriver_is_alpha_monomer(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_int i = 0, j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb,&drv->latt, i,j);

  return mrb_bool_value(ts_chem_driver_is_alpha_monomer(drv, i,j));
}

#define RATE_PREFACTOR_COUNT  TS_CHEMDRIVER_EVT_MAX

static void rate_prefactor_table_initialize(mrb_state *mrb)
{
  int i=0;
  struct tsmrb_rate_prefactor* rtbl;

  rtbl = mrb_malloc(mrb,
      (TS_CHEMDRIVER_EVT_MAX+1)*sizeof(rtbl[0]));

  rtbl[i].sym = TSMRB_SYM(EVT_NONE);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_NONE;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_MONOMER);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_MONOMER;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_LONGITUDINAL);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_LONGITUDINAL;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_LATERAL);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_LATERAL;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_LATERAL_SEAM);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_LATERAL_SEAM;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_HYDROLYSIS);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_HYDROLYSIS;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_POLYMERIZE);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_POLYMERIZE;
  i++;

  rtbl[i].sym = TSMRB_SYM(EVT_K_ON);
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_K_ON;
  i++;

  rtbl[i].sym = 0;
  rtbl[i].evtype = TS_CHEMDRIVER_EVT_MAX;
  i++;

  if (i > TS_CHEMDRIVER_EVT_MAX)
  {
    fprintf(stderr, "\n--- WARNING!  STACK MAY BE HOSED! ---\n");
  }

  mrbtAssert(i <= TS_CHEMDRIVER_EVT_MAX+1);

  tsmrb_classes(mrb)->rate_prefactor_table = rtbl;
}

static inline enum TS_CHEMDRIVER_EVENT_TYPE rate_prefactor_index(mrb_state *mrb, mrb_sym sym)
{
  int i;
  struct tsmrb_rate_prefactor* rtbl = tsmrb_classes(mrb)->rate_prefactor_table;

  for (i=0; rtbl[i].evtype != TS_CHEMDRIVER_EVT_MAX; i++)
  {
    if (sym == rtbl[i].sym) { return rtbl[i].evtype; }
  }

  return TS_CHEMDRIVER_EVT_MAX;
}

static mrb_value cChemDriver_get_rate_prefactors(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  int i, argc = 0;
  mrb_value *argv = NULL;
  mrb_value ret;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "*", &argv, &argc);

  if (argc == 0)
  {
    struct tsmrb_rate_prefactor* rtbl = tsmrb_classes(mrb)->rate_prefactor_table;

    ret = mrb_hash_new_capa(mrb,RATE_PREFACTOR_COUNT);
    for (i=0; rtbl[i].evtype != TS_CHEMDRIVER_EVT_MAX; i++)
    {
      if (! is_valid_prefactor(drv,rtbl[i].evtype))
      {
        tsmrb_raisef(mrb, E_RUNTIME_ERROR, "unknown rate prefactor in table (entry %d)!",i+1);
      }

      mrb_hash_set(mrb, ret,
          mrb_symbol_value(rtbl[i].sym),
          mrb_float_value(mrb, drv->rate_prefactors[rtbl[i].evtype]));
    }
  }
  else
  {
    ret = mrb_ary_new_capa(mrb,argc);
    for (i=0; i < argc; i++) 
    {
      int ind;

      mrb_check_type(mrb, argv[i], MRB_TT_SYMBOL);
      ind = rate_prefactor_index(mrb, mrb_symbol(argv[i]));
      if (! is_valid_prefactor(drv,ind))
        mrb_raisef(mrb, E_ARGUMENT_ERROR,
            "unknown rate prefactor (arg %d)",mrb_fixnum_value(i+1));

      mrb_ary_push(mrb, ret, mrb_float_value(mrb,drv->rate_prefactors[ind]));
    }
  }

  return ret;
}

static mrb_value cChemDriver_set_rate_prefactor(mrb_state *mrb, mrb_value self)
{
  mrb_sym prefactor = 0;
  mrb_float val = 0.0;
  mrb_get_args(mrb, "nf", &prefactor, &val);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);

  int ind = rate_prefactor_index(mrb, prefactor);

  if (! is_valid_prefactor(drv,ind))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown rate prefactor (arg 1)");

  drv->rate_prefactors[ind] = val;

  return self;
}

static mrb_value cChemDriver_set_monomer_energy(mrb_state *mrb, mrb_value self)
{
  mrb_value r_energy_ind = mrb_nil_value();
  mrb_float energy = 0.0;
  mrb_get_args(mrb, "of", &r_energy_ind, &energy);

  ts_chem_driver* latt = tsmrb_to_chem_driver(mrb,self);

  spin_t s_energy_ind = spin_value(mrb, TS_CHEMDRIVER_SITE_MONOMER, 1, r_energy_ind);
  if (!is_valid_monomer_energy(latt, s_energy_ind.val))
    mrb_raisef(mrb, E_RANGE_ERROR,
        "invalid monomer energy index %S", mrb_fixnum_value(s_energy_ind.val));

  latt->monomer_energies[ s_energy_ind.val ] = energy;

  return self;
}

static mrb_value lookup_monomer_energies(mrb_state *mrb, mrb_value self)
{
  mrb_value r_ener_ind = mrb_nil_value();
  mrb_get_args(mrb, "o", &r_ener_ind);

  mrb_value r_drv = mrb_iv_get(mrb, self, mrb_intern_cstr(mrb,"@drv"));
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,r_drv);

  spin_t s_ener_ind = spin_value(mrb, TS_CHEMDRIVER_SITE_MONOMER, 1, r_ener_ind);
  if (!is_valid_monomer_energy(latt, s_ener_ind.val))
    mrb_raisef(mrb, E_RANGE_ERROR,
        "invalid monomer energy index %S", mrb_fixnum_value(s_ener_ind.val));

  return mrb_float_value(mrb, drv->monomer_energies[ s_ener_ind.val ] );
}

static mrb_value cChemDriver_get_monomer_energies(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* latt = tsmrb_to_chem_driver(mrb,self);
  mrb_value energies = mrb_hash_new(mrb);

  for (int i=0; i < NUM_CONST(monomer_states_all); i++)
  {
    mrb_hash_set(mrb, energies, 
        mrb_fixnum_value(monomer_states_all[i]),
        mrb_float_value(mrb, latt->monomer_energies[ monomer_states_all[i] ]));
  }

  mrb_iv_set(mrb, energies, mrb_intern_cstr(mrb,"@drv"), self);
  mrb_define_singleton_method(mrb, mrb_obj_ptr(energies),
      "[]", lookup_monomer_energies, MRB_ARGS_REQ(1));

  return energies;
}

static mrb_value cChemDriver_generate_site_moves(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_check_on_lattice(mrb, (ts_lattice*)drv, i,j);

  int states[TS_CHEMDRIVER_STATE_MAX];
  double rates[TS_CHEMDRIVER_STATE_MAX];
  size_t nmoves = 0;
  tsmrb_handle_error(mrb, 
      ts_chem_driver_generate_site_moves(drv, i,j, states, rates, &nmoves) );

  mrb_value moves = mrb_hash_new(mrb);

  for (size_t i=0; i < nmoves; i++) {
    mrb_hash_set(mrb, moves,
        mrb_fixnum_value(states[i]),
        mrb_float_value(mrb, rates[i]));
  }

  return moves;
}

static mrb_value cChemDriver_resetB(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* latt = tsmrb_to_chem_driver(mrb,self);
  ts_chem_driver_reset(latt);
  return self;
}

static mrb_value cChemDriver_get_barrier_model(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* latt = tsmrb_to_chem_driver(mrb,self);
  switch (latt->barrier_model)
  {
    case TS_CHEMDRIVER_BARRIER_NONE:         return TSMRB_SYM_OBJ(none);
    case TS_CHEMDRIVER_BARRIER_METROPOLIS:   return TSMRB_SYM_OBJ(metropolis);
    case TS_CHEMDRIVER_BARRIER_KAWASAKI:     return TSMRB_SYM_OBJ(kawasaki);
    case TS_CHEMDRIVER_BARRIER_PROPORTIONAL: return TSMRB_SYM_OBJ(proportional);
    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "unknown barrier model %d", (int)latt->barrier_model);
      return mrb_nil_value();
  }
}

struct barrier_model_map { int ind; enum TS_CHEMDRIVER_BARRIER_MODEL model; };

const struct barrier_model_map barrier_models[] = {
  { TSMRB_SYM_IND(none)        , TS_CHEMDRIVER_BARRIER_NONE         },
  { TSMRB_SYM_IND(metropolis)  , TS_CHEMDRIVER_BARRIER_METROPOLIS   },
  { TSMRB_SYM_IND(kawasaki)    , TS_CHEMDRIVER_BARRIER_KAWASAKI     },
  { TSMRB_SYM_IND(proportional), TS_CHEMDRIVER_BARRIER_PROPORTIONAL },
  { TSMRB_SYM_IND(infinite)     , TS_CHEMDRIVER_BARRIER_INFINITE      },
  { TSMRB_SYM_IND(always_accept), TS_CHEMDRIVER_BARRIER_ALWAYS_ACCEPT },
  { TSMRB_SYM_IND(always_reject), TS_CHEMDRIVER_BARRIER_ALWAYS_REJECT },
  { -1, TS_CHEMDRIVER_BARRIER_NONE },
};

static enum TS_CHEMDRIVER_BARRIER_MODEL sym_to_model(mrb_state *mrb, mrb_sym bmodel)
{
  for (int i=0; barrier_models[i].ind != -1; i++)
  {
    if (tsmrb_symtab_at(mrb, barrier_models[i].ind) == bmodel)
      return barrier_models[i].model;
  }

  mrb_raisef(mrb, E_ARGUMENT_ERROR,
      "unknown barrier model %S", mrb_symbol_value(bmodel));
  return TS_CHEMDRIVER_BARRIER_NONE;
}

static mrb_value cChemDriver_set_barrier_model(mrb_state *mrb, mrb_value self)
{
  mrb_sym bmodel = 0;
  mrb_get_args(mrb, "n", &bmodel);

  ts_chem_driver* latt = tsmrb_to_chem_driver(mrb,self);
  latt->barrier_model = sym_to_model(mrb,bmodel);
  return self;
}

static mrb_value cChemDriver_prob_acceptance(mrb_state *mrb, mrb_value self)
{
  mrb_sym rmodel = 0;
  mrb_float dE = 0.0;
  mrb_get_args(mrb, "nf", &rmodel, &dE);

  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  enum TS_CHEMDRIVER_BARRIER_MODEL model = sym_to_model(mrb, rmodel);
  double pacc = ts_chem_driver_prob_acceptance(drv, model, dE);
  return mrb_float_value(mrb,pacc);
}

static mrb_value cChemDriver_get_barrier_param(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_float_value(mrb, drv->barrier_param );
}

static mrb_value cChemDriver_set_barrier_param(mrb_state *mrb, mrb_value self)
{
  mrb_float param = 0.0;
  mrb_get_args(mrb, "f", &param);
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);
  drv->barrier_param = param;
  return self;
}

static mrb_value cChemDriver_get_energy(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value kwargs;
  double energy;

  drv = tsmrb_to_chem_driver(mrb,self);

  kwargs = mrb_nil_value();
  mrb_get_args(mrb, "|H", &kwargs);

  energy = 0.0;
  tsmrb_handle_error(mrb, 
      ts_chem_driver_energy(drv, &energy, NULL));
  
  return mrb_float_value(mrb,energy);
}

static mrb_value cChemDriver_monomer_populations(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv = tsmrb_to_chem_driver(mrb,self);
  mrb_value pops = mrb_hash_new(mrb);
  struct tsch_populations state_pops = { { 0 } };
  double energy = 0.0;
  int i;

  tsmrb_handle_error(mrb, 
      ts_chem_driver_energy(drv, &energy, &state_pops));

  for (i=0; i < NUM_CONST(monomer_states_gtp); i++)
  {
    int ind = monomer_states_gtp[i];
    char c = ts_chem_driver_site_codes[ind];
    mrb_value k = mrb_str_new(mrb,&c,1);
    mrb_hash_set(mrb, pops, k, mrb_fixnum_value((int)state_pops.monomer_pop[ind]));
  }

  return pops;
}

static mrb_value cChemDriver_interface_populations(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver* drv;
  mrb_value pops, key;
  char trip[3];
  int i;
  struct tsch_populations state_pops = { { 0 } };
  double energy = 0.0;

  drv = tsmrb_to_chem_driver(mrb,self);
  pops = mrb_hash_new(mrb);

  tsmrb_handle_error(mrb, 
      ts_chem_driver_energy(drv, &energy, &state_pops));

  for (i=NOT_EMPTY(monomer_states_gtp); i < NUM_CONST(monomer_states_gtp); i++)
  {
    int sti, k;
    sti = monomer_states_gtp[i];
    trip[0] = ts_chem_driver_site_codes[sti];

    for (k=NOT_EMPTY(monomer_states_gtp); k < NUM_CONST(monomer_states_gtp); k++) {
      int stk, j;

      stk = monomer_states_gtp[k];
      trip[2] = ts_chem_driver_site_codes[stk];

      for (j=NOT_EMPTY(longitudinal_states); j < NUM_CONST(longitudinal_states); j++) {
        int stj;

        stj = longitudinal_states[j];
        trip[1] = ts_chem_driver_site_codes[stj];

        key = mrb_str_new(mrb, trip+0,3);
        mrb_hash_set(mrb, pops, key, mrb_fixnum_value((int)state_pops.interface_pop[sti][stj][stk]));
      }

      for (j=NOT_EMPTY(lateral_states); j < NUM_CONST(lateral_states); j++) {
        int stj;

        stj = lateral_states[j];
        trip[1] = ts_chem_driver_site_codes[stj];

        key = mrb_str_new(mrb,trip+0,3);
        mrb_hash_set(mrb,pops, key, mrb_fixnum_value((int)state_pops.interface_pop[sti][stj][stk]));
      }
    }
  }

  trip[0] = trip[1] = trip[2] = ts_chem_driver_site_codes[TS_CHEMDRIVER_STATE_EMPTY];
  key = mrb_str_new(mrb, trip+0,3);
  mrb_hash_set(mrb, pops, key,
      mrb_fixnum_value(
        (int)state_pops.interface_pop[TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY][TS_CHEMDRIVER_STATE_EMPTY]));

  return pops;
}

static mrb_value cChemDriver_energy_change(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_value rs = mrb_nil_value();
  mrb_get_args(mrb, "iio", &i,&j,&rs);
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);

  spin_t s = spin_value(mrb, ts_chem_driver_site_type(drv, i,j), 3, rs);

  if (s.val < 0 || s.val >= PSEUDO_LATERAL_ROTATION)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "spin cannot be negative and cannot be a pseudo-spin");
  }

  double dE = 0.0;
  tsmrb_handle_error(mrb, 
      ts_chem_driver_energy_change(drv, i,j, s, &dE));
  
  return mrb_float_value(mrb,dE);
}

static mrb_value cChemDriver_lateral_energy_change(mrb_state *mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_value rs = mrb_nil_value();
  mrb_get_args(mrb, "iio", &i,&j,&rs);
  ts_chem_driver* drv = tsmrb_to_chem_driver(mrb,self);

  spin_t s = spin_value(mrb, ts_chem_driver_site_type(drv, i,j), 3, rs);

  if (s.val < 0 || s.val >= PSEUDO_LATERAL_ROTATION)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "spin cannot be negative and cannot be a pseudo-spin");
  }

  double dE = 0.0;
  tsmrb_handle_error(mrb, 
      ts_chem_driver_energy_change_lateral(drv, i, j, s, &dE));

  return mrb_float_value(mrb,dE);
}


static void check_ifc_bounds(mrb_state *mrb, int argnum, const char* type, spin_t spin)
{
  if ((spin.val < 0) || (spin.val >= TS_CHEMDRIVER_STATE_MAX))
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "invalid %s state index (arg %d)", type, (int)spin.val);
}

static const char *term2str(mrb_state *mrb, mrb_value term, mrb_int *slenp)
{
  const char *s;

  if (mrb_symbol_p(term))
  {
    s = mrb_sym2name_len(mrb,mrb_symbol(term),slenp);
    return s;
  }

  if (mrb_string_p(term))
  {
    s = RSTRING_PTR(term);
    *slenp = RSTRING_LEN(term);
    return s;
  }

  return NULL;
}

static void decode_interface_triplet(mrb_state *mrb, mrb_value r_triplet, spin_t triplet[3])
{
  const char* s = NULL;
  mrb_int slen = 0;

  if (s = term2str(mrb,r_triplet,&slen), s == NULL)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot convert type into interface energy triplet");
  }

  if (slen != 3) mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid triplet (length != 3)");
  spin_triplet_from_string(mrb, 1, s, triplet);
}

static void decode_bending_triplet(mrb_state *mrb, mrb_value r_triplet, spin_t triplet[3])
{
  const char *s = NULL;
  mrb_int slen = 0;

  if (s = term2str(mrb,r_triplet,&slen), s == NULL)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot convert type into bending energy triplet");
  }

  if (slen != 3) mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid triplet (length != 3)");
  spin_bending_triplet_from_string(mrb, 1, s, triplet);
}

static void decode_coupling_term(mrb_state *mrb, mrb_value r_term, int term[4])
{
  const char *s = NULL;
  mrb_int slen = 0;

  if (s = term2str(mrb,r_term,&slen), s == NULL)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot convert type into coupling energy term");
  }

  if (slen != 4) mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid coupling term (length != 4)");
  spin_coupling_term_from_string(mrb, 1, s, term);
}

static mrb_value
get_threebody_energy(mrb_state *mrb, mrb_value self,
    void (*triplet_decode)(mrb_state *mrb, mrb_value v, spin_t triplet[3]),
    const enum TS_CHEMDRIVER_SITE_TYPE site_types[3],
    const char *site_type_names[3])
{
  ts_chem_driver* drv;
  int argc = 0;
  mrb_value* argv = NULL;
  spin_t triplet[3] = { INVALID_SPIN, INVALID_SPIN, INVALID_SPIN };

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "*", &argv, &argc);

  if (argc == 1)
  {
    triplet_decode(mrb, argv[0], triplet);
  }
  else if (argc == 3)
  {
    triplet[0] = spin_value(mrb, site_types[0], 1, argv[0]);
    triplet[1] = spin_value(mrb, site_types[1], 2, argv[1]);
    triplet[2] = spin_value(mrb, site_types[2], 3, argv[2]);
  }
  else
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "wrong number of arguments (%d but expected 1 or 3)", argc);
  }

  check_ifc_bounds(mrb, 1, site_type_names[0], triplet[0]);
  check_ifc_bounds(mrb, 2, site_type_names[1], triplet[1]);
  check_ifc_bounds(mrb, 3, site_type_names[2], triplet[2]);

  return mrb_float_value(mrb, 
      drv->interface_energies[ triplet[0].val ][ triplet[1].val ][ triplet[2].val ]);
}

static enum TS_CHEMDRIVER_SITE_TYPE ifc_site_types[3] = {
  TS_CHEMDRIVER_SITE_MONOMER,
  TS_CHEMDRIVER_SITE_ANY    ,
  TS_CHEMDRIVER_SITE_MONOMER,
};

static const char *ifc_site_type_names[3] = {
  "monomer",
  "interface",
  "monomer",
};

static mrb_value cChemDriver_get_interface_energy(mrb_state *mrb, mrb_value self)
{
  return get_threebody_energy(mrb, self,
      decode_interface_triplet,
      ifc_site_types,
      ifc_site_type_names);
}

static enum TS_CHEMDRIVER_SITE_TYPE bending_site_types[3] = {
  TS_CHEMDRIVER_SITE_ANY    ,
  TS_CHEMDRIVER_SITE_MONOMER,
  TS_CHEMDRIVER_SITE_ANY    ,
};

static const char *bending_site_type_names[3] = {
  "interface",
  "monomer",
  "interface",
};

static mrb_value cChemDriver_get_bending_energy(mrb_state *mrb, mrb_value self)
{
  return get_threebody_energy(mrb, self,
      decode_bending_triplet,
      bending_site_types,
      bending_site_type_names);
}

static mrb_value
set_threebody_energy(mrb_state *mrb, mrb_value self,
    void (*triplet_decode)(mrb_state *mrb, mrb_value v, spin_t triplet[3]),
    const enum TS_CHEMDRIVER_SITE_TYPE site_types[3],
    const char *site_type_names[3])
{
  ts_chem_driver* drv;
  int argc = 0;
  mrb_value* argv = NULL;
  spin_t triplet[3] = { INVALID_SPIN, INVALID_SPIN, INVALID_SPIN };
  double en = 0.0;

  drv = tsmrb_to_chem_driver(mrb,self);
  mrb_get_args(mrb, "*", &argv, &argc);

  if (argc == 2)
  {
    triplet_decode(mrb, argv[0], triplet);

    mrb_check_type(mrb, argv[1], MRB_TT_FLOAT);
    en = mrb_float( argv[1] );
  }
  else if (argc == 4)
  {
    triplet[0] = spin_value(mrb, site_types[0], 1, argv[0]);
    triplet[1] = spin_value(mrb, site_types[1], 2, argv[1]);
    triplet[2] = spin_value(mrb, site_types[2], 3, argv[2]);

    mrb_check_type(mrb, argv[3], MRB_TT_FLOAT);
    en = mrb_float( argv[3] );
  }
  else
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "wrong number of arguments (%d but expected 2 or 4)", argc);
  }

  check_ifc_bounds(mrb, 1, site_type_names[0], triplet[0]);
  check_ifc_bounds(mrb, 2, site_type_names[1], triplet[1]);
  check_ifc_bounds(mrb, 3, site_type_names[2], triplet[2]);

  drv->interface_energies[ triplet[0].val ][ triplet[1].val ][ triplet[2].val ] = en;

  return self;
}

static mrb_value cChemDriver_set_interface_energy(mrb_state *mrb, mrb_value self)
{
  return set_threebody_energy(mrb, self,
      decode_interface_triplet,
      ifc_site_types,
      ifc_site_type_names);
}

static mrb_value cChemDriver_set_bending_energy(mrb_state *mrb, mrb_value self)
{
  return set_threebody_energy(mrb, self,
      decode_bending_triplet,
      bending_site_types,
      bending_site_type_names);
}

static mrb_value cChemDriver_set_coupling_energy(mrb_state *mrb, mrb_value self)
{
  int terms[4];
  ts_chem_driver *drv;
  mrb_value rterm;
  mrb_float elo,ehi;

  drv = tsmrb_to_chem_driver(mrb,self);

  rterm = mrb_nil_value();
  elo = ehi = 0.0;
  mrb_get_args(mrb, "off", &rterm, &elo, &ehi);

  decode_coupling_term(mrb, rterm, terms);
  tsmrb_handle_error(mrb, ts_chem_driver_add_coupling_energy(drv, terms, elo, ehi));

  return mrb_nil_value();
}

static mrb_value cChemDriver_get_coupling_energy(mrb_state *mrb, mrb_value self)
{
  int terms[4];
  ts_chem_driver *drv;
  mrb_value rterm, energies[2];
  int ind;

  drv = tsmrb_to_chem_driver(mrb,self);

  rterm = mrb_nil_value();
  mrb_get_args(mrb, "o", &rterm);

  decode_coupling_term(mrb, rterm, terms);

  ind = ts_chem_driver_find_coupling_energy(drv, terms);
  if (ind < 0)
  {
    return mrb_nil_value();
  }

  energies[0] = mrb_float_value(mrb, drv->coupling_energies[ind][0]);
  energies[1] = mrb_float_value(mrb, drv->coupling_energies[ind][1]);
  return mrb_ary_new_from_values(mrb, 2, energies);
}

static mrb_value
cChemDriver_add_dissociation_barrier(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  const char *s = NULL;
  spin_t term[3] = { {0} };
  mrb_value rterm;
  mrb_int slen = 0;
  mrb_float en;

  drv = tsmrb_to_chem_driver(mrb,self);

  rterm = mrb_nil_value();
  en = 0.0;
  mrb_get_args(mrb, "of", &rterm, &en);
  if (s = term2str(mrb,rterm,&slen), s == NULL)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot convert type into longitudinal interface");
  }

  if (slen != 3)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid longitudinal interface (length != 3)");
  }

  spin_triplet_from_string(mrb, 1, s, term);

  tsmrb_handle_error(mrb,
      ts_chem_driver_add_dissociation_barrier(drv, term, (double)en));

  return mrb_nil_value();
}

static mrb_value
cChemDriver_get_dissociation_barrier(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  const char *s = NULL;
  spin_t term[3] = { {0} };
  mrb_value rterm;
  int ind;
  mrb_int slen = 0;
  double en;

  drv = tsmrb_to_chem_driver(mrb,self);

  rterm = mrb_nil_value();
  en = 0.0;
  mrb_get_args(mrb, "o", &rterm);
  if (s = term2str(mrb,rterm,&slen), s == NULL)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot convert type into longitudinal interface");
  }

  if (slen != 3)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid longitudinal interface (length != 3)");
  }

  spin_triplet_from_string(mrb, 1, s, term);

  ind = ts_chem_driver_find_dissociation_barrier(drv, term);
  en = (ind >= 0) 
    ? drv->dissociation_barrier_energies[ind]
    : 0.0;

  return mrb_float_value(mrb,en);
}

static mrb_value cChemDriver_set_coupling_edge_distance(mrb_state *mrb, mrb_value self)
{
  mrb_int dist;
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  dist = 0;
  mrb_get_args(mrb, "i", &dist);

  if (dist < 0)
  {
    tsmrb_raisef(mrb, E_RUNTIME_ERROR, "invalid edge coupling distance %d", (int)dist);
  }

  drv->coupling_edge_dist = dist;

  return mrb_nil_value();
}

static mrb_value cChemDriver_get_coupling_edge_distance(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  drv = tsmrb_to_chem_driver(mrb,self);

  return mrb_fixnum_value(drv->coupling_edge_dist);
}

static mrb_value get_cached_row_ranges(mrb_state *mrb, ts_chem_driver *drv)
{
  size_t i, ncol;
  mrb_value ret;

  ncol = drv->latt.ncol;
  ret = mrb_ary_new_capa(mrb, 2*ncol);
  for (i=0; i < ncol; i++)
  {
    mrb_value pair[2];
    int ai;

    ai = mrb_gc_arena_save(mrb);
    pair[0] = mrb_fixnum_value( drv->row_range[2*i+0] );
    pair[1] = mrb_fixnum_value( drv->row_range[2*i+1] );
    mrb_ary_set(mrb, ret, i, mrb_ary_new_from_values(mrb, 2, pair));
    mrb_gc_arena_restore(mrb, ai);
  }

  return ret;
}

static mrb_value cChemDriver_get_cached_row_ranges(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  return get_cached_row_ranges(mrb,drv);
}

static mrb_value cChemDriver_update_row_ranges(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_handle_error(mrb, ts_chem_driver_update_row_range(drv));
  return get_cached_row_ranges(mrb,drv);
}

static mrb_value cChemDriver_print_configuration(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value io;
  FILE *f;

  io = tsmrb_get_stdout(mrb);
  mrb_get_args(mrb, "|o", &io);
  f = mrb_io_file(mrb, io);

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_handle_error(mrb,
      ts_chem_driver_print_configuration(drv, f));

  return mrb_nil_value();
}

enum { NUM_COUNT_BONDS = 16 };
static mrb_value cChemDriver_count_bonds(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  mrb_int argc;
  mrb_value *argv;

  int rv[NUM_COUNT_BONDS];
  int cv[NUM_COUNT_BONDS];
  int nlongv[NUM_COUNT_BONDS];
  int nlatv[NUM_COUNT_BONDS];

  drv = tsmrb_to_chem_driver(mrb,self);

  mrb_get_args(mrb, "*", &argv, &argc);

  if ((argc == 2) && mrb_fixnum_p(argv[0]) && mrb_fixnum_p(argv[1]))
  {
    mrb_value pair[2];

    rv[0] = mrb_fixnum(argv[0]);
    cv[0] = mrb_fixnum(argv[1]);
    nlongv[0] = 0;
    nlatv[0] = 0;

    tsmrb_handle_error(mrb, 
        ts_chem_driver_count_bonds(drv, rv,cv, nlongv,nlatv, 1));

    pair[0] = mrb_fixnum_value(nlongv[0]);
    pair[1] = mrb_fixnum_value(nlatv[0]);
    return mrb_ary_new_from_values(mrb, 2, pair);
  }
  else
  {
    int ind;
    int ai;
    mrb_value ret;

    if (argc == 0) { return mrb_ary_new(mrb); }

    ret = mrb_ary_new_capa(mrb, argc);
    ind = 0;
    while (ind < argc)
    {
      int i, rest;

      rest = argc - ind;
      if (rest > NUM_COUNT_BONDS) { rest = NUM_COUNT_BONDS; }

      for (i=0; i < rest; i++)
      {
        mrb_value arg = argv[i];
        if (!mrb_array_p(arg)
            || (RARRAY_LEN(arg) != 2)
            || !mrb_fixnum_p(RARRAY_PTR(arg)[0])
            || !mrb_fixnum_p(RARRAY_PTR(arg)[1]))
        {
          tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
              "expected an array of two numbers at argument %d", i+1);
        }

        rv[i] = mrb_fixnum(RARRAY_PTR(arg)[0]);
        cv[i] = mrb_fixnum(RARRAY_PTR(arg)[1]);
        nlongv[i] = 0;
        nlatv[i] = 0;
      }

      tsmrb_handle_error(mrb, 
          ts_chem_driver_count_bonds(drv, rv,cv, nlongv,nlatv, rest));

      ai = mrb_gc_arena_save(mrb);
      for (int i=0; i < rest; i++)
      {
        mrb_value pair[2];
        pair[0] = mrb_fixnum_value(nlongv[i]);
        pair[1] = mrb_fixnum_value(nlatv[i]);

        mrb_ary_set(mrb, ret, ind+i, mrb_ary_new_from_values(mrb, 2, pair));
      }
      mrb_gc_arena_restore(mrb, ai);

      ind += rest;
    }

    return ret;
  }
}

static mrb_value cChemDriver_new_ifclog(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_int sz;

  drv = tsmrb_to_chem_driver(mrb,self);

  sz = 0;
  mrb_get_args(mrb, "i", &sz);

  if (sz < 0)
  {
    mrb_raise(mrb, E_RANGE_ERROR, "ifclog size cannot be negavie");
  }

  tsmrb_handle_error(mrb, tsch_ifclog_resize(drv, (size_t)sz));
  return mrb_nil_value();
}

static int is_element(int x, int *vl, size_t n)
{
  size_t i;
  for (i=0; i < n; i++)
  {
    if (x == vl[i])
    {
      return 1;
    }
  }

  return 0;
}

static uint16_t convert_to_ifc(mrb_state *mrb, mrb_value obj)
{
  mrb_int triplen, ifc;
  const char *tripstr;
  spin_t triplet[3];

  switch (mrb_type(obj))
  {
    default:
      mrb_raise(mrb, E_TYPE_ERROR, "expected array of triples or integers");

    case MRB_TT_FIXNUM:
      ifc = mrb_fixnum(obj);
      goto validate_ifc;

    case MRB_TT_STRING:
      tripstr = RSTRING_PTR(obj);
      triplen = RSTRING_LEN(obj);
      break;

    case MRB_TT_SYMBOL:
      tripstr = mrb_sym2name_len(mrb,mrb_symbol(obj),&triplen);
      break;
  }

  if (triplen != 3)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "interface value is not an interface triplet");
  }

  triplet[0].val = triplet[1].val = triplet[2].val = 0;
  spin_triplet_from_string(mrb, -1, tripstr, triplet);
  ifc = tsch_ifclog_pack_ifc(triplet[0].val, triplet[1].val, triplet[2].val);

validate_ifc:
  if (ifc < 0 || ifc > UINT16_MAX)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid interface value %d", (int)ifc);
  }

  return ifc;
}

static mrb_value cChemDriver_get_ifclog(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value ret;
  size_t i,n;
  int sp0, sp1, sp2;
  mrb_value conds;

  mrb_sym sym_plus  = mrb_intern_lit(mrb, "plus");
  mrb_sym sym_minus = mrb_intern_lit(mrb, "minus");
  mrb_sym errsym    = mrb_intern_lit(mrb, "bad_entry");

  int only_valid = 0, tip_pref = 0;
  size_t num_pspins = 0, num_ifcs = 0;
  int *pspins, pspins_buf[8], *ifcs, ifcs_buf[8];
  int rmin,rmax,cmin,cmax;

  conds = mrb_nil_value();
  mrb_get_args(mrb, "|H", &conds);

  drv = tsmrb_to_chem_driver(mrb,self);
  if (!tsch_ifclog_is_logging(drv))
  {
    return mrb_nil_value();
  }

  rmin = INT16_MIN;
  rmax = INT16_MAX;
  cmin = 0;
  cmax = UINT16_MAX;

  if (!mrb_nil_p(conds))
  {
    mrb_value r_pspins, r_ifcs, r_tip;

    mrb_value s_onlyvalid = mrb_symbol_value(mrb_intern_lit(mrb, "onlyvalid"));
    mrb_value s_pspins    = mrb_symbol_value(mrb_intern_lit(mrb, "pspins"));
    mrb_value s_ifcs      = mrb_symbol_value(mrb_intern_lit(mrb, "ifcs"));
    mrb_value s_tip       = mrb_symbol_value(mrb_intern_lit(mrb, "tip"));

    only_valid = mrb_test(mrb_hash_get(mrb, conds, s_onlyvalid));

    r_pspins = mrb_hash_get(mrb, conds, s_pspins);
    if (mrb_test(r_pspins))
    {
      num_pspins = ARRAYLEN(pspins_buf);
      pspins = tsmrb_ary_to_ints(mrb,r_pspins, pspins_buf, &num_pspins);
      if (pspins != &pspins_buf[0])
      {
        tsmrb_memory_ref(mrb, pspins);
      }
    }

    r_ifcs = mrb_hash_get(mrb, conds, s_ifcs);
    if (mrb_test(r_ifcs))
    {
      num_ifcs = mrb_ary_len(mrb,r_ifcs);
      ifcs = &ifcs_buf[0];
      if (num_ifcs > ARRAYLEN(ifcs_buf))
      {
        ifcs = mrb_calloc(mrb, num_ifcs, sizeof *ifcs);
        tsmrb_memory_ref(mrb, ifcs);
      }

      for (i=0; i < num_ifcs; i++)
      {
        mrb_value elt = mrb_ary_ref(mrb, r_ifcs, i);
        ifcs[i] = convert_to_ifc(mrb, elt);
      }
    }


    r_tip = mrb_hash_get(mrb, conds, s_tip);
    if (mrb_symbol_p(r_tip))
    {
      mrb_sym tv = mrb_symbol(r_tip);
      if (tv == sym_plus)
      {
        tip_pref = 2;
      }
      else if (tv == sym_minus)
      {
        tip_pref = 1;
      }
    }
  }


  ret = mrb_ary_new(mrb);
  sp0=sp1=sp2=0;
  for (i=0,n=tsch_ifclog_nentries(drv); i < n; i++)
  {
    int ai;
    mrb_value tuple[5], elt;
    struct tsch_ifclog_entry *ent;
    int is_valid, is_minus, rel_row;

    ai = mrb_gc_arena_save(mrb);
    if (ent = tsch_ifclog_at(drv,i), ent == NULL)
    {
      tsmrb_raisef(mrb, E_RANGE_ERROR, "ifclog entry %zu should be valid but is not", i);
    }

    if (num_pspins > 0 && !is_element(ent->move, pspins, num_pspins))
    {
      continue;
    }

    if (num_ifcs > 0 && !is_element(ent->ifc, ifcs, num_ifcs))
    {
      continue;
    }

    is_minus = 0;
    rel_row = tsch_ifclog_decode_rrow(ent->rel_row, &is_minus);

    if ((tip_pref == 1 && !is_minus) || (tip_pref == 2 && is_minus))
    {
      continue;
    }

    if (rel_row < rmin || rel_row > rmax)
    {
      continue;
    }

    is_valid = tsch_ifclog_unpack_ifc(ent->ifc, &sp0,&sp1,&sp2);
    if (is_valid)
    {
      char trip[3];
      mrb_sym tripsym;

      assert(sp0 >= 0 && sp0 < TS_CHEMDRIVER_STATE_MAX);
      assert(sp1 >= 0 && sp1 < TS_CHEMDRIVER_STATE_MAX);
      assert(sp2 >= 0 && sp2 < TS_CHEMDRIVER_STATE_MAX);

      trip[0] = ts_chem_driver_site_codes[sp0];
      trip[1] = ts_chem_driver_site_codes[sp1];
      trip[2] = ts_chem_driver_site_codes[sp2];

      tripsym = mrb_intern(mrb,trip,3);
      tuple[0] = mrb_symbol_value(tripsym);
      tuple[1] = mrb_fixnum_value(ent->move);
      tuple[2] = mrb_symbol_value(is_minus ? sym_minus : sym_plus);
      tuple[3] = mrb_fixnum_value(rel_row);
      tuple[4] = mrb_fixnum_value(ent->col);
      elt = mrb_ary_new_from_values(mrb, 5, tuple);
    }
    else if (!only_valid)
    {
      tuple[0] = mrb_symbol_value(errsym);
      tuple[1] = mrb_fixnum_value(ent->ifc);
      tuple[2] = mrb_fixnum_value(ent->move);
      tuple[3] = mrb_fixnum_value(ent->rel_row);
      tuple[4] = mrb_fixnum_value(ent->col);
      elt = mrb_ary_new_from_values(mrb, 5, tuple);
    }
    else
    {
      continue;
    }

    mrb_ary_push(mrb, ret, elt);
    mrb_gc_arena_restore(mrb, ai);
  }

  return ret;
}

static mrb_value cChemDriver_clear_ifclog(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  tsch_ifclog_clear(drv);

  return mrb_nil_value();
}

static mrb_value cChemDriver_ifclogQ(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  return mrb_bool_value( tsch_ifclog_is_logging(drv) );
}

static mrb_value cChemDriver_ifclog_size(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  size_t n;

  drv = tsmrb_to_chem_driver(mrb,self);
  n = tsch_ifclog_nentries(drv);
  
  if (n > MRB_INT_MAX)
  {
    mrb_raise(mrb, E_RANGE_ERROR, "size is greater than Fixnum size");
  }

  return mrb_fixnum_value((mrb_int)n);
}

static mrb_value cChemDriver_ifclog_capa(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  size_t capa;

  drv = tsmrb_to_chem_driver(mrb,self);
  capa = tsch_ifclog_capa(drv);

  if (capa > MRB_INT_MAX)
  {
    mrb_raise(mrb, E_RANGE_ERROR, "capacity is greater than Fixnum size");
  }

  return mrb_fixnum_value((mrb_int)capa);
}

static mrb_value cChemDriver_ifclog_zeroB(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);
  tsch_ifclog_zero(drv);
  return mrb_nil_value();
}

static mrb_value cChemDriver_ifclog_unpack_ifc(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_int ifc;
  int sp0,sp1,sp2;
  char trip[3];
  mrb_sym tripsym;

  drv = tsmrb_to_chem_driver(mrb,self);
  ifc = -1;
  mrb_get_args(mrb, "i", &ifc);

  if (ifc < 0 || ifc > UINT16_MAX)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid interface value");
  }

  sp0 = sp1 = sp2 = 0;
  if (!tsch_ifclog_unpack_ifc(ifc, &sp0,&sp1,&sp2))
  {
    return mrb_fixnum_value(ifc);
  }

  assert(sp0 >= 0 && sp0 < TS_CHEMDRIVER_STATE_MAX);
  assert(sp1 >= 0 && sp1 < TS_CHEMDRIVER_STATE_MAX);
  assert(sp2 >= 0 && sp2 < TS_CHEMDRIVER_STATE_MAX);

  trip[0] = ts_chem_driver_site_codes[sp0];
  trip[1] = ts_chem_driver_site_codes[sp1];
  trip[2] = ts_chem_driver_site_codes[sp2];

  tripsym = mrb_intern(mrb,trip,3);
  return mrb_symbol_value(tripsym);
}

static mrb_value cChemDriver_ifclog_add(mrb_state *mrb, mrb_value self)
{
  mrb_value arg;
  mrb_int ifc, mv, r, c;
  ts_chem_driver *drv;
  size_t ind;

  drv = tsmrb_to_chem_driver(mrb,self);

  arg = mrb_nil_value();
  ifc = mv = r = c = -1;
  mrb_get_args(mrb, "oiii", &arg,&mv,&r,&c);

  if (mrb_symbol_p(arg))
  {
    const char *symstr;
    mrb_int symlen;
    spin_t triplet[3] = { {.val = 0}, {.val = 0}, {.val = 0} };

    symlen = 0;
    symstr = mrb_sym2name_len(mrb,mrb_symbol(arg),&symlen);
    if (symlen != 3)
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "interface value is not an interface triplet");
    }

    spin_triplet_from_string(mrb, 1, symstr, triplet);
    ifc = tsch_ifclog_pack_ifc(triplet[0].val, triplet[1].val, triplet[2].val);
  }
  else if (mrb_fixnum_p(arg))
  {
    ifc = mrb_fixnum(arg);
  }
  else
  {
    mrb_raise(mrb,E_TYPE_ERROR, "expected symbol or fixnum for interface");
  }

  if (ifc < 0 || ifc > UINT16_MAX)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid interface value");
  }

  if (mv < 0 || mv > UINT16_MAX)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid move value");
  }

  if (r < 0 || r > UINT16_MAX || r >= drv->latt.nrow)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid row value");
  }

  if (c < 0 || c > UINT16_MAX || c >= drv->latt.ncol)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid column value");
  }

  if (!tsch_ifclog_is_logging(drv))
  {
    return mrb_nil_value();
  }

  ind = tsch_ifclog_add(drv, ifc, mv, r, c);
  if ((~ind) == 0)
  {
    return mrb_nil_value();
  }

  if (ind > MRB_INT_MAX)
  {
    char buf[2 + CHAR_BIT * sizeof ind / 4];
    snprintf(buf,sizeof buf, "%zx", ind);
    return mrb_str_new_cstr(mrb, buf);
  }

  return mrb_fixnum_value(ind);
}


static mrb_value cChemDriver_check_move_consistencyB(mrb_state *mrb, mrb_value self)
{
  size_t rv[16], cv[16];
  TSResult err;
  ts_chem_driver *drv;

  drv = tsmrb_to_chem_driver(mrb,self);

  err = ts_chem_driver_check_move_consistency(drv, rv,cv, ARRAYLEN(rv));
  if (IS_OKAY(err))
  {
    return mrb_nil_value();
  }

  switch (ERRCODE(err))
  {
    case TS_ERR_INVALID_RETURN:

    default:

    case TS_ERR_LATT_BAD_STATE:
    case TS_ERR_INVALID_ARGUMENT:
      tsmrb_handle_error(mrb, err);
      break;

    case TS_ERR_CHEMDRIVER_INCONSISTENT_ROW_RANGE:
    case TS_ERR_LATT_DUPLICATE_MOVE:
      rv[1] = cv[1] = ~(size_t)0;

    case TS_ERR_SIMULATION_STATE:
    case TS_ERR_INSUFFICIENT_SIZE:
      {
        const struct ts_error_state *st;
        mrb_value ret;
        mrb_value pair[2], coord;
        size_t i, sentinel = ~(size_t)0;

        ret = mrb_ary_new(mrb);
        mrb_ary_push(mrb, ret, mrb_fixnum_value(ERRCODE(err)));
        st = ts_error_get_state(err);
        mrb_ary_push(mrb, ret,
            mrb_str_new_cstr(mrb, st->error_message));

        for (i=0; i < ARRAYLEN(rv) && rv[i] != sentinel; i++)
        {
          int ai;
          ai = mrb_gc_arena_save(mrb);
          pair[0] = mrb_fixnum_value(rv[i]);
          pair[1] = mrb_fixnum_value(cv[i]);
          coord = mrb_ary_new_from_values(mrb, 2, pair);
          mrb_ary_push(mrb, ret, coord);
          mrb_gc_arena_restore(mrb,ai);
        }

        return ret;
      }

  }

  return mrb_nil_value();
}

static mrb_value cChemDriver_ifclog_write(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value io;
  FILE *f;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_handle_error(mrb, tsch_ifclog_write(drv, f));
  return mrb_nil_value();
}

static mrb_value cChemDriver_ifclog_read(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value io;
  FILE *f;
  TSResult err;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  drv = tsmrb_to_chem_driver(mrb,self);

  err = tsch_ifclog_read(drv, f);
  if (IS_ERROR(err))
  {
    tsmrb_handle_error(mrb, err);
  }

  return mrb_fixnum_value(tsch_ifclog_nentries(drv));
}

static mrb_value cChemDriver_write_state(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value io;
  FILE *f;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  drv = tsmrb_to_chem_driver(mrb,self);
  tsmrb_handle_error(mrb, tsch_write_state(drv, f));
  return mrb_nil_value();
}

static mrb_value save_header_to_hash(mrb_state *mrb, const struct tsch_save_header *header)
{
  mrb_value ret = mrb_hash_new(mrb);

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_lit(mrb,"version")),
      mrb_fixnum_value(header->version));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_lit(mrb,"flags")),
      mrb_fixnum_value(header->flags));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_lit(mrb,"nsteps")),
      mrb_float_value(mrb,header->nsteps));          

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_lit(mrb,"time")),
      mrb_float_value(mrb,header->time));          

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_lit(mrb,"energy")),
      mrb_float_value(mrb,header->energy));          

  return ret;
}

static mrb_value cChemDriver_read_state_header(mrb_state *mrb, mrb_value mod)
{
  struct tsch_save_header header;
  mrb_value io, rvals[3];
  FILE *f;
  TSResult err;
  uint32_t nrow,ncol;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  memset(&header, 0, sizeof(header));

  nrow=ncol=0;
  err = tsch_read_header(f, &header, &nrow,&ncol);
  if (IS_ERROR(err))
  {
    tsmrb_handle_error(mrb, err);
  }
  else if (IS_IO_EOF(err))
  {
    return mrb_nil_value();
  }

  rvals[0] = save_header_to_hash(mrb, &header);
  rvals[1] = mrb_fixnum_value(nrow);
  rvals[2] = mrb_fixnum_value(ncol);
  return mrb_ary_new_from_values(mrb, 3, rvals);
}

static mrb_value cChemDriver_read_state(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  struct tsch_save_header header;
  mrb_value io, opts;
  FILE *f;
  TSResult err;

  io = mrb_nil_value();
  opts = mrb_nil_value();
  mrb_get_args(mrb, "o|H", &io, &opts);
  f = mrb_io_file(mrb, io);

  drv = tsmrb_to_chem_driver(mrb,self);
  memset(&header, 0, sizeof(header));

  err = tsch_read_state(drv, f, &header);
  if (IS_ERROR(err))
  {
    tsmrb_handle_error(mrb, err);
  }
  else if (IS_IO_EOF(err))
  {
    return mrb_nil_value();
  }

  if (mrb_hash_p(opts))
  {
    mrb_value opt;

    opt = mrb_hash_get(mrb, opts, TSMRB_SYM_OBJ(settime));
    if (mrb_test(opt))
    {
      drv->latt.time = header.time;
    }
  }

  return save_header_to_hash(mrb, &header);
}

static void snapshot_free(mrb_state *mrb, void *p)
{
  tsch_snapshot_free(p);
}

static const mrb_data_type tsch_snapshot_type = {
  .struct_name = "TschSnapshot",
  .dfree = snapshot_free,
};

static mrb_value get_snapshot(mrb_state *mrb, mrb_value self, struct tsch_snapshot **snapp)
{
  ts_chem_driver *driver;
  mrb_value rsnap;
  mrb_sym snap_sym;

  driver = tsmrb_to_chem_driver(mrb,self);
  snap_sym = mrb_intern_lit(mrb, "@snapshot");
  rsnap = mrb_iv_get(mrb, self, snap_sym);

  if (!mrb_nil_p(rsnap))
  {
    *snapp = mrb_data_get_ptr(mrb, rsnap, &tsch_snapshot_type);
  }
  else
  {
    *snapp = NULL;
  }

  return rsnap;
}

static mrb_value make_snapshot(mrb_state *mrb, ts_chem_driver *driver)
{
  mrb_sym snap_sym;
  struct RData *data;
  struct RClass *cSnapshot;
  struct tsch_snapshot *snap;

  snap_sym = mrb_intern_lit(mrb, "@snapshot");

  cSnapshot = mrb_class_get_under(mrb, tsmrb_classes(mrb)->cChemDriver, "Snapshot");
  if (cSnapshot == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "Snapshot class is missing!");
  }

  if (snap = tsch_snapshot_new(driver), snap == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not allocate snapshot");
  }

  data = mrb_data_object_alloc(mrb, cSnapshot, snap, &tsch_snapshot_type);
  return mrb_obj_value(data);
}

static mrb_value cChemDriver_take_snapshot(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *driver;
  struct tsch_snapshot *snap;
  mrb_value rsnap;

  driver = tsmrb_to_chem_driver(mrb,self);

  snap = NULL;
  rsnap = get_snapshot(mrb,self,&snap);

  if (!mrb_nil_p(rsnap))
  {
    tsmrb_handle_error(mrb, tsch_snapshot_update(driver, snap));
  }
  else
  {
    rsnap = make_snapshot(mrb, driver);
    mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@snapshot"), rsnap);
  }

  return rsnap;
}

static mrb_value cChemDriver_write_delta(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *drv;
  mrb_value io;
  FILE *f;
  struct tsch_snapshot *snap;

  drv = tsmrb_to_chem_driver(mrb,self);

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  snap = NULL;
  if (mrb_nil_p(get_snapshot(mrb, self, &snap)))
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "write_delta needs an existing snapshot");
  }

  tsmrb_handle_error(mrb, tsch_snapshot_write_delta(drv, snap, f));
  return mrb_nil_value();
}


static mrb_value translate_config_map(mrb_state *mrb, struct tsch_config *config, mrb_value map);
static mrb_value translate_config_seq(mrb_state *mrb, struct tsch_config *config, mrb_value seq);

static mrb_value translate_config_value(mrb_state *mrb, struct tsch_config *config)
{
  switch (config->type)
  {
    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "invalid config type %d", config->type);
      return mrb_nil_value();

    case TSCH_CFG_NONE:     return mrb_nil_value();
    case TSCH_CFG_STRING:   return mrb_str_new_cstr(mrb,config->s);
    case TSCH_CFG_NUM:      return mrb_float_value(mrb,config->num);
    case TSCH_CFG_BOOL:     return mrb_bool_value(config->b);
    case TSCH_CFG_MAPPING:  return translate_config_map(mrb,config->val,mrb_hash_new(mrb));
    case TSCH_CFG_SEQUENCE: return translate_config_seq(mrb,config->val,mrb_ary_new(mrb));
  }
}

static mrb_value translate_config_seq(mrb_state *mrb, struct tsch_config *config, mrb_value seq)
{
  int ai;
  ai = mrb_gc_arena_save(mrb);

  for(; config; config = config->next)
  {
    mrb_value v = translate_config_value(mrb, config);
    mrb_ary_push(mrb,seq,v);
  }

  mrb_gc_arena_restore(mrb,ai);
  return seq;
}

static mrb_value translate_config_map(mrb_state *mrb, struct tsch_config *config, mrb_value hash)
{
  int ai;

  ai = mrb_gc_arena_save(mrb);
  for (; config; config = config->next)
  {
    mrb_value k,v;
    int ai2;

    if (!config->key) {
      if (config->type != TSCH_CFG_MAPPING) {
        /* report an error! */
        mrb_raise(mrb, E_RUNTIME_ERROR,
            "found an entry without a key that is not a map!");
      }

      translate_config_map(mrb,config->val,hash);
      continue;
    }

    ai2 = mrb_gc_arena_save(mrb);
    k = mrb_str_new_cstr(mrb, config->key);
    v = translate_config_value(mrb,config);
    mrb_hash_set(mrb, hash, k,v);
    mrb_gc_arena_restore(mrb,ai2);
  }

  mrb_gc_arena_restore(mrb,ai);
  return hash;
}

static mrb_value translate_config(mrb_state *mrb, struct tsch_config *config)
{
  return translate_config_map(mrb, config, mrb_hash_new(mrb));
}

static mrb_value config_add_params(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  mrb_value pptr, params, keys;
  const mrb_value *kv;
  mrb_int i,n;

  pptr = params = mrb_nil_value();
  mrb_get_args(mrb, "oH", &pptr, &params);

  p = mrb_cptr(pptr);
  keys = mrb_hash_keys(mrb, params);
  for (i=0, n=RARRAY_LEN(keys), kv=RARRAY_PTR(keys); i < n; i++)
  {
    mrb_value k, v;
    const char *pname;
    char buf[128];
    double pval;

    if (mrb_string_p(kv[i]))
    {
      k = kv[i];
    }
    else if (mrb_symbol_p(kv[i]))
    {
      k = mrb_sym2str(mrb, mrb_symbol(kv[i]));
    }
    else
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "parameter keys must be strings or symbols");
    }

    v = mrb_hash_get(mrb, params, kv[i]);
    if (mrb_fixnum_p(v))
    {
      pval = (double)mrb_fixnum(v);
    }
    else if (mrb_float_p(v))
    {
      pval = (double)mrb_float(v);
    }
    else
    {
      pval = 0.0;
      mrb_raise(mrb, E_ARGUMENT_ERROR, "parameter values must be numeric");
    }

    pname = mrb_str_to_cstr(mrb, k);
    if (pname[0] != TSCH_CFG_PARAM_PREFIX)
    {
      int ret;
      ret = snprintf(buf, sizeof buf, "%c%s", TSCH_CFG_PARAM_PREFIX, pname);

      if (ret < 0)
      {
        tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
            "error prepending '%c' to parameter name '%s': %s",
            TSCH_CFG_PARAM_PREFIX, pname, strerror(errno));
      }
      else if ((size_t)ret > sizeof buf)
      {
        tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
            "parameter name '%s' is too long to prepend '%c' (exceeds %zu characters)",
            pname, TSCH_CFG_PARAM_PREFIX, sizeof buf - 2);
      }

      pname = buf;
    }

    tsmrb_handle_error(mrb, tsch_config_add_parameter(p, pname, pval));
  }

  return mrb_nil_value();
}

static mrb_value cChemDriver_bare_configure(mrb_state *mrb, mrb_value cls)
{
  mrb_value parser_obj, str, params;
  struct tsch_config_parser *p;

  str = params = mrb_nil_value();
  mrb_get_args(mrb, "o|H", &str, &params);

  (void)config_add_params;

  if (mrb_nil_p(str))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "empty configuration");
  }

  parser_obj = mrb_obj_new(mrb,
      tsmrb_chem_configure_class(mrb),
      0, NULL);

  if (!mrb_nil_p(params))
  {
    int ai;

    ai = mrb_gc_arena_save(mrb);
    mrb_funcall_argv(mrb, parser_obj,
        mrb_intern_lit(mrb,"set_params"),
        1, &params);
    mrb_gc_arena_restore(mrb, ai);
  }

  mrb_funcall_argv(mrb, parser_obj,
      mrb_intern_lit(mrb,"parse"),
      1, &str);

  p = tsmrb_to_config_parser(mrb,parser_obj);
  return translate_config(mrb, p->head);
}

static TSResult
parse_and_build_driver(
    struct tsch_config_parser *p,
    const char *s,
    int ignore_missing,
    ts_chem_driver **driverp,
    struct tsch_config **configp)
{
  struct tsch_config *config;

  RETURN_ON_ERROR(tsch_config_initialize(p));
  config = NULL;
  RETURN_ON_ERROR(tsch_parse_config(p,&config,s));

  if (driverp) {
    RETURN_ON_ERROR(
        tsch_config_build_driver(config, driverp, ignore_missing));
  }

  if (configp) {
    *configp = config;
  }

  return TS_OK;
}

static mrb_value cChemDriver_configure(mrb_state *mrb, mrb_value cls)
{
  mrb_value parser_obj, str, params;

  str = params = mrb_nil_value();
  mrb_get_args(mrb, "o|H", &str, &params);

  (void)parse_and_build_driver;

  if (mrb_nil_p(str))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "empty configuration");
  }

  parser_obj = mrb_obj_new(mrb,
      tsmrb_chem_configure_class(mrb),
      0, NULL);

  if (!mrb_nil_p(params))
  {
    int ai;

    ai = mrb_gc_arena_save(mrb);
    mrb_funcall_argv(mrb, parser_obj,
        mrb_intern_lit(mrb,"set_params"),
        1, &params);
    mrb_gc_arena_restore(mrb, ai);
  }


  mrb_funcall_argv(mrb, parser_obj,
      mrb_intern_lit(mrb,"parse"),
      1, &str);

  return mrb_funcall_argv(mrb, parser_obj,
      mrb_intern_lit(mrb,"build"),
      0, NULL);
}


static void sheet_length_bias_free(mrb_state *mrb, void* p)
{
  tsch_sheet_length_bias_delete((struct tsch_sheet_length_bias *)p);
}

static const mrb_data_type sheet_length_bias_type = {
  .struct_name = "SheetLengthBias",
  .dfree = sheet_length_bias_free,
};

struct tsch_sheet_length_bias *tsmrb_to_sheet_length_bias(mrb_state *mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_sheet_length_bias_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a SheetLengthBias object");

  return DATA_PTR(obj);
}

static mrb_value cSheetLengthBias_initialize(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *driver;
  struct tsch_sheet_length_bias *bias;
  void *drvp;
  mrb_value bias_data;
  const mrb_value *entries;
  mrb_int i, len;
  

  drvp = NULL;
  mrb_get_args(mrb, "dA", &drvp, &chem_driver_type, &bias_data);

  driver = drvp;

  len = mrb_ary_len(mrb,bias_data);
  bias = tsch_sheet_length_bias_new(driver, len);
  if (bias == NULL) {
    mrb_sys_fail(mrb, "out of memory");
  }

  DATA_PTR(self) = bias;
  DATA_TYPE(self) = &sheet_length_bias_type;

  entries = RARRAY_PTR(bias_data);
  for (i=0; i < len; i++)
  {
    bias->potential[i] = mrb_to_flo(mrb, entries[i]);
  }

  return self;
}

static mrb_value cSheetLengthBias_get_potential(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;
  mrb_value table;
  size_t i;

  bias = tsmrb_to_sheet_length_bias(mrb, self);

  table = mrb_ary_new_capa(mrb, bias->npts);
  for (i=0; i < bias->npts; i++)
  {
    mrb_ary_push(mrb, table, mrb_float_value(mrb,bias->potential[i]));
  }

  return table;
}

static mrb_value cSheetLengthBias_sheet_length(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;

  bias = tsmrb_to_sheet_length_bias(mrb, self);
  return mrb_float_value(mrb, tsch_sheet_length(bias->driver));
}

static mrb_value cSheetLengthBias_get_scale(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;

  bias = tsmrb_to_sheet_length_bias(mrb, self);
  return mrb_float_value(mrb, bias->scale);
}

static mrb_value cSheetLengthBias_set_scale(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;
  mrb_float scale;

  scale = 1.0;
  mrb_get_args(mrb, "f", &scale);

  if ((scale < 0.0) || !isfinite(scale))
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "scale must be non-negative and finite");
  }

  bias = tsmrb_to_sheet_length_bias(mrb, self);
  bias->scale = scale;

  return mrb_nil_value();
}

static mrb_value cSheetLengthBias_sheet_energy(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;
  double energy;
  mrb_float sheet_length;

  sheet_length = 1.0;
  mrb_get_args(mrb, "f", &sheet_length);

  if ((sheet_length < 0.0) || !isfinite(sheet_length))
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "sheet length must be non-negative and finite");
  }

  bias = tsmrb_to_sheet_length_bias(mrb, self);
  energy = tsch_sheet_length_bias_energy(bias, (double)sheet_length);
  if (isnan(energy))
  {
    tsmrb_raisef(mrb, E_FLOATDOMAIN_ERROR, "energy is NaN");
  }

  return mrb_float_value(mrb, (mrb_float)energy);
}

static mrb_value cSheetLengthBias_energy_change(mrb_state *mrb, mrb_value self)
{
  struct tsch_sheet_length_bias *bias;
  ts_chem_driver *drv;
  mrb_int i,j,st;
  double dE;

  bias = tsmrb_to_sheet_length_bias(mrb, self);
  drv = bias->driver;

  i = j = st = -1;
  mrb_get_args(mrb, "iii", &i, &j, &st);

  dE = 0.0;
  tsmrb_handle_error(mrb, tsch_sheet_length_bias_energy_change(bias, i,j, st, &dE));

  return mrb_float_value(mrb, dE);
}


static void fast_sheet_model_free(mrb_state *mrb, void* p)
{
  tsch_fast_sheet_model_delete((struct tsch_fast_sheet_model *)p);
}

static const mrb_data_type fast_sheet_model_type = {
  .struct_name = "FastSheetModel",
  .dfree = fast_sheet_model_free,
};

struct tsch_fast_sheet_model *tsmrb_to_fast_sheet_model(mrb_state *mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_fast_sheet_model_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a FastSheetModel object");

  return DATA_PTR(obj);
}

static void optional_array_to_floats(mrb_state *mrb, mrb_value arr, const char *opt,
    int nmin, int nmax, double *restrict dbls)
{
  const mrb_value *vals;
  mrb_int i,n;

  if (mrb_nil_p(arr))
  {
    return;
  }

  if (!mrb_array_p(arr))
  {
    tsmrb_raisef(mrb, E_TYPE_ERROR, ":%s expects an array of numbers", opt);
  }


  n = mrb_ary_len(mrb,arr);
  if (n < nmin || n > nmax)
  {
    if (nmin == nmax)
    {
      tsmrb_raisef(mrb, E_TYPE_ERROR,
          ":%s expects an array of %d numbers (%d given)", opt,nmax,n);
    }

    tsmrb_raisef(mrb, E_TYPE_ERROR,
        ":%s expects an array of %d-%d numbers (%d given)", opt,nmin,nmax,n);
  }

  for (vals=RARRAY_PTR(arr), i=0; i < n; i++)
  {
    if (!mrb_numeric_p(vals[i]))
    {
      tsmrb_raisef(mrb, E_TYPE_ERROR,
          ":%s expects an array of numbers (item at index %d is not numeric)", opt, i);
    }

    dbls[i] = mrb_to_flo(mrb, vals[i]);
  }
}

static mrb_value cFastSheetModel_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_value rdriver, kw;
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  void *drvp;
  int npf;

  drvp = NULL;
  rdriver = kw = mrb_nil_value();
  mrb_get_args(mrb, "o|H", &rdriver, &kw);

  driver = tsmrb_to_chem_driver(mrb, rdriver);
  npf = ts_chem_driver_num_protofilaments(driver);

  if (model = tsch_fast_sheet_model_new(driver), model == NULL) {
    mrb_sys_fail(mrb, "out of memory");
  }

  DATA_PTR(self) = model;
  DATA_TYPE(self) = &fast_sheet_model_type;

  if (!mrb_nil_p(kw))
  {
    mrb_value arr;

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"ifc_model"));
    optional_array_to_floats(mrb, arr, "ifc_model", 1,
        sizeof(model->ifc_model)/sizeof(model->ifc_model[0]),
        model->ifc_model+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"existing_model"));
    optional_array_to_floats(mrb, arr, "existing_model", 1,
        sizeof(model->existing_model)/sizeof(model->existing_model[0]),
        model->existing_model+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"relax_coeff"));
    optional_array_to_floats(mrb, arr, "relax_coeff", npf,npf, model->relax_coeff+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"no_tube_strain"));
    optional_array_to_floats(mrb, arr, "no_tube_strain", npf,npf, model->no_tube_strain+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"left_tube_strain"));
    optional_array_to_floats(mrb, arr, "left_tube_strain", npf-1,npf, model->left_tube_strain+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"right_tube_strain"));
    optional_array_to_floats(mrb, arr, "right_tube_strain", npf-1,npf, model->right_tube_strain+0);

    arr = hash_get_value_opt(mrb, kw, mrb_intern_lit(mrb,"both_tube_strain"));
    optional_array_to_floats(mrb, arr, "both_tube_strain", npf-2,npf, model->both_tube_strain+0);
  }

  mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@driver"), rdriver);

  return self;
}

#define PLUS_END  0
#define MINUS_END 1
#define BOTH_ENDS 2
static mrb_value get_tube_rows(mrb_state *mrb, mrb_value self, int end)
{
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  mrb_value tube_rows;
  int j, npf, *work;
  size_t sz;
  struct tsch_fast_sheet_model_wksp wksp;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  driver = model->driver;
  npf = ts_chem_driver_num_protofilaments(driver);

  sz = tsch_fast_sheet_model_nwork(driver);
  work = mrb_malloc(mrb, sz);
  memset(work,0,sz);

  tsch_fast_sheet_model_setup_wksp_extern(model->driver, work, &wksp); 

  tsch_fast_sheet_model_tip_rows(model, wksp.tips);
  tsch_fast_sheet_model_last_tube_rows(model, wksp.tips, &wksp);

  tube_rows = mrb_ary_new_capa(mrb, npf);
  switch (end)
  {
    case PLUS_END:
      tube_rows = mrb_ary_new_capa(mrb, npf);
      for (j=0; j < npf; j++)
      {
        mrb_ary_push(mrb, tube_rows, mrb_fixnum_value(wksp.last_tubes[j]));
      }
      break;

    case MINUS_END:
      tube_rows = mrb_ary_new_capa(mrb, npf);
      for (j=0; j < npf; j++)
      {
        mrb_ary_push(mrb, tube_rows, mrb_fixnum_value(wksp.last_tubes[npf+j]));
      }
      break;

    case BOTH_ENDS:
      tube_rows = mrb_ary_new_capa(mrb, 2*npf);
      for (j=0; j < npf; j++)
      {
        mrb_value elts[2];
        elts[0] = mrb_fixnum_value(wksp.last_tubes[npf+j]);
        elts[1] = mrb_fixnum_value(wksp.last_tubes[j]);
        mrb_ary_push(mrb, tube_rows,
            mrb_ary_new_from_values(mrb, 2, elts));
      }
      break;

    default:
      tube_rows = mrb_nil_value();
      break;
  }

  mrb_free(mrb, work);

  return tube_rows;
}

static mrb_value cFastSheetModel_tube_last_rows(mrb_state *mrb, mrb_value self)
{
  return get_tube_rows(mrb,self,PLUS_END);
}

static mrb_value cFastSheetModel_tube_first_rows(mrb_state *mrb, mrb_value self)
{
  return get_tube_rows(mrb,self,MINUS_END);
}

static mrb_value get_tips(mrb_state *mrb, mrb_value self, int end)
{
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  mrb_value tip_rows;
  int j, npf, tmp[32], *tips;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  driver = model->driver;
  
  npf = ts_chem_driver_num_protofilaments(driver);
  if (2*npf <= ARRAYLEN(tips))
  {
    tips = tmp+0;
  }
  else
  {
    tips = mrb_malloc(mrb, 2*npf*sizeof(*tips));
  }

  memset(tips,0,2*npf*sizeof(*tips));
  tsch_fast_sheet_model_tip_rows(model, tips);

  switch (end)
  {
    case PLUS_END:
      tip_rows = mrb_ary_new_capa(mrb, npf);
      for (j=0; j < npf; j++)
      {
        mrb_ary_push(mrb, tip_rows, mrb_fixnum_value(tips[j]));
      }
      break;

    case MINUS_END:
      tip_rows = mrb_ary_new_capa(mrb, npf);
      for (j=0; j < npf; j++)
      {
        mrb_ary_push(mrb, tip_rows, mrb_fixnum_value(tips[npf+j]));
      }
      break;

    case BOTH_ENDS:
      tip_rows = mrb_ary_new_capa(mrb, 2*npf);
      for (j=0; j < npf; j++)
      {
        mrb_value elts[2];
        elts[0] = mrb_fixnum_value(tips[npf+j]);
        elts[1] = mrb_fixnum_value(tips[j]);
        mrb_ary_push(mrb, tip_rows,
            mrb_ary_new_from_values(mrb, 2, elts));
      }
      break;

    default:
      tip_rows = mrb_nil_value();
      break;
  }

  if (tips != tmp+0)
  {
    mrb_free(mrb, tips);
  }

  return tip_rows;
}

static mrb_value cFastSheetModel_plus_end_tips(mrb_state *mrb, mrb_value self)
{
  return get_tips(mrb, self, PLUS_END);
}

static mrb_value cFastSheetModel_minus_end_tips(mrb_state *mrb, mrb_value self)
{
  return get_tips(mrb, self, MINUS_END);
}

static mrb_value cFastSheetModel_tip_rows(mrb_state *mrb, mrb_value self)
{
  return cFastSheetModel_plus_end_tips(mrb,self);
}

#undef PLUS_END
#undef MINUS_END
#undef BOTH_ENDS

static mrb_value cFastSheetModel_segments(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  mrb_value segments;
  mrb_int row;
  int j, nsegs, npf;
  int *work, *segs;
  struct tsch_fast_sheet_model_wksp wksp;
  size_t sz;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  driver = model->driver;

  row = 0;
  mrb_get_args(mrb, "i", &row);
  if ((row < 1) || (row > ts_chem_driver_num_dimers(driver)))
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid row %d", row);
  }

  npf = ts_chem_driver_num_protofilaments(driver);

  sz = tsch_fast_sheet_model_nwork(driver);
  work = mrb_malloc(mrb, sz);
  memset(work, 0, sz);

  tsch_fast_sheet_model_setup_wksp_extern(model->driver, work, &wksp); 
  segs = wksp.segs0;

  tsch_fast_sheet_model_tip_rows(model, wksp.tips);
  tsch_fast_sheet_model_last_tube_rows(model, wksp.tips, &wksp);
  nsegs = tsch_fast_sheet_model_row_segments(model, row, &wksp, segs);

  segments = mrb_ary_new_capa(mrb, nsegs);
  for (j=0; j < nsegs; j++)
  {
    int ai;
    mrb_value range[2];

    ai = mrb_gc_arena_save(mrb);
    range[0] = mrb_fixnum_value(segs[j]);
    range[1] = mrb_fixnum_value(segs[npf+j]);
    mrb_ary_push(mrb, segments, mrb_ary_new_from_values(mrb, 2, range));
    mrb_gc_arena_restore(mrb, ai);
  }

  mrb_free(mrb, work);

  return segments;
}

static void seg_intervals_to_c(mrb_state *mrb, mrb_value *rsegs, mrb_int nrsegs, int *restrict segs, int n)
{
  mrb_int i;

  for (i=0; i < nrsegs; i++)
  {
    mrb_value beg,end;
    if (!mrb_array_p(rsegs[i]) || (mrb_ary_len(mrb, rsegs[i]) != 2))
    {
      tsmrb_raisef(mrb, E_TYPE_ERROR,
          "expected segment interval %d to be a two item array",i+1);
    }

    beg = mrb_ary_ref(mrb, rsegs[i], 0);
    end = mrb_ary_ref(mrb, rsegs[i], 1);

    if (!mrb_fixnum_p(beg) || !mrb_fixnum_p(end))
    {
      tsmrb_raisef(mrb, E_TYPE_ERROR,
          "expected segment interval %d to be a two item array of fixnums",i+1);
    }

    segs[i]   = mrb_fixnum(beg);
    segs[i+n] = mrb_fixnum(end);
  }
}

static mrb_value cFastSheetModel_map_segments(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  int npf, *restrict segs0, *restrict segs1, *restrict map;
  mrb_int i, nrsegs0, nrsegs1;
  mrb_value *rsegs0, *rsegs1, ret;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  driver = model->driver;
  npf = ts_chem_driver_num_protofilaments(driver);

  nrsegs0 = nrsegs1 = 0;
  segs0 = segs1 = NULL;
  mrb_get_args(mrb, "aa", &rsegs0,&nrsegs0, &rsegs1,&nrsegs1);

  if (nrsegs0 > npf)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "too many segments in row n (%d given, %d max)",
        (int)nrsegs0, (int)npf);
  }

  if (nrsegs1 > npf)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "too many segments in row n+1 (%d given, %d max)",
        (int)nrsegs1, (int)npf);
  }

  if ((nrsegs0 == 0) || (nrsegs1 == 0))
  {
    return mrb_ary_new(mrb);
  }

  segs0 = model->work+0;
  segs1 = segs0 + 2*npf;
  map   = segs1+2*npf;

  seg_intervals_to_c(mrb, rsegs0, nrsegs0, segs0, npf);
  seg_intervals_to_c(mrb, rsegs1, nrsegs1, segs1, npf);

  tsmrb_handle_error(mrb, 
      tsch_fast_sheet_model_map_segments(model,
        segs0, (int)nrsegs0,
        segs1, (int)nrsegs1,
        map)
      );

  ret = mrb_ary_new_capa(mrb, nrsegs0);
  for (i=0; i < nrsegs0; i++)
  {
    mrb_ary_push(mrb, ret, mrb_fixnum_value(map[i]));
  }

  return ret;
}

static mrb_value cFastSheetModel_interface_energy(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  mrb_int nsheet,ntube,nexisting;

  model = tsmrb_to_fast_sheet_model(mrb,self);

  nsheet = ntube = nexisting = 0;
  mrb_get_args(mrb, "iii", &nsheet, &ntube, &nexisting);

  return mrb_float_value(mrb,
      tsch_fast_sheet_model_ifc_energy(model, ntube,nsheet,nexisting));
}

static mrb_value cFastSheetModel_row_energies(mrb_state *mrb, mrb_value self)
{
  ts_chem_driver *driver;
  struct tsch_fast_sheet_model *model;
  mrb_value ret;
  int j, minus_min, minus_max, plus_min, plus_max;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  driver = model->driver;
  ret = mrb_ary_new_capa(mrb, model->plus_maxrow);

  tsmrb_handle_error(mrb,
      tsch_fast_sheet_model_begin_rowscan(model, TSCH_FSM_MINUS_END)
      );

  minus_min = model->minus_minrow; minus_max = model->minus_maxrow;
  plus_min  = model->plus_minrow;  plus_max  = model->plus_maxrow;

  if (plus_min <= minus_min)
  {
    for (j=0; j < plus_min; j++)
    {
      mrb_ary_push(mrb, ret, mrb_float_value(mrb,0.0));
    }

    goto plus_end;
  }

  for (j=0; j < minus_max; j++)
  {
    mrb_ary_push(mrb, ret, mrb_float_value(mrb,0.0));
  }

  for (j=minus_max; j > minus_min; j--)
  {
    double strain, err;
    strain = err = 0.0;
    tsmrb_handle_error(mrb,
        tsch_fast_sheet_model_next_rows(model, j, 1, -1, &strain, &err)
        );
    mrb_ary_set(mrb, ret, j-1, mrb_float_value(mrb,strain));
  }

  for (j=minus_max; j < plus_min; j++)
  {
    mrb_ary_push(mrb, ret, mrb_float_value(mrb,0.0));
  }

plus_end:
  tsmrb_handle_error(mrb,
      tsch_fast_sheet_model_begin_rowscan(model, TSCH_FSM_PLUS_END)
      );

  for (j=plus_min; j < plus_max; j++)
  {
    double strain, err;
    strain = err = 0.0;
    tsmrb_handle_error(mrb,
        tsch_fast_sheet_model_next_rows(model, j, 1, 1, &strain, &err)
        );
    mrb_ary_push(mrb, ret, mrb_float_value(mrb,strain));
  }

  return ret;
}

static mrb_value cFastSheetModel_energy(mrb_state *mrb, mrb_value self)
{
  double strain;
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);

  strain = 0.0;
  tsmrb_handle_error(mrb, tsch_fast_sheet_model_energy(model, &strain));
  return mrb_float_value(mrb, strain);
}

static mrb_value cFastSheetModel_energy_change(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  mrb_int i,j,st1;
  double dE;

  i = j = st1 = -1;
  mrb_get_args(mrb, "iii", &i,&j,&st1);

  model = tsmrb_to_fast_sheet_model(mrb,self);
  dE = 0.0;
  tsmrb_handle_error(mrb,
      tsch_fast_sheet_model_energy_change(model, i,j,st1, &dE));

  return mrb_float_value(mrb, dE);
}

static mrb_value cFastSheetModel_add_movehookB(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);

  tsch_fast_sheet_model_add_movehook(model);

  return mrb_nil_value();
}

static mrb_value cFastSheetModel_num_hook_calls(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);
  return mrb_fixnum_value(model->num_hook_calls);
}

static mrb_value cFastSheetModel_num_hook_accepts(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);
  return mrb_fixnum_value(model->num_hook_accepts);
}

static mrb_value cFastSheetModel_num_hook_rejects(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);
  return mrb_fixnum_value(model->num_hook_rejects);
}

static mrb_value cFastSheetModel_get_scale(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);
  return mrb_float_value(mrb, model->scale);
}

static mrb_value cFastSheetModel_set_scale(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  mrb_float scale;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  scale = 0.0;
  mrb_get_args(mrb, "f", &scale);
  model->scale = scale;

  return mrb_nil_value();
}

static mrb_value cFastSheetModel_get_minus_copy(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  ts_chem_driver* drv_copy;
  mrb_value copy;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  copy = mrb_iv_get(mrb, self, mrb_intern_lit(mrb,"@minus_copy"));

  drv_copy = (!mrb_nil_p(copy))
    ? tsmrb_to_chem_driver(mrb,copy)
    : NULL
    ;

  if (drv_copy != model->minus_copy)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "FastSheetModel: internal minus_copy does not agree with MRuby minus copy object");
  }

  return copy;
}

static mrb_value cFastSheetModel_set_minus_copy(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  ts_chem_driver* drv_copy;
  mrb_value copy;

  model = tsmrb_to_fast_sheet_model(mrb,self);

  copy = mrb_nil_value();
  mrb_get_args(mrb, "o", &copy);

  drv_copy = (!mrb_nil_p(copy))
    ? tsmrb_to_chem_driver(mrb,copy)
    : NULL
    ;

  model->minus_copy = drv_copy;
  mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@minus_copy"), copy);

  return mrb_nil_value();
}

static mrb_value cFastSheetModel_get_sheet_fluctuation_dist(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);

  return mrb_fixnum_value(model->sheet_fluct_dist);
}

static mrb_value cFastSheetModel_set_sheet_fluctuation_dist(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  mrb_int dist;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  dist = -1;
  mrb_get_args(mrb, "i", &dist);

  model->sheet_fluct_dist = dist;
  return mrb_nil_value();
}


static mrb_value cFastSheetModel_get_sheet_fluctuation_penalty(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  model = tsmrb_to_fast_sheet_model(mrb,self);
  return mrb_float_value(mrb, model->sheet_fluct_penalty);
}

static mrb_value cFastSheetModel_set_sheet_fluctuation_penalty(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;
  mrb_float penalty;

  model = tsmrb_to_fast_sheet_model(mrb,self);

  penalty = 0.0;
  mrb_get_args(mrb, "f", &penalty);

  model->sheet_fluct_penalty = penalty;
  return mrb_nil_value();
}


static mrb_value get_model_values(mrb_state *mrb, struct tsch_fast_sheet_model *model,
    const char *prop, const double *vals)
{
  mrb_int wid, npf;
  mrb_bool has_arg;

  npf = ts_chem_driver_num_protofilaments(model->driver);

  wid = 0;
  has_arg = 0;
  mrb_get_args(mrb, "|i?", &wid, &has_arg);

  if (!has_arg)
  {
    return tsmrb_ary_from_doubles(mrb, (size_t)npf, vals);
  }

  if (wid <= 0 || wid > npf)
  {
    tsmrb_raisef(mrb, E_INDEX_ERROR,
        "width for %s must be in [1, %d], %d given", prop, npf, wid);
  }

  return mrb_float_value(mrb, vals[wid-1]);
}

static mrb_value cFastSheetModel_get_relax_coeff(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  return get_model_values(mrb, model, "relax_coeff", model->relax_coeff);
}

static mrb_value cFastSheetModel_get_no_tube_strain(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  return get_model_values(mrb, model, "no_tube_strain", model->no_tube_strain);
}

static mrb_value cFastSheetModel_get_left_tube_strain(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  return get_model_values(mrb, model, "left_tube_strain", model->left_tube_strain);
}

static mrb_value cFastSheetModel_get_right_tube_strain(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  return get_model_values(mrb, model, "right_tube_strain", model->right_tube_strain);
}

static mrb_value cFastSheetModel_get_both_tube_strain(mrb_state *mrb, mrb_value self)
{
  struct tsch_fast_sheet_model *model;

  model = tsmrb_to_fast_sheet_model(mrb,self);
  return get_model_values(mrb, model, "both_tube_strain", model->both_tube_strain);
}


static void chem_configure_free(mrb_state *mrb, void* arg)
{
  struct tsch_config_parser *p = arg;
  tsch_config_finalize(p);
  mrb_free(mrb,p);
}


static const mrb_data_type chem_configure_type = {
  .struct_name = "ChemConfigure",
  .dfree = chem_configure_free,
};

struct tsch_config_parser *tsmrb_to_config_parser(mrb_state *mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_chem_configure_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a ChemConfigure object");

  return DATA_PTR(obj);
}

static mrb_value cChemConfigure_initialize(mrb_state *mrb, mrb_value self)
{
  TSResult err;
  struct tsch_config_parser *p;
  const char *input;

  input = NULL;
  mrb_get_args(mrb, "|z", &input);

  p = mrb_malloc(mrb, sizeof *p);
  err = tsch_config_initialize(p);
  if (IS_ERROR(err))
  {
    mrb_free(mrb,p);
    tsmrb_handle_error(mrb,err);
  }

  DATA_PTR(self) = p;
  DATA_TYPE(self) = &chem_configure_type;

  if (input)
  {
    struct tsch_config *out = NULL;
    tsmrb_handle_error(mrb, tsch_parse_config(p,&out,input));
  }

  return mrb_nil_value();
}

static mrb_value cChemConfigure_parse(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  struct tsch_config *out;
  const char *input;
  TSResult err;

  p = tsmrb_to_config_parser(mrb,self);

  input = NULL;
  mrb_get_args(mrb, "z", &input);

  out = NULL;
  err = tsch_parse_config(p,&out,input);
  tsmrb_handle_error(mrb, err);

  return mrb_nil_value();
}

static mrb_value cChemConfigure_get_state(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;

  p = tsmrb_to_config_parser(mrb,self);
  return translate_config(mrb, p->head);
}

static mrb_value cChemConfigure_build(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  ts_chem_driver *driver;
  struct RData* driver_obj;
  mrb_value user_config;

  p = tsmrb_to_config_parser(mrb,self);
  driver = NULL;

  tsmrb_handle_error(mrb,
      tsch_config_build_driver(p->head, &driver, ignore_missing_config));

  driver_obj = Data_Wrap_Struct(mrb,
      tsmrb_chem_driver_class(mrb), &chem_driver_type, driver);
  user_config = translate_config(mrb, p->head);

  mrb_obj_iv_set(mrb, (struct RObject *)driver_obj,
      mrb_intern_lit(mrb,"@user_config"), user_config);

  return mrb_obj_value(driver_obj);
}

static mrb_value cChemConfigure_set_params(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  mrb_value params, keys;
  const mrb_value *kv;
  mrb_int i,n;

  p = tsmrb_to_config_parser(mrb,self);

  params = mrb_nil_value();
  mrb_get_args(mrb, "H", &params);

  keys = mrb_hash_keys(mrb, params);
  for (i=0, n=RARRAY_LEN(keys), kv=RARRAY_PTR(keys); i < n; i++)
  {
    mrb_value k, v;
    const char *pname;
    char buf[128];
    double pval;

    if (mrb_string_p(kv[i]))
    {
      k = kv[i];
    }
    else if (mrb_symbol_p(kv[i]))
    {
      k = mrb_sym2str(mrb, mrb_symbol(kv[i]));
    }
    else
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "parameter keys must be strings or symbols");
    }

    v = mrb_hash_get(mrb, params, kv[i]);
    if (mrb_fixnum_p(v))
    {
      pval = (double)mrb_fixnum(v);
    }
    else if (mrb_float_p(v))
    {
      pval = (double)mrb_float(v);
    }
    else
    {
      pval = 0.0;
      mrb_raise(mrb, E_ARGUMENT_ERROR, "parameter values must be numeric");
    }

    pname = mrb_str_to_cstr(mrb, k);
    if (pname[0] != TSCH_CFG_PARAM_PREFIX)
    {
      int ret;
      ret = snprintf(buf, sizeof buf, "%c%s", TSCH_CFG_PARAM_PREFIX, pname);

      if (ret < 0)
      {
        tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
            "error prepending '%c' to parameter name '%s': %s",
            TSCH_CFG_PARAM_PREFIX, pname, strerror(errno));
      }
      else if ((size_t)ret > sizeof buf)
      {
        tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
            "parameter name '%s' is too long to prepend '%c' (exceeds %zu characters)",
            pname, TSCH_CFG_PARAM_PREFIX, sizeof buf - 2);
      }

      pname = buf;
    }

    tsmrb_handle_error(mrb, tsch_config_add_parameter(p, pname, pval));
  }

  return mrb_nil_value();
}

static mrb_value cChemConfigure_get_linenum(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  p = tsmrb_to_config_parser(mrb,self);
  return mrb_fixnum_value((mrb_int)tsch_cfg_lex_last_lineno(&p->lex));
}

static mrb_value cChemConfigure_get_col(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  p = tsmrb_to_config_parser(mrb,self);
  return mrb_fixnum_value((mrb_int)tsch_cfg_lex_last_col(&p->lex));
}

static mrb_value cChemConfigure_get_offset(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  p = tsmrb_to_config_parser(mrb,self);
  return mrb_fixnum_value((mrb_int)p->lex.state[0].pos);
}

static mrb_value cChemConfigure_get_loc(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  mrb_value elts[3];

  p = tsmrb_to_config_parser(mrb,self);

  elts[0] = mrb_fixnum_value((mrb_int)tsch_cfg_lex_last_lineno(&p->lex));
  elts[1] = mrb_fixnum_value((mrb_int)tsch_cfg_lex_last_col(&p->lex));
  elts[2] = mrb_fixnum_value((mrb_int)p->lex.state[0].pos);

  return mrb_ary_new_from_values(mrb, 3, elts);
}

static mrb_value cChemConfigure_get_line(mrb_state *mrb, mrb_value self)
{
  struct tsch_config_parser *p;
  const char *lbeg, *lend;

  p = tsmrb_to_config_parser(mrb,self);

  lbeg = tsch_cfg_lex_last_line(&p->lex);
  lend = strchr(lbeg,'\n');
  if (!lend)
  {
    lend = strchr(lbeg, '\0');
  }

  return mrb_str_new(mrb, lbeg, lend-lbeg);
}

static struct RClass *define_chem_driver_class(mrb_state *mrb, struct RClass *mod);
static struct RClass *define_sheet_length_bias_class(mrb_state *mrb, struct RClass *mod);
static struct RClass *define_fast_sheet_model_class(mrb_state *mrb, struct RClass *mod);
static struct RClass *define_chem_configure_class(mrb_state *mrb, struct RClass *mod);

struct RClass *tsmrb_define_chem_driver(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cChemDriver;

  if (tsmrb_chem_driver_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "ChemDriver class has already been initialized!");

  if (tsmrb_sheet_length_bias_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "SheetLengthBias class has already been initialized!");

  if (tsmrb_fast_sheet_model_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "FastSheetModel class has already been initialized!");

  if (tsmrb_lattice_class(mrb) == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Lattice class has not been initialized!");

  rate_prefactor_table_initialize(mrb);

  cChemDriver = define_chem_driver_class(mrb, mod);
  tsmrb_classes(mrb)->cChemDriver = cChemDriver;

  tsmrb_classes(mrb)->cChemConfigure = define_chem_configure_class(mrb, mod);

  tsmrb_classes(mrb)->cSheetLengthBias = define_sheet_length_bias_class(mrb, mod);
  tsmrb_classes(mrb)->cFastSheetModel  = define_fast_sheet_model_class(mrb, mod);

  return cChemDriver;
}

static struct RClass *define_chem_driver_class(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cChemDriver;

  cChemDriver = mrb_define_class_under(mrb, mod, "ChemDriver", tsmrb_lattice_class(mrb));
  MRB_SET_INSTANCE_TT(cChemDriver, MRB_TT_DATA);
  tsmrb_classes(mrb)->cChemDriver = cChemDriver;

  mrb_define_module_function(mrb, cChemDriver, "spin_value", chem_driver_spin_value, MRB_ARGS_REQ(2));
  mrb_define_module_function(mrb, cChemDriver, "char_value", chem_driver_char_value, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, cChemDriver, "debug_updates", chem_driver_get_debug_updates, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, cChemDriver, "debug_updates=", chem_driver_set_debug_updates, MRB_ARGS_REQ(1));

  mrb_define_const(mrb, cChemDriver, "E_FORBIDDEN", mrb_float_value(mrb, TS_CHEMDRIVER_FORBIDDEN_ENERGY));

  mrb_define_const(mrb, cChemDriver, "EMPTY"                 , mrb_fixnum_value(TS_CHEMDRIVER_STATE_EMPTY     ));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GTP_BENT_LEFT" , mrb_fixnum_value(TS_CHEMDRIVER_STATE_BENT_LEFT ));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GTP_BENT_RIGHT", mrb_fixnum_value(TS_CHEMDRIVER_STATE_BENT_RIGHT));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GTP_CURVED"    , mrb_fixnum_value(TS_CHEMDRIVER_STATE_CURVED    ));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GTP_STRAIGHT"  , mrb_fixnum_value(TS_CHEMDRIVER_STATE_STRAIGHT  ));

  mrb_define_const(mrb, cChemDriver, "MONOMER_GDP_BENT_LEFT" , mrb_fixnum_value(TS_CHEMDRIVER_STATE_BENT_LEFT_GDP ));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GDP_BENT_RIGHT", mrb_fixnum_value(TS_CHEMDRIVER_STATE_BENT_RIGHT_GDP));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GDP_CURVED"    , mrb_fixnum_value(TS_CHEMDRIVER_STATE_CURVED_GDP    ));
  mrb_define_const(mrb, cChemDriver, "MONOMER_GDP_STRAIGHT"  , mrb_fixnum_value(TS_CHEMDRIVER_STATE_STRAIGHT_GDP  ));

  mrb_define_const(mrb, cChemDriver, "PSEUDO_LATERAL_ROTATION", mrb_fixnum_value(PSEUDO_LATERAL_ROTATION));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_LATERAL_ZERO"     , mrb_fixnum_value(PSEUDO_LATERAL_ZERO     ));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_LATERAL_SHEETLIKE", mrb_fixnum_value(PSEUDO_LATERAL_SHEETLIKE));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_LATERAL_TUBELIKE" , mrb_fixnum_value(PSEUDO_LATERAL_TUBELIKE ));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_LATERAL_SEAM"    , mrb_fixnum_value(PSEUDO_LATERAL_SEAM));

  mrb_define_const(mrb, cChemDriver, "PSEUDO_K_ON_BENT_LEFT"     , mrb_fixnum_value(PSEUDO_K_ON_BENT_LEFT));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_K_ON_BENT_RIGHT"    , mrb_fixnum_value(PSEUDO_K_ON_BENT_RIGHT));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_K_ON_STRAIGHT_LEFT" , mrb_fixnum_value(PSEUDO_K_ON_STRAIGHT_LEFT));
  mrb_define_const(mrb, cChemDriver, "PSEUDO_K_ON_STRAIGHT_RIGHT", mrb_fixnum_value(PSEUDO_K_ON_STRAIGHT_RIGHT));

  mrb_define_const(mrb, cChemDriver, "PSEUDO_K_OFF"           , mrb_fixnum_value(PSEUDO_K_OFF));

  mrb_define_const(mrb, cChemDriver, "INTERFACE_BENT"    , mrb_fixnum_value(TS_CHEMDRIVER_STATE_BENT_IFC  ));
  mrb_define_const(mrb, cChemDriver, "INTERFACE_CURVED"  , mrb_fixnum_value(TS_CHEMDRIVER_STATE_CURVED    ));
  mrb_define_const(mrb, cChemDriver, "INTERFACE_STRAIGHT", mrb_fixnum_value(TS_CHEMDRIVER_STATE_STRAIGHT  ));

  mrb_define_const(mrb, cChemDriver, "INTERFACE_TUBELIKE"   , mrb_fixnum_value(TS_CHEMDRIVER_STATE_TUBELIKE  ));
  mrb_define_const(mrb, cChemDriver, "INTERFACE_SHEETLIKE"  , mrb_fixnum_value(TS_CHEMDRIVER_STATE_SHEETLIKE ));

  mrb_define_const(mrb, cChemDriver, "STATE_MAX"  , mrb_fixnum_value(TS_CHEMDRIVER_STATE_MAX));

  mrb_define_method(mrb, cChemDriver, "initialize", cChemDriver_initialize, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "num_dimers", cChemDriver_num_dimers, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "alpha_row" , cChemDriver_alpha_row , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cChemDriver, "beta_row"  , cChemDriver_beta_row  , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "pitch_monomers", cChemDriver_get_pitch_monomers, MRB_ARGS_NONE());

# define DEFINE_FLAG(cls,name) do {                       \
    mrb_define_method(mrb, cls, SYMSTR(name),             \
        SYMCAT(cChemDriver_get_,name), MRB_ARGS_NONE());  \
    mrb_define_method(mrb, cls, SYMSTR(name) "=",         \
        SYMCAT(cChemDriver_set_,name), MRB_ARGS_REQ(1));  \
    } while(0)

  DEFINE_FLAG(cChemDriver, seam_allowed_thresh             );
  DEFINE_FLAG(cChemDriver, reversible_lateral_bond_breaking);
  DEFINE_FLAG(cChemDriver, restrict_curved_lateral_bonding );
  DEFINE_FLAG(cChemDriver, disable_ccs_depolymerization    );

  DEFINE_FLAG(cChemDriver, disable_ccs_depolymerization    );
  DEFINE_FLAG(cChemDriver, minus_end_polymerization        );
  DEFINE_FLAG(cChemDriver, enable_lat_long_bending         );
  DEFINE_FLAG(cChemDriver, allow_depolymerization_at_straight_interfaces);
  DEFINE_FLAG(cChemDriver, disallow_polymerization_at_straight_interfaces);
  DEFINE_FLAG(cChemDriver, polym_only_longitudinal         );
  DEFINE_FLAG(cChemDriver, seam_detailed_balance           );
  DEFINE_FLAG(cChemDriver, polym_metropolis_weighted       );

  DEFINE_FLAG(cChemDriver, restrict_sheet_fluctuations     );

  DEFINE_FLAG(cChemDriver, hydrolysis_only_at_sss          );

# undef DEFINE_FLAG

  mrb_define_method(mrb, cChemDriver, "add_dissociation_barrier",
      cChemDriver_add_dissociation_barrier, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cChemDriver, "get_dissociation_barrier",
      cChemDriver_get_dissociation_barrier, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "[]=",  cChemDriver_set , MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cChemDriver, "barrier_model",  cChemDriver_get_barrier_model, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "barrier_model=", cChemDriver_set_barrier_model, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "barrier_param",  cChemDriver_get_barrier_param, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "barrier_param=", cChemDriver_set_barrier_param, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "prob_acceptance", cChemDriver_prob_acceptance, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "reset!", cChemDriver_resetB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "site_type", cChemDriver_site_type, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cChemDriver, "is_alpha_lateral", cChemDriver_is_alpha_lateral, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cChemDriver, "is_alpha_monomer", cChemDriver_is_alpha_monomer, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "get_rate_prefactors", cChemDriver_get_rate_prefactors, MRB_ARGS_ANY());
  mrb_define_method(mrb, cChemDriver, "set_rate_prefactor", cChemDriver_set_rate_prefactor, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "monomer_energies", cChemDriver_get_monomer_energies, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "set_monomer_energy", cChemDriver_set_monomer_energy, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "set_interface_energy", cChemDriver_set_interface_energy,MRB_ARGS_ANY());
  mrb_define_method(mrb, cChemDriver, "get_interface_energy", cChemDriver_get_interface_energy,MRB_ARGS_ANY());

  mrb_define_method(mrb, cChemDriver, "set_bending_energy", cChemDriver_set_bending_energy,MRB_ARGS_ANY());
  mrb_define_method(mrb, cChemDriver, "get_bending_energy", cChemDriver_get_bending_energy,MRB_ARGS_ANY());

  mrb_define_method(mrb, cChemDriver, "coupling_edge_distance=",
      cChemDriver_set_coupling_edge_distance,MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cChemDriver, "coupling_edge_distance",
      cChemDriver_get_coupling_edge_distance,MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "cached_row_ranges",
      cChemDriver_get_cached_row_ranges, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "update_row_ranges!",
      cChemDriver_update_row_ranges, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "set_coupling_energy", cChemDriver_set_coupling_energy,MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cChemDriver, "get_coupling_energy", cChemDriver_get_coupling_energy,MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "energy", cChemDriver_get_energy, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cChemDriver, "energy_change", cChemDriver_energy_change, MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cChemDriver, "lateral_energy_change", cChemDriver_lateral_energy_change, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cChemDriver, "monomer_populations", cChemDriver_monomer_populations, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "interface_populations", cChemDriver_interface_populations, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "generate_site_moves", cChemDriver_generate_site_moves, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cChemDriver, "change_spin", cChemDriver_change_spin, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cChemDriver, "print_configuration",
      cChemDriver_print_configuration, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cChemDriver, "dt" , cChemDriver_get_dt, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "dt=", cChemDriver_set_dt, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "t_max" , cChemDriver_get_t_max, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "t_max=", cChemDriver_set_t_max, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "forbidden_energy" , cChemDriver_get_forbidden_energy, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "forbidden_energy=", cChemDriver_set_forbidden_energy, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "check_move_consistency!", cChemDriver_check_move_consistencyB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "count_bonds", cChemDriver_count_bonds, MRB_ARGS_ANY());
  mrb_define_method(mrb, cChemDriver, "ifclog", cChemDriver_get_ifclog, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cChemDriver, "clear_ifclog", cChemDriver_clear_ifclog, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "new_ifclog", cChemDriver_new_ifclog, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "ifclog?", cChemDriver_ifclogQ, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "ifclog_size", cChemDriver_ifclog_size, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemDriver, "ifclog_capa", cChemDriver_ifclog_capa, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "ifclog_zero!", cChemDriver_ifclog_zeroB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cChemDriver, "ifclog_unpack_ifc", cChemDriver_ifclog_unpack_ifc, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cChemDriver, "ifclog_add", cChemDriver_ifclog_add, MRB_ARGS_REQ(4));

  mrb_define_method(mrb, cChemDriver, "ifclog_write", cChemDriver_ifclog_write, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cChemDriver, "ifclog_read", cChemDriver_ifclog_read, MRB_ARGS_REQ(1));


  mrb_define_method(mrb, cChemDriver, "read_state", cChemDriver_read_state, MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cChemDriver, "write_state", cChemDriver_write_state, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cChemDriver, "take_snapshot", cChemDriver_take_snapshot, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "read_state_header",
      cChemDriver_read_state_header, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "from_yaml",
      chem_driver_from_yaml, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "ignore_missing_config",
      cChemDriver_get_ignore_missing_config, MRB_ARGS_NONE());
  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "ignore_missing_config=",
      cChemDriver_set_ignore_missing_config, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "bare_configure",
      cChemDriver_bare_configure, MRB_ARGS_REQ(1) | MRB_ARGS_OPT(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cChemDriver, "configure",
      cChemDriver_configure, MRB_ARGS_REQ(1));

  mrb_define_class_under(mrb, cChemDriver, "Snapshot", mrb->object_class);

  return cChemDriver;
}

static struct RClass *define_sheet_length_bias_class(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cSheetLengthBias;

  cSheetLengthBias = mrb_define_class_under(mrb, mod, "SheetLengthBias", tsmrb_lattice_class(mrb));
  MRB_SET_INSTANCE_TT(cSheetLengthBias, MRB_TT_DATA);

  mrb_define_method(mrb, cSheetLengthBias, "initialize", cSheetLengthBias_initialize, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cSheetLengthBias, "potential", cSheetLengthBias_get_potential, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSheetLengthBias, "scale", cSheetLengthBias_get_scale, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSheetLengthBias, "scale=", cSheetLengthBias_set_scale, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSheetLengthBias, "sheet_length", cSheetLengthBias_sheet_length, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSheetLengthBias, "sheet_energy", cSheetLengthBias_sheet_energy, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSheetLengthBias, "energy_change", cSheetLengthBias_energy_change, MRB_ARGS_REQ(3));

  return cSheetLengthBias;
}

static struct RClass *define_fast_sheet_model_class(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cFastSheetModel;

  cFastSheetModel = mrb_define_class_under(mrb, mod, "FastSheetModel", mrb->object_class);
  MRB_SET_INSTANCE_TT(cFastSheetModel, MRB_TT_DATA);

  mrb_define_method(mrb, cFastSheetModel, "initialize", cFastSheetModel_initialize,
      MRB_ARGS_REQ(1) | MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cFastSheetModel, "tube_last_rows", cFastSheetModel_tube_last_rows, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "tube_first_rows", cFastSheetModel_tube_first_rows, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "tip_rows", cFastSheetModel_tip_rows, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "plus_end_tips", cFastSheetModel_plus_end_tips, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "minus_end_tips", cFastSheetModel_minus_end_tips, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFastSheetModel, "segments", cFastSheetModel_segments, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cFastSheetModel, "map_segments", cFastSheetModel_map_segments, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cFastSheetModel, "interface_energy", cFastSheetModel_interface_energy, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cFastSheetModel, "row_energies", cFastSheetModel_row_energies, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFastSheetModel, "relax_coeff", cFastSheetModel_get_relax_coeff, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cFastSheetModel, "no_tube_strain", cFastSheetModel_get_no_tube_strain, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cFastSheetModel, "left_tube_strain", cFastSheetModel_get_left_tube_strain, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cFastSheetModel, "right_tube_strain", cFastSheetModel_get_right_tube_strain, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cFastSheetModel, "both_tube_strain", cFastSheetModel_get_both_tube_strain, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cFastSheetModel, "energy", cFastSheetModel_energy, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "energy_change", cFastSheetModel_energy_change, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cFastSheetModel, "add_movehook!", cFastSheetModel_add_movehookB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFastSheetModel, "num_hook_calls"  , cFastSheetModel_num_hook_calls  , MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "num_hook_accepts", cFastSheetModel_num_hook_accepts, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "num_hook_rejects", cFastSheetModel_num_hook_rejects, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFastSheetModel, "scale", cFastSheetModel_get_scale, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "scale=", cFastSheetModel_set_scale, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFastSheetModel, "minus_copy", cFastSheetModel_get_minus_copy, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "minus_copy=", cFastSheetModel_set_minus_copy, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFastSheetModel, "sheet_fluctuation_dist",
      cFastSheetModel_get_sheet_fluctuation_dist, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "sheet_fluctuation_dist=",
      cFastSheetModel_set_sheet_fluctuation_dist, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFastSheetModel, "sheet_fluctuation_penalty",
      cFastSheetModel_get_sheet_fluctuation_penalty, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFastSheetModel, "sheet_fluctuation_penalty=",
      cFastSheetModel_set_sheet_fluctuation_penalty, MRB_ARGS_REQ(1));

  return cFastSheetModel;
}

static struct RClass *define_chem_configure_class(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cChemConfigure;

  cChemConfigure = mrb_define_class_under(mrb, mod, "ChemConfigure", mrb->object_class);
  MRB_SET_INSTANCE_TT(cChemConfigure, MRB_TT_DATA);

  mrb_define_method(mrb, cChemConfigure, "initialize", cChemConfigure_initialize, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cChemConfigure, "parse", cChemConfigure_parse, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cChemConfigure, "state", cChemConfigure_get_state, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "build", cChemConfigure_build, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "set_params", cChemConfigure_set_params, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cChemConfigure, "linenum", cChemConfigure_get_linenum, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "col", cChemConfigure_get_col, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "offset", cChemConfigure_get_offset, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "loc", cChemConfigure_get_loc, MRB_ARGS_NONE());
  mrb_define_method(mrb, cChemConfigure, "line", cChemConfigure_get_line, MRB_ARGS_NONE());

  return cChemConfigure;
}


