#include "tsmrb_dispatch.h"
#include "tsmrb_helpers.h"

#include "ts_dispatch.h"

#define TSMRB_DISPATCH_INIT_FROM_ENV

static mrb_value
mDispatch_enabled_p(mrb_state* mrb, mrb_value mod)
{
  int enabled = 0;
#ifdef TS_DISPATCH_ENABLED
  enabled = 1;
#endif /* 0 */
  return mrb_bool_value(enabled);
}

#if defined(TS_DISPATCH_ENABLED)

static mrb_value mDispatch_initialize(mrb_state* mrb, mrb_value mod)
{
  mrb_int nthread = 0;
  mrb_get_args(mrb, "i", &nthread);

  tsmrb_handle_error(mrb, ts_dispatch_initialize(nthread));

  return mrb_nil_value();
}

static mrb_value mDispatch_initialized_p(mrb_state* mrb, mrb_value mod)
{
  return mrb_bool_value( ts_dispatch_is_initialized() );
}

static mrb_value mDispatch_shutdown(mrb_state* mrb, mrb_value mod)
{
  tsmrb_handle_error(mrb, ts_dispatch_shutdown() );
  return mrb_nil_value();
}

#define DEFAULT_SHUTDOWN_SECS 5
static mrb_value mDispatch_shutdownB(mrb_state* mrb, mrb_value mod)
{
  mrb_int nsec = DEFAULT_SHUTDOWN_SECS;
  mrb_get_args(mrb, "|i",&nsec);

  tsmrb_handle_error(mrb, ts_dispatch_force_shutdown(nsec) );
  return mrb_nil_value();
}

static mrb_value mDispatch_num_threads(mrb_state* mrb, mrb_value mod)
{
  return mrb_fixnum_value( ts_dispatch_num_threads() );
}

static mrb_value mDispatch_active_p(mrb_state* mrb, mrb_value mod)
{
  return mrb_bool_value( ts_dispatch_is_active() );
}

static mrb_value mDispatch_set_active(mrb_state* mrb, mrb_value mod)
{
  mrb_bool tf;
  mrb_get_args(mrb, "b", &tf);
  int old = ts_dispatch_is_active();
  ts_dispatch_set_active( tf );
  return mrb_bool_value(old);
}

static mrb_value mDispatch_foreach_primed_p(mrb_state *mrb, mrb_value mod)
{
  return mrb_bool_value( ts_dispatch_foreach_is_primed() );
}

static mrb_value mDispatch_get_foreach_latency(mrb_state *mrb, mrb_value mod)
{
  return mrb_fixnum_value(ts_dispatch_foreach_latency());
}

static mrb_value mDispatch_set_foreach_latency(mrb_state *mrb, mrb_value mod)
{
  mrb_int lat;

  lat = 0;
  mrb_get_args(mrb, "i", &lat);

  if (lat < 0)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "latency must be non-negative");
  }

  ts_dispatch_foreach_set_latency(lat);
  return mrb_nil_value();
}

#  ifdef TSMRB_DISPATCH_INIT_FROM_ENV
static void init_ignore_cpuset(mrb_state *mrb)
{
  const char *val;
  char *end;
  long ignore_cpuset;

  (void)mrb;

  val = getenv("TSD_IGNORE_CPUSET");
  if (!val || (val[0] == '\0')) {
    /* TS_IGNORE_CPUSET is not set or is empty */
    return;
  }

  end = NULL;
  ignore_cpuset = strtol(val, &end, 0);
  if (!end || *end != '\0') {
    /* invalid value for TS_IGNORE_CPUSET */
    return;
  }

  tsd_ignore_cpuset = !!ignore_cpuset;
}

static void init_dispatch_from_env(mrb_state *mrb)
{
  TSResult err;
  const char *val;
  char *end, buf[128];
  int nthreads;

  init_ignore_cpuset(mrb);

  val = getenv("TSD_NUM_THREADS");
  if ((val == NULL) || (val[0] == '\0')) return;

  // get ENV object from Ruby
  end = NULL;
  nthreads = strtol(val, &end, 0);

  if (end && end[0]) {
    /* TS_NUM_THREADS is not an integer */
    return;
  }

  memset(buf, 0, sizeof(buf));
  if (err = ts_dispatch_initialize(nthreads), !IS_OKAY(err))
  {
    snprintf(buf,sizeof(buf)-1,
        "---[ TS_DISPATCH could not be initialized with num_workers = %d ]---\n"
        "     error code 0x%04x : %s\n",
        nthreads, ERRCODE(err), ts_error_get_state(err)->error_message);
  }
  else
  {
    snprintf(buf,sizeof(buf)-1,
        "---[ TS_DISPATCH initialized with num_workers = %d ]---\n", 
        nthreads);
  }

  mrb_funcall(mrb,
      mrb_obj_value(mrb->object_class), "puts", 1,
      mrb_str_new_cstr(mrb,buf));

}

#  else  /* !TSMRB_DISPATCH_INIT_FROM_ENV */

static void init_dispatch_from_env(mrb_state* mrb)
{
  (void)mrb;
}

#  endif /* TSMRB_DISPATCH_INIT_FROM_ENV */

extern void tsmrb_define_dispatch_tests(mrb_state* mrb, struct RClass* mod);

#endif /* defined(TS_DISPATCH_ENABLED) */


struct RClass* tsmrb_define_dispatch(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* mDispatch = mrb_define_module_under(mrb, mod, "Dispatch");
  mrb_define_module_function(mrb, mDispatch, "enabled?", mDispatch_enabled_p, MRB_ARGS_NONE());

#if defined(TS_DISPATCH_ENABLED)
  mrb_define_module_function(mrb, mDispatch, "initialized?", mDispatch_initialized_p, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mDispatch, "initialize",   mDispatch_initialize, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mDispatch, "shutdown",     mDispatch_shutdown, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mDispatch, "shutdown!",    mDispatch_shutdownB, MRB_ARGS_OPT(1));
  mrb_define_module_function(mrb, mDispatch, "num_threads",  mDispatch_num_threads, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mDispatch, "active?", mDispatch_active_p, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mDispatch, "active=", mDispatch_set_active, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mDispatch, "foreach_primed?", mDispatch_foreach_primed_p, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mDispatch, "foreach_latency", mDispatch_get_foreach_latency, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mDispatch, "foreach_latency=", mDispatch_set_foreach_latency, MRB_ARGS_REQ(1));

  init_dispatch_from_env(mrb);

  tsmrb_define_dispatch_tests(mrb, mDispatch);
#endif /* defined(TS_DISPATCH_ENABLED) */


  return mDispatch;
}

