#include "mruby.h"

#include "tsmrb_helpers.h"

#include "ts_dispatch.h"

#if defined(TS_DISPATCH_ENABLED)

/* Test cases */
static void test_syncpoint_task(void* data)
{
  struct tsd_syncpoint* sp = data;
  ts_dispatch_usleep(200);
  tsd_syncpoint_wait(sp);
}

static mrb_value mTests_test_syncpoint(mrb_state* mrb, mrb_value mod)
{
  int num_threads = ts_dispatch_num_threads();
  mrbtAssert( num_threads > 0 );
  ts_dispatch_stop();

  struct tsd_syncpoint* sp = tsd_syncpoint_new();
  mrbtAssert( sp != NULL );
  mrbtAssert( tsd_syncpoint_count(sp) == 0 );

  for (int i=0; i < num_threads; i++)
  {
    ts_dispatch_add_task(test_syncpoint_task,sp);
  }

  mrbtAssert( tsd_syncpoint_count(sp) == 0 );
  ts_dispatch_start();

  TSD_WAIT_FOR( tsd_syncpoint_count(sp) == num_threads );
  mrbtAssert( tsd_syncpoint_count(sp) == num_threads );
  mrbtAssert( ts_dispatch_num_running() == num_threads );

  tsd_syncpoint_trigger(sp);
  ts_dispatch_sync();

  mrbtAssert( tsd_syncpoint_count(sp) == 0 );

  ts_dispatch_stop();
  tsd_syncpoint_delete(sp);

  return mrb_nil_value();
}

static void counter_task(void* data)
{
  tsd_atomic_s32* count = data;
  // int thr_ind = ts_dispatch_thread_index();
  // fprintf(stderr, "counter_task(%d)\n",thr_ind);
  ts_dispatch_usleep(10);
  MAYBE_UNUSED int32_t prev = tsd_atomic_s32_incr(count);
  // fprintf(stderr, "counter_task(%d) %d\n", 
  //    thr_ind, (int)prev);
}

static mrb_value mTests_test_queue_start_stop(mrb_state* mrb, mrb_value mod)
{
  int num_threads = ts_dispatch_num_threads();
  mrbtAssert( num_threads > 0 );

  tsd_atomic_s32 counter;
  tsd_atomic_s32_init(&counter, 0);

  // fprintf(stderr,"\n\n\n");

  ts_dispatch_stop();
  for (int i=0; i < 2*num_threads; i++)
  {
    ts_dispatch_add_task(counter_task, (void*)&counter);
  }

  ts_dispatch_usleep(100);
  mrbtAssert( tsd_atomic_s32_get(&counter) == 0 );
  mrbtAssert( 2*num_threads == ts_dispatch_num_queued() );
  ts_dispatch_start();

  ts_dispatch_sync();
  // TSD_WAIT_FOR( ts_dispatch_num_queued()  == 0 );
  // TSD_WAIT_FOR( ts_dispatch_num_running() == 0 );

  mrbtAssert( 0 == ts_dispatch_num_running() );
  mrbtAssert( 0 == ts_dispatch_num_queued() );
  // fprintf(stderr, "num_threads = %d  counter = %d\n",
  //     (int)num_threads, (int)tsd_atomic_s32_get(&counter));
  mrbtAssert( tsd_atomic_s32_get(&counter) == 2*num_threads );

  // fprintf(stderr,"\n\n\n");

  ts_dispatch_stop();
  for (int i=0; i < 2*num_threads; i++)
  {
    ts_dispatch_add_task(counter_task, &counter);
  }

  ts_dispatch_usleep(100);
  mrbtAssert( tsd_atomic_s32_get(&counter) == 2*num_threads );
  mrbtAssert( 2*num_threads == ts_dispatch_num_queued() );
  ts_dispatch_start();

  TSD_WAIT_FOR( ts_dispatch_num_queued()  == 0 );
  TSD_WAIT_FOR( ts_dispatch_num_running() == 0 );

  mrbtAssert( tsd_atomic_s32_get(&counter) == 4*num_threads );
  mrbtAssert( 0 == ts_dispatch_num_queued() );

  ts_dispatch_stop();
  tsd_atomic_s32_finalize(&counter);

  return mrb_nil_value();
}

static mrb_value mTests_test_task_queue_wraps_correctly(mrb_state* mrb, mrb_value mod)
{
  int num_threads = ts_dispatch_num_threads();
  mrbtAssert( num_threads > 0 );

  tsd_atomic_s32 counter;
  tsd_atomic_s32_init(&counter, 0);

  ts_dispatch_stop();
  for (int i=0; i < num_threads; i++)
  {
    tsmrb_handle_error(mrb, 
        ts_dispatch_add_task(counter_task, &counter)
        );
  }

  mrbtAssert( tsd_atomic_s32_get(&counter) == 0 );
  mrbtAssert( num_threads == ts_dispatch_num_queued() );

  ts_dispatch_sync();

  mrbtAssert( tsd_atomic_s32_get(&counter) == num_threads );

  for (int i=0; i < TSD_WORKQ_MAX; i++)
  {
    tsmrb_handle_error(mrb, 
        ts_dispatch_add_task(counter_task, &counter)
        );
  }

  ts_dispatch_sync();

  mrbtAssert( 0 == ts_dispatch_num_running() );

  int32_t c = tsd_atomic_s32_get(&counter);
  mrbtAssert( c == TSD_WORKQ_MAX+num_threads );

  tsd_atomic_s32_finalize(&counter);

  return mrb_nil_value();
}

static void test_joinpoint_task(void* v)
{
  struct tsd_joinpoint* jp = v;
  tsd_joinpoint_join(jp);
}

static mrb_value mTests_test_joinpoint(mrb_state* mrb, mrb_value mod)
{
  int num_threads = ts_dispatch_num_threads();
  mrbtAssert( num_threads > 0 );

  struct tsd_joinpoint* jp = tsd_joinpoint_new(num_threads);

  ts_dispatch_stop();
  for (int i=0; i < num_threads-1; i++)
  {
    tsmrb_handle_error(mrb, 
        ts_dispatch_add_task(test_joinpoint_task, jp)
        );
  }

  mrbtAssert( 0 == ts_dispatch_num_running() );
  ts_dispatch_start();

  TSD_WAIT_FOR( num_threads-1 == ts_dispatch_num_running() );
  mrbtAssert( num_threads-1 == ts_dispatch_num_running() );
  mrbtAssert( 0 == ts_dispatch_num_queued() );

  TSD_WAIT_FOR( num_threads-1 == tsd_joinpoint_num_waiting(jp) );
  int nw = tsd_joinpoint_num_woken(jp);
  // fprintf(stderr, "\nnum_woken is %d\n\n",nw);
  mrbtAssert( 0 == nw );

  tsmrb_handle_error(mrb, 
      ts_dispatch_add_task(test_joinpoint_task, jp)
      );

  ts_dispatch_sync();

  mrbtAssert( 0 == ts_dispatch_num_running() );

  mrbtAssert( num_threads == tsd_joinpoint_num_woken(jp) );
  mrbtAssert( 0 == tsd_joinpoint_num_waiting(jp) );

  tsd_joinpoint_delete(jp);

  return mrb_nil_value();
}

static mrb_value mTests_test_dispatch_sync(mrb_state* mrb, mrb_value mod)
{
  int num_threads = ts_dispatch_num_threads();
  mrbtAssert( num_threads > 0 );

  static tsd_atomic_s32 counter;
  tsd_atomic_s32_init(&counter, 0);

  ts_dispatch_stop();
  mrbtAssert( 0 == ts_dispatch_num_running() );
  mrbtAssert( 0 == ts_dispatch_num_queued() );

  for (int i=0; i < num_threads; i++)
  {
    ts_dispatch_add_task(counter_task, &counter);
  }

  ts_dispatch_start();

  tsmrb_handle_error(mrb, 
      ts_dispatch_sync()
      );

  int32_t c = tsd_atomic_s32_get(&counter);
  if (c != num_threads) {
    fprintf(stderr, "c != num_threads: c %d  num_threads %d\n",
        c,num_threads);
  }
  mrbtAssert( tsd_atomic_s32_get(&counter) == num_threads );

  tsd_atomic_s32_finalize(&counter);

  return mrb_nil_value();
}

static TSResult fill_values(void* data, size_t i0, size_t i1, size_t worker, size_t nworkers)
{
  double*restrict p = data;
  for(size_t i = i0; i < i1; i++) {
    p[i] = 2*i+1;
  }
  return TS_OK;
}

static mrb_value mTests_test_foreach(mrb_state* mrb, mrb_value mod)
{
  double values[128] = {0};

  mrbtAssert( 0 == ts_dispatch_num_running() );
  tsmrb_handle_error(mrb, ts_dispatch_foreach(128,fill_values,values+0) );
  mrbtAssert( 0 == ts_dispatch_num_running() );

  for(size_t i=0; i < 128; i++) {
    mrbtAssert( 2*i+1 == values[i] );
  }

  return mrb_nil_value();
}

struct spinlock_data {
  struct tsd_spinlock lock;
  int niter;
  double data[3];
};

static void spinlock_task(void* arg)
{
  struct spinlock_data* data = arg;
  for(int i=0; i < data->niter; i++) {
    tsd_spinlock_lock(&data->lock);
    data->data[0] += 1.0;
    data->data[1] += 1.2;
    data->data[2] += 0.3;
    tsd_spinlock_unlock(&data->lock);
  }
}

static mrb_value mTests_test_spinlock(mrb_state* mrb, mrb_value mod)
{
  struct spinlock_data data;
  data.niter = 10;
  memset(data.data+0,0,sizeof(data.data));
  tsd_spinlock_init(&data.lock);

  int num_threads = ts_dispatch_num_threads();
  double expected[3] = { 0.0 };
  expected[0] = num_threads*10.0;
  expected[1] = num_threads*12.0;
  expected[2] = num_threads*3.0;

  mrbtAssert( 0 == ts_dispatch_num_running() );
  mrbtAssert( 0 == ts_dispatch_num_queued() );

  ts_dispatch_stop();
  for (int i=0; i < num_threads; i++) {
    tsmrb_handle_error(mrb, ts_dispatch_add_task(spinlock_task, &data) );
  }

  ts_dispatch_sync();
  mrbtAssert( 0 == ts_dispatch_num_running() );
  mrbtAssert( 0 == ts_dispatch_num_queued() );

  mrbtAssert( fabs(data.data[0] - num_threads*10.0) < 1e-8 );
  mrbtAssert( fabs(data.data[1] - num_threads*12.0) < 1e-8 );
  mrbtAssert( fabs(data.data[2] - num_threads*3.0 ) < 1e-8 );

  for (int i=0; i < 3; i++) {
    if (fabs(data.data[i] - expected[i]) >= 1e-8) {
      fprintf(stderr, "data.data[0] = %12.8f  expected %12.8f\n",
          data.data[i], expected[i]);
    }
  }

  mrbtAssert( fabs(data.data[0] - expected[0]) < 1e-8 );
  mrbtAssert( fabs(data.data[1] - expected[1]) < 1e-8 );
  mrbtAssert( fabs(data.data[2] - expected[2]) < 1e-8 );

  return mrb_nil_value();
}

#endif /* defined(TS_DISPATCH_ENABLED) */

static inline void add_test(mrb_state* mrb, struct RClass* suite,
    const char* name, mrb_value(*testfunc)(mrb_state* mrb, mrb_value))
{
  mrb_define_module_function(mrb, suite, name, testfunc, MRB_ARGS_NONE());
}

void tsmrb_define_dispatch_tests(mrb_state* mrb, struct RClass* mod)
{
#if defined(TS_DISPATCH_ENABLED)

  struct RClass* mTests = mrb_define_module_under(mrb, mod, "Tests");

  add_test(mrb, mTests, "test_syncpoint"       , mTests_test_syncpoint);
  add_test(mrb, mTests, "test_queue_start_stop", mTests_test_queue_start_stop);
  add_test(mrb, mTests, "test_joinpoint"       , mTests_test_joinpoint);
  add_test(mrb, mTests, "test_dispatch_sync"   , mTests_test_dispatch_sync);
  add_test(mrb, mTests, "test_foreach"         , mTests_test_foreach);
  add_test(mrb, mTests, "test_spinlock"        , mTests_test_spinlock);

  add_test(mrb, mTests, "test_task_queue_wraps_correctly",
      mTests_test_task_queue_wraps_correctly);

#endif /* defined(TS_DISPATCH_ENABLED) */
}



