#include "tsmrb_fenwick.h"

#include "tsmrb_helpers.h"

#include "mruby/data.h"
#include "mruby/variable.h"

#include <stdio.h>

static void free_fenwick(mrb_state *mrb, void *p)
{
  ts_fenwick_tree *tree = p;
  if (tree) { ts_fenwick_tree_finalize(tree); }

  mrb_free(mrb, p);
}

static const mrb_data_type fenwick_tree_type = {
  .struct_name = "FenwickTree",
  .dfree = free_fenwick, // tsmrb_free,
};

#if 0
ts_fenwick_tree* tsrb_to_fenwick_tree(VALUE obj)
{
  ts_fenwick_tree* tree = rbproxy_table_get_data(fenwick_table, obj);
  return tree;
}
#endif /* 0 */

static inline mrb_sym fenwick_elements(mrb_state* mrb)
{
  return mrb_intern_cstr(mrb,"@elements");
}

static mrb_value fenwick_value_get(mrb_state* mrb, mrb_value self, uintptr_t i0)
{
  mrb_int i = (mrb_int) i0;
  mrb_value arr = mrb_iv_get(mrb, self, fenwick_elements(mrb));
  return mrb_ary_ref(mrb, arr, i-1);
}

static void fenwick_value_set(mrb_state* mrb, mrb_value self, uintptr_t i0, mrb_value elt)
{
  mrb_int i = (mrb_int) i0;
  mrb_value arr = mrb_iv_get(mrb, self, fenwick_elements(mrb));
  mrb_ary_set(mrb, arr, i-1, elt);
}

static mrb_int fenwick_value_add(mrb_state* mrb, mrb_value self, mrb_value elt)
{
  mrb_value arr = mrb_iv_get(mrb, self, fenwick_elements(mrb));
  mrb_ary_push(mrb, arr, elt);
  return mrb_ary_len(mrb, arr);
}

static mrb_int fenwick_value_find(mrb_state* mrb, mrb_value self, mrb_value elt0)
{
  mrb_value arr = mrb_iv_get(mrb, self, fenwick_elements(mrb));
  mrb_int n = mrb_ary_len(mrb,arr);
  for(mrb_int i=0; i < n; i++)
  {
    mrb_value elt = mrb_ary_ref(mrb,arr,i);
    if (mrb_obj_equal(mrb,elt,elt0))
      return i+1;
  }
  return 0;
}

ts_fenwick_tree* tsmrb_to_fenwick_tree(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_fenwick_tree_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a FenwickTree object");

  return DATA_PTR(obj);
}

static mrb_value cFenwickTree_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_int sz;
  mrb_get_args(mrb, "i", &sz);

  ts_fenwick_tree* tree = mrb_malloc(mrb, sizeof(*tree));
  memset(tree, 0, sizeof(*tree));

  DATA_PTR(self) = tree;
  DATA_TYPE(self) = &fenwick_tree_type;

  tsmrb_handle_error(mrb,
      ts_fenwick_tree_init(tree, (size_t)sz));

  mrb_value elt_arr = mrb_ary_new_capa(mrb,sz);
  mrb_iv_set(mrb, self, fenwick_elements(mrb), elt_arr);

  return self;
}

static mrb_value cFenwickTree_get_size(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  return mrb_fixnum_value(tree->size);
}

static mrb_value cFenwickTree_get_capacity(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  return mrb_fixnum_value(tree->capacity);
}

static mrb_value cFenwickTree_get_size_2n(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  return mrb_fixnum_value(tree->size_2n);
}

static mrb_value cFenwickTree_get_total_sum(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  return mrb_float_value(mrb,tree->total_sum);
}

static mrb_value cFenwickTree_get_cum_sum(mrb_state* mrb, mrb_value self)
{
  mrb_int ind;
  mrb_get_args(mrb, "i", &ind);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);

  double cum_sum = 0.0;
  tsmrb_handle_error(mrb, 
    ts_fenwick_tree_get_cum_sum(tree, ind, &cum_sum) );
  return mrb_float_value(mrb,cum_sum);
}

static mrb_value cFenwickTree_add(mrb_state* mrb, mrb_value self)
{
  mrb_float freq;
  mrb_value data;
  mrb_get_args(mrb, "fo", &freq, &data);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);

  intptr_t ind = fenwick_value_add(mrb, self, data);

  tsmrb_handle_error(mrb,
      ts_fenwick_tree_add(tree, freq, (void*)ind));

  return self;
}

static mrb_value cFenwickTree_get_element(mrb_state* mrb, mrb_value self)
{
  mrb_int ind;
  mrb_get_args(mrb, "i", &ind);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);

  double freq = 0.0;
  tsmrb_handle_error(mrb, 
    ts_fenwick_tree_get_frequency(tree, ind, &freq));
  intptr_t data_ind = (intptr_t) tree->data[ind];// elements[ind].data;

  mrb_value pair[2] = {
    mrb_float_value(mrb, freq),
    fenwick_value_get(mrb, self, data_ind),
  };
  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value cFenwickTree_get_raw_count(mrb_state* mrb, mrb_value self)
{
  mrb_int ind;
  mrb_get_args(mrb, "i", &ind);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);

  if (ind >= tree->size)
    mrb_raise(mrb, E_RANGE_ERROR, "Invalid index");

  return mrb_float_value(mrb, tree->counts[ind]); // tree->elements[ ind ].count);
}

static mrb_value cFenwickTree_dump(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree *tree = tsmrb_to_fenwick_tree(mrb,self);
  fprintf(stderr, "\n-- dump of Fenwick tree --\n");
  fprintf(stderr, "size:     %lu\n", tree->size);
  fprintf(stderr, "capacity: %lu\n", tree->capacity);
  fprintf(stderr, "\n");
  size_t ind = tree->size - 1;
  while (ind > 0)
  {
    fprintf(stderr, "[%3zu] count %7.4f  %p\n",
        // ind, tree->elements[ind].count, (long)tree->elements[ind].data);
        ind, tree->counts[ind], tree->data[ind]);
    ind--;
  }

  fprintf(stderr, "[%3zu] count %7.4f  %p\n",
      (size_t)0, tree->counts[0], tree->data[0]);
  return self;
}

static mrb_value cFenwickTree_resize(mrb_state* mrb, mrb_value self)
{
  mrb_int sz;
  mrb_get_args(mrb, "i", &sz);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  tsmrb_handle_error(mrb,
      ts_fenwick_tree_resize(tree, sz) );
  return self;
}

static mrb_value cFenwickTree_clearB(mrb_state* mrb, mrb_value self)
{
  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  tsmrb_handle_error(mrb,
      ts_fenwick_tree_clear(tree, NULL, NULL) );
  return self;
}

static mrb_value cFenwickTree_update_count(mrb_state* mrb, mrb_value self)
{
  mrb_int ind;
  mrb_float count;
  mrb_get_args(mrb, "if", &ind, &count);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  tsmrb_handle_error(mrb, 
    ts_fenwick_tree_update_count(tree, ind, count));
  return self;
}

static mrb_value cFenwickTree_remove_at(mrb_state* mrb, mrb_value self)
{
  mrb_int ind;
  mrb_get_args(mrb, "i", &ind);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  if ((ind < 0) || (ind >= tree->size)) {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "invalid index %d (must be in 0...%zu)", tree->size);
  }
  intptr_t data_ind = (intptr_t) tree->data[ind];
  tsmrb_handle_error(mrb,
      ts_fenwick_tree_remove_at(tree, ind));

  fenwick_value_set(mrb, self, data_ind, mrb_nil_value());

  return self;
}

static mrb_value cFenwickTree_replace(mrb_state* mrb, mrb_value self)
{
  mrb_int ind = 0;
  mrb_float val = 0.0;
  mrb_value data = mrb_nil_value();
  mrb_get_args(mrb,"ifo", &ind, &val, &data);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  intptr_t data_ind = (intptr_t) tree->data[ind];
  fenwick_value_set(mrb, self, data_ind, data);

  tsmrb_handle_error(mrb,
      ts_fenwick_tree_replace_at(tree,ind,val,(void*)data_ind));
  return self;
}

static mrb_value cFenwickTree_find_cum_sum(mrb_state* mrb, mrb_value self)
{
  mrb_float sum = 0.0;
  mrb_get_args(mrb, "f", &sum);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  size_t ind = ts_fenwick_tree_find_cum_sum(tree, sum);

  if (ind < tree->size)
    return mrb_fixnum_value(ind);
  else
    return mrb_nil_value();
}

static mrb_value cFenwickTree_find(mrb_state* mrb, mrb_value self)
{
  mrb_value data = mrb_nil_value();
  mrb_get_args(mrb, "o", &data);

  ts_fenwick_tree* tree = tsmrb_to_fenwick_tree(mrb,self);
  mrb_int arr_ind = fenwick_value_find(mrb, self, data);
  if (arr_ind == 0)
    return mrb_nil_value();

  size_t ind = ts_fenwick_tree_find(tree, (void*)((uintptr_t)arr_ind));
  if (ind < tree->size)
    return mrb_fixnum_value(ind);
  else
    return mrb_nil_value();
}

struct RClass* tsmrb_define_fenwick_tree(mrb_state* mrb)
{
  struct RClass* cFenwickTree = mrb_define_class(mrb,
      "FenwickTree", mrb->object_class);
  MRB_SET_INSTANCE_TT(cFenwickTree, MRB_TT_DATA);
  tsmrb_classes(mrb)->cFenwickTree = cFenwickTree;

  mrb_define_method(mrb, cFenwickTree, "initialize", cFenwickTree_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFenwickTree, "clear!", cFenwickTree_clearB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFenwickTree, "resize", cFenwickTree_resize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFenwickTree, "size", cFenwickTree_get_size, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFenwickTree, "capacity", cFenwickTree_get_capacity, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFenwickTree, "size_2n", cFenwickTree_get_size_2n, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFenwickTree, "total_sum", cFenwickTree_get_total_sum, MRB_ARGS_NONE());
  mrb_define_method(mrb, cFenwickTree, "cum_sum", cFenwickTree_get_cum_sum, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cFenwickTree, "raw_count", cFenwickTree_get_raw_count, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFenwickTree, "element", cFenwickTree_get_element, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFenwickTree, "add", cFenwickTree_add, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cFenwickTree, "update_count", cFenwickTree_update_count, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cFenwickTree, "remove_at", cFenwickTree_remove_at, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cFenwickTree, "replace", cFenwickTree_replace, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cFenwickTree, "dump", cFenwickTree_dump, MRB_ARGS_NONE());

  mrb_define_method(mrb, cFenwickTree, "find_cum_sum", cFenwickTree_find_cum_sum, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cFenwickTree, "find", cFenwickTree_find, MRB_ARGS_REQ(1));

  return cFenwickTree;
}
