#include "tsmrb_helpers.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>

#include "mruby/array.h"
#include "mruby/proc.h"
#include "mruby/variable.h"
#include "mruby/class.h"
#include "mruby/string.h"
#include "mruby/error.h"
#include "mruby/data.h"
#include "mruby/numeric.h"

#include "tsmrb_classes.h"

void tsmrb_free(mrb_state* mrb, void* p)
{
  free(p);
}

#if defined(__APPLE__) || defined(__posix) || defined(__unix) || defined(__linux)
#  define PATHSEP "/"
#elif defined(_WIN64) || defined(_WIN32)
#  define PATHSEP "\\/"
#else
#  error ---> unknown path separator for your architecture <--
#endif
static const char* basepath(const char* p)
{
  static const char* sep = PATHSEP;

  const char* bp = p;
  for (int i=0; sep[i] != '\0'; i++)
  {
    const char* tmp = strrchr(p,sep[i]);
    if (tmp && (tmp > bp)) { bp = tmp+1; }
  }

  return bp;
}
#undef PATHSEP

static void push_rb_backtrace(mrb_state* mrb, mrb_value bt)
{
  mrb_value curr_bt = mrb_get_backtrace(mrb);
  if (mrb_type(curr_bt) == MRB_TT_ARRAY)
  {
    for(int i=0; i < RARRAY_LEN(curr_bt); i++)
    {
      mrb_ary_push(mrb, bt, RARRAY_PTR(curr_bt)[i]);
    }
  }
}

static void push_ts_backtrace(mrb_state* mrb, mrb_value bt, const struct ts_error_state* estate)
{
  char buf[256];

  int maxdepth = estate->depth <= STACK_TRACE_LIMIT ? estate->depth : STACK_TRACE_LIMIT;
  for (int i=0; i < maxdepth; i++)
  {
    snprintf(buf, sizeof(buf), "%s:%d:in %s",
        basepath(estate->trace[i].file),
        estate->trace[i].line,
        estate->trace[i].func);

    mrb_value s = mrb_str_new_cstr(mrb, buf);
    mrb_ary_push(mrb, bt, s);
  }
}

static void add_backtrace(mrb_state* mrb, mrb_value exc, const struct ts_error_state* estate)
{
  mrb_value bt = mrb_ary_new(mrb);
  push_ts_backtrace(mrb, bt, estate);
  push_rb_backtrace(mrb, bt);

  mrb_funcall(mrb, exc, "set_backtrace", 1, bt);
}

#define BUFSZ 256
static void raise_error(mrb_state* mrb, struct RClass* etype, TSResult err, const char* mesg)
{
  const struct ts_error_state *estate;
  char buf[BUFSZ];
  struct RClass *eSimError;
  mrb_value exc;

  estate = ts_error_get_state(err);

  if (estate == NULL)
  {
    return;
  }

  if (mesg)
  {
    snprintf(buf, sizeof(buf), "%s: %s", mesg, estate->error_message);
  }
  else
  {
    strncpy(buf, mesg, sizeof(buf));
    buf[sizeof(buf)-1] = '\0';
  }

  eSimError = tsmrb_classes(mrb)->eSimError;
  exc = mrb_exc_new(mrb, etype, buf, strlen(buf));

  if (etype == eSimError)
  {
    add_backtrace(mrb, exc, estate);
    mrb_funcall(mrb, exc, "ecode=", 1, mrb_fixnum_value(ERRCODE(err)));
  }

  mrb_exc_raise(mrb, exc);
}
#undef BUFSZ

static void set_backtrace(mrb_state *mrb, mrb_value exc, mrb_value bt)
{
  mrb_sym mid;
  struct RClass *cl;

  mid = mrb_intern_cstr(mrb, "set_backtrace");
  cl = mrb_obj_class(mrb, exc);
  if (mrb_obj_respond_to(mrb, cl, mid)) {
    mrb_funcall_argv(mrb, exc, mid, 1, &bt);
  } else {
    mrb_sym vid;
    vid = mrb_intern_cstr(mrb, "@backtrace");
    mrb_iv_set(mrb, exc, vid, bt);
  }
}

static void raise_ruby_error(mrb_state *mrb, TSResult err)
{
  const struct ts_error_state* estate;
  mrb_value exc;

  estate = ts_error_get_state(err);
  if (estate == NULL)
  {
    return;
  }

  exc = mrb_obj_value(estate->user_data);
  mrb_value exc_bt = mrb_funcall(mrb, exc, "backtrace", 0);
  if (mrb_type(exc_bt) != MRB_TT_ARRAY)
  {
    raise_error(mrb, E_RUNTIME_ERROR, err, "!!! RUBY ERROR WITH BAD BACKTRACE !!!");
  }

  mrb_value bt = mrb_ary_new(mrb);

  push_rb_backtrace(mrb, bt);
  push_ts_backtrace(mrb, bt, estate);

  set_backtrace(mrb, exc, bt);
  mrb_exc_raise(mrb,exc);
}


#define _ET(ecase, exc, mesg) \
  case SYMCAT(TS_ERR_,ecase): raise_error(mrb, exc, err, mesg); break
#define _E(err, mesg) _ET(err, tsmrb_classes(mrb)->eSimError, mesg)
void tsmrb_handle_error(mrb_state* mrb, TSResult err)
{
  switch(ERRCODE(err))
  {
    case TS_OKAY:
    case TS_STOP_ITERATION: 
    case TS_CANCELLED: 
      break;

    case TS_ERR_IN_RUBY:
      raise_ruby_error(mrb, err);
      break;

    _E(NO_SUCH_PARAMETER,     "no such parameter"                              );

    _E(NULL_POINTER_ARG,      "argument is NULL but not allowed to be NULL"    );
    _E(SIMULATION_STATE,      "error in simulation state!"                     );
    _E(UNSUPPORTED_OPERATION, "operation not supported"                        );
    _E(NOT_A_NUMBER,          "input string expected to be a number"           );
    _ET(INVALID_ARGUMENT,     E_ARGUMENT_ERROR,   "invalid argument"           );
    _ET(INVALID_INDEX,        E_RANGE_ERROR, "invalid object index"            );
    _ET(MAXIMUM_ALLOCATED,    E_RANGE_ERROR, "maximum number of objects are already allocated");
    _E(UNIMPLEMENTED,         "functionality is unimplemented"                 );

    _E(UNKNOWN_POTENTIAL_PARAM,  "unknown parameter to potential"                 );
    _E(PARTICLES_ALREADY_BONDED, "particles are already bonded"                   );
    _E(PARTICLES_NOT_BONDED,     "no bond exists between particles"               );

    _E(INCORRECT_NUM_BODIES,     "error in the number of particle types"          );
    _E(UNSUPPORTED_POTENTIAL,    "potential is not supported"                     );

    _E(IN_FORCE_LOOP,            "unknown error encountered in the force loop"    );
                                
    _E(UNKNOWN_INTEGRATOR,       "unknown integrator specified"                   );
    _E(BAD_DT,                   "bad dt specified"                               );
    _E(BAD_TEMP,                 "bad temperature specified"                      );
                                
    _E(NO_RNG,                   "no RNG has been set"                            );
    _E(WRONG_RNG,                "invalid RNG for called get_info"                );
    _E(RNG_INCONSISTENT_STATE,   "RNG state is inconsistent (possibly uninitialized?)");
                                
    _E(SHAKE_BOND_TOO_LONG,      "SHAKE error: bond too long"                     );
    _E(SHAKE_MAX_ITERATIONS,     "SHAKE error: exceeded max iterations"           );
                                
    _E(OUT_OF_MEMORY,            "OUT OF MEMORY error"                            );

    _E(LATT_NO_BARRIER_MODEL,    "Lattice requires that a barrier model be selected (currently NONE)");
    _E(LATT_UNIMPLEMENTED_BARRIER_MODEL, "Barrier model is not yet implemented"   );
    _E(LATT_BAD_STATE,           "Lattice is in a BAD STATE"                      );
    _E(LATT_RATE_IS_ZERO,        "Rate is zero"                                   );
    _E(LATT_PSEUDOSPIN_NOT_REGISTERED, "Pseudospin has not been registered"       );

    _E(READ_NOT_FINISHED,        "Could not finish reading from file"             );
    _E(WRITE_NOT_FINISHED,       "Could not finish writing to file"               );

    _E(PARSE_ERROR,              "YAML parse error"                          );

    _E(YAML_LOOKAHEAD_EXCEEDED,  "YAML lookahead exceeded (internal error!)"      );
    _E(YAML_INVALID_CONFIG_DOC,  "YAML invalid configuration document"            );
    _E(YAML_INVALID_KEY,         "YAML invalid top-level key"                     );

    _E(YAML_INVALID_PTYPE,       "YAML invalid particle type"                     );
    _E(YAML_INVALID_PTYPE_LIST,  "YAML invalid list of particle types"            );

    _E(YAML_INVALID_RBTYPE,      "YAML invalid rigid body type"                   );
    _E(YAML_INVALID_RBTYPE_LIST, "YAML invalid list of rigid body types"          );

    _E(YAML_INVALID_BONDTYPE,         "YAML invalid bond type"                    );
    _E(YAML_INVALID_BONDTYPE_LIST,    "YAML invalid list of bond types"           );

    _E(YAML_INVALID_DYNAMICS_SECTION, "YAML invalid 'Dynamics' section"           );
    _E(YAML_INVALID_DYNAMICS_KEY,     "YAML invalid key in 'Dynamics' section"    );
    _E(YAML_INVALID_DYNAMICS_VALUE,   "YAML invalid value in 'Dynamics' section"  );

    default:
    {
      char buf[] = "encountered error (code 0x\?\?\?\?)";
      char *boff = strstr(buf, "\?\?\?\?");
      if (boff != NULL)
      {
        i16_to_hex(ERRCODE(err), boff);
      }
      raise_error(mrb, tsmrb_classes(mrb)->eSimError, err, buf);
    }
    break;
  }

}
#undef _ET
#undef _E

void tsmrb_raisef(mrb_state* mrb, struct RClass* etype, const char* fmt, ...)
{
  char buf[128];
  va_list args;

  va_start(args,fmt);
  vsnprintf(buf, sizeof(buf)-1, fmt, args);
  va_end(args);
  
  mrb_raise(mrb, etype, buf);
}

void _mrbtAssert_priv(mrb_state* mrb, int tf, const char* test, const char* fname, int fline)
{

}

mrb_value mrb_funcall_super(mrb_state* mrb, mrb_value self, mrb_sym name, int argc, mrb_value argv[])
{
  struct RClass* klass = mrb_obj_class(mrb, self);
  struct RProc* super_method = mrb_method_search(mrb, klass->super, name);

  if (super_method == NULL)
    mrb_raisef(mrb, E_NOMETHOD_ERROR, "method error %S", mrb_symbol_value(name));

  return mrb_yield_with_class(mrb, 
      mrb_obj_value(super_method),
      argc, argv,
      self,
      klass->super);
}

mrb_value mrb_call_protected(mrb_state* mrb, mrb_func_t func, int argc, mrb_value argv[], int* state)
{
  struct RProc* fproc = mrb_proc_new_cfunc(mrb, func);
  return mrb_call_protected_proc(mrb, fproc, argc, argv, state);
}

struct protected_args {
  struct RProc* proc;
  int argc;
  mrb_value* argv;
};

mrb_value mrb_call_protected_proc(mrb_state* mrb, struct RProc* proc, int argc, mrb_value argv[], int* state)
{
  mrb_value ret, pobj, args, cargs[2];
  int ai;

  ai = mrb_gc_arena_save(mrb);
  pobj = mrb_obj_value(proc);
  args = mrb_ary_new_from_values(mrb, argc, argv);
  cargs[0] = pobj;
  cargs[1] = args;

  ret = mrb_funcall_argv(mrb,
      mrb_obj_value(tsmrb_classes(mrb)->mTubesheet), 
      mrb_intern_lit(mrb, "protected_call"), 2, cargs);
  if (!mrb_array_p(ret) || (RARRAY_LEN(ret) != 2))
  {
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "protected call should return a two-element array");
  }

  *state = !mrb_nil_p(RARRAY_PTR(ret)[1]);
  if (*state)
  {
    mrb->exc = mrb_obj_ptr(RARRAY_PTR(ret)[1]);
    ret = mrb_nil_value();
  }
  else
  {
    ret = RARRAY_PTR(ret)[0];
  }

  mrb_gc_register(mrb, ret);
  mrb_gc_arena_restore(mrb,ai);
  mrb_gc_protect(mrb, ret);
  mrb_gc_unregister(mrb, ret);

  return ret;
}

TSResult tsmrb_handle_mruby_error(mrb_state* mrb)
{
  if (mrb->exc == NULL)
    return TS_OK;

#ifdef MRB_GC_FIXED_ARENA
  if (mrb->gc->arena_idx >= MRB_GC_ARENA_SIZE)
    return TS_NEW_ERROR3(TS_ERR_IN_RUBY,
        "Error in Ruby callback, but arena size exceeded!");
#endif /* MRB_GC_FIXED_ARENA */

  /* Save exception in the arena */
  mrb_gc_protect(mrb, mrb_obj_value(mrb->exc));
  TSResult err = TS_NEW_ERROR3(TS_ERR_IN_RUBY, "Error in Ruby");
  err = ts_error_set_userdata(err, (void*)mrb->exc);
  mrb->exc = NULL;

  return err;
}

static mrb_value eSimError_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value mesg;
  mrb_int ecode = -1;
  mrb_value bt = mrb_nil_value();
  mrb_get_args(mrb, "o|iA", &mesg, &ecode, &bt);

  mrb_funcall_super(mrb,
      self,
      mrb_intern_cstr(mrb,"initialize"),
      1, &mesg);

  mrb_obj_iv_set(mrb, 
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@ecode"),
      mrb_fixnum_value(ecode));

  mrb_obj_iv_set(mrb, 
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@backtrace"),
      bt);

  return self;
}

static mrb_value eSimError_backtrace(mrb_state* mrb, mrb_value self)
{
  mrb_value rest_bt = mrb_obj_iv_get(mrb,
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@backtrace"));

  mrb_value bt = mrb_funcall_super(mrb, self, mrb_intern_cstr(mrb,"backtrace"), 0, NULL);

  if (!mrb_nil_p(rest_bt))
    mrb_ary_concat(mrb, bt, rest_bt);

  return bt;
}

static mrb_value eSimError_set_backtrace(mrb_state* mrb, mrb_value self)
{
  mrb_value bt;
  mrb_get_args(mrb, "A", &bt);

  mrb_obj_iv_set(mrb, 
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@backtrace"),
      bt);

  return self;
}

static mrb_value eSimError_get_ecode(mrb_state* mrb, mrb_value self)
{
  return mrb_obj_iv_get(mrb,
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@ecode"));
}

static mrb_value eSimError_set_ecode(mrb_state* mrb, mrb_value self)
{
  mrb_int ecode = 0;
  mrb_get_args(mrb, "i", &ecode);

  mrb_obj_iv_set(mrb, 
      mrb_obj_ptr(self),
      mrb_intern_cstr(mrb,"@ecode"),
      mrb_fixnum_value(ecode));

  return self;
}

static void define_error_codes(mrb_state *mrb, struct RClass *mod)
{
#define _EC(esym,ecode,emesg) do { \
  if (ecode >= TS_ERROR_OFFSET) {  \
    mrb_define_const(mrb, mod, SYMSTR(esym), mrb_fixnum_value(ecode)); \
  } } while(0);
#include "ts_error_codes.h"
#undef _EC
}

extern const uint8_t tsmrb_helpers_irep[];

void tsmrb_helpers_init(mrb_state *mrb, struct RClass *mod)
{
  int ai;
  struct RClass *eSimError, *cMemory;

  if (tsmrb_classes(mrb)->eSimError != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "SimError already defined");

  ai = mrb_gc_arena_save(mrb);
  eSimError = mrb_define_class_under(mrb, 
      mod, "SimError", mrb_class_get(mrb, "Exception"));

  mrb_define_method(mrb, eSimError, "initialize", eSimError_initialize, MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
  mrb_define_method(mrb, eSimError, "backtrace", eSimError_backtrace, MRB_ARGS_NONE());
  mrb_define_method(mrb, eSimError, "set_backtrace", eSimError_set_backtrace, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, eSimError, "ecode", eSimError_get_ecode, MRB_ARGS_NONE());
  mrb_define_method(mrb, eSimError, "ecode=", eSimError_set_ecode, MRB_ARGS_REQ(1));

  tsmrb_classes(mrb)->eSimError = eSimError;
  mrb_gc_arena_restore(mrb, ai);

  if (tsmrb_classes(mrb)->cMemory != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "Memory already defined");
  ai = mrb_gc_arena_save(mrb);
  cMemory = mrb_define_class(mrb, "Memory", mrb->object_class);
  MRB_SET_INSTANCE_TT(cMemory, MRB_TT_DATA);
  mrb_gc_arena_restore(mrb, ai);

  ai = mrb_gc_arena_save(mrb);
  define_error_codes(mrb, eSimError);
  mrb_gc_arena_restore(mrb, ai);

  ai = mrb_gc_arena_save(mrb);
  mrb_load_irep(mrb, tsmrb_helpers_irep);
  if (mrb->exc) {
    mrb_p(mrb, mrb_obj_value(mrb->exc));
    exit(0);
  }
  mrb_gc_arena_restore(mrb, ai);
}

static const mrb_data_type memory_type = {
  .struct_name = "Memory",
  .dfree = mrb_free,
};

mrb_value tsmrb_memory_ref(mrb_state *mrb, void *p)
{
  struct RData *data;
  data = mrb_data_object_alloc(mrb, mrb_class_get(mrb, "Memory"), p, &memory_type);
  return mrb_obj_value(data);
}

double *tsmrb_ary_to_doubles(mrb_state* mrb, mrb_value rseq, size_t* n_ptr)
{
  mrb_check_type(mrb, rseq, MRB_TT_ARRAY);
  double* vals = mrb_malloc(mrb, RARRAY_LEN(rseq) * sizeof(*vals));
  int elt;
  for (elt=0; elt < RARRAY_LEN(rseq); elt++)
  {
    if (! mrb_numeric_p(RARRAY_PTR(rseq)[elt])) { goto error; }
    vals[elt] = (double)mrb_to_float(mrb, RARRAY_PTR(rseq)[elt]);
  }

  /* success: */
  *n_ptr = RARRAY_LEN(rseq);
  return vals;

error:
  mrb_free(mrb,vals);
  tsmrb_raisef(mrb, E_TYPE_ERROR, "element %d is not a number", elt);
  return NULL;
}

mrb_value tsmrb_ary_from_doubles(mrb_state *mrb, size_t n, const double *vals)
{
  mrb_value ary;
  size_t i;

  ary = mrb_ary_new_capa(mrb, n);
  for (i=0; i < n; i++)
  {
    int ai = mrb_gc_arena_save(mrb);
    mrb_ary_push(mrb, ary, mrb_float_value(mrb, vals[i]));
    mrb_gc_arena_restore(mrb,ai);
  }

  return ary;
}

int *tsmrb_ary_to_ints(mrb_state *mrb, mrb_value ary, int *buf, size_t *szp)
{
  size_t i,n;

  if (!mrb_array_p(ary))
  {
    goto error;
  }

  n = mrb_ary_len(mrb,ary);
  if (buf == NULL || *szp < n)
  {
    buf = mrb_calloc(mrb, n, sizeof *buf);
  }

  for (i=0; i < n; i++)
  {
    mrb_value elt;
    elt = mrb_ary_ref(mrb, ary, i);
    if (!mrb_fixnum_p(elt))
    {
      goto error;
    }

    buf[i] = mrb_fixnum(elt);
  }

  *szp = n;
  return buf;

error:
  mrb_raise(mrb, E_TYPE_ERROR, "expected array of integers");
}

size_t *tsmrb_ary_to_sizes(mrb_state *mrb, mrb_value ary, size_t *nptr)
{
  if (!mrb_array_p(ary))
    mrb_raise(mrb, E_TYPE_ERROR, "expected array of unsigned integers");

  mrb_int n = RARRAY_LEN(ary);
  if (n < 0)
    mrb_raise(mrb, E_TYPE_ERROR, "array size is negative");

  *nptr = n;
  size_t* elts = NULL;
  if (n > 0)
  {
    elts = mrb_calloc(mrb, n, sizeof(*elts));
    for(mrb_int i=0; i < n; i++)
    {
      mrb_value elt = mrb_ary_ref(mrb, ary, i);
      elts[i] = mrb_fixnum(elt);
    }
  }

  return elts;
}

mrb_value tsmrb_ary_from_sizes(mrb_state* mrb, size_t* vals, size_t n)
{
  mrb_value ary = mrb_ary_new_capa(mrb,n);
  for (size_t i=0; i < n; i++)
    mrb_ary_push(mrb, ary, mrb_fixnum_value( (mrb_int) vals[i] ));
  return ary;
}


mrb_value hash_get_value(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc)
{
  mrb_value val = mrb_hash_get(mrb, h, mrb_symbol_value(key));
  if (mrb_nil_p(val))
  {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "%s requires :%s key", desc, mrb_sym2name(mrb, key));
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }
  return val;
}

double hash_get_float(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc)
{
  return mrb_to_flo(mrb,hash_get_value(mrb,h,key,desc));
}

double hash_get_float_opt(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc, double dft)
{
  mrb_value val = mrb_hash_get(mrb, h, mrb_symbol_value(key));
  return mrb_nil_p(val) ? dft : mrb_to_flo(mrb,val);
}

mrb_int hash_get_int(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc)
{
  mrb_value val = mrb_hash_get(mrb, h, mrb_symbol_value(key));
  mrb_check_type(mrb, val, MRB_TT_FIXNUM);
  return mrb_fixnum(val);
}

const char* hash_get_str(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc)
{
  mrb_value val = hash_get_value(mrb,h,key,desc);
  return mrb_string_value_cstr(mrb, &val);
}



mrb_value tsmrb_bond_type_direction(mrb_state* mrb, ts_bond_type bt)
{
  if (bondtype_is_none(bt)) return mrb_nil_value();

  int direc = bondtype_get_direction(bt);
  switch(direc) {
    case BONDTYPE_UNDIRECTED:    return TSMRB_SYM_OBJ(none);
    case BONDTYPE_BACKWARD:      return TSMRB_SYM_OBJ(back);
    case BONDTYPE_FORWARD:       return TSMRB_SYM_OBJ(forw);
    case BONDTYPE_BIDIRECTIONAL: return TSMRB_SYM_OBJ(bidirectional);
    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR, "invalid bond direction (%d)", direc);
  }

  return mrb_nil_value();
}

static uint16_t mrb_num_to_bond_type(mrb_state* mrb, mrb_value val)
{
  mrb_int bt = mrb_fixnum(val);
  tsmrb_check_bondtype(mrb, bt);

  return (uint16_t)bt;
}

static ts_bond_type mrb_ary_to_bond_type(mrb_state* mrb, mrb_value val)
{
  mrb_check_type(mrb, val, MRB_TT_ARRAY);
  if (RARRAY_LEN(val) != 2)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected a two-element array");

  uint16_t bvalue = mrb_num_to_bond_type(mrb, RARRAY_PTR(val)[1]);
  mrb_value r_bdirec = RARRAY_PTR(val)[0];
  mrb_check_type(mrb, r_bdirec, MRB_TT_SYMBOL);

  if (mrb_symbol(r_bdirec) == TSMRB_SYM(none))
    return bondtype_new(BONDTYPE_UNDIRECTED, bvalue);
  else if (mrb_symbol(r_bdirec) == TSMRB_SYM(forw))
    return bondtype_new(BONDTYPE_FORWARD, bvalue);
  else if (mrb_symbol(r_bdirec) == TSMRB_SYM(back))
    return bondtype_new(BONDTYPE_BACKWARD, bvalue);
  else if (mrb_symbol(r_bdirec) == TSMRB_SYM(bidirectional))
    return bondtype_new(BONDTYPE_BIDIRECTIONAL, bvalue);
  else
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid direction specifier: %S", r_bdirec);

  return bondtype_none();
}

ts_bond_type tsmrb_bondtype(mrb_state* mrb, mrb_value val)
{
  switch (mrb_type(val)) {
    case MRB_TT_ARRAY:
      return mrb_ary_to_bond_type(mrb,val);

    case MRB_TT_FIXNUM:
    case MRB_TT_FLOAT:
      return bondtype_new( BONDTYPE_UNDIRECTED, mrb_num_to_bond_type(mrb,val) );

    default:
      mrb_raise(mrb, E_TYPE_ERROR, "expected either an array or a number");
  }

  return bondtype_none();
}

