#ifndef TSMRB_HELPERS_H
#define TSMRB_HELPERS_H

#include "mruby.h"
#include "mruby/hash.h"
#include "mruby/array.h"
#include "mruby/value.h"
#include "mruby/hash.h"

#include "ts_errors.h"
#include "ts_particle.h"

#define MODULE_NAME TSMRB_HELPERS_H

void tsmrb_helpers_init(mrb_state *mrb, struct RClass *mod);

extern void mrb_c_provide(mrb_state *mrb, const char* filename);
extern void tsmrb_free(mrb_state* mrb, void* p);

/* Checks that class variable given in 'cvar' is initialized.  This
 * macro standardizes the error message for the case when 'cvar' is not
 * initialized.
 */
#define MRB_CHECK_CLASS_INITIALIZED(mrb, cvar, cname) do {   \
  if (cvar == NULL) { mrb_raise( mrb, mrb->eException_class, \
      "class " cname " has not yet been initialized!" ); \
  } } while(0)

/* Checks that module variable given in 'mvar' is initialized.  This
 * macro standardizes the error message for the case when 'mvar' is not
 * initialized.
 */
#define MRB_CHECK_MODULE_INITIALIZED(mrb, mvar, mname) do { \
  if (cvar == NULL) { mrb_raise( mrb, mrb->eException_class, \
      "module " mname " has not yet been initialized!" ); \
  } } while(0)

void tsmrb_handle_error(mrb_state* mrb, TSResult err);

#define mrbtAssert( x ) _mrbtAssert_priv(mrb,(int)(x),#x,__FILE__,__LINE__)
void _mrbtAssert_priv(mrb_state* mrb, int tf, const char* test, const char* fname, int fline);


/* Functions that provide some missing MRuby functionality */

void tsmrb_raisef(mrb_state* mrb, struct RClass* etype, const char* fmt, ...);

mrb_value mrb_call_protected(mrb_state* mrb, mrb_func_t func, int argc, mrb_value argv[], int* state);
mrb_value mrb_call_protected_proc(mrb_state* mrb, struct RProc* proc, int argc, mrb_value argv[], int* state);
mrb_value mrb_funcall_super(mrb_state* mrb, mrb_value self, mrb_sym name, int argc, mrb_value argv[]);

TSResult tsmrb_handle_mruby_error(mrb_state* mrb);

mrb_sym mrb_id_this_frame(mrb_state* mrb);

static inline mrb_bool
mrb_numeric_p(mrb_value val)
{ return mrb_float_p(val) || mrb_fixnum_p(val); }

static inline mrb_float
mrb_to_float(mrb_state* mrb, mrb_value val)
{
  if (mrb_float_p(val))
    return mrb_float(val);
  else if (mrb_fixnum_p(val))
    return (mrb_float)mrb_fixnum(val);
  else
    tsmrb_raisef(mrb, E_TYPE_ERROR,
        "expected numeric (float or fixnum) type, found %s",
        mrb_obj_classname(mrb,val));

  return (mrb_float)0.0;  /* suppresses compiler warnings (no return) */
}

static inline mrb_int
mrb_to_fixnum(mrb_state* mrb, mrb_value val)
{
  if (mrb_float_p(val))
    return (mrb_int)mrb_float(val);
  else if (mrb_fixnum_p(val))
    return mrb_fixnum(val);
  else
    tsmrb_raisef(mrb, E_TYPE_ERROR,
        "expected numeric (float or fixnum) type, found %s",
        mrb_obj_classname(mrb,val));

  return (mrb_float)0.0;  /* suppresses compiler warnings (no return) */
}

/* MRuby object that holds a pointer to allocated memory.  used to
 * reclaim memory allocated with mrb_malloc() and friends.
 *
 * The object is allocated in the arena.  Make sure to keep the object
 * in the arena while the memory is still being used.
 */

mrb_value tsmrb_memory_ref(mrb_state *mrb, void *p);

/* Various array helper functions */

double *tsmrb_ary_to_doubles(mrb_state *mrb, mrb_value rseq, size_t* n_ptr);
mrb_value tsmrb_ary_from_doubles(mrb_state *mrb, size_t n, const double* vals);

/* converts array to a C array of ints
 *
 * In parameters:
 *      rseq    MRuby array of Fixnums
 *      buf     fixed-size buffer (may be NULL, if not NULL, pass size
 *              in *szp)
 *      szp     pointer to array size.  if buf is not NULL, *szp should
 *              be initialized to its number of elements
 *
 * Return value and out parameters
 *      return  If buf is not NULL and all elements fit in buf, returns
 *              buf.  Otherwise returns a newly allocated buffer of
 *              sufficient size.
 *
 *      szp     *szp holds the number of integer elements in the array.
 *
 * Throws
 *      OutOfMemoryError        buffer cannot be allocated
 *      ValueError              element of the array is not a fixnum
 */
int *tsmrb_ary_to_ints(mrb_state *mrb, mrb_value ary, int *buf, size_t *szp);

size_t *tsmrb_ary_to_sizes(mrb_state *mrb, mrb_value ary, size_t* nptr);
mrb_value tsmrb_ary_from_sizes(mrb_state *mrb, size_t* vals, size_t n);

mrb_value hash_get_value(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc);

static inline mrb_value
hash_get_value_opt(mrb_state* mrb, mrb_value h, mrb_sym key)
{
  return mrb_hash_get(mrb, h, mrb_symbol_value(key));
}

const char* hash_get_str(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc);

mrb_int hash_get_int(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc);

double hash_get_float(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc);
double hash_get_float_opt(mrb_state* mrb, mrb_value h, mrb_sym key, const char* desc, double dft);

static inline void hash_set_str(mrb_state* mrb, mrb_value h, mrb_sym k, const char* s)
{
  mrb_hash_set(mrb,h, mrb_symbol_value(k), mrb_str_new_cstr(mrb, s));
}

static inline void hash_set_float(mrb_state* mrb, mrb_value h, mrb_sym k, mrb_float val)
{
  mrb_hash_set(mrb,h, mrb_symbol_value(k), mrb_float_value(mrb,val));
}


ts_bond_type tsmrb_bondtype(mrb_state* mrb, mrb_value val);

static inline void tsmrb_check_bondtype(mrb_state* mrb, mrb_int bt)
{
  if (bt > UINT16_MAX)
    mrb_raise(mrb, E_RANGE_ERROR, "bond type exceeds maximum bond type");
}

mrb_value tsmrb_bond_type_direction(mrb_state* mrb, ts_bond_type bt);

static inline mrb_value
tsmrb_full_bondtype_value(mrb_state* mrb, ts_bond_type bt)
{
  mrb_value pair[2] = {
    tsmrb_bond_type_direction(mrb,bt),
    mrb_fixnum_value(bondtype_get_type(bt))
  };

  return mrb_ary_new_from_values(mrb, 2, pair);
}

static inline mrb_value
tsmrb_bondtype_value(mrb_state* mrb, ts_bond_type bt)
{
  if (bondtype_is_undirected(bt))
    return mrb_fixnum_value(bondtype_get_type(bt));
  else
    return tsmrb_full_bondtype_value(mrb,bt);
}


#undef MODULE_NAME

#endif /* TSMRB_HELPERS_H */

