#include "tsmrb_io.h"

#include <ts_io.h>

#include <mruby/variable.h>

#include <numarray_mruby.h>

#include "tsmrb_helpers.h"
#include "tsmrb_sim.h"
#include "tsmrb_veclist.h"

mrb_value tsmrb_get_stdout(mrb_state *mrb)
{
  return mrb_const_get(mrb, 
      mrb_obj_value(mrb->object_class),
      mrb_intern_lit(mrb, "STDOUT"));
}

static mrb_value cSim_write_positions(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* outf;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  outf = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  tsmrb_handle_error(mrb, tspf_write_positions(sim, outf));

  return mrb_nil_value();
}

static mrb_value cSim_write_xyz(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* out;
  TSResult ret;
  const char *comment;

  io = mrb_nil_value();
  comment = NULL;
  mrb_get_args(mrb, "oz", &io, &comment);

  out = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  ret = ts_write_xyz(sim, out, comment);
  tsmrb_handle_error(mrb, ret);

  return mrb_nil_value();
}

static mrb_value cSim_write_pdb(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* out;
  TSResult ret;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  out = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  ret = ts_write_pdb(sim, out);
  tsmrb_handle_error(mrb, ret);

  return mrb_nil_value();
}

static mrb_value cSim_write_psf(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* out;
  TSResult ret;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  out = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  ret = ts_write_psf(sim, out);
  tsmrb_handle_error(mrb, ret);

  return mrb_nil_value();
}

static void tspf_hdr_to_hash(mrb_state *mrb, mrb_value hash,
    const struct tspf_timepoint_header* header)
{
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(n   ), mrb_fixnum_value(header->n));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(t   ), mrb_float_value(mrb,header->t   ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(temp), mrb_float_value(mrb,header->temp));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(KE  ), mrb_float_value(mrb,header->KE  ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(PE  ), mrb_float_value(mrb,header->PE  ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(totE), mrb_float_value(mrb,header->totE));
}

static mrb_value cVectorList_read_tspf(mrb_state *mrb, mrb_value obj)
{
  VectorList *vl;
  FILE *in;
  mrb_value io;
  struct tspf_timepoint_header header;

  vl = tsmrb_to_veclist(mrb, obj);

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);

  memset(&header,0,sizeof(header));
  tsmrb_handle_error(mrb,
      tspf_read_positions_into_veclist(in, &header, vl));

  return obj;
}

static mrb_value cVectorList_from_tspf_timepoint(mrb_state *mrb, mrb_value class)
{
  VectorList *vl;
  FILE *in;
  mrb_value io, rvl;
  struct tspf_timepoint_header header;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);

  memset(&header,0,sizeof(header));

  vl = NULL;
  rvl = tsmrb_alloc_veclist(mrb, &vl);
  tsmrb_handle_error(mrb,
      tspf_read_positions_into_veclist(in, &header, vl));

  return rvl;
}

static mrb_value cVectorList_tspf_read_into_array(mrb_state *mrb, mrb_value mod)
{
  struct tspf_timepoint_header header;
  numarray *A;
  FILE *in;
  mrb_value rio, rhdr, rarr;
  double *restrict xv, *restrict yv, *restrict zv;

  rio = mrb_nil_value();
  rhdr = mrb_nil_value();
  rarr = mrb_nil_value();
  mrb_get_args(mrb, "oHo", &rio, &rhdr, &rarr);

  in = mrb_io_file(mrb, rio);
  A = numarray_unwrap(mrb, rarr);

  if ((A->ndims != 2) || (A->dims[1] != 3))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "array must be an Nx3 matrix");
  }

  xv = A->elts;
  yv = xv + A->dims[0];
  zv = yv + A->dims[0];

  tsmrb_handle_error(mrb,
      tspf_read_positions(in, &header, A->dims[0], xv, yv, zv));

  tspf_hdr_to_hash(mrb, rhdr, &header);

  return mrb_nil_value();
}

static mrb_value mIO_check_tspf_header(mrb_state *mrb, mrb_value mod)
{
  mrb_value io;
  FILE* in;
  TSResult err;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);

  err = tspf_check_file_header(in);

  return mrb_bool_value( IS_OKAY(err) );
}

static mrb_value mIO_write_tspf_header(mrb_state *mrb, mrb_value module)
{
  mrb_value io;
  FILE* outf;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  outf = mrb_io_file(mrb, io);

  TSResult result = tspf_write_header(outf);
  tsmrb_handle_error(mrb, result);

  return mrb_nil_value();
}

static mrb_value cSim_write_all_bonds(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* outf;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  outf = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  ts_write_all_bonds(sim,outf);
  return self;
}

static mrb_value cSim_write(mrb_state *mrb, mrb_value self)
{
  mrb_value io;
  ts_sim* sim;
  FILE* outf;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  outf = mrb_io_file(mrb, io);
  sim = tsmrb_to_sim(mrb, self);

  ts_write(sim,outf);
  return self;
}

static mrb_value cSim_read(mrb_state *mrb, mrb_value klass)
{
  mrb_value io;
  ts_sim* sim;
  FILE* in;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);

  sim = ts_read(in);
  return tsmrb_make_sim(mrb, sim);
}

static mrb_value mIO_write_tstf_header(mrb_state *mrb, mrb_value mod)
{
  mrb_value io;
  FILE* out;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  out = mrb_io_file(mrb, io);

  tsmrb_handle_error(mrb, tstf_write_header(out));

  return mod;
}

static void tstf_hdr_to_hash(mrb_state *mrb, mrb_value hash, const struct tstf_state_header* header)
{
  mrb_value rstate = mrb_hash_new(mrb);
  tspf_hdr_to_hash(mrb, rstate, &header->sim_state);

  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(state), rstate);
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(num_free_particles),
      mrb_fixnum_value(header->num_free_particles));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(num_rigid_bodies),
      mrb_fixnum_value(header->num_rigid_bodies));

  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(flags), mrb_ary_new(mrb));
}

static mrb_value mIO_read_state_header(mrb_state *mrb, mrb_value mod)
{
  struct tstf_state_header header;
  mrb_value io, rhdr;
  FILE* in;
  TSResult err;
  struct tstf_file *f;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);

  err = TS_OK;
  if (f = tstf_open(in, NO_ALLOCATOR, &err), f != NULL) {
    err = tstf_read_state_header(f, &header);
  }
  tstf_close(f);

  tsmrb_handle_error(mrb, err);

  rhdr = mrb_hash_new(mrb);
  tstf_hdr_to_hash(mrb, rhdr, &header);

  return rhdr;
}

static mrb_value mIO_read_bonds_inner(mrb_state *mrb, FILE *in)
{
  mrb_value rbonds;
  TSResult err = TS_OK;
  struct tstf_file *f;

  err = TS_OK;
  f = tstf_open(in, NO_ALLOCATOR, &err);
  if (f == NULL)
  {
    tsmrb_handle_error(mrb, err);
    return mrb_nil_value();
  }

  tsmrb_handle_error(mrb, tstf_start_reading_bonds(f));

  rbonds = mrb_ary_new(mrb);

  err = TS_OK;
  for(;;)
  {
    struct tstf_bond_info bonds[4];
    size_t nbonds = 0;
    err = tstf_read_bonds(f, ARRAYLEN(bonds), bonds, &nbonds);

    if (!IS_OKAY(err))
    {
      goto finished;
    }

    for (size_t i=0; i < nbonds; i++)
    {
      mrb_value tuple[5];
      int ai;

      if (bonds[i].pi == TS_NULL_PARTICLE) { bonds[i].pi = bonds[i].obj_i; }
      if (bonds[i].pj == TS_NULL_PARTICLE) { bonds[i].pj = bonds[i].obj_j; }

      ai = mrb_gc_arena_save(mrb);
      tuple[0] = mrb_fixnum_value((mrb_int)bonds[i].obj_i);
      tuple[1] = mrb_fixnum_value((mrb_int)bonds[i].obj_j);
      tuple[2] = mrb_fixnum_value((mrb_int)bonds[i].pi);
      tuple[3] = mrb_fixnum_value((mrb_int)bonds[i].pj);
      tuple[4] = tsmrb_full_bondtype_value(mrb, bonds[i].bt);

      mrb_ary_push(mrb, rbonds, mrb_ary_new_from_values(mrb, 5, tuple));
      mrb_gc_arena_restore(mrb, ai);
    }
  }

finished:
  if (IS_STOP_ITERATION(err))
  {
    err = tstf_end_reading_bonds(f);
  }
  tstf_close(f);

  tsmrb_handle_error(mrb, err);

  return rbonds;
}

static mrb_value mIO_write_state(mrb_state *mrb, mrb_value mod)
{
  TSResult err;
  ts_sim *sim;
  FILE *out;
  mrb_value rsim, rio;

  rsim = mrb_nil_value();
  rio = mrb_nil_value();
  mrb_get_args(mrb, "oo", &rsim, &rio);
  sim = tsmrb_to_sim(mrb, rsim);
  out = mrb_io_file(mrb, rio);

  err = tstf_write_state(sim, out);
  tsmrb_handle_error(mrb, err);

  return rsim;
}

static mrb_value mIO_read_bonds(mrb_state *mrb, mrb_value mod)
{
  mrb_value io;
  FILE *in;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);
  return mIO_read_bonds_inner(mrb, in);
}

static mrb_value mIO_check_tstf_header(mrb_state *mrb, mrb_value mod)
{
  TSResult result;
  mrb_value io;
  FILE *in;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);

  in = mrb_io_file(mrb, io);
  result = tstf_check_header(in);

  return mrb_bool_value( IS_OKAY(result) );
}

static mrb_value mIO_write_json_state(mrb_state *mrb, mrb_value mod)
{
  TSResult err;
  mrb_value rio, rsim;
  ts_sim* sim;
  FILE* out;

  rio = mrb_nil_value();
  rsim = mrb_nil_value();
  mrb_get_args(mrb, "oo", &rsim, &rio);

  sim = tsmrb_to_sim(mrb, rsim);
  out = mrb_io_file(mrb, rio);

  err = ts_write_json_state(sim, out);
  tsmrb_handle_error(mrb, err);

  return rsim;
}

struct RClass *tsmrb_define_io(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *mIO = mrb_define_module_under(mrb, mod, "IO");
  tsmrb_classes(mrb)->mIO = mIO;

  struct RClass *cSim = tsmrb_sim_class(mrb);
  mrb_define_method(mrb, cSim, "write_positions", cSim_write_positions, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "write_xyz", cSim_write_xyz, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "write_pdb", cSim_write_pdb, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "write_psf", cSim_write_psf, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "write_all_bonds", cSim_write_all_bonds, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "write", cSim_write, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject*)cSim, "read", cSim_read, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mIO, "write_tstf_header",
      mIO_write_tstf_header, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mIO, "check_tstf_header",
      mIO_check_tstf_header, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mIO, "write_json_state", mIO_write_json_state, MRB_ARGS_REQ(2));

  mrb_define_module_function(mrb, mIO, "write_state", mIO_write_state, MRB_ARGS_REQ(2));
  mrb_define_module_function(mrb, mIO, "read_state_header", mIO_read_state_header, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mIO, "read_bonds", mIO_read_bonds, MRB_ARGS_REQ(1));

  {
    struct RClass *cVectorList = tsmrb_veclist_class(mrb);

    mrb_define_singleton_method(mrb, (struct RObject*)cVectorList, "from_tspf_timepoint", 
        cVectorList_from_tspf_timepoint, MRB_ARGS_REQ(1));

    mrb_define_module_function(mrb, mIO, "read_positions_into_veclist",
        cVectorList_from_tspf_timepoint, MRB_ARGS_REQ(1));

    mrb_define_method(mrb, cVectorList, "read_tspf", cVectorList_read_tspf, MRB_ARGS_REQ(1));
  }

  mrb_define_module_function(mrb, mIO, "tspf_read_into_array",
      cVectorList_tspf_read_into_array, MRB_ARGS_REQ(3));

  mrb_define_module_function(mrb, mIO, "write_tspf_header",
      mIO_write_tspf_header, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mIO, "check_tspf_header",
      mIO_check_tspf_header, MRB_ARGS_REQ(1));

  return mIO;
}


