#include "mruby.h"
#include "tsmrb_helpers.h"
#include "tsmrb_classes.h"

#include "mruby/data.h"
#include "mruby/string.h"

#include "ts_json.h"

struct json_wrapper
{
  ts_json_parser* parser;
  ts_json_event event;
};

static inline struct json_wrapper* 
new_json_wrapper(mrb_state* mrb)
{
  struct json_wrapper* wrapper = mrb_malloc(mrb, sizeof(*wrapper));
  wrapper->parser = NULL;
  return wrapper;
}

static void free_json_wrapper(mrb_state* mrb, void* p)
{
  struct json_wrapper* wrapper = p;

  if (wrapper->parser) {
    ts_json_finalize(wrapper->parser);
    mrb_free(mrb,wrapper->parser);
    wrapper->parser = NULL;
  }

  mrb_free(mrb,wrapper);
}

static const mrb_data_type json_type = {
  .struct_name = "JSON",
  .dfree = free_json_wrapper,
};

static inline struct RClass*
tsmrb_json_class(mrb_state* mrb)
{ return tsmrb_classes(mrb)->cJSON; }

static inline struct json_wrapper*
to_json_wrapper(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_json_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a JSON object");

  return DATA_PTR(obj);
}

static void json_free(void* cl, void* p) { mrb_free((mrb_state*)cl, p); }

static mrb_value cJSON_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value sobj = mrb_nil_value();
  mrb_get_args(mrb, "S", &sobj);

  struct json_wrapper* wrapper = new_json_wrapper(mrb);
  DATA_PTR(self) = wrapper;
  DATA_TYPE(self) = &json_type;

  char* origp = RSTRING_PTR(sobj);
  size_t len = RSTRING_LEN(sobj);

  char* bufp = mrb_malloc(mrb, sizeof(char)*len);
  memcpy(bufp, origp, sizeof(char)*len);

  wrapper->parser = mrb_malloc(mrb, sizeof(*(wrapper->parser)));
  tsmrb_handle_error(mrb, 
      ts_json_init_buf(wrapper->parser, bufp, len, json_free, mrb) );

  return self;
}

static inline mrb_value
make_pair(mrb_state* mrb, mrb_sym a, mrb_value b)
{
  mrb_value pair[2] = { mrb_symbol_value(a),b };
  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value event_to_ruby(mrb_state* mrb, ts_json_event* event)
{
  switch (event->type)
  {
    case TS_JSON_NONE:
    case TS_JSON_EOF:
      return mrb_nil_value();

    case TS_JSON_STRING:
      return make_pair(mrb,
          mrb_intern_cstr(mrb,"string"),
          mrb_str_new(mrb,event->value.str.p, event->value.str.len));

    case TS_JSON_NUMBER:
      return make_pair(mrb,
          mrb_intern_cstr(mrb,"number"),
          mrb_float_value(mrb,event->value.num));

    case TS_JSON_ARRAY_BEGIN:
      return mrb_symbol_value( mrb_intern_cstr(mrb, "array_begin") );

    case TS_JSON_ARRAY_END:
      return mrb_symbol_value( mrb_intern_cstr(mrb, "array_end") );

    case TS_JSON_OBJECT_BEGIN:
      return mrb_symbol_value( mrb_intern_cstr(mrb, "object_begin") );

    case TS_JSON_OBJECT_KEY:
      return make_pair(mrb,
          mrb_intern_cstr(mrb,"object_key"),
          mrb_str_new(mrb,event->value.str.p, event->value.str.len));

    case TS_JSON_OBJECT_END:
      return mrb_symbol_value(mrb_intern_cstr(mrb, "object_end"));

    case TS_JSON_LITERAL:
      {
        if (event->value.lit == TS_JSON_NULL)
          return mrb_symbol_value( mrb_intern_cstr(mrb, "null") );
        else if (event->value.lit == TS_JSON_FALSE)
          return mrb_symbol_value( mrb_intern_cstr(mrb, "false") );
        else if (event->value.lit == TS_JSON_TRUE)
          return mrb_symbol_value( mrb_intern_cstr(mrb, "true") );
        else
          tsmrb_raisef(mrb, E_RUNTIME_ERROR,
              "unknown literal value %d", (int)event->value.lit);
      }

    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR, "unknown event type %d", (int)event->type);
  }

  return mrb_nil_value();
}

static mrb_value cJSON_next(mrb_state* mrb, mrb_value self)
{
  struct json_wrapper* wrapper = to_json_wrapper(mrb,self);
  tsmrb_handle_error(mrb, ts_json_next(wrapper->parser, &wrapper->event) );

  return event_to_ruby(mrb,&wrapper->event);
}

static inline mrb_value lit_to_rb(mrb_state* mrb, enum TS_JSON_LITERAL lit)
{
  switch (lit) {
    case TS_JSON_NULL:
      return mrb_nil_value();

    case TS_JSON_FALSE:
      return mrb_bool_value(0);

    case TS_JSON_TRUE:
      return mrb_bool_value(1);
  }

  return mrb_nil_value();
}

static TSResult do_parse_array(mrb_state* mrb, ts_json_parser* parser, ts_json_event* event, mrb_value* valuep);
static TSResult do_parse_object(mrb_state* mrb, ts_json_parser* parser, ts_json_event* event, mrb_value* valuep);

static TSResult do_parse_event(mrb_state* mrb, ts_json_parser* parser, ts_json_event* event, mrb_value* valuep)
{
  switch (event->type) {
    case TS_JSON_NONE:
    case TS_JSON_EOF:
    default:
      return TS_OK;

    case TS_JSON_LITERAL:
      *valuep = lit_to_rb(mrb, event->value.lit);
      return TS_OK;

    case TS_JSON_STRING:
      *valuep = mrb_str_new(mrb, event->value.str.p, event->value.str.len);
      return TS_OK;

    case TS_JSON_NUMBER:
      *valuep = mrb_float_value(mrb, event->value.num);
      return TS_OK;

    case TS_JSON_ARRAY_BEGIN:
      RETURN_WITH_RESULT( do_parse_array(mrb, parser, event, valuep) );

    case TS_JSON_OBJECT_BEGIN:
      RETURN_WITH_RESULT( do_parse_object(mrb, parser, event, valuep) );
  }

}

static TSResult do_parse_array(mrb_state* mrb, ts_json_parser* parser, ts_json_event* event, mrb_value* valuep)
{
  *valuep = mrb_ary_new(mrb);

  for(;;) {
    RETURN_ON_ERROR( ts_json_next(parser, event) );
    if (event->type == TS_JSON_ARRAY_END) { return TS_OK; }

    mrb_value elt = mrb_nil_value();
    RETURN_ON_ERROR( do_parse_event(mrb,parser,event,&elt) );
    mrb_ary_push(mrb, *valuep, elt);
  }
}

static TSResult do_parse_object(mrb_state* mrb, ts_json_parser* parser, ts_json_event* event, mrb_value* valuep)
{
  *valuep = mrb_hash_new(mrb);

  for(;;) {
    RETURN_ON_ERROR( ts_json_next(parser, event) );
    if (event->type == TS_JSON_OBJECT_END) { return TS_OK; }

    if (event->type != TS_JSON_OBJECT_KEY)
      return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN,
          ("expected object key, but found %d", (int)event->type));

    mrb_value key = mrb_str_new(mrb, event->value.str.p, event->value.str.len);
    mrb_value val = mrb_nil_value();

    RETURN_ON_ERROR( ts_json_next(parser, event) );
    RETURN_ON_ERROR( do_parse_event(mrb,parser,event,&val) );
    mrb_hash_set(mrb, *valuep, key, val);
  }
}

static TSResult do_parse(mrb_state* mrb, ts_json_parser* parser, mrb_value* valuep)
{
  ts_json_event event;

  *valuep = mrb_ary_new(mrb);
  for(;;) {
    RETURN_ON_ERROR( ts_json_next(parser, &event) );
    if (event.type == TS_JSON_EOF) { return TS_OK; }

    mrb_value elt = mrb_nil_value();
    RETURN_ON_ERROR( do_parse_event(mrb, parser, &event, &elt) );
    mrb_ary_push(mrb, *valuep, elt);
  }
}

static mrb_value cJSON_parse_string(mrb_state* mrb, mrb_value mod)
{
  const char* origp = NULL;
  int len = 0;
  mrb_get_args(mrb, "s", &origp, &len);

  char* bufp = mrb_malloc(mrb,len*sizeof(*bufp));
  memcpy(bufp, origp, sizeof(char)*len);

  ts_json_parser parser;
  ts_json_init_buf(&parser, bufp, len, json_free, mrb);

  mrb_value ret = mrb_nil_value();
  TSResult err = do_parse(mrb, &parser, &ret);
  ts_json_finalize(&parser);

  tsmrb_handle_error(mrb, err);

  return ret;
}

static mrb_value cJSON_parse(mrb_state* mrb, mrb_value mod)
{
  return cJSON_parse_string(mrb, mod);
}

struct RClass* tsmrb_define_json(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cJSON = mrb_define_class_under(mrb, mod, "JSON", mrb->object_class);
  tsmrb_classes(mrb)->cJSON = cJSON;
  MRB_SET_INSTANCE_TT(cJSON, MRB_TT_DATA);

  mrb_define_module_function(mrb, cJSON, "parse", cJSON_parse, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cJSON, "initialize", cJSON_initialize, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cJSON, "next"      , cJSON_next      , MRB_ARGS_NONE());

  return cJSON;
}
