#include "tsmrb_lattice.h"

#include <stdbool.h>

#include "mruby/data.h"
#include "mruby/khash.h"
#include "mruby/proc.h"
#include "mruby/variable.h"

#include "tsmrb_helpers.h"
#include "tsmrb_rng.h"
#include "tsmrb_io.h"

static void lattice_free(mrb_state* mrb, void* p)
{
  ts_lattice_release(p);
}

struct change_state {
  ts_lattice *latt;
  struct ts_lattice_change_list changes;
  struct ts_lattice_update_list updates;
};

static void changes_free(mrb_state* mrb, void* p)
{
  struct change_state *chg = p;
  ts_lattice_release(chg->latt);
  ts_lattice_change_list_finalize(&chg->changes);
  mrb_free(mrb, chg);
}

static const mrb_data_type latt_type = {
  .struct_name = "Lattice",
  .dfree = lattice_free,
};

static const mrb_data_type changes_type = {
  .struct_name = "Lattice::Changes",
  .dfree = changes_free,
};

void tsmrb_lattice_setup_callback_refs(mrb_state* mrb, mrb_value obj)
{
  mrb_value arr = mrb_ary_new(mrb);
  mrb_iv_set(mrb, obj, mrb_intern_cstr(mrb,"@cbref"), arr);
}

static inline mrb_value get_callback_refs(mrb_state* mrb, mrb_value obj)
{
  return mrb_iv_get(mrb, obj, mrb_intern_cstr(mrb,"@cbref"));
}

static inline void add_callback_ref(mrb_state* mrb, mrb_value obj, mrb_value cb)
{
  mrb_value arr = get_callback_refs(mrb,obj);
  mrb_ary_push(mrb, arr, cb);
}

ts_lattice* tsmrb_to_lattice(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_kind_of(mrb, obj, tsmrb_lattice_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Lattice object");

  return DATA_PTR(obj);
}

void tsmrb_check_on_lattice(mrb_state* mrb, const ts_lattice* latt, mrb_int i, mrb_int j)
{
  if ((i >= latt->nrow) || (j >= latt->ncol))
  { 
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "site (%d,%d) is off the chem_driver", (int)i, (int)j);
  }
}

static mrb_value latt_get_debug_updates(mrb_state *mrb, mrb_value self)
{
  return mrb_fixnum_value(ts_latt_debug_updates);
}

static mrb_value latt_set_debug_updates(mrb_state *mrb, mrb_value self)
{
  mrb_int tf = 0;
  mrb_get_args(mrb, "i", &tf);
  ts_latt_debug_updates = tf;

  return mrb_nil_value();
}

static mrb_value cLattice_ncol(mrb_state* mrb, mrb_value self)
{
  return mrb_fixnum_value( tsmrb_to_lattice(mrb,self)->ncol );
}

static mrb_value cLattice_nrow(mrb_state* mrb, mrb_value self)
{
  return mrb_fixnum_value( tsmrb_to_lattice(mrb,self)->nrow );
}

static mrb_value cLattice_size(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  mrb_value pair[2] = {
    mrb_fixnum_value((mrb_int)latt->nrow),
    mrb_fixnum_value((mrb_int)latt->ncol),
  };
  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value cLattice_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_int nrow, ncol, nrates;
  ts_lattice *latt;
  TSResult err;

  nrow = ncol = nrates = 0;
  mrb_get_args(mrb, "iii", &nrow, &ncol, &nrates);

  err = TS_OK;
  latt = ts_lattice_new(nrow,ncol,nrates,&err);

  if (latt == NULL)
  {
    tsmrb_handle_error(mrb,err);
  }

  DATA_PTR(self) = latt;
  DATA_TYPE(self) = &latt_type;

  tsmrb_lattice_setup_callback_refs(mrb, self);

  return self;
}

static mrb_value cLattice_get_rng(mrb_state* mrb, mrb_value self)
{
  ts_lattice *latt;
  mrb_value r_rng;
  mrb_sym at_rng;

  at_rng = mrb_intern_lit(mrb,"@rng");
  r_rng = mrb_iv_get(mrb, self, at_rng);
  latt = tsmrb_to_lattice(mrb,self);
  if (mrb_nil_p(r_rng) && !latt->rng)
    return mrb_nil_value();

  if (mrb_nil_p(r_rng))
  {
    r_rng = tsmrb_from_rng(mrb,latt->rng);
    mrb_iv_set(mrb, self, at_rng, r_rng);
    return r_rng;
  }

  if (!latt->rng)
  {
    mrb_raisef(mrb, E_RUNTIME_ERROR,
        "inconsistent RNG state (@rng is set but latt->rng is NULL");
  }

  /* sanity check ... they should be consistent! */
  if (latt->rng != tsmrb_to_rng(mrb,r_rng))
  {
    mrb_raisef(mrb, E_RUNTIME_ERROR,
        "inconsistent RNG state (@rng not equal to latt->rng)");
  }

  return r_rng;
}

static mrb_value cLattice_set_rng(mrb_state* mrb, mrb_value self)
{
  mrb_value rrng;
  mrb_get_args(mrb, "o", &rrng);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  struct ts_rng* rng = tsmrb_to_rng(mrb,rrng);
  if (rng != NULL)
    latt->rng = rng;
  mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@rng"), rrng);

  return self;
}

static mrb_value cLattice_get_time(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  return mrb_float_value(mrb, latt->time );
}

static mrb_value cLattice_get_nsteps(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  return mrb_fixnum_value(latt->nsteps);
}

static mrb_value cLattice_set_temp(mrb_state* mrb, mrb_value self)
{
  mrb_float temp;
  mrb_get_args(mrb, "f", &temp);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  ts_lattice_set_temp(latt, temp);
  return self;
}

static mrb_value cLattice_get_temp(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb, self);
  return mrb_float_value(mrb, latt->temp);
}

static mrb_value set_coupling_dist(mrb_state* mrb, mrb_value self, int is_col)
{
  mrb_int coupling_dist;
  ts_lattice *latt = tsmrb_to_lattice(mrb,self);

  latt = tsmrb_to_lattice(mrb,self);
  coupling_dist = 0;

  mrb_get_args(mrb, "i", &coupling_dist);

  if (coupling_dist < 0)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "coupling_dist cannot be negative");
  }

  if (is_col) {
    latt->col_coupling_dist = coupling_dist;
  } else {
    latt->row_coupling_dist = coupling_dist;
  }

  return self;
}

static mrb_value cLattice_set_row_coupling_dist(mrb_state* mrb, mrb_value self)
{
  return set_coupling_dist(mrb, self, 0);
}

static mrb_value cLattice_set_col_coupling_dist(mrb_state* mrb, mrb_value self)
{
  return set_coupling_dist(mrb, self, 1);
}

static mrb_value cLattice_get_row_coupling_dist(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb, self);
  return mrb_fixnum_value(latt->row_coupling_dist);
}

static mrb_value cLattice_get_col_coupling_dist(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb, self);
  return mrb_fixnum_value(latt->col_coupling_dist);
}

static mrb_value cLattice_at(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  tsmrb_check_on_lattice(mrb, latt, i,j);

  return mrb_fixnum_value( ts_lattice_at(latt, i,j).val );
}

static mrb_value cLattice_set(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_int v = 0;
  mrb_get_args(mrb, "iii", &i, &j, &v);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  tsmrb_check_on_lattice(mrb, latt, i,j);

  ts_lattice_ref(latt, i,j)[0] = spin_from_int(v);
  return self;
}

static mrb_value cLattice_wrap_coords(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt;
  mrb_int r,c;
  mrb_value rcoords[2];

  mrb_get_args(mrb, "ii", &r, &c);
  latt = tsmrb_to_lattice(mrb,self);

  ts_lattice_wrap_coord(latt, &r, &c);

  rcoords[0] = mrb_fixnum_value(r);
  rcoords[1] = mrb_fixnum_value(c);

  return mrb_ary_new_from_values(mrb, 2, rcoords);
}

static mrb_value cLattice_get_state_string(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  return mrb_str_new(mrb, (char*)latt->sites, latt->num_sites * sizeof(latt->sites[0]));
}

static mrb_value cLattice_set_state_string(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt;
  char *s;
  size_t expected_size;
  mrb_int len;

  latt = tsmrb_to_lattice(mrb,self);

  s = NULL;
  len = 0;
  mrb_get_args(mrb, "s", &s, &len);

  expected_size = latt->num_sites * sizeof(latt->sites[0]);
  if (!s || (len != expected_size))
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "expected a string of size %zu", expected_size);
  }

  memcpy(latt->sites, s, expected_size);
  return mrb_nil_value();
}

static mrb_value cLattice_change_spin(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_int v = 0;
  mrb_get_args(mrb, "iii", &i, &j, &v);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  tsmrb_handle_error(mrb,
      ts_lattice_change_spin(latt, i,j, spin_from_int(v)));

  return self;
}

static mrb_value cLattice_get_site_moves(mrb_state *mrb, mrb_value self)
{
  ts_lattice *latt;
  mrb_int r = 0;
  mrb_int c = 0;
  ts_lattice_move *mlist;

  latt = tsmrb_to_lattice(mrb,self);
  mrb_get_args(mrb, "ii", &r, &c);

  tsmrb_check_on_lattice(mrb, latt, r, c);

  mrb_value moves = mrb_hash_new(mrb);
  mlist = ts_lattice_get_current_site_moves(latt, r,c);
  while(mlist)
  {
    mrb_hash_set(mrb, moves,
        mrb_fixnum_value(mlist->state.val), mrb_float_value(mrb,mlist->rate));
    mlist = mlist->next;
  }

  return moves;
}

static mrb_value cLattice_get_all_moves(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  ts_lattice_move* moves;
  size_t nmoves;
  double ktot;

  tsmrb_handle_error(mrb, 
      ts_lattice_get_all_moves(latt, &moves, &nmoves, &ktot));

  mrb_value ret = mrb_ary_new_capa(mrb,nmoves);
  for (size_t i=0; i < nmoves; i++)
  {
    mrb_value elts[4];
    elts[0] = mrb_fixnum_value( moves[i].r );
    elts[1] = mrb_fixnum_value( moves[i].c );
    elts[2] = mrb_fixnum_value( moves[i].state.val );
    elts[3] = mrb_float_value(mrb, moves[i].rate);

    mrb_ary_push(mrb, ret,
        mrb_ary_new_from_values(mrb, 4, elts));
  }

  ts_free(NO_ALLOCATOR, moves);

  return ret;
}

static mrb_value cLattice_resetB(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  ts_lattice_reset(latt);
  return self;
}

static mrb_value cLattice_zeroB(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  ts_lattice_zero(latt);
  ts_lattice_rebuild_moves(latt);
  return mrb_nil_value();
}

static mrb_value cLattice_rebuild_movesB(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  tsmrb_handle_error(mrb, ts_lattice_rebuild_moves(latt) );
  return self;
}

static mrb_value call_site_func(mrb_state *mrb, mrb_value self, TSResult (*sitefunc)(ts_lattice *,int,int))
{
  mrb_value *argv;
  mrb_int i, argc;
  ts_lattice* latt;

  mrb_get_args(mrb, "*", &argv,&argc);
  latt = tsmrb_to_lattice(mrb,self);

  for (i=0; i < argc; i++)
  {
    mrb_value e0,e1;
    int r,c;

    mrb_check_type(mrb, argv[i], MRB_TT_ARRAY);
    e0 = mrb_ary_ref(mrb, argv[i], 0);
    e1 = mrb_ary_ref(mrb, argv[i], 1);

    mrb_check_type(mrb, e0, MRB_TT_FIXNUM);
    mrb_check_type(mrb, e1, MRB_TT_FIXNUM);

    r = mrb_fixnum(e0);
    c = mrb_fixnum(e1);

    tsmrb_handle_error(mrb, sitefunc(latt, r,c));
  }

  return self;
}

static mrb_value cLattice_changed_siteB(mrb_state *mrb, mrb_value self)
{
  return call_site_func(mrb, self, &ts_lattice_mark_site_changed);
}

static TSResult update_single_site(ts_lattice *latt, int r, int c)
{
  RETURN_WITH_RESULT(ts_lattice_mark_range_for_update(latt, r,r,c) );
}

static mrb_value cLattice_update_siteB(mrb_state *mrb, mrb_value self)
{
  return call_site_func(mrb, self, &update_single_site);
}

static mrb_value cLattice_update_rangeB(mrb_state *mrb, mrb_value self)
{
  mrb_int r0,r1,c;
  ts_lattice *latt;

  latt = tsmrb_to_lattice(mrb,self);
  r0 = r1 = c = -1;
  mrb_get_args(mrb, "iii", &r0, &r1, &c);

  if (r0 < 0 || r0 >= latt->nrow ||
      r1 < 0 || r1 >= latt->nrow ||
      c < 0 || c >= latt->ncol)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "update range must be on the lattice");
  }

  if (r0 > r1)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "rows out of order (rmin > rmax)");
  }

  tsmrb_handle_error(mrb, ts_lattice_mark_range_for_update(latt, r0,r1,c));

  return mrb_nil_value();
}

static mrb_value cLattice_each_move(mrb_state* mrb, mrb_value self)
{
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  ts_lattice_move_iterator iter;
  tsmrb_handle_error(mrb, 
      ts_lattice_move_iterator_init(latt, &iter) );

  mrb_value elts[4];
  if (!mrb_nil_p(block))
  {
    TSResult ret = TS_OK;
    while (IS_OKAY(ret = ts_lattice_moves_iterate(latt, &iter)))
    {
      elts[0] = mrb_fixnum_value(iter.curr.r);
      elts[1] = mrb_fixnum_value(iter.curr.c);
      elts[2] = mrb_fixnum_value(iter.curr.state.val);
      elts[3] = mrb_float_value(mrb, iter.curr.rate);

      mrb_yield(mrb, block, mrb_ary_new_from_values(mrb, 4, elts));
    }

    tsmrb_handle_error(mrb,ret);
    return self;
  }
  else
  {
    mrb_value arr = mrb_ary_new(mrb);
    TSResult ret = TS_OK;
    while (IS_OKAY(ret = ts_lattice_moves_iterate(latt, &iter)))
    {
      elts[0] = mrb_fixnum_value(iter.curr.r);
      elts[1] = mrb_fixnum_value(iter.curr.c);
      elts[2] = mrb_fixnum_value(iter.curr.state.val);
      elts[3] = mrb_float_value(mrb, iter.curr.rate);

      mrb_ary_push(mrb, arr, mrb_ary_new_from_values(mrb, 4, elts));
    }

    tsmrb_handle_error(mrb,ret);

    return arr;
  }
}

static mrb_value cLattice_get_k_total(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  double ktot = 0.0;
  tsmrb_handle_error(mrb,
      ts_lattice_get_k_total(latt,&ktot));
  return mrb_float_value(mrb,ktot);
}

static mrb_value cLattice_simulate_until(mrb_state* mrb, mrb_value self)
{
  mrb_float t1 = 0.0;
  mrb_get_args(mrb, "f", &t1);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  if (latt->rng == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "Lattice RNG is NULL");

  tsmrb_handle_error(mrb, ts_lattice_simulate_until(latt, t1));
  return self;
}

static mrb_value cLattice_single_step(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  if (latt->rng == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "Lattice RNG is NULL");

  tsmrb_handle_error(mrb, ts_lattice_single_step(latt));

  return self;
}

struct callback_data
{
  mrb_state* mrb;
  struct RData* rlatt;
  struct RProc* proc;

  int nstates;
  spin_t* spins;
  double* rates;
};

static TSResult
pspin_callback_get_return_coords(mrb_state* mrb, mrb_value ret, int i, int* rp, int* cp)
{
  mrb_value r_coord = RARRAY_PTR(ret)[i];

  if (!mrb_array_p(r_coord))
    return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN, 
        ("Element %d in returned value is not an array of coordinates", i));

  if (RARRAY_LEN(r_coord) != 2)
    return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN, 
        ("Element %d in returned value has length != 2 (%d)", i, RARRAY_LEN(r_coord)));

  if (!mrb_fixnum_p(RARRAY_PTR(r_coord)[0]))
    return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN, 
        ("Element %d in returned value has row coord that is not FIXNUM",i));

  if (!mrb_fixnum_p(RARRAY_PTR(r_coord)[1]))
    return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN, 
        ("Element %d in returned value has column coord that is not FIXNUM",i));

  *rp = (int) mrb_fixnum(RARRAY_PTR(r_coord)[0]);
  *cp = (int) mrb_fixnum(RARRAY_PTR(r_coord)[1]);

  return TS_OK;
}

static TSResult pspin_callback(void *data,
    ts_lattice* latt, int r, int c, spin_t pspin)
{
  struct callback_data* closure = data;
  mrb_state* mrb = closure->mrb;

  mrb_value args[4] = {
    mrb_obj_value(closure->rlatt),
    mrb_fixnum_value(r),
    mrb_fixnum_value(c),
    mrb_fixnum_value(pspin.val),
  };

  int state = 0;
  mrb_value ret = mrb_call_protected_proc(mrb,
      closure->proc, 4, args, &state);

  if (state)
    return tsmrb_handle_mruby_error(mrb);

  if (mrb_array_p(ret))
  {
    for (int i=0; i < RARRAY_LEN(ret); i++)
    {
      int ret_r = 0;
      int ret_c = 0;

      RETURN_ON_ERROR( pspin_callback_get_return_coords(mrb, ret, i, &ret_r, &ret_c) );

      if ((ret_r < 0) || (ret_r >= latt->nrow) || (ret_c < 0) || (ret_c >= latt->ncol))
      {
        return TS_NEW_ERROR2(TS_ERR_INVALID_RETURN, 
            ("Element %d of return value: coordinates (%d,%d) out of bounds", i, ret_r,ret_c));
      }

      RETURN_ON_ERROR( ts_lattice_mark_site_changed(latt, ret_r, ret_c) );
    }
  }
  else if (!mrb_nil_p(ret))
  {
    return TS_NEW_ERROR3(TS_ERR_INVALID_RETURN,
        "Returned value is not an array or nil");
  }

  return TS_OK;
}

static mrb_value cLattice_add_pseudospin(mrb_state* mrb, mrb_value self)
{
  mrb_int pspin_val = 0;
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "i&", &pspin_val, &block);

  ts_lattice *latt = tsmrb_to_lattice(mrb,self);
  spin_t pspin = spin_from_int( pspin_val );

  if (mrb_nil_p(block))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Expected block or proc");

  add_callback_ref(mrb, self, block);
  struct callback_data *cb = ts_alloc(NO_ALLOCATOR, sizeof *cb);
  cb->mrb = mrb;
  cb->rlatt = RDATA(self);
  cb->proc = mrb_proc_ptr(block); 

  tsmrb_handle_error(mrb,
      ts_lattice_register_pseudospin(latt, pspin, 
        pspin_callback, cb) );

  return self;
}

static mrb_value cLattice_num_pseudospins(mrb_state* mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  return mrb_fixnum_value( (mrb_int)ts_lattice_num_pseudospins(latt) );
}

static TSResult move_hook_callback(void *data,
    ts_lattice* latt, const ts_lattice_move* move)
{
  struct callback_data* closure = data;
  mrb_state* mrb = closure->mrb;

  mrb_value args[4] = {
    mrb_obj_value(closure->rlatt),
    mrb_fixnum_value(move->r),
    mrb_fixnum_value(move->c),
    mrb_fixnum_value(move->state.val),
  };

  int state = 0;
  mrb_value ret = mrb_call_protected_proc(mrb,
      closure->proc, 4, args, &state);

  if (!state)
  {
    if (mrb_symbol(ret) != TSMRB_SYM(cancel))
    {
      return TS_OK;
    }
    else
    {
      CANCEL_OPERATION();
    }
  }
  else
  {
    return tsmrb_handle_mruby_error(mrb);
  }
}

static mrb_value add_move_hook(mrb_state* mrb, mrb_value self, int call_before_move)
{
  ts_lattice *latt;
  struct callback_data *cb;
  TSResult ret;
  mrb_value block, hooklist;

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);
  latt = tsmrb_to_lattice(mrb,self);

  if (mrb_nil_p(block))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Expected block or proc");

  cb = ts_alloc(NO_ALLOCATOR, sizeof(*cb));
  cb->mrb = mrb;
  cb->rlatt = RDATA(self);
  cb->proc = mrb_proc_ptr(block);

  if (call_before_move)
  {
    ret = ts_lattice_add_move_hook(latt, move_hook_callback, cb);
  }
  else
  {
    ret = ts_lattice_add_post_move_hook(latt, move_hook_callback, cb);
  }

  tsmrb_handle_error(mrb, ret);

  hooklist = mrb_iv_get(mrb, self, mrb_intern_lit(mrb, "@move_hooks"));
  if (mrb_nil_p(hooklist)) {
    hooklist = mrb_ary_new(mrb);
    mrb_iv_set(mrb, self, mrb_intern_lit(mrb, "@move_hooks"), hooklist);
  }

  mrb_ary_push(mrb, hooklist, block);

  return self;
}

static mrb_value cLattice_add_move_hook(mrb_state* mrb, mrb_value self)
{
  int call_before_move = 1;
  return add_move_hook(mrb, self, call_before_move );
}

static mrb_value cLattice_add_post_move_hook(mrb_state* mrb, mrb_value self)
{
  int call_before_move = 0;
  return add_move_hook(mrb, self, call_before_move );
}

static mrb_value cLattice_is_pseudospin(mrb_state *mrb, mrb_value self)
{
  ts_lattice *latt;
  mrb_int spin_value, is_pseudo;
  spin_t spin;

  latt = tsmrb_to_lattice(mrb,self);
  spin_value = 0;
  mrb_get_args(mrb, "i", &spin_value);

  spin = spin_from_int(spin_value);
  is_pseudo = ts_lattice_is_pseudo_spin(latt,spin);
  return mrb_bool_value(is_pseudo);
}

static mrb_value invoke_pseudospin(mrb_state* mrb, mrb_value self, bool do_update)
{
  mrb_int pspin = 0;
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "iii", &pspin, &i, &j);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  if (do_update)
  {
    tsmrb_handle_error(mrb,
        ts_lattice_invoke_pseudospin(latt, spin_from_int(pspin), i,j) );
  }
  else
  {
    tsmrb_handle_error(mrb,
        ts_lattice_invoke_pseudospin_without_update(latt,
          spin_from_int(pspin), i,j) );
  }

  return self;
}

static mrb_value cLattice_invoke_pseudospin(mrb_state *mrb, mrb_value self)
{
  return invoke_pseudospin(mrb, self, true);
}

static mrb_value cLattice_invoke_pseudospin_without_update(mrb_state *mrb, mrb_value self)
{
  return invoke_pseudospin(mrb, self, false);
}

static mrb_value cLattice_rollbackB(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  tsmrb_handle_error(mrb, 
      ts_lattice_rollback_changes(latt));
  return mrb_nil_value();
}

static MAYBE_UNUSED TSResult
callback_resize_states(struct callback_data *cb, int nstates)
{
  if (nstates > cb->nstates)
  {
    spin_t* spins = ts_realloc(NO_ALLOCATOR, cb->spins, nstates*sizeof(cb->spins[0]));
    if (spins == NULL) 
      return TS_NEW_ERROR( TS_ERR_OUT_OF_MEMORY );

    double* rates = ts_realloc(NO_ALLOCATOR, cb->rates, nstates*sizeof(cb->rates[0]));
    if (rates == NULL) 
      return TS_NEW_ERROR( TS_ERR_OUT_OF_MEMORY );

    cb->nstates = nstates;
    cb->spins  = spins;
    cb->rates  = rates;
  }

  return TS_OK;
}

KHASH_DECLARE(ht, mrb_value, mrb_value, 1)

int movetype_callback_func(void *data,
    ts_lattice* latt, const int *restrict rv, const int *restrict cv, size_t nsites,
    spin_t* states, double* rates, size_t *restrict siteinds, int move_max,
    intptr_t* cdata, TSResult* errp)
{
  struct callback_data* closure = data;
  size_t site = *cdata;
  int move = 0;
  mrb_state* mrb = closure->mrb;
  mrb_value args[3];
  int ai;

  ai = mrb_gc_arena_save(mrb);
  args[0] = mrb_obj_value(closure->rlatt);

  while ((site < nsites) && (move < move_max))
  {
    int ai2;
    mrb_value rmoves;

    ai2 = mrb_gc_arena_save(mrb);
    if ((rv[site] < 0) || (rv[site] >= latt->nrow) || (cv[site] < 0) || (cv[site] >= latt->ncol))
    {
      fprintf(stderr, "invalid coord %d, %d\n", rv[site],cv[site]);
    }

    args[1] = mrb_fixnum_value(rv[site]);
    args[2] = mrb_fixnum_value(cv[site]);

    rmoves = mrb_funcall_argv(mrb,
        mrb_obj_value(closure->proc),
        mrb_intern_cstr(mrb,"call"),
        3, args);

    if (mrb_hash_p(rmoves))
    { 
      mrb_value rsites = mrb_hash_keys(mrb, rmoves);
      int ind, nsitemoves;

      nsitemoves = mrb_ary_len(mrb,rsites);

      if (move + nsitemoves > move_max)
      {
        *cdata = site;
        if (move == 0)
        {
          *errp = TS_NEW_ERROR3(TS_ERR_INSUFFICIENT_SIZE,
              "move buffer size is too small to return any moves");
          move = -1;
        }
        else
        {
          move = -move;
        }

        goto finish;
      }

      for (ind = 0; ind < nsitemoves; ind++)
      {
        mrb_value key, val;
        int move_spin;
        double move_rate;

        key = mrb_ary_ref(mrb, rsites, ind);
        val = mrb_hash_get(mrb, rmoves, key);

        if (mrb_fixnum_p(key))
        {
          move_spin = mrb_fixnum(key);
        }
        else
        {
          *errp = TS_NEW_ERRORF(TS_ERR_INVALID_VALUE,
              "new state for site (%d,%d) must be a fixnum, not %s",
              rv[site],cv[site],mrb_obj_classname(mrb,key));
          move = -1;
          goto finish;
        }

        if (mrb_float_p(val))
        {
          move_rate = mrb_float(val);
        }
        else if (mrb_fixnum_p(val))
        {
          move_rate = mrb_fixnum_p(val);
        }
        else
        {
          *errp = TS_NEW_ERRORF(TS_ERR_INVALID_VALUE,
              "rate for site (%d,%d) -> spin %d must be a fixnum or a float, not %s",
              rv[site],cv[site],move_spin,mrb_obj_classname(mrb,val));
          move = -1;
          goto finish;
        }

        states[move+ind] = spin_from_int( move_spin );
        rates[move+ind] = move_rate;
        siteinds[move+ind] = site;
      }

      move += nsitemoves;
    }
    else if (!mrb_nil_p(rmoves))
    {
      mrb_raise(mrb,
          E_TYPE_ERROR, "move function must return a hash or nil");
    }

    mrb_gc_arena_restore(mrb,ai2);
    site++;
  }

finish:
  mrb_gc_arena_restore(mrb,ai);
  return move;
}

static mrb_value cLattice_add_movetype(mrb_state* mrb, mrb_value self)
{
  const char* name = NULL;
  mrb_value rproc = mrb_nil_value();
  mrb_get_args(mrb, "z&", &name, &rproc);

  ts_lattice* latt = tsmrb_to_lattice(mrb,self);

  if (mrb_nil_p(rproc))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Expected block or proc");

  struct callback_data* cb = ts_alloc(NO_ALLOCATOR, sizeof(*cb));
  memset(cb, 0, sizeof(*cb));

  cb->mrb = mrb;
  cb->rlatt = RDATA(self);
  cb->proc = mrb_proc_ptr(rproc); 

  struct ts_lattice_movetype movetype = {
    .name = ts_strdup(NO_ALLOCATOR, name),
    .func = movetype_callback_func,
    .closure = cb
  };

  add_callback_ref(mrb, self, rproc);

  tsmrb_handle_error(mrb, ts_lattice_add_movetype(latt, &movetype));

  return self;
}

static mrb_value cLattice_write_plot(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt;
  mrb_value io;
  FILE *f;

  io = tsmrb_get_stdout(mrb);
  mrb_get_args(mrb, "|o", &io);

  f = mrb_io_file(mrb, io);

  latt = tsmrb_to_lattice(mrb, self);
  tsmrb_handle_error(mrb, ts_lattice_write_plot(latt, f));
  return mrb_nil_value();
}

static mrb_value cLattice_read_plot(mrb_state *mrb, mrb_value self)
{
  ts_lattice* latt;
  mrb_value io;
  FILE *f;

  io = tsmrb_get_stdout(mrb);
  mrb_get_args(mrb, "|o", &io);

  f = mrb_io_file(mrb, io);

  latt = tsmrb_to_lattice(mrb, self);
  tsmrb_handle_error(mrb, ts_lattice_read_plot(latt,f));
  return mrb_nil_value();
}

static mrb_value cLattice_connected_components(mrb_state* mrb, mrb_value self)
{
  mrb_sym opt = 0;
  int has_opt = mrb_get_args(mrb, "|n", &opt);

  ts_lattice* latt = tsmrb_to_lattice(mrb, self);
  /* allocate temporary storage for connected components... */
  size_t working_size = ts_lattice_connected_components_working_size(latt);
  int* scratch = mrb_malloc(mrb, sizeof(scratch[0])*(working_size + latt->num_sites));
  int* working = scratch;
  int* labels  = working + working_size;
  memset(working, 0, working_size*sizeof(working[0]));
  memset(labels, 0, latt->num_sites*sizeof(labels[0]));

  int num = ts_lattice_connected_components(latt, labels, working);

  mrb_value ret;
  if (num > 0)
  {
    ret = mrb_ary_new_capa(mrb, num);
    for (int l=0; l < num; l++) { mrb_ary_push(mrb, ret, mrb_ary_new(mrb)); }

    mrb_value* rlbls = (mrb_value*)RARRAY_PTR(ret);
    size_t nrow = latt->nrow;
    size_t ncol = latt->ncol;
    size_t ind = 0;
    for (size_t j=0; j < ncol; j++) {
      for (size_t i=0; i < nrow; i++) {
        int lbl = labels[ind];
        if (lbl != 0) {
          int ai = mrb_gc_arena_save(mrb);
          mrb_value coords[2];
          coords[0] = mrb_fixnum_value(i);
          coords[1] = mrb_fixnum_value(j);
          mrb_value elt = mrb_ary_new_from_values(mrb, 2, coords);
          mrb_ary_push(mrb, rlbls[lbl-1], elt);
          mrb_gc_arena_restore(mrb,ai);
        }
        ind++;
      }
    }
  }
  else
  {
    ret = mrb_ary_new(mrb);
  }

  if (has_opt)
  {
    mrb_value counts = mrb_ary_new_capa(mrb,num);
    for (int k=0; k < num; k++)
    {
      mrb_ary_push(mrb, counts, mrb_fixnum_value(working[k]));
    }

    mrb_value elts[2] = { ret, counts };
    ret = mrb_ary_new_from_values(mrb, 2, elts);
  }

  mrb_free(mrb, scratch);
  return ret;

}

static mrb_value cLattice_find_row_range(mrb_state *mrb, mrb_value self)
{
  int i, *rowsp, rows[64];
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  TSResult err;
  size_t nelts = 2*latt->ncol;

  mrb_value ret = mrb_ary_new_capa(mrb,latt->ncol);
  if (nelts <= ARRAYLEN(rows))
  {
    rowsp = rows+0;
  }
  else
  {
    rowsp = mrb_malloc(mrb, nelts * sizeof(rowsp[0]));
  }

  err = ts_lattice_find_row_range(latt, rowsp);
  if (IS_ERROR(err))
  {
    if (rowsp != rows+0) { mrb_free(mrb, rowsp); }
    tsmrb_handle_error(mrb,err);
    return mrb_nil_value();
  }

  for(i=0; i < latt->ncol; i++)
  {
    mrb_value rrange[2];
    rrange[0] = mrb_fixnum_value(rowsp[2*i+0]);
    rrange[1] = mrb_fixnum_value(rowsp[2*i+1]);

    mrb_ary_set(mrb, ret, i, 
        mrb_ary_new_from_values(mrb, 2, rrange));
  }

  if (rowsp != rows+0)
  {
    mrb_free(mrb,rowsp);
  }

  return ret;
}

static mrb_value cLattice_find_last_rows(mrb_state *mrb, mrb_value self)
{
  int i, *rowsp, rows[32];
  ts_lattice* latt = tsmrb_to_lattice(mrb,self);
  TSResult err;

  mrb_value ret = mrb_ary_new_capa(mrb,latt->ncol);
  if (latt->ncol <= ARRAYLEN(rows))
  {
    rowsp = rows+0;
  }
  else
  {
    rowsp = mrb_malloc(mrb, latt->ncol * sizeof(rowsp[0]));
  }

  err = ts_lattice_find_last_rows(latt, rowsp);
  if (IS_ERROR(err))
  {
    if (latt->ncol > ARRAYLEN(rows)) { mrb_free(mrb, rowsp); }
    tsmrb_handle_error(mrb,err);
    return mrb_nil_value();
  }

  for(i=0; i < latt->ncol; i++)
  {
    mrb_ary_set(mrb, ret, i, mrb_fixnum_value(rowsp[i]));
  }

  if (latt->ncol > ARRAYLEN(rows))
  {
    mrb_free(mrb,rowsp);
  }

  return ret;
}

static mrb_value cLattice_get_user_config(mrb_state *mrb, mrb_value self)
{
  return mrb_iv_get(mrb, self, mrb_intern_lit(mrb,"@user_config"));
}

static mrb_value cLattice_set_user_config(mrb_state *mrb, mrb_value self)
{
  mrb_value val;
  mrb_get_args(mrb, "o", &val);

  mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@user_config"), val);
  return mrb_nil_value();
}

static mrb_value cLattice_statediffs(mrb_state *mrb, mrb_value self)
{
  mrb_value other, ret;
  ts_lattice *l1, *l2;
  spin_t *restrict st1, *restrict st2;
  size_t off,j,nr,nc;
  mrb_int maxdiff;

  l1 = tsmrb_to_lattice(mrb,self);

  other = mrb_nil_value();
  maxdiff = -1;
  mrb_get_args(mrb, "o|i", &other, &maxdiff);

  l2 = tsmrb_to_lattice(mrb,other);
  if (!l1 || !l2)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "lattice must not be NULL");
  }

  nr = l1->nrow;
  nc = l1->ncol;

  if (l2->nrow != nr || l2->ncol != nc)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "lattice sizes are not equal");
  }

  st1 = l1->sites;
  st2 = l2->sites;

  ret = mrb_ary_new(mrb);
  for (off=0,j=0; j < nc; j++,off+=nr)
  {
    size_t i;
    for (i=0; i < nr; i++)
    {
      if (st1[i+off].val != st2[i+off].val)
      {
        mrb_value coords[2];
        coords[0] = mrb_fixnum_value(i);
        coords[1] = mrb_fixnum_value(j);
        mrb_ary_push(mrb, ret, mrb_ary_new_from_values(mrb, 2, coords));

        if ((maxdiff > 0) && (mrb_ary_len(mrb,ret) > maxdiff))
        {
          return ret;
        }
      }
    }
  }

  return ret;
}

static void append_move(mrb_state *mrb, mrb_value arr, struct ts_lattice_move *mv)
{
  int ai;
  mrb_value tuple[5];

  ai = mrb_gc_arena_save(mrb);
  tuple[0] = mrb_fixnum_value( (mrb_int) mv->r );
  tuple[1] = mrb_fixnum_value( (mrb_int) mv->c );
  tuple[2] = mrb_fixnum_value( (mrb_int) mv->state.val );
  tuple[3] = mrb_float_value(mrb, (mrb_float) mv->rate );
  tuple[4] = mrb_bool_value( mv->next == NULL );

  mrb_ary_push(mrb, arr, mrb_ary_new_from_values(mrb, 5, tuple));
  mrb_gc_arena_restore(mrb,ai);
}

static mrb_value cLattice_history(mrb_state *mrb, mrb_value self)
{
  ts_lattice *latt;
  size_t i,last,first;
  mrb_value ret;

  latt = tsmrb_to_lattice(mrb,self);
  last  = latt->move_history.last;
  if (latt->move_history.moves[last].rate == 0.0) 
  {
    first = 0;
  }
  else
  {
    first = last;
    last = ARRAYLEN(latt->move_history.moves);
  }

  ret = mrb_ary_new_capa(mrb, ARRAYLEN(latt->move_history.moves));

  for (i=first; i < last; i++)
  {
    append_move(mrb, ret, &latt->move_history.moves[i]);
  }

  for (i=0; i < first; i++)
  {
    append_move(mrb, ret, &latt->move_history.moves[i]);
  }

  return ret;
}

static mrb_value cLattice_save_changes(mrb_state *mrb, mrb_value self)
{
  struct RClass *cLatticeChanges;

  cLatticeChanges = mrb_class_get_under(mrb,
      tsmrb_classes(mrb)->cLattice, "Changes");

  return mrb_obj_new(mrb, cLatticeChanges, 1, &self);
}

static mrb_value cLatticeChanges_initialize(mrb_state *mrb, mrb_value self)
{
  ts_lattice *latt;
  mrb_value rlatt;
  struct change_state *chg;

  rlatt = mrb_nil_value();
  mrb_get_args(mrb, "o", &rlatt);
  latt = tsmrb_to_lattice(mrb,rlatt);

  chg = mrb_malloc(mrb, sizeof *chg);
  chg->latt = ts_lattice_retain(latt);
  chg->changes = (struct ts_lattice_change_list){0};
  chg->updates = (struct ts_lattice_update_list){0};

  DATA_PTR(self) = chg;
  DATA_TYPE(self) = &changes_type;

  tsmrb_handle_error(mrb,
      ts_lattice_save_changes(latt, &chg->changes, &chg->updates));

  return mrb_nil_value();
}

static mrb_value cLatticeChanges_restoreB(mrb_state *mrb, mrb_value self)
{
  struct change_state *chg;
  chg = DATA_PTR(self);

  if (chg == NULL || chg->latt == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "invalid Lattice::Changes");
  }

  tsmrb_handle_error(mrb,
      ts_lattice_restore_changes(chg->latt, &chg->changes, &chg->updates));

  return mrb_nil_value();
}

struct RClass* tsmrb_define_lattice(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_lattice_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Lattice class has already been initialized!");

  struct RClass* cLattice = mrb_define_class_under(mrb,
      mod, "Lattice", mrb->object_class);
  MRB_SET_INSTANCE_TT(cLattice, MRB_TT_DATA);
  tsmrb_classes(mrb)->cLattice = cLattice;

  mrb_define_module_function(mrb, cLattice, "debug_updates", latt_get_debug_updates, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, cLattice, "debug_updates=", latt_set_debug_updates, MRB_ARGS_REQ(1));

  mrb_define_const(mrb, cLattice, "PSEUDO_OFFSET", mrb_fixnum_value(TS_LATTICE_PSEUDOSPIN_OFFSET));
  mrb_define_const(mrb, cLattice, "MAX_HISTORY", mrb_fixnum_value(TS_LATTICE_MOVE_HISTORY_SIZE));

  mrb_define_method(mrb, cLattice, "initialize", cLattice_initialize, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cLattice, "size", cLattice_size, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "nrow", cLattice_nrow, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "ncol", cLattice_ncol, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "time", cLattice_get_time, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "nsteps", cLattice_get_nsteps, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "row_coupling_dist",  cLattice_get_row_coupling_dist, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "row_coupling_dist=", cLattice_set_row_coupling_dist, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "col_coupling_dist",  cLattice_get_col_coupling_dist, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "col_coupling_dist=", cLattice_set_col_coupling_dist, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "temp",  cLattice_get_temp, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "temp=", cLattice_set_temp, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "[]",   cLattice_at  , MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cLattice, "[]=",  cLattice_set , MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cLattice, "state_string",  cLattice_get_state_string, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "state_string=", cLattice_set_state_string, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "wrap_coords", cLattice_wrap_coords, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cLattice, "change_spin", cLattice_change_spin, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cLattice, "rng", cLattice_get_rng, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "rng=", cLattice_set_rng, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "rebuild_moves!", cLattice_rebuild_movesB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "changed_site!", cLattice_changed_siteB, MRB_ARGS_REST());
  mrb_define_method(mrb, cLattice, "update_site!", cLattice_update_siteB, MRB_ARGS_REST());
  mrb_define_method(mrb, cLattice, "update_range!", cLattice_update_rangeB, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cLattice, "k_total", cLattice_get_k_total, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "get_site_moves", cLattice_get_site_moves, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cLattice, "all_moves", cLattice_get_all_moves, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "each_move", cLattice_each_move, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "single_step", cLattice_single_step, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "simulate_until", cLattice_simulate_until, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "reset!", cLattice_resetB, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "zero!", cLattice_zeroB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "add_movetype", cLattice_add_movetype, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cLattice, "add_pseudospin", cLattice_add_pseudospin, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cLattice, "num_pseudospins", cLattice_num_pseudospins, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "add_move_hook", cLattice_add_move_hook,
      MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_method(mrb, cLattice, "add_post_move_hook", cLattice_add_post_move_hook,
      MRB_ARGS_NONE() | MRB_ARGS_BLOCK());

  mrb_define_method(mrb, cLattice, "pseudospin?", cLattice_is_pseudospin, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cLattice, "invoke_pseudospin", cLattice_invoke_pseudospin, MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cLattice, "invoke_pseudospin_without_update",
      cLattice_invoke_pseudospin_without_update, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cLattice, "rollback!", cLattice_rollbackB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "find_last_rows", cLattice_find_last_rows, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "find_row_range", cLattice_find_row_range, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "write_plot", cLattice_write_plot, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cLattice, "read_plot", cLattice_read_plot, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cLattice, "connected_components", cLattice_connected_components, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cLattice, "save_changes", cLattice_save_changes, MRB_ARGS_NONE());

  mrb_define_method(mrb, cLattice, "user_config",  cLattice_get_user_config, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLattice, "user_config=", cLattice_set_user_config, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "statediffs", cLattice_statediffs, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cLattice, "history", cLattice_history, MRB_ARGS_NONE());

  struct RClass* cLatticeChanges = mrb_define_class_under(mrb,
      cLattice, "Changes", mrb->object_class);
  MRB_SET_INSTANCE_TT(cLatticeChanges, MRB_TT_DATA);

  mrb_define_method(mrb, cLatticeChanges, "initialize", cLatticeChanges_initialize, MRB_ARGS_NONE());
  mrb_define_method(mrb, cLatticeChanges, "restore!", cLatticeChanges_restoreB, MRB_ARGS_NONE());

  return cLattice;
}

