#include <stdio.h>

#include <mruby.h>
#include <mruby/array.h>

#include <tubesheet.h>
#include <ts_units.h>
#include <ts_common_algo.h>
#include <vector3.h>

#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"
#include "tsmrb_mat3x3.h"
#include "tsmrb_quat.h"
#include "tsmrb_numerical.h"
#include "tsmrb_veclist.h"

#include "tsmrb_rng.h"
#include "tsmrb_sim.h"
#include "tsmrb_potentials.h"

#include "tsmrb_convenience.h"

#include "tsmrb_fenwick.h"
#include "tsmrb_lattice.h"
#include "tsmrb_chem_driver.h"
#include "tsmrb_callbacks.h"
#include "tsmrb_io.h"
#include "tsmrb_stats.h"
#include "tsmrb_units.h"
#include "tsmrb_dispatch.h"

static mrb_value mTubesheet_vector3_tests(mrb_state* mrb, mrb_value obj)
{
  vector3_vector_tests();
  return mrb_nil_value();
}

struct compare_closure
{
  mrb_state* mrb;

  mrb_sym cmp_id;
  mrb_sym eql_id;
  mrb_sym lt_id;
};

static int compare_ruby_values(
    struct compare_closure* cl,
    mrb_value*restrict a,
    mrb_value*restrict b)
{
  if (mrb_respond_to(cl->mrb, *a, cl->cmp_id))
  {
    return mrb_fixnum(
        mrb_funcall_argv(cl->mrb, *a, cl->cmp_id, 1, b));
  }
  else {
    if (mrb_test(mrb_funcall_argv(cl->mrb,*a,cl->lt_id,1,b)))
    {
      return -1;
    }

    return mrb_test(mrb_funcall_argv(cl->mrb,*a,cl->eql_id,1,b))
      ? 0
      : 1;
  }
}

void setup_compare_closure(mrb_state* mrb, struct compare_closure* cl)
{
  cl->mrb = mrb;
  cl->cmp_id = mrb_intern_lit(mrb,"<=>");
  cl->eql_id = mrb_intern_lit(mrb,"==");
  cl->lt_id  = mrb_intern_lit(mrb,"<");
}

enum SORT_TYPE {
  SORT_INSERTION = 1,
  SORT_QUICK     = 2,
  SORT_DEFAULT   = 3,
};

static inline mrb_value mTubesheet_do_sort(mrb_state* mrb, mrb_value mod, enum SORT_TYPE sort_type)
{
  mrb_value* base0 = NULL;
  mrb_int n = 0;

  mrb_get_args(mrb, "a", &base0, &n);

  mrb_value* base1 =  mrb_calloc(mrb, n, sizeof(base1[0]));
  memcpy(base1,base0, n*sizeof(base0[0]));

  struct compare_closure cl;
  setup_compare_closure(mrb,&cl);

  switch(sort_type)
  {
    case SORT_INSERTION:
      insertion_sort_elements(base1, n, sizeof(base1[0]),
          &cl, (GenericCompareFunc)compare_ruby_values);
      break;
    case SORT_QUICK:
      quick_sort_elements(base1, n, sizeof(base1[0]),
          &cl, (GenericCompareFunc)compare_ruby_values);
      break;
    case SORT_DEFAULT:
      sort_elements(base1, n, sizeof(base1[0]),
          &cl, (GenericCompareFunc)compare_ruby_values);
      break;
  }

  mrb_value new_ary = mrb_ary_new_from_values(mrb, n, base1);

  mrb_free(mrb,base1);
  return new_ary;
}

static mrb_value mTubesheet_isort_elements(mrb_state* mrb, mrb_value mod)
{
  return mTubesheet_do_sort(mrb, mod, SORT_INSERTION);
}

static mrb_value mTubesheet_qsort_elements(mrb_state* mrb, mrb_value mod)
{
  return mTubesheet_do_sort(mrb, mod, SORT_QUICK);
}

static mrb_value mTubesheet_sort_elements(mrb_state* mrb, mrb_value mod)
{
  return mTubesheet_do_sort(mrb, mod, SORT_DEFAULT);
}

static mrb_value mTubesheet_isort_sizes(mrb_state *mrb, mrb_value mod)
{
  mrb_value *elts, ret;
  mrb_int i, nelts;
  size_t *restrict list;

  mrb_get_args(mrb, "a", &elts, &nelts);

  list = mrb_malloc(mrb, nelts*sizeof(*list));

  for (i=0; i < nelts; i++)
  {
    list[i] = mrb_fixnum(elts[i]);
  }

  insertion_sort_sizes(list, nelts);

  ret = mrb_ary_new_capa(mrb, nelts);
  for (i=0; i < nelts; i++)
  {
    mrb_ary_push(mrb, ret, mrb_fixnum_value((mrb_int) list[i]));
  }

  mrb_free(mrb, list);
  return ret;
}

static mrb_value mTubesheet_qsort_sizes(mrb_state *mrb, mrb_value mod)
{
  mrb_value *elts, ret;
  mrb_int i, nelts;
  size_t *restrict list;

  mrb_get_args(mrb, "a", &elts, &nelts);

  list = mrb_malloc(mrb, nelts*sizeof(*list));

  for (i=0; i < nelts; i++)
  {
    list[i] = mrb_fixnum(elts[i]);
  }

  quick_sort_sizes(list, nelts);

  ret = mrb_ary_new_capa(mrb, nelts);
  for (i=0; i < nelts; i++)
  {
    mrb_ary_push(mrb, ret, mrb_fixnum_value((mrb_int) list[i]));
  }

  mrb_free(mrb, list);
  return ret;
}

static mrb_value mTubesheet_msort_sizes(mrb_state *mrb, mrb_value mod)
{
  mrb_value *elts, ret;
  mrb_int i, nelts;
  size_t *restrict list;
  size_t *restrict sorted;

  mrb_get_args(mrb, "a", &elts, &nelts);

  list = mrb_malloc(mrb, 2*nelts*sizeof(*list));

  for (i=0; i < nelts; i++)
  {
    list[i] = mrb_fixnum(elts[i]);
  }

  sorted = merge_sort_sizes(list, list+nelts, nelts);

  ret = mrb_ary_new_capa(mrb, nelts);
  for (i=0; i < nelts; i++)
  {
    mrb_ary_push(mrb, ret,
        mrb_fixnum_value((mrb_int) sorted[i]));
  }

  mrb_free(mrb, list);
  return ret;
}

static mrb_value mTubesheet_msort_u32(mrb_state *mrb, mrb_value mod)
{
  mrb_value *elts, ret;
  mrb_int i, nelts;
  uint32_t *restrict list;
  uint32_t *restrict sorted;

  mrb_get_args(mrb, "a", &elts, &nelts);

  list = mrb_malloc(mrb, 2*nelts*sizeof *list);

  for (i=0; i < nelts; i++)
  {
    list[i] = mrb_fixnum(elts[i]);
  }

  sorted = merge_sort_u32(list, list+nelts, nelts);

  ret = mrb_ary_new_capa(mrb, nelts);
  for (i=0; i < nelts; i++)
  {
    mrb_ary_push(mrb, ret,
        mrb_fixnum_value((mrb_int) sorted[i]));
  }

  mrb_free(mrb, list);
  return ret;
}

static mrb_value mTubesheet_trigger_error(mrb_state* mrb, mrb_value mod)
{
  TSResult err = TS_NEW_ERROR(TS_ERR_UNSUPPORTED_OPERATION);
  tsmrb_handle_error(mrb, err);
  return mrb_nil_value();
}

static mrb_value mTubesheet_sort_indices(mrb_state* mrb, mrb_value mod)
{
  mrb_value ary;
  mrb_get_args(mrb, "A", &ary);

  size_t nelts = RARRAY_LEN(ary);
  const mrb_value* base = RARRAY_PTR(ary);

  struct compare_closure cl;
  setup_compare_closure(mrb,&cl);

  size_t*restrict inds = mrb_calloc(mrb, nelts, sizeof(*inds));
  sort_indices( (void*)base, inds, nelts, sizeof(base[0]), 
      &cl, (GenericCompareFunc)compare_ruby_values );

  mrb_value sorted_inds = mrb_ary_new_capa(mrb, nelts);
  mrb_value* vals = (mrb_value*)(RARRAY_PTR(sorted_inds));
  for (size_t i=0; i < nelts; i++) { 
    vals[i] = mrb_fixnum_value(inds[i]);
  }
  RARRAY_LEN(sorted_inds) = nelts;
  mrb_free(mrb,inds);

  return sorted_inds;
}

static mrb_value 
mTubesheet_msb(mrb_state* mrb, mrb_value mod)
{
  mrb_int i;
  mrb_get_args(mrb, "i", &i);
  return mrb_fixnum_value( msb(i) );
}

static mrb_value 
mTubesheet_lsb(mrb_state* mrb, mrb_value mod)
{
  mrb_int i;
  mrb_get_args(mrb, "i", &i);
  return mrb_fixnum_value( lsb(i) );
}

static mrb_value 
mTubesheet_strip_lsb(mrb_state* mrb, mrb_value mod)
{
  mrb_int i;
  mrb_get_args(mrb, "i", &i);
  return mrb_fixnum_value( strip_lsb(i) );
}

static mrb_value
mTubesheet_is_embedded(mrb_state* mrb, mrb_value mod)
{
  return mrb_bool_value(1);
}

enum { ROUNDING_SYM_MAX = 4 };
static const char* rounding_sym[ROUNDING_SYM_MAX+1] = {
  "invalid", "nearest", "tozero", "upward", "downward"
};

static mrb_value mTubesheet_get_rounding(mrb_state* mrb, mrb_value mod) {
  int r = ts_get_rounding();

  if ((r == TS_ROUND_INVALID) || (r < 0) || (r > ROUNDING_SYM_MAX)) {
    tsmrb_raisef(mrb,E_RUNTIME_ERROR,
        "Error in math library!  Invalid rounding mode (%d)", r);
  }

  return mrb_symbol_value(mrb_intern_cstr(mrb,rounding_sym[r]));
}

static mrb_value mTubesheet_set_rounding(mrb_state* mrb, mrb_value mod)
{
  mrb_sym rmode;
  mrb_get_args(mrb, "n", &rmode);
  mrb_value prev = mTubesheet_get_rounding(mrb,mod);

  const char* sym = mrb_sym2name(mrb,rmode);
  if (strcmp(sym,"default") == 0) {
    ts_set_rounding_default();
    return prev;
  }

  for (size_t i=1; i <= ROUNDING_SYM_MAX; i++) {
    if (strcmp(sym,rounding_sym[i]) == 0) {
      tsmrb_handle_error(mrb, ts_set_rounding(i) );
      return prev;
    }
  }

  tsmrb_raisef(mrb,E_ARGUMENT_ERROR, "unknown rounding mode :%s", sym);
  return mrb_nil_value();
}

static mrb_value mTubesheet_ieee_round(mrb_state* mrb, mrb_value mod)
{
  mrb_float num;
  mrb_get_args(mrb, "f", &num);

  double n = num;
  return mrb_float_value(mrb,(mrb_float)rint(n));
}

mrb_value mTubesheet_debug_break(mrb_state *mrb, mrb_value mod)
{
  mrb_int par;

  volatile int value = 0;

  par = 0;
  if (mrb_get_args(mrb,"|i", &par) > 0)
  {
    value = par;
  }

  return mod;
}

static void define_scientific_constants(mrb_state* mrb, struct RClass* module)
{
  mrb_define_const(mrb, module, "GAS_CONSTANT", mrb_float_value(mrb, GAS_CONSTANT));
}

static struct RClass* define_tubesheet_module(mrb_state* mrb)
{
  struct RClass *mTubesheet = mrb_define_module(mrb, "Tubesheet");

  mrb_define_module_function(mrb, mTubesheet, 
      "vector3_tests", mTubesheet_vector3_tests, MRB_ARGS_REQ(0));

  mrb_define_module_function(mrb, mTubesheet, 
      "isort", mTubesheet_isort_elements, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet, 
      "qsort", mTubesheet_qsort_elements, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet, 
      "sort", mTubesheet_sort_elements, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "qsort_sizes", mTubesheet_qsort_sizes, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "isort_sizes", mTubesheet_isort_sizes, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "msort_sizes", mTubesheet_msort_sizes, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "msort_u32", mTubesheet_msort_u32, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "debug_break", mTubesheet_debug_break, MRB_ARGS_OPT(1));

  define_scientific_constants(mrb, mTubesheet);

  mrb_define_module_function(mrb, mTubesheet,
      "sort_indices", mTubesheet_sort_indices, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet, 
      "lsb", mTubesheet_lsb, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mTubesheet, 
      "msb", mTubesheet_msb, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mTubesheet, 
      "strip_lsb", mTubesheet_strip_lsb, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mTubesheet,
      "is_embedded?", mTubesheet_is_embedded, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mTubesheet,
      "trigger_error", mTubesheet_trigger_error, MRB_ARGS_NONE());

  tsmrb_define_dispatch(mrb, mTubesheet);

  return mTubesheet;
}

static const char *banner_string(void)
{
  static char buf[1024] = { 0 };
  if (buf[0] == 0)
  {
    snprintf(buf,sizeof(buf),
        "Tube/Sheet simulation version %s\n"
        "Repository SHA1 %s\n"
        "Copyright (C) 2012-2015, Shannon F. Stewman, Ao Ma, and others\n"
        "See documentation for associated licenses\n"
        "All Rights Reserved\n\n",
        tubesheet_repo_version,tubesheet_repo_sha1);
  }

  return buf;
}

static mrb_value tsmrb_banner_string(mrb_state *mrb)
{
  static const char *banner = NULL;
  static size_t blen = 0;
  if (!banner)
  {
    banner = banner_string();
    blen = strlen(banner);
  }

  return mrb_str_new_static(mrb, banner, blen);
}

static mrb_value mTubesheet_banner_string(mrb_state *mrb, mrb_value mod)
{
  UNUSED_PARAM(mod);
  return tsmrb_banner_string(mrb);
}

static void print_banner(mrb_state *mrb)
{
  UNUSED_PARAM(mrb);
  puts(banner_string());
}

static mrb_value mTubesheet_print_banner(mrb_state *mrb, mrb_value mod)
{
  UNUSED_PARAM(mod);
  print_banner(mrb);
  return mrb_nil_value();
}

extern struct RClass* tsmrb_define_json(mrb_state* mrb, struct RClass* mod);

extern void numarray_mrb_initialize(mrb_state* mrb);

extern struct RClass *tsmrb_define_tsmf(mrb_state *mrb, struct RClass *mod);

void
mrb_mruby_tubesheet_gem_init(mrb_state* mrb)
{
  numarray_mrb_initialize(mrb);

  tsmrb_init(mrb);
  tsmrb_symtab_init(mrb);

  tsmrb_define_fenwick_tree(mrb);

  struct RClass* mod_tubesheet = define_tubesheet_module(mrb);
  tsmrb_classes(mrb)->mTubesheet = mod_tubesheet;

  tsmrb_helpers_init(mrb, mod_tubesheet);

  mrb_define_const(mrb, mod_tubesheet, "TOL", mrb_float_value(mrb,VEC_TOL));
  mrb_define_const(mrb, mod_tubesheet, "SMALL", mrb_float_value(mrb,SMALL));
#if VECTOR_SINGLE_PRECISION
  mrb_define_const(mrb,mod_tubesheet, "PRECISION", mrb_symbol_value(mrb_intern_lit(mrb,"single")));
#else /* VECTOR_SINGLE_PRECISION */
  mrb_define_const(mrb,mod_tubesheet, "PRECISION", mrb_symbol_value(mrb_intern_lit(mrb,"double")));
#endif /* VECTOR_SINGLE_PRECISION */

  mrb_define_const(mrb, mod_tubesheet, "VERSION", 
      mrb_str_new_cstr(mrb, tubesheet_repo_version));
  mrb_define_const(mrb, mod_tubesheet, "VERS_SHA1", 
      mrb_str_new_cstr(mrb, tubesheet_repo_sha1));

  mrb_define_module_function(mrb, mod_tubesheet, "banner_string", mTubesheet_banner_string, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mod_tubesheet, "print_banner", mTubesheet_print_banner, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mod_tubesheet, "rounding", mTubesheet_get_rounding, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mod_tubesheet, "rounding=", mTubesheet_set_rounding, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mod_tubesheet, "ieee_round", mTubesheet_ieee_round, MRB_ARGS_REQ(1));

  tsmrb_define_vec3(mrb, mod_tubesheet);
  tsmrb_define_mat3x3(mrb, mod_tubesheet);
  tsmrb_define_quat(mrb, mod_tubesheet);
  tsmrb_define_veclist(mrb, mod_tubesheet);

  tsmrb_define_json(mrb, mod_tubesheet);

  tsmrb_define_rng(mrb, mod_tubesheet);
  tsmrb_define_sim(mrb, mod_tubesheet);

  tsmrb_define_lattice(mrb, mod_tubesheet);
  tsmrb_define_chem_driver(mrb, mod_tubesheet);

  tsmrb_define_stat_accumulator(mrb, mod_tubesheet);

  tsmrb_define_io(mrb, mod_tubesheet);
  tsmrb_define_tsmf(mrb, mod_tubesheet);

  tsmrb_define_units(mrb, mod_tubesheet);

  tsmrb_define_potentials(mrb, mod_tubesheet);

  tsmrb_define_numerical(mrb, mod_tubesheet);

  tsmrb_define_callbacks(mrb, mod_tubesheet);

  tsmrb_define_convenience(mrb, mod_tubesheet);

  {
    int no_banner;
    const char *setting;
    const char *true_values[] = {
      "1", "T", "t", "TRUE", "true", "True", "YES", "yes", "Yes", NULL
    };

    no_banner=0;
    setting = getenv("TS_NO_BANNER");
    if (setting)
    {
      const char **vp;
      for (vp=true_values; *vp; vp++)
      {
        if (strcmp(setting,*vp) == 0)
        {
          no_banner=1;
          break;
        }
      }
    }

    if (!no_banner)
    {
      print_banner(mrb);
    }
  }

  mrb_c_provide(mrb, "tubesheet");
}

void
mrb_mruby_tubesheet_gem_final(mrb_state* mrb)
{
}

