#include "tsmrb_mat3x3.h"

#include <stdlib.h>
#include <string.h>

#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"
#include "tsmrb_quat.h"

#include "mruby/data.h"
#include "mruby/array.h"
#include "mruby/string.h"

static const mrb_data_type mat3x3_type = {
  .struct_name = "Mat3x3",
  .dfree = tsmrb_free,
};

mat3x3* tsmrb_to_mat3x3(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_mat3x3_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Mat3x3 object");

  return DATA_PTR(obj);
}

static mat3x3* mat3x3_alloc(mrb_state* mrb)
{
  mat3x3* m = mrb_malloc(mrb, sizeof(*m));
  memset(m,0,sizeof(*m));
  return m;
}

static mrb_value mat3x3_wrap(mrb_state* mrb, mat3x3* mat)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_mat3x3_class(mrb), mat, &mat3x3_type);
  return mrb_obj_value(data);
}

static mrb_value create_new_mat3x3(mrb_state* mrb, mat3x3** matp)
{
  *matp = mat3x3_alloc(mrb);
  return mat3x3_wrap(mrb, *matp);
}

mrb_value tsmrb_new_mat3x3(mrb_state* mrb, mat3x3** matp)
{
  return create_new_mat3x3(mrb,matp);
}

mrb_value tsmrb_from_mat3x3(mrb_state* mrb, const mat3x3*restrict mat)
{
  mat3x3* mat_new = NULL;
  mrb_value rmat = create_new_mat3x3(mrb, &mat_new);
  mat3x3_set(mat_new, mat);

  return rmat;
}

static mrb_value cMat3x3_initialize(mrb_state* mrb, mrb_value self)
{
  mat3x3* mat = mat3x3_alloc(mrb);
  mat3x3_zero(mat);

  DATA_TYPE(self) = &mat3x3_type;
  DATA_PTR(self) = mat;

  mrb_value* argv = NULL;
  int argc = 0;
  mrb_get_args(mrb, "*", &argv, &argc);

  if (argc == 0)
  {
    mat3x3_zero(mat);
  }
  else if (argc == 1)
  {
    mrb_check_type(mrb, argv[0], MRB_TT_ARRAY);

    if (RARRAY_LEN(argv[0]) != 3)
      mrb_raise(mrb, E_ARGUMENT_ERROR, "argument must be a 3x3 array");

    for (int i=0; i < 3; i++) {
      mrb_value row = RARRAY_PTR(argv[0])[i];
      mrb_check_type(mrb, row, MRB_TT_ARRAY);
      if (RARRAY_LEN(row) != 3)
        mrb_raise(mrb, E_ARGUMENT_ERROR, "argument must be a 3x3 array");
    }

    for (int i=0; i < 3; i++)
    {
      mrb_value row = RARRAY_PTR(argv[0])[i];
      for (int j=0; j < 3; j++)
        MAT3X3_ELT(mat,i,j) = (VecReal)
          mrb_to_float(mrb, RARRAY_PTR(row)[j]);
    }
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Mat3x3.initialize requires either 0 or 1 argument");
  }

  return self;
}

static mrb_value cMat3x3_identity(mrb_state* mrb, mrb_value mod)
{
  mat3x3* mat = NULL;
  mrb_value rmat = tsmrb_new_mat3x3(mrb,&mat);

  mat3x3_identity(mat);
  return rmat;
}

static enum MAT3X3_ROT_AXIS get_rot_axis(mrb_state* mrb, mrb_sym axis)
{
  if (axis == TSMRB_SYM(X)) { return ROT_X_AXIS; }
  if (axis == TSMRB_SYM(Y)) { return ROT_Y_AXIS; }
  if (axis == TSMRB_SYM(Z)) { return ROT_Z_AXIS; }
  if (axis == TSMRB_SYM(DX)) { return ROT_DX_AXIS; }
  if (axis == TSMRB_SYM(DY)) { return ROT_DY_AXIS; }
  if (axis == TSMRB_SYM(DZ)) { return ROT_DZ_AXIS; }
  if (axis == TSMRB_SYM(DX2)) { return ROT_DX2_AXIS; }
  if (axis == TSMRB_SYM(DY2)) { return ROT_DY2_AXIS; }
  if (axis == TSMRB_SYM(DZ2)) { return ROT_DZ2_AXIS; }
  mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown axis");
  return ROT_NONE;
}

static mrb_value cMat3x3_axis_rotate(mrb_state* mrb, mrb_value kls)
{
  mrb_sym axis;
  mrb_float ang;
  mrb_get_args(mrb, "nf", &axis, &ang);

  mat3x3* mat = NULL;
  mrb_value r_mat = tsmrb_new_mat3x3(mrb,&mat);
  mat3x3_axis_rotate(mat, get_rot_axis(mrb,axis), (VecReal)ang);
  return r_mat;
}

static mrb_value cMat3x3_to_a(mrb_state* mrb, mrb_value self)
{
  mat3x3* mat = tsmrb_to_mat3x3(mrb,self);

  mrb_value elts[3];
  mrb_value rows[3];
  for (int i=0; i < 3; i++)
  {
    for (int j=0; j < 3; j++)
      elts[j] = mrb_float_value(mrb, mat3x3_elt(mat,i,j));
    rows[i] = mrb_ary_new_from_values(mrb, 3, elts);
  }

  return mrb_ary_new_from_values(mrb, 3, rows);
}

static mrb_value cMat3x3_inspect(mrb_state* mrb, mrb_value self)
{
  char buf[64];
  mat3x3* mat = tsmrb_to_mat3x3(mrb,self);

  snprintf(buf,sizeof(buf), "Mat3x3 %p\n", mrb_ptr(self));
  mrb_value s = mrb_str_new(mrb, buf, strlen(buf));

  for (int i=0; i < 3; i++)
  {
    snprintf(buf,sizeof(buf), "  [ %7.4f %7.4f %7.4f ]\n",
        mat3x3_elt(mat, i, 0),
        mat3x3_elt(mat, i, 1),
        mat3x3_elt(mat, i, 2));
    s = mrb_str_cat_cstr(mrb, s, buf);
  }

  return s;
}

static mrb_value cMat3x3_identityB(mrb_state* mrb, mrb_value self)
{
  mat3x3* mat = tsmrb_to_mat3x3(mrb,self);
  mat3x3_identity(mat);
  return self;
}

static inline void check_indices(mrb_state* mrb, int i, int j)
{
  if ((i<0) || (i>3))
    mrb_raise(mrb, E_RANGE_ERROR,
        "index i out of bounds");

  if ((j<0) || (j>3))
    mrb_raise(mrb, E_RANGE_ERROR,
        "index j out of bounds");
}

static mrb_value cMat3x3_get_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  check_indices(mrb,i,j);

  const mat3x3* m = tsmrb_to_mat3x3(mrb,self);

  return mrb_float_value(mrb,
      (mrb_float)MAT3X3_ELT(m,i,j) );
}

static mrb_value cMat3x3_set_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_float elt = 0.0;
  mrb_get_args(mrb,"iif", &i, &j, &elt);

  mat3x3* m = tsmrb_to_mat3x3(mrb,self);

  check_indices(mrb,i,j);

  MAT3X3_ELT(m,i,j) = (VecReal)elt;

  return self;
}

static mrb_value cMat3x3_add(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  const mat3x3* m1 = tsmrb_to_mat3x3(mrb,self);
  const mat3x3* m2 = tsmrb_to_mat3x3(mrb,other);

  mat3x3* mr = NULL;

  mrb_value r_mr = tsmrb_new_mat3x3(mrb,&mr);
  mat3x3_add(mr, m1,m2);

  return r_mr;
}

static mrb_value cMat3x3_sub(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  const mat3x3* m1 = tsmrb_to_mat3x3(mrb,self);
  const mat3x3* m2 = tsmrb_to_mat3x3(mrb,other);

  mat3x3* mr = NULL;

  mrb_value r_mr = tsmrb_new_mat3x3(mrb,&mr);
  mat3x3_sub(mr, m1,m2);

  return r_mr;
}

static mrb_value cMat3x3_add_toB(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  mat3x3* m1 = tsmrb_to_mat3x3(mrb,self);
  const mat3x3* m2 = tsmrb_to_mat3x3(mrb,other);

  mat3x3_add_to(m1,m2);

  return self;
}

static mrb_value cMat3x3_transpose(mrb_state* mrb, mrb_value self)
{
  const mat3x3* m = tsmrb_to_mat3x3(mrb,self);

  mat3x3* mt = NULL;

  mrb_value r_mt = tsmrb_new_mat3x3(mrb,&mt);
  mat3x3_set_transpose(mt, m);

  return r_mt;
}

static mrb_value
new_mat3x3(mrb_state* mrb, mrb_value mod)
{
  mrb_value* rest = NULL;
  int nrest = 0;
  mrb_get_args(mrb, "*", &rest, &nrest);

  return mrb_funcall_argv(mrb,
      mrb_obj_value(tsmrb_mat3x3_class(mrb)),
      mrb_intern_cstr(mrb, "new"),
      nrest,
      rest);
}

static mrb_value cMat3x3_rmulB(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  mat3x3* m1 = tsmrb_to_mat3x3(mrb,self);

  switch ( mrb_type(other) )
  {
    case MRB_TT_FIXNUM:
    case MRB_TT_FLOAT:
      mat3x3_scale_by(m1, mrb_to_float(mrb, other));
      break;

    default:
      mrb_raise(mrb, E_TYPE_ERROR,
          "illegal type for in-place right-multiply");
  }

  return self;
}

static mrb_value cMat3x3_rmul(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  mat3x3* morig = tsmrb_to_mat3x3(mrb,self);
  mrb_value result = mrb_nil_value();

  if (mrb_numeric_p(other))
  {
    mat3x3* mnew = NULL;
    result = tsmrb_new_mat3x3(mrb, &mnew);
    mat3x3_set_scaled(mnew, mrb_to_float(mrb, other), morig);
  }
  else if (mrb_obj_is_instance_of(mrb,other, tsmrb_vec3_class(mrb)))
  {
    vector3* vorig = tsmrb_to_vec3(mrb,other);
    vector3* vnew = NULL;
    result = tsmrb_new_vec3(mrb,&vnew);
    vector3_mat_rmul(vnew, morig,vorig);
  }
  else if (mrb_obj_is_instance_of(mrb,other, tsmrb_mat3x3_class(mrb)))
  {
    mat3x3* mother = tsmrb_to_mat3x3(mrb,other);
    mat3x3* mr = NULL;
    result = tsmrb_new_mat3x3(mrb,&mr);
    if (morig == mother)
      mat3x3_mul_self(mr, morig);
    else
      mat3x3_mul(mr, morig,mother);
  }
  else
  {
    mrb_raise(mrb, E_TYPE_ERROR, "illegal type for right-multiply");
  }

  return result;
}

static mrb_value cMat3x3_rmul_trans(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  mat3x3* morig = tsmrb_to_mat3x3(mrb,self);
  mat3x3* mother = tsmrb_to_mat3x3(mrb,other);

  mat3x3* mr = NULL;
  mrb_value result = tsmrb_new_mat3x3(mrb,&mr);

  if (morig == mother)
  {
    mat3x3_rmul_self_trans(mr, morig);
  }
  else
  {
    mat3x3_rmul_trans(mr, morig, mother);
  }

  return result;
}

static mrb_value cMat3x3_lmul_trans(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  mat3x3* morig = tsmrb_to_mat3x3(mrb,self);
  mat3x3* mother = tsmrb_to_mat3x3(mrb,other);

  mat3x3* mr = NULL;
  mrb_value result = tsmrb_new_mat3x3(mrb,&mr);

  if (morig == mother)
  {
    mat3x3_lmul_self_trans(mr, morig);
  }
  else
  {
    mat3x3_lmul_trans(mr, morig, mother);
  }

  return result;
}

static mrb_value cMat3x3_from_quat(mrb_state* mrb, mrb_value self)
{
  mrb_value rq;
  mrb_get_args(mrb, "o", &rq);

  mat3x3* mr = NULL;
  mrb_value result = tsmrb_new_mat3x3(mrb, &mr);
  quat4* q = tsmrb_to_quat(mrb, rq);

  mat3x3_from_quat(q, mr);
  return result;
}

struct RClass* tsmrb_define_mat3x3(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_mat3x3_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Mat3x3 class has already been initialized!");

  struct RClass* mat3x3_class =
    mrb_define_class_under(mrb, mod, "Mat3x3", mrb->object_class);
  tsmrb_classes(mrb)->cMat3x3 = mat3x3_class;
  MRB_SET_INSTANCE_TT(mat3x3_class, MRB_TT_DATA);

  mrb_define_method(mrb, mat3x3_class, "initialize", cMat3x3_initialize, MRB_ARGS_ANY());

  mrb_define_singleton_method(mrb, (struct RObject*)mat3x3_class,
      "identity", cMat3x3_identity, MRB_ARGS_NONE());

  mrb_define_method(mrb, mat3x3_class, "identity!", cMat3x3_identityB, MRB_ARGS_NONE());

  mrb_define_singleton_method(mrb, (struct RObject*)mat3x3_class,
      "axis_rotate", cMat3x3_axis_rotate, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, mat3x3_class, "[]" , cMat3x3_get_element, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, mat3x3_class, "[]=", cMat3x3_set_element, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, mat3x3_class, "+", cMat3x3_add, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, mat3x3_class, "add_to!", cMat3x3_add_toB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, mat3x3_class, "mul!", cMat3x3_rmulB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, mat3x3_class, "*", cMat3x3_rmul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, mat3x3_class, "-", cMat3x3_sub, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, mat3x3_class, "transpose", cMat3x3_transpose, MRB_ARGS_NONE());
  mrb_define_method(mrb, mat3x3_class, "t", cMat3x3_transpose, MRB_ARGS_NONE());

  mrb_define_method(mrb, mat3x3_class, "to_a", cMat3x3_to_a, MRB_ARGS_NONE());

  mrb_define_method(mrb, mat3x3_class, "inspect", cMat3x3_inspect, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mod, "mat3x3", new_mat3x3, MRB_ARGS_ANY());

  mrb_define_method(mrb, mat3x3_class, "rmul_trans", cMat3x3_rmul_trans, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, mat3x3_class, "lmul_trans", cMat3x3_lmul_trans, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)mat3x3_class,
      "from_quat", cMat3x3_from_quat, MRB_ARGS_REQ(1));

  return mat3x3_class;
}


