#include "tsmrb_numerical.h"
#include "tsmrb_helpers.h"
#include "tsmrb_classes.h"
#include "tsmrb_vec.h"

#include "mruby/array.h"
#include "mruby/data.h"
#include "mruby/proc.h"
#include "mruby/numeric.h"
#include "numarray_mruby.h"

#include "ts_numerical.h"

static void sparse_builder_free(mrb_state* mrb, ts_sparse_builder_t* builder);
static void sparse_free(mrb_state* mrb, ts_sparse_t* matrix);

static void sparse_builder_free_wrapper(mrb_state* mrb, void* p)
{
  sparse_builder_free(mrb, (ts_sparse_builder_t*)p);
}

static void sparse_free_wrapper(mrb_state* mrb, void* p)
{
  sparse_free(mrb, (ts_sparse_t*)p);
}

static const mrb_data_type sparse_builder_type = {
  .struct_name = "SparseBuilder",
  .dfree = sparse_builder_free_wrapper,
};

static const mrb_data_type sparse_type = {
  .struct_name = "Sparse",
  .dfree = sparse_free_wrapper,
};

struct func_closure {
  mrb_state* mrb;
  mrb_value rproc;
};

static double ruby_single_var_func(struct func_closure* cl, double x)
{
  mrb_value ret = 
    mrb_funcall(cl->mrb, cl->rproc,
        "call", 1, mrb_float_value(cl->mrb, x));

  mrb_check_type(cl->mrb, ret, MRB_TT_FLOAT);
  return mrb_float(ret);
}

static double ruby_double_var_func(struct func_closure* cl, double x, double y)
{
  mrb_value ret =
    mrb_funcall(cl->mrb, cl->rproc, 
        "call", 2, mrb_float_value(cl->mrb,x), mrb_float_value(cl->mrb,y));

  return mrb_float(ret);
}

#define DEFAULT_H0 0.1
static mrb_value mNumerical_deriv1(mrb_state* mrb, mrb_value mod)
{
  mrb_value rproc = mrb_nil_value();
  mrb_float x = 0.0;
  mrb_float h0 = DEFAULT_H0;
  mrb_get_args(mrb, "of|f", &rproc, &x, &h0);

  mrb_check_type(mrb, rproc, MRB_TT_PROC);

  struct func_closure cl = {
    .mrb = mrb,
    .rproc = rproc
  };

  double ans = 0.0;
  double err = 0.0;
  TSResult result = deriv_ridders(
        (single_var_func) ruby_single_var_func, &cl, x, h0,
        &ans,
        &err);

  tsmrb_handle_error(mrb, result);

  mrb_value tup[2];
  tup[0] = mrb_float_value(mrb, ans);
  tup[1] = mrb_float_value(mrb, err);

  return mrb_ary_new_from_values(mrb, 2, tup);
}

static mrb_value mNumerical_deriv2(mrb_state* mrb, mrb_value mod)
{
  mrb_value rproc = mrb_nil_value();
  mrb_float x = 0.0;
  mrb_float h0 = DEFAULT_H0;
  mrb_get_args(mrb, "of|f", &rproc, &x, &h0);

  mrb_check_type(mrb, rproc, MRB_TT_PROC);

  struct func_closure cl = {
    .mrb = mrb,
    .rproc = rproc
  };

  double ans = 0.0;
  double err = 0.0;
  TSResult result = deriv2_ridders(
        (single_var_func) ruby_single_var_func, &cl, x,h0,
        &ans,
        &err);

  tsmrb_handle_error(mrb, result);

  mrb_value tup[2];
  tup[0] = mrb_float_value(mrb, ans);
  tup[1] = mrb_float_value(mrb, err);

  return mrb_ary_new_from_values(mrb, 2, tup);
}

static mrb_value mNumerical_deriv_xy(mrb_state* mrb, mrb_value mod)
{
  mrb_value rproc = mrb_nil_value();
  mrb_float x = 0.0;
  mrb_float y = 0.0;
  mrb_float h0 = DEFAULT_H0;
  mrb_get_args(mrb, "offf", &rproc, &x, &y, &h0);

  mrb_check_type(mrb, rproc, MRB_TT_PROC);

  struct func_closure cl = {
    .mrb = mrb,
    .rproc = rproc
  };

  double ans = 0.0;
  double err = 0.0;
  TSResult result = deriv_xy_ridders(
        (double_var_func) ruby_double_var_func, &cl, x,y,h0,
        &ans,
        &err);

  tsmrb_handle_error(mrb, result);

  mrb_value tup[2];
  tup[0] = mrb_float_value(mrb, ans);
  tup[1] = mrb_float_value(mrb, err);

  return mrb_ary_new_from_values(mrb, 2, tup);
}

#define DQW_AT_ONCE 2
static mrb_value mNumerical_dq_w(mrb_state *mrb, mrb_value mod)
{
  mrb_value r_aa, *restv, ret, elts[3];
  mrb_int i,j, restc;
  double ang;
  vector3 aa;
  vector3 dqw1[DQW_AT_ONCE], dqw2[DQW_AT_ONCE], dqw3[DQW_AT_ONCE];
  vector3 dvv[DQW_AT_ONCE];

  restc = 0;
  restv = NULL;
  r_aa = mrb_nil_value();
  mrb_get_args(mrb, "o*", &r_aa, &restv, &restc);

  if (restc < 1)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "dq_w requires at least two arguments (axis-angle and vector)");
  }

  obj_to_vec(mrb, r_aa, &aa);
  ang = vector3_normalize(&aa);

  if (restc == 1)
  {
    obj_to_vec(mrb, restv[0],dvv+0);
    dq_wj(1, dqw1+0, dqw2+0, dqw3+0,
        ang, &aa, dvv+0);

    elts[0] = tsmrb_from_vec3(mrb, dqw1+0);
    elts[1] = tsmrb_from_vec3(mrb, dqw2+0);
    elts[2] = tsmrb_from_vec3(mrb, dqw3+0);

    return mrb_ary_new_from_values(mrb, 3, elts);
  }

  ret = mrb_ary_new_capa(mrb, restc);
  for(i=0; i+DQW_AT_ONCE <= restc; i+=DQW_AT_ONCE)
  {
    for (j=0; j < DQW_AT_ONCE; j++)
    {
      obj_to_vec(mrb, restv[i+j],dvv+j);
    }

    dq_wj(DQW_AT_ONCE, dqw1+0, dqw2+0, dqw3+0,
        ang, &aa, dvv+0);

    for (j=0; j < DQW_AT_ONCE; j++)
    {
      int ai;
      mrb_value item;

      ai = mrb_gc_arena_save(mrb);
      elts[0] = tsmrb_from_vec3(mrb, dqw1+j);
      elts[1] = tsmrb_from_vec3(mrb, dqw2+j);
      elts[2] = tsmrb_from_vec3(mrb, dqw3+j);

      item = mrb_ary_new_from_values(mrb, 3, elts);
      mrb_ary_push(mrb, ret, item);

      mrb_gc_arena_restore(mrb,ai);
    }
  }

  for (; i < restc; i++)
  {
    mrb_value item;

    obj_to_vec(mrb, restv[i],dvv+0);
    dq_wj(1, dqw1+0, dqw2+0, dqw3+0,
        ang, &aa, dvv+0);

    elts[0] = tsmrb_from_vec3(mrb, dqw1+0);
    elts[1] = tsmrb_from_vec3(mrb, dqw2+0);
    elts[2] = tsmrb_from_vec3(mrb, dqw3+0);

    item = mrb_ary_new_from_values(mrb, 3, elts);

    mrb_ary_push(mrb, ret, item);
  }

  return ret;
}

static mrb_value mNumerical_d2q_w(mrb_state* mrb, mrb_value mod)
{
  mrb_value r_aa, r_dv;
  vector3 aa, dv, d2q[D2Q_SIZE];

  mrb_get_args(mrb, "oo", &r_aa, &r_dv);

  obj_to_vec(mrb, r_aa,&aa);
  obj_to_vec(mrb, r_dv,&dv);

  memset(d2q+0, 0, sizeof(d2q));

  d2q_dwi_dwj( d2q, &aa, &dv);

  int dims[3] = {DW_NUM_ELTS, DW_NUM_ELTS, DW_NUM_ELTS};
  numarray *A = NULL;
  mrb_value arr = numarray_new_filled(mrb, 3, dims, 0.0, &A);

  {
    double *restrict hp = A->elts;

    for (size_t k=0; k < DW_NUM_ELTS; k++) {
      for (size_t j=0; j < DW_NUM_ELTS; j++) {
        hp[DW_NUM_ELTS*(DW_NUM_ELTS*k + j) + j] = 
          VECTOR3_ELT( d2q + symm_index(DW_NUM_ELTS,j,j), k );

        for (size_t i=j+1; i < DW_NUM_ELTS; i++) {
          hp[DW_NUM_ELTS*(DW_NUM_ELTS*k + j) + i] = 
            VECTOR3_ELT( d2q + symm_index(DW_NUM_ELTS,i,j), k );

          hp[DW_NUM_ELTS*(DW_NUM_ELTS*k + i) + j] = 
            hp[DW_NUM_ELTS*(DW_NUM_ELTS*k + j) + i];
        }
      }
    }
  }

  return arr;
}

static mrb_value mNumerical_is_symmetric_nelts(mrb_state* mrb, mrb_value mod)
{
  mrb_int n;
  mrb_get_args(mrb, "i", &n);
  return mrb_bool_value(get_symmetric_size(n) > 0);
}

static mrb_value mNumerical_polyfit(mrb_state* mrb, mrb_value self)
{
  mrb_value rxv = mrb_nil_value();
  mrb_value ryv = mrb_nil_value();
  mrb_int ndeg = 0;
  mrb_bool rescale = 0;
  mrb_get_args(mrb, "AAi|b", &rxv, &ryv, &ndeg, &rescale);

  int n = RARRAY_LEN(rxv);
  if (n != RARRAY_LEN(ryv))
    mrb_raise(mrb, E_ARGUMENT_ERROR, 
        "size of x and y arrays must match");

  double* xv = mrb_malloc(mrb, n*sizeof(*xv));
  double* yv = mrb_malloc(mrb, n*sizeof(*xv));
  for (int i=0; i < n; i++)
  {
    xv[i] = mrb_to_float(mrb, mrb_ary_ref(mrb, rxv,i));
    yv[i] = mrb_to_float(mrb, mrb_ary_ref(mrb, ryv,i));
  }

  size_t ncoeff = ndeg+1;
  double* coeffs  = mrb_malloc(mrb, ncoeff*sizeof(*coeffs));
  double* dcoeffs = mrb_malloc(mrb, ncoeff*ncoeff*sizeof(*coeffs));
  double mu[2] = {0};

  TSResult err = polyfit(n, xv, yv, ndeg, coeffs, dcoeffs, mu, rescale);

  mrb_free(mrb,xv);
  mrb_free(mrb,yv);

  if (IS_ERROR(err))
  {
    mrb_free(mrb, coeffs);
    mrb_free(mrb, dcoeffs);
    tsmrb_handle_error(mrb, err);
  }

  mrb_value sd_coeff = mrb_ary_new_capa(mrb, ncoeff);
  for (int i=0; i < ncoeff; i++)
    mrb_ary_push(mrb, sd_coeff,
        mrb_float_value(mrb,
          sqrt(dcoeffs[i+i*ncoeff])));

  mrb_value triple[3] = {
    tsmrb_ary_from_doubles(mrb, ncoeff, coeffs),
    tsmrb_ary_from_doubles(mrb, 2, mu),
    sd_coeff // rcov
  };

  mrb_value ret = mrb_ary_new_from_values(mrb, 3, triple);

  mrb_free(mrb,coeffs);
  mrb_free(mrb,dcoeffs);

  return ret;
}

static mrb_value mNumerical_polyder(mrb_state* mrb, mrb_value mod)
{
  mrb_value rcoeff = mrb_nil_value();
  mrb_get_args(mrb, "A", &rcoeff);

  if (RARRAY_LEN(rcoeff) == 0)
    return rcoeff;

  size_t ncoeff = 0;
  double*restrict coeff  = tsmrb_ary_to_doubles(mrb, rcoeff, &ncoeff);
  double*restrict dcoeff = mrb_calloc(mrb, ncoeff, sizeof(*dcoeff));

  int ndcoeff = polyder(ncoeff-1, coeff, dcoeff);

  mrb_value ret = tsmrb_ary_from_doubles(mrb, ndcoeff+1, dcoeff);

  mrb_free(mrb, coeff);
  mrb_free(mrb, dcoeff);

  return ret;
}

static mrb_value mNumerical_polyval(mrb_state* mrb, mrb_value mod)
{
  mrb_value rcoeff = mrb_nil_value();
  mrb_float x = 0.0;
  mrb_get_args(mrb, "Af", &rcoeff, &x);

  if (RARRAY_LEN(rcoeff) == 0)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "array of polynomial coefficients is empty");

  size_t ncoeff = 0;
  double*restrict coeff  = tsmrb_ary_to_doubles(mrb, rcoeff, &ncoeff);

  double val = polyval(ncoeff-1,coeff, x);

  mrb_free(mrb, coeff);

  return mrb_float_value(mrb, val);
}

static mrb_state* quad_mrb = NULL;
static mrb_value quad_proc;

static double c_quad_proc(double* x)
{
  mrb_value arg = mrb_float_value(quad_mrb, *x);
  mrb_value ret = mrb_funcall_argv(quad_mrb,
      quad_proc, mrb_intern_cstr(quad_mrb,"call"), 1, &arg);

  return mrb_float( ret );
}

static mrb_value mNumerical_quad(mrb_state* mrb, mrb_value mod)
{
  mrb_float x0 = 0.0;
  mrb_float x1 = 0.0;
  mrb_value proc  = mrb_nil_value();
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "ff|o&", &x0, &x1, &proc, &block);

  quad_mrb  = mrb;
  if (!mrb_nil_p(proc) && mrb_obj_is_kind_of(mrb,proc,mrb->proc_class))
    quad_proc = proc;
  else if (!mrb_nil_p(block))
    quad_proc = block;

  if (mrb_nil_p(quad_proc))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected Proc or block");

  double epsabs = 0.0;
  double epsrel = 1e-6;
  double result = 0.0;
  double abserr = 0.0;
  int neval = 0;
  int ier = 0;

  int limit = 10;

  int lenw = 4*limit;
  int last = 0;
  int* iwork = ts_alloc(NO_ALLOCATOR, limit*sizeof(*iwork));
  double* work = ts_alloc(NO_ALLOCATOR, lenw*sizeof(*work));

  dqags(
      c_quad_proc,      //     f      - double precision
                        //              function subprogam defining the integrand
                        //              function f(x). the actual name for f needs to be
                        //              declared e x t e r n a l in the driver program.
                        //
      &x0, &x1,         //     a      - double precision
                        //              lower limit of integration
                        //     b      - double precision
                        //              upper limit of integration
                        //
      &epsabs,          //     epsabs - double precision
                        //              absolute accoracy requested
      &epsrel,          //     epsrel - double precision
                        //              relative accuracy requested
                        //              if  epsabs.le.0
                        //              and epsrel.lt.max(50*rel.mach.acc.,0.5d-28),
                        //              the routine will end with ier = 6.
                        //
                        //  on return
      &result,          //     result - double precision
                        //              approximation to the integral
      &abserr,          //     abserr - double precision
                        //              estimate of the modulus of the absolute error,
                        //              which should equal or exceed abs(i-result)
      &neval,           //     neval  - integer
                        //              number of integrand evaluations
      &ier,             //     ier    - integer
                        //              ier = 0 normal and reliable termination of the
                        //                      routine. it is assumed that the requested
                        //                      accuracy has been achieved.
                        //              ier.gt.0 abnormal termination of the routine
                        //                      the estimates for result and error are
                        //                      less reliable. it is assumed that the
                        //                      requested accuracy has not been achieved.
                        //               error messages
                        //              ier = 1 maximum number of subdivisions allowed
                        //                      has been achieved. one can allow more
                        //                      subdivisions by increasing the value of
                        //                      limit (and taking the according dimension
                        //                      adjustments into account). however, if
                        //                      this yield no improvement it is advised
                        //                      to analyze the integrand in order to
                        //                      determine the integration difficulaties.
                        //                      if the position of a local difficulty can
                        //                      be determined (i.e.singularity,
                        //                      discontinuity within the interval) one
                        //                      will probably gain from splitting up the
                        //                      interval at this point and calling the
                        //                      integrator on the subranges. if possible,
                        //                      an appropriate special-purpose integrator
                        //                      should be used which is designed for
                        //                      handling the type of difficulty involved.
                        //                  = 2 the occurrence of roundoff error is
                        //                      detected, which prevents the requested
                        //                      tolerance from being achieved.
                        //                  = 3 extremely bad integrand behaviour occurs
                        //                      at some points of the integration
                        //                      interval.
                        //                  = 6 the input is invalid, because
                        //                      (epsabs.le.0 and
                        //                       epsrel.lt.max(50*rel.mach.acc.,0.5d-28))
                        //                      or limit.lt.1 or lenw.lt.limit*4.
                        //                      result, abserr, neval, last are set
                        //                      to zero.
                        //                      except when lenw is invalid, iwork(1),
                        //                      work(limit*2+1) and work(limit*3+1) are
                        //                      set to zero, work(1) is set to a and
                        //                      work(limit+1) to b.

                        //  dimensioning parameters
      &limit,           //     limit - integer
                        //             dimensioning parameter for iwork
                        //             limit determines the maximum number of subintervals
                        //             in the partition of the given integration interval
                        //             (a,b), limit.ge.1.
                        //             if limit.lt.1, the routine will end with ier = 6.
      &lenw,            //     lenw  - integer
                        //             dimensioning parameter for work
                        //             lenw must be at least limit*4.
                        //             if lenw.lt.limit*4, the routine will end with
                        //             ier = 6.

      &last,            //     last  - integer
                        //             on return, last equals the number of subintervals
                        //             produced in the subdiviosion process, which
                        //             determines the number of significant elements
                        //             actually in the work arrays.

                        //  work arrays
      iwork,            //     iwork - integer
                        //             vector of dimension at least limit, the first k
                        //             elements of which contain pointers
                        //             to the error estimates over the subintervals
                        //             such that work(limit*3+iwork(1)),... ,
                        //             work(limit*3+iwork(k)) form a decreasing
                        //             sequence, with k = last if last.le.(limit/2+2),
                        //             and k = limit+1-last otherwise

      work              //     work  - double precision
                        //             vector of dimension at least lenw
                        //             on return
                        //             work(1), ..., work(last) contain the left
                        //              end-points of the subintervals in the
                        //              partition of (a,b),
                        //             work(limit+1), ..., work(limit+last) contain
                        //              the right end-points,
                        //             work(limit*2+1), ..., work(limit*2+last) contain
                        //              the integral approximations over the subintervals,
                        //             work(limit*3+1), ..., work(limit*3+last)
                        //              contain the error estimates.
      );

  ts_free(NO_ALLOCATOR, iwork);
  ts_free(NO_ALLOCATOR, work);

  quad_mrb  = NULL;
  quad_proc = mrb_nil_value();

  mrb_value pair[2] = {
    mrb_float_value(mrb, result),
    mrb_float_value(mrb, abserr),
  };

  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value mNumerical_nextafter(mrb_state* mrb, mrb_value mod)
{
  mrb_float f1,f2;
  f1 = f2 = (mrb_float)0.0;
  mrb_get_args(mrb, "ff", &f1,&f2);

  return mrb_float_value(mrb, nextafter(f1, f2));
}

static mrb_value mNumerical_nexttoward(mrb_state* mrb, mrb_value mod)
{
  mrb_float f1,f2;
  f1 = f2 = (mrb_float)0.0;
  mrb_get_args(mrb, "ff", &f1,&f2);

  return mrb_float_value(mrb, nexttoward(f1, f2));
}


ts_sparse_builder_t* tsmrb_to_sparse_builder(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_sparse_builder_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a SparseBuilder object");

  return DATA_PTR(obj);
}

static ts_sparse_builder_t* sparse_builder_alloc(mrb_state* mrb)
{
  ts_sparse_builder_t* builder = mrb_malloc(mrb, sizeof(*builder));
  memset(builder,0,sizeof(*builder));
  return builder;
}

static void sparse_builder_free(mrb_state* mrb, ts_sparse_builder_t* builder)
{
  ts_sparse_builder_finalize(builder);
  mrb_free(mrb, builder);
}

static mrb_value sparse_builder_wrap(mrb_state* mrb, ts_sparse_builder_t* builder)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_sparse_builder_class(mrb), builder, &sparse_builder_type);
  return mrb_obj_value(data);
}

mrb_value tsmrb_new_sparse_builder(mrb_state* mrb, ts_sparse_builder_t** builderp)
{
  ts_sparse_builder_t* builder = sparse_builder_alloc(mrb);
  if (builderp) { *builderp = builder; }

  return sparse_builder_wrap(mrb, builder);
}

static mrb_value cSparseBuilder_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_int m;
  mrb_int n;
  mrb_get_args(mrb, "ii", &m, &n);

  ts_sparse_builder_t* builder = sparse_builder_alloc(mrb);
  TSResult err = ts_sparse_builder_initialize(builder, (size_t)m, (size_t)n);
  if (IS_ERROR(err))
  {
    mrb_free(mrb, builder);
    tsmrb_handle_error(mrb, err);
  }

  DATA_TYPE(self) = &sparse_builder_type;
  DATA_PTR(self) = builder;

  return self;
}

static mrb_value cSparseBuilder_size(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  return mrb_fixnum_value( (mrb_int)ts_sparse_builder_size(builder) );
}

static mrb_value cSparseBuilder_sizes(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  mrb_value elts[2];
  elts[0] = mrb_fixnum_value( (mrb_int) builder->m );
  elts[1] = mrb_fixnum_value( (mrb_int) builder->n );

  return mrb_ary_new_from_values(mrb, 2, elts);
}

static mrb_value cSparseBuilder_m(mrb_state* mrb, mrb_value self)
{ 
  return mrb_fixnum_value( (mrb_int)tsmrb_to_sparse_builder(mrb,self)->m );
}

static mrb_value cSparseBuilder_n(mrb_state* mrb, mrb_value self)
{ 
  return mrb_fixnum_value( (mrb_int)tsmrb_to_sparse_builder(mrb,self)->n );
}

static mrb_value cSparseBuilder_nnz(mrb_state* mrb, mrb_value self)
{ 
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);
  return mrb_fixnum_value( ts_sparse_builder_nnz(builder) );
}

static mrb_value cSparseBuilder_nzelements(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  mrb_value arr =
    mrb_ary_new_capa(mrb,
        (mrb_int)ts_sparse_builder_nnz(builder) );

  ts_sparse_iter iter;
  for(ts_sparse_builder_iter_init(builder, &iter);
      ts_sparse_builder_iter_has_next(&iter);
      ts_sparse_builder_iter_advance(&iter))
  {
    mrb_value elts[3];
    elts[0] = mrb_fixnum_value((mrb_int)iter.i);
    elts[1] = mrb_fixnum_value((mrb_int)iter.j);
    elts[2] = mrb_float_value(mrb, iter.v);

    mrb_ary_push(mrb, arr, 
        mrb_ary_new_from_values(mrb, 3, elts));
  }

  return arr;
}

static void builder_check_indices(mrb_state* mrb, ts_sparse_builder_t* builder, mrb_int i, mrb_int j)
{
  if ((i < 0) || (j < 0) || (i >= builder->m) || (j >= builder->n))
  {
    char buf[64];
    snprintf(buf, sizeof(buf), "invalid index (%ld,%ld)", (long)i,(long)j);
    mrb_raise(mrb, E_INDEX_ERROR, buf);
  }
}

static mrb_value cSparseBuilder_get_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  builder_check_indices(mrb, builder, i,j);

  return mrb_float_value(mrb, ts_sparse_builder_get(builder, i,j) );
}

static mrb_value cSparseBuilder_set_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_float val = 0.0;
  mrb_get_args(mrb, "iif", &i, &j, &val);

  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  builder_check_indices(mrb, builder, i,j);

  tsmrb_handle_error(mrb, 
      ts_sparse_builder_set(builder,
        (size_t)i, (size_t)j, (VecReal)val) );

  return self;
}

static mrb_value cSparseBuilder_add_to_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_float val = 0.0;
  mrb_get_args(mrb, "iif", &i, &j, &val);

  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  builder_check_indices(mrb, builder, i,j);

  tsmrb_handle_error(mrb, 
      ts_sparse_builder_add_to(builder,
        (size_t)i, (size_t)j, (VecReal)val) );

  return self;
}

ts_sparse_t* tsmrb_to_sparse(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_sparse_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Sparse object");

  return DATA_PTR(obj);
}

static ts_sparse_t* sparse_alloc(mrb_state* mrb)
{
  ts_sparse_t* matrix = mrb_malloc(mrb, sizeof(*matrix));
  memset(matrix,0,sizeof(*matrix));
  return matrix;
}

static void sparse_free(mrb_state* mrb, ts_sparse_t* matrix)
{
  ts_sparse_finalize(matrix);
  mrb_free(mrb,matrix);
}

static mrb_value sparse_wrap(mrb_state* mrb, ts_sparse_t* matrix)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_sparse_class(mrb), matrix, &sparse_type);
  return mrb_obj_value(data);
}

mrb_value tsmrb_new_sparse(mrb_state* mrb, ts_sparse_t** matrixp)
{
  ts_sparse_t* matrix = sparse_alloc(mrb);
  if (matrixp) { *matrixp = matrix; }

  return sparse_wrap(mrb, matrix);
}

static mrb_value cSparseBuilder_build(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);
  ts_sparse_t* matrix = NULL;
  mrb_value rmat = tsmrb_new_sparse(mrb, &matrix);

  tsmrb_handle_error(mrb, ts_sparse_builder_build(builder,matrix) );
  return rmat;
}

static mrb_value cSparseBuilder_bands(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  if (builder->m != builder->n)
    mrb_raise(mrb, E_RUNTIME_ERROR, "bands can only be used with square matrices");

  int* bands = mrb_calloc(mrb, 2*builder->m-1, sizeof(*bands));
  int nbands = ts_sparse_builder_get_bands(builder, bands);

  mrb_value r_bands;
  if (nbands > 0)
    r_bands = mrb_ary_new_capa(mrb, nbands);
  else
    r_bands = mrb_ary_new(mrb);

  for (int i=0; i < nbands; i++)
    mrb_ary_push(mrb, r_bands, mrb_fixnum_value(bands[i]));

  mrb_free(mrb,bands);
  return r_bands;
}

static mrb_value cSparseBuilder_as_banded(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);
  if (builder->m != builder->n)
    mrb_raise(mrb, E_RUNTIME_ERROR, "bands can only be used with square matrices");

  mrb_int wh = 0;
  mrb_get_args(mrb, "|i", &wh);

  if (wh != 0) { wh /= abs(wh); }

  int bands[2] = {0,0};
  double* elts = NULL;
  tsmrb_handle_error(mrb,
      ts_sparse_builder_as_banded(builder, bands, &elts, wh) );
  int nb = 1 + bands[0] + bands[1];
  int ncol = builder->n;

  mrb_value r_banded = mrb_ary_new_capa(mrb, nb);
  for (int i=0; i < nb; i++)
  {
    mrb_value r_row = mrb_ary_new_capa(mrb, ncol);
    for (int j=0; j < ncol; j++)
      mrb_ary_push(mrb, r_row, mrb_float_value(mrb, elts[nb*j + i]));

    mrb_ary_push(mrb, r_banded, r_row);
  }

  ts_free(NO_ALLOCATOR, elts);

  mrb_value avals[3];
  avals[0] = mrb_fixnum_value(bands[0]);
  avals[1] = mrb_fixnum_value(bands[1]);
  avals[2] = r_banded;

  return mrb_ary_new_from_values(mrb, 3, avals);
}

static mrb_value cSparseBuilder_each_entry(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb,self);

  mrb_value block;
  int argc = mrb_get_args(mrb, "&", &block);

  mrb_value elts[3];
  mrb_value ret = self;

  ts_sparse_iter iter;
  if (argc > 0)
  {
    for(ts_sparse_builder_iter_init(builder, &iter);
        ts_sparse_builder_iter_has_next(&iter);
        ts_sparse_builder_iter_advance(&iter))
    {
      elts[0] = mrb_fixnum_value((mrb_int)iter.i);
      elts[1] = mrb_fixnum_value((mrb_int)iter.j);
      elts[2] = mrb_float_value(mrb, iter.v);
      mrb_yield(mrb, block, 
          mrb_ary_new_from_values(mrb, 3, elts));
    }
  }
  else
  {
    ret = mrb_ary_new(mrb);
    for(ts_sparse_builder_iter_init(builder, &iter);
        ts_sparse_builder_iter_has_next(&iter);
        ts_sparse_builder_iter_advance(&iter))
    {
      elts[0] = mrb_fixnum_value((mrb_int)iter.i);
      elts[1] = mrb_fixnum_value((mrb_int)iter.j);
      elts[2] = mrb_float_value(mrb, iter.v);
      mrb_ary_push(mrb, ret, 
          mrb_ary_new_from_values(mrb, 3, elts));
    }
  }
  return ret;
}

static mrb_value cSparseBuilder_rcm(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* builder = tsmrb_to_sparse_builder(mrb, self);

  if (builder->m != builder->n)
    mrb_raise(mrb, E_RUNTIME_ERROR, "permute_rcm can only be used with square matrices");

  size_t* perm = mrb_calloc(mrb, builder->n, sizeof(*perm));

  tsmrb_handle_error(mrb, ts_sparse_builder_rcm(builder, perm) );

  mrb_value rperm = mrb_ary_new_capa(mrb, builder->n);
  for (long i=0; i < builder->n; i++)
    mrb_ary_push(mrb, rperm, mrb_fixnum_value((mrb_int)perm[i]));
  mrb_free(mrb,perm);

  return rperm;
}

static mrb_value cSparseBuilder_permute_indices(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* origbld = tsmrb_to_sparse_builder(mrb,self);

  mrb_value rperm;
  mrb_get_args(mrb, "A", &rperm);

  if (origbld->m != origbld->n)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "bands can only be used with square matrices");

  if (RARRAY_LEN(rperm) != origbld->n)
  {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "expected %lu elements in permutation array", origbld->n);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  ts_sparse_builder_t* permbld = NULL;
  mrb_value ret = tsmrb_new_sparse_builder(mrb,&permbld);

  size_t* perm = mrb_calloc(mrb, origbld->n, sizeof(*perm));
  for (size_t i=0; i < origbld->n; i++)
    perm[i] = mrb_fixnum(RARRAY_PTR(rperm)[i]);

  TSResult err = ts_sparse_builder_permute(origbld, permbld, perm);
  mrb_free(mrb,perm);

  tsmrb_handle_error(mrb, err);

  return ret;
}

static mrb_value cSparseBuilder_permute_indicesB(mrb_state* mrb, mrb_value self)
{
  ts_sparse_builder_t* origbld = tsmrb_to_sparse_builder(mrb,self);

  mrb_value rperm;
  mrb_get_args(mrb, "A", &rperm);

  if (origbld->m != origbld->n)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "bands can only be used with square matrices");

  if (RARRAY_LEN(rperm) != origbld->n)
  {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "expected %lu elements in permutation array", origbld->n);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  size_t* perm = mrb_calloc(mrb, origbld->n, sizeof(*perm));
  for (size_t i=0; i < origbld->n; i++)
    perm[i] = mrb_fixnum(RARRAY_PTR(rperm)[i]);

  TSResult err = ts_sparse_builder_permute_self(origbld, perm);
  mrb_free(mrb,perm);

  tsmrb_handle_error(mrb, err);

  return self;
}

static mrb_value cSparseMatrix_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_int m = 0;
  mrb_int n = 0;
  mrb_int nnz = 0;
  mrb_get_args(mrb, "iii", &m, &n, &nnz);

  ts_sparse_t* matrix = sparse_alloc(mrb);
  tsmrb_handle_error(mrb, ts_sparse_initialize(matrix, m, n, nnz) );

  return self;
}

static mrb_value cSparseMatrix_m(mrb_state* mrb, mrb_value self)
{
  return mrb_fixnum_value( tsmrb_to_sparse(mrb,self)->m );
}

static mrb_value cSparseMatrix_n(mrb_state* mrb, mrb_value self)
{
  return mrb_fixnum_value( tsmrb_to_sparse(mrb,self)->n );
}

static mrb_value cSparseMatrix_nnz(mrb_state* mrb, mrb_value self)
{
  return mrb_fixnum_value( tsmrb_to_sparse(mrb,self)->nnz );
}

static mrb_value cSparseMatrix_ir(mrb_state* mrb, mrb_value self)
{
  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);
  mrb_value arr = mrb_ary_new_capa(mrb, matrix->nnz);
  for(size_t i=0; i < matrix->nnz; i++)
    mrb_ary_push(mrb, arr, mrb_fixnum_value(matrix->ir[i]));

  return arr;
}

static mrb_value cSparseMatrix_jc(mrb_state* mrb, mrb_value self)
{
  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);
  mrb_value arr = mrb_ary_new_capa(mrb, matrix->n+1);
  for(size_t i=0; i <= matrix->n; i++)
    mrb_ary_push(mrb, arr, mrb_fixnum_value(matrix->jc[i]));
  return arr;
}

static mrb_value cSparseMatrix_pr(mrb_state* mrb, mrb_value self)
{
  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);
  mrb_value arr = mrb_ary_new_capa(mrb, matrix->nnz);
  for(size_t i=0; i < matrix->nnz; i++)
    mrb_ary_push(mrb, arr, mrb_float_value(mrb, matrix->pr[i]));
  return arr;
}

static void sparse_check_indices(mrb_state* mrb, ts_sparse_t* matrix, long i, long j)
{
  if ((i < 0) || (j < 0) || (i >= matrix->m) || (j >= matrix->n))
  {
    char buf[64];
    snprintf(buf, sizeof(buf)-1, "invalid index (%ld,%ld)", i,j);
    mrb_raise(mrb, E_INDEX_ERROR, buf);
  }
}

static mrb_value cSparseMatrix_get_element(mrb_state* mrb, mrb_value self)
{
  mrb_int i = 0;
  mrb_int j = 0;
  mrb_get_args(mrb, "ii", &i, &j);

  ts_sparse_t* matrix = tsmrb_to_sparse(mrb, self);
  sparse_check_indices(mrb, matrix, i, j);

  return mrb_float_value(mrb, ts_sparse_get(matrix,i,j));
}

static mrb_value cSparseMatrix_full(mrb_state *mrb, mrb_value self)
{
  ts_sparse_t* matrix;
  int dims[2];
  numarray *A;
  mrb_value ret;

  matrix = tsmrb_to_sparse(mrb, self);
  dims[0] = matrix->m;
  dims[1] = matrix->n;

  A = NULL;
  ret = numarray_new_filled(mrb, 2, dims, 0.0, &A);

  ts_sparse_full(matrix, A->elts, TS_ROWMAJOR);

  return ret;
}

static int is_vector(const numarray *A)
{
  if (A->ndims == 1)
  {
    return 1;
  }

  if (A->ndims == 2)
  {
    return (A->dims[0] == 1) || (A->dims[1] == 1);
  }

  return 0;
}

static mrb_value cSparseMatrix_rmul(mrb_state *mrb, mrb_value self)
{
  const numarray *vec0;
  numarray *vec1;
  mrb_value r_vec0, r_vec1;
  int m, n;

  mrb_get_args(mrb, "o", &r_vec0);

  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);
  m = matrix->m;
  n = matrix->n;

  vec0 = numarray_unwrap(mrb, r_vec0);
  if (! is_vector(vec0))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires NumArray to be vector-shaped");
  }

  if (vec0->nelts != n)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "sizes incompatible: right-multiply of vector (%d) with matrix (%d x %d)",
        vec0->nelts, matrix->m, matrix->n);
  }

  vec1 = NULL;
  r_vec1 = numarray_new_filled(mrb, 1, &m, 0.0, &vec1);

  ts_sparse_vec_rmul(matrix, vec0->elts, vec1->elts);
  return r_vec1;
}

static mrb_value cSparseMatrix_lmul(mrb_state *mrb, mrb_value self)
{
  const numarray *vec0;
  numarray *vec1;
  mrb_value r_vec0, r_vec1;
  int m, n;

  mrb_get_args(mrb, "o", &r_vec0);

  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);
  m = matrix->m;
  n = matrix->n;

  vec0 = numarray_unwrap(mrb, r_vec0);
  if (! is_vector(vec0))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires NumArray to be vector-shaped");
  }

  if (vec0->nelts != m)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "sizes incompatible: left-multiply of vector (%d) with matrix (%d x %d)",
        vec0->nelts, matrix->m, matrix->n);
  }

  vec1 = NULL;
  r_vec1 = numarray_new_filled(mrb, 1, &n, 0.0, &vec1);

  ts_sparse_vec_lmul(matrix, vec0->elts, vec1->elts);
  return r_vec1;
}


static mrb_value cSparseMatrix_eigen(mrb_state *mrb, mrb_value self)
{
  int n;
  numarray *V;
  mrb_value r_eigvals;
  struct ts_sparse_eigen eigen;
  ts_sparse_t* matrix;

  matrix = tsmrb_to_sparse(mrb,self);

  n = matrix->n;
  V = NULL;
  r_eigvals = numarray_new_filled(mrb, 1, &n, 0.0, &V);

  eigen.max_iter = TS_SPARSE_EIGEN_MAXITER;
  eigen.nev = matrix->n;

  eigen.matrix = matrix;
  eigen.vals = V->elts;
  eigen.vecs = NULL;

  eigen.num_iter      = 0;
  eigen.num_converged = 0;
  eigen.num_ops       = 0;
  eigen.num_reorth    = 0;

  tsmrb_handle_error(mrb, ts_sparse_eigenvalues_calc(&eigen));
  return r_eigvals;
}

static mrb_value cSparseMatrix_eigensystem(mrb_state *mrb, mrb_value self)
{
  int dims[2];
  numarray *eigvals, *eigvecs;
  mrb_value r_eigvals, r_eigvecs, pair[2];
  struct ts_sparse_eigen eigen;
  ts_sparse_t* matrix;

  matrix = tsmrb_to_sparse(mrb,self);

  dims[0] = dims[1] = matrix->n;
  r_eigvals = numarray_new_filled(mrb, 1, dims, 0.0, &eigvals);
  r_eigvecs = numarray_new_filled(mrb, 2, dims, 0.0, &eigvecs);

  eigen.max_iter = TS_SPARSE_EIGEN_MAXITER;
  eigen.nev = matrix->n;

  eigen.matrix = matrix;
  eigen.vals = eigvals->elts;
  eigen.vecs = eigvecs->elts;

  eigen.num_iter      = 0;
  eigen.num_converged = 0;
  eigen.num_ops       = 0;
  eigen.num_reorth    = 0;

  memset(eigen.vals, 0, matrix->n * sizeof(eigen.vals[0]));
  tsmrb_handle_error(mrb, ts_sparse_eigenvalues_calc(&eigen));

  pair[0] = r_eigvals;
  pair[1] = r_eigvecs;
  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value cSparseMatrix_each_entry(mrb_state* mrb, mrb_value self)
{
  ts_sparse_t* matrix = tsmrb_to_sparse(mrb,self);

  mrb_value block;
  int argc = mrb_get_args(mrb, "&", &block);

  mrb_value elts[3];
  mrb_value ret = self;

  if (argc > 0)
  {
    for (size_t j=0; j < matrix->n; j++)
    {
      for (size_t r = matrix->jc[j]; r < matrix->jc[j+1]; r++)
      {
        elts[0] = mrb_fixnum_value(matrix->ir[r]);
        elts[1] = mrb_fixnum_value(j);
        elts[2] = mrb_float_value(mrb, matrix->pr[r]);

        mrb_yield(mrb, block, 
            mrb_ary_new_from_values(mrb, 3, elts));
      }
    }
  }
  else
  {
    ret = mrb_ary_new_capa(mrb, matrix->nnz);
    for (size_t j=0; j < matrix->n; j++)
    {
      for (size_t r = matrix->jc[j]; r < matrix->jc[j+1]; r++)
      {
        elts[0] = mrb_fixnum_value(matrix->ir[r]);
        elts[1] = mrb_fixnum_value(j);
        elts[2] = mrb_float_value(mrb, matrix->pr[r]);

        mrb_ary_push(mrb, ret, 
            mrb_ary_new_from_values(mrb, 3, elts));
      }
    }
  }

  return ret;
}

#define MAXITER 300
static mrb_value cNumArray_lanczos_upper(mrb_state *mrb, mrb_value self)
{
  mrb_int n, nev, ncv;
  mrb_value r_eigenvals;
  numarray *A, *eigenvals;

  /* various parameters required by dsaupd and dseupd */
  int pad0, lworkl;
  size_t scratch_size;
  double *scratch, *resid, *v, *workd, *workl;

  nev = ncv = 0;
  mrb_get_args(mrb, "ii", &nev, &ncv);

  A = numarray_unwrap(mrb, self);

  if (A->ndims != 2)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires NumArray or rank 2");
  }

  if (A->dims[0] != A->dims[1])
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires square NumArray matrix");
  }

  n = A->dims[0];

  if (nev > n) { nev = n;}

  pad0 = (nev == n);
  if (pad0) { n++; }

  if (ncv < nev) { ncv = nev+1; }
  if (ncv > n) { ncv = n; }

  lworkl = ncv*(ncv+8);

  scratch_size = (n + n*ncv + 3*n+1 + lworkl)*sizeof(scratch[0]);

  scratch = mrb_malloc(mrb, scratch_size);
  resid = scratch;       /* n doubles      */
  v = resid + n;         /* n*ncv doubles  */
  workd = v + (n*ncv);   /* 3*n+1 doubles  */
  workl = workd + 3*n+1; /* lworkl doubles */

  memset(resid, 0,       n*sizeof(*resid));
  memset(v    , 0,   n*ncv*sizeof(*v));
  memset(workd, 0, (3*n+1)*sizeof(*workd));
  memset(workl, 0,  lworkl*sizeof(*workl));

  {
    int ido;
    char bmat;
    const char *which;
    double tol;
    int info;
    int iparam[11], ipntr[11];
    int bmat_len;
    int which_len;

    ido = 0;
    bmat = 'I';
    which = "LM";
    tol = 0.0; // 1e-6;
    info = 0;

    iparam[0] = 1;                // ISHIFT
    iparam[2] = MAXITER;          // max iterations
    iparam[3] = 1;                // NB (blocksize)
    iparam[6] = 1;                // solve A.x == lambda*x for symmetric A

    // parameters that are no longer used
    iparam[1] = 0;
    iparam[5] = 0;

    // outputs for iparam
    iparam[4] = 0;                // number of "converged" Ritz values
    iparam[7] = 0;                // NP: number of shifts the user is to provide
    iparam[8] = 0;                // NUMOP
    iparam[9] = 0;                // NUMOPB
    iparam[10] = 0;               // NUMREO

    bmat_len = 1;
    which_len = 2;

    for(;;) {
      dsaupd_(&ido, 
          &bmat, 
          &n, 
          which, 
          &nev, 
          &tol,
          resid,
          &ncv,
          v,
          &n, 
          iparam,
          ipntr,
          workd,
          workl,
          &lworkl,
          &info,
          &bmat_len,
          &which_len
          );

      if ((1 == ido) || (-1 == ido))
      {
        upper_sym_matmult(pad0 ? n-1 : n,
            workd + ipntr[1]-1,
            A->elts,
            workd + ipntr[0]-1,
            TS_ROWMAJOR);

        if (pad0) { 
          workd[ ipntr[1]+n-1 ] = 0.0;
        }
      } 
      else if (ido == 99)
      {
        break;
      }
      else
      {
        tsmrb_raisef(mrb, E_RUNTIME_ERROR, "unexpected ARPACK value: ido=%d", ido);
      }
    }

    if (info != 0)
    {
      mrb_free(mrb, scratch);

      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "ARPACK encountered error %d in routine dsaupd (check documentation)", info);
    }

    {
      const char* howmny = "A";
      int rvec, *select, ldz, howmny_len;
      double sigma, zdummy;

      rvec = 0;  // no eigenvectors
      select = mrb_malloc(mrb, ncv*sizeof(select[0]));

      eigenvals = NULL;
      r_eigenvals = numarray_new_filled(mrb, 1, &nev, 0.0, &eigenvals);
      memset(eigenvals->elts, 0, nev*sizeof(double));

      ldz = 0;  // Z is not used (*rvec == 0)
      sigma = 0; // should not be referenced

      howmny_len = 1;

      zdummy = 0.0;

      dseupd_(
          &rvec,
          howmny,
          select,
          eigenvals->elts,
          &zdummy,
          &ldz,
          &sigma,

          &bmat, 
          &n, 
          which, 
          &nev, 
          &tol,
          resid,
          &ncv,
          v,
          &n, 
          iparam,
          ipntr,
          workd,
          workl,
          &lworkl,
          &info,
          &howmny_len,
          &bmat_len,
          &which_len);

      if (info != 0) {
        mrb_free(mrb, scratch);
        mrb_free(mrb, select);

        tsmrb_raisef(mrb, E_RUNTIME_ERROR,
            "ARPACK encountered error %d in routine dseupd (check documentation)", 
            info);
      }

      mrb_free(mrb, select);
    }

    mrb_free(mrb, scratch);
  }

  return r_eigenvals;
}
#undef MAXITER

static void check_square_matrix(mrb_state *mrb, const numarray *M)
{
  if (M->ndims != 2)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires NumArray of rank 2");
  }

  if (M->dims[0] != M->dims[1])
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires square NumArray matrix");
  }
}

static void check_vector(mrb_state *mrb, const struct numarray *v)
{
  if (v->ndims != 1)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires a vector");
  }
}

static void check_compatible_vector(mrb_state *mrb, const numarray*restrict v, int n)
{
  check_vector(mrb, v);
  if (v->nelts != n)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "requires vector of size %d", n);
  }
}

static mrb_value cNumArray_upper_sym_mult(mrb_state *mrb, mrb_value self)
{
  mrb_value other, ret;
  numarray *M, *x, *y;
  int n;

  mrb_get_args(mrb, "o", &other);

  M = numarray_unwrap(mrb, self);
  x = numarray_unwrap(mrb, other);

  check_square_matrix(mrb, M);
  n = M->dims[0];

  check_compatible_vector(mrb, x,n);

  y = NULL;
  ret = numarray_new_filled(mrb, 1, &n, 0.0, &y);

  upper_sym_matmult(n, y->elts, M->elts, x->elts, TS_ROWMAJOR);

  return ret;
}

static mrb_value cNumArray_upper_as_vec(mrb_state *mrb, mrb_value self)
{
  mrb_value r_vec;
  numarray *M, *vec;
  int n, nelts;

  M = numarray_unwrap(mrb, self);
  check_square_matrix(mrb, M);

  n = M->dims[0];
  nelts = n*(n+1)/2;

  vec = NULL;
  r_vec = numarray_new_filled(mrb, 1, &nelts, 0.0, &vec);

  {
    double *restrict mdata;
    double *restrict vdata;
    int i,j,c;

    mdata = M->elts;
    vdata = vec->elts;

    for (c=0, j=0; j < n; j++)
    {
      for (i=0; i <= j; i++, c++)
      {
        vdata[c] = mdata[i+n*j];
      }
    }
  }

  return r_vec;
}

static mrb_value cNumArray_eigen(mrb_state *mrb, mrb_value self)
{
  numarray *M, *eigen;
  mrb_value r_eigen;

  M = numarray_unwrap(mrb, self);
  check_square_matrix(mrb, M);

  {
    int n, lda, lwork;
    double *scratch, *copy, *work, *wr, *wi;
    size_t scratch_size;

    n = M->dims[0];
    lda = n;
    lwork = 4*(M->nelts);

    scratch_size = M->nelts + lwork + n + n;
    scratch = mrb_malloc(mrb, scratch_size * sizeof(*scratch));

    copy = scratch;          /* nelts elements */
    work = copy + M->nelts;  /* lwork elements */
    wr = work + lwork;       /* n elements */
    wi = wr + n;             /* n elements */

    /* make a copy to preserve the original matrix (copy is overwritten)
     */
    memcpy(copy, M->elts, M->nelts * sizeof(*copy));

    memset(wr,0,n*sizeof(wr[0]));
    memset(wi,0,n*sizeof(wi[0]));

    {
      int i, ldvl, ldvr, info, edims[2];
      double vl, vr;

      ldvl = 1;
      ldvr = 1;
      vl = 0.0;
      vr = 0.0;

      info = 0;
      /* SUBROUTINE DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR,
       *    $                  LDVR, WORK, LWORK, INFO )
       */
      dgeev( "N", "N", &n, copy, &lda, wr, wi, &vl, &ldvl, &vr, &ldvr, work, &lwork, &info );

      eigen = NULL;
      edims[0] = n;
      edims[1] = 2;
      r_eigen = numarray_new_filled(mrb, 2, edims, 0.0, &eigen);

      for (i=0; i < n; i++) {
        eigen->elts[i+0] = wr[i];
        eigen->elts[i+n] = wi[i];
      }

    }
    mrb_free(mrb, scratch);
  }

  return r_eigen;
}

static mrb_value cNumArray_eigensystem(mrb_state *mrb, mrb_value self)
{
  numarray *M, *eigen, *evecs;
  mrb_value ret, r_eigen, r_evecs;

  M = numarray_unwrap(mrb, self);
  check_square_matrix(mrb, M);

  {
    int n, lda, lwork;
    double *scratch, *copy, *work, *wr, *wi;
    size_t scratch_size;

    n = M->dims[0];
    lda = n;
    lwork = 4*(M->nelts);

    scratch_size = M->nelts + lwork + n + n;
    scratch = mrb_malloc(mrb, scratch_size * sizeof(*scratch));

    copy = scratch;          /* nelts elements */
    work = copy + M->nelts;  /* lwork elements */
    wr = work + lwork;       /* n elements */
    wi = wr + n;             /* n elements */

    /* make a copy to preserve the original matrix (copy is overwritten)
     */
    memcpy(copy, M->elts, M->nelts * sizeof(*copy));

    memset(wr,0,n*sizeof(wr[0]));
    memset(wi,0,n*sizeof(wi[0]));

    {
      int i, ldvl, ldvr, info, edims[2], vdims[2];
      double vl;

      ldvl = 1;
      ldvr = n;
      vl = 0.0;

      vdims[0] = vdims[1] = n;
      evecs = NULL;
      r_evecs = numarray_new_filled(mrb, 2, vdims, 0.0, &evecs);

      info = 0;
      /* SUBROUTINE DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR,
       *    $                  LDVR, WORK, LWORK, INFO )
       */
      dgeev( "N", "V", &n, copy, &lda, wr, wi, &vl, &ldvl, evecs->elts, &ldvr, work, &lwork, &info );

      eigen = NULL;
      edims[0] = n;
      edims[1] = 2;
      r_eigen = numarray_new_filled(mrb, 2, edims, 0.0, &eigen);

      for (i=0; i < n; i++)
      {
        eigen->elts[i+0] = wr[i];
        eigen->elts[i+n] = wi[i];
      }

    }
    mrb_free(mrb, scratch);
  }

  ret = mrb_ary_new_capa(mrb,2);
  mrb_ary_push(mrb, ret, r_eigen);
  mrb_ary_push(mrb, ret, r_evecs);

  return ret;
  // return r_eigen;
}

static mrb_value cNumArray_sym_eigen(mrb_state *mrb, mrb_value self)
{
  numarray *M, *eigen;
  mrb_value r_eigen;
  int n;

  M = numarray_unwrap(mrb, self);
  check_vector(mrb, M);

  n = get_symmetric_size(M->nelts);
  if (n==0)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "vector must be a packed symmetric matrix");
  }

  {
    int lwork;
    double *scratch, *A, *D, *E, *tau, *work;
    size_t scratch_size;

    lwork = 12*n;
    scratch_size = (M->nelts + 3*n-1 + lwork)*sizeof(double);
    scratch = mrb_malloc(mrb, scratch_size);

    // STEP 1: reduce matrix to tridiagonal form...
    A    = scratch;          /* M->nelts elements */
    D    = A + M->nelts;     /* n elements */
    E    = D + n;            /* n elements */
    tau  = E + n;            /* n-1 elements */
    work = tau + n-1;        /* lwork elements */

    // make a copy to preserve the original matrix (A is overwritten)
    memcpy(A, M->elts, (M->nelts) * sizeof(A[0]));

    {
      int info, liwork, *iwork, ldz;
      int il, iu, neigen;
      double vl, vu, d_unused;

      info = 0;

      // Reduce packed symmetric matrix A to tridiagonal matrix...
      dsptrd("U", &n, A, D, E, tau, &info);

      // STEP 2: solve eigenvectors of tridiagonal matrix...
      liwork = 10*n;
      iwork = mrb_malloc(mrb, liwork * sizeof(*iwork));

      vl = vu = 0.0;
      il = iu = 0;

      d_unused = 0.0;

      neigen = 0;
      ldz = 1;

      eigen = NULL;
      r_eigen = numarray_new_filled(mrb, 1, &n, 0.0, &eigen);

      dstegr("N", "A", 
          &n, D,E, 
          &vl, &vu, 
          &il, &iu, 
          &d_unused,                // abstol (no longer used)
          &neigen,                  // M -- number of eigenvalues
          eigen->elts,              // W array (results)
          &d_unused,                // Z array (unused for eigenvalues)
          &ldz,
          &d_unused,                // isuppz (unused for eigenvalues)
          work,
          &lwork,
          iwork,
          &liwork,
          &info);

      mrb_free(mrb, iwork);
    }

    mrb_free(mrb, scratch);
  }

  return r_eigen;
}

static struct RClass* tsmrb_define_sparse_builder(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cSparseBuilder = mrb_define_class(mrb, "SparseBuilder", mrb->object_class );
  tsmrb_classes(mrb)->cSparseBuilder = cSparseBuilder;
  MRB_SET_INSTANCE_TT(cSparseBuilder, MRB_TT_DATA);

  mrb_define_method(mrb, cSparseBuilder, "initialize", cSparseBuilder_initialize, MRB_ARGS_ANY());
  mrb_define_method(mrb, cSparseBuilder, "size", cSparseBuilder_size, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "sizes", cSparseBuilder_sizes, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "m", cSparseBuilder_m, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "n", cSparseBuilder_n, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "nnz", cSparseBuilder_nnz, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "nzelements", cSparseBuilder_nzelements, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseBuilder, "[]" , cSparseBuilder_get_element, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSparseBuilder, "[]=", cSparseBuilder_set_element, MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cSparseBuilder, "add_to", cSparseBuilder_add_to_element, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cSparseBuilder, "build", cSparseBuilder_build, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseBuilder, "bands", cSparseBuilder_bands, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "as_banded", cSparseBuilder_as_banded, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSparseBuilder, "each_entry", cSparseBuilder_each_entry, MRB_ARGS_BLOCK());

  mrb_define_method(mrb, cSparseBuilder, "rcm", cSparseBuilder_rcm, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseBuilder, "permute_indices", cSparseBuilder_permute_indices, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSparseBuilder, "permute_indices!", cSparseBuilder_permute_indicesB, MRB_ARGS_REQ(1));

  return cSparseBuilder;
}

static struct RClass* tsmrb_define_sparse(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cSparseMatrix = mrb_define_class(mrb, "SparseMatrix", mrb->object_class);
  tsmrb_classes(mrb)->cSparseMatrix = cSparseMatrix;
  MRB_SET_INSTANCE_TT(cSparseMatrix, MRB_TT_DATA);

  mrb_define_method(mrb, cSparseMatrix, "initialize", cSparseMatrix_initialize, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cSparseMatrix, "m", cSparseMatrix_m, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseMatrix, "n", cSparseMatrix_n, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseMatrix, "nnz", cSparseMatrix_nnz, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseMatrix, "ir", cSparseMatrix_ir, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseMatrix, "jc", cSparseMatrix_jc, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseMatrix, "pr", cSparseMatrix_pr, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseMatrix, "[]", cSparseMatrix_get_element, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSparseMatrix, "full", cSparseMatrix_full, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseMatrix, "rmul", cSparseMatrix_rmul, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSparseMatrix, "lmul", cSparseMatrix_lmul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSparseMatrix, "eigen", cSparseMatrix_eigen, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSparseMatrix, "eigensystem", cSparseMatrix_eigensystem, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSparseMatrix, "each_entry", cSparseMatrix_each_entry, MRB_ARGS_BLOCK());

  return cSparseMatrix;
}

static void poly_free(mrb_state* mrb, void* ptr)
{
  ts_poly_finalize( (ts_poly*)ptr );
  mrb_free(mrb, ptr);
}

static const mrb_data_type poly_type = {
  .struct_name = "Poly",
  .dfree = poly_free,
};

ts_poly* tsmrb_to_poly(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_poly_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Poly object");

  return DATA_PTR(obj);
}

static inline mrb_value
poly_wrap(mrb_state* mrb, ts_poly* P)
{
  return mrb_obj_value(
      mrb_data_object_alloc(mrb,
        tsmrb_poly_class(mrb), P, &poly_type));
}

static inline ts_poly* poly_alloc(mrb_state* mrb)
{
  ts_poly* P = mrb_malloc(mrb, sizeof(*P));
  memset(P, 0, sizeof(*P));
  return P;
}

static mrb_value cPoly_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value* argv = NULL;
  int argc = 0;
  mrb_get_args(mrb, "*", &argv, &argc);

  ts_poly* P = poly_alloc(mrb);

  DATA_TYPE(self) = &poly_type;
  DATA_PTR(self) = P;

  if (argc > TS_POLY_MAX_COEFF)
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "too many polynomial coefficients (max %S)",
        mrb_fixnum_value(TS_POLY_MAX_COEFF));

  if (argc == 0)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Poly requires at least one coefficient");

  double coeff[TS_POLY_MAX_COEFF];
  for(int i=0; i < argc; i++)
    coeff[i] = mrb_to_float(mrb, argv[i]);

  ts_poly_init(P, argc-1, coeff);

  return self;
}

static mrb_value cPoly_fit(mrb_state* mrb, mrb_value klass)
{
  mrb_int ndeg = 0;
  mrb_value r_xv = mrb_nil_value();
  mrb_value r_yv = mrb_nil_value();
  mrb_get_args(mrb, "iAA", &ndeg, &r_xv, &r_yv);

  if ((ndeg < 1) || (ndeg > TS_POLY_MAX_DEG))
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid polynomial degree %d", mrb_fixnum_value(ndeg));

  size_t npts = 0;
  double*restrict xv = tsmrb_ary_to_doubles(mrb, r_xv, &npts);

  if (npts != RARRAY_LEN(r_yv))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "x and y must be the same size");

  double*restrict yv = tsmrb_ary_to_doubles(mrb, r_yv, &npts);

  ts_poly* P = poly_alloc(mrb);
  TSResult ret = ts_poly_fit(P, ndeg, npts, xv, yv, 0);
  mrb_free(mrb, xv);
  mrb_free(mrb, yv);
  tsmrb_handle_error(mrb, ret);

  return poly_wrap(mrb,P);
}

static mrb_value cPoly_bracket(mrb_state* mrb, mrb_value klass)
{
  mrb_value* argv = NULL;
  int argc = 0;
  mrb_get_args(mrb, "*", &argv, &argc);
  return mrb_funcall_argv(mrb, klass, TSMRB_SYM(new), argc, argv);
}

static mrb_value cPoly_get_coeff(mrb_state* mrb, mrb_value self)
{
  ts_poly* P = tsmrb_to_poly(mrb,self);
  return tsmrb_ary_from_doubles(mrb, P->ncoeff, P->coeff);
}

static mrb_value cPoly_get_sd_coeff(mrb_state* mrb, mrb_value self)
{
  ts_poly* P = tsmrb_to_poly(mrb,self);
  if (P->covar == NULL)
    return mrb_nil_value();

  mrb_value sd_coeff = mrb_ary_new_capa(mrb,P->ncoeff);
  for (int i=0; i < P->ncoeff; i++)
    mrb_ary_push(mrb, sd_coeff,
        mrb_float_value(mrb,
          sqrt(P->covar[i+i*P->ncoeff])));
  return sd_coeff;
}

static mrb_value cPoly_get_cov(mrb_state* mrb, mrb_value self)
{
  ts_poly* P = tsmrb_to_poly(mrb,self);
  size_t ncoeff = P->ncoeff;

  if (P->covar == NULL)
    return mrb_nil_value();

  mrb_value rcov = mrb_ary_new_capa(mrb,ncoeff);
  for (int i=0; i < ncoeff; i++)
  {
    mrb_value row = mrb_ary_new_capa(mrb,ncoeff);
    for (int j=0; j < ncoeff; j++)
      mrb_ary_push(mrb, row, mrb_float_value(mrb, P->covar[i+j*ncoeff]));

    mrb_ary_push(mrb, rcov, row);
  }

  return rcov;
}

static mrb_value cPoly_value(mrb_state* mrb, mrb_value self)
{
  mrb_value* argv = NULL;
  int argc = 0;
  mrb_get_args(mrb, "*", &argv, &argc);
  ts_poly* P = tsmrb_to_poly(mrb,self);

  if (argc == 1)
  {
    double v = ts_poly_value(P, mrb_to_float(mrb,argv[0]));
    return mrb_float_value(mrb,v);
  }
  else
  {
    mrb_value ret = mrb_ary_new_capa(mrb,argc);
    for (int i=0; i < argc; i++)
    {
      if (!mrb_numeric_p(argv[i]))
        mrb_raisef(mrb, E_ARGUMENT_ERROR,
            "argument %S must be a number", mrb_fixnum_value(i));

      int ai = mrb_gc_arena_save(mrb);

      double v = ts_poly_value(P, mrb_to_float(mrb, argv[i]));
      mrb_ary_push(mrb, ret, mrb_float_value(mrb, v));

      mrb_gc_arena_restore(mrb,ai);
    }
    return ret;
  }
}

static mrb_value cPoly_deriv(mrb_state* mrb, mrb_value self)
{
  ts_poly* P = tsmrb_to_poly(mrb, self);
  ts_poly* dP = poly_alloc(mrb);

  tsmrb_handle_error(mrb, ts_poly_deriv(dP, P));

  return poly_wrap(mrb, dP);
}

struct RClass* tsmrb_define_poly(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cPoly = mrb_define_class_under(mrb, mod, "Poly", mrb->object_class);
  tsmrb_classes(mrb)->cPoly = cPoly;
  MRB_SET_INSTANCE_TT(cPoly, MRB_TT_DATA);

  mrb_define_method(mrb, cPoly, "initialize", cPoly_initialize, MRB_ARGS_ANY());

  mrb_define_module_function(mrb, cPoly, "[]", cPoly_bracket, MRB_ARGS_ANY());
  mrb_define_module_function(mrb, cPoly, "fit", cPoly_fit, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cPoly, "deriv", cPoly_deriv, 0);

  mrb_define_method(mrb, cPoly, "coeff"   , cPoly_get_coeff   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cPoly, "sd_coeff", cPoly_get_sd_coeff, MRB_ARGS_NONE());
  mrb_define_method(mrb, cPoly, "cov", cPoly_get_cov, 0);

  mrb_define_method(mrb, cPoly, "[]"   , cPoly_value, MRB_ARGS_ANY());
  mrb_define_method(mrb, cPoly, "value", cPoly_value, MRB_ARGS_ANY());
  mrb_define_method(mrb, cPoly, "call" , cPoly_value, MRB_ARGS_ANY());

  return cPoly;
}

#define VECN_ALIGN 0x20
static double* allocvec(mrb_state *mrb, size_t n, char** origp)
{
  char* p = mrb_malloc(mrb, n*sizeof(double) + VECN_ALIGN-1 );
  int off = ((uintptr_t)p) & (VECN_ALIGN-1);

  double* v = (double*)(p+off);
  if (origp) { *origp = p; }

  return v;
}

static double *cvtvec(mrb_state *mrb, mrb_value rv, size_t *np, char **origp)
{
  size_t i,n;
  double *restrict v; 

  n = RARRAY_LEN(rv);
  v = allocvec(mrb, n, origp);

  for(i=0; i < n; i++)
  {
    v[i] = mrb_to_flo(mrb, RARRAY_PTR(rv)[i]);
  }

  if (np) { *np = n; }
  return v;
}

static mrb_value mVecN_sub(mrb_state *mrb, mrb_value mod)
{
  mrb_value va, vb, ret;
  size_t i,n;
  char *pa, *pb, *pc;
  double *restrict veca, *restrict vecb, *restrict vecc;

  mrb_get_args(mrb, "AA", &va, &vb);

  n = RARRAY_LEN(va);
  if (n != RARRAY_LEN(vb))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "vectors must be the same size");
  }
  ret = mrb_ary_new_capa(mrb, n);

  pa = pb = pc = NULL;
  veca = cvtvec(mrb, va, NULL, &pa);
  vecb = cvtvec(mrb, vb, NULL, &pb);
  vecc = allocvec(mrb, n, &pc);

  subv(n, vecc, veca, vecb);

  mrb_free(mrb, pa);
  mrb_free(mrb, pb);

  for (i=0; i < n; i++) {
    int ai = mrb_gc_arena_save(mrb);
    mrb_ary_push(mrb, ret, mrb_float_value(mrb, vecc[i]));
    mrb_gc_arena_restore(mrb,ai);
  }

  mrb_free(mrb, pc);

  return ret;
}

static mrb_value mVecN_dot(mrb_state *mrb, mrb_value mod)
{
  mrb_value va, vb;
  size_t n;
  char *pa, *pb;
  double *restrict veca, *restrict vecb;
  double dv;

  mrb_get_args(mrb, "AA", &va, &vb);

  n = RARRAY_LEN(va);
  if (n != RARRAY_LEN(vb))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "vectors must be the same size");
  }

  pa = pb = NULL;
  veca = cvtvec(mrb, va, NULL, &pa);
  vecb = cvtvec(mrb, vb, NULL, &pb);

  dv = dotv(n, veca, vecb);

  mrb_free(mrb, pa);
  mrb_free(mrb, pb);

  return mrb_float_value(mrb, dv);
}

static mrb_value mVecN_norm(mrb_state *mrb, mrb_value mod)
{
  mrb_value v;
  size_t n;
  char *p;
  double *restrict vec;
  double nv;

  mrb_get_args(mrb, "A", &v);
  n = RARRAY_LEN(v);

  p = NULL;
  vec = cvtvec(mrb, v, NULL, &p);
  nv = normv(n, vec);

  mrb_free(mrb,p);

  return mrb_float_value(mrb,nv);
}

struct RClass* tsmrb_define_numerical(mrb_state* mrb, struct RClass* mod)
{
  struct RClass *mNumerical;

  mNumerical = mrb_define_module_under(mrb, mod, "Numerical");
  tsmrb_classes(mrb)->mNumerical = mNumerical;

  mrb_define_module_function(mrb, mNumerical, "deriv1", mNumerical_deriv1,
      MRB_ARGS_REQ(2) | MRB_ARGS_OPT(1));

  mrb_define_module_function(mrb, mNumerical, "deriv2", mNumerical_deriv2,
      MRB_ARGS_REQ(2) | MRB_ARGS_OPT(1));

  mrb_define_module_function(mrb, mNumerical, "deriv_xy", mNumerical_deriv_xy,
      MRB_ARGS_REQ(4));

  mrb_define_module_function(mrb, mNumerical, "dq_w", mNumerical_dq_w, MRB_ARGS_REQ(2) | MRB_ARGS_REST());

  mrb_define_module_function(mrb, mNumerical, "d2q_w", mNumerical_d2q_w, MRB_ARGS_REQ(2));

  mrb_define_module_function(mrb, mNumerical, "is_symmetric_nelts", mNumerical_is_symmetric_nelts, MRB_ARGS_REQ(1));

  mrb_define_module_function(mrb, mNumerical, "polyfit", mNumerical_polyfit, 
      MRB_ARGS_REQ(3) | MRB_ARGS_OPT(1));

  mrb_define_module_function(mrb, mNumerical, "polyder", mNumerical_polyder, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, mNumerical, "polyval", mNumerical_polyval, MRB_ARGS_REQ(2));

  mrb_define_module_function(mrb, mNumerical, "quad", mNumerical_quad,
      MRB_ARGS_REQ(2) | MRB_ARGS_OPT(1) | MRB_ARGS_BLOCK());

  mrb_define_module_function(mrb, mNumerical, "nextafter", mNumerical_nextafter,
      MRB_ARGS_REQ(2));

  mrb_define_module_function(mrb, mNumerical, "nexttoward", mNumerical_nexttoward,
      MRB_ARGS_REQ(2));

  tsmrb_define_sparse_builder(mrb, mod);
  tsmrb_define_sparse(mrb, mod);

  tsmrb_define_poly(mrb, mNumerical);

  {
    struct RClass *cNumArray;

    cNumArray = mrb_class_get(mrb, "NumArray");

    mrb_define_method(mrb, cNumArray, "upper_as_vec", cNumArray_upper_as_vec, MRB_ARGS_NONE());
    mrb_define_method(mrb, cNumArray, "lanczos_upper", cNumArray_lanczos_upper, MRB_ARGS_REQ(2));
    mrb_define_method(mrb, cNumArray, "upper_sym_mult", cNumArray_upper_sym_mult, MRB_ARGS_REQ(1));

    mrb_define_method(mrb, cNumArray, "eigen", cNumArray_eigen, MRB_ARGS_NONE());
    mrb_define_method(mrb, cNumArray, "eigensystem", cNumArray_eigensystem, MRB_ARGS_NONE());
    mrb_define_method(mrb, cNumArray, "sym_eigen", cNumArray_sym_eigen, MRB_ARGS_NONE());
  }

  {
    struct RClass *mVecN;
    mVecN = mrb_define_module_under(mrb, mNumerical, "VecN");

    mrb_define_module_function(mrb, mVecN, "sub", mVecN_sub, MRB_ARGS_REQ(2));
    mrb_define_module_function(mrb, mVecN, "dot", mVecN_dot, MRB_ARGS_REQ(2));
    mrb_define_module_function(mrb, mVecN, "norm", mVecN_norm, MRB_ARGS_REQ(1));
  }

  return mNumerical;
}



