#include "tsmrb_sim.h"

#include "mruby/data.h"
#include "mruby/array.h"
#include "mruby/range.h"

#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"

struct tsmrb_particle
{
  ts_sim* sim;
  size_t pind;
};

static const mrb_data_type pwrap_type = {
  .struct_name = "Particle",
  .dfree = tsmrb_free,
};

static inline struct tsmrb_particle*
get_pwrap(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_particle_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Particle object");

  return DATA_PTR(obj);
}

ts_particle* tsmrb_to_particle(mrb_state* mrb, mrb_value obj)
{
  struct tsmrb_particle* pwrap = get_pwrap(mrb,obj);
  return ts_get_particle(pwrap->sim,pwrap->pind);
}

size_t tsmrb_to_particle_index(mrb_state* mrb, mrb_value obj, ts_sim** simptr)
{
  struct tsmrb_particle* pwrap = get_pwrap(mrb,obj);
  if (simptr) { *simptr = pwrap->sim; }
  return pwrap->pind;
}

mrb_value tsmrb_from_particle(mrb_state* mrb, ts_sim* sim, size_t pind)
{
  struct tsmrb_particle* pwrap = mrb_malloc(mrb, sizeof(*pwrap));
  pwrap->sim = sim;
  pwrap->pind = pind;

  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_particle_class(mrb), pwrap, &pwrap_type);
  return mrb_obj_value(data);
}

static mrb_value cParticle_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_raise(mrb, E_RUNTIME_ERROR,
      "Particle must be created through a simulation");
  return self;
}

static mrb_value cParticle_equal(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);
  if (!mrb_obj_is_instance_of(mrb, other, tsmrb_particle_class(mrb)))
    return mrb_bool_value(0);

  struct tsmrb_particle* a = get_pwrap(mrb,self);
  struct tsmrb_particle* b = get_pwrap(mrb,other);

  return mrb_bool_value( (a==b) || ((a->sim == b->sim) && (a->pind == b->pind)) );
}

static mrb_value cParticle_get_index(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t pind = tsmrb_to_particle_index(mrb, self, &sim);
  if (sim == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "sim is NULL");

  return mrb_fixnum_value((mrb_int)pind);
}

static inline void check_pind(mrb_state* mrb, ts_sim* sim, size_t pind)
{
  if (pind >= sim->num_particles)
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "invalid particle index %lu",(unsigned long)pind);
}

#define PARTICLE_VEC_GETTER( rname, mname) \
  static mrb_value SYMCAT(cParticle_get_, rname)             \
      (mrb_state* mrb,mrb_value self)                        \
  { ts_sim* sim = NULL;                                      \
    size_t pind = tsmrb_to_particle_index(mrb, self, &sim);  \
    check_pind(mrb, sim,  pind);                             \
    return tsmrb_vec3_to_ary(mrb, mname(sim, pind) );        \
  }                                                          \
  static mrb_value SYMCAT3(cParticle_get_,rname,v)           \
    (mrb_state* mrb,mrb_value self)                          \
  { ts_sim* sim = NULL;                                      \
    size_t pind = tsmrb_to_particle_index(mrb, self, &sim);  \
    check_pind(mrb, sim,  pind);                             \
    return tsmrb_from_vec3(mrb, mname(sim,pind) ); }         \
  DEFINITION_EAT_SEMI

#define PARTICLE_VEC_SETTER( rname, mname )                \
  static mrb_value SYMCAT(cParticle_set_, rname)           \
    (mrb_state* mrb, mrb_value self)                       \
  { mrb_value rname;                                       \
    ts_sim* sim = NULL;                                    \
    mrb_get_args(mrb, "o", &rname);                        \
    size_t pind = tsmrb_to_particle_index(mrb,self, &sim); \
    check_pind(mrb, sim,  pind);                           \
    obj_to_vec(mrb, rname, mname(sim, pind));              \
    return self; }  DEFINITION_EAT_SEMI

PARTICLE_VEC_GETTER(pos,PARTICLE_POS);
PARTICLE_VEC_SETTER(pos,PARTICLE_POS);

PARTICLE_VEC_GETTER(vel,PARTICLE_VEL);
PARTICLE_VEC_SETTER(vel,PARTICLE_VEL);

PARTICLE_VEC_GETTER(force,PARTICLE_FORCE);
PARTICLE_VEC_SETTER(force,PARTICLE_FORCE);

static mrb_value cParticle_moveby(mrb_state* mrb, mrb_value self)
{
  mrb_float dx = 0.0;
  mrb_float dy = 0.0;
  mrb_float dz = 0.0;
  mrb_get_args(mrb, "fff", &dx, &dy, &dz);

  ts_sim* sim = NULL;
  size_t pind = tsmrb_to_particle_index(mrb, self, &sim);

  vector3_displace( PARTICLE_POS(sim,pind), dx,dy,dz );

  return self;
}

static mrb_value cParticle_set_type(mrb_state* mrb, mrb_value self)
{
  mrb_int ptype_index;
  mrb_get_args(mrb, "i", &ptype_index);

  ts_sim* sim = NULL;
  size_t pind = tsmrb_to_particle_index(mrb, self, &sim);

  if (ptype_index < 0)
    tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle type %d", (int)ptype_index);

  ts_particle_type* ptype = ts_get_particle_type(sim, ptype_index);
  if (ptype == NULL)
    tsmrb_raisef(mrb, E_RANGE_ERROR, "undefined particle type %d", (int)ptype_index);

  TS_PARTICLE(sim,pind)->type = ptype;

  return self;
}

static mrb_value cParticle_get_type(mrb_state* mrb, mrb_value self)
{
  struct tsmrb_particle* pwrap = get_pwrap(mrb,self);
  return mrb_fixnum_value( (mrb_int)TS_PTYPE_IND(pwrap->sim,pwrap->pind) );
}

static mrb_value cParticle_add_bond(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_value rbt;
  int argc = mrb_get_args(mrb, "o|o", &other, &rbt);

  ts_particle* p0 = tsmrb_to_particle(mrb,self); 
  ts_particle* p1 = tsmrb_to_particle(mrb,other);

  ts_bond_type bondtype = (argc > 1)
    ? tsmrb_bondtype(mrb, rbt)
    : bondtype_new( BONDTYPE_UNDIRECTED, 1 )
    ;

  tsmrb_handle_error(mrb, ts_particle_add_bond(p0,p1, bondtype));

  return self;
}

static mrb_value cParticle_add_directed_bond(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_int btype;
  mrb_get_args(mrb, "oi", &other, &btype);

  ts_particle* p0 = tsmrb_to_particle(mrb,self);
  ts_particle* p1 = tsmrb_to_particle(mrb,other);

  tsmrb_check_bondtype(mrb, btype);
  ts_bond_type bt = bondtype_new(BONDTYPE_FORWARD, btype);

  tsmrb_handle_error(mrb, ts_particle_add_bond(p0,p1, bt));

  return self;
}

static mrb_value cParticle_remove_bond(mrb_state* mrb, mrb_value r_p0)
{
  mrb_value r_p1;
  mrb_get_args(mrb, "o", &r_p1);

  ts_particle* p0 = tsmrb_to_particle(mrb,r_p0); 
  ts_particle* p1 = tsmrb_to_particle(mrb,r_p1); 

  tsmrb_handle_error(mrb, ts_particle_remove_bond(p0,p1));

  return r_p0;
}

static mrb_value cParticle_is_bonded_to(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  return mrb_bool_value( ts_particle_is_bonded_to(p0,p1) );
}

static mrb_value cParticle_bond_directional_p(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  ts_bond_type bt = ts_particle_get_bond_type(p0,p1);

  if (bondtype_is_bond(bt))
    return mrb_bool_value( !bondtype_is_undirected(bt) );
  else
    return mrb_nil_value();
}

static mrb_value cParticle_get_bond_direction(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  ts_bond_type bt = ts_particle_get_bond_type(p0,p1);

  return tsmrb_bond_type_direction(mrb, bt);
}

static mrb_value cParticle_get_bond_type(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  int bondtype = bondtype_get_type(ts_particle_get_bond_type(p0,p1));
  if (bondtype > 0)
    return mrb_fixnum_value(bondtype);
  else
    return mrb_nil_value();
}

static mrb_value cParticle_get_bond_info(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  ts_bond_type bondtype = ts_particle_get_bond_type(p0,p1);

  if (bondtype_is_bond(bondtype))
    return tsmrb_full_bondtype_value(mrb, bondtype);
  else
    return mrb_nil_value();
}

static mrb_value cParticle_set_bond_type(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_value bt;
  mrb_get_args(mrb, "oo", &other, &bt);

  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  ts_particle* p1 = tsmrb_to_particle(mrb, other);

  tsmrb_handle_error(mrb, 
      ts_particle_set_bond_type(p0,p1, tsmrb_bondtype(mrb,bt)));

  return self;
}

static double vector_between(mrb_state* mrb, mrb_value self, mrb_value other, vector3*restrict dr)
{
  ts_sim* sim0 = NULL;
  size_t pind0 = tsmrb_to_particle_index(mrb, self, &sim0);

  ts_sim* sim1 = NULL;
  size_t pind1 = tsmrb_to_particle_index(mrb, other, &sim1);

  if (sim0 != sim1)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
      "particles must be from the same simulation!");

  return vector3_distance(PARTICLE_POS(sim0, pind0), PARTICLE_POS(sim1, pind1), dr);
}

static mrb_value cParticle_distance_to(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  vector3 dr;
  return mrb_float_value(mrb, vector_between(mrb,self,other,&dr));
}

static mrb_value cParticle_vector_to(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  vector3 dr;
  vector_between(mrb,self,other,&dr);
  return tsmrb_from_vec3(mrb, &dr);
}

#define PUSH_FLAG(flag, sym) \
  if ((p0->flags & (flag)) > 0) { mrb_ary_push(mrb, flags, TSMRB_SYM_OBJ(sym)); }
static mrb_value cParticle_get_flags(mrb_state* mrb, mrb_value self)
{
  ts_particle *p0 = tsmrb_to_particle(mrb,self); 

  mrb_value flags = mrb_ary_new(mrb);
  if (p0->flags != PARTICLE_FLAG_NONE)
  {
    PUSH_FLAG( PARTICLE_FLAG_FROZEN   , frozen    );
    PUSH_FLAG( PARTICLE_FLAG_RIGID    , rigid     );
    PUSH_FLAG( PARTICLE_FLAG_RIGID_COM, rigid_com );
  }

  return flags;
}
#undef PUSH_FLAG

static mrb_value cParticle_get_flag_bits(mrb_state* mrb, mrb_value self)
{
  ts_particle *p0 = tsmrb_to_particle(mrb,self); 
  return mrb_fixnum_value(p0->flags);
}

static mrb_value cParticle_is_rigid(mrb_state* mrb, mrb_value self)
{
  return mrb_bool_value(ts_particle_is_rigid(tsmrb_to_particle(mrb,self)));
}

static mrb_value cParticle_is_frozen(mrb_state* mrb, mrb_value self)
{
  return mrb_bool_value(ts_particle_is_frozen(tsmrb_to_particle(mrb,self)));
}

static mrb_value cParticle_freeze(mrb_state* mrb, mrb_value self)
{
  ts_particle_set_frozen(tsmrb_to_particle(mrb,self), 1);
  return self;
}

static mrb_value cParticle_unfreeze(mrb_state* mrb, mrb_value self)
{
  ts_particle_set_frozen(tsmrb_to_particle(mrb,self), 0);
  return self;
}

static mrb_value cParticle_get_rigid_body_index(mrb_state* mrb, mrb_value self)
{
  ts_particle* p0 = tsmrb_to_particle(mrb, self); 
  if (p0->rigid_body_index != TS_NULL_RIGID_BODY)
    return mrb_fixnum_value( (mrb_int)p0->rigid_body_index );
  else
    return mrb_nil_value();
}


static size_t define_particles(mrb_state* mrb, ts_sim* sim, size_t n, size_t ptype_index)
{

  ts_particle_type* ptype = ts_get_particle_type(sim, ptype_index);
  if (ptype == NULL) {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "undefined particle type %d", (int)ptype_index);
    mrb_raise(mrb, E_RANGE_ERROR, buf);
  }

  if (ts_num_particles(sim) + n > ts_max_particles(sim)) {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "cannot define %lu particles: not enough particles allocated", n);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  size_t pind = ts_num_particles(sim);

  for (size_t i = 0; i < n; i++) {
    ts_particle* p = ts_define_particle(sim, ptype_index);
    if (p == NULL) {
      char buf[64];
      snprintf(buf,sizeof(buf)-1,
          "error in defining particle %lu", i+1);
      mrb_raise(mrb, E_RUNTIME_ERROR, buf);
    }

  }

  return pind;
}


static mrb_value cSim_define_particle(mrb_state* mrb, mrb_value self)
{
  mrb_int ptype;
  mrb_get_args(mrb, "i", &ptype);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  size_t pind = define_particles(mrb,sim,1,ptype);

  return tsmrb_from_particle(mrb, sim, pind);
}

static mrb_value cSim_define_particles(mrb_state* mrb, mrb_value self)
{
  int argc;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  size_t pind0 = 0;
  size_t n = 0;

  if (argc == 1) {
    mrb_check_type(mrb, argv[0], MRB_TT_ARRAY);
    ts_sim* sim = tsmrb_to_sim(mrb,self);

    pind0 = ts_num_particles(sim);
    n = RARRAY_LEN(argv[0]);

    if (pind0 + n > ts_max_particles(sim)) {
      char buf[64];
      snprintf(buf, sizeof(buf),
          "cannot define %lu particles: not enough particles allocated", n);
      mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
    }

    for (size_t i=0; i < n; i++) {
      size_t ptype_index = mrb_fixnum( RARRAY_PTR(argv[0])[i] );

      ts_particle* p = ts_define_particle(sim, ptype_index);
      if (p == NULL) {
        char buf[64];
        snprintf(buf, sizeof(buf),
            "error in defining particle %lu", i+1);
        mrb_raise(mrb, E_RUNTIME_ERROR, buf);
      }

      // rbtAssert( TS_PIND(rbSim->sim,p) == pind0 + i );
    }
  } else if (argc == 2) { 
    n = mrb_fixnum(argv[0]);
    size_t ptype = mrb_fixnum(argv[1]);

    pind0 = define_particles(mrb,sim,n,ptype);
  } else {
    char buf[64];
    snprintf(buf,sizeof(buf)-1,
        "wrong number of arguments (%d instead of either 1 or 2)", argc);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  return mrb_range_new(mrb, mrb_fixnum_value(pind0), mrb_fixnum_value(pind0+n), 1);
}

static mrb_value cSim_get_particle(mrb_state* mrb, mrb_value self)
{
  mrb_int pind = -1;
  mrb_get_args(mrb, "i", &pind);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if ((pind < 0) || (pind >= sim->num_particles))
  {
    char buf[64];
    snprintf(buf,sizeof(buf),
        "invalid particle index %ld",
        (long)pind);
    mrb_raise(mrb, E_RANGE_ERROR, buf);
  }

  return tsmrb_from_particle(mrb, sim, pind);
}

static mrb_value cSim_all_particles(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb, self);
  size_t nparticle = sim->num_particles;

  mrb_value plist = mrb_ary_new_capa(mrb, nparticle);
  for(size_t i=0; i < nparticle; i++)
  {
    int ai = mrb_gc_arena_save(mrb);
    mrb_ary_push(mrb, plist,
        tsmrb_from_particle(mrb,sim,i));
    mrb_gc_arena_restore(mrb,ai);
  }

  return plist;
}

static mrb_value cSim_index_of_particle(mrb_state* mrb, mrb_value self)
{
  mrb_value rpart;
  mrb_get_args(mrb, "o", &rpart);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  ts_particle* p = tsmrb_to_particle(mrb,rpart);

  if (!ts_has_particle(sim, p))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "simulation does not contain particle");

  return mrb_fixnum_value( (mrb_int)TS_PIND(sim, p) );
}

static mrb_value cSim_each_particle(mrb_state* mrb, mrb_value self)
{
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if (!mrb_nil_p(block))
  {
    for(size_t i=0; i < sim->num_particles; i++)
    {
      int arena = mrb_gc_arena_save(mrb);
      mrb_value pobj = tsmrb_from_particle(mrb, sim, i);
      mrb_yield(mrb, block, pobj);
      mrb_gc_arena_restore(mrb,arena);
    }
    return self;
  }
  else
  {
    mrb_value ary = mrb_ary_new(mrb);
    for(size_t i=0; i < sim->num_particles; i++)
    {
      int arena = mrb_gc_arena_save(mrb);
      mrb_ary_push(mrb, ary,
          tsmrb_from_particle(mrb, sim, i));
      mrb_gc_arena_restore(mrb,arena);
    }
    return ary;
  }
}

static mrb_value cSim_each_particle_with_index(mrb_state* mrb, mrb_value self)
{
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if (!mrb_nil_p(block))
  {
    for(size_t i=0; i < sim->num_particles; i++)
    {
      int arena = mrb_gc_arena_save(mrb);
      mrb_value pair[2] = {
        tsmrb_from_particle(mrb, sim, i),
        mrb_fixnum_value( (mrb_int)i )
      };
      mrb_yield_argv(mrb, block, 2, pair);
      mrb_gc_arena_restore(mrb,arena);
    }
    return self;
  }
  else
  {
    mrb_value ary = mrb_ary_new(mrb);
    for(size_t i=0; i < sim->num_particles; i++)
    {
      int arena = mrb_gc_arena_save(mrb);
      mrb_value pair[2] = {
        tsmrb_from_particle(mrb, sim, i),
        mrb_fixnum_value( (mrb_int)i )
      };
      mrb_ary_push(mrb, ary,
          mrb_ary_new_from_values(mrb, 2, pair));
      mrb_gc_arena_restore(mrb,arena);
    }
    return ary;
  }
}

static void add_methods_to_cSim(mrb_state* mrb, struct RClass* cSim)
{
  mrb_define_method(mrb, cSim, "define_particle"  , cSim_define_particle  , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "define_particles" , cSim_define_particles , MRB_ARGS_ANY());

  mrb_define_method(mrb, cSim, "get_particle"     , cSim_get_particle     , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "index_of_particle", cSim_index_of_particle, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "all_particles"    , cSim_all_particles    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "each_particle"    , cSim_each_particle    , MRB_ARGS_BLOCK());
  mrb_define_method(mrb, cSim, "each_particle_with_index", cSim_each_particle_with_index, MRB_ARGS_BLOCK());
}

struct RClass* tsmrb_define_particle(mrb_state* mrb, struct RClass* mod, struct RClass* cSim)
{
  if (tsmrb_particle_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Particle class has already been initialized!");

  struct RClass* cParticle = mrb_define_class_under(mrb, mod,
      "Particle", mrb->object_class);
  tsmrb_classes(mrb)->cParticle = cParticle;
  MRB_SET_INSTANCE_TT(cParticle, MRB_TT_DATA);

  mrb_define_method(mrb, cParticle, "initialize", cParticle_initialize, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "==", cParticle_equal, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cParticle, "index", cParticle_get_index, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "pos=", cParticle_set_pos, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "pos" , cParticle_get_pos, MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "posv", cParticle_get_posv, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "vel=", cParticle_set_vel, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "vel" , cParticle_get_vel, MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "velv", cParticle_get_velv, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "force=", cParticle_set_force, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "force" , cParticle_get_force, MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "forcev", cParticle_get_forcev, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "moveby!", cParticle_moveby, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cParticle, "type=", cParticle_set_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "type",  cParticle_get_type, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "add_bond"          , cParticle_add_bond          , MRB_ARGS_REQ(1) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cParticle, "add_directed_bond" , cParticle_add_directed_bond , MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cParticle, "remove_bond"       , cParticle_remove_bond       , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "is_bonded_to"      , cParticle_is_bonded_to      , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "bonded_to?"        , cParticle_is_bonded_to      , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "bond_directional?" , cParticle_bond_directional_p, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "get_bond_direction", cParticle_get_bond_direction, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "get_bond_type"     , cParticle_get_bond_type     , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "set_bond_type"     , cParticle_set_bond_type     , MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cParticle, "get_bond_info"     , cParticle_get_bond_info     , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cParticle, "distance_to", cParticle_distance_to, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cParticle, "vector_to"  , cParticle_vector_to  , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cParticle, "flags"    , cParticle_get_flags, MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "flag_bits", cParticle_get_flag_bits, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "rigid?"   , cParticle_is_rigid , MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "frozen?"  , cParticle_is_frozen, MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "freeze!"  , cParticle_freeze   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cParticle, "unfreeze!", cParticle_unfreeze , MRB_ARGS_NONE());

  mrb_define_method(mrb, cParticle, "rigid_body_index", cParticle_get_rigid_body_index, MRB_ARGS_NONE());

#define PARTICLE_FLAG(name) \
  mrb_define_const(mrb, cParticle, SYMSTR(name), \
      mrb_fixnum_value(SYMCAT(PARTICLE_,name)))

  PARTICLE_FLAG(FLAG_FROZEN);
  PARTICLE_FLAG(FLAG_RIGID);
  PARTICLE_FLAG(FLAG_RIGID_COM);

#undef PARTICLE_FLAG

  add_methods_to_cSim(mrb,cSim);

  return cParticle;
}

