#include "tsmrb_potentials.h"

#include "ts_potentials.h"
#include "ts_errors.h"

#include "mruby/data.h"
#include "mruby/khash.h"
#include "mruby/variable.h"
#include "mruby/numeric.h"

#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"
#include "tsmrb_sim.h"

static void free_potential(mrb_state *mrb, void *p);

static const mrb_data_type constant_field_type = {
  .struct_name = "ConstantField",
  .dfree = free_potential,
};

static const mrb_data_type spin_around_particle_type = {
  .struct_name = "SpinAroundParticle",
  .dfree = free_potential,
};

static const mrb_data_type plane_barrier_type = {
  .struct_name = "PlaneBarrier",
  .dfree = free_potential,
};

static const mrb_data_type potential_type = {
  .struct_name = "Potential",
  .dfree = free_potential,
};

static const mrb_data_type harmonic_type = {
  .struct_name = "Harmonic",
  .dfree = free_potential,
};

static const mrb_data_type wca_type = {
  .struct_name = "WCA",
  .dfree = free_potential,
};

static const mrb_data_type angle_harmonic_type = {
  .struct_name = "AngleHarmonic",
  .dfree = free_potential,
};

static const mrb_data_type dihedral_type = {
  .struct_name = "Dihedral",
  .dfree = free_potential,
};

static const mrb_data_type dipole_type = {
  .struct_name = "Dipole",
  .dfree = free_potential,
};

static void free_potential(mrb_state *mrb, void *p)
{
  Potential **potp = p;
  if (*potp) { ts_potential_decref(*potp); }

  tsmrb_free(mrb, potp);
}

static void wrap_potential(mrb_state* mrb, mrb_value self, Potential* pot, const mrb_data_type* dtype)
{
  Potential** potp = mrb_malloc(mrb, sizeof(*potp));
  *potp = pot;

  DATA_TYPE(self) = dtype;
  DATA_PTR(self) = potp;
}

Potential* tsmrb_to_potential(mrb_state* mrb,mrb_value obj)
{
  if (!mrb_obj_is_kind_of(mrb, obj, tsmrb_potential_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Potential object");

  Potential** potp = DATA_PTR(obj);
  return *potp;
}

mrb_value tsmrb_from_potential(mrb_state* mrb, Potential* pot)
{
  Potential** potp = mrb_malloc(mrb, sizeof(*potp));
  *potp = ts_potential_incref(pot);

  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_potential_class(mrb),
      potp,
      &potential_type);

  return mrb_obj_value(data);
}

static mrb_value cPotential_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_raise(mrb, E_RUNTIME_ERROR,
      "invalid attempt to create a Potential object");
  return self;
}

static mrb_value cPotential_equals(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  if (mrb_obj_is_kind_of(mrb, other, tsmrb_classes(mrb)->cPotential))
  {
    Potential* a = tsmrb_to_potential(mrb, self);
    Potential* b = tsmrb_to_potential(mrb, other);
    return mrb_bool_value( a == b );
  }

  return mrb_bool_value(0);
}

static mrb_value cPotential_get_nbodies(mrb_state* mrb, mrb_value self)
{
  Potential* potential = tsmrb_to_potential(mrb, self);
  return mrb_fixnum_value(potential->nbodies);
}

static mrb_value cPotential_get_name(mrb_state* mrb, mrb_value self)
{
  Potential* potential = tsmrb_to_potential(mrb, self);
  return mrb_str_new_cstr(mrb, potential->name);
}

KHASH_DECLARE(ht, mrb_value, mrb_value, 1)
static void set_each_param(mrb_state* mrb, Potential* pot, mrb_value pname, mrb_value pvalue)
{
  mrb_check_type(mrb, pname , MRB_TT_SYMBOL);
  mrb_check_type(mrb, pvalue, MRB_TT_FLOAT);

  struct ts_potential_params param;
  param.name  = mrb_sym2name(mrb, mrb_symbol(pname));
  param.value = mrb_to_flo(mrb,pvalue);

  tsmrb_handle_error(mrb, pot->set_param_func(pot, &param));
}

static mrb_value cPotential_set_parameters(mrb_state* mrb, mrb_value self)
{
  mrb_value newParams;
  mrb_get_args(mrb, "H", &newParams);

  Potential* potential = tsmrb_to_potential(mrb, self);
  if (potential->set_param_func != NULL) {
    struct kh_ht* tbl = mrb_hash_tbl(mrb, newParams);
    for(khiter_t k = kh_begin(tbl); k != kh_end(tbl); k++) {
      if (kh_exist(tbl,k)) {
        mrb_value pname  = kh_key(tbl,k);
        mrb_value pvalue = mrb_hash_get(mrb, newParams, pname);

        set_each_param(mrb, potential, pname, pvalue);
      }
    }
  }

  return self;
}

static mrb_value cPotential_get_parameters(mrb_state* mrb, mrb_value self)
{
  mrb_value rbParams = mrb_hash_new(mrb);
  Potential* potential = tsmrb_to_potential(mrb, self);

  if (potential->get_params_func != NULL) {
    struct ts_potential_params* params = NULL;
    tsmrb_handle_error(mrb, potential->get_params_func(potential, &params));
    if (params)
    {
      for (size_t i=0; params[i].name; i++)
      {
        mrb_sym sym = mrb_intern_cstr(mrb,params[i].name);
        mrb_hash_set(mrb, rbParams,
            mrb_symbol_value(sym), mrb_float_value(mrb, params[i].value));
      }

      ts_free(NO_ALLOCATOR, params);
    }
  }

  return rbParams;
}

static mrb_value cPotential_apply_forces(mrb_state* mrb, mrb_value self)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);

  struct ts_potential_term_data term_data;
  memset(&term_data, 0, sizeof(term_data));

  term_data.potential = tsmrb_to_potential(mrb, self);
  term_data.nbodies = term_data.potential->nbodies;

  if (argc < 2)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "too few arguments (min 2)");

  ts_sim* sim = tsmrb_to_sim(mrb,argv[0]);
  if (sim == NULL) mrb_raise(mrb, E_TYPE_ERROR, "expected Tubesheet::Sim as first argument");

  int nparticles = argc-1;
  if (nparticles > NBODY_LIMIT)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "too many arguments (max %d)",NBODY_LIMIT+1);
  if (nparticles != term_data.potential->nbodies)
    tsmrb_raisef(mrb,E_ARGUMENT_ERROR,
        "potential requires %d particles, but %d given",
        term_data.potential->nbodies, nparticles);

  vector3 posv[NBODY_LIMIT];
  vector3 forcev[NBODY_LIMIT];
  size_t pinds[NBODY_LIMIT];
  for (int i=1; i < argc; i++)
  {
    ts_sim* psim = NULL;
    size_t pi = tsmrb_to_particle_index(mrb, argv[i], &psim);
    pinds[i-1] = pi;

    if (psim != sim)
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "particles must be from the same simulation");
    vector3_set(posv+i-1, PARTICLE_POS(sim,pi));
    vector3_zero(forcev+i-1);
  }

  double energy = 0.0;
  TSResult err = ts_potential_forces(sim, &term_data, posv, forcev, &energy);
  switch(ERRCODE(err))
  {
    case TS_OKAY:
      break;
    case TS_ERR_INCORRECT_NUM_BODIES:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "potential requires %d particles (%d given)", term_data.potential->nbodies, argc);
      break;
    case TS_ERR_UNSUPPORTED_POTENTIAL:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "%d-body potential is currently unsupported", term_data.potential->nbodies);
      break;
    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "unknown error in apply_forces (error code %04x)", ERRCODE(err));
      break;
  }

  for (int i=0; i < nparticles; i++)
  {
    vector3_set(PARTICLE_FORCE(sim,pinds[i]), forcev+i);
  }

  return mrb_float_value(mrb,energy);
}

static inline void check_potential(mrb_state* mrb, Potential* potential, int argc)
{
  if (potential==NULL) mrb_raise(mrb, E_RUNTIME_ERROR, "potential == NULL");
  if (potential->energy_func == NULL) mrb_raise(mrb, E_RUNTIME_ERROR, "energy function == NULL");
  if (potential->forces_func == NULL) mrb_raise(mrb, E_RUNTIME_ERROR, "forces function == NULL");
  if (argc != potential->nbodies)
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "wrong number of arguments (%d for %d)", argc, potential->nbodies);
}

static void setup_potential_arrays(mrb_state* mrb, int argc, mrb_value* argv, vector3* r, vector3* f)
{
  for (int i=0; i < argc; i++)
    obj_to_vec(mrb, argv[i], r+i);

  if (f) {
    for (int i=0; i < argc; i++)
      vector3_zero(f+i);
  }
}

static mrb_value cPotential_energy(mrb_state* mrb, mrb_value self)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);

  Potential* potential = tsmrb_to_potential(mrb, self);
  check_potential(mrb,potential,argc);

  vector3   r[NBODY_LIMIT];
  setup_potential_arrays(mrb, argc,argv, r,NULL);

  double en = potential->energy_func(potential, r);
  return mrb_float_value(mrb, en);
}

static mrb_value cPotential_forces(mrb_state* mrb, mrb_value self)
{
  int argc = 0;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  Potential* potential = tsmrb_to_potential(mrb, self);
  check_potential(mrb,potential,argc);

  vector3 r[NBODY_LIMIT];
  vector3 f[NBODY_LIMIT];
  setup_potential_arrays(mrb, argc,argv, r,f);

  potential->forces_func(potential, r,f);
  if (argc == 1)
  {
    return tsmrb_vec3_to_ary(mrb, f+0);
  }
  else
  {
    mrb_value f_ary = mrb_ary_new_capa(mrb, argc);
    for (int i=0; i < argc; i++)
      mrb_ary_push(mrb, f_ary, tsmrb_vec3_to_ary(mrb, f+i));

    return f_ary;
  }
}

typedef void (*hessian_func)(Potential*, vector3*, VecReal* hess);
static mrb_value call_hessian(mrb_state* mrb, mrb_value self, hessian_func hfunc)
{
  int argc,i;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  Potential* potential = tsmrb_to_potential(mrb, self);
  check_potential(mrb,potential,argc);

  vector3 r[NBODY_LIMIT];
  vector3 f[NBODY_LIMIT];
  for (i=0; i < NBODY_LIMIT; i++)
  {
    vector3_zero(&r[i]);
    vector3_zero(&f[i]);
  }
  setup_potential_arrays(mrb, argc,argv, r,f);

  int num_dof = 3*potential->nbodies;
  VecReal* hess = mrb_calloc(mrb, hessian_size(num_dof), sizeof(*hess));
  for(int i=0; i < hessian_size(num_dof); i++) { hess[i] = 0.0; }

  hfunc(potential, r, hess);

  mrb_value rb_hess = mrb_ary_new_capa(mrb,num_dof);
  for(int i=0; i < num_dof; i++) {
    int ai = mrb_gc_arena_save(mrb);

    mrb_value hrow = mrb_ary_new_capa(mrb,num_dof);
    for(int j=0; j < num_dof; j++) {
      mrb_ary_push(mrb, hrow, mrb_float_value(mrb, hess[ hessian_index(num_dof,i,j) ]));
    }

    mrb_ary_push(mrb, rb_hess, hrow);
    mrb_gc_arena_restore(mrb,ai);
  }

  mrb_free(mrb, hess);
  
  return rb_hess;
}

static mrb_value cPotential_hessian(mrb_state* mrb, mrb_value self)
{
  return call_hessian(mrb, self, ts_potential_hessian);
}

static mrb_value cPotential_numerical_hessian(mrb_state* mrb, mrb_value self)
{
  return call_hessian(mrb, self, ts_potential_numerical_hessian);
}

static mrb_value cConstantField_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  mrb_value rb_cforce = hash_get_value(mrb, kwArgs, TSMRB_SYM(const_force), "ConstantField.initialize" );
  double refPt        = hash_get_float(mrb, kwArgs, TSMRB_SYM(ref_point)  , "ConstantField.initialize" );

  vector3 v = VEC3_ZERO;
  obj_to_vec(mrb, rb_cforce, &v);

  ConstantField* field = tsp_const_field_new(&v, refPt);
  wrap_potential(mrb,self,(Potential*)field, &constant_field_type);

  return self;
}

static mrb_value cConstantField_get_const_force(mrb_state* mrb, mrb_value self)
{
  ConstantField* field = 
    (ConstantField*)tsmrb_to_potential(mrb, self);

  return tsmrb_from_vec3(mrb, &field->const_force);
}

static mrb_value cConstantField_set_const_force(mrb_state* mrb, mrb_value self)
{
  mrb_value r_force = mrb_nil_value();
  mrb_get_args(mrb, "o", &r_force);

  ConstantField* field = 
    (ConstantField*)tsmrb_to_potential(mrb, self);

  obj_to_vec(mrb, r_force, &field->const_force);

  return self;
}

static mrb_value cConstantField_get_ref_point(mrb_state* mrb, mrb_value self)
{
  ConstantField* field = 
    (ConstantField*)tsmrb_to_potential(mrb, self);

  return mrb_float_value(mrb, field->ref_point);
}

static mrb_value cConstantField_set_ref_point(mrb_state* mrb, mrb_value self)
{
  mrb_float refpt = 0.0;
  mrb_get_args(mrb, "f", &refpt);

  ConstantField* field = 
    (ConstantField*)tsmrb_to_potential(mrb, self);

  field->ref_point = refpt;

  return self;
}

static mrb_value cSpinAroundParticle_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  mrb_value rb_particle = hash_get_value(mrb, kwArgs, TSMRB_SYM(particle), "SpinAroundParticle.initialize" );
  mrb_value rb_axis     = hash_get_value(mrb, kwArgs, TSMRB_SYM(axis), "SpinAroundParticle.initialize" );

  ts_sim* sim = NULL;
  size_t pind = tsmrb_to_particle_index(mrb, rb_particle, &sim);
  if (sim == NULL) mrb_raise(mrb, E_RUNTIME_ERROR, "simulation is NULL!");

  vector3 axis;
  obj_to_vec(mrb, rb_axis, &axis);

  SpinAroundParticle* field = tsp_spin_around_particle_new(sim,pind,&axis);
  wrap_potential(mrb,self,(Potential*)field, &spin_around_particle_type);

  return self;
}

static mrb_value cSpinAroundParticle_particle(mrb_state* mrb, mrb_value self)
{
  SpinAroundParticle* field = 
    (SpinAroundParticle*)tsmrb_to_potential(mrb, self);

  return tsmrb_from_particle(mrb, field->sim, field->pind);
}

static mrb_value cSpinAroundParticle_axis(mrb_state* mrb, mrb_value self)
{
  SpinAroundParticle* field = 
    (SpinAroundParticle*)tsmrb_to_potential(mrb, self);

  return tsmrb_from_vec3(mrb, &field->axis);
}

static mrb_value cPlaneBarrier_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  mrb_value r_pos  = hash_get_value(mrb, kwArgs, TSMRB_SYM(pos) , "plane position" );
  mrb_value r_norm = hash_get_value(mrb, kwArgs, TSMRB_SYM(norm), "plane normal" );
  double dist = hash_get_float(mrb, kwArgs, TSMRB_SYM(dist), "potential distance from plane");
  double eps  = hash_get_float(mrb, kwArgs, TSMRB_SYM(eps) , "potential strength");

  vector3 pos  = VEC3_ZERO;
  vector3 norm = VEC3_ZERO;
  obj_to_vec(mrb, r_pos , &pos);
  obj_to_vec(mrb, r_norm, &norm);

  PlaneBarrier* barrier = tsp_plane_barrier_new(&pos, &norm, dist, eps);
  wrap_potential(mrb,self,(Potential*)barrier, &plane_barrier_type);

  return self;
}

static mrb_value cPlaneBarrier_pos(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return tsmrb_from_vec3(mrb, &potential->pos);
}

static mrb_value cPlaneBarrier_norm(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return tsmrb_from_vec3(mrb, &potential->norm);
}

static mrb_value cPlaneBarrier_dist(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->dist);
}

static mrb_value cPlaneBarrier_eps(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->eps);
}

static mrb_value cPlaneBarrier_sigma(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->sigma);
}

static mrb_value cPlaneBarrier_f_coeff(mrb_state* mrb, mrb_value self)
{
  PlaneBarrier* potential =
    (PlaneBarrier*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->f_coeff);
}

static mrb_value cHarmonic_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  HarmonicPotential* potential = tsp_harmonic_new( 0.0, 0.0 );
  wrap_potential(mrb,self,(Potential*)potential, &harmonic_type);

  potential->k = hash_get_float(mrb, kwArgs, TSMRB_SYM(k), "potential parameters");

  mrb_value r_eq = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(r_eq));
  mrb_value req  = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(req));

  if (!mrb_nil_p(req))
    potential->req = mrb_to_flo(mrb,req);
  else if (!mrb_nil_p(r_eq))
    potential->req = mrb_to_flo(mrb,r_eq);
  else
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "equilibrium distance requires either :req or :r_eq key");

  mrb_value name  = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(name));
  if (! mrb_nil_p(name))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "name keyword no longer supported!");

  return self;
}

static mrb_value cHarmonic_get_k(mrb_state* mrb, mrb_value self)
{
  HarmonicPotential* potential =
    (HarmonicPotential*)tsmrb_to_potential(mrb, self);

  if (potential != NULL)
    return mrb_float_value(mrb,potential->k);
  else
    return mrb_nil_value();
}

static mrb_value cHarmonic_get_r_eq(mrb_state* mrb, mrb_value self)
{
  HarmonicPotential* potential =
    (HarmonicPotential*)tsmrb_to_potential(mrb, self);

  if (potential != NULL)
    return mrb_float_value(mrb,potential->req);
  else
    return mrb_nil_value();
}

static mrb_value cWCA_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  WCAPotential* potential = tsp_wca_new( 0.0, 0.0 );
  wrap_potential(mrb,self,(Potential*)potential, &wca_type);

  double sigma = hash_get_float(mrb, kwArgs, TSMRB_SYM(sigma), "potential parameters");
  double eps   = hash_get_float(mrb, kwArgs, TSMRB_SYM(eps)  , "potential parameters");

  tsp_wca_set_params(potential, sigma, eps);

  return self;
}

static mrb_value cWCA_get_sigma(mrb_state* mrb, mrb_value self)
{
  WCAPotential* potential =
    (WCAPotential*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->sigma);
}

static mrb_value cWCA_get_eps(mrb_state* mrb, mrb_value self)
{
  WCAPotential* potential =
    (WCAPotential*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->eps);
}

static mrb_value cWCA_get_r_min(mrb_state* mrb, mrb_value self)
{
  WCAPotential* potential =
    (WCAPotential*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, potential->r_min);
}

static mrb_value cAngleHarmonic_initialize(mrb_state* mrb,mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  AngleHarmonicPotential* potential = tsp_angle_harmonic_new( 0.0, 0.0 );
  wrap_potential(mrb,self,(Potential*)potential, &angle_harmonic_type);

  potential->k        = hash_get_float(mrb, kwArgs, TSMRB_SYM(k)       , "potential parameters");
  potential->theta_eq = hash_get_float(mrb, kwArgs, TSMRB_SYM(theta_eq), "potential parameters");

  return self;
}

static mrb_value cAngleHarmonic_get_k(mrb_state* mrb, mrb_value self)
{
  AngleHarmonicPotential* pot = 
    (AngleHarmonicPotential*)tsmrb_to_potential(mrb, self);

  return mrb_float_value(mrb, pot->k);
}

static mrb_value cAngleHarmonic_get_theta_eq(mrb_state* mrb, mrb_value self)
{
  AngleHarmonicPotential* pot = 
    (AngleHarmonicPotential*)tsmrb_to_potential(mrb, self);

  return mrb_float_value(mrb, pot->theta_eq);
}

static mrb_value cDihedral_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  DihedralPotential* potential = tsp_dihedral_new( DIHEDRAL_TORSIONAL, 0.0, 0.0 );
  wrap_potential(mrb,self,(Potential*)potential, &dihedral_type);

  double k      = hash_get_float(mrb, kwArgs, TSMRB_SYM(k)     , "potential parameters");
  double phi_eq = hash_get_float(mrb, kwArgs, TSMRB_SYM(phi_eq), "potential parameters");

  mrb_sym func_sym;
  mrb_value func_obj = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(func));
  if (mrb_symbol_p(func_obj))
    func_sym = mrb_symbol(func_obj);
  else if (mrb_nil_p(func_obj))
    func_sym = TSMRB_SYM(torsional);
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected :func argument to be a symbol");
    return mrb_nil_value();
  }

  enum DIHEDRAL_FUNC func = DIHEDRAL_TORSIONAL;
  if (func_sym == TSMRB_SYM(torsional))
    func = DIHEDRAL_TORSIONAL;
  else if (func_sym == TSMRB_SYM(harmonic))
    func = DIHEDRAL_HARMONIC;
  else
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unknown dihedral function %S", mrb_symbol_value(func_sym));

  tsp_dihedral_set_params(potential, k, phi_eq, func);
  return self;
}

static mrb_value cDihedral_get_k(mrb_state* mrb, mrb_value self)
{
  DihedralPotential* pot = 
    (DihedralPotential*)tsmrb_to_potential(mrb, self);

  return mrb_float_value(mrb, pot->k);
}

static mrb_value cDihedral_get_phi_eq(mrb_state* mrb, mrb_value self)
{
  DihedralPotential* pot = 
    (DihedralPotential*)tsmrb_to_potential(mrb, self);

  return mrb_float_value(mrb, pot->phi_eq);
}

static mrb_value cDihedral_get_func(mrb_state* mrb, mrb_value self)
{
  DihedralPotential* pot = 
    (DihedralPotential*)tsmrb_to_potential(mrb, self);

  if (pot->func == DIHEDRAL_TORSIONAL)
    return TSMRB_SYM_OBJ(torsional);
  else if (pot->func == DIHEDRAL_HARMONIC)
    return TSMRB_SYM_OBJ(harmonic);
  else
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "unknown dihedral function value %d", (int)pot->func);

  return mrb_nil_value();
}

static mrb_value cDipole_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  DipolePotential* potential = tsp_dipole_new( 0.0, 0.0 );
  wrap_potential(mrb,self,(Potential*)potential, &dipole_type);

  double k        = hash_get_float(mrb, kwArgs, TSMRB_SYM(k)       , "potential parameters");
  double theta_eq = hash_get_float(mrb, kwArgs, TSMRB_SYM(theta_eq), "potential parameters");

  tsp_dipole_set_params(potential, k, theta_eq);
  return self;
}

static mrb_value cDipole_get_k(mrb_state* mrb, mrb_value self)
{
  DipolePotential* pot = (DipolePotential*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, pot->k);
}

static mrb_value cDipole_get_theta_eq(mrb_state* mrb, mrb_value self)
{
  DipolePotential* pot = (DipolePotential*)tsmrb_to_potential(mrb, self);
  return mrb_float_value(mrb, pot->theta_eq);
}

static struct RClass* define_cPotential(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cPotential = mrb_define_class_under(mrb, mod,
      "Potential", mrb->object_class);
  MRB_SET_INSTANCE_TT(cPotential, MRB_TT_DATA);

  tsmrb_classes(mrb)->cPotential = cPotential;

  mrb_define_method(mrb, cPotential, "initialize"    , cPotential_initialize    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cPotential, "=="            , cPotential_equals        , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cPotential, "nbodies"       , cPotential_get_nbodies   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cPotential, "name"          , cPotential_get_name      , MRB_ARGS_NONE());
  mrb_define_method(mrb, cPotential, "parameters"    , cPotential_get_parameters, MRB_ARGS_NONE());
  mrb_define_method(mrb, cPotential, "parameters="   , cPotential_set_parameters, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cPotential, "set_parameters", cPotential_set_parameters, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cPotential, "apply_forces", cPotential_apply_forces, MRB_ARGS_ANY());

  mrb_define_method(mrb, cPotential, "energy" , cPotential_energy , MRB_ARGS_ANY());
  mrb_define_method(mrb, cPotential, "forces" , cPotential_forces , MRB_ARGS_ANY());
  mrb_define_method(mrb, cPotential, "hessian", cPotential_hessian, MRB_ARGS_ANY());
  mrb_define_method(mrb, cPotential, "numerical_hessian", cPotential_numerical_hessian, MRB_ARGS_ANY());

  return cPotential;
}

static struct RClass* define_cConstantField(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cConstantField = mrb_define_class_under(mrb, module,
      "ConstantField", cPotential);
  MRB_SET_INSTANCE_TT(cConstantField, MRB_TT_DATA);

  mrb_define_method(mrb, cConstantField, "initialize", cConstantField_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cConstantField, "const_force", cConstantField_get_const_force, MRB_ARGS_NONE());
  mrb_define_method(mrb, cConstantField, "const_force=", cConstantField_set_const_force, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cConstantField, "ref_point", cConstantField_get_ref_point, MRB_ARGS_NONE());
  mrb_define_method(mrb, cConstantField, "ref_point=", cConstantField_set_ref_point, MRB_ARGS_REQ(1));

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "const_field") ),
      mrb_obj_value(cConstantField) );

  return cConstantField;
}

static struct RClass* define_cSpinAroundParticle(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cSpinAroundParticle = mrb_define_class_under(mrb, module,
      "SpinAroundParticle", cPotential);
  MRB_SET_INSTANCE_TT(cSpinAroundParticle, MRB_TT_DATA);

  mrb_define_method(mrb, cSpinAroundParticle, "initialize", cSpinAroundParticle_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSpinAroundParticle, "particle", cSpinAroundParticle_particle, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSpinAroundParticle, "axis", cSpinAroundParticle_axis, MRB_ARGS_NONE());

  return cSpinAroundParticle;
}

static struct RClass* define_cPlaneBarrier(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cPlaneBarrier = mrb_define_class_under(mrb, module,
      "PlaneBarrier", cPotential);
  MRB_SET_INSTANCE_TT(cPlaneBarrier, MRB_TT_DATA);

  mrb_define_method(mrb, cPlaneBarrier, "initialize", cPlaneBarrier_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cPlaneBarrier, "pos", cPlaneBarrier_pos, MRB_ARGS_NONE());
  mrb_define_method(mrb, cPlaneBarrier, "norm", cPlaneBarrier_norm, MRB_ARGS_NONE());

  mrb_define_method(mrb, cPlaneBarrier, "eps", cPlaneBarrier_eps, MRB_ARGS_NONE());
  mrb_define_method(mrb, cPlaneBarrier, "dist", cPlaneBarrier_dist, MRB_ARGS_NONE());

  mrb_define_method(mrb, cPlaneBarrier, "sigma", cPlaneBarrier_sigma, MRB_ARGS_NONE());
  mrb_define_method(mrb, cPlaneBarrier, "f_coeff", cPlaneBarrier_f_coeff, MRB_ARGS_NONE());

  return cPlaneBarrier;
}

static struct RClass* define_cHarmonic(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cHarmonic = mrb_define_class_under(mrb, module,
      "Harmonic", cPotential);
  tsmrb_classes(mrb)->cHarmonic = cHarmonic;
  MRB_SET_INSTANCE_TT(cHarmonic, MRB_TT_DATA);

  mrb_define_method(mrb, cHarmonic, "initialize", cHarmonic_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cHarmonic, "k", cHarmonic_get_k, MRB_ARGS_NONE());
  mrb_define_method(mrb, cHarmonic, "r_eq", cHarmonic_get_r_eq, MRB_ARGS_NONE());

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "harmonic") ),
      mrb_obj_value(cHarmonic));

  return cHarmonic;
}

static struct RClass* define_cWCA(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cWCA = mrb_define_class_under(mrb,
      module,
      "WCA",
      cPotential);
  MRB_SET_INSTANCE_TT(cWCA, MRB_TT_DATA);

  mrb_define_method(mrb, cWCA, "initialize", cWCA_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cWCA, "sigma", cWCA_get_sigma, MRB_ARGS_NONE());
  mrb_define_method(mrb, cWCA, "eps"  , cWCA_get_eps  , MRB_ARGS_NONE());
  mrb_define_method(mrb, cWCA, "r_min", cWCA_get_r_min, MRB_ARGS_NONE());

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "wca") ),
      mrb_obj_value(cWCA));

  return cWCA;
}

static struct RClass* define_cAngleHarmonic(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cAngleHarmonic = mrb_define_class_under(mrb,
      module, "AngleHarmonic", cPotential);
  MRB_SET_INSTANCE_TT(cAngleHarmonic, MRB_TT_DATA);

  mrb_define_method(mrb, cAngleHarmonic, "initialize", cAngleHarmonic_initialize, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cAngleHarmonic, "k", cAngleHarmonic_get_k, MRB_ARGS_NONE());
  mrb_define_method(mrb, cAngleHarmonic, "theta_eq", cAngleHarmonic_get_theta_eq, MRB_ARGS_NONE());

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "angle") ),
      mrb_obj_value(cAngleHarmonic));

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "angle_harmonic") ),
      mrb_obj_value(cAngleHarmonic));

  return cAngleHarmonic;
}

static struct RClass* define_cDihedral(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cDihedral = mrb_define_class_under(mrb,
        module, "Dihedral", cPotential); 
  MRB_SET_INSTANCE_TT(cDihedral, MRB_TT_DATA);

  mrb_define_method(mrb, cDihedral, "initialize", cDihedral_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cDihedral, "k"     , cDihedral_get_k     , MRB_ARGS_NONE());
  mrb_define_method(mrb, cDihedral, "phi_eq", cDihedral_get_phi_eq, MRB_ARGS_NONE());
  mrb_define_method(mrb, cDihedral, "func"  , cDihedral_get_func  , MRB_ARGS_NONE());

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "dihedral") ),
      mrb_obj_value(cDihedral));

  return cDihedral;
}

static struct RClass* define_cDipole(mrb_state* mrb, struct RClass* module, struct RClass* cPotential)
{
  struct RClass* cDipole = mrb_define_class_under(mrb,
      module, "Dipole", cPotential);
  MRB_SET_INSTANCE_TT(cDipole, MRB_TT_DATA);

  mrb_define_method(mrb, cDipole, "initialize", cDipole_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cDipole, "k", cDipole_get_k, MRB_ARGS_NONE());
  mrb_define_method(mrb, cDipole, "theta_eq", cDipole_get_theta_eq, MRB_ARGS_NONE());

  mrb_hash_set(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials,
      mrb_symbol_value( mrb_intern_cstr(mrb, "dipole") ),
      mrb_obj_value(cDipole));

  return cDipole;
}

static mrb_value new_potential(mrb_state* mrb, mrb_value mod)
{
  mrb_sym name;
  mrb_value kwArgs;
  mrb_get_args(mrb, "nH", &name, &kwArgs);

  mrb_value pot_cls = mrb_hash_get(mrb,
      tsmrb_classes(mrb)->tblNamedPotentials, mrb_symbol_value(name));

  if (mrb_nil_p(pot_cls))
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unknown potential :%S", mrb_symbol_value(name));

  return mrb_funcall(mrb, pot_cls, "new", 1, kwArgs);
}

struct RClass* tsmrb_define_potentials(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_potentials_module(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Potentials module has already been initialized!");

  struct RClass* mPotentials = mrb_define_module_under(mrb, mod, "Potentials");
  tsmrb_classes(mrb)->mPotentials = mPotentials;

  struct RClass* cPotential = define_cPotential(mrb, mPotentials);

  mrb_value tblNamedPotentials = mrb_hash_new(mrb);
  tsmrb_classes(mrb)->tblNamedPotentials = tblNamedPotentials;

  mrb_obj_iv_set(mrb,
      (struct RObject*)cPotential,
      mrb_intern_cstr(mrb, "@potential_table"),
      tblNamedPotentials);

  tsmrb_classes(mrb)->cConstantField = define_cConstantField(mrb, mPotentials, cPotential);
  tsmrb_classes(mrb)->cSpinAroundParticle = define_cSpinAroundParticle(mrb, mPotentials, cPotential);
  tsmrb_classes(mrb)->cPlaneBarrier = define_cPlaneBarrier(mrb, mPotentials, cPotential);

  tsmrb_classes(mrb)->cHarmonic = define_cHarmonic(mrb, mPotentials, cPotential);
  tsmrb_classes(mrb)->cWCA = define_cWCA(mrb, mPotentials, cPotential);

  tsmrb_classes(mrb)->cAngleHarmonic = define_cAngleHarmonic(mrb, mPotentials, cPotential);

  struct RClass* cDihedral = define_cDihedral(mrb, mPotentials, cPotential);
  tsmrb_classes(mrb)->cDihedral = cDihedral;

  mrb_define_const(mrb, mPotentials, "DIHEDRAL_TORSIONAL", mrb_float_value(mrb,DIHEDRAL_TORSIONAL) );
  mrb_define_const(mrb, mPotentials, "DIHEDRAL_HARMONIC" , mrb_float_value(mrb,DIHEDRAL_HARMONIC ) );

  tsmrb_classes(mrb)->cDipole = define_cDipole(mrb, mPotentials, cPotential);

  mrb_define_module_function(mrb, mod, "new_potential",
      new_potential, MRB_ARGS_REQ(2));

  return mPotentials;
}

