#include "tsmrb_quat.h"
#include "ts_units.h"
#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"

#include "mruby/data.h"
#include "mruby/array.h"

static const mrb_data_type quat_type = {
  .struct_name = "Quat",
  .dfree = tsmrb_free,
};

quat4* tsmrb_to_quat(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_quat_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Quat object");

  return DATA_PTR(obj);
}

static inline quat4* quat_alloc(mrb_state* mrb)
{
  quat4* q = mrb_malloc(mrb, sizeof(*q));
  memset(q,0,sizeof(*q));
  return q;
}

static mrb_value quat_wrap(mrb_state* mrb, quat4* q)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_quat_class(mrb), q, &quat_type);
  return mrb_obj_value(data);
}

static mrb_value create_new_quat(mrb_state* mrb, quat4** qp)
{
  *qp = quat_alloc(mrb);
  return quat_wrap(mrb, *qp);
}

mrb_value tsmrb_new_quat(mrb_state* mrb, quat4** qp) { return create_new_quat(mrb,qp); }

mrb_value tsmrb_from_quat(mrb_state* mrb, const quat4*restrict q)
{
  quat4* qnew = NULL;
  mrb_value rq = create_new_quat(mrb, &qnew);
  quat4_set(qnew, q);

  return rq;
}

static mrb_value cQuat_initialize(mrb_state* mrb, mrb_value self)
{
  quat4* q = quat_alloc(mrb);
  quat4_identity(q);

  DATA_TYPE(self) = &quat_type;
  DATA_PTR(self) = q;

  int argc;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  if (argc == 0)
  {
    quat4_identity(q);
  }
  else if (argc == 4)
  {
    quat4_set_elements(q,
        mrb_to_float(mrb, argv[0]),
        mrb_to_float(mrb, argv[1]),
        mrb_to_float(mrb, argv[2]),
        mrb_to_float(mrb, argv[3]));
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "Quat.initialize requires either 0 or 4 arguments");
  }

  return self;
}

static mrb_value cQuat_identity(mrb_state* mrb, mrb_value self)
{
  quat4* q = NULL;
  mrb_value rq = create_new_quat(mrb, &q);
  quat4_identity(q);
  return rq;
}

void ary_to_quat(mrb_state* mrb, mrb_value obj, quat4* q)
{
  mrb_check_type(mrb, obj, MRB_TT_ARRAY);

  if (RARRAY_LEN(obj) != 4)
    mrb_raise(mrb, E_TYPE_ERROR, "expected a three-element array");

  if (q != NULL)
  {
    quat4_set_elements(q,
        mrb_to_float(mrb, RARRAY_PTR(obj)[0]),
        mrb_to_float(mrb, RARRAY_PTR(obj)[1]),
        mrb_to_float(mrb, RARRAY_PTR(obj)[2]),
        mrb_to_float(mrb, RARRAY_PTR(obj)[3]));
  }
}

mrb_value quat_to_ary(mrb_state* mrb, quat4* q)
{
  mrb_value arr[4];
  arr[0] = mrb_float_value(mrb, quat4_w(q));
  arr[1] = mrb_float_value(mrb, quat4_x(q));
  arr[2] = mrb_float_value(mrb, quat4_y(q));
  arr[3] = mrb_float_value(mrb, quat4_z(q));

  return mrb_ary_new_from_values(mrb, 4, arr);
}


static mrb_value cQuat_euler_XYZ(mrb_state* mrb, mrb_value self)
{
  mrb_float phi, theta, psi;
  mrb_get_args(mrb, "fff", &phi, &theta, &psi);

  quat4* q = NULL;
  mrb_value rq = create_new_quat(mrb, &q);

  vector3 euler = VEC3_INIT(
      (VecReal)phi,
      (VecReal)theta,
      (VecReal)psi
      );

  quat4_from_euler_xyz(q, &euler);
  return rq;
}

static mrb_value cQuat_euler_ZXZ(mrb_state* mrb, mrb_value self)
{
  mrb_float phi, theta, psi;
  mrb_get_args(mrb, "fff", &phi, &theta, &psi);

  quat4* q = NULL;
  mrb_value rq = create_new_quat(mrb,&q);

  vector3 euler = VEC3_INIT(
      (VecReal)phi,
      (VecReal)theta,
      (VecReal)psi
      );

  quat4_from_euler_zxz(q, &euler);
  return rq;
}

#define CHECK_TOL 1e-6
static int check_eulers(VecReal eulerA[3], VecReal eulerB[3])
{
  if ( fabs(eulerA[0] - eulerB[0]) >= CHECK_TOL ) return 1;
  if ( fabs(eulerA[1] - eulerB[1]) >= CHECK_TOL ) return 2;
  if ( fabs(eulerA[2] - eulerB[2]) >= CHECK_TOL ) return 3;
  return 0;
}

static int check_quats(const quat4* qA, const quat4* qB)
{
  if ( fabs(QUAT4_ELT(qA,0) - QUAT4_ELT(qB,0)) >= CHECK_TOL ) return 1;
  if ( fabs(QUAT4_ELT(qA,1) - QUAT4_ELT(qB,1)) >= CHECK_TOL ) return 2;
  if ( fabs(QUAT4_ELT(qA,2) - QUAT4_ELT(qB,2)) >= CHECK_TOL ) return 3;
  if ( fabs(QUAT4_ELT(qA,3) - QUAT4_ELT(qB,3)) >= CHECK_TOL ) return 4;
  return 0;
}
#undef CHECK_TOL

static void cQuat_euler_ZXZ_tests(mrb_state* mrb, mrb_int ddeg)
{
  vector3 Xaxis = VEC3_INIT(1.0, 0.0, 0.0);
  vector3 Zaxis = VEC3_INIT(0.0, 0.0, 1.0);

  VecReal arr[3] = { 0.0, 0.0, 0.0 };

  char buf[128];
  for (int phi_deg = -179; phi_deg < 180; phi_deg += ddeg)
  {
    arr[0] = M_PI*phi_deg/180.0;
    quat4 q3;
    quat4_set_axis_angle(&q3, &Zaxis, arr[0]);
    for (int theta_deg = 1; theta_deg < 180; theta_deg += ddeg)
    {
      arr[1] = M_PI*theta_deg/180.0;
      quat4 q2;
      quat4_set_axis_angle(&q2, &Xaxis, arr[1]);
      for (int psi_deg = -179; psi_deg < 180; psi_deg += ddeg)
      {
        arr[2] = M_PI*psi_deg/180.0;
        quat4 q1;
        quat4_set_axis_angle(&q1, &Zaxis, arr[2]);

        quat4 q_rot;
        quat4_set(&q_rot,&q1);
        quat4_lmul(&q_rot,&q2);
        quat4_lmul(&q_rot,&q3);

        quat4 q;
        quat4_from_euler_zxz_array(&q, arr);

        VecReal eulers[3] = { 0.0, 0.0, 0.0 };
        quat4_to_euler_zxz_array(&q_rot, eulers);

        int ind;
        if ((ind = check_quats(&q, &q_rot)) > 0)
        {
          snprintf(buf, sizeof(buf)-1,
              "q_rot disagrees with q at index %d\n"
              "    q_rot (%7.4f, %7.4f, %7.4f, %7.4f)\n"
              "    q     (%7.4f, %7.4f, %7.4f, %7.4f)"
              "  phi %3d deg, theta %3d deg, psi %3d deg",
              ind,
              QUAT4_ELT(&q_rot,0), QUAT4_ELT(&q_rot,1),
              QUAT4_ELT(&q_rot,2), QUAT4_ELT(&q_rot,3),

              QUAT4_ELT(&q,0), QUAT4_ELT(&q,1),
              QUAT4_ELT(&q,2), QUAT4_ELT(&q,3),
              phi_deg, theta_deg, psi_deg);
          mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
        }

        quat4_to_euler_zxz_array(&q, eulers);
        if ((ind = check_eulers(arr, eulers)) > 0)
        {
          snprintf(buf, sizeof(buf)-1,
              "q has wrong Euler angles (%7.4f, %7.4f, %7.4f),\n"
              "  at index %d expected (%7.4f,%7.4f,%7.4f)\n"
              "  phi %3d deg, theta %3d deg, psi %3d deg",
              eulers[0], eulers[1], eulers[2],
              ind, arr[0], arr[1], arr[2],
              phi_deg, theta_deg, psi_deg);
          mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
        }

        if ((ind = check_eulers(arr, eulers)) > 0)
        {
          snprintf(buf, sizeof(buf)-1,
              "q_rot has wrong Euler angles (%7.4f, %7.4f, %7.4f),\n"
              "      at index %d expected (%7.4f,%7.4f,%7.4f)\n"
              "      phi %3d deg, theta %3d deg, psi %3d deg",
              eulers[0], eulers[1], eulers[2],
              ind, arr[0], arr[1], arr[2],
              phi_deg, theta_deg, psi_deg);
          mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
        }
      }
    }
  }
}

static void cQuat_euler_XYZ_tests(mrb_state* mrb, mrb_int ddeg)
{
  vector3 Xaxis = VEC3_INIT(1.0, 0.0, 0.0);
  vector3 Yaxis = VEC3_INIT(0.0, 1.0, 0.0);
  vector3 Zaxis = VEC3_INIT(0.0, 0.0, 1.0);

  VecReal arr[3] = { 0.0, 0.0, 0.0 };

  char buf[128];

  for (int alpha_deg = -179; alpha_deg < 180; alpha_deg += ddeg)
  {
    arr[0] = M_PI*alpha_deg/180.0;
    quat4 q3;
    quat4_set_axis_angle(&q3, &Xaxis, arr[0]);
    for (int beta_deg = -89; beta_deg < 89; beta_deg += ddeg)
    {
      arr[1] = M_PI*beta_deg/180.0;
      quat4 q2;
      quat4_set_axis_angle(&q2, &Yaxis, arr[1]);
      for (int gamma_deg = -179; gamma_deg < 180; gamma_deg += ddeg)
      {
        arr[2] = M_PI*gamma_deg/180.0;
        quat4 q1;
        quat4_set_axis_angle(&q1, &Zaxis, arr[2]);

        quat4 q_rot;
        quat4_set(&q_rot,&q1);
        quat4_lmul(&q_rot,&q2);
        quat4_lmul(&q_rot,&q3);

        quat4 q;
        quat4_from_euler_xyz_array(&q, arr);

        VecReal eulers[3] = { 0.0, 0.0, 0.0 };
        quat4_to_euler_xyz_array(&q_rot, eulers);

        int ind;
        if ((ind = check_quats(&q, &q_rot)) > 0)
        {
          snprintf(buf, sizeof(buf)-1,
              "[%s:%d] "
              "q_rot disagrees with q at index %d\n"
              "    q_rot (%7.4f, %7.4f, %7.4f, %7.4f)\n"
              "    q     (%7.4f, %7.4f, %7.4f, %7.4f)"
              "  alpha %3d deg, beta %3d deg, gamma %3d deg",
              __FILE__, __LINE__,
              ind,
              QUAT4_ELT(&q_rot,0), QUAT4_ELT(&q_rot,1),
              QUAT4_ELT(&q_rot,2), QUAT4_ELT(&q_rot,3),

              QUAT4_ELT(&q,0), QUAT4_ELT(&q,1),
              QUAT4_ELT(&q,2), QUAT4_ELT(&q,3),
              alpha_deg, beta_deg, gamma_deg);
          mrb_raise(mrb, E_RUNTIME_ERROR, buf);
        }

        quat4_to_euler_xyz_array(&q, eulers);
        if ((ind = check_eulers(arr, eulers)) > 0)
        {
          snprintf(buf,sizeof(buf)-1,
              "[%s:%d] "
              "q has wrong Euler angles (%7.4f, %7.4f, %7.4f),\n"
              "  at index %d expected (%7.4f,%7.4f,%7.4f)\n"
              "  alpha %3d deg, beta %3d deg, gamma %3d deg",
              __FILE__, __LINE__,
              eulers[0], eulers[1], eulers[2],
              ind, arr[0], arr[1], arr[2],
              alpha_deg, beta_deg, gamma_deg);
          mrb_raise(mrb, E_RUNTIME_ERROR, buf);
        }

        if ((ind = check_eulers(arr, eulers)) > 0)
        {
          snprintf(buf, sizeof(buf)-1,
              "[%s:%d] "
              "q_rot has wrong Euler angles (%7.4f, %7.4f, %7.4f),\n"
              "      at index %d expected (%7.4f,%7.4f,%7.4f)\n"
              "      alpha %3d deg, beta %3d deg, gamma %3d deg",
              __FILE__, __LINE__,
              eulers[0], eulers[1], eulers[2],
              ind, arr[0], arr[1], arr[2],
              alpha_deg, beta_deg, gamma_deg);
          mrb_raise(mrb, E_RUNTIME_ERROR, buf);
        }
      }
    }
  }
}

static mrb_value cQuat_euler_tests(mrb_state* mrb, mrb_value mod)
{
  mrb_int ddeg = 0;
  mrb_get_args(mrb, "i", &ddeg);

  cQuat_euler_ZXZ_tests(mrb, ddeg);
  cQuat_euler_XYZ_tests(mrb, ddeg);

  return mrb_nil_value();
}

static inline mrb_value
rot_axis(mrb_state* mrb, mrb_float x, mrb_float y, mrb_float z)
{
  mrb_float ang;
  mrb_get_args(mrb, "f", &ang);

  vector3 ax;
  vector3_set_with(&ax, x,y,z);

  quat4* q = NULL;
  mrb_value rq = create_new_quat(mrb,&q);
  quat4_set_axis_angle(q, &ax, ang);

  return rq;
}

static mrb_value cQuat_rotX(mrb_state* mrb, mrb_value mod)
{
  return rot_axis(mrb, 1.0, 0.0, 0.0);
}

static mrb_value cQuat_rotY(mrb_state* mrb, mrb_value mod)
{
  return rot_axis(mrb, 0.0, 1.0, 0.0);
}

static mrb_value cQuat_rotZ(mrb_state* mrb, mrb_value mod)
{
  return rot_axis(mrb, 0.0, 0.0, 1.0);
}

static mrb_value cQuat_set_identityB(mrb_state* mrb, mrb_value self)
{
  quat4_identity(tsmrb_to_quat(mrb,self));
  return self;
}

static mrb_value cQuat_copy(mrb_state* mrb, mrb_value self)
{
  quat4* q = NULL;
  mrb_value rqnew = create_new_quat(mrb,&q);
  quat4_set(q, tsmrb_to_quat(mrb,self));

  return rqnew;
}

static inline mrb_value
cQuat_get_elt(mrb_state* mrb, mrb_value self, int elt)
{
  return mrb_float_value(mrb, QUAT4_ELT(tsmrb_to_quat(mrb,self),elt));
}

static inline mrb_value
cQuat_set_elt(mrb_state* mrb, mrb_value self, int elt)
{
  mrb_float f;
  mrb_get_args(mrb, "f", &f);
  QUAT4_ELT(tsmrb_to_quat(mrb,self),elt) = (VecReal)f;
  return self;
}

static mrb_value cQuat_get_x(mrb_state* mrb, mrb_value self) { return cQuat_get_elt(mrb,self,0); }
static mrb_value cQuat_get_y(mrb_state* mrb, mrb_value self) { return cQuat_get_elt(mrb,self,1); }
static mrb_value cQuat_get_z(mrb_state* mrb, mrb_value self) { return cQuat_get_elt(mrb,self,2); }
static mrb_value cQuat_get_w(mrb_state* mrb, mrb_value self) { return cQuat_get_elt(mrb,self,3); }

static mrb_value cQuat_set_x(mrb_state* mrb, mrb_value self) { return cQuat_set_elt(mrb,self,0); }
static mrb_value cQuat_set_y(mrb_state* mrb, mrb_value self) { return cQuat_set_elt(mrb,self,1); }
static mrb_value cQuat_set_z(mrb_state* mrb, mrb_value self) { return cQuat_set_elt(mrb,self,2); }
static mrb_value cQuat_set_w(mrb_state* mrb, mrb_value self) { return cQuat_set_elt(mrb,self,3); }

static mrb_value cQuat_to_euler(mrb_state* mrb, mrb_value self)
{
  int argc;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  quat4* q = tsmrb_to_quat(mrb,self);
  vector3 euler;

  char buf[64];
  if (argc > 1)
  {
    snprintf(buf, sizeof(buf)-1,
        "wrong number of arguments (%d for 0 or 1)", argc);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  enum TS_ROT_REPR rot_repr = TS_ROT_EULER_ZXZ;
  if (argc == 1)
  {
    mrb_check_type(mrb, argv[0], MRB_TT_SYMBOL);
    if (mrb_symbol(argv[0]) == TSMRB_SYM( ZXZ ))
    {
      rot_repr = TS_ROT_EULER_ZXZ;
    }
    else if (mrb_symbol(argv[0]) == TSMRB_SYM( XYZ ))
    {
      rot_repr = TS_ROT_EULER_XYZ;
    }
    else
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid rotational representation");
    }
  }

  quat4_to_euler(q, rot_repr, &euler);

  return tsmrb_from_vec3(mrb,&euler);
}

static mrb_value cQuat_rmulB(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);
 
  quat4* q = tsmrb_to_quat(mrb,self);

  if (mrb_numeric_p(other))
  {
    quat4_scale_by(q, mrb_to_float(mrb,other));
  }
  else if (mrb_obj_is_instance_of(mrb,other,tsmrb_quat_class(mrb)))
  {
    quat4_rmul(q, tsmrb_to_quat(mrb,other));
  }
  else
  {
    mrb_raise(mrb, E_TYPE_ERROR, "illegal type for right-multiply");
  }

  return self;
}

static mrb_value cQuat_rmul(mrb_state* mrb, mrb_value self)
{
  quat4* q = NULL;
  mrb_value rqnew = create_new_quat(mrb,&q);
  quat4_set(q, tsmrb_to_quat(mrb,self));

  return cQuat_rmulB(mrb, rqnew);
}

static mrb_value cQuat_negate(mrb_state* mrb, mrb_value self)
{
  quat4* qnew = NULL;
  mrb_value rqnew = create_new_quat(mrb,&qnew);

  quat4* q = tsmrb_to_quat(mrb,self);
  quat4_set_elements(qnew,
      -quat4_w(q), -quat4_x(q), -quat4_y(q), -quat4_z(q));

  return rqnew;
}

static mrb_value cQuat_negateB(mrb_state* mrb, mrb_value self)
{
  quat4* q = tsmrb_to_quat(mrb,self);
  QUAT4_W(q) = -QUAT4_W(q);
  QUAT4_X(q) = -QUAT4_X(q);
  QUAT4_Y(q) = -QUAT4_Y(q);
  QUAT4_Z(q) = -QUAT4_Z(q);

  return self;
}

static mrb_value cQuat_conjB(mrb_state* mrb, mrb_value self)
{
  quat4* q = tsmrb_to_quat(mrb,self);
  quat4_conj(q);
  return self;
}

static mrb_value cQuat_conj(mrb_state* mrb, mrb_value self)
{
  return cQuat_conjB(mrb, cQuat_copy(mrb,self) );
}

static mrb_value cQuat_norm(mrb_state* mrb, mrb_value self)
{
  return mrb_float_value(mrb, (mrb_float)quat4_norm(tsmrb_to_quat(mrb,self)));
}

static mrb_value cQuat_vecnorm(mrb_state* mrb, mrb_value self)
{
  return mrb_float_value(mrb, (mrb_float)vector3_norm((vector3*)tsmrb_to_quat(mrb,self)));
}

static mrb_value cQuat_normalizeB(mrb_state* mrb, mrb_value self)
{
  quat4_normalize( tsmrb_to_quat(mrb,self) );
  return self;
}

static mrb_value cQuat_normalize(mrb_state* mrb, mrb_value self)
{
  return cQuat_normalizeB(mrb, cQuat_copy(mrb,self) );
}

static mrb_value cQuat_add(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  quat4* a = tsmrb_to_quat(mrb,self);
  quat4* b = tsmrb_to_quat(mrb,other);

  quat4* result = NULL;
  mrb_value r_result = create_new_quat(mrb,&result);

  quat4_add(result, a,b);
  return r_result;
}

static mrb_value cQuat_sub(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  quat4* result = NULL;
  mrb_value r_result = create_new_quat(mrb,&result);

  quat4* a = tsmrb_to_quat(mrb,self);
  quat4* b = tsmrb_to_quat(mrb,other);

  quat4_sub(result, a,b);
  return r_result;
}

static mrb_value cQuat_rotate(mrb_state* mrb, mrb_value self)
{
  mrb_value rvec;
  mrb_get_args(mrb, "o", &rvec);

  quat4* q = tsmrb_to_quat(mrb,self);

  vector3 v;
  obj_to_vec(mrb, rvec, &v);

  quat4_rotate_vector(q, &v);

  return tsmrb_from_vec3(mrb, &v);
}

static mrb_value cQuat_rotatev(mrb_state* mrb, mrb_value self)
{
  mrb_value rvec = mrb_nil_value();
  mrb_get_args(mrb, "o", &rvec);

  quat4* q = tsmrb_to_quat(mrb,self);

  vector3 v;
  obj_to_vec(mrb, rvec, &v);

  quat4_rotate_vector(q, &v);

  return tsmrb_from_vec3(mrb,&v);
}

static mrb_value cQuat_axis_angle(mrb_state* mrb, mrb_value cls)
{
  int argc;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  vector3 vec;
  double ang = 0.0;

  quat4* q = NULL;
  mrb_value rq = create_new_quat(mrb,&q);

  if (argc == 1)
  {
    obj_to_vec(mrb, argv[0], &vec);
    quat4_from_axis_angle( q, &vec);
  }
  else if (argc == 2)
  {
    obj_to_vec(mrb, argv[0], &vec);
    vector3_normalize(&vec);
    ang = mrb_to_float(mrb, argv[1]);
    quat4_set_axis_angle( q, &vec, ang );
  }
  else
  {
    char buf[64];
    snprintf(buf, sizeof(buf),
        "wrong number of arguments (%d for 1-2)", argc);
    mrb_raise(mrb, E_ARGUMENT_ERROR, buf);
  }

  return rq;
}

static mrb_value cQuat_to_axis_angle(mrb_state* mrb, mrb_value self)
{
  vector3 aa;
  vector3_zero(&aa);
  quat4_to_axis_angle( tsmrb_to_quat(mrb,self), &aa);

  return tsmrb_from_vec3(mrb, &aa);
}

static mrb_value cQuat_to_axis_and_angle(mrb_state* mrb, mrb_value self)
{
  vector3 aa;
  quat4* q = tsmrb_to_quat(mrb,self);
  double ang = quat4_to_axis_angle_separately(q, &aa);

  mrb_value ang_axis_pair[2];
  ang_axis_pair[0] = mrb_float_value(mrb, ang);
  ang_axis_pair[1] = tsmrb_from_vec3(mrb, &aa);
  return mrb_ary_new_from_values(mrb, 2, ang_axis_pair);
}

static mrb_value cQuat_eql_p(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  return mrb_bool_value(
      mrb_obj_is_instance_of(mrb, other, tsmrb_quat_class(mrb)) &&
      quat4_equal(
        tsmrb_to_quat(mrb,self),
        tsmrb_to_quat(mrb,other)
        )
      );
}

static mrb_value cQuat_equals(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  quat4* q = tsmrb_to_quat(mrb,self);

  if (mrb_obj_is_instance_of(mrb, other, tsmrb_quat_class(mrb)))
  {
    quat4* q2 = tsmrb_to_quat(mrb,other);
    return mrb_bool_value(quat4_equal(q,q2));
  }
  else if (mrb_type(other) == MRB_TT_ARRAY)
  {
    quat4 qother;
    ary_to_quat(mrb, other, &qother);

    return mrb_bool_value(quat4_equal(q,&qother));
  }
  else
  {
    return mrb_bool_value(0);
  }
}

static mrb_value cQuat_distance_to(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  quat4* q1 = tsmrb_to_quat(mrb,self);
  quat4* q2 = tsmrb_to_quat(mrb,other);

  return mrb_float_value(mrb, quat4_distance_to(q1,q2));
}

static mrb_value cQuat_to_a(mrb_state* mrb, mrb_value self)
{
  quat4* q = tsmrb_to_quat(mrb,self);
  mrb_value vals[4];
  vals[0] = mrb_float_value(mrb, quat4_w(q) );
  vals[1] = mrb_float_value(mrb, quat4_x(q) );
  vals[2] = mrb_float_value(mrb, quat4_y(q) );
  vals[3] = mrb_float_value(mrb, quat4_z(q) );

  return mrb_ary_new_from_values(mrb, 4, vals);
}

static mrb_value cQuat_inspect(mrb_state* mrb, mrb_value self)
{
  char buf[64];
  quat4* q = tsmrb_to_quat(mrb,self);

  snprintf(buf,sizeof(buf), "Quat [ %7.4f %7.4f %7.4f %7.4f ] @ %p",
      quat4_w(q),
      quat4_x(q),
      quat4_y(q),
      quat4_z(q),
      (void*)q);

  return mrb_str_new_cstr(mrb, buf);
}

struct RClass* tsmrb_define_quat(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_quat_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Quat class has already been initialized!");

  struct RClass* cQuat = mrb_define_class_under(mrb, mod, "Quat", mrb->object_class);
  tsmrb_classes(mrb)->cQuat = cQuat;
  MRB_SET_INSTANCE_TT(cQuat, MRB_TT_DATA);

  mrb_define_method(mrb, cQuat, "initialize", cQuat_initialize, MRB_ARGS_ANY());

  mrb_define_method(mrb, cQuat, "to_a", cQuat_to_a, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "inspect", cQuat_inspect, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "set_identity!", cQuat_set_identityB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "copy", cQuat_copy, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "x", cQuat_get_x, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "y", cQuat_get_y, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "z", cQuat_get_z, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "w", cQuat_get_w, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "x=", cQuat_set_x, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "y=", cQuat_set_y, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "z=", cQuat_set_z, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "w=", cQuat_set_w, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "to_euler", cQuat_to_euler, MRB_ARGS_ANY());

  mrb_define_method(mrb, cQuat, "-@", cQuat_negate, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "negate", cQuat_negate, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "negate!", cQuat_negateB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "conj", cQuat_conj, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "conj!", cQuat_conjB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "vecnorm", cQuat_vecnorm, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "norm", cQuat_norm, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "normalize", cQuat_normalize, MRB_ARGS_NONE());
  mrb_define_method(mrb, cQuat, "normalize!", cQuat_normalizeB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cQuat, "+", cQuat_add, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "-", cQuat_sub, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "mul!", cQuat_rmulB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "*=", cQuat_rmulB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "mul", cQuat_rmul, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "*", cQuat_rmul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "rotate", cQuat_rotate, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "rotatev", cQuat_rotatev, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "eql?", cQuat_eql_p, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "==", cQuat_equals, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cQuat, "distance_to", cQuat_distance_to, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "axis_angle", cQuat_axis_angle, MRB_ARGS_ANY());
  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "identity", cQuat_identity, MRB_ARGS_NONE());

  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "euler", cQuat_euler_ZXZ, MRB_ARGS_REQ(3));
  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "euler_ZXZ", cQuat_euler_ZXZ, MRB_ARGS_REQ(3));
  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "euler_XYZ", cQuat_euler_XYZ, MRB_ARGS_REQ(3));

  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "rotX", cQuat_rotX, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "rotY", cQuat_rotY, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject*)cQuat, "rotZ", cQuat_rotZ, MRB_ARGS_REQ(1));

  mrb_define_singleton_method(mrb, (struct RObject*)cQuat,
      "euler_tests", cQuat_euler_tests, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cQuat, "to_axis_angle", cQuat_to_axis_angle, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cQuat, "to_axis_and_angle", cQuat_to_axis_and_angle, MRB_ARGS_NONE());

  return cQuat;
}

