#include "tsmrb_sim.h"

#include "mruby/data.h"

#include "tsmrb_helpers.h"
#include "tsmrb_vec.h"
#include "tsmrb_quat.h"

#include "tubesheet.h"
#include "ts_bonds.h"
#include "ts_dof_map.h"

#include "lbfgs.h"

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
};

struct tsmrb_rigid
{
  ts_sim* sim;
  size_t rb_ind;
};

static const mrb_data_type rigid_body_type = {
  .struct_name = "RigidBody",
  .dfree = tsmrb_free,
};

static inline struct tsmrb_rigid*
get_rigid_wrap(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_rigid_body_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a RigidBody object");

  struct tsmrb_rigid* wrap = DATA_PTR(obj);
  if (wrap == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "RigidBody internal pointer is NULL");

  return wrap;
}

ts_rigid_body* tsmrb_to_rigid_body(mrb_state* mrb, mrb_value obj)
{
  struct tsmrb_rigid* wrap = get_rigid_wrap(mrb,obj);
  return ts_get_rigid_body(wrap->sim, wrap->rb_ind);
}

size_t tsmrb_to_rigid_body_index(mrb_state* mrb, mrb_value obj, ts_sim** simptr)
{
  struct tsmrb_rigid* wrap = get_rigid_wrap(mrb,obj);
  *simptr = wrap->sim;
  return wrap->rb_ind;
}

mrb_value tsmrb_from_rigid_body(mrb_state* mrb, ts_sim* sim, size_t rb_ind)
{
  struct tsmrb_rigid* wrap = mrb_malloc(mrb, sizeof(*wrap));
  wrap->sim = sim;
  wrap->rb_ind = rb_ind;

  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_rigid_body_class(mrb), wrap, &rigid_body_type);

  return mrb_obj_value(data);
}

static mrb_value cRigidBody_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_raise(mrb, E_RUNTIME_ERROR,
      "RigidBody must be created through a simulation");
  return self;
}

static mrb_value cRigidBody_equal(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  if (mrb_obj_is_instance_of(mrb,other,tsmrb_rigid_body_class(mrb)))
  {
    ts_sim* sim1 = NULL;
    ts_sim* sim2 = NULL;
    size_t rb1 = tsmrb_to_rigid_body_index(mrb, self, &sim1);
    size_t rb2 = tsmrb_to_rigid_body_index(mrb, other, &sim2);

    return mrb_bool_value( (sim1 == sim2) && (rb1 == rb2) );
  }

  return mrb_bool_value(0);
}

static mrb_value cRigidBody_inspect(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* rigid = ts_get_rigid_body(sim,rb);

  char buf[256];
  snprintf(buf,sizeof(buf)-1,
      "#<Tubesheet::RigidBody::%p [Sim %p, Index %lu]>",
      rigid, sim, (unsigned long)rb);

  return mrb_str_new_cstr(mrb, buf);
}

static void check_body_type(mrb_state* mrb, ts_rigid_body* body)
{
  if (body->type_ptr == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "rigid body type pointer is NULL");
}

static mrb_value cRigidBody_get_type(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  ts_rigid_body* rigid = ts_get_rigid_body(sim,rb);
  struct ts_rigid_body_type* rb_type = rigid->type_ptr;

  ptrdiff_t body_type_ind = DATAVECTOR_INDEX_OF(
      struct ts_rigid_body_type, 
      &(sim->rigid_body_types), 
      rb_type);

  return mrb_fixnum_value( (mrb_int)(body_type_ind+1) );
}

static mrb_value cRigidBody_is_frozen(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  mrb_sym which = 0;
  int nargs = mrb_get_args(mrb, "|n", &which);

  int tf = 0;
  if (nargs == 0)
  {
    tf = ts_rigid_body_is_frozen(sim, rb);
  }
  else // argc == 1
  {
    if (which == TSMRB_SYM(translation)) {
      tf = ts_rigid_body_is_frozen_translation(sim, rb);
    } else if (which == TSMRB_SYM(rotation)) {
      tf = ts_rigid_body_is_frozen_rotation(sim, rb);
    } else if (which == TSMRB_SYM(all)) {
      tf = ts_rigid_body_is_frozen(sim, rb);
    } else {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown argument value");
    }
  }

  return mrb_bool_value(tf);
}

#define FREEZE(trans,rot) do { \
    ts_rigid_body_set_frozen(sim,rb,trans,rot); \
  } while(0)
static mrb_value cRigidBody_freezeB(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  mrb_sym which = 0;
  int nargs = mrb_get_args(mrb, "|n", &which);

  if (nargs == 0)
  {
    FREEZE(1,1);
  }
  else // nargs == 1
  {
    if (which == TSMRB_SYM(translation)) {
      int fr_rot = ts_rigid_body_is_frozen_rotation(sim, rb);
      FREEZE(1,fr_rot);
    } else if (which == TSMRB_SYM(all)) {
      FREEZE(1,1);
    } else {
      mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown argument value");
    }
  }

  return self;
}
#undef FREEZE

static mrb_value cRigidBody_unfreezeB(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body_set_frozen(sim, rb, 0, 0);
  return self;
}

static mrb_value cRigidBody_get_body_type(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  check_body_type(mrb,body);

  enum TS_RIGID_BODY_TYPE type = body->type_ptr->type;

  if (type == TS_RIGID_SITE_INTERACTION)
    return TSMRB_SYM_OBJ( site_interaction );
  else if (type == TS_RIGID_NOT_DEFINED)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "rigid body objects should not have type :not_defined");
  else
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "unknown rigid body type %d", (int)(type));

  return mrb_nil_value();
}

static mrb_value cRigidBody_get_rotfric(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  check_body_type(mrb,body);

  return mrb_float_value(mrb, body->type_ptr->rotfric);
}

static mrb_value cRigidBody_get_inv_rotfric(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  check_body_type(mrb,body);

  return mrb_float_value(mrb, body->type_ptr->inv_rotfric);
}

static mrb_value cRigidBody_get_I0(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  check_body_type(mrb,body);

  return mrb_float_value(mrb, body->type_ptr->I0);
}

static mrb_value cRigidBody_get_inv_I0(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb, self);
  check_body_type(mrb,body);

  return mrb_float_value(mrb, body->type_ptr->inv_I0 );
}

static mrb_value cRigidBody_get_bd_rand_coeff(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb, self);
  check_body_type(mrb,body);

  return mrb_float_value(mrb, body->type_ptr->bd_rand_coeff);
}

static mrb_value cRigidBody_get_sim(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  tsmrb_to_rigid_body_index(mrb, self, &sim);

  return tsmrb_from_sim(mrb, sim);
}

static mrb_value cRigidBody_get_index(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  return mrb_fixnum_value( (mrb_int)rb );
}

static mrb_value cRigidBody_get_site_indices(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  return tsmrb_ary_from_sizes(mrb, body->site_indices, body->type_ptr->num_sites);
}

static mrb_value cRigidBody_get_com_particle_index(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb, self);
  return mrb_fixnum_value(body->com_particle);
}

static mrb_value cRigidBody_get_com_particle(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  return tsmrb_from_particle(mrb, sim, body->com_particle);
}

static mrb_value cRigidBody_get_site_particles(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);
  check_body_type(mrb,body);

  mrb_value ary = mrb_ary_new_capa(mrb, body->type_ptr->num_sites);
  for(size_t i=0; i < body->type_ptr->num_sites; i++)
    mrb_ary_push(mrb, ary, tsmrb_from_particle(mrb, sim, body->site_indices[i]));

  return ary;
}

static mrb_value cRigidBody_get_site_particle(mrb_state* mrb, mrb_value self)
{
  mrb_int ind = 0;
  mrb_get_args(mrb, "i", &ind);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim,rb);
  check_body_type(mrb,body);

  if ((ind < 0) && (ind >= body->type_ptr->num_sites))
    tsmrb_raisef(mrb, E_INDEX_ERROR,
        "site particle index %d is invalid", (int)ind);

  return tsmrb_from_particle(mrb, sim, body->site_indices[ind]);
}

static mrb_value cRigidBody_get_reference_vectors(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb, self);
  check_body_type(mrb,body);

  mrb_value ary = mrb_ary_new_capa(mrb, body->type_ptr->num_sites);
  for(size_t i=0; i < body->type_ptr->num_sites; i++)
    mrb_ary_push(mrb, ary, tsmrb_from_vec3(mrb, body->type_ptr->ref_vectors + i));

  return ary;
}

static inline void
check_same_sim(mrb_state* mrb, ts_sim* sim1, ts_sim* sim2, const char* desc)
{
  if (sim1 != sim2)
    mrb_raise(mrb, E_ARGUMENT_ERROR, desc);
}

static mrb_value rb_bonded_to_rb(mrb_state* mrb, mrb_value body1, mrb_value body2)
{
  ts_sim* sim1 = NULL;
  ts_sim* sim2 = NULL;
  size_t rb1 = tsmrb_to_rigid_body_index(mrb, body1, &sim1);
  size_t rb2 = tsmrb_to_rigid_body_index(mrb, body2, &sim2);

  check_same_sim(mrb, sim1, sim2,
    "rigid bodies must be from the same simulation");

  return mrb_bool_value( ts_rigid_body_is_bonded_to(sim1,
        ts_get_rigid_body(sim1, rb1),
        ts_get_rigid_body(sim1, rb2)) );
}

static mrb_value rb_bonded_to_pt(mrb_state* mrb, mrb_value body, mrb_value particle)
{
  ts_sim* sim1 = NULL;
  ts_sim* sim2 = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, body, &sim1);
  size_t pt = tsmrb_to_particle_index(mrb, particle, &sim2);

  check_same_sim(mrb, sim1, sim2,
      "rigid body and particle must be from the same simulation");

  return mrb_bool_value( 
      ts_rigid_body_is_bonded_to_particle(sim1,
        ts_get_rigid_body(sim1, rb),
        ts_get_particle(sim1, pt)));
}

static mrb_value cRigidBody_bonded_to_p(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  if (mrb_obj_is_instance_of(mrb, other, tsmrb_rigid_body_class(mrb)))
    return rb_bonded_to_rb(mrb, self, other);
  else if (mrb_obj_is_instance_of(mrb, other, tsmrb_particle_class(mrb)))
    return rb_bonded_to_pt(mrb, self, other);
  else
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unsupported argument type (%S)", other);

  return mrb_nil_value();
}

static mrb_value convert_rigid_bondinfo(mrb_state* mrb, struct ts_rigid_body_bond_info* info)
{
  mrb_value elts[3] = {
    mrb_fixnum_value( info->pi ),
    mrb_fixnum_value( info->pj ),
    tsmrb_full_bondtype_value(mrb, info->btype)
  };

  return mrb_ary_new_from_values(mrb, 3, elts);
}

#define BOND_INFO_DEFAULT_SIZE 16
static mrb_value cRigidBody_get_bond_info(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_sim* sim1 = NULL;
  ts_sim* sim2 = NULL;
  size_t rb1 = tsmrb_to_rigid_body_index(mrb, self, &sim1);
  size_t rb2 = tsmrb_to_rigid_body_index(mrb, other, &sim2);

  check_same_sim(mrb, sim1, sim2, "rigid bodies must be from the same simulation");

  DATAVECTOR(struct ts_rigid_body_bond_info) info;
  datavector_init( &info, sizeof(struct ts_rigid_body_bond_info), BOND_INFO_DEFAULT_SIZE );

  TSResult err = ts_rigid_body_find_bonds(sim1,
      ts_get_rigid_body(sim1, rb1),
      ts_get_rigid_body(sim1, rb2),
      &info);

  mrb_value binfo = mrb_nil_value();
  if (IS_OKAY(err))
  {
    binfo = mrb_ary_new_capa(mrb,info.size);
    struct ts_rigid_body_bond_info* bond_info = 
      DATAVECTOR_BEGIN(struct ts_rigid_body_bond_info, &info);

    for (size_t i=0; i < info.size; i++)
      mrb_ary_push(mrb,binfo, convert_rigid_bondinfo(mrb,bond_info+i));
  }

  datavector_free(&info);
  tsmrb_handle_error(mrb, err);

  return binfo;
}
#undef BOND_INFO_DEFAULT_SIZE

static mrb_value cRigidBody_get_first_bond(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  ts_sim* sim1 = NULL;
  ts_sim* sim2 = NULL;
  size_t rb1 = tsmrb_to_rigid_body_index(mrb, self, &sim1);
  size_t rb2 = tsmrb_to_rigid_body_index(mrb, other, &sim2);

  check_same_sim(mrb, sim1, sim2, "rigid bodies must be from the same simulation");

  struct ts_rigid_body_bond_info binfo = { 0 };
  int has_bond = ts_rigid_body_find_first_bond(sim1,
      ts_get_rigid_body(sim1, rb1),
      ts_get_rigid_body(sim1, rb2),
      &binfo);

  if (has_bond)
    return convert_rigid_bondinfo(mrb,&binfo);
  else
    return mrb_nil_value();
}

static mrb_value cRigidBody_get_orientation(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  return tsmrb_from_quat(mrb, RIGID_ORI(sim, rb));
}

static mrb_value cRigidBody_get_euler_orientation(mrb_state *mrb, mrb_value self)
{
  vector3 euler;
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  quat4_to_euler_zxz(RIGID_ORI(sim, rb), &euler);
  return tsmrb_from_vec3(mrb, &euler);
}

static mrb_value cRigidBody_set_euler_orientation(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  size_t rb;
  vector3 euler;
  mrb_value reuler;

  sim = NULL;
  rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  reuler = mrb_nil_value();
  mrb_get_args(mrb, "o", &reuler);

  vector3_zero(&euler);
  obj_to_vec(mrb, reuler, &euler);
  quat4_from_euler_zxz(RIGID_ORI(sim, rb), &euler);

  return self;
}

static mrb_value cRigidBody_rotate_byB(mrb_state* mrb, mrb_value self)
{
  mrb_value rq = mrb_nil_value();
  mrb_get_args(mrb, "o", &rq);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  quat4* qrot = tsmrb_to_quat(mrb, rq);

  if (fabs(quat4_norm(qrot)-1.0) > VEC_TOL)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "invalid rotation (quaternion is not a unit quaternion)");

  quat4_lmul(RIGID_ORI(sim,rb), qrot);
  return self;
}

static mrb_value cRigidBody_move_byB(mrb_state* mrb, mrb_value self)
{
  mrb_float dx = 0.0;
  mrb_float dy = 0.0;
  mrb_float dz = 0.0;
  mrb_get_args(mrb, "fff", &dx, &dy, &dz);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  vector3_displace( PARTICLE_POS(sim, body->com_particle), 
      (VecReal)dx, (VecReal)dy, (VecReal)dz );

  return self;
}

static mrb_value cRigidBody_set_orientation(mrb_state* mrb, mrb_value self)
{
  mrb_value ori = mrb_nil_value();
  mrb_get_args(mrb, "o", &ori);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  quat4* qori = tsmrb_to_quat(mrb,ori);
  quat4_set(RIGID_ORI(sim,rb), qori);

  return self;
}

static mrb_value cRigidBody_get_torque(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  return tsmrb_from_vec3(mrb, &body->torque);
}

static mrb_value cRigidBody_get_torquev(mrb_state* mrb, mrb_value self)
{
  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);
  return tsmrb_from_vec3(mrb, &body->torque);
}

static mrb_value cRigidBody_set_torque(mrb_state* mrb, mrb_value self)
{
  mrb_value torque = mrb_nil_value();
  mrb_get_args(mrb, "o", &torque);

  ts_rigid_body* body = tsmrb_to_rigid_body(mrb,self);

  obj_to_vec(mrb, torque, &body->torque);
  return self;
}

static mrb_value cRigidBody_get_forcev(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  return tsmrb_from_vec3(mrb, PARTICLE_FORCE(sim, body->com_particle));
}

static mrb_value cRigidBody_get_force(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  return tsmrb_from_vec3(mrb, PARTICLE_FORCE(sim, body->com_particle));
}

static mrb_value cRigidBody_get_pos(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  return tsmrb_from_vec3(mrb, PARTICLE_POS(sim, body->com_particle) );
}

static mrb_value cRigidBody_get_posv(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  return tsmrb_from_vec3(mrb, PARTICLE_POS(sim, body->com_particle) );
}

static mrb_value cRigidBody_set_pos(mrb_state* mrb, mrb_value self)
{
  mrb_value v = mrb_nil_value();
  mrb_get_args(mrb, "o", &v);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  obj_to_vec(mrb, v, PARTICLE_POS(sim, body->com_particle) );
  ts_update_rigid_body(sim, rb);

  return self;
}

static mrb_value cRigidBody_set_force(mrb_state* mrb, mrb_value self)
{
  mrb_value rforce = mrb_nil_value();
  mrb_get_args(mrb, "o", &rforce);

  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim, rb);

  obj_to_vec(mrb, rforce, PARTICLE_FORCE(sim, body->com_particle) );
  return self;
}

static mrb_value cRigidBody_update(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  ts_update_rigid_body(sim, rb);
  return self;
}

static mrb_value cRigidBody_calculate_force_and_torque(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = NULL;
  size_t rb = tsmrb_to_rigid_body_index(mrb, self, &sim);
  ts_rigid_body* body = ts_get_rigid_body(sim,rb);

  ts_rigid_body_calculate_force_and_torque(sim, body);

  return self;
}

enum { ORI_GRAD_SIZE = 6 };
static mrb_value ori_grad_to_ary(mrb_state *mrb, lbfgsfloatval_t g[ORI_GRAD_SIZE])
{
  size_t i;
  mrb_value ret;

  ret = mrb_ary_new_capa(mrb, ORI_GRAD_SIZE);
  for (i=0; i < ORI_GRAD_SIZE; i++)
  {
    mrb_ary_push(mrb, ret, mrb_float_value(mrb, g[i]));
  }
  return ret;
}

static mrb_value cRigidBody_numerical_orientation_gradient(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  size_t rb;
  lbfgsfloatval_t g[ORI_GRAD_SIZE];
  double en = 0.0;

  sim = NULL;
  rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  tsmrb_handle_error(mrb, ts_calculate_forces(sim,&en, NO_TERM_GROUPS, TERMWISE_FALSE) );

  memset(g, 0, sizeof(g));
  tsmrb_handle_error(mrb, ts_calc_gradient_from_rigid_body_numerical(sim, rb, g) );

  return ori_grad_to_ary(mrb, g);
}

static mrb_value cRigidBody_analytic_orientation_gradient(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  size_t rb;
  lbfgsfloatval_t g[ORI_GRAD_SIZE];
  double en = 0.0;

  sim = NULL;
  rb = tsmrb_to_rigid_body_index(mrb, self, &sim);

  tsmrb_handle_error(mrb, ts_calculate_forces(sim,&en, NO_TERM_GROUPS, TERMWISE_FALSE) );

  memset(g, 0, sizeof(g));
  ts_calc_gradient_from_rigid_body_analytic(sim, rb, g);

  return ori_grad_to_ary(mrb, g);

}

static mrb_value cSim_define_rigid_body(mrb_state* mrb, mrb_value self)
{
  mrb_value bodyDef;
  mrb_get_args(mrb, "H", &bodyDef);
  ts_sim* sim = tsmrb_to_sim(mrb, self);

  mrb_int bd_type = hash_get_int(mrb, bodyDef, TSMRB_SYM(type)        , "rigid body type");
  struct ts_rigid_body_type* type_ptr = ts_get_rigid_body_type(sim, bd_type);
  if (type_ptr == NULL)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid rigid body type %d", (int)bd_type);

  mrb_int bd_com  = hash_get_int(mrb, bodyDef, TSMRB_SYM(com_particle), "rigid body CoM particle");

  mrb_value bd_sites = mrb_hash_get(mrb, bodyDef, TSMRB_SYM_OBJ(sites));
  if (!mrb_array_p(bd_sites))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected :site keyword to be an array of site particles");

  if (RARRAY_LEN(bd_sites) != type_ptr->num_sites)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "length of sites list is not the same as the rigid body type num_sites");

  if (sim->num_rigid_bodies >= sim->max_rigid_bodies)
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "maximum number of rigid bodies (%lu) defined",
        (unsigned long)sim->max_rigid_bodies);

  size_t num_sites = 0;
  size_t* site_inds = tsmrb_ary_to_sizes(mrb, bd_sites, &num_sites);

  ts_rigid_body* body = ts_define_rigid_body(sim, type_ptr, bd_com, site_inds);
  mrb_free(mrb, site_inds);

  if (body == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "error defining rigid body");

  ptrdiff_t body_idx = body - sim->rigid_bodies;
  if ((body_idx < 0) || (body_idx >= sim->num_rigid_bodies))
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "rigid body index (%d) is invalid", (int)body_idx);

  return tsmrb_from_rigid_body(mrb, sim, body_idx);
}

static mrb_value cSim_get_rigid_body(mrb_state* mrb, mrb_value self)
{
  mrb_int ind = 0;
  mrb_get_args(mrb, "i", &ind);
  ts_sim* sim = tsmrb_to_sim(mrb, self);

  if ((ind < 0) || (ind >= sim->num_rigid_bodies))
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid rigid body index %d", (int)ind);

  return tsmrb_from_rigid_body(mrb, sim, ind);
}

static mrb_value cSim_each_rigid_body(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim;
  mrb_value block, ret;

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  sim = tsmrb_to_sim(mrb, self);

  if (!mrb_nil_p(block))
  {
    for(size_t i=0; i < sim->num_rigid_bodies; i++)
    {
      int ai = mrb_gc_arena_save(mrb);
      mrb_value rigid = tsmrb_from_rigid_body(mrb, sim, i);
      mrb_yield(mrb, block, rigid);
      mrb_gc_arena_restore(mrb,ai);
    }

    return self;
  }

  ret = mrb_ary_new_capa(mrb, sim->num_rigid_bodies);
  for(size_t i=0; i < sim->num_rigid_bodies; i++)
  {
    int ai = mrb_gc_arena_save(mrb);
    mrb_value rigid = tsmrb_from_rigid_body(mrb, sim, i);
    mrb_ary_push(mrb, ret, rigid);
    mrb_gc_arena_restore(mrb,ai);
  }

  return ret;
}

static mrb_value cSim_all_rigid_bodies(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb, self);

  mrb_value rlist = mrb_ary_new_capa(mrb, sim->num_rigid_bodies);
  for(size_t i=0; i < sim->num_rigid_bodies; i++)
    mrb_ary_push(mrb, rlist, tsmrb_from_rigid_body(mrb, sim, i) );

  return rlist;
}

static mrb_value cSim_each_rigid_body_with_index(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim;
  mrb_value block, ret;
  size_t i;

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  sim = tsmrb_to_sim(mrb, self);

  if (!mrb_nil_p(block))
  {
    for(i=0; i < sim->num_rigid_bodies; i++)
    {
      int ai;
      mrb_value pair[2];

      ai = mrb_gc_arena_save(mrb);
      pair[0] = tsmrb_from_rigid_body(mrb, sim, i);
      pair[1] = mrb_fixnum_value( (mrb_int)i );
      mrb_yield_argv(mrb, block, 2, pair);
      mrb_gc_arena_restore(mrb,ai);
    }
    return self;
  }

  ret = mrb_ary_new_capa(mrb, sim->num_rigid_bodies);
  for(i=0; i < sim->num_rigid_bodies; i++)
  {
    int ai;
    mrb_value pair[2], elt;

    ai = mrb_gc_arena_save(mrb);

    pair[0] = tsmrb_from_rigid_body(mrb, sim, i);
    pair[1] = mrb_fixnum_value( (mrb_int)i );
    elt = mrb_ary_new_from_values(mrb, 2, pair);
    mrb_ary_push(mrb, ret, elt);

    mrb_gc_arena_restore(mrb,ai);
  }

  return ret;
}

static void add_methods_to_cSim(mrb_state* mrb, struct RClass* cSim)
{
  mrb_define_method(mrb, cSim, "define_rigid_body"  , cSim_define_rigid_body, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "get_rigid_body"     , cSim_get_rigid_body   , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "each_rigid_body"    , cSim_each_rigid_body  ,
      MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_method(mrb, cSim, "all_rigid_bodies"   , cSim_all_rigid_bodies , MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "each_rigid_body_with_index", cSim_each_rigid_body_with_index,
      MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
}

struct RClass* tsmrb_define_rigid_body(mrb_state* mrb, struct RClass* mod, struct RClass* cSim)
{
  struct RClass* cRigidBody = mrb_define_class_under(mrb,
      mod, "RigidBody", mrb->object_class);
  tsmrb_classes(mrb)->cRigidBody = cRigidBody;
  MRB_SET_INSTANCE_TT(cRigidBody, MRB_TT_DATA);

  mrb_define_method(mrb, cRigidBody, "initialize", cRigidBody_initialize    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "=="         , cRigidBody_equal        , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "inspect"    , cRigidBody_inspect      , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "sim"        , cRigidBody_get_sim      , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "index"      , cRigidBody_get_index    , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "body_type"  , cRigidBody_get_body_type, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "type"       , cRigidBody_get_type     , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "frozen?"  , cRigidBody_is_frozen, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cRigidBody, "freeze!"  , cRigidBody_freezeB  , MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cRigidBody, "unfreeze!", cRigidBody_unfreezeB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "rotfric"    , cRigidBody_get_rotfric  , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "inv_rotfric", cRigidBody_get_inv_rotfric, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "I0"         ,  cRigidBody_get_I0      , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "inv_I0"     ,  cRigidBody_get_inv_I0  , MRB_ARGS_NONE());
  
  mrb_define_method(mrb, cRigidBody, "bd_rand_coeff",  cRigidBody_get_bd_rand_coeff, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "site_indices"      , cRigidBody_get_site_indices  , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "com_particle_index", cRigidBody_get_com_particle_index, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "get_site_particle" , cRigidBody_get_site_particle , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRigidBody, "site_particles"    , cRigidBody_get_site_particles, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "com_particle"      , cRigidBody_get_com_particle  , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "bonded_to?", cRigidBody_bonded_to_p, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRigidBody, "get_bond_info"     , cRigidBody_get_bond_info , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRigidBody, "get_first_bond"    , cRigidBody_get_first_bond, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "reference_vectors" , cRigidBody_get_reference_vectors, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "orientation"       , cRigidBody_get_orientation, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "orientation="      , cRigidBody_set_orientation, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "ori_euler"         , cRigidBody_get_euler_orientation    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "ori_euler="        , cRigidBody_set_euler_orientation    , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "rotate_by!", cRigidBody_rotate_byB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRigidBody, "move_by!", cRigidBody_move_byB, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cRigidBody, "torque",  cRigidBody_get_torque, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "torquev",  cRigidBody_get_torquev, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "torque=", cRigidBody_set_torque, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "force",  cRigidBody_get_force, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "forcev",  cRigidBody_get_forcev, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "force=", cRigidBody_set_force, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "pos" ,  cRigidBody_get_pos , MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "posv",  cRigidBody_get_posv, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "pos=",  cRigidBody_set_pos , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRigidBody, "update", cRigidBody_update, MRB_ARGS_NONE());

  mrb_define_method(mrb, cRigidBody, "calculate_force_and_torque", cRigidBody_calculate_force_and_torque, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "numerical_orientation_gradient", cRigidBody_numerical_orientation_gradient, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRigidBody, "analytic_orientation_gradient" , cRigidBody_analytic_orientation_gradient, MRB_ARGS_NONE());

  add_methods_to_cSim(mrb,cSim);

  return cRigidBody;
}

