#include "tsmrb_rng.h"

#include "ts_rng.h"

#include "mruby/data.h"
#include "mruby/array.h"
#include "tsmrb_helpers.h"

static void rng_free(mrb_state *mrb, void *p);

static const mrb_data_type rng_type = {
  .struct_name = "RNG",
  .dfree = rng_free,
};

struct ts_rng* tsmrb_to_rng(mrb_state* mrb, mrb_value obj);

void tsmrb_check_rng(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_rng_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a RNG object");
}

static struct ts_rng** get_rng_handle(mrb_state* mrb, mrb_value obj)
{
  tsmrb_check_rng(mrb, obj);
  return DATA_PTR(obj);
}

struct ts_rng* tsmrb_to_rng(mrb_state* mrb, mrb_value obj)
{
  struct ts_rng** rngHandle = get_rng_handle(mrb,obj);
  return (rngHandle != NULL) ? *rngHandle : NULL;
}

mrb_value tsmrb_from_rng(mrb_state* mrb, struct ts_rng* rng)
{
  struct ts_rng** rngp = mrb_malloc(mrb, sizeof(*rngp));
  *rngp = rng;
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_rng_class(mrb), rngp, &rng_type);
  return mrb_obj_value(data);
}

static mrb_value cRNG_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs;
  mrb_get_args(mrb, "H", &kwArgs);

  struct ts_rng** rngp = mrb_malloc(mrb, sizeof(*rngp));
  *rngp = NULL;

  DATA_TYPE(self) = &rng_type;
  DATA_PTR(self) = rngp;

  mrb_value r_generator = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(generator));
  if (mrb_nil_p(r_generator))
  {
    r_generator = hash_get_value_opt(mrb, kwArgs, TSMRB_SYM(type));
    if (mrb_nil_p(r_generator))
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "RNG constructor require either :generator or :type key");
    }
  }

  mrb_check_type(mrb, r_generator, MRB_TT_SYMBOL);
  mrb_sym generator_sym = mrb_symbol(r_generator);

  if (generator_sym == TSMRB_SYM(mt19937))
  {
    uint32_t seed = (uint32_t) hash_get_int(mrb, kwArgs, TSMRB_SYM(seed), "rng parameters");
    tsmrb_handle_error(mrb, RNG_mt19937_new(seed, rngp) );
  }
  else if (generator_sym == TSMRB_SYM(testrng))
  {
    mrb_value rseq = hash_get_value(mrb, kwArgs, TSMRB_SYM(seq), "rng parameters");

    size_t n = 0;
    double* vals = tsmrb_ary_to_doubles(mrb, rseq, &n);
    TSResult err = RNG_test_new(n, vals, rngp);
    if (IS_ERROR(err))
    {
      mrb_free(mrb,vals);
      tsmrb_handle_error(mrb, err);
    }
  }
  else
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "unknown RNG type %s", mrb_sym2name(mrb, generator_sym));
  }

  return self;
}

static void rng_free(mrb_state *mrb, void *p)
{
  struct ts_rng **rngp = p;
  if (*rngp)
    ts_rng_free(*rngp);

  mrb_free(mrb, rngp);
}

static mrb_value cRNG_equal(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);
  struct ts_rng* a = tsmrb_to_rng(mrb, self);
  struct ts_rng* b = tsmrb_to_rng(mrb, other);

  return mrb_bool_value( a == b );
}

static mrb_value cRNG_get_generator(mrb_state* mrb, mrb_value self)
{
  struct ts_rng* rng = tsmrb_to_rng(mrb,self);

  mrb_sym rtype = 0;

  if (rng == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "RNG state is NULL");
  else if (rng->type == TS_RNG_MT19937)
    rtype = TSMRB_SYM(mt19937);
  else if (rng->type == TS_RNG_TEST)
    rtype = TSMRB_SYM(testrng);
  else
    rtype = TSMRB_SYM(unknown);

  return mrb_symbol_value(rtype);
}

static mrb_value cRNG_get_info(mrb_state* mrb,mrb_value self)
{
  struct ts_rng* rng = tsmrb_to_rng(mrb,self);
  if (rng == NULL) { return mrb_nil_value(); }

  mrb_value params = mrb_hash_new(mrb);

  if (rng->type == TS_RNG_MT19937)
  {
    uint32_t seed = 0;
    tsmrb_handle_error(mrb, RNG_mt19937_get_info(rng, &seed) );

    mrb_hash_set(mrb, params, TSMRB_SYM_OBJ(generator), TSMRB_SYM_OBJ(mt19937));
    mrb_hash_set(mrb, params, TSMRB_SYM_OBJ(seed), mrb_fixnum_value(seed));
  }
  else if (rng->type == TS_RNG_TEST)
  {
    size_t seq_len = 0;
    size_t seq_idx = 0;
    double* seq_vals = NULL;

    tsmrb_handle_error(mrb, 
        RNG_test_get_info(rng, &seq_len, &seq_idx, &seq_vals) );

    mrb_hash_set(mrb, params, TSMRB_SYM_OBJ(generator), TSMRB_SYM_OBJ(testrng));

    mrb_value seq = mrb_nil_value();
    if (seq_vals != NULL) {
      seq = tsmrb_ary_from_doubles(mrb, seq_len, seq_vals);
    } else {
      seq = mrb_ary_new(mrb);
    }
    mrb_hash_set(mrb, params, TSMRB_SYM_OBJ(seq), seq);
  }
  else
  {
    mrb_hash_set(mrb, params, TSMRB_SYM_OBJ(type), TSMRB_SYM_OBJ(unknown));
  }

  return params;
}

static mrb_value get_random_values(mrb_state* mrb, mrb_value self, TSResult (*rng_func)(struct ts_rng*,size_t,double*))
{
  mrb_int n;
  mrb_get_args(mrb, "i", &n);

  struct ts_rng* rng = tsmrb_to_rng(mrb,self);
  if (rng == NULL) { mrb_raise(mrb, E_RUNTIME_ERROR, "RNG state is NULL"); }

  double* vals = mrb_calloc(mrb, n, sizeof(*vals));
  TSResult result = rng_func(rng, n, vals);
  if (IS_ERROR(result))
  {
    mrb_free(mrb, vals);
    tsmrb_handle_error(mrb, result);
  }

  mrb_value ary = tsmrb_ary_from_doubles(mrb, n, vals);
  mrb_free(mrb, vals);

  return ary;
}

static mrb_value cRNG_random(mrb_state* mrb, mrb_value self)
{
  return get_random_values(mrb,self,ts_rng_gen_uniforms);
}

static mrb_value cRNG_randn(mrb_state* mrb, mrb_value self)
{
  return get_random_values(mrb,self,ts_rng_gen_normals);
}

static mrb_value cRNG_randn_seq(mrb_state* mrb, mrb_value self)
{
  return get_random_values(mrb,self,ts_rng_gen_normals_sequential);
}

static mrb_value cRNG_randn_block(mrb_state* mrb, mrb_value self)
{
  return get_random_values(mrb,self,ts_rng_gen_normals_block);
}

struct RClass* tsmrb_define_rng(mrb_state* mrb, struct RClass* mod)
{
  struct RClass* cRNG = mrb_define_class_under(mrb, mod, "RNG", mrb->object_class);
  tsmrb_classes(mrb)->cRNG = cRNG;
  MRB_SET_INSTANCE_TT(cRNG, MRB_TT_DATA);

  mrb_define_method(mrb, cRNG, "initialize", cRNG_initialize   , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRNG, "generator" , cRNG_get_generator, MRB_ARGS_NONE());
  mrb_define_method(mrb, cRNG, "info"      , cRNG_get_info     , MRB_ARGS_NONE());

  mrb_define_method(mrb, cRNG, "=="        , cRNG_equal        , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cRNG, "random"     , cRNG_random     , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRNG, "randn"      , cRNG_randn      , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRNG, "randn_seq"  , cRNG_randn_seq  , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cRNG, "randn_block", cRNG_randn_block, MRB_ARGS_REQ(1));

  return cRNG;
}
