#include "tsmrb_sim.h"

#include <math.h>

#include "ts_common.h"
#include "ts_common_algo.h"

#include "tsmrb_helpers.h"

#include "mruby/data.h"
#include "mruby/hash.h"
#include "mruby/string.h"
#include "mruby/value.h"
#include "mruby/numeric.h"
#include "mruby/proc.h"
#include "mruby/array.h"
#include "mruby/variable.h"

#include "numarray_mruby.h"

#include "tubesheet.h"
#include "ts_bonds.h"
#include "ts_io.h"
#include "ts_io.h"
#include "ts_integrators.h"
#include "ts_dof_map.h"
#include "ts_normalmode.h"
#include "ts_minimize.h"

#include "tsmrb_veclist.h"
#include "tsmrb_vec.h"
#include "tsmrb_potentials.h"

#include "tsmrb_rng.h"

#include "tsmrb_numerical.h"

#define DEFAULT_ROT_REPR TS_ROT_AXIS_ANGLE

enum {
  NO_TERM_GROUPS = 0,
  TERMWISE_FALSE = 0,
  TERMWISE_TRUE  = 1,
};

static void tsmrb_sim_free(mrb_state* mrb, void* p);
static const mrb_data_type sim_type = {
  .struct_name = "Sim",
  .dfree = tsmrb_sim_free,
};

static void sim_weakref_table_setup(mrb_state* mrb, struct RClass* c)
{
  mrb_value backref_tbl = mrb_hash_new(mrb);
  mrb_mod_cv_set(mrb, c, mrb_intern_cstr(mrb,"@@backrefs"), backref_tbl);
}

static mrb_value sim_weakref_table_get(mrb_state* mrb, struct RClass* c)
{
  return mrb_mod_cv_get(mrb, c, mrb_intern_cstr(mrb,"@@backrefs"));
}

static void sim_weakref_add(mrb_state* mrb, ts_sim* sim, struct RObject* obj)
{
  mrb_value p_sim = mrb_cptr_value(mrb, sim);
  mrb_value p_obj = mrb_cptr_value(mrb, obj);

  mrb_value tbl = sim_weakref_table_get(mrb, tsmrb_sim_class(mrb));

  return mrb_hash_set(mrb, tbl, p_sim, p_obj);
}

static struct RObject *sim_weakref_get(mrb_state *mrb, ts_sim *sim)
{
  mrb_value p_sim = mrb_cptr_value(mrb, sim);
  mrb_value tbl = sim_weakref_table_get(mrb, tsmrb_sim_class(mrb));
  mrb_value p_obj = mrb_hash_get(mrb, tbl, p_sim);
  if (!mrb_nil_p(p_obj))
    return mrb_cptr(p_obj);
  else
    return NULL;
}

static void tsmrb_sim_free(mrb_state *mrb, void *p)
{
  ts_sim_delete((ts_sim *)p);
}

ts_sim* tsmrb_to_sim(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_sim_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Sim object");

  return DATA_PTR(obj);
}

mrb_value tsmrb_from_sim(mrb_state* mrb, ts_sim* sim)
{
  struct RObject* obj = sim_weakref_get(mrb, sim);
  if (obj)
    return mrb_obj_value(obj);
  else
    return mrb_nil_value();
}

mrb_value tsmrb_make_sim(mrb_state* mrb, ts_sim *sim)
{
  struct RData *data;

  data = mrb_data_object_alloc(mrb,
      tsmrb_sim_class(mrb), sim, &sim_type);
  sim_weakref_add(mrb, sim, (struct RObject*)data);

  return mrb_obj_value(data);
}

static mrb_value cSim_initialize(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim = ts_sim_new();

  if (sim == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not allocate a Sim object");
  }

  DATA_TYPE(self) = &sim_type;
  DATA_PTR(self) = sim;

  sim_weakref_add(mrb, sim, mrb_obj_ptr(self));

  return self;
}

static mrb_value cSim_alloc_particles(mrb_state* mrb, mrb_value self)
{
  mrb_int np;
  mrb_get_args(mrb, "i", &np);

  tsmrb_handle_error(mrb, 
      ts_alloc_particles(tsmrb_to_sim(mrb,self), np)
      );

  return self;
}

static void print_potential_terms(FILE* outf, const char* title, struct ts_potential_term_list* plist)
{
  if (plist->terms.size == 0) return;

  char banner[81];
  memset(banner, '-', sizeof(banner)-1);
  banner[sizeof(banner)-1] = '\0';

  const char tbrak[4] = { '[', ' ', ' ', ']' };

  size_t title_len = strlen(title); // 4-char buffer for '[ ... ]'
  if (title_len+sizeof(tbrak) < sizeof(banner) - 1) {
    size_t off = (sizeof(banner) - 1 - (title_len+sizeof(tbrak)))/2;

    banner[off] = tbrak[0];
    banner[off+1] = tbrak[1];
    memcpy(banner+off+2,title,title_len);
    banner[off+2+title_len] = tbrak[2];
    banner[off+2+title_len+1] = tbrak[3];
  }

  fprintf(outf, "%s\n", banner);
  size_t ind0 = 0;
  for (size_t i=0; i < plist->potential_indices.size; i++)
  {
    ts_potential_term* term = DATAVECTOR_AT(ts_potential_term, &plist->terms, ind0);
    size_t next_ind = *(DATAVECTOR_AT(size_t, &plist->potential_indices, i));

    fprintf(outf, 
        "POTE|%3ld|%-36s|%-16s|%1d|%11.6f\n",
        next_ind - ind0,
        term->name,
        term->potential->name,
        term->potential->nbodies,
        term->potential->potential_energy);

    ind0 = next_ind;
  }
  fprintf(outf,"\n");
}

static mrb_value cSim_print_potential_terms(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  print_potential_terms(sim->logging_output, "BONDED", &sim->bondlist);
  print_potential_terms(sim->logging_output, "NON-BOND", &sim->nonbond_list);

  return self;
}

static mrb_value cSim_num_particles(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->num_particles);
}

static mrb_value cSim_max_particles(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->max_particles);
}

static mrb_value cSim_num_free_particles(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(ts_sim_num_free_particles(sim));
}

static mrb_value cSim_num_dof(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(ts_sim_num_dof(sim));
}

static mrb_value cSim_num_rigid_body_types(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->rigid_body_types.size);
}

static mrb_value cSim_get_nonbonds_enabled(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_bool_value( sim->nonbonds_enabled );
}

static mrb_value cSim_set_nonbonds_enabled(mrb_state* mrb, mrb_value self)
{
  mrb_bool enabled = 0;
  mrb_get_args(mrb, "b", &enabled);

  ts_sim* sim = tsmrb_to_sim(mrb,self);
  sim->nonbonds_enabled = enabled;
  return self;
}

static mrb_value cSim_get_nonbond_cutoff(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_float_value(mrb, sim->nonbond_cutoff);
}

static mrb_value cSim_set_nonbond_cutoff(mrb_state* mrb, mrb_value self)
{
  mrb_float f = 0.0;
  mrb_get_args(mrb, "f", &f);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  sim->nonbond_cutoff = (double)f;
  return self;
}

static mrb_value cSim_get_nonbond_padding(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_float_value(mrb, sim->nonbond_padding);
}

static mrb_value cSim_set_nonbond_padding(mrb_state* mrb, mrb_value self)
{
  mrb_float f = 0.0;
  mrb_get_args(mrb, "f", &f);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  sim->nonbond_padding = (double)f;
  return self;
}

static mrb_value cSim_count_bond_patterns(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value( ts_count_bond_patterns(sim) );
}

static mrb_value cSim_count_potential_terms(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value( ts_count_potential_terms(sim) );
}

static mrb_value cSim_count_constraints(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value( ts_count_constraints(sim) );
}

static mrb_value cSim_clear_forces(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  ts_clear_forces(sim);
  return self;
}

static mrb_value cSim_build_potential(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  tsmrb_handle_error(mrb, ts_build_potential(sim));
  return self;
}

static mrb_value cSim_get_integrator_steps(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->integrator_steps);
}

static mrb_value cSim_get_time(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_float_value(mrb,sim->time);
}

static mrb_value cSim_get_dt(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_float_value(mrb,sim->dt);
}

static mrb_value cSim_set_dt(mrb_state* mrb, mrb_value self)
{
  mrb_float dt = 0.0;
  mrb_get_args(mrb, "f", &dt);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  sim->dt = (double)dt;
  return self;
}

static mrb_value cSim_get_temp(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_float_value(mrb,sim->temp);
}

static mrb_value cSim_set_temp(mrb_state* mrb, mrb_value self)
{
  mrb_float temp = 0.0;
  mrb_get_args(mrb, "f", &temp);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  sim->temp = (double)temp;
  return self;
}

static mrb_value cSim_single_step(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  tsmrb_handle_error(mrb, ts_sim_single_step(sim));
  return mrb_float_value(mrb, sim->time);
}

static mrb_value cSim_update_rigid_bodies(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  ts_update_rigid_bodies(sim);
  return self;
}

static mrb_value
cSim_calculate_forces_and_torques_on_rigid_bodies(mrb_state* mrb,
    mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  ts_rigid_bodies_calculate_all_forces_and_torques(sim);
  return self;
}

static mrb_value cSim_get_num_rigid_bodies(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->num_rigid_bodies);
}

static mrb_value cSim_get_max_rigid_bodies(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->max_rigid_bodies);
}

static mrb_value cSim_get_net_torque(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return tsmrb_from_vec3(mrb, &sim->net_torque);
}

static mrb_value cSim_get_net_force(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return tsmrb_from_vec3(mrb, &sim->net_force);
}

enum { MAX_TERM_GROUPS = 16 };
static mrb_value cSim_calculate_forces(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim;
  mrb_int max_term_groups = 0;
  int termwise_energies = 0;
  mrb_value arg1,arg2;
  double energies[MAX_TERM_GROUPS+1] = { 0.0 };

  sim = tsmrb_to_sim(mrb,self);

  arg1 = arg2 = mrb_nil_value();
  mrb_get_args(mrb, "|oo", &arg1, &arg2);

  if (mrb_nil_p(arg2) && mrb_hash_p(arg1))
  {
    arg2 = arg1;
    arg1 = mrb_nil_value();
  }

  if (!mrb_nil_p(arg1))
  {
    mrb_check_type(mrb, arg1, MRB_TT_FIXNUM);
    max_term_groups = mrb_int(mrb,arg1);
  }

  if (!mrb_nil_p(arg2))
  {
    mrb_check_type(mrb, arg2, MRB_TT_HASH);
    termwise_energies =
      mrb_bool( mrb_hash_fetch(
            mrb, arg2, TSMRB_SYM_OBJ(termwise), mrb_nil_value()) );
  }

  if (max_term_groups > MAX_TERM_GROUPS)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "only %d term groups are currently supported", MAX_TERM_GROUPS);
  }

  tsmrb_handle_error(mrb,
      ts_calculate_forces(sim, energies, max_term_groups, termwise_energies));

  if (max_term_groups == 0)
    return mrb_float_value(mrb,energies[0]);
  else
    return tsmrb_ary_from_doubles(mrb, max_term_groups+1, energies);
}

static mrb_value cSim_get_free_particle_to_particle_index_map(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  struct ts_dof_map dof_map;

  ts_dof_map_init(&dof_map, sim);

  mrb_value ind_map = mrb_ary_new_capa(mrb, dof_map.num_free_particles);
  for (size_t i=0; i < dof_map.num_free_particles; i++)
    mrb_ary_push(mrb, ind_map, mrb_fixnum_value(dof_map.fp_to_p[i]));

  ts_dof_map_finalize(&dof_map);

  return ind_map;
}

static mrb_value cSim_get_particle_to_free_particle_index_map(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  struct ts_dof_map dof_map;

  ts_dof_map_init(&dof_map, sim);

  mrb_value ind_map = mrb_ary_new_capa(mrb, sim->num_particles);
  for (size_t i=0; i < sim->num_particles; i++)
  {
    if (dof_map.p_to_fp[i] != TS_NULL_PARTICLE)
      mrb_ary_push(mrb, ind_map, mrb_fixnum_value( dof_map.p_to_fp[i] ));
    else
      mrb_ary_push(mrb, ind_map, mrb_nil_value());
  }

  ts_dof_map_finalize(&dof_map);

  return ind_map;
}

static mrb_value cSim_add_particle_type(mrb_state* mrb, mrb_value self)
{
  mrb_value typeDef;
  mrb_get_args(mrb, "H", &typeDef);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  ts_particle_type ptype;

  const char* name = hash_get_str(mrb, typeDef, TSMRB_SYM(name), "particle type description");
  ptype.name     = ts_strdup(NO_ALLOCATOR, name);

  ptype.mass     = hash_get_float(mrb, typeDef, TSMRB_SYM(mass)    , "particle type definition");
  ptype.radius   = hash_get_float(mrb, typeDef, TSMRB_SYM(radius)  , "particle type definition");
  ptype.collfreq = hash_get_float(mrb, typeDef, TSMRB_SYM(collfreq), "particle type definition");

  ptype.charge   = hash_get_float_opt(mrb, typeDef, TSMRB_SYM(charge), "particle type definition", 0.0);

  mrb_value rlongname = mrb_hash_get(mrb, typeDef, mrb_symbol_value(TSMRB_SYM(long_name)));
  if (mrb_nil_p(rlongname))
  {
    ptype.long_name = ts_strdup(NO_ALLOCATOR, name);
  }
  else
  {
    const char* longname = mrb_string_value_cstr(mrb, &rlongname);
    ptype.long_name = ts_strdup(NO_ALLOCATOR, longname);
  }

  size_t pIdx = 0;
  TSResult err = ts_add_particle_type(sim, &ptype, &pIdx);
  tsmrb_handle_error(mrb, err);

  return mrb_fixnum_value(pIdx);
}

mrb_value tsmrb_sim_ptype_to_hash(mrb_state *mrb, ts_particle_type *ptype)
{
  mrb_value rbDef = mrb_hash_new(mrb);

  hash_set_str(mrb, rbDef, TSMRB_SYM(name)     , ptype->name);
  hash_set_str(mrb, rbDef, TSMRB_SYM(long_name), ptype->long_name);

  hash_set_float(mrb, rbDef, TSMRB_SYM(mass)    , ptype->mass    );
  hash_set_float(mrb, rbDef, TSMRB_SYM(radius)  , ptype->radius  );
  hash_set_float(mrb, rbDef, TSMRB_SYM(charge)  , ptype->charge  );
  hash_set_float(mrb, rbDef, TSMRB_SYM(collfreq), ptype->collfreq);

  return rbDef;
}

static mrb_value cSim_get_particle_type(mrb_state* mrb, mrb_value self)
{
  mrb_int ptIdx = -1;
  mrb_get_args(mrb, "i", &ptIdx);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  ts_particle_type* ptype = ts_get_particle_type(sim, ptIdx);
  if (ptype == NULL)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "get_particle_type: invalid particle type %ld", 
        (long) ptIdx);

  return tsmrb_sim_ptype_to_hash(mrb, ptype);
}

static mrb_value cSim_find_particle_type(mrb_state* mrb, mrb_value self)
{
  const char* typeName = NULL;
  mrb_get_args(mrb, "z", &typeName);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  return mrb_fixnum_value(
      (mrb_int) ts_find_particle_type(sim,typeName)
      );
}

static mrb_value cSim_find_rigid_body_type(mrb_state* mrb, mrb_value self)
{
  const char* name = NULL;
  mrb_get_args(mrb, "z", &name);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  for (size_t i=1; i <= ts_count_rigid_body_types(sim); i++) {
    struct ts_rigid_body_type* rbtype = ts_get_rigid_body_type(sim, i);

    if (strcmp(rbtype->name, name) == 0)
      return mrb_fixnum_value(i);
  }

  return mrb_fixnum_value(0);
}

static mrb_value cSim_num_particle_types(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value( sim->particle_types.size );
}


static mrb_value cSim_add_particle_set(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim;
  struct ts_index_set *set;
  mrb_value r_inds;
  mrb_int i,n;
  const mrb_value *elts;
  size_t set_index;

  sim = tsmrb_to_sim(mrb,self);

  r_inds = mrb_nil_value();
  mrb_get_args(mrb, "A", &r_inds);

  n = mrb_ary_len(mrb, r_inds);
  elts = RARRAY_PTR(r_inds);
  for (i=0; i < n; i++)
  {
    mrb_int pind;

    if (!mrb_fixnum_p(elts[i]))
    {
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "element %zu is not a fixnum", (size_t)i);
    }

    pind = mrb_fixnum(elts[i]);
    if (pind < 0 || (size_t)pind >= sim->num_particles)
    {
      tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %d at element %zu", pind, (size_t)i);
    }
  }

  if (set = ts_add_particle_set(sim, (size_t)n), set == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not allocate an particle set");
  }

  if (set_index = ts_lookup_particle_set(sim, set), set_index == 0)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "invalid particle set index");
  }

  for (i=0; i < n; i++)
  {
    // we've already check the values
    mrb_assert(set->len < set->cap);
    set->indices[set->len++] = (size_t)mrb_fixnum(elts[i]);
  }

  quick_sort_sizes(set->indices, set->len);

  return mrb_fixnum_value((mrb_int) set_index);
}

static mrb_value cSim_delete_particle_set(mrb_state* mrb, mrb_value self)
{
  return mrb_nil_value();
}

static mrb_value cSim_get_particle_set(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim;
  struct ts_index_set *set;
  mrb_value r_inds;
  mrb_int set_index, i;

  sim = tsmrb_to_sim(mrb,self);

  set_index = -1;
  mrb_get_args(mrb, "i", &set_index);

  if (set_index < 0)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid particle set index %d", (int)set_index);
  }

  if (set = ts_get_particle_set(sim, (size_t)set_index), set == NULL)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid particle set index %d", (int)set_index);
  }

  // XXX - range check!
  r_inds = mrb_ary_new_capa(mrb, (mrb_int) set->len);
  for (i=0; i < set->len; i++)
  {
    // XXX - range check!
    mrb_ary_push(mrb, r_inds, mrb_fixnum_value(set->indices[i]));
  }

  return r_inds;
}

static mrb_value cSim_add_to_particle_set(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim;
  struct ts_index_set *set;
  mrb_int set_index, pind;

  sim = tsmrb_to_sim(mrb,self);

  set_index = -1;
  pind = -1;
  mrb_get_args(mrb, "ii", &set_index, &pind);

  if (set_index < 0)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid particle set index %d", (int)set_index);
  }

  if (set = ts_get_particle_set(sim, (size_t)set_index), set == NULL)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid particle set index %d", (int)set_index);
  }

  if (pind < 0 || pind >= sim->num_particles)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %d", pind);
  }

  tsmrb_handle_error(mrb, ts_index_set_add(set, (size_t)pind));

  return mrb_nil_value();
}

static mrb_value cSim_num_particle_sets(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value(sim->particle_sets.size);
}

static mrb_value cSim_get_potential(mrb_state* mrb, mrb_value self)
{
  const char* name = NULL;
  mrb_get_args(mrb, "z", &name);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  Potential* pot = ts_get_potential(sim, name);

  if (pot == NULL) {
    return mrb_nil_value();
  }

  return tsmrb_from_potential(mrb,pot);
}

static mrb_value cSim_add_potential(mrb_state* mrb, mrb_value self)
{
  const char* name = NULL;
  mrb_value r_pot = mrb_nil_value();
  mrb_get_args(mrb, "zo", &name, &r_pot);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  Potential* pot = tsmrb_to_potential(mrb,r_pot);
  tsmrb_handle_error(mrb, ts_add_potential(sim, name, pot) );

  return self;
}

static mrb_value cSim_each_potential(mrb_state* mrb, mrb_value self)
{ 
  ts_sim *sim;
  mrb_value block, ret;
  struct ts_sim_potential_iter iter = {0};

  sim = tsmrb_to_sim(mrb,self);

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  ret = mrb_nil_value();
  if (mrb_nil_p(block))
  {
    ret = mrb_ary_new(mrb);
  }

  ts_sim_potential_iter_init(sim, &iter);
  while (ts_sim_potential_iter_next(sim,&iter))
  {
    int ai;
    mrb_value pair[2];

    ai = mrb_gc_arena_save(mrb);
    pair[0] = mrb_str_new_cstr(mrb,iter.key);
    pair[1] = tsmrb_from_potential(mrb,iter.pot);

    if (mrb_nil_p(block))
    {
      mrb_value elt;
      elt = mrb_ary_new_from_values(mrb, 2, pair);
      mrb_ary_push(mrb,ret,elt);
    }
    else
    {
      mrb_yield_argv(mrb, block, 2, pair);
    }

    mrb_gc_arena_restore(mrb,ai);
  }

  return ret;
}

static mrb_value cSim_load_yaml_string(mrb_state* mrb, mrb_value self)
{
  const char* s = NULL;
  int slen = 0;
  mrb_get_args(mrb, "s", &s, &slen);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  tsmrb_handle_error(mrb, ts_yaml_load_string(sim,s,slen));
  return self;
}

static mrb_value cSim_get_particle_type_info(mrb_state* mrb,mrb_value self)
{
  mrb_int ptIdx = -1;
  mrb_get_args(mrb, "i", &ptIdx);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  ts_particle_type* ptype = ts_get_particle_type(sim, ptIdx);
  if (ptype == NULL)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "get_particle_type: invalid particle type %ld", 
        (long) ptIdx);

  mrb_value rbDef = tsmrb_sim_ptype_to_hash(mrb, ptype);

  hash_set_float(mrb, rbDef, TSMRB_SYM(inv_mass)      , ptype->inv_mass      );

  hash_set_float(mrb, rbDef, TSMRB_SYM(inv_collfreq)  , ptype->inv_collfreq  );
  hash_set_float(mrb, rbDef, TSMRB_SYM(coll_coeff1)   , ptype->coll_coeff1   );
  hash_set_float(mrb, rbDef, TSMRB_SYM(coll_coeff2)   , ptype->coll_coeff2   );
  hash_set_float(mrb, rbDef, TSMRB_SYM(rand_coeff)    , ptype->rand_coeff    );

  hash_set_float(mrb, rbDef, TSMRB_SYM(brownian_coeff), ptype->brownian_coeff);
  hash_set_float(mrb, rbDef, TSMRB_SYM(bd_rand_coeff) , ptype->bd_rand_coeff );

  return rbDef;
}

static ts_bond_type decode_bond_type(mrb_state* mrb, mrb_value bondType);
static mrb_value cSim_add_bond_type(mrb_state* mrb, mrb_value self)
{
  const char* btName = NULL;
  mrb_value r_bondType = mrb_nil_value();
  int nargs = mrb_get_args(mrb, "z|o", &btName, &r_bondType);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  mrb_value ret = mrb_nil_value();

  ts_bond_type bondtype = bondtype_none();
  if (nargs == 2) {
    bondtype = decode_bond_type(mrb, r_bondType);
    ret = r_bondType;
  } else {
    uint16_t unused = ts_find_unused_bond_type(sim);

    if (unused > 0)
      bondtype = bondtype_new(BONDTYPE_UNDIRECTED, unused);
    else
      mrb_raise(mrb, E_RANGE_ERROR, "could not find an unused bond type");

    ret = mrb_fixnum_value((int)unused);
  }

  ts_add_bond_type(sim, btName, bondtype);

  return ret;
}

struct bond_list_term {
  const char* name;
  mrb_sym sym;
  enum TS_BOND_TERM term;
};

static void initialize_bond_list_terms(mrb_state* mrb, struct bond_list_term** terms_ptr)
{
  static const struct bond_list_term template[] = {
    { "zero_body" , 0, TS_BOND_TERM_ZEROBODY   },
    { "one_body"  , 0, TS_BOND_TERM_ONEBODY    },
    { "two_body"  , 0, TS_BOND_TERM_TWOBODY    },
    { "three_body", 0, TS_BOND_TERM_THREEBODY  },
    { "four_body" , 0, TS_BOND_TERM_FOURBODY   },
    { "improper"  , 0, TS_BOND_TERM_IMPROPER   },
    { "nonbond"   , 0, TS_BOND_TERM_NONBOND    },
    { "constraint", 0, TS_BOND_TERM_CONSTRAINT },
    { 0 }
  };

  struct bond_list_term* terms = mrb_malloc(mrb, sizeof(template));
  memcpy(terms, template, sizeof(template));

  struct bond_list_term* s = terms;
  while(s->name)
  {
    s->sym = mrb_intern_cstr(mrb, s->name);
    s++;
  }

  *terms_ptr = terms;
}

static mrb_value cSim_get_bonds_enabled(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  mrb_value ret = mrb_hash_new(mrb);

  struct bond_list_term* terms = tsmrb_classes(mrb)->bond_list_terms;

  while (terms->name)
  {
    mrb_hash_set(mrb, ret,
        mrb_symbol_value(terms->sym),
        mrb_bool_value( sim->bonds_enabled[ terms->term ] ));

    terms++;
  }

  return ret;
}

static mrb_value cSim_enable_bonds(mrb_state* mrb, mrb_value self)
{
  mrb_value val;
  mrb_get_args(mrb, "H", &val);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  struct bond_list_term* terms = tsmrb_classes(mrb)->bond_list_terms;
  while (terms->name)
  {
    mrb_value lk = mrb_hash_get(mrb, val, mrb_symbol_value(terms->sym));
    if (!mrb_nil_p(lk))
    {
      sim->bonds_enabled[ terms->term ] = mrb_test(lk);
    }
    terms++;
  }

  return self;
}

static mrb_value cSim_get_bond_type(mrb_state* mrb, mrb_value self)
{
  const char* btName = NULL;
  mrb_get_args(mrb, "z", &btName);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  ts_bond_type bt = ts_get_bond_type(sim,btName);
  return tsmrb_full_bondtype_value(mrb, bt);
}

static mrb_value cSim_each_bond_type(mrb_state* mrb,mrb_value self)
{
  mrb_value block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  mrb_value ret = self;
  if (!mrb_nil_p(block))
  {
    struct ts_bond_type_iter iter = { 0 };
    ts_bond_type_iter_init(sim, &iter);
    while(ts_bond_type_iter_next(sim, &iter))
    {
      mrb_value pair[2] = {
        mrb_str_new_cstr(mrb, iter.key),
        tsmrb_full_bondtype_value(mrb, iter.bt),
      };

      mrb_yield_argv(mrb, block, 2, pair);
    }
  }
  else
  {
    ret = mrb_ary_new(mrb);

    struct ts_bond_type_iter iter = { 0 };
    ts_bond_type_iter_init(sim, &iter);
    while(ts_bond_type_iter_next(sim, &iter))
    {
      mrb_value pair[2] = {
        mrb_str_new_cstr(mrb, iter.key),
        tsmrb_full_bondtype_value(mrb, iter.bt),
      };

      mrb_ary_push(mrb, ret, 
          mrb_ary_new_from_values(mrb, 2, pair));
    }
  }

  return ret; 
}

static mrb_value cSim_num_bond_types(mrb_state* mrb,mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  return mrb_fixnum_value( ts_count_bond_types(sim) );
}

static int sym_to_direc(mrb_state* mrb, mrb_sym sym, enum BONDTYPE_DIRECTION* direcp)
{
  if      (sym == TSMRB_SYM(none))          *direcp = BONDTYPE_UNDIRECTED;
  else if (sym == TSMRB_SYM(forw))          *direcp = BONDTYPE_FORWARD;
  else if (sym == TSMRB_SYM(back))          *direcp = BONDTYPE_BACKWARD;
  else if (sym == TSMRB_SYM(bidirectional)) *direcp = BONDTYPE_BIDIRECTIONAL;
  else
    return 0;

  return 1;
}

static ts_bond_type decode_bond_type(mrb_state* mrb, mrb_value bondType)
{
  ts_bond_type bt = bondtype_none();

  if (mrb_numeric_p(bondType))
  {
    bt = bondtype_new(BONDTYPE_UNDIRECTED, mrb_to_fixnum(mrb,bondType));
  }
  else if (mrb_array_p(bondType))
  {
    if (RARRAY_LEN(bondType) == 1)
    {
      return bondtype_from_int( (uint16_t)(mrb_fixnum(RARRAY_PTR(bondType)[0])) );
    }
    else if (RARRAY_LEN(bondType) == 2)
    {
      mrb_value r_direc = mrb_ary_ref(mrb,bondType,0);
      mrb_value r_type  = mrb_ary_ref(mrb,bondType,1);

      mrb_check_type(mrb, r_direc, MRB_TT_SYMBOL);
      mrb_check_type(mrb, r_type , MRB_TT_FIXNUM);

      enum BONDTYPE_DIRECTION direc = BONDTYPE_UNDIRECTED;
      if (! sym_to_direc(mrb,mrb_symbol(r_direc),&direc))
        mrb_raise(mrb, E_ARGUMENT_ERROR,
            "unknown direction (expected :none, :forw, :back, or :bidirectional)");

      bt = bondtype_new(direc,mrb_fixnum(r_type));
    }
    else
    {
      mrb_raise(mrb, E_RANGE_ERROR, 
          "bondtype entry should be a one- or two-element array: [integer_bond_type] or [direction,type]");
    }
  }
  else
  {
    mrb_raise(mrb, E_TYPE_ERROR, 
        "bondtype entry should be either a number or an array");
  }

  return bt;
}

static void decode_bond_type_list(mrb_state* mrb, mrb_value r_btypes, ts_bond_type btypes[NBOND_LIMIT], int nbodies)
{
  for (int i=0; i < nbodies-1; i++)
  {
    mrb_value entry = mrb_ary_ref(mrb, r_btypes, i);
    btypes[i] = decode_bond_type(mrb, entry);
  }
}

static void decode_ptype_list(mrb_state* mrb, mrb_value r_ptypes, int ptypes[NBODY_LIMIT], int nbodies)
{
  for (int i = 0 ; i < nbodies; i++)
  {
    mrb_value pt = mrb_ary_ref(mrb, r_ptypes, i);
    mrb_check_type(mrb, pt, MRB_TT_FIXNUM);
    ptypes[i] = mrb_fixnum( pt );
  }
}

static mrb_value cSim_add_nonbond_pattern(mrb_state*mrb, mrb_value self)
{
  const char* btName = NULL;
  mrb_value r_potential;
  mrb_value r_ptypes;
  mrb_get_args(mrb, "zoA", &btName, &r_potential, &r_ptypes);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if (RARRAY_LEN(r_ptypes) != 2)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "add_nonbond_pattern requires two particle types, %d given",
        (int)RARRAY_LEN(r_ptypes));

  int ptypes[NBODY_LIMIT];
  decode_ptype_list(mrb, r_ptypes, ptypes, 2);

  struct RClass* cWCA = tsmrb_wca_class(mrb);
  if (cWCA == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR, "WCA non-bond potential not initialized");

  if (!mrb_obj_is_instance_of(mrb, r_potential, cWCA))
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "nonbond potential must be of class %s (arg is %s)",
        mrb_class_name(mrb,cWCA), 
        mrb_obj_classname(mrb,r_potential));

  Potential* potential = tsmrb_to_potential(mrb,r_potential);
  if (potential == NULL)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "internal potential is NULL");

  tsmrb_handle_error(mrb, 
      ts_add_nonbond_pattern(sim, ptypes[0],ptypes[1], btName, potential, NULL));

  return self;
}

static mrb_value cSim_add_bonded_pattern(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs;
  mrb_get_args(mrb, "H", &kwArgs);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  const char* btName    = hash_get_str  (mrb, kwArgs, TSMRB_SYM(name)          , "bonded pattern name");
  mrb_value r_btypes    = hash_get_value(mrb, kwArgs, TSMRB_SYM(bondtypes)     , "bond types"         );
  mrb_value r_potential = hash_get_value(mrb, kwArgs, TSMRB_SYM(potential)     , "bond potential"     );
  mrb_value r_ptypes    = hash_get_value(mrb, kwArgs, TSMRB_SYM(particle_types), "particle types"     );

  mrb_value r_termgroup = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(term_group));

  Potential* potential = tsmrb_to_potential(mrb, r_potential);

  int nbodies = RARRAY_LEN(r_ptypes);
  int ptypes[NBODY_LIMIT] = PTYPES_INIT;
  ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;

  if ((nbodies < 1) || (nbodies > NBODY_LIMIT))
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unsupported number of particles (%lu given)", 
        (unsigned long) RARRAY_LEN(r_ptypes));
  }

  if (potential->nbodies != nbodies)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "potential requires %lu bodies, but %lu were given",
        (unsigned long)potential->nbodies, 
        (unsigned long)nbodies);
  }

  if (RARRAY_LEN(r_btypes) != nbodies-1)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "potential requires %lu bond types, but %lu were given",
        (unsigned long)potential->nbodies-1, 
        (unsigned long)RARRAY_LEN(r_btypes));
  }

  decode_ptype_list(mrb, r_ptypes, ptypes, nbodies);
  decode_bond_type_list(mrb, r_btypes, btypes, nbodies);

  ts_bond_pattern* pat = NULL;
  tsmrb_handle_error(mrb,
      ts_add_bonded_pattern(sim, nbodies, ptypes, btypes, btName, potential, &pat) );

  if (pat && !mrb_nil_p(r_termgroup))
  {
    mrb_check_type(mrb, r_termgroup, MRB_TT_FIXNUM);
    pat->term_group = mrb_fixnum(r_termgroup);
  }

  return self;
}

static mrb_value cSim_add_improper_pattern(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  const char* btName    = hash_get_str  (mrb, kwArgs, TSMRB_SYM(name)          , "bonded pattern name");
  mrb_value r_btypes    = hash_get_value(mrb, kwArgs, TSMRB_SYM(bondtypes)     , "bond types"         );
  mrb_value r_potential = hash_get_value(mrb, kwArgs, TSMRB_SYM(potential)     , "bond potential"     );
  mrb_value r_ptypes    = hash_get_value(mrb, kwArgs, TSMRB_SYM(particle_types), "particle types"     );

  mrb_value r_termgroup = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(term_group));

  Potential* potential = tsmrb_to_potential(mrb, r_potential);
  if (potential == NULL)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "internal potential is NULL");
  if (potential->nbodies != 4)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "improper bond requires 4-body potential (%d-body potential given)",
        potential->nbodies);

  int nbodies = RARRAY_LEN(r_ptypes);
  if (nbodies != 4)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "improper bond requires four particles, %d given", nbodies);

  if (RARRAY_LEN(r_btypes) != 3)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, 
        "improper terms require three bond types, %d given",
        (int)RARRAY_LEN(r_btypes));

  int ptypes[NBODY_LIMIT] = PTYPES_INIT;
  ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;

  decode_ptype_list(mrb, r_ptypes, ptypes, nbodies);
  decode_bond_type_list(mrb, r_btypes, btypes, nbodies);

  ts_bond_pattern* pat = NULL;
  tsmrb_handle_error(mrb,
      ts_add_improper_pattern(sim, ptypes, btypes, btName, potential, &pat));

  if ((pat != NULL) && !mrb_nil_p(r_termgroup))
    pat->term_group = mrb_fixnum(r_termgroup);

  return self;
}

static int symbol_to_bondterm_type(mrb_state* mrb, mrb_sym btId, int* num_ptypes)
{
  int bondType = 0;
  if (btId == TSMRB_SYM(onebody))   { *num_ptypes = 1; bondType = TS_BOND_TERM_ONEBODY  ; }

  if (btId == TSMRB_SYM(twobody))   { *num_ptypes = 2; bondType = TS_BOND_TERM_TWOBODY  ; }
  if (btId == TSMRB_SYM(nonbond))   { *num_ptypes = 2; bondType = TS_BOND_TERM_NONBOND  ; }

  if (btId == TSMRB_SYM(threebody)) { *num_ptypes = 3; bondType = TS_BOND_TERM_THREEBODY; }
  if (btId == TSMRB_SYM(fourbody )) { *num_ptypes = 4; bondType = TS_BOND_TERM_FOURBODY ; }

  if (bondType == 0)
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unknown bond type :%S", mrb_symbol_value(btId));

  return bondType;
}

static void cvt_ptypes(mrb_state* mrb, mrb_value r_ptypes, int num_ptypes, int ptypes[NBODY_LIMIT])
{
  if (RARRAY_LEN(r_ptypes) != num_ptypes)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "array length is %lu, but expected %lu particle types",
        (unsigned long)RARRAY_LEN(r_ptypes),
        (unsigned long)num_ptypes);
  }

  for(int i=0; i < num_ptypes; i++) {
    mrb_value elt = mrb_ary_ref(mrb,r_ptypes,i);
    mrb_check_type(mrb, elt, MRB_TT_FIXNUM);
    ptypes[i] = mrb_fixnum(elt);
  }
}

static void cvt_btypes(mrb_state* mrb, mrb_value r_btypes, int num_btypes, ts_bond_type btypes[NBOND_LIMIT])
{
  mrb_check_type(mrb, r_btypes, MRB_TT_ARRAY);
  if (RARRAY_LEN(r_btypes) != num_btypes)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "array length is %lu, but expected %lu bond types",
        (unsigned long)RARRAY_LEN(r_btypes),
        (unsigned long)(num_btypes));
  }

  decode_bond_type_list(mrb, r_btypes, btypes, num_btypes+1);
}

static inline void init_btypes(int num_btypes, ts_bond_type btypes[NBOND_LIMIT])
{
  for (int i=0; i < num_btypes; i++)
    btypes[i] = bondtype_new( BONDTYPE_UNDIRECTED, 1 );
}

static mrb_value build_bond_pattern_table(mrb_state* mrb, ts_bond_pattern* pat)
{
  if (pat->nbodies > NBODY_LIMIT)
    mrb_raise(mrb, E_RUNTIME_ERROR, "potential nbodies exceeds limit");

  mrb_value table = mrb_hash_new(mrb);

  mrb_hash_set(mrb, table, TSMRB_SYM_OBJ(name)     , mrb_str_new_cstr(mrb, pat->name));
  mrb_hash_set(mrb, table, TSMRB_SYM_OBJ(potential), tsmrb_from_potential(mrb,pat->potential));

  mrb_value rpt[NBODY_LIMIT];
  for (int i=0; i < pat->nbodies; i++)
    rpt[i] = mrb_fixnum_value( pat->ptypes[i] );

  mrb_hash_set(mrb, table, TSMRB_SYM_OBJ(particle_types), 
      mrb_ary_new_from_values(mrb, pat->nbodies, rpt));

  mrb_value rbt[NBOND_LIMIT];
  for (int i=0; i < pat->nbodies-1; i++)
    rbt[i] = tsmrb_bondtype_value(mrb, pat->btypes[i]);

  mrb_hash_set(mrb, table, TSMRB_SYM_OBJ(bondtypes), 
      mrb_ary_new_from_values(mrb, pat->nbodies-1, rbt));

  mrb_hash_set(mrb, table, TSMRB_SYM_OBJ(term_group),
      mrb_fixnum_value(pat->term_group));

  return table;
}

static mrb_value bond_patterns_to_ruby(mrb_state *mrb, DATAVECTOR(ts_bond_pattern)* patlist)
{
  mrb_value rpatterns = mrb_ary_new_capa(mrb, patlist->size);
  ts_bond_pattern* pat = DATAVECTOR_BEGIN(ts_bond_pattern, patlist);
  ts_bond_pattern* end = DATAVECTOR_END(ts_bond_pattern, patlist);
  for(; pat < end; pat++) {
    mrb_value rp = build_bond_pattern_table(mrb, pat);
    mrb_ary_push(mrb, rpatterns, rp);
    pat++;
  }

  return rpatterns;
}

static mrb_value cSim_all_bonds(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim;
  size_t i;
  int j;
  mrb_value list;

  sim = tsmrb_to_sim(mrb,self);
  list = mrb_ary_new(mrb);

  for (i=0; i < sim->num_particles; i++)
  {
    ts_particle *p = TS_PARTICLE(sim,i);
    for (j=0; j < p->num_bonds; j++)
    {
      if (p->bonds[j].pi == p)
      {
        mrb_value binfo[3];
        binfo[0] = mrb_fixnum_value( TS_PIND(sim,p->bonds[j].pi) );
        binfo[1] = mrb_fixnum_value( TS_PIND(sim,p->bonds[j].pj) );
        binfo[2] = mrb_fixnum_value( bondtype_to_int(p->bonds[j].type) );

        mrb_ary_push(mrb, list, mrb_ary_new_from_values(mrb, 3, binfo));
      }
    }
  }

  return list;
}

static mrb_value cSim_all_rigid_bonds(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim;
  size_t i;
  int j;
  mrb_value list;

  sim = tsmrb_to_sim(mrb, self);
  list = mrb_ary_new(mrb);

  for (i=0; i < sim->num_particles; i++)
  {
    ts_particle *p = TS_PARTICLE(sim,i);
    for (j=0; j < p->num_bonds; j++)
    {
      if (p->bonds[j].pi == p)
      {
        if ((p->bonds[j].pi->rigid_body_index != TS_NULL_PARTICLE) &&
            (p->bonds[j].pi->rigid_body_index == p->bonds[j].pj->rigid_body_index))
        {
          continue;
        }

        mrb_value binfo[5];
        binfo[0] = mrb_fixnum_value( TS_PIND(sim,p->bonds[j].pi) );
        binfo[1] = mrb_fixnum_value( p->bonds[j].pi->rigid_body_index );
        binfo[2] = mrb_fixnum_value( TS_PIND(sim,p->bonds[j].pj) );
        binfo[3] = mrb_fixnum_value( p->bonds[j].pj->rigid_body_index );
        binfo[4] = mrb_fixnum_value( bondtype_to_int(p->bonds[j].type) );

        mrb_ary_push(mrb, list, mrb_ary_new_from_values(mrb, 5, binfo));
      }
    }
  }

  return list;
}

static mrb_value cSim_add_extra_term(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim;
  Potential* pot;
  const char *name;
  mrb_value r_pot, r_inds;
  const mrb_value *inds;
  mrb_int i,n;
  const static struct ts_potential_term zero;
  struct ts_potential_term term;

  sim = tsmrb_to_sim(mrb, self);

  name = NULL;

  r_pot = mrb_nil_value();
  r_inds = mrb_nil_value();
  mrb_get_args(mrb, "zoA", &name, &r_pot, &r_inds);

  pot = tsmrb_to_potential(mrb, r_pot);
  n = mrb_ary_len(mrb,r_inds);

  if (pot->nbodies != n)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "potential requires %lu bodies, but %lu were given",
        (unsigned long)pot->nbodies, 
        (unsigned long)n);
  }

  if (n < 1 || n > NBODY_LIMIT)
  {
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid number of bodies: %d", (int)n);
  }

  term = zero;

  term.name = ts_strdup(NO_ALLOCATOR, name);
  term.potential = pot;
  term.pattern = NULL;

  inds = RARRAY_PTR(r_inds);
  for (i=0; i < n; i++)
  {
    mrb_int pi;

    if (!mrb_fixnum_p(inds[i]))
    {
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "element %zu is not a fixnum", (size_t)i);
    }

    pi = mrb_fixnum(inds[i]);
    if (pi < 0 || (size_t)pi >= sim->num_particles)
    {
      tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %d at element %zu", pi, (size_t)i);
    }

    term.particle_indices[i] = pi;
  }

  tsmrb_handle_error(mrb, ts_add_extra_term(sim, &term));

  return mrb_nil_value();
}

static mrb_value cSim_all_bond_patterns(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim = tsmrb_to_sim(mrb, self);
  return bond_patterns_to_ruby(mrb, &sim->bond_patterns1);
}

static mrb_value cSim_all_nonbond_patterns(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim = tsmrb_to_sim(mrb, self);
  return bond_patterns_to_ruby(mrb, &sim->nonbond_patterns1);
}

static mrb_value cSim_get_bond_pattern(mrb_state* mrb, mrb_value self)
{
  int argc;
  mrb_value* argv;
  mrb_get_args(mrb, "*", &argv, &argc);

  ts_sim* sim = tsmrb_to_sim(mrb,self);
  ts_bond_pattern* pat = NULL;

  if ((argc < 1) || (argc > 3))
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "wrong number of arguments (%d for 2 or 3)", argc);

  if (argc == 1)
  {
    mrb_check_type(mrb, argv[0], MRB_TT_STRING);
    const char* name = mrb_string_value_cstr(mrb, argv+0);
    pat = ts_lookup_bond_pattern(sim, name);
  }
  else
  {
    mrb_value r_bondterm_type = argv[0];
    mrb_value r_ptypes = argv[1];

    mrb_check_type(mrb, r_bondterm_type, MRB_TT_SYMBOL);
    mrb_check_type(mrb, r_ptypes       , MRB_TT_ARRAY );

    int ptypes[NBODY_LIMIT] = PTYPES_INIT;

    int num_ptypes = 0;
    int bondterm_type = symbol_to_bondterm_type(mrb, mrb_symbol(r_bondterm_type), &num_ptypes);

    cvt_ptypes(mrb, r_ptypes, num_ptypes, ptypes);

    ts_bond_type btypes[NBOND_LIMIT] = BTYPES_INIT;
    if (argc == 3)
      cvt_btypes(mrb, argv[2], num_ptypes-1, btypes);
    else
      init_btypes(num_ptypes-1,btypes);

    pat = ts_get_bond_pattern(sim, num_ptypes, ptypes, btypes, bondterm_type);
  }

  return pat != NULL
    ? build_bond_pattern_table(mrb,pat)
    : mrb_nil_value();
}

static mrb_value cSim_add_dist_constraint(mrb_state* mrb, mrb_value self)
{
  mrb_value kwArgs = mrb_nil_value();
  mrb_get_args(mrb, "H", &kwArgs);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  const char* name = hash_get_str  (mrb, kwArgs, TSMRB_SYM(name)    , "constraint name"    );
  mrb_float dist   = hash_get_float(mrb, kwArgs, TSMRB_SYM(dist)    , "constraint distance");
  mrb_value r_btype  = hash_get_value(mrb, kwArgs, TSMRB_SYM(bondtype), "bond type"        );
  mrb_value r_ptypes = hash_get_value(mrb, kwArgs, TSMRB_SYM(types)   , "particle types"   );

  if (!mrb_array_p(r_ptypes) || (RARRAY_LEN(r_ptypes) != 2))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "expected :types keyword to be a 2-element array of particle types");

  int ptype1 = mrb_fixnum( mrb_ary_ref(mrb, r_ptypes, 0) );
  int ptype2 = mrb_fixnum( mrb_ary_ref(mrb, r_ptypes, 1) );
  ts_bond_type btype  = decode_bond_type(mrb, r_btype);

  tsmrb_handle_error(mrb, 
      ts_add_distance_constraint(sim, ptype1, ptype2, btype, name, dist));

  return self;
}

static mrb_value cSim_get_bond_list(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  mrb_value r_bondlist = mrb_ary_new_capa(mrb, sim->bondlist.terms.size);

  size_t i;
  for(i=0; i < sim->bondlist.terms.size; i++)
  {
    ts_potential_term* entry = DATAVECTOR_AT(ts_potential_term, &sim->bondlist.terms, i);

    mrb_value r_plist[NBODY_LIMIT];
    for(int j=0; j < entry->potential->nbodies; j++)
    {
      r_plist[j] = tsmrb_from_particle(mrb, sim, entry->particle_indices[j]);
    }

    mrb_value bl_elts[2] = {
      mrb_ary_new_from_values(mrb, entry->potential->nbodies, r_plist),
      tsmrb_from_potential(mrb, entry->potential),
    };

    mrb_ary_push(mrb, r_bondlist, 
        mrb_ary_new_from_values(mrb, 2, bl_elts));
  }

  return r_bondlist;
}

static mrb_value add_bonds_to_sim(mrb_state* mrb, mrb_value self, enum BONDTYPE_DIRECTION direc)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  for(int i=0; i < argc; i++)
  {
    mrb_check_type(mrb, argv[i], MRB_TT_ARRAY);

    size_t alen = RARRAY_LEN(argv[i]);
    if ((alen < 2) && (alen > 3))
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
          "argument %d must be a two or three element array", i);

    size_t pi = mrb_to_fixnum(mrb, mrb_ary_ref(mrb, argv[i], 0));
    size_t pj = mrb_to_fixnum(mrb, mrb_ary_ref(mrb, argv[i], 1));

    int bondtype = (alen == 3) ? mrb_to_fixnum(mrb, mrb_ary_ref(mrb,argv[i],2)) : 1;

    if (pi >= sim->num_particles) { tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %lu", pi); }
    if (pj >= sim->num_particles) { tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %lu", pj); }
    if (pi == pj) { mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot bond particles with the same index"); }

    TSResult err = ts_particle_add_bond(
          ts_get_particle(sim, pi),
          ts_get_particle(sim, pj),
          bondtype_new(direc, bondtype));
    tsmrb_handle_error(mrb, err);
  }

  return self;
}

static mrb_value cSim_change_bonds(mrb_state* mrb, mrb_value self)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  for(int i=0; i < argc; i++)
  {
    mrb_check_type(mrb, argv[i], MRB_TT_ARRAY);
    size_t alen = RARRAY_LEN(argv[i]);
    if (alen != 3)
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
          "argument %d must be a three element array", i);

    mrb_value r_pi = mrb_ary_ref(mrb,argv[i],0);
    mrb_value r_pj = mrb_ary_ref(mrb,argv[i],1);
    mrb_check_type(mrb, r_pi, MRB_TT_FIXNUM);
    mrb_check_type(mrb, r_pj, MRB_TT_FIXNUM);

    mrb_int pi = mrb_fixnum(r_pi);
    mrb_int pj = mrb_fixnum(r_pj);

    if ((pi < 0) || (pi >= sim->num_particles))
      tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %d", (mrb_int)pi);

    if ((pj < 0) || (pj >= sim->num_particles))
      tsmrb_raisef(mrb, E_RANGE_ERROR, "invalid particle index %d", (mrb_int)pj);

    if (pi == pj)
      mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot bond particles with the same index");

    ts_bond_type bt = decode_bond_type(mrb, mrb_ary_ref(mrb,argv[i],2));

    tsmrb_handle_error(mrb,
        ts_particle_modify_bond(ts_get_particle(sim, pi), ts_get_particle(sim, pj), bt));
  }

  return self;
}

static mrb_value cSim_add_bonds(mrb_state* mrb, mrb_value self)
{
  return add_bonds_to_sim(mrb, self, BONDTYPE_UNDIRECTED);
}

static mrb_value cSim_add_directed_bonds(mrb_state* mrb, mrb_value self)
{
  return add_bonds_to_sim(mrb, self, BONDTYPE_FORWARD);
}

static mrb_value cSim_is_bonded_p(mrb_state* mrb, mrb_value self)
{
  mrb_int pi = 0;
  mrb_int pj = 0;
  mrb_get_args(mrb, "ii", &pi, &pj);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if ((pi < 0) || (pi >= sim->num_particles))
    mrb_raise(mrb, E_INDEX_ERROR, "first index is invalid");

  if ((pj < 0) || (pj >= sim->num_particles))
    mrb_raise(mrb, E_INDEX_ERROR, "second index is invalid");

  ts_particle* p0 = ts_get_particle(sim, (size_t)pi);
  ts_particle* p1 = ts_get_particle(sim, (size_t)pj);

  return mrb_bool_value( ts_particle_is_bonded_to(p0,p1) );
}

static mrb_value integrator_to_sym(mrb_state* mrb, int integrator)
{
  switch(integrator) {
    case TS_INTEGRATOR_NONE:              return TSMRB_SYM_OBJ(none);
    case TS_INTEGRATOR_VELOCITY_VERLET:   return TSMRB_SYM_OBJ(velocity_verlet);
    case TS_INTEGRATOR_LEAPFROG_VERLET:   return TSMRB_SYM_OBJ(leapfrog);
    case TS_INTEGRATOR_LEAPFROG_LANGEVIN: return TSMRB_SYM_OBJ(leapfrog_langevin);
    case TS_INTEGRATOR_STOCH_RUNGE_KUTTA: return TSMRB_SYM_OBJ(SRK);

    default: return mrb_nil_value();
  }
}

static int sym_to_integrator(mrb_state* mrb, mrb_sym sym)
{
  if (sym == TSMRB_SYM(none             )) return TS_INTEGRATOR_NONE;
  if (sym == TSMRB_SYM(velocity_verlet  )) return TS_INTEGRATOR_VELOCITY_VERLET;
  if (sym == TSMRB_SYM(leapfrog         )) return TS_INTEGRATOR_LEAPFROG_VERLET;
  if (sym == TSMRB_SYM(leapfrog_langevin)) return TS_INTEGRATOR_LEAPFROG_LANGEVIN;
  if (sym == TSMRB_SYM(SRK              )) return TS_INTEGRATOR_STOCH_RUNGE_KUTTA;

  return -1;
}

static mrb_value cSim_set_integrator(mrb_state* mrb, mrb_value self)
{
  mrb_sym integrator_sym;
  mrb_get_args(mrb, "n", &integrator_sym);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  int integrator = sym_to_integrator(mrb, integrator_sym);
  if (integrator == -1)
  {
    mrb_raisef(mrb, E_RUNTIME_ERROR,
        "invalid integrator symbol :%S",
        mrb_symbol_value(integrator_sym));
  }

  sim->integrator = integrator;
  return self;
}

static mrb_value cSim_get_integrator(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  mrb_value integrator_sym = integrator_to_sym(mrb, sim->integrator);
  if (mrb_nil_p(integrator_sym))
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "unknown integrator setting %d", sim->integrator);

  return integrator_sym;
}

static mrb_value cSim_set_rng(mrb_state* mrb, mrb_value self)
{
  mrb_value r_rng = mrb_nil_value();
  mrb_get_args(mrb, "o", &r_rng);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  struct ts_rng* rng;

  if (!mrb_obj_is_instance_of(mrb, r_rng, tsmrb_rng_class(mrb)))
  {
    mrb_value arg = r_rng;
    r_rng = mrb_funcall(mrb,
        mrb_obj_value(tsmrb_rng_class(mrb)),
        "new",
        1, arg);
    if (! mrb_obj_is_instance_of(mrb, r_rng, tsmrb_rng_class(mrb)))
      mrb_raise(mrb, E_RUNTIME_ERROR, "error creating RNG");
  }

  rng = tsmrb_to_rng(mrb,r_rng);
  if (rng == NULL) { mrb_raise(mrb, E_ARGUMENT_ERROR, "RNG state is NULL"); }
  ts_set_rng(sim, rng);
  mrb_iv_set(mrb, self, mrb_intern_lit(mrb,"@rng"), r_rng);

  return self;
}

static mrb_value cSim_get_rng(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if (sim->rng_state == NULL)
    return mrb_nil_value();
  else
    return tsmrb_from_rng(mrb, sim->rng_state);
}

static mrb_value get_random_values(mrb_state *mrb, mrb_value self,
    TSResult (*rng_func)(struct ts_rng*,size_t,double*))
{
  ts_sim* sim;
  double* v;
  TSResult result;
  mrb_value ary;
  mrb_int i, n;

  sim = tsmrb_to_sim(mrb, self);
  if (sim->rng_state == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "RNG state is NULL");
  }

  n = 1;
  mrb_get_args(mrb, "|i", &n);
  if (n <= 0)
  {
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid number or pRNGs requested (%d)", (int)n);
  }

  v = mrb_malloc(mrb, n * sizeof(v[0]));
  result = rng_func(sim->rng_state, n, v);
  if (!IS_OKAY(result))
  {
    mrb_free(mrb, v);
    tsmrb_handle_error(mrb, result);
  }

  ary = mrb_ary_new_capa(mrb, n);
  for (i = 0; i < n; i++)
  {
    int ai = mrb_gc_arena_save(mrb);
    mrb_ary_push(mrb, ary, mrb_float_value(mrb, v[i]));
    mrb_gc_arena_restore(mrb,ai);
  }

  mrb_free(mrb, v);

  return ary;
}

static mrb_value
cSim_random(mrb_state *mrb, mrb_value self)
{
  return get_random_values(mrb, self, ts_rng_gen_uniforms);
}

static mrb_value
cSim_randn(mrb_state *mrb, mrb_value self)
{
  return get_random_values(mrb, self, ts_rng_gen_normals);
}

static mrb_value
cSim_randn_seq(mrb_state *mrb, mrb_value self)
{
  return get_random_values(mrb, self, ts_rng_gen_normals_sequential);
}

static mrb_value
cSim_randn_block(mrb_state *mrb, mrb_value self)
{
  return get_random_values(mrb, self, ts_rng_gen_normals_block);
}

static mrb_value cSim_many_steps(mrb_state* mrb, mrb_value self)
{
  mrb_int nsteps = 0;
  mrb_get_args(mrb, "i", &nsteps);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  if (nsteps <= 0)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "nsteps must be greater than zero");

  tsmrb_handle_error(mrb, ts_sim_many_steps(sim, nsteps));
  return mrb_float_value(mrb, sim->time);
}

static mrb_value cSim_step_until(mrb_state* mrb, mrb_value self)
{
  mrb_float t_end = 0.0;
  mrb_get_args(mrb, "f", &t_end);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  tsmrb_handle_error(mrb, ts_sim_step_until(sim,t_end));

  return mrb_float_value(mrb, sim->time);
}

static mrb_value hessian_to_numarray(mrb_state *mrb, size_t num_dof, VecReal* hess)
{
  numarray *A;
  int dims[2] = { (int)num_dof, (int)num_dof };
  mrb_value arr;

  A = NULL;
  arr = numarray_new_filled(mrb, 2, dims, 0.0, &A);

  {
    size_t i,j;
    double *restrict hp = A->elts;
    for (i=0; i < num_dof; i++) {
      hp[ i*num_dof + i ] = hess[ hessian_index(num_dof, i,i) ];
      for (j=i+1; j < num_dof; j++) {
        VecReal val = hess[ hessian_index(num_dof, i,j) ];
        hp[ i*num_dof + j ] = val;
        hp[ j*num_dof + i ] = val;
      }
    }
  }

  return arr;
}

static mrb_value sparse_builder_to_numarray(mrb_state *mrb, ts_sparse_builder_t* builder)
{
  numarray *A;
  int dims[2] = { builder->m, builder->n };
  mrb_value arr;

  ts_sparse_t matrix;
  tsmrb_handle_error(mrb, ts_sparse_builder_build(builder,&matrix));

  A = NULL;
  arr = numarray_new_filled(mrb, 2, dims, 0.0, &A);

  ts_sparse_full(&matrix, A->elts, TS_ROWMAJOR);
  ts_sparse_finalize(&matrix);

  return arr;
}

static enum TS_ROT_REPR
sym_to_rot_repr(mrb_state* mrb, mrb_sym sym, int allow_composite)
{
  if      (sym == TSMRB_SYM( euler         )) return TS_ROT_EULER_ZXZ;
  else if (sym == TSMRB_SYM( ZXZ           )) return TS_ROT_EULER_ZXZ;
  else if (sym == TSMRB_SYM( XYZ           )) return TS_ROT_EULER_XYZ;
  else if (sym == TSMRB_SYM( axis_angle    )) return TS_ROT_AXIS_ANGLE;
  else if (sym == TSMRB_SYM( body_relative )) return TS_ROT_BODY_RELATIVE;
  else if (sym == TSMRB_SYM( body_axes     ))
  {
    if (!allow_composite)
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "composite rotational representations are not allowed");
    return TS_ROT_BODY_AXES;
  }
  else
  {
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid rotation representation :%S",
        mrb_symbol_value(sym));
    return -1;
  }
}

static mrb_value
rot_repr_to_sym(mrb_state* mrb, enum TS_ROT_REPR rot_repr)
{
  switch(rot_repr)
  {
    case TS_ROT_NONE:          return TSMRB_SYM_OBJ( none );
    case TS_ROT_EULER_XYZ:     return TSMRB_SYM_OBJ( XYZ );
    case TS_ROT_EULER_ZXZ:     return TSMRB_SYM_OBJ( ZXZ );
    case TS_ROT_AXIS_ANGLE:    return TSMRB_SYM_OBJ( axis_angle );
    case TS_ROT_BODY_AXES:     return TSMRB_SYM_OBJ( body_axes );
    case TS_ROT_BODY_RELATIVE: return TSMRB_SYM_OBJ( body_relative );
    default: return mrb_nil_value();
  }
}

enum TS_ROT_REPR arg_rot_repr(mrb_state* mrb, mrb_value r_repr, int allow_composite)
{
  if (mrb_nil_p(r_repr))
    return DEFAULT_ROT_REPR;
  else
    return sym_to_rot_repr(mrb, mrb_symbol(r_repr), allow_composite);
}

static mrb_value
cSim_debug_break(mrb_state* mrb, mrb_value self)
{
  return self;
}

static mrb_value
cSim_numerical_hessian_check_all_dof(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  VecReal *hess, *hess_err;
  size_t num_dof;
  mrb_value r_rot_repr, pair[2];
  enum TS_ROT_REPR rot_repr;

  r_rot_repr = mrb_nil_value();
  mrb_get_args(mrb, "|o", &r_rot_repr);
  rot_repr = arg_rot_repr(mrb, r_rot_repr, 1);

  sim = tsmrb_to_sim(mrb,self);

  num_dof = 0;
  hess = hess_err = NULL;

  tsmrb_handle_error(mrb,
      ts_sim_numerical_hessian_check_all_dof(
        sim, &hess, &hess_err, &num_dof, rot_repr) );

  pair[0] = hessian_to_numarray(mrb, num_dof, hess);
  pair[1] = hessian_to_numarray(mrb, num_dof, hess_err);

  ts_safe_free(NO_ALLOCATOR, hess);
  ts_safe_free(NO_ALLOCATOR, hess_err);

  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value
cSim_numerical_hessian_by_potential_terms(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  VecReal *hess, *hess_err;
  size_t num_dof;
  mrb_value r_rot_repr, pair[2];
  enum TS_ROT_REPR rot_repr;

  r_rot_repr = mrb_nil_value();
  mrb_get_args(mrb, "|o", &r_rot_repr);
  rot_repr = arg_rot_repr(mrb, r_rot_repr, 1);

  sim = tsmrb_to_sim(mrb,self);

  num_dof = 0;
  hess = hess_err = NULL;

  tsmrb_handle_error(mrb,
      ts_sim_numerical_hessian_by_potential_terms(
        sim, &hess, &hess_err, &num_dof, rot_repr) );

  pair[0] = hessian_to_numarray(mrb, num_dof, hess);
  pair[1] = hessian_to_numarray(mrb, num_dof, hess_err);

  ts_safe_free(NO_ALLOCATOR, hess);
  ts_safe_free(NO_ALLOCATOR, hess_err);

  return mrb_ary_new_from_values(mrb, 2, pair);
}

static mrb_value cSim_sparse_hessian(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  ts_sparse_builder_t* builder = NULL;
  mrb_value r_builder;

  sim = tsmrb_to_sim(mrb,self);
  r_builder = tsmrb_new_sparse_builder(mrb, &builder );

  tsmrb_handle_error(mrb, ts_sim_hessian(sim,builder,NULL));

  return r_builder;
}

static mrb_value build_hessian(mrb_state *mrb, mrb_value self, int mass_weighted)
{
  ts_sim* sim;
  ts_sparse_builder_t builder;
  mrb_value hess;

  sim = tsmrb_to_sim(mrb,self);
  if (mass_weighted)
  {
    tsmrb_handle_error(mrb, ts_sim_mass_weighted_hessian(sim,&builder,NULL) );
  }
  else
  {
    tsmrb_handle_error(mrb, ts_sim_hessian(sim,&builder,NULL) );
  }

  hess = sparse_builder_to_numarray(mrb, &builder);
  ts_sparse_builder_finalize(&builder);
  return hess;
}

static mrb_value cSim_hessian(mrb_state *mrb, mrb_value self)
{
  return build_hessian(mrb, self, 0);
}

static mrb_value cSim_mass_weighted_hessian(mrb_state *mrb, mrb_value self)
{
  return build_hessian(mrb, self, 1);
}

static mrb_value get_dof_map_inner(mrb_state* mrb, ts_dof_map* dof_map)
{
  mrb_value rdof = mrb_hash_new(mrb);
  mrb_hash_set(mrb, rdof,
      TSMRB_SYM_OBJ(num_free_particles),
      mrb_fixnum_value(dof_map->num_free_particles));

  mrb_hash_set(mrb, rdof,
      TSMRB_SYM_OBJ(num_rot_rigids),
      mrb_fixnum_value(dof_map->num_rot_rigids));

  mrb_hash_set(mrb, rdof,
      TSMRB_SYM_OBJ(num_free_rigids),
      mrb_fixnum_value(dof_map->num_free_rigids));

  mrb_hash_set(mrb, rdof,
      TSMRB_SYM_OBJ(num_dof),
      mrb_fixnum_value(dof_map->num_dof));

  mrb_hash_set(mrb, rdof,
      TSMRB_SYM_OBJ(num_objects),
      mrb_fixnum_value(dof_map->num_objects));

  mrb_value rfp = mrb_ary_new_capa(mrb, dof_map->num_free_particles);
  for (size_t i=0; i < dof_map->num_free_particles; i++)
    mrb_ary_push(mrb, rfp, mrb_fixnum_value(dof_map->fp_to_p[i]));

  mrb_hash_set(mrb, rdof, TSMRB_SYM_OBJ(fp_to_p), rfp);

  mrb_value rfr = mrb_ary_new_capa(mrb, dof_map->num_free_rigids);
  for (size_t i=0; i < dof_map->num_free_rigids; i++)
    mrb_ary_push(mrb, rfr, mrb_fixnum_value(dof_map->fr_to_r[i]));

  mrb_hash_set(mrb, rdof, TSMRB_SYM_OBJ(fr_to_r), rfr);

  mrb_value rrr = mrb_ary_new_capa(mrb, dof_map->num_rot_rigids);
  for (size_t i=0; i < dof_map->num_rot_rigids; i++)
    mrb_ary_push(mrb, rrr, mrb_fixnum_value(dof_map->rr_to_r[i]));

  mrb_hash_set(mrb, rdof, TSMRB_SYM_OBJ(rr_to_r), rrr);

  return rdof;
}

static mrb_value cSim_get_dof_map(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim;
  ts_dof_map dof_map;
  mrb_sym opt_map;

  sim = tsmrb_to_sim(mrb,self);
  opt_map = TSMRB_SYM(all);
  mrb_get_args(mrb, "|n", &opt_map);

  if (opt_map == TSMRB_SYM(all))
  {
    tsmrb_handle_error(mrb, ts_dof_map_init(&dof_map, sim));
  }
  else if (opt_map == TSMRB_SYM(rotation))
  {
    tsmrb_handle_error(mrb, ts_dof_map_init_with_type(&dof_map, sim, TS_DOF_ROTATIONAL));
  }
  else if (opt_map == TSMRB_SYM(translation))
  {
    tsmrb_handle_error(mrb, ts_dof_map_init_with_type(&dof_map, sim, TS_DOF_TRANSLATIONAL));
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid DoF map type");
  }

  mrb_value ret = get_dof_map_inner(mrb, &dof_map);
  ts_dof_map_finalize(&dof_map);

  return ret;
}

static mrb_value cSim_get_dof_vec(mrb_state* mrb, mrb_value self)
{
  ts_sim *sim;
  struct ts_dof_map dof_map;
  VecReal *dofv;
  size_t num_obj, num_dof;
  enum TS_ROT_REPR rot_repr, *rot_reps = NULL;
  mrb_value dof_ret, rot_ret, r_rot_repr;

  dof_ret = mrb_nil_value();
  rot_ret = mrb_nil_value();

  r_rot_repr = mrb_nil_value();
  mrb_get_args(mrb, "|o", &r_rot_repr);
  sim = tsmrb_to_sim(mrb,self);
  rot_repr = arg_rot_repr(mrb,r_rot_repr,1);

  tsmrb_handle_error(mrb, ts_dof_map_init(&dof_map, sim));

  num_dof = dof_map.num_dof;
  num_obj = dof_map.num_objects;
  dofv = mrb_calloc(mrb, num_dof, sizeof(*dofv));

  if (rot_repr == TS_ROT_BODY_AXES)
    rot_reps = mrb_calloc(mrb, dof_map.num_objects, sizeof(*rot_reps));

  if (rot_repr != TS_ROT_BODY_AXES)
    ts_dof_to_vec(sim, &dof_map, dofv, rot_repr);
  else
    ts_sim_copy_dof_to_vec_in_body_axes(sim, &dof_map, dofv, rot_reps);

  ts_dof_map_finalize(&dof_map);

  dof_ret = mrb_ary_new_capa(mrb,num_dof);
  for (size_t i=0; i < num_dof; i++)
    mrb_ary_push(mrb, dof_ret, mrb_float_value(mrb,(mrb_float)dofv[i]));

  mrb_free(mrb, dofv);

  if (rot_repr != TS_ROT_BODY_AXES)
  {
    return dof_ret;
  }
  else
  {
    mrb_value pair[2];
    rot_ret = mrb_ary_new_capa(mrb,dof_map.num_objects);
    for (size_t i=0; i < num_obj; i++)
    {
      mrb_ary_push(mrb, rot_ret, rot_repr_to_sym(mrb,rot_reps[i]));
    }

    mrb_free(mrb,rot_reps);

    pair[0] = dof_ret;
    pair[1] = rot_ret;
    return mrb_ary_new_from_values(mrb, 2, pair);
  }
}

static mrb_value cSim_get_grad_vec(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim;
  struct ts_dof_map dof_map;
  size_t num_dof;
  mrb_value r_gradv;
  numarray *arr;
  int dim;

  sim = tsmrb_to_sim(mrb,self);
  num_dof = ts_sim_num_dof(sim);

  if (num_dof > INT_MAX)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "too many degrees of freedom (max is %d)", INT_MAX);
  }

  dim = num_dof;
  arr = NULL;
  r_gradv = numarray_new_uninitialized(mrb, 1, &dim, &arr);

  tsmrb_handle_error(mrb, ts_dof_map_init(&dof_map, sim));
  ts_dof_forces_to_gradient_vec(sim, &dof_map, arr->elts);
  ts_dof_map_finalize(&dof_map);

  return r_gradv;
}

struct set_dof_data {
  ts_sim* sim;
  ts_dof_map dof_map;
  mrb_value rdofv;
  enum TS_ROT_REPR rot_repr;
};

static mrb_value set_dof_vec_inner(mrb_state* mrb, struct set_dof_data* data)
{
  VecReal *dofv, maxdisp[2];

  if (RARRAY_LEN(data->rdofv) != data->dof_map.num_dof)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "expected %lu elements in DoF vector",
        (unsigned long)data->dof_map.num_dof);
  }

  dofv = mrb_calloc(mrb, data->dof_map.num_dof, sizeof(*dofv));

  for (size_t i=0; i < data->dof_map.num_dof; i++)
  {
    dofv[i] = (VecReal) mrb_to_flo(mrb, RARRAY_PTR(data->rdofv)[i]);
  }

  maxdisp[0] = maxdisp[1] = 0.0;

  ts_dof_from_vec(data->sim, &data->dof_map, dofv, maxdisp, data->rot_repr);

  mrb_free(mrb, dofv);

  return mrb_nil_value();
}

static mrb_value cSim_set_dof_vec(mrb_state* mrb, mrb_value self)
{
  mrb_value rdofv = mrb_nil_value();
  mrb_sym rrotr = 0;
  int argc = mrb_get_args(mrb, "A|n", &rdofv, &rrotr);

  struct set_dof_data data = {0};
  data.sim = tsmrb_to_sim(mrb,self);

  data.rdofv = rdofv;

  data.rot_repr = (argc == 2)
    ? sym_to_rot_repr(mrb, rrotr, 0)
    : DEFAULT_ROT_REPR;

  tsmrb_handle_error(mrb, 
      ts_dof_map_init(&(data.dof_map),data.sim) );

  set_dof_vec_inner(mrb,&data);

  ts_dof_map_finalize(&(data.dof_map));

  return self;
}

static mrb_value cSim_set_dof(mrb_state* mrb, mrb_value self)
{
  mrb_value r_dofv0;
  mrb_sym rot_repr_sym;
  mrb_int dof_ind;
  mrb_float val;
  mrb_get_args(mrb, "Anif",
      &r_dofv0, &rot_repr_sym, &dof_ind, &val);

  ts_sim* sim = tsmrb_to_sim(mrb,self);
  struct ts_dof_map dof_map;
  tsmrb_handle_error(mrb, 
      ts_dof_map_init(&dof_map, sim) );

  if (mrb_ary_len(mrb, r_dofv0) != dof_map.num_dof)
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "expected %lu elements in DoF vector", dof_map.num_dof);
  }

  enum TS_ROT_REPR rot_repr = sym_to_rot_repr(mrb,rot_repr_sym,0);

  if ((dof_ind < 0) || (dof_ind >= dof_map.num_dof))
  {
    tsmrb_raisef(mrb, E_RANGE_ERROR,
        "invalid dof index (%d)", (int)dof_ind);
  }

  VecReal* dofv = ts_alloc(NO_ALLOCATOR, dof_map.num_dof * sizeof(dofv[0]));
  if (dofv == NULL)
  {
    tsmrb_raisef(mrb, E_RUNTIME_ERROR, "cannot allocate DoF vector");
  }

  for (size_t i=0; i < dof_map.num_dof; i++)
  {
    dofv[i] = mrb_float(mrb_ary_ref(mrb, r_dofv0, i));
  }

  dofv[dof_ind] = val;

  ts_sim_set_dof(sim, &dof_map, dof_ind, dofv, rot_repr);
  ts_free(NO_ALLOCATOR, dofv);
  ts_dof_map_finalize(&dof_map);

  return self;
}

static mrb_value get_normal_mode_freqs(mrb_state *mrb, mrb_value self,
    VecReal (*freqfunc)(VecReal))
{
  ts_sim* sim;
  int enabled[TS_BOND_TERM_MAX];
  size_t n, off;
  VecReal *freqs;

  sim = tsmrb_to_sim(mrb,self);

  {
    mrb_int argc;
    mrb_value *argv;
    int i;

    argc = 0;
    argv = NULL;
    mrb_get_args(mrb, "*", &argv, &argc);

    memset(enabled, 0, sizeof(enabled));

    int all_syms = 0;
    if (argc == 0)
    {
      all_syms = 1;
    }
    if (argc == 1)
    {
      mrb_check_type(mrb, argv[0], MRB_TT_SYMBOL);
      all_syms = (mrb_symbol(argv[0]) == mrb_intern_lit(mrb, "ALL"));
    }

    if (all_syms)
    {
      for (i=0; i < TS_BOND_TERM_MAX; i++) { enabled[i] = 1; }
    }
    else
    {
      struct bond_list_term *all_terms = tsmrb_classes(mrb)->bond_list_terms;

      for (i=0; i < argc; i++)
      {
        mrb_check_type(mrb, argv[i], MRB_TT_SYMBOL);
        for (struct bond_list_term *terms = all_terms; terms->name != NULL; terms++)
        {
          if (mrb_symbol(argv[i]) == terms->sym)
          {
            enabled[ terms->term ] = 1;
            break;
          }
        }
      }
    }
  }

  {
    size_t skip;
    VecReal *modes;
    double potE;
    TSResult ret;

    n = skip = off = 0;
    potE = 0.0;
    freqs = modes = NULL;

    ret = ts_sim_normal_mode_square_freqs(sim, 
        &freqs, &modes, &n, &potE, enabled, skip, &off);
    ts_safe_free(NO_ALLOCATOR, modes);

    if (!IS_OKAY(ret)) { ts_safe_free(NO_ALLOCATOR, freqs); }
    tsmrb_handle_error(mrb, ret);

    if (freqfunc != NULL) {
      size_t i;
      for (i=0; i < n; i++)
      {
        freqs[i+off] = freqfunc(freqs[i+off]);
      }
    }
  }

  {
    numarray *A;
    mrb_value rA;
    int i, nfreqs = n;

    A = NULL;
    rA = numarray_new_uninitialized(mrb, 1, &nfreqs, &A);

    for (i=0; i < nfreqs; i++)
    {
      A->elts[i] = freqs[off+i];
    }

    ts_safe_free(NO_ALLOCATOR, freqs);

    return rA;
  }
}

static mrb_value cSim_normal_mode_square_freqs(mrb_state *mrb, mrb_value self)
{
  return get_normal_mode_freqs(mrb,self,NULL);
}

static mrb_value cSim_normal_mode_freqs(mrb_state *mrb, mrb_value self)
{
  return get_normal_mode_freqs(mrb,self,sqrt);
}

static mrb_value cSim_normal_mode_free_energy(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  int linear, components;
  mrb_value kwargs;
  TSResult err;
  mrb_value ret;

  sim = tsmrb_to_sim(mrb,self);

  kwargs = mrb_nil_value();
  mrb_get_args(mrb, "|H", &kwargs);

  linear = components = 0;
  if (mrb_hash_p(kwargs))
  {
    linear     = mrb_test(mrb_hash_fetch(mrb, kwargs, TSMRB_SYM_OBJ(linear), mrb_nil_value()));
    components = mrb_test(mrb_hash_fetch(mrb, kwargs, TSMRB_SYM_OBJ(components), mrb_nil_value()));
  }

  ts_sim_nma nma;
  ts_sim_nma_init(&nma, sim);

  nma.freq_skip = linear ? 5 : 6;
  nma.flags = TS_NMA_FL_CALC_FREQS;

  err = ts_sim_nma_free_energy(&nma);
  if (IS_ERROR(err))
  {
    ts_safe_free(NO_ALLOCATOR, nma.freqs);
    ts_sim_nma_finalize(&nma);
    tsmrb_handle_error(mrb, err);
  }

  if (components)
  {
    ret = mrb_hash_new(mrb);
    mrb_hash_set(mrb, ret, TSMRB_SYM_OBJ(enthalpic), mrb_float_value(mrb, nma.enthalpic_energy) );
    mrb_hash_set(mrb, ret, TSMRB_SYM_OBJ(entropic ), mrb_float_value(mrb, nma.entropic_energy ) );
    mrb_hash_set(mrb, ret, TSMRB_SYM_OBJ(free     ), mrb_float_value(mrb, nma.free_energy     ) );
  }
  else
  {
    ret = mrb_float_value(mrb, nma.free_energy);
  }

  ts_sim_nma_finalize(&nma);

  return ret;
}

static mrb_value cSim_system_energies(mrb_state* mrb,mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  mrb_value hash = mrb_hash_new(mrb);
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(KE   ), mrb_float_value(mrb, sim->KE    ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(rotKE), mrb_float_value(mrb, sim->rotKE ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(PE   ), mrb_float_value(mrb, sim->PE    ));
  mrb_hash_set(mrb, hash, TSMRB_SYM_OBJ(totE ), mrb_float_value(mrb, sim->totalE));

  return hash;
}

void collect_term_energies(mrb_state *mrb, ts_sim *sim, 
    const struct ts_potential_term_list *terms,
    mrb_value hash)
{
  struct ts_potential_term_data *term_data;
  size_t *pinds;
  size_t i,n, max_pind;

  term_data = terms->term_data.data;
  pinds = terms->particle_indices.data;

  n = terms->term_data.size;
  max_pind = terms->particle_indices.size;

  for (i=0; i < n; i++)
  {
    mrb_value key, inds[NBODY_LIMIT];
    int j;
    double en;
    size_t off = term_data[i].vec_off;

    if (off + term_data[i].nbodies > max_pind)
    {
      mrb_raise(mrb, E_RUNTIME_ERROR,
          "index into pind array is out of bounds (did you run calculate_energy or calculate_forces?)");
    }

    en = dbldbl_to_double(term_data[i].term_energy);
    if (en == 0.0) { continue; }

    for(j=0; j < term_data[i].nbodies; j++)
    {
      inds[j] = mrb_fixnum_value(pinds[off+j]);
    }

    key = mrb_ary_new_from_values(mrb, term_data[i].nbodies, inds);
    mrb_hash_set(mrb, hash, key, mrb_float_value(mrb,en));
  }
}

static mrb_value
cSim_potential_term_energies(mrb_state *mrb, mrb_value self)
{
  ts_sim *sim = tsmrb_to_sim(mrb,self);
  mrb_value terms = mrb_hash_new(mrb);

  collect_term_energies(mrb, sim, &sim->bondlist, terms);
  collect_term_energies(mrb, sim, &sim->nonbond_list, terms);

  return terms;
}

static mrb_value cSim_add_rigid_body_type(mrb_state* mrb, mrb_value self)
{
  mrb_value typeDef;
  mrb_get_args(mrb, "H", &typeDef);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  struct ts_rigid_body_type rbtype;
  const char* name = hash_get_str(mrb, typeDef, TSMRB_SYM(name), "particle type description");
  rbtype.name    = ts_strdup(NO_ALLOCATOR, name);
  rbtype.I0      = hash_get_float(mrb, typeDef, TSMRB_SYM(I0)     , "symmetric moment of inertia");
  rbtype.rotfric = hash_get_float(mrb, typeDef, TSMRB_SYM(rotfric), "rotational friction");

  mrb_value rb_type = mrb_hash_get(mrb, typeDef, TSMRB_SYM_OBJ(type));
  if (!mrb_symbol_p(rb_type))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected :type keyword to in rigid body definition");

  enum TS_RIGID_BODY_TYPE type = TS_RIGID_NOT_DEFINED;
  if (mrb_symbol(rb_type) == TSMRB_SYM(site_interaction))
    type = TS_RIGID_SITE_INTERACTION;
  else
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "unknown rigid body type :%S", rb_type);

  rbtype.type = type;

  mrb_value rb_sites = mrb_hash_get(mrb, typeDef, TSMRB_SYM_OBJ(reference_vectors));
  if (!mrb_array_p(rb_sites))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected :reference_vectors keyword to be an array of site locations");

  rbtype.num_sites = RARRAY_LEN(rb_sites);

  rbtype.ref_vectors = ts_calloc( NO_ALLOCATOR, rbtype.num_sites, sizeof(rbtype.ref_vectors[0]) );
  for (size_t i=0; i < rbtype.num_sites; i++)
    obj_to_vec(mrb, mrb_ary_ref(mrb, rb_sites, i), rbtype.ref_vectors+i);

  ts_rigid_body_type_update_parameters(sim, &rbtype);

  size_t rbTypeInd = 0;
  tsmrb_handle_error(mrb, 
      ts_add_rigid_body_type(sim, &rbtype, &rbTypeInd));

  return mrb_fixnum_value((mrb_int)rbTypeInd);
}

static mrb_value cSim_set_rigid_body_parameters(mrb_state* mrb, mrb_value self)
{
  mrb_int bd_type_ind = 0;
  mrb_value params = mrb_nil_value();
  mrb_get_args(mrb, "iH", &bd_type_ind, &params);

  ts_sim* sim = tsmrb_to_sim(mrb,self);
  if ((bd_type_ind <= 0) || (bd_type_ind > sim->rigid_body_types.size))
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR, "invalid rigid body type %d", (int)bd_type_ind);

  struct ts_rigid_body_type* bd_type = ts_get_rigid_body_type(sim, bd_type_ind);

  if (bd_type == NULL)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid rigid body index %d", (int)bd_type_ind);

  mrb_value rbI0      = mrb_hash_get(mrb, params, TSMRB_SYM_OBJ(I0));
  mrb_value rbRotFric = mrb_hash_get(mrb, params, TSMRB_SYM_OBJ(rotfric));

  if (!mrb_nil_p(rbI0))      { bd_type->I0 = mrb_float(rbI0); }
  if (!mrb_nil_p(rbRotFric)) { bd_type->rotfric = mrb_float(rbRotFric); }

  ts_rigid_body_type_update_parameters(sim, bd_type);

  return self;
}

mrb_value tsmrb_rigid_body_type_info(mrb_state* mrb, struct ts_rigid_body_type *bd_type, int derived)
{
  mrb_value rb_def = mrb_hash_new(mrb);
  mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(name), mrb_str_new_cstr(mrb,bd_type->name));

  mrb_value type_sym = TSMRB_SYM_OBJ(unknown);
  if (bd_type->type == TS_RIGID_SITE_INTERACTION)
    type_sym = TSMRB_SYM_OBJ(site_interaction);
  mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(type), type_sym);

  mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(I0),            mrb_float_value(mrb,bd_type->I0));
  mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(rotfric),       mrb_float_value(mrb,bd_type->rotfric));

  if (derived)
  {
    mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(inv_I0),        mrb_float_value(mrb,bd_type->inv_I0));
    mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(inv_rotfric),   mrb_float_value(mrb,bd_type->inv_rotfric));
    mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(bd_rand_coeff), mrb_float_value(mrb,bd_type->bd_rand_coeff));
  }

  mrb_value reference_vecs = mrb_ary_new_capa(mrb,bd_type->num_sites);
  for (size_t i=0; i < bd_type->num_sites; i++)
    mrb_ary_push(mrb, reference_vecs, tsmrb_from_vec3(mrb, bd_type->ref_vectors+i));

  mrb_hash_set(mrb, rb_def, TSMRB_SYM_OBJ(reference_vectors), reference_vecs);

  return rb_def;
}

static mrb_value cSim_get_rigid_body_type_info(mrb_state* mrb, mrb_value self)
{
  mrb_int bd_type_ind = 0;
  mrb_get_args(mrb, "i", &bd_type_ind);
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  enum { INCLUDE_DERIVED_INFO = 1 };

  struct ts_rigid_body_type* bd_type = ts_get_rigid_body_type(sim, bd_type_ind);
  if (bd_type == NULL)
    tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
        "get_particle_type: invalid particle type %d",
        (int)bd_type_ind);

  return tsmrb_rigid_body_type_info(mrb, bd_type, INCLUDE_DERIVED_INFO);
}

static mrb_value cSim_alloc_rigid_bodies(mrb_state* mrb, mrb_value self)
{
  mrb_int num = 0;
  mrb_get_args(mrb, "i", &num);
  tsmrb_handle_error(mrb, 
      ts_alloc_rigid_bodies(tsmrb_to_sim(mrb,self), num));
  return self;
}

mrb_value cSim_to_array(mrb_state *mrb, mrb_value self)
{
  int dims[2];
  ts_sim* sim;

  sim = tsmrb_to_sim(mrb, self);
  dims[0] = sim->num_particles;
  dims[1] = 3;

  numarray *A = NULL;
  mrb_value arr = numarray_new_filled(mrb, 2, dims, 0.0, &A);

  if (A == NULL)
  {
    mrb_raisef(mrb, E_RUNTIME_ERROR, "could not allocate %dx3 array for positions",
        (int)(sim->num_particles));
  }

  {
    double*restrict xvec = A->elts;
    double*restrict yvec = xvec + sim->num_particles;
    double*restrict zvec = yvec + sim->num_particles;

    for(size_t i=0; i < sim->num_particles; i++)
    {
      xvec[i] = VECTOR3_X(PARTICLE_POS(sim,i));
      yvec[i] = VECTOR3_Y(PARTICLE_POS(sim,i));
      zvec[i] = VECTOR3_Z(PARTICLE_POS(sim,i));
    }
  }

  return arr;
}

static mrb_value cSim_linked_elements(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  char* links;
  size_t i,j, num_objs;
  mrb_value linkmat;

  sim = tsmrb_to_sim(mrb,self);

  links = NULL;
  num_objs = 0;
  tsmrb_handle_error(mrb,
      ts_sim_numerical_hessian_linked_elements(
        sim, &links, &num_objs) );

  linkmat = mrb_ary_new_capa(mrb, num_objs);
  for (i=0; i < num_objs; i++)
  {
    mrb_value row = mrb_ary_new_capa(mrb, num_objs);
    for (j=0; j < num_objs; j++)
    {
      mrb_ary_push(mrb, row,
          mrb_fixnum_value( links[ hessian_index(num_objs,i,j) ] ));
    }
    mrb_ary_push(mrb, linkmat, row);
  }

  ts_free(NO_ALLOCATOR, links);
  return linkmat;
}

static mrb_value cSim_sum_all_forces(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  vector3 frc;
  vector3_zero(&frc);

  for (size_t i=0; i < sim->num_particles; i++)
    vector3_add_to(&frc, PARTICLE_FORCE(sim, i));

  return tsmrb_from_vec3(mrb, &frc);
}

static mrb_value cSim_energy_is_crazy_p(mrb_state* mrb,mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  return mrb_bool_value(isinf(sim->totalE) || isnan(sim->totalE));
}

static const struct {
  int symtab_index;
  const char* name;
  enum TS_MINIMIZE_METHOD method;
} minimizers[] = {
  { TSMRB_SYM_IND(steepest), "steepest descent", TS_MINIMIZE_STEEPEST },
  { TSMRB_SYM_IND(lbfgs)   , "L-BFGS"          , TS_MINIMIZE_LBFGS    },
  { -1 }
};

static enum TS_MINIMIZE_METHOD 
get_minimization_method(mrb_state *mrb, mrb_value r_method)
{
  if (mrb_nil_p(r_method))
  {
    return TS_MINIMIZE_DEFAULT;
  }

  if (!mrb_symbol_p(r_method))
  {
    mrb_raise(mrb, E_TYPE_ERROR, ":method argument must be a symbol");
  }

  mrb_sym msym = mrb_symbol(r_method);
  for (int i=0; minimizers[i].symtab_index >= 0; i++)
  {
    if (msym == tsmrb_symtab_at(mrb,minimizers[i].symtab_index)) {
      return minimizers[i].method;
    }
  }

  mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid KW argument :method");
  return TS_MINIMIZE_MAXMETHOD;
}

static enum TS_MINIMIZE_WHAT
get_minimization_what(mrb_state *mrb, mrb_value r_what)
{
  if (mrb_nil_p(r_what) || (mrb_symbol(r_what) == TSMRB_SYM(all)))
  {
    return TS_MINIMIZE_ALL;
  }
  else if (mrb_symbol(r_what) == TSMRB_SYM(rotation))
  {
    return TS_MINIMIZE_ORI;
  }
  else if (mrb_symbol(r_what) == TSMRB_SYM(translation))
  {
    return TS_MINIMIZE_POS;
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "unknown value for :what parameters");
  }

  return TS_MINIMIZE_ALL;
}

#define DEFAULT_MINIMIZER_REPORT_FREQ 200
static mrb_value cSim_minimize(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  struct ts_minimize_params params;
  mrb_value kwArgs, r_num_iters, r_disable_nonbonds,
            r_report_freq, r_method, r_epsilon,
            r_maxstep, r_what, r_shifted, r_dimsubsp;

  mrb_get_args(mrb, "|H", &kwArgs);
  if (mrb_nil_p(kwArgs)) { kwArgs = mrb_hash_new(mrb); }

  r_num_iters        = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(num_iterations));
  r_disable_nonbonds = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(disable_nonbonds));
  r_report_freq      = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(report_freq));
  r_method           = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(method));
  r_epsilon          = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(epsilon));
  r_maxstep          = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(maxstep));
  r_what             = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(what));
  r_shifted          = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(shifted));
  r_dimsubsp         = mrb_hash_get(mrb, kwArgs, TSMRB_SYM_OBJ(dimsubsp));

  if (!mrb_numeric_p(r_num_iters))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, ":num_iterations is missing or not a fixnum");
  }

  if (!mrb_numeric_p(r_report_freq) && !mrb_nil_p(r_report_freq))
  {
    tsmrb_raisef(mrb, E_TYPE_ERROR,
        ":report_freq argument is not a fixnum (type %s)",
        mrb_obj_classname(mrb, r_report_freq));
  }

  if (!mrb_float_p(r_epsilon) && !mrb_nil_p(r_epsilon))
  {
    mrb_raise(mrb, E_TYPE_ERROR, ":epsilon argument is not a float");
  }

  ts_minimize_initialize_params(&params);
  params.method = get_minimization_method(mrb, r_method);
  params.what   = get_minimization_what(mrb, r_what);
  params.use_nonbonds = !mrb_test(r_disable_nonbonds);
  params.max_iter = mrb_to_fixnum(mrb, r_num_iters);
  params.rescale_dofs = mrb_test(r_shifted);

  params.report_freq = mrb_numeric_p(r_report_freq)
    ? mrb_to_fixnum(mrb, r_report_freq)
    : DEFAULT_MINIMIZER_REPORT_FREQ
    ;

  if (mrb_float_p(r_epsilon))
  {
    params.eps = mrb_float(r_epsilon);
  }

  if (mrb_numeric_p(r_dimsubsp))
  {
    params.dim_subsp = mrb_to_fixnum(mrb, r_dimsubsp);
  }

  if (mrb_numeric_p(r_maxstep))
  {
    params.maxstep = mrb_to_float(mrb, r_maxstep);
  }

  tsmrb_handle_error(mrb, ts_minimize_energy(sim, &params));

  return self;
}

#define CHECK_COMPONENT(sym,flagval) \
  if (component == TSMRB_SYM(sym)) { if (fl) { *fl = (flagval); } return 1; }
static inline int get_component_flag(mrb_state* mrb, mrb_sym component, uint32_t* fl)
{
  CHECK_COMPONENT(general   , TS_LOG_GENERAL   );
  CHECK_COMPONENT(integrator, TS_LOG_INTEGRATOR);
  CHECK_COMPONENT(minimizer , TS_LOG_MINIMIZER );
  CHECK_COMPONENT(nonbonds  , TS_LOG_NONBONDS  );
  CHECK_COMPONENT(io        , TS_LOG_IO        );

  return 0;
}
#undef CHECK_COMPONENT

#define CHECK_LEVEL(sym,val) \
  if (level == TSMRB_SYM(sym)) { if (loglvl) { *loglvl = (val); } return 1; }
static inline int get_level_flag(mrb_state* mrb, mrb_sym level, enum TS_LOGGING_LEVEL* loglvl)
{
  CHECK_LEVEL( verbose , TS_LOG_VERBOSE );
  CHECK_LEVEL( info    , TS_LOG_INFO    );
  CHECK_LEVEL( warn    , TS_LOG_WARN    );
  CHECK_LEVEL( error   , TS_LOG_ERROR   );
  CHECK_LEVEL( abort   , TS_LOG_ABORT   );

  return 0;
}
#undef CHECK_LEVEL

static mrb_value cSim_is_logging_p(mrb_state* mrb, mrb_value self)
{
  mrb_sym level;
  mrb_sym component;
  mrb_get_args(mrb, "nn", &level, &component);

  ts_sim* sim = tsmrb_to_sim(mrb,self);

  uint32_t cflag = 0;
  enum TS_LOGGING_LEVEL lvl;
  if (!get_level_flag(mrb,level, &lvl))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown level");

  if (!get_component_flag(mrb,component,&cflag))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown component");

  return mrb_bool_value( (cflag != 0) && ts_is_component_logging(sim,lvl,cflag) );
}

static mrb_value cSim_set_logging(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim;
  mrb_sym component;
  mrb_value tf;
  uint32_t cflag;

  component = 0;
  tf = mrb_nil_value();
  mrb_get_args(mrb, "no", &component, &tf);

  if (mrb_symbol_p(tf))
  {
    /* MRuby #define's false to something (grr...), so we have to use
     * tsmrb_symtab_at(mrb, TSMRBSYM_false) here instead of using
     * TSMRB_SYM(false)
     */
    tf = mrb_bool_value( mrb_symbol(tf) != tsmrb_symtab_at(mrb, TSMRBSYM_false) );
  }

  sim = tsmrb_to_sim(mrb,self);

  cflag = 0;
  if (!get_component_flag(mrb,component,&cflag))
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "unknown component");
  }

  ts_set_component_logging(sim, cflag, mrb_test(tf));
  return self;
}

static void print_bond_patterns(FILE* outf, const char* title, DATAVECTOR(ts_bond_pattern)* patterns)
{
  fprintf(outf, "---[ %s ]---\n", title);
  fprintf(outf,"dbg: %p\n", patterns);
  for (size_t i=0; i < patterns->size; i++)
  {
    ts_bond_pattern* pattern = DATAVECTOR_AT(ts_bond_pattern, patterns, i);
    const char* term_type = NULL;
    switch(pattern->type) {
      case TS_BOND_TERM_ONEBODY:    term_type = "1-BODY";   break;
      case TS_BOND_TERM_TWOBODY:    term_type = "2-BODY";   break;
      case TS_BOND_TERM_THREEBODY:  term_type = "3-BODY";   break;

      case TS_BOND_TERM_FOURBODY:   term_type = "4-BODY";   break;
      case TS_BOND_TERM_IMPROPER:   term_type = "IMPROPER"; break;

      case TS_BOND_TERM_NONBOND:    term_type = "NONBOND";  break;
      case TS_BOND_TERM_CONSTRAINT: term_type = "SHAKE";    break;
      default: term_type = "UNKNOWN"; break;

    }
    fprintf(outf, "[%-36s] %-16s %-8s (%d)\n", 
        pattern->name, pattern->potential->name, term_type, 
        pattern->potential->nbodies);
    if (pattern->type != TS_BOND_TERM_IMPROPER) {
      const char* bond_fmt = "--(%d)--%d";

      if (pattern->type == TS_BOND_TERM_NONBOND) {
        bond_fmt = "..(%d)..%d";
      } else if (pattern->type == TS_BOND_TERM_CONSTRAINT) {
        bond_fmt = "==(%d)==%d";
      }

      fprintf(outf, "  %d", pattern->ptypes[0]);
      for (int i=1; i < pattern->potential->nbodies; i++) {
        fprintf(outf, bond_fmt, pattern->btypes[i-1], pattern->ptypes[i]);
      }
    } else {
      char buf[64];
      snprintf(buf, sizeof(buf), "%d--(%d)--%d", 
          pattern->ptypes[1], 
          bondtype_to_int(pattern->btypes[0]), 
          pattern->ptypes[0]);
      size_t buf_len = strlen(buf);

      fprintf(outf, "  %s--(%d)--%d\n", buf, bondtype_to_int(pattern->btypes[1]), pattern->ptypes[2]);

      for(int i=0; i < sizeof(buf); i++) { buf[i] = ' '; };
      buf[sizeof(buf)-1] = '\0'; // zero-terminate
      snprintf(buf+buf_len, sizeof(buf)-buf_len, "`-(%d)--%d", bondtype_to_int(pattern->btypes[2]), pattern->ptypes[3]);
      fprintf(outf, "  %s\n", buf);
    }

    fprintf(outf,"\n");

    {
      struct ts_potential_params* params = NULL;
      ts_potential_get_params(pattern->potential,&params);
      if (params == NULL)
      {
        fprintf(outf, "  <could not get parameters>\n");
      }
      else
      {
        for (size_t i=0; params[i].name != NULL; i++)
        {
          fprintf(outf, "  %12s  %7.4f\n", params[i].name, params[i].value);
        }
        ts_free(NO_ALLOCATOR, params);
      }
    }

    fprintf(outf, "  Current PE  = %7.2f\n", pattern->potential->potential_energy);

  }
  fprintf(outf,"\n");
}

static mrb_value cSim_print_bond_patterns(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  print_bond_patterns(sim->logging_output, "BONDED", &sim->bond_patterns1);
  print_bond_patterns(sim->logging_output, "NON-BOND", &sim->nonbond_patterns1);

  return self;
}

static mrb_value cSim_print_summary(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  tsmrb_handle_error(mrb, ts_sim_print_summary(sim, sim->logging_output));
  return self;
}

static mrb_value cSim_print_potential_summary(mrb_state* mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  tsmrb_handle_error(mrb, ts_print_potential_summary(sim, sim->logging_output));
  return self;
}

static mrb_value cSim_gradient_magnitude(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim = tsmrb_to_sim(mrb,self);
  double en, fmag, tmag, fmin, fmax;
  size_t i, fcount, tcount;
  mrb_value items[4];

  en = 0.0;
  tsmrb_handle_error(mrb, ts_calculate_forces(sim, &en, NO_TERM_GROUPS, TERMWISE_FALSE) );
  tsmrb_handle_error(mrb, ts_rigid_bodies_calculate_all_forces_and_torques(sim) );

  fcount = tcount = 0;
  fmag = tmag = 0.0;

  fmin =  INFINITY;
  fmax = -INFINITY;

  for (i=0; i < sim->num_particles; i++) {
    if (!ts_pind_in_rigid_body(sim, i)) {
      double n = vector3_norm(PARTICLE_FORCE(sim,i));
      fmag += n;
      fcount++;
      if (n > fmax) { fmax = n; }
      if (n < fmin) { fmin = n; }
    }
  }

  for (i=0; i < sim->num_rigid_bodies; i++) {
    double nf,nt;

    nf = vector3_norm(RIGID_FORCE(sim,i));
    fmag += nf;
    fcount++;

    if (nf > fmax) { fmax = nf; }
    if (nf < fmin) { fmin = nf; }

    nt = vector3_norm(RIGID_TORQUE(sim,i));
    tmag += nt;
    tcount++;
  }

  fmag = fcount > 0 ? fmag/fcount : 0;
  tmag = tcount > 0 ? tmag/tcount : 0;

  items[0] = mrb_float_value(mrb, fmag);
  items[1] = mrb_float_value(mrb, tmag);
  items[2] = mrb_float_value(mrb, fmin);
  items[3] = mrb_float_value(mrb, fmax);

  return mrb_ary_new_from_values(mrb, 4, items);
}

static mrb_value cSim_get_parameter(mrb_state* mrb, mrb_value self)
{
  const char* key;
  mrb_get_args(mrb, "z", &key);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  double val = 0.0;
  TSResult err = ts_get_parameter(sim, key, &val);

  if (ERRCODE(err) == TS_ERR_NO_SUCH_PARAMETER)
    return mrb_nil_value();


  tsmrb_handle_error(mrb, err);
  return mrb_float_value(mrb,val);
}

static mrb_value cSim_has_parameter(mrb_state* mrb, mrb_value self)
{ 
  const char* key;
  mrb_get_args(mrb, "z", &key);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  return mrb_bool_value(ts_has_parameter(sim, key));
}

static mrb_value cSim_set_parameter(mrb_state* mrb, mrb_value self)
{ 
  const char* key = NULL;
  mrb_float val = 0.0;
  mrb_get_args(mrb, "zf", &key, &val);
  ts_sim* sim = tsmrb_to_sim(mrb,self);

  tsmrb_handle_error(mrb, ts_set_parameter(sim, key, val));
  return self; 
}

static mrb_value cSim_each_parameter(mrb_state* mrb, mrb_value self)
{ 
  ts_sim *sim;
  mrb_value block, ret;
  struct ts_param_iter iter = {0};

  sim = tsmrb_to_sim(mrb,self);

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  ret = mrb_nil_value();
  if (mrb_nil_p(block))
  {
    ret = mrb_ary_new(mrb);
  }

  ts_param_iter_init(sim->parameters, &iter);
  while (ts_param_iter_next(sim->parameters,&iter))
  {
    int ai;
    mrb_value pair[2];

    ai = mrb_gc_arena_save(mrb);
    pair[0] = mrb_str_new_cstr(mrb,iter.key);
    pair[1] = mrb_float_value(mrb,iter.val);

    if (mrb_nil_p(block))
    {
      mrb_value elt;
      elt = mrb_ary_new_from_values(mrb, 2, pair);
      mrb_ary_push(mrb,ret,elt);
    }
    else
    {
      mrb_yield_argv(mrb, block, 2, pair);
    }

    mrb_gc_arena_restore(mrb,ai);
  }

  return ret;
}

static mrb_value cSim_to_vectorlist(mrb_state *mrb, mrb_value self)
{
  ts_sim* sim;
  mrb_sym vtype;
  mrb_value ret;
  VectorList* vl;
  size_t i;

  sim = tsmrb_to_sim(mrb, self);
  vtype = TSMRB_SYM(pos);
  mrb_get_args(mrb, "|n", &vtype);

  ret = tsmrb_new_veclist(mrb, sim->num_particles, &vl);

  if (vtype == TSMRB_SYM(pos))
  {
    for (i=0; i < sim->num_particles; i++)
    {
      vl->vX[i] = VECTOR3_X( PARTICLE_POS(sim,i) );
      vl->vY[i] = VECTOR3_Y( PARTICLE_POS(sim,i) );
      vl->vZ[i] = VECTOR3_Z( PARTICLE_POS(sim,i) );
    }
  }
  else if (vtype == TSMRB_SYM(vel))
  {
    for (i=0; i < sim->num_particles; i++)
    {
      vl->vX[i] = VECTOR3_X( PARTICLE_VEL(sim,i) );
      vl->vY[i] = VECTOR3_Y( PARTICLE_VEL(sim,i) );
      vl->vZ[i] = VECTOR3_Z( PARTICLE_VEL(sim,i) );
    }
  }
  else if (vtype == TSMRB_SYM(force))
  {
    for (i=0; i < sim->num_particles; i++)
    {
      vl->vX[i] = VECTOR3_X( PARTICLE_FORCE(sim,i) );
      vl->vY[i] = VECTOR3_Y( PARTICLE_FORCE(sim,i) );
      vl->vZ[i] = VECTOR3_Z( PARTICLE_FORCE(sim,i) );
    }
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "invalid vector type (expected :pos, :vel, or :force)");
  }

  return ret;
}

struct RClass* tsmrb_define_sim(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_sim_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Sim class has already been initialized!");

  struct RClass* cSim = mrb_define_class_under(mrb, mod,
      "Sim", mrb->object_class);
  tsmrb_classes(mrb)->cSim = cSim;
  MRB_SET_INSTANCE_TT(cSim, MRB_TT_DATA);

  sim_weakref_table_setup(mrb, cSim);

  mrb_define_method(mrb, cSim, "initialize"  , cSim_initialize, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "alloc_particles"  , cSim_alloc_particles  , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "num_particles"    , cSim_num_particles    , MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "max_particles"    , cSim_max_particles    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "num_free_particles"    , cSim_num_free_particles    , MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "num_dof"          , cSim_num_dof          , MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "particle_to_free_particle_index_map", 
      cSim_get_particle_to_free_particle_index_map, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "free_particle_to_particle_index_map", 
      cSim_get_free_particle_to_particle_index_map, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "add_particle_type", cSim_add_particle_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "get_particle_type", cSim_get_particle_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "find_particle_type", cSim_find_particle_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "num_particle_types", cSim_num_particle_types, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "add_particle_set", cSim_add_particle_set, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "add_to_particle_set", cSim_add_to_particle_set, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "delete_particle_set", cSim_delete_particle_set, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "get_particle_set", cSim_get_particle_set, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "num_particle_sets", cSim_num_particle_sets, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "get_particle_type_info", cSim_get_particle_type_info, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "add_bond_type", cSim_add_bond_type, MRB_ARGS_REQ(1)|MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "get_bond_type", cSim_get_bond_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "num_bond_types", cSim_num_bond_types, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "each_bond_type", cSim_each_bond_type, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "add_nonbond_pattern" , cSim_add_nonbond_pattern , MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cSim, "add_bonded_pattern"  , cSim_add_bonded_pattern  , MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "add_improper_pattern", cSim_add_improper_pattern, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "count_bond_patterns" , cSim_count_bond_patterns , MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "get_bond_pattern"    , cSim_get_bond_pattern    , MRB_ARGS_ANY());

  mrb_define_method(mrb, cSim, "add_extra_term"    , cSim_add_extra_term, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cSim, "all_bond_patterns"      , cSim_all_bond_patterns   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "all_nonbond_patterns"   , cSim_all_nonbond_patterns, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "bonds_enabled"       , cSim_get_bonds_enabled   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "enable_bonds"        , cSim_enable_bonds        , MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "nonbond_cutoff", cSim_get_nonbond_cutoff, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "nonbond_padding", cSim_get_nonbond_padding, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "nonbond_cutoff=", cSim_set_nonbond_cutoff, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "nonbond_padding=", cSim_set_nonbond_padding, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "nonbonds_enabled" , cSim_get_nonbonds_enabled, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "nonbonds_enabled=", cSim_set_nonbonds_enabled, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "add_dist_constraint", cSim_add_dist_constraint, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "build_potential"   , cSim_build_potential, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "bond_list"        , cSim_get_bond_list    , MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "add_bonds"         , cSim_add_bonds         , MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "add_directed_bonds", cSim_add_directed_bonds, MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "change_bonds"      , cSim_change_bonds      , MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "is_bonded?"        , cSim_is_bonded_p       , MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cSim, "count_potential_terms", cSim_count_potential_terms, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "count_constraints", cSim_count_constraints, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "clear_forces"    , cSim_clear_forces    ,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "calculate_forces", cSim_calculate_forces,  MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "calculate_energy", cSim_calculate_forces,  MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSim, "calculate_forces_and_torques_on_rigid_bodies", 
      cSim_calculate_forces_and_torques_on_rigid_bodies,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "numerical_hessian_check_all_dof",
      cSim_numerical_hessian_check_all_dof,  MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "numerical_hessian_by_potential_terms", 
      cSim_numerical_hessian_by_potential_terms,  MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "hessian", cSim_hessian,  MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "sparse_hessian", cSim_sparse_hessian,  MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "mass_weighted_hessian", cSim_mass_weighted_hessian,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "dof_map", cSim_get_dof_map, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "dof_vec", cSim_get_dof_vec, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "grad_vec", cSim_get_grad_vec, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "set_dof_vec", cSim_set_dof_vec, MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "set_dof", cSim_set_dof, MRB_ARGS_REQ(4));

  mrb_define_method(mrb, cSim, "normal_mode_freqs", cSim_normal_mode_freqs, MRB_ARGS_ANY());
  mrb_define_method(mrb, cSim, "normal_mode_square_freqs", cSim_normal_mode_square_freqs, MRB_ARGS_ANY());

  mrb_define_method(mrb, cSim, "normal_mode_free_energy", cSim_normal_mode_free_energy, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSim, "rng="            , cSim_set_rng         ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "rng"             , cSim_get_rng         ,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "random"          , cSim_random          ,  MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "randn"           , cSim_randn           ,  MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "randn_seq"       , cSim_randn_seq       ,  MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cSim, "randn_block"     , cSim_randn_block     ,  MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSim, "integrator="     , cSim_set_integrator  ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "integrator"      , cSim_get_integrator  ,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "dt="             , cSim_set_dt          ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "dt"              , cSim_get_dt          ,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "temp="           , cSim_set_temp        ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "temp"            , cSim_get_temp        ,  MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "time"            , cSim_get_time        ,  MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "integrator_steps", cSim_get_integrator_steps, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "single_step"     , cSim_single_step     ,  MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "many_steps"      , cSim_many_steps      ,  MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "step_until"      , cSim_step_until      ,  MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "system_energies" , cSim_system_energies ,  MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "potential_term_energies", cSim_potential_term_energies, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "add_rigid_body_type", cSim_add_rigid_body_type, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "set_rigid_body_parameters", cSim_set_rigid_body_parameters, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "num_rigid_body_types", cSim_num_rigid_body_types, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "get_rigid_body_type_info", cSim_get_rigid_body_type_info, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "find_rigid_body_type", cSim_find_rigid_body_type, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "alloc_rigid_bodies", cSim_alloc_rigid_bodies, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "num_rigid_bodies", cSim_get_num_rigid_bodies, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "max_rigid_bodies", cSim_get_max_rigid_bodies, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "update_rigid_bodies", cSim_update_rigid_bodies, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "net_force", cSim_get_net_force, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "net_torque", cSim_get_net_torque, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "linked_elements", cSim_linked_elements, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "gradient_magnitude", cSim_gradient_magnitude, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "sum_all_forces", cSim_sum_all_forces, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "to_array", cSim_to_array, MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSim, "print_potential_terms", cSim_print_potential_terms, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "print_bond_patterns", cSim_print_bond_patterns, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "print_summary", cSim_print_summary, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "print_potential_summary", cSim_print_potential_summary, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "energy_is_crazy?", cSim_energy_is_crazy_p, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "minimize", cSim_minimize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "is_logging?", cSim_is_logging_p, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "set_logging", cSim_set_logging, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cSim, "get_parameter", cSim_get_parameter, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "has_parameter", cSim_has_parameter, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "set_parameter", cSim_set_parameter, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "each_parameter", cSim_each_parameter, MRB_ARGS_BLOCK());

  mrb_define_method(mrb, cSim, "get_potential", cSim_get_potential, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cSim, "add_potential", cSim_add_potential, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cSim, "each_potential", cSim_each_potential, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cSim, "to_vectorlist", cSim_to_vectorlist, MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cSim, "load_yaml_string", cSim_load_yaml_string, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cSim, "debug_break", cSim_debug_break, MRB_ARGS_NONE());

  mrb_define_method(mrb, cSim, "all_bonds", cSim_all_bonds, MRB_ARGS_NONE());
  mrb_define_method(mrb, cSim, "all_rigid_bonds", cSim_all_rigid_bonds, MRB_ARGS_NONE());

  initialize_bond_list_terms(mrb, &(tsmrb_classes(mrb)->bond_list_terms));

  tsmrb_define_particle(mrb, mod, cSim);
  tsmrb_define_rigid_body(mrb, mod, cSim);

  return cSim;
}


