#include <ts_io.h>

#include <errno.h>
#include <stdio.h>

#include <mruby.h>
#include <mruby/variable.h>
#include <mruby/irep.h>
#include <mruby/data.h>
#include <mruby/string.h>
#include <mruby/value.h>

#include <numarray_mruby.h>

#include "tsmrb_io.h"
#include "tsmrb_helpers.h"
#include "tsmrb_sim.h"
#include "tsmrb_veclist.h"

static void tsmf_free(mrb_state *mrb, void *p)
{
  tsmf_close(p);
}

static const mrb_data_type tsmf_type = {
  .struct_name = "TSMF",
  .dfree = tsmf_free,
};

static struct tsmf_info *to_tsmf(mrb_state* mrb, mrb_value obj)
{
  struct RClass *cTSMF = tsmrb_classes(mrb)->cTSMF;

  if (!mrb_obj_is_instance_of(mrb, obj, cTSMF))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a TSMF object");

  return DATA_PTR(obj);
}

static mrb_value cTSMF_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_value io,r_sim;
  FILE* f;
  ts_sim *sim;
  struct tsmf_info *tsmf;
  TSResult err;

  r_sim = mrb_nil_value();
  io = mrb_nil_value();
  mrb_get_args(mrb, "oo", &r_sim, &io);

  sim = tsmrb_to_sim(mrb, r_sim);
  f = mrb_io_file(mrb, io);

  err = TS_OK;
  tsmf = tsmf_new(sim, f, TSMF_READWRITE, &err);
  tsmrb_handle_error(mrb, err);

  DATA_TYPE(self) = &tsmf_type;
  DATA_PTR(self) = tsmf;

  return mrb_nil_value();
}

static mrb_value cTSMF_index(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_index *index;
  mrb_value r_index;
  size_t i,n;

  tsmf = to_tsmf(mrb,self);
  index = tsmf->indices;
  if (index == NULL)
  {
    return mrb_nil_value();
  }

  r_index = mrb_ary_new_capa(mrb, index->len);
  for (i=0,n=index->len; i < n; i++)
  {
    int ai;
    mrb_value tup[3], entry;

    ai = mrb_gc_arena_save(mrb);
    tup[0] = mrb_fixnum_value(index->entries[i].type);
    tup[1] = mrb_fixnum_value(index->entries[i].seqno);
    tup[2] = mrb_fixnum_value(index->entries[i].soff);

    entry = mrb_ary_new_from_values(mrb, 3, tup);
    mrb_ary_push(mrb, r_index, entry);

    mrb_gc_arena_restore(mrb,ai);
  }

  return r_index;
}

static mrb_value cTSMF_read_metadata(mrb_state *mrb, mrb_value self);

static mrb_value cTSMF_write_header(mrb_state *mrb, mrb_value self)
{
  return mrb_nil_value();
}

static void encode_metadata(mrb_state *mrb,
    struct tsmf_metadata_info *md, int kcode,
    const mrb_value *items, size_t n)
{
  size_t i;
  enum mrb_vtype vt;

  if (n==0)
  {
    return;
  }

  vt = mrb_type(items[0]);

  /* ensure all values are the same type */
  for (i=1; i < n; i++)
  {
    if (vt != mrb_type(items[i]))
    {
      tsmrb_raisef(mrb, E_ARGUMENT_ERROR,
          "error: kcode %d has hetergeneous values", kcode);
    }
  }

  switch (vt)
  {
    case MRB_TT_FIXNUM:
      {
        int64_t buf[64], *vv = &buf[0];

        if (n > ARRAYLEN(buf))
        {
          vv = mrb_malloc(mrb, n * sizeof *vv);
        }

        for (i=0; i < n; i++) { vv[i] = mrb_fixnum(items[i]); }
        tsmf_metadata_add_short(md, kcode, TSMF_METADATA_VALUE_S64, vv, n);

        if (vv != &buf[0])
        {
          mrb_free(mrb,vv);
        }
      }
      return;

    case MRB_TT_FLOAT:
      {
        double buf[64], *vv = &buf[0];

        if (n > ARRAYLEN(buf))
        {
          vv = mrb_malloc(mrb, n * sizeof *vv);
        }

        for (i=0; i < n; i++) { vv[i] = mrb_float(items[i]); }
        tsmf_metadata_add_short(md, kcode, TSMF_METADATA_VALUE_F64, vv, n);

        if (vv != &buf[0])
        {
          mrb_free(mrb,vv);
        }
      }
      return;


    case MRB_TT_SYMBOL:
    case MRB_TT_STRING:
      {
        mrb_value vs;

        if (n > 1)
        {
          tsmrb_raisef(mrb,
              E_ARGUMENT_ERROR,
              "error: array of string values is unsupported");
        }

        vs = mrb_obj_as_string(mrb, items[0]);

        tsmf_metadata_add_short(md, kcode, TSMF_METADATA_VALUE_STRING,
            RSTRING_PTR(vs), RSTRING_LEN(vs));
      }
      return;

    default:
      {
        struct RClass *cls;
        cls = mrb_class(mrb, items[0]);
        tsmrb_raisef(mrb, E_TYPE_ERROR, "unsupported value type %s",
            mrb_class_name(mrb,cls));
      }
      return;
  }
}

static mrb_value cTSMF_write_metadata(mrb_state *mrb, mrb_value self)
{
  mrb_value entries,keys;
  mrb_int i,nk;
  struct tsmf_info *tsmf;
  struct tsmf_metadata_info md = { 0 };
  struct RClass *eclass = NULL;
  mrb_value errval = mrb_nil_value();
  char errmsg[256] = { 0 };

  entries = mrb_nil_value();
  mrb_get_args(mrb, "H", &entries);
  tsmf = to_tsmf(mrb,self);
  keys = mrb_hash_keys(mrb, entries);

  tsmrb_handle_error(mrb,
      tsmf_metadata_initialize(tsmf, &md));

  for (i=0,nk=RARRAY_LEN(keys); i < nk && !tsmf_metadata_error(&md); i++)
  {
    mrb_value hk,hv;
    int ai;

    ai = mrb_gc_arena_save(mrb);
    hk = RARRAY_PTR(keys)[i];
    hv = mrb_hash_get(mrb, entries, hk);

    switch (mrb_type(hk))
    {
      case MRB_TT_FIXNUM:
        {
          int k = mrb_fixnum(hk);
          switch (k)
          {
            case TSMF_METADATA_COMMENT: 
            case TSMF_METADATA_TIME:    
            case TSMF_METADATA_NSTEPS:  
            case TSMF_METADATA_POT_ENERGY:
            case TSMF_METADATA_KIN_ENERGY:
            case TSMF_METADATA_TOT_ENERGY:
            case TSMF_METADATA_NET_FORCE: 
            case TSMF_METADATA_NET_TORQUE:
              break;
            default:
              eclass = E_RUNTIME_ERROR;
              snprintf(errmsg, sizeof errmsg, "unknown short key 0x%02x", k);
              goto error;
          }

          if (mrb_type(hv) == MRB_TT_ARRAY)
          {
            encode_metadata(mrb, &md, k, RARRAY_PTR(hv), RARRAY_LEN(hv));
          }
          else
          {
            encode_metadata(mrb, &md, k, &hv, 1);
          }
        }
        break;

      case MRB_TT_SYMBOL:
        hk = mrb_str_to_str(mrb,hk);
        /* fallthrough */

      case MRB_TT_STRING:
        {
          mrb_value v = hv;
          switch (mrb_type(hv))
          {
            case MRB_TT_SYMBOL:
              v = mrb_str_to_str(mrb, hv);
              /* fallthrough */

            case MRB_TT_STRING:
              tsmf_metadata_add_kv(&md,
                  (const unsigned char *)RSTRING_PTR(hk), RSTRING_LEN(hk),
                  (const unsigned char *)RSTRING_PTR(hv), RSTRING_LEN(hv));
              break;

            default:
              /* value type error */
              errval = hv;
              goto type_error;
          }
        }
        break;

      default:
        {
          /* type error */
          errval = hk;
          goto type_error;
        }
    }

    mrb_gc_arena_restore(mrb,ai);
  }

  if (!tsmf_metadata_error(&md))
  {
    tsmf_metadata_write(&md);
  }

  if (tsmf_metadata_error(&md))
  {
    const struct ts_error_state *estate;
    estate = ts_error_get_state(md.err);
    eclass = E_RUNTIME_ERROR;
    snprintf(errmsg, sizeof errmsg,
        "error 0x%04x encountered: %s",
        estate->error_code, estate->error_message);
    goto error;
  }

  return mrb_nil_value();

type_error:
  {
    mrb_value s = mrb_inspect(mrb, errval);

    eclass = E_TYPE_ERROR;
    snprintf(errmsg,sizeof errmsg,
        "invalid type: %s", mrb_string_value_cstr(mrb,&s));
  }
  goto error;

error:
  if (md.tsmf)
  {
    tsmf_metadata_finalize(&md);
  }

  assert(eclass != NULL);
  mrb_raise(mrb, eclass, errmsg);
  return mrb_nil_value();
}

static mrb_value cTSMF_read_metadata(mrb_state *mrb, mrb_value self)
{
  mrb_value r_md;
  struct tsmf_metadata_info md;
  struct tsmf_info *tsmf;
  size_t i,n;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb,
      tsmf_metadata_initialize(tsmf,&md));

  tsmrb_handle_error(mrb, tsmf_metadata_read(&md));
  r_md = mrb_hash_new(mrb);

  for (i=0,n=md.slen; i < n; i++)
  {
    int ai;
    mrb_value k,v;
    uint32_t vcount;
    uint8_t vtype;
    struct tsmf_metadata_short_entry *ent;

    ai = mrb_gc_arena_save(mrb);

    ent = md.sdata[i];
    k = mrb_fixnum_value(md.sdata[i]->key_code);
    vtype = md.sdata[i]->vtype;
    vcount = md.sdata[i]->vcount;
    switch (vtype)
    {
      case TSMF_METADATA_VALUE_EMPTY:
        v = mrb_nil_value();
        break;

      case TSMF_METADATA_VALUE_STRING:
      case TSMF_METADATA_VALUE_BYTES:
        v = mrb_str_new(mrb, (const char *)md.sdata[i]->vdata, vcount);
        break;

        // For fixed-width types, the count is implicit in the length
      case TSMF_METADATA_VALUE_S16:
      case TSMF_METADATA_VALUE_U16:
        {
          // XXX - portable for signed?
          uint16_t *dv = (uint16_t *)md.sdata[i]->vdata;
          if (vcount == 1)
          {
            v = mrb_fixnum_value(*dv);
          }
          else
          {
            uint32_t i;
            v = mrb_ary_new_capa(mrb,vcount);
            for (i=0; i < vcount; i++)
            {
              mrb_ary_push(mrb, v, mrb_fixnum_value(dv[i]));
            }
          }
        }
        break;

      case TSMF_METADATA_VALUE_S32:
      case TSMF_METADATA_VALUE_U32:
        {
          // XXX - check if this fits within a Fixnum
          uint32_t *dv = (uint32_t *)md.sdata[i]->vdata;
          if (vcount == 1)
          {
            v = mrb_fixnum_value(*dv);
          }
          else
          {
            uint32_t i;
            v = mrb_ary_new_capa(mrb,vcount);
            for (i=0; i < vcount; i++)
            {
              mrb_ary_push(mrb, v, mrb_fixnum_value(dv[i]));
            }
          }
        }
        break;

      case TSMF_METADATA_VALUE_S64:
      case TSMF_METADATA_VALUE_U64:
        {
          // XXX - check if this fits within a Fixnum
          uint64_t *dv = (uint64_t *)md.sdata[i]->vdata;
          if (vcount == 1)
          {
            v = mrb_fixnum_value(*dv);
          }
          else
          {
            uint64_t i;
            v = mrb_ary_new_capa(mrb,vcount);
            for (i=0; i < vcount; i++)
            {
              mrb_ary_push(mrb, v, mrb_fixnum_value(dv[i]));
            }
          }
        }
        break;

      case TSMF_METADATA_VALUE_F32:
        {
          // XXX - check if this fits within a Fixnum
          float *dv = (float *)md.sdata[i]->vdata;

          if (vcount == 1)
          {
            v = mrb_float_value(mrb,*dv);
          }
          else
          {
            int ai;
            uint32_t i;
            v = mrb_ary_new_capa(mrb,vcount);
            ai = mrb_gc_arena_save(mrb);
            for (i=0; i < vcount; i++)
            {
              mrb_ary_push(mrb, v, mrb_float_value(mrb,dv[i]));
              if (i % 32 == 0)
              {
                mrb_gc_arena_restore(mrb,ai);
                ai = mrb_gc_arena_save(mrb);
              }
            }
            mrb_gc_arena_restore(mrb,ai);
          }
        }
        break;

      case TSMF_METADATA_VALUE_F64:
        {
          // XXX - check if this fits within a Fixnum
          double *dv = (double *)md.sdata[i]->vdata;

          if (vcount == 1)
          {
            v = mrb_float_value(mrb,*dv);
          }
          else
          {
            int ai;
            uint32_t i;
            v = mrb_ary_new_capa(mrb,vcount);
            ai = mrb_gc_arena_save(mrb);
            for (i=0; i < vcount; i++)
            {
              mrb_ary_push(mrb, v, mrb_float_value(mrb,dv[i]));
              if (i % 32 == 0)
              {
                mrb_gc_arena_restore(mrb,ai);
                ai = mrb_gc_arena_save(mrb);
              }
            }
            mrb_gc_arena_restore(mrb,ai);
          }
        }
        break;

      default:
        v = mrb_nil_value();

        tsmrb_raisef(mrb, E_TYPE_ERROR,
            "unknown value type 0x%02x", vtype);
        break;
    }
    mrb_hash_set(mrb, r_md, k, v);

    mrb_gc_arena_restore(mrb,ai);
  }

  for (i=0,n=md.kvlen; i < n; i++)
  {
    int ai;
    uint32_t klen,vlen;
    const char *p;
    mrb_value k,v;

    ai = mrb_gc_arena_save(mrb);

    klen = md.kvdata[i]->klen;
    vlen = md.kvdata[i]->vlen;
    p = (const char *)(&md.kvdata[i]->kvdata[0]);

    k = mrb_str_new(mrb,&p[0], klen);
    v = mrb_str_new(mrb,&p[klen], vlen);

    mrb_hash_set(mrb, r_md, k,v);

    mrb_gc_arena_restore(mrb,ai);
  }

  tsmf_metadata_finalize(&md);

  return r_md;
}

static mrb_value cTSMF_write_coords(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  mrb_sym opt = 0;
  int write_dof;

  mrb_get_args(mrb, "|n", &opt);

  tsmf = to_tsmf(mrb,self);

  if (opt == 0)
  {
    write_dof = 0;
  }
  else if (opt == mrb_intern_lit(mrb, "all"))
  {
    write_dof = 0;
  }
  else if (opt == mrb_intern_lit(mrb, "dof"))
  {
    write_dof = 1;
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid option: must be either :all or :dof");
    write_dof = 0;
  }

  if (write_dof)
  {
    tsmrb_handle_error(mrb, tsmf_coords_write_sim(tsmf));
  }
  else
  {
    tsmrb_handle_error(mrb, tsmf_coords_write_sim_all(tsmf));
  }

  return mrb_nil_value();
}

static mrb_value cTSMF_read_coords_into_sim(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;

  tsmf = to_tsmf(mrb,self);

  if (tsmf->sim == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "TSMF has no associated simulation: cannot use read_coords_into_sim");
  }

  tsmrb_handle_error(mrb, tsmf_coords_read_into_sim(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_read_coords(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_coords_buffer cbuf = {0};
  TSResult err;
  mrb_value ret = mrb_nil_value();
  mrb_value r_arr = mrb_nil_value();
  uint32_t ndof;
  int i,dim;
  numarray *arr = NULL;

  tsmf = to_tsmf(mrb,self);
  err = tsmf_coords_read(tsmf, &cbuf);
  if (IS_ERROR(err))
  {
    goto cleanup;
  }

  ndof = tsmf_coords_buffer_ndof(&cbuf);
  if (ndof > INT_MAX)
  {
    err = TS_NEW_ERROR_MESG(TS_ERR_UNSUPPORTED_OPERATION,
        "number of DoFs is larger than INT_MAX");
    goto cleanup;
  }

  dim = ndof;
  r_arr = numarray_new_uninitialized(mrb, 1, &dim, &arr);
  assert(arr != NULL);
  assert(arr->nelts == ndof);

  {
    const double *restrict src = cbuf.coords;
    double *restrict dst = arr->elts;
    for (i=0; i < dim; i++)
    {
      dst[i] = src[i];
    }
  }

  ret = mrb_hash_new(mrb);
  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_cstr(mrb,"flags")),
      mrb_fixnum_value(cbuf.flags));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_cstr(mrb,"nparticles")),
      mrb_fixnum_value(cbuf.nparticle));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_cstr(mrb,"nrigids")),
      mrb_fixnum_value(cbuf.nrigid));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_cstr(mrb,"nrotate")),
      mrb_fixnum_value(cbuf.nrotate));

  mrb_hash_set(mrb, ret, 
      mrb_symbol_value(mrb_intern_cstr(mrb,"coords")),
      r_arr);

cleanup:
  tsmf_coords_buffer_release(&cbuf);
  tsmrb_handle_error(mrb, err);

  return ret;
}

static mrb_value cTSMF_write_particle_types(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb, tsmf_ptypes_write(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_read_particle_types(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  mrb_value ret;
  struct ts_particle_type *ptypes = NULL;
  uint32_t i,n = 0;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb, tsmf_ptypes_read(tsmf, &ptypes, &n));

  ret = mrb_ary_new_capa(mrb,n);
  for (i=0; i < n; i++)
  {
    int ai;
    mrb_value elt;

    ai = mrb_gc_arena_save(mrb);
    elt = tsmrb_sim_ptype_to_hash(mrb, &ptypes[i]);
    mrb_ary_push(mrb, ret, elt);
    mrb_gc_arena_restore(mrb,ai);
  }

  tsmf_ptypes_free(ptypes, n);

  return ret;
}

static mrb_value cTSMF_write_rigid_body_types(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb, tsmf_rbtypes_write(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_read_rigid_body_types(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_rbtypes_buffer rbbuf = {0};
  mrb_value ret;
  uint32_t i,n;
  TSResult err;
  enum { OMIT_DERIVED_INFO = 0 };

  tsmf = to_tsmf(mrb,self);
  err = tsmf_rbtypes_read(tsmf, &rbbuf);

  n = rbbuf.tlen;
  ret = mrb_ary_new_capa(mrb, n);
  for (i=0; i < n; i++)
  {
    mrb_value elt;
    int ai;

    ai = mrb_gc_arena_save(mrb);
    elt = tsmrb_rigid_body_type_info(mrb, &rbbuf.rbtypes[i], OMIT_DERIVED_INFO);
    mrb_ary_push(mrb, ret, elt);
    mrb_gc_arena_restore(mrb,ai);
  }

  tsmf_btypes_buffer_finalize(&rbbuf);
  tsmrb_handle_error(mrb, err);

  return ret;
}

static mrb_value cTSMF_write_particle_list(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb, tsmf_plist_write(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_read_particle_list(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_plist_buf *plist;
  uint32_t i;
  mrb_value ret;

  tsmf = to_tsmf(mrb,self);

  plist = NULL;
  tsmrb_handle_error(mrb,
      tsmf_plist_read(tsmf, &plist));

  ret = mrb_ary_new_capa(mrb, plist->n);
  for(i=0; i < plist->n; i++)
  {
    mrb_value pair[2];
    pair[0] = mrb_fixnum_value(plist->entries[i].type);
    pair[1] = mrb_fixnum_value(plist->entries[i].flags);
    mrb_ary_push(mrb, ret,
        mrb_ary_new_from_values(mrb, 2, pair));
  }

  return ret;
}

static mrb_value cTSMF_write_rigid_body_list(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;

  tsmf = to_tsmf(mrb,self);
  tsmrb_handle_error(mrb, tsmf_rblist_write(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_read_rigid_body_list(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  ts_sim *sim;
  struct tsmf_rblist_buf *rblist;
  uint32_t i,n;
  mrb_value ret;

  tsmf = to_tsmf(mrb,self);
  if (sim = tsmf->sim, sim == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "TSMF: simulation required");
  }

  rblist = NULL;
  tsmrb_handle_error(mrb, tsmf_rblist_read(tsmf, &rblist));

  n = rblist->nrigid;
  ret = mrb_ary_new_capa(mrb, n);
  for(i=0; i < n; i++)
  {
    struct ts_rigid_body_type *rbtype;
    mrb_value tup[3];
    uint32_t type, pind0, j,np;
    int ai;

    ai = mrb_gc_arena_save(mrb);

    type = rblist->entries[i].type;
    rbtype = ts_get_rigid_body_type(sim, type);
    tup[0] = mrb_fixnum_value(type);
    tup[1] = mrb_fixnum_value(rblist->entries[i].flags);

    np = rbtype->num_sites+1;
    pind0 = rblist->entries[i].pind0;
    tup[2] = mrb_ary_new_capa(mrb, np);

    for(j=0; j < np; j++)
    {
      int ai1;

      ai1 = mrb_gc_arena_save(mrb);
      mrb_ary_push(mrb, tup[2], mrb_fixnum_value(rblist->plist[pind0+j]));
      mrb_gc_arena_restore(mrb,ai1);
    }

    mrb_ary_push(mrb, ret, mrb_ary_new_from_values(mrb, 3, tup));

    mrb_gc_arena_restore(mrb,ai);
  }
  
  tsmf_rblist_buf_delete(rblist);
  return ret;
}

static mrb_value cTSMF_write_connectivity(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  mrb_sym opt;

  tsmf = to_tsmf(mrb,self);
  opt = 0;
  mrb_get_args(mrb, "|n", &opt);

  if (opt == mrb_intern_lit(mrb, "delta"))
  {
    tsmrb_handle_error(mrb, tsmf_conn_write_delta(tsmf));
  }
  else
  {
    tsmrb_handle_error(mrb, tsmf_conn_write(tsmf));
  }

  return mrb_nil_value();
}

static mrb_value cTSMF_read_connectivity(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_conn_list *clist;
  mrb_value bonds,ret;
  uint32_t i,nb;

  tsmf = to_tsmf(mrb,self);
  clist = NULL;
  tsmrb_handle_error(mrb, tsmf_conn_read(tsmf, &clist));

  nb = clist->nbonds;
  bonds = mrb_ary_new_capa(mrb, nb);
  for (i=0; i < nb; i++)
  {
    mrb_value tuple[3];
    int ai;

    ai = mrb_gc_arena_save(mrb);

    tuple[0] = mrb_fixnum_value(clist->bonds[i].pi);
    tuple[1] = mrb_fixnum_value(clist->bonds[i].pj);
    tuple[2] = mrb_fixnum_value(clist->bonds[i].bt);

    mrb_ary_push(mrb, bonds, mrb_ary_new_from_values(mrb, 3, tuple));

    mrb_gc_arena_restore(mrb,ai);
  }

  ret = mrb_hash_new(mrb);

  mrb_hash_set(mrb, ret,
      mrb_symbol_value(mrb_intern_lit(mrb,"flags")),
      mrb_fixnum_value(clist->flags));

  mrb_hash_set(mrb, ret, mrb_symbol_value(mrb_intern_lit(mrb,"bonds")), bonds);

  ts_free(NO_ALLOCATOR, clist);
  return ret;
}

static mrb_value cTSMF_read_header(mrb_state *mrb, mrb_value klass)
{
  struct tsmf_file_header hdr;
  mrb_value io, r_hdr, r_vers;
  FILE* f;

  io = mrb_nil_value();
  mrb_get_args(mrb, "o", &io);
  f = mrb_io_file(mrb, io);

  tsmrb_handle_error(mrb,
      tsmf_read_header(f, &hdr));

  r_hdr = mrb_hash_new(mrb);
  mrb_hash_set(mrb, r_hdr,
      mrb_symbol_value(mrb_intern_cstr(mrb,"magic")),
      mrb_str_new(mrb, hdr.magic, sizeof hdr.magic));

  r_vers = mrb_ary_new(mrb);
  mrb_ary_push(mrb, r_vers, mrb_fixnum_value(hdr.version[0]));
  mrb_ary_push(mrb, r_vers, mrb_fixnum_value(hdr.version[1]));
  mrb_hash_set(mrb, r_hdr,
      mrb_symbol_value(mrb_intern_cstr(mrb,"version")),
      r_vers);

  mrb_hash_set(mrb, r_hdr,
      mrb_symbol_value(mrb_intern_cstr(mrb,"flags")),
      mrb_fixnum_value(hdr.flags));

  mrb_hash_set(mrb, r_hdr,
      mrb_symbol_value(mrb_intern_cstr(mrb,"ioff")),
      mrb_fixnum_value(hdr.ioff));

  mrb_hash_set(mrb, r_hdr,
      mrb_symbol_value(mrb_intern_cstr(mrb,"padding")),
      mrb_str_new(mrb, (char*)hdr.padding, sizeof hdr.padding));

  return r_hdr;
}

static mrb_value cTSMF_close(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  tsmf = to_tsmf(mrb,self);

  tsmrb_handle_error(mrb,
      tsmf_close(tsmf));

  return mrb_nil_value();
}

static mrb_value cTSMF_get_seqno(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  tsmf = to_tsmf(mrb,self);

  return mrb_fixnum_value(tsmf->seqno);
}

static mrb_value cTSMF_set_seqno(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  tsmf = to_tsmf(mrb,self);
  mrb_int v;

  v = 0;
  mrb_get_args(mrb, "i", &v);

  // XXX - check for possible overflow
  if (v < 0)
  {
    mrb_raise(mrb, E_RANGE_ERROR, "seqno must not be negative");
  }

  if (v < tsmf->seqno)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "seqno must monotonically increase");
  }

  tsmf->seqno = v;
  return mrb_nil_value();
}

static mrb_value cTSMF_read_section(mrb_state *mrb, mrb_value self)
{
  struct tsmf_info *tsmf;
  struct tsmf_index *index;
  struct tsmf_index_entry *entry;
  mrb_int i,nrest;
  mrb_value *rest;

  tsmf = to_tsmf(mrb,self);
  i = -1;
  nrest=0;
  rest=NULL;
  mrb_get_args(mrb, "i*", &i, &rest, &nrest);

  index = tsmf->indices;
  if (index == NULL)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "tsmf file has no index");
  }

  if (i < 1 || i > index->len)
  {
    tsmrb_raisef(mrb, E_INDEX_ERROR, "invalid index %d", (int)i);
  }

  entry = &index->entries[i-1];

  if (fseek(tsmf->f, entry->soff, SEEK_SET) != 0)
  {
    tsmrb_raisef(mrb, E_RUNTIME_ERROR,
        "error seeking to section (%d): %s", errno, strerror(errno));
  }

  switch (entry->type)
  {
    case TSMF_SEC_METADATA:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_metadata"), nrest, rest);

    case TSMF_SEC_COORDINATES:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_coords"), nrest, rest);

    case TSMF_SEC_PARTICLE_TYPES:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_particle_types"), nrest, rest);

    case TSMF_SEC_RB_TYPES:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_rigid_body_types"), nrest, rest);

    case TSMF_SEC_PARTICLE_LIST:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_particle_list"), nrest, rest);

    case TSMF_SEC_RB_LIST:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_rigid_body_list"), nrest, rest);

    case TSMF_SEC_CONNECTIVITY:
      return mrb_funcall_argv(mrb, self,
          mrb_intern_cstr(mrb,"read_connectivity"), nrest, rest);

    case TSMF_SEC_INDEX:
    default:
      tsmrb_raisef(mrb, E_RUNTIME_ERROR,
          "reading section type %d is currently unsupported",
          entry->type);
  }

  return mrb_nil_value();
}

extern const uint8_t tsmrb_tsmf_irep[];

struct RClass *tsmrb_define_tsmf(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cTSMF;
  int ai;

  cTSMF = mrb_define_class_under(mrb, mod, "TSMF", mrb->object_class);
  tsmrb_classes(mrb)->cTSMF = cTSMF;

  mrb_define_singleton_method(mrb, (struct RObject *)cTSMF, "read_header",
    cTSMF_read_header, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cTSMF, "initialize",
      cTSMF_initialize, MRB_ARGS_REQ(1) | MRB_ARGS_BLOCK());

  mrb_define_method(mrb, cTSMF, "close", cTSMF_close, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "seqno", cTSMF_get_seqno, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "seqno=", cTSMF_set_seqno, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cTSMF, "index", cTSMF_index, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "read_section",
      cTSMF_read_section, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cTSMF, "write_header", cTSMF_write_header, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cTSMF, "read_metadata", cTSMF_read_metadata, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "write_metadata", cTSMF_write_metadata, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cTSMF, "write_coords", cTSMF_write_coords, MRB_ARGS_OPT(1));
  mrb_define_method(mrb, cTSMF, "read_coords", cTSMF_read_coords, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "read_coords_into_sim", cTSMF_read_coords_into_sim, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "write_particle_types",
      cTSMF_write_particle_types, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "read_particle_types",
      cTSMF_read_particle_types, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "write_rigid_body_types",
      cTSMF_write_rigid_body_types, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "read_rigid_body_types",
      cTSMF_read_rigid_body_types, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "write_particle_list",
      cTSMF_write_particle_list, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "read_particle_list",
      cTSMF_read_particle_list, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "write_rigid_body_list",
      cTSMF_write_rigid_body_list, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "read_rigid_body_list",
      cTSMF_read_rigid_body_list, MRB_ARGS_NONE());

  mrb_define_method(mrb, cTSMF, "write_connectivity",
      cTSMF_write_connectivity, MRB_ARGS_NONE());
  mrb_define_method(mrb, cTSMF, "read_connectivity",
      cTSMF_read_connectivity, MRB_ARGS_NONE());

  mrb_define_const(mrb, cTSMF, "FLAGS_NONE",
      mrb_fixnum_value(TSMF_FLAG_NONE));

  mrb_define_const(mrb, cTSMF, "COORDS_FLAG_ALL",
      mrb_fixnum_value(TSMF_COORDS_FLAG_ALL));

  mrb_define_const(mrb, cTSMF, "CONN_FLAG_DELTA",
      mrb_fixnum_value(TSMF_CONN_FLAGS_DELTA));

  static struct { int sec; const char *name; } sections[] = {
    {TSMF_SEC_INDEX,            "INDEX"                 },
    {TSMF_SEC_METADATA,         "METADATA"              },
    {TSMF_SEC_PARTICLE_TYPES,   "PARTICLE_TYPES"        },
    {TSMF_SEC_RB_TYPES,         "RB_TYPES"              },
    {TSMF_SEC_PARTICLE_LIST,    "PARTICLE_LIST"         },
    {TSMF_SEC_RB_LIST,          "RB_LIST"               },
    {TSMF_SEC_CONNECTIVITY,     "CONNECTIVITY"          },
    {TSMF_SEC_COORDINATES,      "COORDINATES"           },
  };

  size_t i;
  for(i=0; i < ARRAYLEN(sections); i++)
  {
    mrb_define_const(mrb, cTSMF, sections[i].name,
        mrb_fixnum_value(sections[i].sec));
  }

#define METADATA_KEY(name) \
  mrb_define_const(mrb, cTSMF, SYMSTR(name), \
      mrb_fixnum_value(SYMCAT(TSMF_METADATA_,name)))

  METADATA_KEY(COMMENT);
  METADATA_KEY(TIME);
  METADATA_KEY(NSTEPS);
  METADATA_KEY(POT_ENERGY);
  METADATA_KEY(KIN_ENERGY);
  METADATA_KEY(TOT_ENERGY);
  METADATA_KEY(NET_FORCE );
  METADATA_KEY(NET_TORQUE);

#undef METADATA_KEY

  ai = mrb_gc_arena_save(mrb);
  mrb_load_irep(mrb, tsmrb_tsmf_irep);
  if (mrb->exc) {
    mrb_p(mrb, mrb_obj_value(mrb->exc));
    exit(0);
  }
  mrb_gc_arena_restore(mrb, ai);

  return cTSMF;
}
