#include "tsmrb_vec.h"
#include "tsmrb_helpers.h"

#include "mruby/data.h"
#include "mruby/array.h"
#include "mruby/value.h"
#include "mruby/numeric.h"

#include "tsmrb_mat3x3.h"

static const mrb_data_type vec_type = {
  .struct_name = "Vec3",
  .dfree = tsmrb_free,
};

static inline vector3* vec3_alloc(mrb_state* mrb)
{
  vector3* v = mrb_malloc(mrb, sizeof(*v));
  memset(v,0,sizeof(*v));
  return v;
}

static mrb_value vec3_wrap(mrb_state* mrb, vector3* v)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_vec3_class(mrb), v, &vec_type);
  return mrb_obj_value(data);
}

static mrb_value create_new_vec3(mrb_state* mrb, vector3** vp)
{
  *vp = vec3_alloc(mrb);
  return vec3_wrap(mrb, *vp);
}

mrb_value tsmrb_new_vec3(mrb_state* mrb, vector3** vp) { return create_new_vec3(mrb,vp); }

vector3* tsmrb_to_vec3(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_vec3_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a Vec3 object");

  return DATA_PTR(obj);
}

mrb_value tsmrb_from_vec3(mrb_state* mrb, const vector3*restrict v)
{
  vector3* vnew = NULL;
  mrb_value rbvec = create_new_vec3(mrb, &vnew);
  vector3_set(vnew, v);

  return rbvec;
}

static mrb_value
new_vec3(mrb_state* mrb, mrb_value mod)
{
  mrb_value* rest = NULL;
  int nrest = 0;
  mrb_get_args(mrb, "*", &rest, &nrest);

  return mrb_funcall_argv(mrb,
      mrb_obj_value(tsmrb_vec3_class(mrb)),
      mrb_intern_cstr(mrb, "new"),
      nrest,
      rest);
}

mrb_value tsmrb_vec3_to_ary(mrb_state* mrb, vector3* v)
{
  mrb_value elts[3];
  elts[0] = mrb_float_value(mrb, (mrb_float)VECTOR3_X(v));
  elts[1] = mrb_float_value(mrb, (mrb_float)VECTOR3_Y(v));
  elts[2] = mrb_float_value(mrb, (mrb_float)VECTOR3_Z(v));

  return mrb_ary_new_from_values(mrb, 3, elts);
}

void obj_to_vec(mrb_state* mrb, mrb_value rvec, vector3* vec)
{
  enum mrb_vtype tt = mrb_type(rvec);
  if (tt == MRB_TT_ARRAY)
  {
    if (RARRAY_LEN(rvec) != 3)
      mrb_raisef(mrb, E_RANGE_ERROR,
          "expected length of 3, found length of %d", RARRAY_LEN(rvec));

    VECTOR3_X(vec) = mrb_to_flo(mrb, RARRAY_PTR(rvec)[0]);
    VECTOR3_Y(vec) = mrb_to_flo(mrb, RARRAY_PTR(rvec)[1]);
    VECTOR3_Z(vec) = mrb_to_flo(mrb, RARRAY_PTR(rvec)[2]);

    return;
  }
  else if ((tt == MRB_TT_DATA) || (tt == MRB_TT_OBJECT))
  {
    if (mrb_obj_is_kind_of(mrb, rvec, tsmrb_vec3_class(mrb)))
    {
      vector3* v1 = DATA_PTR(rvec);
      vector3_set(vec, v1);

      return;
    }
  }

  mrb_raisef(mrb, E_TYPE_ERROR, 
      "argument type %s could not be coerced into Vec3",
      mrb_obj_classname(mrb, rvec));
}

static mrb_value cVec3_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_value *argv;
  int argc;
  mrb_get_args(mrb, "*", &argv, &argc);

  vector3* v = vec3_alloc(mrb);
  vector3_zero(v);

  DATA_TYPE(self) = &vec_type;
  DATA_PTR(self) = v;

  if (argc > 0)
  {
    if (argc == 1)
    {
      obj_to_vec(mrb,argv[0],v);
    }
    else if (argc == 3)
    {
      vector3_set_with(v, 
          mrb_to_flo(mrb, argv[0]),
          mrb_to_flo(mrb, argv[1]),
          mrb_to_flo(mrb, argv[2]));
    }
    else
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "Vec3.initialize requires either 0, 1, or 3 arguments");
    }
  }

  return self;
}

static mrb_value cVec3_get_x(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  return mrb_float_value(mrb, VECTOR3_X(vec));
}

static mrb_value cVec3_get_y(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  return mrb_float_value(mrb, VECTOR3_Y(vec));
}

static mrb_value cVec3_get_z(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  return mrb_float_value(mrb, VECTOR3_Z(vec));
}

static mrb_value cVec3_get_w(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  return mrb_float_value(mrb, VECTOR3_W(vec));
}



static mrb_value cVec3_set_x(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  mrb_float val = (mrb_float)0.0;
  mrb_get_args(mrb, "f", &val);
  VECTOR3_X(vec) = (VecReal)val;
  return self;
}

static mrb_value cVec3_set_y(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  mrb_float val = (mrb_float)0.0;
  mrb_get_args(mrb, "f", &val);
  VECTOR3_Y(vec) = (VecReal)val;
  return self;
}

static mrb_value cVec3_set_z(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  mrb_float val = (mrb_float)0.0;
  mrb_get_args(mrb, "f", &val);
  VECTOR3_Z(vec) = (VecReal)val;
  return self;
}

static mrb_value cVec3_set_w(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  mrb_float val = (mrb_float)0.0;
  mrb_get_args(mrb, "f", &val);
  VECTOR3_X(vec) = (VecReal)val;
  return self;
}

static mrb_value cVec3_to_a(mrb_state* mrb, mrb_value self)
{
  vector3* vec = DATA_PTR(self);
  return tsmrb_vec3_to_ary(mrb, vec);
}

static mrb_value cVec3_inspect(mrb_state* mrb, mrb_value self)
{
  char buf[64];
  vector3* v = DATA_PTR(self);

  snprintf(buf,sizeof(buf), "Vec3 [ %7.4f %7.4f %7.4f | %7.4f ] @ %p",
      vector3_x(v),
      vector3_y(v),
      vector3_z(v),
      vector3_w(v),
      (void*)v);

  return mrb_str_new_cstr(mrb, buf);
}

static inline int normalize_ruby_ind(mrb_state* mrb, int i)
{
  if (i<0) { i += 4; }

  if ((i < 0) || (i >= 4))
    mrb_raise(mrb, E_RANGE_ERROR, "index out of bounds (0-3)");

  return i;
}

static mrb_value cVec3_get_index(mrb_state* mrb, mrb_value self)
{
  vector3* v = DATA_PTR(self);
  mrb_int i = 0;
  mrb_get_args(mrb,"i",&i);
  i = normalize_ruby_ind(mrb,i);

  return mrb_float_value(mrb, (mrb_float)VECTOR3_ELT(v,i));
}

static mrb_value cVec3_set_index(mrb_state* mrb, mrb_value self)
{
  vector3* v = DATA_PTR(self);
  mrb_int i = 0;
  mrb_float f = 0.0;
  mrb_get_args(mrb,"if",&i,&f);
  i = normalize_ruby_ind(mrb,i);

  VECTOR3_ELT(v, i) = (VecReal)f;

  return self;
}

static mrb_value cVec3_negate(mrb_state* mrb, mrb_value self)
{
  vector3* v = DATA_PTR(self);

  vector3* vnew;
  mrb_value rvnew = create_new_vec3(mrb,&vnew);
  vector3_set_negative(vnew, v);
  return rvnew;
}

static mrb_value cVec3_negateB(mrb_state* mrb, mrb_value self)
{
  vector3* v = DATA_PTR(self);
  vector3_negate(v);
  return self;
}

static mrb_value cVec3_add(mrb_state* mrb, mrb_value ra)
{
  mrb_value rb;
  mrb_get_args(mrb, "o", &rb);

  vector3* a = tsmrb_to_vec3(mrb, ra);
  vector3* b = tsmrb_to_vec3(mrb, rb);

  vector3* c = NULL;
  mrb_value rc = create_new_vec3(mrb, &c);
  vector3_add(c,a,b);

  return rc;
}

static mrb_value cVec3_sub(mrb_state* mrb, mrb_value ra)
{
  mrb_value rb;
  mrb_get_args(mrb, "o", &rb);

  vector3* a = tsmrb_to_vec3(mrb,ra);
  vector3* b = tsmrb_to_vec3(mrb,rb);

  vector3* c = NULL;
  mrb_value rc = create_new_vec3(mrb, &c);
  vector3_sub(c,a,b);

  return rc;
}

static mrb_value cVec3_scale_byB(mrb_state* mrb, mrb_value self)
{
  vector3* v = tsmrb_to_vec3(mrb,self);
  double scl = 0.0;
  mrb_get_args(mrb, "f", &scl);
  vector3_scaleby(v,scl);
  return self;
}

static mrb_value cVec3_mul(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);
  vector3* v = tsmrb_to_vec3(mrb,self);

  mrb_value result = mrb_nil_value();

  switch ( mrb_type(other) )
  {
    case MRB_TT_FIXNUM:
    case MRB_TT_FLOAT:
      {
        vector3* vnew = NULL;
        VecReal scl = (VecReal)mrb_to_flo(mrb, other);

        result = create_new_vec3(mrb,&vnew);
        vector3_set_scaled(vnew, scl, v);
      }
      break;

    case MRB_TT_DATA:
    case MRB_TT_OBJECT:
      {
        if (mrb_obj_is_instance_of(mrb, other, tsmrb_mat3x3_class(mrb)))
        {
          mat3x3* mat = tsmrb_to_mat3x3(mrb,other);
          vector3* vnew = NULL;
          result = tsmrb_new_vec3(mrb,&vnew);
          vector3_mat_lmul(vnew, v, mat);
          break;
        }
      }

    default:
      mrb_raise(mrb, E_TYPE_ERROR, "illegal type for right-multiply");
  }

  return result;
}

static mrb_value cVec3_div(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  vector3* v = tsmrb_to_vec3(mrb,self);

  if (!mrb_numeric_p(other))
    mrb_raise(mrb, E_TYPE_ERROR, "illegal type for vector scalar division");

  double scl = mrb_to_flo(mrb,other);

  vector3* vnew = NULL;
  mrb_value result = tsmrb_new_vec3(mrb,&vnew);

  vector3_set_scaled(vnew, 1.0/scl, v);
  return result;
}

static mrb_value cVec3_equals(mrb_state* mrb, mrb_value a)
{
  mrb_value b;
  mrb_get_args(mrb, "o", &b);

  vector3 va;
  vector3 vb;

  obj_to_vec(mrb, a, &va);
  obj_to_vec(mrb, b, &vb);

  return mrb_bool_value(
      (VECTOR3_X(&va) == VECTOR3_X(&vb)) &&
      (VECTOR3_Y(&va) == VECTOR3_Y(&vb)) &&
      (VECTOR3_Z(&va) == VECTOR3_Z(&vb))
      );
}

static mrb_value cVec3_sqnorm(mrb_state* mrb, mrb_value self)
{
  return mrb_float_value(mrb, vector3_sqnorm(tsmrb_to_vec3(mrb,self)) );
}

static mrb_value cVec3_norm(mrb_state* mrb, mrb_value self)
{
  return mrb_float_value(mrb, vector3_norm(tsmrb_to_vec3(mrb,self)) );
}

static mrb_value cVec3_dist(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  vector3* a = tsmrb_to_vec3(mrb,self);
  vector3* b = tsmrb_to_vec3(mrb,other);

  vector3 vtmp;
  double sqdist = vector3_distance(a,b,&vtmp);
  return mrb_float_value(mrb, sqdist);
}

static mrb_value cVec3_sqdist(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb, "o", &other);

  vector3* a = tsmrb_to_vec3(mrb,self);
  vector3* b = tsmrb_to_vec3(mrb,other);

  vector3 vtmp;
  double sqdist = vector3_sqdistance(a,b,&vtmp);
  return mrb_float_value(mrb, sqdist);
}

static mrb_value cVec3_normalize(mrb_state* mrb, mrb_value v)
{
  const vector3* a = tsmrb_to_vec3(mrb,v);
  vector3 an;
  vector3_set(&an, a);
  vector3_normalize(&an);
  return tsmrb_from_vec3(mrb,&an);
}

static mrb_value cVec3_normalizeB(mrb_state* mrb, mrb_value v)
{
  vector3* a = tsmrb_to_vec3(mrb,v);
  vector3_normalize(a);
  return v;
}

static mrb_value cVec3_cross(mrb_state* mrb, mrb_value ra)
{
  mrb_value rb;
  mrb_get_args(mrb,"o",&rb);

  vector3* a = tsmrb_to_vec3(mrb,ra);
  vector3* b = tsmrb_to_vec3(mrb,rb);

  vector3* c = NULL;
  mrb_value rc = create_new_vec3(mrb,&c);
  vector3_cross(c, a,b);

  return rc;
}

static mrb_value cVec3_distanceB(mrb_state* mrb, mrb_value self)
{
  mrb_value other;
  mrb_get_args(mrb,"o",&other);

  vector3* a = tsmrb_to_vec3(mrb,self);
  vector3* b = tsmrb_to_vec3(mrb,other);

  vector3 dr;
  double dist = vector3_distance(a,b,&dr);

  mrb_value vals[2] = { 
    mrb_float_value(mrb, dist),
    tsmrb_from_vec3(mrb,&dr),
  };

  return mrb_ary_new_from_values(mrb, 2, vals);
}

static mrb_value cVec3_inner(mrb_state* mrb, mrb_value ra)
{
  mrb_value rb;
  mrb_get_args(mrb,"o",&rb);

  vector3* a = tsmrb_to_vec3(mrb,ra);
  vector3* b = tsmrb_to_vec3(mrb,rb);

  return mrb_float_value(mrb, (a == b) ? vector3_sqnorm(a) : vector3_dot(a,b) );
}

static mrb_value cVec3_outer(mrb_state* mrb, mrb_value ra)
{
  mrb_value rb = mrb_nil_value();
  mrb_get_args(mrb, "o", &rb);

  vector3* a = tsmrb_to_vec3(mrb,ra);
  vector3* b = tsmrb_to_vec3(mrb,rb);

  mat3x3* prod = NULL;
  mrb_value rprod = tsmrb_new_mat3x3(mrb,&prod);

  vector3_outer(prod, a,b);
  return rprod;
}

struct RClass* tsmrb_define_vec3(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_vec3_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "Vec3 class has already been initialized!");

  struct RClass* vec3_class = mrb_define_class_under(mrb, mod, "Vec3", mrb->object_class);
  MRB_SET_INSTANCE_TT(vec3_class, MRB_TT_DATA);
  tsmrb_classes(mrb)->cVec3 = vec3_class;

  mrb_define_method(mrb, vec3_class, "initialize", cVec3_initialize, MRB_ARGS_REST());

  mrb_define_method(mrb, vec3_class, "==", cVec3_equals, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "x", cVec3_get_x, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "y", cVec3_get_y, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "z", cVec3_get_z, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "w", cVec3_get_w, MRB_ARGS_NONE());

  mrb_define_method(mrb, vec3_class, "x=", cVec3_set_x, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "y=", cVec3_set_y, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "z=", cVec3_set_z, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "w=", cVec3_set_w, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "at", cVec3_get_index, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "[]", cVec3_get_index, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "[]=", cVec3_set_index, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, vec3_class, "-@", cVec3_negate, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "negate!", cVec3_negateB, MRB_ARGS_NONE());

  mrb_define_method(mrb, vec3_class, "+", cVec3_add, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "-", cVec3_sub, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "*", cVec3_mul, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "mul", cVec3_mul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "/", cVec3_div, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "scl", cVec3_div, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "scale_by!", cVec3_scale_byB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "sqnorm", cVec3_sqnorm, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "norm", cVec3_norm, MRB_ARGS_NONE());

  mrb_define_method(mrb, vec3_class, "sqdist", cVec3_sqdist, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "dist", cVec3_dist, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "distance!", cVec3_distanceB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "normalize", cVec3_normalize, MRB_ARGS_NONE());
  mrb_define_method(mrb, vec3_class, "normalize!", cVec3_normalizeB, MRB_ARGS_NONE());

  mrb_define_method(mrb, vec3_class, "inner", cVec3_inner, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, vec3_class, "dot", cVec3_inner, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "cross", cVec3_cross, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "outer", cVec3_outer, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, vec3_class, "to_a", cVec3_to_a, MRB_ARGS_NONE());

  mrb_define_method(mrb, vec3_class, "inspect", cVec3_inspect, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mod, "vec3", new_vec3, MRB_ARGS_REST());

  return vec3_class;
}

