#include "tsmrb_veclist.h"
#include "tsmrb_helpers.h"
#include "tsmrb_stats.h"

#include "mruby/data.h"
#include "mruby/numeric.h"

#include "numarray_mruby.h"

static void veclist_free(mrb_state* mrb, void* p)
{
  veclist_dealloc((VectorList*)p);
  tsmrb_free(mrb,p);
}

static const mrb_data_type veclist_type = {
  .struct_name = "VectorList",
  .dfree = veclist_free,
};

static VectorList *
alloc_veclist(mrb_state *mrb)
{
  VectorList* vl = mrb_malloc(mrb, sizeof(*vl));
  memset(vl,0,sizeof(*vl));
  return vl;
}

static VectorList *
new_veclist(mrb_state *mrb, mrb_int n)
{
  VectorList *vl = alloc_veclist(mrb);

  if (veclist_init(vl, n) < 0) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not initialize vector list");
  }

  veclist_zero(vl);

  return vl;
}

static mrb_value tsmrb_veclist_wrap(mrb_state* mrb, VectorList* vl)
{
  struct RData* data = mrb_data_object_alloc(mrb,
      tsmrb_veclist_class(mrb), vl, &veclist_type);
  return mrb_obj_value(data);
}

VectorList* tsmrb_to_veclist(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_obj_is_instance_of(mrb, obj, tsmrb_veclist_class(mrb)))
    mrb_raise(mrb, E_TYPE_ERROR, "expected a VectorList object");

  return DATA_PTR(obj);
}

mrb_value tsmrb_alloc_veclist(mrb_state *mrb, VectorList** vlptr)
{
  VectorList *vl;

  vl = alloc_veclist(mrb);
  if (vlptr) { *vlptr = vl; }
  return tsmrb_veclist_wrap(mrb, vl);
}

mrb_value tsmrb_new_veclist(mrb_state* mrb, size_t n, VectorList** vlptr)
{
  VectorList *vl;

  vl = new_veclist(mrb, n);
  if (vlptr) { *vlptr = vl; }
  return tsmrb_veclist_wrap(mrb, vl);
}

static inline void check_same_size(mrb_state *mrb, const VectorList *vA, const VectorList *vB)
{
  if (vA->n != vB->n)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "vector lists are not the same size");
  }
}

static mrb_value cVectorList_initialize(mrb_state *mrb, mrb_value self)
{
  VectorList* vecList;
  mrb_int n;

  mrb_get_args(mrb, "i", &n);
  if (n < 0)
  {
    mrb_raisef(mrb, E_ARGUMENT_ERROR, "VectorList size must not be negative");
  }

  vecList = new_veclist(mrb, n);

  DATA_TYPE(self) = &veclist_type;
  DATA_PTR(self) = vecList;

  return self;
}

static mrb_value cVectorList_copy(mrb_state *mrb, mrb_value self)
{
  VectorList *origList, *newList;

  origList = tsmrb_to_veclist(mrb,self);
  newList = new_veclist(mrb, origList->n);

  memcpy(newList->vX, origList->vX, 3*origList->n*sizeof(*(origList->vX)));

  return tsmrb_veclist_wrap(mrb, newList);
}

static mrb_value cVectorList_size(mrb_state *mrb, mrb_value self)
{
  VectorList* vecList = tsmrb_to_veclist(mrb,self);

  return mrb_fixnum_value(vecList->n);
}

static mrb_value cVectorList_at(mrb_state *mrb, mrb_value self)
{
  mrb_int i;
  mrb_value vec[3];
  VectorList* vecList;

  mrb_get_args(mrb, "i", &i);
  vecList = tsmrb_to_veclist(mrb,self);

  if ((i < 0) || (i >= vecList->n))
  {
    mrb_raisef(mrb, E_INDEX_ERROR, "Vector list index (%d) out of bounds", (int)i);
  }

  vec[0] = mrb_float_value(mrb, vecList->vX[i]);
  vec[1] = mrb_float_value(mrb, vecList->vY[i]);
  vec[2] = mrb_float_value(mrb, vecList->vZ[i]);

  return mrb_ary_new_from_values(mrb, 3, vec);
}

static mrb_value cVectorList_set_at(mrb_state *mrb, mrb_value self)
{
  VectorList* vecList;
  mrb_int ind, argc;
  mrb_value *argv;

  vecList = tsmrb_to_veclist(mrb,self);
  mrb_get_args(mrb, "ia", &ind, &argv, &argc);

  if ((ind < 0) || (ind >= vecList->n))
  {
    mrb_raisef(mrb, E_INDEX_ERROR, "Vector list index (%d) out of bounds", (int)ind);
  }

  if (argc != 3)
  {
    mrb_raisef(mrb, E_TYPE_ERROR, "expected a three-element array");
  }

  vecList->vX[ind] = mrb_to_flo(mrb, argv[0]);
  vecList->vY[ind] = mrb_to_flo(mrb, argv[1]);
  vecList->vZ[ind] = mrb_to_flo(mrb, argv[2]);

  return self;
}

static mrb_value veclist_binop(mrb_state *mrb, mrb_value self,
    void (*binop)(VectorList *restrict vA, const VectorList *restrict vB), const char *selferr)
{
  VectorList *vA, *vB;
  mrb_value other;

  mrb_get_args(mrb, "o", &other);

  vA = tsmrb_to_veclist(mrb,self);
  vB = tsmrb_to_veclist(mrb,other);

  check_same_size(mrb, vA,vB);

  if (selferr)
  {
    if (vA == vB)
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR, selferr);
    }

    binop(vA,vB);
  }
  else
  {
    void (*binop_norestrict)(VectorList *, VectorList *) = (void(*)(VectorList*,VectorList*))binop;
    binop_norestrict(vA,vB);
  }

  return self;
}

static mrb_value cVectorList_addB(mrb_state *mrb, mrb_value self)
{
  return veclist_binop(mrb, self, veclist_add, "cannot add a vector list to itself");
}

static mrb_value cVectorList_subB(mrb_state *mrb, mrb_value self)
{
  return veclist_binop(mrb, self, veclist_sub, "cannot subtract a vector list from itself");
}

static mrb_value cVectorList_setB(mrb_state *mrb, mrb_value self)
{
  VectorList *vA, *vB;
  mrb_value other;
  size_t i,n;

  mrb_get_args(mrb, "o", &other);

  vA = tsmrb_to_veclist(mrb,self);
  vB = tsmrb_to_veclist(mrb,other);

  if (veclist_init(vA,vB->n) < 0) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not initialize vector list");
  }

  for(i=0, n=vA->n; i < n; i++)
  {
    vA->vX[i] = vB->vX[i];
    vA->vY[i] = vB->vY[i];
    vA->vZ[i] = vB->vZ[i];
  }

  return self;
}

static mrb_value cVectorList_scale_byB(mrb_state *mrb, mrb_value self)
{
  VectorList* vecList;
  mrb_float scl;

  vecList = tsmrb_to_veclist(mrb,self);
  mrb_get_args(mrb, "f", &scl);

  veclist_scaleby(vecList, (double)scl);

  return self;
}

static mrb_value cVectorList_normalizeB(mrb_state *mrb, mrb_value self)
{
  VectorList* vecList = tsmrb_to_veclist(mrb,self);

  veclist_normalize(vecList);

  return self;
}

static mrb_value cVectorList_distance_toB(mrb_state *mrb, mrb_value self)
{
  return veclist_binop(mrb, self, veclist_distance_to, NULL);
}

static mrb_value cVectorList_sqdist_toB(mrb_state *mrb, mrb_value self)
{
  return veclist_binop(mrb, self, veclist_sqdist_to, NULL);
}

static mrb_value cVectorList_x_accumulate_stats(mrb_state *mrb, mrb_value self)
{
  struct ts_accumulated_stats stats;
  VectorList *veclist;
  double *beg, *end;
  int first;
  mrb_int argc;
  mrb_value *argv;

  mrb_get_args(mrb, "*", &argv, &argc);
  veclist = tsmrb_to_veclist(mrb,self);
  beg = veclist->vX;
  end = veclist->vX + veclist->n;

  if (argc == 0)
  {
    first = 1;
  }
  else if (argc == 1)
  {
    mrb_check_type(mrb, argv[0], MRB_TT_ARRAY);
    first = (mrb_ary_len(mrb, argv[0]) == 0);
    if (!first)
    {
      tsmrb_accumulated_stats_from_ary(mrb, &stats, argv[0]);
    }
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected 0 or 1 arguments");
  }


  if (first)
  {
    if (beg == end) return mrb_ary_new(mrb);

    ts_accumulated_stats_init(&stats, *beg);
    beg++;
  }

  while (beg != end)
  {
    ts_accumulate_stats(&stats,*beg);
    beg++;
  }

  return tsmrb_accumulated_stats_to_ary(mrb, &stats);
}

static mrb_value cVectorList_dot_withB(mrb_state *mrb, mrb_value self)
{
  return veclist_binop(mrb, self, veclist_dot_with, NULL);
}


#define OP_SQ(x)     ( square(x) )
#define OP_INVERT(x) ( 1.0 / (x) )
#define OP_SQRT(x)   ( sqrt(x) )

#define OP_COS(x)    ( cos(x) )
#define OP_SIN(x)    ( sin(x) )
#define OP_TAN(x)    ( tan(x) )

#define OP_ACOS(x)   ( acos(x) )
#define OP_ASIN(x)   ( asin(x) )
#define OP_ATAN(x)   ( atan(x) )

#define APPLY_OP(op) \
  do { for (size_t i=0; i < vl->n; i++) { double x0 = vl->vX[i]; vl->vX[i] = op( x0 ); } } while(0)
#define IF_OP(sym, op) \
  do { if (op_sym == TSMRB_SYM(sym)) { APPLY_OP( op ); return self; } } while(0)
static mrb_value cVectorList_opB(mrb_state *mrb, mrb_value self)
{
  VectorList* vl;
  mrb_sym op_sym;

  vl = tsmrb_to_veclist(mrb,self);
  mrb_get_args(mrb, "n", &op_sym);

  IF_OP(sq    , OP_SQ    );
  IF_OP(sqrt  , OP_SQRT  );
  IF_OP(invert, OP_INVERT);

  IF_OP(cos   , OP_COS   );
  IF_OP(acos  , OP_ACOS  );

  IF_OP(sin   , OP_SIN   );
  IF_OP(asin  , OP_ASIN  );

  IF_OP(tan   , OP_TAN   );
  IF_OP(atan  , OP_ATAN  );

  mrb_raise(mrb, E_ARGUMENT_ERROR, "invalid operation");
  return self;
}
#undef IF_OP
#undef APPLY_OP

#undef OP_SQ
#undef OP_INVERT
#undef OP_SQRT

#undef OP_COS
#undef OP_ACOS

#undef OP_SIN
#undef OP_ASIN


static mrb_value cVectorList_subset(mrb_state *mrb, mrb_value self)
{
  VectorList *veclist, *sublist;
  mrb_int ind0, ind1, i, n;

  veclist = tsmrb_to_veclist(mrb,self);
  mrb_get_args(mrb, "ii", &ind0, &ind1);

  if (ind0 < 0) { ind0 = 0; }
  if (ind1 > veclist->n) { ind1 = veclist->n; }

  if ((ind1 <= ind0) || (ind0 >= veclist->n)) { return mrb_nil_value(); }

  n = ind1 - ind0;
  sublist = new_veclist(mrb, n);

  {
    const double *restrict vX0 = veclist->vX+ind0;
    const double *restrict vY0 = veclist->vY+ind0;
    const double *restrict vZ0 = veclist->vZ+ind0;

    double *restrict vX1 = sublist->vX;
    double *restrict vY1 = sublist->vY;
    double *restrict vZ1 = sublist->vZ;

    for (i=0; i < n; i++)
    {
      vX1[i] = vX0[i];
      vY1[i] = vY0[i];
      vZ1[i] = vZ0[i];
    }
  }

  return tsmrb_veclist_wrap(mrb, sublist);
}

static mrb_value cVector_to_narray(mrb_state *mrb, mrb_value self)
{
  VectorList* veclist = tsmrb_to_veclist(mrb,self);

  int dims[2];
  dims[0] = veclist->n;
  dims[1] = 3;

  numarray *A = NULL;
  mrb_value arr = numarray_new_filled(mrb, 2, dims, 0.0, &A);

  {
    double *restrict xvec = A->elts;
    double *restrict yvec = xvec + veclist->n;
    double *restrict zvec = yvec + veclist->n;

    memcpy(xvec, veclist->vX, veclist->n);
    memcpy(yvec, veclist->vY, veclist->n);
    memcpy(zvec, veclist->vZ, veclist->n);
  }

  return arr;
}

static mrb_value cVector_to_a(mrb_state *mrb, mrb_value self)
{
  size_t i;
  int ai;
  VectorList* veclist;
  mrb_value ary, vec[3];

  veclist = tsmrb_to_veclist(mrb,self);
  ary = mrb_ary_new_capa(mrb, veclist->n);

  for(i=0; i < veclist->n; i++)
  {
    ai = mrb_gc_arena_save(mrb);

    vec[0] = mrb_float_value(mrb, veclist->vX[i]);
    vec[1] = mrb_float_value(mrb, veclist->vY[i]);
    vec[2] = mrb_float_value(mrb, veclist->vZ[i]);
    mrb_ary_push(mrb, ary, mrb_ary_new_from_values(mrb, 3, vec));

    mrb_gc_arena_restore(mrb,ai);
  }

  return ary;
}

struct RClass* tsmrb_define_veclist(mrb_state* mrb, struct RClass* mod)
{
  if (tsmrb_veclist_class(mrb) != NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "VectorList class has already been initialized!");

  struct RClass* cVectorList =
    mrb_define_class_under(mrb, mod, "VectorList", mrb->object_class);
  tsmrb_classes(mrb)->cVectorList = cVectorList;
  MRB_SET_INSTANCE_TT(cVectorList, MRB_TT_DATA);

  mrb_define_method(mrb, cVectorList, "initialize", cVectorList_initialize, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cVectorList, "size", cVectorList_size, MRB_ARGS_NONE());

  mrb_define_method(mrb, cVectorList, "[]", cVectorList_at, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cVectorList, "[]=", cVectorList_set_at, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cVectorList, "add!", cVectorList_addB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cVectorList, "sub!", cVectorList_subB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cVectorList, "copy", cVectorList_copy, MRB_ARGS_NONE());
  mrb_define_method(mrb, cVectorList, "set!", cVectorList_setB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cVectorList, "scale_by!", cVectorList_scale_byB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cVectorList, "normalize!", cVectorList_normalizeB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cVectorList, "distance_to!", cVectorList_distance_toB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cVectorList, "sqdist_to!", cVectorList_sqdist_toB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cVectorList, "dot_with!", cVectorList_dot_withB, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cVectorList, "op!", cVectorList_opB, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cVectorList, "x_accumulate_stats", cVectorList_x_accumulate_stats, MRB_ARGS_ANY());

  mrb_define_method(mrb, cVectorList, "subset", cVectorList_subset, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cVectorList, "to_narray", cVector_to_narray, MRB_ARGS_NONE());

  mrb_define_method(mrb, cVectorList, "to_a", cVector_to_a, MRB_ARGS_NONE());

  return cVectorList;
}







