require 'lattice'
require 'yaml'

module Tubesheet::Configure
  class ChemDriverConfigurer
    @@ignore_missing = false

    def self.reset_ignore_missing
      @@ignore_missing = false
    end

    def self.ignore_missing=(val)
      if val
        @@ignore_missing = true
      else
        @@ignore_missing = false
      end
    end

    def self.ignore_missing(val)
      @@ignore_missing
    end

    def as_float(val,&block)
      return 0.0 if val.nil? && @@ignore_missing

      if block
        begin
          return Float(val)
        rescue
          return block.call
        end
      else
        return Float(val)
      end
    end

    def as_int(val,&block)
      return 0 if val.nil? && @@ignore_missing

      if block
        begin
          return Integer(val)
        rescue
          return block.call
        end
      else
        return Integer(val)
      end
    end

    def die(mesg)
      raise RuntimeError.new(mesg)
    end

    def from_yaml(data)
      config = YAML::load(data)

      num_pfs    = as_int config['num_protofilaments'] { die "Invalid or missing 'num_protofilaments'" }
      max_dimers = as_int config['max_dimers'] { die "Invalid or missing 'max_dimers'" }

      die "Parameter 'num_protofilaments' must be > 0" unless num_pfs > 0
      die "Parameter 'max_dimers' must be > 0"         unless max_dimers > 0

      nrows = max_dimers*4 + 1
      ncols = num_pfs*2

      driver = ::Tubesheet::ChemDriver.new nrows, ncols

      driver.user_config = config

      driver.t_max = as_float config['t_max'] { die "Invalid or missing t_max" }
      driver.dt    = as_float config['dt']    { die "Invalid or missing dt" }

      rng_seed = as_int config['rng_seed'] { die "Invalid or missing RNG seed" }
      if rng_seed > 0
        driver.rng = Tubesheet::RNG.new :generator => :mt19937, :seed => rng_seed
      else
        die "'rng_seed' must be > 0" unless @@ignore_missing
      end

      if config['seam_allowed_thresh']
        driver.seam_allowed_thresh = as_int config['seam_allowed_thresh'] {
          die "seam_allowed_thresh must be an integer"
        }
      end

      if config['barrier']
        barrier_type = config['barrier']['type'].to_sym
        bparam = config['barrier']['params']
        case bparam
        when Array
          barrier_params = bparam.map do |elt|
            as_float elt { die "barrier parameters must be numbers" }
          end
        when Numeric
          barrier_params = [ bparam.to_f ]
        when nil
          barrier_params = []
        else
          die "Invalid barrier parameters"
        end

        driver.barrier_model = barrier_type
        driver.barrier_param = barrier_params.first unless barrier_params.empty?
      else
        die "Missing 'barrier' section" unless @@ignore_missing
      end

      driver.temp = as_float config['temp'] { die "Invalid or missing temp" }

      driver.forbidden_energy = as_float config['FORBIDDEN'] { die "Invalid or missing 'FORBIDDEN'" }

      if config['rates']
        monomer_rate = as_float config['rates']['monomer'] { die "Missing or invalid monomer rate" }
        long_rate    = as_float config['rates']['longitudinal'] { die "Missing or invalid longitudinal rate" }
        lat_rate     = as_float config['rates']['lateral'] { die "Missing or invalid lateral rate" }
        seam_rate    = config['rates']['seam'].to_f

        driver.set_rate_prefactor :EVT_MONOMER     , monomer_rate
        driver.set_rate_prefactor :EVT_LONGITUDINAL, long_rate
        driver.set_rate_prefactor :EVT_LATERAL     , lat_rate
        driver.set_rate_prefactor :EVT_LATERAL_SEAM, seam_rate unless seam_rate.nil?
      else
        die "Missing 'rates' section" unless @@ignore_missing
      end

      if config['monomer_energies']
        monomer_energies = {}
        config['monomer_energies'].each do |k,v|
          if v == 'FORBIDDEN'
            en = driver.forbidden_energy
          else
            en = as_float(v) { die "Monomer energy #{k} is not a float" }
          end
          driver.set_monomer_energy(k.to_sym, en)
        end
      else
        die "Missing 'monomer_energies' section" unless @@ignore_missing
      end

      if config['interface_energies']
        config['interface_energies'].each do |k,v|
          if v == 'FORBIDDEN'
            en = driver.forbidden_energy
          else
            en = as_float(v) { die "Interface energy #{k} is not a float" }
          end
          driver.set_interface_energy(k.to_sym,en)
        end
      else
        die "Missing 'interface_energies' section" unless @@ignore_missing
      end

      if config['bending_energies']
        config['bending_energies'].each do |k,v|
          if v == 'FORBIDDEN'
            en = driver.forbidden_energy
          else
            en = as_float(v) { die "Bending energy #{k} is not a float" }
          end
          driver.set_bending_energy(k.to_sym,en)
        end
      end

      configure_coupling_from_yaml(driver,config) if config['coupling']

      return driver
    end

    def symbolify_hash(h)
      Hash[ h.map{|k,v| [k.to_sym, (v.is_a? String) ? v.to_sym : v]} ]
    end

    def configure_coupling_from_yaml(driver,config)
      couplingConfig = config['coupling']
      return unless couplingConfig

      ffparams = symbolify_hash( config['force_field'] || {} )
      ffargs = symbolify_hash( config['ffconfig'] || {} )
      builder = ConfigMechanicalCoupling.setup_mechanical(driver,ffparams,ffargs)

      driver.configure do
        mechanical_model(builder) do
          if couplingConfig['states']
            couplingConfig['states'].each do |st,stmap|
              states(st.to_sym, Hash[stmap.map{|k,v| [k.to_sym,v.to_sym]}])
            end
          end

          if couplingConfig['acceptance']
            acceptance(couplingConfig['acceptance'].to_sym)
          end

          if couplingConfig['minimization_params']
            minimization_params(
              Hash[couplingConfig['minimization_params'].map{|k,v| [k.to_sym,v]}]
            )
          end

          if couplingConfig['energy']
            energy(couplingConfig['energy'].to_sym)
          end

          if couplingConfig['mech_temp']
            mech_temp(couplingConfig['mech_temp'])
          end
        end
      end

      if config['logging']
        config['logging'].each do |k,v|
          driver.coupling.sim.set_logging(k.to_sym,v)
          puts "Logging for :#{k} has been #{v ? 'enabled' : 'disabled'}"
        end
      end
    end
  end

  class ConfigMechanicalCoupling
    attr_reader :coupling, :driver

    def initialize(driver,builder)
      @driver = driver

      @builder = builder
      @builder ||= self.class.setup_mechanical(driver)

      @coupling = setup_coupling
      @driver.coupling = @coupling
    end

    def self.setup_mechanical(driver,params=nil,args=nil)
      num_dimers = ((driver.nrow - 1)/4).floor
      num_pfs = (driver.ncol / 2).floor
      max_dimers = num_dimers * num_pfs
      max_monomers = 2*max_dimers

      kwArgs = {
        :rng                => driver.rng,
        :max_monomers       => max_monomers,
        :num_protofilaments => num_pfs,
      }
      kwArgs.merge(args) if args

      kwArgs[:params] = params if params
      builder = ::Tubulin::SimulationBuilder.new(kwArgs)

      return builder
    end

    def setup_coupling
      return ::Tubesheet::ChemCoupling.new @driver, @builder
    end

    def states(site_type, state_mapping)
      @coupling.add_state_mapping(site_type, state_mapping)
    end

    def acceptance(criteria)
      @coupling.fe_acceptance = criteria
    end

    def minimization_params(params)
      @coupling.minimization_params.merge! params
    end

    def energy(energy_type)
      STDERR.puts "WARNING: 'energy' parameter is deprecated. Use 'energy_type'."
      self.energy_type(energy_type)
    end

    def energy_type(energy_type)
      @coupling.energy_type = energy_type
    end

    def mech_temp(temp)
      @coupling.mech_temp = temp
    end

    def link!
      @coupling.link
    end
  end

  class ConfigChemDriverDSL < ConfigLatticeDSL
    def initialize(drv)
      @driver = drv
      @lattice = @driver
    end

    def barrier(btype,param=0.0)
      @driver.barrier_model = btype
      @driver.barrier_param = param
    end

    def rates( prefactors )
      prefactors.each { |k,v| @driver.set_rate_prefactor(k,v) }
    end

    def monomer_energies( energies )
      energies.each { |k,v| @driver.set_monomer_energy(k,v) }
    end

    def interface_energies( energies )
      energies.each do |triplet, energy| @driver.set_interface_energy(triplet, energy) end
    end

    def bending_energies( energies )
      energies.each do |triplet, energy| @driver.set_bending_energy(triplet, energy) end
    end

    def energies(which, energies)
      case which
      when :monomer
        monomer_energies(energies)
      when :interface
        interface_energies(energies)
      when :bending
        bending_energies(energies)
      end
    end

    def make_mechanical_builder(kwArgs={})
      params = kwArgs[:params]
      args   = kwArgs[:args]
      ConfigMechanicalCoupling.setup_mechanical @driver,params,args
    end

    def mechanical_model(builder=nil,&script)
      config = ConfigMechanicalCoupling.new @driver, builder
      config.instance_eval(&script)
      config.coupling
    end
  end

  module DispatchComparisons
    def ==(term) cmp(:==, term) end
    def <(term) cmp(:<, term) end
    def <=(term) cmp(:<=, term) end
    def >(term) cmp(:>, term) end
    def >=(term) cmp(:>=, term) end
    def neq(term) cmp(:neq, term) end

  end

  class MultiCondition
    include DispatchComparisons

    def initialize(*cond)
      @all_conds = cond.to_a
    end

    def cmp(op,term)
      if @all_conds.empty?
        raise RuntimeError.new 'no existing conditions'
      end

      @all_conds.push @all_conds.last.b.cmp(op,term)
      self
    end

    def check(logf)
      tot_fails = 0
      tot_tests = 0
      @all_conds.each { |c|
        num_fails,num_tests = c.check(logf)
        tot_fails += num_fails
        tot_tests += num_tests
      }
      [tot_fails,tot_tests]
    end
  end

  class Condition
    include DispatchComparisons

    attr_reader :a, :op, :b

    def initialize(a, op, b)
      @a = a
      @op = op
      @opstr = op.to_s
      @b = b
      if @op == :neq
        @cond = Proc.new { a.value != b.value }
      else
        @cond = Proc.new { a.value.send op,b.value }
      end
    end

    def eval
      @cond.call
    end

    def to_s
      '%s %s %s' % [@a.pretty_string, @opstr, @b.pretty_string]
    end

    def cmp(op,term)
      MultiCondition.new self, @b.cmp(op,term)
    end

    def check(logf)
      logf.print 'CHECKING condition %s... ' % [ self.to_s ]
      ok = self.eval
      if ok
        logf.print "OKAY\n"
        [0,1]
      else
        logf.print "FAILED\n"
        [1,1]
      end
    end
  end

  class EnergyExpr
    include DispatchComparisons

    def initialize(pstr,&proc)
      @pstr = pstr
      @expr = Proc.new(&proc)
    end

    def coerce(other)
      if other.kind_of? Numeric
        return EnergyTerm.new(@driver, other), self
      end
    end

    def value
      @expr.call
    end

    def pretty_string
      @pstr
    end

    def group_string
      if leaf?
        "%s" % pretty_string
      else
        "(%s)" % pretty_string
      end
    end

    def leaf?
      return false
    end

    def -@
      EnergyExpr.new("-%s" % [self.group_string]) do
        -self.value
      end
    end

    def +@
      self
    end

    def binop(op,term)
      term = EnergyTerm.new(@driver, term) if term.kind_of? Numeric

      EnergyExpr.new("%s %s %s" % [self.group_string,op.to_s,term.group_string]) do
        self.value.send op,term.value
      end
    end

    def +(term) binop(:+,term) end
    def -(term) binop(:-,term) end
    def *(term) binop(:*,term) end
    def /(term) binop(:/,term) end

    def cmp(op,term)
      if term.kind_of? EnergyExpr
        Condition.new self,op,term
      elsif term.kind_of? Numeric
        Condition.new self,op,EnergyTerm.new(@driver,term)
      else
        raise RuntimeError.new 'Invalid type for == comparison'
      end
    end

  end

  class EnergyTerm < EnergyExpr
    attr_reader :value

    def initialize(driver,spec)
      @driver = driver
      @spec = spec
      if spec.kind_of? Symbol
        @value = lookup_energy(spec)
        if @value.nil?
          raise RuntimeError.new 'Unknown energy term specified: %s' % [spec.inspect]
        end
      elsif spec.kind_of? Numeric
        @value = spec
      else
        raise RuntimeError.new 'Invalid energy term specified: %s' % [spec.inspect]
      end
    end

    def lookup_energy(spec)
      s = spec.to_s
      case s.length
      when 1
        @driver.monomer_energies[spec]
      when 3
        @driver.get_interface_energy(spec)
      else
        raise RuntimeError.new 'Invalid spec "%s"' % [s]
      end
    end

    def leaf?
      return true
    end

    def pretty_string
      if (@spec.kind_of? Numeric)
        @spec.to_s
      else
        'en(%s)' % [@spec.to_s]
      end
    end

  end

  class CheckEnergies
    def initialize(driver,logf)
      @driver = driver
      @logf = logf
      @onfail = :warning
      @tot_fails = 0
      @tot_tests = 0
    end

    def en(spec)
      EnergyTerm.new @driver, spec
    end

    def on_failure(err_or_warn)
      case err_or_warn
      when :error
        @onfail = :error
      when :warning
        @onfail = :warning
      when :pause
        @onfail = :pause
      else
        raise RuntimeError.new 'unknown argument to on_failure (expected :error or :warning)'
      end
    end

    def check(desc,*conds)
      @logf.puts "---[ %s ]---\n" % [desc]
      conds.each do |c|
        num_fails,num_tests = c.check(@logf)
        @tot_fails += num_fails
        @tot_tests += num_tests
      end
      @logf.puts "\n"
    end

    def report
      @logf.puts '%d parameter checks run.  %d failed.' % [@tot_tests, @tot_fails]
      if (@tot_fails > 0)
        if @onfail == :error
          raise RuntimeError.new 'Failed %d of %d parameter checks' % [@tot_fails,@tot_tests]
        elsif @onfail == :pause
          if @logf == STDOUT
            @logf.puts 'Press <RETURN> to continue'
            STDIN.gets
          end
        end
      end
    end

    def self.run_checks(driver,logf,&script)
      checker = self.new driver, logf
      checker.instance_eval(&script)

      checker.report
    end
  end
end

module Tubesheet
  class ChemDriver

    def max_dimers
      ((nrow - 1)/4).floor
    end

    def num_protofilaments
      (ncol / 2).floor
    end

    def configure(&script)
      Tubesheet::Configure::ConfigChemDriverDSL.new(self).instance_eval(&script)
    end

    def check_energies(logf, &script)
      Tubesheet::Configure::CheckEnergies.run_checks self,logf,&script
    end

    def latt_as_string(border='',rlim=nil)
      if rlim.nil?
        rlim = 0...nrow
      end

      rows = []
      bsz = border.size
      ssz = ncol+2*bsz
      for r in rlim do
        s = " "*ssz
        if bsz > 0
          off = ssz-bsz
          (0...bsz).each { |i| c = border[i] ; s[i] = c ; s[off] = c }
        end
        for c in 0...ncol do
          s[c+bsz] = ChemDriver.char_value( self[r,c] )
        end
        rows.push s
      end

      return rows.join("\n")
    end

    def default_checks(file_arg=STDOUT,on_fail_arg=:pause)
      check_energies(file_arg) do
        on_failure on_fail_arg

        check 'Bent forms have equal energies',
          en(:L) == en(:R)

        check 'Relative stabilities of the tertiary terms',
          en(:C) < en(:R) < en(:S)

        check 'Bent/bent/bent interfaces have equal energies',
          en(:RBR) == en(:LBL),
          en(:RBR) == en(:RBL),
          en(:RBR) == en(:LBR)

        check 'Relative stabilities of the stable quaternary terms',
          en(:SSS) <= en(:RBR) < en(:CCC)

        check 'Energetics for kinetically-unfavorable states',
          en(:RBR) < en(:RCR),
          en(:RCR) == en(:RCL), en(:RCR) == en(:LCL), en(:RCR) == en(:LCR),
          en(:SSS) < en(:SBS) < en(:SCS)


        check 'CCC -> BBB allosteric energy terms',
          en(:CCC) < en(:RBC) < en(:RCC) < en(0.0),
          en(:RBR) - en(:CCC) < en(:C) - en(:R),
          en(:RCC) < en(:CCR)

        check 'Bent forms in CCC->BBB have same energies',
          en(:CCR) == en(:CCL),
          en(:RCC) == en(:LCC),
          en(:RBC) == en(:LBC)

        check 'BBB -> SSS allosteric energy terms',
          en(:SSS) < en(:RBR) < en(:SSR) < en(:SBR) < en(0.0),
          en(:SSS) < en(:SBS),
          en(:SSS) - en(:RBR) > en(:R) - en(:S),
          en(:SBR) < en(:RBS)

        check 'Bent forms in BBB->SSS have same energies',
          en(:SBR) == en(:SBL),
          en(:SSR) == en(:SSL),
          en(:RBS) == en(:LBS)

        check 'Relationships between CCC->BBB and BBB->SSS terms',
          en(:RCC) < en(:SCC),
          en(:SSR) < en(:SSC),
          en(:SSC) < en(:SBC) > en(:SCC),
          en(:CCR) < en(:CCS)

        check 'Stable lateral interactions',
          en(:StS) < en(:SeS) < en(:LeR) < 0.0,
          en(:StS) < en(:RtL) < en(:LeR) < 0.0

        check 'Stable intermediate forms',
          en(:StS) < en(:RtS) < en(:RtL) < 0.0,
          en(:SeS) < en(:LeS) < en(:LeR) < 0.0,

          en(:RtS) == en(:StL),
          en(:StR) == en(:LtS),
          en(:RtS) <= en(:StR) < 0.0,

          en(:LeS) == en(:SeR),
          en(:SeL) == en(:ReS),
          en(:LeS) <= en(:SeL) < 0.0,

          en(:RtS) < en(:LeS)

        check 'Unstable forms (not involving curved)',
          en(:RtL) < en(:RtR) < en(:LtR),
          en(:RtL) < en(:LtL) < en(:LtR),

          en(:LeR) < en(:LeL) < en(:ReL),
          en(:LeR) < en(:ReR) < en(:ReL)

        check 'Lateral interactions stabilize straight microtubules',
          ((en(:S) - en(:R)) +
           (en(:SSS) - en(:RBR)) +
           (en(:StS) - (en(:RtL) + en(:LeR))/2)) < 0

        check 'Unstable forms involving curved',
          en(:CtC) > en(:CeC) > 0.0,

          en(:CtS) == en(:StC),
          en(:CeS) == en(:SeC),

          en(:StC) < en(:RtC) < en(:CtR),
          en(:SeC) < en(:LeC) < en(:CeL),

          en(:LeC) == en(:CeR),
          en(:RtC) == en(:CtL),

          en(:ReC) == en(:CeL),
          en(:LtC) == en(:CtR),

          en(:RtL) < en(:StC) < en(:RtC),
          en(:LeR) < en(:SeC) < en(:LeC),

          en(:RtR) < en(:RtC),
          en(:LtL) < en(:LtC),
          en(:LeL) < en(:LeC),
          en(:ReL) < en(:ReC)
      end
    end

    def first_row
      return 1
    end

    def last_row
      return self.nrow-2
    end

    def first_col
      return 0
    end

    def last_col
      return self.ncol-1
    end

    EMPTY_STRING = ChemDriver::char_value(ChemDriver::EMPTY)

    def add_ifc_term(ifcs, t1,t2,t3)
      term = raw_ifc_to_symbolic(t1,t2,t3)
      ifcs << term unless term.to_s.include? EMPTY_STRING
    end

    def monomer_terms(r,c)
      if site_type(r,c) != :MONOMER
        raise ArgumentError.new("site %3d,%3d is not a :MONOMER site" % [r,c])
      end

      nrow = self.nrow
      ncol = self.ncol
      st0 = self[r,c]

      ifcs = []
      if r-2 >= first_row
        add_ifc_term(ifcs,self[r-2,c], self[r-1,c], st0)
      end

      if r+2 <= last_row
        add_ifc_term(ifcs,st0, self[r+1,c], self[r+2,c])
      end

      (rm1,cm1) = self.wrap_coords(r,c-1)
      (rm2,cm2) = self.wrap_coords(r,c-2)

      if (rm2 >= first_row) && (rm2 <= last_row)
        add_ifc_term(ifcs,self[rm2,cm2], self[rm1,cm1], st0)
      end

      (rp1,cp1) = self.wrap_coords(r,c+1)
      (rp2,cp2) = self.wrap_coords(r,c+2)

      if (rp2 >= first_row) && (rp2 <= last_row)
        add_ifc_term(ifcs,st0, self[rp1,cp1], self[rp2,cp2])
      end

      ifcs.reject! { |states| states.to_s.include?(ChemDriver::char_value(ChemDriver::EMPTY)) }

      ifcs.sort!
      ifcs
    end

    def get_raw_interface_term(r,c)
      case site_type(r,c)
      when :LONGIFC
        term = [self[r-1,c],self[r,c],self[r+1,c]]
      when :LATIFC
        rp1,cp1 = wrap_coords(r,c+1)
        rm1,cm1 = wrap_coords(r,c-1)
        term = [self[rm1,cm1],self[r,c],self[rp1,cp1]]
      else
        raise ArgumentError.new("site %3d,%3d is not an interface site" % [r,c])
      end
    end

    def raw_ifc_to_symbolic(t1,t2,t3)
      s = "   "
      s[0] = ChemDriver::char_value(t1)
      s[1] = ChemDriver::char_value(t2)
      s[2] = ChemDriver::char_value(t3)
      s.to_sym
    end

    def interface_term(r,c)
      term = get_raw_interface_term(r,c)

      if term.include? ChemDriver::EMPTY
        nil
      else
        raw_ifc_to_symbolic(*term)
      end
    end

    def move_table
      mt = Hash.new do |h,k|
        v = Hash.new do |h1,k1|
          v1 = [0,0.0]
          h1[k1] = v1
          v1
        end
        h[k] = v
        v
      end

      moves = self.all_moves
      moves.each do |(i,j,state1,rate)|
        state0 = self[i,j]
        stype = site_type(i,j)
        case stype
        when :MONOMER
          # gather terms...
          orig_terms = monomer_terms(i,j)

          begin
            orig = self[i,j]
            self[i,j] = state1
            new_terms = monomer_terms(i,j)
          ensure
            self[i,j] = orig
          end

          orig_terms.sort!
          new_terms.sort!

          orig_terms.unshift ChemDriver::char_value(state0)
          new_terms.unshift  ChemDriver::char_value(state1)

        when :LONGIFC
          orig_term = interface_term(i,j)

          begin
            orig = self[i,j]
            self[i,j] = state1
            new_term = interface_term(i,j)
          ensure
            self[i,j] = orig
          end

          orig_terms = [ChemDriver::char_value(state0), orig_term]
          new_terms  = [ChemDriver::char_value(state1), new_term]
        when :LATIFC
          case state1
          when ChemDriver::PSEUDO_LATERAL_ROTATION
            if !is_alpha_lateral(i,j)
              raise RuntimeError.new "Rotation move at non-alpha lateral site %3d,%3d" % [i,j]
            end

            orig_alpha_term = interface_term(i,j)
            orig_beta_term  = interface_term(i+2,j)
            if orig_alpha_term.nil? || orig_beta_term.nil?
              raise RuntimeError.new "Missing alpha or beta state in rotation at lateral site %3d,%3d" % [i,j]
            end

            orig_state = ChemDriver::char_value(self[i,j])
            orig_terms = [orig_alpha_term,orig_beta_term].sort
            orig_terms = [orig_state, *orig_terms]
            if (orig_state == 't') || (orig_state == 'e')
              new_state = orig_state.tr('et','te')
              new_terms = orig_terms.map{|st| st.to_s.tr('et','te').to_sym}.sort
              new_terms.unshift new_state
            else
              raise RuntimeError.new "State at lateral site %3d,%3d is neither 't' nor 'e'" % [i,j]
            end

          when ChemDriver::PSEUDO_LATERAL_SEAM
            if !is_alpha_lateral(i,j)
              raise RuntimeError.new "Rotation move at non-alpha lateral site %3d,%3d" % [i,j]
            end

            orig_alpha_term = get_raw_interface_term(i,j)
            orig_beta_term  = get_raw_interface_term(i+2,j)

            if orig_alpha_term[1] != orig_beta_term[1]
              raise RuntimeError.new \
                "Seam move: alpha/beta lateral interface mismatch at lateral site %3d,%3d" % [i,j]
            end

            new_alpha_term = orig_alpha_term.clone
            new_beta_term  = orig_beta_term.clone
            if orig_alpha_term[1] == ChemDriver::EMPTY
              new_alpha_term[1] = ChemDriver::INTERFACE_TUBELIKE
              new_beta_term[1]  = ChemDriver::INTERFACE_TUBELIKE
            elsif orig_alpha_term[1] == ChemDriver::INTERFACE_TUBELIKE
              new_alpha_term[1] = ChemDriver::EMPTY
              new_beta_term[1]  = ChemDriver::EMPTY
            else
              raise RuntimeError.new "Seam move: lateral site %3d,%3d is neither empty nor 't'" % [i,j]
            end

            orig_state = self[i,j]
            if orig_state == ChemDriver::EMPTY
              orig_state = "_"
              new_state = "t"
            else
              orig_state = ChemDriver::char_value(orig_state)
              new_state = "_"
            end

            orig_terms = [
              orig_state,
              raw_ifc_to_symbolic(*orig_alpha_term),
              raw_ifc_to_symbolic(*orig_beta_term),
            ]

            new_terms = [
              new_state,
              raw_ifc_to_symbolic(*new_alpha_term),
              raw_ifc_to_symbolic(*new_beta_term),
            ]

          else
            raise RuntimeError.new "Move at lateral site %3d,%3d should be a rotation!" % [i,j]
          end

        else
          raise RuntimeError.new "Invalid site type for site %3d,%3d: :%s" % [i,j,stype.to_s]
        end

        orig_terms.unshift(stype)
        curr = mt[orig_terms]
        accum = curr[new_terms]
        accum[0] += 1
        accum[1] += rate
      end
      return mt
    end

    def initialize_lattice_as_sheet(*kwArgs)
      num_pfs = kwArgs[:num_protofilaments] || self.num_protofilaments
      num_dimers = kwArgs[:num_dimers] || self.max_dimers

      for pf in (0...num_pfs) do
        mer_col = 2*pf      
        lat_col = mer_col+1 

        for dimer in (0...num_dimers) do
          alpha_row = 1+(4*dimer)  
          intra_row = alpha_row+1  
          beta_row  = intra_row+1  
          inter_row = beta_row+1   

          long = :B
          mer = (pf.even?) ? (:R) : (:L)
          lat = (pf.even?) ? (:t) : (:e)

          self[alpha_row,mer_col] = mer
          self[intra_row,mer_col] = long
          self[beta_row ,mer_col] = mer

          if dimer < num_dimers-1
            self[inter_row,mer_col] = long
          end

          if pf < num_pfs-1
            self[alpha_row,lat_col] = lat
            self[beta_row,lat_col] = lat
          end

        end
      end
    end

    def initialize_lattice_as_tube(kwArgs={})
      num_pfs = kwArgs[:num_protofilaments] || self.num_protofilaments
      num_dimers = kwArgs[:num_dimers] || self.max_dimers
      seam = (kwArgs[:seam] && true) || false

      if seam && (num_pfs < self.num_protofilaments)
        throw RuntimeError.new "cannot form seam unless all protofilaments are initialized"
      end

      for pf in (0...num_pfs) do
        mer_col = 2*pf      
        lat_col = mer_col+1 

        for dimer in (0...num_dimers) do
          alpha_row = 1+(4*dimer)  
          intra_row = alpha_row+1  
          beta_row  = intra_row+1  
          inter_row = beta_row+1   

          long = :S
          mer = :S
          lat = :t

          begin
            self[alpha_row,mer_col] = mer
            self[intra_row,mer_col] = long
            self[beta_row ,mer_col] = mer
          rescue ArgumentError => e
            puts <<-eos

err : #{e}
alpha_row  : #{alpha_row}
intra_row  : #{intra_row}
beta_row   : #{beta_row}
pf         : #{pf}
dimer      : #{dimer}

nrow       : #{self.nrow}
ncol       : #{self.ncol}
num_dimers : #{num_dimers}
num_dimers : #{self.num_dimers}
max_dimers : #{self.max_dimers}

            eos
          end

          if dimer < num_dimers-1
            self[inter_row,mer_col] = long
          end

          if seam || (pf < num_pfs-1)
            self[alpha_row,lat_col] = lat
            self[beta_row,lat_col] = lat
          end

        end
      end
    end

    class << self
      alias :orig_from_yaml :from_yaml
      def from_yaml(data)
        driver = self.orig_from_yaml(data)
        config = YAML::load(data)
        if config['coupling']
          configurer = Configure::ChemDriverConfigurer.new
          configurer.configure_coupling_from_yaml(driver,config)
        end
        driver.user_config = config
        return driver
      end

      def fsm_from_yaml(driver,data)
        npf = driver.num_protofilaments

        config = driver.user_config['fast_sheet_model']
        return nil unless config

        is_enabled = config['enabled']
        return nil unless is_enabled && is_enabled != false && is_enabled != 0

        scale = config['scale'] ? Float(config['scale']) : 1.0

        [ [ 'ifc_model',          6 ],
          [ 'existing_model',     3 ],
          [ 'relax_coeff',      npf ],
          ['no_tube_strain',    npf ],
          ['left_tube_strain',  npf ],
          ['right_tube_strain', npf ],
          ['both_tube_strain',  npf ]
        ].each do |sec, sz|
          die "Missing 'fast_sheet_model' subsection: '#{sec}'" \
            unless config[sec]

              okay = (Array === config[sec]) \
                && (config[sec].size == sz)  \
                && (config[sec].map { |x| (Numeric === x) }.all?)

              die "'fast_sheet_model' subsection '#{sec}' requires #{sz} numbers" unless okay
        end

        ifc_model         = config['ifc_model'        ]
        existing_model    = config['existing_model'   ]
        relax_coeff       = config['relax_coeff'      ]

        no_tube_strain    = config['no_tube_strain'   ].map { |x| x*strain_scale }
        left_tube_strain  = config['left_tube_strain' ].map { |x| x*strain_scale }
        right_tube_strain = config['right_tube_strain'].map { |x| x*strain_scale }
        both_tube_strain  = config['both_tube_strain' ].map { |x| x*strain_scale }

        model = Tubesheet::FastSheetModel.new(
          driver, 
          :ifc_model      => ifc_model,
          :existing_model => existing_model,
          :relax_coeff    => relax_coeff,
          :no_tube_strain => no_tube_strain,
          :left_tube_strain => left_tube_strain,
          :right_tube_strain => right_tube_strain,
          :both_tube_strain => both_tube_strain
        )

        model.scale = scale

        fluct_dist = config['sheet_fluct_dist']
        fluct_penalty = config['sheet_fluct_penalty']

        model.sheet_fluctuation_dist = fluct_dist if fluct_dist
        model.sheet_fluctuation_penalty = fluct_penalty if fluct_penalty

        model
      end
    end
  end
end

