module Tubesheet
  class ChemExaminer
    attr_reader :driver,:model
    attr_accessor :recordings,:record,:print_args
    attr_accessor :nolineedit

    def initialize(aDriver, aModel)
      @driver = aDriver
      @model  = aModel
      @states = {}
      @recordings = {}
      @record = nil
      @print_args = []
      @registers = {}
      @history = []
      @extra_cmds = {}
      @nolineedit = false
      @emarks = []
      @emarks_auto = nil
      @output_stack = []
    end

    def print_row_energies(n=nil)
      if !model
        puts "No model"
      else
        tot_en = 0.0
        row_energies = model.row_energies
        ind0 = row_energies.find_index { |en| en > 0.0 }
        if n
          row_energies = row_energies[ind0 ... (ind0+n)]
        end
        row_energies.each_with_index do |en,i|
          tot_en += en
          puts "[%2d] %7.4f %7.4f" % [i+1,en,tot_en] if en > 0.0
        end
      end
    end

    def row_range
      rrange = driver.find_row_range
      first_row = nil
      last_row  = nil
      for (rmin,rmax) in rrange
        first_row = rmin if rmin > 0 && (!first_row || rmin < first_row)
        last_row  = rmax if rmax > 0 && (!last_row  || rmax > last_row )
      end
      [first_row,last_row]
    end

    def print_lattice(*args)
      args = @print_args if args.empty?

      first_row = nil
      last_row = nil
      case args.size
      when 0
        (first_row,last_row) = row_range
      when 1
        (first_row,last_row) = row_range
        if args[0] < 0
          last_row = first_row - args[0]
        else
          first_row = last_row - args[0]
        end
      when 2
        (first_row,last_row) = args
      else
        puts <<-eos
Too many arguments!
p
        prints all occupied rows in the lattice

p  <delta>      
        if delta is negative, prints the first <delta> occupied rows
        otherwise, prints the last <delta> occupied rows

p <row0> <row1>
        prints rows row0 through row1
        eos
        return
      end

      if first_row.nil? || last_row.nil?
        puts "\n(empty lattice)\n"
        return
      end

      rows = driver.latt_as_string('|').split("\n")
      anno = rows[first_row .. (last_row+2)].each_with_index.map do |l,di|
        i = di + first_row
        if (driver.site_type(i,0) != :EMPTY) && driver.is_alpha_lateral(i,1)
          "%4d %s alpha %d" % [i,l,1 + (i-1)/4]
        else
          "%4d %s" % [i,l]
        end
      end
      puts anno
      puts "\ntime = #{driver.time}"
    end

    def basic_print_lattice
      anno = driver.latt_as_string('|').split("\n").each_with_index.map do |l,i|
        if (driver.site_type(i,0) != :EMPTY) && driver.is_alpha_lateral(i,1)
          "%4d %s alpha" % [i,l]
        else
          "%4d %s" % [i,l]
        end
      end
      puts anno
    end

    def set_register(reg,cmds)
      val = nil
      if cmds.size == 1
        if cmds[0] == '-'
          @registers.delete(reg)
          puts "removed register '#{reg}'"
          return
        else
          val = asint(cmds[0])
        end
      else
        stack = []
        cmds.each do |c|
          case c
          when '+'
            a = stack.pop
            b = stack.pop
            raise Exception.new "+: expression stack requires two operands" unless a && b
            stack << (a+b)
          when '-'
            a = stack.pop
            b = stack.pop
            raise Exception.new "-: expression stack requires two operands" unless a && b
            stack << (b-a)
          when '*'
            a = stack.pop
            b = stack.pop
            raise Exception.new "*: expression stack requires two operands" unless a && b
            stack << (a*b)
          when '/'
            a = stack.pop
            b = stack.pop
            raise Exception.new "/: expression stack requires two operands" unless a && b
            stack << (b/a)
          when 'neg'
            raise Exception.new "neg: expression stack is empty" if stack.empty?
            stack[-1] = -stack[-1]
          when 'dup'
            raise Exception.new "dup: expression stack is empty" if stack.empty?
            stack << stack.last
          when 'incr'
            raise Exception.new "incr: expression stack is empty" if stack.empty?
            stack[-1] += 1
          when 'decr'
            raise Exception.new "incr: expression stack is empty" if stack.empty?
            stack[-1] -= 1
          else
            stack << asint(c)
          end
        end
        raise Exception.new "expression stack is empty" if stack.empty?
        val = stack.last
        if stack.size > 1
          puts "WARNING: expression stack has more than one item"
        end
      end
      @registers[reg] = val
      puts "register '#{reg}' set to #{val}"
    end

    def save_state(name)
      @states[name] = driver.state_string
      puts "Saved state as: #{name}"
    end

    def load_state(name)
      ss = @states[name]
      if !ss
        puts "No state: #{name}"
      else
        driver.state_string = ss
        driver.rebuild_moves!
        print_lattice
      end
    end

    def save_recordings(fn)
      if !@recordings || @recordings.size == 0
        puts "No recordings to save"
        return
      end

      File.open(fn,"w") do |f|
        puts "Saving recordings to #{fn}"
        f.puts("#{@recordings.size}")
        @recordings.each do |rname,rlist|
          f.puts "#REC #{rlist.size} #{rname}"
          f.puts(rlist.join("\n"))
        end
        puts "Saved #{@recordings.size} recordings"
      end
    end

    def load_recordings(fn)
      File.open(fn) do |f|
        puts "Loading recordings from #{fn}"
        recs = {}
        l0 = f.readline.chomp
        nr = (l0 != '*') ? Integer(l0) : -1

        data = f.read

        recs = data.split(/^#REC/)
        recs.shift if recs[0] == ""
        if nr > 0 && recs.size != nr
          puts "WARNING: expected #{nr} recordings, but found #{rec.size}"
        end

        recs.each do |rec|
          rl = rec.split("\n")
          rc,rn = rl[0].split(' ',2)
          rc = (rc != '*') ? Integer(rc) : -1
          if rc >= 0 && rl.size != rc+1
            puts "WARNING: recording '#{rn}' expects #{rc}, but found #{rl.size-1}"
          end
          @recordings[rn] = rl[1...rl.size]
        end
      end
    end

    def run
      initial_state = driver.state_string
      loop do
        if @nolineedit
          print ">> "
          has_inp = false
          while !has_inp do
            begin
              inp = STDIN.gets
              has_inp = true
            rescue => e
              puts "Error in input: #{e}"
            end
          end
        else
          inp = REPL.readline(">> ")
        end
        break if inp.nil?

        if @record
          @recordings[@record] << inp.chop
        end

        @history << inp

        begin
          return if handle_input(inp) == :q
        rescue Exception => e
          puts "error: #{e}"
        end
      end
    end

    def playback(args)
      if args.size == 0
        puts <<-eos
Usage: playback <name>...
       plays back commands recorded as <name>"
        eos
        return
      end

      args.each_with_index do |rec,rec_ind|
        puts "playback of '#{rec}'"
        list = @recordings[rec]
        if list.nil?
          puts "no recording '#{rec}'"
          return
        end

        if list.empty?
          puts "empty recording '#{rec}'"
          return
        end

        list.each_with_index do |inp, inp_ind|
          puts "#{rec_ind},#{inp_ind}>> #{inp}"
          begin
            handle_input(inp)
          rescue Exception => e
            puts "error: #{e}"
            return
          end
        end
      end
    end

    def asint(val)
      if val && val[0] == '$'
        key = val[1...val.size]
        return @registers[key] if @registers[key]
      end
      return Integer(val)
    end

    def model_energy
      m = model
      return m ? m.energy : 0.0
    end

    def def_cmd(name, &block)
      @extra_cmds[name] = block
    end

    def puts(*args)
      f = @output_stack.last || STDOUT
      f.puts(*args)
    end

    def print(*args)
      f = @output_stack.last || STDOUT
      f.print(*args)
    end

    def list_emarks
      en = @driver.energy
      puts "Energy %16.10f" % [en]
      @emarks.each_with_index do |mark,ind|
        if mark
          en0,info = mark
          puts "%6d\t%16.10f\t%-40s" % [ind+1,en0-en,info]
        end
      end
    end

    def runloop(cmd)
      reg = cmd[0]
      lo = asint(cmd[1])
      hi = asint(cmd[2])

      rest = cmd.drop(3)
      stmts = []
      ind = rest.index(';')
      while ind do
        stmts << rest[0...ind]
        rest = rest.drop(ind+1)
        ind = rest.index(';')
      end
      stmts << rest

      (lo..hi).each do |i|
        @registers[reg] = i
        puts "loop[#{reg} = #{i}]"
        stmts.each do |c|
          handle_cmd(c, c.join(' '))
        end
      end
    end

    def handle_input(inp)
      if inp[0] == ','
        inp = "eval "+inp[1...inp.length]
      end

      cmd = inp.split(/\s+/)
      handle_cmd(cmd,inp)
    end

    def handle_cmd(cmd,inp)
      case cmd[0]
      when 'reset'
        driver.state_string = initial_state
        driver.rebuild_moves!
        print_lattice

      when 'q','quit'
        return :q

      when 'pwin'
        iargs = []
        if cmd.size > 1
          begin
            iargs = cmd.drop(1).map{ |c| asint(c) }
          rescue Exception => e
            puts "arguments must be integers: #{e}"
          end
        end
        puts "print window: #{iargs.join(" ")}"
        @print_args = iargs

      when 'loop'
        runloop(cmd.drop(1))

      when 'p','print'
        iargs = @print_args || []
        if cmd.size > 1
          begin
            iargs = cmd.drop(1).map{ |c| asint(c) }
          rescue Exception => e
            puts "arguments must be integers: #{e}"
          end
        end
        print_lattice(*iargs)

      when 'redirect'
        if cmd.size == 1
          puts "usage: redirect <file>"
        elsif cmd[1] == '-'
          if @output_stack.size > 0
            f = @output_stack.pop
            STDOUT.puts "popped '#{f.path}' from redirect stack"
            f.close
          else
            STDOUT.puts "cannot pop redirection, output stack is empty"
          end
        else
          STDOUT.puts "Redirecting output to '#{cmd[1]}'"
          begin
            f = File.open(cmd[1],'w')
            @output_stack << f
          rescue Exception => e
            STDOUT.puts "Error redirecting to '#{cmd[1]}': #{e}"
          end
        end

      when 'bp'
        basic_print_lattice

      when 'clear'
        driver.reset!

      when 'zero'
        nrow = driver.nrow
        ncol = driver.ncol
        (0...ncol).each do |j|
          (0...nrow).each do |i|
            driver[i,j] = 0
          end
        end
        driver.rebuild_moves!
        print_lattice
      when 'save'
        if cmd.size == 1
          c = 1
          while @states[c.to_s].nil? do c += 1 end
          save_state c.to_s
        else
          save_state cmd[1]
        end
      when 'list_saved'
        @states.keys.sort.each { |k| puts "#{k}" }
      when 'load'
        if cmd.size == 1
          puts "load <name>"
          puts "use 'list_saved' to see the saved states"
        else
          load_state cmd[1]
        end
      when 'emark'
        case cmd[1]
        when nil
          list_emarks

        when "list"
          puts "energy markers"
          @emarks.each_with_index do |mark,ind|
            if mark
              (en,info) = mark
              puts "%6d\t%16.10f\t%-40s" % [ind+1,en,info]
            else
              puts "%6d\t-deleted" % [ind+1]
            end
          end

        when "delta"
          puts "delta of energy markers"
          if @emarks.empty?
            puts "<none>"
          else
            (en0,info0) = @emarks.first
            @emarks.each_with_index do |mark,ind|
              if mark
                (en,info) = mark
                puts "%6d\t%16.10f\t%-40s %-40s" % [ind+1,en-en0,info,info0]
                en0,info0 = en,info
              else
                puts "%6d\t-deleted" % [ind+1]
              end
            end
          end

        when "auto"
          case (cmd[2] || '').downcase
          when 'on', 'yes', '1'
            @emarks_auto = true
          when 'off', 'no', '0'
            @emarks_auto = false
          when ''
            @emarks_auto = !@emarks_auto
          else
            puts "Invalid argument to 'emarks auto'"
          end
          puts "emarks auto set to #{@emarks_auto ? "on" : "off"}"

        when "add"
          info = (cmd.size > 1) ? cmd.drop(2).join(' ') : ""
          @emarks_auto = true if @emarks.empty? && @emarks_auto.nil?
          en = @driver.energy
          @emarks << [en, info]
          puts "added emark %d: %16.10f %-40s\n" % [@emarks.size,en,info]

        when "clear"
          if cmd.size >= 3
            if cmd[2] == '*'
              @emarks = []
              puts 'cleared all emarks'
            else
              begin
                arg = asint(cmd[2])
              rescue Exception => e
                arg = nil
              end

              if Integer === arg
                if arg > 0 && arg <= @emarks.size
                  @emarks[arg-1] = nil
                else
                  puts "invalid index: #{arg} is not a valid index (1 - #{@emarks.size})"
                end
              else
                puts "invalid index: '#{cmd[2]}' is not a number or known register"
              end
            end
          else
            puts "needs an argument.  use '%s %s *' to clear all emarks" % [ cmd[0], cmd[1] ]
          end
        else
          puts "unknown option to %s" % [cmd[0]]
        end

      when 'eval'
        if cmd.size > 1
          (_,s) = inp.split(/\s+/,2)
          puts eval(s).inspect
        else
          in_eval = true
          while in_eval
            print "eval> "; s = gets.strip
            begin
              if s == 'Q' || s == 'q'
                in_eval = false
              else
                puts eval(s).inspect
              end
            rescue Exception => e
              puts "error in eval: #{e}"
            end
          end
        end

      when '@','='
        begin
          r = asint(cmd[1])
          c = asint(cmd[2])
        rescue ArgumentError => err
          puts "invalid argument for @"
          next
        end

        begin
          s = asint(cmd[3])
        rescue ArgumentError => e
          s = cmd[3].to_sym
        end

        chm_en0 = driver.energy
        fsm_en0 = model_energy
        if cmd[0] == '@'
          driver.change_spin(r,c,s)
        else
          driver[r,c] = s
        end
        chm_en1 = driver.energy
        fsm_en1 = model_energy

        print_lattice
        puts <<-eos
chem: #{chm_en1} - #{chm_en0} = #{chm_en1-chm_en0}
fsm : #{fsm_en1} - #{fsm_en0} = #{fsm_en1-fsm_en0}
eos

      when '@@'
        begin
          dimer = asint(cmd[1])
          pf    = asint(cmd[2])
        rescue ArgumentError => err
          puts "invalid argument for @"
          next
        end

        if dimer <= 0
          dimer = dimer.abs
          rr = [ 4*dimer+1, 4*dimer+2, 4*dimer+3, 4*dimer+4 ]
        else
          dimer -= 1
          rr = [ 4*dimer+1, 4*dimer+2, 4*dimer+3, 4*dimer+0 ]
        end


        cc = 2*pf

        case cmd[3]
        when 'S'
          mers = [ :S, :S, :S, :S ]
        when 's'
          mers = [ :S, :S, :s, :S ]
        when 'R', 'r'
          mers = [ :R, :B, :R, :B ]
        when 'L','l'
          mers = [ :L, :B, :L, :B ]
        when 'C'
          mers = [ :C, :C, :C, :C ]
        when 'c'
          mers = [ :C, :C, :c, :C ]
        else
          puts "Unknown argument: #{cmd[3]}"
        end

        rr.zip(mers).each do |r,s|
          puts "#{r},#{cc} = #{s}"
          @driver[r,cc] = s
        end

      when 'reg'
        case cmd.size
        when 1
          puts <<-eos
usage: reg <R> [<value>]

       if <R> is '*', displays all currently set registers

       if <value> is omitted displays the value of register R

       if <value> is '-', clears register R, otherwise sets register R
       to the given value, which must be an integer (and can be given by
       another register)
          eos
        when 2
          reg = cmd[1]
          if reg == '*'
            puts "current registers:"
            @registers.keys.each_slice(5) { |sl|
              puts sl.join("  ")
            }
          else
            puts "register #{reg} is #{@registers[reg]}"
          end
        else
          reg = cmd[1]
          set_register(reg,cmd[2...cmd.size])
        end

      when 'rec'
        @record = nil
        if cmd.size > 1
          @record=cmd[1]
          puts "Recording to '%s'" % record
          @recordings[@record] = []
        end

      when 'listrec'
        rn = cmd[1]
        if rn.nil?
          puts "Recordings"
          puts @recordings.keys.sort.each_slice(5).map{|sl| sl.join("  ")}.join("\n")
        elsif ! @recordings[rn]
          puts "No recording '#{rn}'"
        else
          puts "--- #{rn} ---"
          @recordings[rn].each_with_index do |l,i|
            puts "[%3d] %s" % [i,l]
          end
          puts
        end

      when 'saverecs'
        save_recordings(cmd[1])

      when 'loadrecs'
        load_recordings(cmd[1])

      when 'saverec'
        rec = cmd[1]
        fname = cmd[2]
        if rec.nil? || fname.nil?
          puts 'usage: saverec <rec> <file>'
        elsif @recordings[rec].nil?
          puts "no recording '#{rec}'"
        else
          File.open(fname,"w") { |f| f.puts(@recordings[rec].join("\n")) }
        end

      when 'loadrec'
        rec = cmd[1]
        fname = cmd[2]
        if rec.nil? || fname.nil?
          puts 'usage: loadrec <rec> <file>'
        else
          @recordings[rec] = File.open(fname,"r") do |f|
            f.read.split("\n")
          end
        end

      when 'savehist'
        fname = cmd[1]
        if fname.nil?
          puts 'usage: savehist <file>'
        else
          File.open(fname,"w") { |f| f.puts(@history.join("\n")) }
        end

      when 'rr'
        if cmd[1] == 'c'
          puts 'using cached'
          rr = @driver.cached_row_ranges
        else
          puts 'finding row range'
          rr = @driver.find_row_range
        end

        for i in 0...rr.size do
          puts "[%3d] %4d %4d" % [ i, rr[i][0], rr[i][1] ]
        end

      when 'play'
        if @record
          puts "recording '#{@record}', skipping immediate playback"
        else
          playback(cmd.drop(1))
        end

      when 'tube'
        puts "tubes: #{model.tube_last_rows.join(' ')}"

      when 'e','energy'
        puts <<-eos
chem: #{driver.energy}
fsm : #{model_energy}
eos
      when 'rows'
        if cmd.size>1
          print_row_energies(cmd[1].to_i)
        else
          print_row_energies
        end
      when 'q!'
        Kernel.exit(0)
      else
        c = @extra_cmds[cmd[0]]
        if c
          c.call(self,cmd)
        end
      end

      list_emarks if @emarks_auto
    end

  end

end

