require 'tubesheet'
# require 'pry'

module Tubesheet
  class ChemDriver
    attr_accessor :coupling
  end

  class MechanicalEnergyEvaluation
    attr_accessor :orig_dof, :orig_energy

    def initialize(coupling, rv,cv,stv)
      @coupling = coupling
      @mech = coupling.sim
      @chem = coupling.driver

      @rv = rv
      @cv = cv
      @stv = stv

      @orig_stv = nil

      @orig_dof = nil
      @orig_energy = nil

      @new_energy = nil

      @mech_temp = nil

      save_orig_state
    end

    def save_orig_state
      @orig_stv = Array.new(@rv.size)
      for i in 0...@rv.size
        @orig_stv[i] = @coupling.state_at @rv[i],@cv[i]
      end

      @coupling.current_energy ||= @coupling.calculate_mechanical_energy
      @orig_energy = @coupling.current_energy

      @coupling.current_dof ||= @coupling.get_mechanical_dof
      @orig_dof = @coupling.current_dof
    end

    def calculate_new_energy
      if @coupling.print_energies

        @coupling.sim.calculate_energy
        @coupling.sim.print_summary
      end
      set_bonds! @stv
      @new_energy = @coupling.calculate_mechanical_energy
      @coupling.sim.calculate_energy
      @coupling.sim.print_summary if @coupling.print_energies
      return @new_energy
    end

    def change_bonds?
      calculate_new_energy unless @new_energy
      dE = @new_energy - @orig_energy
      accept = @coupling.accept_energy_change? dE

      if @coupling.print_energies
        puts "E0 = %7.4f  E1 = %7.4f  dE = %7.4f  accept? %s" % [ 
          @orig_energy,
          @new_energy,
          dE,
          accept.to_s
        ]
      end

      return accept
    end

    def set_bonds!(stv)
        rv = @rv
        cv = @cv
        for i in 0...rv.size do
          @coupling.change_bond_type(rv[i],cv[i],stv[i])
          @coupling.do_minimization
        end
    end

    def accept_state!
      if @new_energy
        @coupling.current_energy = @new_energy
        @coupling.current_dof = @coupling.get_mechanical_dof
      else
        throw RuntimeError.new "new_energy or new_dof is nil!"
      end
    end

    def reject_state!
      set_bonds! @orig_stv
      @coupling.set_mechanical_dof @orig_dof
      @new_energy = nil
    end

    def evaluate!
      if change_bonds?
        accept_state!
        return true
      else
        reject_state!
        return false
      end
    end

    def self.evaluate(coupling, rv,cv, stv)
      evaluation = MechanicalEnergyEvaluation.new coupling, rv,cv, stv
      evaluation.evaluate!
    end

  end

  class ChemCoupling
    FE_ACCEPTANCE = :metropolis

    IFC_TUBELIKE  = ::Tubesheet::ChemDriver::INTERFACE_TUBELIKE
    IFC_SHEETLIKE = ::Tubesheet::ChemDriver::INTERFACE_SHEETLIKE
    IFC_NONE      = ::Tubesheet::ChemDriver::EMPTY

    attr_accessor :fe_acceptance
    attr_accessor :mech_temp

    attr_accessor :current_energy, :current_dof

    attr_reader :driver, :builder, :sim
    attr_reader :print_energies

    attr_accessor :debug_move_hook

    def initialize(driver,builder)
      @driver = driver
      @builder = builder
      @sim = @builder.sim

      @state_mapping = {}

      @linked = false

      @fe_acceptance = FE_ACCEPTANCE

      @current_energy = nil
      @current_dof = nil

      @last_count = 0
      @last_rejections = 0

      @count = 0
      @rejections = 0

      @debug_move_hook = false
      @print_energies = false
    end

    def minimization_params=(params)
      builder.minimization_params = params
    end

    def minimization_params
      builder.minimization_params
    end

    def energy_type
      return @builder.energy_type
    end

    def energy_type=(type)
      @builder.energy_type = type
    end

    def get_mechanical_dof
      sim.update_rigid_bodies
      return sim.dof_vec
    end

    def set_mechanical_dof(states)
      sim.set_dof_vec(states)
      sim.update_rigid_bodies
    end

    def acceptance_ratio
      new_count = @count - @last_count
      new_rej = @rejections - @last_rejections

      all_acc = @count - @rejections
      new_acc = new_count - new_rej

      new_ratio = (new_count > 0) ? new_acc.to_f / new_count : 0.0
      all_ratio = (@count > 0) ? all_acc.to_f / @count : 0.0

      @last_count = @count
      @last_rejections = @rejections

      [new_ratio,all_ratio,[new_acc,new_count],[all_acc,@count]]
    end

    def add_state_mapping(site_type, state_bondtype_map)
      @state_mapping[site_type] ||= {}
      @state_mapping[site_type].merge! state_bondtype_map
    end

    def get_state_mapping(site_type, state)
      if state == :" "
        return IFC_NONE 
      else
        map = @state_mapping[site_type]
        return map && map[state]
      end
    end

    def build_mechanical
      0.step(@driver.ncol-1,2) do |c|
        pf = 1 + (c/2)
        nmers = 0
        1.step(@driver.nrow-1,2) do |r|
          if @driver[r,c] != 0
            nmers += 1
          end
        end

        if nmers > 0
          delta = nmers - @builder.protofilaments[pf-1].length

          if delta < 0
            raise RuntimeError.new 'lattice monomers < mechanical monomers'
          elsif delta > 0
            @builder.new_monomers \
              :protofilament => pf,
              :quantity      => delta
          end
        end
      end


      0.step(@driver.ncol-1,2) do |c|
        1.step(@driver.nrow-1,2) do |r|
          mind = (r-1)/2
          if (@driver[r,c] != 0)
            if (@driver[r+1,c] != 0)
              change_bond_type(r+1,c)
            end

            if (c > 0) && (@driver[r,c-1] != 0)
              change_bond_type(r,c-1)
            end
          end
        end
      end
    end

    def monomer_associated_with_site(r,c)
      if @driver.site_type(r,c) != :MONOMER
        raise RuntimeError, "site %d,%d is not a monomer site" % [r,c]
      end
      mer_ind = (r-1)/2
      pf_ind = (c/2)
      pf = @builder.protofilaments[pf_ind]
      if pf.nil?
        raise RuntimeError, "site %d,%d has no matching protofilaments entry"
      end
      m_ind = pf[mer_ind]
      if m_ind.nil?
        raise RuntimeError, "site %d,%d has no matching monomer"
      end
      return @builder.get_monomer(m_ind)
    end

    def change_bond_type(r, c, site_state=nil)
      site_state ||= ChemDriver::char_value(@driver[r,c]).to_sym
      site_type  = @driver.site_type(r, c)
      case site_type
      when :LONGIFC
        mer1 = monomer_associated_with_site(r-1,c)
        mer2 = monomer_associated_with_site(r+1,c)
      when :LATIFC
        (r0,c0) = @driver.wrap_coords(r,c-1)
        (r1,c1) = @driver.wrap_coords(r,c+1)
        mer1 = monomer_associated_with_site(r0,c0)
        mer2 = monomer_associated_with_site(r1,c1)
      else
        return
      end

      bond_type = get_state_mapping site_type, site_state
      if bond_type.nil?
        err = "no valid bond type for [%s, '%s'] at [%3d,%3d]" % [site_type, site_state,r,c]
        raise RuntimeError.new(err)
      end

      bt_new = @builder.bond_types[bond_type]
      (p1,p2,bt_old) = mer1.get_bond_info(mer2).first

      return if bt_old == bt_new

      if p1.nil?
        if site_type == :LATIFC
          pf0 = (c0/2).floor
          pf1 = (c1/2).floor
          @builder.connect_lateral( mer1.ind, mer2.ind, [pf0,pf1] )
        else
          p1 = mer1.com_particle
          p2 = mer2.com_particle
          p1.add_bond p2,bt_new
        end
      elsif bt_new.nil?
        p1 = sim.get_particle(p1)
        p2 = sim.get_particle(p2)
        p1.remove_bond p2
      else
        @sim.change_bonds( [p1,p2, bt_new] )
      end

      @sim.build_potential
    end

    def do_minimization
      params = minimization_params || {}
      @builder.do_minimization(params)
    end

    def calculate_mechanical_energy
      @builder.calculate_mechanical_energy
    end

    def rng_uniform
      return @driver.rng.random(1)[0]
    end

    def accept_energy_change?(dE)
      acc = fe_acceptance
      if @mech_temp 
        dtemp = @driver.temp
        if dtemp != @mech_temp
          ratio = dtemp / @mech_temp
          dE *= ratio
        end
      end
      prob = @driver.prob_acceptance(acc, dE)
      if prob >= 1.0
        return true
      elsif prob <= 0.0
        return false
      else
        return rng_uniform <= prob
      end
    end

    def state_at(r,c)
      ::Tubesheet::ChemDriver::char_value(@driver[r,c]).to_sym
    end

    def flip_lateral_state(st0)
      return IFC_SHEETLIKE if st0 == IFC_TUBELIKE
      return IFC_TUBELIKE  if st0 == IFC_SHEETLIKE

      raise RuntimeError.new \
        "lateral sites at (%3d,%3d) and (%3d,%3d) have unknown state %d" % \
        [r0,c, r1,c]
    end

    def move_allowed?(rv,cv,stv)
      mech_eval = MechanicalEnergyEvaluation.new(self, rv,cv, stv)
      mech_eval.calculate_new_energy
      case fe_acceptance
      when :always_allow
        mech_eval.accept_state!
        return true
      when :always_reject
        mech_eval.reject_state!
        return false
      else
        return mech_eval.evaluate!
      end
    end

    def lateral_rotation_hook(r0,c)
      r1 = r0+2
      st0 = @driver[r0,c]
      if st0 != @driver[r1,c]
        raise RuntimeError.new \
          "lateral sites at (%3d,%3d) and (%3d,%3d) have different states" % \
          [r0,c, r1,c]
      end

      st1 = flip_lateral_state(st0)

      st0 = ::Tubesheet::ChemDriver::char_value(st0).to_sym
      st1 = ::Tubesheet::ChemDriver::char_value(st1).to_sym

      rv  = [r0 ,r1 ]
      cv  = [c  ,c  ]
      stv = [st1,st1]

      allowed = move_allowed?(rv,cv,stv)

      return allowed
    end

    def lateral_change_hook(r0,c, st1)
      r1 = r0+2
      st0 = @driver[r0,c]
      if st0 != @driver[r1,c]
        raise RuntimeError.new \
          "lateral sites at (%3d,%3d) and (%3d,%3d) have different states" % \
          [r0,c, r1,c]
      end

      st0 = ::Tubesheet::ChemDriver::char_value(st0).to_sym
      st1 = ::Tubesheet::ChemDriver::char_value(st1).to_sym

      rv  = [r0 ,r1 ]
      cv  = [c  ,c  ]
      stv = [st1,st1]

      allowed = move_allowed?(rv,cv,stv)

      return allowed
    end

    def lateral_seam_hook(r0,c)
      r1 = r0 + 2
      case @driver[r0,c]
      when IFC_NONE
        st = IFC_TUBELIKE
      when IFC_TUBELIKE
        st = IFC_NONE
      else
        throw RuntimeError.new "Invalid seam state"
      end
      st = ::Tubesheet::ChemDriver::char_value(st).to_sym

      rv  = [r0,r1]
      cv  = [c ,c ]
      stv = [st,st]

      move_allowed?(rv,cv,stv)
    end

    def simple_move_hook(r,c,st)
      st_sym = ::Tubesheet::ChemDriver::char_value(st).to_sym
      move_allowed?([r],[c],[st_sym]) 
    end

    def move_hook(r,c,st)
      return true if @driver.site_type(r,c) == :MONOMER

      @count += 1

      allowed = false

      case st
      when ChemDriver::PSEUDO_LATERAL_ROTATION
        allowed = lateral_rotation_hook r,c

      when ChemDriver::PSEUDO_LATERAL_SHEETLIKE
        allowed = lateral_change_hook r,c, IFC_SHEETLIKE

      when ChemDriver::PSEUDO_LATERAL_TUBELIKE
        allowed = lateral_change_hook r,c, IFC_TUBELIKE

      when ChemDriver::PSEUDO_LATERAL_SEAM
        allowed = lateral_seam_hook r,c

      when ChemDriver::PSEUDO_LATERAL_ZERO
        allowed = lateral_change_hook r,c, IFC_NONE

      else
        if st < Tubesheet::Lattice::PSEUDO_OFFSET
          allowed = simple_move_hook r,c,st
        else
          puts "Unknown PSEUDOSPIN #{st}, rejecting!"
        end
      end

      @rejections += 1 if !allowed
      return allowed || :cancel
    end

    def link
      build_mechanical
      @driver.add_move_hook do |latt,r,c,st|
        move_hook r,c, st
      end

      @current_dof = nil
      @current_energy = nil
      @linked = true
    end

    def linked?
      return @linked
    end

  end
end
