require 'tubesheet'
require 'tubulin'
require 'yaml'

REACTION_INDICES = {
  :BT_LONG_STRAIGHT => 1,
  :BT_LAT_TUBE      => 2,
  :SEAM             => 3,
}


module Tubulin
  class MechFlipper < SimulationBuilder
    attr_accessor :minimize_niters, :first_minimize
    attr_accessor :seam_cutoff, :seam_enabled
    attr_accessor :initial_state

    extend KeywordArguments

    def initialize(params)
      params[:logging] ||= {}
      params[:logging][:nonbonds]  ||= false
      params[:logging][:minimizer] ||= false

      super(params)

      self.first_minimize  = params[:first_minimize] || false
      self.minimize_niters = params[:minimize_niter] || params[:minimize_niters] || 1000
      self.seam_cutoff     = params[:seam_cutoff]    || 1.5
      self.seam_enabled    = params[:seam_enabled]   || false
    end

    def self.from_yaml(data, build_initial=true, &announce_block)
      builder = super data
      builder.build_initial_structure if build_initial
      announce_block.call(builder) if announce_block
      return builder
    end

    def self.construct_sheet(orig_params, &announce_block)
      self.construct(orig_params,&announce_block)
    end

    def self.construct(orig_params, &announce_block)
      params = orig_params.clone

      (num_mers, num_pfs) = self.requiredKeywords params, :num_mers,:num_pfs
      (job, seed0, seed_incr) = self.requiredKeywords params, :job, :seed0, :seed_incr

      params.delete :num_mers
      params.delete :num_pfs
      params.delete :job
      params.delete :seed0
      params.delete :seed_incr

      announce = params[:announce]
      params.delete :announce

      params[:params] ||= {}

      max_monomers = num_mers * num_pfs
      params.merge! \
        :seed0              => seed0           ,
        :seed_incr          => seed_incr       ,
        :job_number         => job             ,
        :max_monomers       => max_monomers    ,
        :num_protofilaments => num_pfs

      builder = self.new params

      builder.build_initial_structure

      announce.call(builder) if announce

      return builder
    end

    def build_initial_structure
      super
      @initial_state = sim.dof_vec
      self.minimize_initial_state if self.first_minimize
    end

    def minimize_initial_state
      puts "Minimizing...\n"
      is_logging_minimizer = sim.is_logging? :info, :minimizer
      sim.set_logging :minimizer, true
      STDOUT.flush
      sim.build_potential

      min_initial = self.minimization_params.clone
      min_initial.delete :nsteepest
      min_initial.delete :nrefine
      min_initial.delete :steepest_epsilon
      min_initial.delete :refine_epsilon

      if (!min_initial[:epsilon] || (min_initial[:epsilon] > 1e-5))
        min_initial[:epsilon] = 1e-5
      end

      if (!min_initial[:num_iterations] || (min_initial[:num_iterations] < 10000))
        min_initial[:num_iterations] = 10000
      end

      min_initial[:nsteepest] = 500
      min_initial[:steepest_epsilon] = 0.01

      self.do_minimization( min_initial )
      puts '... minimization done'
      sim.set_logging :minimizer, is_logging_minimizer
      sim.update_rigid_bodies
      sim.calculate_forces
      puts
      puts "Gradient magnitudes: forces = %7.4e, torques = %7.4e" % sim.gradient_magnitude
      puts "Gradient magnitudes: fmin   = %7.4e, fmax    = %7.4e" % sim.gradient_magnitude[2..3]
      puts
      STDOUT.flush

      sim.update_rigid_bodies
    end

    def write_model(pdb_f,type=:pdb)
      @pdb_ind ||= 0
      @pdb_ind += 1
      case type
      when :pdb
        pdb_f.puts "MODEL %d" % [@pdb_ind]
        sim.write_pdb(pdb_f)
        pdb_f.puts "ENDMDL"
      when :xyz
        sim.write_xyz pdb_f, "step #{@pdb_ind}"
      else
        raise RuntimeError.new "unknown ouput format #{type.inspect}"
      end
    end

    def report_energies(dimer,pf,energies=nil)
      if energies.nil?
        energies = self.calculate_mechanical_energy :components => true
      end

      puts 'DIMER %4d  PF %4d  G %12.7f  H %12.7f  TS %12.7f' % [
        dimer, pf,
        energies[:free],
        energies[:enthalpic],
        energies[:entropic]
      ]
    end

    def report_sim_energy(file=STDOUT)
      sim.update_rigid_bodies
      energies = self.calculate_mechanical_energy :components => true

      file.puts 'Free energy: %12.7f  Enthalpy: %12.7f  Entropy: %12.7f' % \
        [ energies[:free], energies[:enthalpic], energies[:entropic] ]
    end

    def [](mer, pf)
      get_monomer( protofilaments[pf][mer] )
    end

    def do_long_flip(mer1, mer2, new_bt)
      (p1,p2,bt) = mer1.get_bond_info(mer2).first

      if bt[0] == :back
        sim.change_bonds( [p2,p1, bond_types[new_bt]] )
      else
        sim.change_bonds( [p1,p2, bond_types[new_bt]] )
      end
    end

    def flip_single_lateral_bond(mer1, mer2)
      (p1,p2,bt) = mer1.get_bond_info(mer2).first
      if bt && (bt[1] == bond_types[:BT_LAT_SHEET])
        sim.change_bonds( [p1,p2, bond_types[:BT_LAT_TUBE]] )
        true
      else
        false
      end
    end

    def do_lat_flip(ma1, ma2, mb1, mb2 )
      b1 = flip_single_lateral_bond(ma1,ma2)
      b2 = flip_single_lateral_bond(mb1,mb2)
      b1 && b2
    end

    def create_seam_bonds(max_row)
      new_seam = []
      last_pf = num_protofilaments-1
      max_monomer = 2*(max_row+1)
      for i in 0 ... max_monomer
        off = HELICAL_PITCH_NUM_MONOMERS
        mer1 = self[i,0]
        mer2 = self[i+off,last_pf]

        if mer1.ind && mer2.ind
          (p1,p2,bt) = mer1.get_bond_info(mer2).first

          if bt.nil?
            p1 = mer1.com_particle
            p2 = mer2.beta

            dist = p1.distance_to p2
            if dist < seam_cutoff
              puts "SEAM forming: %3d dist = %7.4f, cutoff = %7.4f\n" % [i,dist,seam_cutoff]
              sim.add_bonds [p1.index,p2.index, bond_types[:BT_LAT_TUBE]]
              new_seam << [i, mer1, mer2]
            else
              puts 'SEAM monomer row %3d dist %7.4f' % [i,dist]
            end
          end
        end
      end

      return new_seam
    end

    def output_flip(files, m1,m2, bt)
      sim.build_potential

      do_minimization

      write_model(files[:pdb],:pdb) if files[:pdb]
      write_model(files[:xyz],:xyz) if files[:xyz]

      energies = self.calculate_mechanical_energy :components => true

      rxn_index = REACTION_INDICES[bt] || -1

      files[:ener].puts '%4d %4d %2d %12.7f %12.7f %12.7f' % [ 
        m1.ind, m2.ind,
        rxn_index,
        energies[:free],
        energies[:enthalpic],
        energies[:entropic]
      ]

      return energies
    end

    def report_seam(files, new_seam)
      seam_rows = new_seam.map(&:first)
      puts "SEAM: flipped rows #{seam_rows.join(', ')}"
      (_,m1,m2) = new_seam.last

      output_flip(files, m1,m2, :SEAM)
    end

    def try_creating_seam(files,max_dimer_row,energies)
      return energies if !seam_enabled

      puts 'TRYING TO BUILD SEAM'
      new_seam = create_seam_bonds(max_dimer_row)
      if ! new_seam.empty?
        energies = report_seam(files,new_seam)
      end
      energies
    end

  end
end

