require 'sqlite3'
require 'zlib'
require 'tubesheet'

if Tubesheet::is_embedded?
  require 'md5'
  def md5(s)
    MD5::md5_hex(s)
  end
else
  require 'digest'
  def md5(s)
    Digest::MD5.hexdigest(s)
  end
end

module ParamLib
  class SubstitutionError < RuntimeError
    attr_reader :missing

    def initialize(*missing_keys)
      if Array === missing_keys[0]
        @missing = missing_keys[0]
      else
        @missing = missing_keys
      end
      super("template expected and was not provided keys: #{@missing.join(", ")}")
    end
  end

  attr_accessor :which_process

  def expected_sim_name
    File.basename(File.dirname(File.absolute_path($0)))
  end

  def check_job_name
    STDERR.print "Checking job name ... "
    expected_jobname = expected_sim_name()
    jobname = get_jobname
    if jobname != expected_jobname
      STDERR.puts "failed"
      puts <<-EOS
JOBNAME is #{jobname}, but should be #{expected_jobname}
Stopping!
      EOS
      exit 1
    end

    check = md5(jobname)
    if check != SEED0_PREFIX_MD5
      STDERR.puts "failed"
      puts "You need to update SEED0!"
      puts "Update config.rb:"
      seed_now = (Time.now - Time.at(0)).round
      puts <<-EOS
SEED0=#{seed_now}
SEED0_PREFIX_MD5="#{check}"
      EOS
      exit 1
    end

    STDERR.puts "OK"
  end

  def check_job_sha
    STDERR.print "Checking job SHA ... "

    sim_sha = Tubesheet::VERS_SHA1
    begin
      repo=REPO_DIR
      if repo && repo[0] == '~' && repo[1] == '/'
        repo = ENV['HOME'] + repo[1..-1]
      end
      repo = File.absolute_path(repo)
      git_head = File.open("#{repo}/.git/HEAD") { |f| f.read.chomp }
      case git_head
      when /^ref: (.+)/
        fn = $1
        repo_sha = File.open("#{repo}/.git/#{fn}") { |f| f.read.chomp }
      when /^([0-9a-zA-Z]+)$/
        STDERR.puts "WARNING! git repository has a detached HEAD"
        repo_sha = $1
      else
        STDERR.puts "ERROR! Unrecognized git repository HEAD (neither ref nor SHA)"
        Kernel.exit 1
      end
    rescue Exception => e
      STDERR.puts <<-eos
Error reading SHA from repository #{REPO_DIR}:
#{e.to_s}
Backtrace:
#{e.backtrace.join("\n")}
      eos
      Kernel.exit 1
    end

    if EXPECTED_SHA == ""
      STDERR.puts "failed"
      puts <<-EOS
EXPECTED_SHA in config.rb is empty!
Please update it to match the repository SHA and simulation SHA:
SIM_SHA  #{sim_sha}
REPO_SHA #{repo_sha}

SIM_PROG #{SIM_PROG}
REPO_DIR #{REPO_DIR}
      EOS
      Kernel.exit 1
    elsif (EXPECTED_SHA != sim_sha) || (EXPECTED_SHA != repo_sha)
      STDERR.puts "failed"
      puts <<-EOS
EXPECTED_SHA in config.rb doesn't match the simulation SHA or the repo SHA
Please update it, or fix the repo/sim
EXPECTED_SHA #{EXPECTED_SHA}
SIM_SHA      #{sim_sha}
REPO_SHA     #{repo_sha}

SIM_PROG     #{SIM_PROG}
REPO_DIR     #{REPO_DIR}
      EOS
      Kernel.exit 1
    end

    STDERR.puts "OK"
  end

  def ind2sub(dims,ind0)
    subs = Array.new(dims.size)
    ind = ind0
    dims.each_with_index do |sz,di|
      subs[di] = ind % sz
      ind = ind / sz
    end
    return subs
  end

  def check_max_task
    STDERR.print "Checking that MAX_TASK agrees with number of tasks ... "
    param_sets = @param_sets
    if param_sets.nil?
      param_sets = PARAM_SETS
    end

    if TASKS_PER_SET*param_sets.size != MAX_TASK
      STDERR.puts "failed"
      puts <<-EOS
You need to update MAX_TASK:
MAX_TASK=#{TASKS_PER_SET*param_sets.size}
      EOS
      exit 1
    end
    STDERR.puts "OK"
  end

  def write_param_list(fn)
    if @param_names.nil?
      param_names = PARAM_NAMES
      param_sets = PARAM_SETS
    else
      param_names = @param_names
      param_sets = @param_sets
    end

    File.open(fn,"w") do |f|
      f.write("%12s %6s %8s " % %w(scanname set task))
      f.puts(param_names.map { |name| "%16s" % [name.to_s] }.join(" "))
      task=0
      param_sets.each_with_index do |phash,pind|
        pset=pind+1
        (1..TASKS_PER_SET).each do
          task += 1
          f.write("%12s %6d %8d " % [phash[:scanname],pset,task])
          f.puts(param_names.map { |name| "%16.10f" % [ phash[name] ] }.join(" "))
        end
      end
    end
  end

  def transact(db,&block)
    db.transaction
    succ = false
    begin
      block.call db
      db.commit
      succ = true
    ensure
      db.rollback unless succ
    end
  end

  def write_param_db(fn)
    if @param_names.nil?
      param_names = PARAM_NAMES
      param_sets = PARAM_SETS
    else
      param_names = @param_names
      param_sets = @param_sets
    end

    # Avoid locking
    begin
      STDERR.puts "Writing parameters to database #{fn}"
      paramdb = SQLite3::Database.new "file:#{fn}?nolock=1" # , {}, "unix-none"

      paramdb.execute_batch <<-SQL
DROP TABLE IF EXISTS param_files;
DROP TABLE IF EXISTS jobs;
DROP TABLE IF EXISTS sets;
      SQL

      paramdb.execute_batch <<-SQL
CREATE TABLE sets (
  set_id INTEGER PRIMARY KEY,
  scanname TEXT,
  #{param_names.map{ |pn| "#{pn.to_s} REAL" }.join(",\n")}
);
      SQL

      paramdb.execute_batch <<-SQL
CREATE TABLE jobs (
  job_id INTEGER PRIMARY KEY,
  set_id INTEGER REFERENCES sets(set_id),
  seed INTEGER UNIQUE
);
      SQL

      paramdb.execute_batch <<-SQL
CREATE TABLE param_files (
  job_id INTEGER REFERENCES jobs(job_id),
  compr_yaml BLOB
);
      SQL

      while false do
        print "> "
        ll = gets
        break unless ll
        ll.chomp!

        break if ll == '@q'
        Kernel.exit if ll == '!!!'

        begin
          puts(eval(ll))
        rescue => e
          puts "Exception: #{e}\n#{e.backtrace.join("\n")}\n"
        end
      end

      stmt = <<-SQL
INSERT INTO sets
(
  set_id,
  scanname,
  #{param_names.map { |pn| "#{pn.to_s}" }.join(",\n")}
)
VALUES
(
  ?, ?, #{param_names.map { '?' }.join(', ')}
);
      SQL

      case param_sets.size
      when   0..100  then report_freq = 1
      when 100..1000 then report_freq = 10
      else report_freq = 100
      end

      transact(paramdb) do |db|
        STDERR.puts "Inserting #{param_sets.size} parameter records"
        param_sets.each_with_index do |phash,pind|
          args = [ pind+1, phash[:scanname] ]
          param_names.each { |k| args << phash[k] }
          db.execute_batch(stmt, *args)
          STDERR.print( "%4d / %4d\r" % [ pind+1, param_sets.size ]) if (pind+1)%report_freq == 0
        end
        STDERR.puts "\nDONE"
      end

      stmt = <<-SQL
INSERT INTO jobs
(
  job_id, set_id, seed
)
VALUES 
(
  ?, ? ,?
);
      SQL

      transact(paramdb) do |db|
        STDERR.puts "Inserting #{TASKS_PER_SET} job records for each parameter set"
        param_sets.each_with_index do |phash,pind|
          pset=pind+1
          ind0 = TASKS_PER_SET*pind+1
          ind1 = TASKS_PER_SET*(pind+1)
          for i in (ind0 .. ind1) do
            db.execute_batch(stmt, i, pset, get_seed(i))
          end
          STDERR.print( "Set %4d / %4d\r" % [ pind+1, param_sets.size ]) if (pind+1)%report_freq == 0
        end
        STDERR.puts "\nDONE"
      end

      if @template
        stmt = <<-SQL
INSERT INTO param_files ( job_id, compr_yaml )
VALUES ( ?, ? );
        SQL

        @preprocessed_template = preprocess_template(@template)
        transact(paramdb) do |db|
          STDERR.puts "Inserting #{TASKS_PER_SET} parameter files for each parameter set"
          param_sets.each_with_index do |phash,pind|
            pset=pind+1
            ind0 = TASKS_PER_SET*pind+1
            ind1 = TASKS_PER_SET*(pind+1)
            for i in (ind0 .. ind1) do
              yaml_s = process_template(@template, phash, pind, i)
              cyaml = Zlib.deflate(yaml_s)
              db.execute_batch(stmt, i, cyaml)
            end
            STDERR.print( "Set %4d / %4d\r" % [ pind+1, param_sets.size ]) if (pind+1)%report_freq == 0
          end
          STDERR.puts "\nDONE"
        end
      end

    ensure
      paramdb.close if paramdb
    end
  end

  def get_seed(i)
    SEED0 + i*SEED_INCR
  end

  PARAMLIB_TEMPLATE_RE = /(?<!@)@([\w$.]+)/
  def preprocess_template(tmpl)
    re = PARAMLIB_TEMPLATE_RE
    preprocessed = []

    last_ind = 0

    m = re.match(tmpl,last_ind)
    while m do
      (i0,i1) = m.offset(0)

      if i0 > last_ind
        preprocessed << tmpl[last_ind...i0]
      else
        preprocessed << nil
      end
      preprocessed << m[1].to_sym

      last_ind = i1
      m = re.match(tmpl,last_ind)
    end
    preprocessed << tmpl[last_ind ... tmpl.size] if last_ind < tmpl.size

    preprocessed
  end

  def substitute_template(preprocessed, phash, pind, i)
    seed = get_seed(i)

    outs = []
    missing_keys = {}

    preprocessed.each_slice(2) do |pre,key|
      outs << pre

      case key
      when :JOBNAME
        outs << get_jobname
      when :TASK, :i
        outs << i.to_s
      when :SET, :p
        outs << pind.to_s
      when :SEED, :seed
        outs << seed.to_s
      when nil
        # skip
      else
        if phash.key? key
          outs << "#{phash[key]}"
        elsif phash.key? key.to_s
          outs << "#{phash[key.to_s]}"
        else
          missing_keys[key.to_s] = true
          outs << "<<MISSING:#{key}>>"
        end
      end
    end

    if ! missing_keys.empty?
      raise SubstitutionError.new(missing_keys.keys.sort)
    end

    outs.join('')
  end

  def get_jobname
    JOBNAME
  end

  def new_process_template(tmpl, phash, pind, i)
    preproc = @preprocessed_template || preprocess_template(tmpl)
    substitute_template(preproc, phash, pind, i)
  end

  def process_template(tmpl, phash, pind, i)
    case self.which_process
    when :old
      return old_process_template(tmpl,phash,pind,i)
    when :check
      puts "Checking new and old template processing"
      out_new = new_process_template(tmpl,phash,pind,i)
      out_old = old_process_template(tmpl,phash,pind,i)
      if out_old != out_new
        raise RuntimeError.new "old and new templates do not agree"
      end

      return out_new
    else
      return new_process_template(tmpl,phash,pind,i)
    end
  end

  def old_process_template(tmpl, phash, pind, i)
    seed = get_seed(i)
    re = PARAMLIB_TEMPLATE_RE
    outs = []
    tmpl_rest = tmpl

    last_ind = 0
    missing_keys = {}
    m = re.match(tmpl,last_ind)
    while m do
      (i0,i1) = m.offset(0)
      outs << tmpl[last_ind...i0] if i0 > last_ind

      case m[1]
      when 'JOBNAME'
        outs << get_jobname
      when 'TASK', 'i'
        outs << i.to_s
      when 'SET', 'p'
        outs << pind.to_s
      when 'SEED', 'seed'
        outs << seed.to_s
      else
        mname = m[1]
        if phash.key? mname
          outs << "#{phash[mname]}"
        elsif phash.key? mname.to_sym
          outs << "#{phash[mname.to_sym]}"
        else
          missing_keys[mname] = true
          outs << "<<MISSING:#{mname}>>"
        end
      end
      last_ind = i1
      m = re.match(tmpl,last_ind)
    end
    outs << tmpl[last_ind ... tmpl.size] if last_ind < tmpl.size

    if ! missing_keys.empty?
      raise SubstitutionError.new(missing_keys.keys.sort)
    end

    outs.join('')
  end

  def write_param_files(suffix='.yaml',&block)
    param_sets = @param_sets
    if param_sets.nil?
      param_sets = PARAM_SETS
    end

    param_sets.each_with_index do |phash, pind|
      ind0 = TASKS_PER_SET*pind+1
      ind1 = TASKS_PER_SET*(pind+1)
      for i in (ind0 .. ind1) do
        pfilename = "%s/%s_%d%s" % [OUTDIR,PARAM_FILE_PREFIX,i,suffix]
        File.open(pfilename,'w') do |f|
          if block
            yield(f,i,phash)
          elsif @template
            f.write(process_template(@template, phash, pind, i))
          end
        end
      end

      if (param_sets.size < 100) || ((pind % 100) == 0)
        print "Parameter %5d / %5d\r" % [pind+1,param_sets.size]
      end
    end

    puts "Created #{TASKS_PER_SET*param_sets.size} parameter files"
  end

  def parameters(names_and_defaults)
    raise RuntimeError.new "parameters have already been set" unless @param_names.nil?
    @param_names = []
    @param_defaults = {}
    names_and_defaults.each do |n,dft|
      case n
      when String
        n = n.to_sym
        @param_names << n
        @param_defaults[n] = dft
      when Symbol
        @param_names << n
        @param_defaults[n] = dft
      else
        raise RuntimeError.new "parameters must given a list of symbols or strings"
      end
    end

    @param_sets = []
    @total_sims = 0
    @const_params = {}
  end

  def const_parameters(names_and_defaults)
    @const_params = names_and_defaults
  end

  def linescans(name, param_table)
    param_names = param_table.keys

    if @param_names.nil?
      raise RuntimeError.new "parameter names must be set before a linescan or gridscan"
    end

    psets = []

    param_table.each do |k,vals|
      given_names = ([k] + @param_defaults.keys).uniq

      missing_names = @param_names - given_names
      unless missing_names.empty?
        raise RuntimeError.new "missing parameters: #{missing_names.sort.join(", ")}"
      end

      extra_names = given_names - @param_names
      unless extra_names.empty?
        raise RuntimeError.new "extra parameters: #{extra_names.sort.join(", ")}"
      end

      vals.each do |v|
        phash = Hash.new do |h,k|
          raise RuntimeError.new "missing key #{k}" 
        end

        phash[:scanname] = name
        phash.merge! @const_params
        phash.merge! @param_defaults
        phash[k] = v

        phash = yield phash if block_given?

        # block can return nil to cancel
        if phash
          psets << phash
        end
      end
    end

    @total_sims += psets.size
    @param_sets += psets
  end

  def tablescan(name, kw, &block)
    same_params = kw[:same] || {}
    params = kw[:params]
    values = kw[:values]

    if !(Hash === same_params)
      raise RuntimeError.new("table scan: same_params must be a Hash if given")
    end

    if params.nil? || !(Array === params)
      raise RuntimeError.new("table scan: missing params array")
    end

    if values.nil? || !(Array === values)
      raise RuntimeError.new("table scan: missing values array")
    end

    values.each_with_index do |vlist,ind|
      if vlist.size != params.size
        raise RuntimeError.new(
          "table scan: %d columns, but row %d has %d values" % [
            params.size, ind, vlist.size
        ])
      end

      ptable = {}
      ptable.merge!(same_params)

      params.each_with_index do |n,i|
        ptable[n] = vlist[i]
      end

      if block
        gridscan(name, ptable, &block)
      else
        gridscan(name, ptable)
      end
    end
  end

  def gridscan(name, param_table)
    param_names = param_table.keys
    param_table.keys.each do |k|
      v = param_table[k]
      case v
      when Array
        # nop
      when Range
        param_table[k] = v.to_a
      when String,Numeric
        param_table[k] = [v]
      else
        raise RuntimeError.new "Invalid parameter value for key #{k}: #{v}"
      end
    end

    num_sims = param_table.map{|k,v| v.size}.reduce(&:*)
    dims = param_names.map{|k| param_table[k].size}

    given_names = (param_names + @param_defaults.keys).uniq

    if @param_names.nil?
      raise RuntimeError.new "parameter names must be set before a linescan or gridscan"
    end

    missing_names = @param_names - given_names
    unless missing_names.empty?
      raise RuntimeError.new "missing parameters: #{missing_names.sort.join(", ")}"
    end

    extra_names = given_names - @param_names
    unless extra_names.empty?
      raise RuntimeError.new "extra parameters: #{extra_names.sort.join(", ")}"
    end

    psets = []

    for i in 1..num_sims do
      param_subs = ind2sub(dims,i-1)
      phash = Hash.new do |h,k|
        raise RuntimeError.new "missing key #{k}" 
      end

      phash[:scanname] = name
      phash.merge! @const_params
      phash.merge! @param_defaults
      param_names.zip(param_subs).each do |k,ki|
        phash[k] = param_table[k][ki]
      end

      phash = yield phash if block_given?

      # block can return nil to cancel
      if phash
        psets << phash
      end
    end

    @total_sims += psets.size
    @param_sets += psets
  end

  def template(s)
    @template = s
  end

  def template_file(fn)
    @template = File.open(fn) { |f| f.read }
  end

  def getme
    self
  end
end

$paramlib = true
