require 'sqlite3'

module TSUtils
  TS_SIM_START_TIME = Time.now

  unless Tubesheet.is_embedded?
    at_exit {
      sim_end_time = Time.now
      puts <<-EOS
  Simulation started #{TS_SIM_START_TIME}, finished #{sim_end_time},
    #{sim_end_time-TS_SIM_START_TIME} seconds
      EOS
    }
  end

  def die(mesg)
    STDERR.puts(mesg)
    Kernel.exit(1)
  end

  def driver_from_param_file(fname, klass=Tubesheet::ChemDriver)
    begin
      driver = File.open(fname) { |f| Tubesheet::ChemDriver.from_yaml(f.read) }
    rescue Exception => e
      die "Error loading parameter file #{PARAM_FILE}:\n    #{e}"
    end

    die "RNG not initialized!" if driver.rng.nil?

    return driver
  end

  def is_straight?(spin)
    (spin == MER_GTP_S) || (spin == MER_GDP_S)
  end

  def builder_from_param_file(fname, klass=Tubulin::SimulationBuilder)
    begin
      builder = File.open(fname) do |f| 
        klass.from_yaml(f.read)
      end # closes File.open block
    rescue Exception => e
      die <<-eos
Error loading parameter file #{fname}:
      #{e}
      #{e.backtrace.join("\n")}
      eos
    end
  end

  def write_dof_file(builder, fname)
    sim = builder.sim
    File.open(fname,"w") do |f|
      f.write(sim.dof_vec.pack('G*'))
    end
  end

  def write_bnd_file(builder, fname)
    sim = builder.sim
    rigid_map = builder.rigid_map

    rigid_bonds = sim.all_rigid_bonds

    monomer_bonds = rigid_bonds.map do |pi,ri,pj,rj,bt|
      [ rigid_map[ri], rigid_map[rj], bt ]
    end

    bond_data = monomer_bonds.flatten.pack('N*')
    File.open(fname,'w') do |f|
      f.write(bond_data)
    end
  end

  def read_conn(builder, fname)
    sim = builder.sim

    bond_data = File.open(fname) { |f| f.read }
    monomer_bonds = bond_data.unpack('N*')
    monomer_bonds = monomer_bonds.each_slice(3) do |mi,mj,bt|
      mer_i = builder.get_monomer(mi)
      mer_j = builder.get_monomer(mj)
      (pi,pj,_) = mer_i.get_bond_info(mer_j).first
      sim.change_bonds( [pi,pj, [bt]] )
    end
  end

  def read_dof(builder, fname)
    dofv = File.open(fname) { |f| f.read.unpack('G*') }

    raise RuntimeError.new "file has the wrong number of degrees of freedom" \
      if dofv.size != builder.sim.num_dof

    builder.sim.set_dof_vec dofv
    builder.sim.update_rigid_bodies
  end

  def read_yaml_data(fname)
    yaml_data = File.open(fname) { |f| f.read }

    if yaml_data[0...15] == 'SQLite format 3'
      raise RuntimeError.new "#{$param_yaml_file} is a SQLite 3 database, not a yaml config file"
    elsif yaml_data[0...2] == "\x1F\x8B"
      raise RuntimeError.new "#{$param_yaml_file} is compressed with gzip and must be decompressed to use"
    elsif yaml_data[0...2] == 'BZ'
      raise RuntimeError.new "#{$param_yaml_file} is compressed with bzip2 and must be decompressed to use"
    elsif yaml_data[0...6] == "\xFD7zXZ\0"
      raise RuntimeError.new "#{$param_yaml_file} is compressed with xz and must be decompressed to use"
    end

    
    return yaml_data
  end

  def with_sqlite(dbf, &b)
    db = SQLite3::Database.new(dbf)
    begin
      return b.call(db)
    ensure
      db.close
    end
  end

  def with_sqlite_results(db, stmt, *args)
    r = db.execute(stmt,*args)
    return unless r

    begin
      yield r
    ensure
      r.close
    end
  end

  def read_yaml_from_paramdb(db,jobid)
    yaml_data = nil
    with_sqlite("file:#{db}?nolock=1&mode=ro") do |paramdb|
      stmt = "SELECT compr_yaml FROM param_files WHERE job_id = ?"
      with_sqlite_results(paramdb, stmt, jobid) do |results|
        compr_data = results.next
        if compr_data && compr_data.first
          yaml_data = Zlib.inflate(compr_data.first)
        end
      end
    end

    return yaml_data
  end

  def read_yaml_chem_driver(args)
    case args.size
    when 1
      param_yaml_file = args[0]
      begin
        yaml_data = read_yaml_data(param_yaml_file)
      rescue => e
        die "Cannot read yaml file #{param_yaml_file}, aborting: #{e}"
      end

    when 2
      param_db_file = args[0]
      jobid         = args[1]
      die "Cannot read parameter database #{param_db_file}, aborting." \
        unless File.exists? param_db_file

      begin
        yaml_data = read_yaml_from_paramdb(param_db_file, jobid)
      rescue => e
        die "Error reading data from parameter db #{param_db_file}: #{e}"
      end

      if yaml_data.nil?
        die "Cannot find parameter for job #{jobid}"
      end
    else
      die <<-EOF
Usage: #{$0} <paramdb> <jobid>
   or: #{$0} <yaml_file>
      EOF
    end

    driver = Tubesheet::ChemDriver.from_yaml(yaml_data)
    die "RNG not initialized!" if driver.rng.nil?
    return driver
  end

end


