require 'tubesheet'
require 'angleutils'
require 'kwargs'
require 'yaml'
require 'enumerator' unless Tubesheet::is_embedded?

unless Hash.methods.include?(:freeze)
  class Hash
    def freeze
      return self
    end
  end
end

unless Array.methods.include?(:freeze)
  class Array
    def freeze
      return self
    end
  end
end

module Tubulin

  PARTICLES_PER_MONOMER = 5

  DEFAULT_PARAMS = {
    :num_protofilaments => 13,

    :temp => 1.0,
    :timestep => 0.0001,
    :rng_type => :mt19937,
    :integrator => :SRK,

    :collfreq => 100.0,
    :rotfric  => 10.0,

    :monomer_sigma => 0.8,

    :monomer_z0 => 0.0,

    :klong => 2000.0,
    :req_long => 1.0,

    :klat => 1000.0,
    :req_lat => 1.2,

    :ktheta => 1000.0,
    :theta_curved   =>  12.0,
    :theta_bent     =>   5.0,
    :theta_straight =>   0.0,

    :kalpha => 900.0,

    :kphi => 30.5,
    :phi_eq => 0.0,
  }.freeze

  DEFAULT_MINIMIZATION_PARAMS = {}

  HELICAL_PITCH_PER_MONOMER = 0.3 / Tubesheet::Units::L0_NM
  HELICAL_PITCH_NUM_MONOMERS = 3

  HELICAL_PITCH = HELICAL_PITCH_NUM_MONOMERS * HELICAL_PITCH_PER_MONOMER

  PARTICLE_TYPE_DEFS = [
    [:COM, {
      :name   => "CoM",
      :mass   => 1.0, 
      :radius => 1.0, 
      :charge => 0.0,
      :collfreq => 0.0,
    }],

    [:BETA, {
      :name   => "Beta",
      :mass   => 1.0,
      :radius => 1.0,
      :charge => 0.0,
      :collfreq => 0.0,
    }],

    [:ORI_ETA1, { 
      :name   => "Ori-Eta1",
      :mass   => 1.0,
      :radius => 1.0,
      :charge => 0.0,
      :collfreq => 0.0,
    }], 

    [:ORI_ETA2, {
      :name   => "Ori-Eta2",
      :mass   => 1.0,
      :radius => 1.0,
      :charge => 0.0,
      :collfreq => 0.0,
    }],

    [:ORI_ZETA, { 
      :name   => "Ori-Zeta",
      :mass   => 1.0,
      :radius => 1.0,
      :charge => 0.0,
      :collfreq => 0.0,
    }],
  ].freeze

  BOND_TYPE_DEFS = [
    :BT_ETA1,
    :BT_ETA2,
    :BT_ZETA,
    :BT_BETA,
    [:BT_LONG_CURVED, :forw],
    [:BT_LONG_BENT, :forw],
    [:BT_LONG_STRAIGHT, :forw],
    :BT_LAT_TUBE,
    :BT_LAT_SHEET,
  ].freeze

  BT_LONG_INITIAL = :BT_LONG_BENT
  BT_LAT_INITIAL  = :BT_LAT_SHEET

  SITE_INFO = [
    [ [ 0.0, 0.0,-HELICAL_PITCH ], :BETA    , [[0, :BT_BETA]] ],
    [ [ 0.0, 1.0, 0.0           ], :ORI_ETA1, [[0, :BT_ETA1]] ],
    [ [ 0.0, 1.0,-HELICAL_PITCH ], :ORI_ETA2, [[1, :BT_ETA2]] ],
    [ [ 1.0, 0.0, 0.0           ], :ORI_ZETA, [[0, :BT_ZETA]] ],
  ].freeze

  class Monomer
    attr_reader :ind

    def initialize(builder,ind)
      @builder = builder
      @ind = ind
    end

    def rigid_body
      @builder.sim.get_rigid_body(@ind)
    end

    def pos
      return rigid_body.com_particle.pos
    end

    def posv
      return rigid_body.com_particle.posv
    end

    def beta
      return rigid_body.site_particles[ @builder.site_type_map[:BETA] ]
    end

    def ori_eta1
      return rigid_body.site_particles[ @builder.site_type_map[:ORI_ETA1] ]
    end

    def ori_eta2
      return rigid_body.site_particles[ @builder.site_type_map[:ORI_ETA2] ]
    end

    def evec
      eta_ref1 = rigid_body.reference_vectors[ @builder.site_type_map[:ORI_ETA1] ]
      eta_ref2 = rigid_body.reference_vectors[ @builder.site_type_map[:ORI_ETA2] ]
      if eta_ref1[2] < eta_ref2[2]
        return (ori_eta2.posv - ori_eta1.posv).normalize
      else
        return (ori_eta1.posv - ori_eta2.posv).normalize
      end
    end

    def pos=(pvec)
      rigid_body.com_particle.pos = pvec
      rigid_body.update
    end

    def get_bond_info(mer)
      rigid_body.get_bond_info(mer.rigid_body)
    end

    def bonded_to?(mer)
      rigid_body.bonded_to?(mer.rigid_body)
    end

    def com_particle
      rigid_body.com_particle
    end

    def site_particles
      rigid_body.site_particles
    end

    def all_particles
      rigid_body.site_particles.unshift(rigid_body.com_particle)
    end

  end

  class SimulationBuilder
    include Tubesheet

    attr_reader :sim, :max_monomers
    attr_reader :job_number
    attr_reader :num_monomers, :num_protofilaments
    attr_reader :protofilaments
    attr_reader :particle_types, :bond_types, :site_type_map, :site_info
    attr_reader :rigid_type
    attr_reader :theta_tube, :theta_sheet

    attr_reader :rigid_map

    attr_accessor :prefix, :initial_direction, :long_bondtype, :connect_seam

    attr_reader :params, :config_params
    attr_accessor :energy_type, :minimization_params

    include KeywordArguments
    include AngleUtils

    def self.convert_yaml_entry(entry)
      case entry
      when String
        return entry.to_sym
      when Array
        return entry.map{|x| self.convert_yaml_entry(x)}
      when Hash
        return Hash[ entry.map{|k,v| [k.to_sym,self.convert_yaml_entry(v)]} ]
      else
        return entry
      end
    end

    def self.from_yaml(rawData)
      origData = YAML::load(rawData)
      kwArgs = self.convert_yaml_entry(origData)

      if !kwArgs[:max_monomers]
        if kwArgs[:max_dimers]
          kwArgs[:max_monomers] = 2*kwArgs[:max_dimers]
          kwArgs.delete :max_dimers
        elsif kwArgs[:num_dimers]
          kwArgs[:max_monomers] = 2*kwArgs[:num_dimers]
          kwArgs.delete :num_dimers
        end
      end

      kwArgs[:num_protofilaments] ||= kwArgs[:num_protofils]
      kwArgs[:num_protofilaments] ||= kwArgs[:num_pfs]

      kwArgs[:max_monomers] *= kwArgs[:num_protofilaments]

      kwArgs[:prefix] = kwArgs[:prefix].to_s if kwArgs.include? :prefix

      self.new(kwArgs)
    end

    def make_rng(kwArgs)
      if kwArgs[ :rng ]
        case kwArgs[:rng]
        when Tubesheet::RNG
          @rng = kwArgs[ :rng ]
          @job_number = :none
          @seed0      = :none
          @seed_incr  = :none
          @seed       = @rng.info[:seed] || :none
        when Array
          (rng_type,@seed) = kwArgs[:rng]
          @rng = ::Tubesheet::RNG.new :generator => rng_type, :seed => @seed
        when Hash
          rng_type = kwArgs[:rng][:type]
          @seed = kwArgs[:rng][:seed]
          @rng = ::Tubesheet::RNG.new :generator => rng_type, :seed => @seed
        else
          raise ArgumentError.new "unknown :rng entry in parameters"
        end
      else
        @job_number = Integer( checkKeyword(kwArgs, :job_number) )
        @seed0      = Integer( checkKeyword(kwArgs, :seed0) )
        @seed_incr  = Integer( checkKeyword(kwArgs, :seed_incr) )

        @rng = :none

        @seed = @seed0 + @job_number * @seed_incr
        @rng = ::Tubesheet::RNG.new :generator => :mt19937, :seed => @seed
      end
    end

    def initialize(kwArgs)
      @config_params = kwArgs

      @max_monomers       = Integer( checkKeyword(kwArgs, :max_monomers) )
      @num_protofilaments = Integer( checkKeyword(kwArgs, :num_protofilaments) )

      make_rng(kwArgs)

      @sim = kwArgs[:sim]

      @energy_type = kwArgs[:energy_type] || :free
      @minimization_params = kwArgs[:minimization_params] || DEFAULT_MINIMIZATION_PARAMS.clone

      @initial_direction = vec3(0,0,1)
      @structure = kwArgs[:struct] || :sheet

      @long_bondtype = kwArgs[:long_bondtype] || BT_LONG_INITIAL
      @connect_seam = kwArgs[:connect_seam] || false

      @num_extra_rigid_bodies = kwArgs[:extra_rigid_bodies] || 0
      @num_extra_particles    = kwArgs[:extra_particles] || 0

      @params = DEFAULT_PARAMS.dup
      @params.merge! kwArgs[:params] if kwArgs.include? :params
      @params.merge! kwArgs[:force_field] if kwArgs.include? :force_field

      @prefix = kwArgs[:prefix]

      define_builder_geometry_params
      create_simulation_object

      if kwArgs[:logging]
        kwArgs[:logging].each do |k,v|
          sim.set_logging k,v
          puts "Logging for :#{k} has been #{v ? 'enabled' : 'disabled'}"
        end
      end

      sim.nonbonds_enabled = kwArgs[:enable_nonbonds] if kwArgs.include? :enable_nonbonds
      sim.nonbonds_enabled = kwArgs[:nonbonds]        if kwArgs.include? :nonbonds

      define_particle_types
      define_bond_types
      define_monomer_rigid_body_type
      define_force_field

      setup_builder_structures
    end

    def define_builder_geometry_params
      @proto_x0 = (-0.5*@num_protofilaments + 0.5) * params[ :req_lat ]

      num_pfs = params[ :num_protofilaments ]
      pf_tube_angle = 180.0 * (num_pfs - 2.0) / Float(num_pfs)
      @theta_tube  = 0.5 * pf_tube_angle
      @theta_sheet = 180.0 - @theta_tube
    end

    def create_simulation_object
      if @sim.nil?
        @sim = Tubesheet::Sim.new
      end

      @sim.dt   = params[ :timestep ]
      @sim.temp = params[ :temp ]
      @sim.rng  = @rng
      @sim.integrator = params[ :integrator ]

      nparticles = PARTICLES_PER_MONOMER * @max_monomers + @num_extra_particles
      nrigids = @max_monomers + @num_extra_rigid_bodies

      @sim.alloc_particles(nparticles) if @sim.max_particles == 0
      @sim.alloc_rigid_bodies(nrigids) if @sim.max_rigid_bodies == 0

      avail_particles = @sim.max_particles - @sim.num_particles
      if avail_particles < nparticles
        raise RuntimeError.new("not enough particles")
      end

      avail_rigids    = @sim.max_rigid_bodies - @sim.num_rigid_bodies
      if avail_rigids    < nrigids
        raise RuntimeError.new("not enough rigid bodies")
      end
    end

    def define_particle_types
      @particle_types = {}
      PARTICLE_TYPE_DEFS.each do |ptype, pdata| 
        pdata = pdata.merge( :collfreq => params[:collfreq] )
        @particle_types[ptype] = @sim.add_particle_type( pdata )
      end
    end

    def define_bond_types
      @bond_types = {}
      BOND_TYPE_DEFS.each_with_index do |bt,ind|
        if bt.is_a? Array
          @bond_types[bt[0]] = [bt[1],ind+1]
        else
          @bond_types[bt] = ind+1
        end
      end

      @bond_types.each do |name,defn|
        @sim.add_bond_type name.to_s, defn
      end
    end

    def define_monomer_rigid_body_type
      @site_info = SITE_INFO.clone
      @site_type_map = {}
      @site_info.each_with_index do |info,index|
        @site_type_map[ info[1] ] = index
      end

      @rigid_type = sim.add_rigid_body_type( {
        :name => "Monomer",
        :type => :site_interaction, 
        :rotfric => params[:rotfric],
        :I0 => 1.0,

        :reference_vectors => site_info.map { |info| info[0] }
      } )
    end

    def define_force_field
      @sim.add_nonbond_pattern(
        "Monomer repulsion", 
        potential(:wca, {:sigma => params[:monomer_sigma], :eps => 1.0}),
        [ particle_types[ :COM ], particle_types[ :COM ] ]
      )

      @sim.nonbond_cutoff  = 1.5 * params[:monomer_sigma]
      @sim.nonbond_padding = 1.0 * params[:monomer_sigma]


      req_long = params[:req_long]
      klong_tube  = params[:klong]
      klong_sheet = params[:klong_sheet] || params[:klong]
      add_com_dist_pattern("Long. dist (curved)"  , :BT_LONG_CURVED  , klong_tube , req_long)
      add_com_dist_pattern("Long. dist (bent)"    , :BT_LONG_BENT    , klong_sheet, req_long)
      add_com_dist_pattern("Long. dist (straight)", :BT_LONG_STRAIGHT, klong_tube , req_long)

      req_lat = params[:req_lat]
      add_lateral_dist_pattern( "Lateral dist (sheet)", :BT_LAT_SHEET, req_lat )
      add_lateral_dist_pattern( "Lateral dist (tube)" , :BT_LAT_TUBE , req_lat )


      ktheta = params[:ktheta]
      ktheta_sheet  = params[:ktheta_sheet] || ktheta
      ktheta_curved = params[:ktheta_curved] || ktheta

      theta_curved   = params[:theta_curved]
      theta_bent     = params[:theta_bent]
      theta_straight = params[:theta_straight]

      add_com_bend_patterns("Long. bend (curved)"  , :BT_LONG_CURVED  , ktheta_curved, theta_curved  )
      add_com_bend_patterns("Long. bend (bent)"    , :BT_LONG_BENT    , ktheta_sheet , theta_bent    )
      add_com_bend_patterns("Long. bend (straight)", :BT_LONG_STRAIGHT, ktheta       , theta_straight)

      kalpha_tube = params[:kalpha]
      kalpha_sheet = params[:kalpha_sheet] || kalpha_tube
      add_lateral_bend_pattern( "Lateral bend (tube)" , :BT_LAT_TUBE , kalpha_tube, @theta_tube  )
      add_lateral_bend_pattern( "Lateral bend (sheet)", :BT_LAT_SHEET, kalpha_sheet, @theta_sheet )

      kshift_tube  = params[:kshift] || kalpha_tube
      kshift_sheet = params[:kshift_sheet] || params[:kalpha_sheet] || kshift_tube

      add_lateral_shift_pattern("Lateral shift (tube)",  :BT_LAT_TUBE , kshift_tube)
      add_lateral_shift_pattern("Lateral shift (sheet)", :BT_LAT_SHEET, kshift_sheet)

      kphi_tube  = params[:kphi]
      kphi_sheet = params[:kphi_sheet] || kphi_tube
      kphi_curved = params[:kphi_curved] || kphi_tube

      add_longitudinal_twist_pattern( "Long. twist (curved)"  , :BT_LONG_CURVED  , kphi_curved, params[:phi_eq] )
      add_longitudinal_twist_pattern( "Long. twist (bent)"    , :BT_LONG_BENT    , kphi_sheet , params[:phi_eq] )
      add_longitudinal_twist_pattern( "Long. twist (straight)", :BT_LONG_STRAIGHT, kphi_tube  , params[:phi_eq] )

      add_lateral_twist_pattern( "Lat. twist (tube)" , :BT_LAT_TUBE , kphi_tube , params[:phi_eq] )
      add_lateral_twist_pattern( "Lat. twist (sheet)", :BT_LAT_SHEET, kphi_sheet, params[:phi_eq] )
    end

    def setup_builder_structures
      @num_monomers = 0
      @protofilaments = (0 ... @num_protofilaments).map{ [] }
      @monomers = {}
      @rigid_map = Array.new(@num_monomers)
    end

    def build_initial_structure
      num_mers = self.max_monomers.div(self.num_protofilaments)
      for pf in 1..num_protofilaments do
        self.new_monomers \
          :protofilament => pf,
          :quantity      => num_mers
      end

      sim = self.sim

      sim.build_potential
    end

    def connect_longitudinal(ind0,ind1, btype=@long_bondtype)
      rb0 = @sim.get_rigid_body( ind0 )
      rb1 = @sim.get_rigid_body( ind1 )

      com0 = rb0.com_particle
      com1 = rb1.com_particle

      com0.add_bond( com1, bond_types[btype] )
    end

    def get_lateral_bondtype( pf0, pf1 )
      case @structure
      when :sheet
        return bond_types[ (pf0.odd?) ? :BT_LAT_TUBE : :BT_LAT_SHEET ]
      when :tube
        return bond_types[ :BT_LAT_TUBE ]
      else
        raise RuntimeError.new("unknown value for structure: :#{@structure.to_s}")
      end
    end

    def connect_lateral( mi, mj, pfs )
      rb_i = @sim.get_rigid_body( mi )
      rb_j = @sim.get_rigid_body( mj )

      if pfs[0] < pfs[1]
        pt0 = rb_i.site_particles[ site_type_map[:BETA] ]
        pt1 = rb_j.com_particle
        btype = get_lateral_bondtype(pfs[0], pfs[1])
      else
        pt0 = rb_j.site_particles[ site_type_map[:BETA] ]
        pt1 = rb_i.com_particle
        btype = get_lateral_bondtype(pfs[1], pfs[0])
      end

      pt0.add_bond(pt1,btype) unless pt0.bonded_to?(pt1)
    end

    def make_seam_bonds
      return unless self.connect_seam

      pf0 = 1
      pf1 = @protofilaments.size

      mers0 = @protofilaments[pf0-1]
      mers1 = @protofilaments[pf1-1]

      mers1.drop(HELICAL_PITCH_NUM_MONOMERS).zip(mers0).each do |tup|
        mj,mi = tup

        connect_lateral(mi,mj, [pf1,pf0]) if mi && mj
      end
    end

    def make_lateral_bonds_with_protofilaments(pf0, pf1, monomer_inds)
      mers_i = @protofilaments[ pf0 - 1 ]
      mers_j = @protofilaments[ pf1 - 1 ]

      ni1 = mers_i.length
      ni0 = ni1 - monomer_inds.length
      nj = mers_j.length

      if nj > ni0
        a = [ni0,nj].min
        b = [ni1,nj].min

        (a...b).each do |ind|
          connect_lateral( mers_i[ind], mers_j[ind], [pf0,pf1] )
        end
      end
    end

    def make_lateral_bonds(pf_ind, monomer_inds)
      if pf_ind > 1
        make_lateral_bonds_with_protofilaments(pf_ind, pf_ind-1, monomer_inds)
      end

      if pf_ind < @num_protofilaments
        make_lateral_bonds_with_protofilaments(pf_ind, pf_ind+1, monomer_inds)
      end

      if pf_ind == 1 || pf_ind == num_protofilaments
        make_seam_bonds 
      end
    end

    def pf_minus_end(pf)
      helical_offset = -(pf-1) * HELICAL_PITCH
      dist = params[:req_lat]
      monomer_z0 = params[:monomer_z0]

      xvec = vec3(1,0,0)
      zvec = vec3(0,0,1)
      evec = @initial_direction.normalize

      case @structure
      when :sheet
        x0 = @proto_x0 + (pf-1) * dist
        y0 = 0.0

        if (1 - zvec.dot(evec)).abs < 1e-8
          ori = Quat.identity
        else
          ori  = Quat.axis_angle zvec.cross(evec), Math::acos(zvec.dot(evec))
        end
      when :tube
        turn_per_pf = 2.0*Math::PI/num_protofilaments

        radius = 0.5*dist/Math::cos(0.5*(Math::PI-turn_per_pf))

        ang = (pf-1)*turn_per_pf
        x0 = radius*Math::cos(ang)
        y0 = radius*Math::sin(ang)

        ori = Quat.axis_angle(zvec, 0.5*Math::PI+ang)
      else
        raise RuntimeError.new("invalid value for simulation structure: :#{@structure.to_s}")
      end

      z0 = monomer_z0 + helical_offset
      [x0,y0,z0,ori]
    end

    def new_monomers(kwArgs)
      pf_ind = Integer(checkKeyword(kwArgs, :protofilament))
      n      = Integer(checkKeyword(kwArgs, :quantity))

      if @num_monomers + n > @max_monomers
        raise ArgumentError, 'allocation exceeds the maximum number of monomers'
      end

      pf = @protofilaments[pf_ind-1]
      num_mers_in_pf = pf.length

      monomer_inds = (0 ... n).map {|ind| ind + @num_monomers}

      req_long = params[:req_long]

      if num_mers_in_pf == 0
        (x0,y0,z0,ori) = pf_minus_end(pf_ind)

        xvec = vec3(1,0,0)
        zvec = vec3(0,0,1)
        evec = @initial_direction.normalize
        if (1 - zvec.dot(evec)).abs < 1e-8
          para = xvec
        else
          para = vec3( ori.rotate(xvec) )
        end
        orth = evec.cross para

        mer_pos = para*x0 + orth*y0 + evec*z0
      else
        prev_mer = get_monomer( pf[-1] )
        evec = prev_mer.evec
        ori = prev_mer.rigid_body.orientation
        mer_pos = prev_mer.posv + evec*req_long
      end

      rb = create_monomer( monomer_inds.first, mer_pos)
      rb.orientation = ori
      monomer_inds[1 .. -1].each do |ind|
        mer_pos += evec*req_long

        rb = create_monomer(ind, mer_pos)
        rb.orientation = ori
        connect_longitudinal( ind-1, ind )
      end

      if @protofilaments[pf_ind-1].last
        connect_longitudinal( @protofilaments[pf_ind-1].last, monomer_inds.first )
      end

      @protofilaments[pf_ind-1] += monomer_inds

      make_lateral_bonds(pf_ind, monomer_inds)

      @num_monomers += n
    end

    def get_monomer(ind)
      mer = @monomers[ind]
      if !mer
        mer = Monomer.new(self, ind)
        @monomers[ind] = mer
      end

      return mer
    end

    def each_monomer_inner(&block)
      (0...@num_monomers).each do |i|
        mer = get_monomer(i)
        yield mer
      end
    end

    def each_monomer(&block)
      if block_given?
        each_monomer_inner(&block)
      else
        if Enumerable.constants.include? 'Enumerator'
          return Enumerable::Enumerator.new(self,:each_monomer_inner)
        else
          return self.to_enum(:each_monomer_inner)
        end
      end
    end

    def create_monomer(ind, pos)
      com_p = sim.define_particle( particle_types[ :COM ] )
      com_ind = com_p.index

      site_inds = (com_ind+1)..(com_ind + PARTICLES_PER_MONOMER - 1)
      all_inds = [com_ind] + site_inds.to_a

      site_inds = site_info.map do |pvec, ptype, pbonds|
        p = sim.define_particle( particle_types[ptype] )
        p.pos = com_p.posv + vec3(pvec)
        p.index
      end

      all_inds = [com_ind] + site_inds

      site_inds.each_with_index do |pind,i|
        (pvec,ptype,pbonds) = site_info[i]
        pbonds.each do |bond_ind, bond_type|
          pj = all_inds[bond_ind]
          pi = pind
          bt = bond_types[bond_type]
          if ! @sim.is_bonded?(pi,pj)
            sim.add_bonds [pi, pj, bt]
          end
        end
      end

      rigid = sim.define_rigid_body(
        :type => @rigid_type,
        :sites => site_inds,
        :com_particle => com_ind
      )
      rigid.com_particle.pos = pos

      @rigid_map[ind] = rigid.index
      return rigid
    end

    def potential(type, params)
      case type
      when :harmonic
        return Tubesheet::Potentials::Harmonic.new(params)

      when :wca
        return Tubesheet::Potentials::WCA.new(params)

      when :angle
        return Tubesheet::Potentials::AngleHarmonic.new(params)

      when :dihedral
        return Tubesheet::Potentials::Dihedral.new(params)

      else
        raise ArgumentError, 'unknown potential type '+type.to_s

      end
    end

    def add_com_dist_pattern( name, dtype, k, dist )
      @sim.add_bonded_pattern({
        :name           => name,
        :potential      => potential(:harmonic, { :k => k, :r_eq => dist }),
        :particle_types => [ particle_types[:COM], particle_types[:COM] ],
        :bondtypes      => [ bond_types[ dtype ] ]
      })
    end

    def add_com_bend_patterns( name, btype, ktheta, bend_angle )
      theta_eq = bend_theta(bend_angle)

      @sim.add_bonded_pattern({
        :name           => name + " (Ori-Eta/Forw)",
        :potential      => potential(:angle, { :k => ktheta, :theta_eq => theta_eq }),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:COM]],
        :bondtypes      => [bond_types[:BT_ETA1], bond_types[btype]],
      })

      @sim.add_bonded_pattern({
        :name           => name + " (Ori-Eta/Back)",
        :potential      => potential(:angle, { :k => ktheta, :theta_eq => theta_eq }),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:COM]],
        :bondtypes      => [bond_types[:BT_ETA1], reverse_direction(bond_types[btype])],
      })

      @sim.add_bonded_pattern({
        :name           => name + " (Ori-Zeta/Forw)",
        :potential      => potential(:angle, { :k => ktheta, :theta_eq => 0.5*Math::PI }),
        :particle_types => [particle_types[:ORI_ZETA], particle_types[:COM], particle_types[:COM]],
        :bondtypes      => [bond_types[:BT_ZETA], bond_types[btype]],
      })

      @sim.add_bonded_pattern({
        :name           => name + " (Ori-Zeta/Back)",
        :potential      => potential(:angle, { :k => ktheta, :theta_eq => 0.5*Math::PI }),
        :particle_types => [particle_types[:ORI_ZETA], particle_types[:COM], particle_types[:COM]],
        :bondtypes      => [bond_types[:BT_ZETA], reverse_direction(bond_types[btype])],
      })

    end

    def reverse_direction(bondtype)
      if bondtype.is_a? Numeric
        return bondtype
      elsif bondtype.length == 2
        case bondtype[0]
        when :none
          return [ :none, bondtype[1] ]
        when :back
          return [ :forw, bondtype[1] ]
        when :forw
          return [ :back, bondtype[1] ]
        when :omni
          return [ :omni, bondtype[1] ]
        end
      else
        return bondtype
      end
    end

    def add_lateral_dist_pattern( name, dtype, dist )
      @sim.add_bonded_pattern({
        :name           => name,
        :potential      => potential(:harmonic, { :k => params[:klat], :r_eq => dist }),
        :particle_types => [particle_types[:COM], particle_types[:BETA]],
        :bondtypes      => [bond_types[ dtype ]],
      })
    end

    def add_lateral_bend_pattern( name, btype, kalpha, bend_angle )
      alpha_eq = Math::PI - deg_to_rad(bend_angle)

      @sim.add_bonded_pattern({
        :name           => name + " (Eta1-Com-Beta)",
        :potential      => potential(:angle, { :k => kalpha, :theta_eq => alpha_eq }),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:BETA]],
        :bondtypes      => [bond_types[:BT_ETA1], bond_types[btype]],
      })

      @sim.add_bonded_pattern({
        :name           => name + " (Eta2-Beta-CoM)",
        :potential      => potential(:angle, { :k => kalpha, :theta_eq => alpha_eq }),
        :particle_types => [particle_types[:ORI_ETA2], particle_types[:BETA], particle_types[:COM]],
        :bondtypes      => [bond_types[:BT_ETA2], bond_types[btype]],
      })
    end

    def add_lateral_shift_pattern(name, btype, kshift)
      @sim.add_bonded_pattern({
        :name           => name + " (CoM-Beta-CoM)",
          :potential      => potential(:angle, { :k => kshift, :theta_eq => 0.5*Math::PI }),
          :particle_types => [particle_types[:COM], particle_types[:BETA], particle_types[:COM]],
          :bondtypes      => [bond_types[:BT_BETA], bond_types[btype]],
      })

      @sim.add_bonded_pattern({
        :name           => name + " (Beta-CoM-Beta)",
          :potential      => potential(:angle, { :k => kshift, :theta_eq => 0.5*Math::PI }),
          :particle_types => [particle_types[:BETA], particle_types[:COM], particle_types[:BETA]],
          :bondtypes      => [ bond_types[btype], bond_types[:BT_BETA]], 
      })
    end

    def add_lateral_internal_torque(name, btype0, btype1, bend_angle0, bend_angle1)
      @sim.add_bonded_pattern({
        :name           => name,
        :potential      => Tubesheet::Potentials::Dihedral.new( :k => kphi, :phi_eq => phi_eq ),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:BETA], particle_types[:ORI_ETA2]],
        :bondtypes      => [bond_types[:BT_ETA1], bond_types[btype], bond_types[:BT_ETA2]],
      })
    end

    def bondtype_name(bt)
      bt.to_s.sub(/(LAT|LONG)_/,'')
    end

    def add_longitudinal_twist_pattern( name, btype, kphi, phi_eq )
      @sim.add_bonded_pattern({
        :name           => name,
        :potential      => Tubesheet::Potentials::Dihedral.new( :k => kphi, :phi_eq => phi_eq ),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:COM], particle_types[:ORI_ETA1]],
        :bondtypes      => [bond_types[:BT_ETA1], bond_types[btype], bond_types[:BT_ETA1]],
      })
    end

    def add_lateral_twist_pattern( name, btype, kphi, phi_eq )
      @sim.add_bonded_pattern({
        :name           => name,
        :potential      => Tubesheet::Potentials::Dihedral.new( :k => kphi, :phi_eq => phi_eq ),
        :particle_types => [particle_types[:ORI_ETA1], particle_types[:COM], particle_types[:BETA], particle_types[:ORI_ETA2]],
        :bondtypes      => [bond_types[:BT_ETA1], bond_types[btype], bond_types[:BT_ETA2]],
      })
    end

    def do_minimization(*args)
      kw = args.first || self.minimization_params

      nsteepest = 500
      nrefine = 10000
      report_steepest = nil
      report_refine = nil
      epsilon = nil # use default epsilon

      disable_nonbonds = !sim.nonbonds_enabled
      report_errors = true

      nsteepest = kw[:num_iterations] if kw[:num_iterations]
      nsteepest = kw[:nsteepest]      if kw[:nsteepest]

      nrefine   = kw[:num_iterations] if kw[:num_iterations]
      nrefine   = kw[:nrefine]        if kw[:nrefine]

      disable_nonbonds = kw[:disable_nonbonds] if kw[:disable_nonbonds]

      report_steepest = kw[:report_steepest]
      report_refine   = kw[:report_refine  ]
      report_errors   = kw[:report_errors  ]

      epsilon = kw[ :epsilon ]

      report_steepest ||= nsteepest/5
      report_refine ||= 500
      begin
        if nsteepest > 0
          steepestParams = {
            :method => :steepest,
            :num_iterations => nsteepest,
            :report_freq => report_steepest,
            :disable_nonbonds => disable_nonbonds,
          }

          if kw[ :steepest_epsilon ]
            steepestParams[:epsilon] = kw[ :steepest_epsilon ]
          elsif epsilon
            steepestParams[:epsilon] = epsilon
          end

          sim.minimize(steepestParams)
        end

        begin
          if nrefine > 0
            refineParams = {
              :method => :lbfgs,
              :num_iterations => nrefine,
              :disable_nonbonds => disable_nonbonds,
              :report_freq => report_refine,
            }

            if kw[ :refine_epsilon ]
              refineParams[:epsilon] = kw[ :refine_epsilon ]
            elsif epsilon
              refineParams[:epsilon] = epsilon
            end

            sim.minimize(refineParams)
          end
        rescue ::Tubesheet::SimError => err
          raise err unless err.ecode == Tubesheet::SimError::TS_ERR_MINIMIZATION_INCOMPLETE
          if report_errors
            puts "L-BGFS failed to complete: #{err.message}"
          end
        end
      rescue Exception => e
        puts "Minimizer encountered an error\n", e
        puts "...ABORTING..."
        raise e
      end

      sim.update_rigid_bodies
    end

    def calculate_mechanical_energy(*args)
      kw = (Hash === args.first) ? args.first : {}
      calc_components = kw[:components] || false

      if @energy_type == :none
        if calc_components
          return { :free => 0.0, :enthalpic => 0.0, :entropic => 0.0 }
        else
          return 0.0
        end
      end

      self.do_minimization

      if @energy_type == :enthalpy
        enthalpy = sim.calculate_energy
        if calc_components 
          return { :free => 0.0, :entropic => 0.0, :enthalpic => enthalpy }
        else
          return enthalpy
        end
      end

      if @energy_type == :entropy
        components = sim.normal_mode_free_energy :components => true
        if calc_components
          return components 
        else
          return components[:entropic]
        end
      end

      if @energy_type == :free
        return sim.normal_mode_free_energy :components => calc_components
      end

      throw RuntimeError.new "invalid energy type #{energy_type}"
    end

  end
end
