#include "numarray_mruby.h"

#include <mruby.h>
#include <mruby/data.h>
#include <mruby/array.h>
#include <mruby/range.h>
#include <mruby/data.h>
#include <mruby/class.h>
#include <mruby/numeric.h>
#include <mruby/string.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "numarray_lapack.h"

/* Handy utility functions... */
static mrb_value create_exc(mrb_state* mrb, struct RClass* eclass, const char* fmt, ...)
{
  char buf[128];
  va_list args;
  va_start(args, fmt);
  int buflen = vsnprintf(buf, sizeof(buf), fmt, args);
  va_end(args);

  return mrb_exc_new(mrb, eclass, buf, buflen);
}


/* NumArray module
 *
 * NumArray class : general N-dimension array
 *
 *   get / set individual elements, sub-arrays of elements
 *   
 *   element-wise addition, multiplication,
 *     element-wise functions (log, exp, etc.)
 *
 *     -- add scalar element to all elements in array
 *     -- add two arrays of the same size together
 *     -- take logarithm of each element
 *
 *   reshape array
 *
 *   low priority:
 *     different element types (double/float/int64_t/int32_t/etc.)
 *
 *
 * NumMatrix : 2-D matrix
 *
 *   all NumArray operations for 2-D matrix
 *     (get/set, element-wise operations, reshape)
 *
 *   transpose
 *
 *   matrix-matrix, matrix-vector multiplication
 *
 *   solve systems of equations (given a vector)
 *
 *   useful matrix factorizations (LU, Cholesky, etc.)
 *
 *   solve eigensystems
 *
 *
 * NumVector : 1-D vector
 * 
 *   all NumArray operations for 1-D vector (get/set, element-wise
 *   operations, reshape)
 *
 *   matrix-vector multiplication (either left- or right-)
 *
 *   inner product, outer product
 *
 */

extern void numarray_mrb_initialize(mrb_state* mrb);

static int numarray_dims_compatible(numarray* A, numarray* B)
{
  if (A->ndims != B->ndims) return 0;

  for(int i=0; i < A->ndims; i++)
    if (A->dims[i] != B->dims[i]) return 0;

  return 1;
}

static inline int numarray_compatible_linsize(numarray* A, numarray* B)
{
  return (B->ndims == 1) && (A->nelts == B->nelts);
}

static inline int numarray_matrix_p(numarray* A) { return A->ndims == 2; }
static inline int numarray_vector_p(numarray* A) { return A->ndims == 1; }

static numarray* 
numarray_alloc(mrb_state* mrb, int ndims, int* dims)
{
  if (ndims < 1)
    mrb_raise(mrb, E_RANGE_ERROR, "invalid number of dimensions");
  if (ndims > MAX_DIMS)
    mrb_raise(mrb, E_RANGE_ERROR, "number of dimensions exceeds maximum");

  /* check dimensions */
  size_t nelts = 1;
  for (int i=0; i < ndims; i++)
  {
    if (dims[i] < 0)
      mrb_raisef(mrb, E_RANGE_ERROR,
          "invalid dimension %S", mrb_fixnum_value(i));
    nelts *= dims[i];
  }

  numarray* A = mrb_malloc(mrb, sizeof(*A));
  A->ndims = ndims;
  memcpy(A->dims, dims, ndims*sizeof(*dims));
  A->nelts = nelts;
  A->elts = mrb_calloc(mrb, nelts, sizeof(A->elts[0]));

  return A;
}

static numarray*
numarray_copy(mrb_state* mrb, numarray* A)
{
  numarray* B = numarray_alloc(mrb, A->ndims, A->dims);
  memcpy(B->elts, A->elts, A->nelts*sizeof(A->elts[0]));
  return B;
}

static numarray*
numarray_copy_transpose(mrb_state* mrb, numarray* A)
{
  if (!numarray_matrix_p(A)) return NULL;

  int M = A->dims[0];
  int N = A->dims[1];
  int dims[2] = { N, M };
  numarray* B = numarray_alloc(mrb, 2, dims);

  const double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  for (int i=0; i < M; i++)
  {
    for (int j=0; j < N; j++)
    {
      eltsB[j+N*i] = eltsA[i+M*j];
    }
  }

  return B;
}

static numarray* numarray_eye(mrb_state* mrb, int n)
{
  if (n <= 0) return NULL;

  int dims[2] = { n, n};
  numarray* A = numarray_alloc(mrb, 2, dims);
  double*restrict eltsA = A->elts;
  for (int i=0 ; i < n; i++)
  {
    *eltsA = 1.0;
    eltsA += (n+1);
  }
  return A;
}

static inline void numarray_fill(numarray* A, double val)
{
  if (val == 0.0)
  {
    memset(A->elts, 0, (A->nelts)*sizeof(A->elts[0]));
  }
  else
  {
    double*restrict beg = A->elts;
    double*restrict end = beg+A->nelts;
    while(beg<end) { *(beg++) = val; }
  }
}

enum MATVEC_TYPE {
  MATVEC_ARRAY  = 0,
  MATVEC_MATRIX = 1,
  MATVEC_VECTOR = 2,
};

enum {
  MATVEC_OP_MAT_MAT = 1,
  MATVEC_OP_MAT_VEC = 2,
  MATVEC_OP_VEC_MAT = 3,
  MATVEC_OP_VEC_VEC = 4,
};

static inline enum MATVEC_TYPE matvec_type(numarray* A)
{
  if (numarray_matrix_p(A))
    return MATVEC_MATRIX;
  else if (numarray_vector_p(A))
    return MATVEC_VECTOR;
  else
    return MATVEC_ARRAY;
}

enum {
  MATMUL_OP_MAT_MAT = 1,
  MATMUL_OP_MAT_VEC = 2,
  MATMUL_OP_VEC_MAT = 3,
  MATMUL_OP_VEC_VEC = 4,
};

static inline int matvec_op_type(enum MATVEC_TYPE atype, enum MATVEC_TYPE btype)
{
  //   a   b   ab   ai   bi  2*(ai-1)+bi
  //  --- --- ---- ---- ---- -----------
  //   M   M   MM    1    1    1  (MATVEC_OP_MAT_MAT)
  //   M   V   MV    1    2    2  (MATVEC_OP_MAT_VEC)
  //   V   M   VM    2    1    3  (MATVEC_OP_VEC_MAT)
  //   V   V   VV    2    2    4  (MATVEC_OP_VEC_VEC)

  return 2*(atype-1)+btype;
}

static inline int numarray_matmul_dims_compatible(numarray* A, numarray* B, int matvec)
{
  if ((matvec==MATVEC_OP_MAT_MAT) ||
      (matvec==MATVEC_OP_MAT_VEC))
    return A->dims[1] == B->dims[0];

  if ((matvec == MATVEC_OP_VEC_MAT) ||
      (matvec == MATVEC_OP_VEC_VEC))
    return A->dims[0] == B->dims[0];

  return 0;
}

static inline int 
numarray_matvec_mul_dims_compatible(numarray* A, numarray* B)
{
  return A->dims[1] == B->dims[0];
}

static void numarray_logical_negate(numarray* A)
{
  double*restrict beg = A->elts;
  double*restrict end = beg+A->nelts;
  while(beg<end) { *beg = (*beg == 0.0); beg++; }
}

static int numarray_logical_or(numarray* C, const numarray *A, const numarray *B)
{
  size_t i,n;

  n = A->nelts;
  if (n != B->nelts) { return -1; }
  if (n != C->nelts) { return -2; }

  if ((A == B) && (A == C)) { return -3; }

  if (B == C) { const numarray *tmp = A; A = B; B = tmp; }

  if (A == C)
  {
    const double *restrict bp = B->elts;
    double *restrict cp = C->elts;

    for (i=0; i < n; i++)
    {
      /* FIXME: find a branch-free way to do this */
      cp[i] = ((cp[i] != 0.0) || (bp[i] != 0.0)) ? 1.0 : 0.0;
    }
  }
  else if (A == B)
  {
    const double *restrict bp = B->elts;
    double *restrict cp = C->elts;

    for (i=0; i < n; i++)
    {
      /* FIXME: find a branch-free way to do this */
      cp[i] = (bp[i] != 0.0) ? 1.0 : 0.0;
    }
  }
  else
  {
    const double *restrict ap = A->elts;
    const double *restrict bp = B->elts;
    double *restrict cp = C->elts;

    for (i=0; i < n; i++)
    {
      /* FIXME: find a branch-free way to do this */
      cp[i] = ((ap[i] != 0.0) || (bp[i] != 0.0)) ? 1.0 : 0.0;
    }
  }

  return 0;
}

static void numarray_add_to(numarray* A, numarray* B)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] += eltsB[i];
}

static void numarray_add_scalar(numarray* A, double scl)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] += scl;
}

static void numarray_sub_scalar(numarray* A, double scl)
{
  numarray_add_scalar(A, -scl);
}

static void numarray_mul_scalar(numarray* A, double scl)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] *= scl;
}

static void numarray_div_scalar(numarray* A, double scl)
{
  numarray_mul_scalar(A, 1.0/scl);
}

static void numarray_sub_from(numarray* A, numarray* B)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] -= eltsB[i];
}

static void numarray_mul_by(numarray* A, numarray* B)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] *= eltsB[i];
}

static void numarray_div_by(numarray* A, numarray* B)
{
  size_t n = A->nelts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  for(size_t i=0; i < n; i++)
    eltsA[i] /= eltsB[i];
}

static double* numarray_ref(numarray* A, int inds[])
{
  size_t off = inds[A->ndims-1];
  for(int i=A->ndims-2; i >= 0; i--)
  {
    off *= A->dims[i];
    off += inds[i];
  }

  // fprintf(stderr, "A{ %lu }\n", (unsigned long)off);
  return A->elts+off;
}

static int numarray_matvec_multiply(numarray* C, numarray* A, numarray* B)
{
  if (!numarray_vector_p(C) ||
      !numarray_matrix_p(A) ||
      !numarray_vector_p(B))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  if ((C->dims[0] != A->dims[0]) ||
      (A->dims[1] != B->dims[0]))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  if (C==B)
    return MATRIX_ERROR_DIMS_CANNOT_SELF_MULTIPLY;

  double*restrict eltsC = C->elts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  size_t M = A->dims[0];
  size_t N = A->dims[1];

  memset(eltsC, 0, M*sizeof(*eltsC));

  for(size_t j=0; j < N; j++)
  {
    for(size_t i=0; i < M; i++)
    {
      eltsC[i] += (*eltsA) * (*eltsB);
      eltsA++;
    }
    eltsB++;
  }

  return MATRIX_ERROR_NONE;
}

static int numarray_vecmat_multiply(numarray* C, numarray* A, numarray* B)
{
  if (!numarray_vector_p(C) ||
      !numarray_vector_p(A) ||
      !numarray_matrix_p(B))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  if ((C->dims[0] != B->dims[1]) ||
      (A->dims[0] != B->dims[0]))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  if (C==A)
    return MATRIX_ERROR_DIMS_CANNOT_SELF_MULTIPLY;

  double*restrict eltsC = C->elts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  size_t M = B->dims[0];
  size_t N = B->dims[1];

  memset(eltsC, 0, M*sizeof(*eltsC));

  for(size_t j=0; j < N; j++)
  {
    for(size_t i=0; i < M; i++)
    {
      eltsC[j] += eltsA[i] * eltsB[i+M*j];
    }
  }

  return MATRIX_ERROR_NONE;
}

static int numarray_vecdot(double*restrict dotp, numarray* A, numarray* B)
{
  if (!numarray_vector_p(A) ||
      !numarray_vector_p(B))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  if (A->dims[0] != B->dims[0])
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  double acc = 0.0;
  size_t M = A->dims[0];
  if (A==B)
  {
    double*restrict eltsA = A->elts;
    for(size_t i=0; i < M; i++)
      acc += eltsA[i]*eltsA[i];
  }
  else
  {
    double*restrict eltsA = A->elts;
    double*restrict eltsB = B->elts;
    for(size_t i=0; i < M; i++)
      acc += eltsA[i]*eltsB[i];
  }

  *dotp = acc;

  return MATRIX_ERROR_NONE;
}

static int numarray_matrix_square(numarray* C, numarray* A)
{
  if (C==A)
    return MATRIX_ERROR_DIMS_CANNOT_SELF_MULTIPLY;

  if ((A->dims[0] != A->dims[1]) ||
      (C->dims[0] != A->dims[0]) ||
      (C->dims[1] != A->dims[1]))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  double*restrict eltsC = C->elts;
  memset(eltsC, 0, C->nelts * sizeof(*eltsC));

  double*restrict eltsA = A->elts;

  size_t N = A->dims[0];
  for (size_t i=0; i < N; i++)
  {
    for (size_t j=0; j < N; j++)
    {
      size_t off = i+j*N;
      for (size_t k=0; k < N; k++)
      {
        eltsC[off] += eltsA[i+N*k] * eltsA[k + N*j];
      }
    }
  }

  return MATRIX_ERROR_NONE;
}

static int numarray_matrix_multiply(numarray* C, numarray* A, numarray* B)
{
  if ((C==A) || (C==B))
    return MATRIX_ERROR_DIMS_CANNOT_SELF_MULTIPLY;

  if (A==B) return numarray_matrix_square(C,A);

  if ((A->dims[1] != B->dims[0]) ||
      (C->dims[0] != A->dims[0]) ||
      (C->dims[1] != B->dims[1]))
    return MATRIX_ERROR_DIMS_INCOMPATIBLE;

  double*restrict eltsC = C->elts;
  double*restrict eltsA = A->elts;
  double*restrict eltsB = B->elts;

  // memset(eltsC, 0, C->nelts * sizeof(*eltsC));

  size_t M = A->dims[0];
  size_t L = A->dims[1];
  size_t N = B->dims[1];
  for (size_t k=0; k < N; k++)
  {
    size_t offC = M*k;

    for (size_t i=0; i < M; i++)
    {
      eltsC[i+offC] = 0.0;

      for (size_t j=0; j < L; j++)
      {
        eltsC[i+offC] += eltsA[i+M*j] * eltsB[j+L*k];
      }
    }
  }

  /*
  for (size_t i=0; i < M; i++)
  {
    for (size_t j=0; j < N; j++)
    {
      size_t off = i + M*j;
      for (size_t k=0; k < L; k++)
      {
        eltsC[off] += eltsA[i+L*k] * eltsB[k+L*j];
      }
    }
  }
  */

  return MATRIX_ERROR_NONE;
}

static int numarray_solve(mrb_state* mrb, numarray*restrict A, numarray*restrict X)
{
  int nrhs = 1;
  int* ipiv = mrb_calloc(mrb, A->dims[0], sizeof(*ipiv));
  int info = 0;
  dgesv_(
      A->dims+0,   // N is INTEGER
                   // The number of linear equations, i.e., the order of the
                   // matrix A.  N >= 0.

      &nrhs,       // NRHS is INTEGER
                   // The number of right hand sides, i.e., the number of columns
                   // of the matrix B.  NRHS >= 0.

      A->elts,     // A is DOUBLE PRECISION array, dimension (LDA,N)
                   // On entry, the N-by-N coefficient matrix A.
                   // On exit, the factors L and U from the factorization
                   // A = P*L*U; the unit diagonal elements of L are not stored.

      A->dims+0,   // LDA is INTEGER
                   // The leading dimension of the array A.  LDA >= max(1,N).

      ipiv,        // IPIV is INTEGER array, dimension (N)
                   // The pivot indices that define the permutation matrix P;
                   // row i of the matrix was interchanged with row IPIV(i).

      X->elts,     // B is DOUBLE PRECISION array, dimension (LDB,NRHS)
                   // On entry, the N-by-NRHS matrix of right hand side matrix B.
                   // On exit, if INFO = 0, the N-by-NRHS solution matrix X.

      X->dims+0,   // LDB is INTEGER
                   // The leading dimension of the array B.  LDB >= max(1,N).

      &info        // INFO is INTEGER
                   // = 0:  successful exit
                   // < 0:  if INFO = -i, the i-th argument had an illegal value
                   // > 0:  if INFO = i, U(i,i) is exactly zero.  The factorization
                   //       has been completed, but the factor U is exactly
                   //       singular, so the solution could not be computed.
	);

  mrb_free(mrb, ipiv);

  return info;
}

static inline void slice_finalize(mrb_state* mrb, struct numarray_slice* sl)
{
  if (sl->type == SLICE_LIST)
  {
    mrb_free(mrb, sl->slice.list);
    sl->nelts = 0;
    sl->slice.list = NULL;
  }
}

static inline void slice_list_finalize(mrb_state* mrb, int n, struct numarray_slice* slices)
{
  for(int i=0; i < n; i++)
    slice_finalize(mrb,slices+i);
}

static int slice_index(struct numarray_slice* sl, int ind)
{
  switch (sl->type)
  {
    case SLICE_ALL:
      return ind;

    case SLICE_INDEX:
      return sl->slice.index;

    case SLICE_RANGE:
      return sl->slice.range.beg + ind*sl->slice.range.incr;

    case SLICE_LIST:
      return sl->slice.list[ind];

    default:
      return -1;
  }
}

static double* numarray_ref_slices(numarray* A, struct numarray_slice* slices, size_t elt0)
{
  size_t elt = elt0;

  size_t elt_index = 0;
  size_t multiplier = 1;

  for (int dim=0; dim < A->ndims; dim++)
  {
    ldiv_t ind_and_rest = ldiv(elt, slices[dim].nelts);
    int slice_ind = ind_and_rest.rem;
    elt = ind_and_rest.quot;

    int mat_ind = slice_index(slices+dim,slice_ind);
    elt_index += multiplier*mat_ind;
    multiplier *= A->dims[dim];
  }

  return A->elts+elt_index;
}

static inline void numarray_apply_func(numarray* A, double(*func)(double))
{
  double*restrict beg = A->elts;
  double*restrict end = beg + A->nelts;
  while(beg < end) { *beg = func(*beg); beg++; }
}

static inline double absmax(double a, double b)
{
  b = fabs(b);
  return a < b ? b : a;
}

static inline double absmin(double a, double b)
{
  b = fabs(b);
  return b < a ? b : a;
}

static inline double minpair(double a, double b) { return b < a ? b : a; }
static inline double maxpair(double a, double b) { return a < b ? b : a; } 

static inline double numarray_scalar_reduce(numarray* A, double v0, double(*func)(double,double))
{
  double*restrict beg = A->elts;
  double*restrict end = beg + A->nelts;
  while(beg < end)
  {
    v0 = func(v0,*beg);
    beg++;
  }

  return v0;
}

static inline double numarray_absmin(numarray* A)
{
  return numarray_scalar_reduce(A, fabs(A->elts[0]), absmin);
}

static inline double numarray_min(numarray* A)
{
  return numarray_scalar_reduce(A, A->elts[0], minpair);
}

static inline double numarray_absmax(numarray* A)
{
  return numarray_scalar_reduce(A, 0.0, absmax);
}

static inline double numarray_max(numarray* A)
{
  return numarray_scalar_reduce(A, A->elts[0], maxpair);
}

static double numarray_mean(numarray* A)
{
  double sum,err;
  size_t i;

  if (A->nelts == 0) { return 0.0; }
  if (A->nelts == 1) { return A->elts[0]; }

  sum = A->elts[0];
  err = 0.0;

  /* Use Kahan summation to keep roundoff error to a minimum */
  for (i=1; i < A->nelts; i++)
  {
    double tmp,acc;
    tmp = A->elts[i] - err;
    acc = sum + tmp;

    err = (acc - sum) - tmp;
    sum = acc;
  }

  sum /= A->nelts;

  return sum;
}

static double numarray_var(numarray* A)
{
  double sum, err, mean;
  size_t i;

  if (A->nelts == 0) { return INFINITY; }
  if (A->nelts == 1) { return INFINITY; }

  mean = numarray_mean(A);

  sum = (A->elts[0] - mean);
  sum = sum*sum;
  err = 0.0;

  /* Use Kahan summation to keep roundoff error to a minimum */
  for (i=1; i < A->nelts; i++)
  {
    double val,tmp,acc;
    val = A->elts[i] - mean;
    val = val*val;

    tmp = val - err;
    acc = sum + tmp;

    err = (acc - sum) - tmp;
    sum = acc;
  }

  sum /= (A->nelts-1);

  return sum;
}

/** numarray comparisons **/

static inline double spaceship(double a, double b)
{
  if (a<b) return -1.0;
  if (a>b) return  1.0;
  else     return  0.0;
}

#define SCALAR_CMP(op) do { \
    while(beg < end) {      \
      *beg = (*beg) op val; \
      beg++;                \
  } } while(0)
static void numarray_compare_scalar(numarray*restrict C, double val, enum MATRIX_CMP_OP op)
{
  double*restrict beg = C->elts;
  double*restrict end = beg + C->nelts;
  switch (op)
  {
    case MATRIX_CMP_EQ: SCALAR_CMP(==); break;
    case MATRIX_CMP_NE: SCALAR_CMP(!=); break;
    case MATRIX_CMP_LE: SCALAR_CMP(<=); break;
    case MATRIX_CMP_LT: SCALAR_CMP(< ); break;
    case MATRIX_CMP_GE: SCALAR_CMP(>=); break;
    case MATRIX_CMP_GT: SCALAR_CMP(> ); break;
    case MATRIX_CMP_SPACESHIP:
      {
        while(beg < end)
        {
          *beg = spaceship(*beg,val);
          beg++;
        }
      }
      break;
    default:
      break;
  }
}
#undef SCALAR_OP

#define ARRAY_CMP(op) do { \
    while(eltC < end) {    \
      *eltC = (*eltC) op (*eltA); \
      eltC++; eltA++;      \
  } } while(0)
static void
numarray_compare_arrays(numarray*restrict C,
    const numarray*restrict A, enum MATRIX_CMP_OP op)
{
  double*restrict eltC = C->elts;
  double*restrict end = eltC + C->nelts;

  const double*restrict eltA = A->elts;

  switch (op)
  {
    case MATRIX_CMP_EQ: ARRAY_CMP(==); break;
    case MATRIX_CMP_NE: ARRAY_CMP(!=); break;
    case MATRIX_CMP_LE: ARRAY_CMP(<=); break;
    case MATRIX_CMP_LT: ARRAY_CMP(< ); break;
    case MATRIX_CMP_GE: ARRAY_CMP(>=); break;
    case MATRIX_CMP_GT: ARRAY_CMP(> ); break;
    case MATRIX_CMP_SPACESHIP:
      {
        while(eltC < end)
        {
          *eltC = spaceship(*eltC,*eltA);
          eltC++; eltA++;
        }
      }
      break;
    default: break;
  }
}
#undef ARRAY_OP


/** MRuby-specific bindings **/

static void numarray_free(mrb_state* mrb, void* p)
{
  numarray* A = p;
  mrb_free(mrb, A->elts);
  mrb_free(mrb, A);
}

static const mrb_data_type numarray_type = {
  .struct_name = "NumArray",
  .dfree = numarray_free,
};

static inline struct RClass* get_range_class(mrb_state* mrb)
{
  return mrb_class_get(mrb, "Range");
}

static inline struct RClass* get_numarray_class(mrb_state* mrb)
{
  return mrb_class_get(mrb, "NumArray");
}

static inline int mrb_range_p(mrb_state* mrb, mrb_value obj)
{
  return mrb_obj_is_instance_of(mrb, obj, get_range_class(mrb));
}

static inline int mrb_numarray_p(mrb_state* mrb, mrb_value obj)
{
  return mrb_obj_is_instance_of(mrb, obj, get_numarray_class(mrb));
}

static mrb_value numarray_wrap(mrb_state* mrb, numarray* A)
{
  if (A == NULL)
    mrb_raise(mrb, E_RUNTIME_ERROR,
        "numarray_wrap: A == NULL");

  struct RData* data = mrb_data_object_alloc(mrb, 
      get_numarray_class(mrb), A, &numarray_type);

  return mrb_obj_value(data);
}

numarray* numarray_unwrap(mrb_state* mrb, mrb_value obj)
{
  if (!mrb_numarray_p(mrb,obj))
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected NumArray");

  return DATA_PTR(obj);
}

/* create new array filled with zeros */
// static mrb_value cNumArray_zeros(mrb_state* 

static mrb_value debug_break_hook(mrb_state* mrb, mrb_value mod)
{
  return mrb_nil_value();
}

mrb_value numarray_new_uninitialized(mrb_state *mrb,
    int ndims, int *dims, numarray **arr_ptr)
{
  numarray* A = numarray_alloc(mrb, ndims, dims);
  if (arr_ptr) { *arr_ptr = A; }

  return numarray_wrap(mrb, A);
}

mrb_value numarray_new_filled(mrb_state *mrb,
    int ndims, int *dims, double val, numarray **arr_ptr)
{
  numarray *A;
  mrb_value arr;

  A = NULL;
  arr = numarray_new_uninitialized(mrb, ndims, dims, &A);

  numarray_fill(A, val);
  if (arr_ptr) { *arr_ptr = A; }

  return arr;
}

static mrb_value cNumArray_alloc_filled(mrb_state* mrb, int ndims, mrb_value* rdims, double val)
{
  int dims[ MAX_DIMS ];
  if (ndims > MAX_DIMS)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "number of dimensions exceeds maximum allowed");

  for (int i=0; i < ndims; i++)
  {
    if (!mrb_fixnum_p(rdims[i]))
      mrb_raisef(mrb, E_ARGUMENT_ERROR,
          "dimension %S is not a fixnum", mrb_fixnum_value(i+1));
    dims[i] = mrb_fixnum(rdims[i]);
  }

  return numarray_new_filled(mrb, ndims,dims, val, NULL);
}

static mrb_value cNumArray_zeros(mrb_state* mrb, mrb_value klass)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);
  return cNumArray_alloc_filled(mrb, argc, argv, 0.0);
}

static mrb_value cNumArray_ones(mrb_state* mrb, mrb_value klass)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);
  return cNumArray_alloc_filled(mrb, argc, argv, 1.0);
}

static mrb_value cNumArray_filled(mrb_state* mrb, mrb_value klass)
{
  mrb_value dim_array = mrb_nil_value();
  mrb_float fill_val = 0.0;
  mrb_get_args(mrb, "Af", &dim_array, &fill_val);
  return cNumArray_alloc_filled(mrb,
      RARRAY_LEN(dim_array),
      (mrb_value *)RARRAY_PTR(dim_array),
      fill_val);
}

static mrb_value cNumArray_eye(mrb_state* mrb, mrb_value klass)
{
  mrb_int n = 0;
  mrb_get_args(mrb, "i", &n);
  if (n <= 0)
    mrb_raisef(mrb, E_ARGUMENT_ERROR,
        "invalid matrix rank %S",
          mrb_fixnum_value(n));

  numarray* A = numarray_eye(mrb, n);
  return numarray_wrap(mrb,A);
}

static mrb_value cNumArray_diag(mrb_state* mrb, mrb_value klass)
{
  mrb_value rdiag = mrb_nil_value();
  mrb_get_args(mrb, "A", &rdiag);
  int n = RARRAY_LEN(rdiag);

  if (n == 0)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "expected a non-empty array of diagonal elements");
  }

  for (int i=0; i < n; i++)
  {
    mrb_value elt = RARRAY_PTR(rdiag)[i];

    if (!mrb_fixnum_p(elt) && !mrb_float_p(elt))
      mrb_raisef(mrb, E_ARGUMENT_ERROR,
          "diagonal element %S is not a number",
            mrb_fixnum_value(i));
  }

  int dims[2] = { n, n };
  numarray* A = numarray_alloc(mrb, 2, dims);

  double*restrict eltsA = A->elts;
  for(int i=0; i < n; i++)
  {
    mrb_value elt = RARRAY_PTR(rdiag)[i];
    if (mrb_fixnum_p(elt))
      *eltsA = mrb_fixnum(elt);
    else
      *eltsA = mrb_float(elt);

    eltsA += (n+1);
  }

  return numarray_wrap(mrb, A);
}

static void infer_dimensions(mrb_state* mrb, mrb_value arr, int* ndimsp, int dims[MAX_DIMS])
{
  *ndimsp = 0;

  if (RARRAY_LEN(arr) != 0)
  {
    mrb_value scan_arr = arr;
    while (mrb_array_p(scan_arr))
    {
      if (RARRAY_LEN(scan_arr) == 0)
        mrb_raise(mrb, E_ARGUMENT_ERROR,
            "embedded empty arrays are not allowed");

      dims[ *ndimsp ] = RARRAY_LEN(scan_arr);
      (*ndimsp)++;

      scan_arr = mrb_ary_ref(mrb, scan_arr, 0);
    }
  }
}

static mrb_value cNumArray_infer_dimensions(mrb_state* mrb, mrb_value mod)
{
  mrb_value arr = mrb_nil_value();
  mrb_get_args(mrb, "A", &arr);

  int dims[MAX_DIMS];
  int ndims = 0;

  infer_dimensions(mrb, arr, &ndims, dims);

  mrb_value rdim_elts[MAX_DIMS];
  for (int i=0; i < ndims; i++)
    rdim_elts[i] = mrb_fixnum_value( dims[i] );
  mrb_value rdims = mrb_ary_new_from_values(mrb, ndims, rdim_elts);

  return rdims;
}

static void fill_numarray_from_array(mrb_state* mrb, mrb_value arr, numarray* A, mrb_value* excp)
{
  int indices[MAX_DIMS] = { 0 };
  mrb_value stack[MAX_DIMS+1];

  int ndims = A->ndims;
  stack[0] = arr;
  for(int i=1; i < ndims; i++)
    stack[i] = mrb_ary_ref(mrb,stack[i-1],0);

  while (indices[0] < A->dims[0])
  {
    mrb_value elt = mrb_ary_ref(mrb, stack[ndims-1], indices[ndims-1]);
    double val = 0.0;
    if (mrb_fixnum_p(elt))
      val = (double) mrb_fixnum(elt);
    else if (mrb_float_p(elt))
      val = (double) mrb_float(elt);
    else
    {
      *excp = create_exc(mrb, E_TYPE_ERROR, "invalid type");
      return;
    }

    numarray_ref(A, indices)[0] = val;

    int last_ind = ndims;
    for(int ind=ndims-1; ind >= 0; ind--)
    {
      indices[ind]++;
      if (indices[ind] < A->dims[ind])
      {
        last_ind = ind;
        break;
      }
      else if (ind > 0)
      {
        indices[ind] = 0;
      }
    }

    /* if (ind < ndims-1), then we need to load new arrays for all
     * indices above ind
     */
    if ((last_ind < ndims-1) && (indices[0] < A->dims[0]))
    {
      for (int si = last_ind; si < ndims; si++)
        stack[si+1] = mrb_ary_ref(mrb,stack[si],indices[si]);
    }
  }
}

static numarray* subarray_from_slices(mrb_state* mrb, numarray* A, struct numarray_slice* slices)
{
  int dims[MAX_DIMS] = {0};
  for (int i=0; i < A->ndims; i++)
    dims[i] = slices[i].nelts;

  return numarray_alloc(mrb, A->ndims, dims);
}

static inline void fill_subarray_with_value(numarray* A, size_t nelts, struct numarray_slice* slices, double val)
{
  for (size_t i=0; i < nelts; i++)
    numarray_ref_slices(A, slices, i)[0] = val;
}

static void fill_subarray_with_numarray(mrb_state* mrb, numarray* A,
    size_t nelts, struct numarray_slice* slices, numarray* subA1, mrb_value* excp)
{
  for (size_t i=0; i < nelts; i++)
    numarray_ref_slices(A, slices, i)[0] = subA1->elts[i];
}

static void fill_subarray_with_array(mrb_state* mrb, numarray* A,
    size_t nelts, struct numarray_slice* slices, mrb_value rval, mrb_value* excp)
{
  numarray* A1 = subarray_from_slices(mrb, A, slices);
  fill_numarray_from_array(mrb, rval, A1, excp);
  if (mrb_nil_p(*excp))
    fill_subarray_with_numarray(mrb, A, nelts, slices, A1, excp);
  numarray_free(mrb, A1);
}

static mrb_value cNumArray_from_array(mrb_state* mrb, mrb_value mod)
{
  mrb_value arr = mrb_nil_value();
  mrb_get_args(mrb, "A", &arr);

  int dims[MAX_DIMS];
  int ndims = 0;

  infer_dimensions(mrb, arr, &ndims, dims);

  numarray* A = numarray_alloc(mrb, ndims, dims);

  mrb_value exc = mrb_nil_value();
  fill_numarray_from_array(mrb, arr, A, &exc);

  if (!mrb_nil_p(exc))
  {
    numarray_free(mrb,A);
    mrb_exc_raise(mrb, exc);
  }

  return numarray_wrap(mrb, A);
}

static void
get_indices(mrb_state* mrb, numarray* A, int n, mrb_value* rinds, int inds[MAX_DIMS])
{
  if (n != A->ndims)
    mrb_raise(mrb, E_ARGUMENT_ERROR, "number of indices does not agree with dimension of array");

  for (int i=0; i < n; i++)
  {
    if (!mrb_fixnum_p(rinds[i]))
      mrb_raisef(mrb, E_ARGUMENT_ERROR,
          "dimension %S is not a fixnum", mrb_fixnum_value(i+1));
    inds[i] = mrb_fixnum(rinds[i]);

    // negative indices allowed to wrap (but... only once)
    if (inds[i] < 0) { inds[i] += A->dims[i]; }

    if ((inds[i] < 0) || (inds[i] >= A->dims[i]))
      mrb_raisef(mrb, E_INDEX_ERROR,
          "invalid index %S: %S", mrb_fixnum_value(i), mrb_fixnum_value(inds[i]));
  }
}

static mrb_value cNumArray_initialize(mrb_state* mrb, mrb_value self)
{
  mrb_raise(mrb, E_NOTIMP_ERROR,
      "NumArray cannot be constructed with :new"); /* no returns */
  return self; /* suppresses compiler errors */
}

static mrb_value cNumArray_dims(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  mrb_value dims = mrb_ary_new_capa(mrb, A->ndims);
  for (int i=0; i < A->ndims; i++)
    mrb_ary_push(mrb, dims, mrb_fixnum_value(A->dims[i]));
  return dims;
}

static mrb_value cNumArray_ndims(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_fixnum_value(A->ndims);
}

static mrb_value cNumArray_length(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_fixnum_value(A->nelts);
}

size_t parse_slice_range(mrb_state* mrb, mrb_value rsl, numarray* A, int dim, struct numarray_slice* sl)
{
  mrb_int beg = 0;
  mrb_int len = 0;

  if (!mrb_range_beg_len(mrb, rsl, &beg, &len, A->dims[dim]))
    mrb_raisef(mrb, E_RANGE_ERROR,
        "invalid range at index %S", mrb_fixnum_value(dim));

  sl->type = SLICE_RANGE;
  sl->nelts = len;
  sl->slice.range.beg = beg;
  sl->slice.range.incr = 1;
  sl->slice.range.end = beg+len-1;

  return len;
}

size_t parse_slice_array(mrb_state* mrb, mrb_value rsl, numarray* A, int dim, struct numarray_slice* sl)
{
  mrb_value exc = mrb_nil_value();

  sl->type  = SLICE_LIST;
  sl->nelts = RARRAY_LEN(rsl);
  int* list = mrb_calloc(mrb, sl->nelts, sizeof(sl->slice.list[0]));

  for(int i=0; i < RARRAY_LEN(rsl); i++)
  {
    mrb_value elt = mrb_ary_ref(mrb, rsl, i);
    if (!mrb_fixnum_p(elt))
    {
      exc = create_exc(mrb, E_INDEX_ERROR,
          "expected fixnum in index %d, element %d", dim, i);
      goto error;
    }

    int ind = mrb_fixnum(elt);

    /* negative indices are allowed, as long as:
     *   -A->dims[dim] <= ind
     */
    if (ind < 0) 
      ind += A->dims[dim];

    /* if ind is still negative or greater than the size of the
     * array along that dimension, then the index is bad
     */
    if ((ind < 0) || (ind >= A->dims[dim]))
    {
      exc = create_exc(mrb, E_INDEX_ERROR,
          "invalid value at index %d, element %d: %d",
          dim, i, mrb_fixnum(elt));
      goto error;
    }

    list[i] = ind;
  }

  sl->slice.list = list;
  return sl->nelts;

error:
  mrb_free(mrb, list);
  mrb_exc_raise(mrb, exc); /* never returns */
  return 0; /* suppress compiler error messages */
}

static inline int is_star_sym(mrb_state* mrb, mrb_value rsl)
{
  return mrb_symbol_p(rsl) &&
    (mrb_symbol(rsl) == mrb_intern_lit(mrb,"*"));
}

static inline void slice_all(struct numarray_slice* sl, numarray* A, int dim)
{
  sl->type = SLICE_ALL;
  sl->nelts = A->dims[dim];
}

static size_t parse_slice(mrb_state* mrb, mrb_value rsl, numarray* A, int dim, struct numarray_slice* sl)
{
  enum mrb_vtype tt = mrb_type(rsl);

  if (tt == MRB_TT_FIXNUM)
  {
    int index = mrb_fixnum(rsl);
    if ((index < 0) || (index >= A->dims[dim]))
      mrb_raisef(mrb, E_INDEX_ERROR,
          "invalid index %S: %S",
          mrb_fixnum_value(dim),
          mrb_fixnum_value(index));

    sl->type = SLICE_INDEX;
    sl->nelts = 1;
    sl->slice.index = index;
  }
  else if (tt == MRB_TT_ARRAY)
  {
    parse_slice_array(mrb, rsl, A, dim, sl);
  }
  else if (mrb_nil_p(rsl) || (tt == MRB_TT_TRUE) || is_star_sym(mrb,rsl))
  {
    slice_all(sl,A,dim);
  }
  else if (mrb_range_p(mrb, rsl))
  {
    parse_slice_range(mrb,rsl,A,dim,sl);
  }
  else
  {
    mrb_raisef(mrb, E_INDEX_ERROR, "unknown index type: %S", rsl);
  }

  return sl->nelts;
}

static size_t parse_slices(mrb_state* mrb, int nslices, mrb_value* rslices, numarray* A, struct numarray_slice* slices)
{
  if (nslices != A->ndims)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "number of indices does not agree with dimension of array");

  size_t nelts = 1;
  for (int i=0; i < nslices; i++)
  {
    size_t slice_elts = parse_slice(mrb, rslices[i], A, i, slices+i);
    nelts *= slice_elts;
  }

  return nelts;
}

static mrb_value cNumArray_get(mrb_state* mrb, mrb_value self)
{
  int ninds = 0;
  mrb_value* rinds = NULL;
  mrb_get_args(mrb, "*", &rinds, &ninds);

  numarray* A = numarray_unwrap(mrb,self);

  if (ninds == 1)
  {
    if (mrb_fixnum_p(rinds[0]))
    {
      mrb_int ind = mrb_fixnum(rinds[0]);
      if (ind < 0)
      {
        ind += A->nelts;
      }

      if ((ind < 0) || (ind >= A->nelts))
      {
        mrb_raise(mrb, E_INDEX_ERROR, "invalid linear index");
      }

      return mrb_float_value(mrb, A->elts[ind]);
    }
    else if (mrb_numarray_p(mrb,rinds[0]))
    {
      numarray* B = numarray_unwrap(mrb,rinds[0]);

      if (!numarray_dims_compatible(A,B) &&
          !numarray_compatible_linsize(A,B))
        mrb_raise(mrb, E_ARGUMENT_ERROR,
            "expected equal dimensions or a compatible linear array");

      int count = 0;
      for (size_t i=0; i < B->nelts; i++)
        if (B->elts[i] != 0.0) { ++count; }

      numarray* C = numarray_alloc(mrb, 1, &count);
      double*restrict eltC = C->elts;
      double*restrict eltA = A->elts;
      double*restrict eltB = B->elts;
      double*restrict endA = eltA + A->nelts;
      while (eltA < endA)
      {
        if (*eltB != 0.0)
        {
          *eltC = *eltA;
          eltC++;
        }
        eltA++;
        eltB++;
      }
      return numarray_wrap(mrb, C);
    }
  }

  struct numarray_slice slices[MAX_DIMS];
  memset(slices,0,sizeof(slices));
  size_t nelts = parse_slices(mrb, ninds, rinds, A, slices);

  if (nelts == 1)
  {
    double val = numarray_ref_slices(A, slices, 0)[0];
    slice_list_finalize(mrb, ninds, slices);
    return mrb_float_value(mrb, val);
  }
  else
  {
    numarray* subA = subarray_from_slices(mrb, A, slices);
    for (size_t i=0; i < nelts; i++)
    {
      subA->elts[i] = numarray_ref_slices(A, slices, i)[0];
    }

    slice_list_finalize(mrb, ninds, slices);
    return numarray_wrap(mrb, subA);
  }
}

static mrb_value cNumArray_set(mrb_state* mrb, mrb_value self)
{
  int argc = 0;
  mrb_value* argv = NULL;
  mrb_get_args(mrb, "*", &argv, &argc);

  int ninds = argc-1; /* arg-1 args are the indices */
  mrb_value rval = argv[argc-1]; /* last arg is the value to set */

  numarray* A = numarray_unwrap(mrb,self);

  struct numarray_slice slices[MAX_DIMS];
  memset(slices,0,sizeof(slices));

  if ((argc == 2) && (mrb_numarray_p(mrb,argv[0])))
  {
    numarray* B = numarray_unwrap(mrb,argv[0]);
    int count = 0;
    for(size_t i=0; i < B->nelts; i++)
      if (B->elts[i] != 0.0) ++count;

    if (mrb_fixnum_p(rval) || mrb_float_p(rval))
    {
      double val = mrb_fixnum_p(rval)
        ? (double)mrb_fixnum(rval)
        : (double)mrb_float(rval)
        ;

      double*restrict eltsA = A->elts;
      double*restrict endA  = eltsA + A->nelts;
      double*restrict eltsB = B->elts;

      while(eltsA < endA)
      {
        if (*eltsB != 0.0) { *eltsA = val; }
        eltsA++;
        eltsB++;
      }
    }
    else if (mrb_array_p(rval))
    {
      if (RARRAY_LEN(rval) != count)
        mrb_raisef(mrb, E_ARGUMENT_ERROR,
            "array length expected to be %S", mrb_fixnum_value(count));

      int c = 0;
      double*restrict eltsA = A->elts;
      double*restrict endA  = eltsA + A->nelts;
      double*restrict eltsB = B->elts;

      while(eltsA < endA)
      {
        if (*eltsB != 0.0)
        {
          mrb_value elt = mrb_ary_ref(mrb, rval, c);
          if (mrb_fixnum_p(elt))
            *eltsA = (double)mrb_fixnum(elt);
          else if (mrb_float_p(elt))
            *eltsA = (double)mrb_float(elt);
          else
            mrb_raisef(mrb,E_TYPE_ERROR,
                "expected number at index %S",
                  mrb_fixnum_value(c));
          ++c;
        }

        eltsA++;
        eltsB++;
      }
    }
    else if (mrb_numarray_p(mrb,rval))
    {
      numarray* C = numarray_unwrap(mrb,rval);
      if ((C->ndims != 1) || (C->nelts != count))
        mrb_raisef(mrb, E_ARGUMENT_ERROR,
            "numarray expected to be a vector of length %S",
              mrb_fixnum_value(count));

      double*restrict eltsA = A->elts;
      double*restrict endA  = eltsA + A->nelts;
      double*restrict eltsB = B->elts;

      double*restrict eltsC = C->elts;

      while(eltsA < endA)
      {
        if (*eltsB != 0.0)
        {
          *eltsA = *eltsC;
          eltsC++;
        }

        eltsA++;
        eltsB++;
      }
    }
    else
    {
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "value for assignment should be a number, array, or numarray");
    }

    return self;
  }

  size_t nelts = parse_slices(mrb, ninds, argv, A, slices);

  mrb_value exc = mrb_nil_value();

  if (nelts == 1)
  {
    double* val = numarray_ref_slices(A, slices, 0);
    slice_list_finalize(mrb, ninds, slices);

    if (mrb_fixnum_p(rval))
      *val = (double)(mrb_fixnum(rval));
    else if (mrb_float_p(rval))
      *val = (double)(mrb_float(rval));
    else
      exc = create_exc(mrb, E_ARGUMENT_ERROR,
          "value for assignment should be a number");
  }
  else
  {
    if (mrb_fixnum_p(rval))
      fill_subarray_with_value(A, nelts, slices, (double)mrb_fixnum(rval));
    else if (mrb_float_p(rval))
      fill_subarray_with_value(A, nelts, slices, (double)mrb_float(rval));
    else
    {
      if (mrb_array_p(rval))
        fill_subarray_with_array(mrb, A, nelts, slices, rval, &exc);
      else if (mrb_numarray_p(mrb,rval))
      {
        numarray* A1 = numarray_unwrap(mrb,rval);
        fill_subarray_with_numarray(mrb, A, nelts, slices, A1, &exc);
      }
      else
        exc = create_exc(mrb, E_ARGUMENT_ERROR,
            "value for assignment should be a number, array, or numarray");
    }
  }

  slice_list_finalize(mrb, ninds, slices);
  if (!mrb_nil_p(exc))
    mrb_exc_raise(mrb, exc); /* never returns */

  return self;
}

static mrb_value cNumArray_to_a(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);

  int inds[MAX_DIMS] = { 0 };
  mrb_value ary_stack[MAX_DIMS];

  int nonunitary = 0;
  for (int i=0; i < A->ndims; i++)
  {
    if (A->dims[i] > 1) { nonunitary++; }
  }

  if (nonunitary <= 1)
  {
    size_t ind;

    ary_stack[0] = mrb_ary_new_capa(mrb, A->nelts);
    for (ind = 0; ind < A->nelts; ind++)
    {
      int ai = mrb_gc_arena_save(mrb);
      mrb_ary_push(mrb, ary_stack[0], mrb_float_value(mrb, A->elts[ind]));
      mrb_gc_arena_restore(mrb,ai);
    }
    return ary_stack[0];
  }

  ary_stack[0] = mrb_ary_new_capa(mrb, A->dims[0]);
  for (int i=1; i < A->ndims; i++)
  {
    int ai = mrb_gc_arena_save(mrb);
    ary_stack[i] = mrb_ary_new_capa(mrb, A->dims[i]);
    mrb_ary_push(mrb, ary_stack[i-1], ary_stack[i]);
    mrb_gc_arena_restore(mrb,ai);
  }

  int max_inner = A->dims[A->ndims-1];
  while (inds[0] < A->dims[0])
  {
    int top = A->ndims-1;

    for(inds[top] = 0; inds[top] < max_inner; inds[top]++)
    {
      int ai = mrb_gc_arena_save(mrb);
      mrb_ary_push(mrb,
          ary_stack[top],
          mrb_float_value(mrb,numarray_ref(A, inds)[0]));
      mrb_gc_arena_restore(mrb,ai);
    }

    if (top > 0)
    {
      do {
        inds[top] = 0;
        top--;
        inds[top]++;
      } while ((top>0) && inds[top] >= A->dims[top]);
    }

    if ((top > 0) || (inds[0] < A->dims[0]))
    {
      for(int i=top; i < A->ndims-1; i++)
      {
        int ai = mrb_gc_arena_save(mrb);
        ary_stack[i+1] = mrb_ary_new_capa(mrb, A->dims[i+1]);
        mrb_ary_push(mrb, ary_stack[i], ary_stack[i+1]);
        mrb_gc_arena_restore(mrb,ai);
      }
    }
  }

  return ary_stack[0];
}

static mrb_value cNumArray_matrix_p(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_bool_value( numarray_matrix_p(A) );
}

static mrb_value cNumArray_vector_p(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_bool_value( numarray_vector_p(A) );
}

static mrb_value binary_array_op(mrb_state* mrb, mrb_value self, 
    void(*array_op)(numarray*,numarray*),
    void(*scalar_op)(numarray*,double))
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  numarray* A = numarray_unwrap(mrb,self);
  numarray* C = NULL;

  if (mrb_fixnum_p(other) || mrb_float_p(other))
  {
    if (scalar_op == NULL)
      mrb_raise(mrb, E_ARGUMENT_ERROR, "unsupported operation");

    C = numarray_copy(mrb, A);
    double val;
    if (mrb_fixnum_p(other))
      val = (double)mrb_fixnum(other);
    else
      val = (double)mrb_float(other);

    scalar_op(C, val);
  }
  else if (mrb_numarray_p(mrb,other))
  {
    if (array_op == NULL)
      mrb_raise(mrb, E_ARGUMENT_ERROR, "unsupported operation");

    numarray* B = numarray_unwrap(mrb,other);

    if (!numarray_dims_compatible(A,B))
      mrb_raise(mrb, E_ARGUMENT_ERROR,
          "incompatible numarray dimensions");

    C = numarray_copy(mrb, A);
    array_op(C,B);
  }
  else
  {
    mrb_raise(mrb,E_ARGUMENT_ERROR,
        "expected numarray or number");
  }

  return numarray_wrap(mrb, C);
}

static mrb_value cNumArray_add(mrb_state* mrb, mrb_value self)
{
  return binary_array_op(mrb, self, numarray_add_to, numarray_add_scalar);
}

static mrb_value cNumArray_sub(mrb_state* mrb, mrb_value self)
{
  return binary_array_op(mrb, self, numarray_sub_from, numarray_sub_scalar);
}

static mrb_value cNumArray_mul(mrb_state* mrb, mrb_value self)
{
  return binary_array_op(mrb, self, numarray_mul_by, numarray_mul_scalar);
}

static mrb_value cNumArray_div(mrb_state* mrb, mrb_value self)
{
  return binary_array_op(mrb, self, numarray_div_by, numarray_div_scalar);
}

static void numarray_pow(numarray *A, double p)
{
  size_t i,n;

  for (i=0, n=A->nelts; i < n; i++)
  {
    A->elts[i] = pow(A->elts[i],p);
  }
}

static inline double pow2(double x) { return x*x; }
static inline double pow3(double x) { return (x*x)*x; }
static inline double pow4(double x) { double y = pow2(x); return y*y; }
static inline double pow5(double x) { double y = pow2(x); return y*y*x; }
static inline double pow6(double x) { double y = pow3(x); return y*y; }

static void numarray_ipow(numarray *A, int p)
{
  size_t i,n;

  switch (p)
  {
    case 0:
      numarray_fill(A, 1.0);
      break;

    case 1:
      break;

    case 2:
      for (i=0, n=A->nelts; i < n; i++)
      {
        A->elts[i] = pow2(A->elts[i]);
      }
      break;

    case 3:
      for (i=0, n=A->nelts; i < n; i++)
      {
        A->elts[i] = pow3(A->elts[i]);
      }
      break;

    case 4:
      for (i=0, n=A->nelts; i < n; i++)
      {
        A->elts[i] = pow4(A->elts[i]);
      }
      break;

    case 5:
      for (i=0, n=A->nelts; i < n; i++)
      {
        A->elts[i] = pow5(A->elts[i]);
      }
      break;

    case 6:
      for (i=0, n=A->nelts; i < n; i++)
      {
        A->elts[i] = pow6(A->elts[i]);
      }
      break;

    default:
      numarray_pow(A,(double)p);
      break;
  }
}


static mrb_value cNumArray_pow(mrb_state* mrb, mrb_value self)
{
  mrb_float farg;
  mrb_value obj;
  numarray *A, *newA;

  obj = mrb_nil_value();
  mrb_get_args(mrb, "o", &obj);

  A = numarray_unwrap(mrb,self);

  farg = mrb_to_flo(mrb, obj);
  newA = numarray_copy(mrb, A);

  if (floor(farg) == farg)
  {
    numarray_ipow(newA, (int)farg);
  }
  else
  {
    numarray_pow(newA, (double)farg);
  }

  return numarray_wrap(mrb, newA);
}

/* Matrix-specific functions */

static inline void check_matmul_error(mrb_state* mrb, int err)
{
  if (err != MATRIX_ERROR_NONE)
    mrb_raisef(mrb, E_RUNTIME_ERROR,
        "matrix multiply error %S",
          mrb_fixnum_value(err));
}

static mrb_value cNumArray_negate(mrb_state* mrb, mrb_value self)
{
  numarray *A, *B;
  size_t i;

  A = numarray_unwrap(mrb,self);
  B = numarray_copy(mrb, A);
  for (i = 0; i < B->nelts; i++) { B->elts[i] = -B->elts[i]; }

  return numarray_wrap(mrb, B);
}

static mrb_value cNumArray_matmul(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  numarray* A = numarray_unwrap(mrb,self);
  numarray* B = numarray_unwrap(mrb,other);

  enum MATVEC_TYPE a_matvec = 0;
  enum MATVEC_TYPE b_matvec = 0;

  a_matvec = matvec_type(A);
  b_matvec = matvec_type(B);

  if (a_matvec == MATVEC_ARRAY)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "A % B: A is not a matrix or vector");

  if (b_matvec == MATVEC_ARRAY)
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "A % B: B is not a matrix or vector");

  numarray* C = NULL;

  switch (matvec_op_type(a_matvec,b_matvec))
  {
    case MATMUL_OP_MAT_MAT:
      {
        if (!numarray_matmul_dims_compatible(A,B,1))
          mrb_raise(mrb, E_ARGUMENT_ERROR,
              "A % B: dimensions not compatible");

        int dims[2] = { A->dims[0], B->dims[1] };
        C = numarray_alloc(mrb, 2, dims);

        check_matmul_error(mrb,
            numarray_matrix_multiply(C, A, B));
      }
      break;

    case MATMUL_OP_MAT_VEC:
      {
        if (!numarray_matmul_dims_compatible(A,B,2))
          mrb_raise(mrb, E_ARGUMENT_ERROR,
              "A % B: dimensions not compatible");

        int dims[1] = { A->dims[0] };
        C = numarray_alloc(mrb, 1, dims);

        check_matmul_error(mrb,
            numarray_matvec_multiply(C, A, B));
      }
      break;

    case MATMUL_OP_VEC_MAT:
      {
        if (!numarray_matmul_dims_compatible(A,B,3))
          mrb_raise(mrb, E_ARGUMENT_ERROR,
              "A % B: dimensions not compatible");

        int dims[1] = { B->dims[1] };
        C = numarray_alloc(mrb, 1, dims);

        check_matmul_error(mrb,
            numarray_vecmat_multiply(C, A, B));
      }
      break;

    case MATMUL_OP_VEC_VEC:
      {
        if (!numarray_matmul_dims_compatible(A,B,3))
          mrb_raise(mrb, E_ARGUMENT_ERROR,
              "A % B: dimensions not compatible");
        double dot = 0.0;
        check_matmul_error(mrb,
            numarray_vecdot(&dot, A,B));

        return mrb_float_value(mrb, dot);
      }
      break;

    default:
      mrb_raise(mrb, E_RUNTIME_ERROR,
          "Unknown array/array operation");
      break;
  }

  return numarray_wrap(mrb,C);
}

static mrb_value cNumArray_absmin(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_absmin(A));
}

static mrb_value cNumArray_absmax(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_absmax(A));
}

static mrb_value cNumArray_min(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_min(A));
}

static mrb_value cNumArray_max(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_max(A));
}

static mrb_value cNumArray_mean(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_mean(A));
}

static mrb_value cNumArray_var(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, numarray_var(A));
}

static mrb_value cNumArray_std(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  return mrb_float_value(mrb, sqrt(numarray_var(A)));
}

static mrb_value cNumArray_abs(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  numarray* C = numarray_copy(mrb, A);

  numarray_apply_func(C, fabs);
  return numarray_wrap(mrb, C);
}


/** NumArray comparison functions **/

static mrb_value cNumArray_cmp(mrb_state* mrb, mrb_value self, enum MATRIX_CMP_OP op)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  numarray* A = numarray_unwrap(mrb,self);
  numarray* C = NULL;

  if (mrb_fixnum_p(other) || mrb_float_p(other))
  {
    double val;
    if (mrb_fixnum_p(other))
      val = mrb_fixnum(other);
    else
      val = mrb_float(other);

    C = numarray_copy(mrb, A);
    numarray_compare_scalar(C, val, op);
  }
  else if (mrb_numarray_p(mrb,other))
  {
    numarray* B = numarray_unwrap(mrb,other);
    if (!numarray_dims_compatible(A, B))
      mrb_raise(mrb, E_ARGUMENT_ERROR, "cannot compare arrays of different sizes");

    C = numarray_copy(mrb, A);
    numarray_compare_arrays(C,B,op);
  }
  else
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "expected numarray or scalar");
  }

  return numarray_wrap(mrb, C);
}

static mrb_value cNumArray_eq(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_EQ); }

static mrb_value cNumArray_ne(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_NE); }

static mrb_value cNumArray_ge(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_GE); }

static mrb_value cNumArray_gt(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_GT); }

static mrb_value cNumArray_le(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_LE); }

static mrb_value cNumArray_lt(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_LT); }

static mrb_value cNumArray_spaceship(mrb_state* mrb, mrb_value self)
{ return cNumArray_cmp(mrb, self, MATRIX_CMP_SPACESHIP); }

static mrb_value cNumArray_logical_negate(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  numarray_logical_negate(A);
  return self;
}

static mrb_value cNumArray_logical_or(mrb_state* mrb, mrb_value self)
{
  numarray *A, *B, *C;
  mrb_value rB,rC;

  rB = mrb_nil_value();
  mrb_get_args(mrb, "o", &rB);

  A = numarray_unwrap(mrb,self);
  B = numarray_unwrap(mrb,rB);

  if (A->nelts != B->nelts)
  {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "arrays must have the same number of elements");
  }

  C = NULL;
  rC = numarray_new_filled(mrb, A->ndims, A->dims, 0.0, &C);

  if (numarray_logical_or(C,A,B) < 0)
  {
    mrb_raise(mrb, E_RUNTIME_ERROR, "encountered error OR-ing arrays");
  }

  return rC;
}

static mrb_value cNumArray_anyQ(mrb_state *mrb, mrb_value self)
{
  numarray* A;
  size_t i,n;

  A = numarray_unwrap(mrb,self);
  for (i=0,n=A->nelts; i < n; i++)
  {
    if (A->elts[i] != 0.0) { return mrb_bool_value(1); }
  }

  return mrb_bool_value(0);
}

static mrb_value cNumArray_where(mrb_state *mrb, mrb_value self)
{
  numarray* A;
  size_t i,n;
  mrb_value ret;

  A = numarray_unwrap(mrb,self);
  ret = mrb_ary_new(mrb);
  for (i=0, n=A->nelts; i < n; i++)
  {
    if (A->elts[i] != 0.0)
    {
      mrb_ary_push(mrb, ret, mrb_fixnum_value((mrb_int)i));
    }
  }

  return ret;
}

static mrb_value cNumArray_allQ(mrb_state *mrb, mrb_value self)
{
  numarray* A;
  size_t i,n;

  A = numarray_unwrap(mrb,self);
  for (i=0,n=A->nelts; i < n; i++)
  {
    if (A->elts[i] == 0.0) { return mrb_bool_value(0); }
  }

  return mrb_bool_value(1);
}


/** Matrix routines **/

static mrb_value cNumArray_transpose(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  if (!numarray_matrix_p(A))
    mrb_raise(mrb, E_TYPE_ERROR, "transpose is only supported for matrices");

  numarray* C = numarray_copy_transpose(mrb,A);
  return numarray_wrap(mrb,C);
}

static mrb_value cNumArray_solve(mrb_state* mrb, mrb_value self)
{
  mrb_value other = mrb_nil_value();
  mrb_get_args(mrb, "o", &other);

  numarray* A = numarray_unwrap(mrb,self);
  if (!numarray_matrix_p(A) || (A->dims[0] != A->dims[1]))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "A.solve(b): expected a square matrix for A");

  numarray* B = numarray_unwrap(mrb,other);
  if (!numarray_vector_p(B))
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "A.solve(b): expected a vector for b");

  if (A->dims[0] != B->dims[0])
    mrb_raise(mrb, E_ARGUMENT_ERROR,
        "A.solve(b): sizes of A and b are not equal");

  numarray* X = numarray_copy(mrb,B);

  numarray_solve(mrb, A,X);

  return numarray_wrap(mrb,X);
}

static mrb_value cNumArray_cholesky(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  if (!numarray_matrix_p(A) || (A->dims[0] != A->dims[1]))
    mrb_raise(mrb, E_TYPE_ERROR,
        "A.cholesky: expected a square matrix for A");

  numarray* B = numarray_copy(mrb, A);

  int n = B->dims[0];

  int info = 0;
  dpotrf_( "L", &n, B->elts, &n, &info );

  if (info != 0)
  {
    numarray_free(mrb,B);
    if (info < 0)
      mrb_raisef(mrb, E_RUNTIME_ERROR,
          "error in cholesky factorization (%S)",
            mrb_fixnum_value(info));
    else
      return mrb_nil_value();
  }

  /* fill in upper diagonal */
  double*restrict eltsB = B->elts;
  for (int j=1; j < n; j++)
  {
    for (int i=0; i < j; i++)
    {
      eltsB[i+j*n] = eltsB[j+i*n];
    }
  }

  return numarray_wrap(mrb,B);
}

static mrb_value str_append(mrb_state *mrb, mrb_value str, const char *fmt, ...)
{
  char s[255];
  va_list ap;

  va_start(ap, fmt);
  vsnprintf(s, sizeof(s), fmt, ap);
  va_end(ap);

  if (mrb_nil_p(str))
  {
    return mrb_str_new_cstr(mrb, s);
  }

  return mrb_str_cat_cstr(mrb, str, s);
}

static mrb_value inspect_row(mrb_state *mrb, mrb_value str,
    numarray *A, size_t off, size_t stride, size_t count, int ellipse)
{
  size_t i;

  str = str_append(mrb, str, "[ ");
  for (i=0; i < count; i++, off += stride)
  {
    const char *fmt = "%7.4f ";
    if ((A->elts[off] != 0.0) && (fabs(A->elts[off]) < 1e-3)) { fmt = "%7.4e "; }

    str = str_append(mrb, str, fmt, A->elts[off]);
  }
  return str_append(mrb, str, "]");
}

enum { INSPECT_MAX_COLS = 10, INSPECT_MAX_ROWS = 10 };
static mrb_value cNumArray_inspect(mrb_state* mrb, mrb_value self)
{
  numarray* A = numarray_unwrap(mrb,self);
  int i,nrow,ncol;
  mrb_value ret;
  int row_ellipse,col_ellipse;

  ret = mrb_str_new_cstr(mrb, "");

  nrow = (A->dims[0] <= INSPECT_MAX_ROWS) ? A->dims[0] : INSPECT_MAX_ROWS;
  row_ellipse = (nrow < A->dims[0]);

  if (A->ndims == 1)
  {
    return inspect_row(mrb, ret, A, 0, 1, nrow, row_ellipse);
  }

  ncol = (A->dims[1] <= INSPECT_MAX_COLS) ? A->dims[1] : INSPECT_MAX_COLS;
  col_ellipse = (ncol < A->dims[1]);

  if (A->ndims == 2)
  {
    ret = str_append(mrb, ret, "[ ");
    for (i=0; i < nrow-1; i++)
    {
      ret = inspect_row(mrb, ret, A, i, A->dims[0], ncol, col_ellipse);
      ret = str_append(mrb, ret, "\n  ");
    }
    ret = inspect_row(mrb, ret, A, nrow-1, A->dims[0], ncol, col_ellipse);
    if (row_ellipse)
    {
      return str_append(mrb, ret, "\n  ... ]");
    }

    return str_append(mrb, ret, " ]");
  }

  ret = str_append(mrb, ret, "[NumArray: %d", A->dims[0]);
  for (i=1; i < A->ndims; i++)
  {
    ret = str_append(mrb, ret, "x %d", A->dims[i]);
  }
  return str_append(mrb, ret, "]");
}

/**  Module functions  **/

void numarray_mrb_initialize(mrb_state* mrb)
{
  struct RClass* cNumArray = mrb_define_class(mrb, "NumArray", mrb->object_class);
  MRB_SET_INSTANCE_TT(cNumArray, MRB_TT_DATA);

  mrb_define_module_function(mrb, cNumArray, "debug_break_hook", debug_break_hook, MRB_ARGS_NONE());

  /* array factory methods */
  mrb_define_module_function(mrb, cNumArray, "zeros" , cNumArray_zeros, MRB_ARGS_ANY());
  mrb_define_module_function(mrb, cNumArray, "ones"  , cNumArray_ones  , MRB_ARGS_ANY());
  mrb_define_module_function(mrb, cNumArray, "filled", cNumArray_filled, MRB_ARGS_REQ(2));

  mrb_define_module_function(mrb, cNumArray, "infer_dimensions", cNumArray_infer_dimensions, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, cNumArray, "from_array", cNumArray_from_array, MRB_ARGS_REQ(1));

  /* matrix factory methods */
  mrb_define_module_function(mrb, cNumArray, "eye", cNumArray_eye, MRB_ARGS_REQ(1));
  mrb_define_module_function(mrb, cNumArray, "diag", cNumArray_diag, MRB_ARGS_REQ(1));

  /* NumArray must be created with one of the factory methods, not with
   * 'new'
   */
  mrb_define_method(mrb, cNumArray, "initialize", cNumArray_initialize, MRB_ARGS_ANY());

  mrb_define_method(mrb, cNumArray, "dims", cNumArray_dims, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "ndims", cNumArray_ndims, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "length", cNumArray_length, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "shape", cNumArray_dims, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "sizes", cNumArray_dims, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "rank", cNumArray_ndims, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "[]", cNumArray_get, MRB_ARGS_ANY());
  mrb_define_method(mrb, cNumArray, "[]=", cNumArray_set, MRB_ARGS_ANY());

  mrb_define_method(mrb, cNumArray, "to_a", cNumArray_to_a, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "matrix?", cNumArray_matrix_p, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "vector?", cNumArray_vector_p, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "t", cNumArray_transpose, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "transpose", cNumArray_transpose, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "+", cNumArray_add, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "add", cNumArray_add, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "-", cNumArray_sub, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "sub", cNumArray_sub, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "*", cNumArray_mul, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "mul", cNumArray_mul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "/", cNumArray_div, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "div", cNumArray_div, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "**", cNumArray_pow, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "%", cNumArray_matmul, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "-@", cNumArray_negate, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "abs", cNumArray_abs, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "absmin", cNumArray_absmin, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "min"   , cNumArray_min   , MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "absmax", cNumArray_absmax, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "max"   , cNumArray_max   , MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "mean"  , cNumArray_mean  , MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "var"   , cNumArray_var   , MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "std"   , cNumArray_std   , MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "==", cNumArray_eq, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "!=", cNumArray_ne, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, ">=", cNumArray_ge, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "ge", cNumArray_ge, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, ">" , cNumArray_gt, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "gt", cNumArray_gt, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "<=", cNumArray_le, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "le", cNumArray_le, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "<" , cNumArray_lt, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "lt", cNumArray_lt, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "<=>", cNumArray_spaceship, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "!"  , cNumArray_logical_negate, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "|", cNumArray_logical_or, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cNumArray, "any?", cNumArray_anyQ, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "all?", cNumArray_allQ, MRB_ARGS_NONE());
  mrb_define_method(mrb, cNumArray, "where", cNumArray_where, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "solve", cNumArray_solve, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cNumArray, "cholesky", cNumArray_cholesky, MRB_ARGS_NONE());

  mrb_define_method(mrb, cNumArray, "inspect", cNumArray_inspect, MRB_ARGS_NONE());

  /* silences warnings */
  (void)numarray_matvec_mul_dims_compatible;
  (void)get_indices;
}

#if 0
//
#endif /* 0 */
