#include <ruby.h>

#include "rmpi.h"

#ifdef _POSIX_SOURCE
/* for sleep() */
#  include <unistd.h>
#endif /* _POSIX_SOURCE */

static VALUE mRMPI    = Qnil;
static VALUE cComm    = Qnil;
static VALUE cRequest = Qnil;
static VALUE cStatus  = Qnil;
static VALUE cBuffer  = Qnil;

/** Buffer class **/

static void cBuffer_free(void *p)
{
  struct rmpi_buffer_obj *h= p;
  if (h->buf) {
    rmpi_buffer_decref(h->buf);
    h->buf = NULL;
  }

  free(h);
}

static VALUE cBuffer_allocate_instance(VALUE klass)
{
  struct rmpi_buffer_obj* h = ALLOC(struct rmpi_buffer_obj);
  memset(h, 0, sizeof(*h));

  return Data_Wrap_Struct(klass, 0, cBuffer_free, h);
}

static struct rmpi_buffer_obj *
cBuffer_unwrap(VALUE obj)
{
  struct rmpi_buffer_obj *h;
  Data_Get_Struct(obj, struct rmpi_buffer_obj, h);

  return h;
}

static VALUE 
cBuffer_wrap(struct rmpi_buffer *buf)
{
  VALUE obj;
  struct rmpi_buffer_obj *h;

  obj = cBuffer_allocate_instance(cBuffer);
  h = cBuffer_unwrap(obj);
  rmpi_buffer_incref(buf);
  h->buf = buf;

  return obj;
}

static struct rmpi_buffer_obj *
cBuffer_validate(VALUE obj)
{
  struct rmpi_buffer_obj *h;
  Data_Get_Struct(obj, struct rmpi_buffer_obj, h);

  if (h->buf == NULL) {
    rb_raise(rb_eRuntimeError, "buffer has been released");
  }

  return h;
}

static VALUE
cBuffer_allocate(VALUE klass, VALUE rnum)
{
  size_t n = NUM2SIZET(rnum);

  if (rmpi_buflist_allocate(n) < 0) {
    rb_raise(rb_eNoMemError, "cannot allocate additional buffers");
  }

  return Qnil;
}

static VALUE
cBuffer_num_buffers(VALUE klass)
{
  return SIZET2NUM(rmpi_buflist_count());
}

static VALUE
cBuffer_max_buffers(VALUE klass)
{
  return SIZET2NUM(rmpi_buflist_max());
}

static VALUE 
cBuffer_get_buffer(VALUE klass, VALUE rind)
{
  return Qnil;
}

static VALUE
cBuffer_compactB(VALUE klass)
{
  rmpi_buflist_compact();
  return Qnil;
}

static VALUE
cBuffer_initialize(VALUE self, VALUE rsize)
{
  struct rmpi_buffer_obj *h;
  struct rmpi_buffer *buf;
  size_t size, ind;

  h = cBuffer_unwrap(self);
  size = NUM2SIZET(rsize);

  ind = 0;
  buf = rmpi_buflist_new_buffer(size);
  if (buf == NULL) {
    rb_raise(rb_eNoMemError, "could not allocate a buffer");
  }

  h->buf = buf;

  return self;
}

static VALUE
cBuffer_release(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_unwrap(self);

  if (h->buf != NULL) {
    rmpi_buffer_decref(h->buf);
    h->buf = NULL;
  }

  return Qnil;
}

static VALUE
cBuffer_size(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(self);
  return SIZET2NUM(h->buf->size);
}

static VALUE
cBuffer_capacity(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(self);
  return SIZET2NUM(h->buf->capa);
}

static VALUE
cBuffer_append(VALUE self, VALUE rs)
{
  struct rmpi_buffer_obj *h;
  const char *s;
  size_t len;

  h = cBuffer_validate(self);

  if (rmpi_buffer_is_locked(h->buf)) {
    rb_raise(rb_eRuntimeError, "buffer is locked");
  }

  s = rb_string_value_ptr(&rs);
  len = RSTRING_LEN(rs);

  if (h->buf->size+len > h->buf->capa) {
    rb_raise(rb_eRuntimeError, "buffer is not large enough");
  }

  memcpy(h->buf->ptr+h->buf->size, s, len);
  h->buf->size += len;

  return self;
}

static VALUE
cBuffer_fill(VALUE self, VALUE rs)
{
  struct rmpi_buffer_obj *h;
  const char *s;

  h = cBuffer_validate(self);

  s = rb_string_value_ptr(&rs);
  if (RSTRING_LEN(rs) < 1) {
    rb_raise(rb_eArgError, "argument must be a string of non-zero length");
  }

  if (h->buf->size < h->buf->capa) {
    memset(h->buf->ptr + h->buf->size, s[0], h->buf->capa - h->buf->size);
  }

  h->buf->size = h->buf->capa;

  return self;
}

static VALUE
cBuffer_to_s(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(self);
  return rb_str_new((char*)(h->buf->ptr), h->buf->size);
}

static VALUE
cBuffer_clearB(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(self);
  h->buf->size = 0;

  return self;
}

static VALUE
cBuffer_locked_p(VALUE self)
{
  struct rmpi_buffer_obj *h;
  h = cBuffer_validate(self);

  return rmpi_buffer_is_locked(h->buf) ? Qtrue : Qfalse;
}

static VALUE
cBuffer_ensure_unlock(VALUE self)
{
  struct rmpi_buffer_obj *h;
  h = cBuffer_validate(self);

  rmpi_buffer_unlock(h->buf);
  rmpi_buffer_decref(h->buf);

  return Qnil;
}

static VALUE
cBuffer_lock(VALUE self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(self);
  rb_need_block();

  rmpi_buffer_incref(h->buf);
  if (!rmpi_buffer_lock(h->buf)) {
    rb_raise(rb_eRuntimeError, "lock failed");
  }
  return rb_ensure(rb_yield, self, cBuffer_ensure_unlock, self);
}

/*
cBuffer_inspect
*/

VALUE init_cBuffer(VALUE mod)
{
  cBuffer = rb_define_class_under(mod, "Buffer", rb_cObject);

  rb_define_alloc_func(cBuffer, cBuffer_allocate_instance );

  rb_define_singleton_method(cBuffer, "allocate", cBuffer_allocate, 1);
  rb_define_singleton_method(cBuffer, "max_buffers", cBuffer_max_buffers, 0);
  rb_define_singleton_method(cBuffer, "num_buffers", cBuffer_num_buffers, 0);
  rb_define_singleton_method(cBuffer, "buffer", cBuffer_get_buffer, 1);

  rb_define_singleton_method(cBuffer, "compact!", cBuffer_compactB, 0);

  rb_define_method(cBuffer, "initialize", cBuffer_initialize, 1);

  rb_define_method(cBuffer, "release", cBuffer_release, 0);

  rb_define_method(cBuffer, "size", cBuffer_size, 0);
  rb_define_method(cBuffer, "capacity", cBuffer_capacity, 0);

  rb_define_method(cBuffer, "to_s", cBuffer_to_s, 0);
  /*
  rb_define_method(cBuffer, "inspect", cBuffer_inspect, 0);
  */

  rb_define_method(cBuffer, "locked?", cBuffer_locked_p, 0);

  rb_define_method(cBuffer, "lock", cBuffer_lock, 0);

  rb_define_method(cBuffer, "clear!", cBuffer_clearB, 0);
  rb_define_method(cBuffer, "fill", cBuffer_fill, 1);

  rb_define_method(cBuffer, "<<", cBuffer_append, 1);

  return cBuffer;
}

static void
mpi_check_initialized(void)
{
  int flag;
  flag = 0;
  MPI_Initialized(&flag);
  if (!flag) {
    rb_raise(rb_eRuntimeError, "MPI has not been initialized");
  }
}

/* FIXME: error handling! */
#define CHECKERR(x) (x)

static VALUE
cRequest_wrap(rmpi_request_t *req)
{
  return Data_Wrap_Struct(cRequest, 0, rmpi_request_free, req);
}

#if 0
static VALUE
cRequest_new(rmpi_request_t **reqp, struct rmpi_buffer *buf)
{
  rmpi_request_t *req = request_new(buf);
  if (req == NULL) {
    rb_raise(rb_eNoMemError, "could not allocate a Request");
  }

  if (reqp) { *reqp = req; }

  return cRequest_wrap(req);
}
#endif /* 0 */

static rmpi_request_t *
cRequest_unwrap(VALUE obj)
{
  rmpi_request_t *req;

  req = NULL;
  Data_Get_Struct(obj, rmpi_request_t, req);

  return req;
}

static VALUE
cRequest_initialize(VALUE self)
{
  rb_raise(rb_eRuntimeError, "initialize not implemented");
  return Qnil;
}

static VALUE
cRequest_finished_p(VALUE self)
{
  rmpi_request_t *req = cRequest_unwrap(self);
  return rmpi_request_is_finished(req, 0) ? Qtrue : Qfalse;
}

static VALUE
cRequest_waitB(VALUE self)
{
  rmpi_request_t *req = cRequest_unwrap(self);
  rmpi_request_is_finished(req, 1);
  return Qnil;
}

static VALUE
cRequest_buffer(VALUE self)
{
  rmpi_request_t *req = cRequest_unwrap(self);
  return cBuffer_wrap(req->buf);
}

static void
init_cRequest(VALUE mod)
{
  cRequest = rb_define_class_under(mod, "Request", rb_cObject);

  rb_define_private_method(cRequest, "initialize", cRequest_initialize, 0);

  rb_define_method(cRequest, "finished?", cRequest_finished_p, 0);
  rb_define_method(cRequest, "wait!", cRequest_waitB, 0);

  rb_define_method(cRequest, "buffer", cRequest_buffer, 0);
}

static void
init_cStatus(VALUE mod)
{
  (void)cStatus;
}

static VALUE
cComm_wrap(MPI_Comm comm)
{
  MPI_Comm *commp = ALLOC(MPI_Comm);
  *commp = comm;

  return Data_Wrap_Struct(cComm, 0, xfree, commp);
}

static MPI_Comm
cComm_unwrap(VALUE obj)
{
  MPI_Comm *commp;

  commp = NULL;
  Data_Get_Struct(obj, MPI_Comm, commp);

  return *commp;
}

static VALUE
cComm_rank(VALUE self)
{
  MPI_Comm comm = cComm_unwrap(self);
  int rank;

  CHECKERR( MPI_Comm_rank(comm, &rank) );

  return INT2NUM(rank);
}

static VALUE
cComm_size(VALUE self)
{
  MPI_Comm comm = cComm_unwrap(self);
  int size;

  CHECKERR( MPI_Comm_size(comm, &size) );

  return INT2NUM(size);
}


static VALUE
cComm_initialize(VALUE self)
{
  rb_raise(rb_eRuntimeError, "cannot directly initialize Comm objects");
  return Qnil;
}

static VALUE
cComm_get_world(VALUE klass)
{
  return cComm_wrap(MPI_COMM_WORLD);
}

static VALUE
cComm_send(VALUE self, VALUE rdest, VALUE rtag, VALUE rmsg)
{
  int dest, tag, count;
  void *data;
  MPI_Comm comm;

  comm = cComm_unwrap(self);

  dest = NUM2INT(rdest);
  tag  = NUM2INT(rtag);

  /* FIXME: check rmsg type and handle non-string types */
  Check_Type(rmsg, T_STRING);
  data = rb_string_value_ptr(&rmsg);
  count = RSTRING_LEN(rmsg);

  CHECKERR( MPI_Send(data, count, MPI_BYTE, dest, tag, comm) );

  return Qnil;
}

static VALUE
cComm_recv(int argc, VALUE argv[], VALUE self)
{
  VALUE rsrc, rtag, rdata;
  int src,tag,count;
  void *data;

  MPI_Comm comm;
  MPI_Status status;

  rsrc = rtag = rdata = Qnil;
  rb_scan_args(argc, argv, "21", &rsrc, &rtag, &rdata);

  comm = cComm_unwrap(self);

  src = NUM2INT(rsrc);
  tag  = NUM2INT(rtag);

  if (rdata != Qnil) { 
    /* FIXME: check rmsg type and handle non-string types */
    Check_Type(rdata, T_STRING);
    data = rb_string_value_ptr(&rdata);
    count = RSTRING_LEN(rdata);

    /* FIXME: Use a real status argument */
    CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, &status) );
    return rb_ary_new3(2, INT2NUM(status.MPI_SOURCE), INT2NUM(status.MPI_TAG));
  } else {
    MPI_Status status;
    int code;
    VALUE str;

    memset(&status,0,sizeof(status));
    CHECKERR( code = MPI_Probe(src, tag, comm, &status) );

    count = 0;
    CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

    str = rb_str_buf_new(count);
    data = RSTRING_PTR(str);
    
    CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, MPI_STATUS_IGNORE) );

    rb_str_set_len(str, count);

    return rb_ary_new3(3, str, INT2NUM(status.MPI_SOURCE), INT2NUM(status.MPI_TAG));
  }
}

static VALUE
cComm_probe(VALUE self, VALUE rsrc, VALUE rtag)
{
  int src, tag, count;
  MPI_Comm comm;
  MPI_Status status;

  comm = cComm_unwrap(self);

  src = NUM2INT(rsrc);
  tag  = NUM2INT(rtag);

  memset(&status,0,sizeof(status));
  CHECKERR( MPI_Probe(src, tag, comm, &status) );

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  return rb_ary_new3(3,
      INT2NUM(status.MPI_SOURCE),
      INT2NUM(status.MPI_TAG),
      INT2NUM(count));
}

static VALUE
cComm_iprobe(VALUE self, VALUE rsrc, VALUE rtag)
{
  int src, tag, count, flag;
  MPI_Comm comm;
  MPI_Status status;

  comm = cComm_unwrap(self);

  src = NUM2INT(rsrc);
  tag  = NUM2INT(rtag);

  flag = 0;
  memset(&status,0,sizeof(status));
  CHECKERR( MPI_Iprobe(src, tag, comm, &flag, &status) );

  if (!flag) { return Qnil; }

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  return rb_ary_new3(3,
      INT2NUM(status.MPI_SOURCE),
      INT2NUM(status.MPI_TAG),
      INT2NUM(count));
}

static VALUE
cComm_isend(VALUE self, VALUE rdst, VALUE rtag, VALUE rmsg)
{
  MPI_Comm comm;
  int dst,tag,count;
  rmpi_request_t *req;
  struct rmpi_buffer *buf;

  comm = cComm_unwrap(self);

  dst = NUM2INT(rdst);
  tag = NUM2INT(rtag);

  /* We incref for existing buffers but not for new buffers.  Why?
   * 1. New buffers have a reference count of one.
   * 2. Creating a request with a buffer will lock the buffer and
   *    increase its reference count by one.
   * 3. We want an existing buffer to have a net increase in its
   *    reference count of one, and a new buffer to have a single
   *    reference (the request).
   * 4. After we create the request, a newly-created buffer will
   *    have a reference count of two, so we have to decref.  An
   *    existing buffer will have one additional reference, so when we
   *    decref, it would have no net increase in its reference count.
   *    Incref'ing restores this.
   */
  if (rb_type(rmsg) == RUBY_T_STRING) {
    count = RSTRING_LEN(rmsg);
    buf = rmpi_buflist_new_buffer(count);
    if (buf == NULL) {
      rb_raise(rb_eNoMemError, "could not allocate buffer for string");
    }

    memcpy(buf->ptr, RSTRING_PTR(rmsg), count);
    buf->size = count;
  } else {
    struct rmpi_buffer_obj *h;
    h = cBuffer_validate(rmsg);
    buf = h->buf;
    rmpi_buffer_incref(buf);
  }

  req = rmpi_request_new(buf);
  rmpi_buffer_decref(buf);

  if (req == NULL) {
    rb_raise(rb_eNoMemError, "could not allocate a Request object");
  }

  CHECKERR( MPI_Isend(req->buf->ptr, req->buf->size, MPI_BYTE, dst, tag, comm, &req->req) );

  return cRequest_wrap(req);
}

static VALUE
cComm_irecv(int argc, VALUE argv[], VALUE self)
{
  MPI_Comm comm;
  int src,tag,count;
  void *data;
  VALUE rsrc,rtag,rmsg;
  MPI_Status status;
  VALUE str;
  int flag = 0;

  comm = cComm_unwrap(self);

  rmsg = Qnil;
  rb_scan_args(argc,argv, "21", &rsrc, &rtag, &rmsg);

  src = NUM2INT(rsrc);
  tag = NUM2INT(rtag);

  if (rmsg != Qnil) {
    rb_raise(rb_eRuntimeError, "not yet implemented!");
  }

  memset(&status, 0, sizeof(status));
  CHECKERR( MPI_Iprobe(src,tag, comm, &flag, &status) );
  if (!flag) {
    return Qnil;
  }

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  str = rb_str_buf_new(count);
  data = RSTRING_PTR(str);

  CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, MPI_STATUS_IGNORE) );

  rb_str_set_len(str, count);

  return rb_ary_new3(3, str, INT2NUM(status.MPI_SOURCE), INT2NUM(status.MPI_TAG));
}

static void
init_cComm(VALUE mod)
{
  cComm = rb_define_class_under(mod, "Comm", rb_cObject);

  rb_define_private_method(cComm, "initialize", cComm_initialize, 0);

  rb_define_method(cComm, "rank", cComm_rank, 0);
  rb_define_method(cComm, "size", cComm_size, 0);

  rb_define_method(cComm, "Send", cComm_send, 3);
  rb_define_method(cComm, "Recv", cComm_recv, -1);

  rb_define_method(cComm, "Probe", cComm_probe, 2);
  rb_define_method(cComm, "Iprobe", cComm_iprobe, 2);

  rb_define_method(cComm, "Isend", cComm_isend, 3);
  rb_define_method(cComm, "Irecv", cComm_irecv, -1);

#if 0
  rb_define_method(cComm, "Gather", rb_comm_gather, 3);
  rb_define_method(cComm, "Allgather", rb_comm_allgather, 2);
  rb_define_method(cComm, "Bcast", rb_comm_bcast, 2);
  rb_define_method(cComm, "Scatter", rb_comm_scatter, 3);
  rb_define_method(cComm, "Sendrecv", rb_comm_sendrecv, 6);
  rb_define_method(cComm, "Alltoall", rb_comm_alltoall, 2);
  rb_define_method(cComm, "Reduce", rb_comm_reduce, 4);
  rb_define_method(cComm, "Allreduce", rb_comm_allreduce, 3);
  rb_define_method(cComm, "Errhandler", rb_comm_get_Errhandler, 0);
  rb_define_method(cComm, "Errhandler=", rb_comm_set_Errhandler, 1);
  rb_define_method(cComm, "Barrier", rb_comm_barrier, 0);
#endif /* 0 */

  rb_define_singleton_method(cComm, "world", cComm_get_world, 0);
}

static int is_initialized = 0;
static int is_finalized   = 0;

static void do_finalize(void)
{
  if (is_initialized && !is_finalized) {
    MPI_Finalize();
    is_finalized = 1;
  }
}

static VALUE
mRMPI_init(int argc, VALUE rargv[], VALUE mod)
{
  VALUE progname;
  int i;
  char **argv;

  argv = ALLOC_N(char *, argc+1);

  progname = rb_gv_get("$0");
  argv[0] = StringValueCStr(progname);

  for (i=0; i < argc; i++) {
    if (rb_type(rargv[i]) == T_STRING) {
      argv[i+1] = StringValueCStr(rargv[i]);
    } else {
      argv[i+1] = "";
    }
  }

  argc += 1;
  fprintf(stderr, "\n\n---\nargc = %d\n",argc);
  for(i=0; i < argc; i++) {
    fprintf(stderr, "[%3d] %s\n", i, argv[i]);
  }
  fprintf(stderr, "---\n\n");

  MPI_Init(&argc, &argv);
  if (!is_initialized) {
    is_initialized = 1;
    is_finalized = 0;
    atexit(do_finalize);
  }

  xfree(argv);

  return Qnil;
}

static VALUE
mRMPI_finalize(int argc, VALUE rargv[], VALUE mod)
{
  do_finalize();
  return Qnil;
}

static VALUE mRMPI_rank(VALUE mod)
{
  int rank;
  mpi_check_initialized();

  CHECKERR( MPI_Comm_rank(MPI_COMM_WORLD, &rank) );

  return INT2NUM(rank);
}

static VALUE mRMPI_size(VALUE mod)
{
  int size;
  mpi_check_initialized();

  CHECKERR( MPI_Comm_size(MPI_COMM_WORLD, &size) );

  return INT2NUM(size);
}

static VALUE call_block_by_rank(VALUE mod, int callrank)
{
  int rank = -1;
  VALUE comm;

  mpi_check_initialized();
  rb_need_block();

  CHECKERR( MPI_Comm_rank(MPI_COMM_WORLD, &rank) );

  if ((rank == callrank) || ((callrank < 0) && (rank != 0))) {
    comm = cComm_wrap(MPI_COMM_WORLD);
    rb_yield(comm);
  }

  return mod;
}

static VALUE mRMPI_foreman(VALUE mod)
{
  return call_block_by_rank(mod, 0);
}

static VALUE mRMPI_worker(int argc, VALUE argv[], VALUE mod)
{
  VALUE rnum;

  rnum = Qnil;
  rb_scan_args(argc, argv, "01", &rnum);

  if (rnum == Qnil) {
    return call_block_by_rank(mod, 1);
  }

  return call_block_by_rank(mod, NUM2INT(rnum));
}

static VALUE mRMPI_waitB(VALUE mod)
{
  rmpi_request_finish_waiting();
  return Qnil;
}

static VALUE
mRMPI_debug_hook(VALUE mod)
{
  volatile int flag = 0;
  while (!flag) {
#ifdef _POSIX_SOURCE
    sleep(5);
#endif /* _POSIX_SOURCE */
  }

  return Qnil;
}

void Init_rmpi(void)
{
  mRMPI = rb_define_module("RMPI");

  rb_define_module_function(mRMPI, "Init", mRMPI_init, -1);
  rb_define_module_function(mRMPI, "Finalize", mRMPI_finalize, -1);

  rb_define_module_function(mRMPI, "rank", mRMPI_rank, 0);
  rb_define_module_function(mRMPI, "size", mRMPI_size, 0);

  rb_define_module_function(mRMPI, "foreman", mRMPI_foreman, 0);
  rb_define_module_function(mRMPI, "worker", mRMPI_worker, -1);

  rb_define_module_function(mRMPI, "wait!", mRMPI_waitB, 0);

  rb_define_module_function(mRMPI, "debug_hook", mRMPI_debug_hook, 0);

  // rb_define_module_function(mMPI, "Abort", mRMPI_abort, 2);
  rb_define_const(mRMPI, "VERSION", INT2NUM(MPI_VERSION));
  rb_define_const(mRMPI, "SUBVERSION", INT2NUM(MPI_SUBVERSION));
#if 0
  rb_define_const(mMPI, "SUCCESS", INT2NUM(MPI_SUCCESS));
  rb_define_const(mMPI, "PROC_NULL", INT2NUM(MPI_PROC_NULL));
#endif /* 0 */

  init_cBuffer(mRMPI);
  init_cRequest(mRMPI);
  init_cStatus(mRMPI);
  init_cComm(mRMPI);

  (void)rmpi_buffer_lookup;
}

