#include "rmpi.h"

#include <stdlib.h>
#include <string.h>

/* FIXME: error handling! */
#define CHECKERR(x) (x)

static struct rmpi_buflist *priv_buflists = NULL;

static inline struct rmpi_buflist *
rmpi_buflists(void)
{
  return priv_buflists;
}

static void
rmpi_buflist_append(struct rmpi_buflist *newbufs)
{
  struct rmpi_buflist **listp;

  listp = &priv_buflists;
  while (*listp) {
    listp = &(*listp)->next;
  }

  *listp = newbufs;
}

static struct rmpi_buflist *
reverse_buflists(struct rmpi_buflist *list)
{
  struct rmpi_buflist *head, *item, *next;

  head = NULL;

  item = list;
  while (item) {
    next = item->next;
    item->next = head;
    head = item;

    item = next;
  }

  return head;
}

void
rmpi_buflist_compact(void)
{
  struct rmpi_buflist *head, *orig_head, *list;
  list = orig_head = rmpi_buflists();

  /* walk buffer lists, reversing so the tail is the head */
  head = reverse_buflists(list);

  while (head) {
    size_t j;
    for (j=0; j < head->n; j++) {
      if (head->bufs[j].ptr != NULL) {
        goto cleanup;
      }
    }

    /* all buffers are released, so we can free the block */
    list = head;
    head = head->next;
    free(list);
  }

cleanup:
  if (head) {
    head = reverse_buflists(head);
    /* We have to clear the freelist; it may have entries that point to
     * buflists that we've removed
     */
    head->freelist = NULL;
  }

  priv_buflists = head;
}

static void rmpi_buflist_add_freelist(struct rmpi_buffer *buf)
{
  struct rmpi_buflist *head = rmpi_buflists();
  assert( head != NULL );

  buf->aux_type = RMPI_BUF_FREED;
  buf->aux.next = head->freelist;

  head->freelist = buf;
}

int
rmpi_buflist_allocate(size_t n)
{
  struct rmpi_buflist *newbufs;

  newbufs = malloc(sizeof(*newbufs) + n*sizeof(newbufs->bufs[0]));
  if (newbufs == NULL) {
    return -1;
  }

  newbufs->next = NULL;
  newbufs->freelist = NULL;
  newbufs->n = 0;
  newbufs->capa = n;
  memset(newbufs->bufs, 0, n*sizeof(newbufs->bufs[0]));

  rmpi_buflist_append(newbufs);

  return 0;
}

size_t
rmpi_buflist_max(void)
{
  struct rmpi_buflist *list;
  size_t n;

  for(n=0, list = rmpi_buflists(); list; list = list->next) {
    n += list->capa;
  }

  return n;
}

size_t
rmpi_buflist_count(void)
{
  struct rmpi_buflist *list;
  size_t n;

  for(n=0, list = rmpi_buflists(); list; list = list->next) {
    size_t i;
    for (i=0; i < list->n; i++) {
      if (list->bufs[i].ptr != NULL) { n++; }
    }
  }

  return n;
}

static struct rmpi_buflist *
rmpi_buflist_lookup(size_t *ind)
{
  struct rmpi_buflist *list;

  assert( ind != NULL );

  list = rmpi_buflists();
  while (list && (*ind >= list->capa)) {
    *ind -= list->capa;
    list = list->next;
  }

  assert( *ind >= 0 );
  return list;
}

static struct rmpi_buffer *
rmpi_buflist_alloc(void)
{
  struct rmpi_buflist *list;
  struct rmpi_buffer *buf;

  list = rmpi_buflists();
  if (list == NULL) {
    if (rmpi_buflist_allocate(RMPI_BUFLIST_DEFAULT_SIZE) < 0) {
      return NULL;
    }
    list = rmpi_buflists();
  }

  if (list->freelist) {
    buf = list->freelist;
    list->freelist = buf->aux.next;
    return buf;
  }

  while (list) {
    size_t i;

    if (list->n < list->capa) {
      buf = list->bufs+list->n;
      list->n++;
      return buf;
    }

    for(i=0; i < list->n; i++) {
      if (list->bufs[i].ptr == NULL) {
        return list->bufs+i;
      }
    }

    list = list->next;
  }

  return NULL;
}

static int rmpi_buffer_initialize(struct rmpi_buffer *buf, size_t capa);
static void rmpi_buffer_incref(struct rmpi_buffer *buf);

struct rmpi_buffer *
rmpi_buflist_new_buffer(size_t capa)
{
  struct rmpi_buffer *buf;

  buf = rmpi_buflist_alloc();
  if (buf == NULL) {
    return NULL;
  }

  if (rmpi_buffer_initialize(buf, capa) < 0) {
    rmpi_buflist_add_freelist(buf);
    return NULL;
  }

  rmpi_buffer_incref(buf);

  return buf;
}

struct rmpi_buffer *
rmpi_buffer_lookup(size_t i)
{
  size_t li;
  struct rmpi_buflist *list;

  li = i;
  list = rmpi_buflist_lookup(&li);

  return list ? list->bufs+li : NULL;
}

static int rmpi_buffer_initialize(struct rmpi_buffer *buf, size_t capa)
{
  unsigned char *ptr = malloc(capa);

  if (ptr == NULL) {
    return -1;
  }

  buf->ptr = ptr;
  buf->size = 0;
  buf->capa = capa;

  buf->lock_count = 0;
  buf->aux_type = RMPI_BUF_NO_DATA;
  buf->aux.next = NULL;

  return 0;
}

void
rmpi_buffer_finalize(struct rmpi_buffer *buf)
{
  if (buf->ptr) {
    free(buf->ptr);
    buf->ptr = NULL;
  }

  rmpi_buflist_add_freelist(buf);
}

rmpi_request_t *req_freelist = NULL;
rmpi_request_t *req_waiting = NULL;
struct rmpi_request_block *req_blocks = NULL;

static void
request_check_waiting(void)
{
  rmpi_request_t **reqp;
  reqp = &req_waiting;
  while (*reqp) {
    int flag = 0;
    CHECKERR( MPI_Test(&(*reqp)->req, &flag, MPI_STATUS_IGNORE) );
    if (flag) {
      rmpi_request_t *req = *reqp;
      *reqp = (*reqp)->next;
      req->next = req_freelist;
      req_freelist = req;
    } else {
      reqp = &(*reqp)->next;
    }
  }
}

static rmpi_request_t *
request_alloc(void)
{
  rmpi_request_t *req;
  struct rmpi_request_block *block;

  request_check_waiting();

  if (req_freelist != NULL) {
    req = req_freelist;
    req_freelist = req_freelist->next;
    return req;
  }

  if ((req_blocks == NULL) || (req_blocks->n >= RMPI_REQUEST_BLOCK_SIZE)) {
    /* Try to get another block of requests */
    block = malloc(sizeof(*block));
    if (block == NULL) {
      return NULL;
    }

    block->n = 0;
    block->next = req_blocks;
    req_blocks = block;
  }

  return &req_blocks->block[req_blocks->n++];
}

rmpi_request_t *
rmpi_request_new(struct rmpi_buffer *buf)
{
  rmpi_request_t *req;

  req = request_alloc();
  memset(req,0,sizeof(*req));
  if (buf) {
    req->buf = buf;
    rmpi_buffer_incref(buf);
    rmpi_buffer_lock(buf);
  }

  return req;
}

static void
request_dealloc(rmpi_request_t *req)
{
  if (req->buf) {
    rmpi_buffer_decref(req->buf);
  }

  req->next = req_freelist;
  req_freelist = req;
}

void
rmpi_request_finish_waiting(void)
{
  rmpi_request_t *req, *next;
  req = req_waiting;
  while (req) {
    CHECKERR( MPI_Wait(&req->req, MPI_STATUS_IGNORE) );

    next = req->next;
    request_dealloc(req);

    req = next;
  }
}

void
rmpi_request_free(void *p)
{
  int flag;
  rmpi_request_t *req = p;

  if (req->buf == NULL) {
    request_dealloc(req);
    return;
  }

  flag = 0;
  CHECKERR( MPI_Test(&req->req, &flag, MPI_STATUS_IGNORE) );

  if (flag) {
    request_dealloc(req);
    return;
  }

  req->next = req_waiting;
  req_waiting = req;
}

int
rmpi_request_is_finished(rmpi_request_t *req, int wait)
{
  int completed;

  if (wait) {
    CHECKERR( MPI_Wait(&req->req, MPI_STATUS_IGNORE) );
    completed = 1;
  } else {
    completed = 0;
    CHECKERR( MPI_Test(&req->req, &completed, MPI_STATUS_IGNORE) );
  }

  if (completed && req->buf) {
    rmpi_buffer_unlock(req->buf);
    rmpi_buffer_decref(req->buf);
    req->buf = NULL;
  }

  return completed;
}

