#include "rmpi.h"

#include <mruby.h>
#include <mruby/class.h>
#include <mruby/data.h>
#include <mruby/array.h>
#include <mruby/string.h>

#include <mpi.h>

#include <stdlib.h>
#include <string.h>

#ifdef _POSIX_SOURCE
/* for sleep() */
#  include <unistd.h>
#endif /* _POSIX_SOURCE */

#include <assert.h>

static inline struct RClass *get_rmpi_module(mrb_state *mrb)
{
  return mrb_module_get(mrb, "RMPI");
}

/** Buffer class **/

static inline struct RClass *get_buffer_class(mrb_state *mrb)
{
  return mrb_class_get_under(mrb, get_rmpi_module(mrb), "Buffer");
}

static void cBuffer_free(mrb_state *mrb, void *p)
{
  struct rmpi_buffer_obj *h= p;
  if (h->buf) {
    rmpi_buffer_decref(h->buf);
    h->buf = NULL;
  }

  mrb_free(mrb,h);
}

static const mrb_data_type buffer_type = {
  .struct_name = "Buffer",
  .dfree = cBuffer_free,
};

static struct rmpi_buffer_obj *
cBuffer_unwrap(mrb_value obj)
{
  return DATA_PTR(obj);
}

static mrb_value 
cBuffer_wrap(mrb_state *mrb, struct rmpi_buffer *buf)
{
  struct rmpi_buffer_obj *h;
  struct RData *data;

  h = mrb_malloc(mrb, sizeof(*h));
  h->buf = buf;
  rmpi_buffer_incref(buf);

  data = mrb_data_object_alloc(mrb, get_buffer_class(mrb), h, &buffer_type);

  return mrb_obj_value(data);
}

static struct rmpi_buffer_obj *
cBuffer_validate(mrb_state *mrb, mrb_value obj)
{
  struct rmpi_buffer_obj *h;
  h = cBuffer_unwrap(obj);

  if (h->buf == NULL) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "buffer has been released");
  }

  return h;
}

static mrb_value
cBuffer_allocate(mrb_state *mrb, mrb_value klass)
{
  mrb_int n = 0;
  mrb_get_args(mrb, "i", &n);

  if (rmpi_buflist_allocate(n) < 0) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "cannot allocate additional buffers");
  }

  return mrb_nil_value();
}

static mrb_value
cBuffer_num_buffers(mrb_state *mrb, mrb_value klass)
{
  return mrb_fixnum_value(rmpi_buflist_count());
}

static mrb_value
cBuffer_max_buffers(mrb_state *mrb, mrb_value klass)
{
  return mrb_fixnum_value(rmpi_buflist_max());
}

static mrb_value
cBuffer_compactB(mrb_state *mrb, mrb_value klass)
{
  rmpi_buflist_compact();
  return mrb_nil_value();
}

static mrb_value
cBuffer_initialize(mrb_state *mrb, mrb_value self)
{
  mrb_int size;
  struct rmpi_buffer_obj *h;
  struct rmpi_buffer *buf;

  size = 0;
  mrb_get_args(mrb, "i", &size);
  if (size <= 0) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "buffer size must be > 0");
  }

  h = mrb_malloc(mrb, sizeof(*h));

  buf = rmpi_buflist_new_buffer(size);
  if (buf == NULL) {
    mrb_free(mrb, h);
    mrb_raise(mrb, E_RUNTIME_ERROR, "could not allocate a buffer");
  }

  h->buf = buf;
  DATA_PTR(self) = h;
  DATA_TYPE(self) = &buffer_type;

  return self;
}

static mrb_value
cBuffer_release(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_unwrap(self);

  if (h->buf != NULL) {
    rmpi_buffer_decref(h->buf);
    h->buf = NULL;
  }

  return mrb_nil_value();
}

static mrb_value
cBuffer_size(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(mrb,self);
  return mrb_fixnum_value(h->buf->size);
}

static mrb_value
cBuffer_capacity(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(mrb, self);
  return mrb_fixnum_value(h->buf->capa);
}

static mrb_value
cBuffer_append(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;
  char *s;
  mrb_int len;

  h = cBuffer_validate(mrb, self);
  if (rmpi_buffer_is_locked(h->buf)) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "buffer is locked");
  }

  len = 0;
  s = NULL;
  mrb_get_args(mrb, "s", &s, &len);

  if (h->buf->size+len > h->buf->capa) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "buffer is not large enough");
  }

  memcpy(h->buf->ptr+h->buf->size, s, len);
  h->buf->size += len;

  return self;
}

static mrb_value
cBuffer_fill(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;
  const char *s;
  mrb_int len;

  h = cBuffer_validate(mrb, self);

  len = 0;
  s = NULL;
  mrb_get_args(mrb, "s", &s, &len);

  if (len < 1) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "argument must be a string of non-zero length");
  }

  if (h->buf->size < h->buf->capa) {
    memset(h->buf->ptr + h->buf->size, s[0], h->buf->capa - h->buf->size);
  }

  h->buf->size = h->buf->capa;

  return self;
}

static mrb_value
cBuffer_to_s(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(mrb, self);
  return mrb_str_new(mrb, (const char*)h->buf->ptr, h->buf->size);
}

static mrb_value
cBuffer_clearB(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;

  h = cBuffer_validate(mrb, self);
  h->buf->size = 0;

  return self;
}

static mrb_value
cBuffer_locked_p(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;
  h = cBuffer_validate(mrb,self);

  return mrb_bool_value( rmpi_buffer_is_locked(h->buf) );
}

static mrb_value
cBuffer_lock(mrb_state *mrb, mrb_value self)
{
  struct rmpi_buffer_obj *h;
  struct rmpi_buffer *buf;
  struct mrb_jmpbuf *prevjmp;
  mrb_value block, ret;

  h = cBuffer_validate(mrb, self);
  buf = h->buf;

  block = mrb_nil_value();
  mrb_get_args(mrb, "&", &block);

  if (mrb_nil_p(block)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "requires block");
  }

  rmpi_buffer_incref(buf);
  if (!rmpi_buffer_lock(buf)) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "lock failed");
  }

  prevjmp = mrb->jmp;
  mrb->jmp = NULL;

  ret = mrb_funcall(mrb, block, "call", 0);

  rmpi_buffer_unlock(buf);
  rmpi_buffer_decref(buf);

  mrb->jmp = prevjmp;
  if (mrb->exc) {
    struct RObject *exc = mrb->exc;
    mrb->exc = 0;
    mrb_exc_raise(mrb, mrb_obj_value(exc));
  }

  return ret;
}
#if 0
#endif /* 0 */

struct RClass *init_cBuffer(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cBuffer;

  cBuffer = mrb_define_class_under(mrb, mod, "Buffer", mrb->object_class);
  MRB_SET_INSTANCE_TT(cBuffer, MRB_TT_DATA);

  mrb_define_singleton_method(mrb, (struct RObject*)cBuffer, "allocate", cBuffer_allocate, MRB_ARGS_REQ(1));
  mrb_define_singleton_method(mrb, (struct RObject*)cBuffer, "max_buffers", cBuffer_max_buffers, MRB_ARGS_NONE());
  mrb_define_singleton_method(mrb, (struct RObject*)cBuffer, "num_buffers", cBuffer_num_buffers, MRB_ARGS_NONE());

  mrb_define_singleton_method(mrb, (struct RObject*)cBuffer, "compact!", cBuffer_compactB, MRB_ARGS_NONE());

  mrb_define_method(mrb, cBuffer, "initialize", cBuffer_initialize, MRB_ARGS_REQ(1));
  mrb_define_method(mrb, cBuffer, "release", cBuffer_release, MRB_ARGS_NONE());

  mrb_define_method(mrb, cBuffer, "size", cBuffer_size, MRB_ARGS_NONE());
  mrb_define_method(mrb, cBuffer, "capacity", cBuffer_capacity, MRB_ARGS_NONE());

  mrb_define_method(mrb, cBuffer, "to_s", cBuffer_to_s, MRB_ARGS_NONE());
  /*
  rb_define_method(mrb, cBuffer, "inspect", cBuffer_inspect, MRB_ARGS_NONE());
  */

  mrb_define_method(mrb, cBuffer, "locked?", cBuffer_locked_p, MRB_ARGS_NONE());

  mrb_define_method(mrb, cBuffer, "lock", cBuffer_lock, MRB_ARGS_NONE());

  mrb_define_method(mrb, cBuffer, "clear!", cBuffer_clearB, MRB_ARGS_NONE());
  mrb_define_method(mrb, cBuffer, "fill", cBuffer_fill, MRB_ARGS_REQ(1));

  mrb_define_method(mrb, cBuffer, "<<", cBuffer_append, MRB_ARGS_REQ(1));

  return cBuffer;
}

static void
mpi_check_initialized(mrb_state *mrb)
{
  int flag;
  flag = 0;
  MPI_Initialized(&flag);
  if (!flag) {
    mrb_raise(mrb, E_RUNTIME_ERROR, "MPI has not been initialized");
  }
}

/* FIXME: error handling! */
#define CHECKERR(x) (x)

static void cRequest_free(mrb_state *mrb, void *data)
{
  rmpi_request_free(data);
}

static const mrb_data_type req_type = {
  .struct_name = "Request",
  .dfree = cRequest_free,
};

static inline struct RClass *get_request_class(mrb_state *mrb)
{
  return mrb_class_get_under(mrb, get_rmpi_module(mrb), "Request");
}

static mrb_value
cRequest_wrap(mrb_state *mrb, rmpi_request_t *req)
{
  struct RData* data;

  data = mrb_data_object_alloc(mrb, 
      get_request_class(mrb), req, &req_type);

  return mrb_obj_value(data);
}

static rmpi_request_t *
cRequest_unwrap(mrb_state *mrb, mrb_value obj)
{
  return DATA_PTR(obj);
}

static mrb_value
cRequest_finished_p(mrb_state *mrb, mrb_value self)
{
  rmpi_request_t *req = cRequest_unwrap(mrb, self);
  return mrb_bool_value(rmpi_request_is_finished(req, 0));
}

static mrb_value
cRequest_waitB(mrb_state *mrb, mrb_value self)
{
  rmpi_request_t *req = cRequest_unwrap(mrb, self);
  rmpi_request_is_finished(req, 1);
  return mrb_nil_value();
}

static mrb_value
cRequest_buffer(mrb_state *mrb, mrb_value self)
{
  rmpi_request_t *req = cRequest_unwrap(mrb, self);
  return cBuffer_wrap(mrb, req->buf);
}

static struct RClass *
init_cRequest(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cRequest;

  cRequest = mrb_define_class_under(mrb, mod, "Request", mrb->object_class);

  mrb_undef_method(mrb, cRequest, "initialize");

  mrb_define_method(mrb, cRequest, "finished?", cRequest_finished_p, 0);
  mrb_define_method(mrb, cRequest, "wait!", cRequest_waitB, 0);

  mrb_define_method(mrb, cRequest, "buffer", cRequest_buffer, 0);

  return cRequest;
}

static struct RClass *
init_cStatus(mrb_state *mrb, struct RClass *mod)
{
  /* (void)cStatus; */
  return NULL;
}


static const mrb_data_type comm_type = {
  .struct_name = "Comm",
  .dfree = mrb_free,
};

static inline struct RClass *get_comm_class(mrb_state *mrb)
{
  return mrb_class_get_under(mrb, get_rmpi_module(mrb), "Comm");
}

static mrb_value
cComm_wrap(mrb_state *mrb, MPI_Comm comm)
{
  struct RData* data;
  MPI_Comm *commp;

  commp = mrb_malloc(mrb, sizeof(*commp));
  *commp = comm;

  data = mrb_data_object_alloc(mrb, 
      get_comm_class(mrb), commp, &comm_type);

  return mrb_obj_value(data);
}

static MPI_Comm
cComm_unwrap(mrb_state *mrb, mrb_value obj)
{
  MPI_Comm *commp;

  commp = DATA_PTR(obj);
  return *commp;
}

static mrb_value
cComm_rank(mrb_state *mrb, mrb_value self)
{
  MPI_Comm comm;
  int rank;

  comm = cComm_unwrap(mrb, self);
  CHECKERR( MPI_Comm_rank(comm, &rank) );

  return mrb_fixnum_value(rank);
}

static mrb_value
cComm_size(mrb_state *mrb, mrb_value self)
{
  MPI_Comm comm;
  int size;

  comm = cComm_unwrap(mrb, self);
  CHECKERR( MPI_Comm_size(comm, &size) );

  return mrb_fixnum_value(size);
}

static mrb_value
cComm_get_world(mrb_state *mrb, mrb_value klass)
{
  return cComm_wrap(mrb, MPI_COMM_WORLD);
}

static mrb_value
cComm_send(mrb_state *mrb, mrb_value self)
{
  mrb_int dest, tag, count;
  char *data;
  MPI_Comm comm;

  comm = cComm_unwrap(mrb, self);

  dest = tag = count = 0;
  data = NULL;
  mrb_get_args(mrb, "iis", &dest, &tag, &data, &count);

  CHECKERR( MPI_Send(data, count, MPI_BYTE, dest, tag, comm) );

  return mrb_nil_value();
}

static mrb_value
cComm_recv(mrb_state *mrb, mrb_value self)
{
  mrb_int src,tag,count;
  char *data;
  MPI_Comm comm;
  MPI_Status status;
  mrb_value tuple[3];

  comm = cComm_unwrap(mrb, self);

  src = tag = count = -1;
  data = NULL;
  mrb_get_args(mrb, "ii|s", &src, &tag, &data, &count);

  if (data != NULL) {
    /* FIXME: check rmsg type and handle non-string types */
    /* FIXME: Use a real status argument */

    CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, &status) );
    tuple[0] = mrb_fixnum_value(status.MPI_SOURCE);
    tuple[1] = mrb_fixnum_value(status.MPI_TAG);
    return mrb_ary_new_from_values(mrb, 2, tuple);
  } else {
    MPI_Status status;
    int code;
    mrb_value str;

    memset(&status,0,sizeof(status));
    CHECKERR( code = MPI_Probe(src, tag, comm, &status) );

    count = 0;
    CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

    str = mrb_str_buf_new(mrb, count);
    data = RSTRING_PTR(str);

    CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, MPI_STATUS_IGNORE) );
    RSTR_SET_LEN(RSTRING(str), count) ;

    tuple[0] = str;
    tuple[1] = mrb_fixnum_value(status.MPI_SOURCE);
    tuple[2] = mrb_fixnum_value(status.MPI_TAG);
    return mrb_ary_new_from_values(mrb, 3, tuple);
  }
}

static mrb_value
cComm_probe(mrb_state *mrb, mrb_value self)
{
  int src, tag, count;
  MPI_Comm comm;
  MPI_Status status;
  mrb_value tuple[3];

  comm = cComm_unwrap(mrb, self);

  src = tag = 0;
  mrb_get_args(mrb, "ii", &src, &tag);

  memset(&status,0,sizeof(status));
  CHECKERR( MPI_Probe(src, tag, comm, &status) );

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  tuple[0] = mrb_fixnum_value(status.MPI_SOURCE);
  tuple[1] = mrb_fixnum_value(status.MPI_TAG);
  tuple[2] = mrb_fixnum_value(count);

  return mrb_ary_new_from_values(mrb, 3, tuple);
}

static mrb_value
cComm_iprobe(mrb_state *mrb, mrb_value self)
{
  int src, tag, count, flag;
  MPI_Comm comm;
  MPI_Status status;
  mrb_value tuple[3];

  comm = cComm_unwrap(mrb, self);

  src = tag = 0;
  mrb_get_args(mrb, "ii", &src, &tag);

  flag = 0;
  memset(&status,0,sizeof(status));
  CHECKERR( MPI_Iprobe(src, tag, comm, &flag, &status) );

  if (!flag) { return mrb_nil_value(); }

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  tuple[0] = mrb_fixnum_value(status.MPI_SOURCE);
  tuple[1] = mrb_fixnum_value(status.MPI_TAG);
  tuple[2] = mrb_fixnum_value(count);

  return mrb_ary_new_from_values(mrb, 3, tuple);
}

static mrb_value
cComm_isend(mrb_state *mrb, mrb_value self)
{
  MPI_Comm comm;
  int count;
  mrb_int dst,tag;
  rmpi_request_t *req;
  struct rmpi_buffer *buf;
  mrb_value rmsg;

  comm = cComm_unwrap(mrb, self);

  mrb_get_args(mrb, "iio", &dst, &tag, &rmsg);

  /* We incref for existing buffers but not for new buffers.  Why?
   * 1. New buffers have a reference count of one.
   * 2. Creating a request with a buffer will lock the buffer and
   *    increase its reference count by one.
   * 3. We want an existing buffer to have a net increase in its
   *    reference count of one, and a new buffer to have a single
   *    reference (the request).
   * 4. After we create the request, a newly-created buffer will
   *    have a reference count of two, so we have to decref.  An
   *    existing buffer will have one additional reference, so when we
   *    decref, it would have no net increase in its reference count.
   *    Incref'ing restores this.
   */
  if (mrb_type(rmsg) == MRB_TT_STRING) {
    count = RSTRING_LEN(rmsg);
    buf = rmpi_buflist_new_buffer(count);
    if (buf == NULL) {
      mrb_exc_raise(mrb, mrb_obj_value(mrb->nomem_err));
    }

    memcpy(buf->ptr, RSTRING_PTR(rmsg), count);
    buf->size = count;
  } else {
    struct rmpi_buffer_obj *h;
    h = cBuffer_validate(mrb, rmsg);
    buf = h->buf;
    rmpi_buffer_incref(buf);
  }

  req = rmpi_request_new(buf);
  rmpi_buffer_decref(buf);

  if (req == NULL) {
    mrb_exc_raise(mrb,mrb_obj_value(mrb->nomem_err));
  }

  CHECKERR( MPI_Isend(req->buf->ptr, req->buf->size, MPI_BYTE, dst, tag, comm, &req->req) );

  return cRequest_wrap(mrb, req);
}

static mrb_value
cComm_irecv(mrb_state *mrb, mrb_value self)
{
  int flag;
  mrb_int src,tag,count;
  char *data;
  MPI_Comm comm;
  MPI_Status status;
  mrb_value str, tuple[3];

  comm = cComm_unwrap(mrb, self);

  src = tag = count = -1;
  data = NULL;
  mrb_get_args(mrb, "ii|s", &src, &tag, &data, &count);

  if (data != NULL) {
    mrb_raise(mrb, E_NOTIMP_ERROR, "not yet implemented!");
  }

  flag = 0;
  memset(&status, 0, sizeof(status));
  CHECKERR( MPI_Iprobe(src,tag, comm, &flag, &status) );
  if (!flag) {
    return mrb_nil_value();
  }

  count = 0;
  CHECKERR( MPI_Get_count(&status, MPI_BYTE, &count) );

  str = mrb_str_buf_new(mrb, count);
  data = RSTRING_PTR(str);

  CHECKERR( MPI_Recv(data, count, MPI_BYTE, src, tag, comm, MPI_STATUS_IGNORE) );

  RSTR_SET_LEN(RSTRING(str), count);

  tuple[0] = str;
  tuple[1] = mrb_fixnum_value(status.MPI_SOURCE);
  tuple[2] = mrb_fixnum_value(status.MPI_TAG);
  return mrb_ary_new_from_values(mrb, 3, tuple);
}

static struct RClass *
init_cComm(mrb_state *mrb, struct RClass *mod)
{
  struct RClass *cComm;

  cComm = mrb_define_class_under(mrb, mod, "Comm", mrb->object_class);

  mrb_undef_method(mrb, cComm, "initialize");

  mrb_define_method(mrb, cComm, "rank", cComm_rank, MRB_ARGS_NONE());
  mrb_define_method(mrb, cComm, "size", cComm_size, MRB_ARGS_NONE());

  mrb_define_method(mrb, cComm, "Send", cComm_send, MRB_ARGS_REQ(3));

  mrb_define_method(mrb, cComm, "Recv", cComm_recv, MRB_ARGS_REQ(2) | MRB_ARGS_OPT(1));

  mrb_define_method(mrb, cComm, "Probe", cComm_probe, MRB_ARGS_REQ(2));
  mrb_define_method(mrb, cComm, "Iprobe", cComm_iprobe, MRB_ARGS_REQ(2));

  mrb_define_method(mrb, cComm, "Isend", cComm_isend, MRB_ARGS_REQ(3));
  mrb_define_method(mrb, cComm, "Irecv", cComm_irecv, MRB_ARGS_REQ(2) | MRB_ARGS_OPT(1));

#if 0
  mrb_define_method(mrb, cComm, "Gather", rb_comm_gather, 3);
  mrb_define_method(mrb, cComm, "Allgather", rb_comm_allgather, 2);
  mrb_define_method(mrb, cComm, "Bcast", rb_comm_bcast, 2);
  mrb_define_method(mrb, cComm, "Scatter", rb_comm_scatter, 3);
  mrb_define_method(mrb, cComm, "Sendrecv", rb_comm_sendrecv, 6);
  mrb_define_method(mrb, cComm, "Alltoall", rb_comm_alltoall, 2);
  mrb_define_method(mrb, cComm, "Reduce", rb_comm_reduce, 4);
  mrb_define_method(mrb, cComm, "Allreduce", rb_comm_allreduce, 3);
  mrb_define_method(mrb, cComm, "Errhandler", rb_comm_get_Errhandler, 0);
  mrb_define_method(mrb, cComm, "Errhandler=", rb_comm_set_Errhandler, 1);
  mrb_define_method(mrb, cComm, "Barrier", rb_comm_barrier, 0);
#endif /* 0 */

  mrb_define_singleton_method(mrb, (struct RObject*)cComm, "world", cComm_get_world, MRB_ARGS_NONE());

  return cComm;
}

static inline int
is_initialized(void)
{
  int flag;
  flag = 0;
  MPI_Initialized(&flag);
  return flag;
}

static inline int
is_finalized(void)
{
  int flag;
  flag = 0;
  MPI_Finalized(&flag);
  return flag;
}

static void do_finalize(void)
{
  if (!is_finalized() && is_initialized()) {
    MPI_Finalize();
  }
}

static mrb_value
mRMPI_init(mrb_state *mrb, mrb_value mod)
{
  static int exit_registered = 0;

  if (exit_registered) {
    return mrb_nil_value();
  }

  if (is_initialized()) {
    goto initialized;
  }

  char *prog = "program";
  char **argv = &prog;
  int argc = 1;
  MPI_Init(&argc, &argv);

initialized:
  atexit(do_finalize);
  exit_registered = 1;

  return mrb_nil_value();
}

static mrb_value
mRMPI_finalize(mrb_state *mrb, mrb_value mod)
{
  do_finalize();
  return mrb_nil_value();
}

static mrb_value mRMPI_rank(mrb_state *mrb, mrb_value mod)
{
  int rank;
  mpi_check_initialized(mrb);

  CHECKERR( MPI_Comm_rank(MPI_COMM_WORLD, &rank) );

  return mrb_fixnum_value(rank);
}

static mrb_value mRMPI_size(mrb_state *mrb, mrb_value mod)
{
  int size;
  mpi_check_initialized(mrb);

  CHECKERR( MPI_Comm_size(MPI_COMM_WORLD, &size) );

  return mrb_fixnum_value(size);
}

static mrb_value call_block_by_rank(mrb_state *mrb, mrb_value block, int callrank)
{
  int rank = -1;
  mrb_value comm;

  mpi_check_initialized(mrb);

  if (mrb_nil_p(block)) {
    mrb_raise(mrb, E_ARGUMENT_ERROR, "a block is required");
  }

  CHECKERR( MPI_Comm_rank(MPI_COMM_WORLD, &rank) );

  if ((rank == callrank) || ((callrank < 0) && (rank != 0))) {
    comm = cComm_wrap(mrb, MPI_COMM_WORLD);
    // comm = mrb_nil_value();
    return mrb_yield_argv(mrb, block, 1, &comm);
  }

  return mrb_nil_value();
}

static mrb_value mRMPI_foreman(mrb_state *mrb, mrb_value mod)
{
  mrb_value block;
  mrb_get_args(mrb, "&", &block);

  return call_block_by_rank(mrb, block, 0);
}

static mrb_value mRMPI_worker(mrb_state *mrb, mrb_value mod)
{
  mrb_int rank;
  mrb_value block;

  rank = -1;
  mrb_get_args(mrb, "|i&", &rank, &block);

  return call_block_by_rank(mrb, block, rank);
}

static mrb_value
mRMPI_waitB(mrb_state *mrb, mrb_value mod)
{
  rmpi_request_finish_waiting();
  return mrb_nil_value();
}

#if 0
static mrb_value
mRMPI_debug_hook(mrb_value mod)
{
  volatile int flag = 0;
  while (!flag) {
#ifdef _POSIX_SOURCE
    sleep(5);
#endif /* _POSIX_SOURCE */
  }

  return mrb_nil_value();
}
#endif /* 0 */

void
mrb_mruby_rmpi_gem_init(mrb_state *mrb)
{
  int ai;
  struct RClass *mRMPI;

  ai = mrb_gc_arena_save(mrb);
  mRMPI = mrb_define_module(mrb, "RMPI");
  mrb_gc_arena_restore(mrb,ai);

  mrb_define_module_function(mrb, mRMPI, "Init", mRMPI_init, MRB_ARGS_ANY());
  mrb_define_module_function(mrb, mRMPI, "Finalize", mRMPI_finalize, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mRMPI, "rank", mRMPI_rank, MRB_ARGS_NONE());
  mrb_define_module_function(mrb, mRMPI, "size", mRMPI_size, MRB_ARGS_NONE());

  mrb_define_module_function(mrb, mRMPI, "foreman", mRMPI_foreman, MRB_ARGS_NONE() | MRB_ARGS_BLOCK());
  mrb_define_module_function(mrb, mRMPI, "worker", mRMPI_worker, MRB_ARGS_REQ(0) | MRB_ARGS_OPT(1) | MRB_ARGS_BLOCK());

  mrb_define_module_function(mrb, mRMPI, "wait!", mRMPI_waitB, 0);

  // mrb_define_module_function(mrb, mRMPI, "debug_hook", mRMPI_debug_hook, 0);
  // mrb_define_module_function(mrb, mMPI, "Abort", mRMPI_abort, MRB_ARGS_REQ(2));

  mrb_define_const(mrb, mRMPI, "VERSION", mrb_fixnum_value(MPI_VERSION));
  mrb_define_const(mrb, mRMPI, "SUBVERSION", mrb_fixnum_value(MPI_SUBVERSION));
  // rb_define_const(mMPI, "SUCCESS", INT2NUM(MPI_SUCCESS));
  // rb_define_const(mMPI, "PROC_NULL", INT2NUM(MPI_PROC_NULL));

  init_cBuffer(mrb,mRMPI);
  init_cComm(mrb,mRMPI);
  init_cRequest(mrb, mRMPI);
  init_cStatus(mrb, mRMPI);
}

void
mrb_mruby_rmpi_gem_final(mrb_state *mrb)
{
}


